/*
 * zen/include/zbn.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZBN__
# define __ZBN__

#include <stdio.h>
#include <math.h>
#include "sys.h"

/* \TEX{
   \Section{Big integers layer of \zen}
   \label{zbn} 

   The |ZEN| library was primarily built upon the |BigNum|
   library~\cite{bignum} (version 1.0-b).
   
   Then, we have extended |ZEN| so that it can use |GMP| (from version
   2.0), the Gnu Multiprecision Package~\cite{gmp}. In order to
   achieve this final goal, all the big integers functions were called
   in |ZEN| through a reformatted version of both |BigNum| and |GMP|'
   functions. These macros were prefixed by |ZBN|.
  
   Since only few functions from |BigNum| or |GMP| are needed, we
   finally decided to design a specific big integers layer for |ZEN|,
   the |ZBN| layer (from version 3.0).
} */

/* \TEX{
   \Subsection{Caracteristics of the |ZBN| layer}

   Most of |ZBN| procedures are working on arrays of digits. The
   digits are of type |BigNumDigit|. An array of digits is of type
   |BigNum|. And the size of these arrays is of type
   |BigNumLength|. The sizes of these types depends on the computer
   you are using (generally, $32$ or $64$ bits). |BigNumDigit| must be
   an unsigned type. |BigNumLength| must be a signed.
   } */

/* \TEX{
   \input{zbnconfig}   
   } */
   
/* \TEX{
   \Subsubsection{Memory limitation}
   \label{zbncompil}
   
   Some of the functionalities of the |ZBN| layer, need a lot of
   memory at compilation time. You may decrease the amount of memory
   needed by using the following compilation flag which is set in
   |zbn.h|.
   } */
#ifndef LOW_MEMORY		/* If set do not inline datas
				   in zbn_weight_data.h
				   \CiteC */
# define LOW_MEMORY 0		/* \EndCiteC */
#endif

/* \TEX{
   \Subsubsection{Karatsuba's multiplication on integers}
   
   The Karatsuba's multiplication on big integers is used by default
   in ZEN. This can be disabled using the |KARA| parameter.
   } */
#ifndef KARA			/* Activate Karatsuba operations
				   on integers
				   \CiteC */
# define KARA 1			/* \EndCiteC */
#endif

/*#define KARA_LIM_SET	*/	/* No variable cut-off enabled */

#define ZBN_HAS_INVERSE 0
#define ZBN_NO_INVERSE  1

/* \TEX{
   \Subsection{Constants}
   Three constants are used in |ZEN|:
   \begin{description}
   \item[|SIZE_BLOC|] depends on the arithmetic of
   computers. It is the size in bits of a |BigNumDigit| (usually 32 or 64 bits).
   \item[|SIZE_CHAR|] is the size in bits of a |char| (usually 8 bits).
   \item[|SIZE_SHORT|] is the size in bits of a |short| (usually 16 bits).
   \end{description}
   }
*/
#include "zbnconfig.h"

typedef BigNumDigit* BigNum;

# define SIZE_BLOC  BN_DIGIT_SIZE  
# define SIZE_SHORT sizeof(short)*BN_BYTE_SIZE  
# define SIZE_CHAR  BN_BYTE_SIZE  

/* For compatibility purposes */
# define BN_LT -1
# define BN_EQ  0
# define BN_GT  1

/* \TEX{
   \Subsection{|BigNum| allocation}
   } */

#if 0
/* \TEX{
   \title{Allocation of a |BigNum|}
   \input{A length |nl|.}
   \output{An allocated |BigNum| of size |nl| or |NULL| if an error occured.}
   } */
BigNum ZBNC(nl)
     BigNumLength nl; {}
#endif
#define ZBNC(l) (BigNum)ZENMalloc((l)*sizeof(BigNumDigit))

#if 0
/* \TEX{
   \title{Freeing a |BigNum|}
   \input{An allocated |BigNum|.}
   \side{|n| is freed.}
   } */
void ZBNF(n)
     BigNum n; {}
#endif
#define ZBNF(n) ZENFree(n)

/* \TEX{
   \input{zbndigits}
   } */

/* Basic macros described in digits.c */
#define modSizeBloc(x) ((x) % SIZE_BLOC)
#define divSizeBloc(x) ((x) / SIZE_BLOC)
#define modSizeChar(x) ((x) % SIZE_CHAR)
#define divSizeChar(x) ((x) / SIZE_CHAR)
#define BlocExp2(x) (((unsigned int)(x)) >= SIZE_BLOC ? (BigNumDigit)(0):((BigNumDigit)1<<(x)))
#define ZBNDigit(n,i) ((n)[i])
#define ZBNGetDigit(nn) ZBNDigit((nn), 0)
#define ZBNSetDigit(nn, d) ((void)(ZBNDigit((nn), 0) = (d)))
#define ZBNNumBitsInDigit(d) (SIZE_BLOC-ZBNNumLeadingZeroBitsInDigit(d))

/* Functions described in digits.c */
extern BigNumLength ZBNNumLeadingZeroBitsInDigit ___((BigNumDigit d));
extern BigNumLength ZBNNumDigits ___((BigNum n, BigNumLength nl));

/* \TEX{
   \input{zbnassign}
   } */

/* Functions described in assign.c */
extern void ZBNSetToZero ___((BigNum n, BigNumLength nl));
extern void ZBNSetToOne ___((BigNum n, BigNumLength nl));
extern void ZBNAssign ___((BigNum m, BigNum n, BigNumLength nl));

/* \TEX{
   \input{zbnrandom}
   } */

/* Functions described in random.c */
extern void zbnsrandom       ___((BigNumDigit));
extern BigNumDigit zbnrandom ___((void));
extern void ZBNSetRandom     ___((BigNum n, BigNumLength nl));

/* \TEX{
   \input{zbntest}
   } */

/* Basic macros described in test.c */
#define ZBNAreEqual(mm, ml, nn, nl) !ZBNCompare(mm, ml, nn, nl)

/* Functions described in test.c */
extern int ZBNIsZero ___((BigNum n, BigNumLength nl));
extern int ZBNIsOne ___((BigNum n, BigNumLength nl));
extern int ZBNCompare ___((BigNum m, BigNumLength ml, BigNum n, BigNumLength nl));
extern int ZBNCompareLazy ___((BigNum m, BigNumLength ml, BigNum n, BigNumLength nl));

/* \TEX{
   \input{zbnbinary}
   } */

/* Functions described in binary.c */
extern void ZBNAnd ___((BigNum m, BigNum n, BigNumLength nl));
extern void ZBNOr ___((BigNum m, BigNum n, BigNumLength nl));
extern void ZBNXor ___((BigNum m, BigNum n, BigNumLength nl));

/* \TEX{
   \input{zbnadd}
   } */

/* Functions described in add.c */
extern BigNumDigit ZBNAddCarry ___((BigNum m, BigNumLength ml, BigNumDigit carry));
extern BigNumDigit ZBNAdd ___((BigNum m, BigNumLength ml, BigNum n, BigNumLength nl, BigNumDigit carry));
extern void ZBNComplement ___((BigNum n, BigNumLength nl));
extern BigNumDigit ZBNSubtractBorrow ___((BigNum m, BigNumLength ml, BigNumDigit carry));
extern BigNumDigit ZBNSubtract ___((BigNum m, BigNumLength ml, BigNum n, BigNumLength nl, BigNumDigit carry));

/* \TEX{
   \input{zbnshift}
   } */

/* Functions described in shift.c */
extern BigNumDigit ZBNShiftLeft ___((BigNum n, BigNumLength nl, BigNumLength l));
extern BigNumDigit ZBNShiftRight ___((BigNum n, BigNumLength nl, BigNumLength l));
extern void ZBNAnyShiftLeft ___((BigNum r, BigNumLength *p_rl, BigNum n, BigNumLength nl, unsigned int nnn));
extern void ZBNAnyShiftRight ___((BigNum n, BigNumLength nl, unsigned int nnn));

/* \TEX{
   \input{zbnmul}
   } */

/* Macros described in mul.c */
# define ZBNMultiply(n, nl, a, al, b, bl) ZBNMultiplyPlain(n, nl, a, al, b, bl) 
# define ZBNSquare(n, nl, a, al) ZBNSquarePlain(n, nl, a, al) 

/* Functions described in mul.c */
extern BigNumDigit ZBNMultiplyDigit ___((BigNum m, BigNumLength ml, BigNum n, BigNumLength nl, BigNumDigit d));
extern BigNumDigit ZBNMultiplyPlain ___((BigNum p, BigNumLength pl, BigNum m, BigNumLength ml, BigNum n, BigNumLength nl));
extern void ZBNMult ___((BigNum n, BigNum a, BigNumLength al, BigNum b, BigNumLength bl));
extern BigNumDigit ZBNSquarePlain ___((BigNum p, BigNumLength pl, BigNum n, BigNumLength nl));
extern void ZBNSqu ___((BigNum n, BigNum a, BigNumLength al));

/* \TEX{
   \input{zbnkaramul}
   } */

/* Functions described in karamul.c */
#ifndef KARA_LIM_SET
extern BigNumDigit ZBNMultiplyKaratsuba ___((BigNum x, BigNumLength xl, BigNum a, BigNumLength al, BigNum b, BigNumLength bl));
extern void ZBNMultiplyKaratsubaBuffer ___((BigNum x, BigNum a, BigNumLength al, BigNum b, BigNumLength bl, BigNum buf));
#else
extern BigNumDigit ZBNMultiplyKaratsuba ___((BigNum x, BigNumLength xl, BigNum a, BigNumLength al, BigNum b, BigNumLength bl, BigNumLength lim));
extern void ZBNMultiplyKaratsubaBuffer ___((BigNum x, BigNum a, BigNumLength al, BigNum b, BigNumLength bl, BigNum buf, BigNumLength lim));
#endif
extern BigNumLength ZBNSizeBufKaraM ___((BigNumLength l));

/* \TEX{
   \input{zbnkarasqu}
   } */

/* Functions described in karasqu.c */
#ifndef KARA_LIM_SET
extern BigNumDigit ZBNSquareKaratsuba ___((BigNum x, BigNumLength xl, BigNum a, BigNumLength al));
extern void ZBNSquareKaratsubaBuffer ___((BigNum x, BigNum a, BigNumLength al, BigNum buf));
#else
extern BigNumDigit ZBNSquareKaratsuba ___((BigNum x, BigNumLength xl, BigNum a, BigNumLength al, BigNumLength lim));
extern void ZBNSquareKaratsubaBuffer ___((BigNum x, BigNum a, BigNumLength al, BigNum buf, BigNumLength lim));
#endif
extern BigNumLength ZBNSizeBufKaraS ___((BigNumLength l));

/* \TEX{
   \input{zbnkaracut}
   } */

/* Functions described in karacut.c */
extern BigNumLength ZBNKaratsubaMultiplyCutoff ___((BigNumLength size));
extern BigNumLength ZBNKaratsubaSquareCutoff ___((BigNumLength size));
extern BigNumDigit ZBNMultiplyK ___((BigNum p, BigNumLength pl, BigNum m, BigNumLength ml, BigNum n, BigNumLength nl));
extern BigNumDigit ZBNSquareK ___((BigNum p, BigNumLength pl, BigNum n, BigNumLength nl));


/* \TEX{
   \input{zbndiv}
   } */

/* Functions described in div.c */
extern BigNumDigit ZBNDivideDigit ___((BigNum q, BigNum n, BigNumLength nl, BigNumDigit d));
extern void ZBNDivideHeart ___((BigNum n, BigNumLength nl, BigNum d, BigNumLength dl, BigNumDigit D));
extern void ZBNDivide ___((BigNum n, BigNumLength nl, BigNum d, BigNumLength dl));

/* \TEX{
   \input{zbnlog}
   } */

/* Functions described in log.c */
extern double ZBNLog ___((BigNum n, BigNumLength nl));

/* \TEX{
   \input{zbnsqrt}
   } */

/* Functions described in sqrt.c */
extern BigNumLength ZBNSqrt ___((BigNum r, BigNum a, BigNumLength al));
/*
  \TEX{
  The numbers can be read and write for all bases between 2 and the following constant:
  }
  \CiteC
*/
#define ZBN_IO_MAX_BASIS 36
/* \EndCiteC 
   \input{zbnio}
   } */

/* Basic macros described in io.c */
#define ZBNPrint(n, nl) ZBNPrintToFile(stdout, n, (BigNumLength)(nl), 10)

/* Functions described in io.c */
extern char *ZBNPrintToString ___((BigNum n, BigNumLength nl, int base));
extern int ZBNReadFromStringLazy ___((BigNum m, BigNumLength ml, char *s, int base));
extern int ZBNReadFromString ___((BigNum *p_n, BigNumLength *p_nl, char *s, int base));
extern char *ZBNPutToStringLazy ___((BigNum n, BigNumLength nl));
extern char *ZBNPutToString ___((BigNum n, BigNumLength nl));
extern int ZBNGetFromStringLazy ___((BigNum n, BigNumLength nl, char *s));
extern int ZBNGetFromString ___((BigNum *p_n, BigNumLength *p_nl, char *s));
extern int ZBNPrintToFile ___((FILE *file, BigNum n, BigNumLength nl, int base));
extern int ZBNReadFromFileLazy ___((BigNum n, BigNumLength nl, FILE *file, int base));
extern int ZBNReadFromFile ___((BigNum *p_n, BigNumLength *p_nl, FILE *file, int base));
extern int ZBNPutToFileLazy ___((FILE *file, BigNum n, BigNumLength nl));
extern int ZBNPutToFile ___((FILE *file, BigNum n, BigNumLength nl));
extern int ZBNGetFromFileLazy ___((BigNum n, BigNumLength nl, FILE *file));
extern int ZBNGetFromFile ___((BigNum *p_n, BigNumLength *p_nl, FILE *file));

/* \TEX{
   \input{zbngcd}
   } */

/* Functions described in gcd.c */
extern BigNumDigit ZBNDigitGCD ___((BigNumDigit a_in, BigNumDigit b_in));
extern int ZBNGcd ___((BigNum g, BigNumLength *p_gl, BigNum n, BigNumLength nl, BigNum m, BigNumLength ml));
extern int ZBNEea ___((BigNum g, BigNumLength *p_gl, BigNum a, BigNumLength *p_al, BigNum n, BigNumLength nl, BigNum m, BigNumLength ml));


/* \TEX{
   \input{zbnmodulo}
   } */

/* Basic macros described in modulo.c */
#define ZBNModulo(p_m,n,nl,m,ml) ZBNModuloLazy(p_m,n,ZBNNumDigits(n,nl),m,ZBNNumDigits(m,ml))

/* Functions described in modulo.c */
extern BigNumLength ZBNModuloLazy ___((BigNum *p_m, BigNum n, BigNumLength nl, BigNum m, BigNumLength ml));
extern void ZBNModAdd ___((BigNum a, BigNumLength al, BigNum b, BigNumLength bl, BigNum m, BigNumLength ml));
extern void ZBNModNegate ___((BigNum a, BigNumLength al, BigNum b, BigNumLength bl, BigNum m, BigNumLength ml));
extern void ZBNModSubtract ___((BigNum a, BigNumLength al, BigNum b, BigNumLength bl, BigNum m, BigNumLength ml));
extern void ZBNModMultiply ___((BigNum x, BigNumLength xl, BigNum a, BigNumLength al, BigNum b, BigNumLength bl, BigNum m, BigNumLength ml));
extern void ZBNModSquare ___((BigNum x, BigNumLength xl, BigNum a, BigNumLength al, BigNum m, BigNumLength ml));
extern void ZBNModMultiplyKaratsubaBuffer ___((BigNum x, BigNumLength xl, BigNum a, BigNumLength al, BigNum b, BigNumLength bl, BigNum buf, BigNum m, BigNumLength ml));
extern void ZBNModSquareKaratsubaBuffer ___((BigNum x, BigNumLength xl, BigNum a, BigNumLength al, BigNum buf, BigNum m, BigNumLength ml));


/* \TEX{
   \input{zbnweight}
   } */

/* Variables defined in weight.c */
#if !LOW_MEMORY
extern unsigned int __ZEN_WEIGHT[];
#endif

/* Functions described in weight.c */
extern BigNumLength ZBNWeightDigitFast ___((BigNumDigit d));
extern BigNumLength ZBNWeightDigit ___((BigNum n));
extern BigNumLength ZBNWeight ___((BigNum n, BigNumLength nl));

#endif
