/*
 * zen/include/zef.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZEF__
#define __ZEF__
#include "zen.h"

#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)<(b) ? (b) : (a))

/*
   \TEX{
   The |_ZefFct| functions are the functions on |ZENElt| for fractions.
   The fractions are represented by two bignums and an exponent $\frac{N}{D}=2^{|e|}\frac{(|n|,|nl|)}{(|d|,|dl|)}$.
   This representation is used because operations often lead to results whose binary representation ends with a lot of zeroes. Managing the number |e| can then substantially accelerate the computations.
   Besides, this representation allows computations with approximations (see below).
   Some memory allocations are done silently inside the functions. All are tested, but the errors are sometimes forgotten, as some functions must be void to allow efficient computation in finite rings.
   The operations on fractions are experimental. In fact, they should never be guaranteed as the whole concept of |ZEN| is done to compute in finite sets. Nevertheless, it is sometimes interesting to be able to compute over fractions without having to use another complete package.

   As this is experimental, a number of compilation options are set in |zef.h|. The main goal with this, is to be able to obtain approximate results.
   }
   \CiteC
   */
#define VERBOSE 0		/* Debugging purpose */

#if VERBOSE > 1
# define VERBOSEFCT ZFP
#else
# define VERBOSEFCT ZEP
#endif

#define SIMPLIFY 1		/* If unset, the computations are performed without
				   any attempt to make simplifications between numerators
				   and denominators. */
#if SIMPLIFY
# define SIMPLIFY_LIM 8		/* When numerator or denominator is of size greater
				   than this limit, a simplification is attempted,
				   unless the fraction is already simplified */
#endif
#define APPROXIMATE_RESULTS 0	/* Caution ! This will cause the zentest
				   program failed. */
#if APPROXIMATE_RESULTS
# define APPROXIMATE_LEVEL 10	/* Maximum number of BigNumDigits
				   for an integer. Least significant digits
				   will be considered as zeros and stored in |e|.
				   10 means about 460 significant decimal
				   digits */
# define AFTER_PGCD 0		/* The approximation is performed after
				   the simplification. This is more precise but also
				   (sometimes a lot) slower */
#endif /* Best compromise is reached with approximation after PGCD,
	  precision of 10 digits and KARA_EVERYWHERE unset in zbn.h */
/* \EndCiteC
   \TEX{
   \Section{Ring initialization}
   \input{zefinit}

   \Section{Element operations}
   } 
   */

#define ZEF_ONE     0
#define ZEF_INTEGER 1
#define ZEF_INVERSE 2
#define ZEF_FRAC    3

#if 0
/*
   \TEX{
   \title{Type of the fraction}
   \input{A |ZENElt| and a |ZENRing|.}
   \output{\begin{itemize}
   \item |ZEF_ONE|(=0) if the fraction is one ($N=D=1$).
   \item |ZEF_INTEGER| if the fraction is an integer ($N=1$).
   \item |ZEF_INVERSE| if the fraction is an inverse of an integer ($D=1$).
   \item |ZEF_FRAC| otherwise.
   \end{itemize}}
   \note{This function extracts two bits from a flag.}
   } 
   */
int ZefEltType(F,R)
     ZENElt F;
     ZENRing R; {}
#endif
#define ZefEltFlag(F,R) (ZENElt2Frac(F,R)->flag)
#define ZefEltType(F,R) (ZefEltFlag(F,R)&ZEF_FRAC)
#define ZefEltSetType(F,t,R) ZefEltFlag(F,R)=(((~ZEF_FRAC)&ZefEltFlag(F,R))|(t))

#define ZefEltAdjustM(F,ZEF,R) ((ZefEltFlag(F,R) & ZEF) ? ZBNIsOne(ZefEltM(F,ZEF,R),ZefEltMl(F,ZEF,R)) : 1)
#define ZefEltAdjustType(F,R) ZefEltSetType(F,ZEF_FRAC^( (ZefEltAdjustM(F,ZEF_INVERSE,R)<<1) | (ZefEltAdjustM(F,ZEF_INTEGER,R)) ),R)

#define ZEF_POSITIVE 0
#define ZEF_NEGATIVE 4
#if 0
/*
   \TEX{
   \title{Sign of the fraction}
   \input{A |ZENElt| and a |ZENRing|.}
   \output{\begin{itemize}
   \item |ZEF_POSITIVE|(=0) if the fraction is greater than or equal to 0.
   \item |ZEF_NEGATIVE| otherwise.
   \end{itemize}}
   \note{This function extracts one bit from a flag.}
   } 
   */
int ZefEltSign(F,R)
     ZENElt F;
     ZENRing R; {}
#endif
#define ZefEltSign(F,R) (ZefEltFlag(F,R)&ZEF_NEGATIVE)
#define ZefEltNeg(F,R) ZefEltFlag(F,R) ^= ZEF_NEGATIVE

#define ZEF_TOEXP   8
#define ZEF_TOPGCD 16
#define ZEF_TOSIMPLIFY (ZEF_TOPGCD | ZEF_TOEXP)
#define ZefEltSimplifyFlag(F,R) (ZefEltFlag(F,R)&ZEF_TOSIMPLIFY)
#define ZefEltToSimplify(F,R) ZefEltFlag(F,R) |= ZEF_TOSIMPLIFY

#if 0
/*
   \TEX{
   \title{Numerator of the fraction}
   \input{A |ZENElt| and a |ZENRing|.}
   \output{A |BigNum|}
   \note{This is a macro that returns a field in the structure. You should not disallocate it.}
   } 
   */
BigNum ZefEltN(F,R)
     ZENElt F;
     ZENRing R; {}
#endif
#define ZefEltM(F,i,R) ((ZENElt2Frac(F,R)->n)[(i)-1])
#define ZefEltMl(F,i,R) ((ZENElt2Frac(F,R)->nl)[(i)-1])
#define ZefEltMlm(F,i,R) ((ZENElt2Frac(F,R)->nlm)[(i)-1])
#define ZefEltN(F,R) ZefEltM(F,ZEF_INTEGER,R)

#if 0
/*
   \TEX{
   \title{Length of the numerator of the fraction}
   \input{A |ZENElt| and a |ZENRing|.}
   \output{An integer}
   \note{This is a macro that returns a field in the structure.}
   } 
   */
BigNumLength ZefEltNl(F,R)
     ZENElt F;
     ZENRing R; {}
#endif
#define ZefEltNl(F,R) ZefEltMl(F,ZEF_INTEGER,R)
#define ZefEltNlm(F,R) ZefEltMlm(F,ZEF_INTEGER,R)

#if 0
/*
   \TEX{
   \title{Denominator of the fraction}
   \input{A |ZENElt| and a |ZENRing|.}
   \output{A |BigNum|}
   \note{This is a macro that returns a field in the structure. You should not disallocate it.}
   } 
   */
BigNum ZefEltD(F,R)
     ZENElt F;
     ZENRing R; {}
#endif
#define ZefEltD(F,R) ZefEltM(F,ZEF_INVERSE,R)

#if 0
/*
   \TEX{
   \title{Length of the denominator of the fraction}
   \input{A |ZENElt| and a |ZENRing|.}
   \output{An integer}
   \note{This is a macro that returns a field in the structure.}
   } 
   */
BigNumLength ZefEltDl(F,R)
     ZENElt F;
     ZENRing R; {}
#endif
#define ZefEltDl(F,R)  ZefEltMl(F,ZEF_INVERSE,R)
#define ZefEltDlm(F,R)  ZefEltMlm(F,ZEF_INVERSE,R)

#if 0
/*
   \TEX{
   \title{Exponent of the fraction}
   \input{A |ZENElt| and a |ZENRing|.}
   \output{An integer}
   \note{This is a macro that returns a field in the structure.}
   } 
   */
int ZefEltE(F,R)
     ZENElt F;
     ZENRing R; {}
#endif
#define ZefEltE(F,R) (ZENElt2Frac(F,R)->exp)

#define ZefEltCreate(a,Rg) _ZefEltCreate(&(a),Rg)
#define ZefEltFree(a,Rg) _ZefEltFree(a,Rg)

#define ZefEltSetToZero(a,Rg) _ZefEltSetToZero(&(a),Rg)
#define ZefEltSetToOne(a,Rg) _ZefEltSetToOne(&(a),Rg)

#define ZefEltToZ(p,p_pl,a,R) _ZefEltToZ(p,p_pl,a,R)
#define ZefZToElt(a,p,pl,R) _ZefZToElt(&(a),p,pl,R)

#define ZefEltAdd(a,b,Rg) _ZefEltAdd(&(a),b,Rg)
#define ZefEltSubtract(a,b,Rg) _ZefEltSubtract(&(a),b,Rg)
#define ZefEltMultiply(x,a,b,Rg) _ZefEltMultiply(&(x),a,b,Rg)
#define ZefEltSimplify(a,lim,Rg) _ZefEltSimplify(a,lim,Rg)
#define ZefEltUnsimplify(a,Rg) _ZefEltUnsimplify(a,Rg)
#define ZefEltAssign(a,b,Rg) _ZefEltAssign(&(a),b,Rg)
#define ZefEltNegate(a,b,Rg) _ZefEltNegate(&(a),b,Rg)
#define ZefEltInverse(a,b,Rg) _ZefEltInverse(&(a),b,Rg)
#define ZefEltIsZero(a,Rg) _ZefEltIsZero(a,Rg)

#define ZefEltTrace(r,a,Rg) ZefEltAssign(r,a,Rg)

#define ZefEltPrintToFile(file,a,base,Rg) _ZefEltPrintToFile(file,a,base,Rg)

#if 0
/*
   \TEX{
   \title{Setting a fraction}
   \input{A |ZENElt|, two integers and a fraction |ZENRing|.}
   \side{The element is set to $\frac{|n|}{|d|}$}
   } 
   */
void ZefEltGetFromZs(a,n,d,R)
     ZENElt a;
     int n,d;
     ZENRing R; {}
#endif
#define ZefEltGetFromZs(a,n,d,R) _ZefEltGetFromZs(&(a),n,d,R)

/* Procedure on Rings
 ***********************/
extern int _ZefPrcInitPolyRootsCanonical ___((ZENRing));
extern int _ZefPrcClosePolyRootsCanonical ___((ZENRing));

extern int _ZefPrcInitEltKara ___((ZENRing));
extern int _ZefPrcCloseEltKara ___((ZENRing));

extern int _ZefInit ___((ZENRing));
extern int _ZefClose ___((ZENRing));

/* Procedure on elements
 ***********************/
/* Allocation */
extern int _ZefEltCreate ___((ZENElt *, ZENRing)); /* Create elements */
extern void _ZefEltFree ___((ZENElt, ZENRing)); /* Free elements */

/* Assigning */
extern void _ZefEltAssign ___((ZENElt *, ZENElt, ZENRing)); /* Assign an element with another */
extern void _ZefEltSetToZero ___((ZENElt *, ZENRing)); /* Set an element to zero */
extern void _ZefEltSetToOne ___((ZENElt *, ZENRing)); /* Set an element to one */
extern void _ZefEltSetRandom ___((ZENElt *, ZENRing)); /* Set randomly an element */
extern void _ZefEltSetNext ___((ZENElt *, ZENRing)); /* Set an element to the next */

/* Test */
extern int _ZefEltAreEqual ___((ZENElt, ZENElt, ZENRing)); /* Is an element equal to another */
extern int _ZefEltIsZero ___((ZENElt, ZENRing)); /* Is an element equal to zero */
extern int _ZefEltIsOne ___((ZENElt, ZENRing)); /* Is an element equal to one */

/* Arithmetic */
extern int _ZefEltSimplify ___((ZENElt, int, ZENRing));
extern int _ZefEltUnsimplify ___((ZENElt, ZENRing));
extern void _ZefEltAdd ___((ZENElt *, ZENElt, ZENRing));	/* Addition of elements */
extern void _ZefEltAddMisc ___((ZENElt *, ZENElt, ZENRing));	/* Addition of elements */
extern void _ZefEltNegate ___((ZENElt *, ZENElt, ZENRing));	/* Addition of elements */
extern void _ZefEltSubtract ___((ZENElt *, ZENElt, ZENRing));	/* Addition of elements */
extern void _ZefEltSquare ___((ZENElt *, ZENElt, ZENRing)); /* Squaring of elements */
extern void _ZefEltMultiply ___((ZENElt *, ZENElt, ZENElt, ZENRing)); /* Multiplication of elements */
extern int _ZefEltInverse ___((ZENElt *, ZENElt, ZENRing)); /* Inverse of elements */
extern void _ZefEltTrace ___((ZENElt *, ZENElt, ZENRing));	/* Trace of elements */

/* io */
extern int _ZefEltReadFromString ___((ZENElt *, char *, int, ZENRing)); /* Read a polynomial from a string */
extern char* _ZefEltPrintToString ___((ZENElt, int, ZENRing)); /* Print a polynomial to a string */
extern int _ZefEltReadFromFile ___((ZENElt *, FILE *, int, ZENRing));	/* Read a polynomial from a file */
extern int _ZefEltPrintToFile ___((FILE *, ZENElt, int, ZENRing)); /* Print a polynomial to a file */
extern int _ZefEltGetFromString ___((ZENElt *, char *, ZENRing)); /* Get a polynomial from a string */
extern char* _ZefEltPutToString ___((ZENElt, ZENRing)); /* Put a polynomial to a string */
extern int _ZefEltGetFromFile ___((ZENElt *, FILE *, ZENRing)); /* Get a polynomial from a file */
extern int _ZefEltPutToFile ___((FILE *, ZENElt, ZENRing)); /* Put a polynomial to a file */

extern void _ZefEltToZ ___((BigNum,BigNumLength *,ZENElt,ZENRing)); /* Evaluate in $p$ an element */
extern void _ZefEltGetFromZs ___((ZENElt *, int, int, ZENRing)); /* Evaluate in $p$ an element */
extern void _ZefZToElt ___((ZENElt *,BigNum,BigNumLength,ZENRing));	/* Get the element from an evaluation in $p$ */
extern int _ZefEltConvert ___((ZENElt *,ZENRing,ZENElt,ZENRing));

extern int ZefEltFactorial ___((ZENElt,int,ZENRing));
extern int ZefEltBinomial ___((ZENElt,int,int,ZENRing));

extern int ZFP ___((ZENElt,ZENRing));
/*
  \TEX{
  \input{zefelt_add}
  \input{zefelt_assign}
  \input{zefelt_test}
  \input{zefelt_add}
  \input{zefelt_multiply}
  \input{zefelt_inverse}
  \input{zefelt_trace}
  \input{zefelt_io}
  \input{zefbinomial}
  }
  */
#endif
