/*
 * zen/include/zen.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZEN__
#  define __ZEN__

#define ZEN_VERSION 3
#define ZEN_VERSION_MINOR 0

#include "zentypes.h"

/* \TEX{  

   All the procedures described here form the |libzen.a|
   library. These procedures are all macros defined in |zen.h|. There
   are four types of such macros:

   \begin{itemize}

   \item Macros which are an interface for a field of a |ZENRing| |Rg|.
   The last argument of these macros is |Rg|.

   \item Macros which are an interface for an elliptic curve |ZENEc|
   |E|. The last argument of these macros is |E|.

   \item Macros which are an interface for a function of the
   library. Such functions are prefixed by |_ZEN|.

   \item Macros for basic operations, for instance getting a field of
   a structure described in section \ref{zentypes}.

   \end{itemize} 

   \begin{note}
   {\bf The |ZEN| format is entirely made of macros. It is therefore strictly forbidden to use side effects in parameters of |ZEN| calls.}
   \end{note}
   }
   */
#define ZEN_HAS_INVERSE ZBN_HAS_INVERSE
#define ZEN_NO_INVERSE ZBN_NO_INVERSE

#define POLY_GEN 1
#define MAT_GEN  2
#define SR_GEN   4

#define ZENMAT_HAS_KERNEL  6    
#define ZENMAT_HAS_MAXIMAL_RANK ZEN_HAS_INVERSE

#define ZENFct(R) ((R)->fct)

/*
   \TEX{
     \Section[Finite rings]{Procedures to handle finite rings.}

     A first set of procedures enables the user to allocate or free a
     finite ring.

     \Subsection{Extension allocation.}
     
     The principle of the library is that efficient procedures are
     chosen on the creation of a |ZENRing| by setting a lot of
     pointers on functions inside the |ZENRing| structure. Therefore,
     the use of generic functions is possible together with efficiency
     as the function call only needs one more dereferenciation of
     function pointer.  }

   */

#if 0
/* \TEX{
   \title{Initialization of a ring $\Z/n\Z$.}
   \input{An unallocated |ZENRing|, and a |BigNum| |n| of size |nl|.} 
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{|R| is allocated and set to $\Z/n\Z$.
   If $n=0$, |R| is allocated and set to $\Q$ (experimental feature).}
   \note{No precomputations are done in order to speed up
     some arithmetic operations. If you want some, see
   |ZENRingAddPrc()| and |ZENRingRmPrc()|.}  
 }
   */
int ZENBaseRingAlloc(R, n, nl)
     ZENRing R;
     BigNum n;
     BigNumLength nl;{}
#endif
#define ZENBaseRingCreate(n, nl) _ZENBaseRingCreate(n, nl)
#define ZENBaseRingAlloc(R,n,nl) (((R)=ZENBaseRingCreate(n,nl))==NULL ? ZERR : 0)
#if 0
/* \TEX{
  \title{Initialization of a finite extension over another finite ring.}
   \input{An unallocated |ZENRing|, and a polynomial |P| defined over a finite ring |Rg|.} 
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{The extension |Ex| is allocated and set to $|Rg|[X]/(|P|(X))$.}
   }
   */
int ZENExtRingAlloc(Ex, P, Rg)
     ZENPoly P;
     ZENRing Ex,Rg;{}
#endif
#define ZENExtRingCreate(P, Rg) _ZENExtRingCreate(P, Rg)
#define ZENExtRingAlloc(E,P,R) (((E)=ZENExtRingCreate(P,R))==NULL ? ZERR : 0)

#if 0 
/* \TEX{
   \title{Closing a finite ring.}
   \input{A finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{|Rg| is completely freed.}
   \note{A call to |ZENRingRmPrc(Rg, ZENRingPrcp(Rg))| is
   performed at the beginning of this procedure.}
   }
   */
int ZENRingClose(Rg)            
     ZENRing Rg;{}
#endif
#define ZENRingClose(Rg) _ZENRingClose(Rg)

#if 0
/* \TEX{
   \title{Closing a clone ring.}
   \input{A ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|R| is freed as well as the underlying |ZENRingOrigin(R)|, if it exists. If |R| is a chinese ring, the list of rings is clone freed, but the under-underlying |ZENRingOrigin(R)| is not freed.}
   }
   */
int ZENRingCloneClose(R)		
     ZENRing R;{}
#endif
#define ZENRingCloneClose(Rg) _ZENRingCloneClose(Rg)

#if 0 
/* \TEX{
   \title{Closing a finite ring.}
   \input{A finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{|Rg| is completely freed, as well as the underlying rings.}
   \note{This procedure can be used for freeing a |ZENRing| previously read from a file by |ZENRingReadFromFile|.}
   }
   */
int ZENRingFullClose(Rg)            
     ZENRing Rg;{}
#endif
#define ZENRingFullClose(Rg) _ZENRingFullClose(Rg)

#if 0
/* \TEX{
   \title{Copying a ring.}
   \input{A finite field |R|.}
   \output{|ZENNULL| if an error occurred, a copy of the finite Ring |R|
   without ist precomputation otherwise.} 
   }
   */
ZENRing ZENRingCopy(R)          
     ZENRing R;{}
#endif
#define ZENRingCopy(R) _ZENRingCopy(R)  

#if 0
/*
   \TEX{
   \title{Writing to file to an internal representation.}
   \input{A stream |file|, an allocated |ZENRing| |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |Rg| to |file|.}
   \note{In an extension tower, one only needs to save the last ring as the structure is recursively saved on file.}
   } 
   */
int ZENRingPrintToFile(file, Rg)
     FILE *file;
     ZENRing Rg; {}
#endif
#define ZENRingPrintToFile(file, Rg) _ZENRingPrintToFile(file, Rg)

#if 0
/*
   \TEX{
   \title{Reading from file.}
   \input{A stream |file| and a pointer on a ring |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|Rg| is created according to the datas in |file|.}
   } 
   */
int ZENRingReadFromFile(Rg, file)
     FILE *file;
     ZENRing Rg; {}
#endif
#define ZENRingReadFromFile(Rg, file) _ZENRingReadFromFile(&(Rg), file)

#if 0
/* \TEX{

   \Subsection{Ring parameters}
   
   \title{The ``characteristic''.}
   \input{A |ZENRing| |Rg|.}
   \output{The ``characteristic'' of |Rg|.}
   \note{This procedure is a macro which returns one field of
   |Rg|. You must not disallocate it.}
   }
   */
BigNum ZENRingP(Rg)
     ZENRing Rg;{}
#endif
#define ZENRingP(Rg) (Rg)->p
#if 0
/* \TEX{
   \title{The size of the ``characteristic''.}
   \input{A |ZENRing| |Rg|.}
   \output{The size of the ``characteristic'' of |Rg| in base $2^{|SIZE_BLOC|}$.}
   \note{This procedure is a macro which returns one field of
   |Rg|.}
   }
   */
int ZENRingPl(Rg)
     ZENRing Rg;{}
#endif
#define ZENRingPl(Rg) (Rg)->pl

#if 0
/* \TEX{
   \title{The size of the elements.}
   \input{A |ZENRing| |Rg|.}
   \output{The size of the elements of |Rg| in base $2^{|SIZE_BLOC|}$.}
   \note{This procedure is a macro which returns one field of
   |Rg|. Usually it has the same value as |ZENRingPl(Rg)| except if
   the modulus is a power of $2^{|SIZE_BLOC|}$. In this case it is
   equal to $1$ or $|ZENRingPl(Rg)|-1$.}
   } */
int ZENRingSizeElt(Rg)
     ZENRing Rg;{}
#endif
#define ZENRingSizeElt(Rg) (Rg)->zenBnl
#define ZENRingSZ(Rg) (Rg)->zenBnl2

#if 0
/* \TEX{
   \title{The number of elements of |Rg|.}
   \input{A |ZENRing| |Rg|.}
   \output{The number of elements of |Rg|.}
   \note{This procedure is a macro which returns one field of
   |Rg|. You must not disallocate it.}
   }
   */
BigNum ZENRingQ(Rg)
     ZENRing Rg;{}
#endif
#define ZENRingQ(Rg) (Rg)->p2m
#if 0
/* \TEX{
   \title{The size of the number of elements.}
   \input{A |ZENRing| |Rg|.}
   \output{The size of the number of elements of |Rg| in base $2^{|SIZE_BLOC|}$.}
   \note{This procedure is a macro which returns one field of
   |Rg|.}
   }
   */
int ZENRingQl(Rg)
     ZENRing Rg;{}
#endif
#define ZENRingQl(Rg) (Rg)->p2ml

#if 0
/* \TEX{
   \title{The finite subring.}
   \input{A |ZENRing| |Rg|.}
   \output{|NULL| if |Rg| is $\Z/n\Z$, the finite ring |R| if |Rg| is a
   polynomial finite ring defined over |R|.}
   \note{This procedure is a macro which returns one field of
     |Rg|. You must not disallocate it.}
   }
   */
ZENRing ZENRingDef(Rg)
     ZENRing Rg;{}
#endif
#define ZENRingDef(Rg) (Rg)->rg

#if 0
/* \TEX{
   \title{The definition polynomial.}
   \input{A |ZENRing| |Rg|.}
   \output{|NULL| if |Rg| is $\Z/n\Z$, the polynomial which defines
   |Rg| otherwise.}
   \note{This procedure is a macro which returns one field of
   |Rg|. You must not disallocate it.}
   }
   */
ZENPoly ZENRingPol(Rg)
     ZENRing Rg;{}
#endif
#define ZENRingPol(Rg) (Rg)->pol
#if 0
/* \TEX{
   \title{The degree of finite ring.}
   \input{A |ZENRing| |Rg|.}
   \output{0 if |Rg| is $\Z/n\Z$, the degree of the polynomial which
   defines |Rg| otherwise.}
   \note{This procedure is a macro which returns one field of |Rg|.}
   }
   */
int ZENRingDeg(Rg)
     ZENRing Rg;{}
#endif
#define ZENRingDeg(Rg) (ZENRingPol(Rg)?ZENPolyDeg(ZENRingPol(Rg),ZENRingDef(Rg)):0)

#if 0
/* \TEX{
  \title{The number of finite rings defined over a subfinite ring.}
  \input{A |ZENRing| |Rg|.}
  \output{The number of finite rings defined over |Rg|.}
  \note{This procedure is a macro which returns one field of |Rg|.}
  }
*/
int ZENRingExt(Rg)
     ZENRing Rg;{}
#endif
#define ZENRingExt(Rg) ((Rg)->ext)

#if 0
/* \TEX{
   \title{A factor of a modulo.}
   \input{A finite ring |Rg|.}
   \output{A factor of a modulo which defines |Rg| or a subfinite ring
     of |Rg| if a function of the library returned |ZEN_NO_INVERSE|.}
   \note{This procedure is a macro which returns one field of
   |Rg|. You must not disallocate it.
   This field is set by operations such as |ZENEltInverse| when an inverse
   was impossible to compute. Therefore, a call to this function is pertinent
   only after a |ZEN_NO_INVERSE| return of such a function. }
   }
   */
ZENElt ZENRingFact(Rg)
     ZENRing Rg;{}
#endif
#define ZENRingFact(Rg) ((Rg)->fact)
     
     /*
        Undocumented macros
        */
#define ZENRingP1(Rg) ((Rg)->p_1)
#define ZENRingVar(Rg) ((Rg)->var)
#define ZENRingPms(Rg) ((Rg)->pms)
#define ZENRingArith(Rg) ((Rg)->arithmetic)
#define ZENRingHbuf(Rg) ((Rg)->Hbuf)
#define ZENRingIbuf(Rg) ((Rg)->Ibuf)
#define ZENRingJbuf(Rg) ((Rg)->Jbuf)
#define ZENRingBNbuf(Rg) ((Rg)->bnbuf)
#define ZENRingOddPower(Rg) ((Rg)->OddPower)
#define ZENRingETL__(Rg) ((Rg)->ExpTableL)
#define ZENRingETL_(Rg,lb) (*(ZENRingETL__(Rg)+(lb-4)))
#define ZENRingETL(Rg,lb,i) (*(ZENRingETL_(Rg,lb)+i))
#define ZENRingETU__(Rg) ((Rg)->ExpTableU)
#define ZENRingETU_(Rg,lb) (*(ZENRingETU__(Rg)+(lb-4)))
#define ZENRingETU(Rg,lb,i) (*(ZENRingETU_(Rg,lb)+(i-1)))

#define ZENRingTheta(Rg) ((Rg)->Theta)
#define ZENRingSubDeg(Rg) ((Rg)->k)

#define ZepRingPms(R) ZENRingPms(R).zep
#define ZepRingPmsZemp(R) ZepRingPms(R).zemp
#define ZepRingKaraMLim(R) ZepRingPmsZemp(R).KaraMLim
#define ZepRingKaraSLim(R) ZepRingPmsZemp(R).KaraSLim
#define ZepRingKaraBuf(R) ZepRingPmsZemp(R).KaraBuf

#define ZemRingPms(R) ZENRingPms(R).zem
#define ZemRingPmsZemp(R) ZemRingPms(R).zemp
#define ZemRingKaraMLim(R) ZemRingPmsZemp(R).KaraMLim
#define ZemRingKaraSLim(R) ZemRingPmsZemp(R).KaraSLim
#define ZemRingKaraBuf(R) ZemRingPmsZemp(R).KaraBuf

#define ZextRingPms(R) ZENRingPms(R).zext
#define ZENRingPbuf(R) ZextRingPms(R).Pbuf
#define ZENRingRbuf(R) ZextRingPms(R).Rbuf
#define ZENRingTbuf(R) ZextRingPms(R).Tbuf
#define ZENRingTmp1(R) ZextRingPms(R).tmp1
#define ZENRingTmp2(R) ZextRingPms(R).tmp2
#define ZENRingRec(R) ZextRingPms(R).Rec

#define Ze2mRingNbBits(R) ZextRingPms(R).nbbits
#define Ze2mRingPreDivFast(R) ZextRingPms(R).divfast

/* Precomputations */
#define ZENPrcInitEltMultiplyPtr(Rg) ZENFct(Rg).PrcInitEltMultiply
#define ZENPrcInitEltMultiply(Rg) (*(ZENPrcInitEltMultiplyPtr(Rg)))(Rg)

#define ZENPrcCloseEltMultiplyPtr(Rg) ZENFct(Rg).PrcCloseEltMultiply
#define ZENPrcCloseEltMultiply(Rg) (*(ZENPrcCloseEltMultiplyPtr(Rg)))(Rg)

#define ZENPrcInitTracePtr(Rg) ZENFct(Rg).PrcInitTrace
#define ZENPrcInitTrace(Rg) (*(ZENPrcInitTracePtr(Rg)))(Rg)

#define ZENPrcCloseTracePtr(Rg) ZENFct(Rg).PrcCloseTrace
#define ZENPrcCloseTrace(Rg) (*(ZENPrcCloseTracePtr(Rg)))(Rg)

#define ZENPrcInitPolyRootsCanonicalPtr(Rg) ZENFct(Rg).PrcInitPolyRootsCanonical
#define ZENPrcInitPolyRootsCanonical(Rg) (*(ZENPrcInitPolyRootsCanonicalPtr(Rg)))(Rg)

#define ZENPrcClosePolyRootsCanonicalPtr(Rg) ZENFct(Rg).PrcClosePolyRootsCanonical
#define ZENPrcClosePolyRootsCanonical(Rg) (*(ZENPrcClosePolyRootsCanonicalPtr(Rg)))(Rg)

#define ZENPrcInitEltExpPtr(Rg) ZENFct(Rg).PrcInitEltExp
#define ZENPrcInitEltExp(Rg) (*(ZENPrcInitEltExpPtr(Rg)))(Rg)

#define ZENPrcCloseEltExpPtr(Rg) ZENFct(Rg).PrcCloseEltExp
#define ZENPrcCloseEltExp(Rg) (*(ZENPrcCloseEltExpPtr(Rg)))(Rg)


/*
   \TEX{
   \Section[Elements]{Procedures to handle elements of finite rings}

   These procedures are current operations on elements of finite rings.
   }
   */

/*
   \TEX{
   \Subsection{Allocation}
   }
   */
#if 0
/* \TEX{
   \title{Creation.}
   \input{A finite ring |Rg| and a |ZENElt| |a|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|a| is allocated.}
   }
   */
int ZENEltAlloc(a, Rg)
     ZENElt a;
     ZENRing Rg;{}
#endif
#define ZENEltCreatePtr(R) ZENFct(R).EltCreate
#define ZENEltCreate(a,R) (*(ZENEltCreatePtr(R)))(&(a),R)
#define ZENEltAlloc(a,R) ZENEltCreate(a,R)
#define ZENEltCreateLargePtr(R) ZENFct(R).EltCreateLarge
#define ZENEltCreateLarge(a,R) (*(ZENEltCreateLargePtr(R)))(&(a),R)

#if 0
/* \TEX{
   \title{Freeing.}
   \input{A |ZENElt| |a| of a finite ring |Rg| already allocated with
   |ZENEltAlloc|.}
   \side{|a| is freed.}
   }
   */
void ZENEltFree(a, Rg)
     ZENElt a;
     ZENRing Rg;{}
#endif
#define ZENEltFreePtr(R) ZENFct(R).EltFree
#define ZENEltFree(a,R) (*(ZENEltFreePtr(R)))(a,R)

/*
   \TEX{
   \Subsection{Assignment}
   }
   */

#if 0
/* \TEX{
   \title{Assigning.}
   \input{Two allocated elements |a| and |b| of a finite ring |Rg|.}
   \side{|a| is filled with |b|.}
   }
   */
void ZENEltAssign(a, b, Rg)
     ZENElt a, b;
     ZENRing Rg;{}
#endif
#define ZENEltAssignPtr(R) ZENFct(R).EltAssign
#define ZENEltAssign(a,b,R) (*(ZENEltAssignPtr(R)))(&(a),b,R)
#if 0
/* \TEX{
   \title{Setting to zero.}
   \input{An allocated element |a| of a finite ring |Rg|.}
   \side{|a| is set to zero.}
   }
   */
void ZENEltSetToZero(a, Rg)
     ZENElt a;
     ZENRing Rg;{}
#endif
#define ZENEltSetToZeroPtr(R) ZENFct(R).EltSetToZero
#define ZENEltSetToZero(a,R) (*(ZENEltSetToZeroPtr(R)))(&(a),R)
#if 0
/* \TEX{
   \title{Setting to one.}
   \input{An allocated element |a| of a finite ring |Rg|.}
   \side{|a| is set to one.}
   }
   */
void ZENEltSetToOne(a, Rg)
     ZENElt a;
     ZENRing Rg;{}
#endif
#define ZENEltSetToOnePtr(R) ZENFct(R).EltSetToOne
#define ZENEltSetToOne(a,R) (*(ZENEltSetToOnePtr(R)))(&(a),R)
#if 0
/* \TEX{
   \title{Setting to a generator.}
   \input{An allocated element |a| of a finite ring |Rg|.}
   \side{|a| is set to one if |Rg| is a prime field or set to $t$ if
   $|Rg|=R[t]/((P(t))$ where $P(t)$ is a polynomial which defines an
   extension over a sub-ring $R$.}
   }
   */
void ZENEltSetToGenerator(a, Rg)
     ZENElt a;
     ZENRing Rg;{}
#endif
#define ZENEltSetToGenerator(a, R) (ZENRingDef(R) ? ZENZToElt(a, ZENRingQ(ZENRingDef(R)), ZENRingQl(ZENRingDef(R)), R) : ZENEltSetToOne(a, R))
#if 0
/* \TEX{
   \title{Setting to random.}
   \input{An allocated element |a| of a finite ring |Rg|.}
   \side{|a| is set to random.}
   }
   */
void ZENEltSetRandom(a, Rg)
     ZENElt a;
     ZENRing Rg;{}
#endif
#define ZENEltSetRandomPtr(R) ZENFct(R).EltSetRandom
#define ZENEltSetRandom(a,R) (*(ZENEltSetRandomPtr(R)))(&(a),R)
#if 0
/* \TEX{
   \title{Enumerating.}
   \input{An allocated |ZENElt| |a| of a finite ring |Rg|.}
   \side{|a| is set to another element of |Rg|.}
   \note{This procedure can be seen more or less as 
     |ZENEltToZ(n, nl, a, R)|; |ZBNAddCarry(n, nl, 1)|; |ZENZToElt(a,
     n, nl, R)|.
     After a number of call of this function equal to the number
     of elements |ZENRingQ(Rg)|, one obtain the same element.}
   }
   */
void ZENEltSetNext(a, Rg)
     ZENElt a;
     ZENRing Rg;{}
#endif
#define ZENEltSetNextPtr(R) ZENFct(R).EltSetNext
#define ZENEltSetNext(a,R) (*(ZENEltSetNextPtr(R)))(&(a),R)

#if 0
/*
   \TEX{
   \title{Converting a |BigNum| to a |ZENElt|}
   \input{A |ZENElt|, a |BigNum| and its length and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The big integer is first reduced modulo the characteristic of the field. The result is then assigned to |e| using |ZENZToElt|.}
   } 
   */
int ZENEltFromBigNum(e,n,nl,R)
     ZENElt e;
     BigNum n;
     BigNumLength nl;
     ZENRing R;{}
#endif
#define ZENEltFromBigNum(e,n,nl,R) _ZENEltFromBigNum(&(e),n,nl,R)

#if 0
/*
   \TEX{
   \title{Converting a |ZENElt| to a |ZENElt|}
   \input{Two |ZENElt|s, and two |ZENRing|s.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The element |e1| of |R1| is affected with |e2| after possible modular reduction. |R1| and |R2| must be compatible rings.}
   } 
   */
int ZENEltConvert(e1,R1,e2,R2)
     ZENElt e1,e2;
     ZENRing R1,R2;{}
#endif
#define ZENEltConvertPtr(R) ZENFct(R).EltConvert
#define ZENEltConvert(e1,R1,e2,R2) (*(ZENEltConvertPtr(R1)))(&(e1),R1,e2,R2)

/*
   \TEX{
   \Subsection{Test}
   }
   */

#if 0
/* \TEX{
   \title{Equality.}
   \input{Two allocated elements |a| and |b| of a finite ring |Rg|.}
   \output{The predicate $|a| = |b|$.}
   }
   */
int ZENEltAreEqual(a, b, Rg)
     ZENElt a, b;
     ZENRing Rg;{}
#endif
#define ZENEltAreEqualPtr(R) ZENFct(R).EltAreEqual
#define ZENEltAreEqual(a,b,R) (*(ZENEltAreEqualPtr(R)))(a,b,R)
#if 0
/* \TEX{
   \title{Is Zero.}
   \input{An allocated element |a| of a finite ring |Rg|.}
   \output{The predicate $|a| = 0$.}
   }
   */
int ZENEltIsZero(a, Rg)
     ZENElt a;
     ZENRing Rg;{}
#endif
#define ZENEltIsZeroPtr(R) ZENFct(R).EltIsZero
#define ZENEltIsZero(a,R) (*(ZENEltIsZeroPtr(R)))(a,R)
#if 0
/* \TEX{
   \title{Is one.}
   \input{An allocated element |a| of a finite ring |Rg|.}
   \output{The predicate $|a| = 1$.}
   }
   */
int ZENEltIsOne(a, Rg)
     ZENElt a;
     ZENRing Rg;{}
#endif
#define ZENEltIsOnePtr(R) ZENFct(R).EltIsOne
#define ZENEltIsOne(a,R) (*(ZENEltIsOnePtr(R)))(a,R)
#if 0
/* \TEX{
   \title{Is a square in finite fields.}
   \input{An element |a| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, 1 if |a| is a square, 0 otherwise.}
   \note{This procedure is valid only in finite fields. The algorithm
     used is an exponantiation when the characteristic is odd. In
     characteristic 2 the answer is always 1!}
   }
   */
int ZENEltIsASquare(a, Rg)
     ZENElt a;
     ZENRing Rg;{}
#endif
#define ZENEltIsASquarePtr(R) ZENFct(R).EltIsASquare
#define ZENEltIsASquare(a,R) (*(ZENEltIsASquarePtr(R)))(a,R)

/*
   \TEX{
   \Subsection{Arithmetic}
   }
   */
#if 0
/* \TEX{
   \title{Addition.}
   \input{Two allocated elements |a| and |b| of a finite ring |Rg|.}
   \side{$|b += a|$.}
   }
   */
void ZENEltAdd(b, a, Rg)
     ZENElt b, a;
     ZENRing Rg;{}
#endif
#define ZENEltAddPtr(R) ZENFct(R).EltAdd
#define ZENEltAdd(c,a,R) (*(ZENEltAddPtr(R)))(&(c),a,R)
#if 0
/* \TEX{
   \title{Negation.}
   \input{Two allocated elements |a| and |b| of a finite ring |Rg|.}
   \side{$|b| = -|a|$.}
   }
   */
void ZENEltNegate(b, a, Rg)
     ZENElt b, a;
     ZENRing Rg;{}
#endif
#define ZENEltNegatePtr(R) ZENFct(R).EltNegate
#define ZENEltNegate(c,a,R) (*(ZENEltNegatePtr(R)))(&(c),a,R)
#if 0
/* \TEX{
   \title{Subtract.}
   \input{Two allocated elements |a| and |b| of a finite ring |Rg|.}
   \side{$|b -= a|$.}
   }
   */
void ZENEltSubtract(b, a, Rg)
     ZENElt b, a;
     ZENRing Rg;{}
#endif
#define ZENEltSubtractPtr(R) ZENFct(R).EltSubtract
#define ZENEltSubtract(c,a,R) (*(ZENEltSubtractPtr(R)))(&(c),a,R)
#define ZENEltSubtractLargePtr(R) ZENFct(R).EltSubtractLarge
#define ZENEltSubtractLarge(c,a,R) (*(ZENEltSubtractLargePtr(R)))(&(c),a,R)
#if 0
/* \TEX{
   \title{Squaring.}
   \input{Two allocated elements |a| and |b| of a finite ring |Rg|.}
   \side{$|b| = |a|^2$.}
   }
   */
void ZENEltSquare(b, a, Rg)
     ZENElt b, a;
     ZENRing Rg;{}
#endif
#define ZENEltSquarePtr(R) ZENFct(R).EltSquare
#define ZENEltSquare(c,a,R) (*(ZENEltSquarePtr(R)))(&(c),a,R)
#if 0
/* \TEX{
   \title{Multiplication.}
   \input{Three allocated elements |a|, |b| and |c| of a finite ring |Rg|.}
   \side{$|c| = |a|\times |b|$.}
   \note{ACHTUNG !!! One must have $|c| \neq |a|$. A call to this function with the same element as parameter |c| and |a| can cause segmentation faults. }
   }
   */
void ZENEltMultiply(c, a, b, Rg)
     ZENElt c, a, b;
     ZENRing Rg;{}
#endif
#define ZENEltMultiplyPtr(R) ZENFct(R).EltMultiply
#define ZENEltMultiply(c,a,b,R) (*(ZENEltMultiplyPtr(R)))(&(c),a,b,R)
#if 0
/* \TEX{
   \title{Inverse.}
   \input{Two allocated elements |a| and |b| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|b| = 1/a$ if |a| has an inverse, |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   \note{One can have $|b|=|a|$}
   }
   */
int ZENEltInverse(b, a, Rg)
     ZENElt b, a;
     ZENRing Rg;{}
#endif
#define ZENEltInversePtr(R) ZENFct(R).EltInverse
#define ZENEltInverse(c,a,R) (*(ZENEltInversePtr(R)))(&(c),a,R)
#if 0
/* \TEX{
   \title{Exponentiation.}
   \input{Two allocated elements |R| and |P| of a finite ring |Rg|, a
   |BigNum| |k| of size |kl|.}
   \output{0 if no error occured, |ZENERR| otherwise.}
   \side{$|R| = |P|^{|k|}$.}
   \note{The algorithm used is by default the binary method, but after
     precomputation, it is the m-ary method.}
     }
   */
int ZENEltExp (R, k, kl, P, Rg)
     ZENElt R, P;
     BigNum k;
     BigNumLength kl;
     ZENRing Rg;{}
#endif
#define ZENEltExpPtr(R) ZENFct(R).EltExp
#define ZENEltExp(c,k,kl,a,R) (*(ZENEltExpPtr(R)))(&(c),k,ZBNNumDigits(k,kl),a,R)
#define ZENEltExpPlain(c,k,kl,a,R) _ZENEltExpPlain(&(c),k,ZBNNumDigits(k,kl),a,R)
     
#if 0
/* \TEX{
   \title{``Trace''.}
   \input{Two allocated elements |a| and |b| of a finite ring |Rg|.}
   \side{$|b| = \mbox{Tr}_{|Rg|/R}(|a|)$ where |R| is the base ring
     of |Rg|. If |Rg| is $\Z/n\Z$ where |n| is an integer, $|b|=|a|$.}   
   }
   */
void ZENEltTrace(b, a, Rg)
     ZENElt b, a;
     ZENRing Rg;{}
#endif
#define ZENEltTracePtr(R) ZENFct(R).EltTrace
#define ZENEltTrace(c,a,R) (*(ZENEltTracePtr(R)))(&(c),a,R)

#if 0
/* \TEX{
   \title{``Absolute Trace''.}
   \input{Two allocated elements |e| and |f| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_HAS_INVERSE| otherwise.}
   \side{$|f| = \mbox{Tr}_{|Rg|/(Z/pZ)}(|e|)$ where $p$ is the
   characteristic of |Rg|. If |Rg|=$\Z/p\Z$, $|f|=|e|$.}
   }
   */
int ZENEltAbsoluteTrace(e, f, Rg)
     ZENElt e, f;
     ZENRing Rg;{}
#endif
#define ZENEltAbsoluteTrace(e, f, Rg) _ZENEltAbsoluteTrace(&(e), f, Rg)

#if 0
/* \TEX{
   \title{Square roots in finite fields}.
   \input{Two elements |R| and |P| of a finite field |Rg|.}
   \output{|ZENERR| if an error occurred, 0 if |P| is a square, 1 otherwise.}
   \side{|R| is filled with the square root of |P| if
   0 is returned.}
   \note{This procedure is valid only in finite fields. The algorithm
     of Tonelli and Shanks is used.}
   }
   */
int ZENEltSquareRoot(R, P, Rg)
     ZENElt R, P;
     ZENRing Rg;{}
#endif
#define ZENEltSquareRootPtr(R) ZENFct(R).EltSquareRoot
#define ZENEltSquareRoot(c,a,R) (*(ZENEltSquareRootPtr(R)))(&(c),a,R)


/*
   \TEX{
   \Subsection{Input/Output}
   }
   */
#if 0
/* \TEX{
   \title{Converting from string.}
   \input{A finite ring |Rg|, an allocated element |G| and a string
   |s| representing a |ZENElt| in base $|base|\in \{2,\ldots,
   16\}$.} 
   \output{The number of character read in |s| or |ZENERR| if an error occurred.}
   \side{|G| is filled with |s| if the output is not |ZENERR|.}
   }
*/
int ZENEltReadFromString(G, s, base, Rg)
     ZENElt G;
     char *s;
     int base;
     ZENRing Rg;{}
#endif
#define ZENEltReadFromStringPtr(R) ZENFct(R).EltReadFromString
#define ZENEltReadFromString(a,s,base,R) (*(ZENEltReadFromStringPtr(R)))(&(a),s,base,R)
#if 0
/* \TEX{
   \title{Converting to string.}
   \input{An allocated element |G| of a finite ring |Rg| and a base
   $|base|\in \{2,\ldots, 16\}$.} 
   \output{An allocated string representing |G| in base |base| or |ZENNULL|
   if an error occurred.} 
   }
   */
char *ZENEltPrintToString(G, base, Rg)
     ZENElt G;
     int base;
     ZENRing Rg;{}
#endif
#define ZENEltPrintToStringPtr(R) ZENFct(R).EltPrintToString
#define ZENEltPrintToString(a,base,R) (*(ZENEltPrintToStringPtr(R)))(a,base,R)
#if 0
/* \TEX{
   \title{Reading from file.}
   \input{A stream |file|, an allocated element |G| of a finite ring
   |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|G| is filled with the |ZENElt| read in |file| if no error
   occurred.}
   }
   */
int ZENEltReadFromFile(G, file, base, Rg)
     ZENElt G;
     FILE *file;
     int base;
     ZENRing Rg;{}
#endif
#define ZENEltReadFromFilePtr(R) ZENFct(R).EltReadFromFile
#define ZENEltReadFromFile(a,file,base,R) (*(ZENEltReadFromFilePtr(R)))(&(a),file,base,R)
#if 0
/* \TEX{
   \title{Printing to file.}
   \input{A stream |file|, an allocated element |G| of a finite ring
   |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{Printing a representation of |G| in base |base| to |file|.}
   }
   */
int ZENEltPrintToFile(file, G, base, Rg)
     FILE *file;
     ZENElt G;
     int base;
     ZENRing Rg;{}
#endif
#define ZENEltPrintToFilePtr(R) ZENFct(R).EltPrintToFile
#define ZENEltPrintToFile(file,a,base,R) (*(ZENEltPrintToFilePtr(R)))(file,a,base,R)
#if 0
/* \TEX{
   \title{Converting from a string to an internal representation.}
   \input{An allocated element |G| of a finite ring |Rg| and a string
   |s| representing a |ZENElt| to an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occurred.}
   \side{|G| is filled with |s|}
   }
   */
int ZENEltGetFromString(G, s, Rg)
     char *s;
     ZENElt G;
     ZENRing Rg;{}
#endif
#define ZENEltGetFromStringPtr(R) ZENFct(R).EltGetFromString
#define ZENEltGetFromString(a,s,R) (*(ZENEltGetFromStringPtr(R)))(&(a),s,R)
#if 0
/* \TEX{
   \title{Converting to string to an internal representation.}
   \input{An allocated element |G| of a finite ring |Rg|.}
   \output{An allocated string representing |G| to an internal
   representation or |ZENNULL| if an error occurred.} 
   }
   */
char *ZENEltPutToString(G, Rg)
     ZENElt G;
     ZENRing Rg;{}
#endif
#define ZENEltPutToStringPtr(R) ZENFct(R).EltPutToString
#define ZENEltPutToString(a,R) (*(ZENEltPutToStringPtr(R)))(a,R)
#if 0
/* \TEX{
   \title{Getting from file to an internal representation.}
   \input{A stream |file|, and an element |G| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{|G| is filled with the element read in |file|}
   }
   */
int ZENEltGetFromFile(G, file, Rg)
     ZENElt G;
     FILE *file;
     ZENRing Rg;{}
#endif
#define ZENEltGetFromFilePtr(R) ZENFct(R).EltGetFromFile
#define ZENEltGetFromFile(a,file,R) (*(ZENEltGetFromFilePtr(R)))(&(a),file,R)
#if 0
/* \TEX{
   \title{Writing to file to an internal representation.}
   \input{A stream |file|, an allocated element |G| of a finite ring
   |Rg|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{Printing a representation of |G| to |file|.}
   }
   */
int ZENEltPutToFile(file, G, Rg)
     FILE *file;
     ZENElt G;
     ZENRing Rg;{}
#endif
#define ZENEltPutToFilePtr(R) ZENFct(R).EltPutToFile
#define ZENEltPutToFile(file,a,R) (*(ZENEltPutToFilePtr(R)))(file,a,R)

#if 0
/* \TEX{
   \title{Evaluating an element as a multivariate polynomial over $Z$.}
   \input{A Bignum |p| of allocated size greater or equal than
   |ZENRingQl(Rg)|, a pointer |p_pl| on the real size of |p|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|p| is filled with the value of |G| considered as a
   multivariate polynomial in which we substitute all the variables
   with the ``characteristic'', |*p_pl| contains the real size of |p|.}
   } */
void ZENEltToZ(p, p_pl, G, Rg)
     BigNum p;
     BigNumLength *p_pl;
     ZENElt G;
     ZENRing Rg;{}
#endif
#define ZENEltToZPtr(R) ZENFct(R).EltToZ
#define ZENEltToZ(p,p_pl,a,R) (*(ZENEltToZPtr(R)))(p,p_pl,a,R)

#if 0
/* \TEX{
   \title{Getting an element from its evaluation as a multivariate
   polynomial over $Z$.}
   \input{A Bignum |p| of size |pl|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|G| is filled with the element whose evaluation considered as a
   multivariate polynomial in which we substitute all the variables
   with the ``characteristic'' is equal to (|p|, |pl|).}
   } */
void ZENZToElt(G, p, pl, Rg)
     BigNum p;
     BigNumLength pl;
     ZENElt G;
     ZENRing Rg;{}
#endif
#define ZENZToEltPtr(R) ZENFct(R).ZToElt
#define ZENZToElt(G,p,pl,R) (*(ZENZToEltPtr(R)))(&(G),p,pl,R)
#if 0
/* \TEX{
   \title{Clone conversion.}
   \input{A |ZENElt| |B| of the original ring of |Rg| and a |ZENElt| |C| of the clone |Rg|.}
   \side{|C| is filled with |B|.}
   } */
void ZENEltToClone(C, B, Rg)
     ZENElt C,B;
     ZENRing Rg;{}
#endif
#define ZENEltToClonePtr(R) ZENFct(R).EltToClone
#define ZENEltToClone(C,B,R) (*(ZENEltToClonePtr(R)))(&(C),B,R)

#if 0
/* \TEX{
   \title{Clone conversion.}
   \input{A |ZENElt| |B| of the original ring of |Rg| and a |ZENElt| |C| of the clone |Rg|.}
   \side{|B| is filled with |C|.}
   } */
void ZENCloneToElt(B, C, Rg)
     ZENElt B,C;
     ZENRing Rg;{}
#endif
#define ZENCloneToEltPtr(R) ZENFct(R).CloneToElt
#define ZENCloneToElt(B,C,R) (*(ZENCloneToEltPtr(R)))(&(B),C,R)

/* Undocumented macros */

#define ZENElt2Int(elt,Rg)    (elt).l
#define ZENElt2Bn(elt,Rg)     (elt).b
#define ZENElt2Pol(elt,Rg)    (elt).p
#define ZENElt2Frac(elt,Rg)   (elt).f
#define ZENElt2Elts(elt,Rg)   (elt).c

/* \TEX{
   \Section[Polynomials]{Procedures to handle polynomials over finite rings}

   These procedures are current operations on polynomials over finite rings.
   }
   */

/* \TEX{
   \Subsection{Allocation}
   }
   */
#if 0
/* \TEX{
   \title{Creation.}
   \input{An unallocated |ZENPoly|, a finite ring |Rg| and a degree |deg|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The polynomial is allocated together with its |deg|+1 coefficients.}
   \note{All the coefficients of the polynomial are NOT set to zero. Hence, to create and set a polynomial to $a_{|n|}X^{|n|}+a_{|n|-1}X^{|n|-1}+\cdots +a_0$, one will have to:
   \begin{enumerate}
   \item Create a polynomial with |ZENPolyAlloc(PX, n, Rg)|.
   \item Set it to $X^{|n|}$ in order to initialize it with
     |ZENPolySetToXi(PX, n, Rg)|. 
   \item Set all the non null coefficients $a_i$, $0\leq i \leq |n|$
     using |ZENPolySetCoeff()|. 
   \end{enumerate}}
   }
   */
int ZENPolyAlloc(PX, deg, Rg)
     ZENPoly PX;
     int deg;
     ZENRing Rg;{}
#endif
#define ZENPolyCreatePtr(R) ZENFct(R).PolyCreate
#define ZENPolyCreate(deg,R) (*(ZENPolyCreatePtr(R)))(deg,R)
#define ZENPolyAlloc(P,deg,R) (((P)=ZENPolyCreate(deg,R))==NULL ? ZERR : 0 )
#define ZENPolyCreateLargePtr(R) ZENFct(R).PolyCreateLarge
#define ZENPolyCreateLarge(deg,R) (*(ZENPolyCreateLargePtr(R)))(deg,R)

#if 0
/* \TEX{
   \title{Degree.}
   \input{An allocated polynomial |PX| of a finite ring |Rg|.}
   \output{The degree of |PX|.}
   \note{This procedure is a macro which returns one field of
   |PX|. You can assign |ZENPolyDeg(PX, Rg)|, but beware that all functions assume this field correct.}
   }
   */
int ZENPolyDeg(PX, Rg)
     ZENPoly PX;
     ZENRing Rg;{}
#endif
#define ZENPolyDeg(p,Rg) (p)->deg
#if 0
/* \TEX{
   \title{Length.}
   \input{An allocated polynomial |PX| of a finite ring |Rg|.}
   \output{The maximal degree of |PX|.}
   \note{This procedure is a macro which returns one field of
   |PX|. Assigning |ZENPolyLgt(PX, Rg)| can lead to bugs.}
   }
   */
int ZENPolyLgt(PX, Rg)
     ZENPoly PX;
     ZENRing Rg;{}
#endif
#define ZENPolyLgt(p,Rg) (p)->lgt

#if 0
/* \TEX{
   \title{Copying an allocated polynomial.}
   \input{An allocated polynomial |PX| of a finite ring |Rg|.}
   \output{|ZENNULL| if an error occurred, a copy of |PX| otherwise.}   
   }
   */
ZENPoly ZENPolyCopy (PX, Rg)
     ZENPoly PX;
     ZENRing Rg;{}
#endif
#define ZENPolyCopy(a, Rg) _ZENPolyCopy(a, Rg) 


#if 0
/* \TEX{
   \title{Freeing.}
   \input{An allocated polynomial |PX| of a finite ring |Rg|.}
   \side{|PX| is disallocated.}
   }
   */
void ZENPolyFree(PX,Rg)
     ZENPoly PX;
     ZENRing Rg;{}
#endif
#define ZENPolyFreePtr(R) ZENFct(R).PolyFree
#define ZENPolyFree(p,R) (*(ZENPolyFreePtr(R)))(p,R)

/* \TEX{
   \Subsection{Assigning}
   }
   */
#if 0
/* \TEX{
   \title{Assigning.}
   \input{Two allocated polynomials |PX| and |RX| of a finite ring |Rg|.}
   \side{|RX| is filled with |PX|.}
   }
   */
void ZENPolyAssign(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;{}
#endif
#define ZENPolyAssignPtr(R) ZENFct(R).PolyAssign
#define ZENPolyAssign(r,p,R) (*(ZENPolyAssignPtr(R)))(r,p,R)
#define ZENPolyAssignLargePtr(R) ZENFct(R).PolyAssignLarge
#define ZENPolyAssignLarge(r,p,R) (*(ZENPolyAssignLargePtr(R)))(r,p,R)
#define ZENPolyLazyToVeryLazyPtr(R) ZENFct(R).PolyLazyToVeryLazy
#define ZENPolyLazyToVeryLazy(r,p,R) (*(ZENPolyLazyToVeryLazyPtr(R)))(r,p,R)
#if 0
/* \TEX{
   \title{Setting to zero.}
   \input{An allocated polynomial |RX| of a finite ring |Rg|.}
   \side{|RX| is set to zero.}
   }
   */
void ZENPolySetToZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;{}
#endif
#define ZENPolySetToZeroPtr(R) ZENFct(R).PolySetToZero
#define ZENPolySetToZero(p,R) (*(ZENPolySetToZeroPtr(R)))(p,R)
#define ZENPolySetToZeroLargePtr(R) ZENFct(R).PolySetToZeroLarge
#define ZENPolySetToZeroLarge(p,R) (*(ZENPolySetToZeroLargePtr(R)))(p,R)
#if 0
/* \TEX{
   \title{Setting to $X^i$.}
   \input{An allocated polynomial |RX| of a finite ring |Rg| allocated at least for
   degree $|deg|$.}
   \side{|RX| is set to $X^{|deg|}$.}
   }
   */
void ZENPolySetToXi(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;{}
#endif
#define ZENPolySetToXiPtr(R) ZENFct(R).PolySetToXi
#define ZENPolySetToXi(p,i,R) (*(ZENPolySetToXiPtr(R)))(p,i,R)
#if 0
/* \TEX{
   \title{Setting randomly.}
   \input{An allocated polynomial |RX| of length at least |deg| of a
   finite ring |Rg|.}
   \side{|RX| is set randomly to a polynomial of degree |deg|.}
   }
   */
void ZENPolySetRandom(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;{}
#endif
#define ZENPolySetRandomPtr(R) ZENFct(R).PolySetRandom
#define ZENPolySetRandom(p,i,R) (*(ZENPolySetRandomPtr(R)))(p,i,R)

#if 0
/* \TEX{
   \title{Setting a coefficient.}
   \input{An initialized polynomial |RX| of degree greater or equal to |d|
   and an element |b| of a
   finite ring |Rg|, the degree |d| of the coefficient  to set.}
   \side{The coefficient  of $X^d$ in |RX| is set to |b|.}
   \note{The degree of |RX| is NOT updated by this operation.}
   }
   */
void ZENPolySetCoeff(RX, d, b, Rg)
     ZENPoly RX;
     int d;
     ZENElt b;
     ZENRing Rg;{}
#endif
#define ZENPolySetCoeffPtr(R) ZENFct(R).PolySetCoeff
#define ZENPolySetCoeff(p,i,c,R) (*(ZENPolySetCoeffPtr(R)))(p,i,c,R)
#if 0
/* \TEX{
   \title{Getting a coefficient.}
   \input{An allocated polynomial |RX| of a finite ring |Rg|, the
   degree |d| of the coefficient  to get and an element |b| to assign.} 
   \side{|b| is filled with the coefficient of $X^d$ in |RX|.}
   }
   */
void ZENPolyGetCoeff(b, RX, d, Rg)
     ZENElt b;
     ZENPoly RX;
     int d;
     ZENRing Rg;{}
#endif
#define ZENPolyGetCoeffNewPtr(R) ZENFct(R).PolyGetCoeff
#define ZENPolyGetCoeff(c,p,i,R) (*(ZENPolyGetCoeffNewPtr(R)))(&(c),p,i,R)
#if 0
/* \TEX{
   \title{Extracting a coefficient.}
   \input{An allocated polynomial |RX| of a finite ring |Rg|, the
   degree |d| of the coefficient to get.}
   \output{The pointer to the coefficient  of $X^d$ in |RX|.}
   \side{You must not disallocate the output.}
   }
   */
ZENElt ZENPolyGetCoeffPtr(RX, d, Rg)
     ZENPoly RX;
     int d;
     ZENRing Rg;{}
#endif
#define ZENPolyGetCoeffPtrPtr(R) ZENFct(R).PolyGetCoeffPtr
#define ZENPolyGetCoeffPtr(p,i,R) (*(ZENPolyGetCoeffPtrPtr(R)))(p,i,R)

#if 0
/* \TEX{
   \title{Updating the degree of a polynomial.}
   \input{An allocated polynomial |RX| of a finite ring |Rg|.}
   \side{Resetting the degree of |RX|, assuming it has decreased.}
   }
   */
void ZENPolyUpdateDegree(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;{}
#endif
#define ZENPolyUpdateDegreePtr(R) ZENFct(R).PolyUpdateDegree
#define ZENPolyUpdateDegree(p,R) (*(ZENPolyUpdateDegreePtr(R)))(p,R)

#if 0
/* \TEX{
   \title{Convert.}
   \input{Two |ZENPoly|s and two |ZENRing|s.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The coefficients of |P1| of |R1| are affected with coefficients of |P2| of |R2| after possible modular reduction of those coefficients. |R1| and |R2| must be compatible rings.}
   }
   */
int ZENPolyConvert(P1, R1, P2, R2)
     ZENPoly P1, P2;
     ZENRing R1, R2; {}
#endif
#define ZENPolyConvert(P1, R1, P2, R2) _ZENPolyConvert(P1, R1, P2, R2)
     
/* \TEX{
   \Subsection{Test}
   }
   */
#if 0
/* \TEX{
   \title{Equality.}
   \input{Two allocated polynomials |RX| and |PX| of a finite ring |Rg|.}
   \output{The predicate $|RX| = |PX|$.}
   }
   */
int ZENPolyAreEqual(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;{}
#endif
#define ZENPolyAreEqualPtr(R) ZENFct(R).PolyAreEqual
#define ZENPolyAreEqual(p,q,R) (*(ZENPolyAreEqualPtr(R)))(p,q,R)
#if 0
/* \TEX{
   \title{Is zero.}
   \input{An allocated polynomial |RX| of a finite ring |Rg|.}
   \output{The predicate $|RX| = 0$.}
   }
   */
int ZENPolyIsZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;{}
#endif
#define ZENPolyIsZeroPtr(R) ZENFct(R).PolyIsZero
#define ZENPolyIsZero(p,R) (*(ZENPolyIsZeroPtr(R)))(p,R)
#if 0
/* \TEX{
   \title{Is a polynomial equal to $X^i$.}
   \input{An allocated polynomial |PX| of a finite ring |Rg| and a
     degree |deg|.}
   \output{The predicate $|PX| = X^{|deg|}$.}
   }
   */
int ZENPolyIsXi (PX, deg, Rg)
     ZENPoly PX;
     int deg;
     ZENRing Rg;{}
#endif
#define ZENPolyIsXiPtr(R) ZENFct(R).PolyIsXi
#define ZENPolyIsXi(p,i,R) (*(ZENPolyIsXiPtr(R)))(p,i,R)

/* \TEX{
   \Subsection{Arithmetic}
   }
   */
#if 0
/* \TEX{
   \title{Addition.}
   \input{Two allocated polynomials |RX| and |PX| of a finite ring |Rg|.}
   \side{$|RX| += |PX|$.}
   }
   */
void ZENPolyAdd(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;{}
#endif
#define ZENPolyAddPtr(R) ZENFct(R).PolyAdd
#define ZENPolyAdd(r,p,R) (*(ZENPolyAddPtr(R)))(r,p,R)
#define ZENPolyAddLargePtr(R) ZENFct(R).PolyAddLarge
#define ZENPolyAddLarge(r,p,R) (*(ZENPolyAddLargePtr(R)))(r,p,R)
#if 0
/* \TEX{
   \title{Negation.}
   \input{Two allocated polynomials |RX| and |PX| of a finite ring |Rg|.}
   \side{$|RX| = -|PX|$.}
   }
   */
void ZENPolyNegate(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;{}
#endif
#define ZENPolyNegatePtr(R) ZENFct(R).PolyNegate
#define ZENPolyNegate(r,p,R) (*(ZENPolyNegatePtr(R)))(r,p,R)
#define ZENPolyNegateLargePtr(R) ZENFct(R).PolyNegateLarge
#define ZENPolyNegateLarge(r,p,R) (*(ZENPolyNegateLargePtr(R)))(r,p,R)
#if 0
/* \TEX{
   \title{Subtraction.}
   \input{Two allocated polynomials |RX| and |PX| of a finite ring |Rg|.}
   \side{$|RX| -= |PX|$.}
   }
   */
void ZENPolySubtract(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;{}
#endif
#define ZENPolySubtractPtr(R) ZENFct(R).PolySubtract
#define ZENPolySubtract(r,p,R) (*(ZENPolySubtractPtr(R)))(r,p,R)
#define ZENPolySubtractLargePtr(R) ZENFct(R).PolySubtractLarge
#define ZENPolySubtractLarge(r,p,R) (*(ZENPolySubtractLargePtr(R)))(r,p,R)
#if 0
/* \TEX{
   \title{Squaring.}
   \input{Two allocated polynomial |RX| and |PX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int ZENPolySquare(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{}
#endif
#define ZENPolySquarePtr(R) ZENFct(R).PolySquare
#define ZENPolySquare(r,p,R) (*(ZENPolySquarePtr(R)))(r,p,R)

#define ZENPolySquareLazyPtr(R) ZENFct(R).PolySquareLazy
#define ZENPolySquareLazy(r,p,R) (*(ZENPolySquareLazyPtr(R)))(r,p,R)

#define ZENPolySquareVeryLazyPtr(R) ZENFct(R).PolySquareVeryLazy
#define ZENPolySquareVeryLazy(r,p,R) (*(ZENPolySquareVeryLazyPtr(R)))(r,p,R)
#if 0
/* \TEX{
   \title{Multiplication.}
   \input{Three allocated polynomial |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|$. The Karatsuba's algorithm is
     always used.}
   }
   */
int ZENPolyMultiply(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;{}
#endif
#define ZENPolyMultiplyPtr(R) ZENFct(R).PolyMultiply
#define ZENPolyMultiply(r,p,q,R) (*(ZENPolyMultiplyPtr(R)))(r,p,q,R)
#define ZENPolyMultiplyLazyPtr(R) ZENFct(R).PolyMultiplyLazy
#define ZENPolyMultiplyLazy(r,p,q,R) (*(ZENPolyMultiplyLazyPtr(R)))(r,p,q,R)
#define ZENPolyMultiplyVeryLazyPtr(R) ZENFct(R).PolyMultiplyVeryLazy
#define ZENPolyMultiplyVeryLazy(r,p,q,R) (*(ZENPolyMultiplyVeryLazyPtr(R)))(r,p,q,R)
#if 0
/* \TEX{
   \title{Multiplication by a scalar.}
   \input{Two polynomials |PX|, |RX| and an element |e| of a
     finite ring |Rg|.}
   \note{$|RX| = |e||PX|$.}
   }
         */
void ZENPolyMultiplyScalar(RX, PX, e, Rg)
     ZENPoly RX, PX;
     ZENElt e;
     ZENRing Rg;{}
#endif
#define ZENPolyMultiplyScalarPtr(R) ZENFct(R).PolyMultiplyScalar
#define ZENPolyMultiplyScalar(r,p,e,R) (*(ZENPolyMultiplyScalarPtr(R)))(r,p,e,R)
#if 0
/* \TEX{
   \title{Scalar product.}
   \input{Two polynomials |PX|, |RX| and an allocated element |e| of a
   finite ring |Rg|.}
   \side{$|e| = \sum p_iq_i$ where $p_i$ and $q_i$ are the coefficients of
   |PX| and |QX|.}
   }
   */
void ZENPolyDot(e, PX, QX, Rg)
     ZENElt e;
     ZENPoly PX, QX;
     ZENRing Rg;{}
#endif
#define ZENPolyDotPtr(R) ZENFct(R).PolyDot
#define ZENPolyDot(r,p,q,R) (*(ZENPolyDotPtr(R)))(&(r),p,q,R)
#if 0
/* \TEX{
   \title{Evaluation.}
   \input{A polynomial |PX| and two elements |f| and |e| of a
   finite ring |Rg|.}
   \side{$|f| = |PX|(e)$.}
   }
   */
void ZENPolyEval(f, PX, e, Rg)
     ZENElt f, e;
     ZENPoly PX;
     ZENRing Rg;{}
#endif
#define ZENPolyEvalPtr(R) ZENFct(R).PolyEval
#define ZENPolyEval(r,p,q,R) (*(ZENPolyEvalPtr(R)))(&(r),p,q,R)
#if 0
/* \TEX{
   \title{Make a polynomial monic.}
   \input{Two polynomials |PX| and |RX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise} 
   \side{|R(X)| is filled with |P(X)| divided by its highest coefficient.}
   }
   */
int ZENPolyMakeMonic (RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;{}
#endif
#define ZENPolyMakeMonic(r,p,Rg) _ZENPolyMakeMonic(r,p,Rg) 
#if 0
/* \TEX{
  \title{Divide.}
  \input{Two polynomials |PX| and |QX| to divide in a finite ring
    |Rg|. |RX| will be the quotient and |MX| the remainder.}
  \output{|ZEN_HAS_INVERSE| if no error occurred, |ZEN_NO_INVERSE| if
    a factor of a modulo was discovered, |ZENERR| for an
    error} 
  \side{|RX| is filled with the euclidian quotient of |PX| by |QX|,
    |MX| with the remainder of |PX| by |QX|.}
   }
   */
int ZENPolyDivide(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;{}
#endif
#define ZENPolyDividePtr(R) ZENFct(R).PolyDivide
#define ZENPolyDivide(r,m,p,q,R) (*(ZENPolyDividePtr(R)))(r,m,p,q,R)
#define ZENPolyDivideLazyPtr(R) ZENFct(R).PolyDivideLazy
#define ZENPolyDivideLazy(r,m,p,q,R) (*(ZENPolyDivideLazyPtr(R)))(r,m,p,q,R)
#define ZENPolyDivideVeryLazyPtr(R) ZENFct(R).PolyDivideVeryLazy
#define ZENPolyDivideVeryLazy(r,m,p,q,R) (*(ZENPolyDivideVeryLazyPtr(R)))(r,m,p,q,R)

#define ZENPolyModuloLazyPtr(R) ZENFct(R).PolyModuloLazy
#define ZENPolyModuloLazy(m,p,q,k,R) (*(ZENPolyModuloLazyPtr(R)))(m,p,q,k,R)
#define ZENPolyModuloVeryLazyPtr(R) ZENFct(R).PolyModuloVeryLazy
#define ZENPolyModuloVeryLazy(m,p,q,k,R) (*(ZENPolyModuloVeryLazyPtr(R)))(m,p,q,k,R)

#define ZENPolyReciprocal(r, p, R)  _ZENPolyReciprocal(r, p, R) 
#define ZENPolyBinaryReciprocal(r, p, R)  _ZENPolyBinaryReciprocal(r, p, R) 
#if 0
/* \TEX{
   \title{Gcd of 2 polynomials.}
   \input{Three polynomials |PX|, |QX| and |RX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise} 
   \side{|R|$(X) = gcd($|P|$(X), $|Q|$(X))$ if |ZEN_HAS_INVERSE| is
   returned, |ZENRingFact(Rg)| is filled with a factor of a modulo
   if |ZEN_NO_INVERSE| is returned.}
   }
   */
int ZENPolyGcd (RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;{}
#endif
#define ZENPolyGcd(g, p, q, Rg) _ZENPolyGcd(g, p, q, Rg) 
#if 0
/* \TEX{
   \title{Extended gcd of 2 polynomials.}
   \input{Three polynomials |IX0|, |AX0| and |BX0| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, -2 if the gcd of |AX0| and |BX0|
   is not 1, |ZEN_NO_INVERSE| if a factor of a modulo was found,
   |ZEN_HAS_INVERSE| otherwise} 
   \side{|IX0|$ = 1$/|BX0|$\ \bmod\ $|AX0| if |ZEN_HAS_INVERSE| is
   returned, |IX0| is gcd(|AX0|, |BX0|) if -2 is returned,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int ZENPolyExtGcd (IX0, BX0, AX0, Rg)
     ZENPoly IX0, AX0, BX0;
     ZENRing Rg;{}
#endif
#define ZENPolyExtGcd(g, p, q, Rg) _ZENPolyExtGcd(g, p, q, Rg) 

#if 0
/* \TEX{
   \title{Resultant of 2 polynomials.}
   \input{Three polynomials |*Res|, |A| and |B| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{|*Res|$(X) = resultant($|A|$(X), $|B|$(X))$ if |ZEN_HAS_INVERSE| is
   returned,  |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   \note{Computation of the resultant of two polynomials by the
   sub-resultant algorithm. From P. Gaudry.}
   }
   */
int ZENPolyResultant(Res, A, B, Rg)
     ZENElt Res;
     ZENPoly A, B;
     ZENRing Rg;{}
#endif
#define ZENPolyResultant(Res, A, B, Rg) _ZENPolyResultant(&(Res), A, B, Rg)

/* \TEX{
   \Subsection{Input/Output}
   }
   */
#if 0
/* \TEX{
   \title{Converting from string.}
   \input{A finite ring |Rg|, an unallocated polynomial
   |PX| and a string |s| representing a polynomial in base
   |base|$\in \{2,\ldots, 16\}$.} 
   \output{The number of character read in |s| or |ZENERR| if an error
   occurred.}
   \side{|PX| is allocated and filled with |s| if the output is
   not |ZENERR|}
   \note{Maple's format is used. The biggest monomial must be
   at the beginning of the string}
   }
*/
int ZENPolyReadFromString(PX, s, base, Rg)
     ZENPoly PX;
     char *s;
     int base;
     ZENRing Rg;{}
#endif
#define ZENPolyReadFromString(r,s,base,Rg) _ZENPolyReadFromString(&(r),s,'X',base,Rg) 
#if 0
/* \TEX{
   \title{Converting to string.}
   \input{An allocated polynomial |PX| of a finite ring |Rg| and a base
   |base|$\in \{2,\ldots, 16\}$.}
   \output{An allocated string representing |PX| in base |base| or
   |ZENNULL| if an error occurred.} 
   \note{Maple's format is used.}
   }
   */
char *ZENPolyPrintToString(PX, base, Rg)
     ZENPoly PX;
     int base;
     ZENRing Rg;  {}
#endif
#define ZENPolyPrintToString(r,base,Rg) _ZENPolyPrintToString(r,'X',base,Rg)
#if 0
/* \TEX{
   \title{Reading from file.}
   \input{A stream |file| and an unallocated polynomial
   |PX| of a finite ring |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|PX| is allocated and filled with the polynomial read in |file| if
   no error occurred.}
   \note{Maple's format is used. The biggest monomial must be
   at the beginning of the stream}
   }
   */
int ZENPolyReadFromFile(PX, file, base, Rg)
     ZENPoly PX;
     FILE *file;
     int base;
     ZENRing Rg;{}
#endif
#define ZENPolyReadFromFile(r,file,base,Rg) _ZENPolyReadFromFile(&(r),file,'X',base,Rg) 
#if 0
/* \TEX{
   \title{Printing to file.}
   \input{A stream |file|, an allocated polynomial |PX| of a finite ring
   |Rg| and a base |base|$\in \{2,\ldots, 16\}$.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{Printing a representation of |PX| in base |base| to |file|.}
   \note{Maple's format is used.}
   }
   */
int ZENPolyPrintToFile(file, PX, base, Rg)
     FILE *file;
     ZENPoly PX;
     int base;
     ZENRing Rg;  {}
#endif
#define ZENPolyPrintToFile(file,r,base,Rg) _ZENPolyPrintToFile(file,r,'X',base,Rg)

#if 0
/* \TEX{
   \title{Converting from a string to an internal representation.}
   \input{An unallocated polynomial |PX| of a finite ring |Rg| and a
   string |s| representing a polynomial
   to an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occurred.}
   \side{|PX| is allocated and filled with |s| if the output is different
   from |ZENERR|}         
   }
   */
int ZENPolyGetFromString(PX, s, Rg)
     char *s;
     ZENPoly PX;
     ZENRing Rg;{}
#endif
#define ZENPolyGetFromString(r,s,Rg) _ZENPolyGetFromString(&(r),s,Rg)
#if 0
/* \TEX{
   \title{Converting to string to an internal representation.}
   \input{An allocated polynomials |PX| of a finite ring |Rg|.}
   \output{An allocated string representing |PX| to an internal
     representation or |ZENNULL| if an error occurred.}
   }
   */
char *ZENPolyPutToString(PX, Rg)
     ZENPoly PX;
     ZENRing Rg;{}
#endif
#define ZENPolyPutToString(r,Rg) _ZENPolyPutToString(r,Rg)
#if 0
/* \TEX{
   \title{Getting from file to an internal representation.}
   \input{A stream |file|, and an unallocated polynomial |PX| of a
   finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{|PX| is filled with the polynomial read in |file|}    
   }
   */
int ZENPolyGetFromFile(PX, file, Rg)
     ZENPoly PX;
     FILE *file;
     ZENRing Rg;{}
#endif
#define ZENPolyGetFromFile(r,file,Rg) _ZENPolyGetFromFile(&(r),file,Rg) 
#if 0
/* \TEX{
   \title{Writing to file to an internal representation.}
   \input{A stream |file|, an allocated polynomial |PX| of a finite ring
   |Rg|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{Printing a representation of |PX| to |file|.}    
   }
   */
int ZENPolyPutToFile(file, PX, Rg)
     FILE *file;
     ZENPoly PX;
     ZENRing Rg;  {}
#endif
#define ZENPolyPutToFile(file,r,Rg) _ZENPolyPutToFile(file,r,Rg)
#if 0
/*
   \TEX{
   \title{Compute the derivative of a polynomial}
   \input{Two polynomial |RX| and |PX| defined over a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{$|RX|=|PX|'$.}
   } 
   */
int ZENPolyDerive(RX,PX,Rg)
     ZENPoly RX,PX;
     ZENRing Rg;{}
#endif
#define ZENPolyDerivePtr(Rg) ZENFct(Rg).PolyDerive
#define ZENPolyDerive(DX,PX,Rg) (*(ZENPolyDerivePtr(Rg)))(DX,PX,Rg)
#if 0
/*
   \TEX{
   \title{Roots of a polynomial of degree 2 in finite fields}
   \input{Two |ZENPoly| |P| and |roots| over a |ZENRing| |Rg|.}
   \output{\begin{itemize}
   \item[|ZENERR|] if an error occured,
   \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
   \item[|ZEN_HAS_INVERSE|] if the polynomial has one root.
   \end{itemize}}
   \side{The polynomial |roots| is set to a polynomial of degree -1, 0
   or 1, the coefficients of which are the roots of |P|.}
   \note{|roots| must be allocated for at least degree 1. We must have
   $|P|\neq |roots|$. Valid only in finite fields.}
   } 
   */
int ZENPolyRootsDegree2(roots, P, Rg)
     ZENPoly roots, P;
     ZENRing Rg;{}
#endif
#define ZENPolyRootsDegree2Ptr(Rg) ZENFct(Rg).PolyRootsDegree2
#define ZENPolyRootsDegree2(roots,P,Rg)  (*(ZENPolyRootsDegree2Ptr(Rg)))(roots,P,Rg)

#if 0
/*
   \TEX{
   \title{Roots of $X-X^p=\gamma$.}
   \input{A |ZENPoly| |roots| over a |ZENRing| |Rg| and an element |a|.}
   \output{\begin{itemize}
   \item[|ZENERR|] if an error occured,
   \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
   \item[|ZEN_HAS_INVERSE|] if the polynomial has one root.
   \end{itemize}}
   \side{The polynomial |roots| is set to a polynomial of degree -1
   or $p=$charateristic(|Rg|), the coefficients of which are the
   roots of |P|.}
 \note{|P| must be allocated for at least degree $p-1$. We must have
   $|P|\neq |roots|$. Valid only in finite fields.}
   } 
   */
int ZENPolyRootsCanonical(roots, gamma, Rg)
     ZENPoly roots;
     ZENElt gamma;
     ZENRing Rg;{}
#endif
#define ZENPolyRootsCanonical(roots,P,Rg) _ZENPolyRootsCanonical(roots,P,Rg)  

#define ZENPolyRootsDegree1(roots,P,Rg) _ZENPolyRootsDegree1(roots,P,Rg)  

#if 0
/* \TEX{
   \title{Getting a BigNum from a Polynomial.}
   \input{A pointer on an integer |p_pl|, and a polynomial |P| of
   a |ZENRing| |Rg|.}
   \output{A |BigNum| of size |*p_pl| if no error occured, |ZENNULL|
   otherwise.}
   } */
BigNum ZENPolyToZ(p_pl, P, Rg)
     BigNumLength *p_pl;
     ZENPoly P;
     ZENRing Rg;{}
#endif
#define ZENPolyToZ(p_pl, P, Rg) _ZENPolyToZ(p_pl, P, Rg)

#if 0
/* \TEX{
   \title{Getting a polynomial from a BigNum.}
   \input{A |BigNum| (|p|, |pl|) and a finite Ring |Rg|.}
   \output{|ZENNULL| if an error occured, otherwise a polynomial which once, evaluated in |ZENRingQ(Rg)|, returns (|n|, |nl|).}
   }
   */
ZENPoly ZENZToPoly(p, pl, Rg)
     BigNum p;
     BigNumLength pl;
     ZENRing Rg;{}
#endif
#define ZENZToPoly(p, pl, Rg) _ZENZToPoly(p, pl, Rg)
     
/* Undocumented macros */
#define ZENPolyAllocStruct(Rg) (ZENPoly)ZENMalloc(sizeof(struct zenpoly))

#define ZENPolyDigit(p,Rg) (p)->coeff.b
#define ZENPolyGetDigit(p,i,Rg) ZBNDigit(ZENPolyDigit(p,Rg), i)

#define ZENPolyChar(p,Rg) (p)->coeff.h
#define ZENPolyGetChar(p,i,Rg) ZENPolyChar(p,Rg)[i]

#define ZENPolyShort(p,Rg) (p)->coeff.s
#define ZENPolyGetShort(p,i,Rg) ZENPolyShort(p,Rg)[i]

#define ZENPolyCof(p,Rg) (p)->coeff.c
#define ZENPolyGetCof(p,i,Rg) ZENPolyCof(p,Rg)[i]

/*   \TEX{
   \Section[Matrices]{Procedures to handle matrices over finite rings.}
   \Subsection{Parameters of a matrix}   
   }
   */

/* These codes are used in matrix operations involving matrices types */
#define _CC_ 0                  
#define _CR_ 1
#define _RC_ 2
#define _RR_ 3

#define _CCC_ 0
#define _CCR_ 1
#define _CRC_ 2
#define _CRR_ 3
#define _RCC_ 4
#define _RCR_ 5
#define _RRC_ 6
#define _RRR_ 7

#define ZENMatFct(M,R) ZENFct(M)
#define ZENMatNbPtr(M,R) ((M)->nbptr)
#define ZENMatSize(M,R) ((M)->size)
#define ZENMatNbBn(M,R) ((M)->nbbn)
#define ZENMatNbBnd(M,R) ((M)->nbbnd)
#define ZENMatNbDigit(M,R) ((M)->nbdigit)
#define ZENMatNbBloc(M,R) ((M)->nbbloc)
#define ZENMatType(M,R) ((M)->type_struct)
#define ZENMatFirst(M,R) ((M)->First)

#if 0
/*
   \TEX{
   \title{Testing type of a matrix}
   \input{A |ZENMat| and a |ZENRing|}
   \output{1 if the matrix is of |Row| type, 0 if it is of |Col| type}
   \note{Permutation matrices can be used.}
   }
   */
int ZENMatIsRowType(M,R)
     ZENMat M;
     ZENRing R; {}
#endif
#define ZENMatIsRowType(M,R) (((int)(ZENMatType(M,R)))&1)

#define ZENMatMat(M,R) (((M)->elts).zeg)
#define ZENMatBn(M,l,R) (ZENMatMat(M,R)[l])
#define ZENMatBloc(M,l,c,R) (ZENMatBn(M,l,R)[c])
#define ZENMatCoeff(M,l,c,R) ZENMatBloc(M,ZENMatIsRowType(M,R)?(l):(c),ZENMatIsRowType(M,R)?(c):(l),R)

#define ZENPermutationMat(P,R) (((P)->elts).perm)
#define ZENPermutationBn(P,l,R) (*(ZENPermutationMat(P,R)+(l)))

#define Ze2MatMat(M,R) (((M)->elts).ze2)
#define Ze2MatBn(M,l,R) (*(Ze2MatMat(M,R)+(l)))
#define Ze2MatBloc(M,l,c,R) (*(Ze2MatBn(M,l,R)+(c)))

#define ZepMatMat(M,R) (((M)->elts).zep)
#define ZepMatBn(M,l,R) (ZepMatMat(M,R)[l])
#define ZepMatBloc(M,l,c,R) (ZepMatBn(M,l,R)[c])

#define ZepsMatMat(M,R) (((M)->elts).zeps)
#define ZepsMatBn(M,l,R) (ZepsMatMat(M,R)[l])
#define ZepsMatBloc(M,l,c,R) (ZepsMatBn(M,l,R)[c])

#define ZelogMatMat(M,R) (((M)->elts).zelog)
#define ZelogMatBn(M,l,R) (ZelogMatMat(M,R)[l])
#define ZelogMatBloc(M,l,c,R) (ZelogMatBn(M,l,R)[c])
#define ZelogMatCoeff(M,l,c,R) ZelogMatBloc(M,ZENMatIsRowType(M,R)?(l):(c),ZENMatIsRowType(M,R)?(c):(l),R)

#define ZetabMatMat(M,R) (((M)->elts).zetab)
#define ZetabMatBn(M,l,R) (ZetabMatMat(M,R)[l])
#define ZetabMatBloc(M,l,c,R) (ZetabMatBn(M,l,R)[c])
#define ZetabMatCoeff(M,l,c,R) ZetabMatBloc(M,ZENMatIsRowType(M,R)?(l):(c),ZENMatIsRowType(M,R)?(c):(l),R)
     
     
#if 0
/*
   \TEX{
   \title{Testing type of a matrix}
   \input{A |ZENMat| and a |ZENRing|}
   \output{1 if the matrix is a permutation, 0 if it is a plain matrix}
   }
   */
int ZENMatIsPermutation(M,R)
     ZENMat M;
     ZENRing R; {}
#endif
#define ZENMatIsPermutation(M,R) ((((int)(ZENMatType(M,R)))>>1)&1)
#if 0
/*
   \TEX{
   \title{Testing type of a matrix}
   \input{Two |ZENMat|s and a |ZENRing|}
   \output{1 if the matrix are of same type, 0 otherwise.}
   \note{This function checks also if the matrix is permutation or plain.}
   }
   */
int ZENMatAreSameType(A,B,R)
     ZENMat A,B;
     ZENRing R; {}
#endif
#define ZENMatAreSameType(A,B,R) (ZENMatIsRowType(A,R)==ZENMatIsRowType(B,R))

#if 0
/*
   \TEX{
   \title{Number of rows of a matrix}
   \input{A |ZENMat| and a |ZENRing|}
   \output{The number of rows of the matrix}
   }
   */
Dim ZENMatNbRow(M,R)
     ZENMat M;
     ZENRing R; {}
#endif
#define ZENMatNbRow(M,R) (ZENMatIsRowType(M,R)?ZENMatNbBn(M,R):ZENMatNbDigit(M,R))
#if 0
/*
   \TEX{
   \title{Number of columns of a matrix}
   \input{A |ZENMat| and a |ZENRing|}
   \output{The number of columns of the matrix}
   }
   */
Dim ZENMatNbCol(M,R)
     ZENMat M;
     ZENRing R; {}
#endif
#define ZENMatNbCol(M,R) (ZENMatIsRowType(M,R)?ZENMatNbDigit(M,R):ZENMatNbBn(M,R))

#if 0
/*
   \TEX{
   \Subsection{Allocation}
   
   \title{Allocation of a matrix}
   \input{An unallocated matrix |M|, a number of rows |r|, a number of columns |c| greater than 0 and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{Matrix |M| is allocatedd together with |r| rows and |c| columns.}
   \note{The type of the matrix is chosen to minimize the memory needed.}
   }
   */
int ZENMatAlloc(M,r,c,R)
     ZENMat M;
     Dim r,c;
     ZENRing R; {}
#endif
#define ZENMatCreate(l,c,R) _ZENMatCreate(l,c,R)
#define ZENMatAlloc(M,l,c,R) (((M)=ZENMatCreate(l,c,R))==NULL ? ZERR : 0)
#define ZENMatCreatePtr(R) ZENFct(R).MatCreate
#define _R_ZENMatCreate(M,R) (*(ZENMatCreatePtr(R)))(M,R)
#define ZENMatCreateParamsPtr(R) ZENFct(R).MatCreatePms
#define ZENMatCreateParams(ts,l,c,p_pms,R)  (*(ZENMatCreateParamsPtr(R)))(ts,l,c,p_pms,R)

#if 0
/*
   \TEX{
   \title{Allocation of a matrix}
   \input{An unallocated matrix |M|, a number of rows |r|, a number of columns |c| greater than 0 and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The matrix is allocated |ZENMat| with type |ZENMatTypeRow|  with |r| rows and |c| columns.}
   }
   */
int ZENMatRowAlloc(M,r,c,R)
     ZENMat M;
     Dim r,c;
     ZENRing R; {}
#endif
#define ZENMatRowCreate(l,c,R) _ZENMatRowCreate(l,c,R)
#define ZENMatRowAlloc(M,l,c,R) (((M)=ZENMatRowCreate(l,c,R))==NULL ? ZERR : 0)
#if 0
/*
   \TEX{
   \title{Allocation of a matrix}
   \input{An unallocated matrix |M|, a number of rows |r|, a number of columns |c| greater than 0 and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The matrix is allocated |ZENMat| with type |ZENMatTypeCol|  with |r| rows and |c| columns.}
   }
   */
int ZENMatColAlloc(M,r,c,R)
     ZENMat M;
     Dim r,c;
     ZENRing R; {}
#endif
#define ZENMatColCreate(l,c,R) _ZENMatColCreate(l,c,R)
#define ZENMatColAlloc(M,l,c,R) (((M)=ZENMatColCreate(l,c,R))==NULL ? ZERR : 0)

#define ZENPermutationCreate(n,t,R) _ZENPermutationCreate(n,t,R)
#define ZENPermutationFree(P,R) _ZENPermutationFree(P,R) /* this function
							    isn't documented
							    because it should
							    not be used
							    directly */
#if 0
/*
   \TEX{
   \title{Allocation of a permutation matrix}
   \input{An unallocate |ZENMat|, a size |r| greater than 0 and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{Matrix |PI| is allocated with type |ZENPermutationTypeRow|  with |r| rows and |r| columns.}
   \note{The matrix is set to identity}
   } 
   */
int ZENPermutationRowAlloc(PI,n,R)
     ZENMat PI;
     Dim n;
     ZENRing R; {}
#endif
#define ZENPermutationRowCreate(n,R) ZENPermutationCreate(n,ZENMatTypeRowPermutation,R)
#define ZENPermutationRowAlloc(M,c,R) (((M)=ZENPermutationRowCreate(c,R))==NULL ? ZERR : 0)
#if 0
/*
   \TEX{
   \title{Allocation of a permutation matrix}
   \input{An unallocate |ZENMat|, a size |r| greater than 0 and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{Matrix |PI| is allocated with type |ZENPermutationTypeCol|  with |r| rows and |r| columns.}
   \note{The matrix is set to identity}
   } 
   */
int ZENPermutationColAlloc(PI,n,R)
     ZENMat PI;
     Dim n;
     ZENRing R; {}
#endif
#define ZENPermutationColCreate(n,R) ZENPermutationCreate(n,ZENMatTypeColPermutation,R)
#define ZENPermutationColAlloc(M,c,R) (((M)=ZENPermutationColCreate(c,R))==NULL ? ZERR : 0)
#if 0
/*
   \TEX{
   \title{Freeing a matrix}
   \input{A |ZENMat| and a |ZENRing|}
   \side{The matrix is freed}
   \note{Permutation matrices can be used}
   }
   */
void ZENMatFree(M,R)
     ZENMat M;
     ZENRing R; {}
#endif
#define ZENMatFreePtr(M) ZENFct(M).MatFree
#define ZENMatFree(M,R) (*(ZENMatFreePtr(M)))(M,R)

#if 0
/*
   \TEX{
   \title{Copying a matrix}
   \input{A |ZENMat| and a |ZENRing|}
   \output{A plain |ZENMat| copied from |M| of same type, or |ZENNULL| if an error occured.}
   \note{Permutation matrix can be input, but the output is a plain matrix.}
   }
   */
ZENMat ZENMatCopy(M,R)
     ZENMat M;
     ZENRing R; {}
#endif
#define ZENMatCopy(M,R) _ZENMatCopy(M,R)

/* Assigning */
#if 0
/*
   \TEX{
   \Subsection{Assigning}
   
   \title{Assigning a matrix to another matrix}
   \input{Two |ZENMat| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|A|=|B|.}
   \note{|B| can be a permutation matrix. In this case, |A| can be either a permutation matrix or a plain matrix. Otherwise, it must be a plain matrix.}
   }
   */
int ZENMatAssign(A,B,R)
     ZENMat A,B;
     ZENRing R; {}
#endif
#define ZENMatAssignPtr(R) ZENFct(R).MatAssign
#define _R_ZENMatAssign(A,B,R) (*(ZENMatAssignPtr(R)))(A,B,R)
#define ZENMatAssign(A,B,R)    (*(ZENMatAssignPtr(B)))(A,B,R)

#define ZENPermutation2MatrixPtr(R) ZENFct(R).Perm2Mat
#define ZENPermutation2Matrix(A,B,R) (*(ZENPermutation2MatrixPtr(R)))(A,B,R)

#if 0
/*
   \TEX{
   \title{Assigning to zero}
   \input{A |ZENMat| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|M|=0.}
   \note{|M| must be a plain matrix.}
   }
   */
int ZENMatSetToZero(M,R)
     ZENMat M;
     ZENRing R; {}
#endif
#define ZENMatSetToZeroPtr(R) ZENFct(R).MatSetToZero
#define ZENMatSetToZero(M,R)  (*(ZENMatSetToZeroPtr(R)))(M,R) 

#if 0
/*
   \TEX{
   \title{Assigning to identity}
   \input{A |ZENMat| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|M|_{i,j}=\delta_{i,j}$, with $\delta$ the Kronecker's symbol.}
   \note{Permutation matrix can be used.}
   }
   */
int ZENMatSetToOne(M,R)
     ZENMat M;
     ZENRing R; {}
#endif
#define ZENMatSetToOnePtr(R)   ZENFct(R).MatSetToOne
#define _R_ZENMatSetToOne(M,R) (*(ZENMatSetToOnePtr(R)))(M,R)
#define ZENMatSetToOne(M,R)    (*(ZENMatSetToOnePtr(M)))(M,R)

#if 0
/*
   \TEX{
   \title{Assigning randomly}
   \input{A |ZENMat| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The matrix is randomly set.}
   \note{Permutation matrix can be used.}
   }
   */
int ZENMatSetRandom(M,R)
     ZENMat M;
     ZENRing R; {}
#endif

#define ZENMatSetRandomPtr(R) ZENFct(R).MatSetRandom
#define _R_ZENMatSetRandom(M,R) (*(ZENMatSetRandomPtr(R)))(M,R) 
#define ZENMatSetRandom(M,R)    (*(ZENMatSetRandomPtr(M)))(M,R)
 

#define ZENMatSetBlocPtr(R) ZENFct(R).MatSetBloc
#define ZENMatSetBloc(M,l,c,Z,R) (*(ZENMatSetBlocPtr(R)))(M,l,c,Z,R)
#define ZENMatGetBlocNewPtr(R) ZENFct(R).MatGetBloc
#define ZENMatGetBloc(Z,M,l,c,R) (*(ZENMatGetBlocNewPtr(R)))(&(Z),M,l,c,R)
#define ZENMatSetDigitPtr(R) ZENFct(R).MatSetBloc
#define ZENMatSetDigit(M,l,c,Z,R) (*(ZENMatSetDigitPtr(R)))(M,c,l,Z,R)
#define ZENMatGetDigitNewPtr(R) ZENFct(R).MatGetBloc
#define ZENMatGetDigit(Z,M,l,c,R) (*(ZENMatGetDigitNewPtr(R)))(&(Z),M,c,l,R)
#if 0
/*
   \TEX{
   \title{Assigning a coefficient}
   \input{A |ZENMat|, two indexes, a |ZENElt| and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|M|_{|r|,|c|}=Z$}
   \note{Permutation matrix can be used.}
   }
   */
int ZENMatSetCoeff(M,r,c,Z,R)
     ZENMat M;
     Dim r,c;
     ZENElt Z;
     ZENRing R; {}
#endif
#define ZENMatSetCoeffPtr(M) ZENFct(M).MatSetCoeff
#define ZENMatSetCoeff(M,l,c,Z,R)  (*(ZENMatSetCoeffPtr(M)))(M,l,c,Z,R)

#if 0
/*
   \TEX{
   \title{Getting a coefficient}
   \input{An allocated |ZENElt|, a |ZENMat|, two indexes and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$Z=|M|_{|r|,|c|}$}
   \note{Permutation matrix can be used.}
   }
   */
int ZENMatGetCoeff(Z,M,r,c,R)
     ZENElt Z;
     ZENMat M;
     Dim r,c;
     ZENRing R; {}
#endif
#define ZENMatGetCoeffNewPtr(M) ZENFct(M).MatGetCoeff
#define ZENMatGetCoeff(Z,M,l,c,R)  (*(ZENMatGetCoeffNewPtr(M)))(&(Z),M,l,c,R)

#if 0
/*
   \TEX{
   \title{Assigning a submatrix}
   \input{Two |ZENMat|, two indexes, a |ZENElt| and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The matrix |S| is assigned at position $(|r|,|c|)$ in matrix |M|.}
   \note{Permutation matrix cannot be used.}
   }
   */
int ZENMatSetSubMat(M,r,c,S,R)
     ZENMat M,S;
     Dim r,c;
     ZENRing R; {}
#endif
#define ZENMatSetSubMatPtr(R) ZENFct(R).MatSetSubMat
#define ZENMatSetSubMat(M,l,c,Z,R) (*(ZENMatSetSubMatPtr(R)))(M,l,c,Z,R)

#if 0
/*
   \TEX{
   \title{Getting a submatrix}
   \input{Two allocated |ZENMat|, two indexes and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The submatrix of |M| at position $(|r|,|c|)$ of size those of |S| is extracted and assigned to |S|.}
   \note{Permutation matrix cannot be used.}
   }
   */
int ZENMatGetSubMat(S,M,r,c,R)
     ZENMat M,S;
     Dim r,c;
     ZENRing R; {}
#endif
#define ZENMatGetSubMatPtr(R) ZENFct(R).MatGetSubMat
#define ZENMatGetSubMat(M,l,c,Z,R) (*(ZENMatGetSubMatPtr(R)))(M,l,c,Z,R)

#define ZENMatGetBlocPtrPtr(R) ZENFct(R).MatGetBlocPtr
#define ZENMatGetBlocPtr(M,l,c,R) (*(ZENMatGetBlocPtrPtr(R)))(M,l,c,R)
#define ZENMatGetDigitPtrPtr(R) ZENFct(R).MatGetBlocPtr
#define ZENMatGetDigitPtr(M,l,c,R) (*(ZENMatGetDigitPtrPtr(R)))(M,c,l,R)
#if 0
/*
   \TEX{
   \title{Getting a coefficient pointer}
   \input{A |ZENMat|, two indexes and a |ZENRing|.}
   \output{A pointer on $|M|_{|r|,|c|}$.}
   \note{Do NOT use the result of this function to change the coefficient of the matrix.
   This may not work.
   Use |ZENMatSetCoeff| instead.
   In the same way, do NOT assign the result of this function directly (like |z = ZENMatGetCoeffPtr();|) because further assignments such as |ZENEltAssign(z, x, Rg)| may modify the initial coefficient of the matrix.
   Use |ZENMatGetCoeff| instead.
   This function is intended to do fast read-only access to the coefficients of a matrix.}
   }
   */
int ZENMatGetCoeffPtr(M,r,c,R)
     ZENMat M;
     Dim r,c;
     ZENRing R; {}
#endif
#define ZENMatGetCoeffPtrPtr(M) ZENFct(M).MatGetCoeffPtr
#define ZENMatGetCoeffPtr(M,l,c,R) (*(ZENMatGetCoeffPtrPtr(M)))(M,l,c,R)


#define ZENPermutationAssign(A,B,R)  _ZENPermutationAssign(A,B,R)
#define ZENPermutationSetToOne(A,R)  _ZENPermutationSetToOne(A,R)
#define ZENPermutationIsOne(A,R)  _ZENPermutationIsOne(A,R)
#define ZENPermutationSetRandom(A,R)  _ZENPermutationSetRandom(A,R)
#define ZENPermutationGetBloc(E,P,l,c,R)  _ZENPermutationGetBloc(E,P,l,c,R)
#define ZENPermutationSetBloc(P,l,c,E,R)  _ZENPermutationSetBloc(P,l,c,E,R)
#define ZENPermutationGetDigit(E,P,l,c,R)  _ZENPermutationGetDigit(E,P,l,c,R)
#define ZENPermutationSetDigit(P,l,c,E,R)  _ZENPermutationSetDigit(P,l,c,E,R)
#define ZENPermutationPermuteBn(P,l1,l2,R)  _ZENPermutationPermuteBn(P,l1,l2,R)
#define ZENMatFastPermute(M,l1,l2,R) _ZENMatFastPermuteBn(M,l1,l2,R)

/*
   \TEX{
   \Subsection{Permuting rows or columns}
   } 
   */
#define ZENMatPermuteBnPtr(R) ZENFct(R).MatPermuteBn
#define ZENMatPermuteBn(M,l1,l2,R)    (*(ZENMatPermuteBnPtr(M)))(M,l1,l2,R)
#define _R_ZENMatPermuteBn(M,l1,l2,R) (*(ZENMatPermuteBnPtr(R)))(M,l1,l2,R)

#define ZENMatPermuteDigitPtr(R) ZENFct(R).MatPermuteDigit
#define ZENMatPermuteDigit(M,l1,l2,R) (*(ZENMatPermuteDigitPtr(R)))(M,l1,l2,R)
#if 0
/*
   \TEX{
   \title{Permute rows}
   \input{A |ZENMat|, two indexes and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Rows |r1| and |r2| are permuted.}
   \note{|ZENMatTypeRowPermutation| can be used.}
   }
   */
int ZENMatPermuteRow(M,r1,r2,R)
     ZENMat M;
     Dim r1,r2;     
     ZENRing R; {}
#endif
#define ZENMatPermuteRowPtr(R) ZENFct(R).MatPermuteRow
#define ZENMatPermuteRow(M,l1,l2,R)    (*(ZENMatPermuteRowPtr(M)))(M,l1,l2,R)
#if 0
/*
   \TEX{
   \title{Permute columns}
   \input{A |ZENMat|, two indexes and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Rows |c1| and |c2| are permuted.}
   \note{|ZENMatTypeColPermutation| can be used.}
   }
   */
int ZENMatPermuteCol(M,c1,c2,R)
     ZENMat M;
     Dim c1,c2;
     ZENRing R; {}
#endif
#define ZENMatPermuteColPtr(R) ZENFct(R).MatPermuteCol
#define ZENMatPermuteCol(M,l1,l2,R)    (*(ZENMatPermuteColPtr(M)))(M,l1,l2,R)

/* Test */
#if 0
/*
   \TEX{
   \Subsection{Tests}
   \title{Equality}
   \input{Two |ZENMat| and a |ZENRing|}
   \output{|A == B|.}
   \note{Permutation matrices can be used.}
   }
   */
int ZENMatAreEqual(A,B,R)
     ZENMat A,B;
     ZENRing R; {}
#endif
#define ZENMatAreEqualPtr(R) ZENFct(R).MatAreEqual
#define ZENMatAreEqual(A,B,R)  (*(ZENMatAreEqualPtr(R)))(A,B,R) 
#if 0
/*
   \TEX{
   \title{Equality  to zero}
   \input{A |ZENMat| and a |ZENRing|}
   \output{|M == 0|}
   }
   */
int ZENMatIsZero(M,R)
     ZENMat M;
     ZENRing R; {}
#endif
#define ZENMatIsZeroPtr(R) ZENFct(R).MatIsZero
#define ZENMatIsZero(M,R)  (*(ZENMatIsZeroPtr(R)))(M,R) 
#if 0
/*
   \TEX{
   \title{Equality to identity}
   \input{A |ZENMat| and a |ZENRing|}
   \output{|M == |Identity.}
   \note{The matrix is checked to be square.}
   }
   */
int ZENMatIsOne(M,R)
     ZENMat M;
     ZENRing R; {}
#endif
#define ZENMatIsOnePtr(R) ZENFct(R).MatIsOne
#define _R_ZENMatIsOne(M,R) (*(ZENMatIsOnePtr(R)))(M,R)
#define ZENMatIsOne(M,R)  (*(ZENMatIsOnePtr(M)))(M,R) 

     /* Arithmetic */
#if 0
/*
   \TEX{
   \Subsection{Arithmetic}
   \title{Addition}
   \input{Two |ZENMat| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|A += B|}
   \note{|B| can be a permutation.}
   }
   */
int ZENMatAdd(A,B,R)
     ZENMat A,B;
     ZENRing R; {}
#endif
#define ZENMatAddPtr(R) ZENFct(R).MatAdd
#define ZENMatAdd(A,B,R)  (*(ZENMatAddPtr(R)))(A,B,R) 
#if 0
/*
   \TEX{
   \title{Addition of a scalar}
   \input{A |ZENMat|, a |ZENElt| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{All the elements of |A| are incremented by |b|}
   \note{|A| cannot be a permutation.}
   }
   */
int ZENMatAddScalar(A,b,R)
     ZENMat A;
     ZENElt b;
     ZENRing R; {}
#endif
#define ZENMatAddScalarPtr(R) ZENFct(R).MatAddScalar
#define ZENMatAddScalar(A,B,R)  (*(ZENMatAddScalarPtr(R)))(A,B,R) 
#if 0
/*
   \TEX{
   \title{Negation}
   \input{Two |ZENMat| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|A| = -|B|$.}
   \note{|B| can be a permutation.}
   }
   */
int ZENMatNegate(A,B,R)
     ZENMat A,B;
     ZENRing R; {}
#endif
#define ZENMatNegatePtr(R) ZENFct(R).MatNegate
#define ZENMatNegate(A,B,R)  (*(ZENMatNegatePtr(R)))(A,B,R) 
#if 0
/*
   \TEX{
   \title{Subtraction}
   \input{Two |ZENMat| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|A -= B|}
   \note{|B| can be a permutation.}
   }
   */
int ZENMatSubtract(A,B,R)
     ZENMat A,B;
     ZENRing R; {}
#endif
#define ZENMatSubtractPtr(R) ZENFct(R).MatSubtract
#define ZENMatSubtract(A,B,R)  (*(ZENMatSubtractPtr(R)))(A,B,R) 

#if 0
/*
   \TEX{
   \title{Multiplication}
   \input{Three |ZENMat| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{$|X|=|A||B|$.}
   \note{|A| can be a |ZENMatTypeRowPermutation| and |B| can be a |ZENMatTypeColPermutation|. One can have |A|=|B| but |X| must be distinct from |A| and |B|.}
   }
   */
int ZENMatMultiply(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R; {}
#endif
#define ZENMatMultiply(X,A,B,R) _ZENMatMultiply(X,A,B,R) 
#define ZENMatMultiplyPtr(R) ZENFct(R).MatMultiply
#define _R_ZENMatMultiply(X,A,B,R)  (*(ZENMatMultiplyPtr(R)))(X,A,B,R) 

#if 0
/*
   \TEX{
   \title{Multiplication of a scalar}
   \input{A |ZENMat|, a |ZENElt| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{All the elements of |A| are multiplied by |b|}
   \note{|A| cannot be a permutation.}
   }
   */
int ZENMatMultiplyScalar(A,b,R)
     ZENMat A;
     ZENElt b;
     ZENRing R; {}
#endif
#define ZENMatMultiplyScalarPtr(R) ZENFct(R).MatMultiplyScalar
#define ZENMatMultiplyScalar(A,B,R)  (*(ZENMatMultiplyScalarPtr(R)))(A,B,R) 
#if 0
/*
   \TEX{
   \title{Multiplication}
   \input{Three |ZENMat| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{$|X|=|A||B|$.}
   \note{Permutation cannot be used. The classical algorithm is used.}
   }
   */
int ZENMatMultiplyPlain(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R; {}
#endif
#define ZENMatMultiplyPlainPtr(R) ZENFct(R).MatMultiplyPlain
#define ZENMatMultiplyPlain(X,A,B,R)  (*(ZENMatMultiplyPlainPtr(R)))(X,A,B,R) 
#if 0
/*
   \TEX{
   \title{Multiplication}
   \input{Three |ZENMat| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{$|X|=|A||B|$.}
   \note{Permutation cannot be used. Winograd's algorithm is used.}
   }
   */
int ZENMatWinograd(X,A,B,R)
     ZENRing R; {}
#endif
#define ZENMatWinogradPtr(R) ZENFct(R).MatWinograd
#define ZENMatWinograd(X,A,B,R)  (*(ZENMatWinogradPtr(R)))(X,A,B,R) 

#if 0
/*
   \TEX{
   \title{Gaussian elimination}
   \input{Three |ZENMat| a pointer on a dimension and a |ZENRing|}
   \output{|ZENERR| if an error occurred,  |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned,
   |ZENMAT_HAS_MAXIMAL_RANK| if the matrix is of maximal rank,
   |ZENMAT_HAS_KERNEL| otherwise.}
   \side{|D| is diagonalised. |S| and |P| are modified accordingly (see below). |*p_rk| is set to the rank of the matrix $D$.}
   \note{   \begin{itemize}
   \item
   If |D| is a $r\times c$ |ZENMat| of type |TypeRow|, then |S| must be
   a $r\times r$ |ZENMat| of type |TypeRow| and |P| a  $c\times c$ |ZENMat|
   of type |TypeCol|, or a |TypeColPermutation| |ZENMat| of size $c$.
   Then, $D$ is diagonalized on its rows and $S^{-1} D P^{-1}$ remains
   invariant.
   \item
   If |D| is a $r\times c$ |ZENMat| of type |TypeCol|, then |S| must be
   a $c\times c$ |ZENMat| of type |TypeCol| and |P| a  $r\times r$ |ZENMat|
   of type |TypeRow|, or a |TypeRowPermutation| |ZENMat| of size $r$.
   Then, $D$ is diagonalized on its columns and $P^{-1} D S^{-1} $ remains
   invariant.
   \end{itemize}
   A tricky feature is also possible, but should be used with care: it is possible to change the |ZENMatNbDigit(D,R)| field of |D| before calling a gaussian elimination procedure. The gaussian elimination will therefore be performed only on the specified number of digits, but the corresponding modifications are done on all the matrix, whose real size is known through |ZENMatNbBloc(D,R)|. This allows partial trigonaliation if needed. Don't forget to reset the correct value after the call.
}
   }
   */
int ZENMatGaussPlain(D,S,P,p_rk,R)
     ZENMat D,S,P;
     Dim *p_rk;
     ZENRing R; {}
#endif
#define ZENMatGaussPtr(R) ZENFct(R).MatGauss
#define ZENMatGauss(D,S,P,p_rk,p_det,Perm,R) (*(ZENMatGaussPtr(R)))(D,S,P,p_rk,p_det,Perm,R) 
#define ZENMatGaussPlain(D,S,P,p_rk,R) _ZENMatGaussPlain(D,S,P,p_rk,R)

#if 0
/*
   \TEX{
   \title{Rank of a matrix}
   \input{A |ZENMat| a pointer on a dimension and a |ZENRing|}
   \output{|ZENERR| if an error occurred,  |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned,
   |ZENMAT_HAS_MAXIMAL_RANK| if the matrix is of maximal rank,
   |ZENMAT_HAS_KERNEL| otherwise.}
   \side{|*p_rk| is set to the rank of the matrix $D$.}
   }
   */
int ZENMatRank(D,p_rk,R)
     ZENMat D;
     Dim *p_rk;
     ZENRing R; {}
#endif
#define ZENMatGaussRank(D,p_rk,R) ZENMatGauss(D,NULL,NULL,p_rk,NULL,_ZENMatFastPermuteBn,R)
#define ZENMatRank(D,p_rk,R) _ZENMatRank(D,p_rk,R)

#if 0
/*
   \TEX{
   \title{Determinant of a matrix}
   \input{A |ZENMat| an allocated |ZENElt| and a |ZENRing|}
   \output{|ZENERR| if an error occurred,  |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned,
   |ZENMAT_HAS_INVERSE| otherwise.}
   \side{|*p_det| is set to the determinant of the matrix $D$.}
   }
   */
int ZENMatDet(D,det,R)
     ZENMat D;
     ZENElt det;
     ZENRing R; {}
#endif
#define ZENMatGaussDet(D,p_rk,R) ZENMatGauss(D,NULL,NULL,NULL,p_det,_ZENMatFastPermuteBn,R)
#define ZENMatDet(D,det,R) _ZENMatDet(D,&(det),R)

#define ZENMatGaussLazy(D,P,p_rk,R) _ZENMatGaussLazy(D,P,p_rk,R)
#define ZENMatGaussInverse(D,S,R) _ZENMatGaussInverse(D,S,R)
#define ZENMatGaussTrigInvertible(D,R) ZENMatGauss(D,NULL,NULL,NULL,NULL,_ZENMatFastPermuteBn,R)
#if 0
/*
   \TEX{
   \title{Inverse}
   \input{Two |ZENMat| and a |ZENRing|}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise.}
   \side{$|I| = |M|^{-1}$. |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   \note{|I| and |M| must be different but can be permutations.}
   }
   */
int ZENMatInverse(I,M,R)
     ZENMat I,M;
     ZENRing R; {}
#endif
#define ZENMatInversePtr(R) ZENFct(R).MatInverse
#define _R_ZENMatInverse(I,M,R)  (*(ZENMatInversePtr(R)))(I,M,R) 
#define ZENMatInverse(I,M,R) __ZENMatInverse(I,M,R)
#if 0
/*
   \TEX{
   \title{Kernel}
   \input{An unallocated |ZENMat|, an allocated |ZENMat| and a |ZENRing|}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise.}
   \side{|K| is allocated and set to a basis of the kernel of |M|.
   One has $|M||K|^t=0$.
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   \note{|K| can be returned set to zero.}
   }
   */
int ZENMatKernel(K,M,R)
     ZENRing R; {}
#endif
#define ZENMatKernelPtr(R) ZENFct(R).MatKernel
#define ZENMatKernel(K,M,R)  (*(ZENMatKernelPtr(R)))(&(K),M,R) 

#define ZENMatMultiplyPermutations(X,A,B,R) _ZENMatMultiplyPermutations(X,A,B,R)
#define ZENMatMultiplyRightPermutationPtr(R) ZENFct(R).MatMultiplyRightPermutation
#define ZENMatMultiplyRightPermutation(X,A,B,R) (*(ZENMatMultiplyRightPermutationPtr(R)))(X,A,B,R)
#define ZENMatMultiplyLeftPermutationPtr(R) ZENFct(R).MatMultiplyLeftPermutation
#define ZENMatMultiplyLeftPermutation(X,A,B,R) (*(ZENMatMultiplyLeftPermutationPtr(R)))(X,A,B,R)
#if 0
/*
   \TEX{
   \title{Transpose}
   \input{A |ZENMat| and a |ZENRing|.}
   \side{The matrix is transposed}
   \note{Permutation matrix can be used}
   }
   */
void ZENMatTranspose(M,R)
     ZENMat M;
     ZENRing R; {}
#endif
#define ZENMatTranspose(M,R) _ZENMatTranspose(M,R)
#if 0
/*
   \TEX{
   \Subsection{Input/output}
   
   \title{Printing to string}
   \input{A |ZENMat|, an integer between 2 and 16 and a |ZENRing|.}
   \output{|ZENNULL| if an error occured, a string representing |r| in base |base| otherwise}
   \note{Permutation matrices will be written as plain matrices.}
   }
   */
char *ZENMatPrintToString(r,base,Rg)
     ZENMat r;
     int base;
     ZENRing Rg; {}
#endif
#define ZENMatPrintToString(r,base,Rg) _ZENMatPrintToString(r,base,Rg)
#if 0
/*
   \TEX{
   \title{Reading from string}
   \input{An unallocated |ZENMat|, a string, an integer between 2 and 16 and a |ZENRing|.}
   \output{|ZENERR| if an error occured, the number of read characters  otherwise}
   \side{The matrix is allocated and read from string |s| in base |base|.}
   }
   */
int ZENMatReadFromString(r,s,base,Rg)
     ZENMat r;
     char *s;
     int base;
     ZENRing Rg; {}
#endif
#define ZENMatReadFromString(r,s,base,Rg) _ZENMatReadFromString(&(r),s,base,Rg) 
#if 0
/*
   \TEX{
   \title{Printing to file}
   \input{A file, a |ZENMat|, an integer between 2 and 16 and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \note{Permutation matrices will be written as plain matrices.}
   }
   */
int ZENMatPrintToFile(file,r,base,Rg)
     FILE *file;
     ZENMat r;
     int base;     
     ZENRing Rg; {}
#endif
#define ZENMatPrintToFile(file,r,base,Rg) _ZENMatPrintToFile(file,r,base,Rg)
#if 0
/*
   \TEX{
   \title{Reading from file}
   \input{An unallocated |ZENMat|, a file, an integer between 2 and 16 and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The matrix is allocated and read from file |file| in base |base|.}
   }
   */
int ZENMatReadFromFile(r,file,base,Rg)
     ZENMat r;
     FILE *file;
     int base;          
     ZENRing Rg; {}
#endif
#define ZENMatReadFromFile(r,file,base,Rg) _ZENMatReadFromFile(&(r),file,base,Rg) 

#if 0
/*
   \TEX{
   \title{Printing to string in internal format}
   \input{A |ZENMat|, and a |ZENRing|.}
   \output{|ZENNULL| if an error occured, a string representing |r| otherwise}
   \note{Permutation matrices cannot be used.}
   }
   */
int ZENMatPutToString(r,Rg)
     ZENMat r;
     ZENRing Rg; {}
#endif
#define ZENMatPutToStringPtr(Rg) ZENFct(Rg).MatPutToString
#define ZENMatPutToString(r,Rg) (*(ZENMatPutToStringPtr(Rg)))(r,Rg)
#if 0
/*
   \TEX{
   \title{Reading from string in internal format}
   \input{An unallocated |ZENMat|, a string, and a |ZENRing|.}
   \output{|ZENERR| if an error occured, the number of read characters  otherwise}
   \side{The matrix is allocated and read from string |s|.}
   }
   */
int ZENMatGetFromString(r,s,Rg)
     ZENMat r;
     char *s;
     ZENRing Rg; {}
#endif
#define ZENMatGetFromStringPtr(Rg) ZENFct(Rg).MatGetFromString
#define ZENMatGetFromString(r,s,Rg) (*(ZENMatGetFromStringPtr(Rg)))(&(r),s,Rg)
#if 0
/*
   \TEX{
   \title{Printing to file in internal format}
   \input{A file, a |ZENMat|, and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \note{Permutation matrices cannot be used.}
   }
   */
int ZENMatPutToFile(file,r,Rg)
     FILE *file;
     ZENMat r;
     ZENRing Rg; {}
#endif
#define ZENMatPutToFilePtr(Rg) ZENFct(Rg).MatPutToFile
#define ZENMatPutToFile(file,r,Rg) (*(ZENMatPutToFilePtr(Rg)))(file,r,Rg)
#if 0
/*
   \TEX{
   \title{Reading from file}
   \input{An unallocated |ZENMat|, a file, and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The matrix is allocated and read from file |file|.}
   }
   */
int ZENMatGetFromFile(r,file,Rg)
     ZENMat r;
     FILE *file;
     ZENRing Rg; {}
#endif
#define ZENMatGetFromFilePtr(Rg) ZENFct(Rg).MatGetFromFile
#define ZENMatGetFromFile(r,file,Rg) (*(ZENMatGetFromFilePtr(Rg)))(&(r),file,Rg) 

#if 0
/*
   \TEX{
   \Subsection{Matrix conversion}
   The following functions are intended to allow conversion of matrices between an extension and its definition |ZENRing|.
   
   \title{Simple assignment}
   \input{Two couples (|ZENMat|, |ZENRing|).}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |M1| is assigned with |M2|.}
   \note{The coefficients of |M2| must be elements of |R1|. The two matrices must be of same sizes.}
   } 
   */
int ZENMatConvert(M1,R1,M2,R2)
     ZENMat M1,M2;
     ZENRing R1,R2; {}
#endif
#define ZENMatConvert(M1,R1,M2,R2) _ZENMatConvert(M1,R1,M2,R2) 

/*
   \TEX{
   Now consider the following example: given a vector defined over $\GF{q^n}$
   $$
   m=\pmatrix{
   a_{0,0} + a_{0,1}x + \cdots + a_{0,n-1}x^{n-1} \cr
   \vdots \cr
   a_{k-1,0} + a_{k-1,1}x + \cdots + a_{k-1,n-1}x^{n-1}}
   $$
   we have two matrices over $\GF{q}$ that we can need to consider:
   $$
   M1=\pmatrix{
   a_{0,0} & a_{0,1} & \cdots & a_{0,n-1} \cr
   \vdots  & \vdots  &        & \vdots    \cr
   a_{k-1,0} & a_{k-1,1} & \cdots & a_{k-1,n-1}}
   \mbox{ and }
   M2=\pmatrix{
   a_{0,0} \cr a_{0,1} \cr \vdots \cr a_{0,n-1} \cr
   \vdots  \cr
   a_{k-1,0} \cr a_{k-1,1} \cr \vdots \cr a_{k-1,n-1}}.
   $$
   The following functions allow such conversion.
   The principle is that if the input matrix on, say $\GF{q}$, is of row type the conversion will give a $M1$ type matrix.
   Otherwise, it will be a $M2$ type matrix.
   Sizes of the matrices must be chosen accordingly.
   }
   */

#if 0
/*     
   \TEX{
   \title{Extension conversion}
   \input{Two matrices |ZENMat| and an extension |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |M| is assigned with |m|.}
   \note{If the |ZENRing| is not an extension, this is equivalent to |ZENMatAssign|.}
   } 
   */
int ZENVect2Mat(m,M,R)
     ZENMat M,m;
     ZENRing R; {}
#endif
#define ZENVect2MatPtr(Rg) ZENFct(Rg).Vect2Mat
#define ZENVect2Mat(m,M,Rg) (*(ZENVect2MatPtr(Rg)))(m,M,Rg) 

#if 0
/*     
   \TEX{
   \title{Extension conversion}
   \input{Two matrices |ZENMat| and an extension |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |m| is assigned with |M|.}
   \note{If the |ZENRing| is not an extension, this is equivalent to |ZENMatAssign|.}
   } 
   */
int ZENMat2Vect(M,m,R)
     ZENMat M,m;
     ZENRing R; {}
#endif
#define ZENMat2VectPtr(Rg) ZENFct(Rg).Mat2Vect
#define ZENMat2Vect(M,m,Rg) (*(ZENMat2VectPtr(Rg)))(M,m,Rg) 


/* \TEX{
   \Section[Series]{Procedures to handle series over finite rings}

   These procedures are current operations on series over finite rings.
   }
   */

/* \TEX{
   \Subsection{Allocation}
   }
   */
#if 0
/* \TEX{
   \title{Creation.}
   \input{An unallocated series, a finite ring |Rg| and a length |length|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The series is allocated with EXACTLY |length| allocated coefficients.}
   \note{The series IS NOT set to zero.}
   }
   */
int ZENSrAlloc(S,length, Rg)
     ZENSr S;
     int length;
     ZENRing Rg;{}
#endif
#define ZENSrCreatePtr(R) ZENFct(R).SrCreate
#define ZENSrCreate(deg,R) (*(ZENSrCreatePtr(R)))(deg,R)
#define ZENSrAlloc(S,deg,R) (((S)=ZENSrCreate(deg,R))==NULL ? ZERR : 0)

#if 0
/* \TEX{
   \title{Degree.}
   \input{An allocated series |PX| of a finite ring |Rg|.}
   \output{The degree of |PX|.}
   \note{This procedure is a macro which returns one field of
   |PX|. You can assign |ZENSrDeg(PX, Rg)|.}
   }
   */
int ZENSrDeg(PX, Rg)
     ZENSr PX;
     ZENRing Rg;{}
#endif
#define ZENSrDeg(p,Rg) (p)->deg
#if 0
/* \TEX{
   \title{Valuation.}
   \input{An allocated series |PX| of a finite ring |Rg|.}
   \output{The valuation of |PX|.}
   \note{This procedure is a macro which returns one field of
   |PX|. You can assign |ZENSrVal(PX, Rg)|}
   }
   */
int ZENSrVal(PX, Rg)
     ZENSr PX;
     ZENRing Rg;{}
#endif
#define ZENSrVal(p,Rg) (p)->val
#if 0
/* \TEX{
   \title{Length.}
   \input{An allocated series |PX| of a finite ring |Rg|.}
   \output{The maximal size of |PX|, ie the maximal difference
   |degree(PX)-valuation(PX)|.}
   \note{This procedure is a macro which returns one field of |PX|.}
   }
   */
BigNum ZENSrLgt(PX, Rg)
     ZENSr PX;
     ZENRing Rg;{}
#endif
#define ZENSrLgt(p,Rg) (p)->lgt

#if 0
/* \TEX{
   \title{Copying an allocated series.}
   \input{An allocated series |PX| of a finite ring |Rg|.}
   \output{|ZENNULL| if an error occurred, a copy of |PX| otherwise.}   
   }
   */
ZENSr ZENSrCopy (PX, Rg)
     ZENSr PX;
     ZENRing Rg;{}
#endif
#define ZENSrCopy(a, Rg) _ZENSrCopy(a, Rg) 


#if 0
/* \TEX{
   \title{Freeing.}
   \input{An allocated series |PX| of a finite ring |Rg|.}
   \side{|PX| is disallocated.}
   }
   */
void ZENSrFree(PX,Rg)
     ZENSr PX;
     ZENRing Rg;{}
#endif
#define ZENSrFreePtr(R) ZENFct(R).SrFree
#define ZENSrFree(p,R) (*(ZENSrFreePtr(R)))(p,R)

/* \TEX{
   \Subsection{Assigning}
   }
   */
#if 0
/* \TEX{
   \title{Assigning.}
   \input{Two allocated series |PX| and |RX| of a finite ring |Rg|.}
   \side{|RX| is filled with |PX|.}
   }
   */
void ZENSrAssign(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;{}
#endif
#define ZENSrAssignPtr(R) ZENFct(R).SrAssign
#define ZENSrAssign(r,p,R) (*(ZENSrAssignPtr(R)))(r,p,R)
#if 0
/* \TEX{
   \title{Setting to zero.}
   \input{An allocated series |RX| of a finite ring |Rg|.}
   \side{|RX| is set to zero.}
   }
   */
void ZENSrSetToZero(RX, Rg)
     ZENSr RX;
     ZENRing Rg;{}
#endif
#define ZENSrSetToZeroPtr(R) ZENFct(R).SrSetToZero
#define ZENSrSetToZero(p,R) (*(ZENSrSetToZeroPtr(R)))(p,R)
#if 0
/* \TEX{
   \title{Setting to random.}
   \input{An allocated series |RX|, a valuation |val|, a degree
     |deg| and a finite ring |Rg|.}
   \side{|RX| is set randomly to a series of valuation |val| and
     degree |deg|.}
   }
   */
void ZENSrSetRandom(RX, val, deg, Rg)
     ZENSr RX;
     int deg;
     ZENRing Rg;{}
#endif
#define ZENSrSetRandomPtr(R) ZENFct(R).SrSetRandom
#define ZENSrSetRandom(p,v,d,R) (*(ZENSrSetRandomPtr(R)))(p,v,d,R)

#if 0
/* \TEX{
   \title{Setting a coefficient.}
   \input{An allocated series |RX| and an element |b| of a
   finite ring |Rg|, the degree |d| of the coefficient  to set.}
   \side{The coefficient  of $X^d$ in |RX| is set to |b|.}
   \note{The degree or valuation of |RX| is not recomputed.}
   }
   */
void ZENSrSetCoeff(RX, d, b, Rg)
     ZENSr RX;
     int d;
     ZENElt b;
     ZENRing Rg;{}
#endif
#define ZENSrSetCoeffPtr(R) ZENFct(R).SrSetCoeff
#define ZENSrSetCoeff(p,i,c,R) (*(ZENSrSetCoeffPtr(R)))(p,i,c,R)
#if 0
/* \TEX{
   \title{Getting a coefficient.}
   \input{An allocated series |RX| of a finite ring |Rg|, the
   degree |d| of the coefficient  to get and an element |b| to assign.} 
   \side{|b| is filled with the coefficient of $X^d$ in |RX|.}
   }
   */
void ZENSrGetCoeff(b, RX, d, Rg)
     ZENElt b;
     ZENSr RX;
     int d;
     ZENRing Rg;{}
#endif
#define ZENSrGetCoeffNewPtr(R) ZENFct(R).SrGetCoeff
#define ZENSrGetCoeff(c,p,i,R) (*(ZENSrGetCoeffNewPtr(R)))(&(c),p,i,R)
#if 0
/* \TEX{
   \title{Extracting a coefficient.}
   \input{An allocated series |RX| of a finite ring |Rg|, the
   degree |d| of the coefficient  to set.}
   \output{The pointer to the coefficient  of $X^d$ in |RX|.}
   \side{You must not disallocate the output.}
   }
   */
ZENElt ZENSrGetCoeffPtr(RX, d, Rg)
     ZENSr RX;
     int d;
     ZENRing Rg;{}
#endif
#define ZENSrGetCoeffPtrPtr(R) ZENFct(R).SrGetCoeffPtr
#define ZENSrGetCoeffPtr(p,i,R) (*(ZENSrGetCoeffPtrPtr(R)))(p,i,R)

#if 0
/* \TEX{
   \title{Updating valuation.}
   \input{An allocated series |RX| of a finite ring |Rg|.}
   \side{The valuation is updated if it increased.}
   }
   */
void ZENSrUpdateValuation(RX, Rg)
     ZENSr RX;
     ZENRing Rg;{}
#endif
#define ZENSrUpdateValuationPtr(R) ZENFct(R).SrUpdateValuation
#define ZENSrUpdateValuation(p,R) (*(ZENSrUpdateValuationPtr(R)))(p,R)

/* \TEX{
   \Subsection{Test}
   }
   */
#if 0
/* \TEX{
   \title{Equality.}
   \input{Two allocated series |RX| and |PX| of a finite ring |Rg|.}
   \output{The predicate $|RX| = |PX|$.}
   }
   */
int ZENSrAreEqual(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;{}
#endif
#define ZENSrAreEqualPtr(R) ZENFct(R).SrAreEqual
#define ZENSrAreEqual(p,q,R) (*(ZENSrAreEqualPtr(R)))(p,q,R)
#if 0
/* \TEX{
   \title{Is zero.}
   \input{An allocated series |RX| of a finite ring |Rg|.}
   \output{The predicate $|RX| = 0$.}
   }
   */
int ZENSrIsZero(RX, Rg)
     ZENSr RX;
     ZENRing Rg;{}
#endif
#define ZENSrIsZeroPtr(R) ZENFct(R).SrIsZero
#define ZENSrIsZero(p,R) (*(ZENSrIsZeroPtr(R)))(p,R)

/* \TEX{
   \Subsection{Arithmetic}
   }
   */
#if 0
/* \TEX{
   \title{Addition.}
   \input{Two allocated series |RX| and |PX| of a finite ring |Rg|.}
   \side{$|RX| += |PX|$.}
   }
   */
void ZENSrAdd(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;{}
#endif
#define ZENSrAddPtr(R) ZENFct(R).SrAdd
#define ZENSrAdd(r,p,R) (*(ZENSrAddPtr(R)))(r,p,R)
#if 0
/* \TEX{
   \title{Negation.}
   \input{Two allocated series |RX| and |PX| of a finite ring |Rg|.}
   \side{$|RX| = -|PX|$.}
   }
   */
void ZENSrNegate(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;{}
#endif
#define ZENSrNegatePtr(R) ZENFct(R).SrNegate
#define ZENSrNegate(r,p,R) (*(ZENSrNegatePtr(R)))(r,p,R)
#if 0
/* \TEX{
   \title{Subtraction.}
   \input{Two allocated series |RX| and |PX| of a finite ring |Rg|.}
   \side{$|RX| -= |PX|$.}
   }
   */
void ZENSrSubtract(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;{}
#endif
#define ZENSrSubtractPtr(R) ZENFct(R).SrSubtract
#define ZENSrSubtract(r,p,R) (*(ZENSrSubtractPtr(R)))(r,p,R)
#if 0
/* \TEX{
   \title{Squaring.}
   \input{Two allocated series |RX| and |PX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int ZENSrSquare(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;
{}
#endif
#define ZENSrSquarePtr(R) ZENFct(R).SrSquare
#define ZENSrSquare(r,p,R) (*(ZENSrSquarePtr(R)))(r,p,R)
#if 0
/* \TEX{
   \title{Multiplication.}
   \input{Three allocated series |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|$.}
   }
   */
int ZENSrMultiply(RX, PX, QX, Rg)
     ZENSr RX, PX, QX;
     ZENRing Rg;{}
#endif
#define ZENSrMultiplyPtr(R) ZENFct(R).SrMultiply
#define ZENSrMultiply(r,p,q,R) (*(ZENSrMultiplyPtr(R)))(r,p,q,R)
#if 0
/* \TEX{
   \title{Multiplication by a scalar.}
   \input{Two series |PX|, |RX| and an element |e| of a finite ring
     |Rg|.}
   \side{$|RX|=|e||PX|$.}
   }
   */
void ZENSrMultiplyScalar(RX, PX, e, Rg)
     ZENSr RX, PX;
     ZENElt e;
     ZENRing Rg;{}
#endif
#define ZENSrMultiplyScalarPtr(R) ZENFct(R).SrMultiplyScalar
#define ZENSrMultiplyScalar(r,p,q,R) (*(ZENSrMultiplyScalarPtr(R)))(r,p,q,R)
#if 0
/* \TEX{
   \title{Divide.}
   \input{Two series |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient.}
   \output{|ZEN_HAS_INVERSE| if no error occurred, |ZEN_NO_INVERSE| if
   a factor of a modulo was discovered, |ZENERR| for an
   error} 
   \note{One must have $|valuation(PX)| \ge |valuation(QX)|$ and $|RX| \neq
   |QX|$. Call eventually |ZENSrUpdateValuation(QX, Rg)| to get the
   real valuation of |QX|}
   }
   */
int ZENSrDivide(RX, PX, QX, Rg)
     ZENSr RX, PX, QX;
     ZENRing Rg;{}
#endif
#define ZENSrDividePtr(R) ZENFct(R).SrDivide
#define ZENSrDivide(r,p,q,R) (*(ZENSrDividePtr(R)))(r,p,q,R)

/* \TEX{
   \Subsection{Input/Output}
   }
   */
#if 0
/* \TEX{
   \title{Converting from string.}
   \input{A finite ring |Rg|, an unallocated series
   |PX| and a string |s| representing a series in base
   |base|$\in \{2,\ldots, 16\}$.} 
   \output{The number of character read in |s| or |ZENERR| if an error
   occurred.}
   \side{|PX| is allocated and filled with |s| if the output is
   not |ZENERR|.}
   \note{Maple's format is used. The biggest monomial must be
   at the beginning of the string}
   }
*/
int ZENSrReadFromString(PX, s, base, Rg)
     ZENSr PX;
     char *s;
     int base;
     ZENRing Rg;{}
#endif
#define ZENSrReadFromString(r,s,base,Rg) _ZENSrReadFromString(&(r),s,base,Rg) 
#if 0
/* \TEX{
   \title{Converting to string.}
   \input{An allocated series |PX| of a finite ring |Rg| and a base
   |base|$\in \{2,\ldots, 16\}$.}
   \output{An allocated string representing |PX| in base |base| or
   |ZENNULL| if an error occurred.} 
   \note{Maple's format is used.}
   }
   */
char *ZENSrPrintToString(PX, base, Rg)
     ZENSr PX;
     int base;
     ZENRing Rg;  {}
#endif
#define ZENSrPrintToString(r,base,Rg) _ZENSrPrintToString(r,base,Rg)
#if 0
/* \TEX{
   \title{Reading from file.}
   \input{A stream |file| and an unallocated series
   |PX| of a finite ring |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|PX| is allocated and filled with the series read in |file| if
   no error occurred.}
   \note{Maple's format is used. The biggest monomial must be
   at the beginning of the stream}
   }
   */
int ZENSrReadFromFile(PX, file, base, Rg)
     ZENSr PX;
     FILE *file;
     int base;
     ZENRing Rg;{}
#endif
#define ZENSrReadFromFile(r,file,base,Rg) _ZENSrReadFromFile(&(r),file,base,Rg) 
#if 0
/* \TEX{
   \title{Printing to file.}
   \input{A stream |file|, an allocated series |PX| of a finite ring
   |Rg| and a base |base|$\in \{2,\ldots, 16\}$.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{Printing a representation of |PX| in base |base| to |file|.}
   \note{Maple's format is used.}
   }
   */
int ZENSrPrintToFile(file, PX, base, Rg)
     FILE *file;
     ZENSr PX;
     int base;
     ZENRing Rg;  {}
#endif
#define ZENSrPrintToFile(file,r,base,Rg) _ZENSrPrintToFile(file,r,base,Rg)

#if 0
/* \TEX{
   \title{Converting from a string yo an internal representation.}
   \input{An unallocated series |PX| of a finite ring |Rg| and a
   string |s| representing a series
   to an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occurred.}
   \side{|PX| is allocated and filled with |s| if the output is different
   from |ZENERR|}         
   }
   */
int ZENSrGetFromString(PX, s, Rg)
     char *s;
     ZENSr PX;
     ZENRing Rg;{}
#endif
#define ZENSrGetFromString(r,s,Rg) _ZENSrGetFromString(&(r),s,Rg)
#if 0
/* \TEX{
   \title{Converting to string to an internal representation.}
   \input{An allocated series |PX| of a finite ring |Rg|.}
   \output{An allocated string representing |PX| to an internal
   representation or |ZENNULL| if an error occurred.}    
   }
   */
char *ZENSrPutToString(PX, Rg)
     ZENSr PX;
     ZENRing Rg;{}
#endif
#define ZENSrPutToString(r,Rg) _ZENSrPutToString(r,Rg)
#if 0
/* \TEX{
   \title{Getting from file to an internal representation.}
   \input{A stream |file|, and an unallocated series |PX| of a
   finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{|PX| is filled with the series read in |file|}    
   }
   */
int ZENSrGetFromFile(PX, file, Rg)
     ZENSr PX;
     FILE *file;
     ZENRing Rg;{}
#endif
#define ZENSrGetFromFile(r,file,Rg) _ZENSrGetFromFile(&(r),file,Rg) 
#if 0
/* \TEX{
   \title{Writing to file to an internal representation.}
   \input{A stream |file|, an allocated series |PX| of a finite ring
   |Rg|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{Printing a representation of |PX| to |file|.}    
   }
   */
int ZENSrPutToFile(file, PX, Rg)
     FILE *file;
     ZENSr PX;
     ZENRing Rg;  {}
#endif
#define ZENSrPutToFile(file,r,Rg) _ZENSrPutToFile(file,r,Rg)

/* Undocumented macros */
#define ZENSrAllocStruct(Rg) (ZENSr)ZENMalloc(sizeof(struct zensr))

#define ZENSrDigit(p,Rg) (p)->coeff.b
#define ZENSrGetDigit(p,i,Rg) ZBNDigit(ZENSrDigit(p,Rg), i)

#define ZENSrChar(p,Rg) (p)->coeff.h
#define ZENSrGetChar(p,i,Rg) ZENSrChar(p,Rg)[i]

#define ZENSrCof(p,Rg) (p)->coeff.c
#define ZENSrGetCof(p,i,Rg) ZENSrCof(p,Rg)[i]


/*
   \TEX{
   \Section[Elliptic curves]{Procedures to handle elliptic curves.}

   These routines basically implement the group law defined on an
   elliptic curve. In |zen|, such an elliptic curve is given by its
   Weierstrass parametrization,
   \begin{displaymath}
     Y^2+a_1XY+a_3Y=X^3+a_2X^2+a_4X+a_6.
   \end{displaymath}
   
   For efficieny, |ZEN| procedures do not work directly with these
   parameters but on an internal Weierstrass parametrization. This
   internal parametrization (isomorphic to the user parametrization)
   depends on the ``characteristic'' and on the invariant of the curve.
   This internal parametrization is:
   \begin{description}
   \item[Characteristic 2, invariant $=$ 0:]
     \begin{math}
       Y^2+a_3Y=X^3+a_4X+a_6.
     \end{math}
   \item[Characteristic 2, invariant $\neq$ 0:]
     \begin{math}
       Y^2+XY=X^3+a_4X+a_6.
     \end{math}
   \item[Characteristic 3, invariant $=$ 0:]
     \begin{math}
       Y^2=X^3+a_4X+a_6.
     \end{math}
   \item[Characteristic 3, invariant $\neq$ 0:]
     \begin{math}
       Y^2=X^3+a_2X^2+a_6.
     \end{math}
   \item[Characteristic $>$ 3:]
     \begin{math}
       Y^2=X^3+a_4X+a_6.
     \end{math}
   \end{description}

   The conversion between this internal parametrization and the user
   parametrization is done once at the initialization of the curve with
   |ZENEcInitialize| and each time an I/O operation is performed, for
   instance |ZENEcPtSetX|, |ZENEcPtSetXY|, |ZENEcPrintToString|,\ldots

   \Subsection{Elliptic curve}

   } */
#if 0
/* \TEX{
   \title{Allocation of an elliptic curve.}
   \input{An unallocated elliptic curve and a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The elliptic curve is allocated.}
   }
   */
int ZENEcAlloc(E,Rg)
     ZENEc E;
     ZENRing Rg;{}
#endif
#define ZENEcCreate(Rg) _ZENEcCreate(Rg)
#define ZENEcAlloc(M,R) (((M)=ZENEcCreate(R))==NULL ? ZERR : 0)

#if 0
/* \TEX{
  \title{Initialization of an elliptic curve.}
  \input{Five parameters |a1|,|a2|, |a3|, |a4|, |a6| of an elliptic
    curve the equation of which is $Y^2+a_1XY+a_3Y=X^3+a_2X^2+a4_X+a6$.}
  \output{|ZEN_HAS_INVERSE| if the curve was succesfully initialized,
    |ZEN_NO_INVERSE| if a factor of a modulo was found, |ZENERR| otherwise.}
  \side{The curve |E| is initialized. Its internal parametrization is computed.}
  }
*/
int ZENEcInitialize(E,a1,a2,a3,a4,a6)
     ZENEc E;
     ZENElt a1, a2, a3, a4, a6;{}
#endif
#define ZENEcInitialize(E,a1,a2,a3,a4,a6) _ZENEcInitialize(E,a1,a2,a3,a4,a6)
#define ZENEcInitializeChar2(E) _ZENEcInitializeChar2(E)
#define ZENEcInitializeChar3(E) _ZENEcInitializeChar3(E)
#define ZENEcInitializeCharp(E) _ZENEcInitializeCharp(E)

#if 0
/* \TEX{
   \title{Disallocating an elliptic curve}
   \input{An elliptic curve |E|.}
   \side{|E| is freed.}
   }
   */
void ZENEcFree(E)
     ZENEc E;{}
#endif
#define ZENEcFree(E) _ZENEcFree(E)

#if 0
/* \TEX{
   \title{$a_1$}
   \input{An elliptic curve |E|.}
   \output{The coefficient |a1| of the curve.}
   \note{This procedure is a macro, do not disallocate its output.}
   }
   */
ZENElt ZENEcA1(E)
     ZENEc E;{}
#endif
#define ZENEcA1(E) (E)->a1

#if 0
/* \TEX{
   \title{$a_2$}
   \input{An elliptic curve |E|.}
   \output{The coefficient |a2| of the curve.}
   \note{This procedure is a macro, do not disallocate its output.}
   }
   */
ZENElt ZENEcA2(E)
     ZENEc E;{}
#endif
#define ZENEcA2(E) (E)->a2

#if 0
/* \TEX{
   \title{$a_3$}
   \input{An elliptic curve |E|.}
   \output{The coefficient |a3| of the curve.}
   \note{This procedure is a macro, do not disallocate its output.}
   }
   */
ZENElt ZENEcA3(E)
     ZENEc E;{}
#endif
#define ZENEcA3(E) (E)->a3

#if 0
/* \TEX{
   \title{$a_4$}
   \input{An elliptic curve |E|.}
   \output{The coefficient |a4| of the curve.}
   \note{This procedure is a macro, do not disallocate its output.}
   }
   */
ZENElt ZENEcA4(E)
     ZENEc E;{}
#endif
#define ZENEcA4(E) (E)->a4

#if 0
/* \TEX{
   \title{$a_6$}
   \input{An elliptic curve |E|.}
   \output{The coefficient |a6| of the curve.}
   \note{This procedure is a macro, do not disallocate its output.}
   }
   */
ZENElt ZENEcA6(E)
     ZENEc E;{}
#endif
#define ZENEcA6(E) (E)->a6

#if 0
/* \TEX{
   \title{The discriminant.}
   \input{An elliptic curve |E|.}
   \output{The discrminant of the curve.}
   \note{This procedure is a macro, do not disallocate its output.}
   }
   */
ZENElt ZENEcD(E)
     ZENEc E;{}
#endif
#define ZENEcD(E) (E)->Da

#if 0
/* \TEX{
   \title{The invariant.}
   \input{An elliptic curve |E|.}
   \output{The invariant of the curve.}
   \note{This procedure is a macro, do not disallocate its output.}
   }
   */
ZENElt ZENEcJ(E)
     ZENEc E;{}
#endif
#define ZENEcJ(E) (E)->J

/* \TEX{
   \Subsection{Allocation}
   }
   */
#if 0
/* \TEX{
   \title{Point allocation}
   \input{An unallocated |ZENEcPt| and an elliptic curve |E|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The point is allocated.}
   }
   */
int ZENEcPtAlloc(Pt,E)
     ZENEcPt Pt;
     ZENEc E;{}
#endif
#define ZENEcPtCreate(E) (*((E)->PtCreate))(E)
#define ZENEcPtAlloc(P,E) (((P)=ZENEcPtCreate(E))==NULL ? ZERR : 0)

#if 0
/* \TEX{
   \title{Point freeing}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \side{|P| is disallocated.}
   }
   */
void ZENEcPtFree(P, E)
     ZENEcPt P;
     ZENEc E;{}
#endif
#define ZENEcPtFree(P, E) (*((E)->PtFree))(P, E)

/* \TEX{
   \Subsection{Assigning}
   }
   */
#if 0
/* \TEX{
   \title{Setting a point to zero}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \side{|P| is set to the identity element of the curve.}
   }
   */
void ZENEcPtSetToZero(P,E)
     ZENEcPt P;
     ZENEc E;{}
#endif
#define ZENEcPtSetToZero(P, E) (*((E)->PtSetToZero))(P, E)

#if 0

/* \TEX{
   \title{Enumerating points on a curve.}
   \input{Two points |P| and |Q| of a curve |E|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse of
   a modulo is found , |ZEN_HAS_INVERSE| otherwise.}
   \side{|P| is assigned with a point whose abcissa is obtained from
   the abcissa of |Q| by |ZENEltSetNext|.}
   \note{Valid only in finite fields. This procedure loops for ever if
   there is no point on the curve !}
   }
   */
int ZENEcPtSetNext(P, Q, E)
     ZENEcPt P;
     ZENEcPt Q;
     ZENEc E;{}
#endif
#define ZENEcPtSetNext(P, Q, E) (*((E)->PtSetNext))(P, Q, E)

#if 0
/* \TEX{
   \title{Assigning a point with another point.}
   \input{Two allocated points |P| and |Q| of an elliptic curve |E|.}
   \side{|P| is filled with |Q|.}
   }
   */
void ZENEcPtAssign(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;{}
#endif
#define ZENEcPtAssign(P, Q, E) (*((E)->PtAssign))(P, Q, E)

#if 0
/* \TEX{
   \title{Getting the abcissa.}
   \input{An allocated point |P| of an elliptic curve |E| and an
   element |a|.}
   \side{|a| is filled with the abscissa of |P|.}
   }
   */
void ZENEcPtGetX(a, P, E)
     ZENElt a;
     ZENEcPt P;
     ZENEc E;{}
#endif
#define ZENEcPtGetX(a, P, E) (*((E)->PtGetX))(&(a), P, E)

#if 0
/* \TEX{
   \title{Getting the ordinate.}
   \input{An allocated point |P| of an elliptic curve |E| and an
   element |a|.}
   \side{|a| is filled with the abscissa of |P|.}
   }
   */
void ZENEcPtGetY(a, P, E)
     ZENElt a;
     ZENEcPt P;
     ZENEc E;{}
#endif
#define ZENEcPtGetY(a, P, E) (*((E)->PtGetY))(&(a), P, E)

#if 0
/* \TEX{
   \title{Setting the abcissa and ordinate of a point.}
   \input{An allocated point |P| of an elliptic curve |E| and two
   elements |x|, |y|.}
   \output{The abcissa and ordiante of |P| are filled
   respectively with |x| and |y|.}
   }
   */
void ZENEcPtSetXY(P,x,y,E)
     ZENEcPt P;
     ZENElt x,y;
     ZENEc E;{}
#endif
#define ZENEcPtSetXY(P,x,y,E) (*((E)->PtSetXY))(P,x,y,E)

/* Undocumented macros */
#define ZENEcPtGeneral2Canonical(P,Q,E) (*((E)->PtGeneral2Canonical))(P,Q,E)
#define ZENEcPtCanonical2General(P,Q,E) (*((E)->PtCanonical2General))(P,Q,E)

#define ZENEcB1(E) (E)->b1
#define ZENEcB2(E) (E)->b2
#define ZENEcB3(E) (E)->b3
#define ZENEcB4(E) (E)->b4
#define ZENEcB6(E) (E)->b6

#define ZENEcDb(E) (E)->Db

#define ZENEcAd(E) (E)->Ad
#define ZENEcBd(E) (E)->Bd
#define ZENEcCd(E) (E)->Cd
#define ZENEcDd(E) (E)->Dd
#define ZENEcEd(E) (E)->Ed

#define ZENEcAi(E) (E)->Ai
#define ZENEcBi(E) (E)->Bi
#define ZENEcCi(E) (E)->Ci
#define ZENEcDi(E) (E)->Di
#define ZENEcEi(E) (E)->Ei

/* \TEX{
   \Subsection{Tests}
   }
   */

#if 0
/* \TEX{
   \title{Is a point equal to another point ?}
   \input{Two allocated points |P| and |Q| of an elliptic curve |E|.}
   \output{The predicate $|P|=|Q|$}
   }
   */
int ZENEcPtAreEqual(P, Q, E)
     ZENEcPt P, Q;
     ZENEc E;{}
#endif
#define ZENEcPtAreEqual(P, Q, E) (*((E)->PtAreEqual))(P, Q, E)

#if 0
/* \TEX{
   \title{Is a point on the curve ?}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \output{The predicate $|P|\in|E|$}
   }
   */
int ZENEcPtIsOnEc(P, E)
     ZENEcPt P;
     ZENEc E;{}
#endif
#define ZENEcPtIsOnEc(P, E) (*((E)->PtIsOnEc))(P, E)

#if 0
/* \TEX{
   \title{Is a point equal to zero ?}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \output{The predicate $|P|=0$}
   }
   */
int ZENEcPtIsZero(P, E)
     ZENEcPt P;
     ZENEc E;{}
#endif
#define ZENEcPtIsZero(P, E) (*((E)->PtIsZero))(P, E)

/* \TEX{
   \Subsection{Arithmetic}
   }
*/

#if 0
/* \TEX{
   \title{Addition.}
   \input{Three allocated points |P|, |Q| and |R| of an elliptic curve |E|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|R| = |P|+|Q|$ if |ZEN_HAS_INVERSE| is returned,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   \note{We must have $|R|\neq|P|$}
   }
   */
int ZENEcPtAdd(R, P, Q, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEcPt Q;
     ZENEc E;{}
#endif
#define ZENEcPtAdd(R, P, Q, E) _ZENEcPtAdd(R, P, Q, E) 

#if 0
/* \TEX{
   \title{Subtraction.}
   \input{Three allocated points |P|, |Q| and |R| of an elliptic curve |E|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|R| = |P|-|Q|$ if |ZEN_HAS_INVERSE| is returned,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   \note{We must have $|R|\neq|P|$}
   }
   */
int ZENEcPtSubtract(R, P, Q, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEcPt Q;
     ZENEc E;{}
#endif
#define ZENEcPtSubtract(R, P, Q, E) _ZENEcPtSubtract(R, P, Q, E) 

#if 0
/* \TEX{
   \title{Doubling.}
   \input{Two allocated points |P| and |R| of an elliptic curve |E|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|R| = 2|P|$ if |ZEN_HAS_INVERSE| is returned,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int ZENEcPtDouble(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;{}
#endif
#define ZENEcPtDouble(R, P, E) _ZENEcPtDouble(R, P, E)

#if 0
/* \TEX{
   \title{Opposite.}
   \input{Two allocated points |P| and |R| of an elliptic curve |E|.}
   \side{$|R| = -|P|$ if |ZEN_HAS_INVERSE| is returned,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
void ZENEcPtNegate(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;{}
#endif
#define ZENEcPtNegate(R, P, E) (*((E)->PtNegate))(R, P, E)

#if 0
/* \TEX{
   \title{Multiplication.}
   \input{Two allocated points |P| and |R| of an elliptic curve |E|
   and a large integer |k| of size |kl|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|R| = |k||P|$ if |ZEN_HAS_INVERSE| is returned,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int ZENEcPtMult(R, k, kl, P, E)
     ZENEcPt R;
     BigNum k;
     BigNumLength kl;
     ZENEcPt P;
     ZENEc E;{}
#endif
#define ZENEcPtMult(R, k, kl, P, E) _ZENEcPtMult(R, k, kl, P, E)


#if 0
/*
   \TEX{
   \title{Adding several points at the same time on elliptic curves.}
   \input{2 arrays of |nb| points to add on different curves.
   3 different additions are possible following |type|:
   \begin{itemize}
   \item |type[i]|=|MEc_No|: Nothing is done.
   \item |type[i]|=|MEc_Add|: |*p_P| and |*p_Q| are added.
   \item |type[i]|=|MEc_Double|: |*p_P| is doubled.
   \end{itemize}}
   \output{-1 if an error occured, |ZEN_NO_INVERSE| if a factor is
   found, |ZEN_HAS_INVERSE| otherwise.}
   \side{The result is put in |*p_R|.}
   } 
   */
int ZENMEcPtAdd(p_R, p_P, p_Q, p_E, type, nb, Rg)
     ZENEcPt *p_R, *p_P, *p_Q;
     ZENEc *p_E;
     char *type;
     int nb;
     ZENRing Rg;{}
#endif
#define ZENMEcPtAdd(p_R, p_P, p_Q, p_E, type, nb, Rg) \
        _ZENMEcPtAdd(p_R, p_P, p_Q, p_E, type, nb, Rg)

#if 0
/*
   \TEX{
   \title{Multiplying several points by integers at the same time
   on elliptic curves.}
   \input{An array of points |p_P| and an array of BigNum (|p_k|,
   |p_kl|).} 
   \output{-1 if an error occured, |ZEN_NO_INVERSE| if a factor is
   found, |ZEN_HAS_INVERSE| otherwise.}
   \side{The products of |p_P| by the BigNums |p_k| are put in |p_R|.}
   } 
   */
int ZENMEcPtMult(p_R, p_k, p_kl, p_P, p_E, nb, Rg)
     ZENEcPt *p_R, *p_P;
     BigNum *p_k;
     BigNumLength *p_kl;
     ZENEc *p_E;
     int nb;
     ZENRing Rg;{}
#endif
#define ZENMEcPtMult(p_R, p_k, p_kl, p_P, p_E, nb, Rg) \
        _ZENMEcPtMult(p_R, p_k, p_kl, p_P, p_E, nb, Rg)


/* \TEX{
   \Subsection{Input/Output}
   }
   */

#if 0
/* \TEX{
   \title{Converting from string.}
   \input{An elliptic curve |E|, an allocated point |R| and a string
   |s| representing a point in base $|base|\in \{2,\ldots,
   16\}$.} 
   \output{The number of character read in |s| or |ZENERR| if an error occurred.}
   \side{|R| is filled with |s| if the output is not |ZENERR|.}
   }
*/
int ZENEcPtReadFromString(R, s, base, E)
     ZENEcPt R;
     char *s;
     int base;
     ZENEc E;{}
#endif
#define ZENEcPtReadFromString(R, s, base, E) _ZENEcPtReadFromString(R, s, base, E)

#if 0
/* \TEX{
   \title{Converting to string.}
   \input{An allocated point |R| of an elliptic curve |E| and a base
   $|base|\in \{2,\ldots, 16\}$.} 
   \output{An allocated string representing |R| in base |base| or |ZENNULL|
   if an error occurred.} 
   }
   */
char* ZENEcPtPrintToString(R, base, E)
     ZENEcPt R;
     int base;
     ZENEc E;{}
#endif
#define ZENEcPtPrintToString(R, base, E) _ZENEcPtPrintToString(R, base, E)

#if 0
/* \TEX{
   \title{Reading from file.}
   \input{A stream |file|, an allocated point |R| of an elliptic curve
   |E| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|R| is filled with the point read in |file| if no error
   occurred.}
   }
   */
int ZENEcPtReadFromFile(R, file, base, E)
     ZENEcPt R;
     FILE *file;
     int base;
     ZENEc E;{}
#endif
#define ZENEcPtReadFromFile(R, file, base, E) _ZENEcPtReadFromFile(R, file, base, E)

#if 0
/* \TEX{
   \title{Printing to file.}
   \input{A stream |file|, an allocated point |R| of an elliptic curve
   |E| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{Printing a representation of |R| in base |base| to |file|.}
   }
   */
int ZENEcPtPrintToFile(file, R, base, E)
     FILE *file;
     ZENEcPt R;
     int base;
     ZENEc E;{}
#endif
#define ZENEcPtPrintToFile(file, R, base, E) _ZENEcPtPrintToFile(file, R, base, E)

#if 0
/* \TEX{
   \title{Converting from a string to an internal representation.}
   \input{An allocated point |R| of an elliptic curve |E| and a string
   |s| representing a point to an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occurred.}
   \side{|R| is filled with |s| if the output is different from |ZENNULL|}
   }
   */
int ZENEcPtGetFromString(R, s, E)
     ZENEcPt R;
     char *s;
     ZENEc E;{}
#endif
#define ZENEcPtGetFromString(R, s, E) _ZENEcPtGetFromString(R, s, E)

#if 0
/* \TEX{
   \title{Converting to string to an internal representation.}
   \input{An allocated point |R| of an elliptic curve |E|.}
   \output{An allocated string representing |R| to an internal
   representation or |ZENNULL| if an error occurred.} 
   }
   */
char* ZENEcPtPutToString(R, E)
     ZENEcPt R;
     ZENEc E;{}
#endif
#define ZENEcPtPutToString(R, E) _ZENEcPtPutToString(R, E)

#if 0
/* \TEX{
   \title{Getting from file to an internal representation.}
   \input{A stream |file|, and an point |R| of an elliptic curve |E|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{|R| is filled with the point read in |file|}
   }
   */
int ZENEcPtGetFromFile(R, file, E)
     ZENEcPt R;
     FILE *file;
     ZENEc E;{}
#endif
#define ZENEcPtGetFromFile(R, file, E) _ZENEcPtGetFromFile(R, file, E)

#if 0
/* \TEX{
   \title{Writing to file to an internal representation.}
   \input{A stream |file|, an allocated point |R| of an elliptic curve
   |E|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{Printing a representation of |R| to |file|.}
   }
   */
int ZENEcPtPutToFile(file, R, E)
     FILE *file;
     ZENEcPt R;
     ZENEc E;{}
#endif
#define ZENEcPtPutToFile(file, R, E) _ZENEcPtPutToFile(file, R, E)

/* Undocumented macros */
#define ZENEcRing(E) ((E)->rg)
#define ZENEcAllocStruct(Rg) (ZENEc)ZENMalloc(sizeof(struct zenec))
#define ZENEcXY(E) ((E)->XY)
#define ZENEcX(E) ZENEcPtX(ZENEcXY(E), E)
#define ZENEcY(E) ZENEcPtY(ZENEcXY(E), E)
#define ZENEcZ(E) ((E)->Z)

#define ZENEcPtAllocStruct(E) (ZENEcPt)ZENMalloc(sizeof(struct zenecpt))

#define ZENEcPtX(P, E) ((P)->X)
#define ZENEcPtY(P, E) ((P)->Y)
#define ZENEcPtZ(P, E) ((P)->Z)

#define ZENEcPtPreAdd(P, Q, E) (*((E)->PtPreAdd))(P, Q, E)
#define ZENEcPtPostAdd(R, P, Q, E) (*((E)->PtPostAdd))(R, P, Q, E)
#define ZENEcPtPreDouble(P, E) (*((E)->PtPreDouble))(P, E)
#define ZENEcPtPostDouble(R, P, E) (*((E)->PtPostDouble))(R, P, E)

/*
   \TEX{
     \Section{Optimizations}

     \Subsection{Precomputations}
     \glossary{optimization@{\Large\bf optimization}}
     \label{sc:prc}
     
     Precomputations needed to speed up procedures must be set in a
     |ZENPrc| structure before giving it to |ZENRingAddPrc()| and
     |ZENRingRmPrc()|.
     
     The available flags are:
     }
   */
   
   /* \TEX{|ZENPRC_FINITE_FIELD|: Use some optimizations specific to
      finite fields. This flag is used only to indicate that the ring
      is a finite field. If set, some of the following precomputation
      flags performs better optimizations.} */
#define ZENPRC_FINITE_FIELD 0

   /* \TEX{|ZENPRC_ELT_MULTIPLY|: Speeds up multiplications in finite
     fields.  In modular rings with large modulus try to use
     Karatsuba's divide and conqueer algorithm.  In an extension, try
     to use Newton's method based on the Karatsuba's algorithm to
     compute modulos.} */
#define ZENPRC_ELT_MULTIPLY 1

/* \TEX{ |ZENPRC_ELT_EXP|: Speeds up ZENEltExp(). Provides reduction
  of the exponent modulo the cardinality minus one in finite fields
  (|ZENPRC_FINITE_FIELD| must be set).  } */
#define ZENPRC_ELT_EXP 2

/* \TEX{|ZENPRC_TRACE|: Speeds up traces in finite fields
(|ZENPRC_FINITE_FIELD| must be set).}  */
#define ZENPRC_TRACE 3

/* \TEX{ |ZENPRC_POLY_ROOTS_CANONICAL|: Speeds up the procedure
  PolyRootsCanonical() in finite fields (|ZENPRC_FINITE_FIELD| must be
  set).  } */
#define ZENPRC_POLY_ROOTS_CANONICAL 4

#if 0 
/* \TEX{
   \title{Setting all the precomputation flags.}
   \input{A precomputation structure |Prc|.}
   \side{All the precomputation flags are set in |Prc| EXCEPT the |ZENPRC_FINITE_FIELD| flag.}
   }
   */
void ZENPrcSetAll(Prc)           
     ZENPrc Prc;{}
#endif
#define ZENPrcSetAll(Prc) ((Prc) = ~((ZENPrc)1<<ZENPRC_FINITE_FIELD))

#if 0 
/* \TEX{
   \title{Setting no precomputation flag.}
   \input{A precomputation structure |Prc|.}
   \side{No precomputation flag is set in |Prc|.}
   }
   */
void ZENPrcSetNone(Prc)          
     ZENPrc Prc;{}
#endif
#define ZENPrcSetNone(Prc) ((Prc) = (ZENPrc)0)

#if 0 
/* \TEX{
   \title{Adding a precomputation flag.}
   \input{A precomputation flag |prp| and a structure |Prc|.}
   \side{The precomputation flag |prp| is set in |Prc|.}
   }
   */
void ZENPrcSet(Prc, prp)         
     ZENPrc Prc;
     int prp;{}
#endif
#define ZENPrcSet(Prc, prp) ((Prc) |= ((ZENPrc)1<<prp))

#if 0 
/* \TEX{
   \title{Removing a precomputation flag.}
   \input{A precomputation flag |prp| and a structure |Prc|.}
   \side{The precomputation flag |prp| is suppressed from |Prc|.}
   }
   */
void ZENPrcUnset(Prc, prp)               
     ZENPrc Prc;
     int prp;{}
#endif
#define ZENPrcUnset(Prc, prp) ((Prc) &= ~((ZENPrc)1<<prp))

#if 0 
/* \TEX{
   \title{Is a precomputation flag already set.}
   \input{A precomputation flag |prp| and a structure |Prc|.}
   \output{1 if |prp| is in |Prc|, 0 otherwise.}
   }
   */
int ZENPrcIsSet(Prc, prp)               
     ZENPrc Prc;
     int prp;{}
#endif
#define ZENPrcIsSet(Prc, prp) ((Prc) & ((ZENPrc)1<<prp))

#if 0 
/* \TEX{
   \title{Assigning a precomputation flag in another flag.}
   \input{Two precomputation structures |prc1| and |prc2|.}
   \side{|prc2| is equal to |prc1|.}
   }
   */
void ZENPrcAssign(prc1, prc2)
     ZENPrc prc1, prc2;{}
#endif
#define ZENPrcAssign(prc1, prc2) (prc1) = (prc2)

#if 0 
/* \TEX{
   \title{Intersection of 2 precomputation flags.}
   \input{Two precomputation structures |prc1|, |prc2|.}
   \side{$|prc2| \&= |prc1|$.}
   }
   */
void ZENPrcAnd(prc1, prc2)
     ZENPrc prc1, prc2;{}
#endif
#define ZENPrcAnd( prc1, prc2) (prc1) &= (prc2)

#if 0 
/* \TEX{
   \title{Union of 2 precomputation flags.}
   \input{Two precomputation structures |prc1|, |prc2|.}
   \side{$|prc2| \vert = |prc1|$.}
   }
   */
void ZENPrcOr(prc1, prc2)
     ZENPrc prc1, prc2;{}
#endif
#define ZENPrcOr(prc1, prc2) (prc1) |= (prc2)

#if 0 
/* \TEX{
   \title{Negation of a precomputation flag.}
   \input{A precomputation structure |prc|.}
   \side{$|prc| = \tilde{ } |prc|$.}
   }
   */
void ZENPrcNeg(prc)
     ZENPrc prc;{}
#endif
#define ZENPrcNeg(prc) (prc) = ~(prc)

#if 0
/* \TEX{
   \title{Precomputing.}
   \input{A finite ring |Rg| and precomputations |Prc|.} 
   \output{|ZENERR| if an error occured, |ZEN_NO_INVERSE| if a factor of a
   modulo was discovered, |ZEN_HAS_INVERSE| otherwise.}
   \side{Precomputations asked in the flags of |Prc| are done in order to
     speed up the corresponding procedures. |ZENRingFact(Rg)| is
     filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   \note{Precomputations can take time\ldots}
   }
   */
int ZENRingAddPrc(Rg, Prc)
     ZENRing Rg;
     ZENPrc Prc;{}
#endif
#define ZENRingAddPrc(Rg, Prc)  _ZENRingAddPrc(Rg, Prc) 

#if 0
/* \TEX{
   \title{Suppressing precomputations.}
   \input{A finite ring |Rg| and precomputations |Prc|.} 
   \side{Structures already allocated by  |ZENRingAddPrc()| are
   disallocated.}
   }
   */
void ZENRingRmPrc(Rg, Prc)
     ZENRing Rg;
     ZENPrc Prc;{}
#endif
#define ZENRingRmPrc(Rg, Prc) _ZENRingRmPrc(Rg, Prc)
#if 0
/* \TEX{
   \title{The precomputations of |Rg|.}
   \input{A finite ring |Rg|.}
   \output{The precomputations already done in |Rg|.}
   \note{This procedure is a macro which returns one field of
   |Rg|.}
   }
   */
ZENPrc ZENRingPrcp(Rg)
     ZENRing Rg;{}
#endif
#define ZENRingPrcp(Rg) ((Rg)->prc)


/*
   \TEX{
   \Subsection{Clones}

   Type of clones must be set in a |ZENCln| structure before giving it
   to |ZENRingClone()|.

   The available flags are:
   }
   */

/*
   \TEX{|ZENCLN_LOG|: Use a generator to store results of operations in tables.
   This is possible when the number of elements in the finite field is smaller than $2^{16}$. This overrides |ZENCLN_TABULATE| and |ZENCLN_MONTGOMERY| when several flags are set.}
*/  
#define ZENCLN_LOG 0
/*
   \TEX{|ZENCLN_TABULATE|: Store results of operations in tables
   whenever it's possible,  mainly when the number of elements in the
   ring is smaller than $2^8$.}
*/  
#define ZENCLN_TABULATE 1
/*
   \TEX{|ZENCLN_MONTGOMERY|: In modular rings, use Montgomery's reduction.}
*/  
#define ZENCLN_MONTGOMERY 2

#if 0 
/* \TEX{
   \title{Setting all the clone flags.}
   \input{A clone structure |Cln|.}
   \side{All the clone flags are set in |Cln|.}
   }
   */
void ZENClnSetAll(Cln)           
     ZENCln Cln;{}
#endif
#define ZENClnSetAll(Cln) ((Cln) = ~((ZENCln)0))

#if 0 
/* \TEX{
   \title{Setting no clone flag.}
   \input{A clone structure |Cln|.}
   \side{No clone flag is set in |Cln|.}
   }
   */
void ZENClnSetNone(Cln)          
     ZENCln Cln;{}
#endif
#define ZENClnSetNone(Cln) ((Cln) = 0)

#if 0 
/* \TEX{
   \title{Adding a clone flag.}
   \input{A clone flag |prp| and a structure |Cln|.}
   \side{The clone flag |prp| is set in |Cln|.}
   }
   */
void ZENClnSet(Cln, prp)         
     ZENCln Cln;
     int prp;{}
#endif
#define ZENClnSet(Cln, prp) ((Cln) |= ((ZENCln)1<<prp))

#if 0 
/* \TEX{
   \title{Removing a clone flag.}
   \input{A clone flag |prp| and a structure |Cln|.}
   \side{The clone flag |prp| is suppressed from |Cln|.}
   }
   */
void ZENClnUnset(Cln, prp)               
     ZENCln Cln;
     int prp;{}
#endif
#define ZENClnUnset(Cln, prp) ((Cln) &= ~((ZENCln)1<<prp))

#if 0 
/* \TEX{
   \title{Is a clone flag already set.}
   \input{A clone flag |prp| and a structure |Cln|.}
   \output{1 if |prp| is in |Cln|, 0 otherwise.}
   }
   */
int ZENClnIsSet(Cln, prp)               
     ZENCln Cln;
     int prp;{}
#endif
#define ZENClnIsSet(Cln, prp) ((Cln) & ((ZENCln)1<<prp))

#if 0 
/* \TEX{
   \title{Removing a clone flag.}
   \input{A clone flag |prp| and a structure |Cln|.}
   \side{The clone flag |prp| is suppressed from |Cln|.}
   }
   */
void ZENClnUnset(Cln, prp)
     ZENCln Cln;
     int prp;{}
#endif
#define ZENClnUnset(Cln, prp) ((Cln) &= ~((ZENCln)1<<prp))

#if 0 
/* \TEX{
   \title{Assigning a clone flag in another flag.}
   \input{Two clone structures |cln1| and |cln2|.}
   \side{|cln2| is equal to |cln1|.}
   }
   */
void ZENClnAssign(cln1, cln2)
     ZENCln cln1, cln2;{}
#endif
#define ZENClnAssign(cln1, cln2) (cln1) = (cln2)

#if 0 
/* \TEX{
   \title{Intersection of 2 clone flags.}
   \input{Two clone structures |cln1|, |cln2|.}
   \side{$|cln2| \&= |cln1|$.}
   }
   */
void ZENClnAnd(cln1, cln2)
     ZENCln cln1, cln2;{}
#endif
#define ZENClnAnd( cln1, cln2) (cln1) &= (cln2)

#if 0 
/* \TEX{
   \title{Union of 2 clone flags.}
   \input{Two clone structures |cln1|, |cln2|.}
   \side{$|cln2| \vert= |cln1|$.}
   }
   */
void ZENClnOr(cln1, cln2)
     ZENCln cln1, cln2;{}
#endif
#define ZENClnOr(cln1, cln2) (cln1) |= (cln2)

#if 0 
/* \TEX{
   \title{Negation of a clone flag.}
   \input{A clone structure |cln|.}
   \side{$|cln| = \tilde{ } |cln|$.}
   }
   */
void ZENClnNeg(cln)
     ZENCln cln;{}
#endif
#define ZENClnNeg(cln) (cln) = ~(cln)

#if 0
/* \TEX{
   \title{Clone of a ring.}
   \input{A |ZENRing| |Rg| and the type of the clone |cln|.} 
   \output{A ring |R| if everything was succesfully initialized, |ZENNULL| otherwise.}
   \note{Cloning can take time. If no flag is set in |cln|, a call to
   |ZENRingCopy()| is performed. If several flags are set, this
   procedure can try to make "good choices".}   
   }
   */
ZENRing ZENRingClone(Rg, cln)
     ZENRing Rg;
     ZENCln cln;{}
#endif
#define ZENRingClone(Rg, cln) _ZENRingClone(Rg, cln)

#if 0
/* \TEX{
   \title{Chinese ring.}
   \input{An unallocated |ZENRing|, and an array of |ZENRing|s |PR| of size |N|.} 
   \output{0 if the ring |R| is succesfully initialized, |ZENNULL| otherwise.}
   \note{The obtained ring is the product of the |N| inputs. These rings must be of the same type: either |N| modular rings, or |N| polynomial extensions over a same ring. Nevertheless, some or all of these ring can be clones.}

   In fact, two rings are allocated, the returned one and its |ZENRingOrigin()|. For instance, if |PR[0]| is $\quotient{\Z}{3\Z}$ and |PR[1]| is $\quotient{\Z}{5\Z}$, then |ZENRingChinese(R,2,PR)| will affect to |R| $\quotient{\Z}{3\Z}\times\quotient{\Z}{5\Z}$, while |ZENRingOrigin(R)| will be set to  $\quotient{\Z}{15\Z}$. A subsequent call to |ZENRingFree| will also free this ring, but |R1=ZENRingCopy(R)| can be helpful to keep this original ring.
   }
   */
int ZENRingChinese(R, N, PR)
     int N;
     ZENRing R, *PR;{}
#endif
#define ZENChineseRingCreate(N, PR) _ZENChineseRingCreate(N, PR)
#define ZENRingChinese(R, N, PR) (((R)=ZENChineseRingCreate(N, PR))==NULL ? ZERR : 0)



/* undocumented macros */
#define ZENRingOrigin(Rg) ((Rg)->cl)
#define ZENRingCln(Rg) ((Rg)->nc)

#if 0
/* \TEX{
   \title{Testing whether two rings are equal.}
   \input{Two |ZENRings|.}
   \output{|ZENERR| if an error occured, 1 if the two |ZENRings| are the same, 0 otherwise.}
   \note{A clone is equal to its original ring.}
   }
   */
int ZENRingAreEqual(R1,R2)
     ZENRing R1,R2; {}
#endif
#define ZENRingAreEqual(R1,R2) _ZENRingAreEqual(R1,R2)


#include "zeg.h"

#endif
