/*
 * zen/include/zentypes.h
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#ifndef __ZENTYPES__
# define __ZENTYPES__

/* \TEX{
   \label{zentypes}
   \glossary{types@{\Large\bf types}}   
   
   The few types of |ZEN| are as follows.  First of all, let's recall
   that |ZEN| was originally based on the |BigNum| library, developed jointly by
   INRIA and Digital PRL.
   \begin{center}
   \begin{tabular}{p{0.25\textwidth}@{ }p{0.75\textwidth}} |BigNum| &
   Large integers. Basically a big integer is an array of
   |BigNumDigit| where |BigNumDigit| is simply an integer, generally
   of type |unsigned long|. 
   
   Hence, an integer $n$ will be represented by a couple of two |C|
   variables |(n,nl)| of respective types |(BigNum, BigNumLength)|, where
   the |BigNum| type is a pointer |BigNumDigit *|.
   Value of $n$ will then be
   $$ n=\sum_{i=0}^{|nl|-1}
   |n[i]|\left(2^{|SIZE_BLOC|}\right)^i.
   $$
      
   If you know the size |sl| of your integer in bits, the number of
   |BigNumDigit|s used will be given by the |ZEN| macro
   |divSizeBloc(sl)+1|. The number of bits used in the most
   significant |BigNumDigit| is |modSizeBloc(sl)|.
   \end{tabular}
   \end{center}
   \begin{note}
   It is important to note that due to an evolution in the |BigNum| library, the length of a |BigNum| is now unsigned. This may cause some bugs in previous implementations on top of |BigNum|\footnote{
   For instance, the following loop on a |BigNumLength| |nl| will now infinitely run because |nl| is always larger than zero:
   \begin{center}
   |for(nl=ZBNNumDigits(n,nl)-1;nl>=0;nl--)|
   \end{center}
   Use the following instead:
   \begin{center}
   |for(nl=ZBNNumDigits(n,nl);nl--;)|
   \end{center}
   }.
   \end{note}

   The description of the |ZBN| layer functions is done in appendix~\ref{zbn}.
   } */
#include "zbn.h"
#ifdef ___
struct zenmat;
struct zenring;
struct zenec;
#endif

struct zenfrac {
  BigNum n[2];
  BigNumLength nl[2],nlm[2];
  int exp;
  unsigned int flag;
};

/* \TEX{
   \Subsubsection{The main objects of \zen}

   Afterwards, the following types (|ZENElt|, |ZENPoly|, |ZENSr|
   and |ZENMat|) are allocated in a ``current'' ring of type |ZENRing|.

   \begin{center}
   \begin{tabular}{p{0.25\textwidth}@{  }p{0.75\textwidth}}
   |ZENElt|& Any element of a finite ring. This type is in fact an
   union of elementary types, the size of which are at most the size of
   a |unsigned long| or the size of a pointer. In a first
   approximation, you can see a |ZENElt| like a polynomial or an
   integer.
   \end{tabular}
   \end{center}
  
  }*/

typedef union zenelt {
  BigNumDigit l;
  BigNum b;
  struct zenpoly *p;
  struct zenfrac *f;
  union zenelt *c;
} ZENElt;

/* \TEX{   
   \begin{center}
   \begin{tabular}{p{0.25\textwidth}@{  }p{0.75\textwidth}}
   |ZENPoly|& Any polynomial of a finite ring. This structure can
   be seen as containing the degree of a polynomial, its allocated
   length and an array of |ZENElt|s for its coefficients.
   \end{tabular}
   \end{center}
  }*/

typedef struct zenpoly {
  int deg;
  int lgt;
  union {
    unsigned char *h;
    unsigned short *s;
    BigNum b;
    ZENElt *c;
  } coeff;
} *ZENPoly;

/* \TEX{

   \begin{center}
   \begin{tabular}{p{0.25\textwidth}@{  }p{0.75\textwidth}}
   |ZENSr|& Any truncated series of a |ZENRing|. As algorithms
   for polynomials can be speeded up for series computations, we
   introduced the concept of truncated series in |ZEN|. This structure
   can be seen as containing the valuation of a series, its truncated
   degree, its allocated length and an array of |ZENElt|s for its
   coefficients.
   \end{tabular}
   \end{center}
   
  
  }*/

typedef struct zensr {
  int deg;
  int val;
  int lgt;
  union {
    unsigned char *h;
    BigNum b;
    ZENElt *c;
  } coeff;
} *ZENSr;

/* \TEX{
   \begin{center}
   \begin{tabular}{p{0.25\textwidth}@{  }p{0.75\textwidth}}

   |ZENMat|& Any matrix of a |ZENRing|. It is classically a
   two-dimensional array of elements.  This means that C-implementation
   involves a differentiation between matrices that does not exist in
   the mathematical definition, because the access of the $(r,c)$
   element of a matrix can be done either by looking at the $c$-th
   element of the $r$-th row, or the $r$-th element of the $c$-th
   column.  A |ZENMat| will therefore be of |ZENMatTypeRow| if the
   continuity of rows is kept in the data structure, and of
   |ZENMatTypeCol| otherwise.  Of course, this has no influence on the
   result of the operations (except if indicated).  However, the
   performances may greatly change in certain cases.  For instance, the
   permutation of two rows of a matrix will be faster if the matrix is
   of type |ZENMatTypeRow|.

   Furthermore, a special class of matrices is the class of permutation
   matrix.  A permutation matrix is a square matrix with a single one
   in every row or column.  It can be stored by keeping for every row
   OR every column the column OR row index of the one.  We have
   therefore two additional types for permutation matrices
   |ZENMatTypeColPermutation| and |ZENMatTypeRowPermutation|.
   Permutation matrices should be used only by advanced users.
   
   The dimensions of a matrix are of |Dim| type which is defined in
   |zentypes.h| to be |int|.
   \end{tabular}
   \end{center}
  
  }*/


#define Dim int

typedef enum
  {
    ZENMatTypeCol,
    ZENMatTypeRow,
    ZENMatTypeColPermutation,
    ZENMatTypeRowPermutation
  } ZENMatTypeStruct;

/* Structure for the allocation of a |ZENMat| */



struct zenmatfct {
  void (*MatFree) ___((struct zenmat *, struct zenring *));
  int (*MatAssign) ___((struct zenmat *, struct zenmat *, struct zenring *));
  int (*MatSetToOne) ___((struct zenmat *, struct zenring *));
  int (*MatIsOne) ___((struct zenmat *, struct zenring *));
  int (*MatSetRandom) ___((struct zenmat *, struct zenring *));
  int (*MatSetCoeff) ___((struct zenmat *, Dim, Dim, ZENElt, struct zenring *));
  int (*MatGetCoeff) ___((ZENElt *, struct zenmat *, Dim, Dim, struct zenring *));
  ZENElt (*MatGetCoeffPtr) ___((struct zenmat *, Dim, Dim, struct zenring *));
  int (*MatPermuteBn) ___((struct zenmat *, Dim, Dim, struct zenring *));
  int (*MatPermuteRow) ___((struct zenmat *, Dim, Dim, struct zenring *));
  int (*MatPermuteCol) ___((struct zenmat *, Dim, Dim, struct zenring *));
};

typedef struct zenmat
{
  ZENMatTypeStruct type_struct;
  Dim nbbn,nbbloc,nbdigit,nbptr;
  int nbbnd,size;

  BigNumDigit *First;

  union {
    ZENElt **zeg;
    Dim *perm;
    BigNum *ze2;
    BigNum **zep;
    BigNumDigit **zeps;
    unsigned short **zelog;
    unsigned char **zetab;
  } elts;
  
  struct zenmatfct fct;    
} *ZENMat;


typedef BigNumDigit ZENCln;
typedef BigNumDigit ZENPrc;

/* \TEX{

   As explained previously, we allocate and then handle variables of
   these types in rings. Rings must have the following type.
   \begin{center}
   \begin{tabular}{p{0.25\textwidth}@{  }p{0.75\textwidth}}
   |ZENRing|& This is the main structure of |ZEN|. It represents
   one finite ring, that is to say $\Z/n\Z$ where $n$ is any integer
   (even if it is not prime) or any polynomial finite ring over another
   (even if the definition polynomial is not irreducible). It contains
   some general data about finite rings, mainly its characteristic, its
   degree, its number of elements, its definition polynomial and {\tt
   pointers to functions} able to act on elements of this finite ring.
   So each time the user will call a |ZEN| procedure, it will be
   probably automatically replaced thanks to macros defined in |zen.h|
   by a pointer stored in this structure.
   \end{tabular}
   \end{center}
  
  } */

/***************************************************************
 * Pointers on functions which must be as efficient as possible.*
 ****************************************************************/

struct zen_functions {

  /* Procedure on Rings
   ********************/
  /* Precomputations */
  int (*PrcInitEltMultiply) ___((struct zenring *));	/* EltMultiply precomputations */
  int (*PrcCloseEltMultiply) ___((struct zenring *));
  int (*PrcInitTrace) ___((struct zenring *));	/* EltTrace precomputations */
  int (*PrcCloseTrace) ___((struct zenring *));
  int (*PrcInitPolyRootsCanonical) ___((struct zenring *)); /* PolyRootsCaconical precomputations */
  int (*PrcClosePolyRootsCanonical) ___((struct zenring *));
  int (*PrcInitEltExp) ___((struct zenring *));	/* EltExp precomputations */
  int (*PrcCloseEltExp) ___((struct zenring *));

  /* Procedure on elements
   ***********************/
  /* Allocation */
  int (*EltCreate) ___((ZENElt *, struct zenring *));           /* Create elements */
  int (*EltCreateLarge) ___((ZENElt *, struct zenring *));
  void (*EltFree) ___((ZENElt, struct zenring *));            /* Free elements */

  /* Assigning */
  void (*EltAssign) ___((ZENElt *, ZENElt, struct zenring *));          /* Assign an element with another */
  void (*EltSetToZero) ___((ZENElt *, struct zenring *));       /* Set an element to zero */
  void (*EltSetToOne) ___((ZENElt *, struct zenring *));        /* Set an element to one */
  void (*EltSetRandom) ___((ZENElt *, struct zenring *));       /* Set randomly an element */
  void (*EltSetNext) ___((ZENElt *, struct zenring *));         /* Set an element to the next */
  int (*EltConvert) ___((ZENElt *, struct zenring *, ZENElt, struct zenring *)); /* Convert an element to an element
				   of the current ring */

  /* Test */
  int (*EltAreEqual) ___((ZENElt, ZENElt, struct zenring *));         /* Is an element equal to another */
  int (*EltIsZero) ___((ZENElt, struct zenring *));           /* Is an element equal to zero */
  int (*EltIsOne) ___((ZENElt, struct zenring *));            /* Is an element equal to one */
  int (*EltIsASquare) ___((ZENElt, struct zenring *));        /* Is an element of a finite fields is a square ? */

  /* Arithmetic */
  void (*EltAdd) ___((ZENElt *, ZENElt, struct zenring *));             /* Addition of elements */
  void (*EltNegate) ___((ZENElt *, ZENElt, struct zenring *));          /* Negation of an element */
  void (*EltSubtract) ___((ZENElt *, ZENElt, struct zenring *));        /* Substraction of elements */
  void (*EltSquare) ___((ZENElt *, ZENElt, struct zenring *));           /* Squaring of elements */
  void (*EltMultiply) ___((ZENElt *, ZENElt, ZENElt, struct zenring *));         /* Multiplication of elements */
  int (*EltInverse) ___((ZENElt *, ZENElt, struct zenring *));          /* Inverse of elements */
  void (*EltTrace) ___((ZENElt *, ZENElt, struct zenring *));            /* Trace of elements */
  int (*EltSquareRoot) ___((ZENElt *, ZENElt, struct zenring *));        /* Square root of elements in finite fields */
  int (*EltExp) ___((ZENElt *, BigNum, BigNumLength, ZENElt, struct zenring *));		/* Exponentiation of element */

  /* io */
  int (*EltReadFromString) ___((ZENElt *, char *, int, struct zenring *));	/* Read an element from a string */
  char* (*EltPrintToString) ___((ZENElt, int, struct zenring *));	/* Print an element to a string */
  int (*EltReadFromFile) ___((ZENElt *, FILE *, int, struct zenring *));	/* Read an element from a file */
  int (*EltPrintToFile) ___((FILE *, ZENElt, int, struct zenring *));	/* Print an element to a file */
  int (*EltGetFromString) ___((ZENElt *, char *, struct zenring *)); /* Get an element from a string */
  char* (*EltPutToString) ___((ZENElt, struct zenring *)); /* Put an element to a string */
  int (*EltGetFromFile) ___((ZENElt *, FILE *, struct zenring *)); /* Get an element from a file */
  int (*EltPutToFile) ___((FILE *, ZENElt, struct zenring *)); /* Put an element to a file */

  void (*EltToZ) ___((BigNum, BigNumLength *, ZENElt, struct zenring *)); /* Evaluate in $p$ an element */
  void (*ZToElt) ___((ZENElt *, BigNum, BigNumLength, struct zenring *));	/* Get the element from an evaluation in $p$ */

  void (*EltToClone) ___((ZENElt *, ZENElt, struct zenring *));
  void (*CloneToElt) ___((ZENElt *, ZENElt, struct zenring *));

  /* Procedures on polynomials
   *******************************/
  /* Allocation */
  ZENPoly (*PolyCreate) ___((int, struct zenring *));      /* Create polynomials */
  ZENPoly (*PolyCreateLarge) ___((int, struct zenring *));      /* Create polynomials */
  void (*PolyFree) ___((ZENPoly, struct zenring *));           /* Free polynomials */
  void (*PolySetCoeff) ___((ZENPoly, int, ZENElt, struct zenring *));       /* Set a coefficient of a polynomial */
  ZENElt (*PolyGetCoeffPtr) ___((ZENPoly, int, struct zenring *));  /* Get a pointer on a coefficient*/

  /* Assigning */
  void (*PolyAssign) ___((ZENPoly, ZENPoly, struct zenring *));         /* Assign a polynomial with another */
  void (*PolyLazyToVeryLazy) ___((ZENPoly, ZENPoly, struct zenring *));         /* Assign a polynomial with another */
  void (*PolyAssignLarge) ___((ZENPoly, ZENPoly, struct zenring *));

  void (*PolySetToZero) ___((ZENPoly, struct zenring *));      /* Set a polynomial to zero */
  void (*PolySetToZeroLarge) ___((ZENPoly, struct zenring *));      /* Set a polynomial to zero */
  void (*PolySetToXi) ___((ZENPoly, int, struct zenring *));        /* Set a polynomial to $X^i$ */
  void (*PolySetRandom) ___((ZENPoly, int, struct zenring *));      /* Set randomly a polynomial */
  void (*PolyGetCoeff) ___((ZENElt *, ZENPoly, int, struct zenring *));       /* Get a coefficient of a polynomial */
  void (*PolyUpdateDegree) ___((ZENPoly, struct zenring *));   /* Update the degree of a polynomial */

  /* Test */
  int (*PolyAreEqual) ___((ZENPoly, ZENPoly, struct zenring *));        /* Is a polynomial equal to another */
  int (*PolyIsZero) ___((ZENPoly, struct zenring *));          /* Is a polynomial equal to zero */
  int (*PolyIsXi) ___((ZENPoly, int, struct zenring *));            /* Is a polynomial equal to Xi */

  /* Arithmetic */
  void (*PolyAdd) ___((ZENPoly, ZENPoly, struct zenring *));            /* Addition of polynomials */
  void (*PolyAddLarge) ___((ZENPoly, ZENPoly, struct zenring *));
  void (*PolyNegate) ___((ZENPoly, ZENPoly, struct zenring *));         /* Negation of a polynomial */
  void (*PolyNegateLarge) ___((ZENPoly, ZENPoly, struct zenring *));         /* Negation of a polynomial */
  void (*PolySubtract) ___((ZENPoly, ZENPoly, struct zenring *));       /* Substraction of polynomials */
  void (*PolySubtractLarge) ___((ZENPoly, ZENPoly, struct zenring *));

  int  (*PolySquare) ___((ZENPoly, ZENPoly, struct zenring *));         /* Squaring of polynomials */
  int  (*PolySquareLazy) ___((ZENPoly, ZENPoly, struct zenring *));	/* Lazy Squaring of polynomials */
  int  (*PolySquareVeryLazy) ___((ZENPoly, ZENPoly, struct zenring *));	/* Lazy Squaring of polynomials */

  int (*PolyMultiply) ___((ZENPoly, ZENPoly, ZENPoly, struct zenring *));        /* Multiplication of polynomials */
  int (*PolyMultiplyLazy) ___((ZENPoly, ZENPoly, ZENPoly, struct zenring *));	/* Lazy Multiplication of polynomials */
  int (*PolyMultiplyVeryLazy) ___((ZENPoly, ZENPoly, ZENPoly, struct zenring *));	/* Lazy Multiplication of polynomials */
  void (*PolyMultiplyScalar) ___((ZENPoly, ZENPoly, ZENElt, struct zenring *));  /* Multiplication of polynomials by a scalar */
  void (*PolyDot) ___((ZENElt *, ZENPoly, ZENPoly, struct zenring *));             /* Scalar products of polynomials coefficients*/
  void (*PolyEval) ___((ZENElt *, ZENPoly, ZENElt, struct zenring *));             /* Evaluation of a polynomial */

  int (*PolyDivide) ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, struct zenring *));          /* Division of polynomials */
  int (*PolyDivideLazy) ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, struct zenring *));
  int (*PolyDivideVeryLazy) ___((ZENPoly, ZENPoly, ZENPoly, ZENPoly, struct zenring *));
  int (*PolyModuloLazy) ___((ZENPoly, ZENPoly, ZENPoly, int, struct zenring *));
  int (*PolyModuloVeryLazy) ___((ZENPoly, ZENPoly, ZENPoly, int, struct zenring *));

  int (*PolyDerive) ___((ZENPoly, ZENPoly,struct zenring *));
  int (*PolyRootsDegree2) ___((ZENPoly, ZENPoly, struct zenring *));    /* Roots of polynomials of degree 2 in finite fields */
  int (*PolyRootsCanonical) ___((ZENPoly, ZENElt, struct zenring *));  /* Roots of $X-X^p=a$ in finite fields */
  
  /* Procedures on series
   **************************/
  /* Allocation */
  ZENSr (*SrCreate) ___((int, struct zenring *));          /* Create series */
  void (*SrFree) ___((ZENSr, struct zenring *));             /* Free series  */

  /* Assigning */
  void (*SrAssign) ___((ZENSr, ZENSr, struct zenring *));           /* Assign a series with another */
  void (*SrSetToZero) ___((ZENSr, struct zenring *));        /* Set a series to zero */
  void (*SrSetRandom) ___((ZENSr,int,int,struct zenring *));        /* Set randomly a series */
  void (*SrSetCoeff) ___((ZENSr, int, ZENElt, struct zenring *));         /* Set a coefficient of a series */
  void (*SrGetCoeff) ___((ZENElt *
, ZENSr, int, struct zenring *));         /* Get a coefficient of a series */
  ZENElt (*SrGetCoeffPtr) ___((ZENSr, int, struct zenring *));    /* Get a pointer to a coefficient of a series */
  void (*SrUpdateValuation) ___((ZENSr, struct zenring *));  /* Update valuation of a series */

  /* Test */
  int (*SrAreEqual) ___((ZENSr, ZENSr, struct zenring *));          /* Is a series equal to another */
  int (*SrIsZero) ___((ZENSr, struct zenring *));            /* Is a series equal to zero */

  /* Arithmetic */
  void (*SrAdd) ___((ZENSr, ZENSr, struct zenring *));              /* Addition of series */
  void (*SrNegate) ___((ZENSr, ZENSr, struct zenring *));           /* Negation of series */
  void (*SrSubtract) ___((ZENSr, ZENSr, struct zenring *));         /* Substraction of series */
  int (*SrSquare) ___((ZENSr, ZENSr, struct zenring *));           /* Squaring of series */
  int (*SrMultiply) ___((ZENSr, ZENSr, ZENSr, struct zenring *));         /* Multiplication of series */
  void (*SrMultiplyScalar) ___((ZENSr, ZENSr, ZENElt, struct zenring *));   /* Multiplication of a series by a scalar*/
  int (*SrDivide) ___((ZENSr, ZENSr, ZENSr, struct zenring *));            /* Dividing series */
  
  /* Procedures on matrices
   **************************/
  /* Allocation */
  int (*MatCreatePms) ___((ZENMatTypeStruct,Dim,Dim,ZENMat *,struct zenring *)); /* Create a matrix parameters structure*/
  int (*MatCreate) ___((ZENMat,struct zenring *)); /* Create a matrix */

  /* Assigning */

  int (*MatAssign) ___((ZENMat, ZENMat, struct zenring *));
  int (*Perm2Mat) ___((ZENMat, ZENMat, struct zenring *));
  int (*MatSetToZero) ___((ZENMat, struct zenring *));
  int (*MatSetToOne) ___((ZENMat, struct zenring *));
  int (*MatSetRandom) ___((ZENMat, struct zenring *));
  int (*MatSetBloc) ___((ZENMat, Dim, Dim, ZENElt, struct zenring *));
  int (*MatGetBloc) ___((ZENElt *, ZENMat, Dim, Dim, struct zenring *));
  ZENElt (*MatGetBlocPtr) ___((ZENMat, Dim, Dim, struct zenring *));
  int (*MatGetSubMat) ___((ZENMat, ZENMat, Dim, Dim, struct zenring *));
  int (*MatSetSubMat) ___((ZENMat, Dim, Dim, ZENMat, struct zenring *));

  int (*MatPermuteBn) ___((ZENMat, Dim, Dim, struct zenring *));
  int (*MatPermuteDigit) ___((ZENMat, Dim, Dim, struct zenring *));
  /* Test */
  int (*MatAreEqual) ___((ZENMat, ZENMat, struct zenring *));
  int (*MatIsZero) ___((ZENMat, struct zenring *));
  int (*MatIsOne) ___((ZENMat, struct zenring *));

  /* Arithmetic */

  int (*MatAdd) ___((ZENMat, ZENMat, struct zenring *));
  int (*MatAddScalar) ___((ZENMat, ZENElt, struct zenring *));
  int (*MatSubtract) ___((ZENMat, ZENMat, struct zenring *));
  int (*MatNegate) ___((ZENMat, ZENMat, struct zenring *));
  int (*MatMultiplyScalar) ___((ZENMat, ZENElt, struct zenring *));
  int (*MatMultiply) ___((ZENMat, ZENMat, ZENMat, struct zenring *));
  int (*MatMultiplyPlain) ___((ZENMat, ZENMat, ZENMat, struct zenring *));
  int (*MatWinograd) ___((ZENMat, ZENMat, ZENMat, struct zenring *));
  int (*MatMultiplyRightPermutation) ___((ZENMat, ZENMat, ZENMat, struct zenring *));
  int (*MatMultiplyLeftPermutation) ___((ZENMat, ZENMat, ZENMat, struct zenring *));
  int (*MatInverse) ___((ZENMat, ZENMat, struct zenring *));
  int (*MatKernel) ___((ZENMat *, ZENMat, struct zenring *));

  int (*MatGauss) ___((ZENMat, ZENMat, ZENMat, Dim *, ZENElt *, int (*) ___((ZENMat,int,int,struct zenring *)), struct zenring *));

  /* IO */
  int (*MatGetFromString) ___((ZENMat *, char *, struct zenring *));
  char * (*MatPutToString) ___((ZENMat, struct zenring *));
  int (*MatGetFromFile) ___((ZENMat *,FILE *,struct zenring *));
  int (*MatPutToFile) ___((FILE *, ZENMat, struct zenring *));

  /* Matrix conversion */
  int (*Vect2Mat) ___((ZENMat, ZENMat, struct zenring *));
  int (*Mat2Vect) ___((ZENMat, ZENMat, struct zenring *));
};


struct zen_debug {
  int (*ZENBP) ___((BigNum, BigNumLength));
  int (*ZENBR) ___((BigNum *, BigNumLength *));
  int (*ZENEP) ___((ZENElt, struct zenring *));
  int (*ZENFP) ___((ZENElt, struct zenring *));
  int (*ZENER) ___((ZENElt, struct zenring *));
  int (*ZENPP) ___((ZENPoly, struct zenring *));
  int (*ZENPR) ___((ZENPoly *, struct zenring *));
  int (*ZENMP) ___((ZENMat, struct zenring *));
  int (*ZENMR) ___((ZENMat *, struct zenring *));
  int (*ZENSP) ___((ZENSr, struct zenring *));
  int (*ZENSR) ___((ZENSr *, struct zenring *));
};

struct zemp_params {
  /* Buffer for
     _ZepEltNegate,
     _ZepEltSubtract */
  BigNum Nbuf;

  /* Buffer for
     _ZepEltMultiply */
  BigNum Mbuf; BigNumLength Mbufl;

  /* Buffers used in Karatsuba's operations */
  BigNum KaraBuf;
  BigNumLength KaraMLim,KaraSLim;  

  /* mask for _ZepSetRandom */
  BigNumDigit mask;
};

struct zep_params {
  struct zemp_params zemp;
  
  /* Buffer for
     _ZepEltInverseMisc */
  BigNum Ibuf;
};

struct zem_params {
  struct zemp_params zemp;
  ZENElt One;
  BigNum bn,bn2,bn3,V;
  int bnl,Vl;
};

typedef BigNumDigit ZEN_UL_2SC_Table[1<<(2*SIZE_CHAR)];
struct zeps_params {
  /* Buffer used in:
     ZepsEltAdd().
     */
  ZENElt Ebuf;

  /* Buffers used in:
     ZepsEltMultiplyPrcLargep().
     */
  ZEN_UL_2SC_Table *tab0, *tab1;
  ZEN_UL_2SC_Table *tab2, *tab3;
};

typedef unsigned char ZEN_UC_SC2_Table[1<<SIZE_CHAR][1<<SIZE_CHAR];
typedef unsigned char ZEN_UC_SC_Table[1<<SIZE_CHAR];

struct zetab_params {

  /* The arithmetic tables */
  ZEN_UC_SC2_Table *add, *mult;
  ZEN_UC_SC_Table *neg, *inv, *tr;
  ZENElt *conv;
};

typedef unsigned short ZEN_UC_2SC_Table[1<<(2*SIZE_CHAR)];
struct zelog_params {

  /* The arithmetic tables */
  ZEN_UC_2SC_Table *add, *log, *tr;
  ZENElt *conv;
  
  /* the generator of the ring */
  ZENElt Gen, MinusOne, Hbuf;
};

struct zext_params {
  /* buffers */
  ZENPoly Pbuf;
  ZENElt Rbuf;

  /* Precomputations for :
   * _ ZextEltTrace
   */
  ZENPoly Tbuf;

  /* Buffers for ZextEltToZ and ZextZToElt */
  BigNum tmp1, tmp2;

  /* Precomputations for :
   * _Ze2mEltMultiply
   * _Ze2mEltSquare
   */
  int nbbits;
  unsigned int *divfast;

  /* Precomputations for :
   * _ZextEltMultiplyRec
   * _ZextEltSquareRec
   */
  ZENPoly Rec;
};

struct zec_params {
  int nbsubrings;
  ZENElt *modinvs;
  struct zenring **subrings;
};

typedef struct zenring {

  struct zen_debug debug;
  
  /* The type of the polynomial ring F$(p^m)$ */
  arith arithmetic;

  /* The characteristic */
  BigNum p; BigNumLength pl;
  
  /* The size of elements */
  BigNumLength zenBnl, zenBnl2;

  /* The number of elements : $p^m$ */
  BigNum p2m; BigNumLength p2ml;

  /* p2m-1 */
  BigNum p_1;

  /* The polynomial defining the extension if needed */
  ZENPoly pol;

  /* The smallest $k$ such that pol=$X^m+X^k+$ smallest terms */
  int k;

  /* A factor of the modulo when ZEN_NO_INVERSE is encountered.
   */
  ZENElt fact;

  /* The prime ring if this ring is an extension */
  struct zenring *rg;

  /* Number of extensions defined over this field */
  unsigned int ext;

  /* The origin ring if this ring is a clone */
  struct zenring *cl;

  /* Number of clones defined from this ring */
  unsigned int nc;

  /* The variable to use in extension representation */
  int var;

  union {
    struct zeps_params zeps;
    struct zetab_params zetab;
    struct zelog_params zelog;
    struct zep_params zep;
    struct zext_params zext;
    struct zec_params zec;
    struct zem_params zem;
  } pms;

  struct zen_functions fct;

  /* Buffer used in:
     ZelogEltSubtract(),

     ZepsPolySquare(),
     ZepsPolyMultiply(),
     ZepsPolyDivideLazy(),
     ZepsPolyDot(),
     
     ZextPolySquareCharp(),
     ZextPolyMultiply(),
     ZextPolyDot(),
     ZextPolyDivideLazy(),
     ZextPolyModuloLazy(),

     ZextEltConvert(),

     ZENPolySquareCharp(),
     ZENPolyMultiply(),
     ZENPolyDot(),
     ZENPolyEval(),
     ZENPolyDivideLazy(),
     ZENPolyModuloLazy(),

     ZENMatVect2Mat(),
     
     ZENSrMultiply(),
     ZENSrSquareCharp().     
     */
  ZENElt Hbuf;

  /* Buffer used in:
     ZepsPolyDivideLazy(),
     ZextPolyDivideLazy(),

     ZecEltToZ(),
     ZecZToElt(),
     
     ZENEltCloneConvert(),

     ZENPolyDivideLazy(),
     ZENEltExp().
     */
  ZENElt Ibuf;

  /* Buffer used in:
     ZecCloneToElt,

     ZENEltTrace().

     ZENMatVect2Mat(),
     */
  ZENElt Jbuf;

  /* Buffer of size ZENRingQl(R) used in:
     ZENMatConvert()
     ZecEltToClone().
     */
  BigNum bnbuf;

  /* Buffers used with ZENPRC_ELT_EXP set in ZENEltExp */
  ZENElt *OddPower;
  int **ExpTableL,**ExpTableU;
  
  /* Precomputations already done in Rg */
  ZENPrc prc;

  /* An element of trace 1 over GF(p) if R is a finite field */
  ZENElt Theta;
  
} *ZENRing;


/* \TEX{ Built over a ring, an elliptic curve has type |ZENEc|.

   \begin{center}
   \begin{tabular}{p{0.25\textwidth}@{  }p{0.75\textwidth}}
   |ZENEc|& An elliptic curve of a finite ring. In |ZEN| any elliptic
   curve is a set of couples $(X,Y)$ such that
   \begin{displaymath}
   Y^2+a_1 XY+ a_3Y= X^3+a_2X^2+a_4X+a_6,
   \end{displaymath}
   plus a point at infinity.
   So such a structure contains the coefficients $a_1$, $a_2$, $a_3$,
   $a_4$ and $a_6$, its discriminant $D$ (it must be different from 0)
   and invariant $J$. Moreover, as in a |ZENRing| pointers on functions
   are stored here too to perform operations on elliptic curve
   points. So some other macros defined in |zen.h| replaced calls of
   procedures described in this documentation by calls to fields of
   this structure.
   \end{tabular}
   \end{center}
   
   And finally, a point of an elliptic curve has type |ZENEcPt|.

   \begin{center}
   \begin{tabular}{p{0.25\textwidth}@{  }p{0.75\textwidth}}
   |ZENEcPt|& Any point of an elliptic curve defined over any
   finite ring. It can be the infinity point or a couple $(X, Y)$.
   \end{tabular}
   \end{center}
  
 }*/
typedef struct zenecpt {
  ZENElt X, Y;
  int Z;
} *ZENEcPt;

typedef struct zenec {
  /* The definition ring */
  ZENRing rg;

  /* The coefficients of the curve */
  ZENElt a1, a2, a3, a4, a6;

  /* The coefficients of the canonical curve */
  ZENElt b1, b2, b3, b4, b6;

  /* The coefficients of the substitution */
  ZENElt Ad, Bd, Cd, Dd, Ed;

  /* The coefficients of the inverse substitution */
  ZENElt Ai, Bi, Ci, Di, Ei;

  /* Invariants */
  ZENElt Da, Db, J;

  /* buffers */
  ZENEcPt XY;
  ZENElt Z;
  
  /* Allocation */
  ZENEcPt (*PtCreate) ___((struct zenec *));        /* Create a point on an elliptic curve */
  void (*PtFree) ___((ZENEcPt, struct zenec *));             /* Free a point an elliptic curve */
  
  /* Assigning */
  void (*PtSetToZero) ___((ZENEcPt, struct zenec *));        /* Set a point on an elliptic curve to zero */
  int (*PtSetNext) ___((ZENEcPt, ZENEcPt, struct zenec *));           /* Set randomly a point on an elliptic curve */
  void (*PtAssign) ___((ZENEcPt, ZENEcPt, struct zenec *));           /* Set a point with another */
  void (*PtGetX) ___((ZENElt *, ZENEcPt, struct zenec *));             /* Get the abscissa of a point on an elliptic curve */
  void (*PtGetY) ___((ZENElt *, ZENEcPt, struct zenec *));             /* Get the ordinate of a point on an elliptic curve */
  void (*PtSetXY) ___((ZENEcPt, ZENElt, ZENElt, struct zenec *));		/* Set the abscissa and ordinate of a point on an elliptic curve */
  void (*PtGeneral2Canonical) ___((ZENEcPt, ZENEcPt, struct zenec *)); /* Compute a general point from a canonical */
  void (*PtCanonical2General) ___((ZENEcPt, ZENEcPt, struct zenec *)); /* Compute a canonical point from a general */

  /* Test */
  int (*PtIsZero) ___((ZENEcPt, struct zenec *));            /* Is a point equal to zero */
  int (*PtAreEqual) ___((ZENEcPt, ZENEcPt, struct zenec *));		/* Are two points equal */
  int (*PtIsOnEc) ___((ZENEcPt, struct zenec *));            /* Is a point on an elliptic curve */

  /* Arithmetic */
  void (*PtPreDouble) ___((ZENEcPt, struct zenec *));		/* Fisrt step when doubling 1 point on an elliptic curve */
  void (*PtPostDouble) ___((ZENEcPt, ZENEcPt, struct zenec *));	/* Last step when doubling 1 point on an elliptic curve */

  void (*PtPreAdd) ___((ZENEcPt, ZENEcPt, struct zenec *));		/* First step when adding 2 points on an elliptic curve */
  void (*PtPostAdd) ___((ZENEcPt, ZENEcPt, ZENEcPt, struct zenec *));		/* Last step when adding 2 points on an elliptic curve */

  void (*PtNegate) ___((ZENEcPt, ZENEcPt, struct zenec *));           /* Negation of a point on an elliptic curve */
} *ZENEc;


/* \TEX{
   \Subsubsection{Precomputations and clones}

  We implemented two ways to improve performances in |ZEN|,
  ``precomputations'' and ``clones''.

  Precomputations consist in computing data in advance in order to
  speed up some operations, typically multiplications or
  exponentiations. By default we chose to do no precomputation while
  initializing a ring because it can take a lot of time for large
  rings. But for some specific applications, this time is quite small
  in front of the remaining computations. This is why we provide the
  function |ZENAddPrc(Rg, prc)| which performs and stores the
  precomputations specified by the variable |prc| of type |ZENPrc| in
  a ring |Rg|.

  \begin{center}
  \begin{tabular}{p{0.25\textwidth}@{ }p{0.75\textwidth}}
  |ZENPrc|& Structure for handling precomputations. When working in a
  finite rings, some procedures can be speeded up at the expense (in
  time and space) of specific precomputations. So, once initialized a
  finite ring, you can in |ZEG| add or possibly suppress some
  precomputations in it. This structure enables the user to specify
  which precomputations he needs. Then, he only has to give it to the
  procedures which actually does the precomputations.
  \end{tabular}
  \end{center}
  
  } */

/* \TEX{

   Another way to improve performances is to work with another
   representation of a ring. The default representation is a
   polynomial representation. Elements in these rings are integers
   modulo another integer or polynomials modulo another
   polynomial. There exists numerous other representations. For
   instance, for small rings, you can tabulate everything and then
   handle indexes in tabulars, or compute a normal basis and
   then work in this basis. These representations are called clones in
   |ZEN|. You can compute a clone of any ring |Rg| with the function
   |ZENRingClone(Rg, cln)|. The type of the clone depends on the
   parameter |cln| which is of type |ZENCln|.

   \begin{center}
   \begin{tabular}{p{0.25\textwidth}@{  }p{0.75\textwidth}}
   |ZENCln|& Structure for handling clones. Finite rings can have
   several representation. The default is the polynomial
   representation but |ZEN| can handle other representations, for
   instance by tabulating the operations.
   \end{tabular}
   \end{center}
  } */


/* Some very useful macros */
#define CheckReturn(test,Err,ret)			\
if (test) {						\
  Err;							\
  ret;							\
} else

#define ZENCheckReturn(test,Err,ret) CheckReturn(test,Err,ret)

#define MallocAT(n,nl,type,Err,ret) \
CheckReturn(!(n = (type *)ZENMalloc((nl)*sizeof(type))),ZENSetError(ZEN_LIB_ZEG, ZEN_MALLOC, ZEN_ALLOC); Err,ret)
#define MallocFT(n) CheckReturn(n, ZENFree(n); n = NULL,;)

#define ZBNAT(n,nl,Err,ret) \
     CheckReturn(!(n = ZBNC(nl)),ZENSetError(ZEN_LIB_ZEG, ZEN_MALLOC, ZEN_ALLOC); Err,ret)
#define ZBNFT(n) CheckReturn(n, ZBNF(n); n = NULL,;)

#define ZENEltAT(elt,Rg,Err,ret) \
     ZENCheckReturn(ZENEltCreate(elt, Rg),Err,ret)
#define ZENEltFT(elt, Rg) ZENEltFree(elt, Rg)

#define ZENPolyAT(P,deg,Rg,Err,ret) \
     ZENCheckReturn(!(P = ZENPolyCreate(deg, Rg)),Err,ret)
#define ZENPolyFT(P, Rg) \
     CheckReturn(P, ZENPolyFree(P, Rg); P = NULL,;)

#define ZENEcAT(Ec,Rg,Err,ret) \
     ZENCheckReturn(!(Ec = ZENEcCreate(Rg)),Err,ret)
#define ZENEcFT(Ec) \
     CheckReturn(Ec, ZENEcFree(Ec); Ec = NULL,;)

#define ZENMatAT(M,l,c,Rg,Err,ret) \
     ZENCheckReturn(!(M = ZENMatCreate(l,c,Rg)),Err,ret)
#define ZENMatFT(M,Rg) \
     CheckReturn(M, ZENMatFree(M, Rg); M = NULL,;)

#define ZENEcPtAT(Pt,Ec,Err,ret) \
     ZENCheckReturn(!(Pt = ZENEcPtCreate(Ec)),Err,ret)
#define ZENEcPtFT(Pt, Ec) \
     CheckReturn(Pt, ZENEcPtFree(Pt, Ec); Pt = NULL,;)

#define ZENSrAT(P,deg,Rg,Err,ret) \
ZENCheckReturn(!(P = ZENSrCreate(deg, Rg)),Err,ret)
#define ZENSrFT(P, Rg) \
CheckReturn(P, ZENSrFree(P, Rg); P = NULL,;)

#endif

