/*
 * zen/inputs/example.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

				/* This is a small example of programming in ZEN */
/*
   \TEX{
   The purpose of this program is to perform inversion of some elements in some finite rings. This is a very simple example but it should be of some help for understanding the philosophy of \zen.

   \Subsection{Preliminaries}
   The first thing to do is to include some standard libraries header files.
   } 
   */

				/* \CiteC */
#include <stdlib.h>
#include <stdio.h>
#include "zen.h"		
/* \EndCiteC */

/*
   \TEX{
   We do not need any function in this program, due to its simplicity, but we begin our main function by the declarations we need.
   } 
   */

				/* \CiteC */
main()
{
  BigNum q,r;
  BigNumLength ql,rl;
  ZENElt A,I,E0,E1,EE0;
  ZENRing K,R,E,EE;
  ZENPoly P,Q,PR;		
/* \EndCiteC */

  /*
   \TEX{
   We do not need other objects in this example. That's why we don't have here any |ZENMat|, |ZENSr|, nor |ZENEc|. Nevertheless, the operations that use these types are similar in their syntax to the folowwing ones.

   \Subsection{Modular rings}
   Our example will first consist in building the finite field $\quotient{\Z}{234776683\Z}$. For this purpose we need a |BigNum| representing the modulo. We can for instance use the following function to this purpose.
   } 
   */
				/* \CiteC */
 ZBNReadFromString(&q,&ql,"234776683",10);
/* \EndCiteC */
 /*
   \TEX{
   In order to check the correctness of the implementation, we now print the obtained |BigNum|.
   } 
   */

				/* \CiteC */
 printf("q="); ZBNPrintToFile(stdout,q,ql,10); printf("\n");
/* \EndCiteC */

 /*
   \TEX{
   Let's see now how to build our finite field. The syntax of the function is rather simple, and affects to the |K| variable the structure we want.
   } 
   */

				/* \CiteC */
 ZENBaseRingAlloc(K,q,ql);	
/* \EndCiteC */

 /*
   \TEX{
   If we want to invert 234675 in this field, we will need two elements, one for the data, the other for the result. We first need to allocate them and as they are elements of |K|, the syntax we use is as follows.
   } 
   */

				/* \CiteC */
 ZENEltAlloc(A,K); ZENEltAlloc(I,K);
/* \EndCiteC */

 /*
   \TEX{
   Now, we can set the data and check the value.
   } 
   */

				/* \CiteC */
 ZENEltReadFromString(A,"234675",10,K);
 printf("A="); ZENEltPrintToFile(stdout,A,10,K); printf("\n");
/* \EndCiteC */

 /*
   \TEX{
   In order to obtain the result, we need a call to the suited function.
   } 
   */

				/* \CiteC */
 ZENEltInverse(I,A,K);
 printf("1/A mod q = ");
 ZENEltPrintToFile(stdout,I,10,K);
 printf("\n");
/* \EndCiteC */

 /*
   \TEX{
   As we no more need these values, we mustn't forget to free the memory they use. This point is important for larger programs that could grow indefinitely in memory without such precautions.
   } 
   */

				/* \CiteC */
 ZENEltFree(A,K); ZENEltFree(I,K); 
/* \EndCiteC */

 /*
   \TEX{
   We now take the invert of the same value, but considering it in another structure  $\quotient{\Z}{8745287453\Z}$. The fact is that this is not a field. Hence, in such a ring, we can no more ignore the return values of the inversion function, as this can be the signal of a mathematical incoherence.
   We continue to ignore the returned values of the first functions for simplicity purpose.
   } 
   */

				/* \CiteC */
 ZBNReadFromString(&r,&rl,"8745287453",10);
 printf("r="); ZBNPrintToFile(stdout,r,rl,10); printf("\n");
 ZENBaseRingAlloc(R,r,rl); 

 ZENEltAlloc(A,R); ZENEltReadFromString(A,"234675",10,R);
 printf("A="); ZENEltPrintToFile(stdout,A,10,R); printf("\n");
 ZENEltAlloc(I,R); 
/* \EndCiteC */

 /*
   \TEX{
   Now, we take care of the possible exception. Of course, as the example was chosen, we will find here a factor of the modulus.
   } 
   */

				/* \CiteC */
 if(ZENEltInverse(I,A,R)==ZEN_NO_INVERSE) {
   printf("Non invertible element : modulus factor = ");
   ZENEltPrintToFile(stdout,ZENRingFact(R),10,R); printf("\n"); }
 else {
   printf("1/A mod r = ");
   ZENEltPrintToFile(stdout,I,10,R);
   printf("\n"); } 
/* \EndCiteC */
/*
   \TEX{
   In fact, the set of possibly returned values is slightly larger, but it is unuseful to here detail all the other possibilities. We don't forget to free our variables before proceeding, including this ring.
   } 
   */

				/* \CiteC */
 ZENEltFree(A,R); ZENEltFree(I,R); ZENRingClose(R);
/* \EndCiteC */
  
/*
   \TEX{
   \Subsection{Extension rings}
   We will now work in a slightly more complex structure, namely
   $$
   \quotient{\quotient{\Z}{234776683\Z}[X]}{X^3+3234234X^2+234234X+124123}.
   $$
   This example is such that we have again a finite field.
   The definition of this new |ZENRing| has the same structure as before. We first define a modulus and use it to define the field.
   } 
   */
				/* \CiteC */
 ZENPolyReadFromString(P,
		       "(1)*X^3+(3234234)*X^2+(234234)*X+(124123)",
		       10,K);
 printf("P="); ZENPolyPrintToFile(stdout,P,10,K);
  
 ZENExtRingAlloc(E,P,K); ZENPolyFree(P,K);  
/* \EndCiteC */

 /*
   \TEX{
   Now, if we want to invert $234234t+3234234$ in this field, we just have to write the following.
   } 
   */
				/* \CiteC */
 ZENEltAlloc(E0,E);
 ZENEltReadFromString(E0,"(0)*t^2+(234234)*t+(3234234)",10,E);
 printf("E0="); ZENEltPrintToFile(stdout,E0,10,E); printf("\n");
/* \EndCiteC */
/*
   \TEX{
   The syntax of the strings used for polynomials could appear complex at this point, but the apparently redundant parenthesis will soon show their usefulness in towers of extensions.
   We now perform the inversion.
   } 
   */

				/* \CiteC */
 ZENEltAlloc(E1,E); ZENEltInverse(E1,E0,E);
 printf("1/E0 mod q = ");
 ZENEltPrintToFile(stdout,E1,10,E);
 printf("\n");
/* \EndCiteC */
/*
   \TEX{
   Of course, all this could appear heavy for such a simple operation, but the important point here is that all the strings used to define the parameters of this example could be defined dynamically in the program. In this case, the program is compiled once for all and can work for every structure.


   \Subsection{Double extension}

   We continue to increase complexity. We denote as follows the two structures we have just defined:
   \begin{eqnarray*}
   K &=& \quotient{\Z}{234776683\Z} \cr
   E &=& \quotient{K[t]}{t^3+3234234t^2+234234t+124123}
   \end{eqnarray*}
   and we now build the following structure:
   $$
   \quotient{E[X]}{X^2 + (124123 t^2  + 234234) X + (234234 t + 3234234)}
   $$
We will use here a construction of the polynomial that uses coefficients. It should be here more readable.
We first define the element $124123 t^2  + 234234$ over |E|, and as the element |E0| was kept, we can define our new modulus.
   } 
   */

				/* \CiteC */
 ZENEltReadFromString(E1,"(124123)*t^2+(234234)",10,E);
 printf("E1=");
 ZENEltPrintToFile(stdout,E1,10,E);
 printf("\n");
/* \EndCiteC */
/*
   \TEX{
   We now allocate a polynomial, of maximal degree 2, set it to $X^2$, and then set its coefficients. This procedure is the only guaranteed one to obtain the desired result. In particular, one should not omit to first set the polynomial to its monomial of highest degree, because this is the way to fix the degree of the polynomial. Setting the coefficients, on the contrary, is a fast procedure which does nothing but... setting a coefficient.
   } 
   */

				/* \CiteC */
 ZENPolyAlloc(Q,2,E);
 ZENPolySetToXi(Q,2,E);
 ZENPolySetCoeff(Q,1,E1,E); ZENPolySetCoeff(Q,0,E0,E);
 printf("Q=");
 ZENPolyPrintToFile(stdout,Q,10,E);
 printf("\n");
/* \EndCiteC */
/*
   \TEX{
   We can now build our double extension using the same function as before.
   } 
   */

				/* \CiteC */
 ZENExtRingAlloc(EE,Q,E); ZENPolyFree(Q,E);
/* \EndCiteC */

 /*
   \TEX{
   The following is the setting and inversion of an element in this double extension.
   } 
   */

				/* \CiteC */
 ZENEltAlloc(EE0,EE); ZENPolySetToXi(ZENElt2Pol(EE0,EE),1,E);
 ZENPolySetCoeff(ZENElt2Pol(EE0,EE),1,E1,E);
 ZENPolySetCoeff(ZENElt2Pol(EE0,EE),0,E0,E);
 printf("EE0=");
 ZENEltPrintToFile(stdout,EE0,10,EE);
 printf("\n");
  
 ZENEltFree(E0,E); ZENEltFree(E1,E);
  
 ZENEltAlloc(E1,EE); ZENEltInverse(E1,EE0,EE);
 printf("1/EE0=");
 ZENEltPrintToFile(stdout,E1,10,EE);
 printf("\n");

 ZENEltFree(E1,EE); 
 ZENRingClose(EE); ZENRingClose(E); ZENRingClose(K);
 exit(0);
}
/* \EndCiteC */

#if 0
/*
   \TEX{
   \Subsection{Compilation}
   The compilation of this program will be done by a command like the following one
   |gcc -I../include example.c ../lib/linux/libzen.a -lm|, as the file |example.c| is located in the |zen/inputs| directory.
   The obtained output of the |a.out| executable should then be the following.
   } 
   */
				/* \CiteC */
q=234776683
A=234675
1/A mod q = 11532995
r=8745287453
A=234675
Non invertible element : modulus factor = 7
P=(1)*X^3+(3234234)*X^2+(234234)*X+(124123)E0=(0)*t^2+(234234)*t+(3234234)
1/E0 mod q = (226257462)*t^2+(25554697)*t+(70525059)
E1=(124123)*t^2+(234234)
Q=((0)*t^2+(1))*X^2+((124123)*t^2+(234234))*X+((0)*t^2+(234234)*t+(3234234))
EE0=((124123)*t^2+(234234))*u+((0)*t^2+(234234)*t+(3234234))
1/EE0=((203916487)*t^2+(83557476)*t+(113630396))*u+ \
      ((138383902)*t^2+(180392990)*t+(206406567))
/* \EndCiteC */


#endif
