/*-*-LaTeX-*-
 * zen/inputs/user.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   This is a documentation file of the ZEN library.
   There is no C code in this file, only LaTeX commands.
   
   
   \TEX{
     
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Beginning with ZEN %%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     
     
     \Section{The main principles}
     
     In this section, we describe the main concepts of \zen. Mainly
     the objects it is supposed to handle, the functions working on
     these objects, the way how precomputations can be performed or
     what is a clone of a ring to increase performances.

     \Subsection{The types of \zen}
     \input{zentypes}
     
     \Subsection{The functions of \zen}
     \glossary{rings@{\Large\bf rings}}
     \Subsubsection{Modular rings}
     As we work in finite rings, we must begin by building a modular
     ring.  This is done by using the |BigNum| representation of an
     integer and a call to the |ZENBaseRingAlloc(R,n,nl)| procedure.
     It allocates a |ZENRing R;| and initializes its parameters.

     On success it returns 0 and a non zero exception flag otherwise.
     We will see later (section~\ref{sc:error}) how to use the
     informations of this flag.  This is the normal behavior of all
     |ZEN| functions with the exception of the |void| ones.  In fact,
     if you know that your program should never raise an exception,
     you can simply not use the return values because in debug
     mode, the library will not remain silent in case of exceptions.

     Now, if we want to use objects on this ring we just have for
     instance to allocate a |ZENElt E|: |ZENEltAlloc(E,R)|.  Some
     simple functions with self-contained names allow to initialize
     the element:
     \begin{itemize}
     \item |ZENEltSetToZero(E,R)|.
     \item |ZENEltSetToOne(E,R)|.
     \item |ZENZToElt(E,n,nl,R)| gives the value of the |BigNum| |(n,nl)| to |E|.
     \item |ZENEltReadFromString(E,s,base,R)| reads the string |s| in
       base |base| and affects |E| accordingly.
     \item |ZENEltReadFromFile(E,fd,base,R)| reads the |FILE *fd| in
       base |base| and affects |E| accordingly.
     \end{itemize}
     At the end of computation, one should liberate the memory
     allocated for this element with the function |ZENEltFree(E,R)|.
     The figure~\ref{fig:zen} resume the sequence of operations in a
     modular ring.

\begin{note}
It is possible to build the rational field $\Q$ by using a |BigNum| $|(n,nl)|=0$. This feature is still experimental, and the efficiency is not guarenteed.
\end{note}


     \begin{figure}
       \begin{center}
         \begin{pspicture*}(-1.5,-15)(15,0)
           \rput(2,-1){\sc Sequence of operations}
           \rput(10,-1){\sc Example of function}

           \psframe[shadow=true](-1,-3)(5,-2)
           \rput(2,-2.5){Initialization of a modulus}
           \psframe(6,-3.5)(14,-1.5)
           \rput(10,-2){\ZENF{\begin{tabular}{l}
                 BigNum n; BigNumLength nl; \\
                 \multicolumn{1}{@{\hspace*{5mm}}l}{ZBNReadFromString(\&n,\&nl,} \\
                 \multicolumn{1}{@{\hspace*{10mm}}l}{"4951760157141521099596496921",10);} \\
               \end{tabular} }}
           
           \psline{->}(2,-3)(2,-4)
           
           \psframe[shadow=true](-1,-5)(5,-4)
           \rput(2,-4.5){Initialization of the ring}
           \psframe(6,-5)(14,-4)
           \rput(10,-4.5){\ZENF{\begin{tabular}{l}
                 ZENRing R; \\
                 \multicolumn{1}{@{\hspace*{1cm}}l}{ZENBaseRingAlloc(R,n,nl);} \\
               \end{tabular} }}
           
           \psline{->}(2,-5)(2,-6)
           
           \psframe[shadow=true](-1,-7)(5,-6)
           \rput(2,-6.5){Creation of an element}
           \psframe(6,-7)(14,-6)
           \rput(10,-6.5){\ZENF{\begin{tabular}{l}
                 ZENElt E; \\
                 \multicolumn{1}{@{\hspace*{1cm}}l}{ZENEltAlloc(E,R);} \\
               \end{tabular} }}
           
           \psline{->}(2,-7)(2,-8)
           
           \psframe[shadow=true](-1,-9)(5,-8)
           \rput(2,-8.5){Sequence or arithmetic operations}
           \psframe(6,-9)(14,-8)
           \rput(10,-8.5){\ZENF{\begin{tabular}{l}
                 ZENEltSetToOne(E,R); \\
                 ZENEltAdd(E,E,R); \\
               \end{tabular} }}
           
           \psline{->}(2,-9)(2,-10)
           
           \psframe[shadow=true](-1,-11)(5,-10)
           \rput(2,-10.5){Output of the results}
           \psframe(6,-11)(14,-10)
           \rput(10,-10.5){\ZENF{ZENEltPrintToFile(stdout,E,10,R);}}
           
           \psline{->}(2,-11)(2,-12)
           
           \psframe[shadow=true](-1,-13)(5,-12)
           \rput(2,-12.5){Memory freeing}
           \psframe(6,-13.5)(14,-11.5)
           \rput(10,-12.5){\ZENF{\begin{tabular}{l}
                 ZENEltFree(E,R); \\
                 ZENRingClose(R); \\
                 ZBNF(n); \\
               \end{tabular} }}
         \end{pspicture*}
       \end{center}
       \caption{Sequence of operations in a modular ring.}
       \label{fig:zen}
     \end{figure}

     \Subsubsection{Polynomial extensions}

     The main advantage of |ZEN| is that polynomial extensions are
     very easy to build.  As soon as a |ZENPoly P| is defined with the
     correct value, an extension |ZENRing R2| can be created by the
     command |ZENExtRingAlloc(R2,P,R)|.  This can be repeated as long
     as wanted with the same syntax.  Two elements in a ring |Rx| are
     always added by the function |ZENEltAdd(A,B,Rx)|, whatever the
     ring is.  Of course, for efficiency reasons, it will be faster to
     define a field like $\GF{16}$ as an extension over $\GF{2}$ of
     degree 4 instead of a double extension of degree 2, even if these
     two constructions are isomorphic.

     \Subsubsection{The syntax of \zen}
     \glossary{syntax@{\Large\bf syntax}} These small examples show the
     syntax principles of |ZEN| that are resumed in
     figure~\ref{fig:syntax}

     \begin{figure}
       \begin{center}
         \begin{tabular}{@{$\vert$}ccc@{$\vert$}ccccc@{$\vert$}}\hline
           \multicolumn{3}{@{$\vert$}c@{$\vert$}}{Types} && Modified  & Parameters & Ring & \\ \cline{1-3}
           of arithmetic & of object & of operation && object & & used & \\ \hline
           ZEN&Elt&Multiply&(&X,&A, B,&R&)  \\ \hline
         \end{tabular}
       \end{center}
       \caption{|ZEN| syntax}
       \label{fig:syntax}
     \end{figure}
     \glossary{syntax@{\Large\bf syntax}}

     \Subsection{Error handling}\label{sc:error}
     \input{sys}

     \Section{Optimization}
     \glossary{optimization@{\Large\bf optimization}}

     In most cases, the above functions will be sufficient to obtain
     good performances.  Nevertheless, some applications will need
     much more efficiency, and the following is a way to achieve quite
     good improvements.

     \Subsection{Precomputations}

     Another principle of |ZEN| is to perform only what is asked for.
     Therefore, on a ring creation, no precomputation is done.  But
     some precomputations can be asked for by using a precomputation
     structure |ZENPrc Prc|.  For instance, |ZENPrcSetAll(Prc)|
     activates all the precomputations.  A call to
     |ZENRingAddPrc(R,Prc)| will then perform the precomputations that
     are compatible with the type of ring.  The available flags are
     described in section~\ref{sc:prc}.

     \Subsection{Clones}

     They are several way to represent a ring and this fact yields
     several computer representations. To deal with these numerous
     representations, the user can compute a ``clone'' of a ring with
     the function |ZENRingClone()|. Unfortunataly, only a few clones
     are now available.

     \Subsubsection{Tabulating}
     \input{zetabdoc}
     
     \Subsubsection{Use of logarithms}
     \input{zelogdoc}
     
     \Subsubsection{Use of chinese remainder theorem}
     \input{zecdoc}
     
     \Subsubsection{Use of Montgomery's representation}
     \input{zemdoc}
     
% LocalWords:  ftp tgz gunzip gzip gz xvf cd imakefiles zenmkmf sh doc libzen
% LocalWords:  lib zen dvi adoc cccc sparc SunOS cc gcc Solaris vax mips OSF UX
% LocalWords:  RS AIX linux NetBsd FreeBsd specif def imake Iconfig linestyle
% LocalWords:  txt precomputations BigNum ZENBaseRingAlloc nl ZENRing ZENElt fd
% LocalWords:  ZENEltAlloc ZENEltSetToZero ZENEltSetToOne ZENZToElt ZENEltFree
% LocalWords:  ZENEltReadFromString ZENEltReadFromFile int ZBNReadFromString Rx
% LocalWords:  ZENEltAdd ZENEltPrintToFile stdout ZENRingClose ZBNF ZENPoly
% LocalWords:  ZENExtRingAlloc
}
*/

