/*
 * zen/prgm/c/ec_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Section{Tests on elliptic curves}
   
   These functions are generic tests for points on elliptic curves
   operations.
   The main function is |ZENEcTest()| that calls the other procedures
   of this section.
   }
   */
#include <stdio.h>
#include <ctype.h>
#include <zen.h>
#include <test.h>


/* \TEX{
   \title{Test on elliptic curves.}
   \input{A finite ring |Rg|, the parameters defined in |zentest.c| and the extension level.}
   \output{0 if all the tests were ok, |ZENERR| otherwise.}
   \side{All the following test functions on |ZENEcPt| are invoked.}
   }
   */
int ZENTestEc(Rg, cst, lvl)
     ZENRing Rg;
     int **cst, lvl;
{
  ZENElt a1, a2, a3, a4, a6;
  ZENEc Ec;

  if (ZENEcAlloc(Ec,Rg)) {
    return(ZERR);
  }
  if (ZENEltCreate(a1, Rg)) {
    ZENEcFree(Ec);
    return(ZERR);
  }
  if (ZENEltCreate(a2, Rg)) {
    ZENEcFree(Ec);
    ZENEltFree(a1, Rg);
    return(ZERR);
  }
  if (ZENEltCreate(a3, Rg)) {
    ZENEcFree(Ec);
    ZENEltFree(a1, Rg);
    ZENEltFree(a2, Rg);
    return(ZERR);
  }
  if (ZENEltCreate(a4, Rg)) {
    ZENEcFree(Ec);
    ZENEltFree(a1, Rg);
    ZENEltFree(a2, Rg);
    ZENEltFree(a3, Rg);
    return(ZERR);
  }
  if (ZENEltCreate(a6, Rg)) {
    ZENEcFree(Ec);
    ZENEltFree(a1, Rg);
    ZENEltFree(a2, Rg);
    ZENEltFree(a3, Rg);
    ZENEltFree(a4, Rg);
    return(ZERR);
  }

  do {
#if 1
    ZENEltSetRandom(a1, Rg);
    ZENEltSetRandom(a2, Rg);
    ZENEltSetRandom(a3, Rg);
    ZENEltSetRandom(a4, Rg);
    ZENEltSetRandom(a6, Rg);
#else
    ZENEltReadFromString(a1, "(0)", 10, Rg);
    ZENEltReadFromString(a2, "(1)*t^2+(1)*t+(1)", 10, Rg);
    ZENEltReadFromString(a3, "(1)*t+(1)", 10, Rg);
    ZENEltReadFromString(a4, "(1)*t^2", 10, Rg);
    ZENEltReadFromString(a6, "(1)*t^3+(1)", 10, Rg);
#endif
    if (ZENEcInitialize(Ec, a1, a2, a3, a4, a6) != ZEN_HAS_INVERSE) {
      ZENEcFree(Ec);
      ZENEltFree(a1, Rg);
      ZENEltFree(a2, Rg);
      ZENEltFree(a3, Rg);
      ZENEltFree(a4, Rg);
      return(ZERR);
    }
  } while (ZENEltIsZero(ZENEcD(Ec), Rg) || ZENEltIsZero(ZENEcJ(Ec), Rg));
  
  ZENEltFree(a1, Rg);
  ZENEltFree(a2, Rg);
  ZENEltFree(a3, Rg);
  ZENEltFree(a4, Rg);
  ZENEltFree(a6, Rg);

  if (ZENTestEcPtAlloc(Ec, cst[NB_ZEN_EC_ALLOC][lvl])) {
    return(ZERR);
  }
  if (ZENTestEcPtSetToZero(Ec, cst[NB_ZEN_EC_TEST][lvl])) {
    return(ZERR);
  }
  if (ZENTestEcPtSetNext(Ec, cst[NB_ZEN_EC_ADD][lvl])) {
    return(ZERR);
  }
  if (ZENTestEcPtAdd(Ec, cst[NB_ZEN_EC_ADD][lvl])) {
    return(ZERR);
  }
  if (ZENTestEcPtNegate(Ec, cst[NB_ZEN_EC_NEGATE][lvl])) {
    return(ZERR);
  }
  if (ZENTestEcPtSubtract(Ec, cst[NB_ZEN_EC_SUBTRACT][lvl])) {
    return(ZERR);
  }
  if (ZENTestEcPtMult(Ec, cst[NB_ZEN_EC_MULTIPLY][lvl])) {
    return(ZERR);
  }
  if (ZENTestEcPtPrintToString(Ec, cst[NB_ZEN_EC_PR_STRING][lvl])) {
    return(ZERR);
  }
  if (ZENTestEcPtPrintToFile(Ec, cst[NB_ZEN_EC_PR_FILE][lvl])) {
    return(ZERR);
  }
  if (ZENTestEcPtPutToString(Ec, cst[NB_ZEN_EC_PG_STRING][lvl])) {
    return(ZERR);
  }  
  if (ZENTestEcPtPutToFile(Ec, cst[NB_ZEN_EC_PG_FILE][lvl])) {
    return(ZERR);
  }
  ZENEcFree(Ec);
  return(0);
}

/*\TEX{

  All the following functions take as input a |ZENEc| and the number of tests to perform. The output is |ZENERR| if an error occured and 0 otherwise. We therefore forget in the following these informations.
  
  \title{Allocation test.}
  \note{This test perform many calls to |Malloc|() and |Free|() procedures.}
  } 
*/
int ZENTestEcPtAlloc(Ec, nb)
     ZENEc Ec;
     int nb;
{
  ZENEcPt a;
  int i;
  double tt;

  /* ZENEcPtAlloc() and ZENEcPtFree() */
  printf("\nTesting ZENEcPtAlloc() and ZENEcPtFree()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEcPtAlloc(a,Ec)) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENEcPtFree(a, Ec);
    printf(".");fflush(stdout);
  }
  printf("\nZENEcPtAlloc() and ZENEcPtFree() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Assignment to zero test.}
  \note{This procedure performs calls to |ZENEcPtSetToZero|() and
  |ZENEcPtIsZero|().}
  } 
*/
int ZENTestEcPtSetToZero(Ec, nb)
     ZENEc Ec;
     int nb;
{
  ZENEcPt a;
  int i;
  double tt;

  /* ZENEcPtSetToZero() and ZENEcPtIsZero() */
  printf("\nTesting ZENEcPtSetToZero() and ZENEcPtIsZero()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (!(a = ZENEcPtCreate(Ec))) {
      ZENError();
      printf("\n"); printf("ZENEcPtCreate() failed\n");
      return(ZERR);
    }

    ZENEcPtSetToZero(a, Ec);
    switch(ZENEcPtIsZero(a, Ec)) {
    case 1:
      break;
    default:
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENEcPtFree(a, Ec);
    printf(".");fflush(stdout);
  }
  printf("\nZENEcPtSetToZero() and ZENEcPtIsZero() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Random assignment and equality test.}
  \note{A |ZENEcPt| $a$ is ramdomly set and then assigned in an other
  |ZENEcPt| $b$. The test is ok if $a=b$.}
  } 
*/
int ZENTestEcPtSetNext(Ec, nb)
     ZENEc Ec;
     int nb;
{
  ZENElt elt;
  ZENEcPt a;
  int i;
  double tt;

  /* ZENEcPtSetNext() and ZENEcPtAssign() */
  printf("\nTesting ZENEcPtSetNext() and ZENEcPtIsOnEc()\n");
  if (ZENEltCreate(elt, ZENEcRing(Ec))) {
    return(ZERR);
  }
  tt = runtime();
  if (!(a = ZENEcPtCreate(Ec))) {
    ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  ZENEltSetRandom(elt, ZENEcRing(Ec));
  ZENEcPtSetXY(a, elt, elt, Ec);
  ZENEltFree(elt, ZENEcRing(Ec));
  for(i = 0; i < nb; i++) {
    ZENEcPtSetNext(a, a, Ec);
    
    if (!ZENEcPtIsOnEc(a, Ec)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    printf(".");fflush(stdout);
  }
  ZENEcPtFree(a, Ec);
  printf("\nZENEcPtSetNext()  and ZENEcPtIsOnEc() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  
  return(0);
}

/*\TEX{
  \title{Addition test.}
  \note{The test is $a+b = b+a$.}
  } 
*/
int ZENTestEcPtAdd(Ec, nb)
     ZENEc Ec;
     int nb;
{
  ZENElt elt;
  ZENEcPt a, b;
  int i;
  double tt;

  /* ZENEcPtAdd() */
  printf("\nTesting ZENEcPtAdd()\n");
  if (ZENEltCreate(elt, ZENEcRing(Ec))) {
    return(ZERR);
  }
  tt = runtime();

  if (!(a = ZENEcPtCreate(Ec))) {
    ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  ZENEltSetRandom(elt, ZENEcRing(Ec));
  ZENEcPtSetXY(a, elt, elt, Ec);
  if (nb) ZENEcPtSetNext(a, a, Ec);
  
  if (!(b = ZENEcPtCreate(Ec))) {
    ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  ZENEltSetRandom(elt, ZENEcRing(Ec));
  ZENEcPtSetXY(b, elt, elt, Ec);
  if (nb) ZENEcPtSetNext(b, b, Ec);
  
  for(i = 0; i < nb; i++) {
    switch(ZENEcPtAdd(b, a, b, Ec)) {
    case ZEN_NO_INVERSE:
      if (ZENTestEltNOINVCheck(ZENEcRing(Ec))) {
	printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
      }
      else break;
    case ZENERR:
      printf("\n"); 
      ZENError();
      printf("ZENEltAdd() failed\n");
      return(ZERR);
    case ZEN_HAS_INVERSE:

      if (!ZENEcPtIsOnEc(b, Ec)) {
	printf("The %dth test failed\n", i+1);
	return(ZERR);      
      }
      
      switch(ZENEcPtAdd(a, a, a, Ec)) {
      case ZEN_NO_INVERSE:
	if (ZENTestEltNOINVCheck(ZENEcRing(Ec))) {
	  printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
	}
	else break;
      case ZENERR:
	printf("\n"); 
	ZENError();
	printf("ZENEltAdd() failed\n");
	return(ZERR);
      case ZEN_HAS_INVERSE:
	if (!ZENEcPtIsOnEc(a, Ec)) {
	  printf("The %dth test failed\n", i+1);
	  return(ZERR);      
	}
      }
    }

    printf(".");fflush(stdout);
  }
  ZENEcPtFree(a, Ec);
  ZENEcPtFree(b, Ec);
  printf("\nZENEcPtAdd() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  ZENEltFree(elt, ZENEcRing(Ec));
  return(0);
}
/*\TEX{
  \title{Negation test.}
  \note{The test is $a+(-a) = 0$.}
  } 
*/
int ZENTestEcPtNegate(Ec, nb)
     ZENEc Ec;
     int nb;
{
  ZENElt elt;
  ZENEcPt a, b;
  int i;
  double tt;

  /* ZENEcPtNegate() */
  printf("\nTesting ZENEcPtNegate()\n");
  if (ZENEltCreate(elt, ZENEcRing(Ec))) {
    return(ZERR);
  }
  tt = runtime();
  if (!(a = ZENEcPtCreate(Ec))) {
    ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  ZENEltSetRandom(elt, ZENEcRing(Ec));
  ZENEcPtSetXY(a, elt, elt, Ec);
  if (nb) ZENEcPtSetNext(a, a, Ec);  

  if (!(b = ZENEcPtCreate(Ec))) {
    ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }

  for(i = 0; i < nb; i++) {

    ZENEcPtAssign(b, a, Ec);
    ZENEcPtNegate(b, b, Ec);
    switch (ZENEcPtAdd(a, b, a, Ec))  {
    case ZEN_NO_INVERSE:
      if (ZENTestEltNOINVCheck(ZENEcRing(Ec))) {
	printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
      }
      else break;
    case ZENERR:
      printf("\n"); 
      ZENError();
      printf("ZENEcPtMultiply() failed\n");
      return(ZERR);
    case ZEN_HAS_INVERSE:
	
      if (!ZENEcPtIsZero(a, Ec)) {
	printf("The %dth test failed\n", i+1);
	return(ZERR);      
      }
    }
    ZENEcPtDouble(b, a, Ec);
    printf(".");fflush(stdout);
  }
    
  ZENEcPtFree(a, Ec);
  ZENEcPtFree(b, Ec);
  printf("\nZENEcPtNegate() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  ZENEltFree(elt, ZENEcRing(Ec));
  return(0);
}
/*\TEX{
  \title{Subtraction test.}
  \note{The test is $a-a=0$.}
  } 
*/
int ZENTestEcPtSubtract(Ec, nb)
     ZENEc Ec;
     int nb;
{
  ZENElt elt;
  ZENEcPt a, b;
  int i;
  double tt;

  /* ZENEcPtSubtract() */
  printf("\nTesting ZENEcPtSubtract()\n");
  if (ZENEltCreate(elt, ZENEcRing(Ec))) {
    return(ZERR);
  }
  tt = runtime();
  if (!(a = ZENEcPtCreate(Ec))) {
    ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  ZENEltSetRandom(elt, ZENEcRing(Ec));
  ZENEcPtSetXY(a, elt, elt, Ec);
  if (nb) ZENEcPtSetNext(a, a, Ec);

  if (!(b = ZENEcPtCreate(Ec))) {
    ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  
  for(i = 0; i < nb; i++) {

    ZENEcPtAssign(b, a, Ec);
    switch(ZENEcPtSubtract(b, a, b, Ec)) {
    case ZEN_NO_INVERSE:
      if (ZENTestEltNOINVCheck(ZENEcRing(Ec))) {
	printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
      }
      else break;
    case ZENERR:
      printf("\n"); 
      ZENError();
      printf("ZENEcPtMultiply() failed\n");
      return(ZERR);
    case ZEN_HAS_INVERSE:
      if (!ZENEcPtIsZero(b, Ec)) {
	printf("The %dth test failed\n", i+1);
	return(ZERR);      
      }
    }
    ZENEcPtDouble(a, a, Ec);
    printf(".");fflush(stdout);
  }
  ZENEcPtFree(a, Ec);
  ZENEcPtFree(b, Ec);
  printf("\nZENEcPtSubtract() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  ZENEltFree(elt, ZENEcRing(Ec));
  return(0);
}


/*\TEX{
  \title{Multiplication test.}
  \note{The test is $m\times(n\times P) = m\times n\times P$.}
  } 
*/
int ZENTestEcPtMult(Ec, nb)
     ZENEc Ec;
     int nb;
{
  ZENElt elt;
  ZENEcPt a, b, c;
  int i;
  double tt;
  BigNum n, m, mn;
  BigNumLength nl, ml, mnl;

  /* ZENEcPtAdd() */
  printf("\nTesting ZENEcPtMultiply()\n");
  if (ZENEltCreate(elt, ZENEcRing(Ec))) {
    return(ZERR);
  }
  tt = runtime();

  if (!(a = ZENEcPtCreate(Ec))) {
    ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  ZENEltSetRandom(elt, ZENEcRing(Ec));
  ZENEcPtSetXY(a, elt, elt, Ec);
  if (nb) ZENEcPtSetNext(a, a, Ec);
  
  if (!(b = ZENEcPtCreate(Ec))) {
    ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  if (!(c = ZENEcPtCreate(Ec))) {
    ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  
  if (!(n = ZBNC(ZENRingQl(ZENEcRing(Ec))+1))) {
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  if (!(m = ZBNC(ZENRingQl(ZENEcRing(Ec))+1))) {
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
    
  }
  if (!(mn = ZBNC(2*ZENRingQl(ZENEcRing(Ec))+2))) {
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }

  for(i = 0; i < nb; i++) {
    ZENEltSetRandom(elt, ZENEcRing(Ec));
    ZENEltToZ(n, &nl, elt, ZENEcRing(Ec));

    ZENEltSetRandom(elt, ZENEcRing(Ec));
    ZENEltToZ(m, &ml, elt, ZENEcRing(Ec));

    ZBNSetToZero(mn, ml+nl+1);
    ZBNMultiply(mn, ml+nl+1, m, ml, n, nl);
    mnl = ZBNNumDigits(mn, ml+nl+1);

    switch(ZENEcPtMult(b, m, ml, a, Ec)) {
    case ZEN_NO_INVERSE:
      if (ZENTestEltNOINVCheck(ZENEcRing(Ec))) {
	printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
      }
      else break;
    case ZENERR:
      printf("\n"); 
      ZENError();
      printf("ZENEcPtMult() failed\n");
      return(ZERR);
    case ZEN_HAS_INVERSE:

      if (!ZENEcPtIsOnEc(b, Ec)) {
	printf("The %dth test failed\n", i+1);
	return(ZERR);      
      }
      
      switch(ZENEcPtMult(b, n, nl, b, Ec)) {
      case ZEN_NO_INVERSE:
	if (ZENTestEltNOINVCheck(ZENEcRing(Ec))) {
	  printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
	}
	else break;
      case ZENERR:
	printf("\n"); 
	ZENError();
	printf("ZENEcPtMult() failed\n");
	return(ZERR);
      case ZEN_HAS_INVERSE:
	if (!ZENEcPtIsOnEc(b, Ec)) {
	  printf("The %dth test failed\n", i+1);
	  return(ZERR);      
	}

	switch(ZENEcPtMult(c, mn, mnl, a, Ec)) {
	case ZEN_NO_INVERSE:
	  if (ZENTestEltNOINVCheck(ZENEcRing(Ec))) {
	    printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
	  }
	  else break;
	case ZENERR:
	  printf("\n"); 
	  ZENError();
	  printf("ZENEcPtMult() failed\n");
	  return(ZERR);
	case ZEN_HAS_INVERSE:
	  if (!ZENEcPtIsOnEc(c, Ec)) {
	    printf("The %dth test failed\n", i+1);
	    return(ZERR);      
	  }
	  if (!ZENEcPtAreEqual(b, c, Ec)) {
	    printf("The %dth test failed\n", i+1);
	    return(ZERR);      
	  }
	}
      }
    }

    printf(".");fflush(stdout);
  }
  ZENEcPtFree(a, Ec); ZENEcPtFree(b, Ec); ZENEcPtFree(c, Ec);
  ZBNF(m); ZBNF(n); ZBNF(mn);
  printf("\nZENEcPtAdd() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  ZENEltFree(elt, ZENEcRing(Ec));
  return(0);
}

/*\TEX{
  \title{Print/Read with string test.}
  \note{The test is $a=|ZENEcPtReadFromString|(|ZENEcPtPrintToString|(a))$.}
  } 
*/
int ZENTestEcPtPrintToString(Ec, nb)
     ZENEc Ec;
     int nb;
{
  ZENElt elt;
  ZENEcPt a, b, c;
  int i, j, base1, base2;
  double tt;
  char *str, *str1, *str2;

  /* ZENEcPtPrintToString() and ZENEcPtReadFromString() */
  printf("\nTesting ZENEcPtPrintToString() and ZENEcPtReadFromString()\n");
  if (ZENEltCreate(elt, ZENEcRing(Ec))) {
    return(ZERR);
  }
  tt = runtime();
  if (!(a = ZENEcPtCreate(Ec))) {
    printf("\n"); ZENError();
    printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  ZENEltSetRandom(elt, ZENEcRing(Ec));
  ZENEcPtSetXY(a, elt, elt, Ec);
  if (nb) ZENEcPtSetNext(a, a, Ec);

  if (!(b = ZENEcPtCreate(Ec))) {
    printf("\n"); ZENError();
    printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  if (nb) ZENEcPtDouble(b, a, Ec);
  
  if (!(c = ZENEcPtCreate(Ec))) {
    printf("\n"); ZENError();
    printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }

  for(i = 0; i < nb; i++) {
    ZENEcPtDouble(a, a, Ec);
    base1 = 2+(zbnrandom()%15);
    if (!(str1 = ZENEcPtPrintToString(a, base1, Ec))) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    ZENEcPtAdd(b, b, a, Ec);
    base2 = 2+(zbnrandom()%15);
    if (!(str2 = ZENEcPtPrintToString(b, base2, Ec))) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str1)+strlen(str2)+2)))) {
      fprintf(stderr, "ZENTestEcPt(): malloc(%d) error.\n",
	      strlen(str1)+strlen(str2)+2);
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!strcpy(str, str1) || !strcat(str, " ") || !strcat(str, str2)) {
      fprintf(stderr, "ZENTestEcPt(): strcpy or strcat error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENFree(str1); ZENFree(str2);

    if ((j = ZENEcPtReadFromString(c, str, base1, Ec)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEcPtAreEqual(c, a, Ec)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    if ((ZENEcPtReadFromString(c, str+j, base2, Ec)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEcPtAreEqual(c, b, Ec)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }


    ZENFree(str);

    printf(".");fflush(stdout);
  }
  ZENEcPtFree(c, Ec);
  ZENEcPtFree(b, Ec);
  ZENEcPtFree(a, Ec);
  printf("\nZENEcPtPrintToString() and ZENEcPtReadFromString() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  ZENEltFree(elt, ZENEcRing(Ec));
  return(0);
}
/*\TEX{
  \title{Print/Read with file test.}
  \note{The test is $a=|ZENEcPtReadFromFile|(|ZENEcPtPrintToFile|(a))$.}
  } 
*/
int ZENTestEcPtPrintToFile(Ec, nb)
     ZENEc Ec;
     int nb;
{
  ZENElt elt;
  FILE *fd;
  ZENEcPt a, b, c;
  int i, base1, base2;
  double tt;

  /* ZENEcPtPrintToFile() and ZENEcPtReadFromFile() */
  printf("\nTesting ZENEcPtPrintToFile() and ZENEcPtReadFromFile()\n");
  if (ZENEltCreate(elt, ZENEcRing(Ec))) {
    return(ZERR);
  }
  tt = runtime();

  if (!(fd = tmpfile())) {
    fprintf(stderr, "ZENTestEcPt(): tmpfile() error\n");
    printf("The test failed\n");
    return(ZERR);
  }
  
  if (!(a = ZENEcPtCreate(Ec))) {
    printf("\n"); ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  ZENEltSetRandom(elt, ZENEcRing(Ec));
  ZENEcPtSetXY(a, elt, elt, Ec);
  if (nb) ZENEcPtSetNext(a, a, Ec);
  
  if (!(b = ZENEcPtCreate(Ec))) {
    printf("\n"); ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
    }
  if (nb) ZENEcPtDouble(b, a, Ec);

  if (!(c = ZENEcPtCreate(Ec))) {
    printf("\n"); ZENError();
    printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  
  for(i = 0; i < nb; i++) {
    if (fseek(fd, 0, 0)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestEcPt(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    ZENEcPtDouble(a, a, Ec);
    base1 = 2+(zbnrandom()%15);
    if (ZENEcPtPrintToFile(fd, a, base1, Ec)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (fprintf(fd, " ") < 0) {
      fprintf(stderr, "ZENTestEcPt(): fprintf() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    ZENEcPtAdd(b, a, b, Ec);
    base2 = 2+(zbnrandom()%15);
    if (ZENEcPtPrintToFile(fd, b, base2, Ec)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (fseek(fd, 0, 0)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestEcPt(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENEcPtReadFromFile(c, fd, base1, Ec)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEcPtAreEqual(c, a, Ec)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR); 
    }

    if (ZENEcPtReadFromFile(c, fd, base2, Ec)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEcPtAreEqual(c, b, Ec)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR); 
    }
    printf(".");fflush(stdout);
  }
  ZENEcPtFree(c, Ec);
  ZENEcPtFree(b, Ec);
  ZENEcPtFree(a, Ec);

  if (fclose(fd)) {
    printf("\n");
    fprintf(stderr, "ZENTestEcPt(): fclose() error\n");
    printf("The %dth test failed\n", i+1);
    return(ZERR);
  }

  printf("\nZENEcPtPrintToFile() and ZENEcPtReadFromFile() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  ZENEltFree(elt, ZENEcRing(Ec));
  return(0);
}
/*\TEX{
  \title{Put/Get with string test.}
  \note{The test is $a=|ZENEcPtGetFromString|(|ZENEcPtPutToString|(a))$.}
  } 
*/
int ZENTestEcPtPutToString(Ec, nb)
     ZENEc Ec;
     int nb;
{
  ZENElt elt;
  ZENEcPt a, b, c;
  int i, j;
  double tt;
  char *str, *str1, *str2;

  /* ZENEcPtPutToString() and ZENEcPtGetFromString() */
  printf("\nTesting ZENEcPtPutToString() and ZENEcPtGetFromString()\n");
  if (ZENEltCreate(elt, ZENEcRing(Ec))) {
    return(ZERR);
  }
  tt = runtime();

  if (!(a = ZENEcPtCreate(Ec))) {
    printf("\n"); 
    ZENError();
    printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  ZENEltSetRandom(elt, ZENEcRing(Ec));
  ZENEcPtSetXY(a, elt, elt, Ec);
  if (nb) ZENEcPtSetNext(a, a, Ec);
  
  if (!(b = ZENEcPtCreate(Ec))) {
    printf("\n"); 
    ZENError();
    printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  if (nb) ZENEcPtDouble(b, a, Ec);

  if (!(c = ZENEcPtCreate(Ec))) {
    printf("\n"); ZENError();
    printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }


  for(i = 0; i < nb; i++) {
    ZENEcPtDouble(a, a, Ec);
    if (!(str1 = ZENEcPtPutToString(a, Ec))) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    ZENEcPtAdd(b, a, b, Ec);
    if (!(str2 = ZENEcPtPutToString(b, Ec))) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str1)+strlen(str2)+2)))) {
      printf("\n"); 
      fprintf(stderr, "ZENTestEcPt(): malloc(%d) error.\n",
	      strlen(str1)+strlen(str2)+2);
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (!strcpy(str, str1) || !strcat(str, str2)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestEcPt(): strcpy() or strcat() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
 
   ZENFree(str1); ZENFree(str2);
    for(str1=str;*str1 != 0;str1++)
      if(isspace(*str1)) {
	printf("Spaces in put/get\n");
	return(ZERR);
      }

    if ((j = ZENEcPtGetFromString(c, str, Ec)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEcPtAreEqual(c, a, Ec)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    if ((ZENEcPtGetFromString(c, str+j, Ec)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEcPtAreEqual(c, b, Ec)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENFree(str);
    printf(".");fflush(stdout);
  }
  ZENEcPtFree(c, Ec);
  ZENEcPtFree(b, Ec);
  ZENEcPtFree(a, Ec);
  printf("\nZENEcPtPutToString() and ZENEcPtGetFromString() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  ZENEltFree(elt, ZENEcRing(Ec));
  return(0);
}

/*\TEX{
  \title{Put/Get with file test.}
  \note{The test is $a=|ZENEltGetFromFile|(|ZENEltPutToFile|(a))$.}
  } 
*/
int ZENTestEcPtPutToFile(Ec, nb)
     ZENEc Ec;
     int nb;
{
  ZENElt elt;
  FILE *fd;
  ZENEcPt a, b, c;
  int i;
  double tt;
  
  /* ZENEcPtPutToFile() and ZENEcPtGetFromFile() */
  printf("\nTesting ZENEcPtPutToFile() and ZENEcPtGetFromFile()\n");
  if (ZENEltCreate(elt, ZENEcRing(Ec))) {
    return(ZERR);
  }
  tt = runtime();
  if (!(a = ZENEcPtCreate(Ec))) {
    printf("\n"); 
    ZENError();
    printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  ZENEltSetRandom(elt, ZENEcRing(Ec));
  ZENEcPtSetXY(a, elt, elt, Ec);
  if (nb) ZENEcPtSetNext(a, a, Ec);
  
  if (!(b = ZENEcPtCreate(Ec))) {
    printf("\n"); 
    ZENError();
    printf("\n"); printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  if (nb) ZENEcPtDouble(b, a, Ec);

  if (!(c = ZENEcPtCreate(Ec))) {
    printf("\n"); ZENError();
    printf("ZENEcPtCreate() failed\n");
    return(ZERR);
  }
  
  if (!(fd = tmpfile())) {
    printf("\n"); 
    fprintf(stderr, "ZENTestEcPt(): tmpfile() error\n");
    printf("The test failed\n");
    return(ZERR);
  }

  for(i = 0; i < nb; i++) {    
    if (fseek(fd, 0, 0)) {
      fprintf(stderr, "ZENTestElt(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    ZENEcPtDouble(a, a, Ec);
    if (ZENEcPtPutToFile(fd, a, Ec)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    ZENEcPtAdd(b, a, b, Ec);
    if (ZENEcPtPutToFile(fd, b, Ec)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    if (fseek(fd, 0, 0)) {
      fprintf(stderr, "ZENTestElt(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

# if 0
    /* It's loop forever compiled with icc on OS/2 systems ! */
    for(x=(char)getc(fd);x != EOF; x=(char)getc(fd))
      if(isspace(x)) {
	printf("Spaces in put/get\n");
	fclose(fd);
	return(ZERR);
      }
#endif    

    if (fseek(fd, 0, 0)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestEcPt(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENEcPtGetFromFile(c, fd, Ec)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEcPtAreEqual(c, a, Ec)) {
      printf("\n"); 
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    if (ZENEcPtGetFromFile(c, fd, Ec)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEcPtAreEqual(c, b, Ec)) {
      printf("\n"); 
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    printf(".");fflush(stdout);
  }
  ZENEcPtFree(c, Ec);
  ZENEcPtFree(b, Ec);
  ZENEcPtFree(a, Ec);
  if (fclose(fd)) {
    printf("\n"); 
    fprintf(stderr, "ZENTestEcPt(): fclose() error\n");
    printf("The %dth test failed\n", i+1);
    return(ZERR);
  }
  printf("\nZENEcPtPutToFile() and ZENEcPtGetFromFile() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);  
  ZENEltFree(elt, ZENEcRing(Ec));
  return(0);
}
