/*
 * zen/prgm/c/elt_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Section{Tests on elements functions}
   These functions are generic tests for elements operations.
   The main function is |ZENTestElt()| that calls the other procedures
   of this section.
   }
   */
#include <stdio.h>
#include <ctype.h>
#include <zen.h>
#include <test.h>

/* \TEX{
   \title{Test on elements.}
   \input{A finite ring |Rg|, the parameters defined in |zentest.c| and the extension level.}
   \output{0 if all the tests were ok, |ZENERR| otherwise.}
   \side{All the following test functions on |ZENElt| are invoked.}
   }
   */
int ZENTestElt(Rg, cst, lvl)
     ZENRing Rg;
     int **cst, lvl;
{
  if (ZENTestEltAlloc(Rg, cst[NB_ZEN_ELT_ALLOC][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltSetToZero(Rg, cst[NB_ZEN_ELT_TEST][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltSetToOne(Rg, cst[NB_ZEN_ELT_TEST][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltSetRandom(Rg, cst[NB_ZEN_ELT_TEST][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltSetNext(Rg, cst[NB_ZEN_ELT_NEXT][lvl], 1000)) {
    return(ZERR);
  }
  if (ZENTestEltAdd(Rg, cst[NB_ZEN_ELT_ADD][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltNegate(Rg, cst[NB_ZEN_ELT_NEGATE][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltSubtract(Rg, cst[NB_ZEN_ELT_SUBTRACT][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltMultiply(Rg, cst[NB_ZEN_ELT_MULTIPLY][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltSquare(Rg, cst[NB_ZEN_ELT_SQUARE][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltInverse(Rg, cst[NB_ZEN_ELT_INVERSE][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltTrace(Rg, cst[NB_ZEN_ELT_TRACE][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltExp(Rg, cst[NB_ZEN_ELT_EXP][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltToZ(Rg, cst[NB_ZEN_ELT_TO_Z][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltPrintToString(Rg, cst[NB_ZEN_ELT_PR_STRING][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltPrintToFile(Rg, cst[NB_ZEN_ELT_PR_FILE][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltPutToString(Rg, cst[NB_ZEN_ELT_PG_STRING][lvl])) {
    return(ZERR);
  }
  if (ZENTestEltPutToFile(Rg, cst[NB_ZEN_ELT_PG_FILE][lvl])) {
    return(ZERR);
  }
  return(0);
}
  
/*\TEX{

  All the following functions take as input a |ZENRing| and the number of tests to perform. The output is |ZENERR| if an error occured and 0 otherwise. We therefore forget in the following these informations.
  
  \title{Allocation test.}
  \note{This procedure is only calls to malloc/free.}
  } 
*/

int ZENTestEltAlloc(Rg, nb)
     ZENRing Rg;
     int nb;
{
  int i;
  ZENElt a;
  double tt;

  /* ZENEltAlloc() and ZENEltFree() */
  printf("\nTesting ZENEltAlloc() and ZENEltFree()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltAlloc(a, Rg)) {
      ZENError();
      printf("\n"); fprintf(stderr, "The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENEltFree(a, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENEltAlloc() and ZENEltFree() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Assignment to zero test.}
  \note{This procedure is only calls to |ZENEltSetToZero|() and
  |ZENEltIsZero|().}
  } 
*/
int ZENTestEltSetToZero(Rg, nb)
     ZENRing Rg;
     int nb;
{
  int i;
  ZENElt a;
  double tt;

  /* ZENEltSetToZero() and ZENEltIsZero() */
  printf("\nTesting ZENEltSetToZero() and ZENEltIsZero()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetToZero(a, Rg);
    switch(ZENEltIsZero(a, Rg)) {
    case 1:
      break;
    default:
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENEltFree(a, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENEltSetToZero() and ZENEltIsZero() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Assignment to one test.}
  \note{This procedure is only calls to |ZENEltSetToOne|() and
  |ZENEltIsOne|().}
  } 
*/
int ZENTestEltSetToOne(Rg, nb)
     ZENRing Rg;
     int nb;
{
  int i;
  ZENElt a;
  double tt;

  /* ZENEltSetToOne() and ZENEltIsOne() */
  printf("\nTesting ZENEltSetToOne() and ZENEltIsOne()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetToOne(a, Rg);
    switch(ZENEltIsOne(a, Rg)) {
    case 1:
      break;
    default:
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENEltFree(a, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENEltSetToOne() and ZENEltIsOne() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
    
  return(0);
}


/*\TEX{
  \title{Random assignment and equality test.}
  \note{A |ZENElt| $a$ is ramdomly set and then assigned in an other
  |ZENElt| $b$. The test is ok if $a=b$.}
  } 
*/
int ZENTestEltSetRandom(Rg, nb)
     ZENRing Rg;
     int nb;
{
  ZENElt a, b;
  int i;
  double tt;

  /* ZENEltSetRandom() and ZENEltAssign() */
  printf("\nTesting ZENEltSetRandom() and ZENEltAssign()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    if (ZENEltCreate(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(a, Rg);
    ZENEltAssign(b, a, Rg);
    if (!ZENEltAreEqual(a, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZENEltFree(a, Rg);
    ZENEltFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENEltSetRandom()  and ZENEltAssign() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);

  if(ZENRingDef(Rg) && !ZBNIsZero(ZENRingQ(Rg), ZENRingQl(Rg)))
    {
      ZENElt c;
      
      printf("\nTesting ZENEltConvert()\n");
      tt = runtime();
      if (ZENEltCreate(a, ZENRingDef(Rg))) {
	ZENError();
	printf("\n"); printf("ZENEltCreate() failed\n");
	return(ZERR);
      }
      if (ZENEltCreate(b, Rg)) {
	ZENError();
	printf("\n"); printf("ZENEltCreate() failed\n");
	return(ZERR);
      }
      if (ZENEltCreate(c, ZENRingDef(Rg))) {
	ZENError();
	printf("\n"); printf("ZENEltCreate() failed\n");
	return(ZERR);
      }
      for(i = 0; i < nb; i++) {
	ZENEltSetRandom(a, ZENRingDef(Rg));
	ZENEltConvert(b, Rg, a, ZENRingDef(Rg));
	ZENEltConvert(c, ZENRingDef(Rg),b, Rg);
	if (!ZENEltAreEqual(a, c, ZENRingDef(Rg))) {
	  printf("The %dth test failed\n", i+1);
	  return(ZERR);      
	}
	printf(".");fflush(stdout);
      }
      ZENEltFree(a, ZENRingDef(Rg));
      ZENEltFree(b, Rg);
      ZENEltFree(c, ZENRingDef(Rg));
      printf("\nZENEltConvert() successfully tested in %.2f s\n",
	     (runtime()-tt)/1000.);
    }
  return(0);
}

/*\TEX{
  \title{Enumeration test.}
  \input{The bound |max| on the number of elements in |ZENEltSetNext|.}
  \note{If the ring has less than |max| elements, we enumerate
  all the elements of it and verify that at the end the same element
  is computed. Otherwise, we perform |max| calls to |ZENEltSetNext|()
  and verify that each computed |ZNElt| is different.}
  } 
*/
int ZENTestEltSetNext(Rg, nb, max)
     ZENRing Rg;
     int nb, max;
{
  ZENElt a, b;
  int i, j;
  double tt;

  /* ZENEltSetNext() */
  printf("\nTesting ZENEltSetNext()\n");
  tt = runtime();
  if ((ZENRingQl(Rg) == 1) &&
      (ZBNDigit(ZENRingQ(Rg), 0) < max)) {
    for(i = 0; i < nb; i++) {
      if (ZENEltCreate(a, Rg)) {
	ZENError();
	printf("\n"); printf("ZENEltCreate() failed\n");
	return(ZERR);
      }
      if (ZENEltCreate(b, Rg)) {
	ZENError();
	printf("\n"); printf("ZENEltCreate() failed\n");
	return(ZERR);
      }
      ZENEltSetRandom(a, Rg);
      ZENEltAssign(b, a, Rg);
      for(j = 0 ; j < ZBNDigit(ZENRingQ(Rg), 0)-1; j++) {
	ZENEltSetNext(a, Rg);
	if (ZENEltAreEqual(a, b, Rg)) {
	  printf("The %dth test failed\n", i+1);
	  return(ZERR);      
	}
      }
      ZENEltSetNext(a, Rg);
      if (!ZENEltAreEqual(a, b, Rg)) {
	printf("The %dth test failed\n", i+1);
	return(ZERR);      
      }
      ZENEltFree(a, Rg);
      ZENEltFree(b, Rg);
      printf(".");fflush(stdout);
    }  
  }
  else {
    for(i = 0; i < nb; i++) {
      if (ZENEltCreate(a, Rg)) {
	ZENError();
	printf("\n"); printf("ZENEltCreate() failed\n");
	return(ZERR);
      }
      if (ZENEltCreate(b, Rg)) {
	ZENError();
	printf("\n"); printf("ZENEltCreate() failed\n");
	return(ZERR);
      }
      ZENEltSetRandom(a, Rg);
      ZENEltAssign(b, a, Rg);
      for(j = 0 ; j < max; j++) {
	ZENEltSetNext(a, Rg);
	if (ZENEltAreEqual(a, b, Rg)) {
	  printf("The %dth test failed\n", i+1);
	  return(ZERR);      
	}
      }
      ZENEltFree(a, Rg);
      ZENEltFree(b, Rg);
      printf(".");fflush(stdout);
    }
  }
  printf("\nZENEltSetNext() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Addition test.}
  \note{The test is $p a = 0$ where $p$ is the characteristic of the
  ring and $a+0 = a$ for a random |ZENElt| $a$.}
  } 
*/

int ZENTestEltAdd(Rg, nb)
     ZENRing Rg;
     int nb;
{
  ZENElt a, b;
  BigNumDigit C[1];
  int i, j, kl, kd, ks;
  double tt;

  /* ZENEltAdd() */
  printf("\nTesting ZENEltAdd()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    if (ZENEltCreate(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(a, Rg);

    /* $b = characteristic \times a$ */
    ZENEltAssign(b, a, Rg);
    kl = ZENRingPl(Rg); kd = kl - 1;

    /* Starting the computation */
    ZBNAssign(C, ZENRingP(Rg)+kd, 1);
    
    /* Looking for the most significant bit */
    for(ks = SIZE_BLOC; ks; ks--){
      if(ZBNShiftLeft(C, 1, 1))
	break;
    }
    ks--;
    while(kl >= 1){
      for(j = 0; j < ks; j++){
	ZENEltAdd(b, b, Rg);
	if(ZBNShiftLeft(C, 1, 1)){
	  ZENEltAdd(b, a, Rg);
	}
      }
      kl--;
      if(kl){
	kd--; ks = BN_DIGIT_SIZE;
	ZBNAssign(C, ZENRingP(Rg)+kd, 1);
      }
    }
    if (!ZENEltIsZero(b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENEltAdd(b, a, Rg);
    if (!ZENEltAreEqual(b, a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }


    ZENEltFree(a, Rg);
    ZENEltFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENEltAdd() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);

  return(0);
}
/*\TEX{
  \title{Negation test.}
  \note{The test is $a+(-a)=0$.}
  } 
*/
int ZENTestEltNegate(Rg, nb)
     ZENRing Rg;
     int nb;
{
  ZENElt a, b;
  int i;
  double tt;

  /* ZENEltNegate() */
  printf("\nTesting ZENEltNegate()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    if (ZENEltCreate(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(a, Rg);
    ZENEltAssign(b, a, Rg);
    ZENEltNegate(b, b, Rg);
    ZENEltAdd(a, b, Rg);

    if (!ZENEltIsZero(a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENEltFree(a, Rg);
    ZENEltFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENEltNegate() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Subtraction test.}
  \note{The test is $a-a = 0$.}
  } 
*/

int ZENTestEltSubtract(Rg, nb)
  ZENRing Rg;
  int nb;
{
  ZENElt a, b;
  int i;
  double tt;

  /* ZENEltSubtract() */
  printf("\nTesting ZENEltSubtract()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    if (ZENEltCreate(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(a, Rg);
    ZENEltAssign(b, a, Rg);
    ZENEltSubtract(b, a, Rg);

    if (!ZENEltIsZero(b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZENEltSubtract(a, a, Rg);
    if (!ZENEltIsZero(a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENEltFree(a, Rg);
    ZENEltFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENEltSubtract() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}


/*\TEX{
  \title{Multiplication test.}
  \note{The test is $a(b+c) = ab+ac$.}
  } 
*/

int ZENTestEltMultiply(Rg, nb)
     ZENRing Rg;
     int nb;
{
  ZENElt a, b, c, d, e;
  int i;
  double tt;

  /* ZENEltMultiply() */
  printf("\nTesting ZENEltMultiply()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      printf("\n"); 
      ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(a, Rg);

    if (ZENEltCreate(b, Rg)) {
      printf("\n"); 
      ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(b, Rg);

    if (ZENEltCreate(c, Rg)) {
      printf("\n");       
      ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(c, Rg);

    if (ZENEltCreate(d, Rg)) {
      printf("\n");       
      ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    if (ZENEltCreate(e, Rg)) {
      printf("\n"); 
      ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltMultiply(d, a, b, Rg);
    ZENEltMultiply(e, a, c, Rg);
    ZENEltAdd(d, e, Rg);
    ZENEltAdd(b, c, Rg);  
    ZENEltMultiply(a, b, a, Rg);
    if (!ZENEltAreEqual(a, d, Rg)) {
      printf("\n"); 
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    
    ZENEltFree(e, Rg); ZENEltFree(d, Rg);
    ZENEltFree(c, Rg); ZENEltFree(b, Rg);
    ZENEltFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENEltMultiply() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);

  return(0);
}
/*\TEX{
  \title{Squaring test.}
  \note{The test is $a^2 = a\times a$.}
  } 
*/
int ZENTestEltSquare(Rg, nb)
     ZENRing Rg;
     int nb;
{
  ZENElt a, b, c;
  int i;
  double tt;

  /* ZENEltSquare() */
  printf("\nTesting ZENEltSquare()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      printf("\n"); 
      ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(a, Rg);

    if (ZENEltCreate(b, Rg)) {
      printf("\n"); 
      ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    if (ZENEltCreate(c, Rg)) {
      printf("\n"); 
      ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltAssign(b, a, Rg);

    ZENEltSquare(a, a, Rg);
    ZENEltMultiply(c, b, b, Rg);
    if (!ZENEltAreEqual(a, c, Rg)) {
      printf("\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENEltFree(c, Rg);
    ZENEltFree(b, Rg);
    ZENEltFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENEltSquare() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);

  return(0);
}

/*\TEX{
  \title{Inverse test.}
  \note{The test is $a\times 1/a = 1$}
  } 
*/
int ZENTestEltInverse(Rg, nb)
     ZENRing Rg;
     int nb;
{
  ZENElt a, b;
  int i;
  double tt;

  /* ZENEltInverse() */
  printf("\nTesting ZENEltInverse()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      printf("\n"); 
      ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    
    do {
      ZENEltSetRandom(a, Rg);
    } while(ZENEltIsZero(a, Rg));

    if (ZENEltCreate(b, Rg)) {
      printf("\n"); 
      ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }

    switch(ZENEltInverse(b, a, Rg)) {
    case ZEN_NO_INVERSE:
      if (ZENTestEltNOINVCheck(Rg)) {
	printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
      }
      else break;
    case ZENERR:
      printf("\n"); 
      ZENError();
      printf("ZENEltInverse() failed\n");
      return(ZERR);
    case ZEN_HAS_INVERSE:
      ZENEltMultiply(b, a, b, Rg);
      if (!ZENEltIsOne(b, Rg)) {
	printf("\n"); 
	printf("The %dth test failed\n", i+1);
	return(ZERR);      
      }
    }
    
    ZENEltFree(b, Rg);
    ZENEltFree(a, Rg);
    
    printf(".");fflush(stdout);
  }
  printf("\nZENEltInverse() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);

  return(0);
}
/*\TEX{
  \title{Trace test.}
  \note{The test is $\mbox{Tr}(a+b) = \mbox{Tr}(a)+\mbox{Tr}(b)$.}
  } 
*/
int ZENTestEltTrace(Rg, nb)
     ZENRing Rg;
     int nb;
{
  ZENElt a, b, c, d, e;
  int i;
  double tt;

  /* ZENEltTrace() */
  printf("\nTesting ZENEltTrace()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(a, Rg);

    if (ZENEltCreate(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(b, Rg); 
    if (ZENEltCreate(c, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltAssign(c, a, Rg);
    ZENEltAdd(c, b, Rg);
    if (ZENEltCreate(d, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    if (ZENEltCreate(e, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltTrace(d, a, Rg);
    ZENEltTrace(e, b, Rg);
    ZENEltAdd(d, e, Rg);
    ZENEltTrace(e, c, Rg);
    if (!ZENEltAreEqual(e, d, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    ZENEltFree(c, Rg);
    ZENEltFree(d, Rg);
    ZENEltFree(e, Rg);
    ZENEltFree(b, Rg);
    ZENEltFree(a, Rg);
    
    printf(".");fflush(stdout);
  }
  printf("\nZENEltTrace() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);

  return(0);
}
/*\TEX{
  \title{Print/Read with string test.}
  \note{The test is $a=|ZENEltReadFromString|(|ZENEltPrintToString|(a))$.}
  } 
*/
int ZENTestEltPrintToString(Rg, nb)
     ZENRing Rg;
     int nb;
{
  ZENElt a, b, c;
  int i, j, base2, base1;
  char *str1, *str2, *str;
  double tt;

  /* ZENEltPrintToString() and ZENEltReadFromString() */
  printf("\nTesting ZENEltPrintToString() and ZENEltReadFromString()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(a, Rg);
    base1 = 2+(zbnrandom()%15);
    if (!(str1 = ZENEltPrintToString(a, base1, Rg))) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENEltCreate(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(b, Rg);
    base2 = 2+(zbnrandom()%15);
    if (!(str2 = ZENEltPrintToString(b, base2, Rg))) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str1)+strlen(str2)+2)))) {
      fprintf(stderr, "ZENTestElt(): malloc(%d) error.\n",
	      strlen(str1)+strlen(str2)+2);
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!strcpy(str, str1) || !strcat(str, " ") || !strcat(str, str2)) {
      fprintf(stderr, "ZENTestElt(): strcpy or strcat error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENFree(str1); ZENFree(str2);

    if (ZENEltCreate(c, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    if ((j = ZENEltReadFromString(c, str, base1, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEltAreEqual(c, a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    if ((ZENEltReadFromString(c, str+j, base2, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEltAreEqual(c, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENEltFree(c, Rg);
    ZENFree(str);
    ZENEltFree(b, Rg);
    ZENEltFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENEltPrintToString() and ZENEltReadFromString() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);

  return(0);
}
/*\TEX{
  \title{Print/Read with file test.}
  \note{The test is $a=|ZENEltReadFromFile|(|ZENEltPrintToFile|(a))$.}
  }
*/
int ZENTestEltPrintToFile(Rg, nb)
     ZENRing Rg;
     int nb;
{
  FILE *fd;
  ZENElt a, b, c;
  int i, base1, base2;
  double tt;

  /* ZENEltPrintToFile() and ZENEltReadFromFile() */
  printf("\nTesting ZENEltPrintToFile() and ZENEltReadFromFile()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (!(fd = tmpfile())) {
      fprintf(stderr, "ZENTestElt(): tmpfile() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENEltCreate(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(a, Rg);

    if(i==0) {
      ZENEltPrintToFile(stdout,a,10,Rg);
      printf("\n");
    }
    
    base1 = 2+(zbnrandom()%15);
    if (ZENEltPrintToFile(fd, a, base1, Rg)) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (fprintf(fd, " ") < 0) {
      fprintf(stderr, "ZENTestElt(): fprintf() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENEltCreate(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(b, Rg);
    base2 = 2+(zbnrandom()%15);
    if (ZENEltPrintToFile(fd, b, base2, Rg)) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (fseek(fd, 0, 0)) {
      fprintf(stderr, "ZENTestElt(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENEltCreate(c, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    if (ZENEltReadFromFile(c, fd, base1, Rg)) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEltAreEqual(c, a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    if (ZENEltReadFromFile(c, fd, base2, Rg)) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEltAreEqual(c, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENEltFree(c, Rg);
    ZENEltFree(b, Rg);
    ZENEltFree(a, Rg);

    if (fclose(fd)) {
      fprintf(stderr, "ZENTestElt(): fclose() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    printf(".");fflush(stdout);
  }
  printf("\nZENEltPrintToFile() and ZENEltReadFromFile() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);

  return(0);
}
/*\TEX{
  \title{Put/Get with string test.}
  \note{The test is $a=|ZENEltGetFromString|(|ZENEltPutToString|(a))$.}
  } 
*/
int ZENTestEltPutToString(Rg, nb)
     ZENRing Rg;
     int nb;
{
  ZENElt a, b, c;
  char *str1, *str2, *str;
  int i, j;
  double tt;

  /* ZENEltPutToString() and ZENEltGetFromString() */
  printf("\nTesting ZENEltPutToString() and ZENEltGetFromString()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(a, Rg);
    if (!(str1 = ZENEltPutToString(a, Rg))) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENEltCreate(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(b, Rg);
    if (!(str2 = ZENEltPutToString(b, Rg))) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str1)+strlen(str2)+1)))) {
      fprintf(stderr, "ZENTestElt(): malloc(%d) error.\n",
	      strlen(str1)+strlen(str2)+2);
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if(sprintf(str,"%s%s",str1,str2)<0) {
      fprintf(stderr, "ZENTestElt(): sprintf() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENFree(str1); ZENFree(str2);

    for(str1=str;*str1 != 0;str1++)
      if(isspace(*str1)) {
	printf("Spaces in put/get\n");
	return(ZERR);
      }

    if (ZENEltCreate(c, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    if ((j = ZENEltGetFromString(c, str, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEltAreEqual(c, a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    if ((ZENEltGetFromString(c, str+j, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEltAreEqual(c, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENEltFree(c, Rg);
    ZENFree(str);
    ZENEltFree(b, Rg);
    ZENEltFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENEltPutToString() and ZENEltGetFromString() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);

  return(0);
}
/*\TEX{
  \title{Put/Get with file test.}
  \note{The test is $a=|ZENEltGetFromFile|(|ZENEltPutToFile|(a))$.}
  } 
*/
int ZENTestEltPutToFile(Rg, nb)
     ZENRing Rg;
     int nb;
{
  FILE *fd;
  ZENElt a, b, c;
  int i;
  double tt;

  /* ZENEltPutToFile() and ZENEltGetFromFile() */
  printf("\nTesting ZENEltPutToFile() and ZENEltGetFromFile()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (!(fd = tmpfile())) {
      fprintf(stderr, "ZENTestElt(): tmpfile() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENEltCreate(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(a, Rg);
    if (ZENEltPutToFile(fd, a, Rg)) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (ZENEltCreate(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(b, Rg);
    if (ZENEltPutToFile(fd, b, Rg)) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    if (fseek(fd, 0, 0)) {
      fprintf(stderr, "ZENTestElt(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

#if 0
    /* It's loop forever compiled with icc on OS/2 systems ! */

    for(x=(char)getc(fd);x != EOF; x=(char)getc(fd))
      if(isspace(x)) {
	printf("Spaces in put/get\n");
	fclose(fd);
	return(ZERR);
      }
#endif

    if (fseek(fd, 0, 0)) {
      fprintf(stderr, "ZENTestElt(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENEltCreate(c, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    if (ZENEltGetFromFile(c, fd, Rg)) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEltAreEqual(c, a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    if (ZENEltGetFromFile(c, fd, Rg)) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENEltAreEqual(c, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENEltFree(c, Rg);
    ZENEltFree(b, Rg);
    ZENEltFree(a, Rg);

    if (fclose(fd)) {
      fprintf(stderr, "ZENTestElt(): fclose() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    printf(".");fflush(stdout);
  }
  printf("\nZENEltPutToFile() and ZENEltGetFromFile() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);  
  
  return(0);
}
/*
   \TEX{
   \title{Exponentiation.}
   \note{The test is $(a^n)^{1/n \bmod q-1}=a$.}
   } 
   */
int ZENTestEltExp(Rg, nb)
     ZENRing Rg;
     int nb;
{ZENElt a, b, c;
 BigNum n;
 BigNumLength nl;
 int i;
 double tt;
 ZENPrc prc;
 
 printf("\nTesting ZENEltExp() \n");
 ZENPrcSetNone(prc);
 ZENPrcSet(prc,ZENPRC_ELT_EXP);
 if(ZENRingAddPrc(Rg,prc)) {
   ZENError();
   printf("\n"); printf("ZENRingAddPrc() failed\n");
   return(ZERR);
 }
 nl=ZENRingQl(Rg)+2;
 n=ZBNC(nl);
 if (!n) {
   ZENError();
   printf("\n"); printf("ZBNC() failed\n");
   return(ZERR);
 }
 
 if (ZENEltCreate(a, Rg)) {
   ZENError();
   printf("\n"); printf("ZENEltCreate() failed\n");
   return(ZERR);
 }
 
 if (ZENEltCreate(b, Rg)) {
   ZENError();
   printf("\n"); printf("ZENEltCreate() failed\n");
   return(ZERR);
 }
 
 if (ZENEltCreate(c, Rg)) {
   ZENError();
   printf("\n"); printf("ZENEltCreate() failed\n");
   return(ZERR);
 }
 
 
 tt = runtime();
 for(i = 0; i < nb; i++) {
   
   ZENEltSetRandom(a, Rg);
   ZBNSetRandom(n,nl);
   
   if(ZENEltExpPlain(c,n,nl,a,Rg)) {
     ZENError();
     printf("\n"); printf("ZENEltExpPlain() failed\n");
     return(ZERR);
   }
   if(ZENEltExp(b,n,nl,a,Rg)) {
     ZENError();
     printf("\n"); printf("ZENEltExp() failed\n");
     return(ZERR);
   }
   if(!ZENEltAreEqual(c,b,Rg)) {
     printf("The %dth test failed\n", i+1);
     return(ZERR);
   }
   printf(".");fflush(stdout);
 }
 ZBNF(n);
 ZENEltFree(a,Rg);
 ZENEltFree(b,Rg);
 ZENEltFree(c,Rg);

 printf("\nZENEltExp() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);  
 return(0);
}

/*\TEX{
  \title{EltToZ/ZToElt test.}
  \note{The test is $a=|ZENZToElt|(|ZENEltToZ|(a))$.}
  }
*/
int ZENTestEltToZ(Rg, nb)
     ZENRing Rg;
     int nb;
{
  ZENElt a, b;
  BigNum p;
  BigNumLength pl;
  int i;
  double tt;

  /* ZENEltToZ() and ZENZToElt() */
  printf("\nTesting ZENEltToZ() and ZENZToElt()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (ZENEltCreate(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(a, Rg);

    if (ZENEltCreate(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }

    if (!(p = ZBNC(ZENRingQl(Rg)))) {
      printf("\n"); printf("ZBNC() failed\n");
      return(ZERR);
    }

    ZENEltToZ(p, &pl, a, Rg);
    ZENZToElt(b, p, pl, Rg);

    if (!ZENEltAreEqual(b, a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENEltFree(b, Rg);
    ZENEltFree(a, Rg);
    ZBNF(p);
    
    printf(".");fflush(stdout);
  }
  printf("\nZENEltToZ() and ZENZToElt() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  
  return(0);
}
