/*
 * zen/prgm/c/mat_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include <zen.h>
#include <test.h>

#ifndef min
#  define min(x,y) ((x)<(y)?(x):(y))
#endif

/* \TEX{
   \Section{Tests on Matrices}
   
   These functions are generic tests for matrices operations.
   The main function is |ZENTestMat()| that calls the other procedures
   of this section.

   \title{Test on matrices.}
   \input{A finite ring |Rg|, the parameters defined in |zentest.c| and the extension level.}
   \output{0 if all the tests were ok, |ZENERR| otherwise.}
   \side{All the following test functions on |ZENMat| are invoked.}
   }
   */
int ZENTestMat(Rg, cst, lvl)
     ZENRing Rg;
     int **cst, lvl;
{
  if (ZENTestMatAlloc(Rg, cst[NB_ZEN_MAT_ALLOC][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatSetRandom(Rg, cst[NB_ZEN_MAT_TEST][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatSetCoeff(Rg, cst[NB_ZEN_MAT_COEFF][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatAdd(Rg, cst[NB_ZEN_MAT_ADD][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatNegate(Rg, cst[NB_ZEN_MAT_NEGATE][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatSubtract(Rg, cst[NB_ZEN_MAT_SUBTRACT][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatMultiply(Rg, cst[NB_ZEN_MAT_MULTIPLY][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatWinograd(Rg, cst[NB_ZEN_MAT_WINOGRAD][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatGauss(Rg, cst[NB_ZEN_MAT_GAUSS][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatKernel(Rg, cst[NB_ZEN_MAT_KERNEL][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatConvert(Rg, cst[NB_ZEN_MAT_CONVERT][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatPrintToString(Rg, cst[NB_ZEN_MAT_PR_STRING][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatPrintToFile(Rg, cst[NB_ZEN_MAT_PR_FILE][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatPutToString(Rg, cst[NB_ZEN_MAT_PG_STRING][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestMatPutToFile(Rg, cst[NB_ZEN_MAT_PG_FILE][lvl], cst[MAX_ZEN_MAT_DEG][lvl])) {
    return(ZERR);
  }
  return(0);
}



/*\TEX{
  All the following functions take as input a |ZENRing|, the number of tests to perform and the maximal size |deg| of the matrices. The output is |ZENERR| if an error occured and 0 otherwise. We therefore forget in the following these informations.
  
  \title{Allocation test.}
  \note{This test perform many calls to |Malloc|() and |Free|() procedures.}
  } 
*/
int ZENTestMatAlloc(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat a;
  int i, nc,nl;
  double tt;

  /* ZENMatAlloc() and ZENMatFree() */
  printf("\nTesting ZENMatAlloc() and ZENMatFree()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);
    if (ZENMatAlloc(a,nl,nc, Rg)) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENMatFree(a, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENMatAlloc() and ZENMatFree() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Random assignment and equality test.}
  \note{A |ZENMat| $a$ is ramdomly set and then assigned in an other
  |ZENMat| $b$. The test is ok if $a=b$. Also tested, reset to zero and assignation by submatrices.}
  } 
*/
int ZENTestMatSetRandom(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat a, b;
  int i, nc,nl;
  double tt;

  /* ZENMatSetRandom() and ZENMatAssign() */
  printf("\nTesting ZENMatSetRandom() and ZENMatAssign()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);
    if (!(a = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatSetRandom(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatSetRandom() failed\n");
      return(ZERR);
    }

    if (!(b = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }

    ZENMatAssign(b, a, Rg);
    if (!ZENMatAreEqual(a, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENMatSetToZero(b, Rg);
    switch(ZENMatIsZero(b, Rg)) {
    case 1:
      break;
    default:
      ZENError();
      printf("\n"); printf("reset to zero failed\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    {
      ZENMat bc1,bc2,bc3,bc4;	/* bc1 | bc2
				   ---------
				   bc3 | bc4 */
      int l,c,nbl,nbc;

      c = (zbnrandom()%nc);
      l = (zbnrandom()%nl);

      nbl=l;
      nbc=c;
      if(nbl && nbc)
	{
	  if (!(bc1 = ZENMatCreate(nbl,nbc, Rg))) {
	    ZENError();
	    printf("\n"); printf("ZENMatCreate() failed\n");
	    return(ZERR);
	  }
	}
      else bc1 = NULL;
      
      nbc=nc-c;
      if(nbl && nbc)
	{
	  if (!(bc2 = ZENMatCreate(nbl,nbc, Rg))) {
	    ZENError();
	    printf("\n"); printf("ZENMatCreate() failed\n");
	    return(ZERR);
	  }
	}
      else bc2 = NULL;
      
      nbl=nl-l;
      nbc=c;
      if(nbl && nbc)
	{
	  if (!(bc3 = ZENMatCreate(nbl,nbc, Rg))) {
	    ZENError();
	    printf("\n"); printf("ZENMatCreate() failed\n");
	    return(ZERR);
	  }
	}
      else bc3 = NULL;
      
      nbc=nc-c;
      if(nbl && nbc)
	{
	  if (!(bc4 = ZENMatCreate(nbl,nbc, Rg))) {
	    ZENError();
	    printf("\n"); printf("ZENMatCreate() failed\n");
	    return(ZERR);
	  }
	}
      else bc4 = NULL;
      
      
      if(bc1)
	{
	  if(ZENMatGetSubMat(bc1,a,0,0,Rg)) {
	    ZENError();
	    printf("\n"); printf("ZENMatGetSubMat() failed (1st block)\n");
	    return(ZERR);
	  }
	}
      
      if(bc2)
	{
	  if(ZENMatGetSubMat(bc2,a,0,c,Rg)) {
	    ZENError();
	    printf("\n"); printf("ZENMatGetSubMat() failed (2nd block)\n");
	    return(ZERR);
	  }
	}
      
      if(bc3)
	{
	  if(ZENMatGetSubMat(bc3,a,l,0,Rg)) {
	    ZENError();
	    printf("\n"); printf("ZENMatGetSubMat() failed (3rd block)\n");
	    return(ZERR);
	  }
	}
      
      if(bc4)
	{
	  if(ZENMatGetSubMat(bc4,a,l,c,Rg)) {
	    ZENError();
	    printf("\n"); printf("ZENMatGetSubMat() failed (4th block)\n");
	    return(ZERR);
	  }
	}
      
      
      if(bc2)
	{
	  if(ZENMatSetSubMat(b,0,c,bc2,Rg)) {
	    ZENError();
	    printf("\n"); printf("ZENMatSetSubMat() failed (2nd block)\n");
	    return(ZERR);
	  }
	}
      
      if(bc4)
	{
	  if(ZENMatSetSubMat(b,l,c,bc4,Rg)) {
	    ZENError();
	    printf("\n"); printf("ZENMatSetSubMat() failed (4th block)\n");
	    return(ZERR);
	  }
	}
      
      if(bc3)
	{
	  if(ZENMatSetSubMat(b,l,0,bc3,Rg)) {
	    ZENError();
	    printf("\n"); printf("ZENMatSetSubMat() failed (3rd block)\n");
	    return(ZERR);
	  }
	}
      
      if(bc1)
	{
	  if(ZENMatSetSubMat(b,0,0,bc1,Rg)) {
	    ZENError();
	    printf("\n"); printf("ZENMatSetSubMat() failed (1st block)\n");
	    return(ZERR);
	  }
	}
      
      
      if (!ZENMatAreEqual(a, b, Rg)) {
	printf("The %dth test failed\n", i+1);
	return(ZERR);      
      }

      if(bc1) ZENMatFree(bc1,Rg);
      if(bc2) ZENMatFree(bc2,Rg);
      if(bc3) ZENMatFree(bc3,Rg);
      if(bc4) ZENMatFree(bc4,Rg);
    }

    ZENMatFree(a, Rg);
    ZENMatFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENMatSetRandom()  and ZENMatGetCoeff() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);

}

/*\TEX{
  \title{Coefficient assignment test.}
  \note{This procedure performs calls to |ZENMatSetCoeff| and
  |ZENMatGetCoeff|.}
  } 
*/
int ZENTestMatSetCoeff(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat a;
  ZENElt elt,elv;
  int i, l,c, nl,nc,nlr,ncr;
  double tt;

  
  printf("\nTesting ZENMatSetCoeff() and ZENMatGetCoeff()\n");
  tt = runtime();
  if (ZENEltCreate(elt, Rg)) {
    printf("\n"); ZENError();
    printf("ZENEltCreate() failed\n");
    return(ZERR);
  }
  if (ZENEltCreate(elv, Rg)) {
    printf("\n"); ZENError();
    printf("ZENEltCreate() failed\n");
    return(ZERR);
  }
  for(i = 0; i < nb; i++) {
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);
    
    if (!(a = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetToZero(a,Rg);
    
    ZENEltSetRandom(elt,Rg);
    nlr=(zbnrandom()%nl);
    ncr=(zbnrandom()%nc);
    
    if(ZENMatSetCoeff(a, nlr,ncr,elt, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }
      
    if(ZENMatGetCoeff(elv, a, nlr,ncr, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }

    if(!ZENEltAreEqual(elt,elv,Rg)) {
      printf("\n The %dth test failed\n", i+1);
      return(ZERR);      
    }

    for(l=ZENMatNbRow(a,Rg)-1;l>=0;l--)
      for(c=ZENMatNbCol(a,Rg)-1;c>=0;c--)
	if((l!=nlr)&&(c!=nlr))
	  {
	    ZENMatGetCoeff(elt, a, l,c, Rg);
	    if (!ZENEltIsZero(elt, Rg)) {
	      printf("\nThe %dth test failed\n", i+1);
	      return(ZERR);      
	    }
	  }
    
    ZENMatFree(a, Rg);
    printf(".");fflush(stdout);
  }

  ZENEltFree(elt, Rg);
  ZENEltFree(elv, Rg);

  
  printf("\nZENMatSetCoeff() and ZENMatGetCoeff() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}



/*\TEX{
  \title{Conversion test.}
  \note{This procedure performs calls to |ZENMatConvert|, |ZENVect2Mat| and
  |ZENMat2Vect| and should therefore be invoked only in an extension.}
  } 
*/
int ZENTestMatConvert(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat a,b,b2,c;
  int i, nl,nc;
  double tt;

  
  printf("\nTesting ZENMatConvert() and ZENVect2Mat()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);
    nc=1;
    nl=2;
    if(zbnrandom()%2)
      {
	if (!(a = ZENMatRowCreate(nl,nc, Rg))) {
	  ZENError();
	  printf("\n"); printf("ZENMatCreate() failed\n");
	  return(ZERR);
	}
      }
    else
      {
	if (!(a = ZENMatColCreate(nl,nc, Rg))) {
	  ZENError();
	  printf("\n"); printf("ZENMatCreate() failed\n");
	  return(ZERR);
	}
      }
    if (!(c = ZENMatCreate(nl,nc, ZENRingDef(Rg)))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if (!(b = ZENMatCreate(nl,nc, ZENRingDef(Rg)))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetRandom(b,ZENRingDef(Rg));
    if(ZENMatConvert(a,Rg,b,ZENRingDef(Rg)) ||
       ZENMatConvert(c,ZENRingDef(Rg),a,Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatConvert() failed\n");
      return(ZERR);
    }

    if(!ZENMatAreEqual(b,c,ZENRingDef(Rg))) {
      printf("\n The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZENMatFree(c, ZENRingDef(Rg));
    ZENMatFree(b, ZENRingDef(Rg));
    printf("_");fflush(stdout);

    if (!(c = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if (!(b = ZENMatRowCreate(nl,
			      nc*ZENPolyDeg(ZENRingPol(Rg),
					    ZENRingDef(Rg)),
			      ZENRingDef(Rg)))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if (!(b2 = ZENMatRowCreate(nl,
			       nc*ZENPolyDeg(ZENRingPol(Rg),
					     ZENRingDef(Rg)),
			       ZENRingDef(Rg)))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetRandom(a,Rg);
    if(ZENVect2Mat(b,a,Rg)) {
      ZENError();
      printf("\n"); printf("ZENVect2Mat() failed\n");
      return(ZERR);
    }
    if(ZENMat2Vect(c,b,Rg)) {
      ZENError();
      printf("\n"); printf("ZENMat2Vect() failed\n");
      return(ZERR);
    }
    
    if(!ZENMatAreEqual(a,c,Rg)) {
      printf("\n The %dth test failed\n", i+1);
      return(ZERR);      
    }
    printf("");fflush(stdout);
    if(_ZENCloneVect2Mat(b2,a,Rg)) {
      ZENError();
      printf("\n"); printf("ZENVect2Mat() failed\n");
      return(ZERR);
    }
    if(_ZENCloneMat2Vect(c,b2,Rg)) {
      ZENError();
      printf("\n"); printf("ZENMat2Vect() failed\n");
      return(ZERR);
    }
#if 0				/* when I'll have time, I'll debug that */
    if(!ZENMatAreEqual(b,b2,ZENRingDef(Rg))) {
      printf("\n The %dth test failed\n", i+1);
      return(ZERR);      
    }
#endif
    if(!ZENMatAreEqual(a,c,Rg)) {
      printf("\n The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZENMatFree(b, ZENRingDef(Rg));
    ZENMatFree(b2, ZENRingDef(Rg));
    printf("_");fflush(stdout);

    if (!(b = ZENMatColCreate(nl*ZENPolyDeg(ZENRingPol(Rg),
					    ZENRingDef(Rg)),
			      nc,
			      ZENRingDef(Rg)))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if (!(b2 = ZENMatColCreate(nl*ZENPolyDeg(ZENRingPol(Rg),
					     ZENRingDef(Rg)),
			       nc,
			       ZENRingDef(Rg)))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetRandom(a,Rg);
    if(ZENVect2Mat(b,a,Rg)) {
      ZENError();
      printf("\n"); printf("ZENVect2Mat() failed\n");
      return(ZERR);
    }
    if(ZENMat2Vect(c,b,Rg)) {
      ZENError();
      printf("\n"); printf("ZENMat2Vect() failed\n");
      return(ZERR);
    }
    if(!ZENMatAreEqual(a,c,Rg)) {
      printf("\n The %dth test failed\n", i+1);
      return(ZERR);      
    }
    printf("");fflush(stdout);
    if(_ZENCloneVect2Mat(b2,a,Rg)) {
      ZENError();
      printf("\n"); printf("ZENVect2Mat() failed\n");
      return(ZERR);
    }
    if(_ZENCloneMat2Vect(c,b2,Rg)) {
      ZENError();
      printf("\n"); printf("ZENMat2Vect() failed\n");
      return(ZERR);
    }
#if 0				/* when I'll have time, I'll debug that */
    if(!ZENMatAreEqual(b,b2,ZENRingDef(Rg))) {
      printf("\n The %dth test failed\n", i+1);
      return(ZERR);      
    }
#endif
    if(!ZENMatAreEqual(a,c,Rg)) {
      printf("\n The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZENMatFree(b, ZENRingDef(Rg));
    ZENMatFree(b2, ZENRingDef(Rg));
    
    ZENMatFree(a, Rg);
    ZENMatFree(c, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENMatConvert() and ZENMat2Vect() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}



/*\TEX{
  \title{Addition test.}
  \note{The test is $a+b = b+a$.}
  } 
*/
int ZENTestMatAdd(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat a, b, c, d;
  int i, nl,nc;
  double tt;

  /* ZENMatAdd() */
  printf("\nTesting ZENMatAdd()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);

    if (!(a = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatSetRandom(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatSetRandom() failed\n");
      return(ZERR);
    }

    if (!(b = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatSetRandom(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatSetRandom() failed\n");
      return(ZERR);
    }

    
    if (!(c = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetToZero(c, Rg);
    ZENMatAdd(c, a, Rg);
    ZENMatAdd(c, b, Rg);

    if (!(d = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetToZero(d, Rg);
    ZENMatAdd(d, b, Rg);
    ZENMatAdd(d, a, Rg);

    if (!ZENMatAreEqual(c, d, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENMatFree(a, Rg);
    ZENMatFree(b, Rg);
    ZENMatFree(c, Rg);
    ZENMatFree(d, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENMatAdd() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Negation test.}
  \note{The test is $a+(-a) = 0$.}
  } 
*/
int ZENTestMatNegate(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat a, b;
  int i, nl,nc;
  double tt;

  /* ZENMatNegate() */
  printf("\nTesting ZENMatNegate()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);
    if (!(a = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatSetRandom(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatSetRandom() failed\n");
      return(ZERR);
    }

    if (!(b = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatAssign(b, a, Rg);

    ZENMatNegate(b, b, Rg);
    ZENMatAdd(a, b, Rg);

    if (!ZENMatIsZero(a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENMatFree(a, Rg);
    ZENMatFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENMatNegate() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Subtraction test.}
  \note{The test is $a-a=0$.}
  } 
*/
int ZENTestMatSubtract(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat a, b;
  int i, nl,nc;
  double tt;

  /* ZENMatSubtract() */
  printf("\nTesting ZENMatSubtract()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);

    if (!(a = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatSetRandom(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatSetRandom() failed\n");
      return(ZERR);
    }
    if (!(b = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatAssign(b, a, Rg);
    ZENMatSubtract(b, a, Rg);

    if (!ZENMatIsZero(b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZENMatSubtract(a, a, Rg);
    if (!ZENMatIsZero(a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENMatFree(a, Rg);
    ZENMatFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENMatSubtract() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Multiplication test.}
  \note{The test is $a(b+c) = ab+ac$.}
  } 
*/
int ZENTestMatMultiply(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat a, b, c, d, e;
  int i, nl,nc, ni;
  double tt;

  /* ZENMatMultiply() */
  printf("\nTesting ZENMatMultiplyPlain()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    ni = 1 + (zbnrandom()%deg);
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);
    
    if (!(a = ZENMatCreate(nl,ni, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatSetRandom(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatSetRandom() failed\n");
      return(ZERR);
    }

    if (!(b = ZENMatCreate(ni,nc, Rg))) {
      printf("\n");
      ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatSetRandom(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatSetRandom() failed\n");
      return(ZERR);
    }

    if (!(c = ZENMatCreate(ni,nc, Rg))) {
      printf("\n");
      ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatSetRandom(c, Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatSetRandom() failed\n");
      return(ZERR);
    }

    if (!(d = ZENMatCreate(nl,nc, Rg))) {
      printf("\n");
      ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if (!(e = ZENMatCreate(nl,nc, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatMultiplyPlain(d, a, b, Rg)) {
      printf("\n"); 
      ZENError();
      printf("ZENMatMultiplyPlain() failed\n");
      return(ZERR);
    }
    if(ZENMatMultiplyPlain(e, a, c, Rg)) {
      printf("\n"); 
      ZENError();
      printf(" ZENMatMultiplyPlain() failed\n");
      return(ZERR);
    }
    
    ZENMatAdd(d, e, Rg);
    ZENMatAdd(b, c, Rg);
	
    if(ZENMatMultiplyPlain(e, a, b, Rg)) {
      printf("\n"); 
      ZENError();
      printf("  ZENMatMultiplyPlain() failed\n");
      return(ZERR);
    }
    if (!ZENMatAreEqual(e, d, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENMatFree(e, Rg); ZENMatFree(d, Rg);
    ZENMatFree(c, Rg); ZENMatFree(b, Rg);
    ZENMatFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENMatMultiplyPlain() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Winograd test.}
  \note{The test compare the result from |MultiplyPlain| and |MultiplyWinograd|.}
  } 
*/
int ZENTestMatWinograd(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat a, b,c,d;
  int i, nl,ni,nc;
  double tt;

  /* ZENMatSquare() */
  printf("\nTesting ZENMatWinograd()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);
    ni = 1 + (zbnrandom()%deg);

    if (!(a = ZENMatCreate(nl,ni, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatSetRandom(a, Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatSetRandom() failed\n");
      return(ZERR);
    }

    if (!(b = ZENMatCreate(ni,nc, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatSetRandom(b, Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatSetRandom() failed\n");
      return(ZERR);
    }

    if (!(c = ZENMatCreate(nl,nc, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }

    if (!(d = ZENMatCreate(nl,nc, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatMultiplyPlain(c, a, b, Rg)) {
      printf("\n"); 
      ZENError();
      printf("ZENMatMultiplyPlain() failed\n");
      return(ZERR);
    }
    if(ZENMatWinograd(d, a, b,Rg)) {
      printf("\n"); 
      ZENError();
      printf("ZENMatWinograd() failed\n");
      return(ZERR);
    }
    
    if (!ZENMatAreEqual(c, d, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    
    if(ZENMatMultiply(d, a, b,Rg)) {
      printf("\n"); 
      ZENError();
      printf("ZENMatMultiply() failed\n");
      return(ZERR);
    }
    
    if (!ZENMatAreEqual(c, d, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    
    ZENMatFree(c, Rg);
    ZENMatFree(d, Rg);
    ZENMatFree(b, Rg);
    ZENMatFree(a, Rg);
    
    printf(".");fflush(stdout);
  }
  printf("\nZENMatWinograd() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Inversion test.}
  \note{
  \begin{itemize}
  \item |Gauss Plain| is first tested. The test is $|D| = |S||M||P|$ or $|D| = |P||M||S|$ according to the type of the matrix |M| (see |GaussPlain|).
  \item |Inverse| is tested. The test is $|S||S|^{-1}=|S|^{-1}|S|=Id$.
  \end{itemize}
  }
  } 
*/
int ZENTestMatGauss(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat I,J,M,D,E,B,C,P,Q,S;
  int rk,i, nl,nc,rkl;
  double tt;

  /* ZENMatDivide() */
  printf("\nTesting ZENMatGaussPlain() ZENMatGaussLazy() ZENMatInverse()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);

    
    if (!(M = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    if(ZENMatSetRandom(M, Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatSetRandom() failed\n");
      return(ZERR);
    }
    
    
    D=ZENMatCopy(M,Rg);
    
    if (!D) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    
	
    if (!(S = (ZENMatIsRowType(M,Rg) ? ZENMatRowCreate(nl,nl, Rg) : ZENMatColCreate(nc,nc, Rg)))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    
	
	
	      
    if((zbnrandom()%1))
      {
	if (!(P = (ZENMatIsRowType(M,Rg) ? ZENMatColCreate(nc,nc, Rg) : ZENMatRowCreate(nl,nl, Rg)))) {
	  ZENError();
	  printf("\n"); printf("ZENMatCreate() failed\n");
	  return(ZERR);
	}
	if (!(Q = (ZENMatIsRowType(M,Rg) ? ZENMatColCreate(nc,nc, Rg) : ZENMatRowCreate(nl,nl, Rg)))) {
	  ZENError();
	  printf("\n"); printf("ZENMatCreate() failed\n");
	  return(ZERR);
	}
      }
    else
      {
	if (!(P = (ZENMatIsRowType(M,Rg)
		   ? ZENPermutationCreate(nc, ZENMatTypeColPermutation, Rg)
		   : ZENPermutationCreate(nl,ZENMatTypeRowPermutation, Rg)))) {
	  ZENError();
	  printf("\n"); printf("ZENPermutationCreate() failed\n");
	  return(ZERR);
	}
	if (!(Q = (ZENMatIsRowType(M,Rg)
		   ? ZENPermutationCreate(nc, ZENMatTypeColPermutation, Rg)
		   : ZENPermutationCreate(nl,ZENMatTypeRowPermutation, Rg)))) {
	  ZENError();
	  printf("\n"); printf("ZENPermutationCreate() failed\n");
	  return(ZERR);
	}
      }
    
    ZENCheckError(ZENMatSetToOne(S,Rg), ZERR);
    ZENCheckError(ZENMatSetToOne(P,Rg), ZERR);
    
    
    switch( ZENMatGaussPlain(D, S, P, &rk, Rg) )
      {
      case ZEN_NO_INVERSE:
	if (ZENTestEltNOINVCheck(Rg)) {
	  printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
	}
	else break;
      case ZENERR:
	ZENError();
	printf("ZENMatGaussPlain() failed\n");
	return(ZERR);
      case ZEN_HAS_INVERSE:
	{
	  int vrk;
	  
	  printf("_");fflush(stdout);
	  switch(ZENMatRank(M,&vrk,Rg))
	    {
	    case ZENMAT_HAS_MAXIMAL_RANK:
	      if(vrk != min(ZENMatNbRow(M,Rg),ZENMatNbCol(M,Rg))) {		
		ZENError();
		printf("\n"); printf("ZENMatRank() failed\n");
		return(ZERR);
	      }
	      break;
	    case ZENMAT_HAS_KERNEL:
	      if(vrk >= min(ZENMatNbRow(M,Rg),ZENMatNbCol(M,Rg))) {		
		ZENError();
		printf("\n"); printf("ZENMatRank() failed\n");
		return(ZERR);
	      }
	      break;
	    default:
	      ZENError();
	      printf("\n"); printf("ZENMatRank() failed\n");
	      return(ZERR);
	    }
	  
	  if(vrk != rk) {
	    ZENError();
	    printf("\n"); printf("ZENMatRank() failed. Not same rank.\n");
	    return(ZERR);
	  }
	}
	printf("_");fflush(stdout);
	if (!(B = ZENMatCreate(nl,nc, Rg))) {
	  ZENError();
	  printf("\n"); printf("ZENMatCreate() failed\n");
	  return(ZERR);
	}
	
	if (!(C = ZENMatCreate(nl,nc, Rg))) {
	  ZENError();
	  printf("\n"); printf("ZENMatCreate() failed\n");
	  return(ZERR);
	}
	
	
	if(ZENMatIsRowType(D,Rg))
	  {
	    if(ZENMatMultiply(B, M, P, Rg)) {
	      ZENError();
	      printf(" ZENMatMultiply() failed\n");
	      return(ZERR);
	    }
	    if(ZENMatMultiply(C, S, B, Rg)) {
	      ZENError();
	      printf("ZENMatMultiply() failed\n");
	      return(ZERR);
	    }
	    
	  }
	else
	  {
	    if(ZENMatMultiply(B, M, S, Rg)) {
	      ZENError();
	      printf(" ZENMatMultiply() failed\n");
	      return(ZERR);
	    }
	    if(ZENMatMultiply(C, P, B, Rg)) {
	      ZENError();
	      printf("ZENMatMultiply() failed\n");
	      return(ZERR);
	    }
	  }
		      
	if (!ZENMatAreEqual(C, D, Rg)) {
	  printf("The %dth test failed\n", i+1);
	  return(ZERR);      
	}
		      
	ZENMatFree(C,Rg);
	ZENMatFree(B,Rg);
      
		  
	printf("_");fflush(stdout);
	  
	E=ZENMatCopy(M,Rg);
	
	if (!E) {
	  ZENError();
	  printf("\n"); printf("ZENMatCreate() failed\n");
	  return(ZERR);
	}
	
	switch( ZENMatGaussLazy(E, Q, &rkl, Rg) )
	  {
	  case ZEN_NO_INVERSE:
	    if (ZENTestEltNOINVCheck(Rg)) {
	      printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
	    }
	    else break;
	  case ZENERR:
	    ZENError();
	    printf(" ZENMatMultiply() failed\n");
	    return(ZERR);
	  case ZEN_HAS_INVERSE:
	    if ((!ZENMatAreEqual(E, D, Rg))
		||(!ZENMatAreEqual(P, Q, Rg))
		||(rk != rkl)) {
	      printf("The %dth test failed\n", i+1);
	      return(ZERR);      
	    }
	    
	  }
	printf("_");fflush(stdout);
	    
	if(ZENMatAssign(E,M,Rg)) {
	  ZENError();
	  printf("\n"); printf("ZENMatAssign() failed\n");
	  return(ZERR);
	}
	
	switch(ZENMatGaussTrigInvertible(E,Rg))
	  {
	  case ZENMAT_HAS_KERNEL:
	    if(rk >= min(ZENMatNbRow(M,Rg),ZENMatNbCol(M,Rg))) {
	      printf(" The %dth test failed\n", i+1);
	      return(ZERR);      
	    }
	    break;
	  case ZENMAT_HAS_MAXIMAL_RANK:
	    if(rk < min(ZENMatNbRow(M,Rg),ZENMatNbCol(M,Rg))) {
	      printf(" The %dth test failed\n", i+1);
	      return(ZERR);      
	    }
	    break;
	  default:
	    ZENError();
	    printf(" ZENMatGaussTrigInvertible() failed\n");
	    return(ZERR);
	  }
	
	ZENMatFree(E,Rg);
      
	printf("_");fflush(stdout);	  
	
	if (!(I = ZENMatCreate(ZENMatNbBn(S,Rg),ZENMatNbBn(S,Rg), Rg))) {
	  ZENError();
	  printf("\n"); printf("ZENMatCreate() failed\n");
	  return(ZERR);
	}
	
	switch(ZENMatInverse(I,S,Rg))
	  {
	  case ZEN_NO_INVERSE:
	    if (ZENTestEltNOINVCheck(Rg)) {
	      printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
	    }
	    else break;
	  case ZENERR:
	    ZENError();
	    printf("ZENMatInverse() failed\n");
	    return(ZERR);
	  case ZEN_HAS_INVERSE:
	    
	    if (!(J = ZENMatCreate(ZENMatNbBn(S,Rg),ZENMatNbBn(S,Rg), Rg))) {
	      ZENError();
	      printf("\n"); printf("ZENMatCreate() failed\n");
	      return(ZERR);
	    }
	    if(ZENMatMultiply(J,I,S,Rg)) {
	      printf("The %dth test failed\n", i+1);
	      return(ZERR);      
	    }
	    if (!ZENMatIsOne(J, Rg)) {
	      printf("The %dth test failed\n", i+1);
	      return(ZERR);      
	    }
	    if(ZENMatMultiply(J,S,I,Rg)) {
	      printf("The %dth test failed\n", i+1);
	      return(ZERR);      
	    }
	    if (!ZENMatIsOne(J, Rg)) {
	      printf("The %dth test failed\n", i+1);
	      return(ZERR);      
	    }
	    printf("_");fflush(stdout);	  
	    {
	      ZENElt detI,detJ,detS;
	      
	      if(ZENEltAlloc(detI,Rg) ||ZENEltAlloc(detJ,Rg) || ZENEltAlloc(detS,Rg)) {
		printf("The %dth test failed\n", i+1);
		return(ZERR);      
	      }
	      
	      ZENMatSetRandom(I,Rg);
	      switch(ZENMatDet(I,detI,Rg))
		{
		case ZEN_NO_INVERSE:
		  if (ZENTestEltNOINVCheck(Rg)) {
		    printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
		  }
		  else break;
		case ZENERR:
		  ZENError();
		  printf("ZENMatInverse() failed\n");
		  return(ZERR);
		
		default:
		  ZENMatSetRandom(J,Rg);	
		  switch(ZENMatDet(J,detJ,Rg))
		    {
		    case ZEN_NO_INVERSE:
		      if (ZENTestEltNOINVCheck(Rg)) {
			printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
		      }
		      else break;
		    case ZENERR:
		      ZENError();
		      printf("ZENMatInverse() failed\n");
		      return(ZERR);

		    default:

		      ZENMatMultiply(S,I,J,Rg);
		      switch(ZENMatDet(S,detS,Rg))
			{
			case ZEN_NO_INVERSE:
			  if (ZENTestEltNOINVCheck(Rg)) {
			    printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
			  }
			  else break;
			case ZENERR:
			  ZENError();
			  printf("ZENMatInverse() failed\n");
			  return(ZERR);

			default:
			  ZENEltMultiply(detI,detJ,detI,Rg);
			  if(!ZENEltAreEqual(detS,detI,Rg)) {
			    printf("\n"); printf("Different deteminants\n");
			    return(ZERR);
			  }
			}
		    }
		}
	      ZENEltFree(detI,Rg);
	      ZENEltFree(detJ,Rg);
	      ZENEltFree(detS,Rg);
	    }
	    ZENMatFree(J,Rg);
	  } 
	ZENMatFree(I,Rg);
      }
    
    ZENMatFree(D, Rg);
    ZENMatFree(S, Rg);
    ZENMatFree(M, Rg);
    ZENMatFree(P, Rg);
    ZENMatFree(Q, Rg);
  
    
    printf(".");fflush(stdout);
  }
  printf("\nZENMatGaussPlain(), ZENMatGaussLazy()  and ZENMatInverse() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Kernel test.}
  \note{The test is $|M||K|^{t}=0$.}
  } 
*/
int ZENTestMatKernel(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat M,K,B;
  int i, nl,nc;
  double tt;

  /* ZENMatDivide() */
  printf("\nTesting ZENMatKernel()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);

    if (!(M = ZENMatCreate(nl,nc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    
    if(ZENMatSetRandom(M, Rg)) {
      ZENError();
      printf("\n"); printf("ZENMatSetRandom() failed\n");
      return(ZERR);
    }
    
    switch(ZENMatKernel(K,M,Rg)) {
    case ZEN_NO_INVERSE:
      if (ZENTestEltNOINVCheck(Rg)) {
	printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
      }
      else break;
    case ZENERR:
      ZENError();
      printf("ZENMatKernel() failed\n");
      return(ZERR);
    case ZEN_HAS_INVERSE:
      ZENMatTranspose(K,Rg);
      if (!(B = ZENMatCreate(ZENMatNbRow(M,Rg),ZENMatNbCol(K,Rg), Rg))) {
	ZENError();
	printf("\n"); printf("ZENMatCreate() failed\n");
	return(ZERR);
      }
      if(ZENMatMultiply(B,M,K,Rg)) {
	printf("The %dth test failed\n", i+1);
	return(ZERR);      
      }
      if (!ZENMatIsZero(B, Rg)) {
	printf("The %dth test failed\n", i+1);
	return(ZERR);      
      }
      ZENMatFree(B,Rg);
      ZENMatFree(K,Rg);
    }
    
    ZENMatFree(M,Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENMatKernel() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Print/Read with string test.}
  \note{The test is $a=|ZENMatReadFromString|(|ZENMatPrintToString|(a))$.}
  } 
*/
int ZENTestMatPrintToString(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat a, b, c;
  int i, j, nl,nc,  base1, base2;
  double tt;
  char *str, *str1, *str2;

  deg=(deg<5?deg:5);
  /* ZENMatPrintToString() and ZENMatReadFromString() */
  printf("\nTesting ZENMatPrintToString() and ZENMatReadFromString()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);
    
    if (!(a = ZENMatCreate(nl,nc, Rg))) {
      printf("\n"); ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetRandom(a, Rg);
    if(i)
      base1 = 2+ (zbnrandom()%15);
    else
      base1=10;
    
    if (!(str1 = ZENMatPrintToString(a, base1, Rg))) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if(!i)
      printf("%s\n",str1);

    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);

    if (!(b = ZENMatCreate(nl,nc, Rg))) {
      printf("\n"); ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetRandom(b, Rg);
    base2 = 2+ (zbnrandom()%15);
    
    if (!(str2 = ZENMatPrintToString(b, base2, Rg))) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str1)+strlen(str2)+2)))) {
      fprintf(stderr, "ZENTestMat(): malloc(%d) error.\n",
	      strlen(str1)+strlen(str2)+2);
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!strcpy(str, str1) || !strcat(str, " ") || !strcat(str, str2)) {
      fprintf(stderr, "ZENTestMat(): strcpy or strcat error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENFree(str1); ZENFree(str2);

    if ((j = ZENMatReadFromString(c, str, base1, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENMatAreEqual(c, a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }
    ZENMatFree(c, Rg);

    if ((ZENMatReadFromString(c, str+j, base2, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENMatAreEqual(c, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENMatFree(c, Rg);
    ZENFree(str);
    ZENMatFree(b, Rg);
    ZENMatFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENMatPrintToString() and ZENMatReadFromString() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Print/Read with file test.}
  \note{The test is $a=|ZENMatReadFromFile|(|ZENMatPrintToFile|(a))$.}
  } 
*/
int ZENTestMatPrintToFile(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  FILE *fd;
  ZENMat a, b, c;
  int i, base1, base2, nl,nc;
  double tt;

  /* ZENMatPrintToFile() and ZENMatReadFromFile() */
  printf("\nTesting ZENMatPrintToFile() and ZENMatReadFromFile()\n");
  deg=(deg<5?deg:5);
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (!(fd = tmpfile())) {
      fprintf(stderr, "ZENTestMat(): tmpfile() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);

    if (!(a = ZENMatCreate(nl,nc, Rg))) {
      printf("\n"); ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetRandom(a, Rg);

    base1 = 2+(zbnrandom()%15);
    if (ZENMatPrintToFile(fd, a, base1, Rg)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (fprintf(fd, " ") < 0) {
      fprintf(stderr, "ZENTestMat(): fprintf() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);

    if (!(b = ZENMatCreate(nl,nc, Rg))) {
      printf("\n"); ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetRandom(b, Rg);
    base2 = 2+ (zbnrandom()%15);
    if (ZENMatPrintToFile(fd, b, base2, Rg)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (fseek(fd, 0, 0)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestMat(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENMatReadFromFile(c, fd, base1, Rg)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENMatAreEqual(c, a, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR); 
    }
    ZENMatFree(c, Rg);

    if (ZENMatReadFromFile(c, fd, base2, Rg)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENMatAreEqual(c, b, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENMatFree(c, Rg);
    ZENMatFree(b, Rg);
    ZENMatFree(a, Rg);

    if (fclose(fd)) {
      printf("\n");
      fprintf(stderr, "ZENTestMat(): fclose() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    printf(".");fflush(stdout);
  }
  printf("\nZENMatPrintToFile() and ZENMatReadFromFile() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Put/Get with string test.}
  \note{The test is $a=|ZENMatGetFromString|(|ZENMatPutToString|(a))$.}
  } 
*/
int ZENTestMatPutToString(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENMat a, b, c;
  int i, j, nl,nc;
  double tt;
  char *str, *str1, *str2;

  deg=(deg<5?deg:5);
  /* ZENMatPutToString() and ZENMatGetFromString() */
  printf("\nTesting ZENMatPutToString() and ZENMatGetFromString()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);

    if (!(a = ZENMatCreate(nl,nc, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetRandom(a, Rg);
    if (!(str1 = ZENMatPutToString(a, Rg))) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);

    if (!(b = ZENMatCreate(nl,nc, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }

    ZENMatSetRandom(b, Rg);
    if (!(str2 = ZENMatPutToString(b, Rg))) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str1)+strlen(str2)+1)))) {
      printf("\n"); 
      fprintf(stderr, "ZENTestMat(): malloc(%d) error.\n",
	      strlen(str1)+strlen(str2)+2);
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    if (sprintf(str,"%s%s",str1,str2)<0) {
      printf("\n"); 
      fprintf(stderr, "ZENTestMat(): sprintf() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    ZENFree(str1); ZENFree(str2);
    
    for(str1=str;*str1 != 0;str1++)
      if(isspace(*str1)) {
	printf("Spaces in put/get\n");
	return(ZERR);
      }


    if ((j = ZENMatGetFromString(c, str, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENMatAreEqual(c, a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }
    ZENMatFree(c, Rg);

    if ((ZENMatGetFromString(c, str+j, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENMatAreEqual(c, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENMatFree(c, Rg);
    ZENFree(str);
    ZENMatFree(b, Rg);
    ZENMatFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENMatPutToString() and ZENMatGetFromString() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Put/Get with file test.}
  \note{The test is $a=|ZENEltGetFromFile|(|ZENEltPutToFile|(a))$.}
  } 
*/
int ZENTestMatPutToFile(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  FILE *fd;
  ZENMat a, b, c;
  int i, nl,nc;
  double tt;

  /* ZENMatPutToFile() and ZENMatGetFromFile() */
  deg=(deg<5?deg:5);
  printf("\nTesting ZENMatPutToFile() and ZENMatGetFromFile()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (!(fd = tmpfile())) {
      printf("\n"); 
      fprintf(stderr, "ZENTestMat(): tmpfile() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);

    if (!(a = ZENMatCreate(nl,nc, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetRandom(a, Rg);

    if (ZENMatPutToFile(fd, a, Rg)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    nc = 1 + (zbnrandom()%deg);
    nl = 1 + (zbnrandom()%deg);

    if (!(b = ZENMatCreate(nl,nc, Rg))) {
      printf("\n"); 
      ZENError();
      printf("\n"); printf("ZENMatCreate() failed\n");
      return(ZERR);
    }
    ZENMatSetRandom(b, Rg);

    if (ZENMatPutToFile(fd, b, Rg)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    if (fseek(fd, 0, 0)) {
      fprintf(stderr, "ZENTestElt(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

# if 0
    /* It's loop forever compiled with icc on OS/2 systems ! */
    for(x=(char)getc(fd);x != EOF; x=(char)getc(fd))
      if(isspace(x)) {
	printf("Spaces in put/get\n");
	fclose(fd);
	return(ZERR);
      }
    if (fseek(fd, 0, 0)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestMat(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
#endif    


    if (ZENMatGetFromFile(c, fd, Rg)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENMatAreEqual(c, a, Rg)) {
      printf("\n"); 
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }
    ZENMatFree(c, Rg);

    if (ZENMatGetFromFile(c, fd, Rg)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENMatAreEqual(c, b, Rg)) {
      printf("\n"); 
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENMatFree(c, Rg);
    ZENMatFree(b, Rg);
    ZENMatFree(a, Rg);

    if (fclose(fd)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestMat(): fclose() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    printf(".");fflush(stdout);
  }
  printf("\nZENMatPutToFile() and ZENMatGetFromFile() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);  
  return(0);
}
