/*
 * zen/prgm/c/poly_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Section{Tests on polynomial functions}
   These functions are generic tests for polynomial operations.
   The main function is |ZENPolyTest()| that calls the other procedures
   of this section.
   }
   */
#include <stdio.h>
#include <ctype.h>
#include <zen.h>
#include <test.h>

/* \TEX{
   \title{Test on polynomials.}
   \input{A finite ring |Rg|, the parameters defined in |zentest.c| and the extension level.}
   \output{0 if all the tests were ok, |ZENERR| otherwise.}
   \side{All the following test functions on |ZENPoly| are invoked.}
   }
   */
int ZENTestPoly(Rg, cst, lvl)
     ZENRing Rg;
     int **cst, lvl;
{  
  if (ZENTestPolyAlloc(Rg, cst[NB_ZEN_POLY_ALLOC][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolySetToZero(Rg, cst[NB_ZEN_POLY_TEST][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolySetRandom(Rg, cst[NB_ZEN_POLY_TEST][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolySetToXi(Rg, cst[NB_ZEN_POLY_COEFF][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolyUpdateDegree(Rg, cst[NB_ZEN_POLY_UPDATE][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolyAdd(Rg, cst[NB_ZEN_POLY_ADD][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolyNegate(Rg, cst[NB_ZEN_POLY_NEGATE][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolySubtract(Rg, cst[NB_ZEN_POLY_SUBTRACT][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolyEval(Rg, cst[NB_ZEN_POLY_EVAL][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolyMultiply(Rg, cst[NB_ZEN_POLY_MULTIPLY][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolySquare(Rg, cst[NB_ZEN_POLY_SQUARE][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolyDivide(Rg, cst[NB_ZEN_POLY_DIVIDE][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolyPrintToString(Rg, cst[NB_ZEN_POLY_PR_STRING][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestPolyPrintToFile(Rg, cst[NB_ZEN_POLY_PR_FILE][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  
  if (ZENTestPolyPutToString(Rg, cst[NB_ZEN_POLY_PG_STRING][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  
  if (ZENTestPolyPutToFile(Rg, cst[NB_ZEN_POLY_PG_FILE][lvl], cst[MAX_ZEN_POLY_DEG][lvl])) {
    return(ZERR);
  }
  return(0);
}

/*\TEX{
  All the following functions take as input a |ZENRing|, the number of tests to perform and the maximal degree |deg| of the polynomials. The output is |ZENERR| if an error occured and 0 otherwise. We therefore forget in the following these informations.
  
  \title{Allocation test.}
  \note{This test perform many calls to |Malloc|() and |Free|() procedures.}
  } 
*/
int ZENTestPolyAlloc(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a;
  int i, dega;
  double tt;

  /* ZENPolyAlloc() and ZENPolyFree() */
  printf("\nTesting ZENPolyAlloc() and ZENPolyFree()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = zbnrandom()%deg;
    if (ZENPolyAlloc(a, dega, Rg)) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENPolyFree(a, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENPolyAlloc() and ZENPolyFree() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Assignment to zero test.}
  \note{This procedure performs calls to |ZENPolySetToZero|() and
  |ZENPolyIsZero|().}
  } 
*/
int ZENTestPolySetToZero(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a;
  int i, dega;
  double tt;

  /* ZENPolySetToZero() and ZENPolyIsZero() */
  printf("\nTesting ZENPolySetToZero() and ZENPolyIsZero()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolyDeg(a, Rg) =  dega;

    ZENPolySetToZero(a, Rg);
    switch(ZENPolyIsZero(a, Rg)) {
    case 1:
      break;
    default:
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENPolyFree(a, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENPolySetToZero() and ZENPolyIsZero() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Random assignment and equality test.}
  \note{A |ZENPoly| $a$ is ramdomly set and then assigned in an other
  |ZENPoly| $b$. The test is ok if $a=b$.}
  } 
*/
int ZENTestPolySetRandom(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a, b;
  int i, dega;
  double tt;

  /* ZENPolySetRandom() and ZENPolyAssign() */
  printf("\nTesting ZENPolySetRandom() and ZENPolyAssign()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = ((zbnrandom()%deg));
    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);

    if (!(b = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }

    ZENPolyAssign(b, a, Rg);
    if (!ZENPolyAreEqual(a, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZENPolyFree(a, Rg);
    ZENPolyFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENPolySetRandom()  and ZENPolyGetCoeff() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  if(ZENRingDef(Rg) && !ZBNIsZero(ZENRingQ(Rg), ZENRingQl(Rg)))
    {
      ZENPoly c;
      
      printf("\nTesting ZENPolyConvert()\n");
      tt = runtime();
      for(i = 0; i < nb; i++) {
	dega = ((zbnrandom()%deg));
	if (!(a = ZENPolyCreate(dega, ZENRingDef(Rg)))) {
	  ZENError();
	  printf("\n"); printf("ZENPolyCreate() failed\n");
	  return(ZERR);
	}
	ZENPolySetRandom(a, dega, ZENRingDef(Rg));

	if (!(b = ZENPolyCreate(dega, Rg))) {
	  ZENError();
	  printf("\n"); printf("ZENPolyCreate() failed\n");
	  return(ZERR);
	}
	if (!(c = ZENPolyCreate(dega, ZENRingDef(Rg)))) {
	  ZENError();
	  printf("\n"); printf("ZENPolyCreate() failed\n");
	  return(ZERR);
	}
	ZENPolyConvert(b, Rg, a, ZENRingDef(Rg));
	ZENPolyConvert(c, ZENRingDef(Rg),b, Rg);
	if (!ZENPolyAreEqual(a, c, ZENRingDef(Rg))) {
	  printf("The %dth test failed\n", i+1);
	  return(ZERR);      
	}
	ZENPolyFree(a, ZENRingDef(Rg));
	ZENPolyFree(b, Rg);
	ZENPolyFree(c, ZENRingDef(Rg));
	printf(".");fflush(stdout);
      }
      printf("\nZENPolyConvert() successfully tested in %.2f s\n",
	     (runtime()-tt)/1000.);
    }
  return(0);
}
/*\TEX{
  \title{Assignment to $X^i$ test.}
  \note{This procedure performs calls to |ZENPolySetToXi|($i$) and
  |ZENPolyIsXi|($i$) with $i$ a random integer.}
  } 
*/
int ZENTestPolySetToXi(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a;
  ZENElt elt;
  int i, j, dega;
  double tt;

  /* ZENPolySetToXi(), ZENPolySetCoeff() and ZENPolyGetCoeff() */
  printf("\nTesting ZENPolySetToXi(), ZENPolySetCoeff() and ZENPolyGetCoeff()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = 1+(zbnrandom()%deg);
    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    if (ZENEltCreate(elt, Rg)) {
      printf("\n"); ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }

    ZENPolySetToXi(a, dega, Rg);
    ZENPolyGetCoeff(elt, a, dega, Rg);
    if (!ZENEltIsOne(elt, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZENPolySetCoeff(a, dega-1, elt, Rg);
    ZENPolyGetCoeff(elt, a, dega-1, Rg);
    if (!ZENEltIsOne(elt, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }
    for(j = 0; j < dega-1; j++) {
      ZENPolyGetCoeff(elt, a, j, Rg);
      if (!ZENEltIsZero(elt, Rg)) {
	printf("\nThe %dth test failed\n", i+1);
	return(ZERR);      
      }
    }
    ZENPolyFree(a, Rg);
    ZENEltFree(elt, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENPolySetToXi(), ZENPolySetCoeff() and ZENPolyGetCoeff() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Update of degree test.}
  \note{This test first check if $|ZENPolyUpdateDegree|(a) = a$, Then
  sets a polynomial to $b=X^n+X^{n-2}$ with $n$ random and test if
  $|ZENPolyUpdateDegree|(b) = b$ or if $|ZENPolyUpdateDegree|(b-X^n) = X^{n-2}$.}
  } 
*/
int ZENTestPolyUpdateDegree(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a, b;
  ZENElt elt;
  int i, dega;
  double tt;

  /* ZENPolyUpdateDegree() */
  printf("\nTesting ZENPolyUpdateDegree()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = 2+(zbnrandom()%deg);
    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);

    if (!(b = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolyAssign(b, a, Rg);
    ZENPolyUpdateDegree(a, Rg);
    if (!ZENPolyAreEqual(a, b, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }

    if (ZENEltCreate(elt, Rg)) {
      printf("\n"); ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetToOne(elt, Rg);

    ZENPolySetToXi(a, dega, Rg);
    ZENPolySetCoeff(a, dega-2, elt, Rg);
    ZENPolyUpdateDegree(a, Rg);
    if (ZENPolyDeg(a, Rg) != dega) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }
    
    ZENEltSetToZero(elt, Rg);
    ZENPolySetCoeff(a, dega, elt, Rg);
    ZENPolyUpdateDegree(a, Rg);
    if (ZENPolyDeg(a, Rg) != dega-2) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }
    
    ZENPolySetCoeff(a, dega-2, elt, Rg);
    ZENPolyUpdateDegree(a, Rg);
    if (!ZENPolyIsZero(a, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }
    
    ZENEltFree(elt, Rg);
    ZENPolyFree(a, Rg);
    ZENPolyFree(b, Rg);
    printf(".");fflush(stdout);
  }  
  printf("\nZENPolyUpdateDegree() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Addition test.}
  \note{The test is $a+b = b+a$.}
  } 
*/
int ZENTestPolyAdd(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a, b, c, d;
  int i, dega, degb, degc;
  double tt;

  /* ZENPolyAdd() */
  printf("\nTesting ZENPolyAdd()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);

    degb = (zbnrandom()%deg);
    if (!(b = ZENPolyCreate(degb, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(b, degb, Rg);

    degc = dega > degb ? dega : degb;
    if (!(c = ZENPolyCreate(degc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetToZero(c, Rg);
    ZENPolyAdd(c, a, Rg);
    ZENPolyAdd(c, b, Rg);

    if (!(d = ZENPolyCreate(degc, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetToZero(d, Rg);
    ZENPolyAdd(d, b, Rg);
    ZENPolyAdd(d, a, Rg);

    if (!ZENPolyAreEqual(c, d, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENPolyFree(a, Rg);
    ZENPolyFree(b, Rg);
    ZENPolyFree(c, Rg);
    ZENPolyFree(d, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENPolyAdd() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Negation test.}
  \note{The test is $a+(-a) = 0$.}
  } 
*/
int ZENTestPolyNegate(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a, b;
  int i, dega;
  double tt;

  /* ZENPolyNegate() */
  printf("\nTesting ZENPolyNegate()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);

    if (!(b = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolyAssign(b, a, Rg);

    ZENPolyNegate(b, b, Rg);
    ZENPolyAdd(a, b, Rg);

    if (!ZENPolyIsZero(a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENPolyFree(a, Rg);
    ZENPolyFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENPolyNegate() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Subtraction test.}
  \note{The test is $a-a=0$.}
  } 
*/
int ZENTestPolySubtract(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a, b;
  int i, dega;
  double tt;

  /* ZENPolySubtract() */
  printf("\nTesting ZENPolySubtract()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);
    if (!(b = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolyAssign(b, a, Rg);
    ZENPolySubtract(b, a, Rg);

    if (!ZENPolyIsZero(b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZENPolySubtract(a, a, Rg);
    if (!ZENPolyIsZero(a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENPolyFree(a, Rg);
    ZENPolyFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENPolySubtract() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Evaluation test.}
  \note{The test is $(a+b)(e) = a(e)+b(e)$.}
  } 
*/
int ZENTestPolyEval(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a, b;
  ZENElt e, f, g;
  int i, dega;
  double tt;

  /* ZENPolyAdd() */
  printf("\nTesting ZENPolyEval()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);
    if (!(b = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(b, dega, Rg);

    if (ZENEltCreate(e, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetRandom(e, Rg);
    
    if (ZENEltCreate(f, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    if (ZENEltCreate(g, Rg)) {
      ZENError();
      printf("\n"); printf("ZENEltCreate() failed\n");
      return(ZERR);
    }

    ZENPolyEval(g, a, e, Rg);
    ZENPolyEval(f, b, e, Rg);
    ZENEltAdd(g, f, Rg);
    ZENPolyAdd(a, b, Rg);
    ZENPolyEval(f, a, e, Rg);

    if (!ZENEltAreEqual(g, g, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENPolyFree(a, Rg);
    ZENPolyFree(b, Rg);
    ZENEltFree(e, Rg);
    ZENEltFree(f, Rg);
    ZENEltFree(g, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENPolyEval() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Multiplication test.}
  \note{The test is $a(b+c) = ab+ac$.}
  } 
*/
int ZENTestPolyMultiply(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a, b, c, d, e;
  int i, dega, degb;
  double tt;

  /* ZENPolyMultiply() */
  printf("\nTesting ZENPolyMultiply()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
#if 1
    dega = (zbnrandom()%deg);
    degb = (zbnrandom()%deg);
#else
    dega=10; degb=10;
#endif

    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);

    dega = dega+degb;
    if (!(b = ZENPolyCreate(degb, Rg))) {
      printf("\n");
      ZENError();
      printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(b, degb, Rg);

    if (!(c = ZENPolyCreate(degb, Rg))) {
      printf("\n");
      ZENError();
      printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(c, degb, Rg);

    if (!(d = ZENPolyCreate(dega, Rg))) {
      printf("\n");
      ZENError();
      printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    if (!(e = ZENPolyCreate(dega, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    switch(ZENPolyMultiply(d, a, b, Rg)) {
    case ZENERR:
      printf("\n"); 
      ZENError();
      printf("ZENPolyMultiply() failed\n");
      return(ZERR);
    case 0:
      switch(ZENPolyMultiply(e, a, c, Rg)) {
      case ZENERR:
	printf("\n"); 
	ZENError();
	printf("ZENPolyMultiply() failed\n");
	return(ZERR);
      case 0:
	ZENPolyAdd(d, e, Rg);
	ZENPolyAdd(b, c, Rg);
	ZENPolyAssign(e, a, Rg);
	switch(ZENPolyMultiply(e, b, e, Rg)) {
	case ZENERR:
	  printf("\n"); 
	  ZENError();
	  printf("ZENPolyMultiply() failed\n");
	  return(ZERR);
	case 0:
	  if (!ZENPolyAreEqual(e, d, Rg)) {
	    printf("The %dth test failed\n", i+1);
	    return(ZERR);      
	  }
	}
      }
    }

    ZENPolyFree(e, Rg); ZENPolyFree(d, Rg);
    ZENPolyFree(c, Rg); ZENPolyFree(b, Rg);
    ZENPolyFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENPolyMultiply() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Squaring test.}
  \note{The test is $a^2 = a\times a$.}
  } 
*/
int ZENTestPolySquare(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a, b;
  int i, dega;
  double tt;

  /* ZENPolySquare() */
  printf("\nTesting ZENPolySquare()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
#if 1
    dega = (zbnrandom()%deg);
#else
    dega=0;
#endif
    if (!(a = ZENPolyCreate(2*dega, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);

    if (!(b = ZENPolyCreate(2*dega, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }

    switch(ZENPolyMultiply(b, a, a, Rg)) {
    case ZENERR:
      printf("\n"); 
      ZENError();
      printf("ZENPolyMultiply() failed\n");
      return(ZERR);
    case 0:
      switch(ZENPolySquare(a, a, Rg)) {
      case ZENERR:
	printf("\n"); 
	ZENError();
	printf("ZENPolyMultiply() failed\n");
	return(ZERR);
      case 0:
	if (!ZENPolyAreEqual(a, b, Rg)) {
	  printf("The %dth test failed\n", i+1);
	  return(ZERR);      
	}
      }
    }

    ZENPolyFree(b, Rg);
    ZENPolyFree(a, Rg);
    
    printf(".");fflush(stdout);
  }
  printf("\nZENPolySquare() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Division test.}
  \note{The test is $a=bq+r$ and $\mbox{degree}(r) < \mbox{degree}(b)$
  if $\mbox{degree}(r) \neq 0$ where $q$ and $r$ are the quotient and
  the remainder of $a/b$ obtained with |ZENPolyDivide()|.}
  } 
*/
int ZENTestPolyDivide(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a, b, c, d, e;
  int i, dega, degb;
  double tt;

  /* ZENPolyDivide() */
  printf("\nTesting ZENPolyDivide()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
#if 1
    dega = (zbnrandom()%deg);
    degb = (zbnrandom()%deg);
#else
    dega = 4; degb = 2;
#endif
    zbnsrandom(i);

    if (!(a = ZENPolyCreate(dega, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);

    if (!(b = ZENPolyCreate(degb, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(b, degb, Rg);

    if (!(c = ZENPolyCreate(dega >= degb ? dega-degb : 0, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }

    if (!(d = ZENPolyCreate(degb ? degb-1 : 0, Rg))) {
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }

    switch(ZENPolyDivide(c, d, a, b, Rg)) {
    case ZEN_NO_INVERSE:
      if (ZENTestEltNOINVCheck(Rg)) {
	printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
      }
      else break;
    case ZENERR:
      printf("\n"); 
      ZENError();
      printf("ZENPolyDivide() failed\n");
      return(ZERR);
    case ZEN_HAS_INVERSE:
      if (!(e = ZENPolyCreate(dega > degb ? dega : degb, Rg))) {
	ZENError();
	printf("\n"); printf("ZENPolyCreate() failed\n");
	return(ZERR);
      }
      
      if (ZENPolyDeg(b, Rg) &&
	  (ZENPolyDeg(d, Rg) >= ZENPolyDeg(b, Rg))) {
	printf("The %dth test failed\n", i+1);
	return(ZERR);      
      }

      switch(ZENPolyMultiply(e, b, c, Rg)) {
      case ZENERR:
	printf("\n"); 
	ZENError();
	printf("ZENPolyMultiply() failed\n");
	return(ZERR);
      case 0:
	ZENPolyAdd(e, d, Rg);
	
	if (!ZENPolyAreEqual(a, e, Rg)) {
	  printf("The %dth test failed\n", i+1);
	  return(ZERR);      
	}
	
	ZENPolyFree(e, Rg);
      }
    }
    
    ZENPolyFree(d, Rg);
    ZENPolyFree(c, Rg);
    ZENPolyFree(b, Rg);
    ZENPolyFree(a, Rg);
    
    printf(".");fflush(stdout);
  }
  printf("\nZENPolyDivide() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Print/Read with string test.}
  \note{The test is $a=|ZENPolyReadFromString|(|ZENPolyPrintToString|(a))$.}
  } 
*/
int ZENTestPolyPrintToString(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a, b, c;
  int i, j, dega, degb, base1, base2;
  double tt;
  char *str, *str1, *str2;

  /* ZENPolyPrintToString() and ZENPolyReadFromString() */
  printf("\nTesting ZENPolyPrintToString() and ZENPolyReadFromString()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    if (!(a = ZENPolyCreate(dega, Rg))) {
      printf("\n"); ZENError();
      printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);
    base1 = 2+(zbnrandom()%15);
    if (!(str1 = ZENPolyPrintToString(a, base1, Rg))) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    degb = (zbnrandom()%deg);
    if (!(b = ZENPolyCreate(degb, Rg))) {
      printf("\n"); ZENError();
      printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(b, degb, Rg);
    base2 = 2+(zbnrandom()%15);
    if (!(str2 = ZENPolyPrintToString(b, base2, Rg))) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str1)+strlen(str2)+2)))) {
      fprintf(stderr, "ZENTestPoly(): malloc(%d) error.\n",
	      strlen(str1)+strlen(str2)+2);
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!strcpy(str, str1) || !strcat(str, " ") || !strcat(str, str2)) {
      fprintf(stderr, "ZENTestPoly(): strcpy or strcat error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENFree(str1); ZENFree(str2);

    if ((j = ZENPolyReadFromString(c, str, base1, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENPolyAreEqual(c, a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }
    ZENPolyFree(c, Rg);

    if ((ZENPolyReadFromString(c, str+j, base2, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENPolyAreEqual(c, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENPolyFree(c, Rg);
    ZENFree(str);
    ZENPolyFree(b, Rg);
    ZENPolyFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENPolyPrintToString() and ZENPolyReadFromString() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Print/Read with file test.}
  \note{The test is $a=|ZENPolyReadFromFile|(|ZENPolyPrintToFile|(a))$.}
  } 
*/
int ZENTestPolyPrintToFile(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  FILE *fd;
  ZENPoly a, b, c;
  int i, base1, base2, dega, degb;
  double tt;

  /* ZENPolyPrintToFile() and ZENPolyReadFromFile() */
  printf("\nTesting ZENPolyPrintToFile() and ZENPolyReadFromFile()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (!(fd = tmpfile())) {
      fprintf(stderr, "ZENTestPoly(): tmpfile() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    dega = (zbnrandom()%deg);
    if (!(a = ZENPolyCreate(dega, Rg))) {
      printf("\n"); ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);
    base1 = 2+(zbnrandom()%15);
    if (ZENPolyPrintToFile(fd, a, base1, Rg)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (fprintf(fd, " ") < 0) {
      fprintf(stderr, "ZENTestPoly(): fprintf() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    degb = (zbnrandom()%deg);
    if (!(b = ZENPolyCreate(degb, Rg))) {
      printf("\n"); ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(b, degb, Rg);
    base2 = 2+(zbnrandom()%15);
    if (ZENPolyPrintToFile(fd, b, base2, Rg)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (fseek(fd, 0, 0)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestPoly(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENPolyReadFromFile(c, fd, base1, Rg)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENPolyAreEqual(c, a, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR); 
    }
    ZENPolyFree(c, Rg);

    if (ZENPolyReadFromFile(c, fd, base2, Rg)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENPolyAreEqual(c, b, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENPolyFree(c, Rg);
    ZENPolyFree(b, Rg);
    ZENPolyFree(a, Rg);

    if (fclose(fd)) {
      printf("\n");
      fprintf(stderr, "ZENTestPoly(): fclose() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    printf(".");fflush(stdout);
  }
  printf("\nZENPolyPrintToFile() and ZENPolyReadFromFile() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Put/Get with string test.}
  \note{The test is $a=|ZENPolyGetFromString|(|ZENPolyPutToString|(a))$.}
  } 
*/
int ZENTestPolyPutToString(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENPoly a, b, c;
  int i, j, dega, degb;
  double tt;
  char *str, *str1, *str2;

  /* ZENPolyPutToString() and ZENPolyGetFromString() */
  printf("\nTesting ZENPolyPutToString() and ZENPolyGetFromString()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    if (!(a = ZENPolyCreate(dega, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);
    if (!(str1 = ZENPolyPutToString(a, Rg))) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    degb = (zbnrandom()%deg);
    if (!(b = ZENPolyCreate(degb, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(b, degb, Rg);
    if (!(str2 = ZENPolyPutToString(b, Rg))) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str1)+strlen(str2)+2)))) {
      printf("\n"); 
      fprintf(stderr, "ZENTestPoly(): malloc(%d) error.\n",
	      strlen(str1)+strlen(str2)+2);
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!strcpy(str, str1) || !strcat(str, str2)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestPoly(): strcpy() or strcat() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENFree(str1); ZENFree(str2);

    for(str1=str;*str1 != 0;str1++)
      if(isspace(*str1)) {
	printf("Spaces in put/get\n");
	return(ZERR);
      }

    if ((j = ZENPolyGetFromString(c, str, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENPolyAreEqual(c, a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }
    ZENPolyFree(c, Rg);

    if ((ZENPolyGetFromString(c, str+j, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENPolyAreEqual(c, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENPolyFree(c, Rg);
    ZENFree(str);
    ZENPolyFree(b, Rg);
    ZENPolyFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENPolyPutToString() and ZENPolyGetFromString() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Put/Get with file test.}
  \note{The test is $a=|ZENEltGetFromFile|(|ZENEltPutToFile|(a))$.}
  } 
*/
int ZENTestPolyPutToFile(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  FILE *fd;
  ZENPoly a, b, c;
  int i, dega, degb;
  double tt;

  /* ZENPolyPutToFile() and ZENPolyGetFromFile() */
  printf("\nTesting ZENPolyPutToFile() and ZENPolyGetFromFile()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (!(fd = tmpfile())) {
      printf("\n"); 
      fprintf(stderr, "ZENTestPoly(): tmpfile() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    dega = (zbnrandom()%deg);
    if (!(a = ZENPolyCreate(dega, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(a, dega, Rg);

    if (ZENPolyPutToFile(fd, a, Rg)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    degb = (zbnrandom()%deg);
    if (!(b = ZENPolyCreate(degb, Rg))) {
      printf("\n"); 
      ZENError();
      printf("\n"); printf("ZENPolyCreate() failed\n");
      return(ZERR);
    }
    ZENPolySetRandom(b, degb, Rg);

    if (ZENPolyPutToFile(fd, b, Rg)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (fseek(fd, 0, 0)) {
      fprintf(stderr, "ZENTestElt(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

#if 0
    /* It's loop forever compiled with icc on OS/2 systems ! */

    for(x=(char)getc(fd);x != EOF; x=(char)getc(fd))
      if(isspace(x)) {
	printf("Spaces in put/get\n");
	fclose(fd);
	return(ZERR);
      }
#endif    

    if (fseek(fd, 0, 0)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestPoly(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENPolyGetFromFile(c, fd, Rg)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENPolyAreEqual(c, a, Rg)) {
      printf("\n"); 
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }
    ZENPolyFree(c, Rg);

    if (ZENPolyGetFromFile(c, fd, Rg)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENPolyAreEqual(c, b, Rg)) {
      printf("\n"); 
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENPolyFree(c, Rg);
    ZENPolyFree(b, Rg);
    ZENPolyFree(a, Rg);

    if (fclose(fd)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestPoly(): fclose() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    printf(".");fflush(stdout);
  }
  printf("\nZENPolyPutToFile() and ZENPolyGetFromFile() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);  
  return(0);
}
