/*
 * zen/prgm/c/sr_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Section{Tests on Series}
   
   These functions are generic tests for series operations.
   The main function is |ZENSrTest()| that calls the other procedures
   of this section.
   }
   */
#include <stdio.h>
#include <zen.h>
#include <test.h>


/* \TEX{
   \title{Test on series.}
   \input{A finite ring |Rg|, the parameters defined in |zentest.c| and the extension level.}
   \output{0 if all the tests were ok, |ZENERR| otherwise.}
   \side{All the following test functions on |ZENSr| are invoked.}
   }
   */
int ZENTestSr(Rg, cst, lvl)
     ZENRing Rg;
     int **cst, lvl;
{  
  if (ZENTestSrAlloc(Rg, cst[NB_ZEN_SR_ALLOC][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestSrSetToZero(Rg, cst[NB_ZEN_SR_TEST][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestSrSetRandom(Rg, cst[NB_ZEN_SR_TEST][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestSrSetToCoeff(Rg, cst[NB_ZEN_SR_COEFF][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestSrUpdateValuation(Rg, cst[NB_ZEN_SR_UPDATE][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestSrAdd(Rg, cst[NB_ZEN_SR_ADD][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestSrNegate(Rg, cst[NB_ZEN_SR_NEGATE][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestSrSubtract(Rg, cst[NB_ZEN_SR_SUBTRACT][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestSrMultiply(Rg, cst[NB_ZEN_SR_MULTIPLY][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestSrSquare(Rg, cst[NB_ZEN_SR_SQUARE][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestSrDivide(Rg, cst[NB_ZEN_SR_DIVIDE][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestSrPrintToString(Rg, cst[NB_ZEN_SR_PR_STRING][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  if (ZENTestSrPrintToFile(Rg, cst[NB_ZEN_SR_PR_FILE][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  
  if (ZENTestSrPutToString(Rg, cst[NB_ZEN_SR_PG_STRING][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
  
  if (ZENTestSrPutToFile(Rg, cst[NB_ZEN_SR_PG_FILE][lvl], cst[MAX_ZEN_SR_DEG][lvl])) {
    return(ZERR);
  }
 
  return(0);
}

/*\TEX{
    All the following functions take as input a |ZENRing|, the number of tests to perform and the maximal degree |deg| of the series. The output is |ZENERR| if an error occured and 0 otherwise. We therefore forget in the following these informations.
  

  \title{Allocation test.}
  \note{This test perform many calls to |Malloc|() and |Free|() procedures.}
  } 
*/
int ZENTestSrAlloc(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a;
  int i, vala, dega;
  double tt;

  /* ZENSrAlloc() and ZENSrFree() */
  printf("\nTesting ZENSrAlloc() and ZENSrFree()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    vala = dega ? (zbnrandom()%dega) : 0;
    if (ZENSrAlloc(a, dega-vala+1, Rg)) {
      ZENError();
      printf("\n"); printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENSrFree(a, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENSrAlloc() and ZENSrFree() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Assignment to zero test.}
  \note{This procedure performs calls to |ZENSrSetToZero|() and
  |ZENSrIsZero|().}
  } 
*/
int ZENTestSrSetToZero(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a;
  int i, vala, dega;
  double tt;

  /* ZENSrSetToZero() and ZENSrIsZero() */
  printf("\nTesting ZENSrSetToZero() and ZENSrIsZero()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    vala = dega ? (zbnrandom()%dega) : 0;
    if (!(a = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrVal(a, Rg) = vala;
    ZENSrDeg(a, Rg) =  dega;

    ZENSrSetToZero(a, Rg);
    switch(ZENSrIsZero(a, Rg)) {
    case 1:
      break;
    default:
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENSrFree(a, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENSrSetToZero() and ZENSrIsZero() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Random assignment and equality test.}
  \note{A |ZENSr| $a$ is ramdomly set and then assigned in an other
  |ZENSr| $b$. The test is ok if $a=b$.}
  } 
*/
int ZENTestSrSetRandom(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a, b;
  int i, vala, dega;
  double tt;

  /* ZENSrSetRandom() and ZENSrAssign() */
  printf("\nTesting ZENSrSetRandom() and ZENSrAssign()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = ((zbnrandom()%deg));
    vala = dega ? (zbnrandom()%dega) : 0;
    if (!(a = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, vala, dega, Rg);

    if (!(b = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }

    ZENSrAssign(b, a, Rg);
    if (!ZENSrAreEqual(a, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZENSrFree(a, Rg);
    ZENSrFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENSrSetRandom()  and ZENSrGetCoeff() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Set to coeff test.}
  \note{This procedure performs calls to |ZENSrSetToCoeff|($i$) and
  |ZENSrGetCoeffi|($i$) with $i$ a random integer.}
  } 
*/
int ZENTestSrSetToCoeff(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a;
  ZENElt elt;
  int i, j, vala, dega;
  double tt;

  /* ZENSrSetCoeff() and ZENSrGetCoeff() */
  printf("\nZENSrSetCoeff() angd ZENSrGetCoeff()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = 1+(zbnrandom()%deg);
    vala = dega ? (int)(zbnrandom()%dega) : 0;
    if (!(a = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    if (ZENEltCreate(elt, Rg)) {
      printf("\n"); ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, vala, dega, Rg);

    j = vala+(zbnrandom()%(dega-vala+1));
    ZENEltSetToOne(elt, Rg);
    ZENSrSetCoeff(a, j, elt, Rg);
    ZENSrGetCoeff(elt, a, j, Rg);
    if (!ZENEltIsOne(elt, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENSrFree(a, Rg);
    ZENEltFree(elt, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENSrSetCoeff() and ZENSrGetCoeff() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Update of degree test.}
  \note{This test first check if $|ZENSrUpdateValuation|(a) = a$, Then
  sets a series to $b=X^n+X^{n-2}$ with $n$ random and test if
  $|ZENSrUpdateValuation|(b) = b$ or if $|ZENSrUpdateValuation|(b-X^n) = X^{n-2}$.}
  } 
*/
int ZENTestSrUpdateValuation(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a, b;
  ZENElt elt;
  int i, vala, dega;
  double tt;

  /* ZENSrUpdateValuation() */
  printf("\nTesting ZENSrUpdateValuation()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = 2+(zbnrandom()%deg);
    vala = dega ? (zbnrandom()%dega) : 0;
    dega += 4;
    if (!(a = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, vala, dega, Rg);

    if (!(b = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrAssign(b, a, Rg);
    ZENSrUpdateValuation(a, Rg);
    if (!ZENSrAreEqual(a, b, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }

    if (ZENEltCreate(elt, Rg)) {
      printf("\n"); ZENError();
      printf("ZENEltCreate() failed\n");
      return(ZERR);
    }
    ZENEltSetToOne(elt, Rg);

    ZENSrSetToZero(a, Rg);
    ZENSrSetCoeff(a, dega-2, elt, Rg);
    ZENSrUpdateValuation(a, Rg);
    if (ZENSrVal(a, Rg) != dega-2) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }
    
    ZENSrSetCoeff(a, dega, elt, Rg);
    ZENEltSetToZero(elt, Rg);
    ZENSrSetCoeff(a, dega-2, elt, Rg);
    ZENSrUpdateValuation(a, Rg);
    if (ZENSrVal(a, Rg) != dega) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }
    
    ZENSrSetCoeff(a, dega, elt, Rg);
    ZENSrUpdateValuation(a, Rg);
    if (!ZENSrIsZero(a, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }
    
    ZENEltFree(elt, Rg);
    ZENSrFree(a, Rg);
    ZENSrFree(b, Rg);
    printf(".");fflush(stdout);
  }  
  printf("\nZENSrUpdateValuation() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Addition test.}
  \note{The test is $a+b = b+a$.}
  } 
*/
int ZENTestSrAdd(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a, b, c, d;
  int i, vala, dega, valb, degb, degc;
  double tt;

  /* ZENSrAdd() */
  printf("\nTesting ZENSrAdd()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    vala = dega ? (zbnrandom()%dega) : 0;
    if (!(a = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, vala, dega, Rg);

    degb = (zbnrandom()%deg);
    valb = degb ? (zbnrandom()%degb) : 0;
    if (!(b = ZENSrCreate(degb-valb+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(b, valb, degb, Rg);

    degc = dega > degb ? dega : degb;
    if (!(c = ZENSrCreate(degc+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrDeg(c, Rg) = degc;
    ZENSrVal(c, Rg) = 0;
    ZENSrSetToZero(c, Rg);

    ZENSrAdd(c, a, Rg);
    ZENSrAdd(c, b, Rg);

    if (!(d = ZENSrCreate(degc+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrDeg(d, Rg) = degc;
    ZENSrVal(d, Rg) = 0;
    ZENSrSetToZero(d, Rg);

    ZENSrAdd(d, b, Rg);
    ZENSrAdd(d, a, Rg);

    if (!ZENSrAreEqual(c, d, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENSrFree(a, Rg);
    ZENSrFree(b, Rg);
    ZENSrFree(c, Rg);
    ZENSrFree(d, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENSrAdd() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Negation test.}
  \note{The test is $a+(-a) = 0$.}
  } 
*/
int ZENTestSrNegate(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a, b;
  int i, vala, dega;
  double tt;

  /* ZENSrNegate() */
  printf("\nTesting ZENSrNegate()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    vala = dega ? (zbnrandom()%dega) : 0;
    if (!(a = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, vala, dega, Rg);

    if (!(b = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrAssign(b, a, Rg);

    ZENSrNegate(b, b, Rg);
    ZENSrAdd(a, b, Rg);

    if (!ZENSrIsZero(a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENSrFree(a, Rg);
    ZENSrFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENSrNegate() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Subtraction test.}
  \note{The test is $a-a=0$.}
  } 
*/
int ZENTestSrSubtract(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a, b;
  int i, vala, dega;
  double tt;

  /* ZENSrSubtract() */
  printf("\nTesting ZENSrSubtract()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    vala = dega ? (zbnrandom()%dega) : 0;
    if (!(a = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, vala, dega, Rg);
    if (!(b = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrAssign(b, a, Rg);
    ZENSrSubtract(b, a, Rg);

    if (!ZENSrIsZero(b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZENSrSubtract(a, a, Rg);
    if (!ZENSrIsZero(a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZENSrFree(a, Rg);
    ZENSrFree(b, Rg);
    printf(".");fflush(stdout);
  }
  printf("\nZENSrSubtract() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Multiplication test.}
  \note{The test is $a(b+c) = ab+ac$.}
  } 
*/
int ZENTestSrMultiply(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a, b, c, d, e;
  int i, vala, dega, valb, degb;
  double tt;

  /* ZENSrMultiply() */
  printf("\nTesting ZENSrMultiply()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    vala = dega ? (zbnrandom()%dega) : 0;
    if (!(a = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, vala, dega, Rg);

    degb = (zbnrandom()%deg);
    valb = degb ? (zbnrandom()%degb) : 0;
    dega = dega+degb;
    if (!(b = ZENSrCreate(degb-valb+1, Rg))) {
      printf("\n");
      ZENError();
      printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(b, valb, degb, Rg);

    if (!(c = ZENSrCreate(degb-valb+1, Rg))) {
      printf("\n");
      ZENError();
      printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(c, valb, degb, Rg);

    if (!(d = ZENSrCreate(dega+1, Rg))) {
      printf("\n");
      ZENError();
      printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    if (!(e = ZENSrCreate(dega+1, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    switch(ZENSrMultiply(d, a, b, Rg)) {
    case ZENERR:
      printf("\n"); 
      ZENError();
      printf("ZENSrMultiply() failed\n");
      return(ZERR);
    case 0:
      switch(ZENSrMultiply(e, a, c, Rg)) {
      case ZENERR:
	printf("\n"); 
	ZENError();
	printf("ZENSrMultiply() failed\n");
	return(ZERR);
      case 0:
	ZENSrAdd(d, e, Rg);
	ZENSrAdd(b, c, Rg);
	ZENSrAssign(e, a, Rg);
	switch(ZENSrMultiply(e, b, e, Rg)) {
	case ZENERR:
	  printf("\n"); 
	  ZENError();
	  printf("ZENSrMultiply() failed\n");
	  return(ZERR);
	case 0:
	  ZENSrDeg(e, Rg) = ZENSrDeg(d, Rg);
	  if (ZENSrVal(e, Rg) > ZENSrDeg(e, Rg)) {
	    if (!ZENSrIsZero(d, Rg)) {
	      printf("The %dth test failed\n", i+1);
	      return(ZERR);      
	    }
	  }
	  else {
	    if (!ZENSrAreEqual(e, d, Rg)) {
	      printf("The %dth test failed\n", i+1);
	      return(ZERR);      
	    }
	  }
	}
      }
    }

    ZENSrFree(e, Rg); ZENSrFree(d, Rg);
    ZENSrFree(c, Rg); ZENSrFree(b, Rg);
    ZENSrFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENSrMultiply() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Squaring test.}
  \note{The test is $a^2 = a\times a$.}
  } 
*/
int ZENTestSrSquare(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a, b;
  int i, d, vala, dega;
  double tt;

  /* ZENSrSquare() */
  printf("\nTesting ZENSrSquare()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    vala = dega ? (zbnrandom()%dega) : 0;
    if (!(a = ZENSrCreate(2*dega-2*vala+1, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, vala, dega, Rg);

    if (!(b = ZENSrCreate(2*dega-2*vala+1, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENSrCreate() failed\n");
      return(ZERR);
    }

    switch(ZENSrMultiply(b, a, a, Rg)) {
    case ZENERR:
      printf("\n"); 
      ZENError();
      printf("ZENSrMultiply() failed\n");
      return(ZERR);
    case 0:
      d = ZENSrDeg(a, Rg) + ZENSrVal(a, Rg);
      switch(ZENSrSquare(a, a, Rg)) {
      case ZENERR:
	printf("\n"); 
	ZENError();
	printf("ZENSrMultiply() failed\n");
	return(ZERR);
      case 0:
	ZENSrDeg(a, Rg) = d;
	if (!ZENSrAreEqual(a, b, Rg)) {
	  printf("The %dth test failed\n", i+1);
	  return(ZERR);      
	}
      }
    }

    ZENSrFree(b, Rg);
    ZENSrFree(a, Rg);
    
    printf(".");fflush(stdout);
  }
  printf("\nZENSrSquare() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Division test.}
  \note{The test is $a=bq+r$ and $\mbox{degree}(r) < \mbox{degree}(b)$
  if $\mbox{degree}(r) \neq 0$ where $q$ and $r$ are the quotient and
  the remainder of $a/b$ obtained with |ZENSrDivide()|.}
  } 
*/
int ZENTestSrDivide(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a, b, c, e;
  int i, vala, dega, valb, degb;
  double tt;

  /* ZENSrDivide() */
  printf("\nTesting ZENSrDivide()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    vala = dega ? (zbnrandom()%dega) : 0;
    if (!(a = ZENSrCreate(dega-vala+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, vala, dega, Rg);

    degb = (zbnrandom()%deg);
    valb = degb > vala ?
      (vala ? (zbnrandom()%vala) : 0) :
	(degb ? (zbnrandom()%degb) : 0);
    if (!(b = ZENSrCreate(degb-valb+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(b, valb, degb, Rg);

    if (!(c = ZENSrCreate(dega >= degb ? dega+1 : degb+1, Rg))) {
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }

    ZENSrAssign(c, a, Rg);
    switch(ZENSrDivide(c, c, b, Rg)) {
    case ZEN_NO_INVERSE:
      if (ZENTestEltNOINVCheck(Rg)) {
	printf("\n"); printf("ZENTestEltNOINVCheck() failed\n");
      }
      else break;
    case ZENERR:
      printf("\n"); 
      ZENError();
      printf("ZENSrDivide() failed\n");
      return(ZERR);
    case ZEN_HAS_INVERSE:
      if (!(e = ZENSrCreate(dega >= degb ? dega+1 : degb+1, Rg))) {
	ZENError();
	printf("\n"); printf("ZENSrCreate() failed\n");
	return(ZERR);
      }
      
      switch(ZENSrMultiply(e, b, c, Rg)) {
      case ZENERR:
	printf("\n"); 
	ZENError();
	printf("ZENSrMultiply() failed\n");
	return(ZERR);
      case 0:
	ZENSrDeg(a, Rg) =
	  ZENSrDeg(b, Rg) + ZENSrVal(c, Rg) < ZENSrVal(b, Rg) + ZENSrDeg(c, Rg) ?
	    ZENSrDeg(b, Rg) + ZENSrVal(c, Rg) : ZENSrVal(b, Rg) + ZENSrDeg(c, Rg);
	if (!ZENSrAreEqual(a, e, Rg)) {
	  printf("The %dth test failed\n", i+1);
	  return(ZERR);      
	}
	
	ZENSrFree(e, Rg);
      }
    }
    
    ZENSrFree(c, Rg);
    ZENSrFree(b, Rg);
    ZENSrFree(a, Rg);
    
    printf(".");fflush(stdout);
  }
  printf("\nZENSrDivide() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Print/Read with string test.}
  \note{The test is $a=|ZENSrReadFromString|(|ZENSrPrintToString|(a))$.}
  } 
*/
int ZENTestSrPrintToString(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a, b, c;
  int i, j, dega, degb, base1, base2;
  double tt;
  char *str, *str1, *str2;

  /* ZENSrPrintToString() and ZENSrReadFromString() */
  printf("\nTesting ZENSrPrintToString() and ZENSrReadFromString()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    if (!(a = ZENSrCreate(dega+1, Rg))) {
      printf("\n"); ZENError();
      printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, 0, dega, Rg);
    base1 = 2+(zbnrandom()%15);
    if (!(str1 = ZENSrPrintToString(a, base1, Rg))) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    degb = (zbnrandom()%deg);
    if (!(b = ZENSrCreate(degb+1, Rg))) {
      printf("\n"); ZENError();
      printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(b, 0, degb, Rg);
    base2 = 2+(zbnrandom()%15);
    if (!(str2 = ZENSrPrintToString(b, base2, Rg))) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str1)+strlen(str2)+2)))) {
      fprintf(stderr, "ZENTestSr(): malloc(%d) error.\n",
	      strlen(str1)+strlen(str2)+2);
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!strcpy(str, str1) || !strcat(str, " ") || !strcat(str, str2)) {
      fprintf(stderr, "ZENTestSr(): strcpy or strcat error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENFree(str1); ZENFree(str2);

    if ((j = ZENSrReadFromString(c, str, base1, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENSrAreEqual(c, a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }
    ZENSrFree(c, Rg);

    if ((ZENSrReadFromString(c, str+j, base2, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENSrAreEqual(c, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENSrFree(c, Rg);
    ZENFree(str);
    ZENSrFree(b, Rg);
    ZENSrFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENSrPrintToString() and ZENSrReadFromString() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Print/Read with file test.}
  \note{The test is $a=|ZENSrReadFromFile|(|ZENSrPrintToFile|(a))$.}
  } 
*/
int ZENTestSrPrintToFile(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  FILE *fd;
  ZENSr a, b, c;
  int i, base1, base2, dega, degb;
  double tt;

  /* ZENSrPrintToFile() and ZENSrReadFromFile() */
  printf("\nTesting ZENSrPrintToFile() and ZENSrReadFromFile()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (!(fd = tmpfile())) {
      fprintf(stderr, "ZENTestSr(): tmpfile() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    dega = (zbnrandom()%deg);
    if (!(a = ZENSrCreate(dega+1, Rg))) {
      printf("\n"); ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, 0, dega, Rg);
    base1 = 2+(zbnrandom()%15);
    if (ZENSrPrintToFile(fd, a, base1, Rg)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (fprintf(fd, " ") < 0) {
      fprintf(stderr, "ZENTestSr(): fprintf() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    degb = (zbnrandom()%deg);
    if (!(b = ZENSrCreate(degb+1, Rg))) {
      printf("\n"); ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(b, 0, degb, Rg);
    base2 = 2+(zbnrandom()%15);
    if (ZENSrPrintToFile(fd, b, base2, Rg)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (fseek(fd, 0, 0)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestSr(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENSrReadFromFile(c, fd, base1, Rg)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENSrAreEqual(c, a, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR); 
    }
    ZENSrFree(c, Rg);

    if (ZENSrReadFromFile(c, fd, base2, Rg)) {
      printf("\n"); ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENSrAreEqual(c, b, Rg)) {
      printf("\nThe %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENSrFree(c, Rg);
    ZENSrFree(b, Rg);
    ZENSrFree(a, Rg);

    if (fclose(fd)) {
      printf("\n");
      fprintf(stderr, "ZENTestSr(): fclose() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    printf(".");fflush(stdout);
  }
  printf("\nZENSrPrintToFile() and ZENSrReadFromFile() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}
/*\TEX{
  \title{Put/Get with string test.}
  \note{The test is $a=|ZENSrGetFromString|(|ZENSrPutToString|(a))$.}
  } 
*/
int ZENTestSrPutToString(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  ZENSr a, b, c;
  int i, j, vala, dega, valb, degb;
  double tt;
  char *str, *str1, *str2;

  /* ZENSrPutToString() and ZENSrGetFromString() */
  printf("\nTesting ZENSrPutToString() and ZENSrGetFromString()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    dega = (zbnrandom()%deg);
    vala = dega ? (zbnrandom()%dega) : 0;
    if (!(a = ZENSrCreate(dega-vala+1, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, vala, dega, Rg);
    if (!(str1 = ZENSrPutToString(a, Rg))) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    degb = (zbnrandom()%deg);
    valb = degb ? (zbnrandom()%degb) : 0;
    if (!(b = ZENSrCreate(degb-valb+1, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(b, valb, degb, Rg);
    if (!(str2 = ZENSrPutToString(b, Rg))) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str1)+strlen(str2)+2)))) {
      printf("\n"); 
      fprintf(stderr, "ZENTestSr(): malloc(%d) error.\n",
	      strlen(str1)+strlen(str2)+2);
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!strcpy(str, str1) || !strcat(str, str2)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestSr(): strcpy() or strcat() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    ZENFree(str1); ZENFree(str2);
    for(str1=str;*str1 != 0;str1++)
      if(isspace(*str1)) {
	printf("Spaces in put/get\n");
	return(ZERR);
      }


    if ((j = ZENSrGetFromString(c, str, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENSrAreEqual(c, a, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }
    ZENSrFree(c, Rg);

    if ((ZENSrGetFromString(c, str+j, Rg)) < 0) {
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENSrAreEqual(c, b, Rg)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENSrFree(c, Rg);
    ZENFree(str);
    ZENSrFree(b, Rg);
    ZENSrFree(a, Rg);

    printf(".");fflush(stdout);
  }
  printf("\nZENSrPutToString() and ZENSrGetFromString() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

/*\TEX{
  \title{Put/Get with file test.}
  \note{The test is $a=|ZENEltGetFromFile|(|ZENEltPutToFile|(a))$.}
  } 
*/
int ZENTestSrPutToFile(Rg, nb, deg)
     ZENRing Rg;
     int nb;
     int deg;
{
  FILE *fd;
  ZENSr a, b, c;
  int i, vala, dega, valb, degb;
  double tt;

  /* ZENSrPutToFile() and ZENSrGetFromFile() */
  printf("\nTesting ZENSrPutToFile() and ZENSrGetFromFile()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    if (!(fd = tmpfile())) {
      printf("\n"); 
      fprintf(stderr, "ZENTestSr(): tmpfile() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    dega = (zbnrandom()%deg);
    vala = dega ? (zbnrandom()%dega) : 0;
    if (!(a = ZENSrCreate(dega-vala+1, Rg))) {
      printf("\n"); 
      ZENError();
      printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(a, vala, dega, Rg);

    if (ZENSrPutToFile(fd, a, Rg)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    degb = (zbnrandom()%deg);
    valb = degb ? (zbnrandom()%degb) : 0;
    if (!(b = ZENSrCreate(degb-valb+1, Rg))) {
      printf("\n"); 
      ZENError();
      printf("\n"); printf("ZENSrCreate() failed\n");
      return(ZERR);
    }
    ZENSrSetRandom(b, valb, degb, Rg);

    if (ZENSrPutToFile(fd, b, Rg)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    
    if (fseek(fd, 0, 0)) {
      fprintf(stderr, "ZENTestElt(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

# if 0
    /* It's loop forever compiled with icc on OS/2 systems ! */
    for(x=(char)getc(fd);x != EOF; x=(char)getc(fd))
      if(isspace(x)) {
	printf("Spaces in put/get\n");
	fclose(fd);
	return(ZERR);
      }
#endif    

    if (fseek(fd, 0, 0)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestSr(): fseek() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    if (ZENSrGetFromFile(c, fd, Rg)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENSrAreEqual(c, a, Rg)) {
      printf("\n"); 
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }
    ZENSrFree(c, Rg);

    if (ZENSrGetFromFile(c, fd, Rg)) {
      printf("\n"); 
      ZENError();
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }
    if (!ZENSrAreEqual(c, b, Rg)) {
      printf("\n"); 
      printf("The %dth test failed\n", i+1);
      return(ZERR); 
    }

    ZENSrFree(c, Rg);
    ZENSrFree(b, Rg);
    ZENSrFree(a, Rg);

    if (fclose(fd)) {
      printf("\n"); 
      fprintf(stderr, "ZENTestSr(): fclose() error\n");
      printf("The %dth test failed\n", i+1);
      return(ZERR);
    }

    printf(".");fflush(stdout);
  }
  printf("\nZENSrPutToFile() and ZENSrGetFromFile() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);  
  return(0);
}
