/*
 * zen/prgm/c/zbn_test.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

#define SIZE_MAX 1000
#define NB_MAX 128

#ifndef ZBN_STANDALONE
/* \TEX{
   \Section{Tests on ZBN functions}
   These functions are generic tests for ZBN operations.
   The main function is |ZENTest()| that calls the other procedures
   of this section.
   }
   */

/* \TEX{
   \title{Test ZBN functions.}
   \input{A finite ring |Rg|, the parameters defined in |zentest.c|
   and the extension level.}
   \output{0 if all the tests were ok, |ZENERR| otherwise.}
   } */
int ZENTestZBN()
{
  /* Allocation tests */
  if (ZBNTestAlloc(SIZE_MAX, NB_MAX)) {return(ZERR);}

  /* Set to Zero tests */
  if (ZBNTestZero(SIZE_MAX, NB_MAX)) {return(ZERR);}

  /* Set to One tests */
  if (ZBNTestOne(SIZE_MAX, NB_MAX)) {return(ZERR);}
  
  /* Assign tests */
  if (ZBNTestAssign(SIZE_MAX, NB_MAX)) {return(ZERR);}

  /* Num Digits tests */
  if (ZBNTestNumDigits(SIZE_MAX, NB_MAX)) {return(ZERR);}

  /* Leading Bits tests */
  if (ZBNTestLeadingZeroBitsInDigits(NB_MAX)) {return(ZERR);}

  /* Binary tests */
  if (ZBNTestBinary(SIZE_MAX, NB_MAX)) {return(ZERR);}

  /* Shift tests */
  if (ZBNTestShift(SIZE_MAX, NB_MAX)) {return(ZERR);}

  /* Add tests */
  if (ZBNTestAdd(SIZE_MAX, NB_MAX)) {return(ZERR);}

  /* Subtract tests */
  if (ZBNTestSubtract(SIZE_MAX, NB_MAX)) {return(ZERR);}

  /* Multiply Digit tests */
  if (ZBNTestMultiplyDigit(SIZE_MAX, NB_MAX)) {return(ZERR);}

  /* Divide Digit tests */
  if (ZBNTestDivideDigit(SIZE_MAX, NB_MAX)) {return(ZERR);}

  /* Divide tests */
  if (ZBNTestDivide(SIZE_MAX, NB_MAX)) {return(ZERR);}

  /* IO tests */  
  if (ZBNTestIO(SIZE_MAX/10, NB_MAX)) {return(ZERR);}

  return(0);
}
#else
/* \TEX{ At the compilation time, a test program is compiled. To use
   it, you only have to type |bin/arch/zbntest|. This program tests
   all the functions of the library (and so can take time ...).
   
   For a detailed description of these options, type
   |bin/arch/zbntest -h|.  } */
extern void Usage ___(());

int main(argc, argv)
     int argc;
     char *argv[];
{
  double tt;

  int flagset = 0;
  int allocflag = 0;
  int zeroflag = 0;
  int oneflag = 0;
  int assignflag = 0;
  int numdigitsflag = 0;
  int leadingbitsflag = 0;
  int binflag = 0;
  int shiftflag = 0;
  int addflag = 0;
  int subflag = 0;
  int muldflag = 0;
  int divdflag = 0;
  int divflag = 0;
  int ioflag = 0;

  int c;
  extern char *optarg;

  flagset = 0;
  while ((c = getopt(argc, argv, "p:")) != EOF) {
    switch (c) {
      
    case 'p':
      flagset = 1;
      switch(atoi(optarg)) {
      case 1:
	allocflag = 1;
	break;
      case 2:
	zeroflag = 1;
	break;
      case 3:
	oneflag = 1;
	break;
      case 4:
	assignflag = 1;
	break;
      case 5:
	numdigitsflag = 1;
	break;
      case 6:
	leadingbitsflag = 1;
	break;
      case 7:
	binflag = 1;
	break;
      case 8:
	shiftflag = 1;
	break;
      case 9:
	addflag = 1;
	break;
      case 10:
	subflag = 1;
	break;
      case 11:
	muldflag = 1;
	break;
      case 12:
	divdflag = 1;
	break;
      case 13:
	divflag = 1;
	break;
      case 14:
	ioflag = 1;
	break;
      }
      break;
      
    case 'h':
    case '?':
    default:
      Usage(argv[0]);
      exit(-1);
    }
  }
  if (!flagset) {
    allocflag = 1;
    zeroflag = 1;
    oneflag = 1;
    assignflag = 1;
    numdigitsflag = 1;
    leadingbitsflag = 1;
    binflag = 1;
    shiftflag = 1;
    addflag = 1;
    subflag = 1;
    muldflag = 1;
    divdflag = 1;
    divflag = 1;
    ioflag = 1;
  }

  /* Test loop */
  tt = runtime();

  printf("Computing tests\n");
  printf("===============\n");

  /* Allocation tests */
  if (allocflag) {
    printf("\nTest  1:\n--------\n");
    zbnsrandom(1);
    if (ZBNTestAlloc(SIZE_MAX, NB_MAX)) {exit(-1);}
  }

  /* Set to Zero tests */
  if (zeroflag) {
    printf("\nTest  2:\n--------\n");
    zbnsrandom(2);
    if (ZBNTestZero(SIZE_MAX, NB_MAX)) {exit(-1);}
  }

  /* Set to One tests */
  if (oneflag) {
    printf("\nTest  3:\n--------\n");
    zbnsrandom(3);
    if (ZBNTestOne(SIZE_MAX, NB_MAX)) {exit(-1);}
  }
  
  /* Assign tests */
  if (assignflag) {
    printf("\nTest  4:\n--------\n");
    zbnsrandom(4);
    if (ZBNTestAssign(SIZE_MAX, NB_MAX)) {exit(-1);}
  }

  /* Num Digits tests */
  if (numdigitsflag) {
    printf("\nTest  5:\n--------\n");
    zbnsrandom(5);
    if (ZBNTestNumDigits(SIZE_MAX, NB_MAX)) {exit(-1);}
  }

  /* Leading Bits tests */
  if (leadingbitsflag) {
    printf("\nTest  6:\n--------\n");
    zbnsrandom(6);
    if (ZBNTestLeadingZeroBitsInDigits(NB_MAX)) {exit(-1);}
  }

  /* Binary tests */
  if (binflag) {
    printf("\nTest  7:\n--------\n");
    zbnsrandom(7);
    if (ZBNTestBinary(SIZE_MAX, NB_MAX)) {exit(-1);}
  }

  /* Shift tests */
  if (shiftflag) {
    printf("\nTest  8:\n--------\n");
    zbnsrandom(8);
    if (ZBNTestShift(SIZE_MAX, NB_MAX)) {exit(-1);}
  }

  /* Add tests */
  if (addflag) {
    printf("\nTest  9:\n--------\n");
    zbnsrandom(9);
    if (ZBNTestAdd(SIZE_MAX, NB_MAX)) {exit(-1);}
  }

  /* Subtract tests */
  if (subflag) {
    printf("\nTest 10:\n--------\n");
    zbnsrandom(10);
    if (ZBNTestSubtract(SIZE_MAX, NB_MAX)) {exit(-1);}
  }

  /* Multiply Digit tests */
  if (muldflag) {
    printf("\nTest 11:\n--------\n");
    zbnsrandom(11);
    if (ZBNTestMultiplyDigit(SIZE_MAX, NB_MAX)) {exit(-1);}
  }

  /* Divide Digit tests */
  if (divdflag) {
    printf("\nTest 12:\n--------\n");
    zbnsrandom(12);
    if (ZBNTestDivideDigit(SIZE_MAX, NB_MAX)) {exit(-1);}
  }

  /* Divide tests */
  if (divflag) {
    printf("\nTest 13:\n--------\n");
    zbnsrandom(13);
    if (ZBNTestDivide(SIZE_MAX, NB_MAX)) {exit(-1);}
  }

  if (ioflag) {
    printf("\nTest 14:\n--------\n");
    zbnsrandom(14);
    if (ZBNTestIO(SIZE_MAX, NB_MAX)) {exit(-1);}
  }

  printf("\n");
  printf("All the tests correctly performed in %.2f s\n",
	 (runtime()-tt)/1000.);
  
  exit(0);
}

void Usage(arg)
     char *arg;
{
  fprintf(stderr, "Usage: %s [-p <proc>] [-h]\n", arg);
  fprintf(stderr, "This program perform tests of the library zbn\n");
  fprintf(stderr, "If no option is given, all the functions are tested\n");
  fprintf(stderr, "-p <proc>: To perform tests only on the procedure number <proc>.\n");
}
#endif

int ZBNTestAlloc(lmax, nb)
     int lmax, nb;
{
  BigNum n;
  BigNumLength nl;
  int i;
  double tt;

  printf("\nTesting ZBNC() and ZBNF().\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; nl = !nl ? 1 : nl;
    if (!(n = ZBNC(lmax))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    ZBNF(n);
    printf(".");fflush(stdout);
  }
  printf("\nZBNC() and ZBNF() successfully tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);
}

int ZBNTestZero(lmax, nb)
     int lmax, nb;
{
  BigNum n;
  BigNumLength nl;
  int i;
  double tt;

  printf("\nTesting ZBNSetToZero() and ZBNIsZero().\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; nl = !nl ? 1 : nl;
    if (!(n = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (nl) {n[0] = ~0; n[nl-1] = ~0;}
    ZBNSetToZero(n, nl);
    switch(ZBNIsZero(n, nl)) {
    case 1:
      break;
    default:
      printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    ZBNF(n);
    printf(".");fflush(stdout);
  }
  printf("\nZBNSetToZero() and ZBNIsZero() successfully tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);
}

int ZBNTestOne(lmax, nb)
     int lmax, nb;
{
  BigNum n;
  BigNumLength nl;
  int i;
  double tt;

  printf("\nTesting ZBNSetToOne() and ZBNIsOne().\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; nl = !nl ? 1 : nl;
    if (!(n = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (nl) {n[0] = ~0; n[nl-1] = ~0;}
    ZBNSetToOne(n, nl);
    switch(ZBNIsOne(n, nl)) {
    case 1:
      break;
    default:
      printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    ZBNF(n);
    printf(".");fflush(stdout);
  }
  printf("\nZBNSetToOne() and ZBNIsOne() successfully tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);
}

int ZBNTestAssign(lmax, nb)
     int lmax, nb;
{
  BigNum a, b;
  BigNumLength nl;
  int i;
  double tt;

  /* ZBNSetRandom() and ZBNAssign() */
  printf("\nTesting ZBNAssign() and ZBNSetRandom()\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; nl = !nl ? 1 : nl;
    if (!(a = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(b = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }

    ZBNSetRandom(a, nl);
    ZBNAssign(b, a, nl);
    if (!ZBNAreEqual(b, nl, a, nl)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }

    ZBNSetRandom(b, nl);
    ZBNAssign(a, b, nl);
    if (!ZBNAreEqual(a, nl, b, nl)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZBNF(a); ZBNF(b);
    printf(".");fflush(stdout);
  }
  printf("\nZBNAssign() and ZBNSetRandom() successfully tested in %.2f s\n",
	 (runtime()-tt)/1000.);
  return(0);
}

int ZBNTestNumDigits(lmax, nb)
     int lmax, nb;
{
  BigNum n;
  BigNumLength nl;
  int i;
  double tt;

  printf("\nTesting ZBNNumDigits().\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; nl = !nl ? 1 : nl;
    if (!(n = ZBNC(lmax))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    ZBNSetToZero(n, lmax);
    n[nl] = 1;
    switch(ZBNNumDigits(n, lmax) == nl+1) {
    case 1:
      break;
    default:
      printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    ZBNF(n);
    printf(".");fflush(stdout);
  }
  printf("\nZBNSetToOne() and ZBNIsOne() successfully tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);
}

int ZBNTestLeadingZeroBitsInDigits(nb)
     int nb;
{
  BigNumDigit d;
  BigNumLength nl;
  int i;
  double tt;

  printf("\nTesting ZBNNumLeadingZeroBitsInDigit().\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%BN_DIGIT_SIZE; nl = !nl ? 1 : nl;
    d = zbnrandom() >> (BN_DIGIT_SIZE-nl-1);
    d |= CST_BNDIGIT(1) << nl;
    switch(ZBNNumLeadingZeroBitsInDigit(d) == BN_DIGIT_SIZE-nl-1) {
    case 1:
      break;
    default:
      printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    printf(".");fflush(stdout);
  }
  printf("\nZBNNumLeadingZeroBitsInDigit() successfully tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);
}

int ZBNTestBinary(lmax, nb)
     int lmax, nb;
{
  BigNum a, b, c, d, e, f;
  BigNumLength nl;
  int i;
  double tt;
  
  printf("\nTesting ZBNComplement(), ZBNAnd(), ZBNOr() and ZBNXor().\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; nl = !nl ? 1 : nl;
    if (!(a = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(b = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(c = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(d = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(e = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    ZBNSetRandom(a, nl);
    ZBNSetRandom(b, nl);

    ZBNAssign(c, a, nl);
    ZBNXor(c, b, nl);
    
    ZBNAssign(d, b, nl);
    ZBNComplement(d, nl);
    ZBNAnd(d, a, nl);

    ZBNAssign(e, a, nl);
    ZBNComplement(e, nl);
    ZBNAnd(e, b, nl);

    ZBNOr(e, d, nl);
    
    if (!ZBNAreEqual(e, nl, c, nl)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZBNF(a); ZBNF(b); ZBNF(c); ZBNF(d); ZBNF(e);
    printf(".");fflush(stdout);
  }
  printf("\nZBNComplement(), ZBNAnd(), ZBNOr() and ZBNXor() successfully tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);
}

int ZBNTestShift(lmax, nb)
     int lmax, nb;
{
  BigNum a, b;
  BigNumLength nl, s;
  BigNumDigit e, d; 
  int i;
  double tt;
  
  printf("\nTesting ZBNShiftLeft()  and ZBNShiftRight().\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; nl = !nl ? 1 : nl;
    s = zbnrandom()%BN_DIGIT_SIZE;
    if (!(a = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(b = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    ZBNSetRandom(a, nl);
    ZBNAssign(b, a, nl);
    d = ZBNShiftLeft(b, nl, s);
    e = ZBNShiftRight(b, nl, s);
    b[nl-1] |= d << (BN_DIGIT_SIZE-s);

    if (!ZBNAreEqual(a, nl, b, nl) || e) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZBNF(a); ZBNF(b);
    printf(".");fflush(stdout);
  }
  printf("\nZBNShiftLeft()  and ZBNShiftRight() successfully tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);
}

int ZBNTestAdd(lmax, nb)
     int lmax, nb;
{
  BigNum a, b;
  BigNumLength nl, s;
  BigNumDigit d; 
  int i;
  double tt;

  printf("\nTesting ZBNAddCarry() and ZBNAdd().\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; nl=2; nl = !nl ? 1 : nl;
    if (!(a = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(b = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    ZBNSetRandom(a, nl);
    ZBNAssign(b, a, nl);
    ZBNComplement(b, nl);
    ZBNAddCarry(b, nl, CST_BNDIGIT(1));
    d = ZBNAdd(b, nl, a, nl, CST_BNDIGIT(0));
    
    if ((d != 1) || !ZBNIsZero(b, nl)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZBNF(a); ZBNF(b);
    printf(".");fflush(stdout);
  }
  printf("\nZBNAddCarry() and ZBNAdd() successfully tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);

}

int ZBNTestSubtract(lmax, nb)
     int lmax, nb;
{
  BigNum a, b, c;
  BigNumLength nl, s;
  BigNumDigit d; 
  int i;
  double tt;

  printf("\nTesting ZBNSubtractBorrow() and ZBNSubtract().\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; nl = !nl ? 1 : nl;
    if (!(a = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(b = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(c = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }

    ZBNSetRandom(a, nl);
    ZBNSetRandom(b, nl);

    ZBNAssign(c, a, nl);
    ZBNAddCarry(c, nl, CST_BNDIGIT(1));
    ZBNSubtractBorrow(c, nl, CST_BNDIGIT(0));
    if (!ZBNAreEqual(c, nl, a, nl)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    
    ZBNAssign(c, a, nl);
    ZBNAdd(c, nl, b, nl, CST_BNDIGIT(0));
    ZBNSubtract(c, nl, b, nl, CST_BNDIGIT(1));
    if (!ZBNAreEqual(c, nl, a, nl)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZBNF(a); ZBNF(b);  ZBNF(c);
    printf(".");fflush(stdout);
  }
  printf("\nZBNSubtractBorrow() and ZBNSubtract() tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);

}

int ZBNTestMultiplyDigit(lmax, nb)
     int lmax, nb;
{
  BigNum a, b, c, d, e, f;
  BigNumDigit A;
  BigNumLength nl;
  int i;
  double tt;
  
  printf("\nTesting ZBNMultiplyDigit().\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; nl = !nl ? 1 : nl;
    if (!(a = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(b = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(c = ZBNC(nl+2))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(d = ZBNC(nl+2))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(e = ZBNC(nl+2))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    ZBNSetRandom(a, nl);
    ZBNSetRandom(b, nl);
    A = zbnrandom();

    ZBNSetToZero(c, nl+2);
    ZBNAssign(c, a, nl);
    ZBNAdd(c, nl+1, b, nl, CST_BNDIGIT(0));

    ZBNSetToZero(d, nl+2);
    ZBNMultiplyDigit(d, nl+2, c, nl+1, A);

    ZBNSetToZero(e, nl+2);
    ZBNMultiplyDigit(e, nl+2, a, nl, A);
    ZBNMultiplyDigit(e, nl+2, b, nl, A);

    if (!ZBNAreEqual(e, nl+2, d, nl+2)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZBNF(a); ZBNF(b); ZBNF(c); ZBNF(d); ZBNF(e);
    printf(".");fflush(stdout);
  }
  printf("\nZBNMultiplyDigit() successfully tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);
}

int ZBNTestDivideDigit(lmax, nb)
     int lmax, nb;
{
  BigNum a, b, c;
  BigNumDigit R, D;
  BigNumLength nl;
  int i;
  double tt;
  
  printf("\nTesting ZBNDivideDigit().\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; if (nl < 2) {nl = 2;}
    if (!(a = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(b = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(c = ZBNC(nl-1))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }

    ZBNSetRandom(a, nl);
    D = zbnrandom();
    a[nl-1] = a[nl-1]%D;

    R = ZBNDivideDigit(c, a, nl, D);

    ZBNSetToZero(b, nl);    
    ZBNSetDigit(b, R);

    R = ZBNMultiplyDigit(b, nl, c, nl-1, D);

    if (ZBNCompare(a, nl, b, nl) || R) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZBNF(a); ZBNF(b); ZBNF(c);
    printf(".");fflush(stdout);
  }
  printf("\nZBNDivideDigit() successfully tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);
}

int ZBNTestDivide(lmax, nb)
     int lmax, nb;
{
  BigNum a, b, c, d;
  BigNumDigit R, D;
  BigNumLength nl;
  int i;
  double tt;
  
  printf("\nTesting ZBNDivide().\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; if (nl < 2) {nl = 2;}

    if (!(a = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(b = ZBNC(nl/2))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(c = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
    if (!(d = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }

    ZBNSetRandom(a, nl);
    ZBNSetRandom(b, nl/2);
    a[nl-1] = a[nl-1]%b[nl/2-1];
    ZBNAssign(d, a, nl);

    ZBNDivide(a, nl, b, nl/2);

    ZBNSetToZero(c, nl);    
    ZBNAssign(c, a, nl/2);
    ZBNMultiply(c, nl, a+nl/2, nl-nl/2, b, nl/2);

    if (ZBNCompare(d, nl, c, nl)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZBNF(a); ZBNF(b); ZBNF(c); ZBNF(d);
    printf(".");fflush(stdout);
  }
  printf("\nZBNDivide() successfully tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);
}

int ZBNTestIO(lmax, nb)
     int lmax, nb;
{
  BigNum a, b;
  BigNumLength nl,bl;
  char *str;
  int i, base;
  double tt;
  
  printf("\nTesting ZBNPrintToString() and ZBNReadFromString.\n");
  tt = runtime();
  for(i = 0; i < nb; i++) {
    nl = zbnrandom()%lmax; if (nl < 2) {nl = 2;}

    if (!(a = ZBNC(nl))) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }

    ZBNSetRandom(a, nl);

    base = 2+(zbnrandom()%(ZBN_IO_MAX_BASIS-1));

    str=ZBNPrintToString(a,nl,base);
    if(!str) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }

    if(ZBNReadFromString(&b,&bl,str,base)==ZENERR) {
      printf("\n"); printf("The %dth test failed.\n", i+1);
      return(ZERR);
    }
      
    if (ZBNCompare(a, nl, b, bl)) {
      printf("The %dth test failed\n", i+1);
      return(ZERR);      
    }
    ZBNF(a); ZBNF(b); ZENFree(str);
    printf(".");fflush(stdout);
  }
  printf("\nZBN[ReadFrom|PrintTo]String() successfully tested in %.2f s.\n",
	 (runtime()-tt)/1000.);
  return(0);
}
