/*
 * zen/sys/c/memalloc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
      Following a GMP mechanism, we allow customized allocation functions.
      This mechanism is also intended to help debugging memory leaks and buffer
      overflows. Hence, the behaviour of this feature is partially governed by
      |ZEN_MALLOC_TEST| precompilation directive, which is only set when compiling
      library with |DebuggingMalloc| flag set in file |site.h|.
      }
*/
#include <stdio.h>
#include <stdlib.h>
#include "sys.h"

#if !ZEN_MALLOC_TEST
/* \TEX{
   \subsection{Optimized version (|ZEN_MALLOC_TEST| is not set)}
      }
*/
# if 0
/*
   \TEX{
   \title{Allocating memory}
   \input{The size of memory to allocate}
   \output{A pointer on the allocated array or |NULL| if an error occured.}
   \note{This macro is used in the whole |ZEN| and |ZENFACT| libraries when memory allocation is needed. By default, a call to the native |malloc| function is performed. This can be overwritten by |ZENSetMemoryFunctions|.}
   } 
   */
void * ZENMalloc(size)
     size_t size;{}
# endif
void * (*ZENMallocPtr) ___((size_t))  = malloc;

# if 0
/*
   \TEX{
   \title{Reallocating memory}
   \input{The pointer to a previously allocated memory array, obtained by |ZENMalloc| , and the new size needed.}
   \output{A pointer on the newly allocated array, or |NULL| if an error occured.}
   \side{The values stored in the first array designed by |oldptr| are copied in the new array.}
   \note{This macro is used in the whole |ZEN| and |ZENFACT| libraries when memory reallocation is needed. By default, a call to the native |realloc| function is performed. This can be overwritten by |ZENSetMemoryFunctions|.}
   } 
   */
void * ZENRealloc(oldptr,newsize)
     void * oldptr;
     size_t newsize;{}
# endif
void * (*ZENReallocPtr) ___((void *, size_t))  = realloc;
# if 0
/*
   \TEX{
   \title{Freeing memory}
   \input{A pointer to a previously allocated memory array, obtained by |ZENMalloc|.}
   \side{The memory pointed to by |ptr| is freed.}
   \note{This macro is used in the whole |ZEN| and |ZENFACT| libraries when freeing memory is needed. By default, a call to the native |free| function is performed. This can be overwritten by |ZENSetMemoryFunctions|.}
   } 
   */
void ZENFree(ptr)
     void * ptr; {}
# endif
void (*ZENFreePtr) ___((void *))  = free;

/*
   \TEX{
   \title{Setting custom memory allocation functions}
   \input{The three memory allocation functions performing |malloc|, |realloc|, and |free|.}
   \side{The internal memory allocation functions are customized}
   \note{By default, the standard |malloc|, |realloc|, and |free|
   functions are used. This function should be called at the very
   beginning of a program.}
   } */
void ZENSetMemoryFunctions(ma,re,fr)
     void * (* ma) ___((size_t));
     void * (* re) ___((void *, size_t));
     void  (* fr) ___((void *));
{
  if(ma)
    ZENMallocPtr=ma;
  if(re)
    ZENReallocPtr=re;
  if(fr)
    ZENFreePtr=fr;

}

#else  /* ZEN_MALLOC_TEST */
/* \TEX{
   \subsection{Debugging version (|ZEN_MALLOC_TEST| is set)}
   We here only provide a mean to get information from inside the library but no garbage collector nor memory checker is embedded in ZEN, even if the following entry points should be well suited for that.  

   \title{Allocating memory}
   \input{The size of memory to allocate, a string and an integer}
   \output{A pointer on the allocated array or |NULL| if an error occured.}
   \note{In the whole scope of the library, file and line are set according to the line of the source file where the call is performed. This can be used to track memory leaks and buffer overflows. By default, a call to the native |malloc| function is performed and file and line are just ignored. This can be overwritten by |ZENSetMemoryFunctions|.}
   } 
   */
void * ZENMallocFct(size, file, line)
     size_t size;
     char * file;
     int line;
{
  return(malloc(size));
}
void * (*ZENMallocPtr) ___((size_t, char *, int))  = ZENMallocFct;

/*
   \TEX{
   \title{Reallocating memory}
   \input{The pointer to a previously allocated memory array, obtained by |ZENMalloc| , and the new size needed.}
   \output{A pointer on the newly allocated array, or |NULL| if an error occured.}
   \side{The values stored in the first array designed by |oldptr| are copied in the new array.}
   \note{In the whole scope of the library, file and line are set according to the line of the source file where the call is performed. This can be used to track memory leaks and buffer overflows. By default, a call to the native |realloc| function is performed and file and line are just ignored. This can be overwritten by |ZENSetMemoryFunctions|.}
   } 
   */
void * ZENReallocFct(oldptr, newsize, file, line)
     void * oldptr;
     size_t newsize;
     char * file;
     int line;
{
  return(realloc(oldptr,newsize));
}
void * (*ZENReallocPtr) ___((void *, size_t, char *, int))  = ZENReallocFct;

/*
   \TEX{
   \title{Freeing memory}
   \input{A pointer to a previously allocated memory array, obtained by |ZENMalloc|.}
   \side{The memory pointed to by |ptr| is freed.}
   \note{In the whole scope of the library, file and line are set according to the line of the source file where the call is performed. This can be used to track memory leaks and buffer overflows. By default, a call to the native |free| function is performed and file and line are just ignored. This can be overwritten by |ZENSetMemoryFunctions|.}
   } 
   */
void ZENFreeFct(ptr, file, line)
     void * ptr;
     char * file;
     int line;
{
  return(free(ptr));
}

void (*ZENFreePtr) ___((void *, char *, int))  = ZENFreeFct;

/*
   \TEX{
   \title{Setting custom memory allocation functions}
   \input{The three memory allocation functions performing |malloc|, |realloc|, and |free|.}
   \side{The internal memory allocation functions are customized}
   \note{By default, the standard |ZENMallocFct|, |ZENReallocFct|, and |ZENFreeFct|
   functions are used. This function should be called at the very
   beginning of a program.}
   } */
void ZENSetMemoryFunctions(ma,re,fr)
     void * (* ma) ___((size_t, char *, int));
     void * (* re) ___((void *, size_t, char *, int));
     void  (* fr) ___((void *, char *, int));
{
  if(ma)
    ZENMallocPtr=ma;
  if(re)
    ZENReallocPtr=re;
  if(fr)
    ZENFreePtr=fr;

}
#endif /* ZEN_MALLOC_TEST */
