/*
 * zen/zbn/c/div.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

#if USE_GMP
# include <gmp.h>
#endif

/* \TEX{
   \Subsection{Division}   
   } */

/* \TEX{
   \title{Division by a digit}
   \input{Two allocated |BigNum|s, a |BigNumLength| and a
   |BigNumDigit|.}
   \output{The remainder of the division $|n|/|d|$.}
   \side{The |BigNum| |(q,nl-1)| is filled with the quotient $\frac{|n|} {|d|}$.}
   \note{The value |d| must be greater than
   |ZBNDigit(n,nl-1)|.}
   } */
BigNumDigit ZBNDivideDigit(q, n, nl, d)
     BigNum q, n;
     BigNumLength nl;
     BigNumDigit d;
{
  BigNumDigit r;

#ifdef zbndiv_normalized
  if (d & (CST_BNDIGIT(1) << (BN_DIGIT_SIZE-1))) {
#endif
    nl--;
    r = n[nl];
    if (nl) {
      nl--;
      do {
	zbndiv(q[nl], r, r, n[nl], d);
	nl--;
      } while(nl >= 0);
    }
#ifdef zbndiv_normalized
  }
  else {
    nl--;
    r = n[nl];
    if (nl) {
      BigNumLength shift;

      shift = ZBNNumLeadingZeroBitsInDigit(d);
      d <<= shift;
      r <<= shift;
      nl--;
      r |= n[nl] >> (BN_DIGIT_SIZE-shift);
      if (nl) {
	do {
	  zbndiv(q[nl], r, r,
		  (n[nl] << shift) | (n[nl-1] >> (BN_DIGIT_SIZE-shift)),
		  d);
	  nl--;
	} while(nl);
      }
      zbndiv(q[nl], r, r, (n[nl] << shift), d);
      r >>= shift;
    }
  }
#endif
  return(r);
}

void ZBNDivideHeart(n, nl, d, dl, D)
     BigNum n, d;
     BigNumLength nl, dl;
     BigNumDigit D;
{
  BigNumDigit Q, r;
  BigNumLength pl;
  
  nl--;
  if (nl) {
    pl = nl-dl;    
    r = n[nl];
    nl--;
    do {
      if (n[nl+1] == D) {
	Q = ~CST_BNDIGIT(0);
      }
      else {
	zbndiv(Q, r, r, n[nl], D);
      }
      ZBNMultiplyDigit(n+pl, dl+1, d, dl, Q);
      while(n[nl+1] != Q) {
	ZBNSubtract(n+pl, dl+1, d, dl, CST_BNDIGIT(1));
	Q--;
      }
      r= n[nl];
      nl--;
      pl--;
    }
    while(pl >= 0);
  }
}

/* \TEX{
   \title{Division}
   \input{Two allocated |BigNum|s and their lengths.}
   \side{The quotient and the remainder of $|(n,nl)|/|(d,dl)|$ are
   stored in |n|. The |dl| least significant digits contains the
   remainder, whereas the $|nl|-|dl|$ most significant digits contains
   the quotient.}
   \note{One must have $|nl|>|dl|$ and the value |ZBNDigit(d,dl-1)| must be strictly greater than
   |ZBNDigit(n,nl-1)|.}
   } */
void ZBNDivide(n, nl, d, dl)
     BigNum n, d;
     BigNumLength nl, dl;
{
  BigNum p;
  BigNumDigit D;
  BigNumLength shift;

  if (dl == 0) {
    n[0] = 1%dl;
  }
  else if (dl == 1) {
    n[0] = ZBNDivideDigit(n+1, n, nl, d[0]);
  }
  else {
    shift=ZBNNumLeadingZeroBitsInDigit(d[dl-1]);
    ZBNShiftLeft(d,dl,shift);
    ZBNShiftLeft(n,nl,shift);
    
    D = d[dl-1];
    ZBNComplement(d, dl);
    ZBNAddCarry(d, dl, CST_BNDIGIT(1));

    ZBNDivideHeart(n, nl, d, dl, D);

    ZBNComplement(d, dl);
    ZBNAddCarry(d, dl, CST_BNDIGIT(1));
    
    ZBNShiftRight(d, dl, shift);
    ZBNShiftRight(n, dl, shift);
  }
}

