/*
 * zen/zbn/c/gcd.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

/* \TEX{
   \Subsection{Greatest Common Divisor}
   } */

/* \TEX{
   Two families of procedures are provided. The goal of the first family
   is the computation of greatest common divisor of integers and the goal
   of the second is the computation of extended gcd.
   } */
static int ZBNTwoDigitGCD ___((BigNum, BigNum, BigNum));

static void ZBNParEuc ___((BigNumDigit*, BigNumDigit*,
			   BigNumDigit*, BigNumDigit*,
			   BigNumDigit, BigNumDigit));
static void ZBNParLehmer ___((BigNumDigit *, BigNumDigit *,
			      BigNumDigit *,BigNumDigit *,
			      BigNum, BigNumLength,
			      BigNum, BigNumLength));
static void ZBNParBin ___((BigNumDigit *, BigNumDigit *,
			   BigNumDigit *, BigNumDigit *,
			   BigNum, BigNumLength,
			   BigNum, BigNumLength)); 
static void ZBNEuclidPCext ___((BigNum, BigNumLength*,
				BigNum, BigNumLength*, int*,
				BigNum, BigNumLength,
				BigNum, BigNumLength, int,
				BigNum, BigNumLength,
				BigNum, BigNumLength, int,
				 BigNum, BigNum));
static int ZBNEeaLehmer ___((BigNum, BigNumLength*,
			     BigNum, BigNumLength*,
			     BigNum, BigNumLength,
			     BigNum, BigNumLength));
static int ZBNEeaPlain ___((BigNum, BigNumLength*,
			    BigNum, BigNumLength*,
			    BigNum, BigNumLength,
			    BigNum, BigNumLength));
static int ZBNGcdPlain ___((BigNum, BigNumLength *,
			     BigNum, BigNumLength,
			     BigNum, BigNumLength));
static int ZBNGcdBinary ___((BigNum, BigNumLength *,
			     BigNum, BigNumLength,
			     BigNum, BigNumLength));

/* A certain number of option compilations can be found in
   |gcd.c|. These options should not be changed as they should improve
   performances in (almost) all cases. For completeness here are the
   options.  If |LEHMER| is set, lehmer algorithm is used to compute
   extended gcd and binary algorithm for usual gcd. Furthermore, if
   this file is compiled with |ZBN_DEBUG| set, the variable
   |_zbn_debug_level| gives the depth of debugging messages.  */
#ifndef LEHMER
# define LEHMER  1
#endif
#ifdef ZBN_DEBUG
int _zbn_debug_level = 30;
#endif
/* Is n greater than one */
#define ZBNIsGtOne(n, nl) ((((nl) == 1)  && ((n)[0] <= (BigNumDigit)1)) ? 0 : 1)

#define SMALL(a) (a < ((BigNumDigit)1 << (BN_DIGIT_SIZE>>1)))

/* q_ = a_/b_, r_ = a_ mod b_ */
#if defined(mips)
#define DIV(q_, r_, a_, b_) \
if(1){ \
  if (a_ >= (b_ << 1)) { \
    q_ = a_ / b_; \
    r_ = a_ % b_; \
  } \
  else { \
    q_ = 1; \
    r_ = a_ - b_; \
  } \
}else
#else
#define DIV(q_, r_, a_, b_) \
if(1){ \
  if (a_ >= (b_ << 1)) { \
    q_ = a_ / b_; \
    r_ = a_ - q_ * b_; \
  } \
  else { \
    q_ = 1; \
    r_ = a_ - b_; \
  } \
}else
#endif

/* Gt = abs(y*G1+s*x*G2) where s = +/- 1 */

#define ZBNILCOMB(Gt, Gtsize, G1, G1size, G2, G2size, y, x, s) \
if(1){ \
  if (s > 0) \
    ZBNILAdd(Gt, Gtsize, G1, G1size, G2, G2size, x, y); \
  else { \
    ZBNSetDigit(G1+G1size, (BigNumDigit)0); \
    ZBNILSubtract(Gt, Gtsize, G1, G1size+1, G2, G2size, x, y); \
    if ((Gtsize == G1size+2) &&  \
	(ZBNGetDigit(Gt+G1size+1) & ((BigNumDigit)1 << (BN_DIGIT_SIZE-1)))) { \
      ZBNComplement(Gt, Gtsize); \
      ZBNAddCarry(Gt, Gtsize, (BigNumDigit)1); \
    } \
  } \
}else

/* Ut := U2; Ut := x*Ut+y*U1; */
#define ZBNILAdd(Ut, Utsize, U1, U1size, U2, U2size, x, y) \
if(1){ \
  if (x > 1) { \
    Utsize = (U2size >= U1size ? U2size+1 : U1size+1); \
    ZBNSetToZero(Ut, Utsize); \
    ZBNMultiplyDigit(Ut, U2size+1, U2, U2size, x); \
  } \
  else if (x == 1) { \
    Utsize = (U2size > U1size ? U2size : U1size+1); \
    ZBNAssign(Ut, U2, U2size);  \
    ZBNSetToZero(Ut+U2size, Utsize-U2size); \
  } \
  else { \
    Utsize = U1size+1; \
    ZBNSetToZero(Ut, Utsize); \
  } \
   \
  if ((y > 1) && ZBNMultiplyDigit(Ut, Utsize, U1, U1size, y)) { \
    ZBNSetDigit(Ut+Utsize, (BigNumDigit)1); (Utsize)++; \
  } \
  if ((y == 1) && ZBNAdd(Ut, Utsize, U1, U1size, 0)) { \
    ZBNSetDigit(Ut+Utsize, (BigNumDigit)1); (Utsize)++; \
  }   \
}else

/* Ut := U2; Ut := -x*Ut + y*U1; U1+U1size-1 = 0; U1size >= U2size */
#define ZBNILSubtract(Ut, Utsize, U1, U1size, U2, U2size, x, y) \
if(1){  \
  Utsize = U1size+1; \
  if (x > 1) { \
    ZBNSetToZero(Ut, Utsize); \
    ZBNMultiplyDigit(Ut, U2size+1, U2, U2size, x); \
    ZBNComplement(Ut, Utsize); \
    ZBNAddCarry(Ut, Utsize, (BigNumDigit)1); \
  } \
  else if (x == 1) { \
    ZBNAssign(Ut, U2, U2size);  \
    ZBNSetToZero(Ut+U2size, Utsize-U2size); \
    ZBNComplement(Ut, Utsize); \
    ZBNAddCarry(Ut, Utsize, (BigNumDigit)1); \
  } \
  else { \
    Utsize = U1size+1; \
    ZBNSetToZero(Ut, Utsize); \
  } \
  if (y > 1) \
    ZBNMultiplyDigit(Ut, Utsize, U1, U1size, y); \
  else if (y == 1) \
    ZBNAdd(Ut, Utsize, U1, U1size, (BigNumDigit)0); \
}else

/* Gt = abs(y*G1+s*x*G2) and Vt = abs(s1*y*V1+s*s2*x*V2) with s = +/- 1 */
#define ZBNILCOMBext(Vt, Vtsize, Gt, Gtsize, st, V1, V1size, G1, G1size, s1, V2, V2size, G2, G2size, s2, y, x, s) \
if(1){ \
  st = s1;\
  if (s > 0) \
    ZBNILAdd(Gt, Gtsize, G1, G1size, G2, G2size, x, y); \
  else { \
    ZBNSetDigit(G1+G1size, (BigNumDigit)0); \
    ZBNILSubtract(Gt, Gtsize, G1, G1size+1, G2, G2size, x, y); \
    if ((Gtsize == G1size+2) &&  \
	((ZBNGetDigit(Gt+G1size+1) & ((BigNumDigit)1 << (BN_DIGIT_SIZE-1))))) { \
      ZBNComplement(Gt, Gtsize); \
      ZBNAddCarry(Gt, Gtsize, (BigNumDigit)1); \
      st = -st; \
    } \
  } \
  if (((s1 == s2) && (s < 0)) || \
      ((s1 != s2) && (s > 0))) { \
    ZBNSetDigit(V1+V1size, (BigNumDigit)0); \
    ZBNILSubtract(Vt, Vtsize, V1, V1size+1, V2, V2size, x, y); \
    if ((Vtsize == V1size+2) &&  \
	((ZBNGetDigit(Vt+V1size+1) & ((BigNumDigit)1 << (BN_DIGIT_SIZE-1))))) { \
      ZBNComplement(Vt, Vtsize); \
      ZBNAddCarry(Vt, Vtsize, (BigNumDigit)1); \
      st = -st; \
    } \
  } \
  else \
    ZBNILAdd(Vt, Vtsize, V1, V1size, V2, V2size, x, y); \
}else

#define ZBNShiftCommonGCDPC(digit, bit, A, Asize, B, Bsize, a, b) \
if(1){ \
  digit = bit = 0; \
  while (!(ZBNGetDigit(A) || ZBNGetDigit(B))) { \
    (digit)++; \
    (A)++; (Asize)--; \
    (B)++; (Bsize)--; \
  } \
  a = ZBNGetDigit(A); b = ZBNGetDigit(B); \
  while (((a & 1)==0) && ((b & 1)==0)) {  \
    a >>= 1;  \
    b >>= 1; \
    (bit)++; \
  } \
  ZBNShiftRight((A), Asize, bit); \
  ZBNShiftRight((B), Bsize, bit); \
} else

#define ZBNShiftGCDPC(digit, bit, U, Usize, s) \
if(1){ \
  digit = bit = 0; \
  if (!ZBNIsZero(U, Usize)) {\
  while (!ZBNGetDigit(U)) { \
    digit++; \
    U++; Usize--; \
  } \
  s = ZBNGetDigit(U); \
  while((s & 1)==0) { \
    s >>= 1; \
    bit++; \
  } \
  ZBNShiftRight(U, Usize, bit); \
  }\
}else


/* \TEX{
   \title{Integer gcd of two |BigNumDigit|s}
   \input{Two |BigNumDigit|s |a_in| and |b_in|.}
   \output{The gcd of |a_in| and  |b_in|.}
   \note{A binary algorithm is used.}
   } */
BigNumDigit ZBNDigitGCD(a_in, b_in)
     BigNumDigit a_in, b_in;
{
  BigNumDigit a, b, c;
  int shift_a, shift_b, shift_c;
  
  if(a_in > b_in) {
    a = a_in; b = b_in; 
  }
  else {
    a = b_in; b = a_in;
  }

  if (a == 0) return(0);
  if (b == 0) return(a);

  shift_a = 0;
  while((a & 1) == 0) {
    a = a >> 1;
    shift_a++;
  }
  shift_b = 0;
  while((b & 1) == 0) {
    b = b >> 1;
    shift_b++;
  }
  shift_c = ((shift_a < shift_b) ? shift_a : shift_b);

  if(a < b) {
    c = a; a = b; b = c;
  }

  for(;;) {
    c = a - b;
    if (!c)
      return(a << shift_c);
    while((c & 1) == 0) {
      c = c >> 1; };
    if(c > b)
      a = c;
    else {
      a = b; b = c;
    }
  }
}

/* \_____TEX{
   \title{Integer gcd of two double BigNumDigit.}
   \input{Two BigNum |Ain| and |Bin| of size 2.}
   \output{The size of the gcd of |Ain| and |Bin|}
   \side{|Cin| receives te gcd of |Ain| and |Bin|, it must have size 2.}
   \note{A binary algorithm is used. We can have |Ain| = |Bin| = |Cin|}
   }
   */
int ZBNTwoDigitGCD(Cin, Ain, Bin)
     BigNum Cin, Ain, Bin;
{
  BigNumDigit A[3], B[2], C[2], r, m;
  int shiftA, shiftB, shiftC;
  
  if ((Ain[1] > Bin[1]) || 
      ((Ain[1] == Bin[1]) && (Ain[0] > Bin[0]))) {
    A[0] = Ain[0]; A[1] = Ain[1];
    B[0] = Bin[0]; B[1] = Bin[1];
  }
  else {
    B[0] = Ain[0]; B[1] = Ain[1];
    A[0] = Bin[0]; A[1] = Bin[1];
  }

  if(A[1] == 0) {
    Cin[0] = ZBNDigitGCD(A[0], B[0]);
    return(1);
  }

  if(B[1] == 0) {
    if(B[0] == 0) {
      Cin[1] = A[1]; Cin[0] = A[0];
      return (2);
    }
    
    if (B[0] <= A[1]) {
      A[2] = 0;
      r = ZBNDivideDigit(C, A, 3, B[0]);
    }
    else 
      r = ZBNDivideDigit(C, A, 2, B[0]);
    Cin[0] = ZBNDigitGCD(r, B[0]);
    return(1);
  }

  /* now we know A and B are two words long */
  /* shifting */
  m = (BigNumDigit)1 << (BN_DIGIT_SIZE-1);
  shiftA = 0;
  while((A[0] & 1) == 0) {
    A[0] >>= 1;
    if (A[1] & 1)
      A[0] |= m;
    A[1] >>= 1;
    shiftA++;
  }
  shiftB = 0;
  while((B[0] & 1) == 0){
    B[0] >>= 1;
    if (B[1] & 1)
      B[0] |= m;
    B[1] >>= 1;
    shiftB++;
  };

  shiftC = ((shiftA < shiftB) ? shiftA : shiftB); /* min(shiftA,shiftB) */

  /* test again */
  if ((A[1] < B[1]) || 
      ((A[1] == B[1]) && (A[0] < B[0]))) {
    C[0] = A[0]; C[1] = A[1];
    A[0] = B[0]; A[1] = B[1];
    B[0] = C[0]; B[1] = C[1];
  }

  /* main loop */
  for(;;) { 

    /* see if already small */
    if(A[1] == 0) { 
      Cin[1] = 0;
      Cin[0] = ZBNDigitGCD(A[0], B[0]);
      ZBNShiftLeft(Cin, (BigNumLength)2, shiftC);
      if (Cin[1])
	return(2);
      else
	return(1);
    }

    /* A[1] != 0 */
    if(B[1] == 0) {
      if (B[0] <= A[1]) {
	A[2] = 0;
	r = ZBNDivideDigit(C, A, 3, B[0]);
      }
      else 
	r = ZBNDivideDigit(C, A, 2, B[0]);
      Cin[1] = 0;
      Cin[0] = ZBNDigitGCD(r, B[0]);
      ZBNShiftLeft(Cin, (BigNumLength)2, shiftC);
      if (Cin[1])
	return(2);
      else
	return(1);
    }
  
    /* subtract and shift */
    C[0] = A[0]-B[0];
    C[1] = A[1]-B[1];
    if (C[0] > A[0])
      C[1]--;

    if((C[0] == 0) && (C[1] == 0)) {
      Cin[1] = A[1]; Cin[0] = A[0];
      ZBNShiftLeft(Cin, (BigNumLength)2, shiftC);
      if (Cin[1])
	return(2);
      else
	return(1);
    }
    while((C[0] & 1) == 0) {
      C[0] >>= 1;
      if (C[1] & 1)
	C[0] |= m;
      C[1] >>= 1;
    }

    if ((C[1] > B[1]) || 
	((C[1] == B[1]) && (C[0] > B[0]))) {
      A[0] = C[0]; A[1] = C[1];
    }
    else {
      A[0] = B[0]; A[1] = B[1];
      B[0] = C[0]; B[1] = C[1];
    }
  }
}

#if 1
/* \TEX{
   \title{Gcd of two |BigNum|s}
   \input{Two allocated |BigNum|s |n| and |m| of size |nl| and |ml|, a
   |BigNum| |g| of allocated size greater than |nl| and |ml|.}
   \output{The real size of the gcd or |ZENERR| if an allocation error
   occured.}
   \side{|g| receives the gcd of |n| and |m|.}
   \note{One can have |g|=|n|=|m|.}
   } */
int ZBNGcd(g, p_gl, n, nl, m, ml)
     BigNum g, n, m;
     BigNumLength *p_gl, nl, ml;
{
#if LEHMER
  if(ZBNCompare(n, nl, m, ml) == BN_GT)
    return(ZBNGcdBinary(g, p_gl, n, nl, m, ml));
  else
    return(ZBNGcdBinary(g, p_gl, m, ml, n, nl));
#else
  return(ZBNGcdPlain(g, p_gl, n, nl, m, ml));
#endif
}

/* \_____TEX{
   \title{Gcd of two BigNum with the Euclid's algorithm.}
   \input{Two BigNum |a| and |b| of size |al| and |bl|, a BigNum |g|
   of allocated size greater than |al| and |bl|.}
   \output{The real size of the gcd or |ZENERR| if an allocation error
   occured.}
   \side{|g| receives the gcd of |a| and |b|.}
   \note{We can have |g|=|a|=|b|.}
   } */
static int ZBNGcdPlain(g, p_gl, a, al, b, bl)
     BigNum g, a, b;
     BigNumLength *p_gl, al, bl;
{
  BigNumLength tmpl, nl, ml;
  BigNum tmp, n, m;

  if (ZBNCompare(a, al, b, bl) == BN_GT) {
    if (!(n = ZBNC(al+1))) {
      ZENSetError(ZEN_LIB_ZBZ, ZEN_GCDPLAIN, ZEN_ALLOC);
      return(ZERR);
    }
    if (!(m = ZBNC(bl+1))) {
      ZBNF(n);
      ZENSetError(ZEN_LIB_ZBZ, ZEN_GCDPLAIN, ZEN_ALLOC);
      return(ZERR);
    }
    nl = al; ZBNAssign(n, a, nl);
    ml = bl; ZBNAssign(m, b, ml);
  }
  else {
    if (!(m = ZBNC(al+1))) {
      ZENSetError(ZEN_LIB_ZBZ, ZEN_GCDPLAIN, ZEN_ALLOC);
      return(ZERR);
    }
    if (!(n = ZBNC(bl+1))) {
      ZBNF(m);
      ZENSetError(ZEN_LIB_ZBZ, ZEN_GCDPLAIN, ZEN_ALLOC);
      return(ZERR);
    }
    
    ml = al; ZBNAssign(m, a, ml);
    nl = bl; ZBNAssign(n, b, nl);
  }
  
  /* ordinary euclid */
  while(!ZBNIsZero(m, ml)){
#ifdef ZBN_DEBUG
    if(_zbn_debug_level >= 10){
      printf("m="); ZBNPrintToFile(stdout,m, ml, 10); printf("\n");
      printf("n="); ZBNPrintToFile(stdout,n, nl, 10); printf("\n");
    }
#endif
    /* division of n[0..nl-1][0] by m[0..ml-1] */
    ZBNSetDigit((n+nl), 0);
    ZBNDivide(n, nl+1, m, ml);
    tmpl = ZBNNumDigits(n, ml);
    nl = ml; ml = tmpl; tmp = n;
    n = m; m = tmp;
  }
  ZBNAssign(g, n, nl);
  ZBNF(n); ZBNF(m);
  
#ifdef ZBN_DEBUG
  if(_zbn_debug_level >= 10){printf("g="); ZBNPrintToFile(stdout,g, nl, 10); printf("\n");}
#endif
  *p_gl = nl;
  return(ZBN_HAS_INVERSE);
}

/* \_____TEX{
   \title{Gcd of two BigNum with the Binary algorithm.}
   \input{Two BigNum |A_| and |B_| of size |A_size| and |B_size| such
   that |A_|$>$|B_|, a BigNum |C_| of allocated size greater than
   |B_size|.} 
   \output{The real size of the gcd or |ZENERR| if an allocation error
   occured.}
   \side{|g| receives the gcd of |A_| and |B|.}
   \note{We can have |C-|=|A_|=|B_|.}
   } */
static int ZBNGcdBinary(C_, p_C_size, A_, A_size, B_, B_size)
     BigNum C_, A_, B_;
     BigNumLength *p_C_size, A_size, B_size;
{
  BigNum A, Aadr, B, Badr, U, Uadr, V, Vadr, S;
  int com_bit, com_digit, bit, digit;
  BigNumLength s, Asize, Bsize, Usize, Vsize;
  BigNumDigit t, x, y, xp, yp;
#ifdef ZBN_DEBUG
  int lenA, lenB;
#endif

  /* Get inputs 
   *************/
  Asize = ZBNNumDigits (A_,A_size);
  if (!(Aadr = A = ZBNC(Asize<<1))) {
    ZENSetError(ZEN_LIB_ZBZ, ZEN_GCDBINARY, ZEN_ALLOC);
    return(ZERR);
  }
  ZBNAssign(A, A_, Asize); ZBNSetToZero(A+Asize, Asize);
  Bsize = ZBNNumDigits (B_,B_size);
  if (!(Badr = B = ZBNC(Asize<<1))) {
    ZBNF(A);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_GCDBINARY, ZEN_ALLOC);
    return(ZERR);
  }
  ZBNAssign(B, B_, Bsize); ZBNSetToZero(B+Bsize, (Asize<<1)-Bsize);
  if (!(Uadr = U = ZBNC(Asize<<1))) {
    ZBNF(A); ZBNF(B);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_GCDBINARY, ZEN_ALLOC);
    return(ZERR);
  }
  if (!(Vadr = V = ZBNC(Asize<<1))) {
    ZBNF(A); ZBNF(B); ZBNF(U);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_GCDBINARY, ZEN_ALLOC);
    return(ZERR);
  }

  /* shift A and B and get common zeroes 
   **************************************/
  ZBNShiftCommonGCDPC(com_digit, com_bit, A, Asize, B, Bsize, x, y);
  Asize = ZBNNumDigits(A, Asize);
  Bsize = ZBNNumDigits(B, Bsize);
  
#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 20)
    printf("com_bit := %d; com_digit := %d\n");
#endif

  /* main loop  
   ************/
  for(;;) {

#ifdef ZBN_DEBUG
    if (_zbn_debug_level >= 20) {
      fprintf(stderr, "Before shifting\nA := ");
      ZBNPrintToFile(stderr, A, Asize, 10); fprintf(stderr, "; B := "); 
      ZBNPrintToFile(stderr, B, Bsize, 10); fprintf(stderr, ";\n");
    }
#endif

    /* shift A and B */
    ZBNShiftGCDPC(digit, bit, A, Asize, t);
    ZBNShiftGCDPC(digit, bit, B, Bsize, t);
    Asize = ZBNNumDigits(A, Asize);
    Bsize = ZBNNumDigits(B, Bsize);

    /* see if small */
    if (Asize <= 2) {
      BigNumDigit A2[2], B2[2];

      if (ZBNIsZero(A, Asize)) {
	Usize = Bsize; ZBNAssign(U, B, Usize);
	break;
      }
      if (Asize == 1) {
	ZBNSetDigit(B+Bsize, 0);
	x = ZBNDivideDigit(U, B, Bsize+1, ZBNGetDigit(A));
	ZBNSetDigit(U, ZBNDigitGCD(x, ZBNGetDigit(A)));
	Usize = 1;
	break;
      }
      A2[1] = ZBNGetDigit(A+1);
      A2[0] = ZBNGetDigit(A);
      if (Bsize > 2) {
	ZBNSetDigit(B+Bsize, 0);
	ZBNDivide(B, Bsize+1, A2, 2);
      }
      B2[0] = ZBNGetDigit(B); B2[1] = ZBNGetDigit(B+1);
      Usize = ZBNTwoDigitGCD(U, A2, B2);
      break;
    }

    if (Bsize <= 2) { 
      BigNumDigit A2[2], B2[2];

      if (ZBNIsZero(B, Bsize)) {
	Usize = Asize; ZBNAssign(U, A, Usize);
	break;
      }
      if (Bsize == 1) {
	ZBNSetDigit(A+Asize, 0);
	x = ZBNDivideDigit(U, A, Asize+1, ZBNGetDigit(B));
	ZBNSetDigit(U, ZBNDigitGCD(x, ZBNGetDigit(B)));
	Usize = 1;
	break;
      }
      B2[1] = ZBNGetDigit(B+1);
      B2[0] = ZBNGetDigit(B);
      if (Asize > 2) {
	ZBNSetDigit(A+Asize, 0);
	ZBNDivide(A, Asize+1, B2, 2);
      }
      A2[0] = ZBNGetDigit(A); A2[1] = ZBNGetDigit(A+1);
      Usize = ZBNTwoDigitGCD(U, A2, B2);
      break;
    } 

#ifdef ZBN_DEBUG
    if (_zbn_debug_level >= 10) {
      static int len_A, len_B;

      lenA = BN_DIGIT_SIZE*Asize-ZBNNumLeadingZeroBitsInDigit(ZBNGetDigit(A+Asize-1));
      lenB = BN_DIGIT_SIZE*Bsize-ZBNNumLeadingZeroBitsInDigit(ZBNGetDigit(B+Bsize-1));
      fprintf(stderr, "lenA := %d (%d); lenB := %d (%d);\n",
	      lenA, len_A-lenA, lenB, len_B-lenB);
      len_A = lenA; len_B = lenB; 
    }
    if (_zbn_debug_level >= 15) {
      fprintf(stderr, "A := "); 
      ZBNPrintToFile(stderr, A, Asize, 10); fprintf(stderr, "; B := "); 
      ZBNPrintToFile(stderr, B, Bsize, 10); fprintf(stderr, ";\n");
    }
#endif

    /* Force B to be the smaller */
    if (ZBNCompare(A, Asize, B, Bsize) == BN_LT) {
      S = A; A = B; B = S;
      s = Asize; Asize = Bsize; Bsize = s;
    }
    
    if (Asize > Bsize) { 
      
      /* reduce by euclidian division */
      ZBNSetDigit(A+Asize, 0);
      ZBNDivide(A, Asize+1, B, Bsize);
      Asize = Bsize;
      continue;
    }
    
#ifdef ZBN_DEBUG
    if (_zbn_debug_level >= 10)
      fprintf(stderr, "# BY CONJUGATE\n");
#endif

    ZBNParBin(&xp, &yp, &x, &y, A, Asize, B, Bsize);
    ZBNILCOMB(U, Usize, A, Asize, B, Bsize, xp, yp, -1); 
    ZBNILCOMB(V, Vsize, A, Asize, B, Bsize, x, y, -1); 
    Asize = Usize; S = A; A = U; U = S;
    Bsize = Vsize; S = B; B = V; V = S;

  } /* end of main loop */

  /* the GCD is U * 2^comzeroes */
#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 10) {
    fprintf(stderr, "BisGCD := "); 
    ZBNPrintToFile(stderr, U, Usize, 10);
    fprintf(stderr, ";\n");
  }
#endif

  /* restore common null digits */
  ZBNSetToZero(C_, com_digit);
  
  /* and then common null bits */
  ZBNSetDigit(U+Usize, 0);
  ZBNShiftLeft(U, Usize+1, com_bit);
  Usize = ZBNNumDigits(U, Usize+1);
  ZBNAssign(C_+com_digit, U, Usize);
  Usize += com_digit;

#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 10) {
    fprintf(stderr, "GCD := "); 
    ZBNPrintToFile(stderr, U, Usize, 10);
    fprintf(stderr, ";\n");
  }
#endif

  ZBNF(Aadr); ZBNF(Badr); 
  ZBNF(Uadr); ZBNF(Vadr);
  *p_C_size = Usize;
  return(ZBN_HAS_INVERSE);
}
#else
/* \_____TEX{
   \title{Gcd of two BigNum.}
   \input{Two BigNum |n| and |m| of size |nl| and |ml|, a BigNum |g|
   of allocated size greater than |nl| and |ml|.}
   \output{The real size of the gcd or |ZENERR| if an allocation error occured.}
   \side{|g| receives the gcd of |n| and|m|.}
   \note{We can have |g|=|n|=|m|.}
   } */
int ZBNGcd(g,p_gl,n,nl,m,ml)
     BigNum g,n,m;
     BigNumLength *p_gl,nl,ml;
{
  BigNum nbuf,mbuf;
  int d2;

  if(!(nl && ml))
    return(ZBN_HAS_INVERSE);

  d2=0;
  while((d2<nl)&&(d2<ml)
	&&(ZBNDigit(n,d2)==0)&&(ZBNDigit(m,d2)==0))
    d2++;
  
  nbuf=ZBNC(nl-d2);
  mbuf=ZBNC(ml-d2);
  if((nbuf==NULL)||(mbuf==NULL))
    {
      ZENSetError(ZEN_LIB_ZBN, ZEN_GCD, ZEN_ALLOC);
      return(ZERR);
    }
  nl -= d2;
  ml -= d2;
  ZBNAssign(nbuf,n+d2,nl);
  ZBNAssign(mbuf,m+d2,ml);
  d2 *= SIZE_BLOC;
  while((ZBNDigit(nbuf,0)&((BigNumDigit)1)==0)
	&&(ZBNDigit(mbuf,0)&((BigNumDigit)1)==0))
    {
      d2++;
      ZBNShiftRight(nbuf,nl,1);
      ZBNShiftRight(mbuf,ml,1);
    }
  while(ZBNDigit(nbuf,0)&((BigNumDigit)1)==0)
    ZBNShiftRight(nbuf,nl,1);
  while(ZBNDigit(mbuf,0)&((BigNumDigit)1)==0)
    ZBNShiftRight(mbuf,ml,1);
  nl=ZBNNumDigits(nbuf,nl);
  ml=ZBNNumDigits(mbuf,ml);
  
  switch(ZBNCompareLazy(nbuf,nl,mbuf,ml))
    {
    case BN_GT:
      *p_gl=mpn_gcd(g,mbuf,ml,nbuf,nl);
      break;
    case BN_LT:
      *p_gl=mpn_gcd(g,nbuf,nl,mbuf,ml);
      break;
    case BN_EQ:
      *p_gl=nl;
      ZBNAssign(g,nbuf,nl);
      break;
    }

  ZBNF(nbuf);
  ZBNF(mbuf);
  ZBNAnyShiftLeft(g,p_gl,g,*p_gl,d2);  
  return(ZBN_HAS_INVERSE);
}


#endif

/* \_____TEX{
   \title{Partial cofactors for improved binary GCD algorithm.}
   \input{Four pointers |u0_|, |v0_|, |u1_|, |v1_| on BigNumDigits, two
   integers (|A|, |A_size|), (|B_|, |B_size|).}
   \side{|u0_|, |v0_|, |u1_| and |v1_| are filled with integers such that
   |*u0_||A_|+|*v0_||B_| and |*u1_||A_|+|*v1_||B_| are smaller as
   much as possible than |A_| and |B_|.} 
   \note{Several elementary steps of the binary algorithm are gathered
   together to obtain |u0_|, |v0_|, |u1_| and |v1_|.}
   } */
static void ZBNParBin(u0_, v0_, u1_, v1_, A_, A_size, B_, B_size)
     BigNumDigit *u0_, *v0_, *u1_, *v1_;
     BigNum A_, B_;
     BigNumLength A_size, B_size; 
{ 
  BigNumDigit A[2], B[2];
  BigNumDigit u0, v0, u1, v1;
  int k0, k1, shift, KA, KB;

  shift = ZBNNumLeadingZeroBitsInDigit(ZBNGetDigit(A_+A_size-1));
  if (shift) {
    A[1] = (A_[A_size-1]<<shift) ^ (A_[A_size-2]>>(BN_DIGIT_SIZE-shift));
    B[1] = (B_[A_size-1]<<shift) ^ (B_[A_size-2]>>(BN_DIGIT_SIZE-shift));
  }
  else {
    A[1] = A_[A_size-1];
    B[1] = B_[B_size-1];
  }
  A[0] = A_[0];
  B[0] = B_[0];

#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 10) {
    printf("A := "); ZBNPrintToFile(stdout,A, 2, 10); printf(";\n");
    printf("B := "); ZBNPrintToFile(stdout,B, 2, 10); printf(";\n");
  }
#endif

  KA = BN_DIGIT_SIZE; KB = BN_DIGIT_SIZE;
  u0 = 1; v0 = 0;
  u1 = 0; v1 = 1;
  k0 = k1 = 0;

  while(!(B[0] & 1) && KB) {
    B[0] >>= 1; B[1] >>= 1;
    KB--;
    k0++; 
  }


  while(!(A[0] & 1) && KA) {
    k1++; 
    A[0] >>= 1; A[1] >>= 1;
    KA--;
  }

  for(; A[1] && B[1];) {
#ifdef ZBN_DEBUG
    if (_zbn_debug_level >= 10)
      printf("A := %lu*2^32+%lu; B := %lu*2^32+%lu;\n", A[1], A[0], B[1], B[0]);
#endif

    if ((A[1] > B[1]) || 
	((A[1] == B[1]) && (A[0] > B[0]))) {
      u0 = (u0 << k0) + (u1 << k1);
      v0 = (v0 << k0) + (v1 << k1);
#ifdef ZBN_DEBUG
      if (_zbn_debug_level >= 10)
	printf("(%lu*A-%lu*B)/2^1; #A>B\n", u0, v0);
#endif

      A[1] -= B[1]; A[0] -= B[0];
      k0 = 0;
      while(!(A[0] & 1) && KA) {
        k1++; 
	A[0] >>= 1; A[1] >>= 1;
	KA--;
      }
    }
    else { 

      u1 = (u0 << k0) + (u1 << k1);
      v1 = (v0 << k0) + (v1 << k1);
#ifdef ZBN_DEBUG
      if (_zbn_debug_level >= 10)
	printf("(%lu*A-%lu*B)/2^1; #A<B\n", u1, v1);
#endif

      B[1] -= A[1]; B[0] -= A[0];
      k1 = 0;
      while(!(B[0] & 1) && KB) {
	B[0] >>= 1; B[1] >>= 1;
	KB--; k0++; 
      }
    }
  } 
  (*u0_) = u0; (*v0_) = v0;
  (*u1_) = u1; (*v1_) = v1;
}


/* \TEX{
   \title{Extended gcd of two |BigNum|s}
   \input{Two allocated |BigNum|s (|n|, |nl|) and (|m|, |ml|) with |n|$\ge$|m| and
   |nl|$\ge$|ml|, two |BigNum|s |a| and |g| of allocated size
   $\ge$ |nl|.}
   \output{If the gcd is equal to 1, |ZBN_HAS_INVERSE| is returned and
   |a| is equal to $1/$|m| mod |n|, if the gcd is greater than
   one, |ZBN_NO_INVERSE| is returned and |g| is the
   gcd. When an allocation error occured, |ZENERR| is returned.}
   \note{This porcedure is a front head for |ZBNEeaLehmer| if |LEHMER| is
   set at the compilation or |ZBNEeaPlain| otherwise.}
   } */
int ZBNEea(g, p_gl, a, p_al, n, nl, m, ml)
     BigNum g, a, n, m;
     BigNumLength *p_gl, *p_al, nl, ml;
{
#if LEHMER
  return(ZBNEeaLehmer(g, p_gl, a, p_al, n, nl, m, ml));
#else
  return(ZBNEeaPlain(g, p_gl, a, p_al, n, nl, m, ml));
#endif
}

/* \_____TEX{
   \title{Extended gcd of two BigNum by the Euclid algorithm.}
   \input{Two BigNum (|n0|, |nl|) and (|m0|, |ml|) with |n0|$\ge$|m0| and
   |nl| $\ge$ |ml|, two BigNum |a| and |g| of allocated size
   $\ge$ |nl|.}
   \output{If the gcd is equal to 1, |ZBN_HAS_INVERSE| is returned and
   |a| is equal to $1/$|m0| mod |n0|, if the gcd is greater than
   one, |ZBN_NO_INVERSE| is returned and |g| is the
   gcd. When an allocation error occured, |ZENERR| is returned.}
   } */
static int ZBNEeaPlain(g, p_gl, a, p_al, n0, nl, m0, ml)
     BigNum g, a, n0, m0;
     BigNumLength *p_gl, *p_al, nl, ml;
{
  BigNum n, m, aa, b, gg;
  BigNum tm, ggadr;
  int d = 1, tmp, inv;
  BigNumLength bl = 1, al = 1, ql, ggl = nl;

  if (!(n = ZBNC(nl+1))) {
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEAPLAIN, ZEN_ALLOC);
    return(ZERR);
  }
  if (!(m = ZBNC(ml+1))) {
    ZBNF(n);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEAPLAIN, ZEN_ALLOC);
    return(ZERR);
  }
  if (!(aa = ZBNC(nl+1))) {
    ZBNF(n); ZBNF(m);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEAPLAIN, ZEN_ALLOC);
    return(ZERR);
    
  }
  if (!(b = ZBNC(nl+1))) {
    ZBNF(n); ZBNF(m); ZBNF(aa);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEAPLAIN, ZEN_ALLOC);
    return(ZERR);
  }
  if (!(gg = ZBNC(nl+1))) {
    ZBNF(n); ZBNF(m); ZBNF(aa); ZBNF(b);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEAPLAIN, ZEN_ALLOC);
    return(ZERR);
  }
  
  ZBNAssign(n, n0, nl);
  ZBNAssign(m, m0, ml);
  
#ifdef ZBN_DEBUG
  if(_zbn_debug_level >= 15){
    printf("m="); ZBNPrintToFile(stdout,m, ml, 10); printf("\n");
    printf("n="); ZBNPrintToFile(stdout,n, nl, 10); printf("\n");
  }
#endif
  ZBNSetDigit(aa, 1);
  ZBNAssign(gg, n, nl);
  ggadr = gg;
  while(ZBNIsGtOne(m, ml)){
    d = 0 - d;
    /* division of gg[0..ggl-1][0] by m[0..ml-1] */
    ZBNSetDigit((gg+ggl), 0);
    /* gg = q * m + r = gg[0..ml-1][ml..ggl] */
    ZBNDivide(gg, ggl+1, m, ml);
    /* accumulation in aa */
    ql = 1 + ggl - ml;
    bl = al + ql;
    /* b <- b + q * aa */
    ZBNMultiply(b, bl, aa, al, (gg+ml), ql);
    tmp = ZBNNumDigits(gg, ml); ggl = ml; ml = tmp;
    tmp = ZBNNumDigits(b, bl); bl = al; al = tmp;
    tm = gg; gg = m; m = tm;
    tm = b; b = aa; aa = tm;
  }
  /* m[0]=0 ou 1 */
  inv = ZBN_NO_INVERSE;
  if(ZBNGetDigit(m) == 1){
    /* there is an inverse */
    inv = ZBN_HAS_INVERSE;
    ZBNSetDigit(g, 1);
    ggl = 1;
    if(d == -1){
      ZBNComplement(aa, nl);
      ZBNAdd(aa, nl, n, nl, (BigNumDigit)1);
      al = ZBNNumDigits(aa, nl);
    }
#ifdef ZBN_DEBUG
    if(_zbn_debug_level >= 15)
      {printf("1/m mod n="); ZBNPrintToFile(stdout,aa, al, 10); printf("\n");}
#endif
  }
  else{
    /* gg contains the gcd of n and m */
    ZBNAssign(g, gg, ggl);
  }
    *p_gl = ggl; *p_al = al;
  ZBNAssign(a, aa, al);
  ZBNF(n); ZBNF(m);
  ZBNF(aa); ZBNF(b); ZBNF(ggadr);
  return(inv);
}

/* \_____TEX{
   \title{Extended gcd of two BigNum by the Lehmer algorithm.}
   \input{Two BigNum (|A_|, |A_size|) and (|B_|, |B_size|) with
   |A_|$\ge$|B_|, |A_size| $\ge$ | B_size|, and
   two BigNum |G_| and |V_| of allocated size $\ge$
   |A_size|.}
   \output{If the gcd is equal to 1, |ZBN_HAS_INVERSE| is returned and
   |V_| is equal to $1/$|B_| mod |A_|, if the gcd is greater than
   one, |ZBN_NO_INVERSE| is returned and |G_| is the
   gcd. When an allocation error occured, |ZENERR| is returned.}
   \side{|p_Gsize| and |p_V_size| receive the size of |G_| and |V_| at
   the end of the procedure.}    
   } */
static int ZBNEeaLehmer(G_, p_G_size, V_, p_V_size, A_, A_size, B_, B_size)
     BigNum G_, V_, A_, B_;
     BigNumLength *p_G_size, *p_V_size, A_size, B_size;
{
  /* Buffers */
  BigNum A, Aadr, B, Badr, V1, V2, Vt, Vs, G1, G2, Gt, Gs, S;
  BigNumLength Asize, Bsize;
  BigNumLength V1size, V2size, Vtsize, Vssize;
  BigNumLength G1size, G2size, Gtsize, Gssize;
  /* otherwise */
  BigNumDigit xp, yp, x, y;
  int s, s1, s2, st, ss, com_digit, com_bit;
  BigNumLength t;
#ifdef ZBN_DEBUG
  int lenG1, lenG2;
#endif

   /* get inputs 
   *************/
  Asize = A_size;
  if (!(Aadr = A = ZBNC(Asize))) {
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEALEHMER, ZEN_ALLOC);
    return(ZERR);
  }
  ZBNAssign(A, A_, Asize); 
  Bsize = B_size;
  if (!(Badr = B = ZBNC(Bsize))) {
    ZBNF(A);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEALEHMER, ZEN_ALLOC);
    return(ZERR);
  }
  ZBNAssign(B, B_, Bsize); 

#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 10) {
    fprintf(stderr, "\nExtended Binary GCD\nA_ := ");
    ZBNPrintToFile(stderr, A, Asize, 10);
    fprintf(stderr, ";\nB_ := ");
    ZBNPrintToFile(stderr, B, Bsize, 10);
    fprintf(stderr, ";\n");
  }
#endif
  
  /* shift A and B and get common zeroes 
   **************************************/
  ZBNShiftCommonGCDPC(com_digit, com_bit, A, Asize, B, Bsize, x, y);

#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 10) {
    fprintf(stderr, "A := ");
    ZBNPrintToFile(stderr, A, Asize, 10);
    fprintf(stderr, ";\nB := ");
    ZBNPrintToFile(stderr, B, Bsize, 10);
    fprintf(stderr, ";\n");
  }
#endif

  /* Initialization of the main buffers.
   **************************************/

  /* V1 = 1, G1 = A such that U1*A+V1*B = G1 */
  V1size = 1;
  if (!(V1 = ZBNC(2+(Asize<<1)))) {
    ZBNF(A); ZBNF(B);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEALEHMER, ZEN_ALLOC);
    return(ZERR);
  }
  ZBNSetDigit(V1, 0);
  G1size = Asize;
  if (!(G1 = ZBNC(Asize+3))) {
    ZBNF(A); ZBNF(B);
    ZBNF(V1);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEALEHMER, ZEN_ALLOC);
    return(ZERR);
  }
  ZBNAssign(G1, A, G1size);
  s1 = -1;

  /* V2 = 0, G2 = B such that U2*A+V2*B = G2 */
  V2size = 1;
  if (!(V2 = ZBNC(2+(Asize<<1)))) {
    ZBNF(A); ZBNF(B);
    ZBNF(V1); ZBNF(G1);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEALEHMER, ZEN_ALLOC);
    return(ZERR);
  }
  ZBNSetDigit(V2, 1);
  G2size = Bsize;
  if (!(G2 = ZBNC(Asize+3))) {
    ZBNF(A); ZBNF(B);
    ZBNF(V1); ZBNF(G1);
    ZBNF(V2);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEALEHMER, ZEN_ALLOC);
    return(ZERR);
  }
  ZBNAssign(G2, B, G2size);
  s2 = 1;

  /* We are trying now to find Vt such that Ut*At+Vt*B = GCD(A, B) ... */
  if (!(Vt = ZBNC(2+(Asize<<1)))) {
    ZBNF(A); ZBNF(B);
    ZBNF(V1); ZBNF(G1);
    ZBNF(V2); ZBNF(G2);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEALEHMER, ZEN_ALLOC);
    return(ZERR);
  }
  if (!(Gt = ZBNC(Asize+3))) {
    ZBNF(A); ZBNF(B);
    ZBNF(V1); ZBNF(G1);
    ZBNF(V2); ZBNF(G2);
    ZBNF(Vt);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEALEHMER, ZEN_ALLOC);
    return(ZERR);
  }
  if (!(Vs = ZBNC(2+(Asize<<1)))) {
    ZBNF(A); ZBNF(B);
    ZBNF(V1); ZBNF(G1);
    ZBNF(V2); ZBNF(G2);
    ZBNF(Vt); ZBNF(Gt);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEALEHMER, ZEN_ALLOC);
    return(ZERR);
  }
  if (!(Gs = ZBNC(Asize+3))) {
    ZBNF(A); ZBNF(B);
    ZBNF(V1); ZBNF(G1);
    ZBNF(V2); ZBNF(G2);
    ZBNF(Vt); ZBNF(Gt);
    ZBNF(Vs);
    ZENSetError(ZEN_LIB_ZBZ, ZEN_EEALEHMER, ZEN_ALLOC);
    return(ZERR);
  }

#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 10) {
    fprintf(stderr, "G1 := ");
    ZBNPrintToFile(stderr, G1, G1size, 10);
    fprintf(stderr, ";\n");
    fprintf(stderr, "V1 := ");
    ZBNPrintToFile(stderr, V1, V1size, 10);
    fprintf(stderr, ";\n");
    fprintf(stderr, "U1 := (");
    if (s1 > 0)
      fprintf(stderr, "-");
    fprintf(stderr, "G1+B*V1)/A;\n");
    
    fprintf(stderr, "G2 := ");
    ZBNPrintToFile(stderr, G2, G2size, 10);
    fprintf(stderr, ";\n");
    fprintf(stderr, "V2 := ");
    ZBNPrintToFile(stderr, V2, V2size, 10);
    fprintf(stderr, ";\n");
    fprintf(stderr, "U2 := (");
    if (s2 > 0)
      fprintf(stderr, "-");
    fprintf(stderr, "G2+B*V2)/A;\n");
  }
#endif

  /* main loop  
   ************/
  for(;;) {

    G1size = ZBNNumDigits(G1, G1size);
    G2size = ZBNNumDigits(G2, G2size);
    V1size = ZBNNumDigits(V1, V1size);
    V2size = ZBNNumDigits(V2, V2size);


#ifdef ZBN_DEBUG
    if (_zbn_debug_level >= 10) {
      fprintf(stderr, "G1 := ");
      ZBNPrintToFile(stderr, G1, G1size, 10);
      fprintf(stderr, ";\n");
      fprintf(stderr, "V1 := ");
      ZBNPrintToFile(stderr, V1, V1size, 10);
      fprintf(stderr, ";\n");
      fprintf(stderr, "U1 := (");
      if (s1 > 0)
	fprintf(stderr, "-");
      fprintf(stderr, "G1+B*V1)/A;\n");
      
      fprintf(stderr, "G2 := ");
      ZBNPrintToFile(stderr, G2, G2size, 10);
      fprintf(stderr, ";\n");
      fprintf(stderr, "V2 := ");
      ZBNPrintToFile(stderr, V2, V2size, 10);
      fprintf(stderr, ";\n");
      fprintf(stderr, "U2 := (");
      if (s2 > 0)
	fprintf(stderr, "-");
      fprintf(stderr, "G2+B*V2)/A;\n");
    }
    
    if (_zbn_debug_level >= 20) {
      static len1 = 0, len2 = 0;


      lenG1 = BN_DIGIT_SIZE*G1size-
	ZBNNumLeadingZeroBitsInDigit(ZBNGetDigit(G1+G1size-1));
      lenG2 = BN_DIGIT_SIZE*G2size-
	ZBNNumLeadingZeroBitsInDigit(ZBNGetDigit(G2+G2size-1));

      if (len1)
	fprintf(stderr, "lenG1 := %d (%d); lenG2 := %d; (%d)\n",
		lenG1, len1-lenG1, lenG2, len2-lenG2);
      else
	fprintf(stderr, "lenG1 := %d; lenG2 := %d\n", lenG1, lenG2);
	
      len1 = lenG1; len2 = lenG2;
    }
#endif

    /* Force G2 to be the smaller
     ******************************/
    if (ZBNCompare(G1, G1size, G2, G2size) == BN_LT) {
      S = G1; G1 = G2; G2 = S;
      t = G1size; G1size = G2size; G2size = t;
      S = V1; V1 = V2; V2 = S;
      t = V1size; V1size = V2size; V2size = t;
      s = s1; s1 = s2; s2 = s; 
    }
    
#ifdef ZBN_DEBUG
      if (_zbn_debug_level >= 15) {
      
	if (s1 > 0)
	  fprintf(stderr, "eq1 := -X*UF1 +Y*");
	else
	  fprintf(stderr, "eq1 := X*UF1 -Y*");
	ZBNPrintToFile(stderr, V1, V1size, 10);
	fprintf(stderr, " =\n");
	ZBNPrintToFile(stderr, G1, G1size, 10);
	fprintf(stderr, ";\n");
      
	if (s2 > 0)
	  fprintf(stderr, "eq2 := -X*UF2 +Y*");
	else
	fprintf(stderr, "eq2 := X*UF2 -Y*");
	ZBNPrintToFile(stderr, V2, V2size, 10);
	fprintf(stderr, " =\n");
	ZBNPrintToFile(stderr, G2, G2size, 10);
	fprintf(stderr, ";\n");
      }
#endif
      
    if (ZBNIsZero(G2, G2size)) {
      S = Gt; Gt = G1; G1 = S;
      Gtsize = G1size;
      S = Vt; Vt = V1; V1 = S;
      Vtsize = V1size;
      st = s1;
      break;
    }

    if (G1size <= 1) {
#ifdef ZBN_DEBUG
      if (_zbn_debug_level >= 20)
	fprintf(stderr, "\n#By partial euclidian cofactors\n");
#endif      
      ZBNParEuc(&xp, &yp, &x, &y, ZBNGetDigit(G1), ZBNGetDigit(G2));
      
      ZBNILCOMBext(Vt, Vtsize, Gt, Gtsize, st, 
		   V1, V1size, G1, G1size, s1, 
		   V2, V2size, G2, G2size, s2,
		   xp, yp, -1); 
#ifdef ZBN_DEBUG
      if (_zbn_debug_level >= 15) {
	fprintf(stderr, "eqT := %lu*eq1 - %lu*eq2;\n", xp, yp);
		  
	if (st > 0)
	  fprintf(stderr, "eqT-( -X*UT +Y*");
	else
	  fprintf(stderr, "eqT-( X*UT -Y*");
	ZBNPrintToFile(stderr, Vt, Vtsize, 10);
	fprintf(stderr, " =\n");
	ZBNPrintToFile(stderr, Gt, Gtsize, 10);
	fprintf(stderr, ");\n");
      }
#endif 

      break;
    }
      
    if (G1size == G2size) {
#ifdef ZBN_DEBUG
      if (_zbn_debug_level >= 20)
	fprintf(stderr, "\n#By Generalized Lehmer\n");
#endif
      ZBNParLehmer(&xp, &yp, &x, &y, G1, G1size, G2, G2size);
      
      if (yp) {
	ZBNILCOMBext(Vt, Vtsize, Gt, Gtsize, st, 
		     V1, V1size, G1, G1size, s1, 
		     V2, V2size, G2, G2size, s2,
		     xp, yp, -1); 
	ZBNILCOMBext(Vs, Vssize, Gs, Gssize, ss, 
		     V1, V1size, G1, G1size, s1, 
		     V2, V2size, G2, G2size, s2,
		     x, y, -1); 
#ifdef ZBN_DEBUG
	if (_zbn_debug_level >= 15) {
	  fprintf(stderr, "eqT := %lu*eq1 - %lu*eq2;\n", yp, xp);
	  if (st > 0)
	    fprintf(stderr, "eqT-( -X*UT +Y*");
	  else
	    fprintf(stderr, "eqT-( X*UT -Y*");
	  ZBNPrintToFile(stderr, Vt, Vtsize, 10);
	  fprintf(stderr, " =\n");
	  ZBNPrintToFile(stderr, Gt, Gtsize, 10);
	  fprintf(stderr, ");\n");

	  fprintf(stderr, "eqT := %lu*eq1 - %lu*eq2;\n", y, x);	  
	  if (ss > 0)
	    fprintf(stderr, "eqT-( -X*UT +Y*");
	  else
	    fprintf(stderr, "eqT-( X*UT -Y*");
	  ZBNPrintToFile(stderr, Vs, Vssize, 10);
	  fprintf(stderr, " =\n");
	  ZBNPrintToFile(stderr, Gs, Gssize, 10);
	  fprintf(stderr, ");\n");
	}
#endif 
	G1size = Gtsize; S = G1; G1 = Gt; Gt = S;
	V1size = Vtsize; S = V1; V1 = Vt; Vt = S;
	s1 = st;
	G2size = Gssize; S = G2; G2 = Gs; Gs = S;
	V2size = Vssize; S = V2; V2 = Vs; Vs = S;
	s2 = ss;
	continue;
      }
    }

#ifdef ZBN_DEBUG
    if (_zbn_debug_level >= 20)
      fprintf(stderr, "\n#By Euclid\n");
#endif
    
    /* reduce by euclid division */
    ZBNEuclidPCext(V1, &V1size, G1, &G1size, &s1,
		   V1, V1size, G1, G1size, s1, 
		   V2, V2size, G2, G2size, s2,
		   Gt, Vt);
  }

  /* We have G2 = GCD(A, B);
   * We are now able to find V_ and G_ such that A_*U_+B*V_ = G_.
   ***************************************************************/
#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 10) {
    fprintf(stderr, "\nGt := ");
    ZBNPrintToFile(stderr, Gt, Gtsize, 10);
    fprintf(stderr, ";\n");
    fprintf(stderr, "Vt := ");
    ZBNPrintToFile(stderr, Vt, Vtsize, 10);
    fprintf(stderr, ";\n");
    fprintf(stderr, "Ut := (");
    if (st > 0)
      fprintf(stderr, "-");
    fprintf(stderr, "Gt+B*Vt)/A;\n");
  }
#endif

  if (st < 0) {
    ZBNAssign(V1, A, Asize);
    t = Asize > Vtsize ? Asize : Vtsize;
    ZBNSetToZero(V1+Asize, t-Asize);
    ZBNSubtract(V1, t, Vt, Vtsize, 1);
    ZBNAssign(Vt, V1, t);
    Vtsize = t;
  }
  *p_V_size = ZBNNumDigits(Vt, Vtsize);
  ZBNAssign(V_, Vt, *p_V_size);

  /* restore common null digits */
  ZBNSetToZero(G_, com_digit);

  /* and then common null bits */
  ZBNSetDigit(Gt+Gtsize, 0);
  ZBNShiftLeft(Gt, Gtsize+1, com_bit);
  *p_G_size = ZBNNumDigits(Gt, Gtsize+1);
  ZBNAssign(G_+com_digit, Gt, *p_G_size);
  *p_G_size += com_digit;

  /* Free memory.
   **************/
  ZBNF(Aadr); ZBNF(Badr);
  ZBNF(V1); ZBNF(G1);
  ZBNF(V2); ZBNF(G2);
  ZBNF(Vt); ZBNF(Gt);
  ZBNF(Vs); ZBNF(Gs);

#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 10) {
    fprintf(stderr, "G_ := ");
    ZBNPrintToFile(stderr, G_, *p_G_size, 10);
    fprintf(stderr, ";\n");
    fprintf(stderr, "V_ := ");
    ZBNPrintToFile(stderr, V_, *p_V_size, 10);
    fprintf(stderr, ";\n");
    fprintf(stderr, "U_ := (-G_+B_*V_)/A_;\n");
  }
#endif
  if ((ZBNGetDigit(G_) == 1) && (*p_G_size == 1))
    return(ZBN_HAS_INVERSE);
  return(ZBN_NO_INVERSE);
}

/* \_____TEX{
   \title{Partial cofactors for improved Lehmer GCD algorithm.}
   \input{Four pointers |u0_|, |v0_|, |u1_|, |v1_| on BigNumDigits, two
   integers (|M_|, |M_size|), (|N_|, |N_size|). We must have
   |M_size|$\ge$2 and |M_size|=|N_size|}
   \side{|u0_|, |v0_|, |u1_| and |v1_| are filled with integers such that
   |*u0_||M_|+|*v0_||N_| and |*u1_||M_|+|*v1_||N_| are smaller as
   much as possible than |M_| and |N_|.} 
   \note{Several elementary steps of the Lehmer algorithm are gathered
   together to obtain |u0_|, |v0_|, |u1_| and |v1_|.}
   } */
static void ZBNParLehmer(u0_, v0_, u1_, v1_, M_, M_size, N_, N_size)
     BigNum M_, N_;
     BigNumLength M_size, N_size;
     BigNumDigit *u0_, *v0_, *u1_, *v1_;
{ 
  BigNumDigit M[2], N[2], S[3];
  BigNumDigit a, b, A, B, C, D, T1, Q1, T2, Q2;
  int shift, sgn;

  /* First step
   ************/
  shift = ZBNNumLeadingZeroBitsInDigit(ZBNGetDigit(M_+M_size-1));
  if (shift) {
    if (M_size <= 2) {
      M[1] = (M_[1]<<shift) ^ (M_[0]>>(BN_DIGIT_SIZE-shift));
      M[0] = (M_[0]<<shift);
      N[1] = (N_[1]<<shift) ^ (N_[0]>>(BN_DIGIT_SIZE-shift));
      N[0] = (N_[0]<<shift);
    }
    else {
      M[1] = (M_[M_size-1]<<shift) ^ (M_[M_size-2]>>(BN_DIGIT_SIZE-shift));
      M[0] = (M_[M_size-2]<<shift) ^ (M_[M_size-3]>>(BN_DIGIT_SIZE-shift));
      N[1] = (N_[M_size-1]<<shift) ^ (N_[M_size-2]>>(BN_DIGIT_SIZE-shift));
      N[0] = (N_[M_size-2]<<shift) ^ (N_[M_size-3]>>(BN_DIGIT_SIZE-shift));
    }
  }
  else {
    M[1] = M_[M_size-1]; M[0] = M_[M_size-2];
    N[1] = N_[M_size-1]; N[0] = N_[M_size-2];
  }

#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 20)
    fprintf(stderr, "First step\n");
#endif

  A = 1; B = 0; C = 0; D = 1;
  a = M[1]; b = N[1];

  sgn = 1;
  DIV(Q1, T1, a, b);
  if (!SMALL(T1)) {
    for(;;) {
      DIV(Q2, T2, b, T1);
      if (SMALL(T2)) {sgn = 1; break;}
      
      a = b; b = T1;
      T1 = A+Q1*C; A = C; C = T1;
      T1 = B+Q1*D; B = D; D = T1;
      T1 = T2; Q1 = Q2;
      
      DIV(Q2, T2, b, T1);
      if (SMALL(T2)) {sgn = -1; break;}

      a = b; b = T1;
      T1 = A+Q1*C; A = C; C = T1;
      T1 = B+Q1*D; B = D; D = T1;
      T1 = T2; Q1 = Q2;   
    }
  }

  if ((!B) || (N_size < 1)) {
    (*u0_) = A; (*v0_) = B; (*u1_) = C; (*v1_) = D;
    return;
  }

  /* Second step
   *************/
  if (sgn > 0) {
    S[0] = S[1] = S[2] = 0;
    ZBNMultiplyDigit(S, 3, N, 2, B);
    S[0] = ~S[0]; S[1] = ~S[1]; S[2] = ~S[2];
    if (!(++S[0]))
      if (!(++S[1]))
	++S[2];
    ZBNMultiplyDigit(S, 3, M, 2, A);
    shift = ZBNNumLeadingZeroBitsInDigit(S[1]);
    if (shift)
      a = (S[1]<<shift) ^ (S[0]>>(BN_DIGIT_SIZE-shift));
    else
      a = S[1];
    
    S[0] = S[1] = S[2] = 0;
    ZBNMultiplyDigit(S, 3, M, 2, C);
    S[0] = ~S[0]; S[1] = ~S[1]; S[2] = ~S[2];
    if (!(++S[0]))
      if (!(++S[1]))
	++S[2];
    ZBNMultiplyDigit(S, 3, N, 2, D);
    if (shift)
      b = (S[1]<<shift) ^ (S[0]>>(BN_DIGIT_SIZE-shift));
    else
      b = S[1];
  }
  else {
    S[0] = S[1] = S[2] = 0;
    ZBNMultiplyDigit(S, 3, M, 2, A);
    S[0] = ~S[0]; S[1] = ~S[1]; S[2] = ~S[2];
    if (!(++S[0]))
      if (!(++S[1]))
	++S[2];
    ZBNMultiplyDigit(S, 3, N, 2, B);
    shift = ZBNNumLeadingZeroBitsInDigit(ZBNGetDigit(S+1));
    if (shift)
      a = (S[1]<<shift) ^ (S[0]>>(BN_DIGIT_SIZE-shift));
    else
      a = S[1];
    
    S[0] = S[1] = S[2] = 0;
    ZBNMultiplyDigit(S, 3, N, 2, D);
    S[0] = ~S[0]; S[1] = ~S[1]; S[2] = ~S[2];
    if (!(++S[0]))
      if (!(++S[1]))
	++S[2];
    ZBNMultiplyDigit(S, 3, M, 2, C);
    if (shift)
      b = (S[1]<<shift) ^ (S[0]>>(BN_DIGIT_SIZE-shift));
    else
      b = S[1];
  }

#ifdef ZBN_DEBUG
    if (_zbn_debug_level >= 20)
	fprintf(stderr, "Second step\n");
#endif

  DIV(Q1, T1, a, b);
  if (!SMALL(T1)) {

    for(;;) {

      DIV(Q2, T2, b, T1);
      if (SMALL(T2)) {sgn = 1; break;}
      
      a = b; b = T1;
      T1 = A+Q1*C; A = C; C = T1;
      T1 = B+Q1*D; B = D; D = T1;
      T1 = T2; Q1 = Q2;
    }
  }

  (*u0_) = A; (*v0_) = B; (*u1_) = C; (*v1_) = D;

#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 20)
    fprintf(stderr, "A := %lu; B := -%lu; C := -%lu; D := %lu;\n",
	    *u0_, *v0_, *u1_, *v1_);
#endif
}


/* \_____TEX{
   \title{Partial cofactors for digits.}
   \input{Four pointers |u0_|, |v0_|, |u1_|, |v1_| on BigNumDigits,
   two BigNumDigits |a_| and |b_|.}
   \side{|u0_|, |v0_|, |u1_| and |v1_| are filled with integers such
   that |*u0_||a_|+|*v0_||a_| and |*u1_||a_|+|*v1_||b_| are smaller as
   much as possible than |a_| and |b_|.}
   \note{Several elementary steps of the Euclid's algorithm are
   gathered together to obtain |u0_|, |v0_|, |u1_| and |v1_|.}
   } */
static void ZBNParEuc(u0_, v0_, u1_, v1_, a_, b_)
     BigNumDigit *u0_, *v0_, *u1_, *v1_, a_, b_;
{ 
  BigNumDigit a, b, A, B, C, D, T, Q;

  a = a_; b = b_;
  A = 1; B = 0; C = 0; D = 1; 

  for(;;) {
    if (!b)
      break;

    DIV(Q, T, a, b);
    a = b; b = T;
    T = A+Q*C; A = C; C = T;
    T = B+Q*D; B = D; D = T;
  }

  (*u0_) = A; (*v0_) = B;
  (*u1_) = C; (*v1_) = D;
}


/* \____TEX{
   \title{Partial computation fo improved Lehmer GCD algorithm by
   euclidian division.}
   \note{Internal procedure used by |ZBNEeaLehmer|.}
   } */
static void ZBNEuclidPCext(Vt, p_Vtsize, Gt, p_Gtsize, p_st,
		 V1, V1size, G1, G1size, s1, 
		 V2, V2size, G2, G2size, s2,
		 R, Q)
     int s1, s2, *p_st;
     BigNum Vt, Gt, V1, G1, V2, G2, R, Q;
     BigNumLength *p_Vtsize, *p_Gtsize, V1size, G1size, V2size, G2size;
{
  int Rsize, Qsize;
 
#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 15) {
    if (s1 > 0)
      fprintf(stderr, "eq1 := -X*UF1 +Y*");
    else
      fprintf(stderr, "eq1 := X*UF1 -Y*");
    ZBNPrintToFile(stderr, V1, V1size, 10);
    fprintf(stderr, " = ");
    ZBNPrintToFile(stderr, G1, G1size, 10);
    fprintf(stderr, ";\n");
  }
#endif

#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 15) {
    if (s2 > 0)
      fprintf(stderr, "eq2 := -X*UF2 +Y*");
    else
      fprintf(stderr, "eq2 := X*UF2 -Y*");
    ZBNPrintToFile(stderr, V2, V2size, 10);
    fprintf(stderr, " = ");
    ZBNPrintToFile(stderr, G2, G2size, 10);
    fprintf(stderr, ";\n");
  }
#endif

  /* Gt := abs(G1 mod G2); */
  *p_Gtsize = G1size+1; ZBNAssign(Gt, G1, G1size);
  ZBNSetDigit(Gt+G1size, 0);
  ZBNDivide(Gt, *p_Gtsize, G2, G2size);
  Qsize = ZBNNumDigits(Gt+G2size, (*p_Gtsize)-G2size);
  ZBNAssign(Q, Gt+G2size, Qsize);
  *p_Gtsize = ZBNNumDigits(Gt, G2size);

  /* Vt := (s1 != s2 ) ? ( V1 + Q*V2 ) : abs(V1 - Q*V2); */
  if (s1 != s2) {
    *p_Vtsize = (V1size > V2size+Qsize ? V1size : V2size+Qsize);
    ZBNAssign(Vt, V1, V1size);
    ZBNSetToZero(Vt+V1size, *p_Vtsize-V1size);
    if (ZBNMultiply(Vt, *p_Vtsize, V2, V2size, Q, Qsize)) {
      ZBNSetDigit(Vt+*p_Vtsize, 1);
      (*p_Vtsize)++;
    }
    *p_st = s1;
  }
  else {
    Rsize = V2size+Qsize; ZBNSetToZero(R, Rsize);
    ZBNMultiply(R, Rsize, V2, V2size, Q, Qsize);
    *p_Vtsize = Rsize > V1size ? Rsize : V1size;

    if (ZBNCompare(V1, V1size, R, Rsize) != BN_LT) {
      ZBNAssign(Vt, V1, V1size);
      ZBNSetToZero(Vt+V1size, *p_Vtsize-V1size);
      ZBNSubtract(Vt, *p_Vtsize, R, Rsize, 1);
      *p_st = s1;
    }
    else {
      ZBNSubtract(R, *p_Vtsize, V1, V1size, 1);
      ZBNAssign(Vt, R, *p_Vtsize);
      *p_st = -s1;
    }
  }
  *p_Vtsize = ZBNNumDigits(Vt, *p_Vtsize);

#ifdef ZBN_DEBUG
  if (_zbn_debug_level >= 15) {
      fprintf(stderr, "eqT := eq1 - ");
    ZBNPrintToFile(stderr, Q, Qsize, 10);
    fprintf(stderr, "*eq2;\n");
    if (*p_st > 0)
      fprintf(stderr, "eqT-( -X*US +Y*");
    else
      fprintf(stderr, "eqT-( X*US -Y*");
    ZBNPrintToFile(stderr, Vt, *p_Vtsize, 10);
    fprintf(stderr, " = ");
    ZBNPrintToFile(stderr, Gt, *p_Gtsize, 10);
    fprintf(stderr, ");\n");
  }
#endif
}

