/*
 * zen/zbn/c/io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   Conversions between the internal representation and the ASCII
   representation of a |BigNum| are done by |ZBNPrintToString| and
   |ZBNReadFromString|.  |ZBNPrintToFile| and |ZBNReadFromFile| allows
   to write/read a |BigNum| in a file using these functions.
   
   One can moreover write to and then read from a file a |BigNum| in
   an internal format with |ZBNPutToFile| and |ZBNGetFromFile|.

   The |Lazy| forms of these functions assume that the |BigNum| is of
   known size and perform no allocation.

   The new version of |BigNum| has similar functionalities with
   |BnFromString|, |BnToString|, |BnFillString|, |BnPrintToFile| and
   |BnReadFromFile|.  Of course |GMP| has also such features with the
   |mpz_[inp$\vert$out]_[str$\vert$raw]| functions.  We however keep
   the following functions that were present in the first version of
   |ZEN|. This should ensure compatibility for reading data files
   written with programs that were using first version of \zen.

   \begin{note} These functions cannot manage correctly |BigNum|s of
   length larger than $2^{32}$. 
   } */
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "zbn.h"

#ifdef M_LN2
# undef M_LN2
#endif
#define M_LN2	                0.69314718055994530942

#ifdef M_LN10
# undef M_LN10
#endif
#define M_LN10	                2.30258509299404568402

/*
  \TEX{
  Alphanumeric symbols are used as digit when needed:
  }
  \CiteC
*/
static char *Digit = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
/* \EndCiteC 
  \TEX{
  When reading a number, case comparison is not performed.
   \end{note}
  }
*/
static double BnLog [ZBN_IO_MAX_BASIS+1] = 
{
  0, 
  0,				/* ln (1) */
  M_LN2,			/* ln (2) */
  1.098612,			/* ln (3) */
  1.386294,			/* ln (4) */
  1.609438,			/* ln (5) */
  1.791759,			/* ln (6) */
  1.945910,			/* ln (7) */
  2.079442,			/* ln (8) */
  2.197225,			/* ln (9) */
  M_LN10,			/* ln (10) */
  2.397895,			/* ln (11) */
  2.484907,			/* ln (12) */
  2.564949,			/* ln (13) */
  2.639057,			/* ln (14) */
  2.708050,			/* ln (15) */
  2.772588,			/* ln (16) */
  2.83321334405621608024,
  2.89037175789616469220,
  2.94443897916644046000,
  2.99573227355399099343,
  3.04452243772342299650,
  3.09104245335831585347,
  3.13549421592914969080,
  3.17805383034794561964,
  3.21887582486820074920,
  3.25809653802148204547,
  3.29583686600432907418,
  3.33220451017520392393,
  3.36729582998647402718,
  3.40119738166215537541,
  3.43398720448514624592,
  3.46573590279972654708,
  3.49650756146648023545,
  3.52636052461616138966,
  3.55534806148941367970,
  3.58351893845611000162	/* ln(36) */
};

#define CTOI(c)			(c >= '0' && c <= '9' ? c - '0' :	\
                                 c >= 'a' && c <= 'z' ? c - 'a' + 10:	\
                                 c >= 'A' && c <= 'Z' ? c - 'A' + 10:	\
                                 0)
#define isAdigit(c) (c >= '0' && c <= '9' ? 1 :	\
                     c >= 'a' && c <= 'z' ? 1 :	\
                     c >= 'A' && c <= 'Z' ? 1 :	\
                     0)

/* \TEX{
   \title{Converting to string}
   \input{An allocated |BigNum| |n| of length |nl| and a base
   |base|$\in \{2,\ldots, |ZBN_IO_MAX_BASIS|\}$.}
   \output{An allocated string representing |n| in base |base|, or
   |ZENNULL| if an error occured.}
   } */
char *ZBNPrintToString(n,nl,base)
     BigNumLength nl;
     int base;
     BigNum n;
{
  BigNum y, q, t;
  char *string, *s;
  int sl, sd;
  BigNumDigit r;

  if((base < 2) || (base > ZBN_IO_MAX_BASIS)) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_PRINTTOSTRING, ZEN_INPUT);
    return(ZNULL);
  }

  sd= (int)((M_LN2*BN_DIGIT_SIZE*nl)/BnLog[base]);

  sl = sd+3;

  if (!(string = (char *)ZENMalloc(sizeof(char)*sl))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_PRINTTOSTRING, ZEN_ALLOC);
    return(ZNULL);
  }
  if (!(y = ZBNC(nl+1))) {	
    ZENFree(string);
    ZENSetError(ZEN_LIB_ZBN, ZEN_PRINTTOSTRING, ZEN_ALLOC);
    return(ZNULL);
  }
  if (!(q = ZBNC(nl+1))) {
    ZENFree(string);
    ZBNF(y);
    ZENSetError(ZEN_LIB_ZBN, ZEN_PRINTTOSTRING, ZEN_ALLOC);
    return(ZNULL);
  }
  s = string + sl;
  *(--s) = '\0';
  
  ZBNSetToZero(y, nl+1);
  ZBNSetToZero(q, nl+1);
  ZBNAssign(y, n, nl);

  if (ZBNIsZero(n, nl)) {
    *--s = '0';
  }
  else {
    do {
      r = ZBNDivideDigit (q, y, nl+1, (BigNumDigit)base);
      --s;
      *s = Digit[r];
      t = q,  q = y,  y = t;
    } while (!ZBNIsZero(y, nl+1));
  }
  
  if ((sd = s-string) > 0) {
    while (s < string + sl) {
      *(s-sd) = *s;
      s++;
    }
  }
  ZBNF(y); ZBNF(q);
  return(string);
}

/* \TEX{
   \title{Converting from string}
   \input{An allocated |BigNum| and its length |nl| and
   a string |s| representing a |BigNum| in base |base|$\in
   \{2,\ldots, |ZBN_IO_MAX_BASIS|\}$.}
   \output{The number of character read if no error occured, |ZENERR|
   otherwise}
   \side{|n| is filled with |s|}
   } */
int ZBNReadFromStringLazy(m, ml, s, base)
     BigNum m;
     BigNumLength ml;
     char *s;
     int base;
{
  BigNum n, p, t;
  BigNumLength nl;
  char *begin, *end, *str, *ptr;

  /* Throw away any initial space */
  for(begin = s; isspace(*begin); begin++) ;
  
  /* Go at the end of the string */
  for(end = begin; isAdigit(*end); end++) ;

  if (!(str = (char *)ZENMalloc(sizeof(char)*(end-begin+1)))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMSTRING, ZEN_ALLOC);
    return(ZERR);
  }
  strncpy(str, begin, end-begin);
  *(str+(end-begin)) = '\0';

  nl = ml+1;
  
  if (!(n = ZBNC(nl))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMSTRING, ZEN_ALLOC);
    ZENFree(str);
    return(ZERR);
  }
  
  if (!(p = ZBNC(nl))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMSTRING, ZEN_ALLOC);
    ZENFree(str); ZBNF(n);
    return(ZERR);
  }

  ZBNSetToZero(n, nl);
  for (ptr = str;  *ptr;  ptr++) {
    ZBNSetToZero(p, nl);
    ZBNSetDigit(p, CTOI (*ptr));
    ZBNMultiplyDigit(p, nl, n, nl-1, (BigNumDigit)base);
    t = p;  p = n;  n = t;
  }  
  ZENFree(str);
  ZBNF(p);
  nl = ZBNNumDigits(n, nl);
  if (nl != ml) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMSTRING, ZEN_SYNTAX);
    ZBNF(n);
    return(ZERR);
  }
  ZBNAssign(m, n, ml);
  return(end-s);
}

/* \TEX{
   \title{Converting from string}
   \input{A pointer |p_n| on a non allocated |BigNum|, a pointer
   |p_nl| to a |BigNumLength| and a string |s| representing a |BigNum|
   in base |base|$\in \{2,\ldots, |ZBN_IO_MAX_BASIS|\}$.}
   \output{The number of character read if no error occured, |ZENERR|
   otherwise}
   \side{|p_n| is allocated and is filled with |s|}
   } */
int ZBNReadFromString(p_n, p_nl, s, base)
     BigNum *p_n;
     BigNumLength *p_nl;
     char *s;
     int base;
{
  BigNum n, p, t;
  BigNumLength nl;
  char *begin, *end, *str, *ptr;

  if((base < 2) || (base > ZBN_IO_MAX_BASIS)) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_PRINTTOSTRING, ZEN_INPUT);
    return(ZERR);
  }

  /* Throw away any initial space */
  for(begin = s; isspace(*begin); begin++) ;
  
  /* Go at the end of the string */
  for(end = begin; isAdigit(*end); end++) ;

  if (!(str = (char *)ZENMalloc(sizeof(char)*(end-begin+1)))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMSTRING, ZEN_ALLOC);
    return(ZERR);
  }
  strncpy(str, begin, end-begin);
  *(str+(end-begin)) = '\0';

  
  nl = 2+(int)(BnLog[base]*strlen(str)/(M_LN2*BN_DIGIT_SIZE));
  if (!(n = ZBNC(nl))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMSTRING, ZEN_ALLOC);
    ZENFree(str);
    return(ZERR);
  }
  if (!(p = ZBNC(nl))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMSTRING, ZEN_ALLOC);
    ZENFree(str); ZBNF(n);
    return(ZERR);
  }
  
  ZBNSetToZero(n, nl);
  for (ptr = str;  *ptr;  ptr++) {
    ZBNSetToZero(p, nl);
    ZBNSetDigit(p, CTOI (*ptr));
    ZBNMultiplyDigit(p, nl, n, nl-1, (BigNumDigit)base);  
    t = p;  p = n;  n = t;
  }  
  ZENFree(str);
  ZBNF(p);

  *p_n = n;
  *p_nl = ZBNNumDigits(n, nl);
  
  return(end-s);
}


/* \TEX{
   \title{Converting to string in an internal representation}
   \input{An allocated |BigNum| |n| of length |nl|.}
   \output{An allocated string representing |n| in base |base| if no
   error occured, |ZENNULL| otherwise.}
   \note{The size |nl| is not saved in the string.}
   } */
char *ZBNPutToStringLazy(n,nl)
     BigNum n;
     BigNumLength nl;
{
  int i, j;
  char *string, *ptr;
  BigNumDigit u;

  i=nl*2*sizeof(BigNumDigit)/sizeof(char)+1;
  if (!(string =
	(char *)ZENMalloc(sizeof(char)*i))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_PUTTOSTRING, ZEN_ALLOC);
    return(ZNULL);
  }
  
  for(ptr = string, i = 0; i < nl; i++) {
    u = ZBNDigit(n, i);
    for(j = 0; j < sizeof(BigNumDigit)/sizeof(char); j++) {            
      if (sprintf(ptr, "%.2X", (int)(u & (((BigNumDigit)1 << (sizeof(char)*BN_BYTE_SIZE))-1))) < 0) {
	ZENFree(string);
	ZENSetError(ZEN_LIB_ZBN, ZEN_PUTTOSTRING, ZEN_SPRINTF);
	return(ZNULL);
      }
      ptr += 2;
      u >>= sizeof(char)*BN_BYTE_SIZE;
    }
  }
  
  return(string);
}

/* \TEX{
   \title{Converting to string in an internal representation}
   \input{An allocated |BigNum| |n| of length |nl|.}
   \output{An allocated string representing |n| in base |base| if no
   error occured, |ZENNULL| otherwise.}
   \note{The size of the |BigNum| is saved in the string.}
   } */
char *ZBNPutToString(n,nl)
     BigNum n;
     BigNumLength nl;
{
  BigNumLength i;
  char *string, *ptr;


  if(!(ptr=ZBNPutToStringLazy(n,nl)))
    return(ZNULL);
  
  i=9+ nl*2*sizeof(BigNumDigit)/sizeof(char);
  if (!(string =
	(char *)ZENMalloc(sizeof(char)*i))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_PUTTOSTRING, ZEN_ALLOC);
    return(ZNULL);
  }
  
  if (sprintf(string, "%.8X%s", (unsigned int)(nl*sizeof(BigNumDigit)/sizeof(char)), ptr) < 0) {
    ZENFree(string);
    ZENSetError(ZEN_LIB_ZBN, ZEN_PUTTOSTRING, ZEN_SPRINTF);
    return(ZNULL);
  }

  ZENFree(ptr);
  return(string);
}

/* \TEX{
   \title{Converting from string in an internal representation}
   \input{An allocated |BigNum| and its length |nl| and
   a string |s| representing a |BigNum| in an internal
   representation.}
   \output{The number of character read if no error occured, |ZENERR|
   otherwise}
   \side{|n| is filled with |s|}
   } */
int ZBNGetFromStringLazy(n, nl, s)
     BigNum n;
     BigNumLength nl;
     char *s;
{
  int i, j;
  char *ptr;
  unsigned int u;

  ptr=s;
  for(i = 0; i < nl; i++, n++) {
    *n = 0;
    for(j = 0; j < sizeof(BigNumDigit)/sizeof(char); j++) {
      if (sscanf(ptr, "%2x", &u) == EOF) {
	ZENSetError(ZEN_LIB_ZBN, ZEN_GETFROMSTRING, ZEN_SSCANF);
	return(ZERR);
      }
      for(; isspace(*ptr); ptr++) ;
      ptr+=2;
      *n |= (BigNumDigit)u<<(j*sizeof(char)*BN_BYTE_SIZE);
    }
  }
  return(ptr-s);
}

/* \TEX{
   \title{Converting from string in an internal representation}
   \input{A pointer |p_n| on a non allocated |BigNum|, a pointer
   |p_nl| to an integer and a string |s| representing a |BigNum| in an
   internal representation.}
   \output{The number of character read if no error occured, |ZENERR|
   otherwise}
   \side{|p_n| is allocated and is filled with |s|}
   } */
int ZBNGetFromString(p_n, p_nl, s)
     BigNum *p_n;
     BigNumLength *p_nl;
     char *s;
{
  int i, j;
  char *ptr;
  unsigned int nl;

  /* Throw away any initial space */
  ptr=s;
  if (sscanf(ptr, "%8x", &nl) == EOF) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_GETFROMSTRING, ZEN_SSCANF);
    return(ZERR);
  }
  *p_nl = (BigNumLength)nl;

  /* Becareful, 1 BigNumDigit on a 32 bits computers must be read
     as 1 BigNumDigit on a 64 bits computer ! */
  *p_nl = 1+(*p_nl-1)/sizeof(BigNumDigit)/sizeof(char);
  if (!(*p_n = ZBNC(*p_nl))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_GETFROMSTRING, ZEN_ALLOC);
    return(ZERR);
  }
  ptr+=8;
  for(; isspace(*ptr); ptr++) ;
  i=ptr-s;
  
  if((j=ZBNGetFromStringLazy(*p_n,*p_nl,ptr))<0) {
    ZBNF(*p_n);
    return(ZERR);
    }
  
  return(i+j);
}

/* \TEX[
   The syntax of file functions is the same, using a |FILE *file| instead of a string.

   \title{Printing to file.}
   \input{A stream |file|, an allocated |BigNum| |n| of length |nl|
   and a base |base|$\in \{2,\ldots, |ZBN_IO_MAX_BASIS|\}$.}
   \output{|ZENERR| if an error occured, 0 ootherwise}
   \side{Printing a representation of |n| in base |base| to |file|.}
   \note{In order to be correctly re-read by the |ZBNReadFromFile|
   function, a separating character must be inserted (space or
   carriage return for instance). Otherwise, two consecutive |BigNum|s
   would be concatenated and read as a single number. This can be
   viewed as a feature.}
   } */
int ZBNPrintToFile(file,n,nl,base)
     FILE *file;
     BigNumLength nl;
     int base;
     BigNum n;
{
  char *s;

  if (!(s = ZBNPrintToString(n, nl, base))) {
    return(ZERR);
  }
  if (fprintf(file, "%s", s) < 0) {
    ZENFree(s);
    ZENSetError(ZEN_LIB_ZBN, ZEN_PRINTTOFILE, ZEN_FPRINTF);
    return(ZERR);
  }
  ZENFree(s);
  return(0);
}
 
/* \TEX[
   \title{Reading from file.}
   \input{A stream |file| and an allocated bignum |n| of length |nl|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Fill |n| with |file|.}
   } */
int ZBNReadFromFileLazy(n, nl, file, base)
     BigNum n;
     FILE *file;
     BigNumLength nl;
     int base;
{
  char *buf, *ptr;
  int bufsiz;

  bufsiz = BUFSIZ;
  if (!(buf = (char *)ZENMalloc(sizeof(char)*bufsiz))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMFILE, ZEN_ALLOC);
    return(ZERR); 
  }
  for(*(ptr = buf) = '\0';
      (*ptr = (char)fgetc(file)) != (char)EOF;) {
    if (!isspace((int)*ptr))
      break;
  }
  if (*ptr == (char)EOF) {
    ZENFree(buf);
    ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMFILE, ZEN_SYNTAX);
    return(ZERR); 
  }

  for(;*ptr != (char)EOF; *ptr = (char)fgetc(file)) {
    
    if (!isAdigit((int)*ptr)) {
      ungetc((int)(*ptr), file);
      break;
    }
    
    if (((++ptr)-buf) == bufsiz) {
      bufsiz += BUFSIZ;
      if (!(buf = (char *)ZENRealloc(buf, sizeof(char)*bufsiz))) {
	ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMFILE, ZEN_ALLOC);
	return(ZERR); 
      }
      ptr = buf + bufsiz-BUFSIZ;
      continue;
    }
  }
  if (ptr == buf) {
    if (*ptr) {
      ZENFree(buf);
      ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMFILE, ZEN_SYNTAX);
      return(ZERR);
    }
  }
  *ptr = '\0';
  if (ZBNReadFromStringLazy(n, nl, buf, base) < 0) {
    ZENFree(buf);
    return(ZERR);
  }
  ZENFree(buf);
  return(0);
}
 
/* \TEX[
   \title{Reading from file.}
   \input{A stream |file| and the adress of a bignum |p_n|, |p_nl|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Allocate |*p_nl|, fill it with |file| and assigns the length of |n| to |*p_nl|.}
   } */
int ZBNReadFromFile(p_n, p_nl, file, base)
     BigNum *p_n;
     FILE *file;
     BigNumLength *p_nl;
     int base;
{
  char *buf, *ptr;
  int bufsiz;

  bufsiz = BUFSIZ;
  if (!(buf = (char *)ZENMalloc(sizeof(char)*bufsiz))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMFILE, ZEN_ALLOC);
    return(ZERR); 
  }
  for(*(ptr = buf) = '\0';
      (*ptr = (char)fgetc(file)) != (char)EOF;) {
    if (!isspace((int)*ptr))
      break;
  }
  if (*ptr == (char)EOF) {
    ZENFree(buf);
    ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMFILE, ZEN_SYNTAX);
    return(ZERR); 
  }

  for(;*ptr != (char)EOF; *ptr = (char)fgetc(file)) {
    
    if (!isAdigit((int)*ptr)) {
      ungetc((int)(*ptr), file);
      break;
    }
    
    if (((++ptr)-buf) == bufsiz) {
      bufsiz += BUFSIZ;
      if (!(buf = (char *)ZENRealloc(buf, sizeof(char)*bufsiz))) {
	ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMFILE, ZEN_ALLOC);
	return(ZERR); 
      }
      ptr = buf + bufsiz-BUFSIZ;
      continue;
    }
  }
  if (ptr == buf) {
    if (*ptr) {
      ZENFree(buf);
      ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMFILE, ZEN_SYNTAX);
      return(ZERR);
    }
  }
  *ptr = '\0';
  if (ZBNReadFromString(p_n, p_nl, buf, base) < 0) {
    ZENFree(buf);
    return(ZERR);
  }
  ZENFree(buf);
  return(0);
}
 
/* \TEX[
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |BigNum| |n| of length |nl|.}
   \side{Printing the representation used by |ZBNReadFromFile| of |n| to |file|.}
   } */
int ZBNPutToFileLazy(file,n,nl)
     FILE *file;
     BigNumLength nl;
     BigNum n;
{
  int i, j;
  BigNumDigit u;

 for(i = 0; i < nl; i++) {
   u = ZBNDigit(n, i);
   for(j = 0; j < sizeof(BigNumDigit)/sizeof(char); j++) {
     if (fprintf(file, "%.2X", (int)(u & (((BigNumDigit)1 << (sizeof(char)*BN_BYTE_SIZE))-1))) < 0) {
       ZENSetError(ZEN_LIB_ZBN, ZEN_PUTTOFILE, ZEN_FPRINTF);
       return(ZERR);
     }
     u >>= sizeof(char)*BN_BYTE_SIZE;
   }
 }
 
 return(0);
}

/* \TEX[
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |BigNum| |n| of length |nl|.}
   \side{Printing the representation used by |ZBNReadFromFile| of |n| to |file|.}
   } */
int ZBNPutToFile(file,n,nl)
     FILE *file;
     BigNumLength nl;
     BigNum n;
{

 if (fprintf(file, "%.8X", (unsigned int)(nl*sizeof(BigNumDigit)/sizeof(char))) < 0) {
   ZENSetError(ZEN_LIB_ZBN, ZEN_PUTTOFILE, ZEN_FPRINTF);
   return(ZERR);
 }

 if (ZBNPutToFileLazy(file,n,nl))
   return(ZERR);
 
 return(0);
}

/* \TEX[
   \title{Reading from file in an internal representation.}
   \input{A stream |file| and an allocated |BigNum| of length |nl|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Assigns |n| with |file|.}
   } */
int ZBNGetFromFileLazy(n, nl, file)
     BigNum n;
     BigNumLength nl;
     FILE *file;
{
  int i, j;
  unsigned int u;

  for(i = 0; i < nl; i++, n++) {
    *n = 0;
    for(j = 0; j < sizeof(BigNumDigit)/sizeof(char); j++) {
      if (fscanf(file, "%2x", &u) != 1) {
	ZENSetError(ZEN_LIB_ZBN, ZEN_GETFROMFILE, ZEN_FSCANF);
	return(ZERR);
      }
      *n |= (BigNumDigit)u<<(j*sizeof(char)*BN_BYTE_SIZE);
    }
  }
  return(0);
}
/* \TEX[
   \title{Reading from file in an internal representation.}
   \input{A stream |file| and the adress of a bignum |p_n|, |p_nl|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Assigns the length of |n| to |*p_nl|.}
   } */
int ZBNGetFromFile(p_n, p_nl, file)
     BigNum *p_n;
     BigNumLength *p_nl;
     FILE *file;
{
  unsigned int nl;

  if (fscanf(file, "%8x", &nl) == EOF) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_GETFROMFILE, ZEN_FSCANF);
    return(ZERR);
  }
  *p_nl = (BigNumLength)nl;

  /* Becareful, 1 BigNumDigit on a 32 bits computers must be read
     as 1 BigNumDigit on a 64 bits computer ! */
  *p_nl = 1+(*p_nl-1)/sizeof(BigNumDigit)/sizeof(char);
  if(!(*p_n = ZBNC(*p_nl))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_GETFROMFILE, ZEN_ALLOC);
    return(ZERR);
  }
  if(ZBNGetFromFileLazy(*p_n,*p_nl,file)) {
    ZBNF(*p_n);
    return(ZERR);
  }
  return(0);
}
 
