/*
 * zen/zbn/c/karacut.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdlib.h>
#include "kara_defaults.h"

#define TEST_KARA   0

#define MIN_TIMING  50
#define REF_TIMING  100
#define NB_INIT     128

#define Infinite 1000000.0
#define OPT_BOUND 0.90

#ifndef CHECK_KARA		/* this checks the result of operations */
# define CHECK_KARA 1 /*KARA_DEBUG*/
#endif

/* \TEX{
   \title{Setting of Karatsuba's cutoff}
   \input{The size ($\ge 2$) in |BigNumDigit|s of the numbers to
   multiply}
   \output{|ZENERR| if an error occured, 0 if Karatsuba's
   operations are slower, the cutoff to use otherwise.}
   \note{For all possible values of cutoff, a number of randomized
   multiplications are performed. The fastest parameter is
   returned. Karatsuba's operations are only chosen if a speed
   increase of at least 10\% is observed. This is made because
   optimizers can often include simple operations but not complex
   ones. If difference is tiny between Karatsuba's and standard
   algorithm, standard algorithm often leads to final best results.}
   } */
BigNumLength ZBNKaratsubaMultiplyCutoff(size)
     BigNumLength size;
{
  int nbops;
  double tt, *tab;
  int i, nblim;
  BigNumLength lim;
  BigNum *y,*x,*a,buf;

  /* size=4; */

  /* Initialization */
#ifdef KARAm_LIM
  nblim = 2;
#else
  nblim = size;
#endif
  tab = (double *)(malloc((size_t)(nblim*sizeof(double))));
  a = (BigNum *)(malloc((size_t)((NB_INIT+1)*sizeof(BigNum))));
  x = (BigNum *)(malloc((size_t)(NB_INIT*sizeof(BigNum))));
  y = (BigNum *)(malloc((size_t)(NB_INIT*sizeof(BigNum))));
  buf = ZBNC(ZBNSizeBufKaraM(size));

  if(!(tab && x && y && a && buf)) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_KARAMC, ZEN_ALLOC);
    return(0);
  }
  for(i = NB_INIT-1; i >= 0; i--) {
    a[i] = ZBNC(size); x[i] = ZBNC(2*size); y[i] = ZBNC(2*size);
    if(!(a[i] && x[i] && y[i])) {
      ZENSetError(ZEN_LIB_ZBN, ZEN_KARAMC, ZEN_ALLOC);
      return(0);
    }
    ZBNSetRandom(a[i],size);
    ZBNSetToZero(x[i], 2*size);
    ZBNSetToZero(y[i], 2*size);
  }

  /* How many calls to ZBNMultiplyPlain do we need to have accurate timings ?  */
  for(tt = 0., nbops = 1;
      tt < MIN_TIMING;
      nbops *= 2) {
    tt = runtime();
    for(i = 0; i < nbops; i++) {
      ZBNSetToZero(x[0], 2*size);
      ZBNMultiplyPlain(x[0], 2*size, a[i%NB_INIT], size, a[(i+1)%NB_INIT], size);
    }
    tt = runtime()-tt;
  }
  if (nbops > 1) {nbops /= 2;}
  nbops = (int)((nbops*REF_TIMING)/tt);
  if (!nbops) {nbops = 1;}
  
  /* nbops = 1; */


  /* ZBNMultiplyPlain */
  tt = runtime();
  for(i = nbops-1; i >= 0; i--) {
    ZBNSetToZero(x[i%NB_INIT], 2*size);
    ZBNMultiplyPlain(x[i%NB_INIT],2*size,a[i%NB_INIT],size,a[(i+1)%NB_INIT],size);
  }
  tab[0]=(runtime()-tt)/nbops;
  fprintf(stderr,"%d ZBNMultiplyPlain(): %f ms/ops\n", nbops, tab[0]);
 
  /* ZBNMultiplyKaratsuba */
  for(lim = nblim-1; lim > 0; lim--)
    tab[lim] = Infinite;
  for(lim = 1; lim < nblim; lim *= 2);
  
  for(;lim >  1;) {
    lim >>= 1;
    tt=runtime();
    for(i = nbops-1; i>=0; i--) {
      ZBNSetToZero(y[i%NB_INIT], 2*size);
#ifndef KARAm_LIM
      ZBNMultiplyKaratsubaBuffer(y[i%NB_INIT], a[i%NB_INIT],size,a[(i+1)%NB_INIT],size, buf, lim);
#else
      ZBNMultiplyKaratsubaBuffer(y[i%NB_INIT], a[i%NB_INIT],size,a[(i+1)%NB_INIT],size, buf);
#endif
    }
    tab[lim] = (runtime()-tt)/nbops;
    fprintf(stderr,"%d ZBNMultiplyKaratsuba(%d): %f ms/ops\n",nbops,(int)lim,tab[lim]);

    for(i = NB_INIT-1; i >= 0; i--) {
#if CHECK_KARA
      if(ZBNCompare(x[i], 2*size, y[i], 2*size)) {
	fprintf(stderr,"Error: Different results have been obtained\n");
	exit(-1);
      }
#endif
    }
  }

  /* Selection */
  tt=Infinite;
  for(i = nblim-1; i >= 0; i--) {
    if(tab[i] <= tt)
      tt = tab[(lim=i)];
  }
  if(tab[0]*OPT_BOUND <= tt) {
    lim = 0;
  }

  /* Freeing memory */
 free(tab);
 ZBNF(buf);
 for(i = NB_INIT-1; i >= 0; i--) {
   ZBNF(a[i]); ZBNF(x[i]); ZBNF(y[i]);
 }
 free(a); free(x); free(y);
 
#if TEST_KARA
 lim = 1;
#endif
 if (lim) {
   fprintf(stderr,"ZBNMultiplyKaratsuba_%d chosen.\n", (int)lim);
 }
 else {
   fprintf(stderr,"ZBNMultiplyPlain chosen.\n");
 }
 
 return(lim);
}

/* \TEX{
   \title{Setting of Karatsuba's cutoff}
   \input{The size in |BigNumDigit|s of the numbers to square}
   \output{|ZENERR| if an error occured, 0 if Karatsuba's
   operations are slower than |ZBNSquarePlain|, the cutoff to use
   otherwise.}
   \note{For all possible values of cutoff, a number of randomized
   squares are performed. The fastest parameter is returned.}
   } */
BigNumLength ZBNKaratsubaSquareCutoff(size)
     BigNumLength size;
{
  int nbops;
  double tt, *tab;
  int i, nblim;
  BigNumLength lim;
  BigNum *y,*x,*a,buf;

  /* Initialization */
#ifdef KARAm_LIM
  nblim = 2;
#else
  nblim = size;
#endif
  tab = (double *)(malloc((size_t)(nblim*sizeof(double))));
  a = (BigNum *)(malloc((size_t)((NB_INIT+1)*sizeof(BigNum))));
  x = (BigNum *)(malloc((size_t)(NB_INIT*sizeof(BigNum))));
  y = (BigNum *)(malloc((size_t)(NB_INIT*sizeof(BigNum))));
  buf = ZBNC(ZBNSizeBufKaraS(size));

  if(!(tab && x && y && a && buf)) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_KARAMC, ZEN_ALLOC);
    return(0);
  }
  for(i = NB_INIT-1; i >= 0; i--) {
    a[i] = ZBNC(size); x[i] = ZBNC(2*size); y[i] = ZBNC(2*size);
    if(!(a[i] && x[i] && y[i])) {
      ZENSetError(ZEN_LIB_ZBN, ZEN_KARAMC, ZEN_ALLOC);
      return(0);
    }
    ZBNSetRandom(a[i],size);
    ZBNSetToZero(x[i], 2*size);
    ZBNSetToZero(y[i], 2*size);
  }

  /* How many calls to ZBNSquarePlain do we need to have accurate timings ?  */
  for(tt = 0., nbops = 1;
      tt < MIN_TIMING;
      nbops *= 2) {
    tt = runtime();
    for(i = 0; i < nbops; i++) {
      ZBNSetToZero(x[0], 2*size);
      ZBNSquarePlain(x[0], 2*size, a[i%NB_INIT], size);
    }
    tt = runtime()-tt;
  }
  if (nbops > 1) {nbops /= 2;}
  nbops = (int)((nbops*REF_TIMING)/tt);
  if (!nbops) {nbops = 1;}

  /* ZBNSquarePlain */
  tt = runtime();
  for(i = nbops-1; i >= 0; i--) {
    ZBNSetToZero(x[i%NB_INIT], 2*size);
    ZBNSquarePlain(x[i%NB_INIT],2*size,a[i%NB_INIT],size);
  }
  tab[0]=(runtime()-tt)/nbops;
  fprintf(stderr,"%d ZBNSquarePlain(): %f ms/ops\n", nbops, tab[0]);

  /* ZBNSquareKaratsuba */
  for(lim = nblim-1; lim > 0; lim--)
    tab[lim] = Infinite;
  for(lim = 1; lim < nblim; lim *= 2);
  
  for(;lim >  1;) {
    lim >>= 1;
    tt=runtime();
    for(i = nbops-1; i>=0; i--) {
      ZBNSetToZero(y[i%NB_INIT], 2*size);
#ifndef KARAm_LIM
      ZBNSquareKaratsubaBuffer(y[i%NB_INIT], a[i%NB_INIT],size, buf, lim);
#else
      ZBNSquareKaratsubaBuffer(y[i%NB_INIT], a[i%NB_INIT],size, buf);
#endif
    }
    tab[lim] = (runtime()-tt)/nbops;
    fprintf(stderr,"%d ZBNSquareKaratsuba(%d): %f ms/ops\n",nbops,(int)lim,tab[lim]);

    for(i = 0; i < NB_INIT; i++) {
#if CHECK_KARA
      if(ZBNCompare(x[i], 2*size, y[i], 2*size)) {
	fprintf(stderr,"Error: Different results have been obtained\n");
	exit(-1);
      }
#endif
      ZBNSetToZero(y[i], 2*size);
    }
  }

  /* Selection */
  tt=Infinite;
  for(i = nblim-1; i >= 0; i--) {
    if(tab[i] <= tt)
      tt = tab[(lim=i)];
  }
  if(tab[0]*OPT_BOUND <= tt) {
    lim=0;
  }

  /* Freeing memory */
 free(tab);
 ZBNF(buf);
 for(i = NB_INIT-1; i >= 0; i--) {
   ZBNF(a[i]); ZBNF(x[i]); ZBNF(y[i]);
 }
 free(a); free(x); free(y);
 
#if TEST_KARA
 lim = 1;
#endif
 if (lim) {
   fprintf(stderr,"ZBNSquareKaratsuba_%d chosen.\n", (int)lim);
 }
 else {
   fprintf(stderr,"ZBNSquarePlain chosen.\n");
 }
 
 return(lim);
}

BigNumDigit ZBNSquareK (pp, pl, nn, nl)
     BigNum		pp, nn;
     BigNumLength	pl, nl;
{
#ifndef KARA_LIM_SET
# if defined(KARAs_BOUND)
  return(nl > KARAs_BOUND
	 ?
	 ZBNSquareKaratsuba(pp,pl,nn,nl)
	 :
	 ZBNSquarePlain(pp,pl,nn,nl));
# else
  return(ZBNSquareKaratsuba(pp,pl,nn,nl));
# endif
#else
  return(ZBNSquarePlain(pp,pl,nn,nl));
#endif
}  

BigNumDigit ZBNMultiplyK (pp, pl, mm, ml, nn, nl)
     BigNum		pp, nn, mm;
     BigNumLength	pl, nl, ml;
{
#ifndef KARA_LIM_SET
# ifdef KARAm_BOUND
  if(ml>nl)
    return(nl > KARAm_BOUND
	   ?
	   ZBNMultiplyKaratsuba(pp,pl,mm,ml,nn,nl)
	   :
	   ZBNMultiplyPlain(pp,pl,mm,ml,nn,nl));
  else
    return(ml > KARAm_BOUND
	   ?
	   ZBNMultiplyKaratsuba(pp,pl,mm,ml,nn,nl)
	   :
	   ZBNMultiplyPlain(pp,pl,mm,ml,nn,nl));
# else
  return(ZBNMultiplyKaratsuba(pp,pl,mm,ml,nn,nl));
# endif
#else
  return(ZBNMultiplyPlain(pp,pl,mm,ml,nn,nl));
#endif
}
