/*
 * zen/zbn/c/modulo.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

#if 0
/* \TEX{
   \Subsection{Modular operations}
   
   \title{Reduction of a |BigNum| modulo another}
   \input{Two allocated |BigNum|s and their lengths, and a pointer on
   a non allocated |BigNum|.}
   \output{|ZENERR| if an error occured, the length of the result otherwise.}
   \side{|*p_m| is allocated and set to $|n| \bmod |m| $.}
   } */
BigNumLength ZBNModulo(p_m,n,nl,m,ml)
     BigNum *p_m,n,m;
     BigNumLength nl,ml;{}
#endif

/* \TEX{
   \title{Reduction of a |BigNum| modulo another}
   \input{Two allocated |BigNum|s and their lengths, and a pointer on
   a non allocated |BigNum|.}
   \output{0 if an error occured, the length of the result otherwise.}
   \side{|*p_m| is allocated and set to $|n| \bmod |m| $.}
   \note{One must have $|ZBNDigit(n,nl-1)|\neq 0$ and
   $|ZBNDigit(m,ml-1)|\neq 0$.}
   } */
BigNumLength ZBNModuloLazy(p_m,n,nl,m,ml)
     BigNum *p_m,n,m;
     BigNumLength nl,ml;
{
  switch(ZBNCompareLazy(n,nl,m,ml)) {
  case BN_LT:
    *p_m=ZBNC(nl);
    if(!(*p_m)) {
      ZENSetError(ZEN_LIB_ZBN,ZEN_MODULO,ZEN_ALLOC);
      return(0);
    }
    ZBNAssign(*p_m,n,nl);
    return(nl);
    
  case BN_GT:
    *p_m=ZBNC(nl+1);
    if(!(*p_m)) {
      ZENSetError(ZEN_LIB_ZBN,ZEN_MODULO,ZEN_ALLOC);
      return(0);
    }
    (*p_m)[nl]=0;
    ZBNAssign(*p_m,n,nl);
    ZBNDivide(*p_m,nl+1,m,ml);
    nl=ZBNNumDigits(*p_m,ml);
    return(nl);
  }
  *p_m=ZBNC(1);
  **p_m=0;
  return(1);
}

/* \TEX{
   \title{Modular addition}
   \input{Three allocated |BigNum|s.}
   \side{$|a| = |a| + |b| \bmod |m|$}
   \note{One must have $|ml| = |al|$, $|a| < |m|$, $|b| < |m|$ and
   $|ZBNDigit(m,ml-1)| \neq 0 $. One can have $|a| = |b|$.}
   } */
void ZBNModAdd(a,al,b,bl,m,ml)
     BigNum a,b,m;
     BigNumLength al,bl,ml;
{
  if(ZBNAdd(a,al,b,bl,(BigNumDigit)0)) {
    ZBNSubtract(a, al, m, ml >= al ? al : ml, (BigNumDigit)1);
  }
  else {
    al = ZBNNumDigits(a, al);
    switch(ZBNCompareLazy(a, al, m, ml)) {
    case BN_EQ:
      ZBNSetToZero(a,al);
      break;
      
    case BN_GT:
      ZBNSubtract(a, al, m, ml, (BigNumDigit)1);
      break;
    }
  }
}

/*
   \TEX{
   \title{Modular negation}
   \input{Three distinct allocated |BigNum|s.}
   \side{$|a| = - |b| \bmod (|m|+1)$}
   \note{One must have $|al| \geq |ml|$, $|b| \leq |m|$ and $|*(m+ml)|
   \neq 0 $.}
   } */
void ZBNModNegate(a,al,b,bl,m,ml)
     BigNum a,b,m;
     BigNumLength al,bl,ml;
{
  if(ZBNIsZero(b,bl))
    ZBNSetToZero(a,al);
  else {
    ZBNAssign(a,m,ml);
    ZBNSetToZero(a+ml,al-ml);
    ZBNSubtract(a,ml,b,bl,(BigNumDigit)1);
    al=ZBNNumDigits(a,ml);
    if (ZBNAddCarry(a,al,(BigNumDigit)1)) {
      a[al] = 1;
    }
  }
}

/* \TEX{
   \title{Modular subtraction}
   \input{Three distincts allocated |BigNum|s.}
   \side{$|a| = |a| - |b| \bmod (|m|+1)$}
   \note{One must have $|ml| \geq |al| \geq size of (|m|-1)$, $|a|
   \leq |m|$ and $|b| \leq |m|$.}
   } */
void ZBNModSubtract(a,al,b,bl,m,ml)
     BigNum a,b,m;
     BigNumLength al,bl,ml;
{
  al=ZBNNumDigits(a,al);
  bl=ZBNNumDigits(b,bl);
  switch(ZBNCompareLazy(a,al,b,bl)) {
  case BN_EQ:
    ZBNSetToZero(a,al);
    break;
    
  case BN_GT:
    ZBNSubtract(a,al,b,bl,1);
    break;

  case BN_LT:
    ZBNAdd(a,ml,m,ml,(BigNumDigit)1);
    ZBNSubtract(a,ml,b,bl,(BigNumDigit)1);
    break;
  }
}

/* \TEX{
   \title{Modular multiplication}
   \input{Four allocated |BigNum|s.}
   \side{$|x| = |a| . |b| \bmod |m| $}
   \note{One must have $|xl| \geq |al|+ |bl| + 1$, $|*(a+al-1)|\neq 0$
   and $|*(b+bl-1)|\neq 0$. One can have $|a| = |b|$. }
   } */
void ZBNModMultiply(x,xl,a,al,b,bl,m,ml)
     BigNum x,a,b,m;
     BigNumLength xl,al,bl,ml;
{
  BigNumLength ll;
  
  ZBNSetToZero(x,xl);
  
  ll=al+bl;
  
  ZBNMultiply(x,al+bl,b,bl,a,al);
  
#define ModularReduction()			\
  al=ZBNNumDigits(x,ll);			\
  switch(ZBNCompareLazy(x,al,m,ml)) {		\
   case BN_EQ:					\
     ZBNSetToZero(x,ml);			\
     break;					\
   case BN_LT:					\
     break;					\
   case BN_GT:					\
     ZBNDivide(x,al+1,m,ml);			\
     ZBNSetToZero(x+ml,al+1-ml);		\
   }

  ModularReduction();
}

/* \TEX{
   \title{Modular squaring}
   \input{Three allocated |BigNum|s.}
   \side{$|x| = |a|^2 \bmod |m| $}
   \note{One must have $|xl| \geq 2\times|al| + 1$, $|*(a+al-1)|\neq 0$ and $|*(b+bl-1)|\neq 0$.}
   } */
void ZBNModSquare(x,xl,a,al,m,ml)
     BigNum x,a,m;
     BigNumLength xl,al,ml;
{
  BigNumLength ll;
 
  ZBNSetToZero(x,xl);
  
  ll=al<<1;
  
  ZBNSquare(x,al+al,a,al); 
  
  ModularReduction();
}

/* \_____TEX{
   \title{Modular multiplication using Karatsuba's operations}
   \input{Four allocated |BigNum|s and a buffer of allocated size
   |ZBNSizeBufKaraM(al)|.}
   \side{$|x| = |a| . |b| \bmod |m| $}
   \note{One must have $|xl| \geq |al|+ |bl| + 1$, $|*(a+al-1)|\neq 0$
   and $|*(b+bl-1)|\neq 0$. One can have $|a| = |b|$. }
   } */
void ZBNModMultiplyKaratsubaBuffer(x,xl,a,al,b,bl,buf,m,ml)
     BigNum x,a,b,m,buf;
     BigNumLength xl,al,bl,ml;
{
  BigNumLength ll;

  ZBNSetToZero(x,xl);

  ll=al+bl;
  
  ZBNMultiplyKaratsubaBuffer(x,b,bl,a,al,buf);
  
  ModularReduction();
}

/* \_____TEX{
   \title{Modular squaring}
   \input{Three allocated |BigNum|s and a buffer of allocated size
   |ZBNSizeBufKaraS(al)|.}
   \side{$|x| = |a|^2 \bmod |m| $}
   \note{One must have $|xl| \geq 2\times|al| + 1$, $|*(a+al-1)|\neq
   0$ and $|*(b+bl-1)|\neq 0$. One can have $|a| = |b|$. }
   } */
void ZBNModSquareKaratsubaBuffer(x,xl,a,al,buf,m,ml)
     BigNum x,a,m,buf;
     BigNumLength xl,al,ml;
{
  BigNumLength ll;
 
  ZBNSetToZero(x,xl);
  
  ll=al<<1;
  
  ZBNSquareKaratsubaBuffer(x,a,al,buf); 
  
  ModularReduction();
}
