/*
 * zen/zbn/c/mul.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

#ifdef USE_GMP
# include <gmp.h>
#endif

/* \TEX{
   \Subsection{Multiplication}
   } */

#ifndef USE_GMP
static BigNumDigit ZBNMultiplyEven(m, n, nl, d)
     BigNum m, n;
     BigNumLength nl;
     BigNumDigit d;
{
  BigNumDigit c, h, l;

  c = 0;
  do {
    l = n[nl];
    zbnmuladds(h, m[nl], l, d, c);

    l = n[nl+1];
    zbnmuladds(c, m[nl+1], l, d, h);

    nl += 2; 
  } while (nl);

  return(c);
}

static BigNumDigit ZBNMultiplyOdd(m, n, nl, d)
     BigNum m, n;
     BigNumLength nl;
     BigNumDigit d;
{
  BigNumDigit c, h, l;

  c = 0;
  l = n[nl];
  zbnmuladds(h, m[nl], l, d, c);
  nl++;

  if (nl) {
    do {
      l = n[nl];
      zbnmuladds(c, m[nl], l, d, h);

      l = n[nl+1];
      zbnmuladds(h, m[nl+1], l, d, c);

      nl += 2;
    } while (nl);
  }
  return(h);
}

/* \TEX{
   \title{Multiplication by a digit}
   \input{Two allocated |BigNum|s and a digit.}
   \output{A carry out.}
   \side{$|(m,ml)| += |(n,nl)| \times |d|$.}
   \note{One must have $|ml| \ge |nl|+1$.  If |nl| equals zero, the
   return is 0 and nothing is done.  }
   } */
BigNumDigit ZBNMultiplyDigit(m, ml, n, nl, d)
     BigNum m; BigNumLength ml;
     BigNum n; BigNumLength nl;
     BigNumDigit d;
{
  BigNumDigit l, h;
  BigNumDigit carry;

  if (!d) {
    return(CST_BNDIGIT(0));
  }
  if (d == CST_BNDIGIT(1)) {
    return(ZBNAdd(m, ml, n, nl, CST_BNDIGIT(0)));
  }
  if (nl) {
    m += nl; n += nl;
    nl *= CST_BNLENGTH(-1); ml += nl;

    /* Unrolling loop */
    if (nl%2) {
      carry = ZBNMultiplyOdd(m, n, nl, d);
    }
    else {
      carry = ZBNMultiplyEven(m, n, nl, d);
    }

    d = carry; carry = 0;
    zbnadd(carry, m[0], carry, m[0], 0, d);
    m += ml; ml--;
  }
  else {
    m += ml; carry = 0;
  }
  if (ml && carry) {
    ml *= CST_BNLENGTH(-1);
    do {
      if (++m[ml]) {break;}
      ml++;
    } while(ml);
    carry = ml ? 0 : 1;
  }
  return(carry);
}
#else
BigNumDigit ZBNMultiplyDigit(m, ml, n, nl, d)
     BigNum m; BigNumLength ml;
     BigNum n; BigNumLength nl;
     BigNumDigit d;
{
  return(
	 (nl) ? mpn_add_1(m+nl, m+nl, ml-nl,
			  mpn_addmul_1(m, n, nl, d)) : (BigNumDigit)(0)
	 );
}
#endif

#if 0
/* \TEX{
   \title{Multiplication and Add}
   \input{Three allocated |BigNum|s and their lengths}
   \output{A carry out.}
   \side{$|(p,pl)| += |(m,ml)| \times |(n,nl)|$.}
   \note{One must have $|pl| \ge |ml|+|nl|$. If |ml| or |nl| equals zero,
   there is no side effect and the carry out is 0.}
   } */
BigNumDigit ZBNMultiply(p, pl, m, ml, n, nl)
     BigNum p, m, n;
     BigNumLength pl, ml, nl; {}
#endif
BigNumDigit ZBNMultiplyPlain(p, pl, m, ml, n, nl)
     BigNum p; BigNumLength pl;
     BigNum m; BigNumLength ml;
     BigNum n; BigNumLength nl;
{
  BigNumDigit carry;
  
  carry = 0;
  while (nl-- > 0) {
    carry += ZBNMultiplyDigit(p, pl, m, ml, *n);
    p++; n++; pl--;
  }
  return(carry);
}

/* \TEX{
   \title{Multiplication}
   \input{Three |BigNum|s and their lengths}
   \side{$|(n,nl)| \leftarrow |(a,al)| \times |(b,bl)|$.}
   \note{One must have $|nl| \ge |al|+|bl|$, |al| and |bl| non zeros.}
   } */
void ZBNMult(n, a, al, b, bl)
     BigNum n, a, b;
     BigNumLength al, bl;
{
  ZBNSetToZero(n, al+bl);
  ZBNMultiply(n, al+bl, a, al, b, bl);
}

#if 0
/* \TEX{
   \title{Squaring and Add}
   \input{Two allocated |BigNum|s and their lengths}
   \output{A carry out.}
   \side{$|(n,nl)| += |(a,al)|^2$.}
   \note{One must have $|nl| \ge 2|al|$. If |al| equals zero, there
   is no side effect and the carry out is 0.}
   } */
BigNumDigit ZBNSquare(n, nl, a, al)
     BigNum n, a;
     BigNumLength nl,al; {}
#endif
BigNumDigit ZBNSquarePlain(p, pl, n, nl)
     BigNum p; BigNumLength pl;
     BigNum n; BigNumLength nl;
{
  BigNumDigit carry;
  BigNumLength ml;
  BigNum m;
  
  m = n; ml = nl; carry = 0;
  while (nl-- > 0) {
    carry += ZBNMultiplyDigit(p, pl, m, ml, *n);
    p++, n++, pl--;
  }
  return(carry);
}

/* \TEX{
   \title{Squaring}
   \input{Two |BigNum|s and a length}
   \side{$|(n,nl)| \leftarrow |(a,al)|^2 $.}
   \note{One must have |n| distinct from |a|, $|nl| \ge 2|al|$, and
   $|al| > 0$.}
   } */
void ZBNSqu(n, a, al)
     BigNum n, a;
     BigNumLength al;
{
  ZBNSetToZero(n, al<<1);
  ZBNSquare(n, al<<1, a, al);
}
