/*
 * zen/zbn/c/random.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "zbn.h"

/* \TEX{
   \Subsection{Random assignment}
   } */

/* \TEX{
   In the same spirit as for allocation functions, we allow customized
   random functions.
   } */

/* \TEX{
   \title{Pseudo-Random Generator (PRG) initialization with a BigNumDigit}
   \input{A |BigNumDigit| which serves as seed for the internal PRG.}

   \note{The PRG behavior is determined by this seed. The programs
   included in |ZEN| use this function to initialize the PRG. By
   default, this function calls the native |ZENNativeSrand|
   function. This can be overwritten by |ZENSetRandomFunctions|. This
   function is provided for compatibility purposes with older version of
   ZEN (cf. |ZENNativeSRand|).}
} */
void zbnsrandom(seed)
     BigNumDigit seed;
{
#if SIZE_BLOC == 32
  unsigned char _seed[64/SIZE_CHAR];
#else
  unsigned char _seed[SIZE_BLOC/SIZE_CHAR];
#endif
  BigNumLength i;

  for(i = 0; i < SIZE_BLOC/SIZE_CHAR; i++) {
    _seed[i] = seed;
    seed >>= SIZE_CHAR;
  }
  for(; i < 64/SIZE_CHAR; i++) {_seed[i] = 0;}
  zsrandom(_seed, (size_t)i);
}

/* \TEX{ 
   \title{Pseudo-Random Generation of a BigNumDigit}
   \output{A pseudo-random |BigNumDigit|.}
   \note{This function is used all over the |ZEN| and |ZENFACT|
   libraries when random generation is needed. By default, this
   function calls the native |ZENNativeRand| function. This can be
   overwritten by |ZENSetRandomFunctions|.}
   } */
BigNumDigit zbnrandom()
{
  BigNumLength i;
  BigNumDigit d;
  
  d = 0;
  for(i = 0; i < SIZE_BLOC/SIZE_CHAR; i++) {
    d |= ((BigNumDigit)zrandom()) << (i*SIZE_CHAR);
  }
  return(d);
}

/* \TEX{
   \title{Set a |BigNum| randomly}
   \input{An allocated |BigNum| and its length.}
   \side{The |BigNum| is filled randomly.}
   } */
void ZBNSetRandom(n, nl)
     BigNum n;
     BigNumLength nl;
{
  BigNumLength i;
  
  if (!nl) return;
  for(i = 0; i < nl; i++) {
    ZBNSetDigit(n+i,zbnrandom());
  }

  /* Sordid hack for compatibility purposes with 64-bit computers */
  if ((SIZE_BLOC == 32) && (nl%2)) {zbnrandom();}
}
