/*
 * zen/zbn/c/sqrt.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zbn.h"

/* \TEX{
   \Subsection{Square root}
   } */

/* \TEX{
   \title{Square root of a |BigNum|}
   \input{An allocated |BigNum| |a| of length |al| and
   another |BigNum| |r|.}
   \output{0 if an error occured, the size of
   the square root otherwise}.
   \side{|r| is filled with floor(sqrt(a, al)).}
   } */
BigNumLength ZBNSqrt(r, a, al)
     BigNum r, a;
     BigNumLength al;
{
  BigNum x0, x1, x2;
  BigNumLength x0l = 1, x1l, x2l, b;

  if(ZBNIsOne(a, al) ||
     ((al == 1) && (ZBNDigit(a, 0) <= 3))){

    if(*a==0) {
      ZBNSetDigit(r, 0);
    }
    else {
      ZBNSetDigit(r, 1);
    }
    return(1);
  }

  /* Allocation */
  if (!(x1 =ZBNC(al + 2))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_SQRT, ZEN_ALLOC);
    return(0);
  }
  x1l = al;

  x2l = al+1;
  if (!(x2 =ZBNC(al+2))) {
    ZBNF(x1);
    ZENSetError(ZEN_LIB_ZBN, ZEN_SQRT, ZEN_ALLOC);
    return(0);
  }

  if (!(x0 = ZBNC(al + 2))) {
    ZBNF(x1); ZBNF(x2);
    ZENSetError(ZEN_LIB_ZBN, ZEN_SQRT, ZEN_ALLOC);
    return(0);
  }

  ZBNAssign(x1, a, al);
  if(al == 1){

    /* choose x0=a-1 */
    ZBNAssign(x0, a, al);
    ZBNSubtractBorrow(x0, (BigNumLength)1, (BigNumDigit)0);
  }
  else{

    /* choose x0=B^(al/2) */
    b = BN_DIGIT_SIZE*(al-1)+(!(*(a+al-1)) ? 1 : BN_DIGIT_SIZE-ZBNNumLeadingZeroBitsInDigit(*(a+al-1)));
    b = 1+ (b / 2);
    x0l = b / BN_DIGIT_SIZE;
    ZBNSetToZero(x0, x0l);
    ZBNSetDigit((x0+x0l), 1);
    ZBNShiftLeft((x0+x0l), (BigNumLength)1, (unsigned int)(b % BN_DIGIT_SIZE));
    x0l++;
  }

  while(ZBNCompareLazy(x0, x0l, x1, x1l) == -1){

    /* x2 <- a ou a+x0 : au pire x2[al] contient 1 */
    ZBNAssign(x2, a, al);
    ZBNSetDigit((x2+al), 0);
    if((*x0) && ((BigNumDigit)1)) {
      ZBNAdd(x2, x2l, x0, x0l, (BigNumDigit)0);
    }

    /* x1 <- x0 */
    ZBNAssign(x1, x0, x0l);
    x1l = x0l;

    /* x0 <- 2*x0 */
    ZBNDigit(x0, x0l) = 0;
    x0l++;
    ZBNAdd(x0, x0l, x0, x0l, (BigNumDigit)0);
    if(*(x0+x0l - 1)==0) {
      x0l--;
    }

    /* x2 <- quomod(x2, x0) */
    ZBNSetDigit((x2+x2l), 0);
    ZBNDivide(x2, x2l+1, x0, x0l);

    /* x0 <- x0/4 */
    ZBNShiftRight(x0, x0l, (unsigned int)2);

    /* x0 <- x0+x2[x0l..x2l-1] */
    ZBNAdd(x0, x0l, (x2+x0l), x2l - x0l, (BigNumDigit)0);
    x0l = ZBNNumDigits(x0, x0l);
  }
  ZBNAssign(r, x0, x0l);

  ZBNF(x0); ZBNF(x1); ZBNF(x2);
  return(x0l);
}
