/*
 * zen/ze2/c/elt_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element assignment}
   
   One can set a |ZENElt| to zero or one with |_Ze2EltSetToZero| or
   |_Ze2EltSetToOne|. Then we can assign a |ZENElt| with an other with
   |_Ze2EltAssign|. Furthermore, we can set randomly a |ZENElt| with
   |_Ze2EltSetRandom| or enumerate the elements of a finite ring with
   |_Ze2EltSetNext|. Finally, if we consider that a |ZENElt| is a
   polynomial in $T$, the function |_Ze2EltGetCoeff| and |_Ze2EltSetCoeff|
   enable us to get or set the coefficient of $T^i$, |_Ze2EltGetDegree|
   enable us to get its degree.
   }
   */
#include <stdio.h>
#include "ze2.h"

/* \TEX[
   \title{Assign.}
   \input{A pointer |pa| on a |ZENElt| and a |ZENElt| |b| of a finite ring |Rg|.}
   \side{|*pa| is filled with |b|.}
   }
   */
void _Ze2EltAssign(pa, b, Rg)
     ZENElt *pa, b;
     ZENRing Rg;
{
  Ze2EltAssign(*pa, b, Rg);
}

/* \TEX[
   \title{Set to zero.}
   \input{A pointer |pa| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*pa| is set to zero..}
   }
   */
void _Ze2EltSetToZero(pa, Rg)
     ZENElt *pa;
     ZENRing Rg;
{
  Ze2EltSetToZero(*pa, Rg);
}

/* \TEX[
   \title{Set to one.}
   \input{A pointer |pa| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*pa| is set to one.}
   }
   */
void _Ze2EltSetToOne(pa, Rg)
     ZENElt *pa;
     ZENRing Rg;
{
  Ze2EltSetToOne(*pa, Rg);
}

/* \TEX[
   \title{Set to random.}
   \input{A pointer |pa| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*pa| is set randomly.}
   }
   */
void _Ze2EltSetRandom(pa, Rg)
     ZENElt *pa;
     ZENRing Rg;
{
  Ze2EltSetRandom(*pa, Rg);
}

/* \TEX[
   \title{Enumerate.}
   \input{A pointer |pa| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*pa| is set randomly.}
   \note{After a number of call of this function equal to the number of elements in the finite ring, we obtain the same element.}
   }
   */
void _Ze2EltSetNext(pa, Rg)
     ZENElt *pa;
     ZENRing Rg;
{
  Ze2EltSetNext(*pa, Rg);
}

/* \TEX[
   \title{Conversion.}
   \input{A pointer |pa| on a |ZENElt| of a binary finite ring |Rg|, an element |b| of a modular ring |R|.}
   \side{|*pa| is set to the modular conversion of |b|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   }
   */
int _Ze2EltConvert(pa, Rg, b, R)
     ZENElt *pa,b;
     ZENRing Rg,R;
{
  switch(ZENRingArith(R)) {

  /* Fractions
   *************/
  case ZEN_LIB_ZEF:
    {
      BigNumDigit p[1];
      BigNumLength pl;
      
      ZENErrErrno = 0; pl = 1;
      ZENEltToZ(p, &pl, b, R);
      if ((ZENErrErrno == ZEN_INPUT) && (pl == 1)) {
	
	/* This element is a fraction, nothing can be done */
	ZENSetError(ZEN_LIB_ZE2, ZEN_CONVERT, ZEN_INPUT);
	return(ZERR);
      }
      if(p[0] & 1)
	Ze2EltSetToOne(*pa, R);
      else
	Ze2EltSetToZero(*pa, R);
    }
    break;
    
  /* Specific cases where things can be done faster
   *************************************************/
  case ZEN_LIB_ZE2:
    Ze2EltAssign(*pa, b, Rg);
    break;

    /* Default
     **********/
  default:

    /* R is a finite field of characteristic 2
       or a prime finite field ? */
    if (!ZBNCompare(ZENRingP(R), ZENRingPl(R),
		    ZENRingP(Rg), ZENRingPl(Rg)) ||
	!ZENRingDeg(R)) {
      BigNum p;
      BigNumLength pl;
      
      pl = ZENRingQl(R);
      if (!(p = ZBNC(pl))) {
	ZENSetError(ZEN_LIB_ZE2,ZEN_CONVERT,ZEN_ALLOC);
	return(ZERR);
      }
      ZENEltToZ(p,&pl,b,R);
      if(p[0]&1)
	Ze2EltSetToOne(*pa,Rg);
      else
	Ze2EltSetToZero(*pa,Rg);
      ZBNF(p);
      break;
    }

    /* Otherwise, what can we do ? */
    ZENSetError(ZEN_LIB_ZE2,ZEN_CONVERT,ZEN_INPUT);
    return(ZERR);
  }
  return(ZEN_HAS_INVERSE);
}






