/*
 * zen/ze2/c/elt_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{input/output}
   
   Conversions between the internal representation and the ASCII
   representation of a |ZENElt| are done by |_Ze2EltPrintToString|. |_Ze2EltPrintToFile|
   allows to write a |ZENElt| in a file.
   One can moreover write to and then read from a file a |ZENelt| with |_Ze2EltPutToFile|
   and |_Ze2EltReadFromFile|.
   }
   */
#include <stdio.h>
#include <ctype.h>
#include "ze2.h"

/* \TEX[
   \title{Converting from string.}
   \input{A finite field |Rg|, a pointer on a |ZENElt| |G| and a string
	    |s| representing a |ZENElt| in base $|base|\in \{2,\ldots,
	 16\}$.} 
\output{The number of character read in |s| or |ZENERR| if an error occured.}
\side{|G| is filled with |s| if the output is not |ZENERR|.}
}
*/
int _Ze2EltReadFromString(G, s, base, Rg)
     ZENElt *G;
     char *s;
     int base;
     ZENRing Rg;
{
  char *begin;


  for(begin = s; isspace(*begin); begin++) ;
  switch(*begin) {
  case '0':
    Ze2Elt2Int(*G, Rg) = 0;
    return(begin-s+1);

  case '1':
    Ze2Elt2Int(*G, Rg) = 1;
    return(begin-s+1);
  }
  ZENSetError(ZEN_LIB_ZE2, ZEN_READFROMSTRING, ZEN_SYNTAX);
  return(ZERR);
}

/* \TEX[
   \title{Converting to string.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{An allocated string representing |G| in base |base| or |ZENNULL| if an error occured.}   
   }
   */
char *_Ze2EltPrintToString(G, base, Rg)
     ZENElt G;
     int base;
     ZENRing Rg;  
{
  char *s;
  
  if (!(s = (char *)ZENMalloc(sizeof(char)*2))) {
    ZENSetError(ZEN_LIB_ZE2, ZEN_PRINTTOSTRING, ZEN_ALLOC);
    return(ZENNULL);
  }
  if (Ze2EltIsZero(G, Rg)) s[0] = '0'; else s[0] = '1';
  s[1] = '\0';
  
  return(s);
}


/* \TEX[
   \title{Converting from a string in an internal representation.}
   \input{A finite field |Rg| and a string |s| representing a |ZENElt|
   in an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|G| is filled with s if the output is different from |ZENNULL|}
   }
   */
int _Ze2EltGetFromString(G, s, Rg)
     char *s;
     ZENElt *G;
     ZENRing Rg;
{
  return(_Ze2EltReadFromString(G, s, 10, Rg));
}

/* \TEX[
   \title{Converting to string in an internal representation.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg|.}
   \output{An allocated string representing |G| in an internal representation or |ZENNULL| if an error occured.}   
   }
   */
char *_Ze2EltPutToString(G, Rg)
     ZENElt G;
     ZENRing Rg;
{
  return(_Ze2EltPrintToString(G, 10, Rg));
}


/* \TEX[
   \title{Printing to file.}
   \input{A stream |file|, an allocated |ZENElt| |G| of a finite field
	    |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
	    \output{|ZENERR| if an error occured, 0 otherwise.}
	    \side{Printing a representation of |G| in base |base| to |file|.}
	    }
	    */
int _Ze2EltPrintToFile(file, G, base, Rg)
     FILE *file;
     ZENElt G;
     int base;
     ZENRing Rg;  
{
  
  if (Ze2EltIsZero(G, Rg)) {
    if(fprintf(file, "0") < 0) {
      ZENSetError(ZEN_LIB_ZE2, ZEN_PRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
  }
  else {
    if (fprintf(file, "1") < 0) {
      ZENSetError(ZEN_LIB_ZE2, ZEN_PRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
  }
  return(0);
}

/* \TEX[
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |ZENElt| |G| of a finite field
   |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |G| to |file|.}
   }
   */
int _Ze2EltPutToFile(file, G, Rg)
     FILE *file;
     ZENElt G;
     ZENRing Rg;  
{
  return(_Ze2EltPrintToFile(file, G, 10, Rg));
}

/* \TEX[
   \title{Reading from file.}
   \input{A stream |file| and an element |G| of a finite field |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|G| is filled with the ZENElt read in |file| if no error occurred.}
   }
   */
int _Ze2EltReadFromFile(G, file, base, Rg)
     ZENElt *G;
     FILE *file;
     int base;
     ZENRing Rg;
{
  int c;
  

  while(isspace(c = fgetc(file))) ;

  switch(c) {
  case '0':
    Ze2Elt2Int(*G, Rg) = 0;
    return(0);

  case '1':
    Ze2Elt2Int(*G, Rg) = 1;
    return(0);
  }
  ZENSetError(ZEN_LIB_ZE2, ZEN_READFROMFILE, ZEN_SYNTAX);
  return(ZERR);
}

/* \TEX[
   \title{Getting from file in an internal representation.}
   \input{A stream |file|, and an element |G| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|G| is filled with the element read in |file|}
   }
   */
int _Ze2EltGetFromFile(G, file, Rg)
     ZENElt *G;
     FILE *file;
     ZENRing Rg;
{
  return(_Ze2EltReadFromFile(G, file, 10, Rg));
}


/* \TEX{
   \title{Evaluating an element as a multivariate polynomial over $Z$.}
   \input{A Bignum |p| of allocated size greater or equal than
   |ZENRingQl(Rg)|, a pointer |p_pl| on the real size of |p|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|p| is filled with the value of |G| considered as a
   multivariate polynomial in which we substitute all the variables
   with 2, |*p_pl| contains the real size of |p|.}
   } */
void _Ze2EltToZ(p, p_pl, G, Rg)
     BigNum p;
     BigNumLength *p_pl;
     ZENElt G;
     ZENRing Rg;
{
  Ze2EltToZ(p, p_pl, G, Rg);
}

/* \TEX{
   \title{Getting an element from its evaluation as a multivariate
   polynomial over $Z$.}
   \input{A Bignum |p| of size |pl|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|G| is filled with the element whose evaluation considered as a
   multivariate polynomial in which we substitute all the variables
   with 2 is equal to (|p|, |pl|).}
   } */
void _Ze2ZToElt(G, p, pl, Rg)
     BigNum p;
     BigNumLength pl;
     ZENElt *G;
     ZENRing Rg;
{
  Ze2ZToElt(*G, p, pl, Rg);
}
