/*
 * zen/ze2/c/init.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   A finite ring is completely described by a variable of type
   |ZENRing|. Precomputations are needed to speedup the operations in this
   ring. In the particular case of $ZEN(2)$, this is done with
   |_Ze2Init|. These precomutations are then free with |_Ze2Close|.
   }
   */
#include <stdio.h>
#include "ze2.h"
#include "zed.h"
#include "zeg.h"

/* \TEX{
   \title{Initialization.}
   \input{A partially defined |ZENRing|.}
   \output{|SR_GEN|.}
   \side{The initialization of $\Z/2\Z$ is completed.}
   \note{This procedure is called by |ZENBaseRingCreate| and should not be called otherwise.}
   }
   */
int _Ze2Init(R)
     ZENRing R;
{
  ZENRingArith(R) = ZEN_LIB_ZE2;

  /* Procedure on rings
   **********************/
  ZENPrcInitEltMultiplyPtr(R) = NULL;
  ZENPrcCloseEltMultiplyPtr(R) = NULL;

  ZENPrcInitPolyRootsCanonicalPtr(R) = NULL;
  ZENPrcClosePolyRootsCanonicalPtr(R) = NULL;

  ZENPrcInitEltExpPtr(R) = NULL;
  ZENPrcCloseEltExpPtr(R) = NULL;

  /* Procedure on elements
   ****************************/
  /* Allocation */
  ZENEltCreatePtr(R) = _Ze2EltCreate;	/* Create elements */
  ZENEltCreateLargePtr(R) = _Ze2EltCreate;	/* Create elements */
  ZENEltFreePtr(R) = _Ze2EltFree;	/* Free elements */

  /* Assigning */
  ZENEltAssignPtr(R) = _Ze2EltAssign;	/* Assign an element with another */
  ZENEltSetToZeroPtr(R) = _Ze2EltSetToZero; /* Set an element to zero */
  ZENEltSetToOnePtr(R) = _Ze2EltSetToOne; /* Set an element to one */
  ZENEltSetRandomPtr(R) = _Ze2EltSetRandom; /* Set randomly an element */
  ZENEltSetNextPtr(R) = _Ze2EltSetNext; /* Set an element to the next */
  ZENEltConvertPtr(R) = _Ze2EltConvert;

  /* Test */
  ZENEltAreEqualPtr(R) = _Ze2EltAreEqual; /* Is an element equal to another */
  ZENEltIsZeroPtr(R) = _Ze2EltIsZero;	/* Is an element equal to zero */
  ZENEltIsOnePtr(R) = _Ze2EltIsOne;	/* Is an element equal to one */
  ZENEltIsASquarePtr(R) = _ZENEltIsASquareChar2; /* Is an element a square ? */

  /* Arithmetic */
  ZENEltAddPtr(R) = _Ze2EltAdd;	/* Addition of elements */
  ZENEltNegatePtr(R) = _Ze2EltAssign;	/* Negation of an element */
  ZENEltSubtractPtr(R) = _Ze2EltAdd; /* Substraction of elements */
  ZENEltSquarePtr(R) = _Ze2EltSquare;	/* Squaring of elements */
  ZENEltMultiplyPtr(R) = _Ze2EltMultiply; /* Multiplication of elements */
  ZENEltInversePtr(R) = _Ze2EltInverse; /* Inverse of elements */
  ZENEltTracePtr(R) = _Ze2EltTrace;		/* Trace of elements */
  ZENEltSquareRootPtr(R) = _ZENEltSquareRootChar2; /* Square root of an element */
  
  /* io */
  ZENEltReadFromStringPtr(R) = _Ze2EltReadFromString;	/* Read an element from a string */
  ZENEltPrintToStringPtr(R) = _Ze2EltPrintToString; /* Print an element to a string */
  ZENEltReadFromFilePtr(R) = _Ze2EltReadFromFile; /* Read an element from a file */
  ZENEltPrintToFilePtr(R) = _Ze2EltPrintToFile; /* Print an element to a file */
  ZENEltGetFromStringPtr(R) = _Ze2EltGetFromString; /* Get an element from a string */
  ZENEltPutToStringPtr(R) = _Ze2EltPutToString; /* Put an element to a string */
  ZENEltGetFromFilePtr(R) = _Ze2EltGetFromFile; /* Get an element from a file */
  ZENEltPutToFilePtr(R) = _Ze2EltPutToFile; /* Put an element to a file */

  ZENZToEltPtr(R) = _Ze2ZToElt; /* Evaluate in $p$ an element */
  ZENEltToZPtr(R) = _Ze2EltToZ; /* Get the element from an evaluation in $p$ */

  /* Procedures on polynomials
   *******************************/
  /* Allocation */
  ZENPolyCreatePtr(R) = _Ze2PolyCreate; /* Create polynomials */
  ZENPolyCreateLargePtr(R) = _Ze2PolyCreate; /* Create polynomials */
  ZENPolyFreePtr(R) = _Ze2PolyFree;	/* Free polynomials */

  /* Assigning */
  ZENPolyAssignPtr(R) = _Ze2PolyAssign; /* Assign a polynomial with another */
  ZENPolyAssignLargePtr(R) = _Ze2PolyAssign; /* Assign a polynomial with another */
  ZENPolyLazyToVeryLazyPtr(R) = _Ze2PolyAssign; /* Assign a polynomial with another */
  ZENPolySetToZeroPtr(R) = _Ze2PolySetToZero;	/* Set a polynomial to zero */
  ZENPolySetToZeroLargePtr(R) = _Ze2PolySetToZero;	/* Set a polynomial to zero */
  ZENPolySetToXiPtr(R) = _Ze2PolySetToXi;	/* Set a polynomial to X^i */
  ZENPolySetRandomPtr(R) = _Ze2PolySetRandom;	/* Set randomly a polynomial */
  ZENPolySetCoeffPtr(R) = _Ze2PolySetCoeff; /* Set a coefficient of a polynomial */
  ZENPolyGetCoeffNewPtr(R) = _Ze2PolyGetCoeff; /* Get a coefficient of a polynomial */
  ZENPolyGetCoeffPtrPtr(R) = _Ze2PolyGetCoeffPtr; /* Get a pointer on coefficient of a polynomial */
  ZENPolyUpdateDegreePtr(R) = _Ze2PolyUpdateDegree; /* Get a coefficient of a polynomial */

  /* Test */
  ZENPolyAreEqualPtr(R) = _Ze2PolyAreEqual; /* Is a polynomial equal to another */
  ZENPolyIsZeroPtr(R) = _Ze2PolyIsZero; /* Is a polynomial equal to zero */
  ZENPolyIsXiPtr(R) = _Ze2PolyIsXi; /* Is a polynomial equal to $X^i$ */

  /* Arithmetic */
  ZENPolyAddPtr(R) = _Ze2PolyAdd;	/* Addition of polynomials */
  ZENPolyAddLargePtr(R) = _Ze2PolyAdd;	/* Addition of polynomials */
  ZENPolyNegatePtr(R) = _Ze2PolyAssign; /* Negation of polynomials */
  ZENPolyNegateLargePtr(R) = _Ze2PolyAssign; /* Negation of polynomials */
  ZENPolySubtractPtr(R) = _Ze2PolyAdd; /* Substraction of polynomials */
  ZENPolySubtractLargePtr(R) = _Ze2PolyAdd; /* Substraction of polynomials */

  ZENPolySquarePtr(R) = _Ze2PolySquare; /* Squaring of polynomials */
  ZENPolySquareLazyPtr(R) = _Ze2PolySquare;
  ZENPolySquareVeryLazyPtr(R) = _Ze2PolySquare;

  ZENPolyMultiplyPtr(R) = _Ze2PolyMultiply; /* Multiplication of polynomials */
  ZENPolyMultiplyLazyPtr(R) = _Ze2PolyMultiply;
  ZENPolyMultiplyVeryLazyPtr(R) = _Ze2PolyMultiply;

  ZENPolyMultiplyScalarPtr(R) = _Ze2PolyMultiplyScalar; /* Multiplication of polynomials by a scalar*/
  ZENPolyDotPtr(R) = _Ze2PolyDot; /* Scalar multiplication of polynomials */
  ZENPolyEvalPtr(R) = _Ze2PolyEval; /* Evaluation of a polynomial */
  ZENPolyDividePtr(R) = _Ze2PolyDivide; /* Division of polynomials */
  ZENPolyDivideLazyPtr(R) = _Ze2PolyDivideLazy; /* Lazy division of polynomials */
  ZENPolyDivideVeryLazyPtr(R) = _Ze2PolyDivideLazy; /* Lazy division of polynomials */
  ZENPolyModuloLazyPtr(R) = _Ze2PolyModuloLazy; /* Lazy modulo of polynomials */
  ZENPolyModuloVeryLazyPtr(R) = _Ze2PolyModuloLazy; /* Lazy modulo of polynomials */
  ZENPolyDerivePtr(R) = _Ze2PolyDerive;
  ZENPolyRootsDegree2Ptr(R) = _ZENPolyRootsDegree2Char2; /* Roots of polynomials of degree 2 */

  /* Procedures on matrix
   **************************/
  /* Allocation */
  ZENMatCreateParamsPtr(R) = _Ze2MatCreateParams;	/* Create a matrix */
  ZENMatCreatePtr(R) = __ZedsMatCreate;	/* Create a matrix */

  /* Assigning */
  ZENMatAssignPtr(R) = _Ze2MatAssign;	/* Assign a matrix with another */
  ZENPermutation2MatrixPtr(R) = _Ze2Permutation2Matrix;	/* Assign a matrix with another */
  ZENMatSetToZeroPtr(R) = _Ze2MatSetToZero; /* Set a matrix to zero */
  ZENMatSetToOnePtr(R) = _Ze2MatSetToOne; /* Set a matrix to identity */
  ZENMatSetRandomPtr(R) = _Ze2MatSetRandom; /* Set randomly a matrix */
  ZENMatSetBlocPtr(R) = _Ze2MatSetBloc; /* Set a coefficient of a matrix */
  ZENMatGetBlocNewPtr(R) = _Ze2MatGetBloc; /* Get a coefficient of a matrix */
  ZENMatGetBlocPtrPtr(R) = _Ze2MatGetBlocPtr; /* Get a coefficient of a matrix */
  ZENMatSetSubMatPtr(R) = _Ze2MatSetSubMat; /* Set a submatrix */
  ZENMatGetSubMatPtr(R) = _Ze2MatGetSubMat; /* Get a submatrix */

  ZENMatPermuteBnPtr(R) = _Ze2MatPermuteBn;
  ZENMatPermuteDigitPtr(R) = _Ze2MatPermuteDigit;

  /* Test */
  ZENMatAreEqualPtr(R) = _Ze2MatAreEqual; /* Is a matrix equal to another */
  ZENMatIsZeroPtr(R) = _Ze2MatIsZero;	/* Is a matrix equal to zero */
  ZENMatIsOnePtr(R) = _Ze2MatIsOne;	/* Is a matrix equal to one */

  /* Arithmetic */
  ZENMatAddPtr(R) = _Ze2MatAdd;	/* Addition of Matrix */
  ZENMatAddScalarPtr(R) = _Ze2MatAddScalar;	/* Addition of Matrix */
  ZENMatSubtractPtr(R) = _Ze2MatSubtract; /* Substraction of Matrix */
  ZENMatNegatePtr(R) = _Ze2MatNegate;	/* Squaring of Matrix */
  ZENMatMultiplyPtr(R) = _Ze2MatMultiplyPlain; /* Multiplication of Matrix */
  ZENMatMultiplyScalarPtr(R) = _Ze2MatMultiplyScalar; /* Multiplication of Matrix */
  ZENMatMultiplyPlainPtr(R) = _Ze2MatMultiplyPlain; /* Multiplication of Matrix */
  ZENMatMultiplyLeftPermutationPtr(R) = _Ze2MatMultiplyLeftPermutation;
  ZENMatMultiplyRightPermutationPtr(R) = _Ze2MatMultiplyRightPermutation;
  ZENMatInversePtr(R) = _Ze2MatInverse; /* Inverse of a Matrix */
  ZENMatKernelPtr(R) = _Ze2MatKernel;	/* Kernel of a Matrix */
  ZENMatGaussPtr(R) = _Ze2MatGauss;	
  /* IO */
  ZENMatGetFromFilePtr(R) = _Ze2MatGetFromFile;
  ZENMatPutToFilePtr(R) = _Ze2MatPutToFile;
  ZENMatGetFromStringPtr(R) = _Ze2MatGetFromString;
  ZENMatPutToStringPtr(R) = _Ze2MatPutToString;

  /* An element of trace 1 */
  ZENEltSetToOne(ZENRingTheta(R), R);

  return(SR_GEN);
}

/* \TEX[
   \title{Closing.}
   \input{A partially freed |ZENRing|.}
   \output{0.}
   }
   */
int _Ze2Close(R)		
     ZENRing R;
{
  return(0);
}
