/*
 * zen/ze2/c/mat_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial additions}

   These functions are specific to  $\Z/2\Z$. 
   }
   */
#include <stdio.h>
#include "ze2.h"

/*
   \TEX{
   \title{Addition.}
   \input{Two |Ze2Mat| |A| and |B| of same type of a finite ring |R|.}
   \side{|A| += |B|}
   }
   */

void _Ze2MatSameTypeAdd(A,B,R)
     ZENMat A,B;
     ZENRing R;
{BigNumDigit *ptA,*ptB,*ptrlim;

 ptrlim= (ptA=Ze2MatFirst(A,R)) + Ze2MatNbBnd(A,R);
 ptB=Ze2MatFirst(B,R);

 while(ptA<ptrlim)
   *ptA++ ^= *ptB++;

#if 0
 int l,c;
   
 for(l=Ze2MatNbBn(A,R)-1;l>=0;l--)
   for(c=Ze2MatNbBloc(A,R)-1;c>=0;c--)
     Ze2MatBloc(A,l,c,R) ^= Ze2MatBloc(B,l,c,R);
#endif
 return;
}
/*
   \TEX{
   \title{Addition.}
   \input{Two |Ze2Mat| |A| and |B| of opposite type of a finite ring |R|.}
   \side{|A| += |B|}
   }
   */

void _Ze2MatOppositeTypeAdd(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c,dl;
 BigNumDigit El;
 
 l=Ze2MatNbBn(A,R)-1;
 dl=divSizeBloc(l);
 El=BlocExp2(modSizeBloc(l));
 while(l>=0)
   {
     for(c=Ze2MatNbBn(B,R)-1;c>=0;c--)
       if(Ze2MatLnCoeffTest(B,c,dl,El,R))
	 Ze2MatLnChangeCoeff(A,l,c,R);
     l--;
     El>>=1;
     if(El==0)
       {
	 dl--;
	 El=BlocExp2(SIZE_BLOC-1);
       } 
   }
}

/*
   \TEX[
   \title{Addition.}
   \input{Two |Ze2Mat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| += |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int _Ze2MatAdd(A,B,R)
     ZENMat A,B;
     ZENRing R;
{
  ZENTest(Ze2MatIsPermutation(A,R)
	  ||Ze2MatIsPermutation(B,R),
	  ZEN_ISPERM,ZEN_MATADD,ZEN_LIB_ZE2, ZERR);
 
  if(Ze2MatAreSameType(A,B,R))
    {
      ZENTest((Ze2MatNbBn(A,R) != Ze2MatNbBn(B,R))
	      ||(Ze2MatNbDigit(A,R) != Ze2MatNbDigit(B,R)),
	      ZEN_MAT_DIM,ZEN_MATADD,ZEN_LIB_ZE2, ZERR);
      Ze2MatSameTypeAdd(A,B,R);
    }
  else
    {      
      ZENTest((Ze2MatNbBn(A,R) != Ze2MatNbDigit(B,R))
	      ||(Ze2MatNbDigit(A,R) != Ze2MatNbBn(B,R)),
	      ZEN_MAT_DIM,ZEN_MATADD,ZEN_LIB_ZE2, ZERR);
      Ze2MatOppositeTypeAdd(A,B,R);
    }
  return(0);
}

/*
   \TEX[
   \title{Addition of scalar}
   \input{A |ZENMat|, a |ZENElt| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{All the elements of |A| are incremented by |b|.}
   \note{|A| cannot be a permutation.}
   }
   */
int _Ze2MatAddScalar(A,b,R)
     ZENMat A;
     ZENElt b;
     ZENRing R;
{
  ZENTest(Ze2MatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATADD,ZEN_LIB_ZE2, ZERR);
  Ze2MatAddScalar(A,b,R);
  return(0);
}

/*
   \TEX{
   \title{Addition of scalar}
   \input{A |ZENMat| and a |ZENRing|}
   \side{All the elements of |A| are incremented by one.}
   \note{|A| cannot be a permutation.}
   }
   */
void _Ze2MatAddScalarOne(A,R)
     ZENMat A;
     ZENRing R;
{
  BigNumDigit *ptA,*ptrlim;

  ptrlim= (ptA=Ze2MatFirst(A,R)) + Ze2MatNbBnd(A,R);

  for(;ptA<ptrlim;ptA++)
    *ptA = ~(*ptA);

  return;
}

 /* \TEX{
   \title{Negation.}
   \input{Two |Ze2Mat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = -|B| if no error occured, |zen_errno| is set otherwise.}
   }
   */

int _Ze2MatNegate(A,B,R)
     ZENMat A,B;
     ZENRing R;
{
  return(Ze2MatNegate(A,B,R));
}

/*
   \TEX{
   \title{Subtraction.}
   \input{Two |Ze2Mat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| -= |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int _Ze2MatSubtract(A,B,R)
     ZENMat A,B;
     ZENRing R;
{
  return(Ze2MatAdd(A,B,R));
}

   
