/*
 * zen/ze2/c/mat_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial assignments}

   These functions are specific to $\Z/2\Z$.
   }
   */
#include <stdio.h>
#include "ze2.h"

#ifndef min
#  define min(x,y) ((x)<(y)?(x):(y))
#endif

/*
   \TEX{
   \title{Assignment.}
   \input{Two |Ze2Mat| |A| and |B| of opposite type of a finite ring |R|.}
   \side{|A| = |B|}
   }
   */
   
void _Ze2MatOppositeTypeAssign(A,B,R)
     ZENMat A,B;
     ZENRing R;
{
  Ze2MatSetToZero(A,R);
  Ze2MatOppositeTypeAdd(A,B,R);
}

/*
   \TEX[
   \title{Assignment.}
   \input{Two |Ze2Mat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
int _Ze2MatAssign(A,B,R)
     ZENMat A,B;
     ZENRing R;
{
  ZENTest(Ze2MatIsPermutation(A,R)
	  ||Ze2MatIsPermutation(B,R),
	  ZEN_ISPERM,ZEN_MATASSIGN,ZEN_LIB_ZE2, ZERR);
 
  if(Ze2MatAreSameType(A,B,R))
    {
      ZENTest((Ze2MatNbBn(A,R) != Ze2MatNbBn(B,R))
	      ||(Ze2MatNbDigit(A,R) != Ze2MatNbDigit(B,R)),
	      ZEN_MAT_DIM,ZEN_MATASSIGN,ZEN_LIB_ZE2, ZERR);
      Ze2MatSameTypeAssign(A,B,R);
    }
  else
    {      
      ZENTest((Ze2MatNbBn(A,R) != Ze2MatNbDigit(B,R))
	      ||(Ze2MatNbDigit(A,R) != Ze2MatNbBn(B,R)),
	      ZEN_MAT_DIM,ZEN_MATASSIGN,ZEN_LIB_ZE2, ZERR);
      Ze2MatOppositeTypeAssign(A,B,R);
    }
  return(0);
}

/*
   \TEX[
   \title{Set to zero}
   \input{A |Ze2Mat| |A| of a finite ring |R|.}
   \output{0}
   \side{All the coefficients of |A| are set to zero.}
   }
   */
   
int _Ze2MatSetToZero(A,R)
     ZENMat A;
     ZENRing R;
{
  Ze2MatSetToZero(A,R);
  return(0);
}

/*
   \TEX[
   \title{Set to one}
   \input{A |Ze2Mat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are set to zero except for the coefficients with same indices on rows and columns that are set to one.}
   \note{There is no error if the matrix is not square.}
   }
   */
   
int _Ze2MatSetToOne(A,R)
     ZENMat A;
     ZENRing R;
{int i,di;
 BigNumDigit Emi;

 ZENTest(Ze2MatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETONE,ZEN_LIB_ZE2, ZERR);
 
 Ze2MatSetToZero(A,R);
 
 i=min(Ze2MatNbRow(A,R),Ze2MatNbCol(A,R))-1;
 di=divSizeBloc(i);
 Emi=BlocExp2(modSizeBloc(i));
 
 while(i>=0)
   {
     Ze2MatLnChangeCoeffFast(A,i,di,Emi,R);
     i--;
     Emi>>=1;
     if(Emi == 0)
       {
	 di--;
	 Emi=BlocExp2(SIZE_BLOC-1);
       }
   }
 return(0);
}

/*
   \TEX[
   \title{Set to random}
   \input{A |Ze2Mat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are randomly set.}
   }
   
   */
   
int _Ze2MatSetRandom(A,R)
     ZENMat A;
     ZENRing R;
{int i,j;
 BigNumDigit mask;


 ZENTest(Ze2MatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETRANDOM,ZEN_LIB_ZE2, ZERR);
 
 mask=BlocExp2(modSizeBloc(Ze2MatNbDigit(A,R)))-1;
 
 for(i=Ze2MatNbBn(A,R)-1;i>=0;i--)
   {
     j=Ze2MatNbBloc(A,R)-1;
     if (mask)
       {
	 Ze2MatBloc(A,i,j,R) = zbnrandom() & mask;
	 j--;
       }
     
     for(;j>=0;j--)
       Ze2MatBloc(A,i,j,R) = zbnrandom();
   }

 return(0);
}

/*
   \TEX{
   \title{Get a bloc}
   \input{A |Ze2Mat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{|E| is set to the indexed bloc of |A|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred}
   }
   
   */
   
int _Ze2MatGetBloc(p_E,A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENElt *p_E;
     ZENRing R;
{
  ZENTest(Ze2MatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATGETBLOC,ZEN_LIB_ZE2, ZERR);
  ZENTest((l<0)||(c<0)||(l>=Ze2MatNbBn(A,R))||(c>=Ze2MatNbDigit(A,R)),
	  ZEN_MAT_DIM0,ZEN_MATGETBLOC,ZEN_LIB_ZE2, ZERR);

  if(Ze2MatLnCoeff(A,l,c,R))
    Ze2EltSetToOne(*p_E,R);
  else
    Ze2EltSetToZero(*p_E,R);

  return(0);
}

/*
   \TEX{
   \title{Get a bloc pointer}
   \input{A |Ze2Mat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{|E| is set to the indexed bloc of |A|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred}
   }
   
   */
   
ZENElt _Ze2MatGetBlocPtr(A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENRing R;
{ZENElt e;
 
  if(Ze2MatLnCoeff(A,l,c,R))
    Ze2Elt2Int(e,R)=1;
  else
    Ze2Elt2Int(e,R)=0;

  return(e);
}

/*
   \TEX{
   \title{Set a bloc}
   \input{A |Ze2Mat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The indexed bloc of |A| is set to |E|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred}
   }
   
   */
   
int _Ze2MatSetBloc(A,l,c,E,R)
     ZENMat A;
     Dim l,c;
     ZENElt E;
     ZENRing R;
{
  int dc;
  BigNumDigit Ec;
 
  ZENTest(Ze2MatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATSETBLOC,ZEN_LIB_ZE2, ZERR);
  ZENTest((l<0)||(c<0)||(l>=Ze2MatNbBn(A,R))||(c>=Ze2MatNbDigit(A,R)),
	  ZEN_MAT_DIM0,ZEN_MATSETBLOC,ZEN_LIB_ZE2, ZERR);
 
  dc=divSizeBloc(c);
  Ec=BlocExp2(modSizeBloc(c));
 
 if(opXor(Ze2Elt2Int(E,R),Ze2MatLnCoeffTest(A,l,dc,Ec,R)))
   Ze2MatLnChangeCoeffFast(A,l,dc,Ec,R);

 return(0);
}

/********************************************************/

static int __Ze2MatGetSubMatST(S,A,l,c,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{BigNum n;
 int nl,dc,mc,i;
 BigNumDigit mask;

 dc=divSizeBloc(c);
 mc=modSizeBloc(c);
 nl=Ze2MatNbBloc(A,R)-dc;
 n=ZBNC(nl);

 ZENTest(n==NULL,
	 ZEN_ALLOC,ZEN_MATGETSUBMAT,ZEN_LIB_ZE2, ZERR);
 if(mc)
   for(i=0;i<Ze2MatNbBn(S,R);i++,l++)
     {
       ZBNAssign(n,Ze2MatBn(A,l,R)+dc,nl);
       ZBNShiftRight(n,nl,mc);
       ZBNAssign(Ze2MatBn(S,i,R),n,Ze2MatNbBloc(S,R));
     }
 else
   for(i=0;i<Ze2MatNbBn(S,R);i++,l++)
     ZBNAssign(Ze2MatBn(S,i,R),Ze2MatBn(A,l,R)+dc,Ze2MatNbBloc(S,R));

 mask=BlocExp2(modSizeBloc(Ze2MatNbDigit(S,R)))-1;
 if(mask)
   {
     nl=Ze2MatNbBloc(S,R)-1;
     for(i=Ze2MatNbBn(S,R);i--;)
       Ze2MatBloc(S,i,nl,R) &= mask;
   }

 ZBNF(n);
 return(0);
}

static void __Ze2MatGetSubMatOT(S,A,l,c1,R)
     ZENMat A,S;
     Dim l,c1;
     ZENRing R;
{int dl,k,c,c2,dc;
 BigNumDigit Ec,El;

 Ze2MatSetToZero(S,R);

 l+=(k=Ze2MatNbBn(S,R)-1);
 dl=divSizeBloc(l);
 El=BlocExp2(modSizeBloc(l));
 while(k>=0)
   {
     c=Ze2MatNbDigit(S,R)-1;
     dc=divSizeBloc(c);
     Ec=BlocExp2(modSizeBloc(c));
     c2=c+c1;
     while(c>=0)
       {
	 if(Ze2MatColCoeffTest(A,dl,El,c2,R))
	   Ze2MatLnChangeCoeffFast(S,k,dc,Ec,R);
	 c--;
	 c2--;
	 Ec>>=1;
	 if(Ec==0)
	   {
	     dc--;
	     Ec=BlocExp2(SIZE_BLOC-1);
	   }
       }
     k--;
     El>>=1;
     if(El==0)
       {
	 dl--;
	 El=BlocExp2(SIZE_BLOC-1);
       }
   }
}
 
static int __Ze2MatSetSubMatST(A,l,c,S,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{BigNum n;
 int nl,dc,d2c,mc,i;
 BigNumDigit mask2;

 dc=divSizeBloc(c);
 mc=modSizeBloc(c);
 nl=c+Ze2MatNbDigit(S,R);
 d2c=divSizeBloc(nl-1);
 mask2=BlocExp2(modSizeBloc(nl))-1;
 nl=d2c-dc+1;
 n=ZBNC(nl);
 ZENTest(n==NULL,
	 ZEN_ALLOC,ZEN_MATSETSUBMAT,ZEN_LIB_ZE2, ZERR);

 if(mc)
   {
     BigNumDigit mask1;
     int dnl;

     mask1=BlocExp2(mc)-1;
     dnl=nl-Ze2MatNbBloc(S,R);	/* gives only 0 or 1 */

     if(dnl)
       {
	 if(mask2)		/*  dnl && mask2 */
	   {
	     mask2 = ~mask2;
	     for(i=0;i<Ze2MatNbBn(S,R);i++,l++)
	       {
		 ZBNAssign(n,Ze2MatBn(S,i,R),Ze2MatNbBloc(S,R));
		 n[nl-1]=0;
		 ZBNShiftLeft(n,nl,mc);
		 n[0]^=(Ze2MatBloc(A,l,dc,R) & mask1);
		 n[nl-1] ^= (Ze2MatBloc(A,l,d2c,R) & mask2);
		 ZBNAssign(Ze2MatBn(A,l,R)+dc,n,nl);
	       }
	   }
	 else			/* dnl && !mask2 */
	   for(i=0;i<Ze2MatNbBn(S,R);i++,l++)
	     {
	       ZBNAssign(n,Ze2MatBn(S,i,R),Ze2MatNbBloc(S,R));
	       n[nl-1]=0;
	       ZBNShiftLeft(n,nl,mc);
	       n[0]^=(Ze2MatBloc(A,l,dc,R) & mask1);
	       ZBNAssign(Ze2MatBn(A,l,R)+dc,n,nl);
	     }
       }
     else
       {
	 if(mask2)		/*  mc && !dnl && mask2 */
	   {
	     mask2 = ~mask2;
	     for(i=0;i<Ze2MatNbBn(S,R);i++,l++)
	       {
		 ZBNAssign(n,Ze2MatBn(S,i,R),Ze2MatNbBloc(S,R));
		 ZBNShiftLeft(n,nl,mc);
		 n[0]^=(Ze2MatBloc(A,l,dc,R) & mask1);
		 n[nl-1] ^= (Ze2MatBloc(A,l,d2c,R) & mask2);
		 ZBNAssign(Ze2MatBn(A,l,R)+dc,n,nl);
	       }
	   }
	 else			/* mc && !dnl && !mask2 */
	   for(i=0;i<Ze2MatNbBn(S,R);i++,l++)
	     {
	       ZBNAssign(n,Ze2MatBn(S,i,R),Ze2MatNbBloc(S,R));
	       ZBNShiftLeft(n,nl,mc);
	       n[0]^=(Ze2MatBloc(A,l,dc,R) & mask1);
	       ZBNAssign(Ze2MatBn(A,l,R)+dc,n,nl);
	     }
       }
   }
 else
   {
     if(mask2)		/*  !mc && mask2 */
	   {
	     mask2 = ~mask2;
	     for(i=0;i<Ze2MatNbBn(S,R);i++,l++)
	       {
		 ZBNAssign(n,Ze2MatBn(S,i,R),Ze2MatNbBloc(S,R));
		 n[nl-1] ^= (Ze2MatBloc(A,l,d2c,R) & mask2);
		 ZBNAssign(Ze2MatBn(A,l,R)+dc,n,nl);
	       }
	   }
	 else			/* !mc && !mask2 */
	   for(i=0;i<Ze2MatNbBn(S,R);i++,l++)
	     ZBNAssign(Ze2MatBn(A,l,R)+dc,Ze2MatBn(S,i,R),Ze2MatNbBloc(S,R));
   }

 ZBNF(n);
 return(0);
}

static void __Ze2MatSetSubMatOT(A,l,c1,S,R)
     ZENMat A,S;
     Dim l,c1;
     ZENRing R;
{int dl,k,c,c2,dc;
 BigNumDigit El,Ec;

 l += (k=Ze2MatNbBn(S,R)-1);
 dl=divSizeBloc(l);
 El=BlocExp2(modSizeBloc(l));
 while(k>=0)
   {
     c=Ze2MatNbDigit(S,R)-1;
     dc=divSizeBloc(c);
     Ec=BlocExp2(modSizeBloc(c));
     c2=c+c1;
     while(c>=0)
       {
	 if(opXor(Ze2MatColCoeffTest(A,dl,El,c2,R),
		  Ze2MatLnCoeffTest(S,k,dc,Ec,R)))
	   Ze2MatColChangeCoeffFast(A,dl,El,c2,R);
	 c--;
	 c2--;
	 Ec>>=1;;
	 if(Ec==0)
	   {
	     dc--;
	     Ec=BlocExp2(SIZE_BLOC-1);
	   }
       }
     k--;
     El >>= 1;
     if(El == 0)
       {
	 dl--;
	 El=BlocExp2(SIZE_BLOC-1);
       }
   }	   
}

/*
   \TEX[
   \title{Getting a submatrix}
   \input{Two allocated |ZENMat|, two indexes and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The submatrix of |M| at position $(|r|,|c|)$ of size those of |S| is extracted and assigned to |S|.}
   \note{Permutation matrix cannot be used.}
   }
   */

int _Ze2MatGetSubMat(S,A,l,c,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int cas;
 
 ZENTest((ZENMatIsPermutation(A,R))||(ZENMatIsPermutation(S,R)),
	 ZEN_ISPERM,ZEN_MATGETSUBMAT,ZEN_LIB_ZE2, ZERR);
 ZENTest((l<0)||(c<0)||((l+ZENMatNbRow(S,R))>ZENMatNbRow(A,R))||((c+ZENMatNbCol(S,R))>ZENMatNbCol(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATGETSUBMAT,ZEN_LIB_ZE2, ZERR);

 cas=(Ze2MatIsRowType(A,R)<<1)|(Ze2MatIsRowType(S,R));

 switch(cas)
   {
   case _RR_:
     return(__Ze2MatGetSubMatST(S,A,l,c,R));
   case _RC_:
     __Ze2MatGetSubMatOT(S,A,c,l,R);
     break;
   case _CR_:
     __Ze2MatGetSubMatOT(S,A,l,c,R);
     break;
   case _CC_:
     return(__Ze2MatGetSubMatST(S,A,c,l,R));
   }   
 return(0);
}

/*
   \TEX[
   \title{Assigning a submatrix}
   \input{Two |ZENMat|, two indexes, a |ZENElt| and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The matrix |S| is assigned at position $(|r|,|c|)$ in matrix |M|.}
   \note{Permutation matrix cannot be used.}
   }
   */
int _Ze2MatSetSubMat(A,l,c,S,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int cas;
 
 ZENTest(ZENMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETSUBMAT,ZEN_LIB_ZE2, ZERR);
 ZENTest((l<0)||(c<0)||((l+ZENMatNbRow(S,R))>ZENMatNbRow(A,R))||((c+ZENMatNbCol(S,R))>ZENMatNbCol(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATSETSUBMAT,ZEN_LIB_ZE2, ZERR);
 
 cas=(Ze2MatIsRowType(A,R)<<1)|(Ze2MatIsRowType(S,R));

 switch(cas)
   {
   case _RR_:
     return(__Ze2MatSetSubMatST(A,l,c,S,R));
   case _RC_:
     __Ze2MatSetSubMatOT(A,c,l,S,R);
     break;
   case _CR_:
     __Ze2MatSetSubMatOT(A,l,c,S,R);
     break;
   case _CC_:
     return(__Ze2MatSetSubMatST(A,c,l,S,R));
   }     
  
  return(0);
}

/*
   \TEX[
   \Subsubsection{Matrix conversion}
   The following functions are intended to allow conversion of matrices between an extension and its definition |ZENRing|.
   
   \title{Simple assignment}
   \input{Two couples (|ZENMat|, |ZENRing|).}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |M1| is assigned with |M2|.}
   \note{The coefficients of |M2| must be elements of |R1|. The two matrices must be of same sizes.}
   } 
   */
int _Ze2MatConvert(M1,R1,M2,R2)
     ZENMat M1,M2;
     ZENRing R1,R2;
{
  if(ZENRingArith(R1) == ZEN_LIB_ZE2)
    {
      if(ZENRingArith(R2) == ZEN_LIB_ZE2)
	Ze2MatAssign(M1,M2,R2);
      else
	{
	  int l,c,dc;
	  BigNumDigit Ec;
	  
	  Ze2MatSetToZero(M1,R1);
	  if(Ze2MatIsRowType(M1,R1))
	    {
	      for(l=Ze2MatNbBn(M1,R1);l--;)
		{
		  c=Ze2MatNbDigit(M1,R1)-1;
		  dc=divSizeBloc(c);
		  Ec=BlocExp2(modSizeBloc(c));
		  while(c>=0)
		    {
		      if(ZENEltIsOne(ZENMatGetCoeffPtr(M2,l,c,R2),R2))
			Ze2MatLnChangeCoeffFast(M1,l,dc,Ec,R1);
		      c--;
		      Ec>>=1;
		      if(Ec == 0)
			{
			  dc--;
			  Ec=BlocExp2(SIZE_BLOC-1);
			}
		    }
		}
	    }
	  else
	    {
	      for(c=Ze2MatNbBn(M1,R1);c--;)
		{
		  l=Ze2MatNbDigit(M1,R1)-1;
		  dc=divSizeBloc(l);
		  Ec=BlocExp2(modSizeBloc(l));
		  while(l>=0)
		    {
		      if(ZENEltIsOne(ZENMatGetCoeffPtr(M2,l,c,R2),R2))
			Ze2MatColChangeCoeffFast(M1,dc,Ec,c,R1);
		      l--;
		      Ec>>=1;
		      if(Ec == 0)
			{
			  dc--;
			  Ec=BlocExp2(SIZE_BLOC-1);
			}
		    }
		}
	    }
	}
    }
  else if(ZENRingArith(R2) == ZEN_LIB_ZE2)
    {
      int l,c,dc;
      BigNumDigit Ec;

      ZENEltSetToZero(ZENRingHbuf(R1),R1);
      ZENEltSetToOne(ZENRingIbuf(R1),R1);

      if(Ze2MatIsRowType(M2,R2))
	{
	  for(l=Ze2MatNbBn(M2,R2);l--;)
	    {
	      c=Ze2MatNbDigit(M2,R2)-1;
	      dc=divSizeBloc(c);
	      Ec=BlocExp2(modSizeBloc(c));
	      while(c>=0)
		{
		  if(Ze2MatLnCoeffTest(M2,l,dc,Ec,R2))
		    ZENMatSetCoeff(M1,l,c,ZENRingIbuf(R1),R1);
		  else
		    ZENMatSetCoeff(M1,l,c,ZENRingHbuf(R1),R1);
	      
		  c--;
		  Ec>>=1;
		  if(Ec==0)
		    {
		      dc--;
		      Ec=BlocExp2(SIZE_BLOC-1);
		    }
		}
	    }
	}
      else
	{
	  for(c=Ze2MatNbBn(M2,R2);c--;)
	    {
	      l=Ze2MatNbDigit(M2,R2)-1;
	      dc=divSizeBloc(l);
	      Ec=BlocExp2(modSizeBloc(l));
	      while(l>=0)
		{
		  if(Ze2MatColCoeffTest(M2,dc,Ec,c,R2))
		    ZENMatSetCoeff(M1,l,c,ZENRingIbuf(R1),R1);
		  else
		    ZENMatSetCoeff(M1,l,c,ZENRingHbuf(R1),R1);
	      
		  l--;
		  Ec>>=1;
		  if(Ec == 0)
		    {
		      dc--;
		      Ec=BlocExp2(SIZE_BLOC-1);
		    }
		}
	    }
	}
    }
  else
    return(ZERR);
  return(0);
}


