/*
 * zen/ze2/c/mat_gauss.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Gauss elimination}

   These functions are specific to $\Z/2\Z$.
   }
   */
#include <stdio.h>
#include "ze2.h"
#include "zeg.h"

#ifndef min
#  define min(x,y) ((x)<(y)?(x):(y))
#endif

/*
  \TEX{
   \title{General gaussian elimination}
   \input{Three matrices, a pointer on integer, a pointer on a |ZENElt| and on permutation function to use}
   \output{|ZENERR| if an error occurred,  |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned,
   |ZENMAT_HAS_MAXIMAL_RANK| if the matrix is of maximal rank,
   |ZENMAT_HAS_KERNEL| otherwise.}
   \side{Compute |S|, |P|, |p_rk| or |p_det| only if non |NULL| pointers.}
   \note{
   Permute is the operation to use to permute |D|.
   If |P| and |p_rk| are |NULL|, process stops as soon as the matrix is not of maximal rank.
   If |P| and |S| are |NULL|, trigonalization is performed.
   This function should never appear in a program. Use the encapsulated functions instead.
   }
   }
 */
int _Ze2MatGauss(D,S,P,p_rk,p_det,Permute,R)
     ZENMat D,S,P;
     ZENElt *p_det;
     Dim *p_rk;
     int (*Permute) ___((ZENMat,int,int,ZENRing));
     ZENRing R;
{int k,l,c,dk,flag;
 Dim lim,ret,nbbn;
 BigNumDigit Ek;
 
 nbbn=ZENMatNbBn(D,R)-(((Dim)(ZENMatType(D,R)))>>2); /* Tricky feature:
							not documented !
							I need that... (FC) */
 ZENTest(Ze2MatIsPermutation(D,R),
	 ZEN_ISPERM,ZEN_MATGAUSS,ZEN_LIB_ZE2, ZERR);
 if(p_det) Ze2EltSetToOne(*p_det,R);
 if(S)
   {
     ZENTest(Ze2MatIsPermutation(S,R),
	     ZEN_ISPERM,ZEN_MATGAUSS,ZEN_LIB_ZE2, ZERR);
     ZENTest(!ZENMatAreSameType(S,D,R),
	     ZEN_MAT_TYPE,ZEN_MATGAUSS,ZEN_LIB_ZE2, ZERR);
     ZENTest((nbbn != ZENMatNbBn(S,R))
	     ||(ZENMatNbDigit(S,R) != ZENMatNbBn(S,R)),
	     ZEN_MAT_DIM,ZEN_MATGAUSS,ZEN_LIB_ZE2, ZERR);
   }
 
 if(P)
   {
     ZENTest(ZENMatAreSameType(P,D,R),
	     ZEN_MAT_TYPE,ZEN_MATGAUSS,ZEN_LIB_ZE2, ZERR);
     ZENTest((ZENMatNbDigit(D,R) != ZENMatNbBn(P,R))
	     ||(ZENMatNbDigit(P,R) != ZENMatNbBn(P,R)),
	     ZEN_MAT_DIM,ZEN_MATGAUSS,ZEN_LIB_ZE2, ZERR);
   }
 
 ret=ZENMatNbDigit(D,R);
 lim=min(nbbn,ret);
 flag=ZENMAT_HAS_MAXIMAL_RANK;

 dk=k=0;
 Ek=1;
 for(;;)
   {     
     while (!(Ze2MatLnCoeffTest(D,k,dk,Ek,R)))
       {
	 l=k+1;
	 while(l<nbbn)
	   if(!(Ze2MatLnCoeffTest(D,l,dk,Ek,R)))
	     l++;
	   else
	     {
	       (*Permute)(D,k,l,R);
	       if(S) (*Permute)(S,k,l,R); 
	       goto OK;
	     }
	 if((S)&&(!P))
	   {
	     ZENSetError(ZEN_LIB_ZE2,ZEN_MATGAUSS,ZEN_MAT_NOTINV);
	     return(ZERR);
	   }

	 if (--ret > k)
	   {
	     if(ret<lim)
	       {
		 if(p_det) Ze2EltSetToZero(*p_det,R);
		 if((!p_rk)&&(!P))		/* GaussTrigInvertible */
		   return(ZENMAT_HAS_KERNEL);
		 flag=ZENMAT_HAS_KERNEL;
		 lim=ret;
	       }
	     Ze2MatPermuteDigit(D,k,ret,R);
	     if(P) ZENMatPermuteBn(P,k,ret,R); /* ZEN fct for the case of permutation */
	   }
         else
	   {
	     if(p_det) Ze2EltSetToZero(*p_det,R);
	     flag=ZENMAT_HAS_KERNEL;
	     goto FIN;
	   }
       }			

   OK:
#define nop()
#define ___SubSComputation()                                            \
       for (c=0;c<Ze2MatNbBloc(S,R);c++)	                        \
         Ze2MatBloc(S,l,c,R) ^= Ze2MatBloc(S,k,c,R); 
	 

#define ___Computation(foo)                                               \
     if (Ze2MatLnCoeffTest(D,l,dk,Ek,R))                                  \
       { foo();                                                           \
	 for (c=dk;c<Ze2MatNbBloc(D,R);c++)                               \
	   Ze2MatBloc(D,l,c,R) ^= Ze2MatBloc(D,k,c,R);           \
       }else

  if(S)
    for(l=k+1;l<ZENMatNbBn(D,R);l++)
      ___Computation(___SubSComputation);
  else
    for(l=k+1;l<ZENMatNbBn(D,R);l++)
      ___Computation(nop);
   
     if(++k >= lim)
       goto FIN;
     Ek<<=1;
     if(Ek==0)
       {
	 Ek=1;
	 dk++;
       }
   }
FIN:
 if(p_rk) *p_rk=k;
 lim=k--;
 dk=divSizeBloc(k);
 Ek=BlocExp2(modSizeBloc(k));
	 
 while(k>0)
   {
     if(S)
       {
	 for(l=0;l<k;l++)
	   ___Computation(___SubSComputation);
       }
     else if(P)
       {
	 for(l=0;l<k;l++)
	   ___Computation(nop);
       }

     k--;
     Ek>>=1;;
     if(Ek==0)
       {
	 Ek=BlocExp2(SIZE_BLOC-1);
	 dk--;
       }
   }
 return(flag);
}



