/*
 * zen/ze2/c/mat_inverse.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matrix inversion}

   These functions are specific to $\Z/2\Z$.
   }
   */
#include <stdio.h>
#include "ze2.h"
#include "zeg.h"

extern void _ZedsMatFree ___((ZENMat,ZENRing));

#ifndef min
#  define min(x,y) ((x)<(y)?(x):(y))
#endif

/*
   \TEX{
   \title{Matrix inversion}
   \input{Two |Ze2Mat| over a |Ze2Ring|}
   \output{|ZEN_HAS_INVERSE| or |ZENERR|.}
   \side{$|I|=|M|^{-1}$ if |ZEN_HAS_INVERSE| is returned.}
   }
   */

int _Ze2MatInverse(I,M,R)
     ZENMat I,M;
     ZENRing R;
{
  ZENMat D;
  int ret;
  
  ZENTest(Ze2MatIsPermutation(I,R)
	  ||Ze2MatIsPermutation(M,R),
	  ZEN_ISPERM,ZEN_MATINVERSE,ZEN_LIB_ZE2, ZERR);
  ZENTest((Ze2MatNbBn(I,R) != Ze2MatNbBn(M,R))
	  ||(Ze2MatNbDigit(M,R) != Ze2MatNbBn(M,R))
	  ||(Ze2MatNbDigit(I,R) != Ze2MatNbBn(M,R)),
	  ZEN_MAT_DIM,ZEN_MATINVERSE,ZEN_LIB_ZE2, ZERR);
  
  ZENCheckError(!(D=(Ze2MatIsRowType(I,R)
		     ?Ze2MatRowCreate(Ze2MatNbBn(M,R),Ze2MatNbBn(M,R),R)
		     :Ze2MatColCreate(Ze2MatNbBn(M,R),Ze2MatNbBn(M,R),R))), ZERR);
  
  Ze2MatAssign(D,M,R);
  Ze2MatSetToOne(I,R);
  ret=Ze2MatGaussInverse(D,I,R);
  Ze2MatFree(D,R);
  return(ret);
}

/*
   \TEX{
   \title{Matrix kernel}
   \input{A pointer on an unallocated |Ze2Mat| and a |Ze2Mat| over a |Ze2Ring|}
   \output{|ZEN_HAS_INVERSE| or |ZENERR|.}
   \side{$|I|=|M|^{-1}$ if |ZEN_HAS_INVERSE| is returned.}
   \side{|*p_K| is allocated and set to a basis of the kernel of |M| if |ZEN_HAS_INVERSE| is returned..}
   }
   */

int _Ze2MatKernel(p_K,M,R)
     ZENMat *p_K,M;
     ZENRing R;
{
  int n,rk,i,j,l,di;
  BigNumDigit Ei;
  ZENMat D,P;
  ZENElt un;
  
  Ze2EltSetToOne(un,R);
  n=Ze2MatNbCol(M,R);
  
  if(!(D=Ze2MatRowCreate(Ze2MatNbRow(M,R),n,R)))
    return(ZERR);
  
  ZENCheckError(Ze2MatAssign(D,M,R), ZERR);
  
  if(!(P=ZENPermutationColCreate(n,R)))
    return(ZERR);
  
  ZENCheckError(Ze2MatGaussLazy(D,P,&rk,R), ZERR);
  
#define dim (n<=rk ? 1 : n-rk)
  
  if(!(*p_K=(Ze2MatIsRowType(M,R) ? Ze2MatRowCreate(dim,n,R) : Ze2MatColCreate(dim,n,R))))
    return(ZERR);
  
  i=n-1;
  j=i-rk;
  di=divSizeBloc(i);
  Ei=BlocExp2(modSizeBloc(i));
  while(i>=rk)
    {
      Ze2MatSetCoeff(*p_K,j,ZENPermutationBn(P,i,R),un,R);
      
      for(l=rk-1;l>=0;l--)
	{
	  if(Ze2MatLnCoeffTest(D,l,di,Ei,R))
	   Ze2MatSetCoeff(*p_K,j,ZENPermutationBn(P,l,R),un,R);
       }
     i--;
     j--;
     Ei>>=1;
     if(Ei==0)
       {
	 di--;
	 Ei=BlocExp2(SIZE_BLOC-1);
       }
   }

 Ze2MatFree(D,R);
 ZENPermutationFree(P,R);
 return(ZEN_HAS_INVERSE);
}
