/*
 * zen/ze2/c/mat_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   \Subsection{input/output}
   
   Conversions between the internal representation and the ASCII
   representation of a |Ze2Mat| are done by
   |_Ze2MatPrintToString|. |_Ze2MatPrintToFile| allows to write a
   |Ze2Mat| in a file. One can moreover write to and then read from a
   file a |_Ze2Mat| with |_Ze2MatPutToFile| and |_Ze2MatReadFromFile|.
   }
   */
#include <stdio.h>
#include <ctype.h>
#include "ze2.h"

extern void _ZedsMatFree ___((ZENMat,ZENRing));

/* \TEX[
   \title{Converting from a string in an internal representation.}
   \input{A finite field |Rg| and a string |s| representing a |Ze2Mat|
   in an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|M| is allocated and filled with s if the output is different
   from |ZENERR|}	    
   }
   */
int _Ze2MatGetFromString(M, s, Rg)
     char *s;
     ZENMat *M;
     ZENRing Rg;
{
  ZENMatTypeStruct t;
  char *str;
  int l,nl,nc,nbc;
  
  str=s;
  
  if (sscanf(str,"%2X%8X%8X", &t,&nl,&nc) != 3) { 
    ZENSetError(ZEN_LIB_ZE2, ZEN_MATGETFROMSTRING, ZEN_SYNTAX);
    return(ZERR);
  }
  for(; isspace(*str); str++) ;
  str += 18;
  
  if(!(*M=(t
	   ?Ze2MatRowCreate(nl,nc,Rg)
	   :Ze2MatColCreate(nl,nc,Rg))))
    return(ZERR);

  nbc=str-s;
  if((l=ZBNGetFromStringLazy(Ze2MatFirst(*M,Rg),(BigNumLength)Ze2MatNbBnd(*M,Rg),str))<0) {
    Ze2MatFree(*M,Rg);
    return(ZERR);
  }
  return(nbc+l);
}


/* \TEX[
   \title{Converting to string in an internal representation.}
   \input{An allocated |Ze2Mat| |M| of a finite field |Rg|.}
   \output{An allocated string representing |G| in an internal representation or |ZENNULL| if an error occured.}   
   }
   */
char *_Ze2MatPutToString(M, Rg)
     ZENMat M;
     ZENRing Rg;
{
  char *str, *s1;
  
  s1=ZBNPutToStringLazy(Ze2MatFirst(M,R),(BigNumLength)Ze2MatNbBnd(M,Rg));
  if(s1 == NULL) 
    return(ZNULL);
  
  
  if (!(str = (char *)ZENMalloc(sizeof(char)*(18+strlen(s1)+1)))) {
    ZENSetError(ZEN_LIB_ZE2, ZEN_MATPUTTOSTRING, ZEN_ALLOC);
    return(ZNULL);
  }
  
  if (sprintf(str,"%.2X%.8X%.8X%s",
	      Ze2MatType(M,Rg),
	      Ze2MatNbRow(M,Rg),
	      Ze2MatNbCol(M,Rg),
	      s1)<0) {
    ZENSetError(ZEN_LIB_ZE2, ZEN_MATPUTTOSTRING, ZEN_SPRINTF);
    ZENFree(s1);
    ZENFree(str); 
    return(ZNULL);
  }
  ZENFree(s1);
  return(str);
}

/* \TEX[
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |Ze2Mat| |M| of a finite field
   |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |M| to |file|.}    
   }
   */
int _Ze2MatPutToFile(file, M, Rg)
     FILE *file;
     ZENMat M;
     ZENRing Rg;  
{
  if (fprintf(file,"%.8x%.8x%.8x",
	      Ze2MatType(M,Rg),
	      Ze2MatNbRow(M,Rg),
	      Ze2MatNbCol(M,Rg))<0) {
    ZENSetError(ZEN_LIB_ZE2, ZEN_MATPUTTOFILE, ZEN_FPRINTF);
    return(ZERR);
  }

  if(ZBNPutToFileLazy(file,Ze2MatFirst(M,R),(BigNumLength)Ze2MatNbBnd(M,Rg))) 
    return(ZERR);
  
  return(0);
}
  

/* \TEX[
   \title{Getting from file in an internal representation.}
   \input{A stream |file|, and a pointer on matnomial |M| of a finite
   field |Ze2Mat|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|*M| is filled with the matnomial read in |file|}    
   }
   */
int _Ze2MatGetFromFile(M, file, Rg)
     ZENMat *M;
     FILE *file;
     ZENRing Rg;
{
  ZENMatTypeStruct t;
  int nl,nc;
   
  if (fscanf(file, "%8x%8x%8x", &t, &nl,&nc) != 3) {
    ZENSetError(ZEN_LIB_ZE2, ZEN_MATGETFROMFILE, ZEN_SSCANF);
    return(ZERR);
  }

  if(!(*M=(t
	   ?Ze2MatRowCreate(nl,nc,Rg)
	   :Ze2MatColCreate(nl,nc,Rg))))
    return(ZERR);

  if (ZBNGetFromFileLazy(Ze2MatFirst(*M,Rg),(BigNumLength)Ze2MatNbBnd(*M,Rg),file))
    return(ZERR);
  
  return(0);
}

