/*
 * zen/ze2/c/mat_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial multiplications}

   These functions are specific to $\Z/2\Z$.
   }
   */
#include <stdio.h>
#include "ze2.h"

/*
   \TEX[
   \title{Multiplication of scalar}
   \input{A |ZENMat|, a |ZENElt| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{All the elements of |A| are multiplied by |b|.}
   \note{|A| cannot be a permutation.}
   }
   */
int _Ze2MatMultiplyScalar(A,b,R)
     ZENMat A;
     ZENElt b;
     ZENRing R;
{
  ZENTest(Ze2MatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATMULTIPLY,ZEN_LIB_ZE2, ZERR);
  Ze2MatMultiplyScalar(A,b,R);
  return(0);
}

/*
   \TEX{
   \title{Multiplication.}
   \input{Three |Ze2Mat| |X|, |A| and |B| of $\Z/2\Z$.}
   \output{|ZENERR| if an error occurred, 0 otherwise}
   \side{$|X|= |A| \times |B|$.}
   \note{|X|, |A| and |B| can be of type |ZENMatTypeRow| (|R|) or |ZENMatTypeCol| (|C|). But the efficiency of the computation changes with the types.
   \begin{tabular}{ccc@{ --- }p{.75\textwidth}}
   |X| & |A| & |B| & Algorithm \\ \hline
   |C| & |C| & |C| & |MultiplyRRR| is used. Efficient algorithm. \\
   |C| & |C| & |R| & |MultiplyRCR| is used. Efficient algorithm. \\
   |C| & |R| & |C| & |MultiplyRRC| is used. Good algorithm. \\
   |C| & |R| & |R| & |MultiplyRCC| is used. Low-efficient algorithm. \\
   |R| & |C| & |C| & |MultiplyRCC| is used. Low-efficient algorithm. \\
   |R| & |C| & |R| & |MultiplyRCR| is used. Efficient algorithm. \\
   |R| & |R| & |C| & |MultiplyRRC| is used. Good algorithm. \\
   |R| & |R| & |R| & |MultiplyRRR| is used. Efficient algorithm. 
   \end{tabular}
   }
   }
   */
   
int _Ze2MatMultiplyPlain(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{char cas;
 
 ZENTest(Ze2MatIsPermutation(A,R)||Ze2MatIsPermutation(B,R)||Ze2MatIsPermutation(X,R),
	 ZEN_ISPERM,ZEN_MATMULTIPLY,ZEN_LIB_ZE2, ZERR);
 ZENTest((Ze2MatNbRow(A,R) != Ze2MatNbRow(X,R))
	 ||(Ze2MatNbCol(A,R) != Ze2MatNbRow(B,R))
	 ||(Ze2MatNbCol(B,R) != Ze2MatNbCol(X,R)),
	 ZEN_MAT_DIM,ZEN_MATMULTIPLY,ZEN_LIB_ZE2, ZERR);
 cas=(((int)Ze2MatType(X,R))<<2)|(((int)Ze2MatType(A,R))<<1)|(((int)Ze2MatType(B,R)));

 switch(cas)
   {
   case _CCC_ :
     Ze2MatMultiplyCCC(X,A,B,R);
     break;
   case _CCR_ :
     Ze2MatMultiplyCCR(X,A,B,R);
     break;
   case _CRC_ :
     Ze2MatMultiplyCRC(X,A,B,R);
     break;
   case _CRR_ :
     Ze2MatMultiplyCRR(X,A,B,R);
     break;
   case _RCC_ :
     Ze2MatMultiplyRCC(X,A,B,R);
     break;
   case _RCR_ :
     Ze2MatMultiplyRCR(X,A,B,R);
     break;
   case _RRC_ :
     Ze2MatMultiplyRRC(X,A,B,R);
     break;
   case _RRR_ :
     Ze2MatMultiplyRRR(X,A,B,R);
     break;
   }
 return(0);
}

/*
   \TEX{
   \title{Multiplication.}
   \input{Three |Ze2Mat| |X|, |A| and |B| of $\Z/2\Z$ of types |R|, |R| and |R| or |C|, |C| and |C|.}
   \side{$|X|= |A| \times |B|$.}
   \note{The algorithm uses x-oring on |BigNumDigit|.}
   }
   */

void _Ze2MatMultiplyRRR(X,A,B,R) /* CCC */
     ZENMat X,A,B;
     ZENRing R;
{int i,k,l,dl;
 BigNumDigit El;
 Ze2MatSetToZero(X,R);
 
 l=Ze2MatNbBn(B,R)-1;
 dl=divSizeBloc(l);
 El=BlocExp2(modSizeBloc(l));
 while(l>=0)
   {
     for(i=Ze2MatNbBn(X,R)-1;i>=0;i--)
       if(Ze2MatLnCoeffTest(A,i,dl,El,R))
	 for(k=Ze2MatNbBloc(X,R)-1;k>=0;k--)
	   Ze2MatBloc(X,i,k,R)^=Ze2MatBloc(B,l,k,R);
     l--;
     El >>= 1;
     if(El==0)
       {
	 dl--;
	 El=BlocExp2(SIZE_BLOC-1);
       } 
   }
}

/*
   \TEX{
   \title{Multiplication.}
   \input{Three |Ze2Mat| |X|, |A| and |B| of $\Z/2\Z$ of types |R|, |C| and |R| or |C|, |C| and |R|.}
   \side{$|X|= |A| \times |B|$.}
   \note{The algorithm uses x-oring on |BigNumDigit|.}
   }
   */

void _Ze2MatMultiplyRCR(X,A,B,R) /* CCR */
     ZENMat X,A,B;
     ZENRing R;
{int i,k,l,dl;
 BigNumDigit El;
 
 Ze2MatSetToZero(X,R);
 
 l=Ze2MatNbBn(X,R)-1;
 dl=divSizeBloc(l);
 El=BlocExp2(modSizeBloc(l));
 while(l>=0)
   {
     for(i=Ze2MatNbBn(A,R)-1;i>=0;i--)
       if(Ze2MatLnCoeffTest(A,i,dl,El,R))
	 for(k=Ze2MatNbBloc(X,R)-1;k>=0;k--)
	   Ze2MatBloc(X,l,k,R)^=Ze2MatBloc(B,i,k,R);
     l--;
     El>>=1;
     if(El==0)
       {
	 dl--;
	 El=BlocExp2(SIZE_BLOC-1);
       } 
   }
}

#if 1
/*
   \TEX{
   \title{Multiplication.}
   \input{Three |Ze2Mat| |X|, |A| and |B| of $\Z/2\Z$ of types |R|, |R| and |C| or |C|, |R| and |C|.}
   \side{$|X|= |A| \times |B|$.}
   \note{The algorithm uses x-oring on |BigNumDigit| and weight computation |ZBNWeightDigit|.}
   }
   */
void _Ze2MatMultiplyRRC(X,A,B,R) /* CRC */
     ZENMat X,A,B;
     ZENRing R;
{int i,k,l,dl,pd;
 BigNumDigit El;
 
 Ze2MatSetToZero(X,R);
 
 l=Ze2MatNbBn(B,R)-1;
 dl=divSizeBloc(l);
 El=BlocExp2(modSizeBloc(l));
 while(l>=0)
   {
     for(i=Ze2MatNbBn(A,R)-1;i>=0;i--)
       {
	 pd=0;
	 for(k=0;k<Ze2MatNbBloc(A,R);k++)
	   {
	     BigNumDigit d;
	     
	     d=Ze2MatBloc(A,i,k,R)&Ze2MatBloc(B,l,k,R);
	     pd+=ZBNWeightDigitFast(d);
	   }
	 if(pd&1)
	   Ze2MatLnChangeCoeffFast(X,i,dl,El,R);
       }
     l--;
     El >>= 1;
     if(El==0)
       {
	 dl--;
	 El=BlocExp2(SIZE_BLOC-1);
       } 
   }
}
#else
static int ZBNWeightDigitMod2(pd)
     BigNumDigit pd;
{
  unsigned char *ptr;
  int i;
  BigNumDigit tmp;
  
  ptr=(unsigned char *)(&pd);
  tmp= *(ptr++);
  for(i=1;i<sizeof(BigNumDigit);i++)
    tmp ^= *(ptr++);

  tmp = (tmp & 15)^(tmp >> 4);
  tmp = (tmp & 3)^(tmp >> 2);
  tmp = (tmp & 1)^(tmp >> 1);
  return(tmp);    
}

void _Ze2MatMultiplyRRC(X,A,B,R) /* CRC */
     ZENMat X,A,B;
     ZENRing R;
{int i,k,l,dl;
 BigNumDigit pd,El;
 
 Ze2MatSetToZero(X,R);
 
 l=Ze2MatNbBn(B,R)-1;
 dl=divSizeBloc(l);
 El=BlocExp2(modSizeBloc(l));
 while(l>=0)
   {
     for(i=Ze2MatNbBn(A,R)-1;i>=0;i--)
       {	 
	 pd=0;
	 for(k=0;k<Ze2MatNbBloc(A,R);k++)
	   pd ^= Ze2MatBloc(A,i,k,R)&Ze2MatBloc(B,l,k,R);
	 if(ZBNWeightDigitMod2(pd))
	   Ze2MatLnChangeCoeffFast(X,i,dl,El,R);
       }
     l--;
     El >>= 1;
     if(El==0)
       {
	 dl--;
	 El=BlocExp2(SIZE_BLOC-1);
       } 
   }
}
#endif

/*
   \TEX{
   \title{Multiplication.}
   \input{Three |Ze2Mat| |X|, |A| and |B| of $\Z/2\Z$ of types |R|, |C| and |C| or |C|, |R| and |R|.}
   \side{$|X|= |A| \times |B|$.}
   \note{The algorithm uses one-by-one coefficient changing.}
   }
   */
void _Ze2MatMultiplyRCC(X,A,B,R) /* CRR */
     ZENMat X,A,B;
     ZENRing R;
{int c,k,l,dl,dk;
 BigNumDigit El,Ek;
 
 Ze2MatSetToZero(X,R);
 
 l=Ze2MatNbBn(X,R)-1;
 dl=divSizeBloc(l);
 El=BlocExp2(modSizeBloc(l));
 while(l>=0)
   {
     k=Ze2MatNbBn(A,R)-1;
     dk=divSizeBloc(k);
     Ek=BlocExp2(modSizeBloc(k));
     while(k>=0)
       {
	 for(c=Ze2MatNbBn(B,R)-1;c>=0;c--)
	   if((Ze2MatColCoeffTest(A,dl,El,k,R))
	      &&(Ze2MatColCoeffTest(B,dk,Ek,c,R)))
	     Ze2MatLnChangeCoeff(X,l,c,R);
	 k--;
	 Ek >>= 1;
	 if(Ek==0)
	   {
	     dk--;
	     Ek=BlocExp2(SIZE_BLOC-1);
	   } 
       }
     l--;
     El >>= 1;
     if(El==0)
       {
	 dl--;
	 El=BlocExp2(SIZE_BLOC-1);
       } 
   }
}
