/*
 * zen/ze2/c/mat_permute.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial permutations}

   These functions are specific to $\Z/2\Z$.
   }
   */
#include <stdio.h>
#include "ze2.h"

/*
   \TEX{
   \title{Permute two |Bn|s}
   \input{A |Ze2Mat| |A|, two indexes and a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The specified |Bn|s of |A| are exchanged}
   \note{Except for advanced user, |PermuteRow| and |PermuteCol| functions should be preferred}
   }
   
   */
   
int _Ze2MatPermuteBn(A,l1,l2,R)
     ZENMat A;
     Dim l1,l2;
     ZENRing R;
{
 BigNum b;

 
 ZENTest(Ze2MatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATPERMUTEBNS,ZEN_LIB_ZE2, ZERR);
 ZENTest((l1<0)||(l2<0)||(l1>=Ze2MatNbBn(A,R))||(l2>=Ze2MatNbBn(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATPERMUTEBNS,ZEN_LIB_ZE2, ZERR);
 b=ZBNC(Ze2MatNbBloc(A,R));
 ZENTest(!(b),
	 ZEN_ALLOC,ZEN_MATPERMUTEBNS,ZEN_LIB_ZE2, ZERR);

 ZBNAssign(b,Ze2MatBn(A,l1,R),Ze2MatNbBloc(A,R));
 ZBNAssign(Ze2MatBn(A,l1,R),Ze2MatBn(A,l2,R),Ze2MatNbBloc(A,R));
 ZBNAssign(Ze2MatBn(A,l2,R),b,Ze2MatNbBloc(A,R));

 ZBNF(b);

 return(0);
}

/*
   \TEX{
   \title{Permute two |digit|s}
   \input{A |Ze2Mat| |A|, two indexes and a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The specified |digits| of |A| are exchanged}
   \note{Except for advanced user, |PermuteRow| and |PermuteCol| functions should be preferred}
   }   
   */
   

int _Ze2MatPermuteDigit(A,l1,l2,R)
     ZENMat A;
     Dim l1,l2;
     ZENRing R;
{int l,ds1,ds2;
 BigNumDigit E1,E2;

 ZENTest(Ze2MatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATPERMUTEDG,ZEN_LIB_ZE2, ZERR);
 ZENTest((l1<0)||(l2<0)||(l1>=Ze2MatNbDigit(A,R))||(l2>=Ze2MatNbDigit(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATPERMUTEDG,ZEN_LIB_ZE2, ZERR);

 ds1=divSizeBloc(l1);
 E1=BlocExp2(modSizeBloc(l1));
 ds2=divSizeBloc(l2);
 E2=BlocExp2(modSizeBloc(l2));

 for(l=Ze2MatNbBn(A,R);l--;)
   if(opXor(Ze2MatLnCoeffTest(A,l,ds1,E1,R),Ze2MatLnCoeffTest(A,l,ds2,E2,R)))
     {
       Ze2MatLnChangeCoeffFast(A,l,ds1,E1,R);
       Ze2MatLnChangeCoeffFast(A,l,ds2,E2,R);
     }
 return(0);
}

/*
   \TEX{
   \title{Multiply permutation.}
   \input{A |ZENMat| |X|, a permutation |A|, a  matrix |B| of types RRR and a finite ring |R|.}
   \side{The |Bn|s of |B| are permuted according to |A|. Result is put in |X|.}
   \note{|X| and |B| must be different.}
  }
   */
void _Ze2MatMultiplyLeftPermutationRRR(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{
  int l;

 for(l=Ze2MatNbBn(X,R)-1;l>=0;l--)
   ZBNAssign(Ze2MatBn(X,l,R),Ze2MatBn(B,ZENPermutationBn(A,l,R),R),Ze2MatNbBloc(X,R));
}

/*
   \TEX{
   \title{Multiply permutation.}
   \input{A |ZENMat| |X|, a permutation |A|, a  matrix |B| of types RCR and a finite ring |R|.}
   \side{The |Bn|s of |B| are permuted according to |A|. Result is put in |X|.}
   \note{|X| and |B| must be different.}
  }
   */
void _Ze2MatMultiplyLeftPermutationRCR(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{
  int l;

  for(l=Ze2MatNbBn(X,R)-1;l>=0;l--)
    ZBNAssign(Ze2MatBn(X,ZENPermutationBn(A,l,R),R),Ze2MatBn(B,l,R),Ze2MatNbBloc(X,R));   
}

/*
   \TEX{
   \title{Multiply permutation.}
   \input{The starting point |c|, A |ZENMat| |X|, a permutation |A|, a  matrix |B| of types XRC and a finite ring |R|.}
   \side{The |Digit|s of |B| are permuted according to |A|. Result is put in |X|.}
   \note{If |X| is of type |C|, |c| must be set to |Ze2MatNbDigit(X,R)|, otherwise to |Ze2MatNbBn(X,R)|.}
  }
   */
void __Ze2MatMultiplyLeftPermutationXRC(c,X,A,B,R)
     int c;
     ZENMat X,A,B;
     ZENRing R;
{
  int l,dc,dcp;
  Dim cp;
  BigNumDigit Ec,Ecp;

  Ze2MatSetToZero(X,R);

  c--;
  dc=divSizeBloc(c);
  Ec=BlocExp2(modSizeBloc(c));
  
  while(c>=0)
    {
      cp=ZENPermutationBn(A,c,R);
      dcp=divSizeBloc(cp);
      Ecp=BlocExp2(modSizeBloc(cp));
      
      for(l=Ze2MatNbBn(X,R)-1;l>=0;l--)
	if(Ze2MatLnCoeffTest(B,l,dcp,Ecp,R))
	  Ze2MatLnChangeCoeffFast(X,l,dc,Ec,R);
      
      c--;
      Ec>>=1;
      if(!Ec)
	{
	  dc--;
	  Ec=BlocExp2(SIZE_BLOC-1);
	}
    }
}

/*
   \TEX{
   \title{Multiply permutation.}
   \input{The starting point |c|, A |ZENMat| |X|, a permutation |A|, a  matrix |B| of types XCC and a finite ring |R|.}
   \side{The |Digit|s of |B| are permuted according to |A|. Result is put in |X|.}
   \note{If |X| is of type |C|, |c| must be set to |Ze2MatNbDigit(X,R)|, otherwise to |Ze2MatNbBn(X,R)|.}
  }
   */
void __Ze2MatMultiplyLeftPermutationXCC(c,X,A,B,R)
     int c;
     ZENMat X,A,B;
     ZENRing R;
{
  int l,dc,dcp;
  Dim cp;
  BigNumDigit Ec,Ecp;

  Ze2MatSetToZero(X,R);
  
  c--;
  dc=divSizeBloc(c);
  Ec=BlocExp2(modSizeBloc(c));
 
  while(c>=0)
    {
      cp=ZENPermutationBn(A,c,R);
      dcp=divSizeBloc(cp);
      Ecp=BlocExp2(modSizeBloc(cp));
      
      for(l=Ze2MatNbBn(X,R)-1;l>=0;l--)
	if(Ze2MatLnCoeffTest(B,l,dc,Ec,R))
	  Ze2MatLnChangeCoeffFast(X,l,dcp,Ecp,R);

      c--;
      Ec>>=1;
      if(Ec==0)
	{
	  dc--;
	  Ec=BlocExp2(SIZE_BLOC-1);
	}
    }
}

/*
   \TEX{
   \title{Multiply left permutation.}
   \input{Two |Ze2Mat| |X| and |B|, a |ZENMatTypeRowPermutation| |A| and a finite ring |R|.}
   \output{|ZENERR| if an error occurred, 0 otherwise otherwise}
   \side{$|X|= |A| \times |B|$.}
   }
   */
int _Ze2MatMultiplyLeftPermutation(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{
  char cas;
 
  ZENTest(!Ze2MatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATMULLEFTPERM,ZEN_LIB_ZE2, ZERR);
  ZENTest((Ze2MatNbRow(A,R) != Ze2MatNbRow(X,R))
	  ||(Ze2MatNbCol(A,R) != Ze2MatNbRow(B,R))
	  ||(Ze2MatNbCol(B,R) != Ze2MatNbCol(X,R)),
	  ZEN_MAT_DIM,ZEN_MATMULLEFTPERM,ZEN_LIB_ZE2, ZERR);
  
  cas=(Ze2MatIsRowType(X,R)<<2)|(Ze2MatIsRowType(A,R)<<1)|(Ze2MatIsRowType(B,R));
  
  switch(cas)
    {
    case _CCC_:
      Ze2MatMultiplyLeftPermutationCCC(X,A,B,R);
      return(0);
    case _CCR_:
      Ze2MatMultiplyLeftPermutationCCR(X,A,B,R);
      return(0);
    case _CRC_:
      Ze2MatMultiplyLeftPermutationCRC(X,A,B,R);
      return(0);
    case _CRR_:
      Ze2MatMultiplyLeftPermutationCRR(X,A,B,R);
      return(0);
    case _RCC_:
      Ze2MatMultiplyLeftPermutationRCC(X,A,B,R);
      return(0);
    case _RCR_:
      Ze2MatMultiplyLeftPermutationRCR(X,A,B,R);
      return(0);
    case _RRC_:
      Ze2MatMultiplyLeftPermutationRRC(X,A,B,R);
      return(0);
    case _RRR_:
      Ze2MatMultiplyLeftPermutationRRR(X,A,B,R);
      return(0);
    }
  return(ZERR);
}

/*
   \TEX{
   \title{Multiply right permutation.}
   \input{Two |Ze2Mat| |X| and |A|, a |ZENMatTypeColPermutation| |B| and a finite ring |R|.}
   \output{|ZENERR| if an error occurred, 0 otherwise otherwise}
   \side{$|X|= |A| \times |B|$.}
   }
   */
int _Ze2MatMultiplyRightPermutation(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{
  char cas;
 
  ZENTest(!Ze2MatIsPermutation(B,R),
	  ZEN_ISPERM,ZEN_MATMULRIGHTPERM,ZEN_LIB_ZE2, ZERR);
  ZENTest((Ze2MatNbRow(A,R) != Ze2MatNbRow(X,R))
	  ||(Ze2MatNbCol(A,R) != Ze2MatNbRow(B,R))
	  ||(Ze2MatNbCol(B,R) != Ze2MatNbCol(X,R)),
	  ZEN_MAT_DIM,ZEN_MATMULRIGHTPERM,ZEN_LIB_ZE2, ZERR);
  
  cas=(Ze2MatIsRowType(X,R)<<2)|(Ze2MatIsRowType(A,R)<<1)|(Ze2MatIsRowType(B,R));
  
  switch(cas)
    {
    case _CCC_:
      Ze2MatMultiplyRightPermutationCCC(X,A,B,R);
      return(0);
    case _CCR_:
      Ze2MatMultiplyRightPermutationCCR(X,A,B,R);
      return(0);
    case _CRC_:
      Ze2MatMultiplyRightPermutationCRC(X,A,B,R);
      return(0);
    case _CRR_:
      Ze2MatMultiplyRightPermutationCRR(X,A,B,R);
      return(0);
    case _RCC_:
      Ze2MatMultiplyRightPermutationRCC(X,A,B,R);
      return(0);
    case _RCR_:
      Ze2MatMultiplyRightPermutationRCR(X,A,B,R);
      return(0);
    case _RRC_:
      Ze2MatMultiplyRightPermutationRRC(X,A,B,R);
      return(0);
    case _RRR_:
      Ze2MatMultiplyRightPermutationRRR(X,A,B,R);
      return(0);
    }
  return(ZERR);
}
