/*
 * zen/ze2/c/poly_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial assignment}
   
   One can set a |ZENPoly| to zero with |_Ze2PolySetToZero|. Then we can
   assign a |ZENPoly| with an other with |_Ze2PolyAssign|. Furthermore, we
   can set randomly a |ZENPoly| with |_Ze2PolySetRandom|. 
   
   The functions |_Ze2PolyGetCoeff| and |_Ze2PolySetCoeff|
   enable us to get or set the coefficient of $X^i$,
   |_Ze2PolyUpdateDegree| compute the real degree of a |ZENPoly| if it
   decreased.
   }
   */
#include <stdio.h>
#include "ze2.h"

/* \TEX[
   \title{Assign.}
   \input{Two |ZENPoly| |PX| and |RX| of a finite ring |Rg|.}
   \side{|RX| is filled with |PX|.}
   }
   */
void _Ze2PolyAssign(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int ds;

  ds = 1+divSizeBloc(Ze2PolyDeg(PX, Rg));
  Ze2PolyDeg(RX, Rg) = Ze2PolyDeg(PX, Rg);
  ZBNAssign(Ze2PolyDigit(RX, Rg), Ze2PolyDigit(PX, Rg), ds);
}

/* \TEX[
   \title{Set to zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{|RX| is set to zero.}
   }
   */
void _Ze2PolySetToZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  Ze2PolyDeg(RX, Rg) = 0;
  Ze2PolyGetDigit(RX, 0, Rg) = 0;
}

/* \TEX[
   \title{Set to $X^i$.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg| allocated at least for
   degree $i$.}
   \side{|RX| is set to $|X|^i$.}
   }
   */
void _Ze2PolySetToXi(RX, i, Rg)
     ZENPoly RX;
     int i;
     ZENRing Rg;
{
  int ds, ms;

  Ze2PolyDeg(RX, Rg) = i;

  ds = divSizeBloc(i);  
  ZBNSetToZero(Ze2PolyDigit(RX, Rg), ds+1);

  ms = modSizeBloc(i);
  Ze2PolyGetDigit(RX, ds, Rg) |= ((BigNumDigit)1 << ms);

}

/* \TEX[
   \title{Set to random.}
   \input{A |ZENPoly| |RX| of degree |d| of a finite ring |Rg|.}
   \side{|RX| is set randomly with degree |d|.}
   }
   */
void _Ze2PolySetRandom(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  BigNumLength ds;
  int ms;

  ms = 1+modSizeBloc(deg);
  ds = divSizeBloc(deg);
  ZBNSetRandom(Ze2PolyDigit(RX, Rg), ds);
  Ze2PolyGetDigit(RX, ds, Rg) =  zbnrandom();
  Ze2PolyGetDigit(RX, ds, Rg) >>= (SIZE_BLOC-ms);
  Ze2PolyGetDigit(RX, ds, Rg) |= ((BigNumDigit)1 << (ms-1));
  Ze2PolyDeg(RX, Rg) = deg;
}

/* \TEX[
   \title{Set a coeff.}
   \input{A |ZENPoly| |RX| and a |ZENElt| |b| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \side{The coeff of $X^d$ in |RX| is set to |b|.}
   }
   */
void _Ze2PolySetCoeff(RX, d, b, Rg)
     ZENPoly RX;
     int d;
     ZENElt b;
     ZENRing Rg;
{
  int ds, ms;

  ds = divSizeBloc(d);
  ms = modSizeBloc(d);
  if (Ze2EltIsZero(b, Rg)) {
    Ze2PolyGetDigit(RX, ds, Rg) &= ~((BigNumDigit)1 << ms);
  }
  else {
    Ze2PolyGetDigit(RX, ds, Rg) |= ((BigNumDigit)1 << ms);
  }
}

/* \TEX[
   \title{Get a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The coeff of $X^d$ in |RX|.}
   }
   */
void _Ze2PolyGetCoeff(p_c, RX, d, Rg)
     ZENElt *p_c;
     ZENPoly RX;
     int d;
     ZENRing Rg;
{
  int ds, ms;
  
  ds = divSizeBloc(d);
  ms = modSizeBloc(d);
  if (Ze2PolyGetDigit(RX, ds, Rg) & ((BigNumDigit)1 << ms)) {
    Ze2EltSetToOne(*p_c, Rg);
  }
  else {
    Ze2EltSetToZero(*p_c, Rg);
  }
}

/* \TEX[
   \title{Get a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{A pointer to the coeff of $X^d$ in |RX|.}
   }
   */
ZENElt _Ze2PolyGetCoeffPtr(RX, d, Rg)
     ZENPoly RX;
     int d;
     ZENRing Rg;
{
  ZENElt e;
  
  Ze2Elt2Int(e, Rg) =
    (Ze2PolyGetDigit(RX, divSizeBloc(d), Rg) >>  modSizeBloc(d)) & 1;
  return(e);
}

/* \TEX[
   \title{Update the degree of a polynomial.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{Resetting the degree of |P|, assuming it has decreased.}
   }
   */
void _Ze2PolyUpdateDegree(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  int i, j;
  BigNumDigit a, mask;
  
  mask = (~(BigNumDigit)0);
  mask >>= SIZE_BLOC-1-modSizeBloc(Ze2PolyDeg(RX, Rg));
  i = divSizeBloc(Ze2PolyDeg(RX, Rg));

  if ((a = Ze2PolyGetDigit(RX, i, Rg) & mask)) {
    for(a >>= 1, j = 0; a; a >>= 1, j++) {
      ;
    }
  }
  else {
    for(i--; i > 0; i--) {
      if (Ze2PolyGetDigit(RX, i, Rg)) {
	break;
      }
    }
    if (i >= 0) {
      for(a = Ze2PolyGetDigit(RX, i, Rg)>>1, j = 0; 
	  a; 
	  a >>= 1, j++) {
	;
      }
    }
    else {
      i = j = 0;
    }
  }
  Ze2PolyDeg(RX, Rg) = i*SIZE_BLOC+j;
}
