/*
 * zen/zec/c/elt_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Element assignment}
   
   One can set a |ZENElt| to zero or one with |_ZecEltSetToZero| or
   |_ZecEltSetToOne|. Then we can assign a |ZENElt| with an other with
   |_ZecEltAssign|. Furthermore, we can set randomly a |ZENElt| with
   |_ZecEltSetRandom| or enumerate the elements of a finite ring with
   |_ZecEltSetNext|.
   }
   */
#include <stdio.h>
#include "zec.h"

/* \TEX[
   \title{Assign.}
   \input{A pointer |p_a| on a |ZENElt| and a |ZENElt| |b| of a finite ring |Rg|.}
   \side{|*p_a| is filled with |b|.}
   }
   */
void _ZecEltAssign(p_a, b, Rg)
     ZENElt *p_a, b;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltAssign(ZecElt2Elt(*p_a,i,Rg),ZecElt2Elt(b,i,Rg),ZecRingSubring(Rg,i));
}

/* \TEX[
   \title{Assign.}
   \input{A pointer |p_a| on a |ZENElt| of a clone |Rg| and a |ZENElt| |b| of the original ring of |Rg|.}
   \side{|*p_a| is filled with |b|.}
   \note{This function performs modular reduction for integer domains}
   }
   */
void _ZecEltToClone(p_a, b, Rg)
     ZENElt *p_a, b;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltConvert(ZecElt2Elt(*p_a,i,Rg),ZecRingSubring(Rg,i),
		  b,ZENRingOrigin(Rg));
}

/* \TEX[
   \title{Assign.}
   \input{A pointer |p_a| on a |ZENElt| of the original ring of |Rg| and a |ZENElt| |b| of |Rg|.}
   \side{|*p_a| is filled with |b|.}
   }
   */
void _ZecCloneToElt(p_a, b, Rg)
     ZENElt *p_a, b;
     ZENRing Rg;
{
  int i;
  ZENEltSetToZero(*p_a,ZENRingOrigin(Rg));
  
  for(i=ZecRingNbS(Rg);i--;)
    {
      ZENEltConvert(ZENRingJbuf(ZENRingOrigin(Rg)),ZENRingOrigin(Rg),
		    ZecElt2Elt(b,i,Rg),ZecRingSubring(Rg,i));		     
      ZENEltMultiply(ZENRingJbuf(ZENRingOrigin(Rg)),
		     ZecRingModinv(Rg,i),
		     ZENRingJbuf(ZENRingOrigin(Rg)),
		     ZENRingOrigin(Rg));
      ZENEltAdd(*p_a,ZENRingJbuf(ZENRingOrigin(Rg)),ZENRingOrigin(Rg));
    }
}

/* \TEX[
   \title{Set to zero.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set to zero..}
   }
   */
void _ZecEltSetToZero(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltSetToZero(ZecElt2Elt(*p_a,i,Rg),ZecRingSubring(Rg,i));
}

/* \TEX[
   \title{Set to one.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set to one.}
   }
   */
void _ZecEltSetToOne(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltSetToOne(ZecElt2Elt(*p_a,i,Rg),ZecRingSubring(Rg,i));
}

/* \TEX[
   \title{Set to random.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set randomly.}
   }
   */
void _ZecEltSetRandom(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltSetRandom(ZecElt2Elt(*p_a,i,Rg),ZecRingSubring(Rg,i));
}

/* \TEX[
   \title{Enumerate.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set randomly.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \note{After a number of call of this function equal to the number of elements in the finite ring, we obtain the same element.}
   }
   */
void _ZecEltSetNext(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  int i;

  for(i=ZecRingNbS(Rg);i--;)
    ZENEltSetNext(ZecElt2Elt(*p_a,i,Rg),ZecRingSubring(Rg,i));
}








