/*
 * zen/zec/c/init.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zec.h"
#include "zeg.h"
#include "zed.h"

/* \TEX{
   \title{Initialization of tables for chinese remainder theorem.}
   \input{A |ZENRing| |Rg| defined as product of modular |ZENRing|s.}
   \output{|ZENERR| if an error occured, $|POLY_GEN| \vert |SR_GEN| \vert |MAT_GEN|$ otherwise.}
   \side{Precomputation of iomorphsm of the clone |Rg| is performed.}
   \note{This function is called by |ZENChineseRingCreate| and should not be called otherwise.}
   }
   */
int _ZecInit(Rg)
     ZENRing Rg;
{
  int i;

  ZENRingArith(Rg) = ZEN_LIB_ZEC;

  /**** Computation of the reciprocal isomorphism ****/
  for(i=ZecRingNbS(Rg);i--;)
    {
      if(ZENEltConvert(ZENRingHbuf(ZecRingSubring(Rg,i)),ZecRingSubring(Rg,i),
		       ZecRingModinv(Rg,i),ZENRingOrigin(Rg)))
	return(ZERR);
      if(ZENEltInverse(ZENRingHbuf(ZecRingSubring(Rg,i)),
		       ZENRingHbuf(ZecRingSubring(Rg,i)),
		       ZecRingSubring(Rg,i)))
	return(ZERR);
      if(ZENEltConvert(ZENRingHbuf(ZENRingOrigin(Rg)),ZENRingOrigin(Rg),
		       ZENRingHbuf(ZecRingSubring(Rg,i)),ZecRingSubring(Rg,i)))
	return(ZERR);
      ZENEltMultiply(ZecRingModinv(Rg,i),
		     ZENRingHbuf(ZENRingOrigin(Rg)),
		     ZecRingModinv(Rg,i),
		     ZENRingOrigin(Rg));
    }
  /************ new pointers of functions are used ***************/

  /* Procedure on rings
   ****************************/
  ZENPrcInitEltMultiplyPtr(Rg) = NULL;
  ZENPrcCloseEltMultiplyPtr(Rg) = NULL;

  ZENPrcInitTracePtr(Rg) = NULL;
  ZENPrcCloseTracePtr(Rg) = NULL;

  ZENPrcInitPolyRootsCanonicalPtr(Rg) = NULL;
  ZENPrcClosePolyRootsCanonicalPtr(Rg) = NULL;

  ZENPrcInitEltExpPtr(Rg) = _ZecPrcInitEltExp;
  ZENPrcCloseEltExpPtr(Rg) = _ZecPrcCloseEltExp;

  /* Procedure on elements
   ****************************/
  /* Allocation */
  ZENEltCreatePtr(Rg) = _ZecEltCreate;
  ZENEltCreateLargePtr(Rg) = _ZecEltCreate;
  ZENEltFreePtr(Rg) = _ZecEltFree;
  
  /* Assigning */
  ZENEltAssignPtr(Rg) = _ZecEltAssign;
  ZENEltSetToZeroPtr(Rg) = _ZecEltSetToZero;
  ZENEltSetToOnePtr(Rg) = _ZecEltSetToOne;
  ZENEltSetRandomPtr(Rg) = _ZecEltSetRandom;
  ZENEltSetNextPtr(Rg) = _ZecEltSetNext;

  ZENEltToClonePtr(Rg) = _ZecEltToClone;
  ZENCloneToEltPtr(Rg) = _ZecCloneToElt;
  ZENEltConvertPtr(Rg) = _ZENEltCloneConvert;
  
  /* Test */
  ZENEltAreEqualPtr(Rg) = _ZecEltAreEqual;
  ZENEltIsZeroPtr(Rg) = _ZecEltIsZero;
  ZENEltIsOnePtr(Rg) = _ZecEltIsOne;
  ZENEltIsASquarePtr(Rg) = _ZENEltIsASquareCharp;

  /* Arithmetic */
  ZENEltAddPtr(Rg) = _ZecEltAdd;
  ZENEltNegatePtr(Rg) = _ZecEltNegate;
  ZENEltSubtractPtr(Rg) = _ZecEltSubtract;
  ZENEltSquarePtr(Rg) = _ZecEltSquare;
  ZENEltMultiplyPtr(Rg) = _ZecEltMultiply;
  ZENEltInversePtr(Rg) = _ZecEltInverse;
  ZENEltTracePtr(Rg) = _ZecEltTrace;
  ZENEltSquareRootPtr(Rg) = _ZENEltSquareRootCharp;

  /* io */
  ZENZToEltPtr(Rg) = _ZecZToElt;
  ZENEltToZPtr(Rg) = _ZecEltToZ;

  ZENEltReadFromStringPtr(Rg) = _ZENCloneEltReadFromString;
  ZENEltPrintToStringPtr(Rg) = _ZENCloneEltPrintToString;
  ZENEltReadFromFilePtr(Rg) = _ZENCloneEltReadFromFile;
  ZENEltPrintToFilePtr(Rg) = _ZENCloneEltPrintToFile;
  
  ZENEltGetFromStringPtr(Rg) = _ZENCloneEltGetFromString;
  ZENEltPutToStringPtr(Rg) = _ZENCloneEltPutToString;
  ZENEltGetFromFilePtr(Rg) = _ZENCloneEltGetFromFile;
  ZENEltPutToFilePtr(Rg) = _ZENCloneEltPutToFile;

  /* Procedures on matrices
   **************************/
  /* Allocation */
  ZENMatCreateParamsPtr(Rg) = _ZedMatCreateParams; /* Create a matrix */
  ZENMatCreatePtr(Rg) = __ZedMatCreate;	/* Create a matrix */

  return(POLY_GEN | MAT_GEN | SR_GEN);
}


/* \TEX[
   \title{Desallocating tables for small extensions.}
   \input{A |ZENRing| of cardinality less than $2^8$.}
   \output{|ZENERR| if an error occured, |ZEN_HAS_INVERSE| otherwise.}
   \note{All tables allocated with |_ZecInit| are
   freed.}
   }
   */
int _ZecClose(Rg)
     ZENRing Rg;
{
  if(ZENRingClose(ZENRingOrigin(Rg)))
    return(ZERR);
  return(ZEN_HAS_INVERSE);
}

