/*
 * zen/zed/c/mat_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial assignments}

   These functions are default ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions.

   
   }
   */
#include <stdio.h>
#include "zed.h"


/*
   \TEX{
   \title{Get a bloc}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{|E| is set to the indexed bloc of |A|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred. This function must be chosen together with |_ZedMatCreateParams|.}
   }
   
   */
   
int _ZedMatGetBloc(p_E,A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENElt *p_E;
     ZENRing R;
{
  ZENTest(ZENMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATGETBLOC,ZEN_LIB_ZED, ZERR);
  ZENTest((l<0)||(c<0)||(l>=ZENMatNbBn(A,R))||(c>=ZENMatNbDigit(A,R)),
	  ZEN_MAT_DIM0,ZEN_MATGETBLOC,ZEN_LIB_ZED, ZERR);
  
  ZENEltAssign(*p_E,ZedMatBloc(A,l,c,R),R);
  return(0);
}

/*
   \TEX{
   \title{Get a bloc pointer}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{|E| is set to the indexed bloc of |A|}
   \note{This function must be chosen together with |_ZedMatCreateParams|.}
   }
   
   */
   
ZENElt _ZedMatGetBlocPtr(A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENRing R;
{
  return(ZedMatBloc(A,l,c,R));
}

/*
   \TEX{
   \title{Set a bloc}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The indexed bloc of |A| is set to |E|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred.
   This function must be chosen together with |_ZedMatCreateParams|.}
   }
   
   */
   
int _ZedMatSetBloc(A,l,c,E,R)
     ZENMat A;
     Dim l,c;
     ZENElt E;
     ZENRing R;
{
  ZENTest(ZENMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATSETBLOC,ZEN_LIB_ZED, ZERR);
  ZENTest((l<0)||(c<0)||(l>=ZENMatNbBn(A,R))||(c>=ZENMatNbDigit(A,R)),
	  ZEN_MAT_DIM0,ZEN_MATSETBLOC,ZEN_LIB_ZED, ZERR);

  ZENEltAssign(ZedMatBloc(A,l,c,R),E,R);
  return(0);
}

/************************************************/

static void __ZedMatGetSubMatST(S,A,l,c,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int i,j,k;
 
 for(i=0;i<ZENMatNbBn(S,R);i++,l++)
   for(j=0,k=c;j<ZENMatNbDigit(S,R);j++,k++)    
     ZENEltAssign(ZedMatBloc(S,i,j,R),ZedMatBloc(A,l,k,R),R);
}

static void __ZedMatGetSubMatOT(S,A,l,c,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int i,j,k;
 
 for(i=0;i<ZENMatNbBn(S,R);i++,l++)
   for(j=0,k=c;j<ZENMatNbDigit(S,R);j++,k++)    
     ZENEltAssign(ZedMatBloc(S,i,j,R),ZedMatBloc(A,k,l,R),R);
}

static void __ZedMatSetSubMatST(A,l,c,S,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int i,j,k;
 
 for(i=0;i<ZENMatNbBn(S,R);i++,l++)
   for(j=0,k=c;j<ZENMatNbDigit(S,R);j++,k++)    
     ZENEltAssign(ZedMatBloc(A,l,k,R),ZedMatBloc(S,i,j,R),R);
}

static void __ZedMatSetSubMatOT(A,l,c,S,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int i,j,k;
 
 for(i=0;i<ZENMatNbBn(S,R);i++,l++)
   for(j=0,k=c;j<ZENMatNbDigit(S,R);j++,k++)    
     ZENEltAssign(ZedMatBloc(A,k,l,R),ZedMatBloc(S,i,j,R),R);
}

/*
   \TEX[
   \title{Getting a submatrix}
   \input{Two allocated |ZENMat|, two indexes and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The submatrix of |M| at position $(|r|,|c|)$ of size those of |S| is extracted and assigned to |S|.}
   \note{Permutation matrix cannot be used.}
   }
   */

int _ZedMatGetSubMat(S,A,l,c,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int cas;
 
 ZENTest((ZENMatIsPermutation(A,R))||(ZENMatIsPermutation(S,R)),
	 ZEN_ISPERM,ZEN_MATGETSUBMAT,ZEN_LIB_ZED, ZERR);
 ZENTest((l<0)||(c<0)||((l+ZENMatNbRow(S,R))>ZENMatNbRow(A,R))||((c+ZENMatNbCol(S,R))>ZENMatNbCol(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATGETSUBMAT,ZEN_LIB_ZED, ZERR);

 cas=(ZENMatIsRowType(A,R)<<1)|(ZENMatIsRowType(S,R));

 switch(cas)
   {
   case _RR_:
     __ZedMatGetSubMatST(S,A,l,c,R);
     break;
   case _RC_:
     __ZedMatGetSubMatOT(S,A,c,l,R);
     break;
   case _CR_:
     __ZedMatGetSubMatOT(S,A,l,c,R);
     break;
   case _CC_:
     __ZedMatGetSubMatST(S,A,c,l,R);
     break;
   }   
 return(0);
}

/*
   \TEX[
   \title{Assigning a submatrix}
   \input{Two |ZENMat|, two indexes, a |ZENElt| and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The matrix |S| is assigned at position $(|r|,|c|)$ in matrix |M|.}
   \note{Permutation matrix cannot be used.}
   }
   */
int _ZedMatSetSubMat(A,l,c,S,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int cas;
 
 ZENTest(ZENMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETSUBMAT,ZEN_LIB_ZED, ZERR);
 ZENTest((l<0)||(c<0)||((l+ZENMatNbRow(S,R))>ZENMatNbRow(A,R))||((c+ZENMatNbCol(S,R))>ZENMatNbCol(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATSETSUBMAT,ZEN_LIB_ZED, ZERR);
 
 cas=(ZENMatIsRowType(A,R)<<1)|(ZENMatIsRowType(S,R));

 switch(cas)
   {
   case _RR_:
     __ZedMatSetSubMatST(A,l,c,S,R);
     break;
   case _RC_:
     __ZedMatSetSubMatOT(A,c,l,S,R);
     break;
   case _CR_:
     __ZedMatSetSubMatOT(A,l,c,S,R);
     break;
   case _CC_:
     __ZedMatSetSubMatST(A,c,l,S,R);
     break;
   }     
  
  return(0);
}

