/*
 * zen/zed/c/mat_create.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Allocating a matrix}
   }
   */


#include <stdio.h>
#include "zed.h"
#include "zeg.h"

/* \TEX{
   \title{Computation of the matrix parameters}
   \input{A |ZENMatTypeStruct|, the dimensions of the desired matrix, a pointer on a |ZENMat| and a |ZENRing|.}
   \output{0 if no error occured, |ZENERR| otherwise.}
   \side{The |ZENMat| structure is allocated and the parameters computed according to the type of the desired matrix. Array is not allocated. Allocation of the coefficients will be performed separately.}
   \note{This function should not be directly used (see |ZENMatCreate|). If chosen in a |ZENRing| initialization  routine, |_ZedGetBloc|, |_ZedGetBlocPtr| and |_ZedSetBloc| must also be chosen.}
   }
   */   

int _ZedMatCreateParams(ts,l,c,p_pms,R)
     Dim l,c;
     ZENMatTypeStruct ts;
     ZENMat *p_pms;
     ZENRing R;
{
  (*p_pms)=(ZENMat)ZENMalloc(sizeof(struct zenmat));

  ZENTest(*p_pms==NULL,
	  ZEN_ALLOC,ZEN_MATCREPMS,ZEN_LIB_ZED, ZERR);
  ZENTest((l<=0)||(c<=0),
	  ZEN_MAT_DIM0,ZEN_MATCREPMS,ZEN_LIB_ZED, ZERR);

  ZENMatFreePtr(*p_pms)=_ZedMatFree;
  ZENMatAssignPtr(*p_pms)=__ZENMatAssign;
  ZENMatSetToOnePtr(*p_pms)=__ZENMatSetToOne;
  ZENMatIsOnePtr(*p_pms)=__ZENMatIsOne;
  ZENMatSetRandomPtr(*p_pms)=__ZENMatSetRandom;
  ZENMatPermuteBnPtr(*p_pms)=__ZENMatPermuteBn;
  switch(ts)
    {
    case ZENMatTypeRow:
      ZENMatType(*p_pms,R)=ZENMatTypeRow;
      ZENMatNbBn(*p_pms,R)=l;
      ZENMatNbDigit(*p_pms,R)=c;
      ZENMatNbBloc(*p_pms,R)=c;

      ZENMatGetCoeffNewPtr(*p_pms)=__ZENMatGetBloc;
      ZENMatGetCoeffPtrPtr(*p_pms)=__ZENMatGetBlocPtr;
      ZENMatSetCoeffPtr(*p_pms)=__ZENMatSetBloc;     
      ZENMatPermuteRowPtr(*p_pms)=__ZENMatPermuteBn;     
      ZENMatPermuteColPtr(*p_pms)=__ZENMatPermuteDigit;     
      break;
      
    case ZENMatTypeCol:
      ZENMatType(*p_pms,R)=ZENMatTypeCol;
      ZENMatNbBn(*p_pms,R)=c;
      ZENMatNbDigit(*p_pms,R)=l;
      ZENMatNbBloc(*p_pms,R)=l;

      ZENMatGetCoeffNewPtr(*p_pms)=__ZENMatGetDigit;
      ZENMatGetCoeffPtrPtr(*p_pms)=__ZENMatGetDigitPtr;
      ZENMatSetCoeffPtr(*p_pms)=__ZENMatSetDigit;     
      ZENMatPermuteRowPtr(*p_pms)=__ZENMatPermuteDigit;     
      ZENMatPermuteColPtr(*p_pms)=__ZENMatPermuteBn;     
      break;

    default:
      ZENTest(1,
	      ZEN_UNKNOWN,ZEN_MATCREPMS,ZEN_LIB_ZED, ZERR);
    }

  ZENMatNbBnd(*p_pms,R)=0;
  ZENMatSize(*p_pms,R)=0;
  ZENMatNbPtr(*p_pms,R)=l*c;
  return(0);
}
/*
   \TEX{
   \title{Allocating a matrix}
   \input{A parametrized but unallocated |ZENMat| and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The matrix coefficients are allocated.}
   \note{|ZENMatCreate| calls this function after call to |ZENMatCreateParams|.Should not be used by itself.}
   } 
   */
int __ZedMatCreate(M,R)
     ZENMat M;
     ZENRing R;
{
  ZENElt *ptmat;
  int l,c;
 
  ZENMatMat(M,R)=(ZENElt **)ZENMalloc(ZENMatNbBn(M,R)*sizeof(BigNum *));
  ZENTest(ZENMatMat(M,R)==NULL,
	  ZEN_ALLOC,ZEN_MAT_CREATE,ZEN_LIB_ZEG, ZERR);

  ptmat=(ZENElt *)ZENMalloc((ZENMatNbPtr(M,R)+ZENMatNbBnd(M,R))*sizeof(ZENElt));
  ZENMatFirst(M,R)=(BigNumDigit *)ptmat+ZENMatNbPtr(M,R);
  ZENTest(ZENMatFirst(M,R)==NULL,
	  ZEN_ALLOC,ZEN_MAT_CREATE,ZEN_LIB_ZEG, ZERR);
  
  for(l=0;l<ZENMatNbBn(M,R);l++,ptmat += ZENMatNbBloc(M,R))
    ZENMatBn(M,l,R)=ptmat;

  if(ZENMatNbBnd(M,R))
    for(l=ZENMatNbBn(M,R)-1,ptmat=(ZENElt *)ZENMatFirst(M,R);
	l>=0;
	l--)
      for(c=ZENMatNbBloc(M,R)-1;
	  c>=0;
	  c--, ptmat += ZENMatSize(M,R))
	ZENElt2Bn(ZENMatBloc(M,l,c,R),R)=(BigNumDigit *)ptmat;
  else    
    for(l=ZENMatNbBn(M,R)-1;
	l>=0;
	l--)
      for(c=ZENMatNbBloc(M,R)-1;
	  c>=0;
	  c--)
	{
	  ZENCheckError(ZENEltCreate(ZENMatBloc(M,l,c,R),R), ZERR);
	  ZENEltSetToZero(ZENMatBloc(M,l,c,R),R);
	}	   	
  
  return(0);  
}
/*
   \TEX{
   \title{Freeing a matrix}
   \input{A |ZENMat| over a |ZENRing|.}
   \side{The matrix is completely freed.}
   } 
   */
void _ZedMatFree(M,R)
     ZENMat M;
     ZENRing R;
{
  int l,c;
  
  if(ZENMatNbBnd(M,R)==0)    
    for(l=ZENMatNbBn(M,R)-1;
	l>=0;
	l--)
      for(c=ZENMatNbBloc(M,R)-1;
	  c>=0;
	  c--)
	ZENEltFree(ZENMatBloc(M,l,c,R),R);
  
  ZENMatFirst(M,R)-=ZENMatNbPtr(M,R);
  ZENFree(ZENMatFirst(M,R));
  ZENFree(ZENMatMat(M,R));
  ZENFree(M);
}

/* \TEX{
   \title{Computation of the matrix parameters}
   \input{A |ZENMatTypeStruct|, the dimensions of the desired matrix, a pointer on a |ZENMat| structure and a |ZedRing|.}
   \output{0 if no error occured, |ZENERR| otherwise.}
   \side{The |ZENMat| structure is allocated and the parameters computed according to the type of the desired matrix. Array is not allocated. Allocation of the coefficients is unnecessary (e.g. |ZEPS|, |ZETAB|, |ZELOG|).}
   \note{This function should not be directly used (see |ZENMatCreate|). If chosen in a |ZENRing| initialization  routine, |_ZedGetBloc|, |_ZedGetBlocPtr| and |_ZedSetBloc| must also be chosen.}
   }
   */   

int _ZedsMatCreateParams(ts,l,c,p_pms,R)
     Dim l,c;
     ZENMatTypeStruct ts;
     ZENMat *p_pms;
     ZENRing R;
{
  (*p_pms)=(ZENMat)ZENMalloc(sizeof(struct zenmat));

  ZENTest(*p_pms==NULL,
	  ZEN_ALLOC,ZEN_MATCREPMS,ZEN_LIB_ZED, ZERR);
  ZENTest((l<=0)||(c<=0),
	  ZEN_MAT_DIM0,ZEN_MATCREPMS,ZEN_LIB_ZED, ZERR);
  
  ZENMatFreePtr(*p_pms)=_ZedsMatFree;
  ZENMatAssignPtr(*p_pms)=__ZENMatAssign;
  ZENMatSetToOnePtr(*p_pms)=__ZENMatSetToOne;
  ZENMatIsOnePtr(*p_pms)=__ZENMatIsOne;
  ZENMatSetRandomPtr(*p_pms)=__ZENMatSetRandom;
  ZENMatPermuteBnPtr(*p_pms)=__ZENMatPermuteBn;     
  switch(ts)
    {
    case ZENMatTypeRow:
      ZENMatType(*p_pms,R)=ZENMatTypeRow;
      ZENMatNbBn(*p_pms,R)=l;
      ZENMatNbDigit(*p_pms,R)=c;
      ZENMatNbBloc(*p_pms,R)=c;

      ZENMatGetCoeffNewPtr(*p_pms)=__ZENMatGetBloc;
      ZENMatGetCoeffPtrPtr(*p_pms)=__ZENMatGetBlocPtr;
      ZENMatSetCoeffPtr(*p_pms)=__ZENMatSetBloc;     
      ZENMatPermuteRowPtr(*p_pms)=__ZENMatPermuteBn;     
      ZENMatPermuteColPtr(*p_pms)=__ZENMatPermuteDigit;     
      break;
      
    case ZENMatTypeCol:
      ZENMatType(*p_pms,R)=ZENMatTypeCol;
      ZENMatNbBn(*p_pms,R)=c;
      ZENMatNbDigit(*p_pms,R)=l;
      ZENMatNbBloc(*p_pms,R)=l;

      ZENMatGetCoeffNewPtr(*p_pms)=__ZENMatGetDigit;
      ZENMatGetCoeffPtrPtr(*p_pms)=__ZENMatGetDigitPtr;
      ZENMatSetCoeffPtr(*p_pms)=__ZENMatSetDigit;     
      ZENMatPermuteRowPtr(*p_pms)=__ZENMatPermuteDigit;     
      ZENMatPermuteColPtr(*p_pms)=__ZENMatPermuteBn;     
      break;

    default:
      ZENTest(1,
	      ZEN_UNKNOWN,ZEN_MATCREPMS,ZEN_LIB_ZED, ZERR);
    }

  ZENMatNbBnd(*p_pms,R)=l*c;
  ZENMatSize(*p_pms,R)=0;
  ZENMatNbPtr(*p_pms,R)=0;
  return(0);
}
/*
   \TEX{
   \title{Allocating a matrix}
   \input{A parametrized but unallocated |ZENMat| and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The matrix coefficients are allocated.}
   \note{|ZENMatCreate| calls this function after call to |ZENMatCreateParams|. It should not be used by itself.}
   } 
   */
int __ZedsMatCreate(M,R)
     ZENMat M;
     ZENRing R;
{
  BigNumDigit *ptmat;
  int l;
  
  ZENMatMat(M,R)=(ZENElt **)ZENMalloc(ZENMatNbBn(M,R)*sizeof(BigNum *));
  ZENTest(ZENMatMat(M,R)==NULL,
	  ZEN_ALLOC,ZEN_MAT_CREATE,ZEN_LIB_ZED, ZERR);
  ptmat=ZENMatFirst(M,R)=(BigNumDigit *)ZENMalloc(ZENMatNbBnd(M,R)*sizeof(BigNumDigit));
  ZENTest(ZENMatFirst(M,R)==NULL,
	  ZEN_ALLOC,ZEN_MAT_CREATE,ZEN_LIB_ZED, ZERR);

  for(l=0;l<ZENMatNbBn(M,R);l++,ptmat += ZENMatNbBloc(M,R))
    Ze2MatBn(M,l,R)=ptmat;	/* Here, MatBns point directly on coeffs,
				   so we take a pointer on BigNumDigits */
  
  
  ZENCheckError(ZENMatSetToZero(M,R), ZERR);

  return(0);
}
     
/*
   \TEX[
   \title{Freeing a matrix}
   \input{A |ZENMat| over a |ZENRing|.}
   \side{The matrix is completely freed.}
   } 
   */
void _ZedsMatFree(M,R)
     ZENMat M;
     ZENRing R;
{
  ZENFree(ZENMatFirst(M,R));
  ZENFree(ZENMatMat(M,R));
  ZENFree(M);
}








