/*
 * zen/zed/c/poly_alloc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial creation}
   
   |_ZedPolyCreate| allocates one |ZENPoly| and |_ZedPolyFree| frees then it.
   }
   */
#include <stdio.h>
#include "zed.h"

/* \TEX{
   \title{Creation.}
   \input{A finite ring |Rg| and a degree |deg|.}
   \output{An allocated |ZENPoly| set to 0 with |deg| UNALLOCATED coefficients, or |ZENNULL| if an error occured. }
   \note{As |Zelog|, |Zetab|, |Zeps| use elements stored on one computer, word the further allocation of the coefficients is innecessary.}
   }
   */
ZENPoly _ZedsPolyCreate(deg, Rg)
     int deg;
     ZENRing Rg;
{
  ZENPoly PX;
  
  if (!(PX = ZedPolyAllocStruct(Rg))) {
    ZENSetError(ZEN_LIB_ZED, ZEN_POLYCREATE, ZEN_ALLOC);
    return(ZNULL);
  }
  if (!(ZedPolyCof(PX, Rg) = (ZENElt *)ZENMalloc(sizeof(ZENElt)*(deg+1)))) {
    ZENFree(PX);
    ZENSetError(ZEN_LIB_ZED, ZEN_POLYCREATE, ZEN_ALLOC);
    return(ZNULL);
  }
  ZENPolyDeg(PX, Rg) =0;
  ZedPolyLgt(PX, Rg) = deg;
  return(PX);
}

/* \TEX[
   \title{Free.}
   \input{a |ZENPoly| |PX| of a finite ring |Rg| already allocated with
   |_ZedsPolyCreate|.}
   \side{|PX| is freed.}
   }
   */
void _ZedsPolyFree(PX,Rg)
     ZENPoly PX;
     ZENRing Rg;
{
  ZENFree(ZedPolyCof(PX, Rg));
  ZENFree(PX);
}


/* \TEX{
   \title{Creation.}
   \input{A finite ring |Rg| and a degree |deg|.}
   \output{An allocated |ZENPoly|, set to 0, with |deg| coefficients allocated, or |ZENNULL| if an error occured. }
   }
   */
ZENPoly _ZedPolyCreate(deg, Rg)
     int deg;
     ZENRing Rg;
{
  ZENPoly PX;
  int i;

  
  if (!(PX = _ZedsPolyCreate(deg,Rg))) {
    return(ZNULL);
  }

  for(i = 0; i <= deg; i++) {
    if (ZENEltCreate(ZedPolyGetCof(PX, i, Rg), Rg)) {
      for(i--; i>= 0; i--) {
	ZENEltFree(ZedPolyGetCof(PX, i, Rg), Rg);
      }
      _ZedsPolyFree(PX,Rg);
      return(ZNULL);
    }
  }

  return(PX);
}

/* \TEX[
   \title{Free.}
   \input{a |ZENPoly| |PX| of a finite ring |Rg| already allocated with
   |_ZedPolyCreate|.}
   \side{|PX| is freed.}
   }
   */
void _ZedPolyFree(PX,Rg)
     ZENPoly PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZENPolyLgt(PX, Rg); i++) {
    ZENEltFree(ZedPolyGetCof(PX, i, Rg), Rg);
  }
  _ZedsPolyFree(PX,Rg);
}
/* \TEX{
   \Subsection{Polynomial assignment}   
   }
   */

/* \TEX[
   \title{Setting a coeff.}
   \input{A |ZENPoly| |RX| and a |ZENElt| |b| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \side{The coeff of $X^d$ in |RX| is set to |b|.}
   }
   */
void _ZedPolySetCoeff(RX, d, b, Rg)
     ZENPoly RX;
     int d;
     ZENElt b;
     ZENRing Rg;
{
  ZENEltAssign(ZedPolyGetCof(RX, d, Rg), b, Rg);
}


/* \TEX[
   \title{Extracting a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The pointer to the coeff of $X^d$ in |RX|.}
   }
   */
ZENElt _ZedPolyGetCoeffPtr(RX, d, Rg)
     ZENPoly RX;
     int d;
     ZENRing Rg;
{
  return(ZedPolyGetCof(RX, d, Rg));
}
