/*
 * zen/zed/c/prc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   Precomputations are handled by a |ZENPrc| structure. Users can
   easely "add" to or "remove" from a |ZENRing| these
   precomputations with |ZENRingAddPrc| and |ZextRingRmPrc|.
   }
   */
#include <stdio.h>
#include <zen.h>


/* \TEX[
   \title{Adding precomputations.}
   \input{A finite ring |Rg| and precomputations |Prc|.} 
   \output{|ZEN_HAS_INVERSE|.}
   \side{Precomputations asked in |Prc| are done in order to speed up
   the corresponding procedures. |ZENRingFact(Rg)| is filled with a
   factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   \note{Precomputations can take time.}
   } */
int _ZENRingAddPrc(Rg, Prc)		
     ZENRing Rg;
     ZENPrc Prc;
{
#define ZENPrcAddNew(flag, ptr, function)			\
if(1) {								\
  ret = ZEN_HAS_INVERSE;					\
  if (ZENPrcIsSet(Prc, flag) &&					\
      (!ZENPrcIsSet(ZENRingPrcp(Rg), flag)) &&			\
      !(ptr(Rg) &&						\
       ((ret = function(Rg)) == ZEN_HAS_INVERSE ? 0 : 1)	\
       )							\
      ) {							\
    ZENPrcSet(ZENRingPrcp(Rg), flag);				\
  }								\
  if ((ret != ZEN_HAS_INVERSE)) {				\
    return(ret);						\
  }								\
} else

  int ret;

  ret = ZEN_HAS_INVERSE;

  if (ZENPrcIsSet(Prc, ZENPRC_FINITE_FIELD))
    ZENPrcSet(ZENRingPrcp(Rg), ZENPRC_FINITE_FIELD);

  /* Multiplication
  ***********************************/
  ZENPrcAddNew(ZENPRC_ELT_MULTIPLY,
	       ZENPrcInitEltMultiplyPtr,
	       ZENPrcInitEltMultiply);

  /* Exponentiation
   ***********************************/
  if(ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_FINITE_FIELD)) {
    ZENPrcAddNew(ZENPRC_ELT_EXP,
		 ZENPrcInitEltExpPtr,
		 ZENPrcInitEltExp);
  }

  /* An element of trace 1.
   ***********************************/
  if(ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_FINITE_FIELD)) {
    ZENPrcAddNew(ZENPRC_POLY_ROOTS_CANONICAL,
		 ZENPrcInitPolyRootsCanonicalPtr,
		 ZENPrcInitPolyRootsCanonical);
  }

  /* Precomputations for EltTrace.
   ************************************/
  if(ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_FINITE_FIELD)) {
    ZENPrcAddNew(ZENPRC_TRACE,
		 ZENPrcInitTracePtr,
		 ZENPrcInitTrace);
  }

  return(ZEN_HAS_INVERSE);
}

/* \TEX[
   \title{Removing precomputations.}
   \input{A finite ring |Rg| and precomputations |Prc|.}
   \output{|ZENERR| if an error occured, |ZEN_NO_INVERSE| if a factor of a
   modulo was discovered, |ZEN_HAS_INVERSE| otherwise.}
   \side{Structures already allocated by |_ZextRingAddPrc()| are
   disallocated. |ZENRingFact(Rg)| is filled with a factor of a modulo
   if |ZEN_NO_INVERSE| is returned.}
   } */
int _ZENRingRmPrc(Rg, Prc)
     ZENRing Rg;
     ZENPrc Prc;
{
#define ZENPrcRmOld(flag, ptr, function)				\
if(1) {								\
  if (ZENPrcIsSet(Prc, flag) &&					\
      ZENPrcIsSet(ZENRingPrcp(Rg), flag) &&			\
      !(ptr(Rg) &&							\
       ((ret = function(Rg)) == ZEN_HAS_INVERSE ? 0 : 1)	\
       )							\
      ) {							\
    ZENPrcUnset(ZENRingPrcp(Rg), flag);				\
  }								\
  if ((ret != ZEN_HAS_INVERSE)) {				\
    return(ret);						\
  }								\
} else

  int ret;

  ret = ZEN_HAS_INVERSE;

  /* Precomputations for EltMultiply.
  ************************************/
  ZENPrcRmOld(ZENPRC_ELT_MULTIPLY,
	      ZENPrcCloseEltMultiplyPtr,
	      ZENPrcCloseEltMultiply);

  /* Precomputations for EltTrace.
  ************************************/
  ZENPrcRmOld(ZENPRC_TRACE,
	       ZENPrcCloseTracePtr,
	       ZENPrcCloseTrace);

  /* An element of trace 1.
   ***********************************/
  ZENPrcRmOld(ZENPRC_POLY_ROOTS_CANONICAL,
	       ZENPrcClosePolyRootsCanonicalPtr,
	       ZENPrcClosePolyRootsCanonical);
    
  /* Exponentiation
   ***********************************/
  ZENPrcRmOld(ZENPRC_ELT_EXP,
	       ZENPrcCloseEltExpPtr,
	       ZENPrcCloseEltExp);

  return(ZEN_HAS_INVERSE);
}

