/*
 * zen/zed/c/sr_alloc.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Series creation}
   
   |_ZedSrCreate| allocates one |ZENSr| and |_ZedSrFree| frees then it.
   }
   */
#include <stdio.h>
#include "zed.h"


/* \TEX[
   \title{Creation.}
   \input{A finite ring |Rg| and a length |length|.}
   \output{An allocated series set to 0 with at most |length|
   coefficients.}
   }
   */
ZENSr _ZedSrCreate(length, Rg)
     int length;
     ZENRing Rg;
{
  ZENSr PX;
  int i;

  
  if (!(PX = ZedSrAllocStruct(Rg))) {
    ZENSetError(ZEN_LIB_ZED, ZEN_SRCREATE, ZEN_ALLOC);
    return(ZNULL);
  }
  if (!length) {
    ZENFree(PX);
    ZENSetError(ZEN_LIB_ZED, ZEN_SRCREATE, ZEN_ALLOC);
    return(ZNULL);
  }
  if (!(ZedSrCof(PX, Rg) = (ZENElt *)ZENMalloc(sizeof(ZENElt)*(length)))) {
    ZENFree(PX);
    ZENSetError(ZEN_LIB_ZED, ZEN_SRCREATE, ZEN_ALLOC);
    return(ZNULL);
  }
  for(i = 0; i < length; i++) {
    if (ZENEltCreate(ZedSrGetCof(PX, i, Rg), Rg)) {
      for(i--; i>= 0; i--) {
	ZENEltFree(ZedSrGetCof(PX, i, Rg), Rg);
      }
      ZENFree(ZedSrCof(PX, Rg));
      ZENFree(PX);
      return(ZNULL);
    }
  }
  ZENSrLgt(PX, Rg) = length;
  return(PX);
}


/* \TEX[
   \title{Freeing.}
   \input{An allocated series |PX| of a finite ring |Rg|.}
   \side{|PX| is disallocated.}
   }
   */
void _ZedSrFree(PX,Rg)
     ZENSr PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i < ZENSrLgt(PX, Rg); i++) {
    ZENEltFree(ZedSrGetCof(PX, i, Rg), Rg);
  }
  ZENFree(ZedSrCof(PX, Rg));
  ZENFree(PX);
}

/* \TEX{
   \Subsection{Series assignment}   
   }
   */

/* \TEX[
   \title{Setting a coeff.}
   \input{An allocated series |RX| and an element |b| of a
   finite ring |Rg|, the degree |d| of the coeff to set.}
   \side{The coeff of $X^d$ in |RX| is set to |b|.}
   \note{The degree or valuation of |RX| is not recomputed.}
   }
   */
void _ZedSrSetCoeff(RX, d, b, Rg)
     ZENSr RX;
     int d;
     ZENElt b;
     ZENRing Rg;
{
  ZENEltAssign(ZedSrGetCof(RX, d-ZENSrVal(RX, Rg), Rg), b, Rg);
}


/* \TEX[
   \title{Extracting a coeff.}
   \input{An allocated series |RX| of a finite ring |Rg|, the
   degree |d| of the coeff to set.}
   \output{The pointer to the coeff of $X^d$ in |RX|.}
   \side{You must not disallocate the output.}
   }
   */
ZENElt _ZedSrGetCoeffPtr(RX, d, Rg)
     ZENSr RX;
     int d;
     ZENRing Rg;
{
  return(ZedSrGetCof(RX, d-ZENSrVal(RX, Rg), Rg));
}
