/*
 * zen/zef/c/binomial.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include "zef.h"

/***************************************************************/
static int __Zs(a,n,ZEF,Rg)
     ZENElt a;
     int n,ZEF;
     ZENRing Rg;
{
  if(ZefEltMlm(a,ZEF,Rg)==0)
    {
      ZefEltM(a,ZEF,Rg)=ZBNC(1);
      if(ZefEltM(a,ZEF,Rg)==NULL)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	  return(ZERR);
	}
      ZefEltMlm(a,ZEF,Rg)=1;
    }
  ZefEltM(a,ZEF,Rg)[0]=n;
  ZefEltMl(a,ZEF,Rg)=1;
  return(0);
}
 
/*
   \TEX{
   \Subsection{Useful features}
   We provide two additional functions that may be useful in some applications.

   \title{Setting a fraction}
   \input{A pointer on a |ZENElt|, two integers and a fraction |ZENRing|.}
   \side{The element is set to $\frac{|n|}{|d|}$}
   } 
   */
void _ZefEltGetFromZs(pa, n, d, Rg)    
     int n,d;
     ZENElt *pa;
     ZENRing Rg;
{
  ZefEltE(*pa,Rg)=0;
  
  __Zs(*pa,n,ZEF_INTEGER,Rg);
  __Zs(*pa,d,ZEF_INVERSE,Rg);
  ZefEltFlag(*pa,Rg) = ZEF_TOSIMPLIFY | ZEF_FRAC;
  ZefEltAdjustType(*pa,Rg);
}

/*
   \TEX{
   \title{Factorial}
   \input{An allocated fraction, an integer and the |ZENRing| representing $\Q$.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|G| is filled with $n!$.}
   \note{The results are stored inside the function. Hence, if the computation of $n!$ is performed, further calls to this function with inferior values will take no extra time. A call to this function with $|n|= -1$ frees the already computed factorials.}
   } 
   */
int ZefEltFactorial(G,n,R)
     ZENElt G;
     int n;
     ZENRing R;
{
  static ZENElt *F=NULL;
  static int nmax;
  int i;

  if(n == -1)
    {
      for(i=0;i<nmax;i++)
	ZefEltFree(F[i],R);
      ZENFree(F);
      F=NULL;
      return(0);
    }
  if(n<0)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEF_FACTORIAL,ZEN_DIVBYZERO);
      return(ZERR);
    }
  if(n<=nmax)
    {
      ZefEltAssign(G,F[n],R);
      return(0);
    }
  
  if(F==NULL)
    {
      F=(ZENElt *)ZENMalloc((n+1)*sizeof(ZENElt));
      if(F==NULL)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEF_FACTORIAL,ZEN_ALLOC);
	  return(ZERR);
	}
      for(i=0;i<=n;i++)
	if(ZefEltCreate(F[i],R))
	  {
	    for(;i--;)
	      ZefEltFree(F[i],R);
	    ZENFree(F);
	    F=NULL;
	    return(ZERR);
	  }
      ZefEltSetToOne(F[0],R);
      nmax=0;
    }
  else
    {
      ZENElt *g;

      g=F;
      F=(ZENElt *)ZENMalloc((n+1)*sizeof(ZENElt));
      if(F==NULL)
	{
	  F=g;
	  ZENSetError(ZEN_LIB_ZEF,ZEF_FACTORIAL,ZEN_ALLOC);
	  return(ZERR);
	}
      for(i=0;i<=nmax;i++)
	F[i]=g[i];
      ZENFree(g);
      for(;i<=n;i++)
	if(ZefEltCreate(F[i],R))
	  {
	    for(;i--;)
	      ZefEltFree(F[i],R);
	    ZENFree(F);
	    F=NULL;
	    return(ZERR);
	  }
    }

  for(i=nmax;i<n;i++)
    {
      ZefEltGetFromZs(ZENRingHbuf(R),i+1,1,R);
      ZefEltMultiply(F[i+1],F[i],ZENRingHbuf(R),R);
    }
  nmax=n;
  ZefEltAssign(G,F[nmax],R);
  return(0);  
}

/*
   \TEX{
   \title{Binomial coefficient}
   \input{An allocated fraction, two integers and the |ZENRing| representing $\Q$.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|G| is filled with $\ncr{|n|}{|p|}$. By convention, the binomial coefficient is set to zero if $|p|<|n|$.}
   \note{This function makes calls to |ZefEltFactorial|.}
   } 
   */

int ZefEltBinomial(F,n,p,R)
     ZENElt F;
     int n,p;
     ZENRing R;
{
  if(n<p)
    {
      ZefEltSetToZero(F,R);
      return(0);
    }
  if((n==p)||(p==0))
    {
      ZefEltSetToOne(F,R);
      return(0);
    }

  if(ZefEltFactorial(F,p,R))
    return(ZERR);
  if(ZefEltFactorial(ZENRingHbuf(R),n-p,R))
    return(ZERR);  
  ZefEltMultiply(F,ZENRingHbuf(R),F,R);
  if(ZefEltInverse(F,F,R))
    return(ZERR);
  if(ZefEltFactorial(ZENRingHbuf(R),n,R))
    return(ZERR);  
  ZefEltMultiply(F,ZENRingHbuf(R),F,R);
  return(0);
}
