/*
 * zen/zef/c/elt_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{input/output}
   }
   */
#include <stdio.h>
#include <ctype.h>
#include "zef.h"


/* \TEX[
   \title{Converting from string.}
   \input{A finite field |Rg|, a pointer on a |ZENElt| |G| and a string
   |s| representing a |ZENElt| in base $|base|\in \{2,\ldots,
   16\}$.} 
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|G| is filled with |s| if the output is not |ZENERR|.}
   }
*/
int _ZefEltReadFromString(G, s, base, Rg)
     ZENElt *G;
     char *s;
     int base;
     ZENRing Rg;
{
  char *begin;
  int ret;
  
  for(begin=s; isspace(*begin); begin++);

  if(*begin == '-')
    {
      ZefEltFlag(*G,Rg) |= ZEF_NEGATIVE;
      for(begin++; isspace(*begin); begin++);
    }
  else
    ZefEltFlag(*G,Rg) &= ~ZEF_NEGATIVE;

  ZefEltFlag(*G,Rg) |= ZEF_FRAC;
  
  if(ZefEltNlm(*G,Rg))
    ZBNF(ZefEltN(*G,Rg));
  ret=ZBNReadFromString(&(ZefEltN(*G,Rg)),&(ZefEltNlm(*G,Rg)),begin,base);
  if(ret<0)
    return(ZERR);
  ZefEltNl(*G,Rg)=ZBNNumDigits(ZefEltN(*G,Rg),ZefEltNlm(*G,Rg));
  for(begin+=ret;isspace(*begin);begin++);
  if(*begin !='/')
    {
      ZENSetError(ZEN_LIB_ZEF, ZEN_READFROMSTRING, ZEN_SYNTAX);
      return(ZERR);
    }
  for(begin++; isspace(*begin); begin++);

  if(ZefEltDlm(*G,Rg))
    ZBNF(ZefEltD(*G,Rg));
  ret=ZBNReadFromString(&(ZefEltD(*G,Rg)),&(ZefEltDlm(*G,Rg)),begin,base);
  if(ret<0)
    return(ZERR);
  ZefEltDl(*G,Rg)=ZBNNumDigits(ZefEltD(*G,Rg),ZefEltDlm(*G,Rg));
  
  ZefEltAdjustType(*G,Rg);
  ZefEltFlag(*G,Rg) |= ZEF_TOSIMPLIFY;
  ZefEltE(*G,Rg)=0;
  return(begin+ret-s);
}


/* \TEX[
   \title{Converting to string.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{An allocated string representing |G| in base |base| or |ZENNULL| if an error occured.}   
   }
   */
char *_ZefEltPrintToString(GG, base, Rg)
     ZENElt GG;
     int base;
     ZENRing Rg;  
{
  char *ret,*s1,*s2,sgn[3];
  ZENElt G;

  
  if(ZefEltCreate(G,Rg))
    return(ZNULL);
  ZefEltAssign(G,GG,Rg);
  
  if(ZefEltSimplify(G,0,Rg)) {
    ZefEltFree(G,Rg);
    return(ZNULL);
  }
  if(ZefEltUnsimplify(G,Rg)) {
    ZefEltFree(G,Rg);    
    return(ZNULL);
  }
  if(ZefEltSign(G,Rg))
    {
      sgn[0]='-';
      sgn[1]=' ';
      sgn[2]=0;
    }
  else
    sgn[0]=0;
  if(ZefEltType(G,Rg)&ZEF_INTEGER)
    s1=ZBNPrintToString(ZefEltN(G,Rg),ZefEltNl(G,Rg),base);
  else
    {
      s1=ZENMalloc(2);
      if(s1==NULL)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_PRINTTOSTRING,ZEN_ALLOC);
	  ZefEltFree(G,Rg);    
	  return(ZNULL);
	}
      s1[0]='1';
      s1[1]=0;
    }
  if(ZefEltType(G,Rg)&ZEF_INVERSE)
    s2=ZBNPrintToString(ZefEltD(G,Rg),ZefEltDl(G,Rg),base);
  else
    {
      s2=ZENMalloc(2);
      if(s2==NULL)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_PRINTTOSTRING,ZEN_ALLOC);
	  ZefEltFree(G,Rg);    
	  return(ZNULL);
	}
      s2[0]='1';
      s2[1]=0;
    }
  if((s1 == NULL) ||
     (s2 == NULL)) {
    ZefEltFree(G,Rg);    
    return(ZNULL);
  }
  ret=ZENMalloc(strlen(s1)+strlen(s2)+strlen(sgn)+4);
  if(ret == NULL)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_PRINTTOSTRING,ZEN_ALLOC);
      ZefEltFree(G,Rg);    
      return(ZNULL);
    }
  if(sprintf(ret,"%s%s / %s",sgn,s1,s2)<0)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_PRINTTOSTRING,ZEN_SPRINTF);
      ZefEltFree(G,Rg);    
      return(ZNULL);
    }
  ZENFree(s1);
  ZENFree(s2);
  ZefEltFree(G,Rg);    

  return(ret);
}


/* \TEX[
   \title{Converting from a string in an internal representation.}
   \input{A finite field |Rg| and a string |s| representing a |ZENElt|
   in an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|G| is filled with s.}
   }
   */
int _ZefEltGetFromString(G, s, Rg)
     char *s;
     ZENElt *G;
     ZENRing Rg;
{
  char *ptr;
  int ret;
  
  for(ptr=s;isspace(*ptr);ptr++);
  if(sscanf(ptr,"%8x%8x",&(ZefEltE(*G,Rg)),&(ZefEltFlag(*G,Rg))) == EOF)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_GETFROMSTRING,ZEN_SSCANF);
      return(ZERR);
    }
  for(ptr+=16;isspace(*ptr);ptr++);
  if(ZefEltFlag(*G,Rg) & ZEF_INTEGER)
    {
      if(ZefEltNlm(*G,Rg))
	ZBNF(ZefEltN(*G,Rg));
      ret=ZBNGetFromString(&(ZefEltN(*G,Rg)),&(ZefEltNlm(*G,Rg)),ptr);
      if(ret<0)
	return(ZERR);
      ZefEltNl(*G,Rg)=ZBNNumDigits(ZefEltN(*G,R),ZefEltNlm(*G,R));
      for(ptr+=ret;isspace(*ptr);ptr++);
    }
  if(ZefEltFlag(*G,Rg) & ZEF_INVERSE)
    {
      if(ZefEltDlm(*G,Rg))
	ZBNF(ZefEltD(*G,Rg));
      ret=ZBNGetFromString(&(ZefEltD(*G,Rg)),&(ZefEltDlm(*G,Rg)),ptr);
      if(ret<0)
	return(ZERR);
      ZefEltDl(*G,Rg)=ZBNNumDigits(ZefEltD(*G,R),ZefEltDlm(*G,R));
      for(ptr+=ret;isspace(*ptr);ptr++);
    }
  return(ptr-s);
}


/* \TEX[
   \title{Converting to string in an internal representation.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg|.}
   \output{An allocated string representing |G| in an internal representation or |ZENNULL| if an error occured.}   
   }
   */
char *_ZefEltPutToString(GG, Rg)
     ZENElt GG;
     ZENRing Rg;
{
  char *str,in[1024],*s1,*s2;
  ZENElt G;

  
  if(ZefEltCreate(G,Rg))
    return(ZNULL);
  ZefEltAssign(G,GG,Rg);
    
  if(ZefEltSimplify(G,0,Rg)) {
    ZefEltFree(G,Rg);
    return(ZNULL);
  }
  if(sprintf(in,"%.8x%.8x",(unsigned int)ZefEltE(G,Rg),ZefEltFlag(G,Rg))<0)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_PUTTOSTRING,ZEN_SPRINTF);
      ZefEltFree(G,Rg);    
      return(ZNULL);
    }
  if(ZefEltFlag(G,Rg)&ZEF_INTEGER)
    {
      s1=ZBNPutToString(ZefEltN(G,Rg),ZefEltNl(G,Rg));
      if(!s1) {
	ZefEltFree(G,Rg);
	return(ZNULL);
      }
    }
  else
    {
      s1=(char *)ZENMalloc(1);
      if(!s1)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_PUTTOSTRING,ZEN_ALLOC);
	    ZefEltFree(G,Rg);    
	  return(ZNULL);
	}
      s1[0]=0;
    }
  if(ZefEltFlag(G,Rg)&ZEF_INVERSE)
    {
      s2=ZBNPutToString(ZefEltD(G,Rg),ZefEltDl(G,Rg));
      if(!s2) {
	ZefEltFree(G,Rg);
	return(ZNULL);
      }
    }
  else
    {
      s2=(char *)ZENMalloc(1);
      if(!s2)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_PUTTOSTRING,ZEN_ALLOC);
	  ZefEltFree(G,Rg);    
	  return(ZNULL);
	}
      s2[0]=0;
    }
  str=(char *)ZENMalloc(strlen(in)+strlen(s1)+strlen(s2)+1);
  if(!str)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_PUTTOSTRING,ZEN_ALLOC);
      ZefEltFree(G,Rg);    
      return(ZNULL);
    }
  if(sprintf(str,"%s%s%s",in,s1,s2)<0)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_PUTTOSTRING,ZEN_SPRINTF);
      ZefEltFree(G,Rg);    
      return(ZNULL);
    }
  ZENFree(s1);
  ZENFree(s2);
  ZefEltFree(G,Rg);    
  return(str);
}


/* \TEX[
   \title{Printing to file.}
   \input{A stream |file|, an allocated |ZENElt| |G| of a finite field
	    |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
	    \output{|ZENERR| if an error occured, 0 otherwise.}
	    \side{Printing a representation of |G| in base |base| to |file|.}
	    }
	    */
int _ZefEltPrintToFile(file, GG, base, Rg)
     FILE *file;
     ZENElt GG;
     int base;
     ZENRing Rg;  
{
  ZENElt G;
  
  if(ZefEltCreate(G,Rg))
    return(ZERR);
  ZefEltAssign(G,GG,Rg);
    
  if(ZefEltSimplify(G,0,Rg)) {
    ZefEltFree(G,Rg);
    return(ZERR);
  }
  if(ZefEltUnsimplify(G,Rg)) {
    ZefEltFree(G,Rg);
    return(ZERR);
  }
  if((ZefEltSign(G,Rg))
     &&(fprintf(file,"- ")<0))
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_PRINTTOFILE,ZEN_FPRINTF);
      ZefEltFree(G,Rg);    
      return(ZERR);
    }
  
  if(ZefEltType(G,Rg)&ZEF_INTEGER)
    {
      if(ZBNPrintToFile(file,ZefEltN(G,Rg),ZefEltNl(G,Rg),base)) {
	ZefEltFree(G,Rg);
	return(ZERR);
      }
    }
  else if(fprintf(file,"1")<0)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_PRINTTOFILE,ZEN_FPRINTF);
      ZefEltFree(G,Rg);    
      return(ZERR);
    }
    
  if(fprintf(file," / ")<0)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_PRINTTOFILE,ZEN_FPRINTF);
      ZefEltFree(G,Rg);    
      return(ZERR);
    }
  if(ZefEltType(G,Rg)&ZEF_INVERSE)
    {
      if(ZBNPrintToFile(file,ZefEltD(G,Rg),ZefEltDl(G,Rg),base)) {
	ZefEltFree(G,Rg);
	return(ZERR);
      }
    }
  else if(fprintf(file,"1")<0)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_PRINTTOFILE,ZEN_FPRINTF);
      ZefEltFree(G,Rg);    
      return(ZERR);
    }
  ZefEltFree(G,Rg);    
  return(0);
}

/* \TEX[
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |ZENElt| |G| of a finite field
   |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |G| to |file|.}
   }
   */
int _ZefEltPutToFile(file, GG, Rg)
     FILE *file;
     ZENElt GG;
     ZENRing Rg;  
{
  ZENElt G;
    
  if(ZefEltCreate(G,Rg))
    return(ZERR);
  ZefEltAssign(G,GG,Rg);
  

  if(ZefEltSimplify(G,0,Rg)) {
    ZefEltFree(G,Rg);
    return(ZERR);
  }
  if(fprintf(file,"%.8x%.8x",(unsigned int)ZefEltE(G,Rg),ZefEltFlag(G,Rg))<0)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_PUTTOFILE,ZEN_FPRINTF);
      ZefEltFree(G,Rg);    
      return(ZERR);
    }
  if(ZefEltFlag(G,Rg)&ZEF_INTEGER)
    {
      if(ZBNPutToFile(file,ZefEltN(G,Rg),ZefEltNl(G,Rg))) {
	ZefEltFree(G,Rg);
	return(ZERR);
      }
    }
  if(ZefEltFlag(G,Rg)&ZEF_INVERSE)
    {
      if(ZBNPutToFile(file,ZefEltD(G,Rg),ZefEltDl(G,Rg))) {
	ZefEltFree(G,Rg);
	return(ZERR);
      }
    }
  ZefEltFree(G,Rg);    
  return(0);
}

/* \TEX[
   \title{Reading from file.}
   \input{A stream |file| and an element |G| of a finite field |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|G| is filled with the ZENElt read in |file| if no error occurred.}
   }
   */
int _ZefEltReadFromFile(G, file, base, Rg)
     ZENElt *G;
     FILE *file;
     int base;
     ZENRing Rg;
{
  int c,bufsiz;
  char *buf,*ptr;

  bufsiz = BUFSIZ;
  if (!(buf = (char *)ZENMalloc(sizeof(char)*bufsiz))) {
    ZENSetError(ZEN_LIB_ZBN, ZEN_READFROMFILE, ZEN_ALLOC);
    return(ZERR); 
  }
 
  for(c=fgetc(file);isspace(c);c=fgetc(file));
  if(c == '-')
    {
      ZefEltFlag(*G,Rg) |= ZEF_NEGATIVE;
      for(c=fgetc(file);isspace(c);c=fgetc(file));
    }
  else
    ZefEltFlag(*G,Rg) &= ~ZEF_NEGATIVE;

  ZefEltFlag(*G,Rg) |= ZEF_FRAC;
  
  ptr=buf;
  for(;isxdigit(c);c=fgetc(file))
    {
      *(ptr++)=c;
      if(ptr>=buf+bufsiz)
	{
	  bufsiz += BUFSIZ;
	  if (!(buf = (char *)ZENRealloc(buf, sizeof(char)*bufsiz))) {
	    ZENSetError(ZEN_LIB_ZEF, ZEN_READFROMFILE, ZEN_ALLOC);
	    return(ZERR); 
	  }
	  ptr = buf + bufsiz-BUFSIZ;
	}
    }
  *ptr = 0;
  if(ZefEltNlm(*G,Rg))
    ZBNF(ZefEltN(*G,Rg));
  if(ZBNReadFromString(&(ZefEltN(*G,Rg)),&(ZefEltNlm(*G,Rg)),buf,base)<0)
    return(ZERR);
  ZefEltNl(*G,Rg)=ZBNNumDigits(ZefEltN(*G,Rg),ZefEltNlm(*G,Rg));
  ZBNF(buf);
  for(;isspace(c);c=fgetc(file));
  if(c !='/')
    {
      ZENSetError(ZEN_LIB_ZEF, ZEN_READFROMSTRING, ZEN_SYNTAX);
      return(ZERR);
    }

  if(ZefEltDlm(*G,Rg))
    ZBNF(ZefEltD(*G,Rg));
  if(ZBNReadFromFile(&(ZefEltD(*G,Rg)),&(ZefEltDlm(*G,Rg)),file,base))
    return(ZERR);
  ZefEltDl(*G,Rg)=ZBNNumDigits(ZefEltD(*G,Rg),ZefEltDlm(*G,Rg));
  
  ZefEltAdjustType(*G,Rg);
  ZefEltFlag(*G,Rg) |= ZEF_TOSIMPLIFY;
  ZefEltE(*G,Rg)=0;
  return(0);
}

/* \TEX[
   \title{Getting from file in an internal representation.}
   \input{A stream |file|, and an element |G| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|G| is filled with the element read in |file|}
   }
   */
int _ZefEltGetFromFile(G, file, Rg)
     ZENElt *G;
     FILE *file;
     ZENRing Rg;
{
   if(fscanf(file,"%8x%8x",&(ZefEltE(*G,Rg)),&(ZefEltFlag(*G,Rg))) == EOF)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_GETFROMSTRING,ZEN_SSCANF);
      return(ZERR);
    }
  if(ZefEltFlag(*G,Rg) & ZEF_INTEGER)
    {
      if(ZefEltNlm(*G,Rg))
	ZBNF(ZefEltN(*G,Rg));
      if(ZBNGetFromFile(&(ZefEltN(*G,Rg)),&(ZefEltNlm(*G,Rg)),file))
	return(ZERR);
      ZefEltNl(*G,Rg)=ZBNNumDigits(ZefEltN(*G,R),ZefEltNlm(*G,R));
    }
  if(ZefEltFlag(*G,Rg) & ZEF_INVERSE)
    {
      if(ZefEltDlm(*G,Rg))
	ZBNF(ZefEltD(*G,Rg));
      if(ZBNGetFromFile(&(ZefEltD(*G,Rg)),&(ZefEltDlm(*G,Rg)),file))
	return(ZERR);
      ZefEltDl(*G,Rg)=ZBNNumDigits(ZefEltD(*G,R),ZefEltDlm(*G,R));
    }
  return(0);
}

/* \TEX{
   \title{Evaluating an element over $Z$.}
   \input{A Bignum |p| of allocated size greater or equal than
   the numerator of |GG|, a pointer |p_pl| on the real size of |p|, and
   an infinite ring |Rg|.}
   \side{|p| is filled with the value of |GG| numerator,
   |*p_pl| contains the real size of |p|.}
   \note{Error flag is raised if the fraction is not a natural integer.}
   } */
void _ZefEltToZ(p, p_pl, GG, Rg)
     BigNum p;
     BigNumLength *p_pl;
     ZENElt GG;
     ZENRing Rg;
{
  ZENElt G;
  int ret;
  BigNum buf;
  BigNumLength bufl;
    
  if(ZefEltCreate(G,Rg)) {
    return;
  }
  ZefEltAssign(G,GG,Rg);
  
  if(ZefEltSimplify(G,0,Rg)) {
    ZefEltFree(G,Rg);
    return;
  }
  if(ZefEltType(G,Rg)&ZEF_INVERSE)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_ELTTOZ,ZEN_INPUT);
      ZefEltFree(G,Rg);
      return;
    }
  if(ZefEltType(G,Rg)==ZEF_ONE)
    {
      if(ZefEltNlm(G,Rg)<1)
	{
	  ZefEltN(G,Rg)=ZBNC((BigNumLength)1);
	  if(ZefEltN(G,Rg)==NULL)
	    {
	      ZENSetError(ZEN_LIB_ZEF,ZEN_ELTTOZ,ZEN_ALLOC);
	      ZefEltFree(G,Rg);
	      return;
	    }
	  ZefEltNlm(G,Rg)=ZefEltNl(G,Rg)=(BigNumLength)1;
	}
      ZBNDigit(ZefEltN(G,Rg),0)=(BigNumDigit)1;
      ZefEltNl(G,Rg)=1;
    }

  if(!(ZefEltFlag(G,Rg) & ZEF_TOEXP))
    {
      if(ZefEltE(G,Rg)<0)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_ELTTOZ,ZEN_INPUT);
	  ZefEltFree(G,Rg);
	  return;
	}
      bufl=ZefEltNl(G,Rg)+divSizeBloc(ZefEltE(G,Rg))+1;
      buf=ZBNC(bufl);
      if(buf==NULL)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_ELTTOZ,ZEN_ALLOC);
	  ZefEltFree(G,Rg);
	  return;
	}

      ZBNAnyShiftLeft(buf,&bufl,ZefEltN(G,Rg),ZefEltNl(G,Rg),ZefEltE(G,Rg));
    }
  else
    {
      bufl=ZefEltNl(G,Rg);
      buf=ZBNC(bufl);
      if(buf==NULL)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_ELTTOZ,ZEN_ALLOC);
	  ZefEltFree(G,Rg);
	  return;
	}
      ZBNAssign(buf,ZefEltN(G,Rg),ZefEltNl(G,Rg));
    }

  if(bufl>*p_pl)
    {
      ZENSetError(ZEN_LIB_ZEF,ZEN_ELTTOZ,ZEN_INPUT);
      ZBNAssign(p,buf,*p_pl);
      *p_pl=bufl;      
    }
  else
    {
      *p_pl=bufl;      
      ZBNAssign(p,buf,*p_pl);
    }
  ZBNF(buf);
  ZefEltFree(G,Rg);
}

/* \TEX{
   \title{Getting an element from its evaluation over $Z$.}
   \input{A Bignum |p| of size |pl|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|G| is filled with the integer |Z|.}
   } */
void _ZefZToElt(pa, p, pl, Rg)
     BigNum p;
     BigNumLength pl;
     ZENElt *pa;
     ZENRing Rg;
{
  ZefEltE(*pa,Rg)=0;

  if(ZefEltNlm(*pa,Rg)==0)
    {
      if(ZefEltDlm(*pa,Rg)>0)
	{
	  ZefEltN(*pa,Rg)=ZefEltD(*pa,Rg);
	  ZefEltNlm(*pa,Rg)=ZefEltDlm(*pa,Rg);
	  ZefEltDlm(*pa,Rg)=0;
	}
    }

  if(ZefEltNlm(*pa,Rg)<pl)
    {
      ZefEltN(*pa,Rg)=ZBNC(pl);
      if(ZefEltN(*pa,Rg)==NULL)
	{
	  ZENSetError(ZEN_LIB_ZEF,ZEN_UNKNOWN,ZEN_ALLOC);
	  return;
	}
      ZefEltNlm(*pa,Rg)=pl;
    }

  ZBNAssign(ZefEltN(*pa,Rg),p,pl);
  ZefEltNl(*pa,Rg)=pl;      
  ZefEltFlag(*pa,Rg)=ZEF_INTEGER;
}

/* \TEX{
   \Subsubsection{Debugging purpose function}

   The following function can be called under gdb to print fractions.
   \title{Printing to stdout.}
   \input{An allocated |ZENElt| |M| of $\Q$.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing the strucutre of |M| and a representation of |M| in base 10 to |stdout|.}    
   }
   */
int ZFP(M, Rg)
     ZENElt M;
     ZENRing Rg;  
{
  printf("N[%d][%d]=",(int)ZefEltNl(M,Rg),(int)ZefEltNlm(M,Rg));
  if(ZefEltN(M,Rg))
    ZBP(ZefEltN(M,Rg),ZefEltNl(M,Rg));
  else
    printf("NULL\n");
  printf("D[%d][%d]=",(int)ZefEltDl(M,Rg),(int)ZefEltDlm(M,Rg));
  if(ZefEltD(M,Rg))
    ZBP(ZefEltD(M,Rg),ZefEltDl(M,Rg));
  else
    printf("NULL\n");
  printf("exp=%d ",ZefEltE(M,Rg));
  printf("flag=%d\n",ZefEltFlag(M,Rg));
  if(ZefEltFlag(M,R) & ZEF_NEGATIVE) printf ("NEGATIVE ");
  if(ZefEltFlag(M,R) & ZEF_INTEGER) printf ("INTEGER ");
  if(ZefEltFlag(M,R) & ZEF_INVERSE) printf ("INVERSE ");
  if(ZefEltFlag(M,R) & ZEF_TOEXP) printf ("TOEXP ");
  if(ZefEltFlag(M,R) & ZEF_TOPGCD) printf ("TOPGCD ");
  printf("\n");
  return(ZEP(M, Rg));
}
