/*
 * zen/zeg/c/clone_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{input/output}
   
   Conversions between the internal clone representation and the ASCII
   representation of a |ZENElt| are done by |_ZENClonePrintToString|. |_ZENClonePrintToFile|
   allows to write a |ZENElt| in a file.
   One can moreover write to and then read from a file a |ZENelt| with |_ZENClonePutToFile|
   and |_ZENCloneReadFromFile|. Extensive use of |ZENEltToClone| and |ZENCloneToElt|.
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX[
   \title{Converting from string.}
   \input{A finite field |Rg|, a pointer on a |ZENElt| |G| and a string
   |s| representing a |ZENElt| in base $|base|\in \{2,\ldots,16\}$.} 
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|G| is filled with |s| if the output is not |ZENERR|.}
   }
   */
int _ZENCloneEltReadFromString(G, s, base, Rg)
     ZENElt *G;
     char *s;
     int base;
     ZENRing Rg;
{  
  int ret;
  ZENElt elt;

  if (ZENEltCreate(elt, ZENRingOrigin(Rg))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_READFROMSTRING, ZEN_ALLOC);
    return(ZERR);
  }
  
  if ((ret = ZENEltReadFromString(elt, s, base, ZENRingOrigin(Rg))) < 0) {
    ZENEltFree(elt, ZENRingOrigin(Rg));
    return(ret);
  }
  
  ZENEltToClone(*G, elt, Rg);
  ZENEltFree(elt, ZENRingOrigin(Rg));

  return(ret);
}

/* \TEX[
   \title{Converting to string.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{An allocated string representing |G| in base |base| or NULL if an error occured.}   
   }
   */
char *_ZENCloneEltPrintToString(G, base, Rg)
     ZENElt G;
     int base;
     ZENRing Rg;  
{ 
  ZENElt elt;
  char * ret;

  if (ZENEltCreate(elt, ZENRingOrigin(Rg))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_PRINTTOSTRING, ZEN_ALLOC);
    return(NULL);
  }
  
  ZENCloneToElt(elt, G, Rg);
  
  if (!(ret = ZENEltPrintToString(elt, base, ZENRingOrigin(Rg)))) {
    ZENEltFree(elt, ZENRingOrigin(Rg));
    return(NULL);
  }
  
  ZENEltFree(elt, ZENRingOrigin(Rg));
  return(ret);
}


/* \TEX[
   \title{Converting from a string in an internal representation.}
   \input{A finite field |Rg| and a string |s| representing a |ZENElt|
   in an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|G| is filled with s if the output is different from |NULL|}
   }
   */
int _ZENCloneEltGetFromString(G, s, Rg)
     char *s;
     ZENElt *G;
     ZENRing Rg;
{
  int ret;
  ZENElt elt;

  if (ZENEltCreate(elt, ZENRingOrigin(Rg))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_READFROMSTRING, ZEN_ALLOC);
    return(ZERR);
  }
  
  if ((ret = ZENEltGetFromString(elt, s, ZENRingOrigin(Rg))) < 0) {
    ZENEltFree(elt, ZENRingOrigin(Rg));
    return(ret);
  }
  
  ZENEltToClone(*G, elt, Rg);
  ZENEltFree(elt, ZENRingOrigin(Rg));
  
  return(ret);
}

/* \TEX[
   \title{Converting to string in an internal representation.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg|.}
   \output{An allocated string representing |G| in an internal representation or NULL if an error occured.}   
   }
   */
char *_ZENCloneEltPutToString(G, Rg)
     ZENElt G;
     ZENRing Rg;
{
  ZENElt elt;
  char * ret;

  if (ZENEltCreate(elt, ZENRingOrigin(Rg))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_PRINTTOSTRING, ZEN_ALLOC);
    return(NULL);
  }
  
  ZENCloneToElt(elt, G, Rg);
  
  if (!(ret = ZENEltPutToString(elt, ZENRingOrigin(Rg)))) {
    ZENEltFree(elt, ZENRingOrigin(Rg));
    return(NULL);
  }
  
  ZENEltFree(elt, ZENRingOrigin(Rg));
  return(ret);
}


/* \TEX[
   \title{Printing to file.}
   \input{A stream |file|, an allocated |ZENElt| |G| of a finite field
	    |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
	    \output{|ZENERR| if an error occured, 0 otherwise.}
	    \side{Printing a representation of |G| in base |base| to |file|.}
	    }
	    */
int _ZENCloneEltPrintToFile(file, G, base, Rg)
     FILE *file;
     ZENElt G;
     int base;
     ZENRing Rg;  
{  
  ZENElt elt;
  int ret;

  if (ZENEltCreate(elt, ZENRingOrigin(Rg))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_PRINTTOSTRING, ZEN_ALLOC);
    return(ZERR);
  }
  
  ZENCloneToElt(elt, G, Rg);
  
  if ((ret = ZENEltPrintToFile(file, elt, base, ZENRingOrigin(Rg)))) {
    ZENEltFree(elt, ZENRingOrigin(Rg));
    return(ZERR);
  }
  
  ZENEltFree(elt, ZENRingOrigin(Rg));
  return(ret);
}

/* \TEX[
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |ZENElt| |G| of a finite field
   |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |G| to |file|.}
   }
   */
int _ZENCloneEltPutToFile(file, G, Rg)
     FILE *file;
     ZENElt G;
     ZENRing Rg;  
{
  ZENElt elt;
  int ret;

  if (ZENEltCreate(elt, ZENRingOrigin(Rg))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_PRINTTOSTRING, ZEN_ALLOC);
    return(ZERR);
  }
  
  ZENCloneToElt(elt, G, Rg);
  
  if ((ret = ZENEltPutToFile(file, elt, ZENRingOrigin(Rg)))) {
    ZENEltFree(elt, ZENRingOrigin(Rg));
    return(ZERR);
  }
  
  ZENEltFree(elt, ZENRingOrigin(Rg));
  return(ret);
}

/* \TEX[
   \title{Reading from file.}
   \input{A stream |file| and an element |G| of a finite field |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|G| is filled with the ZENElt read in |file| if no error occurred.}
   }
   */
int _ZENCloneEltReadFromFile(G, file, base, Rg)
     ZENElt *G;
     FILE *file;
     int base;
     ZENRing Rg;
{ 
  int ret;
  ZENElt elt;

  if (ZENEltCreate(elt, ZENRingOrigin(Rg))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_READFROMSTRING, ZEN_ALLOC);
    return(ZERR);
  }
  
  if ((ret = ZENEltReadFromFile(elt, file, base, ZENRingOrigin(Rg))) < 0) {
    ZENEltFree(elt, ZENRingOrigin(Rg));
    return(ret);
  }
  
  ZENEltToClone(*G, elt, Rg);
  ZENEltFree(elt, ZENRingOrigin(Rg));

  return(ret);
}

/* \TEX[
   \title{Getting from file in an internal representation.}
   \input{A stream |file|, and an element |G| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|G| is filled with the element read in |file|}
   }
   */
int _ZENCloneEltGetFromFile(G, file, Rg)
     ZENElt *G;
     FILE *file;
     ZENRing Rg;
{
  int ret;
  ZENElt elt;

  if (ZENEltCreate(elt, ZENRingOrigin(Rg))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_READFROMSTRING, ZEN_ALLOC);
    return(ZERR);
  }
  
  if ((ret = ZENEltGetFromFile(elt, file, ZENRingOrigin(Rg))) < 0) {
    ZENEltFree(elt, ZENRingOrigin(Rg));
    return(ret);
  }
  
  ZENEltToClone(*G, elt, Rg);
  ZENEltFree(elt, ZENRingOrigin(Rg));

  return(ret);
}




