/*
 * zen/zeg/c/debug.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   \Subsection{Debugging purpose functions}

   The following functions can be called under gdb to print or enter from stdin parameters of the program.
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{Printing to stdout.}
   \input{An allocated |BigNum| and its length.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |n| in base 10 to |stderr|.}    
   }
   */
int ZBP(n, nl)
     BigNum n;
     BigNumLength nl;
{
  int ret;
  
  if ((ret = ZBNPrintToFile(stderr, n, nl, 10)) < 0) {
    return(ret);
  }
  if (fprintf(stderr, ";\n") < 0) {
    return(ZERR);
  }
  return(ret);
}

/* \TEX{
   \title{Reading from stdin.}
   \input{An allocated |ZENElt| |M| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Reading a representation of |M| in base 10 from |stdin|.}    
   }
   */
int ZBR(p,pl)
     BigNum *p;
     BigNumLength *pl;
{
  return(ZBNReadFromFile(p,pl, stdin, 10));
}

/* \TEX{
   \title{Printing to stderr.}
   \input{An allocated |ZENElt| |M| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |M| in base 10 to |stderr|.}    
   }
   */
int ZEP(M, Rg)
     ZENElt M;
     ZENRing Rg;  
{
  int ret;
  
  if ((ret = ZENEltPrintToFile(stderr, M, 10, Rg)) < 0) {
    return(ret);
  }
  if (fprintf(stderr, "\n") < 0) {
    return(ZERR);
  }
  return(ret);
}

/* \TEX{
   \title{Reading from stdin.}
   \input{An allocated |ZENElt| |M| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Reading a representation of |M| in base 10 from |stdin|.}    
   }
   */
int ZER(M, Rg)
     ZENElt M;
     ZENRing Rg;  
{
  return(ZENEltReadFromFile(M, stdin, 10, Rg));
}

/* \TEX{
   \title{Printing to stderr.}
   \input{An allocated |ZENPoly| |M| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |M| in base 10 to |stderr|.}    
   }
   */
int ZPP(M, Rg)
     ZENPoly M;
     ZENRing Rg;  
{
  int ret;
  
  if ((ret = ZENPolyPrintToFile(stderr, M, 10, Rg)) < 0) {
    return(ret);
  }
  if (fprintf(stderr, "\n") < 0) {
    return(ZERR);
  }
  return(ret);
}

/* \TEX{
   \title{Reading from stdin.}
   \input{An allocated |ZENPoly| |M| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Reading a representation of |M| in base 10 from |stdin|.}    
   }
   */
int ZPR(M, Rg)
     ZENPoly *M;
     ZENRing Rg;  
{
  return(ZENPolyReadFromFile(*M, stdin, 10, Rg));
}

/* \TEX{
   \title{Printing to stderr.}
   \input{An allocated |ZENMat| |M| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |M| in base 10 to |stderr|.}    
   }
   */
int ZMP(M, Rg)
     ZENMat M;
     ZENRing Rg;  
{
  return(ZENMatPrintToFile(stderr, M, 10, Rg));
}

/* \TEX{
   \title{Reading from stdin.}
   \input{An allocated |ZENMat| |M| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Reading a representation of |M| in base 10 from stdin.}    
   }
   */
int ZMR(M, Rg)
     ZENMat *M;
     ZENRing Rg;  
{
  return(ZENMatReadFromFile(*M, stdin, 10, Rg));
}


/* \TEX{
   \title{Printing to stderr.}
   \input{An allocated |ZENSr| |M| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |M| in base 10 to |stderr|.}    
   }
   */
int ZSP(M, Rg)
     ZENSr M;
     ZENRing Rg;  
{
  int ret;

  if ((ret = ZENSrPrintToFile(stderr, M, 10, Rg)) < 0) {
    return(ret);
  }
  if (fprintf(stderr, "\n") < 0) {
    return(ZERR);
  }
  return(ret);
}

/* \TEX{
   \title{Reading from stdin.}
   \input{An allocated |ZENSr| |M| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Reading a representation of |M| in base 10 from |stdin|.}    
   }
   */
int ZSR(M, Rg)
     ZENSr *M;
     ZENRing Rg;  
{
  return(ZENSrReadFromFile(*M, stdin, 10, Rg));
}



