/*
 * zen/zeg/c/ec_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   \Subsection{input/output}
   
   Conversions between the internal representation and the ASCII
   representation of a |ZENEcPt| are done by
   |_ZENEcPtPrintToString|. |_ZENEcPtPrintToFile| allows to write a
   |ZENEcPt| in a file. One can moreover write to and then read from a
   file a |_ZENEcPt| with |_ZENEcPtPutToFile| and |_ZENEcPtReadFromFile|.
   }
   */
#include <stdio.h>
#include <ctype.h>
#include "zen.h"


/* \TEX{
   \title{Converting from string.}
   \input{An elliptic curve |E|, an allocated point |R| and a string
   |s| representing a point in base $|base|\in \{2,\ldots,
   16\}$.} 
   \output{The number of character read in |s| or |ZENERR| if an error occurred.}
   \side{|R| is filled with |s| if the output is not |ZENERR|.}
   }
*/
int _ZENEcPtReadFromString(R, s, base, E)
     ZENEcPt R;
     char *s;
     int base;
     ZENEc E;
{
  int nbc;
  char *ptr;

  ptr = s;
  for(; isspace(*ptr); ptr++) ;
  if (*ptr != '[') {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTREADFROMSTRING, ZEN_SYNTAX);
    return(ZERR);
  }
    
  ptr++;
  if ((nbc = ZENEltReadFromString(ZENEcPtX(R, E), ptr, base, ZENEcRing(E))) < 0) {
    return(ZERR);
  }
  
  for(ptr += nbc; isspace(*ptr); ptr++) ;
  if (*ptr != ',') {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTREADFROMSTRING, ZEN_SYNTAX);
    return(ZERR);
  }
   
  ptr++;
  if ((nbc = ZENEltReadFromString(ZENEcPtY(R, E), ptr, base, ZENEcRing(E))) < 0) {
    return(ZERR);
  }

  for(ptr += nbc; isspace(*ptr); ptr++) ;
  if (*ptr != ',') {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTREADFROMSTRING, ZEN_SYNTAX);
    return(ZERR);
  }
  
  for(ptr++; isspace(*ptr); ptr++) ;
  switch (*ptr) {
  case '0':
    ZENEcPtZ(R, E) = 0;
    break;
  case '1':
    ZENEcPtZ(R, E) = 1;
    break;
  default:
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTREADFROMSTRING, ZEN_SYNTAX);
    return(ZERR);
  }
  for(ptr ++; isspace(*ptr); ptr++) ;
  if (*ptr != ']') {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTREADFROMSTRING, ZEN_SYNTAX);
    return(ZERR);
  }

  ZENEcPtGeneral2Canonical(R, R, E);
  return(ptr+1-s);
}

/* \TEX{
   \title{Converting to string.}
   \input{An allocated point |R| of an elliptic curve |E| and a base
   $|base|\in \{2,\ldots, 16\}$.} 
   \output{An allocated string representing |R| in base |base| or |NULL|
   if an error occurred.} 
   }
   */
char* _ZENEcPtPrintToString(R, base, E)
     ZENEcPt R;
     int base;
     ZENEc E;
{
  char *sX, *sY, *str;
  
  ZENEcPtCanonical2General(ZENEcXY(E), R, E);

  if (ZENEcPtIsZero(ZENEcXY(E), E)) {
    ZENEltSetToZero(ZENEcX(E), ZENEcRing(E));
    ZENEltSetToOne(ZENEcY(E), ZENEcRing(E));
  }

  if (!(sX = ZENEltPrintToString(ZENEcX(E), base, ZENEcRing(E)))) {
    return(NULL);
  }
  if (!(sY = ZENEltPrintToString(ZENEcY(E), base, ZENEcRing(E)))) {
    ZENFree(sX);
    return(NULL);
  }
  if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(sX)+strlen(sY)+6)))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTPRINTTOSTRING, ZEN_ALLOC);
    ZENFree(sX);
    ZENFree(sY);
    return(NULL);
  }
  if (sprintf(str, "[%s,%s,%d]", sX, sY,ZENEcPtZ(R, E)) < 0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTPRINTTOSTRING, ZEN_SPRINTF);
    ZENFree(sX); ZENFree(sY); ZENFree(str);
    return(NULL);
  }
  
  ZENFree(sX);
  ZENFree(sY);
  return(str);
}

/* \TEX{
   \title{Reading from file.}
   \input{A stream |file|, an allocated point |R| of an elliptic curve
   |E| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|R| is filled with the point read in |file| if no error
   occurred.}
   }
   */
int _ZENEcPtReadFromFile(R, file, base, E)
     ZENEcPt R;
     FILE *file;
     int base;
     ZENEc E;
{
  int c;

  c = fgetc(file);
  for(; isspace(c); c = fgetc(file)) ;

  if (c != '[') {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTREADFROMFILE, ZEN_SYNTAX);
    return(ZERR);
  }
    
  if (ZENEltReadFromFile(ZENEcPtX(R, E), file, base, ZENEcRing(E))) {
    return(ZERR);
  }
  
  for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
  if (c != ',') {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTREADFROMFILE, ZEN_SYNTAX);
    return(ZERR);
  }

  if (ZENEltReadFromFile(ZENEcPtY(R, E), file, base, ZENEcRing(E))) {
    return(ZERR);
  }

  for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
  if (c != ',') {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTREADFROMFILE, ZEN_SYNTAX);
    return(ZERR);
  }

  if ((fscanf(file, "%d", &c) != 1) || 
      ((c != 0) &&
       (c != 1))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTREADFROMFILE, ZEN_SYNTAX);
    return(ZERR);
  }
  ZENEcPtZ(R, E) = c;

  for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
  if (c != ']') {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTREADFROMSTRING, ZEN_SYNTAX);
    return(ZERR);
  }

  ZENEcPtGeneral2Canonical(R, R, E);
  return(0);
}

/* \TEX{
   \title{Printing to file.}
   \input{A stream |file|, an allocated point |R| of an elliptic curve
   |E| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{Printing a representation of |R| in base |base| to |file|.}
   }
   */
int _ZENEcPtPrintToFile(file, R, base, E)
     FILE *file;
     ZENEcPt R;
     int base;
     ZENEc E;
{
  ZENEcPtCanonical2General(ZENEcXY(E), R, E);

  if (ZENEcPtIsZero(ZENEcXY(E), E)) {
    ZENEltSetToZero(ZENEcX(E), ZENEcRing(E));
    ZENEltSetToOne(ZENEcY(E), ZENEcRing(E));
  }

  if (fprintf(file, "[") < 0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTPRINTTOFILE, ZEN_FPRINTF);
    return(ZERR);
  }
  if (ZENEltPrintToFile(file, ZENEcX(E), base, ZENEcRing(E))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTPRINTTOFILE, ZEN_FPRINTF);
    return(ZERR);
  }
  if (fprintf(file, ",") < 0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTPRINTTOFILE, ZEN_FPRINTF);
    return(ZERR);
  }
  if (ZENEltPrintToFile(file, ZENEcY(E), base, ZENEcRing(E))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTPRINTTOFILE, ZEN_FPRINTF);
    return(ZERR);
  }
  if (fprintf(file, ",%d]",ZENEcPtZ(R, E)) < 0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTPRINTTOFILE, ZEN_FPRINTF);
    return(ZERR);
  }
  return(0);
}

/* \TEX{
   \title{Converting from a string in an internal representation.}
   \input{An allocated point |R| of an elliptic curve |E| and a string
   |s| representing a point in an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occurred.}
   \side{|R| is filled with |s| if the output is different of |ZENERR|}
   }
   */
int _ZENEcPtGetFromString(R, s, E)
     ZENEcPt R;
     char *s;
     ZENEc E;
{
  char *ptr;
  int nbc;

  for(ptr = s; isspace(*ptr); ptr++) ;
  if ((nbc = ZENEltGetFromString(ZENEcPtX(R, E), ptr, ZENEcRing(E))) < 0) {
    return(ZERR);
  }
  ptr += nbc;
  if ((nbc = ZENEltGetFromString(ZENEcPtY(R, E), ptr, ZENEcRing(E))) < 0) {
    return(ZERR);
  }
  ptr += nbc;
  if (sscanf(ptr, "%1x", &nbc) != 1) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTGETFROMSTRING, ZEN_SSCANF);
    return(ZERR);
  }
  ZENEcPtZ(R, E) = nbc;

  ZENEcPtGeneral2Canonical(R, R, E);
  return(ptr+1-s);
}

/* \TEX{
   \title{Converting to string in an internal representation.}
   \input{An allocated point |R| of an elliptic curve |E|.}
   \output{An allocated string representing |R| in an internal
   representation or |NULL| if an error occurred.} 
   }
   */
char* _ZENEcPtPutToString(R, E)
     ZENEcPt R;
     ZENEc E;
{
  char *sX, *sY, *str;

  ZENEcPtCanonical2General(ZENEcXY(E), R, E);
  if (ZENEcPtIsZero(R, E)) {
    ZENEltSetToZero(ZENEcX(E), ZENEcRing(E));
    ZENEltSetToOne(ZENEcY(E), ZENEcRing(E));
  }

  if (!(sX = ZENEltPutToString(ZENEcX(E), ZENEcRing(E)))) {
    return(NULL);
  }
  if (!(sY = ZENEltPutToString(ZENEcY(E), ZENEcRing(E)))) {
    ZENFree(sX);
    return(NULL);
  }
  if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(sX)+strlen(sY)+2)))) {
    ZENFree(sX);
    ZENFree(sY);
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTPRINTTOSTRING, ZEN_ALLOC);
    return(NULL);
  }
  if (sprintf(str, "%s%s%.1x", sX, sY,ZENEcPtZ(R, E)) < 0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTPRINTTOSTRING, ZEN_SPRINTF);
    ZENFree(sX); ZENFree(sY); ZENFree(str);
    return(NULL);
  }
  
  ZENFree(sX);
  ZENFree(sY);
  return(str);
}

/* \TEX{
   \title{Getting from file in an internal representation.}
   \input{A stream |file|, and an point |R| of an elliptic curve |E|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{|R| is filled with the point read in |file|}
   }
   */
int _ZENEcPtGetFromFile(R, file, E)
     ZENEcPt R;
     FILE *file;
     ZENEc E;
{
  int c;

  if (ZENEltGetFromFile(ZENEcPtX(R, E), file, ZENEcRing(E))) {
    return(ZERR);
  }
  if (ZENEltGetFromFile(ZENEcPtY(R, E), file, ZENEcRing(E))) {
    return(ZERR);
  }
  if ((c = fgetc(file)) == EOF) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTGETFROMFILE, ZEN_SYNTAX);
    return(ZERR);
  }  
  ZENEcPtZ(R, E) = c-'0';
  ZENEcPtGeneral2Canonical(R, R, E);
  return(0);
}

/* \TEX{
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated point |R| of an elliptic curve
   |E|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{Printing a representation of |R| to |file|.}
   }
   */
int _ZENEcPtPutToFile(file, R, E)
     FILE *file;
     ZENEcPt R;
     ZENEc E;
{
  ZENEcPtCanonical2General(ZENEcXY(E), R, E);
  if (ZENEcPtIsZero(R, E)) {
    ZENEltSetToZero(ZENEcX(E), ZENEcRing(E));
    ZENEltSetToOne(ZENEcY(E), ZENEcRing(E));
  }

  if (ZENEltPutToFile(file, ZENEcX(E), ZENEcRing(E))) {
    return(ZERR);
  }
  if (ZENEltPutToFile(file, ZENEcY(E), ZENEcRing(E))) {
    return(ZERR);
  }
  if (fprintf(file, "%.1x", ZENEcPtZ(R, E)) < 0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTPUTTOFILE, ZEN_FPRINTF);
    return(ZERR);
  }
  return(0);
}
