/*
 * zen/zeg/c/ecpt_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Elliptic curve points additions}   
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{Addition.}
   \input{Three allocated points |P|, |Q| and |R| of an elliptic curve |E|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|R| = |P|+|Q|$ if |ZEN_HAS_INVERSE| is returned,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZENEcPtAdd(R, P, Q, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEcPt Q;
     ZENEc E;
{
  int ret;

  if (ZENEcPtIsZero(P, E)) {
    ZENEcPtAssign(R, Q, E);
    return(ZEN_HAS_INVERSE);
  }
  if (ZENEcPtIsZero(Q, E)) {
    ZENEcPtAssign(R, P, E);
    return(ZEN_HAS_INVERSE);
  }
  if (ZENEltAreEqual(ZENEcPtX(P, E), ZENEcPtX(Q, E), ZENEcRing(E))) {
    if (ZENEltAreEqual(ZENEcPtY(P, E), ZENEcPtY(Q, E), ZENEcRing(E))) {
      return(ZENEcPtDouble(R, P, E));
    }
    ZENEcPtSetToZero(R, E);
    return(ZEN_HAS_INVERSE);
  }
  ZENEcPtPreAdd(P, Q, E);
  if ((ret = ZENEltInverse(ZENEcX(E), ZENEcX(E), ZENEcRing(E)))
      != ZEN_HAS_INVERSE) {
    return(ret);
  }
  ZENEcPtPostAdd(R, P, Q, E);
  ZENEcPtZ(R, E) = 1;

  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Doubling.}
   \input{Two allocated points |P| and |R| of an elliptic curve |E|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|R| = 2|P|$ if |ZEN_HAS_INVERSE| is returned,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZENEcPtDouble(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;
{
  int ret;

  if (ZENEcPtIsZero(P, E)) {
    ZENEcPtSetToZero(R, E);
    return(ZEN_HAS_INVERSE);
  }  
  ZENEcPtPreDouble(P, E);
  if (ZENEltIsZero(ZENEcY(E), ZENEcRing(E))) {
    ZENEcPtSetToZero(R, E);
    return(ZEN_HAS_INVERSE);
  }
  if ((ret = ZENEltInverse(ZENEcY(E), ZENEcY(E), ZENEcRing(E)))
      != ZEN_HAS_INVERSE) {
    return(ret);
  }
  ZENEcPtPostDouble(R, P, E);
  ZENEcPtZ(R, E) = 1;
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Subtraction.}
   \input{Three allocated points |P|, |Q| and |R| of an elliptic curve |E|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|R| = |P|-|Q|$ if |ZEN_HAS_INVERSE| is returned,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   \note{We must have $|R|\neq|P|$}
   }
   */
int _ZENEcPtSubtract(R, P, Q, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEcPt Q;
     ZENEc E;
{
  ZENEcPtNegate(R, Q, E);
  return(ZENEcPtAdd(R, P, R, E));
}

/* \TEX{
   \title{First step for adding 2 points.}
   \input{Two allocated points |P| and |Q| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| are set.}
   }
   */
void _ZENEcPtPreAdd(P, Q, E)
     ZENEcPt P;
     ZENEcPt Q;
     ZENEc E;
{ 
  ZENEltAssign(ZENEcX(E), ZENEcPtX(Q, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcX(E), ZENEcPtX(P, E), ZENEcRing(E));  

}

/* \TEX{
   \title{Last step for adding 2 points in any finite fields.}
   \input{Three allocated points |R|, |P| and |Q| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| must be correctly initialized.}
   }
   */
void _ZENEcPtPostAddAllChar(R, P, Q, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEcPt Q;
     ZENEc E;
{
  ZENEltAssign(ZENEcY(E), ZENEcPtY(Q, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcY(E), ZENEcPtY(P, E), ZENEcRing(E));

  ZENEltMultiply(ZENEcY(E), ZENEcX(E), ZENEcY(E), ZENEcRing(E));
    
  ZENEltNegate(ZENEcPtY(R, E), ZENEcPtY(P, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcPtY(R, E), ZENEcA3(E), ZENEcRing(E));
  ZENEltMultiply(ZENEcX(E), ZENEcY(E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcPtY(R, E), ZENEcX(E), ZENEcRing(E));
  
  ZENEltAssign(ZENEcX(E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcPtX(Q, E), ZENEcRing(E));
  ZENEltNegate(ZENEcPtX(R, E), ZENEcX(E), ZENEcRing(E));
  ZENEltSubtract(ZENEcPtX(R, E), ZENEcA2(E), ZENEcRing(E));

  ZENEltAssign(ZENEcX(E), ZENEcA1(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcY(E), ZENEcRing(E));
  ZENEltMultiply(ZENEcY(E), ZENEcX(E), ZENEcY(E), ZENEcRing(E));
  ZENEltAdd(ZENEcPtX(R, E), ZENEcY(E), ZENEcRing(E));

  ZENEltMultiply(ZENEcY(E), ZENEcX(E), ZENEcPtX(R, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcPtY(R, E), ZENEcY(E), ZENEcRing(E));
}

/* \TEX{
   \title{First step for doubling 1 point in any finite field.}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| are set.}
   }
   */
void _ZENEcPtPreDoubleAllChar(P, E)
     ZENEcPt P;
     ZENEc E;
{
  ZENEltMultiply(ZENEcY(E), ZENEcA1(E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcY(E), ZENEcA3(E), ZENEcRing(E));
  ZENEltAdd(ZENEcY(E), ZENEcPtY(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcY(E), ZENEcPtY(P, E), ZENEcRing(E));
}

/* \TEX{
   \title{Last step for doubling 1 point in any finite field.}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \side{The buffers |ZENEcX(E)| and |ZENEcY(E)| must be correctly initialized.}
   }
   */

void _ZENEcPtPostDoubleAllChar(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;
{
  ZENEltMultiply(ZENEcX(E), ZENEcA2(E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcX(E), ZENEcRing(E));
  ZENEltSquare(ZENEcZ(E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcZ(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcZ(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcZ(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcA4(E), ZENEcRing(E));
  ZENEltMultiply(ZENEcZ(E), ZENEcA1(E), ZENEcPtY(P, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcX(E), ZENEcZ(E), ZENEcRing(E));

  ZENEltMultiply(ZENEcY(E), ZENEcX(E), ZENEcY(E), ZENEcRing(E));
  
  ZENEltNegate(ZENEcPtY(R, E), ZENEcPtY(P, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcPtY(R, E), ZENEcA3(E), ZENEcRing(E));
  ZENEltMultiply(ZENEcX(E), ZENEcY(E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcPtY(R, E), ZENEcX(E), ZENEcRing(E));
  
  ZENEltAssign(ZENEcX(E), ZENEcA1(E), ZENEcRing(E));
  ZENEltAdd(ZENEcX(E), ZENEcY(E), ZENEcRing(E));
  ZENEltMultiply(ZENEcY(E), ZENEcX(E), ZENEcY(E), ZENEcRing(E));

  ZENEltNegate(ZENEcPtX(R, E), ZENEcPtX(P, E), ZENEcRing(E));
  ZENEltAdd(ZENEcPtX(R, E), ZENEcPtX(R, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcPtX(R, E), ZENEcA2(E), ZENEcRing(E));
  ZENEltAdd(ZENEcPtX(R, E), ZENEcY(E), ZENEcRing(E));

  ZENEltMultiply(ZENEcY(E), ZENEcX(E), ZENEcPtX(R, E), ZENEcRing(E));
  ZENEltSubtract(ZENEcPtY(R, E), ZENEcY(E), ZENEcRing(E));
}

/* \TEX{
   \title{General Opposite.}
   \input{Two allocated points |P| and |R| of an elliptic curve |E|.}
   \side{$|R| = -|P|$.} 
   }
   */
void _ZENEcPtNegateAllChar(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;
{
  ZENEcPtAssign(R, P, E);

  if (!ZENEcPtIsZero(R, E)) {
    ZENEltAdd(ZENEcPtY(R, E), ZENEcA3(E), ZENEcRing(E));
    ZENEltMultiply(ZENEcX(E), ZENEcA1(E), ZENEcPtX(P, E), ZENEcRing(E));
    ZENEltAdd(ZENEcPtY(R, E), ZENEcX(E), ZENEcRing(E));
    ZENEltNegate(ZENEcPtY(R, E), ZENEcPtY(R, E), ZENEcRing(E));
  }
}

/*
   \TEX{
   \title{Adding several points at the same time on elliptic curves.}
   \input{2 arrays of |nb| points to add on different curves.
   3 different additions are possible following |type|:
   \begin{itemize}
   \item $type[i]=|MEc_No|$ Nothing is done.
   \item $type[i]=|MEc_Add|$ |*p_P| and |*p_Q| are added.
   \item $type[i]=|MEc_Double|$ |*p_P| is doubled.
   \end{itemize}}
   \output{-1 if an error occured, |ZEN_NO_INVERSE| if a factor is
   found, |ZEN_HAS_INVERSE| otherwise.}
   \side{The result is put in |*p_R|.}
   } 
   */
int _ZENMEcPtAdd(p_R, p_P, p_Q, p_E, type, nb, Rg)
     ZENEcPt *p_R, *p_P, *p_Q;
     ZENEc *p_E;
     char *type;
     int nb;
     ZENRing Rg;
{
  ZENElt *inv, *v, w;
  char *mytype;
  int i, nbinv, index, ret;

  if (!nb) {
    return(ZEN_HAS_INVERSE);
  }
  
  if (!(mytype = (char *)ZENMalloc(sizeof(char)*nb))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTADD, ZEN_ALLOC);
    return(ZERR);
  }
  nbinv = 0;
  for(i = 0; i < nb; i++) {
    switch(type[i]) {
    case MEc_Add:
      if (ZENEcPtIsZero(p_P[i], p_E[i])) {
	ZENEcPtAssign(p_R[i], p_Q[i], p_E[i]);
	mytype[i] = MEc_No;
      }
      else if (ZENEcPtIsZero(p_Q[i], p_E[i])) {
	ZENEcPtAssign(p_R[i], p_P[i], p_E[i]);
	mytype[i] = MEc_No;
      }
      else if (ZENEltAreEqual(ZENEcPtX(p_P[i], p_E[i]),
			      ZENEcPtX(p_Q[i], p_E[i]),
			      Rg)) {
	if (ZENEltAreEqual(ZENEcPtY(p_P[i], p_E[i]),
			   ZENEcPtY(p_Q[i], p_E[i]),
			   Rg)) {
	  mytype[i] = MEc_Double;
	  nbinv++;
	}
	else {
	  ZENEcPtSetToZero(p_R[i], p_E[i]);
	  mytype[i] = MEc_No;
	}
      }
      else {
	mytype[i] = MEc_Add;
	nbinv++;
      }
      break;
    case MEc_Double:
      if (ZENEcPtIsZero(p_P[i], p_E[i])) {
	ZENEcPtSetToZero(p_P[i], p_E[i]);
	mytype[i] = MEc_No;
      }  
      else {
	mytype[i] = MEc_Double;
	nbinv++;
      }
      break;
    default:
      mytype[i] = MEc_No;
      break;
    }
  }
  if (nbinv) {
    if (ZENEltCreate(w, Rg)) {
      ZENFree(mytype);
      return(ZERR);
    }

    if (!(inv = (ZENElt *)ZENMalloc(sizeof(ZENElt)*nbinv))) {
      ZENFree(mytype);
      ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTADD, ZEN_ALLOC);
      return(ZERR);
    }
    
    if (!(v = (ZENElt *)ZENMalloc(sizeof(ZENElt)*nbinv))) {
      ZENFree(inv); ZENFree(mytype);
      ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTADD, ZEN_ALLOC);    
      return(ZERR); 
    }
    
    for(i = 0; i < nbinv; i++) {
      if (ZENEltCreate(inv[i], Rg)) {
	for(i--; i >= 0; i--) {
	  ZENEltFree(inv[i], Rg);
	  ZENEltFree(v[i], Rg);
	} ZENFree(inv); ZENFree(v); ZENFree(mytype);
	return(ZERR);
      }
      if (ZENEltCreate(v[i], Rg)) {
	ZENEltFree(inv[i], Rg);
	for(i--; i >= 0; i--) {
	  ZENEltFree(inv[i], Rg);
	  ZENEltFree(v[i], Rg);
	} ZENFree(inv); ZENFree(v); ZENFree(mytype);
	return(ZERR);
      } 
    }
    
    index = 0;
    for(i = 0; i < nb; i++) {
      switch(mytype[i]) {
      case MEc_Add:
	ZENEcPtPreAdd(p_P[i], p_Q[i], p_E[i]);
	ZENEltAssign(inv[index], ZENEcX(p_E[i]), Rg);
	index++;
	break;
      case MEc_Double:
	ZENEcPtPreDouble(p_P[i], p_E[i]);
	if (ZENEltIsZero(ZENEcY(p_E[i]), Rg)) {
	  ZENEcPtSetToZero(p_R[i], p_E[i]);
	  mytype[i] = MEc_No;
	}
	else {
	  ZENEltAssign(inv[index], ZENEcY(p_E[i]), Rg);
	  index++;
	}
	break;
      }
    }
    if (index) {
      ZENEltAssign(v[0], inv[0], Rg);
      for(i = 1; i < index; i++) {
	ZENEltMultiply(v[i], inv[i], v[i-1], Rg);
      }
      if ((ret = ZENEltInverse(w, v[index-1], Rg)) != ZEN_HAS_INVERSE) {
	for(i = 0; i < nbinv; i++) {
	  ZENEltFree(inv[i], Rg);
	  ZENEltFree(v[i], Rg);
	}
	ZENFree(v); ZENFree(inv);
	ZENEltFree(w, Rg);
	ZENFree(mytype);
	return(ret);
      }
      for(i = index-1; i > 0; i--) {
	ZENEltMultiply(v[i], v[i-1], w, Rg);
	ZENEltMultiply(w, inv[i], w, Rg);
      }
      ZENEltAssign(v[0], w, Rg);
    }

    index = 0;
    for(i = 0; i < nb; i++) {
      switch(mytype[i]) {
      case MEc_Add:
	ZENEltAssign(ZENEcX(p_E[i]), v[index], Rg);
	ZENEcPtPostAdd(p_R[i], p_P[i], p_Q[i], p_E[i]);
	ZENEcPtZ(p_R[i], p_E[i]) = 1;
	index++;
	break;
      case MEc_Double:
	ZENEltAssign(ZENEcY(p_E[i]), v[index], Rg);
	ZENEcPtPostDouble(p_R[i], p_P[i], p_E[i]);
	ZENEcPtZ(p_R[i], p_E[i]) = 1;
	index++;
	break;
      }
    }
    
    for(i = 0; i < nbinv; i++) {
      ZENEltFree(inv[i], Rg);
      ZENEltFree(v[i], Rg);
    }
    ZENFree(v); ZENFree(inv);
    ZENEltFree(w, Rg);
  }
  ZENFree(mytype);
  return(ZEN_HAS_INVERSE);
}

/*
   \TEX{
   \title{Multiplying several points by integers at the same time
   on elliptic curves.}
   \input{An array of points |p_P| and an array of BigNum (|p_k|,
   |p_kl|).} 
   \output{-1 if an error occured, |ZEN_NO_INVERSE| if a factor is
   found, |ZEN_HAS_INVERSE| otherwise.}
   \side{The products of |p_P| by the BigNums |p_k| are put in |p_R|.}
   } 
   */
int _ZENMEcPtMult(p_R, p_k, p_kl, p_P, p_E, nb, Rg)
     ZENEcPt *p_R, *p_P;
     BigNum *p_k;
     BigNumLength *p_kl;
     ZENEc *p_E;
     int nb;
     ZENRing Rg;
{
  BigNumDigit *p_C;
  ZENEcPt *p_Q;
  int *p_ks;
  int i, ksmax;
  BigNumLength klmax;
  char *status;
  int ret;

  if (!nb) return(ZEN_HAS_INVERSE);
  
  if (!(p_Q = (ZENEcPt *)ZENMalloc(sizeof(ZENEcPt)*nb))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTMULTIPLY, ZEN_ALLOC);
    return(ZERR);
  }
  if (!(p_C = (BigNumDigit *)ZENMalloc(sizeof(BigNumDigit)*nb))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTMULTIPLY, ZEN_ALLOC);
    ZENFree(p_Q);
    return(ZERR);
  }
  if (!(p_ks = (int *)ZENMalloc(sizeof(int)*nb))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTMULTIPLY, ZEN_ALLOC);
    ZENFree(p_Q); ZENFree(p_C);
    return(ZERR);
  }
  if (!(status = (char *)ZENMalloc(sizeof(char)*nb))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTMULTIPLY, ZEN_ALLOC);
    ZENFree(p_Q); ZENFree(p_C); ZENFree(p_ks);
    return(ZERR);
  }
  klmax = 0;
  for(i = 0; i < nb; i++) {
    if (!(p_Q[i] = ZENEcPtCreate(p_E[i]))) {
      for(i--; i >= 0; i--) { ZENEcPtFree(p_Q[i], p_E[i]);}
      ZENFree(p_Q); ZENFree(p_C); ZENFree(p_ks);
      return(ZERR);
    }
    ZENEcPtAssign(p_Q[i], p_P[i], p_E[i]);
    if (p_kl[i] > klmax) {
      klmax = p_kl[i];
    }
  }
  
  for(; klmax; klmax--) {

    ksmax = 0;
    for(i = 0; i < nb; i++) {
      if (p_kl[i] < klmax) {
	p_ks[i] = 0;
      }
      else if (p_kl[i] == klmax) {
	p_C[i] = p_k[i][klmax-1];
	for(p_ks[i] = BN_DIGIT_SIZE; p_ks[i]; p_ks[i]--){
	  if(ZBNShiftLeft(&(p_C[i]), 1, 1))
	    break;
	}
	p_ks[i]--;
	if (p_ks[i] > ksmax) ksmax = p_ks[i];
      }
      else {
	p_C[i] = p_k[i][klmax-1];
	ksmax = p_ks[i] = BN_DIGIT_SIZE;
      }
    }

    for(; ksmax; ksmax--){
      for(i = 0; i < nb; i++) {
	if (p_ks[i] < ksmax) {
	  status[i] = MEc_No;
	}
	else {
	  status[i] = MEc_Double;
	}
      }
      if ((ret = ZENMEcPtAdd(p_Q, p_Q, p_Q, p_E, status, nb, Rg)) != ZEN_HAS_INVERSE) {
	goto end;
      }
      for(i = 0; i < nb; i++) {
	if (status[i] == MEc_Double) {
	  if(ZBNShiftLeft(&(p_C[i]), 1, 1)){
	    status[i] = MEc_Add;
	  }
	  else {
	    status[i] = MEc_No;
	  }
	}
      }
      if ((ret = ZENMEcPtAdd(p_Q, p_Q, p_P, p_E, status, nb, Rg)) != ZEN_HAS_INVERSE) {
	goto end;
      }
    }
  }
  for(i = 0; i < nb; i++) {
    if (ZBNIsZero(p_k[i], p_kl[i])) {
      ZENEcPtSetToZero(p_R[i], p_E[i]);
    }
    else {
      ZENEcPtAssign(p_R[i], p_Q[i], p_E[i]);
    }
  }
  ret = ZEN_HAS_INVERSE;

end:
  for(i = 0; i < nb; i++) {ZENEcPtFree(p_Q[i], p_E[i]);}
  ZENFree(p_Q); ZENFree(p_C), ZENFree(p_ks); ZENFree(status);
  return(ret);
}
