/*
 * zen/zeg/c/ecpt_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Points on elliptic curve assigning}
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{Setting a point to zero}
   \input{An allocated point |P| of an elliptic curve |E|.}
   \side{|P| is set to the identity element of the curve.}
   }
   */
void _ZENEcPtSetToZero(P,E)
     ZENEcPt P;
     ZENEc E;
{
  ZENEcPtZ(P, E) = 0;
}

/* \TEX{
   \title{Assigning a point with another.}
   \input{Two allocated points |P| and |Q| of an elliptic curve |E|.}
   \side{|P| is filled with |Q|.}
   }
   */
void _ZENEcPtAssign(R, P, E)
     ZENEcPt R;
     ZENEcPt P;
     ZENEc E;
{
  if ((ZENEcPtZ(R, E) = ZENEcPtZ(P, E))) {
    ZENEltAssign(ZENEcPtX(R, E), ZENEcPtX(P, E), ZENEcRing(E));
    ZENEltAssign(ZENEcPtY(R, E), ZENEcPtY(P, E), ZENEcRing(E));
  }
}

/* \TEX{
   \title{Enumerating points on a curve.}
   \input{Two points |P| and |Q| of a curve |E|.}
   \output{|ZENERR| if an error occurred, |ZEN_HAS_INVERSE| otherwise.}
   \side{|P| is assigned with a point on the curve |E| whose abcissa
   is obtained from the abcissa of |Q| by |ZENEltSetNext|.}
   \note{Valid only in finite fields. This procedure loops forever if
   there is no point on the curve !}
   }
   */
int _ZENEcPtSetNext(P, Q, E)
     ZENEcPt P;
     ZENEcPt Q;
     ZENEc E;
{
  ZENPoly Pol, roots;
  ZENElt t, u, elt;
  int ret;

  if (!(Pol = ZENPolyCreate(2, ZENEcRing(E)))) {
    return(ZERR);
  }
  
  if (!(roots = ZENPolyCreate(1, ZENEcRing(E)))) {
    ZENPolyFree(Pol, ZENEcRing(E));
    return(ZERR);
  }

  ZENEltAssign(ZENEcPtX(P, E), ZENEcPtX(Q, E), ZENEcRing(E));  
  for(;;) {
    ZENEltSetNext(ZENEcPtX(P, E), ZENEcRing(E));
    
    t = ZENPolyGetCoeffPtr(Pol, 2, ZENEcRing(E));
    u = ZENPolyGetCoeffPtr(Pol, 1, ZENEcRing(E));
    elt = ZENPolyGetCoeffPtr(Pol, 0, ZENEcRing(E));
    ZENEltAssign(elt, ZENEcA6(E), ZENEcRing(E));
    ZENEltMultiply(t, ZENEcA4(E), ZENEcPtX(P, E), ZENEcRing(E));
    ZENEltAdd(elt, t, ZENEcRing(E));
    ZENEltSquare(u, ZENEcPtX(P, E), ZENEcRing(E));
    ZENEltMultiply(t, u, ZENEcA2(E), ZENEcRing(E));
    ZENEltAdd(elt, t, ZENEcRing(E));
    ZENEltMultiply(t, ZENEcPtX(P, E), u, ZENEcRing(E));
    ZENEltAdd(elt, t, ZENEcRing(E));
    ZENEltNegate(elt, elt, ZENEcRing(E));
    ZENPolySetCoeff(Pol, 0, elt, ZENEcRing(E));

    elt = ZENPolyGetCoeffPtr(Pol, 1, ZENEcRing(E));
    ZENEltMultiply(elt, ZENEcA1(E), ZENEcPtX(P, E), ZENEcRing(E));
    ZENEltAdd(elt, ZENEcA3(E), ZENEcRing(E));
    ZENPolySetCoeff(Pol, 1, elt, ZENEcRing(E));

    elt = ZENPolyGetCoeffPtr(Pol, 2, ZENEcRing(E));
    ZENEltSetToOne(elt, ZENEcRing(E));
    ZENPolySetCoeff(Pol, 2, elt, ZENEcRing(E));

    ZENPolyDeg(Pol, ZENEcRing(E)) = 2;

    if ((ret = ZENPolyRootsDegree2(roots, Pol, ZENEcRing(E))) != ZEN_HAS_INVERSE) {
      ZENPolyFree(Pol, ZENEcRing(E));
      ZENPolyFree(roots, ZENEcRing(E));
      return(ret);
    }

    if (ZENPolyDeg(roots, ZENEcRing(E)) >= 0) {
      break;
    }
  }
  ZENEltAssign(ZENEcPtY(P, E), ZENPolyGetCoeffPtr(roots, 0, ZENEcRing(E)), ZENEcRing(E));
  ZENEcPtZ(P, E) = 1;
  ZENPolyFree(Pol, ZENEcRing(E));
  ZENPolyFree(roots, ZENEcRing(E));
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Enumerating points on a canonical curve.}
   \input{Two points |P| and |Q| of a curve |E|.}
   \output{|ZENERR| if an error occurred, |ZEN_HAS_INVERSE| otherwise.}
   \side{|P| is assigned with a point on the curve |E| whose abcissa
   is obtained from the abcissa of |Q| by |ZENEltSetNext|.}
   \note{Valid only in finite fields. This procedure loops forever if
   there is no point on the curve !}
   }
   */
int _ZENEcPtSetNextChar23p(P, Q, E)
     ZENEcPt P, Q;
     ZENEc E;
{
  ZENPoly Pol, roots;
  ZENElt t, u, elt;
  int ret;

  if (!(Pol = ZENPolyCreate(2, ZENEcRing(E)))) {
    return(ZERR);
  }
  
  if (!(roots = ZENPolyCreate(1, ZENEcRing(E)))) {
    ZENPolyFree(Pol, ZENEcRing(E));
    return(ZERR);
  }

  ZENEltAssign(ZENEcPtX(P, E), ZENEcPtX(Q, E), ZENEcRing(E));  
  for(;;) {
    ZENEltSetNext(ZENEcPtX(P, E), ZENEcRing(E));
    
    t = ZENPolyGetCoeffPtr(Pol, 2, ZENEcRing(E));
    u = ZENPolyGetCoeffPtr(Pol, 1, ZENEcRing(E));
    elt = ZENPolyGetCoeffPtr(Pol, 0, ZENEcRing(E));
    ZENEltAssign(elt, ZENEcB6(E), ZENEcRing(E));
    ZENEltMultiply(t, ZENEcB4(E), ZENEcPtX(P, E), ZENEcRing(E));
    ZENEltAdd(elt, t, ZENEcRing(E));
    ZENEltSquare(u, ZENEcPtX(P, E), ZENEcRing(E));
    ZENEltMultiply(t, u, ZENEcB2(E), ZENEcRing(E));
    ZENEltAdd(elt, t, ZENEcRing(E));
    ZENEltMultiply(t, ZENEcPtX(P, E), u, ZENEcRing(E));
    ZENEltAdd(elt, t, ZENEcRing(E));
    ZENEltNegate(elt, elt, ZENEcRing(E));
    ZENPolySetCoeff(Pol, 0, elt, ZENEcRing(E));

    elt = ZENPolyGetCoeffPtr(Pol, 1, ZENEcRing(E));
    ZENEltMultiply(elt, ZENEcB1(E), ZENEcPtX(P, E), ZENEcRing(E));
    ZENEltAdd(elt, ZENEcB3(E), ZENEcRing(E));
    ZENPolySetCoeff(Pol, 1, elt, ZENEcRing(E));

    elt = ZENPolyGetCoeffPtr(Pol, 2, ZENEcRing(E));
    ZENEltSetToOne(elt, ZENEcRing(E));
    ZENPolySetCoeff(Pol, 2, elt, ZENEcRing(E));

    ZENPolyDeg(Pol, ZENEcRing(E)) = 2;

    if ((ret = ZENPolyRootsDegree2(roots, Pol, ZENEcRing(E))) != ZEN_HAS_INVERSE) {
      ZENPolyFree(Pol, ZENEcRing(E));
      ZENPolyFree(roots, ZENEcRing(E));
      return(ret);
    }

    if (ZENPolyDeg(roots, ZENEcRing(E)) >= 0) {
      break;
    }
  }
  ZENEltAssign(ZENEcPtY(P, E), ZENPolyGetCoeffPtr(roots, 0, ZENEcRing(E)), ZENEcRing(E));
  ZENEcPtZ(P, E) = 1;
  ZENPolyFree(Pol, ZENEcRing(E));
  ZENPolyFree(roots, ZENEcRing(E));
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Getting the abcissa.}
   \input{An allocated point |P| of an elliptic curve |E| and an
   element |a|.}
   \side{|a| is filled with the abscissa of |P|.}
   }
   */
void _ZENEcPtGetX(a, P, E)
     ZENElt *a;
     ZENEcPt P;
     ZENEc E;
{
  ZENEltAssign(*a, ZENEcPtX(P, E), ZENEcRing(E));
}

/* \TEX{
   \title{Getting the abcissa from a canonical point.}
   \input{An allocated point |P| of an elliptic curve |E| and an
   element |a|.}
   \side{|a| is filled with the abscissa of |P|.}
   }
   */
void _ZENEcPtGetXChar23p(a, P, E)
     ZENElt *a;
     ZENEcPt P;
     ZENEc E;
{
  ZENEcPtCanonical2General(ZENEcXY(E), P, E);
  ZENEltAssign(*a, ZENEcPtX(ZENEcXY(E), E), ZENEcRing(E));
}

/* \TEX{
   \title{Getting the ordinate.}
   \input{An allocated point |P| of an elliptic curve |E| and an
   element |a|.}
   \side{|a| is filled with the abscissa of |P|.}
   }
   */
void _ZENEcPtGetY(a, P, E)
     ZENElt *a;
     ZENEcPt P;
     ZENEc E;
{
  ZENEltAssign(*a, ZENEcPtY(P, E), ZENEcRing(E));
}

/* \TEX{
   \title{Getting the ordinate from a canonical point.}
   \input{An allocated point |P| of an elliptic curve |E| and an
   element |a|.}
   \side{|a| is filled with the ordinate of |P|.}
   }
   */
void _ZENEcPtGetYChar23p(a, P, E)
     ZENElt *a;
     ZENEcPt P;
     ZENEc E;
{
  ZENEcPtCanonical2General(ZENEcXY(E), P, E);
  ZENEltAssign(*a, ZENEcPtY(ZENEcXY(E), E), ZENEcRing(E));
}

/* \TEX{
   \title{Setting a point.}
   \input{An allocated point |P| of an elliptic curve |E| and couple
   (|x|, |y|).}
   \output{The point |P| is filled with (|x|, |y|).}
   }
   */
void _ZENEcPtSetXY(P,x,y,E)
     ZENEcPt P;
     ZENElt x,y;
     ZENEc E;
{
  ZENEltAssign(ZENEcPtX(P, E), x, ZENEcRing(E));
  ZENEltAssign(ZENEcPtY(P, E), y, ZENEcRing(E));
  ZENEcPtZ(P, E) = 1;
}

/* \TEX{
   \title{Setting a point on a canonical curve.}
   \input{An allocated point |P| of an elliptic curve |E| and an
   element |a|.}
   \output{The point |P| is filled with (|x|, |y|).}
   }
   */
void _ZENEcPtSetXYChar23p(P,x,y,E)
     ZENEcPt P;
     ZENElt x,y;
     ZENEc E;
{
  ZENEltAssign(ZENEcPtX(ZENEcXY(E), E), x, ZENEcRing(E));
  ZENEltAssign(ZENEcPtY(ZENEcXY(E), E), y, ZENEcRing(E));
  ZENEcPtZ(ZENEcXY(E), E) = 1;
  ZENEcPtGeneral2Canonical(P, ZENEcXY(E), E);
}

/* \TEX{
   \title{Computing the image of an general point.}
   \input{Two allocated points |Pg| and |Pc| of an elliptic
   curve |E| and its canonical image.}
   \side{The image of |Pc| is assigned in |Pg| by the isomorphism.}
   }
   */
void _ZENEcPtGeneral2Canonical(Pg, Pc, E)
     ZENEcPt Pg, Pc;
     ZENEc E;
{
  if (ZENEcPtIsZero(Pc, E)) {
    ZENEcPtSetToZero(Pg, E);
  }
  else {
    /* y = Ci*X+Di*Y+Ei */
    ZENEltMultiply(ZENEcPtY(Pg, E), ZENEcDi(E), ZENEcPtY(Pc, E), ZENEcRing(E));
    ZENEltMultiply(ZENEcZ(E), ZENEcCi(E), ZENEcPtX(Pc, E), ZENEcRing(E));
    ZENEltAdd(ZENEcPtY(Pg, E),  ZENEcZ(E), ZENEcRing(E));
    ZENEltAdd(ZENEcPtY(Pg, E),  ZENEcEi(E), ZENEcRing(E));
    
    /* x = Ai*X+Bi */
    ZENEltMultiply(ZENEcPtX(Pg, E), ZENEcAi(E), ZENEcPtX(Pc, E), ZENEcRing(E));
    ZENEltAdd(ZENEcPtX(Pg, E),  ZENEcBi(E), ZENEcRing(E));

    /* z = 1 */
    ZENEcPtZ(Pg, E) = 1;
  }
}

/* \TEX{
   \title{Computing the image of an canonical point.}
   \input{Two allocated points |Pg| and |Pc| of an elliptic
   curve |E| and its canonical image.}
   \side{The image of |Pg| is assigned in |Pg| by the inverse isomorphism.}
   }
   */
void _ZENEcPtCanonical2General(Pc, Pg, E)
     ZENEcPt Pc, Pg;
     ZENEc E;
{
  if (ZENEcPtIsZero(Pg, E)) {
    ZENEcPtSetToZero(Pc, E);
  }
  else {
    /* Y = Cd*x+Dd*y+Ed */
    ZENEltMultiply(ZENEcPtY(Pc, E), ZENEcDd(E), ZENEcPtY(Pg, E), ZENEcRing(E));
    ZENEltMultiply(ZENEcZ(E), ZENEcCd(E), ZENEcPtX(Pg, E), ZENEcRing(E));
    ZENEltAdd(ZENEcPtY(Pc, E),  ZENEcZ(E), ZENEcRing(E));
    ZENEltAdd(ZENEcPtY(Pc, E),  ZENEcEd(E), ZENEcRing(E));
    
    /* X = Ad*x+Bd */
    ZENEltMultiply(ZENEcPtX(Pc, E), ZENEcAd(E), ZENEcPtX(Pg, E), ZENEcRing(E));
    ZENEltAdd(ZENEcPtX(Pc, E),  ZENEcBd(E), ZENEcRing(E));

    /* z = 1 */
    ZENEcPtZ(Pc, E) = 1;
  }
}
