/*
 * zen/zeg/c/ecpt_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Elliptic curve exponentiation.}
   }
   */
#include <stdio.h>
#include "zen.h"

/* For debugging */
#define ZENECPTDEBUGLEVEL 0

#if ZENECPTDEBUGLEVEL >= 25
#define MyZENEcPtPrintToFile(file,P,base,E)		\
(ZENEcPtZ(P, E) ? (fprintf(file, "[ "),					\
ZENEltPrintToFile(file, ZENEcPtX(P, E), base, ZENEcRing(E)),	\
fprintf(file, " , "),					\
ZENEltPrintToFile(file,ZENEcPtY(P, E), base, ZENEcRing(E)),	\
fprintf(file, " , %d ]", ZENEcPtZ(P, E))):fprintf(file, "[0, 1, 0]"));

static void ZBZPrintToFile(file, z)
     FILE *file;
     BigZ z;
{
  char *s;

  s = ZBZToString(z, 10);
  fprintf(file, "%s", s);
  ZENFree(s);
}
#endif

/* \TEX{
   \title{Multiplication with the binary method.}
   \input{Two allocated points |P| and |R| of an elliptic curve |E|
   and a lare integer |k| of size |kl|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|R| = |k||P|$ if |ZEN_HAS_INVERSE| is returned, |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   }
   */
static int _ZENEcPtMultPlain(R, k, kl, P, E)
     ZENEcPt R;
     BigNum k;
     BigNumLength kl;
     ZENEcPt P;
     ZENEc E;
{
  BigNumDigit C[1];
  ZENEcPt Q;
  int i, kd, ks, ret;

  /* Starting the computation */
  for(kd = kl-1; kd >= 0; kd--) {
    ZBNAssign(C, (k+kd), 1);
    
    /* Looking for the most significant bit */
    for(ks = BN_DIGIT_SIZE; ks; ks--){
      if(ZBNShiftLeft(C, 1, 1))
	break;
    }
    if(ks) {
      break;
    } 
  }
  if (kd < 0) {
    ZENEcPtSetToZero(R, E);
    return(ZEN_HAS_INVERSE);
  }
    
  if (!(Q = ZENEcPtCreate(E))) {
    return(ZERR);
  }
  ZENEcPtAssign(Q, P, E);

  ks--;
  while(kl >= 1){
    for(i = 0; i < ks; i++) {
      
      if ((ret = ZENEcPtDouble(Q, Q, E)) != ZEN_HAS_INVERSE) {
	ZENEcPtFree(Q, E);
	return(ret);
      }
      
      if(ZBNShiftLeft(C, 1, 1)){
	if ((ret = ZENEcPtAdd(Q, P, Q, E)) != ZEN_HAS_INVERSE) {
	  ZENEcPtFree(Q, E);
	  return(ret);
	}
      }
    }
    kl--;
    if(kl){
      kd--; ks = BN_DIGIT_SIZE;
      ZBNAssign(C, (k+kd), 1);
    }
  }

  ZENEcPtAssign(R, Q, E);
  ZENEcPtFree(Q, E);
  return(ZEN_HAS_INVERSE);
}

#if 1
/* \TEX{
   \title{Multiplication with substraction-addition chains.}
   \input{Two allocated points |P| and |R| of an elliptic curve |E|
   and a lare integer |k| of size |kl|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|R| = |k||P|$ if |ZEN_HAS_INVERSE| is returned, |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   }
   */
static int _ZENEcPtMultSigned(R, k, kl, P, E)
     ZENEcPt R, P;
     ZENEc E;
     BigNum k;
     BigNumLength kl;
{
  BigNumDigit C[1];
  int b;
  int ks, kd, ret;
  char simple_done, double_done;
  ZENEcPt QBuf;

#if ZENECPTDEBUGLEVEL >= 25
  fprintf(stderr, "# -> ZENEcPtMultSigned(");
  ZBNPrintToFile(stderr, k, kl, 10);
  fprintf(stderr, ")\n");
#endif

  if (ZBNIsZero(k, kl)) {
    ZENEcPtSetToZero(R, E);
    return(ZEN_HAS_INVERSE);
  }

  if (!(QBuf = ZENEcPtCreate(E))) {
    return(ZERR);
  }

  /* Initialization */
  b = simple_done = double_done = 0;

  /* For each digit */
  for(kd = 0; kd < kl; kd++) {
    C[0] = ZBNGetDigit(k+kd);

    /* For each bit in this digit */
    /* 2 cases; an usual digit and a header digit */
    for((kd == kl-1) ? (ks = ZBNNumBitsInDigit(C[0])) : (ks = BN_DIGIT_SIZE);
	ks;
	ks--) {
      
      if (ZBNShiftRight(C, 1, 1) == 0) {

	/* This bit is 0 */
	switch(b) {
	case 2:
#if ZENECPTDEBUGLEVEL >= 45
	  fprintf(stderr, "#O ");
#endif
	  if ((kd == kl-1) && (ks == 2)) {
#if ZENECPTDEBUGLEVEL >= 45
	    fprintf(stderr, "I\n");
#endif
	    if ((ret = ZENEcPtAdd(R, QBuf, R, E))) {
	      ZENEcPtFree(QBuf, E);
	      return(ret);
	    }
	    if ((ret = ZENEcPtDouble(QBuf, QBuf, E))) {
	      ZENEcPtFree(QBuf, E);
	      return(ret);
	    }
	    b = 0;
	  }
	  else
	    b = 1;
	  break;
	case 1:
#if ZENECPTDEBUGLEVEL >= 45
	    fprintf(stderr, "I\n");
#endif
	  if (simple_done) {
	    if ((ret = ZENEcPtAdd(R, QBuf, R, E))) {
	      ZENEcPtFree(QBuf, E);
	      return(ret);
	    }
	  }
	  else {
	    if (double_done)
	      ZENEcPtAssign(R, QBuf, E);
	    else
	      ZENEcPtAssign(R, P, E);
	    simple_done = 1;
	  }

	  if (double_done) {
	    if ((ret = ZENEcPtDouble(QBuf, QBuf, E)))  {
	      ZENEcPtFree(QBuf, E);
	      return(ret);
	    }
	  }
	  else {
	    if ((ret = ZENEcPtDouble(QBuf, P, E)))  {
	      ZENEcPtFree(QBuf, E);
	      return(ret);
	    }
	    double_done = 1;
	  }
	  b = 0;
	case 0:
#if ZENECPTDEBUGLEVEL >= 45
	  fprintf(stderr, "#O O\n");
#endif
	  if (double_done) {
	    if ((ret = ZENEcPtDouble(QBuf, QBuf, E)))  {
	      ZENEcPtFree(QBuf, E);
	      return(ret);
	    }
	  }
	  else {
	    if ((ret = ZENEcPtDouble(QBuf, P, E)))  {
	      ZENEcPtFree(QBuf, E);
	      return(ret);
	    }
	    double_done = 1;
	  }
	}
      }
      else {
	
	/* The bit is 1 */
	switch(b) {
	case 0:
#if ZENECPTDEBUGLEVEL >= 45
	  fprintf(stderr, "#I ");
#endif
	  if ((kd == kl-1) && (ks == 2)) {
#if ZENECPTDEBUGLEVEL >= 45
	    fprintf(stderr, "I\n");
#endif
	    if (simple_done) {
	      if ((ret = ZENEcPtAdd(R, QBuf, R, E)))  {
		ZENEcPtFree(QBuf, E);
		return(ret);
	      }
	      if ((ret = ZENEcPtDouble(QBuf, QBuf, E)))  {
		ZENEcPtFree(QBuf, E);
		return(ret);
	      }
	    }
	    else {
	      if (double_done) {
		ZENEcPtAssign(R, QBuf, E);
		if ((ret = ZENEcPtDouble(QBuf, QBuf, E)))  {
		  ZENEcPtFree(QBuf, E);
		  return(ret);
		}
	      }
	      else {
		ZENEcPtAssign(R, P, E);
		if ((ret = ZENEcPtDouble(QBuf, P, E)))  {
		  ZENEcPtFree(QBuf, E);
		  return(ret);
		}
	      }
	      simple_done = double_done = 1;
	    }
	  }
	  else
	    b = 1;
	  break;
	case 1:
#if ZENECPTDEBUGLEVEL >= 45
	  fprintf(stderr, "T\n");
#endif
	  if (simple_done) {
	    if ((ret = ZENEcPtSubtract(R, QBuf, R, E))) {
	      ZENEcPtFree(QBuf, E);
	      return(ret);
	    }
	    ZENEcPtNegate(R, R, E);
	    if ((ret = ZENEcPtDouble(QBuf, QBuf, E))) {
	      ZENEcPtFree(QBuf, E);
	      return(ret);
	    }
	  }
	  else {
	    if (double_done) {
	      ZENEcPtNegate(R, QBuf, E);
	      if ((ret = ZENEcPtDouble(QBuf, QBuf, E))) {
		ZENEcPtFree(QBuf, E);
	      return(ret);
	      }
	    }
	    else {
	      if ((ret = ZENEcPtDouble(QBuf, P, E))) {
		ZENEcPtFree(QBuf, E);
		return(ret);
	      }
	      double_done = 1;
	      ZENEcPtNegate(R, P, E);
	    }
	    simple_done = 1;
	  }

	  b = 2;
	case 2:
#if ZENECPTDEBUGLEVEL >= 45
	  fprintf(stderr, "#I 0\n");
#endif
	  if ((ret = ZENEcPtDouble(QBuf, QBuf, E)))  {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	}
      }
    }
  }
  if (double_done) {
    if (simple_done) {
      if ((ret = ZENEcPtAdd(R, QBuf, R, E)))  {
	ZENEcPtFree(QBuf, E);
	return(ret);
      }
    }
    else
      ZENEcPtAssign(R, QBuf, E);
  }
  else
    ZENEcPtAssign(R, P, E);
  
#if ZENECPTDEBUGLEVEL >= 45
  if (b == 1)
    fprintf(stderr, "I\n");
  else
    fprintf(stderr, "#O I\n");
#endif
#if ZENECPTDEBUGLEVEL >= 25
  fprintf(stderr, "# <- ZENEcPtMultSigned()\n");
#endif
  ZENEcPtFree(QBuf, E);
  return(ZEN_HAS_INVERSE);
}
#else
static int _ZENEcPtMultSigned(R, k, kl, P, E)
     ZENEcPt R, P;
     ZENEc E;
     BigNum k;
     BigNumLength kl;
{
  BigNumDigit C[1];
  int b;
  int ks, kd, ret;
  ZENEcPt QBuf;

  if (ZBNIsZero(k, kl)) {
    ZENEcPtSetToZero(R, E);
    return(ZEN_HAS_INVERSE);
  }

  if (!(QBuf = ZENEcPtCreate(E))) {
    return(ZERR);
  }

  /* Initialization */
  b = 0;
  ZENEcPtAssign(QBuf, P, E);
  ZENEcPtSetToZero(R, E);

  /* For each digit */
  for(kd = kl-1; kd >= 0; kd--) {
    C[0] = ZBNGetDigit(k+kd);

    /* For each bit in this digit */
    /* 2 cases; an usual digit and a header digit */
    if (kd == kl-1) {
      ks = ZBNNumBitsInDigit(C[0]);
      ZBNShiftLeft(C, 1, BN_DIGIT_SIZE-ks);
    }
    else {
      ks = BN_DIGIT_SIZE;
    }

    for(; ks; ks--) {
      
      if (ZBNShiftLeft(C, 1, 1) == 0) {

	/* This bit is 0 */
	switch(b) {
	case 3:
	  ZENEcPtNegate(QBuf, QBuf, E);
	  if ((ret = ZENEcPtAdd(R, QBuf, R, E))) {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  ZENEcPtNegate(QBuf, QBuf, E);
	  if ((ret = ZENEcPtDouble(R, R, E)))  {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  if ((ret = ZENEcPtDouble(R, R, E)))  {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  b = 0;
	  break;
	case 2:
	  b = 3;
	  break;
	case 1:
	  if ((ret = ZENEcPtDouble(R, R, E)))  {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  if ((ret = ZENEcPtAdd(R, QBuf, R, E))) {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  if ((ret = ZENEcPtDouble(R, R, E)))  {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  b = 0;
	  break;
	case 0:
	  if ((ret = ZENEcPtDouble(R, R, E)))  {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	}
      }
      else {
	
	/* The bit is 1 */
	switch(b) {
	case 0:
	  b = 1;
	  break;
	case 1:	  
	  if ((ret = ZENEcPtAdd(R, QBuf, R, E))) {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  if ((ret = ZENEcPtDouble(R, R, E))) {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  if ((ret = ZENEcPtDouble(R, R, E))) {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  b = 2;
	  break;
	case 2:
	  if ((ret = ZENEcPtDouble(R, R, E)))  {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  break;
	case 3:
	  if ((ret = ZENEcPtDouble(R, R, E))) {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  ZENEcPtNegate(QBuf, QBuf, E);
	  if ((ret = ZENEcPtAdd(R, QBuf, R, E))) {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  ZENEcPtNegate(QBuf, QBuf, E);
	  if ((ret = ZENEcPtDouble(R, R, E))) {
	    ZENEcPtFree(QBuf, E);
	    return(ret);
	  }
	  b = 2;
	  break;
	}
      }
    }
  }

  switch(b) {
  case 3:
    ZENEcPtNegate(QBuf, QBuf, E);
    if ((ret = ZENEcPtAdd(R, QBuf, R, E))) {
      ZENEcPtFree(QBuf, E);
      return(ret);
    }
    if ((ret = ZENEcPtDouble(R, R, E)))  {
      ZENEcPtFree(QBuf, E);
      return(ret);
    }
    break;
  case 2:
    ZENEcPtNegate(QBuf, QBuf, E);
    if ((ret = ZENEcPtAdd(R, QBuf, R, E)))  {
      ZENEcPtFree(QBuf, E);
      return(ret);
    }
    break;
  case 1:
    if ((ret = ZENEcPtDouble(R, R, E)))  {
      ZENEcPtFree(QBuf, E);
      return(ret);
    }
    if ((ret = ZENEcPtAdd(R, QBuf, R, E)))  {
      ZENEcPtFree(QBuf, E);
      return(ret);
    }
    break;
  }

  ZENEcPtFree(QBuf, E);
  return(ZEN_HAS_INVERSE);
}
#endif

static int ZBNWindowSignedTransform(n, nl, d, T, it)
     BigNum n;
     int nl, d;
     int *T;
     int *it;
{
  BigNumDigit C[1];
  int nbits, nword;
  int kl, M, J, X, U, Y, V, W, Z;
  char *B;

  ZBNAssign(C, (n+nl-1), 1);

  /* looking for the most significant bit */
  for(nbits = BN_DIGIT_SIZE; nbits; nbits--)
    if(ZBNShiftLeft(C, 1, 1))
      break;

  if (!(B = (char *)ZENMalloc(sizeof(char)*(1+nl*BN_DIGIT_SIZE)))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTCREATE, ZEN_ALLOC);
    return(ZERR);
  }
  M = J = Y = X = U = V = W = Z = 0;
  for(nword = 0; nword < nl ; nword++) {
    for(kl = (nword == nl-1) ? (nbits-1):(BN_DIGIT_SIZE); kl > 0; kl--) {

      B[X] = ZBNShiftRight(n+nword, 1, 1)?1:0;
      if (B[X]) Y++; else Y--;
      X++;

      
      if (!M) {
	if ((Y-Z) >= d) {
	  while (J < W) {
	    T[J] = B[J]; J++;
	  }
	  T[J++] = -1; V = Y; U = X; M = 1;
	}
	else
	  if (Y < Z) {
	    Z = Y; W = X;
	  }
      }
      else {
	if ((V-Y) >= d) {
	  while (J < U) {
	    T[J] = B[J] - 1; J++;
	  }
	  T[J++] = 1; Z = Y; W = X; M = 0;
	}
	else
	  if (Y > V) {
	    V = Y; U = X;
	  }
      }	
    }
  }

  if ((!M) || ((M == 1) && (V <= Y))) {
    while (J < X) {
      T[J] = B[J] - M; J++;
    }
    T[J++] = 1 - M; 
    if (M)
      T[J++] = M;
  }
  else {
    while (J < U) {
      T[J] = B[J] - 1; J++; 
    }
    T[J++] = 1;
    while (J < X) {
      T[J] = B[J]; J++; 
    }
    T[J++] = 1;
  }

  *it = J;
#if ZENECPTDEBUGLEVEL >= 40
  fprintf(stderr, "#B = 1");
  for(kl = X-1; kl >=0; kl--)
    fprintf(stderr, "%d", B[kl]);
  fprintf(stderr, "\n");
#endif

  ZENFree(B);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Multiplication with the $m$-window substraction-addition chains.}
   \input{Two allocated points |P| and |R| of an elliptic curve |E|
   and a lare integer |k| of size |kl|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|R| = |k||P|$ if |ZEN_HAS_INVERSE| is returned, |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   }
   */
static int _ZENEcPtMultWindowSigned(R, m, d, k, kl, P, E)
     ZENEcPt R, P;
     ZENEc E;
     BigNum k;
     BigNumLength kl;
     int m, d;
{
  ZENEcPt *PrePoint, QBuf;
  BigNum n;
  int di, ui, nl, it, i, ret;
  int twopowerminusone;
  int *T;
  char b, started;
#if ZENECPTDEBUGLEVEL >= 25
  BigZ p, r, *pre;

  fprintf(stderr, "\n# -> MecmMultWindowSigned(");
  ZBNPrintToFile(stderr, k, kl, 10);
  fprintf(stderr, ")\n");
#endif

  if (ZBNIsZero(k, kl)) {
    ZENEcPtSetToZero(R, E);
    return(ZEN_HAS_INVERSE);
  }

  /* Initialization */
  if (!(QBuf = ZENEcPtCreate(E))) {
    return(ZERR);
  }
  twopowerminusone = 1 << (m-1);
  if ((m > 1) && ((ret = ZENEcPtDouble(QBuf, P, E)))) {
    ZENEcPtFree(QBuf, E);
    return(ret);
  }
  
  if (!(PrePoint = (ZENEcPt *)ZENMalloc(twopowerminusone*sizeof(ZENEcPt)))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTCREATE, ZEN_ALLOC);
    ZENEcPtFree(QBuf, E);
    return(ZERR);
  }
  if (!(PrePoint[0] = ZENEcPtCreate(E))) {
    ZENEcPtFree(QBuf, E); ZENFree(PrePoint);
    return(ZERR);
  }
  ZENEcPtAssign(PrePoint[0], P, E);

  for(i = 1; i < twopowerminusone; i++) {
    if (!(PrePoint[i] = ZENEcPtCreate(E))) {
      ZENEcPtFree(QBuf, E);
      for(i--; i >= 0; i--) {
	ZENEcPtFree(PrePoint[i], E);
      }
      ZENFree(PrePoint);
      return(ret);
    }
    if ((ret = ZENEcPtAdd(PrePoint[i], QBuf, PrePoint[i-1], E))) {
      ZENEcPtFree(QBuf, E);
      for(; i >= 0; i--) {
	ZENEcPtFree(PrePoint[i], E);
      }
      ZENFree(PrePoint);
      return(ret);
    }
  }

#if ZENECPTDEBUGLEVEL >= 25
  /* Precomputation */
  p = ZBZFromString("1", 10);
  pre = (BigZ *)ZENMalloc(sizeof(BigZ)*twopowerminusone);
#  if ZENECPTDEBUGLEVEL >= 30
  fprintf(stderr, "P := ");
  MyZENEcPtPrintToFile(stderr, PrePoint[0], 10, E);
  fprintf(stderr, ";\n");
#  endif
  if (m > 1) {
    r = ZBZAdd(p, p); 
#  if ZENECPTDEBUGLEVEL >= 30
    fprintf(stderr, "EcmMult(");
    ZBZPrintToFile(stderr, r);
    fprintf(stderr, ", P, Cv, N) - ");
    MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
    fprintf(stderr, ";\n");
#  endif
  }
  pre[0] = ZBZFromString("1", 10);
  for(i = 1; i < twopowerminusone; i++) {
    pre[i] = ZBZAdd(pre[i-1], r);
#  if ZENECPTDEBUGLEVEL >= 30
    fprintf(stderr, "EcmMult(");
    ZBZPrintToFile(stderr, pre[i]);
    fprintf(stderr, ", P, Cv, N) - ");
    MyZENEcPtPrintToFile(stderr, PrePoint[i], 10, E);
    fprintf(stderr, ";\n");
#  endif
  }
  ZBZFree(r);
#endif

  /* Automaton transformation */
  nl = ZBNNumDigits(k, kl); 
  if (!(n = ZBNC(nl))) {
    ZENEcPtFree(QBuf, E);
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTCREATE, ZEN_ALLOC);
    for(i = 0; i < twopowerminusone; i++) {
      ZENEcPtFree(PrePoint[i], E);
    }
    ZENFree(PrePoint);
    
#if ZENECPTDEBUGLEVEL >= 25
    for(i = 0; i < twopowerminusone; i++)
      ZBZFree(pre[i]);
    ZENFree(pre);
    ZBZFree(r); ZBZFree(p);
#endif
    return(ZERR);
  }
  ZBNAssign(n, k, nl);
  if (!(T = (int *)ZENMalloc(sizeof(int)*(1+nl*BN_DIGIT_SIZE)))) {
    ZENEcPtFree(QBuf, E);
    ZENSetError(ZEN_LIB_ZEG, ZEN_ECPTCREATE, ZEN_ALLOC);
    ZBNF(n);
    for(i = 0; i < twopowerminusone; i++) {
      ZENEcPtFree(PrePoint[i], E);
    }
    ZENFree(PrePoint);
    
#if ZENECPTDEBUGLEVEL >= 25
    for(i = 0; i < twopowerminusone; i++)
      ZBZFree(pre[i]);
    ZENFree(pre);
    ZBZFree(r); ZBZFree(p);
#endif
    
    return(ZERR);
  }
  if (ZBNWindowSignedTransform(n, nl, d, T, &it)) {
    ZENEcPtFree(QBuf, E);
    ZENFree(T); ZBNF(n);
    for(i = 0; i < twopowerminusone; i++) {
      ZENEcPtFree(PrePoint[i], E);
    }
    ZENFree(PrePoint);
    
#if ZENECPTDEBUGLEVEL >= 25
    for(i = 0; i < twopowerminusone; i++)
      ZBZFree(pre[i]);
    ZENFree(pre);
    ZBZFree(r); ZBZFree(p);
#endif
    
    return(ZERR);
  }

#if ZENECPTDEBUGLEVEL >= 40
  fprintf(stderr, "#T = ");
  for(i = it-1; i >=0; i--)
    fprintf(stderr, "%d", (T[i]<0)?2:T[i]);
  fprintf(stderr, "\n");
#endif
  

  /* Let's go */
  ZENEcPtAssign(QBuf, P, E);

  b = started = 0; 
  for(it--; it >= 0; it--) {

    if (b == 0) {
      if (T[it]) {
	di = T[it];
	ui = 0;
	b = 1;
      }
      else {
	if ((ret = ZENEcPtDouble(QBuf, QBuf, E))) {
	  ZENEcPtFree(QBuf, E);
	  ZENFree(T); ZBNF(n);
	  for(i = 0; i < twopowerminusone; i++) {
	    ZENEcPtFree(PrePoint[i], E);
	  }
	  ZENFree(PrePoint);

#if ZENECPTDEBUGLEVEL >= 25
	  for(i = 0; i < twopowerminusone; i++)
	    ZBZFree(pre[i]);
	  ZENFree(pre);
	  ZBZFree(r); ZBZFree(p);
#endif

	  return(ret);
	}

#if ZENECPTDEBUGLEVEL >= 25
	r = ZBZAdd(p, p); ZBZFree(p); p = r;
#  if ZENECPTDEBUGLEVEL >= 30
	fprintf(stderr, "EcmMult(");
	ZBZPrintToFile(stderr, p);
	fprintf(stderr, ", P, Cv, N) - ");
	MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
	fprintf(stderr, ";\n");
#  endif
#endif
      }  
    } 
    else {
      b++;
      if (T[it]) {
	for(i = 0; i < ui; i++)
	  di <<= 1;
	ui = 0;
	  di <<= 1;
	di += T[it];
      }
      else
	ui++;

      if (b == m) {
	if (!started) {
	  started = 1;

	  ZENEcPtAssign(QBuf, PrePoint[di>>1], E);
#if ZENECPTDEBUGLEVEL >= 25
	  ZBZFree(p); p = ZBZCopy(pre[di>>1]);
#  if ZENECPTDEBUGLEVEL >= 30
	  fprintf(stderr, "EcmMult(");
	  ZBZPrintToFile(stderr, p);
	  fprintf(stderr, ", P, Cv, N) - ");
	  MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
	  fprintf(stderr, ";\n");
#  endif
#endif	    
	  for (i = 0; i < ui ; i++) {
	    if ((ret = ZENEcPtDouble(QBuf, QBuf, E))) {
	      ZENEcPtFree(QBuf, E);
	      ZENFree(T); ZBNF(n);
	      for(i = 0; i < twopowerminusone; i++) {
		ZENEcPtFree(PrePoint[i], E);
	      }
	      ZENFree(PrePoint);      
#if ZENECPTDEBUGLEVEL >= 25
	      for(i = 0; i < twopowerminusone; i++)
		ZBZFree(pre[i]);
	      ZENFree(pre);
	      ZBZFree(r); ZBZFree(p);
#endif
	      return(ret);
	    }
#if ZENECPTDEBUGLEVEL >= 25
	    r = ZBZAdd(p, p); ZBZFree(p); p = r;
#  if ZENECPTDEBUGLEVEL >= 30
	    fprintf(stderr, "EcmMult(");
	    ZBZPrintToFile(stderr, p);
	    fprintf(stderr, ", P, Cv, N) - ");
	    MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
	    fprintf(stderr, ";\n");
#  endif	    
#endif	    
	  }
	}
	else {
	  for(i = 0; i < m-ui; i++) {
	    if ((ret = ZENEcPtDouble(QBuf, QBuf,  E))) {
	      ZENEcPtFree(QBuf, E);
	      ZENFree(T); ZBNF(n);
	      for(i = 0; i < twopowerminusone; i++) {
		ZENEcPtFree(PrePoint[i], E);
	      }
	      ZENFree(PrePoint);      
#if ZENECPTDEBUGLEVEL >= 25
	      for(i = 0; i < twopowerminusone; i++)
		ZBZFree(pre[i]);
	      ZENFree(pre);
	      ZBZFree(r); ZBZFree(p);
#endif
	      return(ret);
	    }
#if ZENECPTDEBUGLEVEL >= 25
	    r = ZBZAdd(p, p); ZBZFree(p); p = r;
#  if ZENECPTDEBUGLEVEL >= 30
	    fprintf(stderr, "EcmMult(");
	    ZBZPrintToFile(stderr, p);
	    fprintf(stderr, ", P, Cv, N) - ");
	    MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
	    fprintf(stderr, ";\n");
#  endif	    
#endif	    
	  }
	  if (di < 0) {
	    if ((ret = ZENEcPtSubtract(QBuf, PrePoint[(-di)>>1], QBuf, E))) {
	      ZENEcPtFree(QBuf, E);
	      ZENFree(T); ZBNF(n);
	      for(i = 0; i < twopowerminusone; i++) {
		ZENEcPtFree(PrePoint[i], E);
	      }
	      ZENFree(PrePoint);      
#if ZENECPTDEBUGLEVEL >= 25
	      for(i = 0; i < twopowerminusone; i++)
		ZBZFree(pre[i]);
	      ZENFree(pre);
	      ZBZFree(r); ZBZFree(p);
#endif
	      return(ret);
	    }
	    ZENEcPtNegate(QBuf, QBuf, E);

#if ZENECPTDEBUGLEVEL >= 25
	    r = ZBZSubtract(p, pre[(-di)>>1]); ZBZFree(p); p = r;
#  if ZENECPTDEBUGLEVEL >= 30
	    fprintf(stderr, "EcmMult(");
	    ZBZPrintToFile(stderr, p);
	    fprintf(stderr, ", P, Cv, N) - ");
	    MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
	    fprintf(stderr, ";\n");
#  endif	    
#endif	    
	  }
	  else {
	    if ((ret = ZENEcPtAdd(QBuf, QBuf,  PrePoint[di>>1], E))) {
	      ZENEcPtFree(QBuf, E);
	      ZENFree(T); ZBNF(n);
	      for(i = 0; i < twopowerminusone; i++) {
		ZENEcPtFree(PrePoint[i], E);
	      }
	      ZENFree(PrePoint);      
#if ZENECPTDEBUGLEVEL >= 25
	      for(i = 0; i < twopowerminusone; i++)
		ZBZFree(pre[i]);
	      ZENFree(pre);
	      ZBZFree(r); ZBZFree(p);
#endif
	      return(ret);
	    }
#if ZENECPTDEBUGLEVEL >= 25
	      r = ZBZAdd(p, pre[di>>1]); ZBZFree(p); p = r;
#  if ZENECPTDEBUGLEVEL >= 30
	    fprintf(stderr, "EcmMult(");
	    ZBZPrintToFile(stderr, p);
	    fprintf(stderr, ", P, Cv, N) - ");
	    MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
	    fprintf(stderr, ";\n");
#  endif	    
#endif	    
	  }
	  
	  for (i = 0 ; i < ui ; i++) {
	    if ((ret = ZENEcPtDouble(QBuf, QBuf,  E))) {
	      ZENEcPtFree(QBuf, E);
	      ZENFree(T); ZBNF(n);
	      for(i = 0; i < twopowerminusone; i++) {
		ZENEcPtFree(PrePoint[i], E);
	      }
	      ZENFree(PrePoint);      
#if ZENECPTDEBUGLEVEL >= 25
	      for(i = 0; i < twopowerminusone; i++)
		ZBZFree(pre[i]);
	      ZENFree(pre);
	      ZBZFree(r); ZBZFree(p);
#endif
	      return(ret);
	    }
#if ZENECPTDEBUGLEVEL >= 25
	    r = ZBZAdd(p, p); ZBZFree(p); p = r;
#  if ZENECPTDEBUGLEVEL >= 30
	    fprintf(stderr, "EcmMult(");
	    ZBZPrintToFile(stderr, p);
	    fprintf(stderr, ", P, Cv, N) - ");
	    MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
	    fprintf(stderr, ";\n");
#  endif	    
#endif	    
	  }
	}
	di = b = ui = 0;
      }
    }
  }

  if (b) {
    if (!started) {

      ZENEcPtAssign(QBuf, PrePoint[di>>1], E);

#if ZENECPTDEBUGLEVEL >= 25
      ZBZFree(p); p = ZBZCopy(pre[di>>1]);
#  if ZENECPTDEBUGLEVEL >= 30
      fprintf(stderr, "EcmMult(");
      ZBZPrintToFile(stderr, p);
      fprintf(stderr, ", P, Cv, N) - ");
      MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
      fprintf(stderr, ";\n");
#  endif	    
#endif	    
      for (i = 0; i < ui ; i++) {
	if ((ret = ZENEcPtDouble(QBuf, QBuf, E))) {
	  ZENEcPtFree(QBuf, E);
	  ZENFree(T); ZBNF(n);
	  for(i = 0; i < twopowerminusone; i++) {
	    ZENEcPtFree(PrePoint[i], E);
	  }
	  ZENFree(PrePoint);      
#if ZENECPTDEBUGLEVEL >= 25
	  for(i = 0; i < twopowerminusone; i++)
	    ZBZFree(pre[i]);
	  ZENFree(pre);
	  ZBZFree(r); ZBZFree(p);
#endif
	  return(ret);
	}
#if ZENECPTDEBUGLEVEL >= 25
	  r = ZBZAdd(p, p); ZBZFree(p); p = r;
#  if ZENECPTDEBUGLEVEL >= 30
	fprintf(stderr, "EcmMult(");
	ZBZPrintToFile(stderr, p);
	fprintf(stderr, ", P, Cv, N) - ");
	MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
	fprintf(stderr, ";\n");
#  endif
#endif	    
      }
    }
    else {
      for (i = 0; i < b-ui ; i++) {
	if ((ret = ZENEcPtDouble(QBuf, QBuf, E))) {
	  ZENEcPtFree(QBuf, E);
	  ZENFree(T); ZBNF(n);
	  for(i = 0; i < twopowerminusone; i++) {
	    ZENEcPtFree(PrePoint[i], E);
	  }
	  ZENFree(PrePoint);      
#if ZENECPTDEBUGLEVEL >= 25
	  for(i = 0; i < twopowerminusone; i++)
	    ZBZFree(pre[i]);
	  ZENFree(pre);
	  ZBZFree(r); ZBZFree(p);
#endif
	  return(ret);
	}
#if ZENECPTDEBUGLEVEL >= 25
	r = ZBZAdd(p, p); ZBZFree(p); p = r;
#  if ZENECPTDEBUGLEVEL >= 30
	fprintf(stderr, "EcmMult(");
	ZBZPrintToFile(stderr, p);
	fprintf(stderr, ", P, Cv, N) - ");
	MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
	fprintf(stderr, ";\n");
#  endif
#endif	    
      }
      
      if (di < 0) {
	if ((ret = ZENEcPtSubtract(QBuf, PrePoint[(-di)>>1], QBuf, E))) {
	  ZENEcPtFree(QBuf, E);
	  ZENFree(T); ZBNF(n);
	  for(i = 0; i < twopowerminusone; i++) {
	    ZENEcPtFree(PrePoint[i], E);
	  }
	  ZENFree(PrePoint);      
#if ZENECPTDEBUGLEVEL >= 25
	  for(i = 0; i < twopowerminusone; i++)
	    ZBZFree(pre[i]);
	  ZENFree(pre);
	  ZBZFree(r); ZBZFree(p);
#endif
	  return(ret);
	}
	ZENEcPtNegate(QBuf, QBuf, E);

#if ZENECPTDEBUGLEVEL >= 25
	r = ZBZSubtract(p, pre[(-di)>>1]); ZBZFree(p); p = r;
#  if ZENECPTDEBUGLEVEL >= 30
	fprintf(stderr, "EcmMult(");
	ZBZPrintToFile(stderr, p);
	fprintf(stderr, ", P, Cv, N) - ");
	MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
	fprintf(stderr, ";\n");
#  endif
#endif	    
      }
      else {
	if ((ret = ZENEcPtAdd(QBuf, QBuf,  PrePoint[di>>1], E))) {
	  ZENEcPtFree(QBuf, E);
	  ZENFree(T); ZBNF(n);
	  for(i = 0; i < twopowerminusone; i++) {
	    ZENEcPtFree(PrePoint[i], E);
	  }
	  ZENFree(PrePoint);      
#if ZENECPTDEBUGLEVEL >= 25
	  for(i = 0; i < twopowerminusone; i++)
	    ZBZFree(pre[i]);
	  ZENFree(pre);
	  ZBZFree(r); ZBZFree(p);
#endif
	  return(ret);
	}
#if ZENECPTDEBUGLEVEL >= 25
	r = ZBZAdd(p, pre[di>>1]); ZBZFree(p); p = r;
#  if ZENECPTDEBUGLEVEL >= 30
	fprintf(stderr, "EcmMult(");
	ZBZPrintToFile(stderr, p);
	fprintf(stderr, ", P, Cv, N) - ");
	MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
	fprintf(stderr, ";\n");
#  endif
#endif 
      }
    
      for (i = 0 ; i < ui ; i++) {
	if ((ret = ZENEcPtDouble(QBuf, QBuf, E))) {
	  ZENEcPtFree(QBuf, E);
	  ZENFree(T); ZBNF(n);
	  for(i = 0; i < twopowerminusone; i++) {
	    ZENEcPtFree(PrePoint[i], E);
	  }
	  ZENFree(PrePoint);      
#if ZENECPTDEBUGLEVEL >= 25
	  for(i = 0; i < twopowerminusone; i++)
	    ZBZFree(pre[i]);
	  ZENFree(pre);
	  ZBZFree(r); ZBZFree(p);
#endif
	  return(ret);
	}
#if ZENECPTDEBUGLEVEL >= 25
	r = ZBZAdd(p, p); ZBZFree(p); p = r;
#  if ZENECPTDEBUGLEVEL >= 30
	fprintf(stderr, "EcmMult(");
	ZBZPrintToFile(stderr, p);
	fprintf(stderr, ", P, Cv, N) - ");
	MyZENEcPtPrintToFile(stderr, QBuf, 10, E);
	fprintf(stderr, ";\n");
#  endif
#endif	    
      }
    }
  }

  ZENFree(T); ZBNF(n);
  for(i = 0; i < twopowerminusone; i++) {
    ZENEcPtFree(PrePoint[i], E);
  }
  ZENFree(PrePoint);
       
#if ZENECPTDEBUGLEVEL >= 25
  for(i = 0; i < twopowerminusone; i++)
    ZBZFree(pre[i]);
  ZENFree(pre);
  ZBZFree(r); ZBZFree(p);
#endif


  ZENEcPtAssign(R, QBuf, E);

#if ZENECPTDEBUGLEVEL >= 25
  for(i = 0; i < twopowerminusone; i++)
    ZBZFree(pre[i]);
  ZENFree(pre);
  
  r = ZBZFromBigNum(k, kl);
  if (ZBZCompare(p, r))
    fprintf(stderr, " -> **************** ERROR ****************\n");
  ZBZFree(p);
  ZBZFree(r);
#endif

#if ZENECPTDEBUGLEVEL >= 25
  fprintf(stderr, "# <- ZENEcPtMultWindowSigned()\n");
#endif
  ZENEcPtFree(QBuf, E);

  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Multiplication.}
   \input{Two allocated points |P| and |R| of an elliptic curve |E|
   and a lare integer |k| of size |kl|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|R| = |k||P|$ if |ZEN_HAS_INVERSE| is returned, |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZENEcPtMult(R, k, kl, P, E)
     ZENEcPt R, P;
     ZENEc E;
     BigNum k;
     BigNumLength kl;
{
  int size, ret;

  size = kl*BN_DIGIT_SIZE;
  if (size <= 32)
    ret = ZENEcPtMultSigned(R, k, kl, P, E);
  else if (size <= 64)
    ret = ZENEcPtMultWindowSigned(R, 3, 5, k, kl, P, E);
  else if (size <= 256)
    ret = ZENEcPtMultWindowSigned(R, 4, 5, k, kl, P, E);
  else if (size <= 736)
    ret = ZENEcPtMultWindowSigned(R, 5, 4, k, kl, P, E);
  else if (size <= 1952)
    ret = ZENEcPtMultWindowSigned(R, 6, 3, k, kl, P, E);
  else if (size <= 5088)
    ret = ZENEcPtMultWindowSigned(R, 7, 3, k, kl, P, E);
  else if (size <= 12352)
    ret = ZENEcPtMultWindowSigned(R, 8, 3, k, kl, P, E);
  else if (size <= 30720)
    ret = ZENEcPtMultWindowSigned(R, 9, 3, k, kl, P, E);
  else
    ret = ZENEcPtMultWindowSigned(R, 10, 3, k, kl, P, E);

#if ZENECPTDEBUGLEVEL >= 10
  fprintf(stderr, "EcmMult(");
  ZBNPrintToFile(stderr, k, kl, 10);
  fprintf(stderr, ", ");
  MyZENEcPtPrintToFile(stderr, P, 10, E);
  fprintf(stderr, ", [ ");
  ZEP(ZENEcB4(E), ZENEcRing(E));
  fprintf(stderr, ", ");
  ZEP(ZENEcB6(E), ZENEcRing(E));
  fprintf(stderr, " ], ");
  ZBNPrintToFile(stderr, ZENRingQ(ZENEcRing(E)), ZENRingQl(ZENEcRing(E)), 10);
  fprintf(stderr, ") - ");
  MyZENEcPtPrintToFile(stderr, R, 10, E);
  fprintf(stderr, ";\n");
#endif

  return(ret);
}
