/*
 * zen/zeg/c/elt_exp.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element exponentiation}
   This section is inspired from the file |exp.c| of the BigMod library, written by F.~Morain.
   }
   */
#include <stdio.h>
#include <zeg.h>

#define EXP_LIMIT(x) (1+divSizeBloc(BlocExp2(x)*x))
#define MARYMAX         7
#define TWOPOWERMARYMAX 128 /* 2**MARYMAX */
#define LOCAL_VERBOSE 0

/* \TEX[
   \title{Exponentiation.}
   \input{A pointer on an element |R| and an element |P| of a finite ring |Rg|, a bignum (|k|, |kl|).}
   \output{0 (no error can occur).}
   \side{$|*R| = |P|^{(|k|, |kl|)}$.}
   \note{One must have $|*(k+kl-1)| \neq 0$.}
   }
   */
int _ZENEltExpPlain (R, k, kl, P, Rg)
     ZENElt *R, P;
     BigNum k;
     BigNumLength kl;
     ZENRing Rg;
{
  BigNumDigit C[1];
  int i, kd, ks;
  
  if (ZBNIsOne(k, kl)) {
    ZENEltAssign(*R, P, Rg);
    return(0);
  }
  if(ZENEltIsZero(P,Rg)) {
    ZENEltSetToZero(*R, Rg);
    return(0);
  }
  if ((ZBNIsZero(k, kl)) ||
      (ZENEltIsOne(P,Rg))) {
    ZENEltSetToOne(*R, Rg);
    return(0);
  }
  
  ZENEltAssign(ZENRingIbuf(Rg), P, Rg);
  kd = kl - 1;
  /* Starting the computation */
  ZBNAssign(C, (k+kd), 1);

  /* Looking for the most significant bit */
  for(ks = BN_DIGIT_SIZE; ks; ks--){
    if(ZBNShiftLeft(C, 1, 1))
      break;
  }
  ks--;
  while(kl >= 1){
    for(i = 0; i < ks; i++){
      ZENEltSquare(ZENRingIbuf(Rg), ZENRingIbuf(Rg), Rg);
      if(ZBNShiftLeft(C, 1, 1)){
	ZENEltMultiply(ZENRingIbuf(Rg), P, ZENRingIbuf(Rg), Rg);
      }
    }
    kl--;
    if(kl){
      kd--; ks = BN_DIGIT_SIZE;
      ZBNAssign(C, (k+kd), 1);
    }
  }
  ZENEltAssign(*R, ZENRingIbuf(Rg), Rg);
  return(0);
}

/********** Blocks **********/

/* After the article by H. Cohen and A. K. Lenstra: 
   "Implementation of a new primality test",
   Math. of Comp., volume 48, number 177, January 1987, pp. 103-121.*/

/* For 0 <= i < 2**m, i=(2**li)*ui, with ui odd and 0 <= li < m. Then:
ExpTableL[i]=li and ExpTableU[i]=(ui >> 1) is the address of a**ui
in OddPower (that's a hack to save space).
*/
static void MakeExpTables(m,R)
     int m;
     ZENRing R;
{int i, twopowerm = (1 << m), li, ui;

 ZENRingETL(R,m,0) = 0;
 for(i = 1; i < twopowerm; i++)
   {
     ui = i; li = 0;
     while(!(ui & 1)){ui >>= 1; li++;}
     ZENRingETL(R,m,i) = li;
     ZENRingETU(R,m,i) = (ui >> 1);
   }
}

/* \TEX{
   \title{Adding precomputations.}
   \input{The function pointers to use for the type of finite ring |R|.} 
   \output{|ZENERR| if an allocation error occured, 0 otherwise.}
   \side{Buffers are allocated and initialized for |_ZENEltExpBlock|.}
   \note{If the ring is a finite field (|ZENPRC_FINITE_FIELD| is set), the exponent is first reduced modulo |Q-1|. CAUTION: Results may be wrong if this flag is wrongly set.}
   }
   */
int __ZENPrcInitEltExp(EltExp,EltExpMod,EltExpMod4,EltExpMod5,EltExpMod6,R)
     int (*EltExp) ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));
     int (*EltExpMod) ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));
     int (*EltExpMod4) ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));
     int (*EltExpMod5) ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));
     int (*EltExpMod6) ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));
     ZENRing R;
{int i,tpw;


 ZENRingETL__(R)=(int **)ZENMalloc(3*sizeof(int *));
 if(ZENRingETL__(R)==NULL)
   {
     ZENSetError(ZEN_LIB_ZEG, ZEN_EXPPRC, ZEN_ALLOC);
     return(ZERR);
   }

 ZENRingETU__(R)=(int **)ZENMalloc(3*sizeof(int *));
 if(ZENRingETU__(R)==NULL)
   {
     ZENSetError(ZEN_LIB_ZEG, ZEN_EXPPRC, ZEN_ALLOC);
     ZENFree(ZENRingETL__(R));
     return(ZERR);
   }

 for(i=4;i<=6;i++)
   {
     tpw=(1<<i);
     ZENRingETL_(R,i)=(int *)ZENMalloc(tpw*sizeof(int));
     if(ZENRingETL_(R,i)==NULL)
       {
	 ZENSetError(ZEN_LIB_ZEG, ZEN_EXPPRC, ZEN_ALLOC);
	 for(i--;i>=4;i--)
	   ZENFree(ZENRingETL_(R,i));
	 ZENFree(ZENRingETU__(R));
	 ZENFree(ZENRingETL__(R));
	 return(ZERR);
       }

     ZENRingETU_(R,i)=(int *)ZENMalloc((tpw-1)*sizeof(int));
     if(ZENRingETU_(R,i)==NULL)
       {
	 ZENSetError(ZEN_LIB_ZEG, ZEN_EXPPRC, ZEN_ALLOC);
	 ZENFree(ZENRingETL_(R,i));
	 for(i--;i>=4;i--)
	   {
	     ZENFree(ZENRingETL_(R,i));
	     ZENFree(ZENRingETU_(R,i));
	   }	 
	 ZENFree(ZENRingETU__(R));
	 ZENFree(ZENRingETL__(R));
	 return(ZERR);
       }
     MakeExpTables(i,R);
   }

 ZENRingOddPower(R)=(ZENElt *)ZENMalloc(TWOPOWERMARYMAX*sizeof(ZENElt));
 if(ZENRingOddPower(R)==NULL)
   {
     ZENSetError(ZEN_LIB_ZEG, ZEN_EXPPRC, ZEN_ALLOC);
     for(i=6;i>=4;i--)
	   {
	     ZENFree(ZENRingETL_(R,i));
	     ZENFree(ZENRingETU_(R,i));
	   }
     ZENFree(ZENRingETU__(R));
     ZENFree(ZENRingETL__(R));
     return(ZERR);
   }
 for(i=0;i<TWOPOWERMARYMAX;i++)
   if(ZENEltCreate(ZENRingOddPower(R)[i],R))
     {
       for(i--;i>=0;i--)
	 ZENEltFree(ZENRingOddPower(R)[i],R);
       for(i=6;i>=4;i--)
	 {
	   ZENFree(ZENRingETL_(R,i));
	   ZENFree(ZENRingETU_(R,i));
	 }
       ZENFree(ZENRingETU__(R));
       ZENFree(ZENRingETL__(R));
       ZENFree(ZENRingOddPower(R));
       return(ZERR);
     }

 if(ZENPrcIsSet(ZENRingPrcp(R), ZENPRC_FINITE_FIELD))
   {
     BigNum e;
     int el;

     el = ZENRingSizeElt(R);
     e=ZBNC(el);
     if(!e) {
       for(i=TWOPOWERMARYMAX-1;i>=0;i--)
	 ZENEltFree(ZENRingOddPower(R)[i],R);
       for(i=6;i>=4;i--)
	 {
	   ZENFree(ZENRingETL_(R,i));
	   ZENFree(ZENRingETU_(R,i));
	 }
       ZENFree(ZENRingETU__(R));
       ZENFree(ZENRingETL__(R));
       ZENFree(ZENRingOddPower(R));
       return(ZERR);
     }
     ZBNAssign(e,ZENRingP1(R),el);
     ZBNSubtractBorrow(e, el, 0);
     el = ZBNNumDigits(e, el);
     if(el <= EXP_LIMIT(4))
       {
	 ZENEltExpPtr(R) = EltExpMod4;
#if LOCAL_VERBOSE
	 fprintf(stderr,"Modular exponentiation (binary method)\n");
#endif
       }
     else
       {
	 if(el <= EXP_LIMIT(5))
	   {
	     ZENEltExpPtr(R) = EltExpMod5;
#if LOCAL_VERBOSE
	     fprintf(stderr,"Modular exponentiation (block <= 5)\n");
#endif
	   }
       else
	 {
	 if(el <=  EXP_LIMIT(6))
	   {
	     ZENEltExpPtr(R) = EltExpMod6;
#if LOCAL_VERBOSE
	     fprintf(stderr,"Modular exponentiation (block <= 6)\n");
#endif
	   }
	 else
	   {
	     ZENEltExpPtr(R) = EltExpMod;
#if LOCAL_VERBOSE
	     fprintf(stderr,"Modular exponentiation (block)\n");
#endif
	   }
	 }
       }
     ZBNF(e);
   }
 else
   {
     ZENEltExpPtr(R) = EltExp;
#if LOCAL_VERBOSE
     fprintf(stderr,"No modular exponentiation (only block)\n");
#endif
   }

 return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Adding precomputations.}
   \input{A finite ring |R|.} 
   \output{|ZENERR| if an allocation error occured, 0 otherwise.}
   \side{Buffers are allocated and initialized for |_ZENEltExpBlock|.}
   \note{If the ring is a finite field, the exponent is first reduced modulo |Q-1|.}
   }
   */
int _ZENPrcInitEltExp(R)
     ZENRing R;
{
  return(__ZENPrcInitEltExp(_ZENEltExp,_ZENEltExpMod,_ZENEltExpMod4,_ZENEltExpMod5,_ZENEltExpMod6,R));
}
/* \TEX{
   \title{Freeing precomputations.}
   \input{The function pointers to use for the type of finite ring |R|.} 
   \side{Buffers for |_ZENEltExpBlock| are freed.}
   }
   */
int __ZENPrcCloseEltExp(EltExp,R)
     int (*EltExp) ___((ZENElt *,BigNum,BigNumLength,ZENElt,ZENRing));
     ZENRing R;
{int i;

 for(i=0;i<TWOPOWERMARYMAX;i++)
   ZENEltFree(ZENRingOddPower(R)[i],R);
 ZENFree(ZENRingOddPower(R));
 
 for(i=4;i<=6;i++)
   {
     ZENFree(ZENRingETL_(R,i));
     ZENFree(ZENRingETU_(R,i));
   }
 ZENFree(ZENRingETL__(R));
 ZENFree(ZENRingETU__(R));

 ZENEltExpPtr(R) = EltExp;
 return(ZEN_HAS_INVERSE);
}
/* \TEX{
   \title{Freeing precomputations.}
   \input{A finite ring |R|.} 
   \side{Buffers for |_ZENEltExpBlock| are freed.}
   }
   */
int _ZENPrcCloseEltExp(R)
     ZENRing R;
{
  return(__ZENPrcCloseEltExp(_ZENEltExpPlain,R));
}
/* ---------------------------------------------------------------------------
 .SSection "Block Exponentiation"
---------------------------------------------------------------------------*/

/* Computes p=a**e mod n. Write: e=e[t]*(2**m)**t + ... + e[0] and */
 /* e[i]=(2**li)*ui, e[t] <> 0.*/
/*
   \TEX{
   \title{Exponentiation precomputations}
   \input{A table of exponentiation precomputations to fill, a |ZENElt|, the size of table and a |ZENRing|.}
   \side{Exponentiation precomputations are done}
   \note{This procedure is externally defined to optimize exponentiations with the same |ZENElt| but several exponents.}
   } 
   */

static void _ZENEltExpPrc(OddPower,a,twopowermminus1,R)
     ZENElt *OddPower,a;
     int twopowermminus1;
     ZENRing R;
{int i;

 ZENEltSquare(ZENRingIbuf(R),a,R);
 ZENEltAssign(OddPower[0],a,R);

 for(i = 1; i < twopowermminus1; i++)
   ZENEltMultiply(OddPower[i],
		  ZENRingIbuf(R),OddPower[i-1],R);

}


/*
   \TEX{
   \title{Block exponentiation}
   \input{A pointer on a |ZENElt|, an exponent |(e,el)|, a |ZENElt|, a |ZENRing|, the length of block to use, the pre-computations on the exponent, and the mask on exponent.}
   \side{$|*y| = |a|^{(|e|, |el|)}$.}
   \note{This procedure is externally defined to optimize exponentiations with the same |ZENElt| but several exponents.}
   } 
   */
static void _ZENEltExpBlock2(y, e, el, a, R, lbloc,OddPower,mask)
     ZENElt *y,a,*OddPower;
     BigNumDigit mask;
     BigNum e;
     ZENRing R;
     BigNumLength el;
     int lbloc;
{BigNumDigit C[2];
 int i,nzeros,nbits,qs,rs,ds,ms,et;
 
 nzeros=ZBNNumLeadingZeroBitsInDigit(e[el-1]);
 nbits=el*BN_DIGIT_SIZE -nzeros;

 qs=nbits/lbloc;
 rs=nbits%lbloc;
 if(!rs)
   {
     qs--;
     rs=lbloc;
   }
 ds=qs*lbloc;
 ms=modSizeBloc(ds);
 ds=divSizeBloc(ds);

 if(ms+rs>BN_DIGIT_SIZE)
   {
     ZBNAssign(C,e+ds,2);
     ZBNShiftLeft(C,2,BN_DIGIT_SIZE-ms);
     et=C[1];
   }
 else
   {
     if(ms)
       {
	 ZBNAssign(C,e+ds,1);
	 et=ZBNShiftLeft(C,1,BN_DIGIT_SIZE-ms);
       }
     else
       {
	 C[0]=0;
	 et=e[ds];
       }
   }

 ZENEltAssign(ZENRingIbuf(R),
	      OddPower[ZENRingETU(R,lbloc,et)],
	      R);
 
 /* Et <- Et**(2**lt) */
 for(i = ZENRingETL(R,lbloc,et); i ; i--)
   ZENEltSquare(ZENRingIbuf(R),
		ZENRingIbuf(R),
		R);
 /* loop */
 while(qs>0)
   {
     /* get current digit e[i]=(2**li)*ui */
     qs--;
     ms -= lbloc;
     if(ms<0)
       {
	 ds--;
	 ZBNAssign(C,e+ds,2);
	 ZBNShiftLeft(C,2,-ms);
	 et=C[1]&mask;
	 ms += BN_DIGIT_SIZE;
       }
     else
       et=ZBNShiftLeft(C,1,lbloc);
     
     for(i = lbloc-ZENRingETL(R,lbloc,et); i ; i--)
       ZENEltSquare(ZENRingIbuf(R),
		    ZENRingIbuf(R),
		    R);
     
     if(et)
       {
	 /* Et <- Et*a**vt */
	 ZENEltMultiply(ZENRingIbuf(R),
			OddPower[ZENRingETU(R,lbloc,et)],
			ZENRingIbuf(R),
			R);
	 /* Et <- Et**(2**lt) */
	 for(i = ZENRingETL(R,lbloc,et); i; i--)
	   ZENEltSquare(ZENRingIbuf(R),
			ZENRingIbuf(R),
			R);

       }
   }
 
 ZENEltAssign(*y,ZENRingIbuf(R),R);
}


static void _ZENEltExpBlock(y, e, el, a, R, lbloc)
     ZENElt *y,a;
     BigNum e;
     ZENRing R;
     BigNumLength el;
     int lbloc;
{BigNumDigit mask;
 int twopowermminus1;
 
 if (ZBNIsOne(e, el)) {
   ZENEltAssign(*y, a, R);
   return;
 }

 if(ZENEltIsZero(a,R)) {
   ZENEltSetToZero(*y, R);
   return;
 }

 if ((ZBNIsZero(e, el)) ||
     (ZENEltIsOne(a,R))) {
   ZENEltSetToOne(*y, R);
   return;
 }
 
 
 mask=(1<<lbloc)-1;
 twopowermminus1=(1<<(lbloc-1));
 _ZENEltExpPrc(ZENRingOddPower(R),a,twopowermminus1,R);
 _ZENEltExpBlock2(y,e,el,a,R,lbloc,ZENRingOddPower(R),mask);
}

/****************/

/* \TEX{
   \title{Exponentiation.}
   \input{A pointer on an element |R| and an element |P| of a finite ring |Rg|, a bignum (|k|, |kl|).}
   \side{$|*R| = |P|^{(|k|, |kl|)}$.}
   \output{0 (no error can occur).}
   \note{One must have $|*(k+kl-1)| \neq 0$. This procedure use the best algorithm according to the size of the exponent.}
   }
   */
int _ZENEltExp (R, e, el, P, Rg)
     ZENElt *R, P;
     BigNum e;
     BigNumLength el;
     ZENRing Rg;
{
  if(el <= EXP_LIMIT(4))
    _ZENEltExpPlain(R, e, el, P, Rg);
  else{
    if(el <= EXP_LIMIT(5))
      _ZENEltExpBlock(R, e, el, P, Rg, 4);
    else{
      if(el <=  EXP_LIMIT(6))
	_ZENEltExpBlock(R, e, el, P, Rg, 5);
      else
	_ZENEltExpBlock(R, e, el, P, Rg, 6);
    }
  }
  return(0);
}

/* \TEX{
   \title{Exponentiation.}
   \input{A pointer on an element |R| and an element |P| of a finite ring |Rg|, a bignum (|k|, |kl|).}
   \side{$|*R| = |P|^{(|k|, |kl|)}$.}
   \output{|ZENERR| if an allocation error occurs, 0 otherwise.}
   \note{One must have $|*(k+kl-1)| \neq 0$. The exponent is first reduced modulo $q-1$, with $q$ the cardianlity of |R|. This procedure use the best algorithm according to the size of the exponent.}
   }
   */
int _ZENEltExpMod (R, e, el, P, Rg)
     ZENElt *R, P;
     BigNum e;
     BigNumLength el;
     ZENRing Rg;
{
  BigNum p;

  switch(ZBNCompareLazy(e,el,ZENRingP1(Rg),ZENRingSizeElt(Rg)))
    {
    case BN_EQ:
      if(ZENEltIsZero(P,Rg))
	ZENEltSetToZero(*R,Rg);
      else
	ZENEltSetToOne(*R,Rg);
      return(0);
    case BN_LT:
      p=e;
      e=NULL;
      break;
    case BN_GT:
      p=ZBNC(el+1);
      if(!p)
       {
	 ZENSetError(ZEN_LIB_ZEG,ZEN_EXP,ZEN_ALLOC);
	 return(ZERR);
       }
      ZBNAssign(p,e,el);
      p[el]=0;
      ZBNDivide(p,el+1,ZENRingP1(Rg),ZENRingSizeElt(Rg));
      el=ZBNNumDigits(p,ZENRingSizeElt(Rg));
      break;
    }
  
  if(el <= EXP_LIMIT(4))
    _ZENEltExpPlain(R, p, el, P, Rg);
  else{
    if(el <= EXP_LIMIT(5))
      _ZENEltExpBlock(R, p, el, P, Rg, 4);
    else{
      if(el <=  EXP_LIMIT(6))
	_ZENEltExpBlock(R, p, el, P, Rg, 5);
      else
	_ZENEltExpBlock(R, p, el, P, Rg, 6);
    }
  }
  if(e)
    ZBNF(p);
  return(0);
}
				/* The following are faster, but no need to
				   document them */
int _ZENEltExpMod4(R, e, el, P, Rg)
     ZENElt *R, P;
     BigNum e;
     BigNumLength el;
     ZENRing Rg;
{
  BigNum p;

  el=ZBNModuloLazy(&p,e,el,ZENRingP1(Rg),ZENRingSizeElt(Rg));
  if(!el)
    return(ZERR);
  _ZENEltExpPlain(R, p, el, P, Rg);
  ZBNF(p);
  return(0);
}

int _ZENEltExpMod5(R, e, el, P, Rg)
     ZENElt *R, P;
     BigNum e;
     BigNumLength el;
     ZENRing Rg;
{
  BigNum p;

  el=ZBNModuloLazy(&p,e,el,ZENRingP1(Rg),ZENRingSizeElt(Rg));
  if(!el)
    return(ZERR);
  if(el <= EXP_LIMIT(4))
    _ZENEltExpPlain(R, p, el, P, Rg);
  else
    _ZENEltExpBlock(R, p, el, P, Rg, 4);
  
  ZBNF(p);
  return(0);
}

int _ZENEltExpMod6(R, e, el, P, Rg)
     ZENElt *R, P;
     BigNum e;
     BigNumLength el;
     ZENRing Rg;
{
  BigNum p;

  el=ZBNModuloLazy(&p,e,el,ZENRingP1(Rg),ZENRingSizeElt(Rg));
  if(!el)
    return(ZERR);
  if(el <= EXP_LIMIT(4))
    _ZENEltExpPlain(R, p, el, P, Rg);
  else{
    if(el <= EXP_LIMIT(5))
      _ZENEltExpBlock(R, p, el, P, Rg, 4);
    else
      _ZENEltExpBlock(R, p, el, P, Rg, 5);
  }
  ZBNF(p);
  return(0);
}
