/*
 * zen/zeg/c/elt_sqrt.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element square root}
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{Is a square in finite fields of odd characteristic.}
   \input{An element |P| of a finite ring |Rg|).}
   \output{|ZENERR| if an error occurred, 1 if |P| is a square, 0 otherwise.}
   \note{This procedure is valid only in finite fields.}
   }
   */
int _ZENEltIsASquareCharp(P, Rg)
     ZENElt P;
     ZENRing Rg;
{
  BigNum C;
  ZENElt R;

  if (ZENEltIsZero(P, Rg)) {
    return(1);
  }

  if (!(C = ZBNC(ZENRingQl(Rg)))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ZENELTISASQRCP, ZEN_ALLOC);
    return(ZERR);
  }
  if (ZENEltCreate(R, Rg)) {
    ZBNF(C);
    return(ZERR);
  }

  ZBNAssign(C, ZENRingQ(Rg), ZENRingQl(Rg));
  ZBNShiftRight(C, ZENRingQl(Rg), 1);
  if(ZENEltExp(R, C, ZENRingQl(Rg), P, Rg)) {
    ZBNF(C);
    ZENEltFree(R, Rg);
    return(ZERR);
  }
  ZBNF(C);
  if (ZENEltIsOne(R, Rg)) {
    ZENEltFree(R, Rg);
    return(1);
  }
  ZENEltFree(R, Rg);
  return(0);
}

/* \TEX{
   \title{Is a square in finite fields of characteristic 2.}
   \input{An element |P| of a finite ring |Rg|).}
   \output{1.}
   \note{This procedure is valid only in finite fields, in such a case
   this procedure always returns 1.}
   }
   */
int _ZENEltIsASquareChar2(P, Rg)
     ZENElt P;
     ZENRing Rg;
{
  return(1);
}

/* \TEX{
   \title{Square roots in finite fields of odd characteristic.}
   \input{An pointer to an element |R| and an element |P| of a finite
   field |Rg|).}
   \output{|ZENERR| if an error occurred, 0 if |P| is a square, 1 otherwise.}
   \note{This procedure is valid only in finite fields.}
   }
   */
int _ZENEltSquareRootCharp(R, P, Rg)
     ZENElt *R, P;
     ZENRing Rg;
{
  int i, ret, r, m, init_done;
  BigNum q;
  BigNumDigit C;
  int ql;
  ZENElt x, b, y;

  if (ZENEltIsZero(P, Rg)) {
    ZENEltSetToZero(*R, Rg);
    return(0);
  }

  if (!(q = ZBNC(ZENRingQl(Rg)))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ZENELTSQRTCP, ZEN_ALLOC);
    return(ZERR);
  }

  if (ZENEltCreate(x, Rg)) {
    ZBNF(q);
    return(ZERR);    
  }

  if (ZENEltCreate(b, Rg)) {
    ZBNF(q);
    ZENEltFree(x, Rg);
    return(ZERR);    
  }

  if (ZENEltCreate(y, Rg)) {
    ZBNF(q);
    ZENEltFree(x, Rg);
    ZENEltFree(b, Rg);
    return(ZERR);    
  }

  ZBNAssign(q, ZENRingQ(Rg), ZENRingQl(Rg));
  ZBNShiftRight(q, ZENRingQl(Rg), 1);
  for(r = 1; !(ZBNDigit(q, 0) & 1); r++) {
    ZBNShiftRight(q, ZENRingQl(Rg), 1);
  }
  ql = ZBNNumDigits(q, ZENRingQl(Rg));

  C = ZBNDigit(q, 0);
  ZBNShiftRight(q, ZENRingQl(Rg), 1);
  if(ZENEltExp(x, q, ql, P, Rg)) {
    ZBNF(q);
    ZENEltFree(x, Rg);
    ZENEltFree(y, Rg);
    ZENEltFree(b, Rg);
    return(ZERR);
  }
  ZBNShiftLeft(q, ZENRingQl(Rg), 1);
  ZBNDigit(q, 0) = C;

  ZENEltSquare(b, x, Rg);
  ZENEltMultiply(b, P, b, Rg);

  ZENEltMultiply(x, P, x, Rg);

  init_done = 0;
  while (!ZENEltIsOne(b, Rg)) {
    if (!init_done) {
      ZENEltSetRandom(y, Rg);
      while (ZENEltIsZero(y, Rg) || (ret = ZENEltIsASquare(y, Rg))) {
	ZENEltSetNext(y, Rg);
      }
      if (ret == -1) {
	ZBNF(q);
	ZENEltFree(x, Rg);
	ZENEltFree(b, Rg);
	ZENEltFree(y, Rg);
	return(ZERR);
      }
      if(ZENEltExp(y, q, ql, y, Rg)) {
	ZBNF(q);
	ZENEltFree(x, Rg);
	ZENEltFree(b, Rg);
	ZENEltFree(y, Rg);
	return(ZERR);
      }
      init_done = 1;
    }

    ZENEltAssign(*R, b, Rg);
    for(m = 0; !ZENEltIsOne(*R, Rg); m++) {
      ZENEltSquare(*R, *R, Rg);
    }

    if (m == r) {
      ZBNF(q);
      ZENEltFree(x, Rg);
      ZENEltFree(b, Rg);
      ZENEltFree(y, Rg);
      return(1);
    }

    for(i = 0; i < r-m-1; i++) {
      ZENEltSquare(y, y, Rg);
    }
    ZENEltMultiply(x, y, x, Rg);
    ZENEltSquare(y, y, Rg);
    r = m;
    ZENEltMultiply(b, y, b, Rg);
  }

  ZENEltAssign(*R, x, Rg);
  ZBNF(q);
  ZENEltFree(x, Rg);
  ZENEltFree(b, Rg);
  ZENEltFree(y, Rg);
  return(0);
}

/* \TEX{
   \title{Square roots in finite fields of odd characteristic.}
   \input{A pointer to an element |R| and an element |P| of a finite
   field |Rg|).}
   \output{0.}
   \note{This procedure is valid only in finite fields.}
   }
   */
int _ZENEltSquareRootChar2(R, P, Rg)
     ZENElt *R, P;
     ZENRing Rg;
{
  int nl;
  BigNum n;

  if (!(n = ZBNC(ZENRingQl(Rg)))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_ZENELTSQRTCP, ZEN_ALLOC);
    return(ZERR);
  }
  ZBNAssign(n, ZENRingQ(Rg), nl=ZENRingQl(Rg));
  ZBNShiftRight(n, ZENRingQl(Rg), 1);
  nl = ZBNNumDigits(n, ZENRingQl(Rg));
  if(ZENEltExp(*R, n, nl, P, Rg)) {
    ZBNF(n);
    return(ZERR);
  }
  ZBNF(n);
  
  return(0);
}
