/*
 * zen/zeg/c/mat_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial additions}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions. 
   }
   */
#include <stdio.h>
#include "zen.h"

/*
   \TEX[
   \title{Addition.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| += |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int _ZENMatAdd(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;
 ZENElt S;

 ZENTest(ZENMatIsPermutation(A,R)
	 ||ZENMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATADD,ZEN_LIB_ZEG, ZERR);

 if(ZENMatAreSameType(A,B,R))
   {

     ZENTest((ZENMatNbBn(A,R) != ZENMatNbBn(B,R))
	     ||(ZENMatNbDigit(A,R) != ZENMatNbDigit(B,R)),
	     ZEN_MAT_DIM,ZEN_MATADD,ZEN_LIB_ZEG, ZERR);

     ZENCheckError(ZENEltCreate(S,R), ZERR);

     
     for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZENMatNbDigit(A,R)-1;c>=0;c--)
	 {
	   ZENEltAssign(S,ZENMatGetBlocPtr(A,l,c,R),R);
	   ZENEltAdd(S,ZENMatGetBlocPtr(B,l,c,R),R);
	   ZENMatSetBloc(A,l,c,S,R);
	 }
     ZENEltFree(S,R);
   }
 else
   {

     ZENTest((ZENMatNbBn(A,R) != ZENMatNbDigit(B,R))
	     ||(ZENMatNbDigit(A,R) != ZENMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATADD,ZEN_LIB_ZEG, ZERR);

     ZENCheckError(ZENEltCreate(S,R), ZERR);

     
     for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZENMatNbDigit(A,R)-1;c>=0;c--)
	 {
	   ZENEltAssign(S,ZENMatGetBlocPtr(A,l,c,R),R);
	   ZENEltAdd(S,ZENMatGetBlocPtr(B,c,l,R),R);
	   ZENMatSetBloc(A,l,c,S,R);
	 }
     ZENEltFree(S,R);
   }

 return(0);
}

   
/*
   \TEX[
   \title{Addition of scalar}
   \input{A |ZENMat|, a |ZENElt| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{All the elements of |A| are incremented by |b|.}
   \note{|A| cannot be a permutation.}
   }
   */
int _ZENMatAddScalar(A,b,R)
     ZENMat A;
     ZENElt b;
     ZENRing R;
{
  int l,c;
  ZENElt S;

  ZENTest(ZENMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATADD,ZEN_LIB_ZEG, ZERR);

  ZENCheckError(ZENEltCreate(S,R), ZERR);

     
  for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
    for(c=ZENMatNbDigit(A,R)-1;c>=0;c--)
      {
	ZENEltAssign(S,ZENMatGetBlocPtr(A,l,c,R),R);
	ZENEltAdd(S,b,R);
	ZENMatSetBloc(A,l,c,S,R);
      }
  ZENEltFree(S,R);
  
  return(0);
}

 /* \TEX[
   \title{Negation.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = -|B| if no error occured, |zen_errno| is set otherwise.}
   }
   */

int _ZENMatNegate(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;
 ZENElt S;

 ZENTest(ZENMatIsPermutation(A,R)
	 ||ZENMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATNEGATE,ZEN_LIB_ZEG, ZERR);

 if(ZENMatAreSameType(A,B,R))
   {

     ZENTest((ZENMatNbBn(A,R) != ZENMatNbBn(B,R))
	     ||(ZENMatNbDigit(A,R) != ZENMatNbDigit(B,R)),
	     ZEN_MAT_DIM,ZEN_MATNEGATE,ZEN_LIB_ZEG, ZERR);

     ZENCheckError(ZENEltCreate(S,R), ZERR);

     
     for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZENMatNbDigit(A,R)-1;c>=0;c--)
	 {
	   ZENEltNegate(S,ZENMatGetBlocPtr(B,l,c,R),R);
	   ZENMatSetBloc(A,l,c,S,R);
	 }
     ZENEltFree(S,R);
   }
 else
   {

     ZENTest((ZENMatNbBn(A,R) != ZENMatNbDigit(B,R))
	     ||(ZENMatNbDigit(A,R) != ZENMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATNEGATE,ZEN_LIB_ZEG, ZERR);

     ZENCheckError(ZENEltCreate(S,R), ZERR);

     
     for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZENMatNbDigit(A,R)-1;c>=0;c--)
	 {
	   ZENEltNegate(S,ZENMatGetBlocPtr(B,c,l,R),R);
	   ZENMatSetBloc(A,l,c,S,R);
	 }
     ZENEltFree(S,R);
   }

 return(0);
}

/*
   \TEX[
   \title{Subtraction.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| -= |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int _ZENMatSubtract(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;
 ZENElt S;

 ZENTest(ZENMatIsPermutation(A,R)
	 ||ZENMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATSUBTRACT,ZEN_LIB_ZEG, ZERR);

 if(ZENMatAreSameType(A,B,R))
   {

     ZENTest((ZENMatNbBn(A,R) != ZENMatNbBn(B,R))
	     ||(ZENMatNbDigit(A,R) != ZENMatNbDigit(B,R)),
	     ZEN_MAT_DIM,ZEN_MATSUBTRACT,ZEN_LIB_ZEG, ZERR);

     ZENCheckError(ZENEltCreate(S,R), ZERR);

     
     for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZENMatNbDigit(A,R)-1;c>=0;c--)
	 {
	   ZENEltAssign(S,ZENMatGetBlocPtr(A,l,c,R),R);
	   ZENEltSubtract(S,ZENMatGetBlocPtr(B,l,c,R),R);
	   ZENMatSetBloc(A,l,c,S,R);
	 }
     ZENEltFree(S,R);
   }
 else
   {

     ZENTest((ZENMatNbBn(A,R) != ZENMatNbDigit(B,R))
	     ||(ZENMatNbDigit(A,R) != ZENMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATSUBTRACT,ZEN_LIB_ZEG, ZERR);

     ZENCheckError(ZENEltCreate(S,R), ZERR);

     
     for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZENMatNbDigit(A,R)-1;c>=0;c--)
	 {
	   ZENEltAssign(S,ZENMatGetBlocPtr(A,l,c,R),R);
	   ZENEltSubtract(S,ZENMatGetBlocPtr(B,c,l,R),R);
	   ZENMatSetBloc(A,l,c,S,R);
	 }
     ZENEltFree(S,R);
   }

 return(0);
}

   
