/*
 * zen/zeg/c/mat_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial assignments}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions.

   }
   */
#include <stdio.h>
#include "zeg.h"

/*
   \TEX[
   
   \title{Assignment.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int _ZENMatAssign(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;

 ZENTest(ZENMatIsPermutation(A,R)
	 ||ZENMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATASSIGN,ZEN_LIB_ZEG, ZERR);
 
 if(ZENMatAreSameType(A,B,R))
   {
     ZENTest((ZENMatNbBn(A,R) != ZENMatNbBn(B,R))
	     ||(ZENMatNbDigit(A,R) != ZENMatNbDigit(B,R)),
	     ZEN_MAT_DIM,ZEN_MATASSIGN,ZEN_LIB_ZEG, ZERR);
     
     for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZENMatNbBloc(A,R)-1;c>=0;c--)
	 ZENMatSetBloc(A,l,c,ZENMatGetBlocPtr(B,l,c,R),R);
   }
 else
   {
     ZENTest((ZENMatNbBn(A,R) != ZENMatNbDigit(B,R))
	     ||(ZENMatNbDigit(A,R) != ZENMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATASSIGN,ZEN_LIB_ZEG, ZERR);
     
     for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZENMatNbBloc(A,R)-1;c>=0;c--)
	 ZENMatSetBloc(A,l,c,ZENMatGetBlocPtr(B,c,l,R),R);
   }
return(0);
}

/*
   \TEX[
   \title{Set to zero}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are set to zero.}
   }
   */
   
int _ZENMatSetToZero(A,R)
     ZENMat A;
     ZENRing R;
{int l,c;
 ZENElt S;

 ZENTest(ZENMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETZERO,ZEN_LIB_ZEG, ZERR);
 ZENCheckError(ZENEltCreate(S,R), ZERR);
 ZENEltSetToZero(S,R);
 
 for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZENMatNbBloc(A,R)-1;c>=0;c--)
     ZENMatSetBloc(A,l,c,S,R);

 ZENEltFree(S,R);
 return(0);
}

/*
   \TEX[
   \title{Set to one}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are set to zero except for the coefficients with same indices on rows and columns that are set to one.}
   \note{There is no error if the matrix is not square.}
   }
   */
   
int _ZENMatSetToOne(A,R)
     ZENMat A;
     ZENRing R;
{int l,c;
 ZENElt Z,S;

 ZENTest(ZENMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETONE,ZEN_LIB_ZEG, ZERR);
 ZENCheckError(ZENEltCreate(S,R), ZERR);
 ZENCheckError(ZENEltCreate(Z,R), ZERR);

 ZENEltSetToZero(Z,R);
 ZENEltSetToOne(S,R);
 
 for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZENMatNbBloc(A,R)-1;c>=0;c--)
     if(l==c)
       ZENMatSetBloc(A,l,c,S,R);
     else
       ZENMatSetBloc(A,l,c,Z,R);
       
 ZENEltFree(Z,R);
 ZENEltFree(S,R);
 return(0);
}

/*
   \TEX[
   \title{Set to random}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are randomly set.}
   }
   
   */
   
int _ZENMatSetRandom(A,R)
     ZENMat A;
     ZENRing R;
{int l,c;
 ZENElt S;

 ZENTest(ZENMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETRANDOM,ZEN_LIB_ZEG, ZERR);
 ZENCheckError(ZENEltCreate(S,R), ZERR);
 
 for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZENMatNbBloc(A,R)-1;c>=0;c--)
     {
       ZENEltSetRandom(S,R);
       ZENMatSetBloc(A,l,c,S,R);
     }

 ZENEltFree(S,R);
 return(0);
}

/*
   \TEX[
   \title{Transpose}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \side{$|A| = |A|^{t}$}
   }   
   */
   
void _ZENMatTranspose(M,R)
     ZENMat M;
     ZENRing R;
{int (*ptr) ___((ZENMat, Dim, Dim, ZENRing));
 
 ZENMatType(M,R) = (ZENMatTypeStruct)(((int)(ZENMatType(M,R)))^ 1) ;
 ptr=ZENMatPermuteColPtr(M);
 ZENMatPermuteColPtr(M)=ZENMatPermuteRowPtr(M);
 ZENMatPermuteRowPtr(M)=ptr;

 if(ZENMatIsPermutation(M,R))
   {
     if(ZENMatIsRowType(M,R))
       {
	 ZENMatGetCoeffNewPtr(M)=_ZENPermutationGetBloc;
	 ZENMatSetCoeffPtr(M)=_ZENPermutationSetBloc;     
       }
     else
       {
	 ZENMatGetCoeffNewPtr(M)=_ZENPermutationGetDigit;
	 ZENMatSetCoeffPtr(M)=_ZENPermutationSetDigit;     
       }
   }
 else
  {
     if(ZENMatIsRowType(M,R))
       {
	 ZENMatGetCoeffNewPtr(M)=__ZENMatGetBloc;
	 ZENMatGetCoeffPtrPtr(M)=__ZENMatGetBlocPtr;
	 ZENMatSetCoeffPtr(M)=__ZENMatSetBloc;     
       }
     else
       {
	 ZENMatGetCoeffNewPtr(M)=__ZENMatGetDigit;
	 ZENMatGetCoeffPtrPtr(M)=__ZENMatGetDigitPtr;
	 ZENMatSetCoeffPtr(M)=__ZENMatSetDigit;     
       }
  }
 return;
}


