/*
 * zen/zeg/c/mat_convert.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zen.h"

extern int _Ze2MatConvert ___((ZENMat,ZENRing,ZENMat,ZENRing));

/*
   \TEX[
   \Subsection{Matrix conversion}
   
   \title{Simple assignment}
   \input{Two couples (|ZENMat|, |ZENRing|).}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |M1| is assigned with |M2|.}
   \note{The coefficients of |M2| must be elements of |R1|. The two matrices must be of same sizes.}
   } 
   */
int _ZENMatConvert(M1,R1,M2,R2)
     ZENMat M1,M2;
     ZENRing R1,R2;
{int l,c;
 BigNum n;
 BigNumLength nl,pl;


 ZENTest((ZENMatNbRow(M1,R1) != ZENMatNbRow(M2,R2))
	 ||(ZENMatNbCol(M1,R1) != ZENMatNbCol(M2,R2)),
	 ZEN_MAT_DIM,ZEN_MATCONVERT,ZEN_LIB_ZEG, ZERR);
 ZENTest(ZENMatIsPermutation(M1,R1)
	 ||ZENMatIsPermutation(M2,R2),
	 ZEN_ISPERM,ZEN_MATCONVERT,ZEN_LIB_ZEG, ZERR);
 

 if((ZENRingArith(R1) == ZEN_LIB_ZE2) ||
    (ZENRingArith(R2) == ZEN_LIB_ZE2))
   return(_Ze2MatConvert(M1,R1,M2,R2));
 
 nl=ZENRingQl(R1);
 n=ZENRingBNbuf(R1);
 if(ZENRingQl(R2)>nl)
   {
     nl=ZENRingQl(R2);
     n=ZENRingBNbuf(R2);
   } 

 for(l=ZENMatNbRow(M1,R1);l--;)
   for(c=ZENMatNbCol(M1,R1);c--;)
     {
       pl=nl;
       ZENEltToZ(n,&pl,ZENMatGetCoeffPtr(M2,l,c,R2),R2);	       
       ZENZToElt(ZENRingHbuf(R1),n,pl,R1);
       ZENMatSetCoeff(M1,l,c,ZENRingHbuf(R1),R1);       
     }
 return(0);
}


/*     
   \TEX[
      The following functions are intended to allow conversion of matrices between an extension and its definition |ZENRing|. In all the following functions, matrix |M| is defined over the extension |R| and |m| is defined over the underlevel extension |ZENRingDef(R)|.
These are the general functions that work even if the extension is a clone.

   \title{Extension conversion}
   \input{Two matrices |ZENMat| and a possibly tabulated extension |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |m| is assigned with |M|.}
   } 
   */
int _ZENCloneMat2Vect(M,m,R)
     ZENMat M,m;
     ZENRing R;
{int l,c,cc,k,n;
 ZENRing RO;

 if(ZENRingOrigin(R))
   RO=ZENRingOrigin(R);
 else
   RO=R;
 
 ZENTest(ZENRingDef(R)==NULL,
	 ZEN_NOT_EXT,ZEN_MAT2VECT,ZEN_LIB_ZEG, ZERR);

 ZENTest(ZENMatIsPermutation(M,R)
	 ||ZENMatIsPermutation(m,ZENRingDef(R)),
	 ZEN_ISPERM,ZEN_MAT2VECT,ZEN_LIB_ZEG, ZERR);
     
 n=ZENPolyDeg(ZENRingPol(R),ZENRingDef(R));
     
 if(ZENMatIsRowType(m,ZENRingDef(R)))
   {
     ZENTest((ZENMatNbRow(M,R) != ZENMatNbBn(m,ZENRingDef(R)))
	     ||(ZENMatNbCol(M,R)*n != ZENMatNbDigit(m,ZENRingDef(R))),
	     ZEN_MAT_DIM,ZEN_MAT2VECT,ZEN_LIB_ZEG, ZERR);
     
     for(l=0;l<ZENMatNbBn(m,ZENRingDef(R));l++)
       for(c=0,cc=0;c<ZENMatNbCol(M,R);c++)
	 {
	   for(k=0;k<n;k++,cc++)
	     if(!ZENEltIsZero(ZENMatGetBlocPtr(m,l,cc,ZENRingDef(R)),
			      ZENRingDef(R)))
	       break;
	   if(k<n)
	     {
	       ZENPolySetToXi(ZENElt2Pol(ZENRingHbuf(RO),
					 RO),
			      n-k-1,
			      ZENRingDef(R));
	       for(;k<n;k++,cc++)
		 ZENPolySetCoeff(ZENElt2Pol(ZENRingHbuf(RO),
					    RO),
				 n-k-1,
				 ZENMatGetBlocPtr(m,l,cc,ZENRingDef(R)),
				 ZENRingDef(R));
	       
	       ZENEltToClone(ZENRingHbuf(R),ZENRingHbuf(RO),R);
	     }
	   else
	     ZENEltSetToZero(ZENRingHbuf(R),R);
	   
	   ZENMatSetCoeff(M,l,c,ZENRingHbuf(R),R);
	 }
   }
 else
   {
     ZENTest((ZENMatNbRow(M,R)*n != ZENMatNbDigit(m,ZENRingDef(R)))
	     ||(ZENMatNbCol(M,R) != ZENMatNbBn(m,ZENRingDef(R))),
	     ZEN_MAT_DIM,ZEN_MAT2VECT,ZEN_LIB_ZEG, ZERR);
     
     for(c=0;c<ZENMatNbBn(m,ZENRingDef(R));c++)
       for(l=0,cc=0;l<ZENMatNbRow(M,R);l++)
	 {
	   for(k=0;k<n;k++,cc++)
	     if(!ZENEltIsZero(ZENMatGetBlocPtr(m,c,cc,ZENRingDef(R)),
			      ZENRingDef(R)))
	       break;
	   if(k<n)
	     {
	       ZENPolySetToXi(ZENElt2Pol(ZENRingHbuf(RO),
					 RO),
			      n-k-1,
			      ZENRingDef(R));
	       for(;k<n;k++,cc++)
		 ZENPolySetCoeff(ZENElt2Pol(ZENRingHbuf(RO),
					    RO),
				 n-k-1,
				 ZENMatGetBlocPtr(m,c,cc,ZENRingDef(R)),
				 ZENRingDef(R));

	       ZENEltToClone(ZENRingHbuf(R),ZENRingHbuf(RO),R);
	     }
	   else
	     ZENEltSetToZero(ZENRingHbuf(R),R);
	       
	   ZENMatSetCoeff(M,l,c,ZENRingHbuf(R),R);
	 }
   } 
 return(0);
}

/*     
   \TEX[
   \title{Extension conversion}
   \input{Two matrices |ZENMat| and a possibly tabulated extension |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Matrix |M| is assigned with |m|.}
   \note{If the |ZENRing| is not an extension, this is equivalent to |ZENMatAssign|.}
   } 
   */
int _ZENCloneVect2Mat(m,M,R)
     ZENMat M,m;
     ZENRing R;
{int l,c,cc,k,n;
 ZENRing RO;

 if(ZENRingOrigin(R))
   RO=ZENRingOrigin(R);
 else
   RO=R;
 
 ZENTest(ZENRingDef(R)==NULL,
	 ZEN_NOT_EXT,ZEN_VECT2MAT,ZEN_LIB_ZEG, ZERR);

 ZENTest(ZENMatIsPermutation(M,R)
	 ||ZENMatIsPermutation(m,ZENRingDef(R)),
	 ZEN_ISPERM,ZEN_VECT2MAT,ZEN_LIB_ZEG, ZERR);
 
 n=ZENPolyDeg(ZENRingPol(R),ZENRingDef(R));
 ZENEltSetToZero(ZENRingJbuf(ZENRingDef(R)),ZENRingDef(R));
 
 if(ZENMatIsRowType(m,ZENRingDef(R)))
   {
     ZENTest((ZENMatNbRow(M,R) != ZENMatNbBn(m,ZENRingDef(R)))
	     ||(ZENMatNbCol(M,R)*n != ZENMatNbDigit(m,ZENRingDef(R))),
	     ZEN_MAT_DIM,ZEN_VECT2MAT,ZEN_LIB_ZEG, ZERR);
     
     for(l=0;l<ZENMatNbBn(m,ZENRingDef(R));l++)
       for(c=0,cc=0;c<ZENMatNbCol(M,R);c++)
	 {
	   ZENCloneToElt(ZENRingHbuf(RO),
			 ZENMatGetCoeffPtr(M,l,c,R),
			 R);
	   
	   for(k=0;n-k-1 > ZENPolyDeg(ZENElt2Pol(ZENRingHbuf(RO),
						 RO),
				      ZENRingDef(R));k++,cc++)
	     ZENMatSetBloc(m,l,cc,ZENRingJbuf(ZENRingDef(R)),
			   ZENRingDef(R));
	   
	   for(;k<n;k++,cc++)
	     ZENMatSetBloc(m,
			   l,
			   cc,
			   ZENPolyGetCoeffPtr(ZENElt2Pol(ZENRingHbuf(RO),
							 RO),
					      n-k-1,
					      ZENRingDef(R)),
			   ZENRingDef(R));
	 }
   }
 else
   {
     ZENTest((ZENMatNbRow(M,R)*n != ZENMatNbDigit(m,ZENRingDef(R)))
	     ||(ZENMatNbCol(M,R) != ZENMatNbBn(m,ZENRingDef(R))),
	     ZEN_MAT_DIM,ZEN_VECT2MAT,ZEN_LIB_ZEG, ZERR);
     
     for(c=0;c<ZENMatNbBn(m,ZENRingDef(R));c++)
       for(l=0,cc=0;l<ZENMatNbRow(M,R);l++)
	 {
	   ZENCloneToElt(ZENRingHbuf(RO),
			 ZENMatGetCoeffPtr(M,l,c,R),
			 R);
	   
	   for(k=0;n-k-1 > ZENPolyDeg(ZENElt2Pol(ZENRingHbuf(RO),
						 RO),
				      ZENRingDef(R));k++,cc++)
	     ZENMatSetBloc(m,c,cc,ZENRingJbuf(ZENRingDef(R)),
			   ZENRingDef(R));
	   
	   for(;k<n;k++,cc++)
	     ZENMatSetBloc(m,
			   c,
			   cc,
			   ZENPolyGetCoeffPtr(ZENElt2Pol(ZENRingHbuf(RO),
							 RO),
					      n-k-1,
					      ZENRingDef(R)),
			   ZENRingDef(R));
	 }
   }
 return(0);
}

