/*
 * zen/zeg/c/mat_gauss.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Gauss elimination}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions. 
   }
   */
#include <stdio.h>
#include "zeg.h"

#ifndef min
#  define min(x,y) ((x)<(y)?(x):(y))
#endif

/*
  \TEX{
   \title{General gaussian elimination}
   \input{Three matrices, a pointer on integer, a pointer on a |ZENElt| and on permutation function to use}
   \output{|ZENERR| if an error occurred,  |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned,
   |ZENMAT_HAS_MAXIMAL_RANK| if the matrix is of maximal rank,
   |ZENMAT_HAS_KERNEL| otherwise.}
   \side{Compute |S|, |P|, |p_rk| or |p_det| only if non |NULL| pointers.}
   \note{
   Permute is the operation to use to permute |D|.
   If |P| and |p_rk| are |NULL|, process stops as soon as the matrix is not of maximal rank.
   If |P| and |S| are |NULL|, trigonalization is performed.
   This function should never appear in a program. Use the encapsulated functions instead.
   }
   }
 */
int _ZENMatGauss(D,S,P,p_rk,p_det,Permute,R)
     ZENMat D,S,P;
     ZENElt *p_det;
     Dim *p_rk;
     int (*Permute) ___((ZENMat,int,int,ZENRing));
     ZENRing R;
{int k,l,c,flag;
 Dim ret,lim,nbbn;
 ZENElt z,g,h;

 nbbn=ZENMatNbBn(D,R)-(((int)(ZENMatType(D,R)))>>2); /* Tricky feature:
							not documented !
							I need that... */
 ZENTest(ZENMatIsPermutation(D,R),
	 ZEN_ISPERM,ZEN_MATGAUSS,ZEN_LIB_ZEG, ZERR);
 if(p_det) ZENEltSetToOne(*p_det,R);
 if(S)
   {
     ZENTest(ZENMatIsPermutation(S,R),
	     ZEN_ISPERM,ZEN_MATGAUSS,ZEN_LIB_ZEG, ZERR);
     ZENTest(!ZENMatAreSameType(S,D,R),
	     ZEN_MAT_TYPE,ZEN_MATGAUSS,ZEN_LIB_ZEG, ZERR);
     ZENTest((nbbn != ZENMatNbBn(S,R))
	     ||(ZENMatNbBn(S,R) != ZENMatNbDigit(S,R)),
	     ZEN_MAT_DIM,ZEN_MATGAUSS,ZEN_LIB_ZEG, ZERR);
   }
   
 if(P)
   {
     ZENTest(ZENMatAreSameType(P,D,R),
	     ZEN_MAT_TYPE,ZEN_MATGAUSS,ZEN_LIB_ZEG, ZERR);
     ZENTest((ZENMatNbDigit(D,R) != ZENMatNbBn(P,R))
	     ||(ZENMatNbDigit(P,R) != ZENMatNbBn(P,R)),
	     ZEN_MAT_DIM,ZEN_MATGAUSS,ZEN_LIB_ZEG, ZERR);
   }
 
 ZENCheckError(ZENEltCreate(z,R), ZERR);
 ZENCheckError(ZENEltCreate(g,R), ZERR);
 ZENCheckError(ZENEltCreate(h,R), ZERR);
 
#define FreeReturn(ret) if(1) { \
  ZENEltFree(g,R);                 \
  ZENEltFree(h,R);                 \
  ZENEltFree(z,R);                 \
  return(ret);                     \
  } else return(0)
  
 ret=ZENMatNbDigit(D,R);
 lim=min(nbbn,ret);
 flag=ZENMAT_HAS_MAXIMAL_RANK;
 
 for(k=0;k<lim;k++)
   {
     while(ZENEltIsZero(ZENMatGetBlocPtr(D,k,k,R),R))
       {
         l=k+1;
	 
         while(l<nbbn)
           if(ZENEltIsZero(ZENMatGetBlocPtr(D,l,k,R),R))
             l++;
           else
             {
               (*Permute)(D,k,l,R);
	       if(S) (*Permute)(S,k,l,R);
	       if(p_det) ZENEltNegate(*p_det,*p_det,R);
               goto OK;
             }

	 if((S)&&(!P))			/* GaussInverse */
	   {
	     ZENSetError(ZEN_LIB_ZEG,ZEN_MATGAUSS,ZEN_MAT_NOTINV);
	     FreeReturn(ZERR);
	   }

         if(--ret>k)
           {
	     if(ret<lim)
	       {
		 if(p_det) ZENEltSetToZero(*p_det,R);
		 if((!p_rk)&&(!P))		/* GaussTrigInvertible */
		   FreeReturn(ZENMAT_HAS_KERNEL);
		 flag=ZENMAT_HAS_KERNEL;
		 lim=ret;
	       }
             ZENMatPermuteDigit(D,k,ret,R);
             if(P) ZENMatPermuteBn(P,k,ret,R); /* ZEN fct for the case of permutation */
	   }
         else
	   {
	     if(p_det) ZENEltSetToZero(*p_det,R);
	     flag=ZENMAT_HAS_KERNEL;
	     goto FIN;
	   }
       }

     OK :
       switch(ZENEltInverse(h,ZENMatGetBlocPtr(D,k,k,R),R))
	 {
	 case ZEN_HAS_INVERSE:
	   break; 
	 case ZEN_NO_INVERSE:
	     if(p_rk) *p_rk= -1;
	   FreeReturn(ZEN_NO_INVERSE); 
	 default:
	   FreeReturn(ZERR);
	 }
     
     if(p_det) ZENEltMultiply(*p_det,ZENMatGetBlocPtr(D,k,k,R),*p_det,R);

	 
#define nop()
#define ___SubSComputation()                                            \
       for(c=0;c<ZENMatNbDigit(S,R);c++)                                  \
          { ZENEltMultiply(g,z,ZENMatGetBlocPtr(S,k,c,R),R);          \
            ZENEltAdd(g,ZENMatGetBlocPtr(S,l,c,R),R);                   \
	    ZENMatSetBloc(S,l,c,g,R); }                                 \
	 

#define ___Computation(foo)                                               \
     if(!(ZENEltIsZero(ZENMatGetBlocPtr(D,l,k,R),R)))                     \
       { ZENEltNegate(g,ZENMatGetBlocPtr(D,l,k,R),R);                     \
	 ZENEltMultiply(z,h,g,R);                                         \
 	 foo();                                                           \
	 ZENEltSetToZero(g,R);                                            \
	 ZENMatSetBloc(D,l,k,g,R);                                        \
         for(c=k+1;c<ZENMatNbBloc(D,R);c++)                              \
           if(!(ZENEltIsZero(ZENMatGetBlocPtr(D,k,c,R),R)))               \
             { ZENEltMultiply(g,z,ZENMatGetBlocPtr(D,k,c,R),R);         \
               ZENEltAdd(g,ZENMatGetBlocPtr(D,l,c,R),R);                  \
	       ZENMatSetBloc(D,l,c,g,R); }                                \
       } else

  if(S)
    for(l=k+1;l<ZENMatNbBn(D,R);l++)
      ___Computation(___SubSComputation);
  else
    for(l=k+1;l<ZENMatNbBn(D,R);l++)
      ___Computation(nop);
	
   }
   
FIN:
 if(p_rk) *p_rk=k;             
 lim=k;
#undef ___Computation
#define ___Computation(foo)                                               \
     if(!(ZENEltIsZero(ZENMatGetBlocPtr(D,l,k,R),R)))                     \
       { ZENEltNegate(z,ZENMatGetBlocPtr(D,l,k,R),R);                     \
 	 foo();                                                           \
	 ZENEltSetToZero(g,R);                                            \
	 ZENMatSetBloc(D,l,k,g,R);                                        \
         for(c=lim;c<ZENMatNbBloc(D,R);c++)                              \
           if(!(ZENEltIsZero(ZENMatGetBlocPtr(D,k,c,R),R)))               \
             { ZENEltMultiply(g,z,ZENMatGetBlocPtr(D,k,c,R),R);         \
               ZENEltAdd(g,ZENMatGetBlocPtr(D,l,c,R),R);                  \
	       ZENMatSetBloc(D,l,c,g,R); }                                \
       } else
	 
#define ___Normalization \
      ZENCheckError(ZENEltInverse(z,ZENMatGetBlocPtr(D,k,k,R),R),ZERR); \
      ZENEltSetToOne(h,R); \
      ZENMatSetBloc(D,k,k,h,R); \
      for(c=lim;c<ZENMatNbBloc(D,R);c++) { \
	ZENEltMultiply(g,z,ZENMatGetBlocPtr(D,k,c,R),R); \
        ZENMatSetBloc(D,k,c,g,R); }
	

 if(S)
   for(;k--;)
     {
       ___Normalization;
       for(c=0;c<ZENMatNbDigit(S,R);c++)
	 {
	   ZENEltMultiply(g,z,ZENMatGetBlocPtr(S,k,c,R),R);
	   ZENMatSetBloc(S,k,c,g,R);
	 }
       for(l=0;l<k;l++)
	 ___Computation(___SubSComputation);
     }
 else if(P)
   for(;k--;)
     {
       ___Normalization;
       for(l=0;l<k;l++)
	 ___Computation(nop);
     }
 FreeReturn(flag);
}


/*
   \TEX{
   \title{Gaussian elimination}
   \input{Three |ZENMat| over a |ZENRing|}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{A gausian elimination is performed on |D|, and the rank
   |*p_rk| of the matrix is set. |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   \note{
   \begin{itemize}
   \item
   If |D| is a $r\times c$ |ZENMat| of type |TypeRow|, then |S| must be
   a $r\times r$ |ZENMat| of type |TypeRow| and |P| a  $c\times c$ |ZENMat|
   of type |TypeCol|, or a |TypeColPermutation| |ZENMat| of size $c$.
   Then, $D$ is diagonalized on its rows and $S^{-1} D P^{-1}$ remains
   invariant.
   \item
   If |D| is a $r\times c$ |ZENMat| of type |TypeCol|, then |S| must be
   a $c\times c$ |ZENMat| of type |TypeCol| and |P| a  $r\times r$ |ZENMat|
   of type |TypeRow|, or a |TypeRowPermutation| |ZENMat| of size $r$.
   Then, $D$ is diagonalized on its columns and $P^{-1} D S^{-1} $ remains
   invariant.
   \end{itemize}
   }
   }
   */

int _ZENMatGaussPlain(D,S,P,p_rk,R)
     ZENMat D,S,P;
     Dim *p_rk;
     ZENRing R;
{
  switch(ZENMatGauss(D,S,P,p_rk,NULL,ZENMatPermuteBnPtr(R),R))
    {
    case ZENERR:
      return(ZERR);
    case ZEN_NO_INVERSE:
      return(ZEN_NO_INVERSE);
    default:
      return(ZEN_HAS_INVERSE);
    }
}
/*
   \TEX{
   \title{Gaussian elimination}
   \input{Two |ZENMat| over a |ZENRing|}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{A gausian elimination is performed on |D|, and the rank
   |*p_rk| of the matrix is set. |ZENRingFact(Rg)| is filled with a
   factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   \note{
   \begin{itemize}
   \item
   If |D| is a $r\times c$ |ZENMat| of type |TypeRow|, then |P| must be a  $c\times c$ |ZENMat|
   of type |TypeCol|, or a |TypeColPermutation| |ZENMat| of size $c$.
   Then, $D$ is diagonalized on its rows.
   \item
   If |D| is a $r\times c$ |ZENMat| of type |TypeCol|, then |P| must be a  $r\times r$ |ZENMat|
   of type |TypeRow|, or a |TypeRowPermutation| |ZENMat| of size $r$.
   Then, $D$ is diagonalized on its columns.
   \end{itemize}
   }
   }
   */

int _ZENMatGaussLazy(D,P,p_rk,R)
     ZENMat D,P;
     Dim *p_rk;
     ZENRing R;
{
  switch(ZENMatGauss(D,NULL,P,p_rk,NULL,ZENMatPermuteBnPtr(R),R))
    {
    case ZENERR:
      return(ZERR);
    case ZEN_NO_INVERSE:
      return(ZEN_NO_INVERSE);
    default:
      return(ZEN_HAS_INVERSE);
    }
}

/*
   \TEX{
   \title{Gaussian elimination}
   \input{Two |ZENMat| over a |ZENRing|}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{A gausian elimination is performed on |D|, and the rank
   |*p_rk| of the matrix is set.  |ZENRingFact(Rg)| is filled with a
   factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   \note{
   \begin{itemize}
   \item
   If |D| is a $r\times c$ |ZENMat| of type |TypeRow|, then |S| must be
   a $r\times r$ |ZENMat| of type |TypeRow|.
   \item
   If |D| is a $r\times c$ |ZENMat| of type |TypeCol|, then |S| must be
   a $c\times c$ |ZENMat| of type |TypeCol|.
   \end{itemize}
   This function provides faster inversion assuming the matrices square, of same type and invertible, but |ZENMatFastPermute| is used.
   }
   }
   */

int _ZENMatGaussInverse(D,S,R)
     ZENMat D,S;
     ZENRing R;
{
 switch(ZENMatGauss(D,S,NULL,NULL,NULL,_ZENMatFastPermuteBn,R))
   {
   case ZENERR:
     return(ZERR);
   case ZEN_NO_INVERSE:
     return(ZEN_NO_INVERSE);
   default:
     return(ZEN_HAS_INVERSE);
   }
}

/*
   \TEX{
   \title{Rank of a matrix}
   \input{A |ZENMat|, a pointer to a dimension and a |ZENRing|.}
   \output{|ZENERR| if an error occurred,  |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned,
   |ZENMAT_HAS_MAXIMAL_RANK| if the matrix is of maximal rank,
   |ZENMAT_HAS_KERNEL| otherwise.}
   \side{|*p_rk| is set to the rank of the matrix $D$.}
   \note{|D| remained unchanged}
   } 
   */
int _ZENMatRank(D,p_rk,R)
     ZENMat D;
     int *p_rk;
     ZENRing R;
{ZENMat M;
 int ret;

 M=ZENMatCopy(D,R);
 ZENTest(M==NULL,
	 ZEN_ALLOC,ZEN_MATRANK,ZEN_LIB_ZEG, ZERR);

 ret=ZENMatGaussRank(M,p_rk,R);
 ZENMatFree(M,R);
 return(ret);
}

/*
   \TEX{
   \title{Determinant of a matrix}
   \input{A |ZENMat|, a pointer to an allocated |ZENElt| and a |ZENRing|.}
   \output{|ZENERR| if an error occurred,  |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned,
   |ZENMAT_HAS_MAXIMAL_RANK| if the matrix is of maximal rank,
   |ZENMAT_HAS_KERNEL| otherwise.}
   \side{|*p_det| is set to the determinant of the matrix $D$.}
   \note{|D| remained unchanged}
   } 
   */
int _ZENMatDet(D,p_det,R)
     ZENMat D;
     ZENElt *p_det;
     ZENRing R;
{ZENMat M;
 int ret;

 M=ZENMatCopy(D,R);
 ZENTest(M==NULL,
	 ZEN_ALLOC,ZEN_MATRANK,ZEN_LIB_ZEG, ZERR);

 switch(ZENMatGaussDet(M,p_det,R))
   {
   case ZENERR:
     ret=ZERR;
     break;
   case ZEN_NO_INVERSE:
     ret=ZEN_NO_INVERSE;
     break;
   default:
     ret=ZEN_HAS_INVERSE;
     break;
   }
 ZENMatFree(M,R);
 return(ret);
}

