/*
 * zen/zeg/c/mat_inverse.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matrix inversion}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions. 
   }
   */
#include <stdio.h>
#include "zeg.h"

#ifndef min
#  define min(x,y) ((x)<(y)?(x):(y))
#endif

#define FreeReturn(ret) if(1) {    \
  ZENMatFree(S,R);                 \
  ZENMatFree(D,R);                 \
  ZENEltFree(z,R);                 \
  ZENEltFree(n,R);                 \
  return(ret);                     \
  } else return(0)



#define ZENCheckInverse(X,A,R) switch(ZENEltInverse(X,A,R)) { \
	case ZEN_HAS_INVERSE: break; \
	case ZEN_NO_INVERSE: FreeReturn(ZEN_NO_INVERSE); \
	default: ZENTest(ZENErrErrno == ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_MATINVERSE,ZEN_LIB_ZEG, ZERR); FreeReturn(ZERR);}

#define ZENCheckGaussInverse(X,A,R) switch(ZENMatGaussInverse(X,A,R)) { \
	case ZEN_HAS_INVERSE: break; \
	case ZEN_NO_INVERSE: FreeReturn(ZEN_NO_INVERSE); \
	default: ZENTest(ZENErrErrno == ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_MATINVERSE,ZEN_LIB_ZEG, ZERR); FreeReturn(ZERR);}


/*
   \TEX{
   \title{Matrix inversion}
   \input{Two |ZENMat| over a |ZENRing|}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|I|=|M|^{-1}$. If an inverse was impossible to compute,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */

int __ZENMatInverse(I,M,R)
     ZENMat I,M;
     ZENRing R;
{
  if(ZENMatIsPermutation(M,R))
    {
      ZENMatTranspose(M,R);
      ZENCheckError(ZENMatAssign(I,M,R),ZERR);
      ZENMatTranspose(M,R);
      return(0);
    }
  return(_R_ZENMatInverse(I,M,R));
}

/*
   \TEX{
   \title{Matrix inversion}
   \input{Two non permutation |ZENMat| over a |ZENRing|}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|I|=|M|^{-1}$. If an inverse was impossible to compute,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */

int _ZENMatInverse(I,M,R)
     ZENMat I,M;
     ZENRing R;
{ZENMat D,S;
 int j,k;
 ZENElt z,n;
 
 ZENTest(ZENMatIsPermutation(I,R)
	 ||ZENMatIsPermutation(M,R),
	 ZEN_ISPERM,ZEN_MATINVERSE,ZEN_LIB_ZEG, ZERR);
 ZENTest((ZENMatNbBn(I,R) != ZENMatNbBn(M,R))
	 ||(ZENMatNbDigit(M,R) != ZENMatNbBn(M,R))
	 ||(ZENMatNbDigit(I,R) != ZENMatNbBn(M,R)),
	 ZEN_MAT_DIM,ZEN_MATINVERSE,ZEN_LIB_ZEG, ZERR);
 
 ZENCheckError(ZENEltCreate(z,R), ZERR);
 ZENCheckError(ZENEltCreate(n,R), ZERR);

 if(!(D=ZENMatCopy(M,R)))
   return(ZERR);
 
 if(!(S=(ZENMatIsRowType(D,R)? ZENMatRowCreate(ZENMatNbBn(D,R),ZENMatNbBn(D,R),R) : ZENMatColCreate(ZENMatNbBn(D,R),ZENMatNbBn(D,R),R))))
   return(ZERR);
 
 ZENCheckError(ZENMatSetToOne(S,R), ZERR);
 ZENCheckGaussInverse(D,S,R);

 if(ZENMatAreSameType(I,D,R))
   for(k=ZENMatNbDigit(D,R)-1;k>=0;k--)
     {
       ZENCheckInverse(n,ZENMatGetBlocPtr(D,k,k,R),R);
       for(j=ZENMatNbBn(D,R)-1;j>=0;j--)
	 {
	   ZENEltMultiply(z,n,ZENMatGetBlocPtr(S,k,j,R),R);
	   ZENMatSetBloc(I,k,j,z,R);
	 }
     }
 else
   for(k=ZENMatNbDigit(D,R)-1;k>=0;k--)
     {
       ZENCheckInverse(n,ZENMatGetBlocPtr(D,k,k,R),R);
       for(j=ZENMatNbBn(D,R)-1;j>=0;j--)
	 {
	   ZENEltMultiply(z,n,ZENMatGetBlocPtr(S,k,j,R),R);
	   ZENMatSetBloc(I,j,k,z,R);
	 }
     }
#undef ZENCheckInverse
#undef ZENCheckGaussInverse

 FreeReturn(ZEN_HAS_INVERSE);
#undef FreeReturn
 
}

#define FreeReturn(ret) if(1) {    \
  ZENMatFree(P,R);                 \
  ZENMatFree(D,R);                 \
  ZENEltFree(z,R);                 \
  ZENEltFree(I,R);                 \
  return(ret);                     \
  } else return(0)


#define ZENCheckInverse(X,A,R) switch(ZENEltInverse(X,A,R)) { \
	case ZEN_HAS_INVERSE: break; \
	case ZEN_NO_INVERSE: FreeReturn(ZEN_NO_INVERSE); \
	default: ZENTest(ZENErrErrno == ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_MATINVERSE,ZEN_LIB_ZEG, ZERR); FreeReturn(ZERR);}

/*
   \TEX{
   \title{Matrix kernel}
   \input{A pointer on an unallocated |ZENMat| and a |ZENMat| over a |ZENRing|}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{|*p_K| is allocated and set to a basis of the kernel of |M|.
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */

int _ZENMatKernel(p_K,M,R)
     ZENMat *p_K,M;
     ZENRing R;
{
  int n,rk,i,j;
  Dim l;
  ZENMat D,P;
  ZENElt z,I;
  
  n=ZENMatNbCol(M,R);
  
  if(!(D=ZENMatRowCreate(ZENMatNbRow(M,R),n,R)))
    return(ZERR);
  
  ZENCheckError(ZENMatAssign(D,M,R), ZERR);
  
  if(!(P=ZENPermutationColCreate(n,R)))
    return(ZERR);
  
  ZENCheckError(ZENEltCreate(z,R), ZERR);
  ZENCheckError(ZENEltCreate(I,R), ZERR);
  
  switch(ZENMatGaussLazy(D,P,&rk,R))
    {
    case ZEN_HAS_INVERSE:
      break; 
    case ZEN_NO_INVERSE:
      FreeReturn(ZEN_NO_INVERSE);
      
    default:
      ZENTest(ZENErrErrno == ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_MATKERNEL,ZEN_LIB_ZEG, ZERR);
      FreeReturn(ZERR);
    }
  
#define dim (n<=rk ? 1 : n-rk)
  
  if(!(*p_K=(ZENMatIsRowType(M,R) ? ZENMatRowCreate(dim,n,R) : ZENMatColCreate(dim,n,R))))
    return(ZERR);
  
  ZENCheckError(ZENMatSetToZero(*p_K,R), ZERR);
  
  for(i=0;i<n-rk;i++)
    for(j=0;j<n;j++)
      {
	l=ZENPermutationBn(P,j,R);
	if(j<rk)
	  {
	    ZENCheckInverse(I,ZENMatGetBlocPtr(D,j,j,R),R);
	    ZENEltNegate(z,I,R);
	    ZENEltMultiply(I,z,ZENMatGetCoeffPtr(D,j,i+rk,R),R);
	    ZENMatSetCoeff(*p_K,i,l,I,R);
	  }
	else if (j==rk+i)
	  {
	    ZENEltSetToOne(I,R);
	    ZENMatSetCoeff(*p_K,i,l,I,R);
	  }
      }
  
  FreeReturn(ZEN_HAS_INVERSE);
  
}
