/*
 * zen/zeg/c/mat_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   \Subsection{input/output}
   
   Conversions between the internal representation and the ASCII
   representation of a |ZENMat| are done by
   |_ZENMatPrintToString|. |_ZENMatPrintToFile| allows to write a
   |ZENMat| in a file. One can moreover write to and then read from a
   file a |_ZENMat| with |_ZENMatPutToFile| and |_ZENMatReadFromFile|.
   }
   */
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "zen.h"


/* \TEX[
   \title{Converting from string.}
   \input{A finite field |R|, a pointer on |ZENMat| |M| and a string
   |s| representing a |ZENMat| in base |base|$\in \{2,\ldots,
   16\}$.} 
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|M| is allocated and filled with |s| if the output is not |ZENERR|.}
   }
*/
int _ZENMatReadFromString(M, s, base, Rg)
     ZENMat *M;
     char *s;
     int base;
     ZENRing Rg;
{
  ZENElt elt;
  char *ptr;
  int l,c,nl,nc,nbc;

  /* Alocating a buffer */
  if (ZENEltCreate(elt, Rg)) {
    return(ZERR);
  }
  ptr = s;
  
  /* Parsing */
  for(; isspace(*ptr); ptr++) ;
  if(strncmp(ptr,"array",5)) { /* array */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
    
  for(ptr+=5; isspace(*ptr); ptr++) ;
  
  if (*ptr != '(') {		/* array( */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  
  for(ptr++; isspace(*ptr); ptr++) ;
  if (*ptr != '1') {		/* array(1 */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  
  for(ptr++; isspace(*ptr); ptr++) ;
  if (*ptr != '.') {		/* array(1. */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  ptr++;
  
  if (*ptr != '.') {		/* array(1.. */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  
  for(ptr++; isspace(*ptr); ptr++) ;
  
  if (sscanf(ptr, "%d", &nl) != 1) { /* array(1..nl */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  
  for(ptr++; isdigit(*ptr); ptr++) ;
  for(; isspace(*ptr); ptr++) ;
  
  if (*ptr != ',') {		/* array(1..nl, */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  
  for(ptr++; isspace(*ptr); ptr++) ;
  if (*ptr != '1') {		/* array(1..nl,1 */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  
  for(ptr++; isspace(*ptr); ptr++) ;
  if (*ptr != '.') {		/* array(1..nl,1. */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  ptr++;
  
  if (*ptr != '.') {		/* array(1..nl,1.. */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  
  for(ptr++; isspace(*ptr); ptr++) ;
  
  if (sscanf(ptr, "%d", &nc) != 1) { /* array(1..nl,1..nc */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  
  for(ptr++; isdigit(*ptr); ptr++) ;
  for(; isspace(*ptr); ptr++) ;
  
  if (*ptr != ',') {		/* array(1..nl,1..nc, */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  
  for(ptr++; isspace(*ptr); ptr++) ;
  
  if (*ptr != '[') {		/* array(1..nl,1..nc,[ */
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  
  
  if(!(*M=ZENMatCreate(nl,nc,Rg))) {
    ZENEltFree(elt, Rg);
    return(ZERR);
  }

  for(l=0;l<nl;l++)
    {
      if(l)
	{
	  for(ptr++; isspace(*ptr); ptr++) ;
	  
	  if (*ptr != ',') {		
	    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
	    ZENEltFree(elt, Rg);
	    ZENMatFree(*M,Rg);
	    return(ZERR);
	  }
	  
	}
      
      for(ptr++; isspace(*ptr); ptr++) ;
      
      if (*ptr != '[') {		
	ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
	ZENEltFree(elt, Rg);
	ZENMatFree(*M,Rg);
	return(ZERR);
      }
      
      for(c=0;c<nc;c++)
	{
	  if(c)
	    {
	      for(ptr++; isspace(*ptr); ptr++) ;
	      
	      if (*ptr != ',') {		
		ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		ZENEltFree(elt, Rg);
		ZENMatFree(*M,Rg);
		return(ZERR);
	      }
	    }
	  
	  ptr++;
	  
	  if ((nbc = ZENEltReadFromString(elt, ptr, base, Rg)) < 0) {
	    ZENEltFree(elt, Rg);
	    ZENMatFree(*M,Rg);
	    return(ZERR);
	  }
	  
	  if(ZENMatSetCoeff(*M,l,c,elt,Rg)) {
	    ZENEltFree(elt, Rg);
	    ZENMatFree(*M,Rg);
	    return(ZERR);
	  }
	  ptr+=nbc-1;
	}
      
      for(ptr++; isspace(*ptr); ptr++) ;
      
      if (*ptr != ']') {		
	ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
	ZENEltFree(elt, Rg);
	ZENMatFree(*M,Rg);
	return(ZERR);
      }
      
    }
  
  for(ptr++; isspace(*ptr); ptr++) ;
  
  if (*ptr != ']') {		
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    ZENMatFree(*M,Rg);
    return(ZERR);
  }
  for(ptr++; isspace(*ptr); ptr++) ;
  
  if (*ptr != ')') {		
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    ZENMatFree(*M,Rg);
    return(ZERR);
  }
  for(ptr++; isspace(*ptr); ptr++) ;
  
  if (*ptr != ';') {		
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    ZENMatFree(*M,Rg);
    return(ZERR);
  }
  ptr++;
  
  ZENEltFree(elt, Rg);
  return(ptr-s);
}

/* \TEX[
   \title{Converting to string.}
   \input{An allocated |ZENMat| |G| of a finite field |R| and a base |base|$\in \{2,\ldots, 16\}$.}
   \output{An allocated string representing |G| in base |base| or NULL if an error occured.}
   }
   */
char *_ZENMatPrintToString(M, base, Rg)
     ZENMat M;
     int base;
     ZENRing Rg;  
{  
  int l,c, nl,nc;
  char *str, *s1, *s2;
  ZENElt elt;
  
  nl=ZENMatNbRow(M,Rg);
  nc=ZENMatNbCol(M,Rg);
  
  if (!(str = (char *)ZENMalloc(sizeof(char)*256))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_ALLOC);
    return(NULL);
  }
  if (sprintf(str,"array(1..%d,1..%d,\n[",nl,nc)<0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_SPRINTF);
    ZENFree(str); 
    return(NULL);
  }

  if (ZENEltCreate(elt, Rg)) {
    ZENFree(str); 
    return(NULL);
  }

  for(l=0;l<nl;l++)
    {
      if(l)
	{
	  s2 = str;
	  if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+4)))) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_ALLOC);
	    ZENEltFree(elt, Rg);
	    ZENFree(s2);
	    return(NULL);
	  }
	  if (sprintf(str, "%s,\n ", s2) < 0) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_SPRINTF);
	    ZENFree(str); ZENFree(s2);
	    ZENEltFree(elt, Rg);
	    return(NULL);
	  }
	  ZENFree(s2);
	}
      
      s2 = str;
      if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+2)))) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_ALLOC);
	ZENFree(s2);
	ZENEltFree(elt, Rg);
	return(NULL);
      }
      if (sprintf(str, "%s[", s2) < 0) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_SPRINTF);
	ZENFree(str); ZENFree(s2);
	ZENEltFree(elt, Rg);
	return(NULL);
      }
      ZENFree(s2);

      for(c=0;c<nc;c++)
	{
	  if(c)
	    {
	      s2 = str;
	      if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+3)))) {
		ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_ALLOC);
		ZENFree(s2);
		ZENEltFree(elt, Rg);
		return(NULL);
	      }
	      if (sprintf(str, "%s, ", s2) < 0) {
		ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_SPRINTF);
		ZENFree(str); ZENFree(s2);
		ZENEltFree(elt, Rg);
		return(NULL);
	      }
	      ZENFree(s2);
	    }
	  
	  s1 = str;

	  if (ZENMatGetCoeff(elt,M,l,c,Rg)) {
	    ZENEltFree(elt, Rg);
	    ZENFree(str);
	    return(NULL);
	  }
	  
	  if (!(s2 = ZENEltPrintToString(elt, base, Rg))) {
	    ZENEltFree(elt, Rg);
	    ZENFree(str);
	    return(NULL);
	  }
	  
	  if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(s1)+strlen(s2)+1)))) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_ALLOC);
	    ZENFree(s1); ZENFree(s2);
	    ZENEltFree(elt, Rg);
	    return(NULL);
	  }
	  if (sprintf(str, "%s%s", s1,s2) < 0) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_SPRINTF);
	    ZENFree(str); ZENFree(s1); ZENFree(s2);
	    ZENEltFree(elt, Rg);
	    return(NULL);
	  }
	  
	  ZENFree(s1);
	  ZENFree(s2);
	  
	}
      s2 = str;
      if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+2)))) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_ALLOC);
	ZENFree(s2);
	ZENEltFree(elt, Rg);
	return(NULL);
      }
      if (sprintf(str, "%s]", s2) < 0) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_SPRINTF);
	ZENFree(str); ZENFree(s2);
	ZENEltFree(elt, Rg);
	return(NULL);
      }
      ZENFree(s2);

    }
  s2 = str;
  ZENEltFree(elt, Rg);
	    
  if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+5)))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_ALLOC);
    ZENFree(s2);
    return(NULL);
  }
  if (sprintf(str, "%s]);\n", s2) < 0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_SPRINTF);
    ZENFree(str); ZENFree(s2);
    return(NULL);
  }
  ZENFree(s2);

  return(str);
}


/* \TEX[
   \title{Printing to file.}
   \input{A stream |file|, an allocated |ZENMat| |M| of a finite field
   |Rg| and a base |base|$\in \{2,\ldots, 16\}$.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |M| in base |base| to |file|.}    
   }
   */
int _ZENMatPrintToFile(file, M, base, Rg)
     FILE *file;
     ZENMat M;
     int base;
     ZENRing Rg;  
{ 
  int l,c, nl,nc;
  ZENElt elt;
  
  nl=ZENMatNbRow(M,Rg);
  nc=ZENMatNbCol(M,Rg);
  
  if(fprintf(file,"array(1..%d,1..%d,\n[",nl,nc)<0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOFILE, ZEN_FPRINTF);
    return(ZERR);
  }
  
  if (ZENEltCreate(elt, Rg))
    return(ZERR);
  
  for(l=0;l<nl;l++)
    {
      if(l)
	{
	  if (fprintf(file, ",\n ") < 0) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_FPRINTF);
	    ZENEltFree(elt, Rg);
	    return(ZERR);
	  }
	}
      
      if (fprintf(file, "[") < 0) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_FPRINTF);
	ZENEltFree(elt, Rg);
	return(ZERR);
      }
      

      for(c=0;c<nc;c++)
	{
	  if(c)
	    {
	      if (fprintf(file, ", ") < 0) {
		ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_FPRINTF);
		ZENEltFree(elt, Rg);
		return(ZERR);
	      }
	      
	    }
	  
	  if (ZENMatGetCoeff(elt,M,l,c,Rg)) {
	    ZENEltFree(elt, Rg);
	    return(ZERR);
	  }
	  
	  if (ZENEltPrintToFile(file,elt, base, Rg)) {
	    ZENEltFree(elt, Rg);
	    return(ZERR);
	  }
	  
	}
      if (fprintf(file, "]") < 0) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_FPRINTF);
	ZENEltFree(elt, Rg);
	return(ZERR);
      }
      

    }
  ZENEltFree(elt, Rg);
  
  if (fprintf(file, "]);\n") < 0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPRINTTOSTRING, ZEN_FPRINTF);
    return(ZERR);
  }
  return(0);
}

/* \TEX[
   \title{Reading from file.}
   \input{A stream |file| and a pointer on a matnomial |M| of a finite field |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|M| is allocated and filled with the |ZENMat| read in |file| if
   no error occurred.}
   }
   */

  int _ZENMatReadFromFile(M, file, base, Rg)
     ZENMat *M;
     FILE *file;
     int base;
     ZENRing Rg;
{
  ZENElt elt;
  char ch;
  int l,c,nl,nc;

  /* Alocating a buffer */
  if (ZENEltCreate(elt, Rg)) {
    return(ZERR);
  }
  ch=fgetc(file);
  
  for(; isspace(ch); ch=fgetc(file)) ;
  if (ch != 'a') {
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  ch=fgetc(file);
  if (ch != 'r') {
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  ch=fgetc(file);
  if (ch != 'r') {
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  ch=fgetc(file);
  if (ch != 'a') {
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  ch=fgetc(file);
  if (ch != 'y') {
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  
  for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
  
  if (ch != '(') {		/* array( */
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  
  for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
  if (ch != '1') {		/* array(1 */
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  
  for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
  if (ch != '.') {		/* array(1. */
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  ch=fgetc(file);
  
  if (ch != '.') {		/* array(1.. */
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  
  if (fscanf(file, "%d", &nl) != 1) { /* array(1..nl */
				      ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
				      ZENEltFree(elt, Rg);
				      return(ZERR);
				    }
  
  for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
  
  if (ch != ',') {		/* array(1..nl, */
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  
  for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
  if (ch != '1') {		/* array(1..nl,1 */
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  
  for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
  if (ch != '.') {		/* array(1..nl,1. */
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  ch=fgetc(file);
  
  if (ch != '.') {		/* array(1..nl,1.. */
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  
  if (fscanf(file, "%d", &nc) != 1) { /* array(1..nl,1..nc */
				      ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
				      ZENEltFree(elt, Rg);
				      return(ZERR);
				    }
  
  for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
  
  if (ch != ',') {		/* array(1..nl,1..nc, */
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  
  for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
  
  if (ch != '[') {		/* array(1..nl,1..nc,[ */
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   return(ZERR);
		 }
  
  
  if(!(*M=ZENMatCreate(nl,nc,Rg))) {
				     ZENEltFree(elt, Rg);
				     return(ZERR);
				   }
  
  for(l=0;l<nl;l++)
    {
      if(l)
	{
	  for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;

	  if (ch != ',') {		
			   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
			   ZENEltFree(elt, Rg);
			   ZENMatFree(*M,Rg);
			   return(ZERR);
			 }
	    
	}
    
      for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
	
      if (ch != '[') {		
		       ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		       ZENEltFree(elt, Rg);
		       ZENMatFree(*M,Rg);
		       return(ZERR);
		     }
	
      for(c=0;c<nc;c++)
	{
	  if(c)
	    {
	      for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
		
	      if (ch != ',') {		
			       ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
			       ZENEltFree(elt, Rg);
			       ZENMatFree(*M,Rg);
			       return(ZERR);
			     }
	    }
	    
	  if (ZENEltReadFromFile(elt, file, base, Rg)) {
							 ZENEltFree(elt, Rg);
							 ZENMatFree(*M,Rg);
							 return(ZERR);
						       }
	    
	  if(ZENMatSetCoeff(*M,l,c,elt,Rg)) {
					      ZENEltFree(elt, Rg);
					      ZENMatFree(*M,Rg);
					      return(ZERR);
					    }
	    
	}
	
      for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
	
      if (ch != ']') {		
		       ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		       ZENEltFree(elt, Rg);
		       ZENMatFree(*M,Rg);
		       return(ZERR);
		     }
	
    }
  
  for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
  
  if (ch != ']') {		
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   ZENMatFree(*M,Rg);
		   return(ZERR);
		 }
  for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
  
  if (ch != ')') {		
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   ZENMatFree(*M,Rg);
		   return(ZERR);
		 }
  for(ch=fgetc(file); isspace(ch); ch=fgetc(file)) ;
  
  if (ch != ';') {		
		   ZENSetError(ZEN_LIB_ZEG, ZEN_MATREADFROMSTRING, ZEN_SYNTAX);
		   ZENEltFree(elt, Rg);
		   ZENMatFree(*M,Rg);
		   return(ZERR);
		 }
  
  ZENEltFree(elt, Rg);
  return(0);
}


/* \TEX[
   \title{Converting from a string in an internal representation.}
   \input{A finite field |Rg| and a string |s| representing a |ZENMat|
   in an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|M| is allocated and filled with s if the output is different
   from |ZENERR|}	    
   }
   */
int _ZENMatGetFromString(M, s, Rg)
     char *s;
     ZENMat *M;
     ZENRing Rg;
{
  ZENElt elt;
  char *ptr;
  int l,c,nl,nc,nbc;

  /* Alocating a buffer */
  if (ZENEltCreate(elt, Rg)) {
    return(ZERR);
  }
  ptr = s;
  
  if (sscanf(ptr,"%8x%8x", &nl,&nc) != 2) { 
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATGETFROMSTRING, ZEN_SYNTAX);
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  for(; isspace(*ptr); ptr++) ;
  ptr += 16;
  
  if(!(*M=ZENMatCreate(nl,nc,Rg))) {
    ZENEltFree(elt, Rg);
    return(ZERR);
  }

  for(l=0;l<nl;l++)
    for(c=0;c<nc;c++)
      {
	if ((nbc = ZENEltGetFromString(elt, ptr, Rg)) < 0) {
	  ZENEltFree(elt, Rg);
	  ZENMatFree(*M,Rg);
	  return(ZERR);
	}
	    
	ZENMatSetCoeff(*M,l,c,elt,Rg);
	ptr+=nbc;
      }
	
    ZENEltFree(elt, Rg);
    return(ptr-s);
}


/* \TEX[
   \title{Converting to string in an internal representation.}
   \input{An allocated |ZENMat| |M| of a finite field |Rg|.}
   \output{An allocated string representing |G| in an internal representation or NULL if an error occured.}   
   }
   */
char *_ZENMatPutToString(M, Rg)
     ZENMat M;
     ZENRing Rg;
{
  int l,c, nl,nc;
  char *str, *s1, *s2;
  ZENElt elt;
  
  nl=ZENMatNbRow(M,Rg);
  nc=ZENMatNbCol(M,Rg);
  
  if (!(str = (char *)ZENMalloc(sizeof(char)*256))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPUTTOSTRING, ZEN_ALLOC);
    return(NULL);
  }
  if (sprintf(str,"%.8x%.8x",nl,nc)<0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPUTTOSTRING, ZEN_SPRINTF);
    ZENFree(str);
    return(NULL);
  }

  if (ZENEltCreate(elt, Rg)) {
    ZENFree(str);
    return(NULL);
  }

  for(l=0;l<nl;l++)
    for(c=0;c<nc;c++)
      {
	s1 = str;
	
	if (ZENMatGetCoeff(elt,M,l,c,Rg)) {
	    ZENEltFree(elt, Rg);
	    ZENFree(str);
	    return(NULL);
	  }
	  
	if (!(s2 = ZENEltPutToString(elt, Rg))) {
	  return(NULL);
	}
	if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(s1)+strlen(s2)+1)))) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_MATPUTTOSTRING, ZEN_ALLOC);
	  ZENFree(s1); ZENFree(s2);
	  return(NULL);
	}
	if (sprintf(str, "%s%s", s1,s2) < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_MATPUTTOSTRING, ZEN_SPRINTF);
	  ZENFree(str); ZENFree(s1); ZENFree(s2);
	  return(NULL);
	}
	
	ZENFree(s1);
	ZENFree(s2);
      }
  ZENEltFree(elt,Rg);
  return(str);
}


/* \TEX[
   \title{Getting from file in an internal representation.}
   \input{A stream |file|, and a pointer on matnomial |M| of a finite
   field |ZENMat|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|*M| is filled with the matnomial read in |file|}    
   }
   */
int _ZENMatGetFromFile(M, file, Rg)
     ZENMat *M;
     FILE *file;
     ZENRing Rg;
{
  int l,c,nl,nc;
  ZENElt elt;
  
  if (fscanf(file, "%8x%8x", &nl,&nc) != 2) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATGETFROMFILE, ZEN_SSCANF);
    return(ZERR);
  }

  if (ZENEltCreate(elt, Rg)) 
    return(ZERR);
  
  if(!(*M=ZENMatCreate(nl,nc,Rg))) {
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  
  for(l=0;l<nl;l++)
    for(c=0;c<nc;c++)
      {
	if (ZENEltGetFromFile(elt,file, Rg)) {
	  ZENEltFree(elt, Rg);
	  ZENMatFree(*M, Rg);
	  return(ZERR);
	}
	ZENMatSetCoeff(*M,l,c,elt,Rg);
      }
  ZENEltFree(elt, Rg);

  return(0);
}

/* \TEX[
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |ZENMat| |M| of a finite field
   |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |M| to |file|.}    
   }
   */
int _ZENMatPutToFile(file, M, Rg)
     FILE *file;
     ZENMat M;
     ZENRing Rg;  
{
  int l,c,nl,nc;
  ZENElt elt;
  
  nl=ZENMatNbRow(M,Rg);
  nc=ZENMatNbCol(M,Rg);
  
  if (fprintf(file, "%.8x%.8x", nl,nc) < 0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_MATPUTTOFILE, ZEN_FPRINTF);
    return(ZERR);
  }
  
  if (ZENEltCreate(elt, Rg))
    return(ZERR);

  for(l=0;l<nl;l++)
    for(c=0;c<nc;c++)
      {
	if (ZENMatGetCoeff(elt,M,l,c,Rg)) {
	  ZENEltFree(elt, Rg);
	  return(ZERR);
	}
	
	if (ZENEltPutToFile(file, elt, Rg)) {
	  ZENEltFree(elt, Rg);
	  return(ZERR);
	}
      }

  ZENEltFree(elt,Rg);
	
  return(0);
}
  
