/*
 * zen/zeg/c/mat_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial multiplications}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions. 
   }
   */
#include <stdio.h>
#include "zen.h"
/*
   \TEX{
   \title{Multiplication.}
   \input{Three |ZENMat| |X|, |A| and |B| of a finite ring |R|.}
   \output{|ZENERR| if an error occurred, 0 otherwise}
   \side{$|X|= |A| \times |B|$.}
   \note{Algorithm used is the general one. This procedure can call |MultiplyPermutations| |MultiplyRightPermutation| or |MultiplyLeftPermutation| if needed.}
   }
   */
   
int _ZENMatMultiply(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{
  if(ZENMatIsPermutation(X,R))
    return(ZENMatMultiplyPermutations(X,A,B,R));
  else if(ZENMatIsPermutation(A,R))
    return(ZENMatMultiplyLeftPermutation(X,A,B,R));
  else if(ZENMatIsPermutation(B,R))
    return(ZENMatMultiplyRightPermutation(X,A,B,R));
  else
    return(_R_ZENMatMultiply(X,A,B,R));
}

/*
   \TEX{
   \title{Multiplication.}
   \input{Three |ZENMat| |X|, |A| and |B| of a finite ring |R|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{$|X|= |A| \times |B|$.}
   \note{Algorithm used is the general one.}
   }
   */
   
int _ZENMatMultiplyPlain(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{int l,c,k;
 ZENElt S,g;

 ZENTest(ZENMatIsPermutation(A,R)||ZENMatIsPermutation(B,R)||ZENMatIsPermutation(X,R),
	 ZEN_ISPERM,ZEN_MATMULTIPLY,ZEN_LIB_ZEG, ZERR);
 ZENTest((ZENMatNbRow(A,R) != ZENMatNbRow(X,R))
	 ||(ZENMatNbCol(A,R) != ZENMatNbRow(B,R))
	 ||(ZENMatNbCol(B,R) != ZENMatNbCol(X,R)),
	 ZEN_MAT_DIM,ZEN_MATMULTIPLY,ZEN_LIB_ZEG, ZERR);

 ZENCheckError(ZENEltCreate(S,R), ZERR);
 ZENCheckError(ZENEltCreate(g,R), ZERR);

 for(l=ZENMatNbRow(X,R)-1;l>=0;l--)
   for(c=ZENMatNbCol(X,R)-1;c>=0;c--)
     {
       k=ZENMatNbCol(A,R)-1;
       ZENEltMultiply(S,ZENMatGetCoeffPtr(A,l,k,R),ZENMatGetCoeffPtr(B,k,c,R),R);
       for(k--;k>=0;k--)
         {
	   ZENEltMultiply(g,ZENMatGetCoeffPtr(A,l,k,R),ZENMatGetCoeffPtr(B,k,c,R),R);
           ZENEltAdd(S,g,R);
         }
       ZENMatSetCoeff(X,l,c,S,R);
     }
 ZENEltFree(g,R);
 ZENEltFree(S,R);
 return(0);
}

/*
   \TEX{
   \title{Winograd multiplication.}
   \input{Three |ZENMat| |X|, |A| and |B| of a finite ring |R|.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \side{$|X|= |A| \times |B|$.}
   \note{Algorithm used is Winograd's.}
   }
   */


int _ZENMatWinograd(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{int i,j,k;
 ZENElt S,a,g,x,y;
 ZENMat b;

 
 if ((ZENMatNbBn(A,R)<2)||
     (ZENMatNbDigit(A,R)<2)||
     (ZENMatNbBn(B,R)<2)||
     (ZENMatNbDigit(B,R)<2))
   return(_ZENMatMultiplyPlain(X,A,B,R));
 else
   {
     ZENTest(ZENMatIsPermutation(A,R)||ZENMatIsPermutation(B,R)||ZENMatIsPermutation(X,R),
	 ZEN_ISPERM,ZEN_MATWINOGRAD,ZEN_LIB_ZEG, ZERR);
     ZENTest((ZENMatNbRow(A,R) != ZENMatNbRow(X,R))
	     ||(ZENMatNbCol(A,R) != ZENMatNbRow(B,R))
	     ||(ZENMatNbCol(B,R) != ZENMatNbCol(X,R)),
	     ZEN_MAT_DIM,ZEN_MATWINOGRAD,ZEN_LIB_ZEG, ZERR);
     ZENCheckError(ZENEltCreate(S,R), ZERR);
     ZENCheckError(ZENEltCreate(a,R), ZERR);
     ZENCheckError(ZENEltCreate(g,R), ZERR);
     ZENCheckError(ZENEltCreate(x,R), ZERR);
     ZENCheckError(ZENEltCreate(y,R), ZERR);

     b=ZENMatRowCreate(1,ZENMatNbCol(X,R),R);
     ZENTest(!b,ZEN_ALLOC,ZEN_MATWINOGRAD,ZEN_LIB_ZEG, ZERR);
       


     for (k=0;k<ZENMatNbCol(X,R);k++)     /* Calcul des b */
       {
	 ZENEltMultiply(S,ZENMatGetCoeffPtr(B,0,k,R),ZENMatGetCoeffPtr(B,1,k,R),R);
         for(j=1;j<ZENMatNbCol(A,R)/2;j++)
           {
	     ZENEltMultiply(g,ZENMatGetCoeffPtr(B,2*j,k,R),ZENMatGetCoeffPtr(B,2*j+1,k,R),R);
             ZENEltAdd(S,g,R);
           }
	   ZENMatSetBloc(b,0,k,S,R);
       }
 
     if (ZENMatNbCol(A,R)&1)              /* si n impair */
       for (i=0;i<ZENMatNbRow(X,R);i++)   /* Calcul de z(i,k) */
         {                      /* Calcul de a_i */
	   ZENEltMultiply(a,ZENMatGetCoeffPtr(A,i,1,R),ZENMatGetCoeffPtr(A,i,0,R),R);	     
           for(j=1;j<ZENMatNbCol(A,R)/2;j++)
             {
	       ZENEltMultiply(g,ZENMatGetCoeffPtr(A,i,2*j+1,R),ZENMatGetCoeffPtr(A,i,2*j,R),R);		 
               ZENEltAdd(a,g,R);
             }
   
           
           for(k=0;k<ZENMatNbCol(X,R);k++)
             {
               /* x=A(i,2j)+B(2j-1,k) y=A(i,2j)+B(2j,k) */
	       ZENEltAssign(x,ZENMatGetCoeffPtr(B,0,k,R),R);
               ZENEltAdd(x,ZENMatGetCoeffPtr(A,i,1,R),R);
               ZENEltAssign(y,ZENMatGetCoeffPtr(B,1,k,R),R);
               ZENEltAdd(y,ZENMatGetCoeffPtr(A,i,0,R),R);
	       ZENEltMultiply(S,x,y,R);
		 
               
               for(j=1;j<ZENMatNbCol(A,R)/2;j++)
                 {               
                   /* x=A(i,2j)+B(2j-1,k) y=A(i,2j)+B(2j,k) */
                   ZENEltAssign(x,ZENMatGetCoeffPtr(B,2*j,k,R),R);
                   ZENEltAdd(x,ZENMatGetCoeffPtr(A,i,2*j+1,R),R);
                   ZENEltAssign(y,ZENMatGetCoeffPtr(B,2*j+1,k,R),R);
                   ZENEltAdd(y,ZENMatGetCoeffPtr(A,i,2*j,R),R);
		   ZENEltMultiply(g,x,y,R);
		     
                   ZENEltAdd(S,g,R);
                 }
	       
               ZENEltAssign(g,ZENMatGetBlocPtr(b,0,k,R),R);
               ZENEltAdd(g,a,R);
	       
	       ZENEltSubtract(S,g,R);
	       
	       ZENEltMultiply(g,ZENMatGetCoeffPtr(A,i,ZENMatNbCol(A,R)-1,R),ZENMatGetCoeffPtr(B,ZENMatNbCol(A,R)-1,k,R),R);
	       
               ZENEltAdd(S,g,R);
	       ZENMatSetCoeff(X,i,k,S,R);
	     }
         }
     else                       /* si n pair */
       for (i=0;i<ZENMatNbRow(X,R);i++)   /* Calcul de z(i,k) */
         {                      /* Calcul de a_i */
	   ZENEltMultiply(a,ZENMatGetCoeffPtr(A,i,1,R),ZENMatGetCoeffPtr(A,i,0,R),R);
           for(j=1;j<ZENMatNbCol(A,R)/2;j++)
             {
	       ZENEltMultiply(g,ZENMatGetCoeffPtr(A,i,2*j+1,R),ZENMatGetCoeffPtr(A,i,2*j,R),R);
               ZENEltAdd(a,g,R);
             }
           
           for(k=0;k<ZENMatNbCol(X,R);k++)
             {
               /* x=A(i,2j)+B(2j-1,k) y=A(i,2j)+B(2j,k) */
               ZENEltAssign(x,ZENMatGetCoeffPtr(B,0,k,R),R);
               ZENEltAdd(x,ZENMatGetCoeffPtr(A,i,1,R),R);
               ZENEltAssign(y,ZENMatGetCoeffPtr(B,1,k,R),R);
               ZENEltAdd(y,ZENMatGetCoeffPtr(A,i,0,R),R);
	       ZENEltMultiply(S,x,y,R);
		 
               
               for(j=1;j<ZENMatNbCol(A,R)/2;j++)
                 {               
                   /* x=A(i,2j)+B(2j-1,k) y=A(i,2j)+B(2j,k) */
                   ZENEltAssign(x,ZENMatGetCoeffPtr(B,2*j,k,R),R);
                   ZENEltAdd(x,ZENMatGetCoeffPtr(A,i,2*j+1,R),R);
                   ZENEltAssign(y,ZENMatGetCoeffPtr(B,2*j+1,k,R),R);
                   ZENEltAdd(y,ZENMatGetCoeffPtr(A,i,2*j,R),R);
		   ZENEltMultiply(g,x,y,R);
		     
                   ZENEltAdd(S,g,R);
                 }
               
               ZENEltAssign(g,ZENMatGetBlocPtr(b,0,k,R),R);
               ZENEltAdd(g,a,R);
               
               ZENEltSubtract(S,g,R);
	       ZENMatSetCoeff(X,i,k,S,R);
             }
         }
     ZENEltFree(a,R);
     ZENEltFree(g,R);
     ZENEltFree(x,R);
     ZENEltFree(y,R);
     ZENEltFree(S,R);

     ZENMatFree(b,R);
   }
 return(0);
}
/*
   \TEX[
   \title{Multiplication of scalar}
   \input{A |ZENMat|, a |ZENElt| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{All the elements of |A| are multiplied by |b|.}
   \note{|A| cannot be a permutation.}
   }
   */
int _ZENMatMultiplyScalar(A,b,R)
     ZENMat A;
     ZENElt b;
     ZENRing R;
{
  int l,c;
  ZENElt S;

  ZENTest(ZENMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATMULTIPLY,ZEN_LIB_ZEG, ZERR);

  ZENCheckError(ZENEltCreate(S,R), ZERR);

     
  for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
    for(c=ZENMatNbDigit(A,R)-1;c>=0;c--)
      {
	ZENEltMultiply(S,b,ZENMatGetBlocPtr(A,l,c,R),R);
	ZENMatSetBloc(A,l,c,S,R);
      }
  ZENEltFree(S,R);
  
  return(0);
}

