/*
 * zen/zeg/c/mat_permute.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial permutations}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions.
   }
   */
#include <stdio.h>
#include "zen.h"


/*
   \TEX{
   \title{Permute two |Bn|s}
   \input{A |ZENMat| |A|, two indexes and a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The specified |Bns| of |A| are exchanged}
   \note{Except for advanced user, |PermuteRow| and |PermuteCol| functions should be preferred}
   }
   
   */
   
int _ZENMatPermuteBn(A,l1,l2,R)
     ZENMat A;
     Dim l1,l2;
     ZENRing R;
{int c;
 ZENElt S;
 
 ZENTest(ZENMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATPERMUTEBNS,ZEN_LIB_ZEG, ZERR);
 ZENTest((l1<0)||(l2<0)||(l1>=ZENMatNbBn(A,R))||(l2>=ZENMatNbBn(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATPERMUTEBNS,ZEN_LIB_ZEG, ZERR);
 ZENCheckError(ZENEltCreate(S,R), ZERR);

 for(c=ZENMatNbDigit(A,R)-1;c>=0;c--)
   {
     ZENMatGetBloc(S,A,l1,c,R);
     ZENMatSetBloc(A,l1,c,ZENMatGetBlocPtr(A,l2,c,R),R);
     ZENMatSetBloc(A,l2,c,S,R);
   }
 ZENEltFree(S,R);
 return(0);
}

/*
   \TEX{
   \title{Permute two |digit|s}
   \input{A |ZENMat| |A|, two indexes and a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The specified |digits| of |A| are exchanged}
   \note{Except for advanced user, |PermuteRow| and |PermuteCol| functions should be preferred}
   }   
   */
   

int _ZENMatPermuteDigit(A,l1,l2,R)
     ZENMat A;
     Dim l1,l2;
     ZENRing R;
{int c;
 ZENElt S;
 
 ZENTest(ZENMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATPERMUTEDG,ZEN_LIB_ZEG, ZERR);
 ZENTest((l1<0)||(l2<0)||(l1>=ZENMatNbDigit(A,R))||(l2>=ZENMatNbDigit(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATPERMUTEDG,ZEN_LIB_ZEG, ZERR);
 ZENCheckError(ZENEltCreate(S,R), ZERR);

 for(c=ZENMatNbBn(A,R)-1;c>=0;c--)
   {
     ZENMatGetBloc(S,A,c,l1,R);
     ZENMatSetBloc(A,c,l1,ZENMatGetBlocPtr(A,c,l2,R),R);
     ZENMatSetBloc(A,c,l2,S,R);
   }
 ZENEltFree(S,R);
 return(0);
}

/*
   \TEX{
   \title{Permute two |Bn|s}
   \input{A permutation |ZENMat| |A|, two indexes and a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The specified |Bns| of |A| are exchanged}
   \note{Should be used WITH CARE. Except for advanced user, |PermuteRow| and |PermuteCol| functions should be preferred. This function can be used on non permutation matrix to permute the |Bn| pointers. It is used in gaussian elimination to efficiently compute rank, determinant and inverse of a matrix.}
   }   
   */
int _ZENMatFastPermuteBn(A,l1,l2,R)
     ZENMat A;
     Dim l1,l2;
     ZENRing R;
{
  ZENElt *S;
 
  ZENTest((l1<0)||(l2<0)||(l1>=ZENMatNbBn(A,R))||(l2>=ZENMatNbBn(A,R)),
	  ZEN_MAT_DIM0,ZEN_PERMPERMUTEBNS,ZEN_LIB_ZEG, ZERR);
  ZENTest(ZENMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_PERMPERMUTEBNS,ZEN_LIB_ZEG, ZERR);


  S=ZENMatBn(A,l1,R);
  ZENMatBn(A,l1,R)=ZENMatBn(A,l2,R);
  ZENMatBn(A,l2,R)=S;
   
  return(0);
}

/*
   \TEX{
   \title{Multiply left permutation.}
   \input{Two |ZENMat| |X| and |B|, a |ZENMatTypeRowPermutation| |A| and a finite ring |R|.}
   \output{|ZENERR| if an error occurred, 0 otherwise otherwise}
   \side{$|X|= |A| \times |B|$.}
   }
   */
int _ZENMatMultiplyLeftPermutation(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{
  int l,c;
  Dim lp;
 
  ZENTest(!(ZENMatIsPermutation(A,R)),
	  ZEN_ISPERM,ZEN_MATMULLEFTPERM,ZEN_LIB_ZEG, ZERR);
  ZENTest((ZENMatNbRow(A,R) != ZENMatNbRow(X,R))
	  ||(ZENMatNbCol(A,R) != ZENMatNbRow(B,R))
	  ||(ZENMatNbCol(B,R) != ZENMatNbCol(X,R)),
	  ZEN_MAT_DIM,ZEN_MATMULLEFTPERM,ZEN_LIB_ZEG, ZERR);
  
  if(ZENMatIsRowType(A,R))
    for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
      {
	lp=ZENPermutationBn(A,l,R);
	for(c=ZENMatNbCol(X,R)-1;c>=0;c--)
	  ZENMatSetCoeff(X,l,c,ZENMatGetCoeffPtr(B,lp,c,R),R);
      }
  else
    for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
      {
	lp=ZENPermutationBn(A,l,R);
	for(c=ZENMatNbCol(X,R)-1;c>=0;c--)
	  ZENMatSetCoeff(X,lp,c,ZENMatGetCoeffPtr(B,l,c,R),R);
      }
  return(0);
}


/*
   \TEX{
   \title{Multiply right permutation.}
   \input{Two |ZENMat| |X| and |A|, a permutation |B| and a finite ring |R|.}
   \output{|ZENERR| if an error occurred, 0 otherwise otherwise}
   \side{$|X|= |A| \times |B|$.}
   }
   */
int _ZENMatMultiplyRightPermutation(X,A,B,R)
     ZENMat X,A,B;
     ZENRing R;
{
  int l,c;
  Dim lp;
 
  ZENTest((!(ZENMatIsPermutation(B,R))),
	  ZEN_ISPERM,ZEN_MATMULRIGHTPERM,ZEN_LIB_ZEG, ZERR);
  ZENTest((ZENMatNbRow(A,R) != ZENMatNbRow(X,R))
	  ||(ZENMatNbCol(A,R) != ZENMatNbRow(B,R))
	  ||(ZENMatNbCol(B,R) != ZENMatNbCol(X,R)),
	  ZEN_MAT_DIM,ZEN_MATMULRIGHTPERM,ZEN_LIB_ZEG, ZERR);
  
  if(ZENMatIsRowType(B,R))
    for(l=ZENMatNbBn(B,R)-1;l>=0;l--)
      {
	lp=ZENPermutationBn(B,l,R);
	for(c=ZENMatNbRow(X,R)-1;c>=0;c--)
	  ZENMatSetCoeff(X,c,lp,ZENMatGetCoeffPtr(A,c,l,R),R);
      }
  else
    for(l=ZENMatNbBn(B,R)-1;l>=0;l--)
      {
	lp=ZENPermutationBn(B,l,R);
	for(c=ZENMatNbRow(X,R)-1;c>=0;c--)
	  ZENMatSetCoeff(X,c,l,ZENMatGetCoeffPtr(A,c,lp,R),R);
      }
  return(0);
}
