/*
 * zen/zeg/c/permutation_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Permutation assignments}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions. 
   }
   */
#include <stdio.h>
#include "zen.h"

/*
   \TEX{
   \title{Assignment.}
   \input{A |ZENMat| |A| and a permutation matrix |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = |B| if no error occured, |zen_errno| is set otherwise.}
   \note{|A| can be a permutation matrix.}
   }
   */
   
int _ZENPermutationAssign(A,B,R)
     ZENMat A,B;
     ZENRing R;
{
  int l;

  ZENTest(!(ZENMatIsPermutation(B,R)),
	  ZEN_NOTPERM,ZEN_PERMASSIGN,ZEN_LIB_ZEG, ZERR);
  ZENTest((ZENMatNbBn(A,R) != ZENMatNbBn(B,R))
	  ||(ZENMatNbDigit(A,R) != ZENMatNbDigit(B,R)),
	  ZEN_MAT_DIM,ZEN_PERMASSIGN,ZEN_LIB_ZEG, ZERR);
  
  if(ZENMatIsPermutation(A,R))
    {
      if(ZENMatAreSameType(A,B,R))
	ZBNAssign(ZENMatFirst(A,R),ZENMatFirst(B,R),ZENMatNbBnd(B,R));
      else
	for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
	  ZENPermutationBn(A,ZENPermutationBn(B,l,R),R)=(Dim)l;
      return(0);
    }
  
  return(ZENPermutation2Matrix(A,B,R));
}



/*
   \TEX{
   \title{Assignment.}
   \input{A |ZENMat| |A| and a permutation matrix |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   
int _ZENPermutation2Matrix(A,B,R)
     ZENMat A,B;
     ZENRing R;
{
  int l;
  ZENElt S;
  
  ZENTest(!(ZENMatIsPermutation(B,R)),
	  ZEN_NOTPERM,ZEN_PERMASSIGN,ZEN_LIB_ZEG, ZERR);
  ZENTest(ZENMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_PERMASSIGN,ZEN_LIB_ZEG, ZERR);
  ZENTest((ZENMatNbBn(A,R) != ZENMatNbBn(B,R))
	  ||(ZENMatNbDigit(A,R) != ZENMatNbDigit(B,R)),
	 ZEN_MAT_DIM,ZEN_PERMASSIGN,ZEN_LIB_ZEG, ZERR);
  
  ZENCheckError(ZENEltCreate(S,R), ZERR);
  ZENEltSetToOne(S,R);
  ZENMatSetToZero(A,R);
  if(ZENMatAreSameType(A,B,R))
    for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
      ZENMatSetBloc(A,l,ZENPermutationBn(B,l,R),S,R);
  else
    for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
      ZENMatSetBloc(A,ZENPermutationBn(B,l,R),l,S,R);
  
  ZENEltFree(S,R);
  return(0);
}

/*
   \TEX{
   \title{Set to one}
   \input{A permutation matrix |A| and a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| is set to the identity matrix.}
   \note{The |ZENRing| is not used.}
   }
   */
   
int _ZENPermutationSetToOne(A,R)
     ZENMat A;
     ZENRing R;
{
  int l;

  ZENTest(!(ZENMatIsPermutation(A,R)),
	  ZEN_NOTPERM,ZEN_PERMSETONE,ZEN_LIB_ZEG, ZERR);
  for(l=ZENMatNbBn(A,R)-1;
      l>=0;
      l--)
    ZENPermutationBn(A,l,R)=(Dim)l;
  
  return(0);
}

/*
   \TEX{
   \title{Is one}
   \input{A permutation matrix |A| and a finite ring |R|.}
   \output{1 if |A| is identity, |ZENERR| if |A| is not a permutation, 0 otherwise}
   \note{The |ZENRing| is not used.}
   }
   */
   
int _ZENPermutationIsOne(A,R)
     ZENMat A;
     ZENRing R;
{
  int l;

  ZENTest(!(ZENMatIsPermutation(A,R)),
	  ZEN_NOTPERM,ZEN_PERMISONE,ZEN_LIB_ZEG, ZERR);
  for(l=ZENMatNbBn(A,R)-1;
      l>=0;
      l--)
    if(ZENPermutationBn(A,l,R)!=(Dim)l)
      return(0);
  
  return(1);
}

/*
   \TEX{
   \title{Set to random}
   \input{A permutation |A| and a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{The permutation |A| is randomly set.}
   }
   
   */
   
int _ZENPermutationSetRandom(A,R)
     ZENMat A;
     ZENRing R;
{
  Dim *tab;
  int l,c;
 
  ZENTest(!(ZENMatIsPermutation(A,R)),
	 ZEN_NOTPERM,ZEN_PERMSETRANDOM,ZEN_LIB_ZEG, ZERR);
  tab=(Dim *)ZENMalloc(ZENMatNbBn(A,R)*sizeof(Dim));
  ZENTest(!tab,
	  ZEN_ALLOC,ZEN_PERMSETRANDOM,ZEN_LIB_ZEG, ZERR);

  for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
    tab[l]=l;

  for(l=ZENMatNbBn(A,R)-1;l>=0;l--)
    {
      c=(zbnrandom()%(l+1));
      ZENPermutationBn(A,l,R)=tab[c];
      for(c++;c<=l;c++)
	tab[c-1]=tab[c];
    }
  ZENFree(tab);
  return(0);
}

/*
   \TEX{
   \title{Get a bloc}
   \input{A permutation |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{|E| is set to the indexed bloc of |A|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred}
   }
   
   */
   
int _ZENPermutationGetBloc(p_E,A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENElt *p_E;
     ZENRing R;
{
  ZENTest(!(ZENMatIsPermutation(A,R)),
	  ZEN_NOTPERM,ZEN_PERMGETBLOC,ZEN_LIB_ZEG, ZERR);
  ZENTest((l<0)||(c<0)||(l>=ZENMatNbBn(A,R))||(c>=ZENMatNbDigit(A,R)),
	  ZEN_MAT_DIM0,ZEN_PERMGETBLOC,ZEN_LIB_ZEG, ZERR);
  if(ZENPermutationBn(A,l,R)==c)
    ZENEltSetToOne(*p_E,R);
  else
    ZENEltSetToZero(*p_E,R);

  return(0);
}

/*
   \TEX{
   \title{Set a bloc}
   \input{A permutation |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The indexed bloc of |A| is set to |E|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred}
   }
   
   */
   
int _ZENPermutationSetBloc(A,l,c,E,R)
     ZENMat A;
     Dim l,c;
     ZENElt E;
     ZENRing R;
{
  ZENTest(!(ZENMatIsPermutation(A,R)),
	  ZEN_NOTPERM,ZEN_PERMSETBLOC,ZEN_LIB_ZEG, ZERR);
  ZENTest((l<0)||(c<0)||(l>=ZENMatNbBn(A,R))||(c>=ZENMatNbDigit(A,R)),
	  ZEN_MAT_DIM0,ZEN_MATSETBLOC,ZEN_LIB_ZEG, ZERR);
  if(ZENEltIsOne(E,R))
    ZENPermutationBn(A,l,R)=c;
  else
    ZENTest(!(ZENEltIsZero(E,R)),
	    ZEN_ISPERM,ZEN_PERMSETBLOC,ZEN_LIB_ZEG, ZERR);
  return(0);
}


/*
   \TEX{
   \title{Get a digit}
   \input{A permutation |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{|E| is set to the indexed digit of |A|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred}
   }
   
   */
   
int _ZENPermutationGetDigit(p_E,A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENElt *p_E;
     ZENRing R;
{
  ZENTest(!(ZENMatIsPermutation(A,R)),
	  ZEN_NOTPERM,ZEN_PERMGETDIGIT,ZEN_LIB_ZEG, ZERR);
  ZENTest((l<0)||(c<0)||(l>=ZENMatNbBn(A,R))||(c>=ZENMatNbDigit(A,R)),
	  ZEN_MAT_DIM0,ZEN_PERMGETDIGIT,ZEN_LIB_ZEG, ZERR);
  if(ZENPermutationBn(A,c,R)==l)
    ZENEltSetToOne(*p_E,R);
  else
    ZENEltSetToZero(*p_E,R);

  return(0);
}

/*
   \TEX{
   \title{Set a digit}
   \input{A permutation |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The indexed digit of |A| is set to |E|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred}
   }
   
   */
   
int _ZENPermutationSetDigit(A,l,c,E,R)
     ZENMat A;
     Dim l,c;
     ZENElt E;
     ZENRing R;
{
  ZENTest(!(ZENMatIsPermutation(A,R)),
	  ZEN_NOTPERM,ZEN_PERMSETDIGIT,ZEN_LIB_ZEG, ZERR);
  ZENTest((l<0)||(c<0)||(l>=ZENMatNbBn(A,R))||(c>=ZENMatNbDigit(A,R)),
	  ZEN_MAT_DIM0,ZEN_MATSETDIGIT,ZEN_LIB_ZEG, ZERR);
  if(ZENEltIsOne(E,R))
    ZENPermutationBn(A,c,R)=l;
  else
    ZENTest(!(ZENEltIsZero(E,R)),
	    ZEN_ISPERM,ZEN_PERMSETDIGIT,ZEN_LIB_ZEG, ZERR);
  return(0);
}

/*
   \TEX{
   \title{Permute two |Bn|s}
   \input{A permutation |ZENMat| |A|, two indexes and a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The specified |Bns| of |A| are exchanged}
   \note{Except for advanced user, |PermuteRow| and |PermuteCol| functions should be preferred.}
   }
   
   */
   
int _ZENPermutationPermuteBn(A,l1,l2,R)
     ZENMat A;
     Dim l1,l2;
     ZENRing R;
{
  Dim S;
 
  ZENTest((l1<0)||(l2<0)||(l1>=ZENMatNbBn(A,R))||(l2>=ZENMatNbBn(A,R)),
	  ZEN_MAT_DIM0,ZEN_PERMPERMUTEBNS,ZEN_LIB_ZEG, ZERR);
  ZENTest(!ZENMatIsPermutation(A,R),
	  ZEN_NOTPERM,ZEN_PERMPERMUTEBNS,ZEN_LIB_ZEG, ZERR);
	  

  S=ZENPermutationBn(A,l1,R);
  ZENPermutationBn(A,l1,R)=ZENPermutationBn(A,l2,R);
  ZENPermutationBn(A,l2,R)=S;
   
  return(0);
}

