/*
 * zen/zeg/c/permutation_create.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Permutations}
   A permutation matrix is a square matrix with a single one in every row and every column.
   It can be stored by keeping for every row OR every column the column OR row index of the one.
   We have therefore two types of permutation matrices |ZENMatTypeRowPermutation| and |ZENMatTypeColPermutation| .
   In first case, we keep for each row the column index of the one.
   This means that multiplying this matrix left to another one |M| will permute the rows of |M| as indicated by the stored indexes.

   A permutation matrix can be used as input in a few functions namely |PermutationAssign|, |PermutationSetToOne|, |PermutationSetRandom|, |PermutationGetCoeff|, |Gauss|, |MultiplyPermutation| and |InversePermutation|.
   }
   */


#include <stdio.h>
#include "zeg.h"

/* \TEX{
   \title{Allocating a  permutation matrix.}
   \input{The dimension of the desired matrix, its type and a |ZENRing|.}
   \output{NULL if an error occured, a permutation |ZENMat| of the desired dimensions otherwise.}
   \note{The |ZENRing| is not used}
   }
   */   

ZENMat _ZENPermutationCreate(n,ts,R)
     Dim n;
     ZENMatTypeStruct ts;
     ZENRing R;
{
  ZENMat ret;
  BigNumDigit *ptr;
  
  ZENTest(n<=0,
	  ZEN_MAT_DIM0,ZEN_PERMCREATE,ZEN_LIB_ZEG,ZNULL);

  ret=(ZENMat)(ZENMalloc(sizeof(struct zenmat)));
  ZENTest(ret==NULL,
	  ZEN_ALLOC,ZEN_PERMCREATE,ZEN_LIB_ZEG, ZNULL);
  ZENMatType(ret,R)=(ZENMatTypeStruct)(((int)ZENMatTypeColPermutation)^(((int)ts)&1));

  ZENMatNbBn(ret,R)=n;
  ZENMatNbDigit(ret,R)=n;
  ZENMatNbBloc(ret,R)=0;

  n=divSizeBloc(n*sizeof(Dim)*SIZE_CHAR-1)+1;
  ZENMatNbBnd(ret,R)=n;

  ptr=(BigNumDigit *)ZENMalloc(n*sizeof(BigNumDigit));
  ZENPermutationMat(ret,R)=(Dim *)ptr;
  ptr[n-1]=(BigNumDigit)0;
    
  if(ZENPermutationMat(ret,R)==NULL)
    {
      ZENFree(ret);
      ZENSetError(ZEN_LIB_ZEG,ZEN_PERMCREATE,ZEN_ALLOC);
      return(ZNULL);
    }
  ZENMatFirst(ret,R)=(BigNumDigit *)ZENMatMat(ret,R);
  
  ZENMatFreePtr(ret)=_ZENPermutationFree;
  ZENMatAssignPtr(ret)=_ZENPermutationAssign;
  ZENMatSetToOnePtr(ret)=_ZENPermutationSetToOne;
  ZENMatIsOnePtr(ret)=_ZENPermutationIsOne;
  ZENMatSetRandomPtr(ret)=_ZENPermutationSetRandom;
  ZENMatPermuteBnPtr(ret)=_ZENPermutationPermuteBn;

  switch(ZENMatType(ret,R))
    {
    case ZENMatTypeRowPermutation:
      ZENMatGetCoeffNewPtr(ret)=_ZENPermutationGetBloc;
      ZENMatSetCoeffPtr(ret)=_ZENPermutationSetBloc;     
      ZENMatPermuteRowPtr(ret)=_ZENPermutationPermuteBn;
      break;
      
    case ZENMatTypeColPermutation:
      ZENMatGetCoeffNewPtr(ret)=_ZENPermutationGetDigit;
      ZENMatSetCoeffPtr(ret)=_ZENPermutationSetDigit;     
      ZENMatPermuteColPtr(ret)=_ZENPermutationPermuteBn;
      break;
      
    default:
      ZENSetError(ZEN_LIB_ZEG,ZEN_PERMCREATE,ZEN_UNKNOWN);
      return(ZNULL);
    }
  _ZENPermutationSetToOne(ret,R);
  return(ret);
}

/*
   \TEX{
   \title{Freeing a permutation matrix}
   \input{A |ZENMat| and a |ZENRing|.}
   \side{Free the matrix.}
   \note{This function is also used whenever there is no allocation of coefficients in matrices (e.g. |ze2|, |zeps|, |zetab|, |zelog|)}
   } 
   */
void _ZENPermutationFree(P,R)
     ZENMat P;
     ZENRing R;
{
  ZENFree(ZENPermutationMat(P,R));
  ZENFree(P);
}
