/*
 * zen/zeg/c/poly_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial assignment}
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{   
   One can set a |ZENPoly| to zero with |_ZENPolySetToZero|. Then we can
   assign a |ZENPoly| with an other with |_ZENPolyAssign|. Furthermore, we
   can set randomly a |ZENPoly| with |_ZENPolySetRandom|. 
   
   The functions |_ZENPolyGetCoeff| and |_ZENPolySetCoeff|
   enable us to get or set the coefficient of $X^i$,
   |_ZENPolyUpdateDegree| compute the real degree of a |ZENPoly| if it
   decreased.
   }
   */

/* \TEX{
   \title{Assign.}
   \input{Two |ZENPoly| |PX| and |RX| of a finite ring |Rg|.}
   \side{|RX| is filled with |PX|.}
   }
   */
void _ZENPolyAssign(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZENPolyDeg(PX, Rg); i++) {
    ZENPolySetCoeff(RX, i, ZENPolyGetCoeffPtr(PX, i, Rg), Rg);
  }
  ZENPolyDeg(RX, Rg) =  ZENPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Copying a polynomial.}
   \input{A polynomial |p| of a finite ring |R|.}
   \output{|ZNULL| if an error occured, a copy of |p| otherwise.}   
   }
   */
ZENPoly _ZENPolyCopy (p, R)
     ZENPoly p;
     ZENRing R;
{
  ZENPoly ret;
  
  if (!(ret = ZENPolyCreate(ZENPolyLgt(p, R), R))) {
    return(ZNULL);
  }
  ZENPolyAssign(ret, p, R);
  return(ret);
}

/* \TEX{
   \title{Set to zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{|RX| is set to zero..}
   }
   */
void _ZENPolySetToZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  ZENElt elt;

  ZENPolyDeg(RX, Rg) =  0;
  elt = ZENPolyGetCoeffPtr(RX, 0, Rg);
  ZENEltSetToZero(elt, Rg);
  ZENPolySetCoeff(RX, 0, elt, Rg);
}

/* \TEX{
   \title{Set to $X^i$.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg| allocated at least for
   degree $|deg|$.}
   \side{|RX| is set to $X^{|deg|}$.}
   }
   */
void _ZENPolySetToXi(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  ZENElt elt;
  int i;

  ZENPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    elt = ZENPolyGetCoeffPtr(RX, i, Rg);
    ZENEltSetToZero(elt, Rg);
    ZENPolySetCoeff(RX, i, elt, Rg);
  }
  elt = ZENPolyGetCoeffPtr(RX, deg, Rg);
  ZENEltSetToOne(elt, Rg);
  ZENPolySetCoeff(RX, deg, elt, Rg);
}

/* \TEX{
   \title{Set to random.}
   \input{A |ZENPoly| |RX| of degree |deg| of a finite ring |Rg|.}
   \side{|RX| is set randomly with degree |deg|.}
   }
   */
void _ZENPolySetRandom(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  ZENElt elt;
  int i;

  ZENPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    elt = ZENPolyGetCoeffPtr(RX, i, Rg);
    ZENEltSetRandom(elt, Rg);
    ZENPolySetCoeff(RX, i, elt, Rg);
  }
  elt = ZENPolyGetCoeffPtr(RX, deg, Rg);
  ZENEltSetRandom(elt, Rg);
  while(ZENEltIsZero(elt, Rg)) {
    ZENEltSetRandom(elt, Rg);
  }
  ZENPolySetCoeff(RX, deg, elt, Rg);
}

/* \TEX{
   \title{Get a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The coeff of $X^d$ in |RX|.}
   }
   */
void _ZENPolyGetCoeff(p_c, RX, d, Rg)
     ZENElt *p_c;
     ZENPoly RX;
     int d;
     ZENRing Rg;
{
  ZENEltAssign(*p_c, ZENPolyGetCoeffPtr(RX, d, Rg), Rg);
}

/* \TEX{
   \title{Update the degree of a polynomial.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{Resetting the degree of |P|, assuming it has decreased.}
   }
   */
void _ZENPolyUpdateDegree(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  int i;

  for(i = ZENPolyDeg(RX, Rg); i >= 0; i--) {
    if (!ZENEltIsZero(ZENPolyGetCoeffPtr(RX, i, Rg), Rg)) {
      break;
    }
  }
  if (i < 0) {
    ZENPolySetToZero(RX, Rg);
  }
  else {
    ZENPolyDeg(RX, Rg) =  i;
  }
}

/* \TEX{
   \title{Convert.}
   \input{Two allocated |ZENPoly|s of same degree and two |ZENRing|s.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The coefficients of |P1| of |R1| are affected with coefficients of |P2| of |R2| after possible modular reduction of those coefficients. |R1| and |R2| must be compatible rings.}
   }
   */
int _ZENPolyConvert(P1, R1, P2, R2)
     ZENPoly P1, P2;
     ZENRing R1, R2;
{
  int i;

  ZENPolyDeg(P1, R1) =  ZENPolyDeg(P2, R2);
  for(i = 0; i <= ZENPolyDeg(P2, R2); i++)
    {
      if(ZENEltConvert(ZENRingIbuf(R1),R1,
		       ZENPolyGetCoeffPtr(P2, i, R2),R2))
	return(ZERR);
      ZENPolySetCoeff(P1,i,ZENRingIbuf(R1),R1);
    }
  ZENPolyUpdateDegree(P1, R1);
  return(ZEN_HAS_INVERSE);
}

