/*
 * zen/zeg/c/poly_gcd.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial gcd}
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{Gcd of 2 polynomials.}
   \input{Three polynomials |PX|, |QX| and |RX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{|R|$(X) = gcd($|P|$(X), $|Q|$(X))$ if |ZEN_HAS_INVERSE| is
   returned,  |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZENPolyGcd (RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int ret;
  ZENPoly px, qx;

  if (ZENPolyDeg(PX, Rg) > ZENPolyDeg(QX, Rg)) {
    if (!(px = ZENPolyCopy(PX, Rg))) {
      return(ZERR);
    }
    if (!(qx = ZENPolyCopy(QX, Rg))) {
      ZENPolyFree(px, Rg);
      return(ZERR);
    }
  }
  else {
    if (!(px = ZENPolyCopy(QX, Rg))) {
      return(ZERR);
    }
    if (!(qx = ZENPolyCopy(PX, Rg))) {
      ZENPolyFree(px, Rg);
      return(ZERR);
    }
  }

  while(ZENPolyDeg(qx, Rg) > 0){
    if ((ret = ZENPolyDivide(RX, RX, px, qx, Rg)) != ZEN_HAS_INVERSE) {
      ZENPolyFree(px, Rg);
      ZENPolyFree(qx, Rg);
      return(ret);
    }
    ZENPolyAssign(px, qx, Rg);
    ZENPolyAssign(qx, RX, Rg);
  }
  if (ZENPolyIsZero(qx, Rg)) {
    ZENPolyAssign(RX, px, Rg);
  }
  else {
    ZENPolySetToXi(RX, 0, Rg);
  }

  ZENPolyFree(px, Rg);
  ZENPolyFree(qx, Rg);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Extended gcd of 2 polynomials.}
   \input{Three polynomials |IX0|, |AX0| and |BX0| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, -2 if the gcd of |AX0| and |BX0|
   is not 1, |ZEN_NO_INVERSE| if a factor of a modulo was found,
   |ZEN_HAS_INVERSE| otherwise} 
   \side{|IX0|$ = 1$/|BX0|$ \bmod $|AX0| if |ZEN_HAS_INVERSE| is
   returned, |IX0| is gcd(|AX0|, |BX0|) if -2 is returned,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZENPolyExtGcd (IX0, BX0, AX0, Rg)
     ZENPoly IX0, AX0, BX0;
     ZENRing Rg;
{
  int d, ret;
  ZENPoly AX, BX, IX, RX, SX, QX, TX, pol;
  ZENElt elt, tmp;
  
  if (!(AX = ZENPolyCopy(AX0, Rg))) {
    return(ZERR);
  }
  if (!(BX = ZENPolyCopy(BX0, Rg))) {
    ZENPolyFree(AX, Rg);
    return(ZERR);
  }

  if (ZENPolyDeg(AX, Rg) <= ZENPolyDeg(BX, Rg)) {
    ZENPolyDivide(BX, BX, BX, AX, Rg);
  }
  IX = ZENPolyCreate(ZENPolyDeg(AX, Rg), Rg);
  RX = ZENPolyCreate(ZENPolyDeg(AX, Rg), Rg);
  SX = ZENPolyCreate(ZENPolyDeg(AX, Rg), Rg);
  QX = ZENPolyCreate(ZENPolyDeg(AX, Rg), Rg);
  TX = ZENPolyCreate(ZENPolyDeg(AX, Rg), Rg);
  d = -1;

  ZENPolySetToXi(SX, 0, Rg);
  ZENPolySetToZero(IX, Rg);
  while(ZENPolyDeg(AX, Rg) > 0){
    d = -d;
    if ((ret = ZENPolyDivide(QX, RX, BX, AX, Rg)) != ZEN_HAS_INVERSE) {
      ZENPolyFree(AX, Rg); ZENPolyFree(BX, Rg);
      ZENPolyFree(IX, Rg);
      ZENPolyFree(RX, Rg); ZENPolyFree(QX, Rg);
      ZENPolyFree(SX, Rg); ZENPolyFree(TX, Rg);
      return(ret);
    }
    if ((ret = ZENPolyMultiply(TX, IX, QX, Rg))) {
      ZENPolyFree(AX, Rg); ZENPolyFree(BX, Rg);
      ZENPolyFree(IX, Rg);
      ZENPolyFree(RX, Rg); ZENPolyFree(QX, Rg);
      ZENPolyFree(SX, Rg); ZENPolyFree(TX, Rg);
      return(ret);
    }
    if ((ret = ZENPolyDivide(TX, TX, TX, AX0, Rg)) != ZEN_HAS_INVERSE) {
      ZENPolyFree(AX, Rg); ZENPolyFree(BX, Rg);
      ZENPolyFree(IX, Rg);
      ZENPolyFree(RX, Rg); ZENPolyFree(QX, Rg);
      ZENPolyFree(SX, Rg); ZENPolyFree(TX, Rg);
      return(ret);
    }
    ZENPolyAdd(TX, SX, Rg);

    pol = SX; SX = IX; IX = TX; TX = pol;
    pol = BX; BX = AX; AX = RX; RX = pol;
  }
  if(ZENPolyIsZero(AX, Rg)) {

    /* Non inversible */
    ZENPolyAssign(IX0, BX, Rg);

    ZENPolyFree(AX, Rg); ZENPolyFree(BX, Rg);
    ZENPolyFree(IX, Rg);
    ZENPolyFree(RX, Rg); ZENPolyFree(QX, Rg);
    ZENPolyFree(SX, Rg); ZENPolyFree(TX, Rg);
    return(-2);
  }
  else {

    /* IX * BX0 = AX[0] mod (Modu, AX0) */
    elt = ZENPolyGetCoeffPtr(AX, 0, Rg);

    if (!ZENEltIsOne(elt, Rg)) {
      if (ZENEltCreate(tmp, Rg)) {
	ZENPolyFree(AX, Rg); ZENPolyFree(BX, Rg);
	ZENPolyFree(IX, Rg);
	ZENPolyFree(RX, Rg); ZENPolyFree(QX, Rg);
	ZENPolyFree(SX, Rg); ZENPolyFree(TX, Rg);
	return(ZERR);
      } 
      if ((ret = ZENEltInverse(tmp, elt, Rg)) != ZEN_HAS_INVERSE) {
	ZENEltFree(tmp, Rg);
	ZENPolyFree(AX, Rg); ZENPolyFree(BX, Rg);
	ZENPolyFree(IX, Rg);
	ZENPolyFree(RX, Rg); ZENPolyFree(QX, Rg);
	ZENPolyFree(SX, Rg); ZENPolyFree(TX, Rg);
	
	return(ret);
      }
      ZENPolyMultiplyScalar(IX, IX, tmp, Rg);
      ZENEltFree(tmp, Rg);
    }
    ZENPolyAssign(IX0, IX, Rg);
    if(d < 0) {
      ZENPolyNegate(IX0, IX0, Rg);
    }
  }

  ZENPolyFree(AX, Rg); ZENPolyFree(BX, Rg);
  ZENPolyFree(IX, Rg);
  ZENPolyFree(RX, Rg); ZENPolyFree(QX, Rg);
  ZENPolyFree(SX, Rg); ZENPolyFree(TX, Rg);
  return(ZEN_HAS_INVERSE);
}
