/*
 * zen/zeg/c/poly_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   \Subsection{input/output}
   
   Conversions between the internal representation and the ASCII
   representation of a |ZENPoly| are done by
   |_ZENPolyPrintToString|. |_ZENPolyPrintToFile| allows to write a
   |ZENPoly| in a file. One can moreover write to and then read from a
   file a |_ZENPoly| with |_ZENPolyPutToFile| and |_ZENPolyReadFromFile|.
   }
   */
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "zen.h"

/* \TEX{
   \title{Converting from string.}
   \input{A finite field |R|, a pointer on |ZENPoly| |PX| and a string
   |s| representing a |ZENPoly| in base |base|$\in \{2,\ldots,
   16\}$ with the character |var| as variable (usually |'X'|).} 
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|PX| is allocated and filled with |s| if the output is not |ZENERR|.}
   }
*/
int _ZENPolyReadFromString(PX, s, var, base, Rg)
     ZENPoly *PX;
     char *s;
     int var;
     int base;
     ZENRing Rg;
{
  ZENElt elt;
  char *ptr;
  int i, nbc;


  /* Alocating a buffer */
  if (ZENEltCreate(elt, Rg)) {
    return(ZERR);
  }
  *PX = NULL;
  ptr = s;
  
  /* Parsing */
  for(;;) {
    for(; isspace(*ptr); ptr++) ;
    if (*ptr != '(') {
      ZENSetError(ZEN_LIB_ZEG, ZEN_POLYREADFROMSTRING, ZEN_SYNTAX);
      ZENEltFree(elt, Rg);
      if (*PX) {
	ZENPolyFree(*PX, Rg);
      }
      return(ZERR);
    }
    
    ptr++;
    if ((nbc = ZENEltReadFromString(elt, ptr, base, Rg)) < 0) {
      ZENEltFree(elt, Rg);
      if (*PX) {
	ZENPolyFree(*PX, Rg);
      }
      return(ZERR);
    }

    for(ptr += nbc; isspace(*ptr); ptr++) ;
    if (*ptr != ')') {
      ZENSetError(ZEN_LIB_ZEG, ZEN_POLYREADFROMSTRING, ZEN_SYNTAX);
      ZENEltFree(elt, Rg);
      if (*PX) {
	ZENPolyFree(*PX, Rg);
      }
      return(ZERR);
    }
    
    for(ptr++; isspace(*ptr); ptr++) ;
    if (*ptr == '*') {
      for(ptr++; isspace(*ptr); ptr++) ;
      if (*ptr == var) {
	for(ptr++; isspace(*ptr); ptr++) ;
	switch (*ptr) {
	case '^':
	  for(ptr++; isspace(*ptr); ptr++) ;
	  if (sscanf(ptr, "%d", &i) != 1) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYREADFROMSTRING, ZEN_SYNTAX);
	    ZENEltFree(elt, Rg);
	    if (*PX) {
	      ZENPolyFree(*PX, Rg);
	    }
	    return(ZERR);
	  }
	  if (!*PX) {
	    if (!(*PX = ZENPolyCreate(i, Rg))) {
	      ZENEltFree(elt, Rg);
	      return(ZERR);
	    }
	    ZENPolySetToXi(*PX, i, Rg);
	    ZENPolySetToZero(*PX, Rg);
	  }
	  if (!ZENEltIsZero(elt, Rg)) {
	    ZENPolySetCoeff(*PX, i, elt, Rg);
	    if (i > ZENPolyDeg(*PX, Rg)) {
	      ZENPolyDeg(*PX, Rg) = i;
	    }
	  }
	  for(ptr++; isdigit(*ptr); ptr++) ;
	  for(; isspace(*ptr); ptr++) ;
	  break;
	case '+':
	  if (!*PX) {
	    if (!(*PX = ZENPolyCreate(1, Rg))) {
	      ZENEltFree(elt, Rg);
	      return(ZERR);
	    }
	    ZENPolySetToXi(*PX, 1, Rg);
	    ZENPolySetToZero(*PX, Rg);
	  }
	  if (!ZENEltIsZero(elt, Rg)) {
	    ZENPolySetCoeff(*PX, 1, elt, Rg);
	    if (1 > ZENPolyDeg(*PX, Rg)) {
	      ZENPolyDeg(*PX, Rg) = 1;
	    }
	  }
	  break;
	default:
	  if (!*PX) {
	    if (!(*PX = ZENPolyCreate(1, Rg))) {
	      ZENEltFree(elt, Rg);
	      return(ZERR);
	    }
	    ZENPolySetToXi(*PX, 1, Rg);
	    ZENPolySetToZero(*PX, Rg);
	  }
	  if (!ZENEltIsZero(elt, Rg)) {
	    ZENPolySetCoeff(*PX, 1, elt, Rg);
	    if (1 > ZENPolyDeg(*PX, Rg)) {
	      ZENPolyDeg(*PX, Rg) = 1;
	    }
	  }
	  ZENEltFree(elt, Rg);
	  return(ptr-s);
	}
      }
      else {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYREADFROMSTRING, ZEN_SYNTAX);
	ZENEltFree(elt, Rg);
	if (*PX) {
	  ZENPolyFree(*PX, Rg);
	}
	return(ZERR);
      }
    }
    else {
      if (!*PX) {
	if (!(*PX = ZENPolyCreate(0, Rg))) {
	  ZENEltFree(elt, Rg);
	  return(ZERR);
	}
	ZENPolyDeg(*PX, Rg) =  0;
      }
      ZENPolySetCoeff(*PX, 0, elt, Rg);
      ZENEltFree(elt, Rg);
      return(ptr-s);
    }
    
    if (*ptr != '+') {
      break;
    }
    ptr++;
  }
  
  ZENEltFree(elt, Rg);
  return(ptr-s);
}

/* \TEX{
   \title{Converting to string.}
   \input{An allocated |ZENPoly| |G| of a finite field |R| a character (usually |'X'|) and a base |base|$\in \{2,\ldots, 16\}$.}
   \output{An allocated string representing |G| in base |base| or NULL if an error occured.}
   }
   */
char *_ZENPolyPrintToString(PX, var, base, Rg)
     ZENPoly PX;
     int var;
     int base;
     ZENRing Rg;  
{  
  int i, dgt, nb_dgt;
  char *str, *s1, *s2;
  ZENElt elt;


  str = NULL;
  if (ZENPolyLgt(PX, Rg) > ZENPolyDeg(PX, Rg)) {

    for(nb_dgt = 1, dgt = 1;
	10*dgt <= ZENPolyLgt(PX, Rg);
	nb_dgt++, dgt *= 10) {
      ;
    }

    switch(ZENPolyLgt(PX, Rg)) {

    case 1:
      elt = ZENPolyGetCoeffPtr(PX, ZENPolyLgt(PX, Rg), Rg);
      ZENEltSetToZero(elt, Rg);

      if (!(str = ZENEltPrintToString(elt, base, Rg))) {
	return(NULL);
      }
      s2 = str;
      if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+5)))) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_ALLOC);
	ZENFree(s2);
	return(NULL);
      }
      if (sprintf(str, "(%s)*%c", s2,var) < 0) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_SPRINTF);
	ZENFree(str); ZENFree(s2);
	return(NULL);
      }
      ZENFree(s2);
      break;
      
    default:
      elt = ZENPolyGetCoeffPtr(PX, ZENPolyLgt(PX, Rg), Rg);
      ZENEltSetToZero(elt, Rg);
      if (!(str = ZENEltPrintToString(elt, base, Rg))) {
	return(NULL);
      }
      s2 = str;
      if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+6+nb_dgt)))) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_ALLOC);
	ZENFree(s2);
	return(NULL);
      }
      if (sprintf(str, "(%s)*%c^%d", s2, var,ZENPolyLgt(PX, Rg)) < 0) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_SPRINTF);
	ZENFree(str); ZENFree(s2);
	return(NULL);
      }
      ZENFree(s2);
      break;
    }
  }
  
  for(nb_dgt = 1, dgt = 1;
      10*dgt <= ZENPolyDeg(PX, Rg);
      nb_dgt++, dgt *= 10) {
    ;
  }


  for(i = ZENPolyDeg(PX, Rg); i >= 0; i--) {
    
    if (i < dgt) {
      nb_dgt--;
      dgt /= 10;
    }

    elt = ZENPolyGetCoeffPtr(PX, i, Rg);

    switch(i) {
    case 0:
      if (!str) {
	if (!(str = ZENEltPrintToString(elt, base, Rg))) {
	  return(NULL);
	}
	s2 = str;
	if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+3)))) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_ALLOC);
	  ZENFree(s2);
	  return(NULL);
	}
	if (sprintf(str, "(%s)", s2) < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_SPRINTF);
	  ZENFree(str); ZENFree(s2);
	  return(NULL);
	}
	ZENFree(s2);
      }
      else if (!ZENEltIsZero(elt, Rg)) {
	if (!(s1 = ZENEltPrintToString(elt, base, Rg))) {
	  ZENFree(str);
	  return(NULL);
	}
	s2 = str;
	if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+strlen(s1)+4)))) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_ALLOC);
	  ZENFree(s1); ZENFree(s2);
	  return(NULL);
	}
	if (sprintf(str, "%s+(%s)", s2, s1) < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_SPRINTF);
	  ZENFree(s1); ZENFree(s2);
	  return(NULL);
	}
	ZENFree(s1); ZENFree(s2);
      }
      break;
      
    case 1:
      if (!ZENEltIsZero(elt, Rg)) {
	if (!str) {
	  if (!(str = ZENEltPrintToString(elt, base, Rg))) {
	    return(NULL);
	  }
	  s2 = str;
	  if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+5)))) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_ALLOC);
	    ZENFree(s2);
	    return(NULL);
	  }
	  if (sprintf(str, "(%s)*%c", s2,var) < 0) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_SPRINTF);
	    ZENFree(str); ZENFree(s2);
	    return(NULL);
	  }
	  ZENFree(s2);
	}
	else {
	  if (!(s1 = ZENEltPrintToString(elt, base, Rg))) {
	    ZENFree(str);
	    return(NULL);
	  }
	  s2 = str;
	  if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+strlen(s1)+6)))) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_ALLOC);
	    ZENFree(s1); ZENFree(s2);
	    return(NULL);
	  }
	  if (sprintf(str, "%s+(%s)*%c", s2, s1,var) < 0) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_SPRINTF);
	    ZENFree(s1); ZENFree(s2);
	    return(NULL);
	  }
	  ZENFree(s1); ZENFree(s2);
	}
      }
      break;
      
    default:
      if (!ZENEltIsZero(elt, Rg)) {
	if (!str) {
	  if (!(str = ZENEltPrintToString(elt, base, Rg))) {
	    return(NULL);
	  }
	  s2 = str;
	  if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+6+nb_dgt)))) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_ALLOC);
	    ZENFree(s2);
	    return(NULL);
	  }
	  if (sprintf(str, "(%s)*%c^%d", s2, var, i) < 0) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_SPRINTF);
	    ZENFree(str); ZENFree(s2);
	    return(NULL);
	  }
	  ZENFree(s2);
	  
	}
	else {
	  if (!(s1 = ZENEltPrintToString(elt, base, Rg))) {
	    ZENFree(str);
	    return(NULL);
	  }
	  s2 = str;
	  if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+strlen(s1)+7+nb_dgt)))) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_ALLOC);
	    ZENFree(s1); ZENFree(s2);
	    return(NULL);
	  }
	  if (sprintf(str, "%s+(%s)*%c^%d", s2, s1, var,i) < 0) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOSTRING, ZEN_SPRINTF);
	    ZENFree(s1); ZENFree(s2);
	    return(NULL);
	  }
	  ZENFree(s1); ZENFree(s2);
	}
      }
      break;
    }
  }

  return(str);
}


/* \TEX{
   \title{Converting from a string in an internal representation.}
   \input{A finite field |Rg| and a string |s| representing a |ZENPoly|
   in an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|PX| is allocated and filled with s if the output is different
   from |ZENERR|}	    
   }
   */
int _ZENPolyGetFromString(PX, s, Rg)
     char *s;
     ZENPoly *PX;
     ZENRing Rg;
{
  ZENElt elt;
  int deg, length, i, nbc;
  char *ptr;
  

  ptr=s;
  if (sscanf(ptr, "%8x%8x", &length, &deg) != 2) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYGETFROMSTRING, ZEN_SSCANF);
    return(ZERR);
  }
  if (ZENEltCreate(elt, Rg)) {
    return(ZERR);
  }
  if (!(*PX = ZENPolyCreate(length, Rg))) {
    ZENEltFree(elt, Rg);
    return(ZERR);
  }

  for(; isspace(*ptr); ptr++) ;
  ptr += 16;
  for(i = 0; i <= deg; i++) {
    if ((nbc = ZENEltGetFromString(elt, ptr, Rg)) < 0) {
      ZENEltFree(elt, Rg);
      ZENPolyFree(*PX, Rg);
      return(ZERR);
    }
    ZENPolySetCoeff(*PX, i, elt, Rg);
    ptr += nbc;
  }
  ZENPolyDeg(*PX, Rg) =  deg;
  ZENEltFree(elt, Rg);
  return(ptr-s);
}

/* \TEX{
   \title{Converting to string in an internal representation.}
   \input{An allocated |ZENPoly| |PX| of a finite field |Rg|.}
   \output{An allocated string representing |G| in an internal representation or NULL if an error occured.}   
   }
   */
char *_ZENPolyPutToString(PX, Rg)
     ZENPoly PX;
     ZENRing Rg;
{
  ZENElt elt;
  char *str, *s1;
  int i;


  if (!(str = (char *)ZENMalloc((1+4*sizeof(int))*sizeof(char)))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPUTTOSTRING, ZEN_ALLOC);
    return(NULL);
  }
  if (sprintf(str, "%.8x%.8x", ZENPolyLgt(PX, Rg), ZENPolyDeg(PX, Rg)) < 0) {
    ZENFree(str);
    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPUTTOSTRING, ZEN_SPRINTF);
    return(NULL);
  }
  for(i = 0; i <= ZENPolyDeg(PX, Rg); i++) {
    elt = ZENPolyGetCoeffPtr(PX, i, Rg);

    if (!(s1 = ZENEltPutToString(elt, Rg))) {
      ZENFree(str);
      return(NULL);
    }
    
    if (!(str = (char *)ZENRealloc(str, sizeof(char)*(strlen(str)+strlen(s1)+1)))) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPUTTOSTRING, ZEN_ALLOC);
      ZENFree(s1);
      return(NULL);
    }
    
    if (!strcat(str, s1)) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPUTTOSTRING, ZEN_STRCAT);
      ZENFree(str); ZENFree(s1);
      return(NULL);
    }
    
    ZENFree(s1);
  }
  return(str);
}


/* \TEX{
   \title{Printing to file.}
   \input{A stream |file|, an allocated |ZENPoly| |PX| of a finite field
   |Rg|, a character |var| (usually 'X') and a base |base|$\in \{2,\ldots, 16\}$.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |PX| in base |base| to |file|.}    
   }
   */
int _ZENPolyPrintToFile(file, PX, var, base, Rg)
     FILE *file;
     ZENPoly PX;
     int var;
     int base;
     ZENRing Rg;  
{ 
  int i, first;
  ZENElt elt;


  first = 1;
  if (ZENPolyLgt(PX, Rg) > ZENPolyDeg(PX, Rg)) {
    first = 0;
    switch(ZENPolyLgt(PX, Rg)) {
    case 1:
      if (fprintf(file, "(") < 0) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	return(ZERR);
      }

      elt = ZENPolyGetCoeffPtr(PX, ZENPolyLgt(PX, Rg), Rg);
      ZENEltSetToZero(elt, Rg);
      if (ZENEltPrintToFile(file, elt, base, Rg)) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	return(ZERR);
      }

      if (fprintf(file, ")*%c",var) < 0) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	return(ZERR);
      }
      break;
      
    default:
      if (fprintf(file, "(") < 0) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	return(ZERR);
      }

      elt = ZENPolyGetCoeffPtr(PX, ZENPolyLgt(PX, Rg), Rg);
      ZENEltSetToZero(elt, Rg);
      if (ZENEltPrintToFile(file, elt, base, Rg)) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	return(ZERR);
      }

      if (fprintf(file, ")*%c^%d", var,ZENPolyLgt(PX, Rg)) < 0) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	return(ZERR);
      }
    }
  }

  for(i = ZENPolyDeg(PX, Rg); i >= 0; i--) {
    
    elt = ZENPolyGetCoeffPtr(PX, i, Rg);

    switch(i) {
    case 0:
      if (!first && !ZENEltIsZero(elt, Rg)) {
	if (fprintf(file, "+") < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
      }
      if (first || !ZENEltIsZero(elt, Rg)) {
	first = 0;
	if (fprintf(file, "(") < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
	if (ZENEltPrintToFile(file, elt, base, Rg)) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
	if (fprintf(file, ")") < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
      }
      break;
      
    case 1:
      if (!ZENEltIsZero(elt, Rg)) {
	if (!first) {
	  if (fprintf(file, "+") < 0) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	    return(ZERR);
	  }
	}
	else {
	  first = 0;
	}

	if (fprintf(file, "(") < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
	if (ZENEltPrintToFile(file, elt, base, Rg)) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
	if (fprintf(file, ")*%c",var) < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
      }
      break;
      
    default:
      if (!ZENEltIsZero(elt, Rg)) {
	if (!first) {
	  if (fprintf(file, "+") < 0) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	    return(ZERR);
	  }
	}
	else {
	  first = 0;
	}

	if (fprintf(file, "(") < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
	if (ZENEltPrintToFile(file, elt, base, Rg)) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
	if (fprintf(file, ")*%c^%d", var, i) < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
      }
    }
  }
  return(0);
}

/* \TEX{
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |ZENPoly| |PX| of a finite field
   |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |PX| to |file|.}    
   }
   */
int _ZENPolyPutToFile(file, PX, Rg)
     FILE *file;
     ZENPoly PX;
     ZENRing Rg;  
{
  ZENElt elt;
   int i;

 
  if (fprintf(file, "%.8x%.8x", ZENPolyLgt(PX, Rg), ZENPolyDeg(PX, Rg)) < 0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYPUTTOFILE, ZEN_FPRINTF);
    return(ZERR);
  }
  for(i = 0; i <= ZENPolyDeg(PX, Rg); i++) {
    elt = ZENPolyGetCoeffPtr(PX, i, Rg);

    if (ZENEltPutToFile(file, elt, Rg)) {
      return(ZERR);
    }
  }
  return(0);
}

/* \TEX{
   \title{Reading from file.}
   \input{A stream |file| and a pointer on a polynomial |PX| of a finite field |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|PX| is allocated and filled with the |ZENPoly| read in |file| if
   no error occurred.}
   }
   */
int _ZENPolyReadFromFile(PX, file, var, base, Rg)
     ZENPoly *PX;
     FILE *file;
     int var;
     int base;
     ZENRing Rg;
{
  ZENElt elt;
  int c;
  int i;


  /* Alocating a buffer */
  if (ZENEltCreate(elt, Rg)) {
    return(ZERR);
  }
  *PX = NULL;
  c = fgetc(file);
  
  /* Parsing */
  for(;;) {
    for(; isspace(c); c = fgetc(file)) ;
    if (c != '(') {
      ZENSetError(ZEN_LIB_ZEG, ZEN_POLYREADFROMFILE, ZEN_SYNTAX);
      ZENEltFree(elt, Rg);
      if (*PX) {
	ZENPolyFree(*PX, Rg);
      }
      return(ZERR);
    }
    
    if (ZENEltReadFromFile(elt, file, base, Rg)) {
      ZENEltFree(elt, Rg);
      if (*PX) {
	ZENPolyFree(*PX, Rg);
      }
      return(ZERR);
    }

    for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
    if (c != ')') {
      ZENSetError(ZEN_LIB_ZEG, ZEN_POLYREADFROMFILE, ZEN_SYNTAX);
      ZENEltFree(elt, Rg);
      if (*PX) {
	ZENPolyFree(*PX, Rg);
      }
      return(ZERR);
    }
    
    for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
    if (c == '*') {
      for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
      if (c == var) {
	for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
	switch (c) {
	case '^':
	  if (fscanf(file, "%d", &i) != 1) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYREADFROMFILE, ZEN_SYNTAX);
	    ZENEltFree(elt, Rg);
	    if (*PX) {
	      ZENPolyFree(*PX, Rg);
	    }
	    return(ZERR);
	  }
	  if (!*PX) {
	    if (!(*PX = ZENPolyCreate(i, Rg))) {
	      ZENEltFree(elt, Rg);
	      return(ZERR);
	    }
	    ZENPolySetToXi(*PX, i, Rg);
	    ZENPolySetToZero(*PX, Rg);
	  }	  
	  if (!ZENEltIsZero(elt, Rg)) {
	    ZENPolySetCoeff(*PX, i, elt, Rg);
	    if (i > ZENPolyDeg(*PX, Rg)) {
	      ZENPolyDeg(*PX, Rg) = i;
	    }
	  }	
	  for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
	  break;
	case '+':
	  if (!*PX) {
	    if (!(*PX = ZENPolyCreate(1, Rg))) {
	      ZENEltFree(elt, Rg);
	      return(ZERR);
	    }
	    ZENPolySetToXi(*PX, 1, Rg);
	    ZENPolySetToZero(*PX, Rg);
	  }
	  if (!ZENEltIsZero(elt, Rg)) {
	    ZENPolySetCoeff(*PX, 1, elt, Rg);
	    if (1 > ZENPolyDeg(*PX, Rg)) {
	      ZENPolyDeg(*PX, Rg) = 1;
	    }
	  }
	  break;
	default:
	  if (!*PX) {
	    if (!(*PX = ZENPolyCreate(1, Rg))) {
	      ZENEltFree(elt, Rg);
	      return(ZERR);
	    }
	    ZENPolySetToXi(*PX, 1, Rg);
	    ZENPolySetToZero(*PX, Rg);
	  }
	  if (!ZENEltIsZero(elt, Rg)) {
	    ZENPolySetCoeff(*PX, 1, elt, Rg);
	    if (1 > ZENPolyDeg(*PX, Rg)) {
	      ZENPolyDeg(*PX, Rg) = 1;
	    }
	  }
	  ZENEltFree(elt, Rg);
	  if ((c != EOF) && (ungetc(c, file) == EOF)) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYREADFROMFILE, ZEN_SYNTAX);
	    if (*PX) {
	      ZENPolyFree(*PX, Rg);
	    }
	    return(ZERR);
	  }
	  return(0);
	}
      }
      else {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYREADFROMFILE, ZEN_SYNTAX);
	ZENEltFree(elt, Rg);
	if (*PX) {
	  ZENPolyFree(*PX, Rg);
	}
	return(ZERR);
      }
    }
    else {
      if (!*PX) {
	if (!(*PX = ZENPolyCreate(0, Rg))) {
	  ZENEltFree(elt, Rg);
	  return(ZERR);
	}
	ZENPolyDeg(*PX, Rg) =  0;
      }
      ZENPolySetCoeff(*PX, 0, elt, Rg);
      ZENEltFree(elt, Rg);
      if ((c != EOF) && (ungetc(c, file) == EOF)) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_POLYREADFROMFILE, ZEN_SYNTAX);
	if (*PX) {
	  ZENPolyFree(*PX, Rg);
	}
	return(ZERR);
      }
      return(0);
    }
    
    if (c != '+') {
      break;
    }
    c = fgetc(file);
  }

  ZENEltFree(elt, Rg);
  if ((c != EOF) && (ungetc(c, file) == EOF)) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYREADFROMFILE, ZEN_SYNTAX);
    if (*PX) {
      ZENPolyFree(*PX, Rg);
    }
    return(ZERR);
  }
  return(0);
}

/* \TEX{
   \title{Getting from file in an internal representation.}
   \input{A stream |file|, and a pointer on polynomial |PX| of a finite
   field |ZENPoly|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|*PX| is filled with the polynomial read in |file|}    
   }
   */
int _ZENPolyGetFromFile(PX, file, Rg)
     ZENPoly *PX;
     FILE *file;
     ZENRing Rg;
{
  ZENElt elt;
  int deg, length, i;

  if (fscanf(file, "%8x%8x", &length, &deg) != 2) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYGETFROMFILE, ZEN_SSCANF);
    return(ZERR);
  }

  if (ZENEltCreate(elt, Rg)) {
    return(ZERR);
  }
  if (!(*PX = ZENPolyCreate(length, Rg))) {
    ZENEltFree(elt, Rg);
    return(ZERR);
  }

  for(i = 0; i <= deg; i++) {
    if (ZENEltGetFromFile(elt, file, Rg)) {
      ZENEltFree(elt, Rg);
      ZENPolyFree(*PX, Rg);
      return(ZERR);
    }
    ZENPolySetCoeff(*PX, i, elt, Rg);
  }
  ZENPolyDeg(*PX, Rg) =  deg;
  ZENEltFree(elt, Rg);
  return(0);
}


/* \TEX{
   \title{Getting a BigNum from a Polynomial.}
   \input{A pointer on an integer |p_pl|, and a polynomial |P| of
   a |ZENRing| |Rg|.}
   \output{A |BigNum| of size |*p_pl| if an error occured, |ZNULL|
   otherwise.}
   } */
BigNum _ZENPolyToZ(p_pl, P, Rg)
     BigNumLength *p_pl;
     ZENPoly P;
     ZENRing Rg;
{
  BigNum p, tmp1;
  int i;
  BigNumLength pl;

  pl = (ZENPolyDeg(P, Rg)+1)*ZENRingQl(Rg);  
  if (!(p = ZBNC(pl+1))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYTOZ, ZEN_ALLOC);
    return(ZNULL);
  } 
  if (!(tmp1 = ZBNC(pl+1))) {
    ZBNF(p);
    ZENSetError(ZEN_LIB_ZEG, ZEN_POLYTOZ, ZEN_ALLOC);
    return(ZNULL);
  }
  
  ZENEltToZ(p, p_pl, ZENPolyGetCoeffPtr(P, ZENPolyDeg(P, Rg), Rg), Rg);
  ZBNSetToZero(p+(*p_pl), pl-(*p_pl));

  for(i = ZENPolyDeg(P, Rg)-1;
      i >= 0 ;
      i--) {  
    ZBNSetToZero(tmp1, pl);
    ZBNMultiply(tmp1, pl,
		ZENRingQ(Rg), ZENRingQl(Rg),
		p, *p_pl);
	      
    ZENEltToZ(p, p_pl, ZENPolyGetCoeffPtr(P, i, Rg), Rg); 

    ZBNSetToZero(p+(*p_pl), pl-(*p_pl));
    ZBNAdd(p, pl, tmp1, pl, 0);
    *p_pl = ZBNNumDigits(p, pl);
  }

  ZBNF(tmp1);
  return(p);
}

/* \TEX{
   \title{Getting a polynomial from a BigNum.}
   \input{A |BigNum| (|p|, |pl|) and a finite Ring |Rg|.}
   \output{|ZNULL| if an error occured, otherwise a polynomial which once, evaluated in |ZENRingQ(Rg)|, returns (|n|, |nl|).}
   }
   */
ZENPoly _ZENZToPoly(p, pl, Rg)
     BigNum p;
     BigNumLength pl;
     ZENRing Rg;
{
  int deg, i;
  ZENPoly P;
  BigNum tmp1, tmp2;
  ZENElt elt;
  BigNumLength pbisl;

  if (ZBNIsZero(p, pl)) {
    deg = 0;
  }
  else {
    deg = 3+(int)(((pl-1)*SIZE_BLOC+ZBNNumBitsInDigit(p[pl-1]))
		  /
		  ((ZENRingQl(Rg)-1)*SIZE_BLOC+ZBNNumBitsInDigit(ZENRingQ(Rg)[ZENRingQl(Rg)-1])-1));
  }

  if (!(P = ZENPolyCreate(deg, Rg))) {
    return(ZNULL);
  }
  
  if (!(tmp1 = ZBNC(pl+1))) {
    ZENPolyFree(P, Rg);
    ZENSetError(ZEN_LIB_ZEG, ZEN_ZTOPOLY, ZEN_ALLOC);
    return(ZNULL);
  }
  
  if (!(tmp2 = ZBNC(pl+1))) {
    ZENPolyFree(P, Rg); ZBNF(tmp1);
    ZENSetError(ZEN_LIB_ZEG, ZEN_ZTOPOLY, ZEN_ALLOC);
    return(ZNULL);
  }
  
  ZBNAssign(tmp2, p ,pl);
  pbisl = ZBNNumDigits(tmp2, pl);
  for(i = 0; ; i++) {  
    if (ZBNCompare(tmp2, pbisl,
		   ZENRingQ(Rg), ZENRingQl(Rg)) == BN_LT) {
      break;
    }

    ZBNAssign(tmp1, tmp2, pbisl);
    ZBNDigit(tmp1, pbisl) = 0;
    ZBNDivide(tmp1, pbisl+1,
	      ZENRingQ(Rg), ZENRingQl(Rg));
    
    elt = ZENPolyGetCoeffPtr(P, i, Rg);
    ZENZToElt(elt, tmp1, ZENRingSizeElt(Rg), Rg);
    ZENPolySetCoeff(P,i,elt,Rg);
    
    pbisl = ZBNNumDigits(tmp1+ZENRingQl(Rg),
			 pbisl+1-ZENRingQl(Rg));
    ZBNAssign(tmp2, tmp1+ZENRingQl(Rg), pbisl);
  }
  
  elt = ZENPolyGetCoeffPtr(P, i, Rg);
  ZENZToElt(elt, tmp2, pbisl, Rg);
  ZENPolySetCoeff(P, i, elt, Rg);
  ZENPolyDeg(P, Rg) = i;
  ZENPolyUpdateDegree(P, Rg);
  
  ZBNF(tmp1); ZBNF(tmp2);
  return(P);
}

