/*
 * zen/zeg/c/poly_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Polynomial multiplication}
   
   Two functions are provided to do multiplications: |_ZENPolySquare| to square a
   |ZENPoly| and |_ZENPolyMultily| to compute the multiplication of two |ZENPoly|.
   }
   */
#include <stdio.h>
#include "zeg.h"

/* \TEX{
   \title{Squaring in characteristic 2.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZENPolySquareChar2(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  ZENElt elt;
  int i;

  for(i = ZENPolyDeg(PX, Rg); i; i--) {

    elt = ZENPolyGetCoeffPtr(RX, 2*i, Rg);
    ZENEltSquare(elt,
		 ZENPolyGetCoeffPtr(PX, i, Rg),
		 Rg);
    ZENPolySetCoeff(RX, 2*i, elt, Rg);

    elt = ZENPolyGetCoeffPtr(RX, 2*i-1, Rg);
    ZENEltSetToZero(elt, Rg);
    ZENPolySetCoeff(RX, 2*i-1, elt, Rg);
  }
  elt = ZENPolyGetCoeffPtr(RX, 0, Rg);
  ZENEltSquare(elt,
	       ZENPolyGetCoeffPtr(PX, 0, Rg),
	       Rg);
  ZENPolySetCoeff(RX, 0, elt, Rg);

  ZENPolyDeg(RX, Rg) = 2*ZENPolyDeg(PX, Rg);
  ZENPolyUpdateDegree(RX, Rg);
  return(0);
}


/* \TEX{
   \title{Squaring in characteristic $p\neq 2$.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZENPolySquareCharp(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, j;
  ZENElt e;

  for(i = ZENPolyDeg(PX, Rg)+1; i <= 2*ZENPolyDeg(PX, Rg); i++) {
    e = ZENPolyGetCoeffPtr(RX, i, Rg);
    ZENEltSetToZero(e, Rg);
    ZENPolySetCoeff(RX, i, e, Rg);
  }

  if ((i = ZENPolyDeg(PX, Rg))) {
    ZENEltSquare(ZENRingHbuf(Rg), ZENPolyGetCoeffPtr(PX, i, Rg), Rg);
    ZENPolySetCoeff(RX, 2*i, ZENRingHbuf(Rg), Rg);
    
    for(j = i-1; j >= 0; j--) {
      ZENEltMultiply(ZENRingHbuf(Rg),
		     ZENPolyGetCoeffPtr(PX, j, Rg),
		     ZENPolyGetCoeffPtr(PX, i, Rg),
		     Rg);
      ZENEltAdd(ZENRingHbuf(Rg), ZENRingHbuf(Rg), Rg);
      ZENPolySetCoeff(RX, i+j, ZENRingHbuf(Rg), Rg);
    }
  }

  for(i = ZENPolyDeg(PX, Rg)-1; i > 0; i--) {
    
    ZENEltSquare(ZENRingHbuf(Rg), ZENPolyGetCoeffPtr(PX, i, Rg), Rg);
    e = ZENPolyGetCoeffPtr(RX, 2*i, Rg);
    ZENEltAdd(e, ZENRingHbuf(Rg), Rg);
    ZENPolySetCoeff(RX, 2*i, e, Rg);
    
    for(j = i-1; j > 0; j--) {
      ZENEltMultiply(ZENRingHbuf(Rg),
		     ZENPolyGetCoeffPtr(PX, j, Rg),
		     ZENPolyGetCoeffPtr(PX, i, Rg),
		     Rg);
      ZENEltAdd(ZENRingHbuf(Rg), ZENRingHbuf(Rg), Rg);
      e = ZENPolyGetCoeffPtr(RX, i+j, Rg);
      ZENEltAdd(e, ZENRingHbuf(Rg), Rg);
      ZENPolySetCoeff(RX, i+j, e, Rg);
    }
    ZENEltMultiply(ZENRingHbuf(Rg),
		   ZENPolyGetCoeffPtr(PX, j, Rg),
		   ZENPolyGetCoeffPtr(PX, i, Rg),
		   Rg);
    ZENEltAdd(ZENRingHbuf(Rg), ZENRingHbuf(Rg), Rg);
    ZENPolySetCoeff(RX, i+j, ZENRingHbuf(Rg), Rg);
  }
  e = ZENPolyGetCoeffPtr(RX, 0, Rg);
  ZENEltSquare(e,
	       ZENPolyGetCoeffPtr(PX, 0, Rg),
	       Rg);
  ZENPolySetCoeff(RX, 0, e, Rg);

  ZENPolyDeg(RX, Rg) = 2*ZENPolyDeg(PX, Rg);
  ZENPolyUpdateDegree(RX, Rg);
  return(0);
}

/* \TEX{
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |QX|\times|PX|$.}
   \note{One must have $|RX|\neq|QX|$.}
   }
   */
int _ZENPolyMultiply(RX, QX, PX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i, j;
  ZENElt e;

  if (ZENPolyIsZero(PX, Rg)) {
    ZENPolyAssign(RX, PX, Rg);
    return(0);
  }

  if (ZENPolyIsZero(QX, Rg)) {
    ZENPolyAssign(RX, QX, Rg);
    return(0);
  }

  i = ZENPolyDeg(PX, Rg);
  for(j = ZENPolyDeg(QX, Rg); j >= 0; j--) {
    ZENEltMultiply(ZENRingHbuf(Rg),
		   ZENPolyGetCoeffPtr(QX, j, Rg),
		   ZENPolyGetCoeffPtr(PX, i, Rg),
		   Rg);
    ZENEltMultiply(ZENRingHbuf(Rg),
		   ZENPolyGetCoeffPtr(QX, j, Rg),
		   ZENPolyGetCoeffPtr(PX, i, Rg),
		   Rg);
    ZENPolySetCoeff(RX, i+j, ZENRingHbuf(Rg), Rg);
  }

  for(i = ZENPolyDeg(PX, Rg)-1; i >= 0; i--) {
    for(j = ZENPolyDeg(QX, Rg); j; j--) {
      ZENEltMultiply(ZENRingHbuf(Rg),
		     ZENPolyGetCoeffPtr(QX, j, Rg),
		     ZENPolyGetCoeffPtr(PX, i, Rg),
		     Rg);
      e = ZENPolyGetCoeffPtr(RX, i+j, Rg);
      ZENEltAdd(e, ZENRingHbuf(Rg), Rg);
      ZENPolySetCoeff(RX, i+j, e, Rg);
    }
    ZENEltMultiply(ZENRingHbuf(Rg),
		   ZENPolyGetCoeffPtr(QX, j, Rg),
		   ZENPolyGetCoeffPtr(PX, i, Rg),
		   Rg);
    ZENPolySetCoeff(RX, i+j, ZENRingHbuf(Rg), Rg);
  }
  ZENPolyDeg(RX, Rg) = ZENPolyDeg(PX, Rg) + ZENPolyDeg(QX, Rg);
  ZENPolyUpdateDegree(RX, Rg);
  return(0);
}

/* \TEX{
   \title{Multiplication by a scalar.}
   \input{Two polynomials |PX|, |RX| and a |ZENElt| |e| of a finite ring |Rg|.}
   }
	 */
void _ZENPolyMultiplyScalar(RX, PX, e, Rg)
     ZENPoly RX, PX;
     ZENElt e;
     ZENRing Rg;
{
  ZENElt elt;
  int i;

  if (ZENEltIsZero(e, Rg)) {
    ZENPolySetToZero(RX, Rg);
    return;
  }
  if (ZENEltIsOne(e, Rg)) {
    ZENPolyAssign(RX, PX, Rg);
    return;
  }
  for(i = 0; i <= ZENPolyDeg(PX, Rg); i++) {
    elt = ZENPolyGetCoeffPtr(RX, i, Rg);
    ZENEltMultiply(elt, e,
		   ZENPolyGetCoeffPtr(PX, i, Rg),
		   Rg);
    ZENPolySetCoeff(RX, i, elt, Rg);
  }
  ZENPolyDeg(RX, Rg) = ZENPolyDeg(PX, Rg);
  ZENPolyUpdateDegree(RX, Rg);
}

/* \TEX{
   \title{Scalar product.}
   \input{Two polynomials |PX|, |RX| and a pointer to a |ZENElt| |e| of a
   finite ring |Rg|.}
   \side{$|*e| = \sum p_iq_i$}
   }
   */
void _ZENPolyDot(e, PX, QX, Rg)
     ZENElt *e;
     ZENPoly PX, QX;
     ZENRing Rg;
{
  int i, deg;

  
  ZENEltMultiply(*e,
		 ZENPolyGetCoeffPtr(PX, 0, Rg),
		 ZENPolyGetCoeffPtr(QX, 0, Rg),
		 Rg);

  deg = ZENPolyDeg(PX, Rg) < ZENPolyDeg(QX, Rg) ?
    ZENPolyDeg(PX, Rg) :  ZENPolyDeg(QX, Rg);

  for(i = 1; i <= deg; i++) {
    ZENEltMultiply(ZENRingHbuf(Rg),
		   ZENPolyGetCoeffPtr(PX, i, Rg),
		   ZENPolyGetCoeffPtr(QX, i, Rg),
		   Rg);
    ZENEltAdd(*e, ZENRingHbuf(Rg), Rg);
  }
}



/* \TEX{
   \title{Evaluation.}
   \input{A polynomial |PX| and two elements |f| and |e| of a
   finite ring |Rg|.}
   \side{$|f| = |PX|(e)$.}
   }
   */
void _ZENPolyEval(f, PX, e, Rg)
     ZENElt *f, e;
     ZENPoly PX;
     ZENRing Rg;
{
  int i;

  if (!ZENPolyDeg(PX, Rg)) {
    ZENEltAssign(*f, ZENPolyGetCoeffPtr(PX, 0, Rg), Rg);
    return;
  }

  ZENEltMultiply(ZENRingHbuf(Rg), e, ZENPolyGetCoeffPtr(PX, ZENPolyDeg(PX, Rg), Rg), Rg);
  for(i = ZENPolyDeg(PX, Rg)-1; i >= 1; i--) {
    ZENEltAdd(ZENRingHbuf(Rg), ZENPolyGetCoeffPtr(PX, i, Rg), Rg);
    ZENEltMultiply(ZENRingHbuf(Rg), e, ZENRingHbuf(Rg), Rg);
  }
  ZENEltAdd(ZENRingHbuf(Rg), ZENPolyGetCoeffPtr(PX, 0, Rg), Rg);
  ZENEltAssign(*f, ZENRingHbuf(Rg), Rg);
}

/* \TEX{
   \title{Lazy division.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZEN_NO_INVERSE| if a factor of one definition polynomial was dicovered, |ZENERR| for an error}
   \note{|PX| is destroyed.}
   \side{|ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZENPolyDivideLazy(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int i, j, deg, IsNotOne;
  ZENElt e;


  if(ZENEltCreate(e,Rg))
    return(ZERR);
  
  if (!ZENEltIsOne(ZENPolyGetCoeffPtr(QX, ZENPolyDeg(QX, Rg), Rg), Rg)) {
    IsNotOne = 1;
    
    switch(ZENEltInverse(ZENRingIbuf(Rg),ZENPolyGetCoeffPtr(QX, ZENPolyDeg(QX, Rg), Rg), Rg)) {
    case ZENERR:
      ZENEltFree(e,Rg);
      return(ZERR);
    case ZEN_NO_INVERSE:
      ZENEltFree(e,Rg);
      return(ZEN_NO_INVERSE);
    }
  }
  else {
    IsNotOne = 0;
  }

  for(i = ZENPolyDeg(PX, Rg);
      i >= ZENPolyDeg(QX, Rg);
      i--) {
    if (IsNotOne) {
      ZENPolyGetCoeff(e,PX, i, Rg);
      ZENEltMultiply(e, ZENRingIbuf(Rg),
		     ZENPolyGetCoeffPtr(PX, i, Rg),
		     Rg);
      ZENPolySetCoeff(PX, i, e, Rg);
    }
    
    for(j =  ZENPolyDeg(QX, Rg)-1;
	j >= 0;
	j--) {
      ZENEltMultiply(ZENRingHbuf(Rg),
		     ZENPolyGetCoeffPtr(PX, i, Rg),
		     ZENPolyGetCoeffPtr(QX, j, Rg),
		     Rg);
      ZENPolyGetCoeff(e,PX,i+j-ZENPolyDeg(QX, Rg),Rg);
      ZENEltSubtract(e, ZENRingHbuf(Rg), Rg);
      ZENPolySetCoeff(PX, i+j-ZENPolyDeg(QX, Rg), e, Rg);
    }
  }
  deg =  ZENPolyDeg(QX, Rg);
  for(i = ZENPolyDeg(PX, Rg); i >= ZENPolyDeg(QX, Rg); i--) {
    ZENPolySetCoeff(RX, i-ZENPolyDeg(QX, Rg),
		    ZENPolyGetCoeffPtr(PX, i, Rg),
		    Rg);
  }
  ZENPolyDeg(RX, Rg) = ZENPolyDeg(PX, Rg)-ZENPolyDeg(QX, Rg);
  if (ZENPolyDeg(RX, Rg) < 0) {
    ZENPolySetToZero(RX, Rg);
  }
  for(; i >= 0; i--) {
    if (!ZENEltIsZero(ZENPolyGetCoeffPtr(PX, i, Rg), Rg)) {
      break;
    }
  }
  for(deg = i; i >= 0; i--) {
    ZENPolySetCoeff(MX, i, ZENPolyGetCoeffPtr(PX, i, Rg), Rg);
  }
  if (deg < 0) {
    ZENPolySetToZero(MX, Rg);
  }
  else {
    ZENPolyDeg(MX, Rg) = deg;
  }

  ZENEltFree(e,Rg);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Divide.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZEN_NO_INVERSE| if a factor of one definition polynomial was dicovered, |ZENERR| for an error}
   }
   */
int _ZENPolyDivide(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int ret;
  ZENPoly Pbuf;

  if (!(Pbuf = ZENPolyCopy(PX, Rg))) {
    return(ZERR);
  }
  ret = ZENPolyDivideLazy(RX, MX, Pbuf, QX, Rg);
  ZENPolyFree(Pbuf, Rg);
  return(ret);
}

/* \TEX{
   \title{Make a polynomial monic.}
   \input{Two polynomials |PX| and |RX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was
   impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{|R(X)| is filled with |P(X)| divided by its highest
   coefficient. |ZENRingFact(Rg)| is filled with a factor of a modulo
   if |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZENPolyMakeMonic (RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  ZENElt inv;

  if (ZENEltIsOne(ZENPolyGetCoeffPtr(PX, ZENPolyDeg(PX, Rg), Rg), Rg)) {
    ZENPolyAssign(RX, PX, Rg);
    return(0);
  }

  if (ZENEltCreate(inv, Rg)) {
    return(ZERR);
  }
  
  switch(ZENEltInverse(inv, ZENPolyGetCoeffPtr(PX, ZENPolyDeg(PX, Rg), Rg), Rg)) {
  case ZEN_NO_INVERSE:
    ZENEltFree(inv, Rg);
    return(ZEN_NO_INVERSE);
  case ZENERR:
    ZENEltFree(inv, Rg);
    return(ZERR);
  case ZEN_HAS_INVERSE:
    ZENPolyMultiplyScalar(RX, PX, inv, Rg);
  }
  ZENEltFree(inv, Rg);
  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Lazy mudulo computation.}
   \input{A polynomial |PX| to divide by
   |QX|$=X^{\mbox{degree}(QX)}+sum_{i=0}^{|k|}q_iX^i$} in a finite
   ring |Rg|. |MX| will be the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| for an error}
   \note{|PX| is destroyed.}
   */
int _ZENPolyModuloLazy(MX, PX, QX, k, Rg)
     ZENPoly MX, PX, QX;
     int k;
     ZENRing Rg;
{
  int i, j, deg;
  ZENElt e;

  if(ZENEltCreate(e,Rg))
    return(ZERR);
  
  for(i = ZENPolyDeg(PX, Rg);
      i >= ZENPolyDeg(QX, Rg);
      i--) {
    for(j =  k;
	j >= 0;
	j--) {
      ZENEltMultiply(ZENRingHbuf(Rg),
		     ZENPolyGetCoeffPtr(PX, i, Rg),
		     ZENPolyGetCoeffPtr(QX, j, Rg),
		     Rg);
      ZENPolyGetCoeff(e,PX,i+j-ZENPolyDeg(QX, Rg),Rg);
      ZENEltSubtract(e, ZENRingHbuf(Rg), Rg);
      ZENPolySetCoeff(PX, i+j-ZENPolyDeg(QX, Rg), e, Rg);
    }
  }
  for( ; i >= 0; i--) {
    if (!ZENEltIsZero(ZENPolyGetCoeffPtr(PX, i, Rg), Rg)) {
      break;
    }
  }
  for(deg = i; i >= 0; i--) {
    ZENPolySetCoeff(MX, i, ZENPolyGetCoeffPtr(PX, i, Rg), Rg);
  }
  if (deg < 0) {
    ZENPolySetToZero(MX, Rg);
  }
  else {
    ZENPolyDeg(MX, Rg) = deg;
  }

  ZENEltFree(e,Rg);
  return(ZEN_HAS_INVERSE);
}



/* \TEX{
   \title{Reciprocal of a polynomial whose degree is a power of 2.}
   \input{Two |ZENPoly| |RX| and |AX| of a finite ring |Rg|.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| for an error}
   \side{$|RX| = \left[X^{2d}/|PX|\right]$ where $d$ is the degree of
   $|RX|$, $d$ is a power of 2.}
   }
   */
int _ZENPolyBinaryReciprocal(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int k, i;
  ZENElt e, t, h, elt;
  ZENPoly pbuf1, pbuf2, pbuf3;

  if (ZENEltCreate(t, Rg)) {
    return(ZERR);
  }
  if (ZENEltCreate(e, Rg)) {
    ZENEltFree(t, Rg);
    return(ZERR);
  }
  if (ZENEltCreate(h, Rg)) {
    ZENEltFree(t, Rg); ZENEltFree(e, Rg);
    return(ZERR);
  }
  if (!(pbuf1 = ZENPolyCreate(ZENPolyDeg(PX, Rg), Rg))) {
    ZENEltFree(t, Rg); ZENEltFree(e, Rg); ZENEltFree(h, Rg);
    return(ZERR);
  }
  if (!(pbuf2 = ZENPolyCreate(ZENPolyDeg(PX, Rg), Rg))) {
    ZENEltFree(t, Rg); ZENEltFree(e, Rg); ZENEltFree(h, Rg);
    ZENPolyFree(pbuf1, Rg);
    return(ZERR);
  }
  if (!(pbuf3 = ZENPolyCreate(2*ZENPolyDeg(PX, Rg), Rg))) {
    ZENEltFree(t, Rg); ZENEltFree(e, Rg); ZENEltFree(h, Rg);
    ZENPolyFree(pbuf1, Rg); ZENPolyFree(pbuf2, Rg);
    return(ZERR);
  }

  switch(ZENEltInverse(h, ZENPolyGetCoeffPtr(PX, ZENPolyDeg(PX, Rg), Rg), Rg)) {
  case ZENERR:
    ZENEltFree(t, Rg); ZENEltFree(e, Rg); ZENEltFree(h, Rg);
    ZENPolyFree(pbuf1, Rg); ZENPolyFree(pbuf2, Rg); ZENPolyFree(pbuf3, Rg);
    return(ZERR);
  case ZEN_NO_INVERSE:
    ZENEltFree(t, Rg); ZENEltFree(e, Rg); ZENEltFree(h, Rg);
    ZENPolyFree(pbuf1, Rg); ZENPolyFree(pbuf2, Rg); ZENPolyFree(pbuf3, Rg);
    return(ZEN_NO_INVERSE);
  }
  ZENEltMultiply(e, h, ZENPolyGetCoeffPtr(PX, ZENPolyDeg(PX, Rg)-1, Rg), Rg);
  ZENEltNegate(e, e, Rg);
  
  ZENPolyDeg(pbuf1, Rg) = 0;
  ZENPolySetCoeff(pbuf1, 0, h, Rg);
  for(k = 2; k <= PX->deg ; k <<= 1) {
    ZENPolySquare(pbuf2, pbuf1, Rg);

    ZENPolyDeg(pbuf3, Rg) = k-1;
    for(i =  0; i <= k-1; i++) {
      ZENPolySetCoeff(pbuf3, i, ZENPolyGetCoeffPtr(PX, ZENPolyDeg(PX, Rg)-k+1+i, Rg), Rg);
    } 
    ZENPolyMultiply(pbuf3, pbuf2, pbuf3, Rg);
    
    ZENEltSquare(e, e, Rg);
    if (k != 2) {
      ZENEltMultiply(t,
		     ZENPolyGetCoeffPtr(pbuf3, k-3, Rg),
		     ZENPolyGetCoeffPtr(PX, ZENPolyDeg(PX, Rg), Rg),
		     Rg);
      ZENEltSubtract(e, t, Rg);
    }
    ZENEltMultiply(t,
		   ZENPolyGetCoeffPtr(PX, ZENPolyDeg(PX, Rg)-k, Rg),
		   h, Rg);
    ZENEltSubtract(e, t, Rg);

    for(i = ZENPolyDeg(pbuf1, Rg); i >= 0; i--) {
      elt = ZENPolyGetCoeffPtr(pbuf1, i, Rg);
      ZENEltAdd(elt,  elt, Rg);
      ZENPolySetCoeff(pbuf1, i+k/2, elt, Rg);
    }
    for(i = 0; i < k/2; i++) {
      elt = ZENPolyGetCoeffPtr(pbuf1, i, Rg);
      ZENEltSetToZero(elt, Rg);
      ZENPolySetCoeff(pbuf1, i, elt, Rg);
    }
    for(i = 0; i <= k-1; i++) {
      elt =  ZENPolyGetCoeffPtr(pbuf1, i, Rg);
      ZENEltSubtract(elt, ZENPolyGetCoeffPtr(pbuf3, k-2+i, Rg), Rg);
      ZENPolySetCoeff(pbuf1, i, elt, Rg);
    }
    ZENPolyDeg(pbuf1, Rg) = k-1;
  }
  
  elt = ZENPolyGetCoeffPtr(RX, 0, Rg);
  ZENEltMultiply(elt, e, h, Rg);
  ZENPolySetCoeff(RX, 0, elt, Rg);
  for(i = 1; i <= ZENPolyDeg(PX, Rg); i++) {
    ZENPolySetCoeff(RX, i, ZENPolyGetCoeffPtr(pbuf1, i-1, Rg), Rg);
  }
  ZENPolyDeg(RX, Rg) = ZENPolyDeg(PX, Rg);

  ZENEltFree(t, Rg); ZENEltFree(e, Rg); ZENEltFree(h, Rg);
  ZENPolyFree(pbuf1, Rg); ZENPolyFree(pbuf2, Rg); ZENPolyFree(pbuf3, Rg);
  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Reciprocal of a polynomial.}
   \input{Two |ZENPoly| |RX| and |AX| of a finite ring |Rg|.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| for an error}
   \side{$|RX| = \left[X^{2d}/|PX|\right]$ where $d$ is the degree of
   $|RX|$.}
   }
   */
int ZENPolyReciprocal(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int l, i, d; 
  ZENPoly pbuf;
  ZENElt elt;
  
  l = 1;
  while (l <  ZENPolyDeg(PX, Rg))
    l <<= 1;

  if (l ==  ZENPolyDeg(PX, Rg)) {
    return(ZENPolyBinaryReciprocal(RX, PX, Rg));
  }

  d = l - ZENPolyDeg(PX, Rg);
  
  if (!(pbuf = ZENPolyCreate(l, Rg))) {
    return(ZERR);
  }

  for(i = 0; i < d; i++) {
    elt = ZENPolyGetCoeffPtr(pbuf, i, Rg);
    ZENEltSetToZero(elt, Rg);
    ZENPolySetCoeff(pbuf, i, elt, Rg);
  }
  for(i = ZENPolyDeg(PX, Rg); i >= 0; i--) {
    ZENPolySetCoeff(pbuf, i+d, ZENPolyGetCoeffPtr(PX, i, Rg), Rg);
  }
  ZENPolyDeg(pbuf, Rg) = l;
  switch(ZENPolyBinaryReciprocal(pbuf, pbuf, Rg)) {
  case ZENERR:
    ZENPolyFree(pbuf, Rg);
    return(ZERR);
  case ZEN_NO_INVERSE:
    ZENPolyFree(pbuf, Rg);
    return(ZEN_NO_INVERSE);
  }
  for(i = 0; i <= ZENPolyDeg(PX, Rg); i++) {
    ZENPolySetCoeff(RX, i, ZENPolyGetCoeffPtr(pbuf, i+d, Rg), Rg);
  }
  ZENPolyDeg(RX, Rg) = ZENPolyDeg(PX, Rg);
  ZENPolyFree(pbuf, Rg);
  return(ZEN_HAS_INVERSE);
}

