/*
 * zen/zeg/c/poly_roots.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
  \TEX{
  \Subsection{Roots of polynomials.}

  For need to other part of the library, procedures to find roots of
  polynomials of degree smaller or equal than 2 and polynomials
  $X^p-X-\gamma$ are provided.
  
  } */

#include <stdio.h>
#include "zen.h"

/*
   \TEX{
   \title{Roots of a polynomial of degree 2 in finite fields of
   odd characteristic}
   \input{Two |ZENPoly| |P| and |roots| over a |ZENRing| |Rg|.}
   \output{\begin{itemize}
   \item[|ZENERR|] if an error occured,
   \item[|ZEN_HAS_INVERSE|] if the polynomial has one root.
   \end{itemize}}
   \side{The polynomial |roots| is set to a polynomial of degree -1, 0
   or 1, the coefficients of which are the roots of |P|.}
   \note{|P| must be allocated for at least degree 1.We must have
   $|P|\neq |roots|$. Valid only in finite fields.}
   } 
   */
int _ZENPolyRootsDegree2Charp(roots, P, Rg)
     ZENPoly roots, P;
     ZENRing Rg;
{
  int ret;
  ZENElt elt, t;

  ZENPolyDeg(roots, Rg) = 1;

  if (ZENEltCreate(t, Rg)) {
    return(ZERR);
  }

  ZENEltSquare(t, ZENPolyGetCoeffPtr(P, 1, Rg), Rg);
  elt = ZENPolyGetCoeffPtr(roots, 0, Rg);
  ZENEltMultiply(elt, ZENPolyGetCoeffPtr(P, 2, Rg),
		 ZENPolyGetCoeffPtr(P, 0, Rg),Rg);
  ZENEltSubtract(t, elt, Rg);
  ZENEltSubtract(t, elt, Rg);
  ZENEltSubtract(t, elt, Rg);
  ZENEltSubtract(t, elt, Rg);

  if ((ret = ZENEltSquareRoot(t, t, Rg))) {
    if (ret == 1) {
      ZENPolyDeg(roots, rg) = -1;
    }
    else {
      goto argh;
    }
  }
  ZENEltNegate(elt, ZENPolyGetCoeffPtr(P, 1, Rg), Rg);
  ZENEltSubtract(elt, t, Rg);
  ZENPolySetCoeff(roots, 0, elt, Rg);

  if (ZENEltIsZero(t, Rg)) {
    ZENPolyDeg(roots, Rg) = 0;
  }
  else {
    elt = ZENPolyGetCoeffPtr(roots, 1, Rg);
    ZENEltNegate(elt, ZENPolyGetCoeffPtr(P, 1, Rg), Rg);
    ZENEltAdd(elt, t, Rg);
  }
  
  ZENEltAssign(t, ZENPolyGetCoeffPtr(P, 2, Rg), Rg);
  ZENEltAdd(t, t, Rg);
  if (ZENEltInverse(t, t, Rg) != ZEN_HAS_INVERSE) {
    goto argh;
  }
  if (ZENPolyDeg(roots, Rg)) {
    ZENEltMultiply(elt, t, elt, Rg);
    ZENPolySetCoeff(roots, 1, elt, Rg);
  }

  elt = ZENPolyGetCoeffPtr(roots, 0, Rg);
  ZENEltMultiply(elt, t, elt, Rg);
  ZENPolySetCoeff(roots, 0, elt, Rg);


  ZENEltFree(t, Rg);
  return(ZEN_HAS_INVERSE);

 argh:
  ZENEltFree(t, Rg);
  return(ZERR);
}

/*
   \TEX{
   \title{Roots of a polynomial of degree 2 in finite fields of
   characteristic 2}
   \input{Two |ZENPoly| |P| and |roots| over a |ZENRing| |Rg|.}
   \output{\begin{itemize}
   \item[|ZENERR|] if an error occured,
   \item[|ZEN_HAS_INVERSE|] if the polynomial has one root.
   \end{itemize}}
   \side{The polynomial |roots| is set to a polynomial of degree 0, 1
   or two, the coefficients of which are the roots of |P|.}
   \note{|P| must be allocated for at least degree 2.We must have
   $|P|\neq |roots|$. Valid only in finite fields.}
   } 
   */
int _ZENPolyRootsDegree2Char2(roots, P, Rg)
     ZENPoly roots, P;
     ZENRing Rg;
{
  int ret;
  ZENElt elt, t1, t2;

  if (ZENEltCreate(t1, Rg)) {
    return(ZERR);
  }
  if (ZENEltCreate(t2, Rg)) {
    ZENEltFree(t1, Rg);
    return(ZERR);
  }

  if (ZENEltIsZero(ZENPolyGetCoeffPtr(P, 1, Rg), Rg)) {
    if ((ret = ZENEltInverse(t2, ZENPolyGetCoeffPtr(P, 2, Rg), Rg)) !=
	ZEN_HAS_INVERSE) {
      goto argh;
    }
    ZENEltMultiply(t1, ZENPolyGetCoeffPtr(P, 0, Rg), t2,
			      Rg);
    elt = ZENPolyGetCoeffPtr(roots, 0, Rg);
    if ((ret = ZENEltSquareRoot(elt, t1, Rg))) {
      goto argh;
    }

    ZENPolyDeg(roots, Rg) = 0;
    ZENPolySetCoeff(roots, 0, elt, Rg);
    ZENEltFree(t1, Rg);
    ZENEltFree(t2, Rg);
    return(ZEN_HAS_INVERSE);
  }

  /* t1 := A*C/B^2; */
  ZENEltSquare(t2, ZENPolyGetCoeffPtr(P, 1, Rg), Rg);
  if ((ret = ZENEltInverse(t1, t2, Rg)) != ZEN_HAS_INVERSE) {
    if (ret == ZEN_NO_INVERSE) {
      ZENPolySetCoeff(roots, 0, t1, Rg);
    }
    goto argh;
  }

  ZENEltMultiply(t1, ZENPolyGetCoeffPtr(P, 2, Rg), t1, Rg);
  ZENEltMultiply(t1, ZENPolyGetCoeffPtr(P, 0, Rg), t1, Rg);
  elt = ZENPolyGetCoeffPtr(roots, 0, Rg);
  if ((ret = ZENPolyRootsCanonical(roots, t1, Rg))) {
    goto argh;
  }

  if (ZENPolyDeg(roots, Rg) < 0) {
    ZENEltFree(t1, Rg);
    ZENEltFree(t2, Rg);
    return(ZEN_HAS_INVERSE);
  }

  /* The roots are B/A*roots[0], B/A*roots[1] */
  if ((ret = ZENEltInverse(t1, ZENPolyGetCoeffPtr(P, 2, Rg), Rg)) !=
      ZEN_HAS_INVERSE) {
    goto argh;
  }

  elt = ZENPolyGetCoeffPtr(roots, 0, Rg);
  ZENEltMultiply(elt, ZENPolyGetCoeffPtr(P, 1, Rg), elt, Rg);
  ZENEltMultiply(elt, t1, elt, Rg);
  ZENPolySetCoeff(roots, 0, elt, Rg);

  elt = ZENPolyGetCoeffPtr(roots, 1, Rg);
  ZENEltMultiply(elt, ZENPolyGetCoeffPtr(P, 1, Rg), elt, Rg);
  ZENEltMultiply(elt, t1, elt, Rg);
  ZENPolySetCoeff(roots, 1, elt, Rg);

  ZENPolyDeg(roots, Rg) = 1;

  ZENEltFree(t1, Rg);
  ZENEltFree(t2, Rg);
  return(ZEN_HAS_INVERSE);

 argh:
  ZENEltFree(t1, Rg);
  ZENEltFree(t2, Rg);
  return(ret);
}

/*
   \TEX{
   \title{Roots of a polynomial of degree 1}
   \input{Two |ZENPoly| |P| and |roots| over a |ZENRing| |R|.}
   \output{\begin{itemize}
   \item[|ZENERR|] if an error occured,
   \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
   \item[|ZEN_HAS_INVERSE|] if the polynomial has one root.
   \end{itemize}}
   \side{The polynomial |roots| is set to a polynomial of degree 0 the
   coefficient of which is the root of |P| if |ZEN_HAS_INVERSE| is
   returned. |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   \note{|P| must be allocated for at least degree 0. We must have
   $|P|\neq |roots|$.}
   } 
   */
int _ZENPolyRootsDegree1(roots, P, R)
     ZENPoly roots, P;
     ZENRing R;
{
  int ret;
  ZENElt elt;

  elt = ZENPolyGetCoeffPtr(roots, 0, R);
  ZENPolyDeg(roots, R) = 0;
  if ((ret = ZENEltInverse(elt, ZENPolyGetCoeffPtr(P, 1, R), R))
      != ZEN_HAS_INVERSE) {
    return(ret);
  }
  ZENEltMultiply(elt, ZENPolyGetCoeffPtr(P, 0, R),
		 elt, R);
  ZENEltNegate(elt, elt, R);
  ZENPolySetCoeff(roots, 0, elt, R);
  return(ZEN_HAS_INVERSE);
}

/*
   \TEX{
   \title{Roots of $X-X^p=\gamma$.}
   \input{A |ZENPoly| |roots| over a |ZENRing| |Rg| and an element |a|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The polynomial |roots| is set to a polynomial of degree 0, 1,
   \ldots, charateristic(|Rg)|, the coefficients of which are the
   roots of |P|.}
   \note{|P| must be allocated for at least the characteristic of the
   field. We must have $|P|\neq |roots|$. Valid only in finite fields.}
   } 
   */
int _ZENPolyRootsCanonical(roots, gamma, Rg)
     ZENPoly roots;
     ZENElt gamma;
     ZENRing Rg;
{
  ZENElt t, u, v, w, elt;
  int i, deg;
  ZENRing rg, last_rg;
  BigNum n;
  BigNumLength nl;

  if (ZENEltCreate(t, Rg)) {
    return(ZERR);
  }
  ZENEltTrace(t, gamma, Rg);
  deg = 1;
  for(last_rg = Rg, rg = ZENRingDef(last_rg);
      rg;
      last_rg = rg, rg = ZENRingDef(last_rg)) {
    deg *= ZENRingDeg(last_rg);
    if (!(n = ZBNC(ZENRingQl(last_rg)))) {
      ZENEltFree(t, last_rg);
      return(ZERR);
    }
    ZENEltToZ(n, &nl, t, last_rg);

    if (ZBNCompareLazy(n, nl, ZENRingQ(rg), ZENRingQl(rg)) != BN_LT ) {
      ZENEltFree(t, last_rg);
      ZBNF(n);
      return(ZERR);
    }
    ZENEltFree(t, last_rg);
    if (ZENEltCreate(u, rg)) {
      ZBNF(n);
      return(ZERR);
    }
    ZENZToElt(u, n, nl, rg);
    ZBNF(n);
    
    ZENEltTrace(u, u, rg);
    t = u;
  }
  if (!ZENEltIsZero(t, last_rg)) {
    ZENEltFree(t, last_rg);
    ZENPolyDeg(roots, Rg) = -1;
    return(0);
  }
  ZENEltFree(t, last_rg);

  ZENPolyDeg(roots, Rg) = ZBNDigit(ZENRingP(Rg), 0)-1;
  if (!ZENPrcIsSet(ZENRingPrcp(Rg), ZENPRC_POLY_ROOTS_CANONICAL)) {
    ZENPrc prc;
  
    ZENPrcSetNone(prc);
    ZENPrcSet(prc, ZENPRC_POLY_ROOTS_CANONICAL);
    ZENPrcSet(prc, ZENPRC_FINITE_FIELD);
    switch (ZENRingAddPrc(Rg, prc)) {
    case ZENERR:
      return(ZERR);
    case ZEN_NO_INVERSE:
      return(ZEN_NO_INVERSE);
    case ZEN_HAS_INVERSE:
      if (!ZENPrcIsSet(ZENRingPrcp(Rg),
		       ZENPRC_POLY_ROOTS_CANONICAL)) {
	return(ZERR);
      }
      break;
    }
  }

  elt = ZENPolyGetCoeffPtr(roots, 0, Rg);
  if (ZENEltCreate(t, Rg)) {
    return(ZERR);
  }  
  if (ZENEltCreate(u, Rg)) {
    ZENEltFree(t, Rg);
    return(ZERR);
  }  
  if (ZENEltCreate(v, Rg)) {
    ZENEltFree(t, Rg);
    ZENEltFree(u, Rg);
    return(ZERR);
  }  
  if (ZENEltCreate(w, Rg)) {
    ZENEltFree(t, Rg);
    ZENEltFree(u, Rg);
    ZENEltFree(v, Rg);
    return(ZERR);
  }  
  ZENEltAssign(t, gamma, Rg);
  ZENEltAssign(u, gamma, Rg);
  ZENEltAssign(elt, gamma, Rg);
  if(ZENEltExp(v, ZENRingP(Rg), ZENRingPl(Rg), ZENRingTheta(Rg), Rg)) {
    ZENEltFree(t, Rg);
    ZENEltFree(u, Rg);
    ZENEltFree(v, Rg);
    ZENEltFree(w, Rg);
    return(ZERR);
  }
  ZENEltMultiply(elt, v, t, Rg);

  for(i = 1; i < deg-1; i++) {
    if(ZENEltExp(v, ZENRingP(Rg), ZENRingPl(Rg), v, Rg)) {
      ZENEltFree(t, Rg);
      ZENEltFree(u, Rg);
      ZENEltFree(v, Rg);
      ZENEltFree(w, Rg);
      return(ZERR);
    }
    if(ZENEltExp(u, ZENRingP(Rg), ZENRingPl(Rg), u, Rg)) {
      ZENEltFree(t, Rg);
      ZENEltFree(u, Rg);
      ZENEltFree(v, Rg);
      ZENEltFree(w, Rg);
      return(ZERR);
    }
    ZENEltAdd(t, u, Rg);
    ZENEltMultiply(w, v, t, Rg);
    ZENEltAdd(elt, w, Rg);
  }
  ZENEltFree(v, Rg);
  ZENEltFree(u, Rg);
  ZENEltFree(w, Rg);

  ZENPolySetCoeff(roots, 0, elt, Rg);  
  ZENEltSetToOne(t, Rg);
  for(i = 1; i <= ZENPolyDeg(roots, Rg); i++) {
    elt = ZENPolyGetCoeffPtr(roots, i, Rg);
    ZENEltAssign(elt, ZENPolyGetCoeffPtr(roots, i-1, Rg), Rg);
    ZENEltAdd(elt, t, Rg);
    ZENPolySetCoeff(roots, i, elt, Rg);
  }
  ZENEltFree(t, Rg);
  return(0);
}


