/*
 * zen/zeg/c/sr_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Series assignment}
   
   One can set a |ZENSr| to zero with |_ZENSrSetToZero|. Then we can
   assign a |ZENSr| with an other with |_ZENSrAssign|. Furthermore, we
   can set randomly a |ZENSr| with |_ZENSrSetRandom|. 
   
   The function |_ZENSrGetCoeff|
   enables us to get or set the coefficient of $X^i$,
   |_ZENSrUpdateValuation| computes the real valuation of a |ZENSr| if it
   decreased.
   }
   */
#include <stdio.h>
#include "zen.h"


/* \TEX{
   \title{Assigning.}
   \input{Two allocated series |PX| and |RX| of a finite ring |Rg|.}
   \side{|RX| is filled with |PX|.}
   }
   */
void _ZENSrAssign(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;
{
  int i;

  ZENSrDeg(RX, Rg) =  ZENSrDeg(PX, Rg);
  ZENSrVal(RX, Rg) =  ZENSrVal(PX, Rg);
  for(i = ZENSrVal(PX, Rg); i <= ZENSrDeg(PX, Rg); i++) {
    ZENSrSetCoeff(RX, i, ZENSrGetCoeffPtr(PX, i, Rg), Rg);
  }
}


/* \TEX{
   \title{Copying an allocated series.}
   \input{An allocated series |PX| of a finite ring |Rg|.}
   \output{|NULL| if an error occurred, a copy of |PX| otherwise.}   
   }
   */
ZENSr _ZENSrCopy (p, R)
     ZENSr p;
     ZENRing R;
{
  ZENSr ret;
  
  if (!(ret = ZENSrCreate(ZENSrLgt(p, R), R))) {
    return(NULL);
  }
  ZENSrAssign(ret, p, R);
  return(ret);
}

/* \TEX{
   \title{Setting to zero.}
   \input{An allocated series |RX| of a finite ring |Rg|.}
   \side{|RX| is set to zero.}
   }
   */
void _ZENSrSetToZero(RX, Rg)
     ZENSr RX;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  for(i= ZENSrVal(RX, Rg); i <= ZENSrDeg(RX, Rg); i++) {
    elt = ZENSrGetCoeffPtr(RX, i, Rg);
    ZENEltSetToZero(elt, Rg);
    ZENSrSetCoeff(RX, i, elt, Rg);
  }
}

/* \TEX{
   \title{Setting to random.}
   \input{An allocated series |RX| with valuation |val| and degree|deg| of an
   extension |Rg|.}
   \side{|RX| is set randomly to a series of valuation |val| and degree|deg|.}
   }
   */
void _ZENSrSetRandom(RX, val, deg, Rg)
     ZENSr RX;
     int val, deg;
     ZENRing Rg;
{
  ZENElt elt;
  int i;

  ZENSrDeg(RX, Rg) =  deg;
  ZENSrVal(RX, Rg) =  val;

  elt = ZENSrGetCoeffPtr(RX, val, Rg);
  ZENEltSetRandom(elt, Rg);
  while(ZENEltIsZero(elt, Rg)) {
    ZENEltSetRandom(elt, Rg);
  }
  ZENSrSetCoeff(RX, val, elt, Rg);
  for(i = val+1; i <= deg; i++) {
    elt = ZENSrGetCoeffPtr(RX, i, Rg);
    ZENEltSetRandom(elt, Rg);
    ZENSrSetCoeff(RX, i, elt, Rg);
  }
}


/* \TEX{
   \title{Getting a coeff.}
   \input{An allocated series |RX| of a finite ring |Rg|, the
   degree |d| of the coeff to get and an element |b| to assign.} 
   \side{|b| is filled with the coefficient of $X^d$ in |RX|.}
   }
   */
void _ZENSrGetCoeff(p_c, RX, d, Rg)
     ZENElt *p_c;
     ZENSr RX;
     int d;
     ZENRing Rg;
{ 
  ZENElt elt;

  elt = ZENSrGetCoeffPtr(RX, d, Rg);
  ZENEltAssign(*p_c, elt, Rg);
}

/* \TEX{
   \title{Update the degree of a Snomial.}
   \input{A |ZENSr| |RX| of a finite ring |Rg|.}
   \side{Resetting the degree of |P|, assuming it has decreased.}
   }
   */
void _ZENSrUpdateValuation(RX, Rg)
     ZENSr RX;
     ZENRing Rg;
{
  int i, j;
  
  for(i = ZENSrVal(RX, Rg); i <= ZENSrDeg(RX, Rg); i++) {
    if(!ZENEltIsZero(ZENSrGetCoeffPtr(RX, i, Rg), Rg))
      break;
  }
  if (i > ZENSrDeg(RX, Rg)) {
    ZENSrVal(RX, Rg) = ZENSrDeg(RX, Rg);
    return;
  }

  if (i != ZENSrVal(RX, Rg)) {
    for(j = i; j <= ZENSrDeg(RX, Rg); j++) {
      ZENSrSetCoeff(RX,
		    j+ZENSrVal(RX, Rg)-i,
		    ZENSrGetCoeffPtr(RX, j, Rg),
		    Rg);
    }
    ZENSrVal(RX, Rg) = i;
  }
}

