/*
 * zen/zeg/c/sr_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   \Subsection{input/output}
   
   Conversions between the internal representation and the ASCII
   representation of a |ZENSr| are done by
   |_ZENSrPrintToString|. |_ZENSrPrintToFile| allows to write a
   |ZENSr| in a file. One can moreover write to and then read from a
   file a |_ZENSr| with |_ZENSrPutToFile| and |_ZENSrReadFromFile|.
   }
   */
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "zen.h"

/* \TEX{
   \title{Converting from string.}
   \input{A finite field |R|, a pointer on |ZENSr| |PX| and a string
   |s| representing a |ZENSr| in base |base|$\in \{2,\ldots,
   16\}$.} 
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|PX| is allocated and filled with |s| if the output is not |ZENERR|.}
   }
*/
int _ZENSrReadFromString(PX, s, base, Rg)
     ZENSr *PX;
     char *s;
     int base;
     ZENRing Rg;
{
  ZENElt elt;
  char *ptr;
  int i, nbc;


  /* Alocating a buffer */
  if (ZENEltCreate(elt, Rg)) {
    return(ZERR);
  }
  *PX = NULL;
  ptr = s;
  
  /* Parsing */
  for(;;) {
    for(; isspace(*ptr); ptr++) ;
    if (*ptr != '(') {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRREADFROMSTRING, ZEN_SYNTAX);
      ZENEltFree(elt, Rg);
      if (*PX) {
	ZENSrFree(*PX, Rg);
      }
      return(ZERR);
    }
    
    ptr++;
    if ((nbc = ZENEltReadFromString(elt, ptr, base, Rg)) < 0) {
      ZENEltFree(elt, Rg);
      if (*PX) {
	ZENSrFree(*PX, Rg);
      }
      return(ZERR);
    }

    for(ptr += nbc; isspace(*ptr); ptr++) ;
    if (*ptr != ')') {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRREADFROMSTRING, ZEN_SYNTAX);
      ZENEltFree(elt, Rg);
      if (*PX) {
	ZENSrFree(*PX, Rg);
      }
      return(ZERR);
    }
    
    for(ptr++; isspace(*ptr); ptr++) ;
    if (*ptr == '*') {
      for(ptr++; isspace(*ptr); ptr++) ;
      if (*ptr == 'X') {
	for(ptr++; isspace(*ptr); ptr++) ;
	switch (*ptr) {
	case '^':
	  for(ptr++; isspace(*ptr); ptr++) ;
	  if (sscanf(ptr, "%d", &i) != 1) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_SRREADFROMSTRING, ZEN_SYNTAX);
	    ZENEltFree(elt, Rg);
	    if (*PX) {
	      ZENSrFree(*PX, Rg);
	    }
	    return(ZERR);
	  }
	  if (!*PX) {
	    if (!(*PX = ZENSrCreate(i+1, Rg))) {
	      ZENEltFree(elt, Rg);
	      return(ZERR);
	    }
	    ZENSrVal(*PX, Rg) = 0;
	    ZENSrDeg(*PX, Rg) = i;
	    ZENSrSetToZero(*PX, Rg);
	  }
	  ZENSrSetCoeff(*PX, i, elt, Rg);
	  for(ptr++; isdigit(*ptr); ptr++) ;
	  for(; isspace(*ptr); ptr++) ;
	  break;
	case '+':
	  if (!*PX) {
	    if (!(*PX = ZENSrCreate(2, Rg))) {
	      ZENEltFree(elt, Rg);
	      return(ZERR);
	    }
	    ZENSrVal(*PX, Rg) = 0;
	    ZENSrDeg(*PX, Rg) = 1;
	    ZENSrSetToZero(*PX, Rg);
	  }
	  ZENSrSetCoeff(*PX, 1, elt, Rg);
	  break;
	default:
	  if (!*PX) {
	    if (!(*PX = ZENSrCreate(2, Rg))) {
	      ZENEltFree(elt, Rg);
	      return(ZERR);
	    }
	    ZENSrVal(*PX, Rg) = 0;
	    ZENSrDeg(*PX, Rg) = 1;
	    ZENSrSetToZero(*PX, Rg);
	  }
	  ZENSrSetCoeff(*PX, 1, elt, Rg);
	  ZENEltFree(elt, Rg);
	  return(ptr-s);
	}
      }
      else {
	ZENSetError(ZEN_LIB_ZEG, ZEN_SRREADFROMSTRING, ZEN_SYNTAX);
	ZENEltFree(elt, Rg);
	if (*PX) {
	  ZENSrFree(*PX, Rg);
	}
	return(ZERR);
      }
    }
    else {
      if (!*PX) {
	if (!(*PX = ZENSrCreate(1, Rg))) {
	  ZENEltFree(elt, Rg);
	  return(ZERR);
	}
	ZENSrVal(*PX, Rg) = ZENSrDeg(*PX, Rg) =  0;
      }
      ZENSrSetCoeff(*PX, 0, elt, Rg);
      ZENEltFree(elt, Rg);
      return(ptr-s);
    }
    
    if (*ptr != '+') {
      break;
    }
    ptr++;
  }
  
  ZENEltFree(elt, Rg);
  return(ptr-s);
}

/* \TEX{
   \title{Converting to string.}
   \input{An allocated |ZENSr| |G| of a finite field |R| and a base |base|$\in \{2,\ldots, 16\}$.}
   \output{An allocated string representing |G| in base |base| or NULL if an error occured.}
   }
   */
char *_ZENSrPrintToString(PX, base, Rg)
     ZENSr PX;
     int base;
     ZENRing Rg;  
{  
  int i, dgt, nb_dgt;
  char *str, *s1, *s2;
  ZENElt elt;


  str = NULL;
  for(nb_dgt = 1, dgt = 1;
      10*dgt <= ZENSrDeg(PX, Rg);
      nb_dgt++, dgt *= 10) {
    ;
  }

  switch(ZENSrDeg(PX, Rg)) {  
  case 0:
    elt = ZENSrGetCoeffPtr(PX, ZENSrDeg(PX, Rg), Rg);
    
    if (!(str = ZENEltPrintToString(elt, base, Rg))) {
      return(NULL);
    }
    s2 = str;
    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+3)))) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOSTRING, ZEN_ALLOC);
      ZENFree(s2);
      return(NULL);
    }
    if (sprintf(str, "(%s)", s2) < 0) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOSTRING, ZEN_SPRINTF);
      ZENFree(str); ZENFree(s2);
      return(NULL);
    }
    ZENFree(s2);
    break;
    
  case 1:
    elt = ZENSrGetCoeffPtr(PX, ZENSrDeg(PX, Rg), Rg);
    
    if (!(str = ZENEltPrintToString(elt, base, Rg))) {
      return(NULL);
    }
    s2 = str;
    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+5)))) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOSTRING, ZEN_ALLOC);
      ZENFree(s2);
      return(NULL);
    }
    if (sprintf(str, "(%s)*X", s2) < 0) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOSTRING, ZEN_SPRINTF);
      ZENFree(str); ZENFree(s2);
      return(NULL);
    }
    ZENFree(s2);
    break;
    
  default:
    elt = ZENSrGetCoeffPtr(PX, ZENSrDeg(PX, Rg), Rg);
    if (!(str = ZENEltPrintToString(elt, base, Rg))) {
      return(NULL);
    }
    s2 = str;
    if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+6+nb_dgt)))) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOSTRING, ZEN_ALLOC);
      ZENFree(s2);
      return(NULL);
    }
    if (sprintf(str, "(%s)*X^%d", s2, ZENSrDeg(PX, Rg)) < 0) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOSTRING, ZEN_SPRINTF);
      ZENFree(str); ZENFree(s2);
      return(NULL);
    }
    ZENFree(s2);
    break;
  }
  
  for(nb_dgt = 1, dgt = 1;
      10*dgt <= ZENSrDeg(PX, Rg)-1;
      nb_dgt++, dgt *= 10) {
    ;
  }
  
  for(i = ZENSrDeg(PX, Rg)-1; i >= ZENSrVal(PX, Rg); i--) {
    
    if (i < dgt) {
      nb_dgt--;
      dgt /= 10;
    }
    
    elt = ZENSrGetCoeffPtr(PX, i, Rg);

    switch(i) {
    case 0:
      if (!ZENEltIsZero(elt, Rg)) {
	if (!(s1 = ZENEltPrintToString(elt, base, Rg))) {
	  ZENFree(str);
	  return(NULL);
	}
	s2 = str;
	if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+strlen(s1)+4)))) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOSTRING, ZEN_ALLOC);
	  ZENFree(s1); ZENFree(s2);
	  return(NULL);
	}
	if (sprintf(str, "%s+(%s)", s2, s1) < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOSTRING, ZEN_SPRINTF);
	  ZENFree(s1); ZENFree(s2);
	  return(NULL);
	}
	ZENFree(s1); ZENFree(s2);
      }
      break;
      
    case 1:
      if (!ZENEltIsZero(elt, Rg)) {
	if (!(s1 = ZENEltPrintToString(elt, base, Rg))) {
	  ZENFree(str);
	  return(NULL);
	}
	s2 = str;
	if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+strlen(s1)+6)))) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOSTRING, ZEN_ALLOC);
	  ZENFree(s1); ZENFree(s2);
	  return(NULL);
	}
	if (sprintf(str, "%s+(%s)*X", s2, s1) < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOSTRING, ZEN_SPRINTF);
	  ZENFree(s1); ZENFree(s2);
	  return(NULL);
	}
	ZENFree(s1); ZENFree(s2);
      }
      break;
      
    default:
      if (!ZENEltIsZero(elt, Rg)) {
	if (!(s1 = ZENEltPrintToString(elt, base, Rg))) {
	  ZENFree(str);
	  return(NULL);
	}
	s2 = str;
	if (!(str = (char *)ZENMalloc(sizeof(char)*(strlen(str)+strlen(s1)+7+nb_dgt)))) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOSTRING, ZEN_ALLOC);
	  ZENFree(s1); ZENFree(s2);
	  return(NULL);
	}
	if (sprintf(str, "%s+(%s)*X^%d", s2, s1, i) < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOSTRING, ZEN_SPRINTF);
	  ZENFree(s1); ZENFree(s2);
	  return(NULL);
	}
	ZENFree(s1); ZENFree(s2);
      }
      break;
    }
  }

  return(str);
}


/* \TEX{
   \title{Converting from a string in an internal representation.}
   \input{A finite field |Rg| and a string |s| representing a |ZENSr|
   in an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|PX| is allocated and filled with s if the output is different
   from |ZENERR|.}	    
   }
   */
int _ZENSrGetFromString(PX, s, Rg)
     char *s;
     ZENSr *PX;
     ZENRing Rg;
{
  ZENElt elt;
  int deg, length, i, nbc, val;
  char *ptr;
  

  if (sscanf(s, "%8x%8x%8x", &length, &val, &deg) != 3) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_SRGETFROMSTRING, ZEN_SSCANF);
    return(ZERR);
  }
  if (ZENEltCreate(elt, Rg)) {
    return(ZERR);
  }
  if (!(*PX = ZENSrCreate(length, Rg))) {
    ZENEltFree(elt, Rg);
    return(ZERR);
  }
  ZENSrDeg(*PX, Rg) =  deg;
  ZENSrVal(*PX, Rg) =  val;

  for(ptr = s; isspace(*ptr); ptr++) ;
  ptr += 24;
  for(i = val; i <= deg; i++) {
    if ((nbc = ZENEltGetFromString(elt, ptr, Rg)) < 0) {
      ZENEltFree(elt, Rg);
      ZENSrFree(*PX, Rg);
      return(ZERR);
    }
    ZENSrSetCoeff(*PX, i, elt, Rg);
    ptr += nbc;
  }
  ZENEltFree(elt, Rg);
  return(ptr-s);
}

/* \TEX{
   \title{Converting to string in an internal representation.}
   \input{An allocated |ZENSr| |PX| of a finite field |Rg|.}
   \output{An allocated string representing |G| in an internal representation or NULL if an error occured.}   
   }
   */
char *_ZENSrPutToString(PX, Rg)
     ZENSr PX;
     ZENRing Rg;
{
  ZENElt elt;
  char *str, *s1;
  int i;

  if (!(str = (char *)ZENMalloc((1+6*sizeof(int))*sizeof(char)))) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_SRPUTTOSTRING, ZEN_ALLOC);
    return(NULL);
  }
  if (sprintf(str, "%.8x%.8x%.8x",
	      ZENSrLgt(PX, Rg),
	      ZENSrVal(PX, Rg),
	      ZENSrDeg(PX, Rg)) < 0) {
    ZENFree(str);
    ZENSetError(ZEN_LIB_ZEG, ZEN_SRPUTTOSTRING, ZEN_SPRINTF);
    return(NULL);
  }
  for(i = ZENSrVal(PX, Rg); i <= ZENSrDeg(PX, Rg); i++) {
    elt = ZENSrGetCoeffPtr(PX, i, Rg);

    if (!(s1 = ZENEltPutToString(elt, Rg))) {
      ZENFree(str);
      return(NULL);
    }
    
    if (!(str = (char *)ZENRealloc(str, sizeof(char)*(strlen(str)+strlen(s1)+1)))) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPUTTOSTRING, ZEN_ALLOC);
      ZENFree(s1);
      return(NULL);
    }
    
    if (!strcat(str, s1)) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPUTTOSTRING, ZEN_STRCAT);
      ZENFree(str); ZENFree(s1);
      return(NULL);
    }
    
    ZENFree(s1);
  }
  return(str);
}


/* \TEX{
   \title{Printing to file.}
   \input{A stream |file|, an allocated |ZENSr| |PX| of a finite field
   |Rg| and a base |base|$\in \{2,\ldots, 16\}$.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |PX| in base |base| to |file|.}    
   }
   */
int _ZENSrPrintToFile(file, PX, base, Rg)
     FILE *file;
     ZENSr PX;
     int base;
     ZENRing Rg;  
{ 
  int i;
  ZENElt elt;


  switch(ZENSrDeg(PX, Rg)) {
  case 0:
    if (fprintf(file, "(") < 0) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
    
    elt = ZENSrGetCoeffPtr(PX, ZENSrDeg(PX, Rg), Rg);
    if (ZENEltPrintToFile(file, elt, base, Rg)) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
    
    if (fprintf(file, ")") < 0) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
    break;
    
  case 1:
    if (fprintf(file, "(") < 0) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
    
    elt = ZENSrGetCoeffPtr(PX, ZENSrDeg(PX, Rg), Rg);
    if (ZENEltPrintToFile(file, elt, base, Rg)) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
    
    if (fprintf(file, ")*X") < 0) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
    break;
    
  default:
    if (fprintf(file, "(") < 0) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
    
    elt = ZENSrGetCoeffPtr(PX, ZENSrDeg(PX, Rg), Rg);
    if (ZENEltPrintToFile(file, elt, base, Rg)) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
    
    if (fprintf(file, ")*X^%d", ZENSrDeg(PX, Rg)) < 0) {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
  }

  for(i = ZENSrDeg(PX, Rg)-1; i >= ZENSrVal(PX, Rg); i--) {
    
    elt = ZENSrGetCoeffPtr(PX, i, Rg);

    switch(i) {
    case 0:
      if (!ZENEltIsZero(elt, Rg)) {
	if (fprintf(file, "+(") < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
	if (ZENEltPrintToFile(file, elt, base, Rg)) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
	if (fprintf(file, ")") < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
      }
      break;
      
    case 1:
      if (!ZENEltIsZero(elt, Rg)) {
	if (fprintf(file, "+(") < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
	if (ZENEltPrintToFile(file, elt, base, Rg)) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
	if (fprintf(file, ")*X") < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
      }
      break;
      
    default:
      if (!ZENEltIsZero(elt, Rg)) {
	if (fprintf(file, "+(") < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
	if (ZENEltPrintToFile(file, elt, base, Rg)) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
	if (fprintf(file, ")*X^%d", i) < 0) {
	  ZENSetError(ZEN_LIB_ZEG, ZEN_SRPRINTTOFILE, ZEN_FPRINTF);
	  return(ZERR);
	}
      }
    }
  }
  return(0);
}

/* \TEX{
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |ZENSr| |PX| of a finite field
   |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |PX| to |file|.}    
   }
   */
int _ZENSrPutToFile(file, PX, Rg)
     FILE *file;
     ZENSr PX;
     ZENRing Rg;  
{
  ZENElt elt;
   int i;

 
  if (fprintf(file, "%.8x%.8x%.8x", ZENSrLgt(PX, Rg), ZENSrVal(PX, Rg), ZENSrDeg(PX, Rg)) < 0) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_SRPUTTOFILE, ZEN_FPRINTF);
    return(ZERR);
  }
  for(i = ZENSrVal(PX, Rg); i <= ZENSrDeg(PX, Rg); i++) {
    elt = ZENSrGetCoeffPtr(PX, i, Rg);

    if (ZENEltPutToFile(file, elt, Rg)) {
      return(ZERR);
    }
  }
  return(0);
}

/* \TEX{
   \title{Reading from file.}
   \input{A stream |file| and a pointer on a series |PX| of a finite field |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|PX| is allocated and filled with the |ZENSr| read in |file| if
   no error occurred.}
   }
   */
int _ZENSrReadFromFile(PX, file, base, Rg)
     ZENSr *PX;
     FILE *file;
     int base;
     ZENRing Rg;
{
  ZENElt elt;
  int c;
  int i;


  /* Alocating a buffer */
  if (ZENEltCreate(elt, Rg)) {
    return(ZERR);
  }
  *PX = NULL;
  c = fgetc(file);
  
  /* Parsing */
  for(;;) {
    for(; isspace(c); c = fgetc(file)) ;
    if (c != '(') {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRREADFROMFILE, ZEN_SYNTAX);
      ZENEltFree(elt, Rg);
      if (*PX) {
	ZENSrFree(*PX, Rg);
      }
      return(ZERR);
    }
    
    if (ZENEltReadFromFile(elt, file, base, Rg)) {
      ZENEltFree(elt, Rg);
      if (*PX) {
	ZENSrFree(*PX, Rg);
      }
      return(ZERR);
    }

    for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
    if (c != ')') {
      ZENSetError(ZEN_LIB_ZEG, ZEN_SRREADFROMFILE, ZEN_SYNTAX);
      ZENEltFree(elt, Rg);
      if (*PX) {
	ZENSrFree(*PX, Rg);
      }
      return(ZERR);
    }
    
    for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
    if (c == '*') {
      for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
      if (c == 'X') {
	for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
	switch (c) {
	case '^':
	  if (fscanf(file, "%d", &i) != 1) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_SRREADFROMFILE, ZEN_SYNTAX);
	    ZENEltFree(elt, Rg);
	    if (*PX) {
	      ZENSrFree(*PX, Rg);
	    }
	    return(ZERR);
	  }
	  if (!*PX) {
	    if (!(*PX = ZENSrCreate(i+1, Rg))) {
	      ZENEltFree(elt, Rg);
	      return(ZERR);
	    }
	    ZENSrVal(*PX, Rg) = 0;
	    ZENSrDeg(*PX, Rg) = i;
	    ZENSrSetToZero(*PX, Rg);
	  }	  
	  ZENSrSetCoeff(*PX, i, elt, Rg);
	  for(c = fgetc(file); isspace(c); c = fgetc(file)) ;
	  break;
	case '+':
	  if (!*PX) {
	    if (!(*PX = ZENSrCreate(2, Rg))) {
	      ZENEltFree(elt, Rg);
	      return(ZERR);
	    }
	    ZENSrVal(*PX, Rg) = 0;
	    ZENSrDeg(*PX, Rg) = 1;
	    ZENSrSetToZero(*PX, Rg);
	  }
	  ZENSrSetCoeff(*PX, 1, elt, Rg);
	  break;
	default:
	  if (!*PX) {
	    if (!(*PX = ZENSrCreate(2, Rg))) {
	      ZENEltFree(elt, Rg);
	      return(ZERR);
	    }
	    ZENSrVal(*PX, Rg) = 0;
	    ZENSrDeg(*PX, Rg) = 1;
	    ZENSrSetToZero(*PX, Rg);
	  }
	  ZENSrSetCoeff(*PX, 1, elt, Rg);
	  ZENEltFree(elt, Rg);
	  if ((c != EOF) && (ungetc(c, file) == EOF)) {
	    ZENSetError(ZEN_LIB_ZEG, ZEN_SRREADFROMFILE, ZEN_SYNTAX);
	    if (*PX) {
	      ZENSrFree(*PX, Rg);
	    }
	    return(ZERR);
	  }
	  return(0);
	}
      }
      else {
	ZENSetError(ZEN_LIB_ZEG, ZEN_SRREADFROMFILE, ZEN_SYNTAX);
	ZENEltFree(elt, Rg);
	if (*PX) {
	  ZENSrFree(*PX, Rg);
	}
	return(ZERR);
      }
    }
    else {
      if (!*PX) {
	if (!(*PX = ZENSrCreate(1, Rg))) {
	  ZENEltFree(elt, Rg);
	  return(ZERR);
	}
	ZENSrVal(*PX, Rg) = 0;
	ZENSrDeg(*PX, Rg) =  0;
      }
      ZENSrSetCoeff(*PX, 0, elt, Rg);
      ZENEltFree(elt, Rg);
      if ((c != EOF) && (ungetc(c, file) == EOF)) {
	ZENSetError(ZEN_LIB_ZEG, ZEN_SRREADFROMFILE, ZEN_SYNTAX);
	if (*PX) {
	  ZENSrFree(*PX, Rg);
	}
	return(ZERR);
      }
      return(0);
    }
    
    if (c != '+') {
      break;
    }
    c = fgetc(file);
  }

  ZENEltFree(elt, Rg);
  if ((c != EOF) && (ungetc(c, file) == EOF)) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_SRREADFROMFILE, ZEN_SYNTAX);
    if (*PX) {
      ZENSrFree(*PX, Rg);
    }
    return(ZERR);
  }
  return(0);
}

/* \TEX{
   \title{Getting from file in an internal representation.}
   \input{A stream |file|, and a pointer on series |PX| of a finite
   field |ZENSr|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|*PX| is filled with the series read in |file|}    
   }
   */
int _ZENSrGetFromFile(PX, file, Rg)
     ZENSr *PX;
     FILE *file;
     ZENRing Rg;
{
  ZENElt elt;
  int deg, length, i, val;


  if (fscanf(file, "%8x%8x%8x", &length, &val, &deg) != 3) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_SRGETFROMFILE, ZEN_SSCANF);
    return(ZERR);
  }

  if (ZENEltCreate(elt, Rg)) {
    return(ZERR);
  }
  if (!(*PX = ZENSrCreate(length, Rg))) {
    ZENEltFree(elt, Rg);
    return(ZERR);
  }

  ZENSrDeg(*PX, Rg) =  deg;
  ZENSrVal(*PX, Rg) =  val;
  ZENSrLgt(*PX, Rg) =  length;
  for(i = val; i <= deg; i++) {
    if (ZENEltGetFromFile(elt, file, Rg)) {
      ZENEltFree(elt, Rg);
      ZENSrFree(*PX, Rg);
      return(ZERR);
    }
    ZENSrSetCoeff(*PX, i, elt, Rg);
  }
  ZENEltFree(elt, Rg);
  return(0);
}

