/*
 * zen/zeg/c/sr_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Series multiplication}
   
   Two functions are provided to do multiplications: |_ZENSrSquare| to square a
   |ZENSr| and |_ZENSrMultily| to compute the multiplication of two |ZENSr|.
   }
   */
#include <stdio.h>
#include "zen.h"

/* \TEX{
   \title{Squaring .}
   \input{Two allocated series |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   \note{|RX| has got twice more terms than |PX|.}
   }
   */
int _ZENSrSquareChar2(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;
{
  ZENElt elt;
  int i, vp, vr;

  /* valid terms are 2*vs...2*ds */
  vp = ZENSrVal(PX, Rg);
  ZENSrVal(RX, Rg) = ZENSrVal(PX, Rg)<<1;
  ZENSrDeg(RX, Rg) = ZENSrDeg(PX, Rg)<<1;
  vr = ZENSrVal(PX, Rg);
  for(i = ZENSrDeg(RX, Rg); i >= ZENSrVal(RX, Rg); i--){
    if (i & 1) {
      elt = ZENSrGetCoeffPtr(RX, i, Rg);
      ZENEltSetToZero(elt, Rg);
      ZENSrSetCoeff(RX, i, elt, Rg);
    }
    else {
      elt = ZENSrGetCoeffPtr(RX, i, Rg);
      ZENSrVal(PX, Rg) = vp; 
      ZENEltSquare(elt,
		   ZENSrGetCoeffPtr(PX, i>>1, Rg),
		   Rg);
      ZENSrVal(PX, Rg) = vr;
      ZENSrSetCoeff(RX, i, elt, Rg);
    }
  }
  return(0);
}


/* \TEX{
   \title{Squaring in characteristic greater than 2.}
   \input{Two |ZENSr| |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZENSrSquareCharp(RX, PX, Rg)
     ZENSr RX, PX;
     ZENRing Rg;
{
  int i, j, vp, dp;
  ZENElt e;

  /* valid terms are 2*vs... vs+ds */
  vp = ZENSrVal(PX, Rg); dp = ZENSrDeg(PX, Rg);
  ZENSrVal(PX, Rg) = 0; ZENSrDeg(PX, Rg) = dp-vp;

  ZENSrVal(RX, Rg) = 0;
  ZENSrDeg(RX, Rg) = ZENSrDeg(PX, Rg); 

  for(i = ZENSrDeg(PX, Rg); i > 0; i--) {
    
    if (2*i <= ZENSrDeg(RX, Rg)) {
      ZENEltSquare(ZENRingHbuf(Rg), ZENSrGetCoeffPtr(PX, i, Rg), Rg);
      e = ZENSrGetCoeffPtr(RX, 2*i, Rg);
      ZENEltAdd(e, ZENRingHbuf(Rg), Rg);
      ZENSrSetCoeff(RX, 2*i, e, Rg);
    }
    for(j = ZENSrDeg(PX, Rg)-i < i-1 ?
	ZENSrDeg(PX, Rg)-i : i-1;
	j > 0;
	j--) {
      ZENEltMultiply(ZENRingHbuf(Rg),
		     ZENSrGetCoeffPtr(PX, j, Rg),
		     ZENSrGetCoeffPtr(PX, i, Rg),
		     Rg);
      ZENEltAdd(ZENRingHbuf(Rg), ZENRingHbuf(Rg), Rg);
      e = ZENSrGetCoeffPtr(RX, i+j, Rg);
      ZENEltAdd(e, ZENRingHbuf(Rg), Rg);
      ZENSrSetCoeff(RX, i+j, e, Rg);
    }
    ZENEltMultiply(ZENRingHbuf(Rg),
		   ZENSrGetCoeffPtr(PX, j, Rg),
		   ZENSrGetCoeffPtr(PX, i, Rg),
		   Rg);
    ZENEltAdd(ZENRingHbuf(Rg), ZENRingHbuf(Rg), Rg);
    ZENSrSetCoeff(RX, i+j, ZENRingHbuf(Rg), Rg);
  }
  e = ZENSrGetCoeffPtr(RX, 0, Rg);
  ZENEltSquare(e,
	       ZENSrGetCoeffPtr(PX, 0, Rg),
	       Rg);
  ZENSrSetCoeff(RX, 0, e, Rg);
  
  ZENSrVal(PX, Rg) = vp;
  ZENSrDeg(PX, Rg) = dp;

  ZENSrVal(RX, Rg) = 2*vp;
  ZENSrDeg(RX, Rg) =  dp+vp;

  return(0);
}

/* \TEX{
   \title{Multiplication.}
   \input{Three allocated series |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|QX|$.}
   }
   */
int _ZENSrMultiply(RX, QX, PX, Rg)
     ZENSr RX, QX, PX;
     ZENRing Rg;
{
  ZENElt e;
  int i, j, vp, dp, vq, dq;


  vp = ZENSrVal(PX, Rg); dp = ZENSrDeg(PX, Rg);
  vq = ZENSrVal(QX, Rg); dq = ZENSrDeg(QX, Rg);

  ZENSrVal(PX, Rg) = 0; ZENSrDeg(PX, Rg) = dp-vp;
  ZENSrVal(QX, Rg) = 0; ZENSrDeg(QX, Rg) = dq-vq;

  ZENSrVal(RX, Rg) = 0;
  if (dp-vp <  dq-vq) {
    ZENSrDeg(RX, Rg) = ZENSrDeg(QX, Rg) = ZENSrDeg(PX, Rg) = dp-vp; 
  }
  else {
    ZENSrDeg(RX, Rg) = ZENSrDeg(PX, Rg) = ZENSrDeg(QX, Rg) = dq-vq; 
  }

  for(i = ZENSrDeg(PX, Rg); i >= 0; i--) {
    for(j = ZENSrDeg(QX, Rg)-i; j; j--) {
      ZENEltMultiply(ZENRingHbuf(Rg),
		     ZENSrGetCoeffPtr(QX, j, Rg),
		     ZENSrGetCoeffPtr(PX, i, Rg),
		     Rg);
      e = ZENSrGetCoeffPtr(RX, i+j, Rg);
      ZENEltAdd(e, ZENRingHbuf(Rg), Rg);
      ZENSrSetCoeff(RX, i+j, e, Rg);
    }
    ZENEltMultiply(ZENRingHbuf(Rg),
		   ZENSrGetCoeffPtr(QX, j, Rg),
		   ZENSrGetCoeffPtr(PX, i, Rg),
		   Rg);
    ZENSrSetCoeff(RX, i+j, ZENRingHbuf(Rg), Rg);
  }
  
  ZENSrVal(PX, Rg) = vp;
  ZENSrDeg(PX, Rg) = dp;

  ZENSrVal(QX, Rg) = vq;
  ZENSrDeg(QX, Rg) = dq;
  
  ZENSrVal(RX, Rg) = vp+vq;
  ZENSrDeg(RX, Rg) =  dp+vq < dq+vp ? dp+vq : dq+vp;

  return(0);
}

/* \TEX{
   \title{Multiplication by a scalar.}
   \input{Two series |PX|, |RX| and an element |e| of a finite ring |Rg|.}
   }
	 */
void _ZENSrMultiplyScalar(RX, PX, e, Rg)
     ZENSr RX, PX;
     ZENElt e;
     ZENRing Rg;
{
  ZENElt elt;
  int i;

  ZENSrDeg(RX, Rg) = ZENSrDeg(PX, Rg);
  ZENSrVal(RX, Rg) = ZENSrVal(PX, Rg);
  if (ZENEltIsZero(e, Rg)) {
    ZENSrSetToZero(RX, Rg);
    return;
  }
  if (ZENEltIsOne(e, Rg)) {
    ZENSrAssign(RX, PX, Rg);
    return;
  }

  for(i = ZENSrVal(PX, Rg); i <= ZENSrDeg(PX, Rg); i++) {
    elt = ZENSrGetCoeffPtr(RX, i, Rg);
    ZENEltMultiply(elt, e,
		   ZENSrGetCoeffPtr(PX, i, Rg),
		   Rg);
    ZENSrSetCoeff(RX, i, elt, Rg);
  }
}

/* \TEX{
   \title{Divide.}
   \input{Two series |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient.}
   \output{|ZEN_HAS_INVERSE| if no error occurred, |ZEN_NO_INVERSE| if
   a factor of a modulo was discovered, |ZENERR| for an
   error} 
   \note{One must have $|valuation(PX)\ge valuation(QX)|$ and $|RX| \neq
   |QX|$.}
   \side{|ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZENSrDivide(RX, PX, QX, Rg)
     ZENSr RX, PX, QX;
     ZENRing Rg;
{
  ZENElt sum, prod, inverse, e;
  int k, n, IsOne, vs, vr, df;
  int ret;

  IsOne = 1;
  if (!ZENEltIsOne(ZENSrGetCoeffPtr(QX, ZENSrVal(QX, Rg), Rg), Rg)) {
    if (ZENEltCreate(inverse, Rg)) {
      return(ZERR);
    }
    if ((ret = ZENEltInverse(inverse, ZENSrGetCoeffPtr(QX, ZENSrVal(QX, Rg), Rg), Rg)) != ZEN_HAS_INVERSE) {
      ZENEltFree(inverse, Rg);
      return(ret);
    }
    IsOne = 0;
  }
  
  if (ZENEltCreate(sum, Rg)) {
    if (!IsOne) {
      ZENEltFree(inverse, Rg);
    }
    return(ZERR);
  }
  if (ZENEltCreate(prod, Rg)) {
    ZENEltFree(sum, Rg);
    if (!IsOne) {
      ZENEltFree(inverse, Rg);
    }
    return(ZERR);
  }
  vs  = ZENSrVal(PX, Rg);
  ZENSrVal(RX, Rg) = ZENSrVal(PX, Rg) - ZENSrVal(QX, Rg);
  vr  = ZENSrVal(PX, Rg);
  ZENSrDeg(RX, Rg) = ZENSrDeg(PX, Rg)-ZENSrVal(QX, Rg) <
    ZENSrDeg(QX, Rg)+ vs - 2*ZENSrVal(QX, Rg) ?
      ZENSrDeg(PX, Rg) - ZENSrVal(QX, Rg) :
	ZENSrDeg(QX, Rg) + vs -2*ZENSrVal(QX, Rg);
  df = vs-ZENSrVal(RX, Rg);

  for(n = ZENSrVal(RX, Rg); n <= ZENSrDeg(RX, Rg); n++){
    ZENSrVal(PX, Rg) = vs;
    ZENEltAssign(sum, ZENSrGetCoeffPtr(PX, n+df, Rg), Rg);
    ZENSrVal(PX, Rg) = vr;
    for(k = 1; k <= n-ZENSrVal(RX, Rg); k++){
      ZENEltMultiply(prod,
		     ZENSrGetCoeffPtr(QX, k+ZENSrVal(QX, Rg), Rg),
		     ZENSrGetCoeffPtr(RX, n-k, Rg),
		     Rg);
      ZENEltSubtract(sum, prod, Rg);
    }
    e = ZENSrGetCoeffPtr(RX, n, Rg);
    if (IsOne) {
      ZENEltAssign(e, sum, Rg);
    }
    else {
      ZENEltMultiply(e, sum, inverse, Rg);
    }
    ZENSrSetCoeff(RX, n, e, Rg);
  }
  if (!IsOne) {
    ZENEltFree(inverse, Rg);
  }
  ZENEltFree(sum, Rg);
  ZENEltFree(prod, Rg);
  return(ZEN_HAS_INVERSE);
}
