/*
 * zen/zelog/c/init.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zelog.h"
#include "zeg.h"
#include "zed.h"

/* \TEX{
   \title{Initialization of logs for small extensions.}
   \input{A |ZENRing| |R| of cardinality less than $2^8$.}
   \output{|ZENERR| if an error occured, $|SR_GEN| \vert |MAT_GEN|$ otherwise.}
   \side{Tabulation of the clone |Rg| is performed.}
   \note{All the arithmetic operations are stored.
   This function is called by |ZENRingClone| and should not be called otherwise.}
   }
   */
int _ZelogInit(Rg)
     ZENRing Rg;
{
  int SmallFactors[2*SIZE_CHAR];
  int ne, nb_fact, n;
  BigNumDigit C[1];
  ZENElt a, b, c;
  int i;
  BigNumLength k;

  ZENRingArith(Rg) = ZEN_LIB_ZELOG;

  /* Log table */
  if (!(ZelogLogTablePtr(Rg) = (ZEN_UC_2SC_Table *)ZENMalloc(sizeof(ZEN_UC_2SC_Table)))) {
    ZENSetError(ZEN_LIB_ZELOG, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  /* Addition */
  if (!(ZelogAddTablePtr(Rg) = (ZEN_UC_2SC_Table *)ZENMalloc(sizeof(ZEN_UC_2SC_Table)))) {
    ZENSetError(ZEN_LIB_ZELOG, ZEN_INIT, ZEN_ALLOC);
    ZENFree(ZelogLogTablePtr(Rg));
    return(ZERR);
  }

  /* Trace */
  if (!(ZelogTrTablePtr(Rg) = (ZEN_UC_2SC_Table *)ZENMalloc(sizeof(ZEN_UC_2SC_Table)))) {
    ZENFree(ZelogAddTablePtr(Rg));
    ZENFree(ZelogLogTablePtr(Rg));
    ZENSetError(ZEN_LIB_ZELOG, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  if (ZENEltCreate(a, ZelogRingOrigin(Rg))) {
    ZENFree(ZelogAddTablePtr(Rg));
    ZENFree(ZelogLogTablePtr(Rg));
    ZENFree(ZelogTrTablePtr(Rg));
    ZENSetError(ZEN_LIB_ZELOG, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  if (ZENEltCreate(b, ZelogRingOrigin(Rg))) {
    ZENFree(ZelogAddTablePtr(Rg));
    ZENFree(ZelogLogTablePtr(Rg));
    ZENFree(ZelogTrTablePtr(Rg));
    ZENEltFree(b, ZelogRingOrigin(Rg));
    ZENSetError(ZEN_LIB_ZELOG, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  if (ZENEltCreate(c, ZelogRingOrigin(Rg))) {
    ZENFree(ZelogAddTablePtr(Rg));
    ZENFree(ZelogLogTablePtr(Rg));
    ZENFree(ZelogTrTablePtr(Rg));
    ZENEltFree(a, ZelogRingOrigin(Rg));
    ZENEltFree(b, ZelogRingOrigin(Rg));
    ZENSetError(ZEN_LIB_ZELOG, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

 /* The size of the field */
  ne = (int)ZENRingQ(Rg)[0];
  
  /* Conversion */
  if (!(ZelogConvTablePtr(Rg) = (ZENElt *)ZENMalloc(ne*sizeof(ZENElt)))) {
    ZENFree(ZelogAddTablePtr(Rg));
    ZENFree(ZelogLogTablePtr(Rg));
    ZENFree(ZelogTrTablePtr(Rg));
    ZENEltFree(a, ZelogRingOrigin(Rg));
    ZENEltFree(b, ZelogRingOrigin(Rg));
    ZENEltFree(c, ZelogRingOrigin(Rg));
    ZENSetError(ZEN_LIB_ZELOG, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

   /************ Filling the tables *********************/

  /* factorizing ne-1 */
  nb_fact = 0;
  n = ne-1;
  for(n = ne-1; !(n%2); n /= 2) {
    ;
  }
  if (n < ne-1) {
    SmallFactors[nb_fact++] = 2;
  }
  for(i = 3; i < n; i += 2) {
    if (!(n%i)) {
      for(; !(n%i); n /= i) {
	;
      }
      SmallFactors[nb_fact++] = i; 
    }
  }
  if (n > 1) {
    SmallFactors[nb_fact++] = n; 
  }

  /* finding a generator */
  if (ZENEltCreate(ZelogGenerator(Rg), ZelogRingOrigin(Rg))) {
    ZENFree(ZelogAddTablePtr(Rg));
    ZENFree(ZelogLogTablePtr(Rg));
    ZENFree(ZelogTrTablePtr(Rg));
    ZENFree(ZelogConvTablePtr(Rg));
    ZENEltFree(a, ZelogRingOrigin(Rg));
    ZENEltFree(b, ZelogRingOrigin(Rg));
    ZENEltFree(c, ZelogRingOrigin(Rg));
    ZENSetError(ZEN_LIB_ZELOG, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }
  ZENEltSetToOne(ZelogGenerator(Rg), ZelogRingOrigin(Rg));
  for(ZENEltSetNext(ZelogGenerator(Rg), ZelogRingOrigin(Rg));
      !ZENEltIsZero(ZelogGenerator(Rg), ZelogRingOrigin(Rg));
      ZENEltSetNext(ZelogGenerator(Rg), ZelogRingOrigin(Rg))) {
    for(i = 0; i < nb_fact; i++) {
      C[0] = (ne-1)/SmallFactors[i];
      _ZENEltExpPlain(&a, C, 1, ZelogGenerator(Rg), ZelogRingOrigin(Rg));
      if (ZENEltIsOne(a, ZelogRingOrigin(Rg))) {
	break;
      }
    }
    if (i == nb_fact) {
      break;
    }
  }
  if (ZENEltIsZero(ZelogGenerator(Rg), ZelogRingOrigin(Rg))) {
      /* This is not a finite field */

    ZENFree(ZelogAddTablePtr(Rg));
    ZENFree(ZelogLogTablePtr(Rg));
    ZENFree(ZelogTrTablePtr(Rg));
    ZENFree(ZelogConvTablePtr(Rg));
    ZENEltFree(a, ZelogRingOrigin(Rg));
    ZENEltFree(b, ZelogRingOrigin(Rg));
    ZENEltFree(c, ZelogRingOrigin(Rg));
    ZENEltFree(ZelogGenerator(Rg), ZelogRingOrigin(Rg));
    return(ZENWGINPUT);
  }

  if (ZENEltIsZero(ZelogGenerator(Rg), ZelogRingOrigin(Rg))) {
    ZENFree(ZelogAddTablePtr(Rg));
    ZENFree(ZelogLogTablePtr(Rg));
    ZENFree(ZelogTrTablePtr(Rg));
    ZENFree(ZelogConvTablePtr(Rg));
    ZENEltFree(a, ZelogRingOrigin(Rg));
    ZENEltFree(b, ZelogRingOrigin(Rg));
    ZENEltFree(c, ZelogRingOrigin(Rg));
    ZENEltFree(ZelogGenerator(Rg), ZelogRingOrigin(Rg));
    ZENSetError(ZEN_LIB_ZELOG, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  /* Filling the table of log, log(g^x)=x+1 */
  
  ZelogLogTable(Rg, 0) = 0;
  if (ZENEltCreate(ZelogConvTable(Rg, 0), ZelogRingOrigin(Rg))) {
    ZENFree(ZelogAddTablePtr(Rg));
    ZENFree(ZelogLogTablePtr(Rg));
    ZENFree(ZelogTrTablePtr(Rg));
    ZENEltFree(a, ZelogRingOrigin(Rg));
    ZENEltFree(b, ZelogRingOrigin(Rg));
    ZENEltFree(c, ZelogRingOrigin(Rg));
    ZENEltFree(ZelogGenerator(Rg), ZelogRingOrigin(Rg));
    ZENFree(ZelogConvTablePtr(Rg));
    ZENSetError(ZEN_LIB_ZELOG, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }
  ZENEltSetToZero(ZelogConvTable(Rg, 0), ZelogRingOrigin(Rg));
  
  if (ZENEltCreate(ZelogConvTable(Rg, 1), ZelogRingOrigin(Rg))) {
    ZENFree(ZelogAddTablePtr(Rg));
    ZENFree(ZelogLogTablePtr(Rg));
    ZENFree(ZelogTrTablePtr(Rg));
    ZENEltFree(a, ZelogRingOrigin(Rg));
    ZENEltFree(b, ZelogRingOrigin(Rg));
    ZENEltFree(c, ZelogRingOrigin(Rg));
    ZENEltFree(ZelogConvTable(Rg, 0), ZelogRingOrigin(Rg));
    ZENFree(ZelogConvTablePtr(Rg));
    ZENSetError(ZEN_LIB_ZELOG, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }
  ZENEltSetToOne(a, ZelogRingOrigin(Rg));
  ZENEltSetToOne(ZelogConvTable(Rg, 1), ZelogRingOrigin(Rg));

  for(i = 0; i < ne; i++) {
    ZelogLogTable(Rg, i) = 0;
  }
  ZelogLogTable(Rg, 1) = 1;

  for(i = 2; i < ne; i++) {
    ZENEltMultiply(a, ZelogGenerator(Rg), a, ZelogRingOrigin(Rg));

    if (ZENEltCreate(ZelogConvTable(Rg, i), ZelogRingOrigin(Rg))) {
      ZENFree(ZelogAddTablePtr(Rg));
      ZENFree(ZelogLogTablePtr(Rg));
      ZENFree(ZelogTrTablePtr(Rg));
      ZENEltFree(a, ZelogRingOrigin(Rg));
      ZENEltFree(b, ZelogRingOrigin(Rg));
      ZENEltFree(c, ZelogRingOrigin(Rg));
      for(i--;i>=0;i--)
	ZENEltFree(ZelogConvTable(Rg, i), ZelogRingOrigin(Rg));
      ZENFree(ZelogConvTablePtr(Rg));
      ZENSetError(ZEN_LIB_ZELOG, ZEN_INIT, ZEN_ALLOC);
      return(ZERR);
    }
    ZENEltAssign(ZelogConvTable(Rg, i), a, ZelogRingOrigin(Rg));
    
    ZENEltToZ(C, &k, a, ZelogRingOrigin(Rg));
    if (ZelogLogTable(Rg, C[0])) {
      /* This is not a finite field */

      ZENFree(ZelogAddTablePtr(Rg));
      ZENFree(ZelogLogTablePtr(Rg));
      ZENFree(ZelogTrTablePtr(Rg));
      ZENEltFree(a, ZelogRingOrigin(Rg));
      ZENEltFree(b, ZelogRingOrigin(Rg));
      ZENEltFree(c, ZelogRingOrigin(Rg));
      for(;i>=0;i--)
	ZENEltFree(ZelogConvTable(Rg, i), ZelogRingOrigin(Rg));
      ZENFree(ZelogConvTablePtr(Rg));
      return(ZENWGINPUT);
    }
    ZelogLogTable(Rg, C[0]) = i;
  }

  /* Filling the other tables */
  ZelogAddTable(Rg, 0) = 1;
  ZelogTrTable(Rg, 0) = 0;

  ZENEltSetToOne(a, ZelogRingOrigin(Rg));

  ZENEltSetToOne(b, ZelogRingOrigin(Rg));
  ZENEltAdd(b, a, ZelogRingOrigin(Rg));
  ZENEltToZ(C, &k, b, ZelogRingOrigin(Rg));
  ZelogAddTable(Rg, 1) = ZelogLogTable(Rg, C[0]);

  ZENEltTrace(c, a, ZelogRingOrigin(Rg));
  ZENEltToZ(C, &k, c, ZelogRingOrigin(Rg));
  ZelogTrTable(Rg, 1) = ZelogLogTable(Rg, C[0]);

  for(i = 1; i < ne-1; i++) {

    ZENEltMultiply(a, ZelogGenerator(Rg), a, ZelogRingOrigin(Rg));
    
    /* Filling the table of log(1+g^x) */
    ZENEltSetToOne(b, ZelogRingOrigin(Rg));
    ZENEltAdd(b, a, ZelogRingOrigin(Rg));
    ZENEltToZ(C, &k, b, ZelogRingOrigin(Rg));
    ZelogAddTable(Rg, i+1) = ZelogLogTable(Rg, C[0]);

    /* filling the table of trace(g^x) */
    ZENEltTrace(c, a, ZelogRingOrigin(Rg));
    ZENEltToZ(C, &k, c, ZelogRingOrigin(Rg));
    ZelogTrTable(Rg, i+1) = ZelogLogTable(Rg, C[0]);

    if (ZENEltIsOne(c, ZelogRingOrigin(Rg))) {
      ZelogElt2Int(ZENRingTheta(Rg), Rg) = i+1;
    }
  }
  ZENEltFree(a, ZelogRingOrigin(Rg));
  ZENEltFree(b, ZelogRingOrigin(Rg));
  ZENEltFree(c, ZelogRingOrigin(Rg));


  /************ new pointers of functions are used ***************/

  /* Procedure on rings
   ****************************/
  ZENPrcInitEltMultiplyPtr(Rg) = NULL;
  ZENPrcCloseEltMultiplyPtr(Rg) = NULL;

  ZENPrcInitTracePtr(Rg) = NULL;
  ZENPrcCloseTracePtr(Rg) = NULL;

  ZENPrcInitPolyRootsCanonicalPtr(Rg) = _ZelogPrcInitPolyRootsCanonical;
  ZENPrcClosePolyRootsCanonicalPtr(Rg) = _ZelogPrcClosePolyRootsCanonical;

  ZENPrcInitEltExpPtr(Rg) = _ZENPrcInitEltExp;
  ZENPrcCloseEltExpPtr(Rg) = _ZENPrcCloseEltExp;

  /* Procedure on elements
   ****************************/
  /* Allocation */
  ZENEltCreatePtr(Rg) = _ZelogEltCreate;
  ZENEltCreateLargePtr(Rg) = _ZelogEltCreate;
  ZENEltFreePtr(Rg) = _ZelogEltFree;
  
  /* Assigning */
  ZENEltAssignPtr(Rg) = _ZelogEltAssign;
  ZENEltSetToZeroPtr(Rg) = _ZelogEltSetToZero;
  ZENEltSetToOnePtr(Rg) = _ZelogEltSetToOne;
  ZENEltSetRandomPtr(Rg) = _ZelogEltSetRandom;
  ZENEltSetNextPtr(Rg) = _ZelogEltSetNext;

  ZENEltToClonePtr(Rg) = _ZelogEltToClone;
  ZENCloneToEltPtr(Rg) = _ZelogCloneToElt;

  /* Test */
  ZENEltAreEqualPtr(Rg) = _ZelogEltAreEqual;
  ZENEltIsZeroPtr(Rg) = _ZelogEltIsZero;
  ZENEltIsOnePtr(Rg) = _ZelogEltIsOne;

  /* Arithmetic */
  ZENEltAddPtr(Rg) = _ZelogEltAdd;
  ZENEltNegatePtr(Rg) = _ZelogEltNegate;
  ZENEltSubtractPtr(Rg) = _ZelogEltSubtract;
  ZENEltSquarePtr(Rg) = _ZelogEltSquare;
  ZENEltMultiplyPtr(Rg) = _ZelogEltMultiply;
  ZENEltInversePtr(Rg) = _ZelogEltInverse;
  ZENEltTracePtr(Rg) = _ZelogEltTrace;

  /* io */
  ZENZToEltPtr(Rg) = _ZelogZToElt;
  ZENEltToZPtr(Rg) = _ZelogEltToZ;

  ZENEltReadFromStringPtr(Rg) = _ZENCloneEltReadFromString;
  ZENEltPrintToStringPtr(Rg) = _ZENCloneEltPrintToString;
  ZENEltReadFromFilePtr(Rg) = _ZENCloneEltReadFromFile;
  ZENEltPrintToFilePtr(Rg) = _ZENCloneEltPrintToFile;
  
  ZENEltGetFromStringPtr(Rg) = _ZENCloneEltGetFromString;
  ZENEltPutToStringPtr(Rg) = _ZENCloneEltPutToString;
  ZENEltGetFromFilePtr(Rg) = _ZENCloneEltGetFromFile;
  ZENEltPutToFilePtr(Rg) = _ZENCloneEltPutToFile;
  
  /* Procedures on polynomials
   *******************************/
  /* Allocation */
  ZENPolyCreatePtr(Rg) = _ZelogPolyCreate;
  ZENPolyCreateLargePtr(Rg) = _ZelogPolyCreate;
  ZENPolyFreePtr(Rg) = _ZelogPolyFree;

  /* Assigning */
  ZENPolyAssignPtr(Rg) = _ZelogPolyAssign;
  ZENPolyAssignLargePtr(Rg) = _ZelogPolyAssign;
  ZENPolyLazyToVeryLazyPtr(Rg) = _ZelogPolyAssign;
  ZENPolySetToZeroPtr(Rg) = _ZelogPolySetToZero;
  ZENPolySetToZeroLargePtr(Rg) = _ZelogPolySetToZero;
  ZENPolySetToXiPtr(Rg) = _ZelogPolySetToXi;
  ZENPolySetRandomPtr(Rg) = _ZelogPolySetRandom;
  ZENPolySetCoeffPtr(Rg) = _ZelogPolySetCoeff;
  ZENPolyGetCoeffNewPtr(Rg) = _ZelogPolyGetCoeff;
  ZENPolyGetCoeffPtrPtr(Rg) = _ZelogPolyGetCoeffPtr;
  ZENPolyUpdateDegreePtr(Rg) = _ZelogPolyUpdateDegree;

  /* Test */
  ZENPolyAreEqualPtr(Rg) = _ZelogPolyAreEqual;
  ZENPolyIsZeroPtr(Rg) = _ZelogPolyIsZero;
  ZENPolyIsXiPtr(Rg) = _ZelogPolyIsXi;

  /* Arithmetic */
  ZENPolyAddPtr(Rg) = _ZelogPolyAdd;
  ZENPolyAddLargePtr(Rg) = _ZelogPolyAdd;
  ZENPolyNegatePtr(Rg) = _ZelogPolyNegate;
  ZENPolyNegateLargePtr(Rg) = _ZelogPolyNegate;
  ZENPolySubtractPtr(Rg) = _ZelogPolySubtract;
  ZENPolySubtractLargePtr(Rg) = _ZelogPolySubtract;

  if (ZBNDigit(ZENRingP(Rg), 0) == 2) {
    ZENPolySquarePtr(Rg) = _ZelogPolySquareChar2;
    ZENPolySquareLazyPtr(Rg) = _ZelogPolySquareChar2;
    ZENPolySquareVeryLazyPtr(Rg) = _ZelogPolySquareChar2;
  }
  else {
    ZENPolySquarePtr(Rg) = _ZelogPolySquareKaratsuba;
    ZENPolySquareLazyPtr(Rg) = _ZelogPolySquareKaratsubaLazy;
    ZENPolySquareVeryLazyPtr(Rg) = _ZelogPolySquareKaratsubaLazy;
  }

  ZENPolyMultiplyPtr(Rg) = _ZelogPolyMultiply;
  ZENPolyMultiplyLazyPtr(Rg) = _ZelogPolyMultiplyLazy;
  ZENPolyMultiplyVeryLazyPtr(Rg) = _ZelogPolyMultiplyLazy;

  ZENPolyMultiplyScalarPtr(Rg) = _ZelogPolyMultiplyScalar;
  ZENPolyDotPtr(Rg) = _ZelogPolyDot;
  ZENPolyEvalPtr(Rg) = _ZENPolyEval; /* Evaluation of a polynomial */
  ZENPolyDividePtr(Rg) = _ZelogPolyDivide;
  ZENPolyDivideLazyPtr(Rg) = _ZelogPolyDivideLazy;
  ZENPolyDivideVeryLazyPtr(Rg) = _ZelogPolyDivideLazy;
  ZENPolyModuloLazyPtr(Rg) = _ZelogPolyModuloLazy;
  ZENPolyModuloVeryLazyPtr(Rg) = _ZelogPolyModuloLazy;

  /* -1 in Rg */
  if ((ZENRingPl(Rg) == 1) && (ZBNDigit(ZENRingP(Rg), 0) == 2)) {
    ZelogElt2Int(ZelogMinusOne(Rg), Rg) = 1;
  }
  else {
    ZelogElt2Int(ZelogMinusOne(Rg), Rg) = 1+(ZBNDigit(ZENRingQ(Rg), 0)-1)/2;
  }
  
  /* Procedures on matrices
   **************************/
  /* Allocation */
  ZENMatCreateParamsPtr(Rg) = _ZelogMatCreateParams; /* Create a matrix */
  ZENMatCreatePtr(Rg) = __ZedsMatCreate;	/* Create a matrix */

  /* Assigning */
  ZENMatAssignPtr(Rg) = _ZelogMatAssign; /* Assign a matrix with another */
  ZENPermutation2MatrixPtr(Rg) = _ZENPermutation2Matrix;	/* Assign a matrix with another */
  ZENMatSetToZeroPtr(Rg) = _ZelogMatSetToZero; /* Set a matrix to zero */
  ZENMatSetToOnePtr(Rg) = _ZelogMatSetToOne; /* Set a matrix to identity */
  ZENMatSetRandomPtr(Rg) = _ZelogMatSetRandom; /* Set randomly a matrix */
  ZENMatSetBlocPtr(Rg) = _ZelogMatSetBloc; /* Set a coefficient of a matrix */
  ZENMatGetBlocNewPtr(Rg) = _ZelogMatGetBloc; /* Get a coefficient of a matrix */
  ZENMatGetBlocPtrPtr(Rg) = _ZelogMatGetBlocPtr; /* Get a coefficient of a matrix */
  ZENMatSetSubMatPtr(Rg) = _ZelogMatSetSubMat; /* Set a submatrix */
  ZENMatGetSubMatPtr(Rg) = _ZelogMatGetSubMat; /* Get a submatrix */
  
  ZENMatPermuteBnPtr(Rg) = _ZelogMatPermuteBn;
  ZENMatPermuteDigitPtr(Rg) = _ZelogMatPermuteDigit;
  
  /* Test */
  ZENMatAreEqualPtr(Rg) = _ZelogMatAreEqual; /* Is a matrix equal to another */
  ZENMatIsZeroPtr(Rg) = _ZelogMatIsZero;	/* Is a matrix equal to zero */
  ZENMatIsOnePtr(Rg) = _ZelogMatIsOne;	/* Is a matrix equal to one */
  
  /* Arithmetic */
  ZENMatAddPtr(Rg) = _ZelogMatAdd;	/* Addition of Matrix */
  ZENMatAddScalarPtr(Rg) = _ZelogMatAddScalar;	/* Addition of Matrix */
  ZENMatMultiplyScalarPtr(Rg) = _ZelogMatMultiplyScalar;	/* Multiplication of Matrix */
  ZENMatSubtractPtr(Rg) = _ZelogMatSubtract; /* Substraction of Matrix */
  ZENMatNegatePtr(Rg) = _ZelogMatNegate;	/* Squaring of Matrix */
  ZENMatWinogradPtr(Rg) = _ZENMatWinograd; /* Multiplication of Matrix */
  ZENMatMultiplyPtr(Rg) = _ZelogMatMultiplyPlain; /* Multiplication of Matrix */
  ZENMatMultiplyPlainPtr(Rg) = _ZelogMatMultiplyPlain; /* Multiplication of Matrix */
  ZENMatMultiplyLeftPermutationPtr(Rg) = _ZENMatMultiplyLeftPermutation;
  ZENMatMultiplyRightPermutationPtr(Rg) = _ZENMatMultiplyRightPermutation;
  ZENMatInversePtr(Rg) = _ZelogMatInverse; /* Inverse of a Matrix */
  ZENMatKernelPtr(Rg) = _ZelogMatKernel;	/* Kernel of a Matrix */  
  ZENMatGaussPtr(Rg) = _ZelogMatGauss;
  
  /* IO */
  ZENMatGetFromFilePtr(Rg) = _ZENMatGetFromFile;
  ZENMatPutToFilePtr(Rg) = _ZENMatPutToFile;
  ZENMatGetFromStringPtr(Rg) = _ZENMatGetFromString;
  ZENMatPutToStringPtr(Rg) = _ZENMatPutToString;

  /* Matrix conversion */
  
  if(ZENRingArith(ZENRingOrigin(Rg))==ZEN_LIB_ZEXT)
    {
      if(ZENRingArith(ZENRingDef(Rg))==ZEN_LIB_ZE2)
	{
	  ZENVect2MatPtr(Rg) = _Zelog2mVect2Mat;
	  ZENMat2VectPtr(Rg) = _Zelog2mMat2Vect;
	}
      else
	{
	  ZENVect2MatPtr(Rg) = _ZENCloneVect2Mat;
	  ZENMat2VectPtr(Rg) = _ZENCloneMat2Vect;
	}
    }
  else
    {
      ZENVect2MatPtr(Rg) = ZENMatAssignPtr(Rg);
      ZENMat2VectPtr(Rg) = ZENMatAssignPtr(Rg);
    }
  return(SR_GEN);
}


/* \TEX{
   \title{Desallocating logles for small extensions.}
   \input{A |ZENRing| of cardinality less than $2^8$.}
   \output{|ZENERR| if an error occured, |ZEN_HAS_INVERSE| otherwise.}
   \note{All logles allocated with |_ZelogPrcInitEltTabulate| are
   freed.}
   }
   */
int _ZelogClose(Rg)
     ZENRing Rg;
{int i;

 i=(int)ZENRingQ(Rg)[0];
 
 for(i--;i>=0;i--)
   ZENEltFree(ZelogConvTable(Rg, i), ZelogRingOrigin(Rg));
 ZENFree(ZelogConvTablePtr(Rg));

 /* Restoring the allocated memory */
 ZENFree(ZelogAddTablePtr(Rg));
 ZENFree(ZelogLogTablePtr(Rg));
 ZENFree(ZelogTrTablePtr(Rg));
 ZENEltFree(ZelogGenerator(Rg), ZelogRingOrigin(Rg));

 return(ZEN_HAS_INVERSE);
}


