/*
 * zen/zelog/c/mat_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial additions}
   }
   */
#include <stdio.h>
#include "zelog.h"

/*
   \TEX{
   \title{Addition.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| += |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int _ZelogMatAdd(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;

 ZENTest(ZelogMatIsPermutation(A,R)
	 ||ZelogMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATADD,ZEN_LIB_ZELOG, ZERR);

 if(ZelogMatAreSameType(A,B,R))
   {

     ZENTest((ZelogMatNbBn(A,R) != ZelogMatNbBn(B,R))
	     ||(ZelogMatNbDigit(A,R) != ZelogMatNbDigit(B,R)),
	     ZEN_MAT_DIM,ZEN_MATADD,ZEN_LIB_ZELOG, ZERR);

     for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
	 ZelogAnyEltAdd(ZelogMatBloc(A,l,c,R),ZelogMatBloc(B,l,c,R),R);
   }
 else
   {

     ZENTest((ZelogMatNbBn(A,R) != ZelogMatNbDigit(B,R))
	     ||(ZelogMatNbDigit(A,R) != ZelogMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATADD,ZEN_LIB_ZELOG, ZERR);
     
     for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
	 ZelogAnyEltAdd(ZelogMatBloc(A,l,c,R),ZelogMatBloc(B,c,l,R),R);

   }

 return(0);
}

   
/*
   \TEX{
   \title{Addition of scalar}
   \input{A |ZENMat|, a |ZENElt| and a |ZENRing|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{All the elements of |A| are incremented by |b|.}
   \note{|A| cannot be a permutation.}
   }
   */
int _ZelogMatAddScalar(A,b,R)
     ZENMat A;
     ZENElt b;
     ZENRing R;
{
  int l,c;

  ZENTest(ZelogMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATADD,ZEN_LIB_ZELOG, ZERR);

     
  for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
    for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
      ZelogAnyEltAdd(ZelogMatBloc(A,l,c,R),ZelogElt2Int(b,R),R);
  
  return(0);
}

 /* \TEX{
   \title{Negation.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = -|B| if no error occured, |zen_errno| is set otherwise.}
   }
   */

int _ZelogMatNegate(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;

 ZENTest(ZelogMatIsPermutation(A,R)
	 ||ZelogMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATNEGATE,ZEN_LIB_ZELOG, ZERR);

 if(ZelogMatAreSameType(A,B,R))
   {

     ZENTest((ZelogMatNbBn(A,R) != ZelogMatNbBn(B,R))
	     ||(ZelogMatNbDigit(A,R) != ZelogMatNbDigit(B,R)),
	     ZEN_MAT_DIM,ZEN_MATNEGATE,ZEN_LIB_ZELOG, ZERR);

     for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
	   ZelogAnyEltNegate(ZelogMatBloc(A,l,c,R),ZelogMatBloc(B,l,c,R),R);

   }
 else
   {

     ZENTest((ZelogMatNbBn(A,R) != ZelogMatNbDigit(B,R))
	     ||(ZelogMatNbDigit(A,R) != ZelogMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATNEGATE,ZEN_LIB_ZELOG, ZERR);

     for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
	 ZelogAnyEltNegate(ZelogMatBloc(A,l,c,R),ZelogMatBloc(B,c,l,R),R);
	 
   }

 return(0);
}

/*
   \TEX{
   \title{Subtraction.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| -= |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int _ZelogMatSubtract(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;

 ZENTest(ZelogMatIsPermutation(A,R)
	 ||ZelogMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATSUBTRACT,ZEN_LIB_ZELOG, ZERR);

 if(ZelogMatAreSameType(A,B,R))
   {

     ZENTest((ZelogMatNbBn(A,R) != ZelogMatNbBn(B,R))
	     ||(ZelogMatNbDigit(A,R) != ZelogMatNbDigit(B,R)),
	     ZEN_MAT_DIM,ZEN_MATSUBTRACT,ZEN_LIB_ZELOG, ZERR);

     for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
	 ZelogAnyEltSubtract(ZelogMatBloc(A,l,c,R),ZelogMatBloc(B,l,c,R),R);
   }
 else
   {

     ZENTest((ZelogMatNbBn(A,R) != ZelogMatNbDigit(B,R))
	     ||(ZelogMatNbDigit(A,R) != ZelogMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATSUBTRACT,ZEN_LIB_ZELOG, ZERR);

     for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
	 ZelogAnyEltSubtract(ZelogMatBloc(A,l,c,R),ZelogMatBloc(B,c,l,R),R);
   }

 return(0);
}

   
