/*
 * zen/zelog/c/mat_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial assignments}

   }
   */
#include <stdio.h>
#include "zelog.h"

/*
   \TEX{
   \title{Get a bloc}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{|E| is set to the indexed bloc of |A|}
   }
   */
   
int _ZelogMatGetBloc(p_E,A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENElt *p_E;
     ZENRing R;
{
  ZENTest(ZelogMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATGETBLOC,ZEN_LIB_ZELOG, ZERR);
  ZENTest((l<0)||(c<0)||(l>=ZelogMatNbBn(A,R))||(c>=ZelogMatNbDigit(A,R)),
	  ZEN_MAT_DIM0,ZEN_MATGETBLOC,ZEN_LIB_ZELOG, ZERR);
  
  ZelogAnyEltAssign(ZelogElt2Int(*p_E,R),ZelogMatBloc(A,l,c,R),R);
  return(0);
}

/*
   \TEX{
   \title{Set a bloc}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The indexed bloc of |A| is set to |E|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred.
   This function must be chosen together with |_ZelogMatCreateParams|.}
   }
   
   */
   
int _ZelogMatSetBloc(A,l,c,E,R)
     ZENMat A;
     Dim l,c;
     ZENElt E;
     ZENRing R;
{
  ZENTest(ZelogMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATSETBLOC,ZEN_LIB_ZELOG, ZERR);
  ZENTest((l<0)||(c<0)||(l>=ZelogMatNbBn(A,R))||(c>=ZelogMatNbDigit(A,R)),
	  ZEN_MAT_DIM0,ZEN_MATSETBLOC,ZEN_LIB_ZELOG, ZERR);

  ZelogAnyEltAssign(ZelogMatBloc(A,l,c,R),ZelogElt2Int(E,R),R);
  return(0);
}

/*
   \TEX{
   \title{Get a bloc pointer}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{|E| is set to the indexed bloc of |A|}
   \note{This function must be chosen together with |_ZedMatCreateParams|.}
   }
   
   */
   
ZENElt _ZelogMatGetBlocPtr(A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENRing R;
{
  ZENElt ret;

  ZelogAnyEltAssign(ZelogElt2Int(ret,R),ZelogMatBloc(A,l,c,R),R);
  return(ret);
}

/************************************************/

static void __ZelogMatGetSubMatST(S,A,l,c,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int i,j,k;
 
 for(i=0;i<ZelogMatNbBn(S,R);i++,l++)
   for(j=0,k=c;j<ZelogMatNbDigit(S,R);j++,k++)    
     ZelogAnyEltAssign(ZelogMatBloc(S,i,j,R),ZelogMatBloc(A,l,k,R),R);
}

static void __ZelogMatGetSubMatOT(S,A,l,c,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int i,j,k;
 
 for(i=0;i<ZelogMatNbBn(S,R);i++,l++)
   for(j=0,k=c;j<ZelogMatNbDigit(S,R);j++,k++)    
     ZelogAnyEltAssign(ZelogMatBloc(S,i,j,R),ZelogMatBloc(A,k,l,R),R);
}

static void __ZelogMatSetSubMatST(A,l,c,S,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int i,j,k;
 
 for(i=0;i<ZelogMatNbBn(S,R);i++,l++)
   for(j=0,k=c;j<ZelogMatNbDigit(S,R);j++,k++)    
     ZelogAnyEltAssign(ZelogMatBloc(A,l,k,R),ZelogMatBloc(S,i,j,R),R);
}

static void __ZelogMatSetSubMatOT(A,l,c,S,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int i,j,k;
 
 for(i=0;i<ZelogMatNbBn(S,R);i++,l++)
   for(j=0,k=c;j<ZelogMatNbDigit(S,R);j++,k++)    
     ZelogAnyEltAssign(ZelogMatBloc(A,k,l,R),ZelogMatBloc(S,i,j,R),R);
}

/*
   \TEX{
   \title{Getting a submatrix}
   \input{Two allocated |ZENMat|, two indexes and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The submatrix of |M| at position $(|r|,|c|)$ of size those of |S| is extracted and assigned to |S|.}
   \note{Permutation matrix cannot be used.}
   }
   */

int _ZelogMatGetSubMat(S,A,l,c,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int cas;
 
 ZENTest((ZelogMatIsPermutation(A,R))||(ZelogMatIsPermutation(S,R)),
	 ZEN_ISPERM,ZEN_MATGETSUBMAT,ZEN_LIB_ZELOG, ZERR);
 ZENTest((l<0)||(c<0)||((l+ZelogMatNbRow(S,R))>ZelogMatNbRow(A,R))||((c+ZelogMatNbCol(S,R))>ZelogMatNbCol(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATGETSUBMAT,ZEN_LIB_ZELOG, ZERR);

 cas=(ZelogMatIsRowType(A,R)<<1)|(ZelogMatIsRowType(S,R));

 switch(cas)
   {
   case _RR_:
     __ZelogMatGetSubMatST(S,A,l,c,R);
     break;
   case _RC_:
     __ZelogMatGetSubMatOT(S,A,c,l,R);
     break;
   case _CR_:
     __ZelogMatGetSubMatOT(S,A,l,c,R);
     break;
   case _CC_:
     __ZelogMatGetSubMatST(S,A,c,l,R);
     break;
   }   
 return(0);
}

/*
   \TEX{
   \title{Assigning a submatrix}
   \input{Two |ZENMat|, two indexes, a |ZENElt| and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The matrix |S| is assigned at position $(|r|,|c|)$ in matrix |M|.}
   \note{Permutation matrix cannot be used.}
   }
   */
int _ZelogMatSetSubMat(A,l,c,S,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int cas;
 
 ZENTest(ZelogMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETSUBMAT,ZEN_LIB_ZELOG, ZERR);
 ZENTest((l<0)||(c<0)||((l+ZelogMatNbRow(S,R))>ZelogMatNbRow(A,R))||((c+ZelogMatNbCol(S,R))>ZelogMatNbCol(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATSETSUBMAT,ZEN_LIB_ZELOG, ZERR);
 
 cas=(ZelogMatIsRowType(A,R)<<1)|(ZelogMatIsRowType(S,R));

 switch(cas)
   {
   case _RR_:
     __ZelogMatSetSubMatST(A,l,c,S,R);
     break;
   case _RC_:
     __ZelogMatSetSubMatOT(A,c,l,S,R);
     break;
   case _CR_:
     __ZelogMatSetSubMatOT(A,l,c,S,R);
     break;
   case _CC_:
     __ZelogMatSetSubMatST(A,c,l,S,R);
     break;
   }     
  
  return(0);
}

/*
   \TEX{
   
   \title{Assignment.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int _ZelogMatAssign(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;

 ZENTest(ZelogMatIsPermutation(A,R)
	 ||ZelogMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATASSIGN,ZEN_LIB_ZELOG, ZERR);
 
 if(ZelogMatAreSameType(A,B,R))
   ZelogMatSameTypeAssign(A,B,R);
 else
   {
     ZENTest((ZelogMatNbBn(A,R) != ZelogMatNbDigit(B,R))
	     ||(ZelogMatNbDigit(A,R) != ZelogMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATASSIGN,ZEN_LIB_ZELOG, ZERR);
     
     for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
	 ZelogAnyEltAssign(ZelogMatBloc(A,l,c,R),ZelogMatBloc(B,c,l,R),R);
   }
 return(0);
}

/*
   \TEX{
   \title{Set to zero}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are set to zero.}
   }
   */
   
int _ZelogMatSetToZero(A,R)
     ZENMat A;
     ZENRing R;
{
 ZENTest(ZelogMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETZERO,ZEN_LIB_ZELOG, ZERR);

 ZelogMatSetToZero(A,R);
 
 return(0);
}

/*
   \TEX{
   \title{Set to one}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are set to zero except for the coefficients with same indices on rows and columns that are set to one.}
   \note{There is no error if the matrix is not square.}
   }
   */
   
int _ZelogMatSetToOne(A,R)
     ZENMat A;
     ZENRing R;
{int l,c;

 ZENTest(ZelogMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETONE,ZEN_LIB_ZELOG, ZERR);

 for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
     if(l==c)
       ZelogAnyEltSetToOne(ZelogMatBloc(A,l,c,R),R);
     else
       ZelogAnyEltSetToZero(ZelogMatBloc(A,l,c,R),R);
     
 return(0);
}

/*
   \TEX{
   \title{Set to random}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are randomly set.}
   }
   
   */
   
int _ZelogMatSetRandom(A,R)
     ZENMat A;
     ZENRing R;
{int l,c;

 ZENTest(ZelogMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETRANDOM,ZEN_LIB_ZELOG, ZERR);
 for(l=ZelogMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZelogMatNbDigit(A,R)-1;c>=0;c--)
     ZelogAnyEltSetRandom(ZelogMatBloc(A,l,c,R),R);

 return(0);
}

