/*
 * zen/zelog/c/mat_inverse.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matrix inversion}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions. 
   }
   */
#include <stdio.h>
#include "zelog.h"
#include "zeg.h"

#ifndef min
#  define min(x,y) ((x)<(y)?(x):(y))
#endif

#define FreeReturn(ret) if(1) {    \
  ZelogMatFree(S,R);                 \
  ZelogMatFree(D,R);                 \
  return(ret);                     \
  } else return(0)



#define ZelogAnyCheckInverse(X,A,R) switch(ZelogAnyEltInverse(X,A,R)) { \
	case ZEN_HAS_INVERSE: break; \
	case ZEN_NO_INVERSE: FreeReturn(ZEN_NO_INVERSE); \
	default: ZENTest(ZENErrErrno == ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_MATINVERSE,ZEN_LIB_ZELOG, ZERR); FreeReturn(ZERR);}

#define ZelogCheckGaussInverse(X,A,R) switch(ZelogMatGaussInverse(X,A,R)) { \
	case ZEN_HAS_INVERSE: break; \
	case ZEN_NO_INVERSE: FreeReturn(ZEN_NO_INVERSE); \
	default: ZENTest(ZENErrErrno == ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_MATINVERSE,ZEN_LIB_ZELOG, ZERR); FreeReturn(ZERR);}


/*
   \TEX{
   \title{Matrix inversion}
   \input{Two |ZENMat| over a |ZENRing|}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|I|=|M|^{-1}$. If an inverse was impossible to compute,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */

int _ZelogMatInverse(I,M,R)
     ZENMat I,M;
     ZENRing R;
{ZENMat D,S;
 int j,k;
 ZENElt n;
 
 ZENTest(ZelogMatIsPermutation(I,R)
	 ||ZelogMatIsPermutation(M,R),
	 ZEN_ISPERM,ZEN_MATINVERSE,ZEN_LIB_ZELOG, ZERR);
 ZENTest((ZelogMatNbBn(I,R) != ZelogMatNbBn(M,R))
	 ||(ZelogMatNbDigit(M,R) != ZelogMatNbBn(M,R))
	 ||(ZelogMatNbDigit(I,R) != ZelogMatNbBn(M,R)),
	 ZEN_MAT_DIM,ZEN_MATINVERSE,ZEN_LIB_ZELOG, ZERR);
 
 if(!(D=ZelogMatCopy(M,R)))
   return(ZERR);
 
 if(!(S=(ZelogMatIsRowType(D,R)? ZelogMatRowCreate(ZelogMatNbBn(D,R),ZelogMatNbBn(D,R),R) : ZelogMatColCreate(ZelogMatNbBn(D,R),ZelogMatNbBn(D,R),R))))
   return(ZERR);
 
 ZENCheckError(ZelogMatSetToOne(S,R), ZERR);
 ZelogCheckGaussInverse(D,S,R);

 if(ZelogMatAreSameType(I,D,R))
   for(k=ZelogMatNbDigit(D,R)-1;k>=0;k--)
     {
       ZelogAnyCheckInverse(ZelogElt2Int(n,R),ZelogMatBloc(D,k,k,R),R);
       for(j=ZelogMatNbBn(D,R)-1;j>=0;j--)
	ZelogAnyEltMultiply(ZelogMatBloc(I,k,j,R),ZelogElt2Int(n,R),ZelogMatBloc(S,k,j,R),R);
     }
 else
   for(k=ZelogMatNbDigit(D,R)-1;k>=0;k--)
     {
       ZelogAnyCheckInverse(ZelogElt2Int(n,R),ZelogMatBloc(D,k,k,R),R);
       for(j=ZelogMatNbBn(D,R)-1;j>=0;j--)
	ZelogAnyEltMultiply(ZelogMatBloc(I,j,k,R),ZelogElt2Int(n,R),ZelogMatBloc(S,k,j,R),R);
	 
     }
#undef ZelogCheckInverse
#undef ZelogCheckGaussInverse

 FreeReturn(ZEN_HAS_INVERSE);
#undef FreeReturn
 
}

#define FreeReturn(ret) if(1) {    \
  ZelogMatFree(P,R);                 \
  ZelogMatFree(D,R);                 \
  return(ret);                     \
  } else return(0)


#define ZelogAnyCheckInverse(X,A,R) switch(ZelogAnyEltInverse(X,A,R)) { \
	case ZEN_HAS_INVERSE: break; \
	case ZEN_NO_INVERSE: FreeReturn(ZEN_NO_INVERSE); \
	default: ZENTest(ZENErrErrno == ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_MATINVERSE,ZEN_LIB_ZELOG, ZERR); FreeReturn(ZERR);}

/*
   \TEX{
   \title{Matrix kernel}
   \input{A pointer on an unallocated |ZENMat| and a |ZENMat| over a |ZENRing|}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{|*p_K| is allocated and set to a basis of the kernel of |M|.
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */

int _ZelogMatKernel(p_K,M,R)
     ZENMat *p_K,M;
     ZENRing R;
{
  int n,rk,i,j;
  Dim l;
  ZENMat D,P;
  ZENElt z,I,elt;
  
  n=ZelogMatNbCol(M,R);
  
  if(!(D=ZelogMatRowCreate(ZelogMatNbRow(M,R),n,R)))
    return(ZERR);
  
  ZENCheckError(ZelogMatAssign(D,M,R), ZERR);
  
  if(!(P=ZENPermutationColCreate(n,R)))
    return(ZERR);
  
  switch(ZelogMatGaussLazy(D,P,&rk,R))
    {
    case ZEN_HAS_INVERSE:
      break; 
    case ZEN_NO_INVERSE:
      FreeReturn(ZEN_NO_INVERSE);
      
    default:
      ZENTest(ZENErrErrno == ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_MATKERNEL,ZEN_LIB_ZELOG, ZERR);
      FreeReturn(ZERR);
    }

#define dim (n<=rk ? 1 : n-rk)
 
  if(!(*p_K=(ZelogMatIsRowType(M,R) ? ZelogMatRowCreate(dim,n,R) : ZelogMatColCreate(dim,n,R))))
    return(ZERR);
 
  ZelogMatSetToZero(*p_K,R);
  
  for(i=0;i<n-rk;i++)
    for(j=0;j<n;j++)
      {
	l=ZENPermutationBn(P,j,R);
	if(j<rk)
	  {
	    ZelogAnyCheckInverse(ZelogElt2Int(I,R),ZelogMatBloc(D,j,j,R),R);
	    ZelogEltNegate(z,I,R);
	    elt=ZelogMatGetCoeffPtr(D,j,i+rk,R);
	    ZelogEltMultiply(I,z,elt,R);
	    ZelogAnyEltAssign(ZelogMatCoeff(*p_K,i,l,R),ZelogElt2Int(I,R),R);
	  }
	else if (j==rk+i)
	  ZelogAnyEltSetToOne(ZelogMatCoeff(*p_K,i,l,R),R);
      }

  FreeReturn(ZEN_HAS_INVERSE);
  
}
