/*
 * zen/zelog/c/poly_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial assignment}
   
   One can set a |ZENPoly| to zero with |_ZelogPolySetToZero|. Then we can
   assign a |ZENPoly| with an other with |_ZelogPolyAssign|. Furthermore, we
   can set randomly a |ZENPoly| with |_ZelogPolySetRandom|. 
   
   The functions |_ZelogPolyGetCoeff| and |_ZelogPolySetCoeff|
   enable us to get or set the coefficient of $X^i$,
   |_ZelogPolyUpdateDegree| compute the real degree of a |ZENPoly| if it
   decreased.
   }
   */
#include <stdio.h>
#include "zelog.h"

/* \TEX{
   \title{Assign.}
   \input{Two |ZENPoly| |PX| and |RX| of a finite ring |Rg|.}
   \side{|RX| is filled with |PX|.}
   }
   */
void _ZelogPolyAssign(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZelogPolyDeg(PX, Rg); i++) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, i, Rg),
		ZelogPolyGetCof(PX, i, Rg),
		Rg);
  }
  ZelogPolyDeg(RX, Rg) =  ZelogPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Set to zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{|RX| is set to zero..}
   }
   */
void _ZelogPolySetToZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  ZelogPolyDeg(RX, Rg) =  0;
  ZelogAnyEltSetToZero(ZelogPolyGetCof(RX, 0, Rg), Rg);
}

/* \TEX{
   \title{Set to $X^i$.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg| allocated at least for
   degree $|deg|$.}
   \side{|RX| is set to $X^{|deg|}$.}
   }
   */
void _ZelogPolySetToXi(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  int i;

  ZelogPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    ZelogAnyEltSetToZero(ZelogPolyGetCof(RX, i, Rg), Rg);
  }
  ZelogAnyEltSetToOne(ZelogPolyGetCof(RX, deg, Rg), Rg);
}

/* \TEX{
   \title{Set to random.}
   \input{A |ZENPoly| |RX| of degree |deg| of a finite ring |Rg|.}
   \side{|RX| is set randomly with degree |deg|.}
   }
   */
void _ZelogPolySetRandom(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  int i;

  ZelogPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    ZelogAnyEltSetRandom(ZelogPolyGetCof(RX, i, Rg), Rg);
  }
  ZelogAnyEltSetRandom(ZelogPolyGetCof(RX, deg, Rg), Rg);
  while(ZelogAnyEltIsZero(ZelogPolyGetCof(RX, deg, Rg), Rg)) {
    ZelogAnyEltSetRandom(ZelogPolyGetCof(RX, deg, Rg), Rg);
  }
}

/* \TEX{
   \title{Set a coeff.}
   \input{A |ZENPoly| |RX| and a |ZENElt| |b| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \side{The coeff of $X^d$ in |RX| is set to |b|.}
   }
   */
void _ZelogPolySetCoeff(RX, d, b, Rg)
     ZENPoly RX;
     int d;
     ZENElt b;
     ZENRing Rg;
{
  ZelogPolySetCoeff(RX, d, b, Rg);
}

/* \TEX{
   \title{Get a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The coeff of $X^d$ in |RX|.}
   }
   */
void _ZelogPolyGetCoeff(p_c, RX, d, Rg)
     ZENElt *p_c;
     ZENPoly RX;
     int d;
     ZENRing Rg;
{ 
  ZelogPolyGetCoeff(*p_c, RX, d, Rg);
}

/* \TEX{
   \title{Extract a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The pointer to the coeff of $X^d$ in |RX|.}
   }
   */
ZENElt _ZelogPolyGetCoeffPtr(RX, d, Rg)
     ZENPoly RX;
     int d;
     ZENRing Rg;
{
  ZENElt elt;

  ZelogElt2Int(elt, Rg) = (BigNumDigit)ZelogPolyGetCof(RX,d,Rg);
  return(elt);
}

/* \TEX{
   \title{Update the degree of a polynomial.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{Resetting the degree of |P|, assuming it has decreased.}
   }
   */
void _ZelogPolyUpdateDegree(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  int i;

  for(i = ZelogPolyDeg(RX, Rg); i >= 0; i--) {
    if (ZelogPolyGetCof(RX,i,Rg)) {
      break;
    }
  }
  if (i < 0) {
    ZelogPolySetToZero(RX, Rg);
  }
  else {
    ZelogPolyDeg(RX, Rg) =  i;
  }
}
