/*
 * zen/zelog/c/poly_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Polynomial multiplication}
   
   Two functions are provided to do multiplications: |_ZelogPolySquare| to square a
   |ZENPoly| and |_ZelogPolyMultily| to compute the multiplication of two |ZENPoly|.
   }
   */
#include <stdio.h>
#include "zelog.h"
#include "zed.h"

#define ZELOG_POLY_MULTIPLY_CUTOFF 1
#define ZELOG_POLY_SQUARE_CUTOFF 4

/* \TEX{
   \title{Squaring in characteristic 2.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZelogPolySquareChar2(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;
  
  for(i = ZelogPolyDeg(PX, Rg); i; i--) {
    ZelogAnyEltSquare(ZelogPolyGetCof(RX, 2*i, Rg),
		  ZelogPolyGetCof(PX, i, Rg),
		  Rg);
    ZelogAnyEltSetToZero(ZelogPolyGetCof(RX, 2*i-1, Rg), Rg);
  }
  ZelogAnyEltSquare(ZelogPolyGetCof(RX, 0, Rg),
		ZelogPolyGetCof(PX, 0, Rg),
		Rg);
  ZelogPolyDeg(RX, Rg) = 2*ZelogPolyDeg(PX, Rg);
  return(0);
}

/* \TEX{
   \title{Squaring when the characteristic is odd.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZelogPolySquareCharp(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, j;

  for(i = ZelogPolyDeg(PX, Rg)+1; i <= 2*ZelogPolyDeg(PX, Rg); i++) {
    ZelogAnyEltSetToZero(ZelogPolyGetCof(RX, i, Rg), Rg);
  }

  for(i = ZelogPolyDeg(PX, Rg); i > 0; i--) {
    
    ZelogAnyEltSquare(ZelogElt2Int(ZENRingHbuf(Rg), Rg), ZelogPolyGetCof(PX, i, Rg), Rg);
    ZelogAnyEltAdd(ZelogPolyGetCof(RX, 2*i, Rg), ZelogElt2Int(ZENRingHbuf(Rg), Rg), Rg);
    
    for(j = i-1; j > 0; j--) {
      ZelogAnyEltMultiply(ZelogElt2Int(ZENRingHbuf(Rg), Rg),
		      ZelogPolyGetCof(PX, j, Rg),
		      ZelogPolyGetCof(PX, i, Rg),
		      Rg);
      ZelogEltAdd(ZENRingHbuf(Rg), ZENRingHbuf(Rg), Rg);
      ZelogAnyEltAdd(ZelogPolyGetCof(RX, i+j, Rg), ZelogElt2Int(ZENRingHbuf(Rg), Rg), Rg);
    }
    ZelogAnyEltMultiply(ZelogElt2Int(ZENRingHbuf(Rg), Rg),
		    ZelogPolyGetCof(PX, j, Rg),
		    ZelogPolyGetCof(PX, i, Rg),
		    Rg);
    ZelogEltAdd(ZENRingHbuf(Rg), ZENRingHbuf(Rg), Rg);
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, i+j, Rg), ZelogElt2Int(ZENRingHbuf(Rg), Rg), Rg);
  }
  ZelogAnyEltSquare(ZelogPolyGetCof(RX, 0, Rg), ZelogPolyGetCof(PX, 0, Rg), Rg);
    
  ZelogPolyDeg(RX, Rg) = 2*ZelogPolyDeg(PX, Rg);
  return(0);
}

/* \TEX{
   \title{Squaring with Karatsuba.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZelogPolySquareKaratsuba(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{  
  int i;
  ZENPoly RBuf;

  if (!(RBuf = ZelogPolyCreateLarge(4*ZelogPolyDeg(PX, Rg)+1, Rg))) {
    return(ZERR);
  }
  ZelogPolySquareKaratsubaLazy(RBuf, PX, Rg);
  ZelogPolyAssign(RX, RBuf, Rg);
  ZelogPolyFree(RBuf, Rg);
  return(ZEN_HAS_INVERSE);
}

static void ZelogPolySquareKaratsubaLazyDegree(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  struct zenpoly UP1, UP3;
  ZENPoly up1, up3;
  int m, n, i, j;
  
  /* Initialization ***********/
  up1 = &UP1; up3 = &UP3;
  m = (ZelogPolyDeg(PX, Rg)+1)/2; 
  n = ZelogPolyDeg(PX, Rg)+1-m;
  
  /* let's go with the recursivity **********/
  
  /* PX- ^ 2 -> RX */
  ZelogPolyDeg(up1, Rg) = m-1;
  ZelogPolyCof(up1, Rg) = ZelogPolyCof(PX, Rg);
  ZelogPolyCof(up3, Rg) = ZelogPolyCof(RX, Rg);
  ZelogPolySquareKaratsubaLazy(up3, up1, Rg);
  
  /* PX- + PX+ -> RX+4*m+2*n-2 */
  for(j = 0, i = 4*m+2*n-2;
      j < m;
      j++, i++) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, i, Rg),
		  ZelogPolyGetCof(PX, j+m, Rg),
		  Rg);
    ZelogAnyEltAdd(ZelogPolyGetCof(RX, i, Rg), ZelogPolyGetCof(PX, j, Rg), Rg);
  }
  if (n > m) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, i, Rg),
		  ZelogPolyGetCof(PX, j+m, Rg),
		  Rg);
  }
  
  /* (PX- + PX+) ^ 2 -> RX+2*m-1 */
  ZelogPolyDeg(up1, Rg) = n-1;
  ZelogPolyCof(up1, Rg) = ZelogPolyCof(RX, Rg)+4*m+2*n-2;
  ZelogPolyCof(up3, Rg) = ZelogPolyCof(RX, Rg)+2*m-1;
  ZelogPolySquareKaratsubaLazy(up3, up1, Rg);
  
  /* (PX- + PX+) ^ 2 - PX- ^ 2  -> RX+2*m-1 */
  for(j = 0, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZelogAnyEltSubtract(ZelogPolyGetCof(RX, i, Rg), ZelogPolyGetCof(RX, j, Rg), Rg);
  }
  
  /* [ PX- ^ 2, (PX- + PX+) ^ 2 - PX- ^ 2 ] -> RX+m */
  for(j = m, i = 2*m-1; j < 2*m-1; j++, i++) {
    ZelogAnyEltAdd(ZelogPolyGetCof(RX, j, Rg), ZelogPolyGetCof(RX, i, Rg), Rg);
  }
  for(;j < n+2*m-1; j++, i++) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, j, Rg), ZelogPolyGetCof(RX, i, Rg), Rg);
  }
  
  /* PX+ ^ 2 -> RX+ */
  ZelogPolyDeg(up1, Rg) = n-1;
  ZelogPolyCof(up1, Rg) = ZelogPolyCof(PX, Rg)+m;
  ZelogPolyCof(up3, Rg) = ZelogPolyCof(RX, Rg)+2*m+n-1;
  ZelogPolySquareKaratsubaLazy(up3, up1, Rg);

  /* [ PX- ^ 2, (PX- + PX+) ^ 2 - PX- ^2 - PX+ ^ 2 , PX+ ^ 2 ] -> RX */
  for(j = m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZelogAnyEltSubtract(ZelogPolyGetCof(RX, j, Rg), ZelogPolyGetCof(RX, i, Rg), Rg);
  }
  for(j = 2*m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZelogAnyEltAdd(ZelogPolyGetCof(RX, j, Rg), ZelogPolyGetCof(RX, i, Rg), Rg);
  }
  for(; j < 2*m+2*n-1; j++, i++) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, j, Rg), ZelogPolyGetCof(RX, i, Rg), Rg);
  }

  ZelogPolyDeg(RX, Rg) = 2*ZelogPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Lazy Squaring with Karatsuba.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0 for compatibylity.}
   \side{$|RX| = |PX|^2$.}
   \note{One must have $|RX|\neq|PX|$. |PX| is destroyed at the return.}
   }
   */
int ZelogPolySquareKaratsubaLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  if (ZelogPolyDeg(PX, Rg) < ZELOG_POLY_SQUARE_CUTOFF) {
    ZelogPolySquareCharp(RX, PX, Rg);
    return(ZEN_HAS_INVERSE);
  }
  ZelogPolySquareKaratsubaLazyDegree(RX, PX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|$.}
   }
   */
int _ZelogPolyMultiply(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZENPoly RBuf;
  int n, m, q, r, L, k;

  if (ZelogPolyDeg(PX, Rg) > ZelogPolyDeg(QX, Rg)) {
    n = ZelogPolyDeg(PX, Rg)+1; m = ZelogPolyDeg(QX, Rg)+1;
    q = n/m; r = n%m;
  }
  else {
    n = ZelogPolyDeg(QX, Rg)+1; m = ZelogPolyDeg(PX, Rg)+1;
    q = n/m; r = n%m;
  }
  switch(q) {
  case 1:
    L = r ? 4*m-4+2*r : 4*m-3;
    break;
  case 2:
    L = 2*r > (m+1) ? 5*m-5+2*r : 6*m-4;
    break;
  default:
    L = 2*r > m ?  (3+q)*m-4+2*r: (4+q)*m-4;
    break;
  }
  if (!(RBuf = ZelogPolyCreateLarge(L, Rg))) {
    return(ZERR);
  }
  ZelogPolyMultiplyLazy(RBuf, PX, QX, Rg);
  ZelogPolyAssign(RX, RBuf, Rg);
  ZelogPolyFree(RBuf, Rg);

  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Lazy multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0 for compatibylity.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|\neq|QX|$.}
   }
   */
int _ZelogPolyMultiplyLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  struct zenpoly sxzp, rxzp, pxzp, qxzp;
  ZENPoly sx, rx, px, qx;
  int i, dp, dq, quo, rem;

  /* Switch
  **********/
  if (ZelogPolyDeg(PX, Rg) == ZelogPolyDeg(QX, Rg)) {
    ZelogPolyMultiplyKaratsubaLazy(RX, PX, QX, Rg);
    return(ZEN_HAS_INVERSE);
  }
  if (ZelogPolyDeg(PX, Rg) > ZelogPolyDeg(QX, Rg)) {
    if ((ZelogPolyDeg(QX, Rg)) <= ZELOG_POLY_MULTIPLY_CUTOFF) {
      ZelogPolyMultiplyPlain(RX, PX, QX, Rg);
      return(ZEN_HAS_INVERSE);
    }
    px = &pxzp;
    ZelogPolyCof(px, Rg) = ZelogPolyCof(PX, Rg);
    dp = ZelogPolyDeg(PX, Rg);

    qx = &qxzp;
    ZelogPolyCof(qx, Rg) = ZelogPolyCof(QX, Rg);
    dq = ZelogPolyDeg(QX, Rg);
  }
  else {
    if (ZelogPolyDeg(PX, Rg) <= ZELOG_POLY_MULTIPLY_CUTOFF) {
      ZelogPolyMultiplyPlain(RX, PX, QX, Rg);
      return(ZEN_HAS_INVERSE);
    }
    px = &pxzp;
    ZelogPolyCof(px, Rg) = ZelogPolyCof(QX, Rg);
    dp = ZelogPolyDeg(QX, Rg);

    qx = &qxzp;
    ZelogPolyCof(qx, Rg) = ZelogPolyCof(PX, Rg);
    dq = ZelogPolyDeg(PX, Rg);
  }

  /* We have to cut px in pieces 
    **********************************/  
  /* Initialization */
  ZelogPolyDeg(RX, Rg) = dp+dq;

  rx = &rxzp;
  ZelogPolyCof(rx, Rg) = ZelogPolyCof(RX, Rg);

  sx = &sxzp;
  ZelogPolyCof(sx, Rg) = ZelogPolyCof(RX, Rg);

  quo = (dp+1)/(dq+1);
  rem = (dp+1)%(dq+1);

  /* First multiplication */
  ZelogPolyDeg(px, Rg) = dq;
  ZelogPolyDeg(qx, Rg) = dq;
  ZelogPolyMultiplyKaratsubaLazy(rx, qx, px, Rg);
  if ((quo == 1) && !rem) {
    return(ZEN_HAS_INVERSE);
  }

  /* Second multiplication */
  ZelogPolyCof(px, Rg) += dq+1;
  ZelogPolyDeg(px, Rg) = (quo == 1 ? rem-1: dq);
  ZelogPolyCof(rx, Rg) += 2*dq+1;
  if (quo == 1) {
    ZelogPolyMultiplyLazy(rx, qx, px, Rg);
  }
  else {
    ZelogPolyMultiplyKaratsubaLazy(rx, qx, px, Rg);
  }
  ZelogPolyDeg(sx, Rg) = dq-1;
  ZelogPolyCof(sx, Rg) += dq+1;
  ZelogPolyAdd(sx, rx, Rg);
  if (quo == 1) {
    return(ZEN_HAS_INVERSE);
  }

  for(i = 2; i < quo; i++) {
    ZelogPolyCof(px, Rg) += (dq+1);
    ZelogPolyCof(rx, Rg) += (dq+1);

    ZelogPolyMultiplyKaratsubaLazy(rx, qx, px, Rg);
    
    ZelogPolyDeg(sx, Rg) = dq-1;
    ZelogPolyCof(sx, Rg) += dq+1;
    ZelogPolyAdd(sx, rx, Rg);
  }

  if (!rem) {
    return(ZEN_HAS_INVERSE);
  }

  ZelogPolyDeg(px, Rg) = rem-1;
  ZelogPolyCof(px, Rg) += (dq+1);
  ZelogPolyCof(rx, Rg) += (dq+1);
  ZelogPolyMultiplyLazy(rx, qx, px, Rg);
    
  ZelogPolyDeg(sx, Rg) = dq-1;
  ZelogPolyCof(sx, Rg) += dq+1;
  ZelogPolyAdd(sx, rx, Rg);

  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|$.}
   }
   */
int ZelogPolyMultiplyPlain(RX, QX, PX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{

  int i, j;

  if (ZelogPolyIsZero(PX, Rg) || ZelogPolyIsZero(QX, Rg)) {
    ZelogPolySetToZero(RX, Rg);
    return(0);
  }

  for(i = ZelogPolyDeg(PX, Rg)+1;
      i <= ZelogPolyDeg(PX, Rg)+ZelogPolyDeg(QX, Rg);
      i++) {
    ZelogAnyEltSetToZero(ZelogPolyGetCof(RX, i, Rg), Rg);
  }

  for(i = ZelogPolyDeg(PX, Rg); i >= 0; i--) {
    for(j = ZelogPolyDeg(QX, Rg); j; j--) {
      ZelogAnyEltMultiply(ZelogElt2Int(ZENRingHbuf(Rg), Rg),
		      ZelogPolyGetCof(QX, j, Rg),
		      ZelogPolyGetCof(PX, i, Rg),
		      Rg);
      ZelogAnyEltAdd(ZelogPolyGetCof(RX, i+j, Rg), ZelogElt2Int(ZENRingHbuf(Rg), Rg), Rg);
    }
    ZelogAnyEltMultiply(ZelogElt2Int(ZENRingHbuf(Rg), Rg),
		    ZelogPolyGetCof(QX, j, Rg),
		    ZelogPolyGetCof(PX, i, Rg),
		    Rg);
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, i+j, Rg), ZelogElt2Int(ZENRingHbuf(Rg), Rg), Rg);
  }
  ZelogPolyDeg(RX, Rg) = ZelogPolyDeg(PX, Rg) + ZelogPolyDeg(QX, Rg);
  return(0);
}

/* \TEX{
   \title{Polynomial multiplication with the Karatsuba method.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an errpr occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|$. If $\deg(|PX|) \ge deg(|QX|)$,
   one polynomial buffer of size |PX| is allocated, otherwise 2
   buffers of size |QX| are allocated.}
   }
   */
int _ZelogPolyMultiplyKaratsuba(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i;
  ZENPoly QBuf, RBuf;
  ZENElt elt;

  if (ZelogPolyDeg(PX, Rg) >= ZelogPolyDeg(QX, Rg)) {
    if (!(QBuf = ZelogPolyCreate(ZelogPolyDeg(PX, Rg), Rg))) {
      return(ZERR);
    }
    ZelogPolyAssign(QBuf, QX, Rg);
    for(i = ZelogPolyDeg(QX, Rg)+1;
	i <= ZelogPolyDeg(PX, Rg);
	i++) {
      elt = ZelogPolyGetCoeffPtr(QBuf, i, Rg);
      ZelogEltSetToZero(elt, Rg);
      ZelogPolySetCoeff(QBuf, i, elt, Rg);
    }

    if (!(RBuf = ZelogPolyCreateLarge(4*ZelogPolyDeg(PX, Rg)+1, Rg))) {
      return(ZERR);
    }

    ZelogPolyMultiplyKaratsubaLazy(RBuf, PX, QBuf, Rg);
  }
  else {
    if (!(QBuf = ZelogPolyCreate(ZelogPolyDeg(QX, Rg), Rg))) {
      return(ZERR);
    }
    ZelogPolyAssign(QBuf, PX, Rg);
    for(i = ZelogPolyDeg(PX, Rg)+1;
	i <= ZelogPolyDeg(QX, Rg);
	i++) {
      elt = ZelogPolyGetCoeffPtr(QBuf, i, Rg);
      ZelogEltSetToZero(elt, Rg);
      ZelogPolySetCoeff(QBuf, i, elt, Rg);
    }

    if (!(RBuf = ZelogPolyCreateLarge(4*ZelogPolyDeg(PX, Rg)+1, Rg))) {
      return(ZERR);
    }

    ZelogPolyMultiplyKaratsubaLazy(RBuf, QX, QBuf, Rg);
  }
  ZelogPolyAssign(RX, RBuf, Rg);
  ZelogPolyFree(QBuf, Rg); ZelogPolyFree(RBuf, Rg);
  return(ZEN_HAS_INVERSE);
}


static void ZelogPolyMultiplyKaratsubaLazyDegree0(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZelogAnyEltMultiply(ZelogPolyGetCof(RX, 0, Rg),
		   ZelogPolyGetCof(PX, 0, Rg),
		   ZelogPolyGetCof(QX, 0, Rg), Rg);
  ZelogPolyDeg(RX, Rg) = 0;
}

static void ZelogPolyMultiplyKaratsubaLazyDegree1(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZelogAnyEltAssign(ZelogPolyGetCof(RX, 0, Rg), ZelogPolyGetCof(PX, 0, Rg), Rg);
  ZelogAnyEltAdd(ZelogPolyGetCof(RX, 0, Rg), ZelogPolyGetCof(PX, 1, Rg), Rg);
  
  ZelogAnyEltAssign(ZelogPolyGetCof(RX, 2, Rg), ZelogPolyGetCof(QX, 0, Rg), Rg);
  ZelogAnyEltAdd(ZelogPolyGetCof(RX, 2, Rg), ZelogPolyGetCof(QX, 1, Rg), Rg);

  ZelogAnyEltMultiply(ZelogPolyGetCof(RX, 1, Rg),
		   ZelogPolyGetCof(RX, 0, Rg),
		   ZelogPolyGetCof(RX, 2, Rg), Rg);
  
  ZelogAnyEltMultiply(ZelogPolyGetCof(RX, 0, Rg),
		   ZelogPolyGetCof(PX, 0, Rg),
		   ZelogPolyGetCof(QX, 0, Rg), Rg);
  
  ZelogAnyEltMultiply(ZelogPolyGetCof(RX, 2, Rg),
		   ZelogPolyGetCof(PX, 1, Rg),
		   ZelogPolyGetCof(QX, 1, Rg), Rg);
  
  ZelogAnyEltSubtract(ZelogPolyGetCof(RX, 1, Rg), ZelogPolyGetCof(RX, 0, Rg), Rg);
  ZelogAnyEltSubtract(ZelogPolyGetCof(RX, 1, Rg), ZelogPolyGetCof(RX, 2, Rg), Rg);

  ZelogPolyDeg(RX, Rg) = 2;
}

static void ZelogPolyMultiplyKaratsubaLazyHeart ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

static void ZelogPolyMultiplyKaratsubaLazyDegree(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  struct zenpoly UP1, UP2, UP3;
  ZENPoly up1, up2, up3;
  int m, n, i, j;
  
  /* Initialization ***********/
  up1 = &UP1; up2 = &UP2; up3 = &UP3;
  up2->deg = 0;
  m = (ZelogPolyDeg(PX, Rg)+1)/2; 
  n = ZelogPolyDeg(PX, Rg)+1-m;
  
  /* let's go with the recursivity **********/
  
  /* PX- * QX- -> RX */
  ZelogPolyDeg(up1, Rg) = m-1;
  ZelogPolyCof(up1, Rg) = ZelogPolyCof(PX, Rg);
  ZelogPolyCof(up2, Rg) = ZelogPolyCof(QX, Rg);
  ZelogPolyCof(up3, Rg) = ZelogPolyCof(RX, Rg);
  ZelogPolyMultiplyKaratsubaLazyHeart(up3, up1, up2, Rg);
  
  /* PX- + PX+ -> RX+4*m+2*n-2 */
  for(j = 0, i = 4*m+2*n-2;
      j < m;
      j++, i++) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, i, Rg),
		  ZelogPolyGetCof(PX, j+m, Rg),
		  Rg);
    ZelogAnyEltAdd(ZelogPolyGetCof(RX, i, Rg), ZelogPolyGetCof(PX, j, Rg), Rg);
  }
  if (n > m) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, i, Rg),
		  ZelogPolyGetCof(PX, j+m, Rg),
		  Rg);
  }
  
  /* QX- + QX+ -> RX+4*m+3*n-2 */
  for(j = 0, i = 4*m+3*n-2;
      j < m;
      j++, i++) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, i, Rg),
		  ZelogPolyGetCof(QX, j+m, Rg),
		  Rg);
    ZelogAnyEltAdd(ZelogPolyGetCof(RX, i, Rg), ZelogPolyGetCof(QX, j, Rg), Rg);
  }
  if (n > m) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, i, Rg),
		  ZelogPolyGetCof(QX, j+m, Rg),
		  Rg);
  }
  
  /* (PX- + PX+) * (QX- + QX+) -> RX+2*m-1 */
  ZelogPolyDeg(up1, Rg) = n-1;
  ZelogPolyCof(up1, Rg) = ZelogPolyCof(RX, Rg)+4*m+2*n-2;
  ZelogPolyCof(up2, Rg) = ZelogPolyCof(RX, Rg)+4*m+3*n-2;
  ZelogPolyCof(up3, Rg) = ZelogPolyCof(RX, Rg)+2*m-1;
  ZelogPolyMultiplyKaratsubaLazyHeart(up3, up1, up2, Rg);
  
  /* (PX- + PX+) * (QX- + QX+) - PX- * QX-  -> RX+2*m-1 */
  for(j = 0, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZelogAnyEltSubtract(ZelogPolyGetCof(RX, i, Rg), ZelogPolyGetCof(RX, j, Rg), Rg);
  }
  
  /* [ PX- * QX-, (PX- + PX+) * (QX- + QX+) - PX- * QX- ] -> RX+m */
  for(j = m, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZelogAnyEltAdd(ZelogPolyGetCof(RX, j, Rg), ZelogPolyGetCof(RX, i, Rg), Rg);
  }
  for(;j < n+2*m-1; j++, i++) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, j, Rg), ZelogPolyGetCof(RX, i, Rg), Rg);
  }
  
  /* PX+ * QX+ -> RX+ */
  ZelogPolyDeg(up1, Rg) = n-1;
  ZelogPolyCof(up1, Rg) = ZelogPolyCof(PX, Rg)+m;
  ZelogPolyCof(up2, Rg) = ZelogPolyCof(QX, Rg)+m;
  ZelogPolyCof(up3, Rg) = ZelogPolyCof(RX, Rg)+2*m+n-1;
  ZelogPolyMultiplyKaratsubaLazyHeart(up3, up1, up2, Rg);

  /* [ PX- * QX-, (PX- + PX+) * (QX- + QX+) - PX- * QX- - PX+ * QX+ , PX+ * QX+ ] -> RX */
  for(j = m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZelogAnyEltSubtract(ZelogPolyGetCof(RX, j, Rg), ZelogPolyGetCof(RX, i, Rg), Rg);
  }
  for(j = 2*m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZelogAnyEltAdd(ZelogPolyGetCof(RX, j, Rg), ZelogPolyGetCof(RX, i, Rg), Rg);
  }
  for(; j < 2*m+2*n-1; j++, i++) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, j, Rg), ZelogPolyGetCof(RX, i, Rg), Rg);
  }

  ZelogPolyDeg(RX, Rg) = ZelogPolyDeg(PX, Rg)+ZelogPolyDeg(QX, Rg);
}

static void ZelogPolyMultiplyKaratsubaLazyHeart(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  switch (ZelogPolyDeg(PX, Rg)) {

    /* compute the product of 2 polynoms whose degree is 0 */
  case 0:
    ZelogPolyMultiplyKaratsubaLazyDegree0(RX, PX, QX, Rg);
    break;

    /* compute the product of 2 polynoms whose degree is 1 */
  case 1: 
    ZelogPolyMultiplyKaratsubaLazyDegree1(RX, PX, QX, Rg); 
    break;

    /* compute the product of 2 polynoms whose degree is greater than 2*/
  default : 
    ZelogPolyMultiplyKaratsubaLazyDegree(RX, PX, QX, Rg); 
  }
}

int _ZelogPolyMultiplyKaratsubaLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  if (ZelogPolyDeg(PX, Rg) >= ZelogPolyDeg(QX, Rg)) {
    for(i = ZelogPolyDeg(QX, Rg)+1;
	i <= ZelogPolyDeg(PX, Rg);
	i++) {
      elt = ZelogPolyGetCoeffPtr(QX, i, Rg);
      ZelogEltSetToZero(elt, Rg);
      ZelogPolySetCoeff(QX, i, elt, Rg);
    }
  }
  else {
    for(i = ZelogPolyDeg(PX, Rg)+1;
	i <= ZelogPolyDeg(QX, Rg);
	i++) {
      elt = ZelogPolyGetCoeffPtr(PX, i, Rg);
      ZelogEltSetToZero(elt, Rg);
      ZelogPolySetCoeff(PX, i, elt, Rg);
    }
  }
  ZelogPolyMultiplyKaratsubaLazyHeart(RX, PX, QX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Multiplication by a scalar.}
   \input{Two polynomials |PX|, |RX| and a |ZENElt| |e| of a finite ring |Rg|.}
   }
   */
void _ZelogPolyMultiplyScalar(RX, PX, e, Rg)
     ZENPoly RX, PX;
     ZENElt e;
     ZENRing Rg;
{
  int i;
  
  if (ZelogEltIsZero(e, Rg)) {
    ZelogPolySetToZero(RX, Rg);
    return;
  }
  if (ZelogEltIsOne(e, Rg)) {
    ZelogPolyAssign(RX, PX, Rg);
    return;
  }
  for(i = 0; i <= ZelogPolyDeg(PX, Rg); i++) {
    ZelogAnyEltMultiply(ZelogPolyGetCof(RX, i, Rg),
		    ZelogElt2Int(e, Rg),
		    ZelogPolyGetCof(PX, i, Rg),
		    Rg);
  }
  ZENPolyDeg(RX, Rg) = ZENPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Scalar product.}
   \input{Two polynomials |PX|, |RX| and a pointer to a |ZENElt| |e| of a
   finite ring |Rg|.}
   \side{$|*e| = \sum p_iq_i$}
   }
   */
void _ZelogPolyDot(e, PX, QX, Rg)
     ZENElt *e;
     ZENPoly PX, QX;
     ZENRing Rg;
{
  int i, deg;
  
  ZelogAnyEltMultiply(ZelogElt2Int(*e, Rg),
		  ZelogPolyGetCof(PX, 0, Rg),
		  ZelogPolyGetCof(QX, 0, Rg),
		  Rg);

  deg = ZelogPolyDeg(PX, Rg) < ZelogPolyDeg(QX, Rg) ?
    ZelogPolyDeg(PX, Rg) :  ZelogPolyDeg(QX, Rg);

  for(i = 1; i <= deg; i++) {
    ZelogAnyEltMultiply(ZelogElt2Int(ZENRingHbuf(Rg), Rg),
		    ZelogPolyGetCof(PX, i, Rg),
		    ZelogPolyGetCof(QX, i, Rg),
		    Rg);
    ZelogEltAdd(*e, ZENRingHbuf(Rg), Rg);
  }
}

/* \title{Divide.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured,  |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   \note{|PX| is destroyed.}
   */
int _ZelogPolyDivideLazy(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int i, j, deg, IsNotOne;

  if (!ZelogAnyEltIsOne(ZelogPolyGetCof(QX, ZelogPolyDeg(QX, Rg), Rg), Rg)) {
    IsNotOne = 1;
    
    if (ZelogAnyEltInverse(ZelogElt2Int(ZENRingIbuf(Rg), Rg),
		       ZelogPolyGetCof(QX, ZelogPolyDeg(QX, Rg), Rg),
		       Rg) == ZEN_NO_INVERSE) {
      return(ZEN_NO_INVERSE);
    }
  }
  else {
    IsNotOne = 0;
  }
  
  for(i = ZelogPolyDeg(PX, Rg);
      i >= ZelogPolyDeg(QX, Rg);
      i--) {
    if (IsNotOne) {
      ZelogAnyEltMultiply(ZelogPolyGetCof(PX, i, Rg),
		      ZelogElt2Int(ZENRingIbuf(Rg), Rg), ZelogPolyGetCof(PX, i, Rg),
		      Rg);
    }
    
    for(j =  ZelogPolyDeg(QX, Rg)-1;
	j >= 0;
	j--) {
      ZelogAnyEltMultiply(ZelogElt2Int(ZENRingHbuf(Rg), Rg),
		      ZelogPolyGetCof(PX, i, Rg),
		      ZelogPolyGetCof(QX, j, Rg),
		      Rg);
      ZelogAnyEltSubtract(ZelogPolyGetCof(PX,
				     i+j-ZelogPolyDeg(QX, Rg),
				     Rg), ZelogElt2Int(ZENRingHbuf(Rg), Rg), Rg);
    }
  }
  deg =  ZelogPolyDeg(QX, Rg);
  for(i = ZelogPolyDeg(PX, Rg); i >= ZelogPolyDeg(QX, Rg); i--) {
    ZelogAnyEltAssign(ZelogPolyGetCof(RX, i-ZelogPolyDeg(QX, Rg), Rg),
		   ZelogPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZelogPolyDeg(RX, Rg) = ZelogPolyDeg(PX, Rg)-ZelogPolyDeg(QX, Rg);
  if (ZelogPolyDeg(RX, Rg) < 0) {
    ZelogPolySetToZero(RX, Rg);
  }
  for(; i >= 0; i--) {
    if (!ZelogAnyEltIsZero(ZelogPolyGetCof(PX, i, Rg), Rg)) {
      break;
    }
  }
  for(deg = i; i >= 0; i--) {
    ZelogAnyEltAssign(ZelogPolyGetCof(MX, i, Rg),
		   ZelogPolyGetCof(PX, i, Rg),
		   Rg);
  }
  if (deg < 0) {
    ZelogPolySetToZero(MX, Rg);
  }
  else {
    ZelogPolyDeg(MX, Rg) = deg;
  }
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Divide.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZEN_NO_INVERSE| if
   a factor of the modulo was dicovered, |ZENERR| for an error.}
   \side{ |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZelogPolyDivide(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int ret;
  ZENPoly Pbuf;
  
  if (!(Pbuf = ZelogPolyCopy(PX, Rg))) {
    return(ZERR);
  }
  ret = ZelogPolyDivideLazy(RX, MX, Pbuf, QX, Rg);
  ZelogPolyFree(Pbuf, Rg);
  return(ret);
}


/* \TEX{
   \title{Lazy mudulo computation.}
   \input{A polynomial |PX| to divide by
   |QX|$=X^{\mbox{degree}(QX)}+sum_{i=0}^{|k|}q_iX^i$} in a finite
   ring |Rg|. |MX| will be the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| for an error}
   \note{|PX| is destroyed.}
   */
int _ZelogPolyModuloLazy(MX, PX, QX, k, Rg)
     ZENPoly MX, PX, QX;
     int k;
     ZENRing Rg;
{
  int i, j, deg;
  
  for(i = ZelogPolyDeg(PX, Rg);
      i >= ZelogPolyDeg(QX, Rg);
      i--) {
    for(j =  k;
	j >= 0;
	j--) {
      ZelogAnyEltMultiply(ZelogElt2Int(ZENRingHbuf(Rg), Rg),
		      ZelogPolyGetCof(PX, i, Rg),
		      ZelogPolyGetCof(QX, j, Rg),
		      Rg);
      ZelogAnyEltSubtract(ZelogPolyGetCof(PX,
				     i+j-ZelogPolyDeg(QX, Rg),
				     Rg), ZelogElt2Int(ZENRingHbuf(Rg), Rg), Rg);
    }
  }
  for( ; i >= 0; i--) {
    if (!ZelogAnyEltIsZero(ZelogPolyGetCof(PX, i, Rg), Rg)) {
      break;
    }
  }
  for(deg = i; i >= 0; i--) {
    ZelogPolyGetCof(MX, i, Rg) = ZelogPolyGetCof(PX, i, Rg);
  }
  if (deg < 0) {
    ZelogPolySetToZero(MX, Rg);
  }
  else {
    ZelogPolyDeg(MX, Rg) = deg;
  }

  return(ZEN_HAS_INVERSE);
}
