/*
 * zen/zem/c/elt_exp.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsubsection{Element exponentiation}
   This section is inspired from the file |exp.c| of the BigMod library, written by F.~Morain. It introduces the same functions as the |elt_exp.c| file of the |zep| arithmetic.
   }
   */
#include <stdio.h>
#include <zem.h>

#define EXP_LIMIT(x) divSizeBloc(BlocExp2(x)*x)
#define MARYMAX         7
#define TWOPOWERMARYMAX 128 /* 2**MARYMAX */

extern int _ZENPrcInitEltExp ___((ZENRing));
extern int _ZENPrcCloseEltExp ___((ZENRing));

/* \TEX[
   \title{Exponentiation.}
   \input{A pointer on an element |R| and an element |P| of a finite ring |Rg|, a bignum (|k|, |kl|).}
   \output{0 (no error can occur).}
   \side{$|*R| = |P|^{(|k|, |kl|)}$.}
   \note{One must have $|*(k+kl-1)| \neq 0$.}
   }
   */
int _ZemEltExpPlain (R, k, kl, P, Rg)
     ZENElt *R, P;
     BigNum k;
     BigNumLength kl;
     ZENRing Rg;
{
  BigNumDigit C[1];
  int i, kd, ks;
  
  if ((ZBNIsZero(k, kl)) ||
      (ZemEltIsOne(P,Rg))) {
    ZemEltSetToOne(*R, Rg);
    return(0);
  }
  if (ZBNIsOne(k, kl)) {
    ZemEltAssign(*R, P, Rg);
    return(0);
  }
  if(ZemEltIsZero(P,Rg)) {
    ZemEltSetToZero(*R, Rg);
    return(0);
  }
  
  ZemEltAssign(ZENRingIbuf(Rg), P, Rg);
  kd = kl - 1;
  /* Starting the computation */
  ZBNAssign(C, (k+kd), 1);

  /* Looking for the most significant bit */
  for(ks = BN_DIGIT_SIZE; ks; ks--){
    if(ZBNShiftLeft(C, 1, 1))
      break;
  }
  ks--;
  while(kl >= 1){
    for(i = 0; i < ks; i++){
      ZemEltSquare(ZENRingIbuf(Rg), ZENRingIbuf(Rg), Rg);
      if(ZBNShiftLeft(C, 1, 1)){
	ZemEltMultiply(ZENRingIbuf(Rg), P, ZENRingIbuf(Rg), Rg);
      }
    }
    kl--;
    if(kl){
      kd--; ks = BN_DIGIT_SIZE;
      ZBNAssign(C, (k+kd), 1);
    }
  }
  ZemEltAssign(*R, ZENRingIbuf(Rg), Rg);
  return(0);
}
/********** Blocks **********/

/* After the article by H. Cohen and A. K. Lenstra: 
   "Implementation of a new primality test",
   Math. of Comp., volume 48, number 177, January 1987, pp. 103-121.*/

/* \TEX[
   \title{Adding precomputations.}
   \input{A finite ring |R|.} 
   \output{|ZENERR| if an allocation error occured, 0 otherwise.}
   \side{Buffers are allocated and initialized for |_ZENEltExpBlock|.}
   }
   */
int _ZemPrcInitEltExp(R)
     ZENRing R;
{
 return(__ZENPrcInitEltExp(_ZemEltExp,_ZemEltExpMod,_ZemEltExpMod4,_ZemEltExpMod5,_ZemEltExpMod6,R));
}

/* \TEX[
   \title{Freeing precomputations.}
   \input{A finite ring |R|.} 
   \side{Buffers for |_ZENEltExpBlock| are freed.}
   }
   */
int _ZemPrcCloseEltExp(R)
     ZENRing R;
{
  return(__ZENPrcCloseEltExp(_ZemEltExpPlain,R));
}
/* ---------------------------------------------------------------------------
 .SSection "Block Exponentiation"
---------------------------------------------------------------------------*/

/* Computes p=a**e mod n. Write: e=e[t]*(2**m)**t + ... + e[0] and */
 /* e[i]=(2**li)*ui, e[t] <> 0.*/
/*
   \TEX[
   \title{Exponentiation precomputations}
   \input{A table of exponentiation precomputations to fill, a |ZENElt|, the size of table and a |ZENRing|.}
   \side{Exponentiation precomputations are done}
   \note{This procedure is externally defined to optimize exponentiations with the same |ZENElt| but several exponents.}
   } 
   */

static void _ZemEltExpPrc(OddPower,a,twopowermminus1,R)
     ZENElt *OddPower,a;
     int twopowermminus1;
     ZENRing R;
{int i;

 ZemEltSquare(ZENRingIbuf(R),a,R);
 ZemEltAssign(OddPower[0],a,R);

 for(i = 1; i < twopowermminus1; i++)
   ZemEltMultiply(OddPower[i],
		  ZENRingIbuf(R),OddPower[i-1],R);

}

/*
   \TEX[
   \title{Block exponentiation}
   \input{A pointer on a |ZENElt|, an exponent |(e,el)|, a |ZENElt|, a |ZENRing|, the length of block to use, the pre-computations on the exponent, and the mask on exponent.}
   \side{$|*y| = |a|^{(|e|, |el|)}$.}
   \note{This procedure is externally defined to optimize exponentiations with the same |ZENElt| but several exponents.}
   } 
   */
static void _ZemEltExpBlock2(y, e, el, a, R, lbloc,OddPower,mask)
     ZENElt *y,a,*OddPower;
     BigNumDigit mask;
     BigNum e;
     ZENRing R;
     BigNumLength el;
     int lbloc;
{BigNumDigit C[2];
 int i,nzeros,nbits,qs,rs,ds,ms,et;
 
 nzeros=ZBNNumLeadingZeroBitsInDigit(e[el-1]);
 nbits=el*BN_DIGIT_SIZE -nzeros;

 qs=nbits/lbloc;
 rs=nbits%lbloc;
 if(!rs)
   {
     qs--;
     rs=lbloc;
   }
 ds=qs*lbloc;
 ms=modSizeBloc(ds);
 ds=divSizeBloc(ds);

 if(ms+rs>BN_DIGIT_SIZE)
   {
     ZBNAssign(C,e+ds,2);
     ZBNShiftLeft(C,2,BN_DIGIT_SIZE-ms);
     et=C[1];
   }
 else
   {
     if(ms)
       {
	 ZBNAssign(C,e+ds,1);
	 et=ZBNShiftLeft(C,1,BN_DIGIT_SIZE-ms);
       }
     else
       {
	 C[0]=0;
	 et=e[ds];
       }
   }

 ZemEltAssign(ZENRingIbuf(R),
	      OddPower[ZENRingETU(R,lbloc,et)],
	      R);
 
 /* Et <- Et**(2**lt) */
 for(i = ZENRingETL(R,lbloc,et); i ; i--)
   ZemEltSquare(ZENRingIbuf(R),
		ZENRingIbuf(R),
		R);
 /* loop */
 while(qs>0)
   {
     /* get current digit e[i]=(2**li)*ui */
     qs--;
     ms -= lbloc;
     if(ms<0)
       {
	 ds--;
	 ZBNAssign(C,e+ds,2);
	 ZBNShiftLeft(C,2,-ms);
	 et=C[1]&mask;
	 ms += BN_DIGIT_SIZE;
       }
     else
       et=ZBNShiftLeft(C,1,lbloc);
     
     for(i = lbloc-ZENRingETL(R,lbloc,et); i ; i--)
       ZemEltSquare(ZENRingIbuf(R),
		    ZENRingIbuf(R),
		    R);
     
     if(et)
       {
	 /* Et <- Et*a**vt */
	 ZemEltMultiply(ZENRingIbuf(R),
			OddPower[ZENRingETU(R,lbloc,et)],
			ZENRingIbuf(R),
			R);
	 /* Et <- Et**(2**lt) */
	 for(i = ZENRingETL(R,lbloc,et); i; i--)
	   ZemEltSquare(ZENRingIbuf(R),
			ZENRingIbuf(R),
			R);

       }
   }
 
 ZemEltAssign(*y,ZENRingIbuf(R),R);
}


static void _ZemEltExpBlock(y, e, el, a, R, lbloc)
     ZENElt *y,a;
     BigNum e;
     ZENRing R;
     BigNumLength el;
     int lbloc;
{BigNumDigit mask;
 int twopowermminus1;
 
 if ((ZBNIsZero(e, el)) ||
     (ZemEltIsOne(a,R))) {
   ZemEltSetToOne(*y, R);
   return;
 }
 
 if (ZBNIsOne(e, el)) {
   ZemEltAssign(*y, a, R);
   return;
 }

 if(ZemEltIsZero(a,R)) {
   ZemEltSetToZero(*y, R);
   return;
 }

 mask=(1<<lbloc)-1;
 twopowermminus1=(1<<(lbloc-1));
 _ZemEltExpPrc(ZENRingOddPower(R),a,twopowermminus1,R);
 _ZemEltExpBlock2(y,e,el,a,R,lbloc,ZENRingOddPower(R),mask);
}

/****************/

/* \TEX[
   \title{Exponentiation.}
   \input{A pointer on an element |R| and an element |P| of a finite ring |Rg|, a bignum (|k|, |kl|).}
   \side{$|*R| = |P|^{(|k|, |kl|)}$.}
   \output{0 (no error can occur).}
   \note{One must have $|*(k+kl-1)| \neq 0$. This procedure use the best algorithm according to the size of the exponent.}
   }
   */
int _ZemEltExp (R, e, el, P, Rg)
     ZENElt *R, P;
     BigNum e;
     BigNumLength el;
     ZENRing Rg;
{
  if(el <= EXP_LIMIT(4))
    _ZemEltExpPlain(R, e, el, P, Rg);
  else{
    if(el <= EXP_LIMIT(5))
      _ZemEltExpBlock(R, e, el, P, Rg, 4);
    else{
      if(el <=  EXP_LIMIT(6))
	_ZemEltExpBlock(R, e, el, P, Rg, 5);
      else
	_ZemEltExpBlock(R, e, el, P, Rg, 6);
    }
  }
  return(0);
}

/* \TEX[
   \title{Exponentiation.}
   \input{A pointer on an element |R| and an element |P| of a finite ring |Rg|, a bignum (|k|, |kl|).}
   \side{$|*R| = |P|^{(|k|, |kl|)}$.}
   \output{|ZENERR| if an allocation error occurs, 0 otherwise.}
   \note{One must have $|*(k+kl-1)| \neq 0$. The exponent is first reduced modulo $q-1$, with $q$ the cardianlity of |R|. This procedure use the best algorithm according to the size of the exponent.}
   }
   */
int _ZemEltExpMod (R, e, el, P, Rg)
     ZENElt *R, P;
     BigNum e;
     BigNumLength el;
     ZENRing Rg;
{
  BigNum p;

  switch(ZBNCompareLazy(e,el,ZENRingP1(Rg),ZENRingSizeElt(Rg)))
    {
    case BN_EQ:
      if(ZemEltIsZero(P,Rg))
	ZemEltSetToZero(*R,Rg);
      else
	ZemEltSetToOne(*R,Rg);
      return(0);
    case BN_LT:
      p=e;
      e=NULL;
      break;
    case BN_GT:
      p=ZBNC(el+1);
      if(!p)
       {
	 ZENSetError(ZEN_LIB_ZEM,ZEN_EXP,ZEN_ALLOC);
	 return(ZERR);
       }
      ZBNAssign(p,e,el);
      p[el]=0;
      ZBNDivide(p,el+1,ZENRingP1(Rg),ZENRingSizeElt(Rg));
      el=ZBNNumDigits(p,ZENRingSizeElt(Rg));
      break;
    }

  if(el <= EXP_LIMIT(4))
    _ZemEltExpPlain(R, p, el, P, Rg);
  else{
    if(el <= EXP_LIMIT(5))
      _ZemEltExpBlock(R, p, el, P, Rg, 4);
    else{
      if(el <=  EXP_LIMIT(6))
        _ZemEltExpBlock(R, p, el, P, Rg, 5);
      else
        _ZemEltExpBlock(R, p, el, P, Rg, 6);
    }
  }
  if(e)
    ZBNF(p);
  return(0);
}
				/* The following are faster, but no need to
				   document them */
int _ZemEltExpMod4(R, e, el, P, Rg)
     ZENElt *R, P;
     BigNum e;
     BigNumLength el;
     ZENRing Rg;
{
  BigNum p;

  el=ZBNModuloLazy(&p,e,el,ZENRingP1(Rg),ZENRingSizeElt(Rg));
  if(!el)
    return(ZERR);
  _ZemEltExpPlain(R, p, el, P, Rg);
  ZBNF(p);
  return(0);
}

int _ZemEltExpMod5(R, e, el, P, Rg)
     ZENElt *R, P;
     BigNum e;
     BigNumLength el;
     ZENRing Rg;
{
  BigNum p;

  el=ZBNModuloLazy(&p,e,el,ZENRingP1(Rg),ZENRingSizeElt(Rg));
  if(!el)
    return(ZERR);
  if(el <= EXP_LIMIT(4))
    _ZemEltExpPlain(R, p, el, P, Rg);
  else
    _ZemEltExpBlock(R, p, el, P, Rg, 4);
  
  ZBNF(p);
  return(0);
}

int _ZemEltExpMod6(R, e, el, P, Rg)
     ZENElt *R, P;
     BigNum e;
     BigNumLength el;
     ZENRing Rg;
{
  BigNum p;

  el=ZBNModuloLazy(&p,e,el,ZENRingP1(Rg),ZENRingSizeElt(Rg));
  if(!el)
    return(ZERR);
  if(el <= EXP_LIMIT(4))
    _ZemEltExpPlain(R, p, el, P, Rg);
  else{
    if(el <= EXP_LIMIT(5))
      _ZemEltExpBlock(R, p, el, P, Rg, 4);
    else
      _ZemEltExpBlock(R, p, el, P, Rg, 5);
  }
  ZBNF(p);
  return(0);
}
