/*
 * zen/zem/c/elt_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element multiplication}
   
   Two functions are provided to do multiplications: |_ZemEltSquare| to square a
   |ZENElt| and |_ZemEltMultily| to compute the multiplication of two |ZENElt|.
   }
   */
#include <stdio.h>
#include "zem.h"

/* \TEX[
   \title{Squaring.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{$|*p_c| = a^2$.}
   \note{One can have |*p_c| = |a|.}
   }
   */
void _ZemEltSquare(p_c, a, R)
     ZENElt *p_c, a;
     ZENRing R;
{
  _ZemModSquare(ZemRingMbuf(R),
		     ZENElt2Bn(a,R),ZENRingSizeElt(R),
		     R);
  ZBNAssign(ZENElt2Bn(*p_c,R),ZemRingMbuf(R),ZENRingSizeElt(R));
}

/* \TEX[
   \title{Multiplication.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |R|.}
   \side{$|*p_c| = a\times b$.}
   \note{One can have |*p_c| = |b|.}
   }
   */
void _ZemEltMultiply(p_c, a, b, R)
     ZENElt *p_c, a, b;
     ZENRing R;
{
  _ZemModMultiply(ZemRingMbuf(R),
		 ZENElt2Bn(a,R),ZENRingSizeElt(R),
		 ZENElt2Bn(b,R),ZENRingSizeElt(R),
		 R);
  ZBNAssign(ZENElt2Bn(*p_c,R),ZemRingMbuf(R),ZENRingSizeElt(R));
}


/*********************************************************
  \TEX[
  \title{Adding precomputations.}
  \input{A finite ring |R| and the cutoffs.} 
  \output{|ZENERR| if an allocation error occured, 0 otherwise.}
  \side{Buffers are allocated and initialized for Karatsuba's operations.}
  \note{No test is performed}
  }
  */
static int __ZemPrcInitEltKara(R,cutM,cutS)
     ZENRing R;
     BigNumLength cutM,cutS;
{
  ZemRingKaraMLim(R) = cutM;

  if(!(ZemRingBn3(R)=ZBNC(ZemRingBnl(R)))) {
    ZENSetError(ZEN_LIB_ZEM, ZEN_KARAPRC, ZEN_ALLOC);
    return(ZERR);
  }
  if(ZemRingKaraMLim(R)>0)
    {
      ZemRingKaraBuf(R)=ZBNC(ZBNSizeBufKaraM(ZENRingSizeElt(R)));
      if(!ZemRingKaraBuf(R))
	{
	  ZENSetError(ZEN_LIB_ZEM, ZEN_KARAPRC, ZEN_ALLOC);
	  return(ZERR);
	}
      ZENEltMultiplyPtr(R) = _ZemEltMultiplyKaratsuba; /* Multiplication of elements */
    }
  else
    ZemRingKaraBuf(R)=NULL;
  
  ZemRingKaraSLim(R) = cutS;

  if(ZemRingKaraSLim(R)>0)
    {
      if(!ZemRingKaraBuf(R))
	{
	  ZemRingKaraBuf(R)=ZBNC(ZBNSizeBufKaraS(ZENRingSizeElt(R)));
	  if(!ZemRingKaraBuf(R))
	    {
	      ZENSetError(ZEN_LIB_ZEM, ZEN_KARAPRC, ZEN_ALLOC);
	      return(ZERR);
	    }
	}
      ZENEltSquarePtr(R) = _ZemEltSquareKaratsuba; /* Squaring of elements */
    }


  return(0);

}

/* \TEX[
   \title{Adding precomputations.}
   \input{A finite ring |R|.} 
   \output{|ZENERR| if an allocation error occured, 0 otherwise.}
   \side{Buffers are allocated and initialized for Karatsuba's operations.}
   }
   */
int _ZemPrcInitEltKara(R)
     ZENRing R;
{
  return(__ZemPrcInitEltKara(R,
			     ZBNKaratsubaMultiplyCutoff(ZENRingSizeElt(R)),
			     ZBNKaratsubaSquareCutoff(ZENRingSizeElt(R))));
}

/* \TEX[
   \title{Freeing precomputations.}
   \input{A finite ring |R|.} 
   \side{Buffers for Karatsuba's operations are freed.}
   }
   */
int _ZemPrcCloseEltKara(R)
     ZENRing R;
{
  ZBNF(ZemRingBn3(R));
  
  if((ZemRingKaraMLim(R)>0) ||
     (ZemRingKaraSLim(R)>0))
    ZBNF(ZemRingKaraBuf(R));

  ZENEltMultiplyPtr(R) = _ZemEltMultiply; /* Multiplication of elements */
  ZENEltSquarePtr(R) = _ZemEltSquare;      /* Squaring of elements */
  return(0);
}

/* \TEX[
   \title{Squaring.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{$|*p_c| = a^2$.}
   \note{One can have |*p_c| = |a|.}
   }
   */
void _ZemEltSquareKaratsuba(p_c, a, R)
     ZENElt *p_c, a;
     ZENRing R;
{
 
  _ZemModSquareKaratsubaBuffer(ZemRingMbuf(R),
			       ZENElt2Bn(a,R),ZENRingSizeElt(R),
			       ZemRingKaraBuf(R),
			       R);
  ZBNAssign(ZENElt2Bn(*p_c,R),ZemRingMbuf(R),ZENRingSizeElt(R));
}

/* \TEX[
   \title{Multiplication.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |R|.}
   \side{$|*p_c| = a\times b$.}
   \note{One can have |*p_c| = |b|.}
   }
   */
void _ZemEltMultiplyKaratsuba(p_c, a, b, R)
     ZENElt *p_c, a, b;
     ZENRing R;
{
  _ZemModMultiplyKaratsubaBuffer(ZemRingMbuf(R),
				 ZENElt2Bn(a,R),ZENRingSizeElt(R),
				 ZENElt2Bn(b,R),ZENRingSizeElt(R),
				 ZemRingKaraBuf(R),			
				 R);
  ZBNAssign(ZENElt2Bn(*p_c,R),ZemRingMbuf(R),ZENRingSizeElt(R));
}

