/*
 * zen/zem/c/init.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   A finite ring is completely described by a variable of type
   |ZENRing|. Precomputations are needed to speedup the operations in this
   ring. In the particular case of $ZEN(p)$, this is done with
   |_ZemInit|. These precomutations are then free with |_ZemClose|.
   }
   */
#include <stdio.h>
#include "zem.h"
#include "zeg.h"
#include "zed.h"


/* \TEX[
   \title{Initialization.}
   \input{A partially defined |ZENRing|.}
   \output{|ZENERR| if an error occured, $|SR_GEN|\vert |MAT_GEN|$ otherwise.}
   \side{The initialization of $\Z/p\Z$ is completed.}
   \note{This procedure is called by |ZENBaseRingCreate| and should not be called otherwise.}
   }
   */
int _ZemInit(R)
     ZENRing R;
{
  ZENRing Rg;
  ZENElt V;
  int ret;
  
  ZENRingArith(R) = ZEN_LIB_ZEM;

  ZemRingMask(R) = BlocExp2(SIZE_BLOC-ZBNNumLeadingZeroBitsInDigit(ZENRingP1(R)[ZENRingSizeElt(R)-1]))-1;

  /* Procedure on rings
   **********************/
  ZENPrcInitTracePtr(R) = NULL;
  ZENPrcCloseTracePtr(R) = NULL;

#if KARA
  ZENPrcInitEltMultiplyPtr(R) = _ZemPrcInitEltKara;
  ZENPrcCloseEltMultiplyPtr(R) = _ZemPrcCloseEltKara;
#else
  ZENPrcInitEltMultiplyPtr(R) = NULL;
  ZENPrcCloseEltMultiplyPtr(R) = NULL;
#endif

  ZENPrcInitPolyRootsCanonicalPtr(R) = _ZemPrcInitPolyRootsCanonical;
  ZENPrcClosePolyRootsCanonicalPtr(R) = _ZemPrcClosePolyRootsCanonical;

  ZENPrcInitEltExpPtr(R) = _ZemPrcInitEltExp;
  ZENPrcCloseEltExpPtr(R) = _ZemPrcCloseEltExp;

  /* Procedure on elements
   ****************************/
  /* Allocation */
  ZENEltCreatePtr(R) = _ZepEltCreate;	/* Create elements */
  ZENEltCreateLargePtr(R) = _ZepEltCreateLarge;	/* Create elements */
  ZENEltFreePtr(R) = _ZepEltFree;	/* Free elements */

  /* Assigning */
  ZENEltAssignPtr(R) = _ZepEltAssign;	/* Assign an element with another */
  ZENEltSetToZeroPtr(R) = _ZepEltSetToZero; /* Set an element to zero */
  ZENEltSetToOnePtr(R) = _ZemEltSetToOne; /* Set an element to one */
  ZENEltSetRandomPtr(R) = _ZepEltSetRandom; /* Set randomly an element */
  ZENEltSetNextPtr(R) = _ZemEltSetNext; /* Set an element to the next */

  ZENEltToClonePtr(R) = _ZemEltToClone;
  ZENCloneToEltPtr(R) = _ZemCloneToElt;

  /* Test */
  ZENEltAreEqualPtr(R) = _ZepEltAreEqual; /* Is an element equal to another */
  ZENEltIsZeroPtr(R) = _ZepEltIsZero;	/* Is an element equal to zero */
  ZENEltIsOnePtr(R) = _ZemEltIsOne;	/* Is an element equal to one */
  ZENEltIsASquarePtr(R) = _ZENEltIsASquareCharp; /* Is an element a square ? */

  /* Arithmetic */
  ZENEltAddPtr(R) = _ZepEltAdd;	/* Addition of elements */
  ZENEltNegatePtr(R) = _ZepEltNegate;	/* Negation of an element */
  ZENEltSubtractPtr(R) = _ZepEltSubtract; /* Substraction of elements */
  ZENEltSquarePtr(R) = _ZemEltSquare;	/* Squaring of elements */
  ZENEltMultiplyPtr(R) = _ZemEltMultiply; /* Multiplication of elements */
  ZENEltInversePtr(R) = _ZemEltInverse; /* Inverse of elements */
  ZENEltTracePtr(R) = _ZepEltTrace;		/* Trace of elements */
  ZENEltSquareRootPtr(R) = _ZENEltSquareRootCharp; /* Square root of an element */
  ZENEltExpPtr(R) = _ZENEltExpPlain;	/* Exponentiation */

  /* io */
  ZENEltReadFromStringPtr(R) = _ZENCloneEltReadFromString;	/* Read an element from a string */
  ZENEltPrintToStringPtr(R) = _ZENCloneEltPrintToString; /* Print an element to a string */
  ZENEltReadFromFilePtr(R) = _ZENCloneEltReadFromFile; /* Read an element from a file */
  ZENEltPrintToFilePtr(R) = _ZENCloneEltPrintToFile; /* Print an element to a file */
  ZENEltGetFromStringPtr(R) = _ZepEltGetFromString; /* Get an element from a string */
  ZENEltPutToStringPtr(R) = _ZepEltPutToString; /* Put an element to a string */
  ZENEltGetFromFilePtr(R) = _ZepEltGetFromFile; /* Get an element from a file */
  ZENEltPutToFilePtr(R) = _ZepEltPutToFile; /* Put an element to a file */
  
  ZENZToEltPtr(R) = _ZemZToElt; /* Evaluate in $p$ an element */
  ZENEltToZPtr(R) = _ZemEltToZ; /* Get the element from an evaluation in $p$ */
  ZENEltConvertPtr(R) = _ZemEltConvert;

  /* Procedures on polynomials
   ****************************/
  /* Allocation */
  ZENPolyCreatePtr(R) = _ZepPolyCreate;
  ZENPolyCreateLargePtr(R) = _ZepPolyCreateLarge;
  ZENPolyFreePtr(R) = _ZepPolyFree;
  
  /* Assigning */
  ZENPolyAssignPtr(R) = _ZepPolyAssign;
  ZENPolyAssignLargePtr(R) = _ZepPolyAssignLarge;
  ZENPolyLazyToVeryLazyPtr(R) = _ZemPolyLazyToVeryLazy;
  ZENPolySetToZeroPtr(R) = _ZepPolySetToZero;
  ZENPolySetToZeroLargePtr(R) = _ZepPolySetToZeroLarge;
  ZENPolySetToXiPtr(R) = _ZemPolySetToXi;
  ZENPolySetRandomPtr(R) = _ZepPolySetRandom;
  ZENPolySetCoeffPtr(R) = _ZedPolySetCoeff;
  ZENPolyGetCoeffNewPtr(R) = _ZepPolyGetCoeff;
  ZENPolyGetCoeffPtrPtr(R) = _ZepPolyGetCoeffPtr;
  ZENPolyUpdateDegreePtr(R) = _ZepPolyUpdateDegree;
  
  /* Test */
  ZENPolyAreEqualPtr(R) = _ZepPolyAreEqual;
  ZENPolyIsZeroPtr(R) = _ZepPolyIsZero;
  ZENPolyIsXiPtr(R) = _ZemPolyIsXi;
  
  /* Arithmetic */
  ZENPolyAddPtr(R) = _ZepPolyAdd;
  ZENPolyAddLargePtr(R) = _ZepPolyAddLarge;
  ZENPolyNegatePtr(R) = _ZepPolyNegate;
  ZENPolyNegateLargePtr(R) = _ZepPolyNegateLarge;
  ZENPolySubtractPtr(R) = _ZepPolySubtract;
  ZENPolySubtractLargePtr(R) = _ZepPolySubtractLarge;

  ZENPolyAddPtr(R) = _ZepPolyAdd;	/* Addition of polynomials */
  ZENPolyAddLargePtr(R) = _ZepPolyAddLarge;	/* Addition of polynomials */
  ZENPolyNegatePtr(R) = _ZepPolyNegate; /* Negation of polynomials */
  ZENPolyNegateLargePtr(R) = _ZepPolyNegateLarge; /* Negation of polynomials */
  ZENPolySubtractPtr(R) = _ZepPolySubtract; /* Substraction of polynomials */
  ZENPolySubtractLargePtr(R) = _ZepPolySubtractLarge; /* Substraction of polynomials */

  ZENPolySquarePtr(R) = _ZemPolySquarePlain; /* Squaring of polynomials */
  ZENPolySquareLazyPtr(R) = _ZemPolySquareKaratsubaLazy;
  ZENPolySquareVeryLazyPtr(R) = _ZepPolySquareKaratsubaVeryLazy;

  ZENPolyMultiplyPtr(R) = _ZemPolyMultiplyPlain; /* Multiplication of polynomials */
  ZENPolyMultiplyLazyPtr(R) = _ZemPolyMultiplyLazy;
  ZENPolyMultiplyVeryLazyPtr(R) = _ZepPolyMultiplyVeryLazy;

  ZENPolyMultiplyScalarPtr(R) = _ZemPolyMultiplyScalar; /* Multiplication of polynomials by a scalar*/
  ZENPolyDotPtr(R) = _ZemPolyDot; /* Scalar multiplication of polynomials */
  ZENPolyEvalPtr(R) = _ZENPolyEval; /* Evaluation of a polynomial */
  ZENPolyDividePtr(R) = _ZemPolyDivide; /* Division of polynomials */
  ZENPolyDivideLazyPtr(R) = _ZemPolyDivideLazy; /* Lazy division of polynomials */
  ZENPolyDivideVeryLazyPtr(R) = _ZemPolyDivideVeryLazy; /* Lazy division of polynomials */
  ZENPolyModuloLazyPtr(R) = _ZemPolyModuloVeryLazy; /* Lazy modulo of polynomials */
  ZENPolyModuloVeryLazyPtr(R) = _ZemPolyModuloVeryLazy; /* Lazy modulo of polynomials */
  ZENPolyDerivePtr(R) = _ZENPolyDerive;
  ZENPolyRootsDegree2Ptr(R) = _ZENPolyRootsDegree2Charp; /* Roots of polynomials of degree 2 */

  /* Procedures on matrices
   **************************/
  /* Allocation */
  ZENMatCreateParamsPtr(R) = _ZepMatCreateParams; /* Create a matrix */
  ZENMatCreatePtr(R) = __ZedMatCreate;	/* Create a matrix */

  /* Assigning */
  ZENMatAssignPtr(R) = _ZepMatAssign; /* Assign a matrix with another */
  ZENPermutation2MatrixPtr(R) = _ZENPermutation2Matrix;	/* Assign a matrix with another */
  ZENMatSetToZeroPtr(R) = _ZepMatSetToZero; /* Set a matrix to zero */
  ZENMatSetToOnePtr(R) = _ZepMatSetToOne; /* Set a matrix to identity */
  ZENMatSetRandomPtr(R) = _ZENMatSetRandom; /* Set randomly a matrix */
  ZENMatSetBlocPtr(R) = _ZedMatSetBloc; /* Set a coefficient of a matrix */
  ZENMatGetBlocNewPtr(R) = _ZedMatGetBloc; /* Get a coefficient of a matrix */
  ZENMatGetBlocPtrPtr(R) = _ZedMatGetBlocPtr; /* Get a coefficient of a matrix */
  ZENMatSetSubMatPtr(R) = _ZedMatSetSubMat; /* Set a submatrix */
  ZENMatGetSubMatPtr(R) = _ZedMatGetSubMat; /* Get a submatrix */
  
  ZENMatPermuteBnPtr(R) = _ZENMatPermuteBn;
  ZENMatPermuteDigitPtr(R) = _ZENMatPermuteDigit;
  
  /* Test */
  ZENMatAreEqualPtr(R) = _ZepMatAreEqual; /* Is a matrix equal to another */
  ZENMatIsZeroPtr(R) = _ZepMatIsZero;	/* Is a matrix equal to zero */
  ZENMatIsOnePtr(R) = _ZepMatIsOne;	/* Is a matrix equal to one */
  
  /* Arithmetic */
  ZENMatAddPtr(R) = _ZENMatAdd;	/* Addition of Matrix */
  ZENMatAddScalarPtr(R) = _ZENMatAddScalar;	/* Addition of Matrix */
  ZENMatMultiplyScalarPtr(R) = _ZENMatMultiplyScalar;	/* Multiplication of Matrix */
  ZENMatSubtractPtr(R) = _ZENMatSubtract; /* Substraction of Matrix */
  ZENMatNegatePtr(R) = _ZENMatNegate;	/* Squaring of Matrix */
  ZENMatWinogradPtr(R) = _ZENMatWinograd; /* Multiplication of Matrix */
  ZENMatMultiplyPtr(R) = _ZENMatWinograd; /* Multiplication of Matrix */
  ZENMatMultiplyPlainPtr(R) = _ZENMatMultiplyPlain; /* Multiplication of Matrix */
  ZENMatMultiplyLeftPermutationPtr(R) = _ZENMatMultiplyLeftPermutation;
  ZENMatMultiplyRightPermutationPtr(R) = _ZENMatMultiplyRightPermutation;
  ZENMatInversePtr(R) = _ZENMatInverse; /* Inverse of a Matrix */
  ZENMatKernelPtr(R) = _ZENMatKernel;	/* Kernel of a Matrix */  
  ZENMatGaussPtr(R) = _ZENMatGauss;
  
  /* IO */
  ZENMatGetFromFilePtr(R) = _ZENMatGetFromFile;
  ZENMatPutToFilePtr(R) = _ZENMatPutToFile;
  ZENMatGetFromStringPtr(R) = _ZENMatGetFromString;
  ZENMatPutToStringPtr(R) = _ZENMatPutToString;

  ZENVect2MatPtr(R) = ZENMatAssignPtr(R);
  ZENMat2VectPtr(R) = ZENMatAssignPtr(R);

				/* ********* */

  ZemRingMbufl(R) = 2+ZENRingSizeElt(R)*ZENRingSizeElt(R);
  if (!(ZemRingMbuf(R) = ZBNC(ZemRingMbufl(R)))) {
    ZENSetError(ZEN_LIB_ZEM, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  if (!(ZemRingNbuf(R) = ZBNC(ZENRingSizeElt(R)))) {
    ZBNF(ZemRingMbuf(R));
    ZENSetError(ZEN_LIB_ZEM, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  ZemRingBnl(R)=ZENRingPl(R)+ZENRingSizeElt(ZENRingOrigin(R))+1;
  if(!(ZemRingBn(R)=ZBNC(ZemRingBnl(R)))) {
    ZBNF(ZemRingMbuf(R));
    ZBNF(ZemRingNbuf(R));
    ZENSetError(ZEN_LIB_ZEM, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }
  if(!(ZemRingBn2(R)=ZBNC(ZemRingBnl(R)))) {
    ZBNF(ZemRingBn(R));
    ZBNF(ZemRingMbuf(R));
    ZBNF(ZemRingNbuf(R));
    ZENSetError(ZEN_LIB_ZEM, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  ZBNSetToZero(ZemRingBn(R),ZemRingBnl(R));
  ZBNDigit(ZemRingBn(R),ZENRingPl(R))=1; /* this is $R$ */
  if(!ZBNModuloLazy(&(ZENElt2Bn(ZemRingOne(R),R)),
		   ZemRingBn(R),ZENRingPl(R)+1,
		   ZENRingP(R),ZENRingPl(R))) {
    ZBNF(ZemRingMbuf(R));
    ZBNF(ZemRingNbuf(R));
    ZBNF(ZemRingBn(R));
    ZBNF(ZemRingBn2(R));
    return(ZERR);
  }
  
  if(ZENBaseRingAlloc(Rg,ZemRingBn(R),ZENRingPl(R)+1)) {
    ZemEltFree(ZemRingOne(R),R);
    ZBNF(ZemRingMbuf(R));
    ZBNF(ZemRingNbuf(R));
    ZBNF(ZemRingBn(R));
    ZBNF(ZemRingBn2(R));
    return(ZERR);
  }
  if(ZENEltCreate(V,Rg)) {
    ZemEltFree(ZemRingOne(R),R);
    ZBNF(ZemRingMbuf(R));
    ZBNF(ZemRingNbuf(R));
    ZBNF(ZemRingBn(R));
    ZBNF(ZemRingBn2(R));
    return(ZERR);
  }
  ZENZToElt(V,ZENRingP(R),ZENRingPl(R),Rg);
  if((ret=ZENEltInverse(V,V,Rg)))
    {
      ZENEltFree(V,Rg);
      ZemEltFree(ZemRingOne(R),R);
      ZBNF(ZemRingMbuf(R));
      ZBNF(ZemRingNbuf(R));
      ZBNF(ZemRingBn(R));
      ZBNF(ZemRingBn2(R));
      ZENSetError(ZEN_LIB_ZEM, ZEN_INIT, ZEN_INPUT);
      return(ZERR);      
    }
  ZENEltNegate(V,V,Rg);
  ZemRingV(R)=ZENElt2Bn(V,Rg);
  ZemRingVl(R)=ZBNNumDigits(ZemRingV(R),ZENRingSizeElt(Rg));
  if(ZENRingClose(Rg)) {
    ZENEltFree(V,Rg);
    ZemEltFree(ZemRingOne(R),R);
    ZBNF(ZemRingMbuf(R));
    ZBNF(ZemRingNbuf(R));
    ZBNF(ZemRingBn(R));
    ZBNF(ZemRingBn2(R));
    return(ZERR);      
  }
#if MONTGOMERY_VERBOSE
  printf("N:=");  ZBP(ZENRingP(R),ZENRingPl(R));
  printf("R:=2^(%d*%d);\n",ZENRingPl(R),SIZE_BLOC);
  printf("v:=");  ZBP(ZemRingV(R),ZemRingVl(R));
  printf("1+v*N mod R;\n");
  printf("One:=");  ZEP(ZemRingOne(R),ZENRingOrigin(R)); printf(";\n");
  printf("One - R mod N;\n");
#endif
  return(SR_GEN);
}

/* \TEX[
   \title{Closing.}
   \input{A partially freed |ZENRing|.}
   \output{0.}
   }
   */
int _ZemClose(R)		
     ZENRing R;
{
  /* The buffers */
  ZemEltFree(ZemRingOne(R),R);
  ZBNF(ZemRingBn(R));
  ZBNF(ZemRingBn2(R));
  ZBNF(ZemRingMbuf(R));
  ZBNF(ZemRingNbuf(R));
  ZBNF(ZemRingV(R));
  return(0);
}

