/*
 * zen/zem/c/modulo.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Modular reduction}
   } 
   */

#include "zem.h"
#ifndef min
#define min(x,y) ((x)<(y)? (x) : (y))
#endif

#define OPTIMIZE 1
/*
   \TEX{
   \title{Montgomery's modular reduction}
   \input{A |BigNum| and its length and a Montgomery's clone |Rg|.}
   \side{|z| is filled with the modular reduction $|z|R^{-1} \bmod N$.}
   \output{The size of the result.}
   \note{We denote $N$ the odd modulus of |ZENRingOrigin(Rg)|, and $R$ the smallest power of $2^{|SIZE_BLOC|}$ greater than $N$. Following Montgomery, the input integer $z=(|z|,|zl|)$ is supposed to be less than $RN$. |z| must be allocated with size at least |ZemRingBnl(Rg)| and the most significant digits from |*p_zl| to $|ZemRingBnl(Rg)|-1$ accordingly set to zero.}
   } 
   */
BigNumLength _ZemEltRedMod(z,Rg)
     BigNum z;
     ZENRing Rg;
{
  BigNumLength vl,bnl,ml,pl,l;
  BigNum pt,v,bn2,p;

  bnl=ZemRingBnl(Rg);
#if MONTGOMERY_VERBOSE
  printf("z:=");  ZBP(z,bnl);
  printf("N*R-z;\n");
#endif

  v=ZemRingV(Rg);
  vl=ZemRingVl(Rg);
  bn2=ZemRingBn2(Rg);
  p=ZENRingP(Rg);
  pl=ZENRingPl(Rg);
  ml=ZBNNumDigits(z,pl);

#if OPTIMIZE
  {
    BigNum vd;

    l=pl+1;
    ZBNSetToZero(bn2,l);
    vd=z;
    pt=bn2;
    while(ml--) {
      ZBNMultiplyDigit(pt,l,v,vl,*vd);
      pt++; l--; vl--; vd++;
    }
    
  }
#else
  l=ml+vl;
  ZBNSetToZero(bn2,l);
  ZBNMultiply(bn2,l,z,ml,v,vl);
#endif
  
  l=ZBNNumDigits(bn2,pl);
#if MONTGOMERY_VERBOSE
  printf("m:=");      ZBP(bn2,l);
  printf("m - (((z mod R)*v)mod R);\n");
#endif
  ZBNMultiply(z,bnl,p,pl,bn2,l);
  pt=z+pl;
  l=ZBNNumDigits(pt,bnl-pl);
#if MONTGOMERY_VERBOSE
  printf("t:=");      ZBP(pt,l);
  printf("t - ((z + m*N)/ R);\n");
#endif
  switch(ZBNCompareLazy(pt,l,p,pl))
    {
    case BN_EQ:
      z[0]=0;
      return(1);
    case BN_GT:
      ZBNSubtract(pt,l,p,pl,1);
      l=ZBNNumDigits(pt,l);
      while (ZBNCompareLazy(pt,l,p,pl) == BN_GT) {
	ZBNSubtract(pt,l,p,pl,1);
	l=ZBNNumDigits(pt,l);
      }
    case BN_LT:
      break;
    }

  ZBNAssign(z,pt,l);
#if MONTGOMERY_VERBOSE
  printf("zr:=");  ZBP(z,l);
  printf("((zr*R) mod N)-(z mod N);\n");
#endif
  return(l);
}


/*
   \TEX{
   \title{Montgomery's modular reduction}
   \input{A |BigNum| and its length and a Montgomery's clone |Rg|.}
   \side{|z| is filled with the modular reduction $|z|R^{-1} \bmod N$.}
   \output{The size of the result.}
   \note{We denote $N$ the odd modulus of |ZENRingOrigin(Rg)|, and $R$ the smallest power of $2^{|SIZE_BLOC|}$ greater than $N$. Following Montgomery, the input integer $z=(|z|,|zl|)$ is supposed to be less than $RN$. |z| must be allocated with size at least |ZemRingBnl(Rg)| and the most significant digits from |*p_zl| to $|ZemRingBnl(Rg)|-1$ accordingly set to zero. Karatsuba's multiplication algorithm is used.}
   } 
   */
BigNumLength _ZemEltRedModKaratsuba(z,buf,Rg)
     BigNum z,buf;
     ZENRing Rg;
{
  BigNumLength vl,bnl,ml,pl,l;
  BigNum pt,v,bn2,p;

  bnl=ZemRingBnl(Rg);
#if MONTGOMERY_VERBOSE
  printf("z:=");  ZBP(z,bnl);
  printf("N*R-z;\n");
#endif

  v=ZemRingV(Rg);
  vl=ZemRingVl(Rg);
  bn2=ZemRingBn2(Rg);
  p=ZENRingP(Rg);
  pl=ZENRingPl(Rg);
  ml=ZBNNumDigits(z,pl);

#if OPTIMIZE
  {
    BigNum vd;

    l=pl+1;
    ZBNSetToZero(bn2,l);
    vd=z;
    pt=bn2;
    while(ml--) {
      ZBNMultiplyDigit(pt,l,v,vl,*vd);
      pt++; l--; vl--; vd++;
    }
  }
#else
  l=ml+vl;
  ZBNSetToZero(bn2,l);
  ZBNMultiplyKaratsubaBuffer(bn2,z,ml,v,vl,buf);
#endif
  
  l=ZBNNumDigits(bn2,pl);
#if MONTGOMERY_VERBOSE
  printf("m:=");      ZBP(bn2,l);
  printf("m - (((z mod R)*v)mod R);\n");
#endif
#if OPTIMIZE
  {
    BigNum bn3;

    bn3=ZemRingBn3(Rg);
    ZBNSetToZero(bn3,l+pl);
    ZBNMultiplyKaratsubaBuffer(bn3,p,pl,bn2,l,buf);
    pt=z+pl;
    if(ZBNIsZero(bn3,pl))
      ZBNAdd(pt,bnl-pl,bn3+pl,l,(BigNumDigit)0);
    else
      ZBNAdd(pt,bnl-pl,bn3+pl,l,(BigNumDigit)1);
  }
#else
  ZBNMultiply(z,bnl,p,pl,bn2,l);
  pt=z+pl;
#endif
  l=ZBNNumDigits(pt,bnl-pl);
#if MONTGOMERY_VERBOSE
  printf("t:=");      ZBP(pt,l);
  printf("t - ((z + m*N)/ R);\n");
#endif
  switch(ZBNCompareLazy(pt,l,p,pl))
    {
    case BN_EQ:
      z[0]=0;
      return(1);
    case BN_GT:
      ZBNSubtract(pt,l,p,pl,1);
      l=ZBNNumDigits(pt,l);
    case BN_LT:
      break;
    }

  ZBNAssign(z,pt,l);
#if MONTGOMERY_VERBOSE
  printf("zr:=");  ZBP(z,l);
  printf("((zr*R) mod N)-(z mod N);\n");
#endif
  return(l);
}
/*
   \TEX[
   \title{Modular multiplication}
   \input{Three allocated |BigNum|s.}
   \side{$|x| = |a| . |b| \bmod |m| $}
   \note{One must have $|xl| = ZemRingBnl(R)$, $|*(a+al-1)|\neq 0$ and $|*(b+bl-1)|\neq 0$. One can have $|a| = |b|$. }
   } 
   */
void _ZemModMultiply(x,a,al,b,bl,R)
     BigNum x,a,b;
     BigNumLength al,bl;
     ZENRing R;
{
  BigNumLength ll;

  ZBNSetToZero(x,ZemRingBnl(R));
  
  ll=al+bl;
  ZBNMultiply(x,ll,b,bl,a,al);
#define ModularReduction() \
  ll=_ZemEltRedMod(x,R); \
  ZBNSetToZero(x+ll,ZENRingSizeElt(R)-ll);
  ModularReduction();
}  


/*
   \TEX[
   \title{Modular squaring}
   \input{Two allocated |BigNum|s.}
   \side{$|x| = |a|^2 \bmod |m| $}
   \note{One must have $|xl| = |ZemRingBnl(R)|$, and $|*(a+al-1)|\neq 0$.}
   } 
   */
void _ZemModSquare(x,a,al,R)
     BigNum x,a;
     BigNumLength al;
     ZENRing R;
{
  BigNumLength ll;
 
  ZBNSetToZero(x,ZemRingBnl(R));

  ll=al<<1;

  ZBNSquare(x,ll,a,al); 

  ModularReduction();
}

/*
   \TEX[
   \title{Modular multiplication}
   \input{Three allocated |BigNum|s.}
   \side{$|x| = |a| . |b| \bmod |m| $}
   \note{One must have $|xl| \geq |al|+ |bl| + 1$, $|*(a+al-1)|\neq 0$ and $|*(b+bl-1)|\neq 0$. One can have $|a| = |b|$. }
   } 
   */
void _ZemModMultiplyKaratsubaBuffer(x,a,al,b,bl,buf,R)
     BigNum x,a,b,buf;
     BigNumLength al,bl;
     ZENRing R;
{
  BigNumLength ll;

  
  ZBNSetToZero(x,ZemRingBnl(R));
  ll=al+bl;

  ZBNMultiplyKaratsubaBuffer(x,b,bl,a,al,buf);
#undef ModularReduction
#define ModularReduction() \
  ll=_ZemEltRedModKaratsuba(x,buf,R); \
  ZBNSetToZero(x+ll,ZENRingSizeElt(R)-ll);

  ModularReduction();
}

/*
   \TEX[
   \title{Modular squaring}
   \input{Two allocated |BigNum|s.}
   \side{$|x| = |a|^2 \bmod |m| $}
   \note{One must have $|xl| \geq 2\times|al| + 1$, $|*(a+al-1)|\neq 0$ and $|*(b+bl-1)|\neq 0$. One can have $|a| = |b|$. }
   } 
   */
void _ZemModSquareKaratsubaBuffer(x,a,al,buf,R)
     BigNum x,a,buf;
     BigNumLength al;
     ZENRing R;
{
  BigNumLength ll;
 
  ZBNSetToZero(x,ZemRingBnl(R));

  ll=al<<1;

  ZBNSquareKaratsubaBuffer(x,a,al,buf); 

  ModularReduction()
}
