/*
 * zen/zem/c/poly_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsubsection{Polynomial additions}
   
   One function is provided to do additions: |_ZepPolyAdd| to add a
   |ZENPoly| to another.
   }
   */
#include <stdio.h>
#include "zem.h"

#define ZBNPolyGetCof(P,i,R) ZENElt2Bn(ZemPolyGetCof((P),(i),(R)),(R))
#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

/* \TEX[
   \title{Addition.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| += |PX|.}
   }
   */
void _ZemPolyAdd(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int mindeg, i;

  if (ZemPolyDeg(RX, Rg) < ZemPolyDeg(PX, Rg)) {
    mindeg = ZemPolyDeg(RX, Rg);
    ZemPolyDeg(RX, Rg) =  ZemPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZemPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    ZemEltAdd(ZemPolyGetCof(RX, i, Rg), ZemPolyGetCof(PX, i, Rg), Rg);
  }
  for(i = mindeg+1; i <= ZemPolyDeg(PX, Rg); i++) {
    ZemEltAssign(ZemPolyGetCof(RX, i, Rg),
		   ZemPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZemPolyUpdateDegree(RX, Rg);
}

/* For internal use only, cf. poly_multiply. */
void _ZemPolyAddLarge(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int mindeg, i;
  
  mindeg = min(ZemPolyDeg(RX, Rg), ZemPolyDeg(PX, Rg));
  for(i = 0; i <= mindeg; i++) {
    ZBNAdd(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg),
	   ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), 0);
  }
  for(i = mindeg+1; i <= ZemPolyDeg(PX, Rg); i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg));
  }
  for(i = max(i-1, ZemPolyDeg(RX, Rg)); i >= 0; i--) {
    if (!ZBNIsZero(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg))) {
      break;
    }
  }
  ZemPolyDeg(RX, Rg) = i;
}

/* \TEX[
   \title{Negation.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| = -|PX|.}
   }
   */
void _ZemPolyNegate(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZemPolyDeg(PX, Rg); i++) {
    ZemEltNegate(ZemPolyGetCof(RX, i, Rg), ZemPolyGetCof(PX, i, Rg), Rg);
  }
  ZemPolyDeg(RX, Rg) =  ZemPolyDeg(PX, Rg);
}
/* For internal use only, cf. poly_multiply. */
void _ZemPolyNegateLarge(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZemPolyDeg(PX, Rg); i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg), ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg));
    ZBNComplement(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg));
    ZBNAddCarry(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 1);
  }
  ZemPolyDeg(RX, Rg) = ZemPolyDeg(PX, Rg);
}

/* \TEX[
   \title{Subtraction.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| -= |PX|.}
   }
   */
void _ZemPolySubtract(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, mindeg;

  if (ZemPolyDeg(RX, Rg) < ZemPolyDeg(PX, Rg)) {
    mindeg = ZemPolyDeg(RX, Rg);
    ZemPolyDeg(RX, Rg) =  ZemPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZemPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    ZemEltSubtract(ZemPolyGetCof(RX, i, Rg), ZemPolyGetCof(PX, i, Rg), Rg);
  }
  for(i = mindeg+1; i <= ZemPolyDeg(PX, Rg); i++) {
    ZemEltNegate(ZemPolyGetCof(RX, i, Rg),
		   ZemPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZemPolyUpdateDegree(RX, Rg);
}


/* For internal use only, cf. poly_multiply. */
void _ZemPolySubtractLarge(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, mindeg;

  mindeg = min(ZemPolyDeg(RX, Rg), ZemPolyDeg(PX, Rg));
  for(i = 0; i <= mindeg; i++) {
    ZBNSubtract(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg),
		ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), 1);
  }
  for(i = mindeg+1; i <= ZemPolyDeg(PX, Rg); i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg), ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg));
    ZBNComplement(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg));
    ZBNAddCarry(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 1);
  }
  for(i = max(i-1, ZemPolyDeg(RX, Rg)); i >= 0; i--) {
    if (!ZBNIsZero(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg))) {
      break;
    }
  }
  ZemPolyDeg(RX, Rg) = i;
}

