/*
 * zen/zem/c/poly_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsubsection{Polynomial multiplication}
   
   Two functions are provided to do multiplications: |_ZepPolySquare| to square a
   |ZENPoly| and |_ZepPolyMultily| to compute the multiplication of two |ZENPoly|.
   }
   */
#include <stdio.h>
#include "zem.h"
#include "zeg.h"

#define max(a,b) ((a) > (b) ? (a) : (b))
#define min(a,b) ((a) < (b) ? (a) : (b))

#define ZBNPolyGetCof(P,i,R) ZENElt2Bn(ZemPolyGetCof((P),(i),(R)),(R))
#define ZBNRingHbuf(R) ZENElt2Bn(ZENRingHbuf(R),(R))

/* Some miscaellenous functions
********************************/

static void ZemQuickMod(e, n, nl, Rg)
     ZENElt e;
     BigNum n;
     BigNumLength nl;
     ZENRing Rg;
{ 
  BigNumLength ll;

  ll = ZemEltRedMod(n, Rg);
  ZBNSetToZero(n+ll, ZENRingSizeElt(Rg)-ll);  
  ZBNAssign(ZENElt2Bn(e,Rg), n, ZENRingSizeElt(Rg));
}

/* Usual squaring routines
***************************/

/* \TEX[
   \title{Squaring in characteristic $p\neq 2$.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZemPolySquarePlain(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, j, k;

  for(k = ZemPolyDeg(PX, Rg); k > 0; k--) {

    /* coefficient of X^(2k) */
    ZBNSetToZero(ZBNRingHbuf(Rg), ZENRingSZ(Rg)+1);
    for(i = min(ZemPolyDeg(PX, Rg), 2*k);
	i > k;
	i--) {
      j = 2*k - i;
      ZBNMultiply(ZBNRingHbuf(Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(PX, j, Rg), ZENRingSizeElt(Rg));
    }
    ZBNShiftLeft(ZBNRingHbuf(Rg), ZENRingSZ(Rg), 1);
    ZBNSquare(ZBNRingHbuf(Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(PX, k, Rg), ZENRingSizeElt(Rg));
    ZemQuickMod(ZemPolyGetCof(RX, 2*k, Rg),
		ZBNRingHbuf(Rg),
		ZENRingSZ(Rg), Rg);

    /* coefficient of X^(2k-1) */
    ZBNSetToZero(ZBNRingHbuf(Rg), ZENRingSZ(Rg)+1);
    for(i = min(ZemPolyDeg(PX, Rg), 2*k-1);
	i >= k;
	i--) {
      j = 2*k - i - 1;
      ZBNMultiply(ZBNRingHbuf(Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(PX, j, Rg), ZENRingSizeElt(Rg));
    }
    ZBNShiftLeft(ZBNRingHbuf(Rg), ZENRingSZ(Rg), 1);
    ZemQuickMod(ZemPolyGetCof(RX, 2*k-1, Rg),
		ZBNRingHbuf(Rg),
		ZENRingSZ(Rg), Rg);
  }

  /* coefficient of X^(0) */
  ZBNSetToZero(ZBNRingHbuf(Rg), ZENRingSZ(Rg)+1);
  ZBNSquare(ZBNRingHbuf(Rg), ZENRingSZ(Rg),
	    ZBNPolyGetCof(PX, 0, Rg), ZENRingSizeElt(Rg));
  ZemQuickMod(ZemPolyGetCof(RX, 0, Rg),
	      ZBNRingHbuf(Rg),
	      ZENRingSZ(Rg), Rg);
  
  ZENPolyDeg(RX, Rg) = 2*ZENPolyDeg(PX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* Karatsuba squaring routines
*******************************/

/* \TEX[
   \title{Squaring with Karatsuba.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZemPolySquareKaratsuba(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;
  ZENPoly RBuf;

  if (!(RBuf = ZemPolyCreateLarge(4*ZemPolyDeg(PX, Rg)+1, Rg))) {
    return(ZERR);
  }
  ZemPolySquareKaratsubaVeryLazy(RBuf, PX, Rg);
  for(i = ZemPolyDeg(RBuf, Rg); i >= 0; i--) {
    ZemQuickMod(ZemPolyGetCof(RBuf, i, Rg),
		ZBNPolyGetCof(RBuf, i, Rg), 
		2*ZENRingSizeElt(Rg)+1, Rg);
  }
  ZemPolyAssign(RX, RBuf, Rg);
  ZemPolyFree(RBuf, Rg);
  return(ZEN_HAS_INVERSE);
}

int _ZemPolySquareKaratsubaLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  ZemPolySquareKaratsubaVeryLazy(RX, PX, Rg);
  for(i = ZemPolyDeg(RX, Rg); i >= 0; i--) {
    ZemQuickMod(ZemPolyGetCof(RX, i, Rg),
		ZBNPolyGetCof(RX, i, Rg), 
		2*ZENRingSizeElt(Rg)+1, Rg);
  }
  return(ZEN_HAS_INVERSE);
}

/* General multiplication routines
**********************************/

/* \TEX[
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   }
   */
int _ZemPolyMultiply(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZENPoly RBuf;
  int n, m, q, r, L, k;

  if (ZemPolyDeg(PX, Rg) > ZemPolyDeg(QX, Rg)) {
    n = ZemPolyDeg(PX, Rg)+1; m = ZemPolyDeg(QX, Rg)+1;
    q = n/m; r = n%m;
  }
  else {
    n = ZemPolyDeg(QX, Rg)+1; m = ZemPolyDeg(PX, Rg)+1;
    q = n/m; r = n%m;
  }
  switch(q) {
  case 1:
    L = r ? 4*m-4+2*r : 4*m-3;
    break;
  case 2:
    L = 2*r > (m+1) ? 5*m-5+2*r : 6*m-4;
    break;
  default:
    L = 2*r > m ?  (3+q)*m-4+2*r: (4+q)*m-4;
    break;
  }
  if (!(RBuf = ZemPolyCreateLarge(L, Rg))) {
    return(ZERR);
  }
  ZemPolyMultiplyVeryLazy(RBuf, PX, QX, Rg);
  for(k = ZemPolyDeg(RBuf, Rg); k >= 0; k--) {
    ZemQuickMod(ZemPolyGetCof(RBuf, k, Rg),
		ZBNPolyGetCof(RBuf, k, Rg), 
		2*ZENRingSizeElt(Rg)+1, Rg);
  }
  ZemPolyAssign(RX, RBuf, Rg);
  ZemPolyFree(RBuf, Rg);

  return(ZEN_HAS_INVERSE);
}

int _ZemPolyMultiplyLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int k;

  ZemPolyMultiplyVeryLazy(RX, PX, QX, Rg);
  for(k = ZemPolyDeg(RX, Rg); k >= 0; k--) {
    ZemQuickMod(ZemPolyGetCof(RX, k, Rg),
		ZBNPolyGetCof(RX, k, Rg), 
		2*ZENRingSizeElt(Rg)+1, Rg);
  }
  return(ZEN_HAS_INVERSE);
}

/* \TEX[
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|\times|QX|$.}
   }
   */
int _ZemPolyMultiplyPlain(RX, QX, PX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i, j, k;

  if (ZemPolyIsZero(PX, Rg) || ZemPolyIsZero(QX, Rg)) {
    ZBNSetToZero(ZBNPolyGetCof(RX, 0, Rg), ZENRingSizeElt(Rg));
    ZENPolyDeg(RX, Rg) = 0;
    return(ZEN_HAS_INVERSE);
  }

  for(k = ZemPolyDeg(PX, Rg)+ZemPolyDeg(QX, Rg);
      k >= 0;
      k--) {
    ZBNSetToZero(ZBNRingHbuf(Rg), ZENRingSZ(Rg)+1);
    for(i = min(ZemPolyDeg(PX, Rg), k);
	i >= max(0, k - ZemPolyDeg(QX, Rg));
	i--) {
      j = k - i;
      ZBNMultiply(ZBNRingHbuf(Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(QX, j, Rg), ZENRingSizeElt(Rg));
    }
    ZemQuickMod(ZemPolyGetCof(RX, k, Rg),
		ZBNRingHbuf(Rg),
		ZENRingSZ(Rg), Rg);
  }
  ZemPolyDeg(RX, Rg) = ZemPolyDeg(PX, Rg) + ZemPolyDeg(QX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* Specific Karatsuba multiplication routines
*********************************************/

/* \TEX[
   \title{Polynomial multiplication with the Karatsuba method.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an errpr occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   }
   */
int _ZemPolyMultiplyKaratsuba(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i;
  ZENPoly QBuf, RBuf;
  ZENElt elt;

  if (ZemPolyDeg(PX, Rg) >= ZemPolyDeg(QX, Rg)) {
    if (!(QBuf = ZemPolyCreate(ZemPolyDeg(PX, Rg), Rg))) {
      return(ZERR);
    }
    ZemPolyAssign(QBuf, QX, Rg);
    for(i = ZemPolyDeg(QX, Rg)+1;
	i <= ZemPolyDeg(PX, Rg);
	i++) {
      elt = ZemPolyGetCoeffPtr(QBuf, i, Rg);
      ZemEltSetToZero(elt, Rg);
      ZemPolySetCoeff(QBuf, i, elt, Rg);
    }

    if (!(RBuf = ZemPolyCreateLarge(4*ZemPolyDeg(PX, Rg)+1, Rg))) {
      return(ZERR);
    }

    ZemPolyMultiplyKaratsubaVeryLazy(RBuf, PX, QBuf, Rg);
  }
  else {
    if (!(QBuf = ZemPolyCreate(ZemPolyDeg(QX, Rg), Rg))) {
      return(ZERR);
    }
    ZemPolyAssign(QBuf, PX, Rg);
    for(i = ZemPolyDeg(PX, Rg)+1;
	i <= ZemPolyDeg(QX, Rg);
	i++) {
      elt = ZemPolyGetCoeffPtr(QBuf, i, Rg);
      ZemEltSetToZero(elt, Rg);
      ZemPolySetCoeff(QBuf, i, elt, Rg);
    }

    if (!(RBuf = ZemPolyCreateLarge(4*ZemPolyDeg(PX, Rg)+1, Rg))) {
      return(ZERR);
    }

    ZemPolyMultiplyKaratsubaVeryLazy(RBuf, QX, QBuf, Rg);
  }
  for(i = ZemPolyDeg(RBuf, Rg); i >= 0; i--) {
    ZemQuickMod(ZemPolyGetCof(RBuf, i, Rg),
		ZBNPolyGetCof(RBuf, i, Rg), 
		2*ZENRingSizeElt(Rg)+1, Rg);
  }
  ZemPolyAssign(RX, RBuf, Rg);
  ZemPolyFree(QBuf, Rg); ZemPolyFree(RBuf, Rg);
  return(ZEN_HAS_INVERSE);
}

int _ZemPolyMultiplyKaratsubaLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  if (ZemPolyDeg(PX, Rg) >= ZemPolyDeg(QX, Rg)) {
    for(i = ZemPolyDeg(QX, Rg)+1;
	i <= ZemPolyDeg(PX, Rg);
	i++) {
      elt = ZemPolyGetCoeffPtr(QX, i, Rg);
      ZemEltSetToZero(elt, Rg);
      ZemPolySetCoeff(QX, i, elt, Rg);
    }


  }
  else {
    for(i = ZemPolyDeg(PX, Rg)+1;
	i <= ZemPolyDeg(QX, Rg);
	i++) {
      elt = ZemPolyGetCoeffPtr(PX, i, Rg);
      ZemEltSetToZero(elt, Rg);
      ZemPolySetCoeff(PX, i, elt, Rg);
    }
  }
  ZemPolyMultiplyKaratsubaVeryLazy(RX, PX, QX, Rg);
  for(i = ZemPolyDeg(RX, Rg); i >= 0; i--) {
    ZemQuickMod(ZemPolyGetCof(RX, i, Rg),
		ZBNPolyGetCof(RX, i, Rg), 
		2*ZENRingSizeElt(Rg)+1, Rg);
  }
  return(ZEN_HAS_INVERSE);
}

/* Division routines
***********************/

/* \TEX[
   \title{VeryLazy division.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZEN_NO_INVERSE| if
   the leading coefficient of |QX| is a factor of a modulo, |ZENERR| for an
   error}
   \note{|PX| is destroyed. |ZENRingFact(Rg)| is filled with a factor
   of a modulo if |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZemPolyDivideLazy(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  ZemPolyLazyToVeryLazy(PX, PX, Rg);
  return(ZemPolyDivideVeryLazy(RX, MX, PX, QX, Rg));
}

int _ZemPolyDivideVeryLazy(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int i, j, deg, IsNotOne;

  if (!ZemEltIsOne(ZemPolyGetCof(QX, ZemPolyDeg(QX, Rg), Rg), Rg)) {
    IsNotOne = 1;
    
    switch(ZemEltInverse(ZENRingIbuf(Rg),ZemPolyGetCof(QX, ZemPolyDeg(QX, Rg), Rg), Rg)) {
    case ZENERR:
      return(ZERR);
    case ZEN_NO_INVERSE:
      return(ZEN_NO_INVERSE);
    }
  }
  else {
    IsNotOne = 0;
  }

  for(i = ZemPolyDeg(PX, Rg);
      i >= ZemPolyDeg(QX, Rg);
      i--) {
      int neg;

      neg = 0;
      if (ZBNPolyGetCof(PX, i, Rg)[ZENRingSZ(Rg)-1] >> (BN_DIGIT_SIZE-1)) {
	ZBNComplement(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg));
	ZBNAddCarry(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), 1);
	neg = 1;
      }
      ZemQuickMod(ZemPolyGetCof(PX, i, Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), Rg);
      
      if (IsNotOne) {
	ZemEltMultiply(ZemPolyGetCof(PX, i, Rg),
		       ZENRingIbuf(Rg), ZemPolyGetCof(PX, i, Rg),
		       Rg);
      }
      if (neg) {
	ZemEltNegate(ZemPolyGetCof(PX, i, Rg), ZemPolyGetCof(PX, i, Rg), Rg);
      }
    
    for(j =  ZemPolyDeg(QX, Rg)-1;
	j >= 0;
	j--) {
      ZBNSetToZero(ZemRingMbuf(Rg), ZENRingSZ(Rg)+1);
      ZBNMultiply(ZemRingMbuf(Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(QX, j, Rg), ZENRingSizeElt(Rg));
      ZBNSubtract(ZBNPolyGetCof(PX, i+j-ZemPolyDeg(QX, Rg), Rg), ZENRingSZ(Rg),
		  ZemRingMbuf(Rg), ZENRingSZ(Rg), 1);
    }
  }
  deg =  ZemPolyDeg(QX, Rg);
  for(i = ZemPolyDeg(PX, Rg); i >= ZemPolyDeg(QX, Rg); i--) {
    ZemEltAssign(ZemPolyGetCof(RX, i-ZemPolyDeg(QX, Rg), Rg),
		 ZemPolyGetCof(PX, i, Rg),
		 Rg);
  }
  ZemPolyDeg(RX, Rg) = ZemPolyDeg(PX, Rg)-ZemPolyDeg(QX, Rg);
  if (ZemPolyDeg(RX, Rg) < 0) {
    ZemPolySetToZero(RX, Rg);
  }
  deg = -1;
  for(; i >= 0; i--) {
    int neg;
    
    neg = 0;
    if (ZBNPolyGetCof(PX, i, Rg)[ZENRingSZ(Rg)-1] >> (BN_DIGIT_SIZE-1)) {
      ZBNComplement(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg));
      ZBNAddCarry(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), 1);
      neg = 1;
    }
    ZemQuickMod(ZemPolyGetCof(PX, i, Rg), ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), Rg);
    if (!ZemEltIsZero(ZemPolyGetCof(PX, i, Rg), Rg)) {
      if (neg) {
	ZemEltNegate(ZemPolyGetCof(MX, i, Rg), ZemPolyGetCof(PX, i, Rg), Rg);
      }
      else {
	ZemEltAssign(ZemPolyGetCof(MX, i, Rg), ZemPolyGetCof(PX, i, Rg), Rg);
      }
      if (deg == -1) {
	deg = i;
      }
    }
    else if (deg >= 0) {
      ZemEltSetToZero(ZemPolyGetCof(MX, i, Rg), Rg);
    }
  }
  if (deg < 0) {
    ZemPolySetToZero(MX, Rg);
  }
  else {
    ZemPolyDeg(MX, Rg) = deg;
  }
  return(ZEN_HAS_INVERSE);
}


/* \TEX[
   \title{Divide.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZEN_NO_INVERSE| if
   a factor of a modulo was dicovered, |ZENERR| for an error}
   \side{|ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZemPolyDivide(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int i, ret;
  ZENPoly Pbuf;

  if (!(Pbuf = ZemPolyCreateLarge(ZemPolyDeg(PX, Rg), Rg))) {
    return(ZERR);
  }
  ZemPolyLazyToVeryLazy(Pbuf, PX, Rg);
  ret = ZemPolyDivideVeryLazy(RX, MX, Pbuf, QX, Rg);
  ZemPolyFree(Pbuf, Rg);
  return(ret);
}


/* \TEX[
   \title{VeryLazy mudulo computation.}
   \input{A polynomial |PX| to divide by
   |QX|$=X^{\mbox{degree}(QX)}+sum_{i=0}^{|k|}q_iX^i$} in a finite
   ring |Rg|. |MX| will be the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| for an error}
   \note{|PX| is destroyed.}
   */
int _ZemPolyModuloVeryLazy(MX, PX, QX, k, Rg)
     ZENPoly MX, PX, QX;
     int k;
     ZENRing Rg;
{
  int i, j, deg, IsNotOne;

  for(i = ZemPolyDeg(PX, Rg);
      i >= ZemPolyDeg(QX, Rg);
      i--) {
    int neg;
    
    neg = 0;
    if (ZBNPolyGetCof(PX, i, Rg)[ZENRingSZ(Rg)-1] >> (BN_DIGIT_SIZE-1)) {
      ZBNComplement(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg));
      ZBNAddCarry(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), 1);
      neg = 1;
    }
    ZemQuickMod(ZemPolyGetCof(PX, i, Rg),
		ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), Rg);
    if (neg) {
      ZemEltNegate(ZemPolyGetCof(PX, i, Rg), ZemPolyGetCof(PX, i, Rg), Rg);
    }
      
    for(j = k;
	j >= 0;
	j--) {
      ZBNSetToZero(ZemRingMbuf(Rg), ZENRingSZ(Rg)+1);
      ZBNMultiply(ZemRingMbuf(Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(QX, j, Rg), ZENRingSizeElt(Rg));
      ZBNSubtract(ZBNPolyGetCof(PX, i+j-ZemPolyDeg(QX, Rg), Rg), ZENRingSZ(Rg),
		  ZemRingMbuf(Rg), ZENRingSZ(Rg), 1);
    }
  }

  deg = -1;
  for(; i >= 0; i--) {
    int neg;

    neg = 0;
    if (ZBNPolyGetCof(PX, i, Rg)[ZENRingSZ(Rg)-1] >> (BN_DIGIT_SIZE-1)) {
      ZBNComplement(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg));
      ZBNAddCarry(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), 1);
      neg = 1;
    }
    ZemQuickMod(ZemPolyGetCof(PX, i, Rg), ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), Rg);
    if (!ZENEltIsZero(ZemPolyGetCof(PX, i, Rg), Rg)) {
      if (neg) {
	ZemEltNegate(ZemPolyGetCof(MX, i, Rg), ZemPolyGetCof(PX, i, Rg), Rg);
      }
      else {
	ZemEltAssign(ZemPolyGetCof(MX, i, Rg), ZemPolyGetCof(PX, i, Rg), Rg);
      }
      if (deg == -1) {
	deg = i;
      }
    }
    else if (deg >= 0) {
      ZemEltSetToZero(ZemPolyGetCof(MX, i, Rg), Rg);
    }
  }
  if (deg < 0) {
    ZemPolySetToZero(MX, Rg);
  }
  else {
    ZemPolyDeg(MX, Rg) = deg;
  }
  return(ZEN_HAS_INVERSE);
}

/* Other routines
*********************************************/

/* \TEX[
   \title{Multiplication by a scalar.}
   \input{Two polynomials |PX|, |RX| and a |ZENElt| |e| of a finite ring |Rg|.}
   }
	 */
void _ZemPolyMultiplyScalar(RX, PX, e, Rg)
     ZENPoly RX, PX;
     ZENElt e;
     ZENRing Rg;
{
  int i;

  if (ZemEltIsZero(e, Rg)) {
    ZemPolySetToZero(RX, Rg);
    return;
  }
  if (ZemEltIsOne(e, Rg)) {
    ZemPolyAssign(RX, PX, Rg);
    return;
  }
  for(i = 0; i <= ZemPolyDeg(PX, Rg); i++) {
    ZemEltMultiply(ZemPolyGetCof(RX, i, Rg),
		    e,
		    ZemPolyGetCof(PX, i, Rg),
		    Rg);
  }
  ZemPolyDeg(RX, Rg) = ZemPolyDeg(PX, Rg);
}

/* \TEX[
   \title{Scalar product.}
   \input{Two polynomials |PX|, |RX| and a pointer to a |ZENElt| |e| of a
   finite ring |Rg|.}
   \side{$|*e| = \sum p_iq_i$}
   }
   */
void _ZemPolyDot(e, PX, QX, Rg)
     ZENElt *e;
     ZENPoly PX, QX;
     ZENRing Rg;
{
  int i, deg;

  ZBNSetToZero(ZemRingMbuf(Rg), ZENRingSZ(Rg)+1);
  ZBNMultiply(ZemRingMbuf(Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(PX, 0, Rg), ZENRingSizeElt(Rg),
	      ZBNPolyGetCof(QX, 0, Rg), ZENRingSizeElt(Rg));


  deg = ZemPolyDeg(PX, Rg) < ZemPolyDeg(QX, Rg) ?
    ZemPolyDeg(PX, Rg) :  ZemPolyDeg(QX, Rg);

  for(i = 1; i <= deg; i++) {
    ZBNMultiply(ZemRingMbuf(Rg), ZENRingSZ(Rg),
		ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		ZBNPolyGetCof(QX, i, Rg), ZENRingSizeElt(Rg));
  }
  ZemQuickMod(*e, ZemRingMbuf(Rg), ZENRingSZ(Rg), Rg);
}

