/*
 * zen/zep/c/elt_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element assignment}
   
   One can set a |ZENElt| to zero or one with |_ZepEltSetToZero| or
   |_ZepEltSetToOne|. Then we can assign a |ZENElt| with an other with
   |_ZepEltAssign|. Furthermore, we can set randomly a |ZENElt| with
   |_ZepEltSetRandom| or enumerate the elements of a finite ring with
   |_ZepEltSetNext|. Finally, if we consider that a |ZENElt| is a
   polynomial in $T$, the function |_ZepEltGetCoeff| and |_ZepEltSetCoeff|
   enable us to get or set the coefficient of $T^i$, |_ZepEltGetDegree|
   enable us to get its degree.
   }
   */
#include <stdio.h>
#include "zep.h"

/* \TEX{
   \title{Assign.}
   \input{A pointer |pa| on a |ZENElt| and a |ZENElt| |b| of a finite ring |Rg|.}
   \side{|*pa| is filled with |b|.}
   }
   */
void _ZepEltAssign(pa, b, Rg)
     ZENElt *pa, b;
     ZENRing Rg;
{
  ZepEltAssign(*pa, b, Rg);
}

/* \TEX{
   \title{Set to zero.}
   \input{A pointer |pa| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*pa| is set to zero..}
   }
   */
void _ZepEltSetToZero(pa, Rg)
     ZENElt *pa;
     ZENRing Rg;
{
  ZepEltSetToZero(*pa, Rg);
}

/* \TEX{
   \title{Set to one.}
   \input{A pointer |pa| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*pa| is set to one.}
   }
   */
void _ZepEltSetToOne(pa, Rg)
     ZENElt *pa;
     ZENRing Rg;
{
  ZepEltSetToOne(*pa, Rg);
}

/* \TEX{
   \title{Set to random.}
   \input{A pointer |pa| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*pa| is set randomly.}
   }
   */
void _ZepEltSetRandom(pa, Rg)
     ZENElt *pa;
     ZENRing Rg;
{BigNumLength nl;

 ZBNSetRandom(ZepElt2Bn(*pa,Rg),ZENRingSizeElt(Rg));
 ZepElt2Bn(*pa,Rg)[ZENRingSizeElt(Rg)-1] &= ZepRingMask(Rg);
 nl=ZBNNumDigits(ZepElt2Bn(*pa,Rg), ZENRingSizeElt(Rg));
 switch(ZBNCompareLazy(ZepElt2Bn(*pa,Rg),nl,ZENRingP(Rg),ZENRingPl(Rg)))
   {
   case BN_EQ:
     ZepEltSetToZero(*pa,Rg);
     break;

   case BN_GT:
     ZBNSubtract(ZepElt2Bn(*pa,Rg),nl,ZENRingP(Rg),ZENRingPl(Rg),1);
     break;

   case BN_LT:
     break;
   }
}

/* \TEX{
   \title{Enumerate.}
   \input{A pointer |pa| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*pa| is set randomly.}
   \note{After a number of call of this function equal to the number of elements in the finite ring, we obtain the same element.}
   }
   */
void _ZepEltSetNext(pa, Rg)
     ZENElt *pa;
     ZENRing Rg;
{BigNumLength al;

 al=ZBNNumDigits(ZepElt2Bn(*pa,Rg),ZENRingSizeElt(Rg));
 if(!ZBNCompareLazy(ZepElt2Bn(*pa,Rg),al,ZepRingP1(Rg),ZENRingSizeElt(Rg)))
   ZepEltSetToZero(*pa,Rg);
 else
   ZBNAddCarry(ZepElt2Bn(*pa,Rg),al,(BigNumDigit)1);
}

/* \TEX{
   \title{Conversion.}
   \input{A pointer |pa| on a |ZENElt| of a modular ring |Rg|,
   an element |b| of a modular ring |R|.}
   \side{|*pa| is set to the modular conversion of |b|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   }
   */
int _ZepEltConvert(pa, Rg, b, R)
     ZENElt *pa,b;
     ZENRing Rg,R;
{
  BigNum p;
  BigNumLength pl;
  
  switch(ZENRingArith(R)) {
    
  /* Fractions
   *************/
  case ZEN_LIB_ZEF:
    pl = ZENRingSizeElt(Rg);

    if (!(p = ZBNC(pl))) {
      ZENSetError(ZEN_LIB_ZEP, ZEN_CONVERT, ZEN_ALLOC);
      return(ZERR);
    }
    ZENEltToZ(p, &pl, b, R);

    if ((ZENErrErrno == ZEN_INPUT) && (pl == ZENRingSizeElt(Rg))) {
      
      /* This element is a fraction, nothing can be done */
      ZENSetError(ZEN_LIB_ZEP, ZEN_CONVERT, ZEN_INPUT);
      return(ZERR);
    }


    ZBNF(p);
    if (!(p = ZBNC(pl+1))) {
      ZENSetError(ZEN_LIB_ZEP, ZEN_CONVERT, ZEN_ALLOC);
      return(ZERR);
    }
    ZENEltToZ(p, &pl, b, R);
    break;
    
    /* Specific cases where things can be done faster
     *************************************************/
  case ZEN_LIB_ZE2:
    if(ZENEltIsZero(b, R))
      ZepEltSetToZero(*pa, Rg);
    else
      ZepEltSetToOne(*pa, Rg);
    return(ZEN_HAS_INVERSE);

    /* Default
     **********/
  default:
    /* R is a finite field of same characteristic
       or a a prime finite field ? */
    if (!ZBNCompare(ZENRingP(R), ZENRingPl(R),
		    ZENRingP(Rg), ZENRingPl(Rg)) ||
	!ZENRingDeg(R)) {
      if (!(p = ZBNC(ZENRingQl(R)+1))) {
	ZENSetError(ZEN_LIB_ZEP, ZEN_CONVERT, ZEN_ALLOC);
	return(ZERR);
      }
      ZENEltToZ(p, &pl, b, R);
      break;
    }

    /* Otherwise, what can we do ? */
    ZENSetError(ZEN_LIB_ZEP, ZEN_CONVERT, ZEN_INPUT);
    return(ZERR);
  }    

  switch(ZBNCompareLazy(p, pl,
			ZENRingP(Rg), ZENRingPl(Rg))) {
  case BN_EQ:
    ZepEltSetToZero(*pa, Rg);
    break;
  case BN_LT:
    ZepZToElt(*pa, p, pl, Rg);
    break;
  default:
    p[pl] = 0;
    ZBNDivide(p, pl+1, ZENRingP(Rg), ZENRingPl(Rg));
    pl = ZBNNumDigits(p, ZENRingPl(Rg));
    ZepZToElt(*pa, p, pl, Rg);
  }
  ZBNF(p);
  return(ZEN_HAS_INVERSE);
}
