/*
 * zen/zep/c/elt_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{input/output}
   
   Conversions between the internal representation and the ASCII
   representation of a |ZENElt| are done by |_ZepEltPrintToString|. |_ZepEltPrintToFile|
   allows to write a |ZENElt| in a file.
   One can moreover write to and then read from a file a |ZENelt| with |_ZepEltPutToFile|
   and |_ZepEltReadFromFile|.
   }
   */
#include <stdio.h>
#include "zep.h"

#define GetReadFromStringFile(fct)                    \
 if(ret<0) return(ZERR);                                \
 if(bl>ZENRingSizeElt(Rg)) {                              \
   ZBNF(b);                                           \
   ZENSetError(ZEN_LIB_ZEP, fct, ZEN_SYNTAX);         \
   return(ZERR); }                                      \
 ZBNAssign(ZENElt2Bn(*G,Rg),b,bl);                    \
 ZBNF(b);                                             \
 if(bl == ZENRingSizeElt(Rg)) return(ret);                \
 ZBNSetToZero(ZENElt2Bn(*G,Rg)+bl,ZENRingSizeElt(Rg)-bl); \
 return(ret)


/* \TEX{
   \title{Converting from string.}
   \input{A finite field |Rg|, a pointer on a |ZENElt| |G| and a string
   |s| representing a |ZENElt| in base $|base|\in \{2,\ldots,
   16\}$.} 
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|G| is filled with |s| if the output is not |ZENERR|.}
}
*/
int _ZepEltReadFromString(G, s, base, Rg)
     ZENElt *G;
     char *s;
     int base;
     ZENRing Rg;
{
  BigNum b;
  BigNumLength bl;
  int ret;

  ret=ZBNReadFromString(&b,&bl,s,base);
  GetReadFromStringFile(ZEN_READFROMSTRING);
}


/* \TEX{
   \title{Converting to string.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{An allocated string representing |G| in base |base| or |ZENNULL| if an error occured.}   
   }
   */
char *_ZepEltPrintToString(G, base, Rg)
     ZENElt G;
     int base;
     ZENRing Rg;  
{
  return(ZBNPrintToString(ZENElt2Bn(G,Rg),ZBNNumDigits(ZENElt2Bn(G,Rg),ZENRingSizeElt(Rg)),base));
}


/* \TEX{
   \title{Converting from a string in an internal representation.}
   \input{A finite field |Rg| and a string |s| representing a |ZENElt|
   in an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|G| is filled with s.}
   }
   */
int _ZepEltGetFromString(G, s, Rg)
     char *s;
     ZENElt *G;
     ZENRing Rg;
{
  return(ZBNGetFromStringLazy(ZENElt2Bn(*G,Rg),ZENRingSizeElt(Rg),s));
}


/* \TEX{
   \title{Converting to string in an internal representation.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg|.}
   \output{An allocated string representing |G| in an internal representation or |ZENNULL| if an error occured.}   
   }
   */
char *_ZepEltPutToString(G, Rg)
     ZENElt G;
     ZENRing Rg;
{
  return(ZBNPutToStringLazy(ZENElt2Bn(G,Rg),ZENRingSizeElt(Rg)));
}


/* \TEX{
   \title{Printing to file.}
   \input{A stream |file|, an allocated |ZENElt| |G| of a finite field
	    |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
	    \output{|ZENERR| if an error occured, 0 otherwise.}
	    \side{Printing a representation of |G| in base |base| to |file|.}
	    }
	    */
int _ZepEltPrintToFile(file, G, base, Rg)
     FILE *file;
     ZENElt G;
     int base;
     ZENRing Rg;  
{
  return(ZBNPrintToFile(file,ZENElt2Bn(G,Rg),ZBNNumDigits(ZENElt2Bn(G,Rg),ZENRingSizeElt(Rg)),base));
}

/* \TEX{
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |ZENElt| |G| of a finite field
   |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |G| to |file|.}
   }
   */
int _ZepEltPutToFile(file, G, Rg)
     FILE *file;
     ZENElt G;
     ZENRing Rg;  
{
  return(ZBNPutToFileLazy(file,ZENElt2Bn(G,Rg),ZENRingSizeElt(Rg)));
}

/* \TEX{
   \title{Reading from file.}
   \input{A stream |file| and an element |G| of a finite field |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|G| is filled with the ZENElt read in |file| if no error occurred.}
   }
   */
int _ZepEltReadFromFile(G, file, base, Rg)
     ZENElt *G;
     FILE *file;
     int base;
     ZENRing Rg;
{BigNum b;
 BigNumLength bl;
 int ret;

 ret=ZBNReadFromFile(&b,&bl,file,base);
 GetReadFromStringFile(ZEN_READFROMFILE);
}

/* \TEX{
   \title{Getting from file in an internal representation.}
   \input{A stream |file|, and an element |G| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|G| is filled with the element read in |file|}
   }
   */
int _ZepEltGetFromFile(G, file, Rg)
     ZENElt *G;
     FILE *file;
     ZENRing Rg;
{
  return(ZBNGetFromFileLazy(ZENElt2Bn(*G,Rg),ZENRingSizeElt(Rg),file));
}

/* \TEX{
   \title{Evaluating an element as a multivariate polynomial over $Z$.}
   \input{A Bignum |p| of allocated size greater or equal than
   |ZENRingQl(Rg)|, a pointer |p_pl| on the real size of |p|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|p| is filled with the value of |G| considered as a
   multivariate polynomial in which we substitute all the variables
   with the characteristic, |*p_pl| contains the real size of |p|.}
   } */
void _ZepEltToZ(p, p_pl, G, Rg)
     BigNum p;
     BigNumLength *p_pl;
     ZENElt G;
     ZENRing Rg;
{
  ZepEltToZ(p, p_pl, G, Rg);
}

/* \TEX{
   \title{Getting an element from its evaluation as a multivariate
   polynomial over $Z$.}
   \input{A Bignum |p| of size |pl|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|G| is filled with the element whose evaluation considered as a
   multivariate polynomial in which we substitute all the variables
   with the characteristic is equal to (|p|, |pl|).}
   } */
void _ZepZToElt(G, p, pl, Rg)
     BigNum p;
     BigNumLength pl;
     ZENElt *G;
     ZENRing Rg;
{
  ZepZToElt(*G, p, pl, Rg);
}
