/*
 * zen/zep/c/elt_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element multiplication}
   
   Two functions are provided to do multiplications: |_ZepEltSquare| to square a
   |ZENElt| and |_ZepEltMultily| to compute the multiplication of two |ZENElt|.
   }
   */
#include <stdio.h>
#include "zep.h"

/* \TEX{
   \title{Squaring.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{$|*p_c| = a^2$.}
   \note{One can have |*p_c| = |a|.}
   }
   */
void _ZepEltSquare(p_c, a, R)
     ZENElt *p_c, a;
     ZENRing R;
{
  ZBNModSquare(ZepRingMbuf(R),ZepRingMbufl(R),
		 ZENElt2Bn(a,R),ZENRingSizeElt(R),
		 ZENRingP(R),ZENRingPl(R));
  ZBNAssign(ZENElt2Bn(*p_c,R),ZepRingMbuf(R),ZENRingSizeElt(R));
}

/* \TEX{
   \title{Multiplication.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |R|.}
   \side{$|*p_c| = a\times b$.}
   \note{One can have |*p_c| = |b|.}
   }
   */
void _ZepEltMultiply(p_c, a, b, R)
     ZENElt *p_c, a, b;
     ZENRing R;
{
  ZBNModMultiply(ZepRingMbuf(R),ZepRingMbufl(R),
		 ZENElt2Bn(a,R),ZENRingSizeElt(R),
		 ZENElt2Bn(b,R),ZENRingSizeElt(R),
		 ZENRingP(R),ZENRingPl(R));
  ZBNAssign(ZENElt2Bn(*p_c,R),ZepRingMbuf(R),ZENRingSizeElt(R));
}


/*********************************************************
  \TEX{
  \title{Adding precomputations.}
  \input{A finite ring |R| and the cutoffs.} 
  \output{|ZENERR| if an allocation error occured, 0 otherwise.}
  \side{Buffers are allocated and initialized for Karatsuba's operations.}
  \note{No test is performed}
  }
  */
static int __ZepPrcInitEltKara(R,cutM,cutS)
     ZENRing R;
     BigNumLength cutM,cutS;
{
  ZepRingKaraMLim(R) = cutM;

  if(ZepRingKaraMLim(R)>0)
    {
      ZepRingKaraBuf(R)=ZBNC(ZBNSizeBufKaraM(ZENRingSizeElt(R)));
      if(!ZepRingKaraBuf(R))
	{
	  ZENSetError(ZEN_LIB_ZEP, ZEN_KARAPRC, ZEN_ALLOC);
	  return(ZERR);
	}
      ZENEltMultiplyPtr(R) = _ZepEltMultiplyKaratsuba; /* Multiplication of elements */
    }
  else
    ZepRingKaraBuf(R)=NULL;
  
  ZepRingKaraSLim(R) = cutS;

  if(ZepRingKaraSLim(R)>0)
    {
      if(!ZepRingKaraBuf(R))
	{
	  ZepRingKaraBuf(R)=ZBNC(ZBNSizeBufKaraS(ZENRingSizeElt(R)));
	  if(!ZepRingKaraBuf(R))
	    {
	      ZENSetError(ZEN_LIB_ZEP, ZEN_KARAPRC, ZEN_ALLOC);
	      return(ZERR);
	    }
	}
      ZENEltSquarePtr(R) = _ZepEltSquareKaratsuba; /* Squaring of elements */
    }

  return(0);

}

/* \TEX{
   \title{Adding precomputations.}
   \input{A finite ring |R|.} 
   \output{|ZENERR| if an allocation error occured, 0 otherwise.}
   \side{Buffers are allocated and initialized for Karatsuba's operations.}
   }
   */
int _ZepPrcInitEltKara(R)
     ZENRing R;
{
  return(__ZepPrcInitEltKara(R,
			     ZBNKaratsubaMultiplyCutoff(ZENRingSizeElt(R)),
			     ZBNKaratsubaSquareCutoff(ZENRingSizeElt(R))));
}

/* \TEX{
   \title{Freeing precomputations.}
   \input{A finite ring |R|.} 
   \side{Buffers for Karatsuba's operations are freed.}
   }
   */
int _ZepPrcCloseEltKara(R)
     ZENRing R;
{
  if((ZepRingKaraMLim(R)>0) ||
     (ZepRingKaraSLim(R)>0))
    ZBNF(ZepRingKaraBuf(R));

  ZENEltMultiplyPtr(R) = _ZepEltMultiply; /* Multiplication of elements */
  ZENEltSquarePtr(R) = _ZepEltSquare;     /* Squaring of elements */
  return(0);
}

/* \TEX{
   \title{Squaring.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{$|*p_c| = a^2$.}
   \note{One can have |*p_c| = |a|.}
   }
   */
void _ZepEltSquareKaratsuba(p_c, a, R)
     ZENElt *p_c, a;
     ZENRing R;
{
 
  ZBNModSquareKaratsubaBuffer(ZepRingMbuf(R),ZepRingMbufl(R),
			      ZENElt2Bn(a,R),ZENRingSizeElt(R),
			      ZepRingKaraBuf(R),
			      ZENRingP(R),ZENRingPl(R));
  ZBNAssign(ZENElt2Bn(*p_c,R),ZepRingMbuf(R),ZENRingSizeElt(R));
}

/* \TEX{
   \title{Multiplication.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |R|.}
   \side{$|*p_c| = a\times b$.}
   \note{One can have |*p_c| = |b|.}
   }
   */
void _ZepEltMultiplyKaratsuba(p_c, a, b, R)
     ZENElt *p_c, a, b;
     ZENRing R;
{
  ZBNModMultiplyKaratsubaBuffer(ZepRingMbuf(R),ZepRingMbufl(R),
				ZENElt2Bn(a,R),ZENRingSizeElt(R),
				ZENElt2Bn(b,R),ZENRingSizeElt(R),
				ZepRingKaraBuf(R),
				ZENRingP(R),ZENRingPl(R));
  ZBNAssign(ZENElt2Bn(*p_c,R),ZepRingMbuf(R),ZENRingSizeElt(R));
}

