/*
 * zen/zep/c/init.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   A finite ring is completely described by a variable of type
   |ZENRing|. Precomputations are needed to speedup the operations in this
   ring. In the particular case of $ZEN(p)$, this is done with
   |_ZepInit|. These precomutations are then free with |_ZepClose|.
   }
   */
#include <stdio.h>
#include "zep.h"
#include "zeg.h"
#include "zed.h"


/* \TEX{
   \title{Initialization.}
   \input{A partially defined |ZENRing|.}
   \output{|ZENERR| if an error occured, $|SR_GEN|\vert |MAT_GEN|$ otherwise.}
   \side{The initialization of $\Z/p\Z$ is completed.}
   \note{This procedure is called by |ZENBaseRingCreate| and should not be called otherwise.}
   }
   */
int _ZepInit(R)
     ZENRing R;
{
  ZENRingArith(R) = ZEN_LIB_ZEP;

  ZepRingMask(R) = BlocExp2(SIZE_BLOC-ZBNNumLeadingZeroBitsInDigit(ZENRingP1(R)[ZENRingSizeElt(R)-1]))-1;
  
  /* Procedure on rings
   **********************/
#if KARA
  ZENPrcInitEltMultiplyPtr(R) = _ZepPrcInitEltKara;
  ZENPrcCloseEltMultiplyPtr(R) = _ZepPrcCloseEltKara;
#else
  ZENPrcInitEltMultiplyPtr(R) = NULL;
  ZENPrcCloseEltMultiplyPtr(R) = NULL;
#endif

  ZENPrcInitPolyRootsCanonicalPtr(R) = _ZepPrcInitPolyRootsCanonical;
  ZENPrcClosePolyRootsCanonicalPtr(R) = _ZepPrcClosePolyRootsCanonical;

  ZENPrcInitEltExpPtr(R) = _ZepPrcInitEltExp;
  ZENPrcCloseEltExpPtr(R) = _ZepPrcCloseEltExp;

  /* Procedure on elements
   ****************************/
  /* Allocation */
  ZENEltCreatePtr(R) = _ZepEltCreate;	/* Create elements */
  ZENEltCreateLargePtr(R) = _ZepEltCreateLarge;
  ZENEltFreePtr(R) = _ZepEltFree;	/* Free elements */

  /* Assigning */
  ZENEltAssignPtr(R) = _ZepEltAssign;	/* Assign an element with another */
  ZENEltSetToZeroPtr(R) = _ZepEltSetToZero; /* Set an element to zero */
  ZENEltSetToOnePtr(R) = _ZepEltSetToOne; /* Set an element to one */
  ZENEltSetRandomPtr(R) = _ZepEltSetRandom; /* Set randomly an element */
  ZENEltSetNextPtr(R) = _ZepEltSetNext; /* Set an element to the next */

  /* Test */
  ZENEltAreEqualPtr(R) = _ZepEltAreEqual; /* Is an element equal to another */
  ZENEltIsZeroPtr(R) = _ZepEltIsZero;	/* Is an element equal to zero */
  ZENEltIsOnePtr(R) = _ZepEltIsOne;	/* Is an element equal to one */
  ZENEltIsASquarePtr(R) = _ZENEltIsASquareCharp; /* Is an element a square ? */

  /* Arithmetic */
  ZENEltAddPtr(R) = _ZepEltAdd;	/* Addition of elements */
  ZENEltNegatePtr(R) = _ZepEltNegate;	/* Negation of an element */
  ZENEltSubtractPtr(R) = _ZepEltSubtract; /* Substraction of elements */
  ZENEltSquarePtr(R) = _ZepEltSquare;	/* Squaring of elements */
  ZENEltMultiplyPtr(R) = _ZepEltMultiply; /* Multiplication of elements */
  ZENEltInversePtr(R) = _ZepEltInverse; /* Inverse of elements */
  ZENEltTracePtr(R) = _ZepEltTrace;		/* Trace of elements */
  ZENEltSquareRootPtr(R) = _ZENEltSquareRootCharp; /* Square root of an element */
  ZENEltExpPtr(R) = _ZENEltExpPlain;	/* Exponentiation */

  /* io */
  ZENEltReadFromStringPtr(R) = _ZepEltReadFromString;	/* Read an element from a string */
  ZENEltPrintToStringPtr(R) = _ZepEltPrintToString; /* Print an element to a string */
  ZENEltReadFromFilePtr(R) = _ZepEltReadFromFile; /* Read an element from a file */
  ZENEltPrintToFilePtr(R) = _ZepEltPrintToFile; /* Print an element to a file */
  ZENEltGetFromStringPtr(R) = _ZepEltGetFromString; /* Get an element from a string */
  ZENEltPutToStringPtr(R) = _ZepEltPutToString; /* Put an element to a string */
  ZENEltGetFromFilePtr(R) = _ZepEltGetFromFile; /* Get an element from a file */
  ZENEltPutToFilePtr(R) = _ZepEltPutToFile; /* Put an element to a file */
  
  ZENZToEltPtr(R) = _ZepZToElt; /* Evaluate in $p$ an element */
  ZENEltToZPtr(R) = _ZepEltToZ; /* Get the element from an evaluation in $p$ */
  ZENEltConvertPtr(R) = _ZepEltConvert;

  /* Procedures on polynomials
   *******************************/
  /* Allocation */
  ZENPolyCreatePtr(R) = _ZepPolyCreate; /* Create polynomials */
  ZENPolyCreateLargePtr(R) = _ZepPolyCreateLarge; /* Create polynomials */
  ZENPolyFreePtr(R) = _ZepPolyFree;	/* Free polynomials */

  /* Assigning */
  ZENPolyAssignPtr(R) = _ZepPolyAssign; /* Assign a polynomial with another */
  ZENPolyAssignLargePtr(R) = _ZepPolyAssignLarge; /* Assign a polynomial with another */
  ZENPolyLazyToVeryLazyPtr(R) = _ZepPolyLazyToVeryLazy; /* Assign a polynomial with another */
  ZENPolySetToZeroPtr(R) = _ZepPolySetToZero;	/* Set a polynomial to zero */
  ZENPolySetToZeroLargePtr(R) = _ZepPolySetToZeroLarge;	/* Set a polynomial to zero */
  ZENPolySetToXiPtr(R) = _ZepPolySetToXi;	/* Set a polynomial to X^i */
  ZENPolySetRandomPtr(R) = _ZepPolySetRandom;	/* Set randomly a polynomial */
  ZENPolySetCoeffPtr(R) = _ZepPolySetCoeff; /* Set a coefficient of a polynomial */
  ZENPolyGetCoeffNewPtr(R) = _ZepPolyGetCoeff; /* Get a coefficient of a polynomial */
  ZENPolyGetCoeffPtrPtr(R) = _ZepPolyGetCoeffPtr; /* Get a pointer on coefficient of a polynomial */
  ZENPolyUpdateDegreePtr(R) = _ZepPolyUpdateDegree; /* Get a coefficient of a polynomial */

  /* Test */
  ZENPolyAreEqualPtr(R) = _ZepPolyAreEqual; /* Is a polynomial equal to another */
  ZENPolyIsZeroPtr(R) = _ZepPolyIsZero; /* Is a polynomial equal to zero */
  ZENPolyIsXiPtr(R) = _ZepPolyIsXi; /* Is a polynomial equal to $X^i$ */

  /* Arithmetic */
  ZENPolyAddPtr(R) = _ZepPolyAdd;	/* Addition of polynomials */
  ZENPolyAddLargePtr(R) = _ZepPolyAddLarge;	/* Addition of polynomials */
  ZENPolyNegatePtr(R) = _ZepPolyNegate; /* Negation of polynomials */
  ZENPolyNegateLargePtr(R) = _ZepPolyNegateLarge; /* Negation of polynomials */
  ZENPolySubtractPtr(R) = _ZepPolySubtract; /* Substraction of polynomials */
  ZENPolySubtractLargePtr(R) = _ZepPolySubtractLarge; /* Substraction of polynomials */

  ZENPolySquarePtr(R) = _ZepPolySquarePlain; /* Squaring of polynomials */
  ZENPolySquareLazyPtr(R) = _ZepPolySquareKaratsubaLazy;
  ZENPolySquareVeryLazyPtr(R) = _ZepPolySquareKaratsubaVeryLazy;

  ZENPolyMultiplyPtr(R) = _ZepPolyMultiplyPlain; /* Multiplication of polynomials */
  ZENPolyMultiplyLazyPtr(R) = _ZepPolyMultiplyLazy;
  ZENPolyMultiplyVeryLazyPtr(R) = _ZepPolyMultiplyVeryLazy;

  ZENPolyMultiplyScalarPtr(R) = _ZepPolyMultiplyScalar; /* Multiplication of polynomials by a scalar*/
  ZENPolyDotPtr(R) = _ZepPolyDot; /* Scalar multiplication of polynomials */
  ZENPolyEvalPtr(R) = _ZENPolyEval; /* Evaluation of a polynomial */
  ZENPolyDividePtr(R) = _ZepPolyDivide; /* Division of polynomials */
  ZENPolyDivideLazyPtr(R) = _ZepPolyDivideLazy; /* Lazy division of polynomials */
  ZENPolyDivideVeryLazyPtr(R) = _ZepPolyDivideVeryLazy; /* Lazy division of polynomials */
  ZENPolyModuloLazyPtr(R) = _ZepPolyModuloVeryLazy; /* Lazy modulo of polynomials */
  ZENPolyModuloVeryLazyPtr(R) = _ZepPolyModuloVeryLazy; /* Lazy modulo of polynomials */
  ZENPolyDerivePtr(R) = _ZENPolyDerive;
  ZENPolyRootsDegree2Ptr(R) = _ZENPolyRootsDegree2Charp; /* Roots of polynomials of degree 2 */

  /* Procedures on matrices
   **************************/
  /* Allocation */
  ZENMatCreateParamsPtr(R) = _ZepMatCreateParams; /* Create a matrix */
  ZENMatCreatePtr(R) = __ZedMatCreate;	/* Create a matrix */

  /* Assigning */
  ZENMatAssignPtr(R) = _ZepMatAssign; /* Assign a matrix with another */
  ZENPermutation2MatrixPtr(R) = _ZENPermutation2Matrix;	/* Assign a matrix with another */
  ZENMatSetToZeroPtr(R) = _ZepMatSetToZero; /* Set a matrix to zero */
  ZENMatSetToOnePtr(R) = _ZepMatSetToOne; /* Set a matrix to identity */
  ZENMatSetRandomPtr(R) = _ZENMatSetRandom; /* Set randomly a matrix */
  ZENMatSetBlocPtr(R) = _ZedMatSetBloc; /* Set a coefficient of a matrix */
  ZENMatGetBlocNewPtr(R) = _ZedMatGetBloc; /* Get a coefficient of a matrix */
  ZENMatGetBlocPtrPtr(R) = _ZedMatGetBlocPtr; /* Get a coefficient of a matrix */
  ZENMatSetSubMatPtr(R) = _ZedMatSetSubMat; /* Set a submatrix */
  ZENMatGetSubMatPtr(R) = _ZedMatGetSubMat; /* Get a submatrix */
  
  ZENMatPermuteBnPtr(R) = _ZENMatPermuteBn;
  ZENMatPermuteDigitPtr(R) = _ZENMatPermuteDigit;
  
  /* Test */
  ZENMatAreEqualPtr(R) = _ZepMatAreEqual; /* Is a matrix equal to another */
  ZENMatIsZeroPtr(R) = _ZepMatIsZero;	/* Is a matrix equal to zero */
  ZENMatIsOnePtr(R) = _ZepMatIsOne;	/* Is a matrix equal to one */
  
  /* Arithmetic */
  ZENMatAddPtr(R) = _ZENMatAdd;	/* Addition of Matrix */
  ZENMatAddScalarPtr(R) = _ZENMatAddScalar;	/* Addition of Matrix */
  ZENMatMultiplyScalarPtr(R) = _ZENMatMultiplyScalar;	/* Multiplication of Matrix */
  ZENMatSubtractPtr(R) = _ZENMatSubtract; /* Substraction of Matrix */
  ZENMatNegatePtr(R) = _ZENMatNegate;	/* Squaring of Matrix */
  ZENMatWinogradPtr(R) = _ZENMatWinograd; /* Multiplication of Matrix */
  ZENMatMultiplyPtr(R) = _ZENMatWinograd; /* Multiplication of Matrix */
  ZENMatMultiplyPlainPtr(R) = _ZENMatMultiplyPlain; /* Multiplication of Matrix */
  ZENMatMultiplyLeftPermutationPtr(R) = _ZENMatMultiplyLeftPermutation;
  ZENMatMultiplyRightPermutationPtr(R) = _ZENMatMultiplyRightPermutation;
  ZENMatInversePtr(R) = _ZENMatInverse; /* Inverse of a Matrix */
  ZENMatKernelPtr(R) = _ZENMatKernel;	/* Kernel of a Matrix */  
  ZENMatGaussPtr(R) = _ZENMatGauss;
  
  /* IO */
  ZENMatGetFromFilePtr(R) = _ZENMatGetFromFile;
  ZENMatPutToFilePtr(R) = _ZENMatPutToFile;
  ZENMatGetFromStringPtr(R) = _ZENMatGetFromString;
  ZENMatPutToStringPtr(R) = _ZENMatPutToString;

  ZENVect2MatPtr(R) = ZENMatAssignPtr(R);
  ZENMat2VectPtr(R) = ZENMatAssignPtr(R);

				/* ********* */
  ZepRingMbufl(R) = 3+ZENRingSizeElt(R)*ZENRingSizeElt(R);
  if (!(ZepRingMbuf(R) = ZBNC(ZepRingMbufl(R)))) {
    ZENSetError(ZEN_LIB_ZEP, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  if (!(ZepRingNbuf(R) = ZBNC(ZENRingSizeElt(R)))) {
    ZBNF(ZepRingMbuf(R));
    ZENSetError(ZEN_LIB_ZEP, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  if(ZENRingSizeElt(R)!=ZENRingPl(R)) {
    if (!(ZepRingIbuf(R) = ZBNC(ZENRingPl(R)))) {
      ZBNF(ZepRingMbuf(R)); ZBNF(ZepRingNbuf(R));
      ZENSetError(ZEN_LIB_ZEP, ZEN_INIT, ZEN_ALLOC);
      return(ZERR);
    }
    ZENEltInversePtr(R)=_ZepEltInverseMisc;
    ZENEltAddPtr(R)=_ZepEltAddMisc;
    }
  
  
  return(SR_GEN);
}

/* \TEX{
   \title{Closing.}
   \input{A partially freed |ZENRing|.}
   \output{0.}
   }
   */
int _ZepClose(R)		
     ZENRing R;
{
  /* The buffers */
  ZBNF(ZepRingMbuf(R));
  ZBNF(ZepRingNbuf(R));

  if(ZENRingSizeElt(R)!=ZENRingPl(R))
    ZBNF(ZepRingIbuf(R));
      
  return(0);
}
