/*
 * zen/zep/c/poly_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsubsection{Polynomial additions}
   
   One function is provided to do additions: |_ZepPolyAdd| to add a
   |ZENPoly| to another.
   }
   */
#include <stdio.h>
#include "zep.h"

#define ZBNPolyGetCof(P,i,R) ZENElt2Bn(ZepPolyGetCof((P),(i),(R)),(R))
#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

/* \TEX{
   \title{Addition.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| += |PX|.}
   }
   */
void _ZepPolyAdd(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int mindeg, i;

  if (ZepPolyDeg(RX, Rg) < ZepPolyDeg(PX, Rg)) {
    mindeg = ZepPolyDeg(RX, Rg);
    ZepPolyDeg(RX, Rg) =  ZepPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZepPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    ZepEltAdd(ZepPolyGetCof(RX, i, Rg), ZepPolyGetCof(PX, i, Rg), Rg);
  }
  for(i = mindeg+1; i <= ZepPolyDeg(PX, Rg); i++) {
    ZepEltAssign(ZepPolyGetCof(RX, i, Rg),
		   ZepPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZepPolyUpdateDegree(RX, Rg);
}

/* For internal use only, cf. poly_multiply. */
void _ZepPolyAddLarge(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int mindeg, i;
  
  mindeg = min(ZepPolyDeg(RX, Rg), ZepPolyDeg(PX, Rg));
  for(i = 0; i <= mindeg; i++) {
    ZBNAdd(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg),
	   ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), 0);
  }
  for(i = mindeg+1; i <= ZepPolyDeg(PX, Rg); i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg)+1);
  }
  for(i = max(i-1, ZepPolyDeg(RX, Rg)); i >= 0; i--) {
    if (!ZBNIsZero(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg))) {
      break;
    }
  }
  ZepPolyDeg(RX, Rg) = i;
}

/* \TEX{
   \title{Negation.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| = -|PX|.}
   }
   */
void _ZepPolyNegate(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZepPolyDeg(PX, Rg); i++) {
    ZepEltNegate(ZepPolyGetCof(RX, i, Rg), ZepPolyGetCof(PX, i, Rg), Rg);
  }
  ZepPolyDeg(RX, Rg) =  ZepPolyDeg(PX, Rg);
}
/* For internal use only, cf. poly_multiply. */
void _ZepPolyNegateLarge(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZepPolyDeg(PX, Rg); i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg), ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg)+1);
    ZBNComplement(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg));
    ZBNAddCarry(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 1);
  }
  ZepPolyDeg(RX, Rg) = ZepPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Subtraction.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| -= |PX|.}
   }
   */
void _ZepPolySubtract(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, mindeg;

  if (ZepPolyDeg(RX, Rg) < ZepPolyDeg(PX, Rg)) {
    mindeg = ZepPolyDeg(RX, Rg);
    ZepPolyDeg(RX, Rg) =  ZepPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZepPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    ZepEltSubtract(ZepPolyGetCof(RX, i, Rg), ZepPolyGetCof(PX, i, Rg), Rg);
  }
  for(i = mindeg+1; i <= ZepPolyDeg(PX, Rg); i++) {
    ZepEltNegate(ZepPolyGetCof(RX, i, Rg),
		   ZepPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZepPolyUpdateDegree(RX, Rg);
}


/* For internal use only, cf. poly_multiply. */
void _ZepPolySubtractLarge(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, mindeg;

  mindeg = min(ZepPolyDeg(RX, Rg), ZepPolyDeg(PX, Rg));
  for(i = 0; i <= mindeg; i++) {
    ZBNSubtract(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg),
		ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), 1);
  }
  for(i = mindeg+1; i <= ZepPolyDeg(PX, Rg); i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg), ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg)+1);
    ZBNComplement(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg));
    ZBNAddCarry(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 1);
  }
  for(i = max(i-1, ZepPolyDeg(RX, Rg)); i >= 0; i--) {
    if (!ZBNIsZero(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg))) {
      break;
    }
  }
  ZepPolyDeg(RX, Rg) = i;
}

