/*
 * zen/zep/c/poly_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsubsection{Polynomial assignment}
   
   One can set a |ZENPoly| to zero with |_ZepPolySetToZero|. Then we can
   assign a |ZENPoly| with an other with |_ZepPolyAssign|. Furthermore, we
   can set randomly a |ZENPoly| with |_ZepPolySetRandom|. 
   
   The functions |_ZepPolyGetCoeff| and |_ZepPolySetCoeff|
   enable us to get or set the coefficient of $X^i$,
   |_ZepPolyUpdateDegree| compute the real degree of a |ZENPoly| if it
   decreased.
   }
   */
#include <stdio.h>
#include "zep.h"

#define ZBNPolyGetCof(P,i,R) ZENElt2Bn(ZepPolyGetCof((P),(i),(R)),(R))

/* \TEX{
   \title{Assign.}
   \input{Two |ZENPoly| |PX| and |RX| of a finite ring |Rg|.}
   \side{|RX| is filled with |PX|.}
   }
   */
void _ZepPolyAssign(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZepPolyDeg(PX, Rg); i++) {
    ZepEltAssign(ZepPolyGetCof(RX, i, Rg),
		ZepPolyGetCof(PX, i, Rg),
		Rg);
  }
  ZepPolyDeg(RX, Rg) =  ZepPolyDeg(PX, Rg);
}

/* For internal use only, cf. poly_multiply. */
void _ZepPolyAssignLarge(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZepPolyDeg(PX, Rg); i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(PX, i, Rg),
	      ZENRingSZ(Rg)+1);
  }
  ZepPolyDeg(RX, Rg) =  ZepPolyDeg(PX, Rg);
}
/* For internal use only, cf. poly_multiply. */
void _ZepPolyLazyToVeryLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZepPolyDeg(PX, Rg); i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(PX, i, Rg),
	      ZENRingSizeElt(Rg));
    ZBNSetToZero(ZBNPolyGetCof(RX, i, Rg)+ZENRingSizeElt(Rg),
		 ZENRingSZ(Rg)+1-ZENRingSizeElt(Rg));
  }
  for(; i <= ZepPolyLgt(RX, Rg); i++) {
    ZBNSetToZero(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg)+1);
    
  }
  ZepPolyDeg(RX, Rg) =  ZepPolyDeg(PX, Rg);
}


/* \TEX{
   \title{Set to zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{|RX| is set to zero..}
   }
   */
void _ZepPolySetToZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  ZepPolyDeg(RX, Rg) =  0;
  ZepEltSetToZero(ZepPolyGetCof(RX, 0, Rg), Rg);
}

void _ZepPolySetToZeroLarge(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  int i;
  for(i = 0; i <= ZepPolyLgt(RX, Rg); i++) {
    ZBNSetToZero(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg)+1);
  }
  ZepPolyDeg(RX, Rg) =  0;
}

/* \TEX{
   \title{Set to $X^i$.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg| allocated at least for
   degree $|deg|$.}
   \side{|RX| is set to $X^{|deg|}$.}
   }
   */
void _ZepPolySetToXi(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  int i;

  ZepPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    ZepEltSetToZero(ZepPolyGetCof(RX, i, Rg), Rg);
  }
  ZepEltSetToOne(ZepPolyGetCof(RX, deg, Rg), Rg);
}

/* \TEX{
   \title{Set to random.}
   \input{A |ZENPoly| |RX| of degree |deg| of a finite ring |Rg|.}
   \side{|RX| is set randomly with degree |deg|.}
   }
   */
void _ZepPolySetRandom(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  int i;

  ZepPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    ZepEltSetRandom(ZepPolyGetCof(RX, i, Rg), Rg);
  }
  ZepEltSetRandom(ZepPolyGetCof(RX, deg, Rg), Rg);
  while(ZepEltIsZero(ZepPolyGetCof(RX, deg, Rg), Rg)) {
    ZepEltSetRandom(ZepPolyGetCof(RX, deg, Rg), Rg);
  }
}

/* \TEX{
   \title{Set a coeff.}
   \input{A |ZENPoly| |RX| and a |ZENElt| |b| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \side{The coeff of $X^d$ in |RX| is set to |b|.}
   }
   */
void _ZepPolySetCoeff(RX, d, b, Rg)
     ZENPoly RX;
     int d;
     ZENElt b;
     ZENRing Rg;
{
  ZepPolySetCoeff(RX, d, b, Rg);
}

/* \TEX{
   \title{Get a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The coeff of $X^d$ in |RX|.}
   }
   */
void _ZepPolyGetCoeff(p_c, RX, d, Rg)
     ZENElt *p_c;
     ZENPoly RX;
     int d;
     ZENRing Rg;
{ 
  ZepPolyGetCoeff(*p_c, RX, d, Rg);
}

/* \TEX{
   \title{Extract a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The pointer to the coeff of $X^d$ in |RX|.}
   }
   */
ZENElt _ZepPolyGetCoeffPtr(RX, d, Rg)
     ZENPoly RX;
     int d;
     ZENRing Rg;
{
  return(ZepPolyGetCoeffPtr(RX,d,Rg));
}

/* \TEX{
   \title{Update the degree of a polynomial.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{Resetting the degree of |P|, assuming it has decreased.}
   }
   */
void _ZepPolyUpdateDegree(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  int i;

  for(i = ZepPolyDeg(RX, Rg); i >= 0; i--) {
    if (!ZepEltIsZero(ZepPolyGetCoeffPtr(RX,i,Rg), Rg)) {
      break;
    }
  }
  if (i < 0) {
    ZepPolySetToZero(RX, Rg);
  }
  else {
    ZepPolyDeg(RX, Rg) =  i;
  }
}
