/*
 * zen/zep/c/poly_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsubsection{Polynomial multiplication}
   
   Two functions are provided to do multiplications: |_ZepPolySquare| to square a
   |ZENPoly| and |_ZepPolyMultily| to compute the multiplication of two |ZENPoly|.
   }
   */
#include <stdio.h>
#include "zep.h"
#include "zeg.h"

#define max(a,b) ((a) > (b) ? (a) : (b))
#define min(a,b) ((a) < (b) ? (a) : (b))

#define ZEP_POLY_MULTIPLY_CUTOFF 1
#define ZEP_POLY_SQUARE_CUTOFF 4

#define ZBNPolyGetCof(P,i,R) ZENElt2Bn(ZepPolyGetCof((P),(i),(R)),(R))
#define ZBNRingHbuf(R) ZENElt2Bn(ZENRingHbuf(R),(R))

/* Some miscaellenous functions
********************************/

static void ZepQuickMod(e, n, nl, Rg)
     ZENElt e;
     BigNum n;
     BigNumLength nl;
     ZENRing Rg;
{
  nl = ZBNNumDigits(n, nl);
  switch(ZBNCompareLazy(n, nl, ZENRingP(Rg),ZENRingPl(Rg))) {
  case BN_EQ:
    ZBNSetToZero(ZENElt2Bn(e,Rg), ZENRingSizeElt(Rg));
    break;
  case BN_LT:
    ZBNAssign(ZENElt2Bn(e,Rg), n, ZENRingSizeElt(Rg));
    break;
  default:
    ZBNDivide(n, nl+1,
	      ZENRingP(Rg), ZENRingPl(Rg));
    ZBNAssign(ZENElt2Bn(e,Rg), n, ZENRingSizeElt(Rg));
  }
}


/* Usual squaring routines
***************************/

int _ZepPolySquarePlainVeryLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, j, k;

  for(k = ZepPolyDeg(PX, Rg); k > 0; k--) {

    /* coefficient of X^(2k) */
    ZBNSetToZero(ZBNPolyGetCof(RX, 2*k, Rg), ZENRingSZ(Rg)+1);
    for(i = min(ZepPolyDeg(PX, Rg), 2*k);
	i > k;
	i--) {
      j = 2*k - i;
      ZBNMultiply(ZBNPolyGetCof(RX, 2*k, Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(PX, j, Rg), ZENRingSizeElt(Rg));
    }
    ZBNShiftLeft(ZBNPolyGetCof(RX, 2*k, Rg), ZENRingSZ(Rg), 1);
    ZBNSquare(ZBNPolyGetCof(RX, 2*k, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(PX, k, Rg), ZENRingSizeElt(Rg));

    /* coefficient of X^(2k-1) */
    ZBNSetToZero(ZBNPolyGetCof(RX, 2*k-1, Rg), ZENRingSZ(Rg)+1);
    for(i = min(ZepPolyDeg(PX, Rg), 2*k-1);
	i >= k;
	i--) {
      j = 2*k - i - 1;
      ZBNMultiply(ZBNPolyGetCof(RX, 2*k-1, Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(PX, j, Rg), ZENRingSizeElt(Rg));
    }
    ZBNShiftLeft(ZBNPolyGetCof(RX, 2*k-1, Rg), ZENRingSZ(Rg), 1);
  }

  /* coefficient of X^(0) */
  ZBNSetToZero(ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg)+1);
  ZBNSquare(ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg),
	    ZBNPolyGetCof(PX, 0, Rg), ZENRingSizeElt(Rg));

  ZENPolyDeg(RX, Rg) = 2*ZENPolyDeg(PX, Rg);
  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Squaring in characteristic $p\neq 2$.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZepPolySquarePlain(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, j, k;

  for(k = ZepPolyDeg(PX, Rg); k > 0; k--) {

    /* coefficient of X^(2k) */
    ZBNSetToZero(ZBNRingHbuf(Rg), ZENRingSZ(Rg)+1);
    for(i = min(ZepPolyDeg(PX, Rg), 2*k);
	i > k;
	i--) {
      j = 2*k - i;
      ZBNMultiply(ZBNRingHbuf(Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(PX, j, Rg), ZENRingSizeElt(Rg));
    }
    ZBNShiftLeft(ZBNRingHbuf(Rg), ZENRingSZ(Rg), 1);
    ZBNSquare(ZBNRingHbuf(Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(PX, k, Rg), ZENRingSizeElt(Rg));
    ZepQuickMod(ZepPolyGetCof(RX, 2*k, Rg),
		ZBNRingHbuf(Rg),
		ZENRingSZ(Rg), Rg);

    /* coefficient of X^(2k-1) */
    ZBNSetToZero(ZBNRingHbuf(Rg), ZENRingSZ(Rg)+1);
    for(i = min(ZepPolyDeg(PX, Rg), 2*k-1);
	i >= k;
	i--) {
      j = 2*k - i - 1;
      ZBNMultiply(ZBNRingHbuf(Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(PX, j, Rg), ZENRingSizeElt(Rg));
    }
    ZBNShiftLeft(ZBNRingHbuf(Rg), ZENRingSZ(Rg), 1);
    ZepQuickMod(ZepPolyGetCof(RX, 2*k-1, Rg),
		ZBNRingHbuf(Rg),
		ZENRingSZ(Rg), Rg);
  }

  /* coefficient of X^(0) */
  ZBNSetToZero(ZBNRingHbuf(Rg), ZENRingSZ(Rg)+1);
  ZBNSquare(ZBNRingHbuf(Rg), ZENRingSZ(Rg),
	    ZBNPolyGetCof(PX, 0, Rg), ZENRingSizeElt(Rg));
  ZepQuickMod(ZepPolyGetCof(RX, 0, Rg),
	      ZBNRingHbuf(Rg),
	      ZENRingSZ(Rg), Rg);
  
  ZENPolyDeg(RX, Rg) = 2*ZENPolyDeg(PX, Rg);
  return(ZEN_HAS_INVERSE);
}

static void ZepPolySquareKaratsubaVeryLazyBis ___((ZENPoly,ZENPoly,ZENRing));

/* Karatsuba squaring routines
*******************************/

/* \TEX{
   \title{Squaring with Karatsuba.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZepPolySquareKaratsuba(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;
  ZENPoly RBuf;

  if (!(RBuf = ZepPolyCreateLarge(4*ZepPolyDeg(PX, Rg)+1, Rg))) {
    return(ZERR);
  }
  ZepPolySquareKaratsubaVeryLazy(RBuf, PX, Rg);
  for(i = ZepPolyDeg(RBuf, Rg); i >= 0; i--) {
    ZepQuickMod(ZepPolyGetCof(RBuf, i, Rg),
		ZBNPolyGetCof(RBuf, i, Rg), 
		2*ZENRingSizeElt(Rg)+1, Rg);
  }
  ZepPolyAssign(RX, RBuf, Rg);
  ZepPolyFree(RBuf, Rg);
  return(ZEN_HAS_INVERSE);
}

int _ZepPolySquareKaratsubaLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  ZepPolySquareKaratsubaVeryLazy(RX, PX, Rg);
  for(i = ZepPolyDeg(RX, Rg); i >= 0; i--) {
    ZepQuickMod(ZepPolyGetCof(RX, i, Rg),
		ZBNPolyGetCof(RX, i, Rg), 
		2*ZENRingSizeElt(Rg)+1, Rg);
  }
  return(ZEN_HAS_INVERSE);
}

static void ZepPolySquarePlainVeryLazyBis(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, j, k;

  for(k = ZepPolyDeg(PX, Rg); k > 0; k--) {

    /* coefficient of X^(2k) */
    ZBNSetToZero(ZBNPolyGetCof(RX, 2*k, Rg), ZENRingSZ(Rg)+1);
    for(i = min(ZepPolyDeg(PX, Rg), 2*k);
	i > k;
	i--) {
      j = 2*k - i;
      ZBNMultiply(ZBNPolyGetCof(RX, 2*k, Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg)+1,
		  ZBNPolyGetCof(PX, j, Rg), ZENRingSizeElt(Rg)+1);
    }
    ZBNShiftLeft(ZBNPolyGetCof(RX, 2*k, Rg), ZENRingSZ(Rg), 1);
    ZBNSquare(ZBNPolyGetCof(RX, 2*k, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(PX, k, Rg), ZENRingSizeElt(Rg)+1);

    /* coefficient of X^(2k-1) */
    ZBNSetToZero(ZBNPolyGetCof(RX, 2*k-1, Rg), ZENRingSZ(Rg)+1);
    for(i = min(ZepPolyDeg(PX, Rg), 2*k-1);
	i >= k;
	i--) {
      j = 2*k - i - 1;
      ZBNMultiply(ZBNPolyGetCof(RX, 2*k-1, Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg)+1,
		  ZBNPolyGetCof(PX, j, Rg), ZENRingSizeElt(Rg)+1);
    }
    ZBNShiftLeft(ZBNPolyGetCof(RX, 2*k-1, Rg), ZENRingSZ(Rg), 1);
  }

  /* coefficient of X^(0) */
  ZBNSetToZero(ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg)+1);
  ZBNSquare(ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg),
	    ZBNPolyGetCof(PX, 0, Rg), ZENRingSizeElt(Rg)+1);
  
  ZepPolyDeg(RX, Rg) = 2*ZepPolyDeg(PX, Rg);
}

static void ZepPolySquareKaratsubaVeryLazyDegreeBis(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  struct zenpoly UP1, UP3;
  ZENPoly up1, up3;
  int m, n, i, j;
  
  /* Initialization ***********/
  up1 = &UP1; up3 = &UP3;
  m = (ZepPolyDeg(PX, Rg)+1)/2; 
  n = ZepPolyDeg(PX, Rg)+1-m;
  
  /* let's go with the recursivity **********/
  
  /* PX- ^ 2 -> RX */
  ZepPolyDeg(up1, Rg) = m-1;
  ZepPolyCof(up1, Rg) = ZepPolyCof(PX, Rg);
  ZepPolyCof(up3, Rg) = ZepPolyCof(RX, Rg);
  ZepPolySquareKaratsubaVeryLazyBis(up3, up1, Rg);
  
  /* PX- + PX+ -> RX+4*m+2*n-2 */
  for(j = 0, i = 4*m+2*n-2;
      j < m;
      j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(PX, j+m, Rg),
	      ZENRingSizeElt(Rg)+1);
    ZBNAdd(ZBNPolyGetCof(RX, i, Rg), ZENRingSizeElt(Rg)+1,
	   ZBNPolyGetCof(PX, j, Rg), ZENRingSizeElt(Rg)+1, 0);
  }
  if (n > m) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(PX, j+m, Rg),
	      ZENRingSizeElt(Rg)+1);
  }
  
  /* (PX- + PX+) ^ 2 -> RX+2*m-1 */
  ZepPolyDeg(up1, Rg) = n-1;
  ZepPolyCof(up1, Rg) = ZepPolyCof(RX, Rg)+4*m+2*n-2;
  ZepPolyCof(up3, Rg) = ZepPolyCof(RX, Rg)+2*m-1;
  ZepPolySquareKaratsubaVeryLazyBis(up3, up1, Rg);
  
  /* (PX- + PX+) ^ 2 - PX- ^ 2 -> RX+2*m-1 */
  for(j = 0, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZBNSubtract(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg),
		ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg), 1);
  }
  
  /* [ PX- ^ 2, (PX- + PX+) ^ 2 - PX- ^ 2 ] -> RX+m */
  for(j = m, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZBNAdd(ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg),
	   ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 0);
  }
  for(;j < n+2*m-1; j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, j, Rg), ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg));
  }
  
  /* PX+ ^ 2 -> RX+ */
  ZepPolyDeg(up1, Rg) = n-1;
  ZepPolyCof(up1, Rg) = ZepPolyCof(PX, Rg)+m;
  ZepPolyCof(up3, Rg) = ZepPolyCof(RX, Rg)+2*m+n-1;
  ZepPolySquareKaratsubaVeryLazyBis(up3, up1, Rg);

  /* [ PX- ^ 2, (PX- + PX+) ^ 2 - PX- ^2 - PX+ ^ 2 , PX+ ^ 2 ] -> RX */
  for(j = m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZBNSubtract(ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg),
		ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 1);
  }
  for(j = 2*m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZBNAdd(ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg),
	   ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 0);
  }
  for(; j < 2*m+2*n-1; j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, j, Rg),
	      ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg));
  }

  ZepPolyDeg(RX, Rg) = 2*ZepPolyDeg(PX, Rg);
}

static void ZepPolySquareKaratsubaVeryLazyDegree(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  struct zenpoly UP1, UP3;
  ZENPoly up1, up3;
  int m, n, i, j;
  
  /* Initialization ***********/
  up1 = &UP1; up3 = &UP3;
  m = (ZepPolyDeg(PX, Rg)+1)/2; 
  n = ZepPolyDeg(PX, Rg)+1-m;
  
  /* let's go with the recursivity **********/
  
  /* PX- ^ 2 -> RX */
  ZepPolyDeg(up1, Rg) = m-1;
  ZepPolyCof(up1, Rg) = ZepPolyCof(PX, Rg);
  ZepPolyCof(up3, Rg) = ZepPolyCof(RX, Rg);
  ZepPolySquareKaratsubaVeryLazy(up3, up1, Rg);
  
  /* PX- + PX+ -> RX+4*m+2*n-2 */
  for(j = 0, i = 4*m+2*n-2;
      j < m;
      j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(PX, j+m, Rg),
	      ZENRingSizeElt(Rg));
    ZBNPolyGetCof(RX, i, Rg)[ZENRingSizeElt(Rg)] = 0;
    ZBNAdd(ZBNPolyGetCof(RX, i, Rg), ZENRingSizeElt(Rg)+1,
	   ZBNPolyGetCof(PX, j, Rg), ZENRingSizeElt(Rg), 0);
  }
  if (n > m) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(PX, j+m, Rg),
	      ZENRingSizeElt(Rg));
    ZBNPolyGetCof(RX, i, Rg)[ZENRingSizeElt(Rg)] = 0;
  }
  
  /* (PX- + PX+) ^ 2 -> RX+2*m-1 */
  ZepPolyDeg(up1, Rg) = n-1;
  ZepPolyCof(up1, Rg) = ZepPolyCof(RX, Rg)+4*m+2*n-2;
  ZepPolyCof(up3, Rg) = ZepPolyCof(RX, Rg)+2*m-1;
  ZepPolySquareKaratsubaVeryLazyBis(up3, up1, Rg);
  
  /* (PX- + PX+) ^ 2 - PX- ^ 2  -> RX+2*m-1 */
  for(j = 0, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZBNSubtract(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg),
		ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg), 1);
  }
  
  /* [ PX- ^ 2, (PX- + PX+) ^ 2 - PX- ^ 2 ] -> RX+m */
  for(j = m, i = 2*m-1; j < 2*m-1; j++, i++) {
    ZBNAdd(ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg),
	   ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 0);
  }
  for(;j < n+2*m-1; j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, j, Rg), ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg));
  }
  
  /* PX+ ^ 2 -> RX+ */
  ZepPolyDeg(up1, Rg) = n-1;
  ZepPolyCof(up1, Rg) = ZepPolyCof(PX, Rg)+m;
  ZepPolyCof(up3, Rg) = ZepPolyCof(RX, Rg)+2*m+n-1;
  ZepPolySquareKaratsubaVeryLazy(up3, up1, Rg);

  /* [ PX- ^ 2, (PX- + PX+) ^ 2 - PX- ^2 - PX+ ^ 2 , PX+ ^ 2 ] -> RX */
  for(j = m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZBNSubtract(ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg),
		ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 1);
  }
  for(j = 2*m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZBNAdd(ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg),
	   ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 0);
  }
  for(; j < 2*m+2*n-1; j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, j, Rg),
	      ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg));
  }

  ZepPolyDeg(RX, Rg) = 2*ZepPolyDeg(PX, Rg);
}

/* \TEX{
   \title{VeryLazy Squaring with Karatsuba.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0 for compatibylity.}
   \side{$|RX| = |PX|^2$.}
   \note{One must have $|RX|\neq|PX|$. |PX| is destroyed at the return.}
   }
   */
int ZepPolySquareKaratsubaVeryLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  if (ZepPolyDeg(PX, Rg) < ZEP_POLY_SQUARE_CUTOFF) {
    ZepPolySquarePlainVeryLazy(RX, PX, Rg);
    return(ZEN_HAS_INVERSE);
  }
  ZepPolySquareKaratsubaVeryLazyDegree(RX, PX, Rg);
  return(ZEN_HAS_INVERSE);
}

static void ZepPolySquareKaratsubaVeryLazyBis(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  if (ZepPolyDeg(PX, Rg) < ZEP_POLY_SQUARE_CUTOFF) {
    ZepPolySquarePlainVeryLazyBis(RX, PX, Rg);
    return;
  }
  ZepPolySquareKaratsubaVeryLazyDegreeBis(RX, PX, Rg);
}


/* General multiplication routines
**********************************/

/* \TEX{
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   }
   */
int _ZepPolyMultiply(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZENPoly RBuf;
  int n, m, q, r, L, k;

  if (ZepPolyDeg(PX, Rg) > ZepPolyDeg(QX, Rg)) {
    n = ZepPolyDeg(PX, Rg)+1; m = ZepPolyDeg(QX, Rg)+1;
    q = n/m; r = n%m;
  }
  else {
    n = ZepPolyDeg(QX, Rg)+1; m = ZepPolyDeg(PX, Rg)+1;
    q = n/m; r = n%m;
  }
  switch(q) {
  case 1:
    L = r ? 4*m-4+2*r : 4*m-3;
    break;
  case 2:
    L = 2*r > (m+1) ? 5*m-5+2*r : 6*m-4;
    break;
  default:
    L = 2*r > m ?  (3+q)*m-4+2*r: (4+q)*m-4;
    break;
  }
  if (!(RBuf = ZepPolyCreateLarge(L, Rg))) {
    return(ZERR);
  }
  ZepPolyMultiplyVeryLazy(RBuf, PX, QX, Rg);
  for(k = ZepPolyDeg(RBuf, Rg); k >= 0; k--) {
    ZepQuickMod(ZepPolyGetCof(RBuf, k, Rg),
		ZBNPolyGetCof(RBuf, k, Rg), 
		2*ZENRingSizeElt(Rg)+1, Rg);
  }
  ZepPolyAssign(RX, RBuf, Rg);
  ZepPolyFree(RBuf, Rg);

  return(ZEN_HAS_INVERSE);
}

int _ZepPolyMultiplyLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int k;

  ZepPolyMultiplyVeryLazy(RX, PX, QX, Rg);
  for(k = ZepPolyDeg(RX, Rg); k >= 0; k--) {
    ZepQuickMod(ZepPolyGetCof(RX, k, Rg),
		ZBNPolyGetCof(RX, k, Rg), 
		2*ZENRingSizeElt(Rg)+1, Rg);
  }
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{VeryLazy multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0 for compatibylity.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|\neq|QX|$.}
   }
   */
int _ZepPolyMultiplyVeryLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  struct zenpoly sxzp, rxzp, pxzp, qxzp;
  ZENPoly sx, rx, px, qx;
  int i, dp, dq, quo, rem;

  /* Switch
  **********/
  if (ZepPolyDeg(PX, Rg) == ZepPolyDeg(QX, Rg)) {
    ZepPolyMultiplyKaratsubaVeryLazy(RX, PX, QX, Rg);
    return(ZEN_HAS_INVERSE);
  }

  if (ZepPolyDeg(PX, Rg) > ZepPolyDeg(QX, Rg)) {
    if ((ZepPolyDeg(QX, Rg)) <= ZEP_POLY_MULTIPLY_CUTOFF) {
      ZepPolyMultiplyPlainVeryLazy(RX, PX, QX, Rg);
      return(ZEN_HAS_INVERSE);
    }
    px = &pxzp;
    ZepPolyCof(px, Rg) = ZepPolyCof(PX, Rg);
    dp = ZepPolyDeg(PX, Rg);

    qx = &qxzp;
    ZepPolyCof(qx, Rg) = ZepPolyCof(QX, Rg);
    dq = ZepPolyDeg(QX, Rg);
  }
  else {
    if (ZepPolyDeg(PX, Rg) <= ZEP_POLY_MULTIPLY_CUTOFF) {
      ZepPolyMultiplyPlainVeryLazy(RX, PX, QX, Rg);
      return(ZEN_HAS_INVERSE);
    }
    px = &pxzp;
    ZepPolyCof(px, Rg) = ZepPolyCof(QX, Rg);
    dp = ZepPolyDeg(QX, Rg);

    qx = &qxzp;
    ZepPolyCof(qx, Rg) = ZepPolyCof(PX, Rg);
    dq = ZepPolyDeg(PX, Rg);
  }

  /* We have to cut px in pieces 
    **********************************/  
  /* Initialization */
  ZepPolyDeg(RX, Rg) = dp+dq;

  rx = &rxzp;
  ZepPolyCof(rx, Rg) = ZepPolyCof(RX, Rg);

  sx = &sxzp;
  ZepPolyCof(sx, Rg) = ZepPolyCof(RX, Rg);

  quo = (dp+1)/(dq+1);
  rem = (dp+1)%(dq+1);

  /* First multiplication */
  ZepPolyDeg(px, Rg) = dq;
  ZepPolyDeg(qx, Rg) = dq;
  ZepPolyMultiplyKaratsubaVeryLazy(rx, qx, px, Rg);
  if ((quo == 1) && !rem) {
    return(ZEN_HAS_INVERSE);
  }

  /* Second multiplication */
  ZepPolyCof(px, Rg) += dq+1;
  ZepPolyDeg(px, Rg) = (quo == 1 ? rem-1: dq);
  ZepPolyCof(rx, Rg) += 2*dq+1;
  if (quo == 1) {
    ZepPolyMultiplyVeryLazy(rx, qx, px, Rg);
  }
  else {
    ZepPolyMultiplyKaratsubaVeryLazy(rx, qx, px, Rg);
  }
  ZepPolyDeg(sx, Rg) = dq-1;
  ZepPolyCof(sx, Rg) += dq+1;
  ZepPolyAddLarge(sx, rx, Rg);
  if (quo == 1) {
    return(ZEN_HAS_INVERSE);
  }

  for(i = 2; i < quo; i++) {
    ZepPolyCof(px, Rg) += (dq+1);
    ZepPolyCof(rx, Rg) += (dq+1);

    ZepPolyMultiplyKaratsubaVeryLazy(rx, qx, px, Rg);
    
    ZepPolyDeg(sx, Rg) = dq-1;
    ZepPolyCof(sx, Rg) += dq+1;
    ZepPolyAddLarge(sx, rx, Rg);
  }

  if (!rem) {
    return(ZEN_HAS_INVERSE);
  }

  ZepPolyDeg(px, Rg) = rem-1;
  ZepPolyCof(px, Rg) += (dq+1);
  ZepPolyCof(rx, Rg) += (dq+1);
  ZepPolyMultiplyVeryLazy(rx, qx, px, Rg);
    
  ZepPolyDeg(sx, Rg) = dq-1;
  ZepPolyCof(sx, Rg) += dq+1;
  ZepPolyAddLarge(sx, rx, Rg);

  return(ZEN_HAS_INVERSE);
}

/* Specific classical multiplication routines
*********************************************/

int ZepPolyMultiplyPlainVeryLazy(RX, QX, PX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i, j, k;

  if (ZepPolyIsZero(PX, Rg) || ZepPolyIsZero(QX, Rg)) {
    ZBNSetToZero(ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg)+1);
    ZENPolyDeg(RX, Rg) = 0;
    return(ZEN_HAS_INVERSE);
  }

  for(k = ZepPolyDeg(PX, Rg)+ZepPolyDeg(QX, Rg);
      k >= 0;
      k--) {
    ZBNSetToZero(ZBNPolyGetCof(RX, k, Rg), ZENRingSZ(Rg)+1);
    for(i = min(ZepPolyDeg(PX, Rg), k);
	i >= max(0, k - ZepPolyDeg(QX, Rg));
	i--) {
      j = k - i;
      ZBNMultiply(ZBNPolyGetCof(RX, k, Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(QX, j, Rg), ZENRingSizeElt(Rg));
    }
  }
  ZepPolyDeg(RX, Rg) = ZepPolyDeg(PX, Rg) + ZepPolyDeg(QX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|\times|QX|$.}
   }
   */
int _ZepPolyMultiplyPlain(RX, QX, PX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i, j, k;

  if (ZepPolyIsZero(PX, Rg) || ZepPolyIsZero(QX, Rg)) {
    ZBNSetToZero(ZBNPolyGetCof(RX, 0, Rg), ZENRingSizeElt(Rg));
    ZENPolyDeg(RX, Rg) = 0;
    return(ZEN_HAS_INVERSE);
  }

  for(k = ZepPolyDeg(PX, Rg)+ZepPolyDeg(QX, Rg);
      k >= 0;
      k--) {
    ZBNSetToZero(ZBNRingHbuf(Rg), ZENRingSZ(Rg)+1);
    for(i = min(ZepPolyDeg(PX, Rg), k);
	i >= max(0, k - ZepPolyDeg(QX, Rg));
	i--) {
      j = k - i;
      ZBNMultiply(ZBNRingHbuf(Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(QX, j, Rg), ZENRingSizeElt(Rg));
    }
    ZepQuickMod(ZepPolyGetCof(RX, k, Rg),
		ZBNRingHbuf(Rg),
		ZENRingSZ(Rg), Rg);
  }
  ZepPolyDeg(RX, Rg) = ZepPolyDeg(PX, Rg) + ZepPolyDeg(QX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* Specific Karatsuba multiplication routines
*********************************************/

static void ZepPolyMultiplyKaratsubaVeryLazyBis ___((ZENPoly,ZENPoly,ZENPoly,ZENRing));

/* \TEX{
   \title{Polynomial multiplication with the Karatsuba method.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an errpr occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   }
   */
int _ZepPolyMultiplyKaratsuba(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i;
  ZENPoly QBuf, RBuf;
  ZENElt elt;

  if (ZepPolyDeg(PX, Rg) >= ZepPolyDeg(QX, Rg)) {
    if (!(QBuf = ZepPolyCreate(ZepPolyDeg(PX, Rg), Rg))) {
      return(ZERR);
    }
    ZepPolyAssign(QBuf, QX, Rg);
    for(i = ZepPolyDeg(QX, Rg)+1;
	i <= ZepPolyDeg(PX, Rg);
	i++) {
      elt = ZepPolyGetCoeffPtr(QBuf, i, Rg);
      ZepEltSetToZero(elt, Rg);
      ZepPolySetCoeff(QBuf, i, elt, Rg);
    }

    if (!(RBuf = ZepPolyCreateLarge(4*ZepPolyDeg(PX, Rg)+1, Rg))) {
      return(ZERR);
    }

    ZepPolyMultiplyKaratsubaVeryLazy(RBuf, PX, QBuf, Rg);
  }
  else {
    if (!(QBuf = ZepPolyCreate(ZepPolyDeg(QX, Rg), Rg))) {
      return(ZERR);
    }
    ZepPolyAssign(QBuf, PX, Rg);
    for(i = ZepPolyDeg(PX, Rg)+1;
	i <= ZepPolyDeg(QX, Rg);
	i++) {
      elt = ZepPolyGetCoeffPtr(QBuf, i, Rg);
      ZepEltSetToZero(elt, Rg);
      ZepPolySetCoeff(QBuf, i, elt, Rg);
    }

    if (!(RBuf = ZepPolyCreateLarge(4*ZepPolyDeg(PX, Rg)+1, Rg))) {
      return(ZERR);
    }

    ZepPolyMultiplyKaratsubaVeryLazy(RBuf, QX, QBuf, Rg);
  }
  for(i = ZepPolyDeg(RBuf, Rg); i >= 0; i--) {
    ZepQuickMod(ZepPolyGetCof(RBuf, i, Rg),
		ZBNPolyGetCof(RBuf, i, Rg), 
		2*ZENRingSizeElt(Rg)+1, Rg);
  }
  ZepPolyAssign(RX, RBuf, Rg);
  ZepPolyFree(QBuf, Rg); ZepPolyFree(RBuf, Rg);
  return(ZEN_HAS_INVERSE);
}

int _ZepPolyMultiplyKaratsubaLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  if (ZepPolyDeg(PX, Rg) >= ZepPolyDeg(QX, Rg)) {
    for(i = ZepPolyDeg(QX, Rg)+1;
	i <= ZepPolyDeg(PX, Rg);
	i++) {
      elt = ZepPolyGetCoeffPtr(QX, i, Rg);
      ZepEltSetToZero(elt, Rg);
      ZepPolySetCoeff(QX, i, elt, Rg);
    }


  }
  else {
    for(i = ZepPolyDeg(PX, Rg)+1;
	i <= ZepPolyDeg(QX, Rg);
	i++) {
      elt = ZepPolyGetCoeffPtr(PX, i, Rg);
      ZepEltSetToZero(elt, Rg);
      ZepPolySetCoeff(PX, i, elt, Rg);
    }
  }
  ZepPolyMultiplyKaratsubaVeryLazy(RX, PX, QX, Rg);
  for(i = ZepPolyDeg(RX, Rg); i >= 0; i--) {
    ZepQuickMod(ZepPolyGetCof(RX, i, Rg),
		ZBNPolyGetCof(RX, i, Rg), 
		2*ZENRingSizeElt(Rg)+1, Rg);
  }
  return(ZEN_HAS_INVERSE);
}

static void ZepPolyMultiplyKaratsubaVeryLazyDegree0(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZBNSetToZero(ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg)+1);
  ZBNMultiply(ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(PX, 0, Rg), ZENRingSizeElt(Rg),
	      ZBNPolyGetCof(QX, 0, Rg), ZENRingSizeElt(Rg));
  ZepPolyDeg(RX, Rg) = 0;
}

static void ZepPolyMultiplyKaratsubaVeryLazyDegree0Bis(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZBNSetToZero(ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg)+1);
  ZBNMultiply(ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(PX, 0, Rg), ZENRingSizeElt(Rg)+1,
	      ZBNPolyGetCof(QX, 0, Rg), ZENRingSizeElt(Rg)+1);
  ZepPolyDeg(RX, Rg) = 0;
}

static void ZepPolyMultiplyKaratsubaVeryLazyDegree1(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZepEltAssign(ZepPolyGetCof(RX, 0, Rg), ZepPolyGetCof(PX, 0, Rg), Rg);
  ZBNSetToZero(ZBNPolyGetCof(RX, 0, Rg)+ZENRingSizeElt(Rg),
	       ZENRingSZ(Rg)+1-ZENRingSizeElt(Rg));
  ZBNAdd(ZBNPolyGetCof(RX, 0, Rg), ZENRingSizeElt(Rg)+1,
	 ZBNPolyGetCof(PX, 1, Rg), ZENRingSizeElt(Rg), 0);
  
  ZepEltAssign(ZepPolyGetCof(RX, 2, Rg), ZepPolyGetCof(QX, 0, Rg), Rg);
  ZBNSetToZero(ZBNPolyGetCof(RX, 2, Rg)+ZENRingSizeElt(Rg),
	       ZENRingSZ(Rg)+1-ZENRingSizeElt(Rg));
  ZBNAdd(ZBNPolyGetCof(RX, 2, Rg), ZENRingSizeElt(Rg)+1,
	 ZBNPolyGetCof(QX, 1, Rg), ZENRingSizeElt(Rg), 0);
  
  ZBNSetToZero(ZBNPolyGetCof(RX, 1, Rg), ZENRingSZ(Rg)+1);
  ZBNMultiply(ZBNPolyGetCof(RX, 1, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(RX, 0, Rg), ZENRingSizeElt(Rg)+1,
	      ZBNPolyGetCof(RX, 2, Rg), ZENRingSizeElt(Rg)+1);
  
  ZBNSetToZero(ZBNPolyGetCof(RX, 0, Rg), ZENRingSizeElt(Rg)+1);
  ZBNMultiply(ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(PX, 0, Rg), ZENRingSizeElt(Rg),
	      ZBNPolyGetCof(QX, 0, Rg), ZENRingSizeElt(Rg));
  
  ZBNSetToZero(ZBNPolyGetCof(RX, 2, Rg), ZENRingSizeElt(Rg)+1);
  ZBNMultiply(ZBNPolyGetCof(RX, 2, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(PX, 1, Rg), ZENRingSizeElt(Rg),
	      ZBNPolyGetCof(QX, 1, Rg), ZENRingSizeElt(Rg));
  
  ZBNSubtract(ZBNPolyGetCof(RX, 1, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg), 1);
  ZBNSubtract(ZBNPolyGetCof(RX, 1, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(RX, 2, Rg), ZENRingSZ(Rg), 1);

  ZepPolyDeg(RX, Rg) = 2;
}

static void ZepPolyMultiplyKaratsubaVeryLazyDegree1Bis(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZBNAssign(ZBNPolyGetCof(RX, 0, Rg),
	    ZBNPolyGetCof(PX, 0, Rg),
	    ZENRingSizeElt(Rg)+1);
  ZBNSetToZero(ZBNPolyGetCof(RX, 0, Rg)+ZENRingSizeElt(Rg)+1,
	       ZENRingSZ(Rg)-ZENRingSizeElt(Rg));
  ZBNAdd(ZBNPolyGetCof(RX, 0, Rg), ZENRingSizeElt(Rg)+1,
	 ZBNPolyGetCof(PX, 1, Rg), ZENRingSizeElt(Rg)+1, 0);
  
  ZBNAssign(ZBNPolyGetCof(RX, 2, Rg),
	    ZBNPolyGetCof(QX, 0, Rg),
	    ZENRingSizeElt(Rg)+1);
  ZBNSetToZero(ZBNPolyGetCof(RX, 2, Rg)+ZENRingSizeElt(Rg)+1,
	       ZENRingSZ(Rg)-ZENRingSizeElt(Rg));
  ZBNAdd(ZBNPolyGetCof(RX, 2, Rg), ZENRingSizeElt(Rg)+1,
	 ZBNPolyGetCof(QX, 1, Rg), ZENRingSizeElt(Rg)+1, 0);
  
  ZBNSetToZero(ZBNPolyGetCof(RX, 1, Rg), ZENRingSZ(Rg)+1);
  ZBNMultiply(ZBNPolyGetCof(RX, 1, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(RX, 0, Rg), ZENRingSizeElt(Rg)+1,
	      ZBNPolyGetCof(RX, 2, Rg), ZENRingSizeElt(Rg)+1);
  
  ZBNSetToZero(ZBNPolyGetCof(RX, 0, Rg), ZENRingSizeElt(Rg)+1);
  ZBNMultiply(ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(PX, 0, Rg), ZENRingSizeElt(Rg)+1,
	      ZBNPolyGetCof(QX, 0, Rg), ZENRingSizeElt(Rg)+1);
  
  ZBNSetToZero(ZBNPolyGetCof(RX, 2, Rg), ZENRingSizeElt(Rg)+1);
  ZBNMultiply(ZBNPolyGetCof(RX, 2, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(PX, 1, Rg), ZENRingSizeElt(Rg)+1,
	      ZBNPolyGetCof(QX, 1, Rg), ZENRingSizeElt(Rg)+1);
  
  ZBNSubtract(ZBNPolyGetCof(RX, 1, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(RX, 0, Rg), ZENRingSZ(Rg), 1);
  ZBNSubtract(ZBNPolyGetCof(RX, 1, Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(RX, 2, Rg), ZENRingSZ(Rg), 1);

  ZepPolyDeg(RX, Rg) = 2;
}

static void ZepPolyMultiplyKaratsubaVeryLazyDegreeBis(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  struct zenpoly UP1, UP2, UP3;
  ZENPoly up1, up2, up3;
  int m, n, i, j;
  
  /* Initialization ***********/
  up1 = &UP1; up2 = &UP2; up3 = &UP3;
  up2->deg = 0;
  m = (ZepPolyDeg(PX, Rg)+1)/2; 
  n = ZepPolyDeg(PX, Rg)+1-m;
  
  /* let's go with the recursivity **********/
  
  /* PX- * QX- -> RX */
  ZepPolyDeg(up1, Rg) = m-1;
  ZepPolyCof(up1, Rg) = ZepPolyCof(PX, Rg);
  ZepPolyCof(up2, Rg) = ZepPolyCof(QX, Rg);
  ZepPolyCof(up3, Rg) = ZepPolyCof(RX, Rg);
  ZepPolyMultiplyKaratsubaVeryLazyBis(up3, up1, up2, Rg);
  
  /* PX- + PX+ -> RX+4*m+2*n-2 */
  for(j = 0, i = 4*m+2*n-2;
      j < m;
      j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(PX, j+m, Rg),
	      ZENRingSizeElt(Rg)+1);
    ZBNAdd(ZBNPolyGetCof(RX, i, Rg), ZENRingSizeElt(Rg)+1,
	   ZBNPolyGetCof(PX, j, Rg), ZENRingSizeElt(Rg)+1, 0);
  }
  if (n > m) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(PX, j+m, Rg),
	      ZENRingSizeElt(Rg)+1);
  }
  
  /* QX- + QX+ -> RX+4*m+3*n-2 */
  for(j = 0, i = 4*m+3*n-2;
      j < m;
      j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(QX, j+m, Rg),
	      ZENRingSizeElt(Rg)+1);
    ZBNAdd(ZBNPolyGetCof(RX, i, Rg), ZENRingSizeElt(Rg)+1,
	   ZBNPolyGetCof(QX, j, Rg), ZENRingSizeElt(Rg)+1, 0);
  }
  if (n > m) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(QX, j+m, Rg),
	      ZENRingSizeElt(Rg)+1);
  }
  
  /* (PX- + PX+) * (QX- + QX+) -> RX+2*m-1 */
  ZepPolyDeg(up1, Rg) = n-1;
  ZepPolyCof(up1, Rg) = ZepPolyCof(RX, Rg)+4*m+2*n-2;
  ZepPolyCof(up2, Rg) = ZepPolyCof(RX, Rg)+4*m+3*n-2;
  ZepPolyCof(up3, Rg) = ZepPolyCof(RX, Rg)+2*m-1;
  ZepPolyMultiplyKaratsubaVeryLazyBis(up3, up1, up2, Rg);
  
  /* (PX- + PX+) * (QX- + QX+) - PX- * QX-  -> RX+2*m-1 */
  for(j = 0, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZBNSubtract(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg),
		ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg), 1);
  }
  
  /* [ PX- * QX-, (PX- + PX+) * (QX- + QX+) - PX- * QX- ] -> RX+m */
  for(j = m, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
	ZBNAdd(ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg),
	       ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 0);
  }
  for(;j < n+2*m-1; j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, j, Rg), ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg));
  }
  
  /* PX+ * QX+ -> RX+ */
  ZepPolyDeg(up1, Rg) = n-1;
  ZepPolyCof(up1, Rg) = ZepPolyCof(PX, Rg)+m;
  ZepPolyCof(up2, Rg) = ZepPolyCof(QX, Rg)+m;
  ZepPolyCof(up3, Rg) = ZepPolyCof(RX, Rg)+2*m+n-1;
  ZepPolyMultiplyKaratsubaVeryLazyBis(up3, up1, up2, Rg);

  /* [ PX- * QX-, (PX- + PX+) * (QX- + QX+) - PX- * QX- - PX+ * QX+ , PX+ * QX+ ] -> RX */
  for(j = m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZBNSubtract(ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg),
		ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 1);
  }
  for(j = 2*m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZBNAdd(ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg),
	   ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 0);
  }
  for(; j < 2*m+2*n-1; j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, j, Rg),
	      ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg));
  }

  ZepPolyDeg(RX, Rg) = ZepPolyDeg(PX, Rg)+ZepPolyDeg(QX, Rg);
}

static void ZepPolyMultiplyKaratsubaVeryLazyDegree(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  struct zenpoly UP1, UP2, UP3;
  ZENPoly up1, up2, up3;
  int m, n, i, j;
  
  /* Initialization ***********/
  up1 = &UP1; up2 = &UP2; up3 = &UP3;
  up2->deg = 0;
  m = (ZepPolyDeg(PX, Rg)+1)/2; 
  n = ZepPolyDeg(PX, Rg)+1-m;
  
  /* let's go with the recursivity **********/
  
  /* PX- * QX- -> RX */
  ZepPolyDeg(up1, Rg) = m-1;
  ZepPolyCof(up1, Rg) = ZepPolyCof(PX, Rg);
  ZepPolyCof(up2, Rg) = ZepPolyCof(QX, Rg);
  ZepPolyCof(up3, Rg) = ZepPolyCof(RX, Rg);
  ZepPolyMultiplyKaratsubaVeryLazy(up3, up1, up2, Rg);
  
  /* PX- + PX+ -> RX+4*m+2*n-2 */
  for(j = 0, i = 4*m+2*n-2;
      j < m;
      j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(PX, j+m, Rg),
	      ZENRingSizeElt(Rg));
    ZBNPolyGetCof(RX, i, Rg)[ZENRingSizeElt(Rg)] = 0;
    ZBNAdd(ZBNPolyGetCof(RX, i, Rg), ZENRingSizeElt(Rg)+1,
	   ZBNPolyGetCof(PX, j, Rg), ZENRingSizeElt(Rg), 0);
  }
  if (n > m) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(PX, j+m, Rg),
	      ZENRingSizeElt(Rg));
    ZBNPolyGetCof(RX, i, Rg)[ZENRingSizeElt(Rg)] = 0;
  }
  
  /* QX- + QX+ -> RX+4*m+3*n-2 */
  for(j = 0, i = 4*m+3*n-2;
      j < m;
      j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(QX, j+m, Rg),
	      ZENRingSizeElt(Rg));
    ZBNPolyGetCof(RX, i, Rg)[ZENRingSizeElt(Rg)] = 0;
    ZBNAdd(ZBNPolyGetCof(RX, i, Rg), ZENRingSizeElt(Rg)+1,
	   ZBNPolyGetCof(QX, j, Rg), ZENRingSizeElt(Rg), 0);
  }
  if (n > m) {
    ZBNAssign(ZBNPolyGetCof(RX, i, Rg),
	      ZBNPolyGetCof(QX, j+m, Rg),
	      ZENRingSizeElt(Rg));
    ZBNPolyGetCof(RX, i, Rg)[ZENRingSizeElt(Rg)] = 0;
  }
  
  /* (PX- + PX+) * (QX- + QX+) -> RX+2*m-1 */
  ZepPolyDeg(up1, Rg) = n-1;
  ZepPolyCof(up1, Rg) = ZepPolyCof(RX, Rg)+4*m+2*n-2;
  ZepPolyCof(up2, Rg) = ZepPolyCof(RX, Rg)+4*m+3*n-2;
  ZepPolyCof(up3, Rg) = ZepPolyCof(RX, Rg)+2*m-1;
  ZepPolyMultiplyKaratsubaVeryLazyBis(up3, up1, up2, Rg);
  
  /* (PX- + PX+) * (QX- + QX+) - PX- * QX-  -> RX+2*m-1 */
  for(j = 0, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZBNSubtract(ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg),
		ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg), 1);
  }
  
  /* [ PX- * QX-, (PX- + PX+) * (QX- + QX+) - PX- * QX- ] -> RX+m */
  for(j = m, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
	ZBNAdd(ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg),
	       ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 0);
  }
  for(;j < n+2*m-1; j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, j, Rg), ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg));
  }
  
  /* PX+ * QX+ -> RX+ */
  ZepPolyDeg(up1, Rg) = n-1;
  ZepPolyCof(up1, Rg) = ZepPolyCof(PX, Rg)+m;
  ZepPolyCof(up2, Rg) = ZepPolyCof(QX, Rg)+m;
  ZepPolyCof(up3, Rg) = ZepPolyCof(RX, Rg)+2*m+n-1;
  ZepPolyMultiplyKaratsubaVeryLazy(up3, up1, up2, Rg);

  /* [ PX- * QX-, (PX- + PX+) * (QX- + QX+) - PX- * QX- - PX+ * QX+ , PX+ * QX+ ] -> RX */
  for(j = m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZBNSubtract(ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg),
		ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 1);
  }
  for(j = 2*m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZBNAdd(ZBNPolyGetCof(RX, j, Rg), ZENRingSZ(Rg),
	   ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg), 0);
  }
  for(; j < 2*m+2*n-1; j++, i++) {
    ZBNAssign(ZBNPolyGetCof(RX, j, Rg),
	      ZBNPolyGetCof(RX, i, Rg), ZENRingSZ(Rg));
  }

  ZepPolyDeg(RX, Rg) = ZepPolyDeg(PX, Rg)+ZepPolyDeg(QX, Rg);
}

/* \TEX{
   \title{VeryLazy polynomial multiplication with the Karatsuba method.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX| \neq|PX|,|QX|$. The degrees of |PX| and
   |QX| must be equal. |RX| must be four times larger than |PX|.}
   }
   */
int ZepPolyMultiplyKaratsubaVeryLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  switch (ZepPolyDeg(PX, Rg)) {

    /* compute the product of 2 polynoms whose degree is 0 */
  case 0:
    ZepPolyMultiplyKaratsubaVeryLazyDegree0(RX, PX, QX, Rg);
    break;

    /* compute the product of 2 polynoms whose degree is 1 */
  case 1: 
    ZepPolyMultiplyKaratsubaVeryLazyDegree1(RX, PX, QX, Rg); 
    break;

    /* compute the product of 2 polynoms whose degree is greater than 2*/
  default : 
    ZepPolyMultiplyKaratsubaVeryLazyDegree(RX, PX, QX, Rg); 
  }
  return(ZEN_HAS_INVERSE);
}

static void ZepPolyMultiplyKaratsubaVeryLazyBis(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  switch (ZepPolyDeg(PX, Rg)) {

    /* compute the product of 2 polynoms whose degree is 0 */
  case 0:
    ZepPolyMultiplyKaratsubaVeryLazyDegree0Bis(RX, PX, QX, Rg);
    break;

    /* compute the product of 2 polynoms whose degree is 1 */
  case 1: 
    ZepPolyMultiplyKaratsubaVeryLazyDegree1Bis(RX, PX, QX, Rg); 
    break;

    /* compute the product of 2 polynoms whose degree is greater than 2*/
  default : 
    ZepPolyMultiplyKaratsubaVeryLazyDegreeBis(RX, PX, QX, Rg); 
  }
}

/* Division routines
***********************/

/* \TEX{
   \title{VeryLazy division.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZEN_NO_INVERSE| if
   the leading coefficient of |QX| is a factor of a modulo, |ZENERR| for an
   error}
   \note{|PX| is destroyed. |ZENRingFact(Rg)| is filled with a factor
   of a modulo if |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZepPolyDivideVeryLazy(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int i, j, deg, IsNotOne;

  if (!ZepEltIsOne(ZepPolyGetCof(QX, ZepPolyDeg(QX, Rg), Rg), Rg)) {
    IsNotOne = 1;
    
    switch(ZepEltInverse(ZENRingIbuf(Rg),ZepPolyGetCof(QX, ZepPolyDeg(QX, Rg), Rg), Rg)) {
    case ZENERR:
      return(ZERR);
    case ZEN_NO_INVERSE:
      return(ZEN_NO_INVERSE);
    }
  }
  else {
    IsNotOne = 0;
  }

  for(i = ZepPolyDeg(PX, Rg);
      i >= ZepPolyDeg(QX, Rg);
      i--) {
      int neg;

      neg = 0;
      if (ZBNPolyGetCof(PX, i, Rg)[ZENRingSZ(Rg)-1] >> (BN_DIGIT_SIZE-1)) {
	ZBNComplement(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg));
	ZBNAddCarry(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), 1);
	neg = 1;
      }
      ZepQuickMod(ZepPolyGetCof(PX, i, Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), Rg);
      
      if (IsNotOne) {
	ZepEltMultiply(ZepPolyGetCof(PX, i, Rg),
		       ZENRingIbuf(Rg), ZepPolyGetCof(PX, i, Rg),
		       Rg);
      }
      if (neg) {
	ZepEltNegate(ZepPolyGetCof(PX, i, Rg), ZepPolyGetCof(PX, i, Rg), Rg);
      }
    
    for(j =  ZepPolyDeg(QX, Rg)-1;
	j >= 0;
	j--) {
      ZBNSetToZero(ZepRingMbuf(Rg), ZENRingSZ(Rg)+1);
      ZBNMultiply(ZepRingMbuf(Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(QX, j, Rg), ZENRingSizeElt(Rg));
      ZBNSubtract(ZBNPolyGetCof(PX, i+j-ZepPolyDeg(QX, Rg), Rg), ZENRingSZ(Rg),
		  ZepRingMbuf(Rg), ZENRingSZ(Rg), 1);
    }
  }
  deg =  ZepPolyDeg(QX, Rg);
  for(i = ZepPolyDeg(PX, Rg); i >= ZepPolyDeg(QX, Rg); i--) {
    ZepEltAssign(ZepPolyGetCof(RX, i-ZepPolyDeg(QX, Rg), Rg),
		 ZepPolyGetCof(PX, i, Rg),
		 Rg);
  }
  ZepPolyDeg(RX, Rg) = ZepPolyDeg(PX, Rg)-ZepPolyDeg(QX, Rg);
  if (ZepPolyDeg(RX, Rg) < 0) {
    ZepPolySetToZero(RX, Rg);
  }
  deg = -1;
  for(; i >= 0; i--) {
    int neg;
    
    neg = 0;
    if (ZBNPolyGetCof(PX, i, Rg)[ZENRingSZ(Rg)-1] >> (BN_DIGIT_SIZE-1)) {
      ZBNComplement(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg));
      ZBNAddCarry(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), 1);
      neg = 1;
    }
    ZepQuickMod(ZepPolyGetCof(PX, i, Rg), ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), Rg);
    if (!ZepEltIsZero(ZepPolyGetCof(PX, i, Rg), Rg)) {
      if (neg) {
	ZepEltNegate(ZepPolyGetCof(MX, i, Rg), ZepPolyGetCof(PX, i, Rg), Rg);
      }
      else {
	ZepEltAssign(ZepPolyGetCof(MX, i, Rg), ZepPolyGetCof(PX, i, Rg), Rg);
      }
      if (deg == -1) {
	deg = i;
      }
    }
    else if (deg >= 0) {
      ZepEltSetToZero(ZepPolyGetCof(MX, i, Rg), Rg);
    }
  }
  if (deg < 0) {
    ZepPolySetToZero(MX, Rg);
  }
  else {
    ZepPolyDeg(MX, Rg) = deg;
  }
  return(ZEN_HAS_INVERSE);
}

int _ZepPolyDivideLazy(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  ZepPolyLazyToVeryLazy(PX, PX, Rg);
  return(ZepPolyDivideVeryLazy(RX, MX, PX, QX, Rg));
}

/* \TEX{
   \title{Divide.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZEN_NO_INVERSE| if
   a factor of a modulo was dicovered, |ZENERR| for an error}
   \side{|ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZepPolyDivide(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int i, ret;
  ZENPoly Pbuf;

  if (!(Pbuf = ZepPolyCreateLarge(ZepPolyDeg(PX, Rg), Rg))) {
    return(ZERR);
  }
  ZepPolyLazyToVeryLazy(Pbuf, PX, Rg);
  ret = ZepPolyDivideLazy(RX, MX, Pbuf, QX, Rg);
  ZepPolyFree(Pbuf, Rg);
  return(ret);
}


/* \TEX{
   \title{VeryLazy mudulo computation.}
   \input{A polynomial |PX| to divide by
   |QX|$=X^{\mbox{degree}(QX)}+sum_{i=0}^{|k|}q_iX^i$} in a finite
   ring |Rg|. |MX| will be the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| for an error}
   \note{|PX| is destroyed.}
   */
int _ZepPolyModuloVeryLazy(MX, PX, QX, k, Rg)
     ZENPoly MX, PX, QX;
     int k;
     ZENRing Rg;
{
  int i, j, deg, IsNotOne;

  for(i = ZepPolyDeg(PX, Rg);
      i >= ZepPolyDeg(QX, Rg);
      i--) {
    int neg;
    
    neg = 0;
    if (ZBNPolyGetCof(PX, i, Rg)[ZENRingSZ(Rg)-1] >> (BN_DIGIT_SIZE-1)) {
      ZBNComplement(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg));
      ZBNAddCarry(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), 1);
      neg = 1;
    }
    ZepQuickMod(ZepPolyGetCof(PX, i, Rg),
		ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), Rg);
    if (neg) {
      ZepEltNegate(ZepPolyGetCof(PX, i, Rg), ZepPolyGetCof(PX, i, Rg), Rg);
    }
      
    for(j = k;
	j >= 0;
	j--) {
      ZBNSetToZero(ZepRingMbuf(Rg), ZENRingSZ(Rg)+1);
      ZBNMultiply(ZepRingMbuf(Rg), ZENRingSZ(Rg),
		  ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		  ZBNPolyGetCof(QX, j, Rg), ZENRingSizeElt(Rg));
      ZBNSubtract(ZBNPolyGetCof(PX, i+j-ZepPolyDeg(QX, Rg), Rg), ZENRingSZ(Rg),
		  ZepRingMbuf(Rg), ZENRingSZ(Rg), 1);
    }
  }

  deg = -1;
  for(; i >= 0; i--) {
    int neg;

    neg = 0;
    if (ZBNPolyGetCof(PX, i, Rg)[ZENRingSZ(Rg)-1] >> (BN_DIGIT_SIZE-1)) {
      ZBNComplement(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg));
      ZBNAddCarry(ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), 1);
      neg = 1;
    }
    ZepQuickMod(ZepPolyGetCof(PX, i, Rg), ZBNPolyGetCof(PX, i, Rg), ZENRingSZ(Rg), Rg);
    if (!ZepEltIsZero(ZepPolyGetCof(PX, i, Rg), Rg)) {
      if (neg) {
	ZepEltNegate(ZepPolyGetCof(MX, i, Rg), ZepPolyGetCof(PX, i, Rg), Rg);
      }
      else {
	ZepEltAssign(ZepPolyGetCof(MX, i, Rg), ZepPolyGetCof(PX, i, Rg), Rg);
      }
      if (deg < 0) {
	deg = i;
      }
    }
    else if (deg >= 0) {
      ZepEltSetToZero(ZepPolyGetCof(MX, i, Rg), Rg);
    }
  }
  if (deg < 0) {
    ZepPolySetToZero(MX, Rg);
  }
  else {
    ZepPolyDeg(MX, Rg) = deg;
  }
  return(ZEN_HAS_INVERSE);
}

/* Other routines
*********************************************/


/* \TEX{
   \title{Multiplication by a scalar.}
   \input{Two polynomials |PX|, |RX| and a |ZENElt| |e| of a finite ring |Rg|.}
   }
	 */
void _ZepPolyMultiplyScalar(RX, PX, e, Rg)
     ZENPoly RX, PX;
     ZENElt e;
     ZENRing Rg;
{
  int i;

  if (ZepEltIsZero(e, Rg)) {
    ZepPolySetToZero(RX, Rg);
    return;
  }
  if (ZepEltIsOne(e, Rg)) {
    ZepPolyAssign(RX, PX, Rg);
    return;
  }
  for(i = 0; i <= ZepPolyDeg(PX, Rg); i++) {
    ZepEltMultiply(ZepPolyGetCof(RX, i, Rg),
		    e,
		    ZepPolyGetCof(PX, i, Rg),
		    Rg);
  }
  ZepPolyDeg(RX, Rg) = ZepPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Scalar product.}
   \input{Two polynomials |PX|, |RX| and a pointer to a |ZENElt| |e| of a
   finite ring |Rg|.}
   \side{$|*e| = \sum p_iq_i$}
   }
   */
void _ZepPolyDot(e, PX, QX, Rg)
     ZENElt *e;
     ZENPoly PX, QX;
     ZENRing Rg;
{
  int i, deg;

  ZBNSetToZero(ZepRingMbuf(Rg), ZENRingSZ(Rg)+1);
  ZBNMultiply(ZepRingMbuf(Rg), ZENRingSZ(Rg),
	      ZBNPolyGetCof(PX, 0, Rg), ZENRingSizeElt(Rg),
	      ZBNPolyGetCof(QX, 0, Rg), ZENRingSizeElt(Rg));


  deg = ZepPolyDeg(PX, Rg) < ZepPolyDeg(QX, Rg) ?
    ZepPolyDeg(PX, Rg) :  ZepPolyDeg(QX, Rg);

  for(i = 1; i <= deg; i++) {
    ZBNMultiply(ZepRingMbuf(Rg), ZENRingSZ(Rg),
		ZBNPolyGetCof(PX, i, Rg), ZENRingSizeElt(Rg),
		ZBNPolyGetCof(QX, i, Rg), ZENRingSizeElt(Rg));
  }
  ZepQuickMod(*e, ZepRingMbuf(Rg), ZENRingSZ(Rg), Rg);
}

