/*
 * zen/zeps/c/elt_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Element assignment}
   
   One can set a |ZENElt| to zero or one with |_ZepsEltSetToZero| or
   |_ZepsEltSetToOne|. Then we can assign a |ZENElt| with an other with
   |_ZepsEltAssign|. Furthermore, we can set randomly a |ZENElt| with
   |_ZepsEltSetRandom| or enumerate the elements of a finite ring with
   |_ZepsEltSetNext|.
   }
   */
#include <stdio.h>
#include "zeps.h"

/* \TEX{
   \title{Assign.}
   \input{A pointer |p_a| on a |ZENElt| and a |ZENElt| |b| of a finite ring |Rg|.}
   \side{|*p_a| is filled with |b|.}
   }
   */
void _ZepsEltAssign(p_a, b, Rg)
     ZENElt *p_a, b;
     ZENRing Rg;
{
  ZepsEltAssign(*p_a, b, Rg);
}

/* \TEX{
   \title{Set to zero.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set to zero..}
   }
   */
void _ZepsEltSetToZero(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZepsEltSetToZero(*p_a, Rg);
}

/* \TEX{
   \title{Set to one.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set to one.}
   }
   */
void _ZepsEltSetToOne(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZepsEltSetToOne(*p_a, Rg);
}

/* \TEX{
   \title{Set to random.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set randomly.}
   }
   */
void _ZepsEltSetRandom(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZepsEltSetRandom(*p_a, Rg);
}

/* \TEX{
   \title{Enumerate.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set randomly.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \note{After a number of call of this function equal to the number of elements in the finite ring, we obtain the same element.}
   }
   */
void _ZepsEltSetNext(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZepsEltSetNext(*p_a, Rg);
}

/* \TEX{
   \title{Conversion.}
   \input{A pointer |pa| on a |ZENElt| of a modular ring |Rg|, an element |b| of a modular ring |R|.}
   \side{|*pa| is set to the modular conversion of |b|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   }
   */
int _ZepsEltConvert(pa, Rg, b, R)
     ZENElt *pa, b;
     ZENRing Rg, R;
{ 
  BigNum p;
  BigNumLength pl;

  switch(ZENRingArith(R)) {
    
  /* Fractions
   *************/
  case ZEN_LIB_ZEF:
    pl = ZENRingSizeElt(Rg);

    if (!(p = ZBNC(pl))) {
      ZENSetError(ZEN_LIB_ZEPS, ZEN_CONVERT, ZEN_ALLOC);
      return(ZERR);
    }
    ZENEltToZ(p, &pl, b, R);
    if ((ZENErrErrno == ZEN_INPUT) && (pl == ZENRingSizeElt(Rg))) {
      
      /* This element is a fraction, nothing can be done */
      ZENSetError(ZEN_LIB_ZEPS, ZEN_CONVERT, ZEN_INPUT);
      return(ZERR);
    }
    ZBNF(p);
    if (!(p = ZBNC(pl+1))) {
      ZENSetError(ZEN_LIB_ZEPS, ZEN_CONVERT, ZEN_ALLOC);
      return(ZERR);
    }
    ZENEltToZ(p, &pl, b, R);
    break;
    
    /* Specific cases where things can be done faster
     *************************************************/
  case ZEN_LIB_ZE2:
    if(ZENEltIsZero(b, R))
      ZepsEltSetToZero(*pa, Rg);
    else
      ZepsEltSetToOne(*pa, Rg);
    return(ZEN_HAS_INVERSE);

    /* Default
     **********/
  default:
    /* R is a finite field of same characteristic
       or a a prime finite field ? */
    if (!ZBNCompare(ZENRingP(R), ZENRingPl(R),
		    ZENRingP(Rg), ZENRingPl(Rg)) ||
	!ZENRingDeg(R)) {
      if (!(p = ZBNC(ZENRingQl(R)+1))) {
	ZENSetError(ZEN_LIB_ZEPS, ZEN_CONVERT, ZEN_ALLOC);
	return(ZERR);
      }
      ZENEltToZ(p, &pl, b, R);
      break;
    }

    /* Otherwise, what can we do ? */
    ZENSetError(ZEN_LIB_ZEPS, ZEN_CONVERT, ZEN_INPUT);
    return(ZERR);
  }    

  switch(ZBNCompareLazy(p, pl,
			ZENRingP(Rg), ZENRingPl(Rg))) {
  case BN_EQ:
    ZepsEltSetToZero(*pa, Rg);
    break;
  case BN_LT:
    ZepsZToElt(*pa, p, pl, Rg);
    break;
  default:
    p[pl] = 0;
    ZBNDivide(p, pl+1, ZENRingP(Rg), ZENRingPl(Rg));
    pl = ZBNNumDigits(p, ZENRingPl(Rg));
    ZepsZToElt(*pa, p, pl, Rg);
  }
  ZBNF(p);
  return(ZEN_HAS_INVERSE);
}
