/*
 * zen/zeps/c/elt_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{input/output}
   
   Conversions between the internal representation and the ASCII
   representation of a |ZENElt| are done by |_ZepsEltPrintToString|. |_ZepsEltPrintToFile|
   allows to write a |ZENElt| in a file.
   One can moreover write to and then read from a file a |ZENelt| with |_ZepsEltPutToFile|
   and |_ZepsEltReadFromFile|.
   }
   */
#include <stdio.h>
#include <ctype.h>
#include "zeps.h"

/* \TEX{
   \title{Converting from string.}
   \input{A finite field |Rg|, a pointer on a |ZENElt| |G| and a string
	    |s| representing a |ZENElt| in base $|base|\in \{2,\ldots,
	 16\}$.} 
\output{The number of character read in |s| or |ZENERR| if an error occured.}
\side{|G| is filled with |s| if the output is not |ZENERR|.}
}
*/
int _ZepsEltReadFromString(G, s, base, Rg)
     ZENElt *G;
     char *s;
     int base;
     ZENRing Rg;
{
  BigNum n;
  BigNumLength nl;
  int ret;

  if ((ret = ZBNReadFromString(&n, &nl, s, base)) < 0) {
    return(ret);
  }
  if ((nl > 1) || (ZBNDigit(n, 0) > ZBNDigit(ZENRingP(Rg), 0))) {
    ZENSetError(ZEN_LIB_ZEPS, ZEN_READFROMSTRING, ZEN_SYNTAX);
    ZBNF(n);
    return(ZERR);
  }
  ZepsElt2Int(*G, Rg) = ZBNDigit(n, 0);
  ZBNF(n);
  return(ret);
}

/* \TEX{
   \title{Converting to string.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{An allocated string representing |G| in base |base| or |ZENNULL| if an error occured.}   
   }
   */
char *_ZepsEltPrintToString(G, base, Rg)
     ZENElt G;
     int base;
     ZENRing Rg;  
{
  return(ZepsEltPrintToString(G, base, Rg));
}


/* \TEX{
   \title{Converting from a string in an internal representation.}
   \input{A finite field |Rg| and a string |s| representing a |ZENElt|
   in an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|G| is filled with s}
   }
   */
int _ZepsEltGetFromString(G, s, Rg)
     char *s;
     ZENElt *G;
     ZENRing Rg;
{
  int i;
  char *ptr;
  unsigned int c;
  BigNumDigit x;
  
  for(ptr = s; isspace(*ptr); ptr++) ;

  c = x = 0;
  for(i = 0; i < sizeof(BigNumDigit); i++) {
    if (sscanf(ptr+2*i, "%2x", &c) != 1) {
      ZENSetError(ZEN_LIB_ZEPS, ZEN_POLYPUTTOSTRING, ZEN_SPRINTF);
      return(ZERR);
    }
    x |= ((BigNumDigit)c) << (8*i);
  }
  ZepsElt2Int(*G, Rg) = x;
  return(ptr+2*sizeof(BigNumDigit)-s);
}

/* \TEX{
   \title{Converting to string in an internal representation.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg|.}
   \output{An allocated string representing |G| in an internal representation or NULL if an error occured.}   
   }
   */
char *_ZepsEltPutToString(G, Rg)
     ZENElt G;
     ZENRing Rg;
{
  int i;
  char *str;
  BigNumDigit x;

  if (!(str = (char *)ZENMalloc((1+2*sizeof(BigNumDigit))*sizeof(char)))) {
    ZENSetError(ZEN_LIB_ZEPS, ZEN_POLYPUTTOSTRING, ZEN_ALLOC);
    return(ZENNULL);
  }
  x = ZepsElt2Int(G, Rg);
  for(i = 0; i < sizeof(BigNumDigit); i++) {
    if (sprintf(str+2*i, "%.2x", (unsigned char)(x & ((BigNumDigit)255))) < 0) {
      ZENFree(str);
      ZENSetError(ZEN_LIB_ZEPS, ZEN_POLYPUTTOSTRING, ZEN_SPRINTF);
      return(ZENNULL);
    }
    x >>= 8;
  }
  return(str);
}


/* \TEX{
   \title{Printing to file.}
   \input{A stream |file|, an allocated |ZENElt| |G| of a finite field
	    |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
	    \output{|ZENERR| if an error occured, 0 otherwise.}
	    \side{Printing a representation of |G| in base |base| to |file|.}
	    }
	    */
int _ZepsEltPrintToFile(file, G, base, Rg)
     FILE *file;
     ZENElt G;
     int base;
     ZENRing Rg;  
{  
  return(ZepsEltPrintToFile(file, G, base, Rg));
}

/* \TEX{
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |ZENElt| |G| of a finite field
   |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |G| to |file|.}
   }
   */
int _ZepsEltPutToFile(file, G, Rg)
     FILE *file;
     ZENElt G;
     ZENRing Rg;  
{
  int i;
  BigNumDigit x;

  x = ZepsElt2Int(G, Rg);
  for(i = 0; i < sizeof(BigNumDigit); i++) {
    if (fprintf(file, "%.2x", (unsigned char)(x & ((BigNumDigit)255))) < 0) {
      ZENSetError(ZEN_LIB_ZEPS, ZEN_POLYPUTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
    x >>= 8;
  }
  return(0);
}

/* \TEX{
   \title{Reading from file.}
   \input{A stream |file| and an element |G| of a finite field |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|G| is filled with the ZENElt read in |file| if no error occurred.}
   }
   */
int _ZepsEltReadFromFile(G, file, base, Rg)
     ZENElt *G;
     FILE *file;
     int base;
     ZENRing Rg;
{
  BigNum n;
  BigNumLength nl;
  int ret;
  
  if ((ret = ZBNReadFromFile(&n, &nl, file, base)) < 0) {
    return(ret);
  }
  if ((nl > 1) || (ZBNDigit(n, 0) > ZBNDigit(ZENRingP(Rg), 0))) {
    ZBNF(n);
    ZENSetError(ZEN_LIB_ZEPS, ZEN_READFROMFILE, ZEN_SYNTAX);
    return(ZERR);
  }
  ZepsElt2Int(*G, Rg) = ZBNDigit(n, 0);
  ZBNF(n);
  return(ret);
}

/* \TEX{
   \title{Getting from file in an internal representation.}
   \input{A stream |file|, and an element |G| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|G| is filled with the element read in |file|}
   }
   */
int _ZepsEltGetFromFile(G, file, Rg)
     ZENElt *G;
     FILE *file;
     ZENRing Rg;
{
  int i;
  unsigned int c;
  BigNumDigit x;
  
  c = x = 0;
  for(i = 0; i < sizeof(BigNumDigit); i++) {
    if (fscanf(file, "%2x", &c) != 1) {
      ZENSetError(ZEN_LIB_ZEPS, ZEN_POLYGETFROMSTRING, ZEN_SSCANF);
      return(ZERR);
    }
    x |= ((BigNumDigit)c) << (8*i);
  }
  ZepsElt2Int(*G, Rg) = x;
  return(0);
}


/* \TEX{
   \title{Printing to stdout.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |G| in base |10| to |stdout|.}
   \note{Usefull for debugging ...}
   }
   */
int _ZepsEP(G, Rg)
     ZENElt G;
     ZENRing Rg;  
{
  int ret;

  if (!(ret = ZepsEltPrintToFile(stdout, G, 10, Rg))) {
    fprintf(stdout, "\n");
  }
  return(ret);
}

/* \TEX{
   \title{Reading from stdin.}
   \input{A pointer to an element |G| of a finite field |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|G| is filled with the |ZENElt| read in |stdin| in base 10 if no error occurred.}
   \note{Usefull for debugging ...}
   }
   */
int _ZepsER(G, Rg)
     ZENElt *G;
     ZENRing Rg;
{
  int ret;

  if (!(ret = ZepsEltReadFromFile(*G, stdin, 10, Rg))) {
    fprintf(stdout, "\n");
  }
  return(ret);
}


/* \TEX{
   \title{Evaluating an element as a multivariate polynomial over $Z$.}
   \input{A Bignum |p| of allocated size greater or equal than
   |ZENRingQl(Rg)|, a pointer |p_pl| on the real size of |p|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|p| is filled with the value of |G| considered as a
   multivariate polynomial in which we substitute all the variables
   with the characteristic, |*p_pl| contains the real size of |p|.}
   } */
void _ZepsEltToZ(p, p_pl, G, Rg)
     BigNum p;
     BigNumLength *p_pl;
     ZENElt G;
     ZENRing Rg;
{
  ZepsEltToZ(p, p_pl, G, Rg);
}

/* \TEX{
   \title{Getting an element from its evaluation as a multivariate
   polynomial over $Z$.}
   \input{A Bignum |p| of size |pl|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|G| is filled with the element whose evaluation considered as a
   multivariate polynomial in which we substitute all the variables
   with the characteristic is equal to (|p|, |pl|).}
   } */
void _ZepsZToElt(G, p, pl, Rg)
     BigNum p;
     BigNumLength pl;
     ZENElt *G;
     ZENRing Rg;
{
  ZepsZToElt(*G, p, pl, Rg);
}

