/*
 * zen/zeps/c/elt_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Element multiplication}
   
   Two functions are provided to do multiplications: |_ZepsEltSquare| to square a
   |ZENElt| and |_ZepsEltMultily| to compute the multiplication of two |ZENElt|.
   }
   */
#include <stdio.h>
#include "zeps.h"

/* \TEX{
   \title{Squaring for a modulo larger than $2^{|BN_DIGIT_SIZE|/2}$.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{$|*p_c| = a^2$.}
   \note{For debugging purposes.}
   }
   */
void _ZepsEltSquareLargep1(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  ZENElt x, y, r;

  ZepsEltAssign(x, a, Rg);
  ZepsEltAssign(y, a, Rg);
  if (ZepsElt2Int(x, Rg) & 1) {
    ZepsEltAssign(r, y, Rg);
  }
  else {
    ZepsEltSetToZero(r, Rg);
  }
  for(ZepsElt2Int(x, Rg) >>= 1; ZepsElt2Int(x, Rg) > 0; ZepsElt2Int(x, Rg) >>= 1) {
    ZepsEltAdd(y, y, Rg);
    if (ZepsElt2Int(x, Rg) & 1) {
      ZepsEltAdd(r, y, Rg);
    }
  }
  ZepsEltAssign(*p_c, r, Rg);
}


/* \TEX{
   \title{Squaring for a modulo larger than $2^{|BN_DIGIT_SIZE|/2}$.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{$|*p_c| = a^2$.}
   }
   */
void _ZepsEltSquareLargep(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  BigNumDigit T[2], U[2];

  U[0] = U[1] = 0;
  ZBNMultiplyDigit(U, 2, &(ZepsElt2Int(a, Rg)), 1, ZepsElt2Int(a, Rg));
  
  ZepsElt2Int(*p_c, Rg) = ZBNDivideDigit(T, U, 2, ZBNDigit(ZENRingP(Rg), 0));

#  if 0
  {
    ZENElt c;

    _ZepsEltSquareLargep1(&c, a, Rg);
    if (ZepsElt2Int(*p_c, Rg) != ZepsElt2Int(c, Rg)) {
      printf("ARGH a bug in ZepsEltSquareLargep() !!!\n");
      exit(-1);
    }
  }
#  endif
}

/* \TEX{
   \title{Squaring for a modulo larger than $2^{|BN_DIGIT_SIZE|/2}$.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{$|*p_c| = a^2$.}
   \note{Used with precomputations.}
   }
   */
void _ZepsEltSquarePrcLargep(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
#define BASE ((BigNumDigit)1 << (BN_DIGIT_SIZE/2))
#define BASE1 ((BigNumDigit)1 << (2*SIZE_CHAR))

  BigNumDigit al, ah, c, d, tl, th, t, p;

  p = ZBNDigit(ZENRingP(Rg), 0);
  
  /* a*b -> (c, d) */
  ah = ZepsElt2Int(a, Rg)/BASE;
  al = ZepsElt2Int(a, Rg)%BASE;
  
  c = (al*al);
  d = ah*ah;
  
  t = al*ah;
  th = t/BASE;
  d += 2*th;
  tl = ((t%BASE)*BASE);
  c += tl;
  if (c < tl) d++;
  c += tl;
  if (c < tl) d++;

  /* (c, d) % p -> *p_c */
  if (c >= p) c %= p;
  
  t = d%BASE1;
  c += ZepsMultTable0(Rg, t);
  if ((c < ZepsMultTable0(Rg, t)) || (c >= p)) c -= p;

  if (BN_DIGIT_SIZE == 8*SIZE_CHAR) {
    d = d/BASE1;
    t = d%BASE1;
    c += ZepsMultTable2(Rg, t);
    if ((c < ZepsMultTable2(Rg, t)) || (c >= p)) c -= p;

    d = d/BASE1;
    t = d%BASE1;
    c += ZepsMultTable3(Rg, t);
    if ((c < ZepsMultTable3(Rg, t)) || (c >= p)) c -= p;
  }

  d = d/BASE1;
  c += ZepsMultTable1(Rg, d);
  if ((c < ZepsMultTable1(Rg, d)) || (c >= p)) c -= p;

  ZepsElt2Int(*p_c, Rg) = c;

#  if 0
  {
    ZENElt c;

    _ZepsEltSquareLargep1(&c, a, Rg);
    if (ZepsElt2Int(*p_c, Rg) != ZepsElt2Int(c, Rg)) {
      printf("ARGH a bug in ZepsEltMultiplyLargep() !!!\n");
      exit(-1);
    }
  }
#  endif
}

/* \TEX{
   \title{Squaring for a modulo smaller than $2^{|BN_DIGIT_SIZE|/2}$.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |R|.}
   \side{$|*p_c| = a^2$.}
   }
   */
void _ZepsEltSquareSmallp(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  ZepsElt2Int(*p_c, Rg) =
    ((ZepsElt2Int(a, Rg)*ZepsElt2Int(a, Rg)))%ZBNDigit(ZENRingP(Rg), 0);
}

/* \TEX{
   \title{Multiplication for modulo smaller than $2^{|BN_DIGIT_SIZE|/2}$.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |Rg|.}
   \side{$|*p_c| = a\times b$.}
   \note{this procedure is used only for small modulos.}
   }
   */
void _ZepsEltMultiplySmallp(p_c, a, b, Rg)
     ZENElt *p_c, a, b;
     ZENRing Rg;
{
  ZepsElt2Int(*p_c, Rg) =
    ((ZepsElt2Int(a, Rg)*ZepsElt2Int(b, Rg)))%ZBNDigit(ZENRingP(Rg), 0);
}


/* \TEX{
   \title{Multiplication for modulo larger than $2^{|BN_DIGIT_SIZE|/2}$.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |Rg|.}
   \side{$|*p_c| = a\times b$.}
   \note{this procedure is used only for large modulos and for debugging.}
   }
   */
void _ZepsEltMultiplyLargep1(p_c, a, b, Rg)
     ZENElt *p_c, a, b;
     ZENRing Rg;
{
  ZENElt x, y, r;

  ZepsEltAssign(x, a, Rg);
  ZepsEltAssign(y, b, Rg);
  if (ZepsElt2Int(x, Rg) & 1) {
    ZepsEltAssign(r, y, Rg);
  }
  else {
    ZepsEltSetToZero(r, Rg);
  }
  for(ZepsElt2Int(x, Rg) >>= 1; ZepsElt2Int(x, Rg) > 0; ZepsElt2Int(x, Rg) >>= 1) {
    ZepsEltAdd(y, y, Rg);
    if (ZepsElt2Int(x, Rg) & 1) {
      ZepsEltAdd(r, y, Rg);
    }
  }
  ZepsEltAssign(*p_c, r, Rg);
}


/* \TEX{
   \title{Multiplication for modulo larger than $2^{|BN_DIGIT_SIZE|/2}$.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |Rg|.}
   \side{$|*p_c| = a\times b$.}
   \note{this procedure is used only for large modulos.}
   }
   */
void _ZepsEltMultiplyLargep(p_c, a, b, Rg)
     ZENElt *p_c, a, b;
     ZENRing Rg;
{
  BigNumDigit T[2], U[2];

  U[0] = U[1] = 0;
  ZBNMultiplyDigit(U, 2, &(ZepsElt2Int(a, Rg)), 1, ZepsElt2Int(b, Rg));
  
  ZepsElt2Int(*p_c, Rg) = ZBNDivideDigit(T, U, 2, ZBNDigit(ZENRingP(Rg), 0));

#  if 0
  {
    ZENElt c;

    _ZepsEltMultiplyLargep1(&c, a, b, Rg);
    if (ZepsElt2Int(*p_c, Rg) != ZepsElt2Int(c, Rg)) {
      printf("ARGH a bug in ZepsEltMultiplyLargep() !!!\n");
      exit(-1);
    }
  }
#  endif
}

/* \TEX{
   \title{Multiplication for modulo larger than $2^{|BN_DIGIT_SIZE|/2}$.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |Rg|.}
   \side{$|*p_c| = a\times b$.}
   \note{this procedure is used only for large modulos with
   precomputations for 32 and 64 bits computers.}
   }
   */
void _ZepsEltMultiplyPrcLargep(p_c, a, b, Rg)
     ZENElt *p_c, a, b;
     ZENRing Rg;
{
#define BASE ((BigNumDigit)1 << (BN_DIGIT_SIZE/2))
#define BASE1 ((BigNumDigit)1 << (2*SIZE_CHAR))

  BigNumDigit al, ah, bl, bh, c, d, tl, th, t, p;

  p = ZBNDigit(ZENRingP(Rg), 0);
  
  /* a*b -> (c, d) */
  ah = ZepsElt2Int(a, Rg)/BASE;
  al = ZepsElt2Int(a, Rg)%BASE;
  bh = ZepsElt2Int(b, Rg)/BASE;
  bl = ZepsElt2Int(b, Rg)%BASE;
  
  c = (al*bl);
  d = ah*bh;
  
  t = al*bh;
  th = t/BASE;
  d += th;
  tl = ((t%BASE)*BASE);
  c += tl;
  if (c < tl) d++;

  t = ah*bl;
  th = t/BASE;
  d += th;
  tl = ((t%BASE)*BASE);
  c += tl;
  if (c < tl) d++;

  /* (c, d) % p -> *p_c */
  if (c >= p) c %= p;
  
  t = d%BASE1;
  c += ZepsMultTable0(Rg, t);
  if ((c < ZepsMultTable0(Rg, t)) || (c >= p)) c -= p;

  if (BN_DIGIT_SIZE == 8*SIZE_CHAR) {
    d = d/BASE1;
    t = d%BASE1;
    c += ZepsMultTable2(Rg, t);
    if ((c < ZepsMultTable2(Rg, t)) || (c >= p)) c -= p;

    d = d/BASE1;
    t = d%BASE1;
    c += ZepsMultTable3(Rg, t);
    if ((c < ZepsMultTable3(Rg, t)) || (c >= p)) c -= p;
  }

  d = d/BASE1;
  c += ZepsMultTable1(Rg, d);
  if ((c < ZepsMultTable1(Rg, d)) || (c >= p)) c -= p;

  ZepsElt2Int(*p_c, Rg) = c;

#  if 0
  {
    ZENElt c;

    _ZepsEltMultiplyLargep1(&c, a, b, Rg);
    if (ZepsElt2Int(*p_c, Rg) != ZepsElt2Int(c, Rg)) {
      printf("ARGH a bug in ZepsEltMultiplyLargep() !!!\n");
      exit(-1);
    }
  }
#  endif
}

/* \TEX{
   \title{Precomputations.}
   \input{A finite ring |Rg|.}
   \output{|ZENERR| if an error occured, |ZEN_HAS_INVERSE| otherwise.}
   \side{The finite field |Rg| is filled with precomputations for
   speeding up element multiplication.}
   }
   */
int _ZepsPrcInitEltMultiply(Rg)
     ZENRing Rg;
{
  BigNumDigit T[3], U[3], k, p;

  /* Allocation */
  if (!(ZepsMultTable0Ptr(Rg) = (ZEN_UL_2SC_Table *)ZENMalloc(sizeof(ZEN_UL_2SC_Table)))) {
    return(ZERR);
  }
  if (!(ZepsMultTable1Ptr(Rg) = (ZEN_UL_2SC_Table *)ZENMalloc(sizeof(ZEN_UL_2SC_Table)))) {
    ZENFree(ZepsMultTable0Ptr(Rg));
    return(ZERR);
  }
  if (BN_DIGIT_SIZE == 8*SIZE_CHAR) {
    if (!(ZepsMultTable2Ptr(Rg) = (ZEN_UL_2SC_Table *)ZENMalloc(sizeof(ZEN_UL_2SC_Table)))) {
      ZENFree(ZepsMultTable0Ptr(Rg));
      ZENFree(ZepsMultTable1Ptr(Rg));
      return(ZERR);
    }
    if (!(ZepsMultTable3Ptr(Rg) = (ZEN_UL_2SC_Table *)ZENMalloc(sizeof(ZEN_UL_2SC_Table)))) {
      ZENFree(ZepsMultTable0Ptr(Rg));
      ZENFree(ZepsMultTable1Ptr(Rg));
      ZENFree(ZepsMultTable2Ptr(Rg));
      return(ZERR);
    }
  }

  p = ZBNDigit(ZENRingP(Rg), 0);
  T[0] = T[1] = T[2] = 0;
  for(k = 0; k < (1<<(2*SIZE_CHAR)); k++) {
    T[1] = k;
    ZepsMultTable0(Rg, k) = ZBNDivideDigit(U, T, 3, p);

    if (BN_DIGIT_SIZE == 8*SIZE_CHAR) {
      T[1] <<=  2*SIZE_CHAR;
      ZepsMultTable2(Rg, k) = ZBNDivideDigit(U, T, 3, p);

      T[1] <<=  2*SIZE_CHAR;
      ZepsMultTable3(Rg, k) = ZBNDivideDigit(U, T, 3, p);
    }
    T[1] <<=  2*SIZE_CHAR;
    ZepsMultTable1(Rg, k) = ZBNDivideDigit(U, T, 3, p);
  }

  ZENEltMultiplyPtr(Rg) = _ZepsEltMultiplyPrcLargep;
  ZENEltSquarePtr(Rg) = _ZepsEltSquarePrcLargep;
  return(ZEN_HAS_INVERSE);
  
}

/* \TEX{
   \title{Free precomputations for element multiplication.}
   \input{A finite ring |Rg|.}
   \output{|ZEN_HAS_INVERSE|}.
   \side{Memory is desallocated.}
   }
   */
int _ZepsPrcCloseEltMultiply(Rg)
     ZENRing Rg;
{
  ZENFree(ZepsMultTable0Ptr(Rg));
  ZENFree(ZepsMultTable1Ptr(Rg));

  if (BN_DIGIT_SIZE == 8*SIZE_CHAR) {
    ZENFree(ZepsMultTable2Ptr(Rg));
    ZENFree(ZepsMultTable3Ptr(Rg));
  }

  ZENEltMultiplyPtr(Rg) = _ZepsEltMultiplyLargep;
  ZENEltSquarePtr(Rg) = _ZepsEltSquareLargep;
  return(ZEN_HAS_INVERSE);
}

