/*
 * zen/zeps/c/init.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   
   A finite ring is completely described by a variable of type
   |ZENRing|. Precomputations are needed to speedup the operations in this
   ring. In the particular case of $ZEN(p)$, with $p$ small, this is done with
   |_ZepsInit|. These precomutations are then free with |_ZepsClose|.
   }
   */
#include <stdio.h>
#include "zeps.h"
#include "zeg.h"
#include "zed.h"

/* \TEX{
   \title{Initialization.}
   \input{A partially defined |ZENRing|.}
   \output{$|SR_GEN|\vert |MAT_GEN|$.}
   \side{The initialization of $\Z/p\Z$ is completed.}
   \note{This procedure is called by |ZENBaseRingCreate| and should not be called otherwise.}
   }
   */
int _ZepsInit(R)
     ZENRing R;
{
  ZENRingArith(R) = ZEN_LIB_ZEPS;

  /* Procedure on rings
   **********************/
  if (ZBNDigit(ZENRingP(R), 0) > ((BigNumDigit)1 << (BN_DIGIT_SIZE>>1))) {
    ZENPrcInitEltMultiplyPtr(R) = _ZepsPrcInitEltMultiply;
    ZENPrcCloseEltMultiplyPtr(R) = _ZepsPrcCloseEltMultiply;
  }
  else {
    ZENPrcInitEltMultiplyPtr(R) = NULL;
    ZENPrcCloseEltMultiplyPtr(R) = NULL;
  }

  ZENPrcInitPolyRootsCanonicalPtr(R) = NULL;
  ZENPrcClosePolyRootsCanonicalPtr(R) = NULL;

  ZENPrcInitEltExpPtr(R) = _ZENPrcInitEltExp;
  ZENPrcCloseEltExpPtr(R) = _ZENPrcCloseEltExp;

  /* Procedure on elements
   ****************************/
  /* Allocation */
  ZENEltCreatePtr(R) = _ZepsEltCreate;	/* Create elements */
  ZENEltCreateLargePtr(R) = _ZepsEltCreate;
  ZENEltFreePtr(R) = _ZepsEltFree;	/* Free elements */

  /* Assigning */
  ZENEltAssignPtr(R) = _ZepsEltAssign;	/* Assign an element with another */
  ZENEltSetToZeroPtr(R) = _ZepsEltSetToZero; /* Set an element to zero */
  ZENEltSetToOnePtr(R) = _ZepsEltSetToOne; /* Set an element to one */
  ZENEltSetRandomPtr(R) = _ZepsEltSetRandom; /* Set randomly an element */
  ZENEltSetNextPtr(R) = _ZepsEltSetNext; /* Set an element to the next */

  /* Test */
  ZENEltAreEqualPtr(R) = _ZepsEltAreEqual; /* Is an element equal to another */
  ZENEltIsZeroPtr(R) = _ZepsEltIsZero;	/* Is an element equal to zero */
  ZENEltIsOnePtr(R) = _ZepsEltIsOne;	/* Is an element equal to one */
  ZENEltIsASquarePtr(R) = _ZENEltIsASquareCharp; /* Is an element a square in a finite field ? */

  /* Arithmetic */
  ZENEltAddPtr(R) = _ZepsEltAdd;	/* Addition of elements */
  ZENEltNegatePtr(R) = _ZepsEltNegate;	/* Negation of an element */
  ZENEltSubtractPtr(R) = _ZepsEltSubtract; /* Substraction of elements */
  if (ZBNDigit(ZENRingP(R), 0) > ((BigNumDigit)1 << (BN_DIGIT_SIZE>>1)))
    {
      ZENEltSquarePtr(R) = _ZepsEltSquareLargep;	/* Squaring of elements */
      ZENEltMultiplyPtr(R) = _ZepsEltMultiplyLargep; /* Multiplication of elements */
      /*
      if (ZBNDigit(ZENRingP(R), 0) < ((BigNumDigit)1 << (BN_DIGIT_SIZE-1)))
	{
	  ZepsElt2Int(ZENRingHbuf(R), R) = ((BigNumDigit)1 << (BN_DIGIT_SIZE-1))%ZBNDigit(ZENRingP(R), 0);
	  ZepsElt2Int(ZENRingHbuf(R), R) *= 2;
	  if (ZepsElt2Int(ZENRingHbuf(R), R) >= ZBNDigit(ZENRingP(R), 0))
	    {
	      ZepsElt2Int(ZENRingHbuf(R), R) -= ZBNDigit(ZENRingP(R), 0);
	    }
	}
      else
	{
	  ZepsElt2Int(ZENRingHbuf(R), R) = -ZBNDigit(ZENRingP(R), 0);
	}
	*/
    }
  else
    {
      ZENEltSquarePtr(R) = _ZepsEltSquareSmallp;	/* Squaring of elements */
      ZENEltMultiplyPtr(R) = _ZepsEltMultiplySmallp; /* Multiplication of elements */
    }
  ZENEltInversePtr(R) = _ZepsEltInverse; /* Inverse of elements */
  ZENEltTracePtr(R) = _ZepsEltTrace;		/* Trace of elements */
  ZENEltSquareRootPtr(R) = _ZENEltSquareRootCharp; /* Square root of an element in finite fields */
  ZENEltExpPtr(R) = _ZENEltExpPlain;	/* Exponentiation */

  /* io */
  ZENEltReadFromStringPtr(R) = _ZepsEltReadFromString;	/* Read an element from a string */
  ZENEltPrintToStringPtr(R) = _ZepsEltPrintToString; /* Print an element to a string */
  ZENEltReadFromFilePtr(R) = _ZepsEltReadFromFile; /* Read an element from a file */
  ZENEltPrintToFilePtr(R) = _ZepsEltPrintToFile; /* Print an element to a file */
  ZENEltGetFromStringPtr(R) = _ZepsEltGetFromString; /* Get an element from a string */
  ZENEltPutToStringPtr(R) = _ZepsEltPutToString; /* Put an element to a string */
  ZENEltGetFromFilePtr(R) = _ZepsEltGetFromFile; /* Get an element from a file */
  ZENEltPutToFilePtr(R) = _ZepsEltPutToFile; /* Put an element to a file */

  ZENZToEltPtr(R) = _ZepsZToElt; /* Evaluate in $p$ an element */
  ZENEltToZPtr(R) = _ZepsEltToZ; /* Get the element from an evaluation in $p$ */
  ZENEltConvertPtr(R) = _ZepsEltConvert;
  
  /* Procedures on polynomials
   *******************************/
  /* Allocation */
  ZENPolyCreatePtr(R) = _ZedsPolyCreate; /* Create polynomials */
  ZENPolyCreateLargePtr(R) = _ZedsPolyCreate; /* Create polynomials */
  ZENPolyFreePtr(R) = _ZedsPolyFree;	/* Free polynomials */

  /* Assigning */
  ZENPolyAssignPtr(R) = _ZepsPolyAssign; /* Assign a polynomial with another */
  ZENPolyAssignLargePtr(R) = _ZepsPolyAssign; /* Assign a polynomial with another */
  ZENPolyLazyToVeryLazyPtr(R) = _ZepsPolyAssign; /* Assign a polynomial with another */
  ZENPolySetToZeroPtr(R) = _ZepsPolySetToZero;	/* Set a polynomial to zero */
  ZENPolySetToZeroLargePtr(R) = _ZepsPolySetToZero;	/* Set a polynomial to zero */
  ZENPolySetToXiPtr(R) = _ZepsPolySetToXi;	/* Set a polynomial to X^i */
  ZENPolySetRandomPtr(R) = _ZepsPolySetRandom;	/* Set randomly a polynomial */
  ZENPolySetCoeffPtr(R) = _ZepsPolySetCoeff; /* Set a coefficient of a polynomial */
  ZENPolyGetCoeffNewPtr(R) = _ZepsPolyGetCoeff; /* Get a coefficient of a polynomial */
  ZENPolyGetCoeffPtrPtr(R) = _ZepsPolyGetCoeffPtr; /* Get a pointer on coefficient of a polynomial */
  ZENPolyUpdateDegreePtr(R) = _ZepsPolyUpdateDegree; /* Get a coefficient of a polynomial */

  /* Test */
  ZENPolyAreEqualPtr(R) = _ZepsPolyAreEqual; /* Is a polynomial equal to another */
  ZENPolyIsZeroPtr(R) = _ZepsPolyIsZero; /* Is a polynomial equal to zero */
  ZENPolyIsXiPtr(R) = _ZepsPolyIsXi; /* Is a polynomial equal to $X^i$ */

  /* Arithmetic */
  ZENPolyAddPtr(R) = _ZepsPolyAdd;	/* Addition of polynomials */
  ZENPolyAddLargePtr(R) = _ZepsPolyAdd;	/* Addition of polynomials */
  ZENPolyNegatePtr(R) = _ZepsPolyNegate; /* Negation of polynomials */
  ZENPolyNegateLargePtr(R) = _ZepsPolyNegate; /* Negation of polynomials */
  ZENPolySubtractPtr(R) = _ZepsPolySubtract; /* Substraction of polynomials */
  ZENPolySubtractLargePtr(R) = _ZepsPolySubtract; /* Substraction of polynomials */

  ZENPolySquarePtr(R) = _ZepsPolySquareKaratsuba; /* Squaring of polynomials */
  ZENPolySquareLazyPtr(R) = _ZepsPolySquareKaratsubaLazy;
  ZENPolySquareVeryLazyPtr(R) = _ZepsPolySquareKaratsubaLazy;

  ZENPolyMultiplyPtr(R) = _ZepsPolyMultiply; /* Multiplication of polynomials */
  ZENPolyMultiplyLazyPtr(R) = _ZepsPolyMultiplyLazy;
  ZENPolyMultiplyVeryLazyPtr(R) = _ZepsPolyMultiplyLazy;

  ZENPolyMultiplyScalarPtr(R) = _ZepsPolyMultiplyScalar; /* Multiplication of polynomials by a scalar*/
  ZENPolyDotPtr(R) = _ZepsPolyDot; /* Scalar multiplication of polynomials */
  ZENPolyEvalPtr(R) = _ZENPolyEval; /* Evaluation of a polynomial */
  ZENPolyDividePtr(R) = _ZepsPolyDivide; /* Division of polynomials */
  ZENPolyDivideLazyPtr(R) = _ZepsPolyDivideLazy; /* Lazy division of polynomials */
  ZENPolyDivideVeryLazyPtr(R) = _ZepsPolyDivideLazy; /* Lazy division of polynomials */
  ZENPolyModuloLazyPtr(R) = _ZepsPolyModuloLazy; /* Lazy modulo of polynomials */
  ZENPolyModuloVeryLazyPtr(R) = _ZepsPolyModuloLazy; /* Lazy modulo of polynomials */
  ZENPolyDerivePtr(R) = _ZENPolyDerive;
  ZENPolyRootsDegree2Ptr(R) = _ZENPolyRootsDegree2Charp; /* Roots of polynomials of degree 2 in finite fields */

  ZENMatCreateParamsPtr(R) = _ZedsMatCreateParams;	
  ZENMatCreatePtr(R) = __ZedsMatCreate;

  /* An element of trace 1 */
  ZENEltSetToOne(ZENRingTheta(R), R);

  ZENMatMultiplyPtr(R) = _ZENMatMultiplyPlain; 
  return(SR_GEN | MAT_GEN);
}

/* \TEX{
   \title{Closing.}
   \input{A partially freed |ZENRing|.}
   \output{0.}
   }
   */
int _ZepsClose(R)		
     ZENRing R;
{
  return(0);
}
