/*
 * zen/zeps/c/poly_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial assignment}
   
   One can set a |ZENPoly| to zero with |_ZepsPolySetToZero|. Then we can
   assign a |ZENPoly| with an other with |_ZepsPolyAssign|. Furthermore, we
   can set randomly a |ZENPoly| with |_ZepsPolySetRandom|. 
   
   The functions |_ZepsPolyGetCoeff| and |_ZepsPolySetCoeff|
   enable us to get or set the coefficient of $X^i$,
   |_ZepsPolyUpdateDegree| compute the real degree of a |ZENPoly| if it
   decreased.
   }
   */
#include <stdio.h>
#include "zeps.h"

#define min(a,b) ((a) < (b) ? (a) : (b))

/* \TEX{
   \title{Assign.}
   \input{Two |ZENPoly| |PX| and |RX| of a finite ring |Rg|.}
   \side{|RX| is filled with |PX|.}
   }
   */
void _ZepsPolyAssign(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZepsPolyDeg(PX, Rg); i++) {
    ZepsEltAssign(ZepsPolyGetCof(RX, i, Rg),
		  ZepsPolyGetCof(PX, i, Rg),
		  Rg);
  }
  ZepsPolyDeg(RX, Rg) =  ZepsPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Set to zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{|RX| is set to zero..}
   }
   */
void _ZepsPolySetToZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  ZepsPolyDeg(RX, Rg) =  0;
  ZepsEltSetToZero(ZepsPolyGetCof(RX, 0, Rg), Rg);
}

/* \TEX{
   \title{Set to $X^i$.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg| allocated at least for
   degree $|deg|$.}
   \side{|RX| is set to $X^{|deg|}$.}
   }
   */
void _ZepsPolySetToXi(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  int i;

  ZepsPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    ZepsEltSetToZero(ZepsPolyGetCof(RX, i, Rg), Rg);
  }
  ZepsEltSetToOne(ZepsPolyGetCof(RX, deg, Rg), Rg);
}

/* \TEX{
   \title{Set to random.}
   \input{A |ZENPoly| |RX| of degree |deg| of a finite ring |Rg|.}
   \side{|RX| is set randomly with degree |deg|.}
   }
   */
void _ZepsPolySetRandom(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  int i;

  ZepsPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    ZepsEltSetRandom(ZepsPolyGetCof(RX, i, Rg), Rg);
  }
  ZepsEltSetRandom(ZepsPolyGetCof(RX, deg, Rg), Rg);
  while(ZepsEltIsZero(ZepsPolyGetCof(RX, deg, Rg), Rg)) {
    ZepsEltSetRandom(ZepsPolyGetCof(RX, deg, Rg), Rg);
  }
}

/* \TEX{
   \title{Set a coeff.}
   \input{A |ZENPoly| |RX| and a |ZENElt| |b| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \side{The coeff of $X^d$ in |RX| is set to |b|.}
   }
   */
void _ZepsPolySetCoeff(RX, d, b, Rg)
     ZENPoly RX;
     int d;
     ZENElt b;
     ZENRing Rg;
{
  ZepsPolySetCoeff(RX, d, b, Rg);
}

/* \TEX{
   \title{Get a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The coeff of $X^d$ in |RX|.}
   }
   */
void _ZepsPolyGetCoeff(p_c, RX, d, Rg)
     ZENElt *p_c;
     ZENPoly RX;
     int d;
     ZENRing Rg;
{ 
  ZepsPolyGetCoeff(*p_c, RX, d, Rg);
}

/* \TEX{
   \title{Extract a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The pointer to the coeff of $X^d$ in |RX|.}
   }
   */
ZENElt _ZepsPolyGetCoeffPtr(RX, d, Rg)
     ZENPoly RX;
     int d;
     ZENRing Rg;
{
  return(ZepsPolyGetCoeffPtr(RX,d,Rg));
}

/* \TEX{
   \title{Update the degree of a polynomial.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{Resetting the degree of |P|, assuming it has decreased.}
   }
   */
void _ZepsPolyUpdateDegree(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  int i;

  for(i = ZepsPolyDeg(RX, Rg); i >= 0; i--) {
    if (!ZepsEltIsZero(ZepsPolyGetCoeffPtr(RX,i,Rg), Rg)) {
      break;
    }
  }
  if (i < 0) {
    ZepsPolySetToZero(RX, Rg);
  }
  else {
    ZepsPolyDeg(RX, Rg) =  i;
  }
}
