/*
 * zen/zeps/c/poly_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Polynomial multiplication}
   
   Two functions are provided to do multiplications: |_ZepsPolySquare| to square a
   |ZENPoly| and |_ZepsPolyMultily| to compute the multiplication of two |ZENPoly|.
   }
   */
#include <stdio.h>
#include "zeps.h"
#include "zeg.h"
#include "zed.h"

#define ZEPS_POLY_MULTIPLY_CUTOFF 1
#define ZEPS_POLY_SQUARE_CUTOFF 4

/* \TEX{
   \title{Squaring.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZepsPolySquarePlain(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, j;

  for(i = ZepsPolyDeg(PX, Rg)+1; i <= 2*ZepsPolyDeg(PX, Rg); i++) {
    ZepsEltSetToZero(ZepsPolyGetCof(RX, i, Rg), Rg);
  }

  for(i = ZepsPolyDeg(PX, Rg); i > 0; i--) {
    
    ZepsEltSquare(ZENRingHbuf(Rg), ZepsPolyGetCof(PX, i, Rg), Rg);
    ZepsEltAdd(ZepsPolyGetCof(RX, 2*i, Rg), ZENRingHbuf(Rg), Rg);
    
    for(j = i-1; j > 0; j--) {
      ZepsEltMultiply(ZENRingHbuf(Rg),
		      ZepsPolyGetCof(PX, j, Rg),
		      ZepsPolyGetCof(PX, i, Rg),
		      Rg);
      ZepsEltAdd(ZENRingHbuf(Rg), ZENRingHbuf(Rg), Rg);
      ZepsEltAdd(ZepsPolyGetCof(RX, i+j, Rg), ZENRingHbuf(Rg), Rg);
    }
    ZepsEltMultiply(ZENRingHbuf(Rg),
		    ZepsPolyGetCof(PX, j, Rg),
		    ZepsPolyGetCof(PX, i, Rg),
		    Rg);
    ZepsEltAdd(ZENRingHbuf(Rg), ZENRingHbuf(Rg), Rg);
    ZepsEltAssign(ZepsPolyGetCof(RX, i+j, Rg), ZENRingHbuf(Rg), Rg);
  }
  ZepsEltSquare(ZepsPolyGetCof(RX, 0, Rg), ZepsPolyGetCof(PX, 0, Rg), Rg);
    
  ZepsPolyDeg(RX, Rg) = 2*ZepsPolyDeg(PX, Rg);
  return(0);
}

/* \TEX{
   \title{Squaring with Karatsuba.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZepsPolySquareKaratsuba(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{  
  int i;
  ZENPoly RBuf;

  if (!(RBuf = ZepsPolyCreateLarge(4*ZepsPolyDeg(PX, Rg)+1, Rg))) {
    return(ZERR);
  }
  ZepsPolySquareKaratsubaLazy(RBuf, PX, Rg);
  ZepsPolyAssign(RX, RBuf, Rg);
  ZepsPolyFree(RBuf, Rg);
  return(ZEN_HAS_INVERSE);
}

static void ZepsPolySquareKaratsubaLazyDegree(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  struct zenpoly UP1, UP3;
  ZENPoly up1, up3;
  int m, n, i, j;
  
  /* Initialization ***********/
  up1 = &UP1; up3 = &UP3;
  m = (ZepsPolyDeg(PX, Rg)+1)/2; 
  n = ZepsPolyDeg(PX, Rg)+1-m;
  
  /* let's go with the recursivity **********/
  
  /* PX- ^ 2 -> RX */
  ZepsPolyDeg(up1, Rg) = m-1;
  ZepsPolyCof(up1, Rg) = ZepsPolyCof(PX, Rg);
  ZepsPolyCof(up3, Rg) = ZepsPolyCof(RX, Rg);
  ZepsPolySquareKaratsubaLazy(up3, up1, Rg);
  
  /* PX- + PX+ -> RX+4*m+2*n-2 */
  for(j = 0, i = 4*m+2*n-2;
      j < m;
      j++, i++) {
    ZepsEltAssign(ZepsPolyGetCof(RX, i, Rg),
		  ZepsPolyGetCof(PX, j+m, Rg),
		  Rg);
    ZepsEltAdd(ZepsPolyGetCof(RX, i, Rg), ZepsPolyGetCof(PX, j, Rg), Rg);
  }
  if (n > m) {
    ZepsEltAssign(ZepsPolyGetCof(RX, i, Rg),
		  ZepsPolyGetCof(PX, j+m, Rg),
		  Rg);
  }
  
  /* (PX- + PX+) ^ 2 -> RX+2*m-1 */
  ZepsPolyDeg(up1, Rg) = n-1;
  ZepsPolyCof(up1, Rg) = ZepsPolyCof(RX, Rg)+4*m+2*n-2;
  ZepsPolyCof(up3, Rg) = ZepsPolyCof(RX, Rg)+2*m-1;
  ZepsPolySquareKaratsubaLazy(up3, up1, Rg);
  
  /* (PX- + PX+) ^ 2 - PX- ^ 2  -> RX+2*m-1 */
  for(j = 0, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZepsEltSubtract(ZepsPolyGetCof(RX, i, Rg), ZepsPolyGetCof(RX, j, Rg), Rg);
  }
  
  /* [ PX- ^ 2, (PX- + PX+) ^ 2 - PX- ^ 2 ] -> RX+m */
  for(j = m, i = 2*m-1; j < 2*m-1; j++, i++) {
    ZepsEltAdd(ZepsPolyGetCof(RX, j, Rg), ZepsPolyGetCof(RX, i, Rg), Rg);
  }
  for(;j < n+2*m-1; j++, i++) {
    ZepsEltAssign(ZepsPolyGetCof(RX, j, Rg), ZepsPolyGetCof(RX, i, Rg), Rg);
  }
  
  /* PX+ ^ 2 -> RX+ */
  ZepsPolyDeg(up1, Rg) = n-1;
  ZepsPolyCof(up1, Rg) = ZepsPolyCof(PX, Rg)+m;
  ZepsPolyCof(up3, Rg) = ZepsPolyCof(RX, Rg)+2*m+n-1;
  ZepsPolySquareKaratsubaLazy(up3, up1, Rg);

  /* [ PX- ^ 2, (PX- + PX+) ^ 2 - PX- ^2 - PX+ ^ 2 , PX+ ^ 2 ] -> RX */
  for(j = m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZepsEltSubtract(ZepsPolyGetCof(RX, j, Rg), ZepsPolyGetCof(RX, i, Rg), Rg);
  }
  for(j = 2*m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZepsEltAdd(ZepsPolyGetCof(RX, j, Rg), ZepsPolyGetCof(RX, i, Rg), Rg);
  }
  for(; j < 2*m+2*n-1; j++, i++) {
    ZepsEltAssign(ZepsPolyGetCof(RX, j, Rg), ZepsPolyGetCof(RX, i, Rg), Rg);
  }

  ZepsPolyDeg(RX, Rg) = 2*ZepsPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Lazy Squaring with Karatsuba.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0 for compatibylity.}
   \side{$|RX| = |PX|^2$.}
   \note{One must have $|RX|\neq|PX|$. |PX| is destroyed at the return.}
   }
   */
int ZepsPolySquareKaratsubaLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  if (ZepsPolyDeg(PX, Rg) < ZEPS_POLY_SQUARE_CUTOFF) {
    ZepsPolySquarePlain(RX, PX, Rg);
    return(ZEN_HAS_INVERSE);
  }
  ZepsPolySquareKaratsubaLazyDegree(RX, PX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|$.}
   }
   */
int _ZepsPolyMultiply(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZENPoly RBuf;
  int n, m, q, r, L, k;

  if (ZepsPolyDeg(PX, Rg) > ZepsPolyDeg(QX, Rg)) {
    n = ZepsPolyDeg(PX, Rg)+1; m = ZepsPolyDeg(QX, Rg)+1;
    q = n/m; r = n%m;
  }
  else {
    n = ZepsPolyDeg(QX, Rg)+1; m = ZepsPolyDeg(PX, Rg)+1;
    q = n/m; r = n%m;
  }
  switch(q) {
  case 1:
    L = r ? 4*m-4+2*r : 4*m-3;
    break;
  case 2:
    L = 2*r > (m+1) ? 5*m-5+2*r : 6*m-4;
    break;
  default:
    L = 2*r > m ?  (3+q)*m-4+2*r: (4+q)*m-4;
    break;
  }
  if (!(RBuf = ZepsPolyCreateLarge(L, Rg))) {
    return(ZERR);
  }
  ZepsPolyMultiplyLazy(RBuf, PX, QX, Rg);
  ZepsPolyAssign(RX, RBuf, Rg);
  ZepsPolyFree(RBuf, Rg);

  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Lazy multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0 for compatibylity.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|\neq|QX|$.}
   }
   */
int _ZepsPolyMultiplyLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  struct zenpoly sxzp, rxzp, pxzp, qxzp;
  ZENPoly sx, rx, px, qx;
  int i, dp, dq, quo, rem;

  /* Switch
  **********/
  if (ZepsPolyDeg(PX, Rg) == ZepsPolyDeg(QX, Rg)) {
    ZepsPolyMultiplyKaratsubaLazy(RX, PX, QX, Rg);
    return(ZEN_HAS_INVERSE);
  }
  if (ZepsPolyDeg(PX, Rg) > ZepsPolyDeg(QX, Rg)) {
    if ((ZepsPolyDeg(QX, Rg)) <= ZEPS_POLY_MULTIPLY_CUTOFF) {
      ZepsPolyMultiplyPlain(RX, PX, QX, Rg);
      return(ZEN_HAS_INVERSE);
    }
    px = &pxzp;
    ZepsPolyCof(px, Rg) = ZepsPolyCof(PX, Rg);
    dp = ZepsPolyDeg(PX, Rg);

    qx = &qxzp;
    ZepsPolyCof(qx, Rg) = ZepsPolyCof(QX, Rg);
    dq = ZepsPolyDeg(QX, Rg);
  }
  else {
    if (ZepsPolyDeg(PX, Rg) <= ZEPS_POLY_MULTIPLY_CUTOFF) {
      ZepsPolyMultiplyPlain(RX, PX, QX, Rg);
      return(ZEN_HAS_INVERSE);
    }
    px = &pxzp;
    ZepsPolyCof(px, Rg) = ZepsPolyCof(QX, Rg);
    dp = ZepsPolyDeg(QX, Rg);

    qx = &qxzp;
    ZepsPolyCof(qx, Rg) = ZepsPolyCof(PX, Rg);
    dq = ZepsPolyDeg(PX, Rg);
  }

  /* We have to cut px in pieces 
    **********************************/  
  /* Initialization */
  ZepsPolyDeg(RX, Rg) = dp+dq;

  rx = &rxzp;
  ZepsPolyCof(rx, Rg) = ZepsPolyCof(RX, Rg);

  sx = &sxzp;
  ZepsPolyCof(sx, Rg) = ZepsPolyCof(RX, Rg);

  quo = (dp+1)/(dq+1);
  rem = (dp+1)%(dq+1);

  /* First multiplication */
  ZepsPolyDeg(px, Rg) = dq;
  ZepsPolyDeg(qx, Rg) = dq;
  ZepsPolyMultiplyKaratsubaLazy(rx, qx, px, Rg);
  if ((quo == 1) && !rem) {
    return(ZEN_HAS_INVERSE);
  }

  /* Second multiplication */
  ZepsPolyCof(px, Rg) += dq+1;
  ZepsPolyDeg(px, Rg) = (quo == 1 ? rem-1: dq);
  ZepsPolyCof(rx, Rg) += 2*dq+1;
  if (quo == 1) {
    ZepsPolyMultiplyLazy(rx, qx, px, Rg);
  }
  else {
    ZepsPolyMultiplyKaratsubaLazy(rx, qx, px, Rg);
  }
  ZepsPolyDeg(sx, Rg) = dq-1;
  ZepsPolyCof(sx, Rg) += dq+1;
  ZepsPolyAdd(sx, rx, Rg);
  if (quo == 1) {
    return(ZEN_HAS_INVERSE);
  }

  for(i = 2; i < quo; i++) {
    ZepsPolyCof(px, Rg) += (dq+1);
    ZepsPolyCof(rx, Rg) += (dq+1);

    ZepsPolyMultiplyKaratsubaLazy(rx, qx, px, Rg);
    
    ZepsPolyDeg(sx, Rg) = dq-1;
    ZepsPolyCof(sx, Rg) += dq+1;
    ZepsPolyAdd(sx, rx, Rg);
  }

  if (!rem) {
    return(ZEN_HAS_INVERSE);
  }

  ZepsPolyDeg(px, Rg) = rem-1;
  ZepsPolyCof(px, Rg) += (dq+1);
  ZepsPolyCof(rx, Rg) += (dq+1);
  ZepsPolyMultiplyLazy(rx, qx, px, Rg);
    
  ZepsPolyDeg(sx, Rg) = dq-1;
  ZepsPolyCof(sx, Rg) += dq+1;
  ZepsPolyAdd(sx, rx, Rg);

  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|$.}
   }
   */
int _ZepsPolyMultiplyPlain(RX, QX, PX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{

  int i, j;

  if (ZepsPolyIsZero(PX, Rg) || ZepsPolyIsZero(QX, Rg)) {
    ZepsPolySetToZero(RX, Rg);
    return(0);
  }

  for(i = ZepsPolyDeg(PX, Rg)+1;
      i <= ZepsPolyDeg(PX, Rg)+ZepsPolyDeg(QX, Rg);
      i++) {
    ZepsEltSetToZero(ZepsPolyGetCof(RX, i, Rg), Rg);
  }

  for(i = ZepsPolyDeg(PX, Rg); i >= 0; i--) {
    for(j = ZepsPolyDeg(QX, Rg); j; j--) {
      ZepsEltMultiply(ZENRingHbuf(Rg),
		      ZepsPolyGetCof(QX, j, Rg),
		      ZepsPolyGetCof(PX, i, Rg),
		      Rg);
      ZepsEltAdd(ZepsPolyGetCof(RX, i+j, Rg), ZENRingHbuf(Rg), Rg);
    }
    ZepsEltMultiply(ZENRingHbuf(Rg),
		    ZepsPolyGetCof(QX, j, Rg),
		    ZepsPolyGetCof(PX, i, Rg),
		    Rg);
    ZepsEltAssign(ZepsPolyGetCof(RX, i+j, Rg), ZENRingHbuf(Rg), Rg);
  }
  ZepsPolyDeg(RX, Rg) = ZepsPolyDeg(PX, Rg) + ZepsPolyDeg(QX, Rg);
  return(0);
}

/* \TEX{
   \title{Polynomial multiplication with the Karatsuba method.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an errpr occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|$. If $\deg(|PX|) \ge deg(|QX|)$,
   one polynomial buffer of size |PX| is allocated, otherwise 2
   buffers of size |QX| are allocated.}
   }
   */
int _ZepsPolyMultiplyKaratsuba(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i;
  ZENPoly QBuf, RBuf;
  ZENElt elt;

  if (ZepsPolyDeg(PX, Rg) >= ZepsPolyDeg(QX, Rg)) {
    if (!(QBuf = ZepsPolyCreate(ZepsPolyDeg(PX, Rg), Rg))) {
      return(ZERR);
    }
    ZepsPolyAssign(QBuf, QX, Rg);
    for(i = ZepsPolyDeg(QX, Rg)+1;
	i <= ZepsPolyDeg(PX, Rg);
	i++) {
      elt = ZepsPolyGetCoeffPtr(QBuf, i, Rg);
      ZepsEltSetToZero(elt, Rg);
      ZepsPolySetCoeff(QBuf, i, elt, Rg);
    }

    if (!(RBuf = ZepsPolyCreateLarge(4*ZepsPolyDeg(PX, Rg)+1, Rg))) {
      return(ZERR);
    }

    ZepsPolyMultiplyKaratsubaLazy(RBuf, PX, QBuf, Rg);
  }
  else {
    if (!(QBuf = ZepsPolyCreate(ZepsPolyDeg(QX, Rg), Rg))) {
      return(ZERR);
    }
    ZepsPolyAssign(QBuf, PX, Rg);
    for(i = ZepsPolyDeg(PX, Rg)+1;
	i <= ZepsPolyDeg(QX, Rg);
	i++) {
      elt = ZepsPolyGetCoeffPtr(QBuf, i, Rg);
      ZepsEltSetToZero(elt, Rg);
      ZepsPolySetCoeff(QBuf, i, elt, Rg);
    }

    if (!(RBuf = ZepsPolyCreateLarge(4*ZepsPolyDeg(PX, Rg)+1, Rg))) {
      return(ZERR);
    }

    ZepsPolyMultiplyKaratsubaLazy(RBuf, QX, QBuf, Rg);
  }
  ZepsPolyAssign(RX, RBuf, Rg);
  ZepsPolyFree(QBuf, Rg); ZepsPolyFree(RBuf, Rg);
  return(ZEN_HAS_INVERSE);
}


static void ZepsPolyMultiplyKaratsubaLazyDegree0(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZepsEltMultiply(ZepsPolyGetCof(RX, 0, Rg),
		   ZepsPolyGetCof(PX, 0, Rg),
		   ZepsPolyGetCof(QX, 0, Rg), Rg);
  ZepsPolyDeg(RX, Rg) = 0;
}

static void ZepsPolyMultiplyKaratsubaLazyDegree1(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZepsEltAssign(ZepsPolyGetCof(RX, 0, Rg), ZepsPolyGetCof(PX, 0, Rg), Rg);
  ZepsEltAdd(ZepsPolyGetCof(RX, 0, Rg), ZepsPolyGetCof(PX, 1, Rg), Rg);
  
  ZepsEltAssign(ZepsPolyGetCof(RX, 2, Rg), ZepsPolyGetCof(QX, 0, Rg), Rg);
  ZepsEltAdd(ZepsPolyGetCof(RX, 2, Rg), ZepsPolyGetCof(QX, 1, Rg), Rg);

  ZepsEltMultiply(ZepsPolyGetCof(RX, 1, Rg),
		   ZepsPolyGetCof(RX, 0, Rg),
		   ZepsPolyGetCof(RX, 2, Rg), Rg);
  
  ZepsEltMultiply(ZepsPolyGetCof(RX, 0, Rg),
		   ZepsPolyGetCof(PX, 0, Rg),
		   ZepsPolyGetCof(QX, 0, Rg), Rg);
  
  ZepsEltMultiply(ZepsPolyGetCof(RX, 2, Rg),
		   ZepsPolyGetCof(PX, 1, Rg),
		   ZepsPolyGetCof(QX, 1, Rg), Rg);
  
  ZepsEltSubtract(ZepsPolyGetCof(RX, 1, Rg), ZepsPolyGetCof(RX, 0, Rg), Rg);
  ZepsEltSubtract(ZepsPolyGetCof(RX, 1, Rg), ZepsPolyGetCof(RX, 2, Rg), Rg);

  ZepsPolyDeg(RX, Rg) = 2;
}

static void ZepsPolyMultiplyKaratsubaLazyHeart ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

static void ZepsPolyMultiplyKaratsubaLazyDegree(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  struct zenpoly UP1, UP2, UP3;
  ZENPoly up1, up2, up3;
  int m, n, i, j;
  
  /* Initialization ***********/
  up1 = &UP1; up2 = &UP2; up3 = &UP3;
  m = (ZepsPolyDeg(PX, Rg)+1)/2; 
  n = ZepsPolyDeg(PX, Rg)+1-m;
  
  /* let's go with the recursivity **********/
  
  /* PX- * QX- -> RX */
  ZepsPolyDeg(up1, Rg) = m-1;
  ZepsPolyCof(up1, Rg) = ZepsPolyCof(PX, Rg);
  ZepsPolyCof(up2, Rg) = ZepsPolyCof(QX, Rg);
  ZepsPolyCof(up3, Rg) = ZepsPolyCof(RX, Rg);
  ZepsPolyMultiplyKaratsubaLazyHeart(up3, up1, up2, Rg);
  
  /* PX- + PX+ -> RX+4*m+2*n-2 */
  for(j = 0, i = 4*m+2*n-2;
      j < m;
      j++, i++) {
    ZepsEltAssign(ZepsPolyGetCof(RX, i, Rg),
		  ZepsPolyGetCof(PX, j+m, Rg),
		  Rg);
    ZepsEltAdd(ZepsPolyGetCof(RX, i, Rg), ZepsPolyGetCof(PX, j, Rg), Rg);
  }
  if (n > m) {
    ZepsEltAssign(ZepsPolyGetCof(RX, i, Rg),
		  ZepsPolyGetCof(PX, j+m, Rg),
		  Rg);
  }
  
  /* QX- + QX+ -> RX+4*m+3*n-2 */
  for(j = 0, i = 4*m+3*n-2;
      j < m;
      j++, i++) {
    ZepsEltAssign(ZepsPolyGetCof(RX, i, Rg),
		  ZepsPolyGetCof(QX, j+m, Rg),
		  Rg);
    ZepsEltAdd(ZepsPolyGetCof(RX, i, Rg), ZepsPolyGetCof(QX, j, Rg), Rg);
  }
  if (n > m) {
    ZepsEltAssign(ZepsPolyGetCof(RX, i, Rg),
		  ZepsPolyGetCof(QX, j+m, Rg),
		  Rg);
  }
  
  /* (PX- + PX+) * (QX- + QX+) -> RX+2*m-1 */
  ZepsPolyDeg(up1, Rg) = n-1;
  ZepsPolyCof(up1, Rg) = ZepsPolyCof(RX, Rg)+4*m+2*n-2;
  ZepsPolyCof(up2, Rg) = ZepsPolyCof(RX, Rg)+4*m+3*n-2;
  ZepsPolyCof(up3, Rg) = ZepsPolyCof(RX, Rg)+2*m-1;
  ZepsPolyMultiplyKaratsubaLazyHeart(up3, up1, up2, Rg);
  
  /* (PX- + PX+) * (QX- + QX+) - PX- * QX-  -> RX+2*m-1 */
  for(j = 0, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZepsEltSubtract(ZepsPolyGetCof(RX, i, Rg), ZepsPolyGetCof(RX, j, Rg), Rg);
  }
  
  /* [ PX- * QX-, (PX- + PX+) * (QX- + QX+) - PX- * QX- ] -> RX+m */
  for(j = m, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZepsEltAdd(ZepsPolyGetCof(RX, j, Rg), ZepsPolyGetCof(RX, i, Rg), Rg);
  }
  for(;j < n+2*m-1; j++, i++) {
    ZepsEltAssign(ZepsPolyGetCof(RX, j, Rg), ZepsPolyGetCof(RX, i, Rg), Rg);
  }
  
  /* PX+ * QX+ -> RX+ */
  ZepsPolyDeg(up1, Rg) = n-1;
  ZepsPolyCof(up1, Rg) = ZepsPolyCof(PX, Rg)+m;
  ZepsPolyCof(up2, Rg) = ZepsPolyCof(QX, Rg)+m;
  ZepsPolyCof(up3, Rg) = ZepsPolyCof(RX, Rg)+2*m+n-1;
  ZepsPolyMultiplyKaratsubaLazyHeart(up3, up1, up2, Rg);

  /* [ PX- * QX-, (PX- + PX+) * (QX- + QX+) - PX- * QX- - PX+ * QX+ , PX+ * QX+ ] -> RX */
  for(j = m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZepsEltSubtract(ZepsPolyGetCof(RX, j, Rg), ZepsPolyGetCof(RX, i, Rg), Rg);
  }
  for(j = 2*m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZepsEltAdd(ZepsPolyGetCof(RX, j, Rg), ZepsPolyGetCof(RX, i, Rg), Rg);
  }
  for(; j < 2*m+2*n-1; j++, i++) {
    ZepsEltAssign(ZepsPolyGetCof(RX, j, Rg), ZepsPolyGetCof(RX, i, Rg), Rg);
  }

  ZepsPolyDeg(RX, Rg) = ZepsPolyDeg(PX, Rg)+ZepsPolyDeg(QX, Rg);
}

static void ZepsPolyMultiplyKaratsubaLazyHeart(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  switch (ZepsPolyDeg(PX, Rg)) {

    /* compute the product of 2 polynoms whose degree is 0 */
  case 0:
    ZepsPolyMultiplyKaratsubaLazyDegree0(RX, PX, QX, Rg);
    break;

    /* compute the product of 2 polynoms whose degree is 1 */
  case 1: 
    ZepsPolyMultiplyKaratsubaLazyDegree1(RX, PX, QX, Rg); 
    break;

    /* compute the product of 2 polynoms whose degree is greater than 2*/
  default : 
    ZepsPolyMultiplyKaratsubaLazyDegree(RX, PX, QX, Rg); 
  }
}

int _ZepsPolyMultiplyKaratsubaLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  if (ZepsPolyDeg(PX, Rg) >= ZepsPolyDeg(QX, Rg)) {
    for(i = ZepsPolyDeg(QX, Rg)+1;
	i <= ZepsPolyDeg(PX, Rg);
	i++) {
      elt = ZepsPolyGetCoeffPtr(QX, i, Rg);
      ZepsEltSetToZero(elt, Rg);
      ZepsPolySetCoeff(QX, i, elt, Rg);
    }
  }
  else {
    for(i = ZepsPolyDeg(PX, Rg)+1;
	i <= ZepsPolyDeg(QX, Rg);
	i++) {
      elt = ZepsPolyGetCoeffPtr(PX, i, Rg);
      ZepsEltSetToZero(elt, Rg);
      ZepsPolySetCoeff(PX, i, elt, Rg);
    }
  }
  ZepsPolyMultiplyKaratsubaLazyHeart(RX, PX, QX, Rg);
  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Multiplication by a scalar.}
   \input{Two polynomials |PX|, |RX| and a |ZENElt| |e| of a finite ring |Rg|.}
   }
   */
void _ZepsPolyMultiplyScalar(RX, PX, e, Rg)
     ZENPoly RX, PX;
     ZENElt e;
     ZENRing Rg;
{
  int i;
  
  if (ZepsEltIsZero(e, Rg)) {
    ZepsPolySetToZero(RX, Rg);
    return;
  }
  if (ZepsEltIsOne(e, Rg)) {
    ZepsPolyAssign(RX, PX, Rg);
    return;
  }
  for(i = 0; i <= ZepsPolyDeg(PX, Rg); i++) {
    ZepsEltMultiply(ZepsPolyGetCof(RX, i, Rg),
		    e,
		    ZepsPolyGetCof(PX, i, Rg),
		    Rg);
  }
  ZENPolyDeg(RX, Rg) = ZENPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Scalar product.}
   \input{Two polynomials |PX|, |RX| and a pointer to a |ZENElt| |e| of a
   finite ring |Rg|.}
   \side{$|*e| = \sum p_iq_i$}
   }
   */
void _ZepsPolyDot(e, PX, QX, Rg)
     ZENElt *e;
     ZENPoly PX, QX;
     ZENRing Rg;
{
  int i, deg;
  
  ZepsEltMultiply(*e,
		  ZENPolyGetCof(PX, 0, Rg),
		  ZENPolyGetCof(QX, 0, Rg),
		  Rg);

  deg = ZepsPolyDeg(PX, Rg) < ZepsPolyDeg(QX, Rg) ?
    ZepsPolyDeg(PX, Rg) :  ZepsPolyDeg(QX, Rg);

  for(i = 1; i <= deg; i++) {
    ZepsEltMultiply(ZENRingHbuf(Rg),
		    ZepsPolyGetCof(PX, i, Rg),
		    ZepsPolyGetCof(QX, i, Rg),
		    Rg);
    ZENEltAdd(*e, ZENRingHbuf(Rg), Rg);
  }
}

/* \title{Divide.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured,  |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   \note{|PX| is destroyed.}
   */
int _ZepsPolyDivideLazy(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int i, j, deg, IsNotOne;

  if (!ZepsEltIsOne(ZepsPolyGetCof(QX, ZepsPolyDeg(QX, Rg), Rg), Rg)) {
    IsNotOne = 1;
    
    if (ZepsEltInverse(ZENRingIbuf(Rg),
		       ZepsPolyGetCof(QX, ZepsPolyDeg(QX, Rg), Rg),
		       Rg) == ZEN_NO_INVERSE) {
      return(ZEN_NO_INVERSE);
    }
  }
  else {
    IsNotOne = 0;
  }
  
  for(i = ZepsPolyDeg(PX, Rg);
      i >= ZepsPolyDeg(QX, Rg);
      i--) {
    if (IsNotOne) {
      ZepsEltMultiply(ZepsPolyGetCof(PX, i, Rg),
		      ZENRingIbuf(Rg), ZepsPolyGetCof(PX, i, Rg),
		      Rg);
    }
    
    for(j =  ZepsPolyDeg(QX, Rg)-1;
	j >= 0;
	j--) {
      ZepsEltMultiply(ZENRingHbuf(Rg),
		      ZepsPolyGetCof(PX, i, Rg),
		      ZepsPolyGetCof(QX, j, Rg),
		      Rg);
      ZepsEltSubtract(ZepsPolyGetCof(PX,
				     i+j-ZepsPolyDeg(QX, Rg),
				     Rg), ZENRingHbuf(Rg), Rg);
    }
  }
  deg =  ZepsPolyDeg(QX, Rg);
  for(i = ZepsPolyDeg(PX, Rg); i >= ZepsPolyDeg(QX, Rg); i--) {
    ZepsEltAssign(ZepsPolyGetCof(RX, i-ZepsPolyDeg(QX, Rg), Rg),
		   ZepsPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZepsPolyDeg(RX, Rg) = ZepsPolyDeg(PX, Rg)-ZepsPolyDeg(QX, Rg);
  if (ZepsPolyDeg(RX, Rg) < 0) {
    ZepsPolySetToZero(RX, Rg);
  }
  for(; i >= 0; i--) {
    if (!ZepsEltIsZero(ZepsPolyGetCof(PX, i, Rg), Rg)) {
      break;
    }
  }
  for(deg = i; i >= 0; i--) {
    ZepsEltAssign(ZepsPolyGetCof(MX, i, Rg),
		   ZepsPolyGetCof(PX, i, Rg),
		   Rg);
  }
  if (deg < 0) {
    ZepsPolySetToZero(MX, Rg);
  }
  else {
    ZepsPolyDeg(MX, Rg) = deg;
  }
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Divide.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZEN_NO_INVERSE| if
   a factor of the modulo was dicovered, |ZENERR| for an error.}
   \side{ |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZepsPolyDivide(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int ret;
  ZENPoly Pbuf;
  
  if (!(Pbuf = ZepsPolyCopy(PX, Rg))) {
    return(ZERR);
  }
  ret = ZepsPolyDivideLazy(RX, MX, Pbuf, QX, Rg);
  ZepsPolyFree(Pbuf, Rg);
  return(ret);
}


/* \TEX{
   \title{Lazy mudulo computation.}
   \input{A polynomial |PX| to divide by
   |QX|$=X^{\mbox{degree}(QX)}+sum_{i=0}^{|k|}q_iX^i$} in a finite
   ring |Rg|. |MX| will be the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| for an error}
   \note{|PX| is destroyed.}
   */
int _ZepsPolyModuloLazy(MX, PX, QX, k, Rg)
     ZENPoly MX, PX, QX;
     int k;
     ZENRing Rg;
{
  int i, j, deg;
  
  for(i = ZepsPolyDeg(PX, Rg);
      i >= ZepsPolyDeg(QX, Rg);
      i--) {
    for(j =  k;
	j >= 0;
	j--) {
      ZepsEltMultiply(ZepsRingEbuf(Rg),
		      ZepsPolyGetCof(PX, i, Rg),
		      ZepsPolyGetCof(QX, j, Rg),
		      Rg);
      ZepsEltSubtract(ZepsPolyGetCof(PX,
				     i+j-ZepsPolyDeg(QX, Rg),
				     Rg), ZepsRingEbuf(Rg), Rg);
    }
  }
  for(; i >= 0; i--) {
    if (!ZepsEltIsZero(ZepsPolyGetCof(PX, i, Rg), Rg)) {
      break;
    }
  }
  for(deg = i; i >= 0; i--) {
    ZepsPolySetCoeff(MX, i, ZepsPolyGetCof(PX, i, Rg), Rg);
  }
  if (deg < 0) {
    ZepsPolySetToZero(MX, Rg);
  }
  else {
    ZepsPolyDeg(MX, Rg) = deg;
  }

  return(ZEN_HAS_INVERSE);
}
