/*
 * zen/zer/c/chinese_ring.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   \Section{Chinese remainder ring initialization}
   }
   */
#include <stdio.h>
#include <zen.h>
#include <zer.h>
#include <zec.h>

extern int _ZENRingAreEqual ___((ZENRing, ZENRing));

/********************************************************/

/* \TEX{
   \title{Initialization of a ring using chinese remainder theorem.}
   \input{An array of |ZENRing|s of size |N|.} 
   \output{A ring |R| if everything was succesfully initialized, |ZENNULL| otherwise.}
   \note{The obtained ring is the product of the |N| inputs. These rings must be of the same type: either |N| modular rings, or |N| polynomial extensions over a same ring. Nevertheless, some or all of these ring can be clones.}

   In fact, two rings are allocated, the returned one and its |ZENRingOrigin()|. For instance, if |PR[0]| is $\quotient{\Z}{3\Z}$ and |PR[1]| is $\quotient{\Z}{5\Z}$, then |R=ZENChineseRingCreate(2,PR)| will affect to |R| $\quotient{\Z}{3\Z}\times\quotient{\Z}{5\Z}$, while |ZENRingOrigin(R)| will be set to  $\quotient{\Z}{15\Z}$. A subsequent call to |ZENRingFree| will also free this ring, but |R1=ZENRingCopy(R)| can be helpful to keep this original ring.

   In the extension case, the subring used for the creation of |ZENRingOrigin(R)| is the subring of |PR[0]|.
   The |ZENRing| table |PR| is duplicated, but not its rings. Hence, it can be freed but the input |ZENRing|s must be conserved.
   }
   */
ZENRing _ZENChineseRingCreate(N, PR)
     ZENRing *PR;
     int N;
{
  ZENRing R,Rg;
  BigNumLength nl;
  int i,j,ret,TYPE;
  arith ar;
  ZENPoly Q;
#if ZEC_AVAILABLE
# define UNKNOWN 0
# define BASE 1
# define EXT 2
# define ERROR -1
  TYPE=UNKNOWN;

  if(N<1)
    {
      ZENSetError(ZEN_LIB_ZER,ZEN_INIT,ZEN_INPUT);
      return(ZNULL);
    }
  
  nl=0;
  for(i=N;i--;)
    {
      ar=ZENRingArith(PR[i]);
      switch(ar)
	{
	case ZEN_LIB_ZEF:
	  ZENSetError(ZEN_LIB_ZER,ZEN_INIT,ZEN_INPUT);
	  return(ZNULL);
	  	  
	case ZEN_LIB_ZETAB:
	case ZEN_LIB_ZELOG:
	case ZEN_LIB_ZEC:
	case ZEN_LIB_ZEM:
	  ar=ZENRingArith(ZENRingOrigin(PR[i]));
	  break;
	default:
	  ;
	}
      switch(ar)
	{	 
	case ZEN_LIB_ZEP:
	case ZEN_LIB_ZE2:
	case ZEN_LIB_ZEPS:
	  if(!TYPE)
	    TYPE=BASE;
	  else if(TYPE!=BASE)
	    {
	      ZENSetError(ZEN_LIB_ZER,ZEN_INIT,ZEN_INPUT);
	      return(ZNULL);
	    }
	  nl+=ZENRingPl(PR[i]);  
	  break;
	case ZEN_LIB_ZEXT:
	  if(!TYPE)
	    TYPE=EXT;
	  else if(TYPE!=EXT)
	    {
	      ZENSetError(ZEN_LIB_ZER,ZEN_INIT,ZEN_INPUT);
	      return(ZNULL);
	    }
	  nl+=ZENPolyDeg(ZENRingPol(PR[i]),ZENRingDef(PR[i]));  
	  break;
	default:
	  ZENSetError(ZEN_LIB_ZER,ZEN_INIT,ZEN_UNKNOWN);
	  return(ZNULL);
	}
    }
  
  if(TYPE==BASE)
    {
      BigNum n,n1,tmp;

      n=ZBNC(nl);
      if(n==NULL)
	{
	  ZENSetError(ZEN_LIB_ZER,ZEN_INIT,ZEN_ALLOC);
	  return(ZNULL);
	}
      n1=ZBNC(nl);
      if(n1==NULL)
	{
	  ZENSetError(ZEN_LIB_ZER,ZEN_INIT,ZEN_ALLOC);
	  return(ZNULL);
	}
      i=N-1;
      ZBNAssign(n,ZENRingP(PR[i]),ZENRingPl(PR[i]));
      ZBNSetToZero(n+ZENRingPl(PR[i]),nl-ZENRingPl(PR[i]));
      for(;i--;)
	{
	  ZBNSetToZero(n1,nl);
	  ZBNMultiply(n1,nl,
		      n,ZBNNumDigits(n,nl),
		      ZENRingP(PR[i]),ZENRingPl(PR[i]));
	  tmp=n1;n1=n;n=tmp;
	}
      ZBNF(n1);
      nl=ZBNNumDigits(n,nl);
      if ((R = __ZENRingCreate(n,nl)) == ZENNULL)
	{
	  ZBNF(n);
	  return(ZNULL);
	}
      
      ZENRingQl(R) = nl;
      if (!(ZENRingQ(R) = ZBNC(nl))) {
	ZBNF(n);
	ZBNF(ZENRingP(R)); ZENFree(R);
	ZENSetError(ZEN_LIB_ZER, ZEN_INIT, ZEN_ALLOC);
	return(ZNULL);
      }
      ZBNAssign(ZENRingQ(R), n, nl);

      ZENRingPol(R) = NULL;
      ZENRingDef(R) = NULL;
      if((ZENRingOrigin(R) =  ZENBaseRingCreate(n,nl))==NULL)
	{
	  ZBNF(n);
	  ZBNF(ZENRingP(R)); ZBNF(ZENRingQ(R)); ZENFree(R);	  
	  return(ZNULL);
	}
      ZBNF(n);
      ZENRingVar(R)=0;

    }
  else if(TYPE==EXT)
    {
      ZENPoly P;

      Rg=ZENRingDef(PR[0]);
      if ((R = __ZENRingCreate(ZENRingP(Rg),ZENRingPl(Rg))) == ZENNULL)
	return(ZNULL);

      P=ZENPolyCreate(nl,Rg);
      if(P==NULL)
	{
	  ZBNF(ZENRingP(R));
	  ZENFree(R);
	  return(ZNULL);
	}
      ZENPolyAssign(P,ZENRingPol(PR[0]),Rg);
      for(i=N-1;i;i--)
	{
	  BigNum pp;
	  BigNumLength ppl;
	  
	  if(!ZENRingAreEqual(ZENRingDef(PR[i]),Rg))
	    {
	      ZENPolyFree(P,Rg);
	      ZBNF(ZENRingP(R));
	      ZENFree(R);
	      return(ZNULL);
	    }
	  pp=ZENPolyToZ(&ppl,ZENRingPol(PR[i]),ZENRingDef(PR[i]));
	  if(pp==NULL)
	    {
	      ZENPolyFree(P,Rg);
	      ZBNF(ZENRingP(R));
	      ZENFree(R);
	      return(ZNULL);
	    }
	  Q=ZENZToPoly(pp,ppl,Rg);
	  ZBNF(pp);
	  if(Q==NULL)
	    {
	      ZENPolyFree(P,Rg);
	      ZBNF(ZENRingP(R));
	      ZENFree(R);
	      return(ZNULL);
	    }
	  ZENPolyMultiply(P,Q,P,Rg);
	  ZENPolyFree(Q,Rg);
	}
      if((ZENRingOrigin(R) =  ZENExtRingCreate(P,Rg))==NULL)
	{
	  ZENPolyFree(P,Rg);
	  ZBNF(ZENRingP(R));
	  ZENFree(R);
	  return(ZNULL);
	}
      ZENPolyFree(P,Rg);
      ZENRingVar(R)=ZENRingVar(ZENRingOrigin(R));
      ZENRingQl(R)=ZENRingQl(ZENRingOrigin(R));
      if (!(ZENRingQ(R) = ZBNC(ZENRingQl(R)))) {
	ZBNF(ZENRingP(R)); ZENFree(R);
	ZENSetError(ZEN_LIB_ZEC, ZEN_INIT, ZEN_ALLOC);
	return(ZNULL);
      }
      ZBNAssign(ZENRingQ(R), ZENRingQ(ZENRingOrigin(R)), ZENRingQl(R));
      ZENRingDef(R)=Rg;
      if(!(ZENRingPol(R)=ZENPolyCopy(ZENRingPol(ZENRingOrigin(R)),ZENRingDef(R)))) {
	ZBNF(ZENRingP(R)); ZBNF(ZENRingQ(R)); ZENFree(R);
	ZENSetError(ZEN_LIB_ZEC, ZEN_INIT, ZEN_ALLOC);
	return(ZNULL);
      }      
    }

  ZENRingSizeElt(R)=ZENRingSizeElt(ZENRingOrigin(R));
  ZENRingSZ(R) = 2*ZENRingSizeElt(R)+1;
  if (!(ZENRingP1(R) = ZBNC(ZENRingSizeElt(R)))) {
    ZENSetError(ZEN_LIB_ZEC, ZEN_INIT, ZEN_ALLOC);
    ZBNF(ZENRingP(R));
    ZBNF(ZENRingQ(R));
    ZENRingClose(ZENRingOrigin(R));
    ZENFree(R);
    return(ZNULL);
  }
  ZBNAssign(ZENRingP1(R), ZENRingP1(ZENRingOrigin(R)), ZENRingSizeElt(R));

  /* We allocate the tables */
  ZecRingNbS(R)=N;
  ZecRingSubrings(R)=(ZENRing *)ZENMalloc(N*sizeof(ZENRing));
  if(ZecRingSubrings(R)==NULL)
    {
      ZENSetError(ZEN_LIB_ZEC, ZEN_INIT, ZEN_ALLOC);
      ZBNF(ZENRingP(R));
      ZBNF(ZENRingP1(R));
      ZBNF(ZENRingQ(R));
      ZENRingClose(ZENRingOrigin(R));
      ZENFree(R);
      return(ZNULL);
    }
  for(i=0;i<N;i++)
    ZecRingSubring(R,i)=PR[i];
  ZecRingModinvs(R)=(ZENElt *)ZENMalloc(N*sizeof(ZENElt));
  if(ZecRingModinvs(R)==NULL)
    {
      ZENFree(ZecRingSubrings(R));
      ZBNF(ZENRingP(R));
      ZBNF(ZENRingP1(R));
      ZBNF(ZENRingQ(R));
      ZENRingClose(ZENRingOrigin(R));
      ZENFree(R);
      return(ZNULL);
    }
  for(i=N;i--;)
    if(ZENEltCreate(ZecRingModinv(R,i),ZENRingOrigin(R)))
      {
	for(i++;i<N;i++)
	  ZENEltFree(ZecRingModinv(R,i),ZENRingOrigin(R));
	ZENFree(ZecRingModinvs(R));
	ZENFree(ZecRingSubrings(R));
	ZBNF(ZENRingP(R));
	ZBNF(ZENRingP1(R));
	ZBNF(ZENRingQ(R));
	ZENRingClose(ZENRingOrigin(R));
	ZENFree(R);
	return(ZNULL);
      }
  
  /* We now complete the construction */
  if(TYPE==BASE)		/* we compute for each ring its
				   complementary product */
    {
      ZENEltSetToOne(ZecRingModinv(R,0),ZENRingOrigin(R));
      for(i=1;i<N;i++)
	ZENZToElt(ZecRingModinv(R,i),
		  ZENRingP(PR[0]),ZENRingPl(PR[0]),
		  ZENRingOrigin(R));
      for(i=1;i<N;i++)
	{
	  ZENZToElt(ZENRingHbuf(ZENRingOrigin(R)),
		    ZENRingP(PR[i]),ZENRingPl(PR[i]),
		    ZENRingOrigin(R));
	  for(j=0;j<N;j++)
	    if(j!=i)
	      ZENEltMultiply(ZecRingModinv(R,j),
			     ZENRingHbuf(ZENRingOrigin(R)),
			     ZecRingModinv(R,j),
			     ZENRingOrigin(R));
	}
    }
  else if(TYPE==EXT)
    {
      BigNum p;
      BigNumLength pl;

      ZENEltSetToOne(ZecRingModinv(R,0),ZENRingOrigin(R));
      for(i=1;i<N;i++)
	ZENPolyAssign(ZENElt2Pol(ZecRingModinv(R,i),ZENRingOrigin(R)),
		      ZENRingPol(PR[0]),
		      Rg);
      for(i=1;i<N;i++)
	{
	  p=ZENPolyToZ(&pl,ZENRingPol(PR[i]),ZENRingDef(PR[i]));
	  if(p==NULL)
	    {
	      for(i=N;i--;)
		ZENEltFree(ZecRingModinv(R,i),ZENRingOrigin(R));
	      ZENFree(ZecRingSubrings(R));
	      ZENFree(ZecRingModinvs(R));
	      ZBNF(ZENRingP(R));
	      ZBNF(ZENRingP1(R));
	      ZBNF(ZENRingQ(R));
	      ZENRingClose(ZENRingOrigin(R));
	      ZENFree(R);
	      return(ZNULL);
	    }
	  Q=ZENZToPoly(p,pl,Rg);
	  ZBNF(p);
	  if(Q==NULL)
	    {	      
	      for(i=N;i--;)
		ZENEltFree(ZecRingModinv(R,i),ZENRingOrigin(R));
	      ZENFree(ZecRingSubrings(R));
	      ZENFree(ZecRingModinvs(R));
	      ZBNF(ZENRingP(R));
	      ZBNF(ZENRingP1(R));
	      ZBNF(ZENRingQ(R));
	      ZENRingClose(ZENRingOrigin(R));
	      ZENFree(R);
	      return(ZNULL);
	    }
	  ZENPolyAssign(ZENElt2Pol(ZENRingHbuf(ZENRingOrigin(R)),ZENRingOrigin(R)),Q,Rg);
	  ZBNF(Q);
	  for(j=0;j<N;j++)
	    if(j!=i)
	      ZENEltMultiply(ZecRingModinv(R,j),
			     ZENRingHbuf(ZENRingOrigin(R)),
			     ZecRingModinv(R,j),
			     ZENRingOrigin(R));
	}
    }
#if CHINESE_VERBOSE
  fprintf(stderr, "Modulo: ");
  ZBNPrintToFile(stderr,ZENRingP(R),ZENRingPl(R),10);
  fprintf(stderr, "\n");
  for(j=0;j<N;j++)
    {
      fprintf(stderr, "PR%d :",j);
      ZEP(ZecRingModinv(R,j),ZENRingOrigin(R));
    }
#endif
  ret=_ZecInit(R);
  if(ret<0)
    {
      for(i=N;i--;)
	ZENEltFree(ZecRingModinv(R,i),ZENRingOrigin(R));
      ZENFree(ZecRingSubrings(R));
      ZENFree(ZecRingModinvs(R));
      ZBNF(ZENRingP(R));
      ZBNF(ZENRingP1(R));
      ZBNF(ZENRingQ(R));
      ZENRingClose(ZENRingOrigin(R));
      ZENFree(R);
      return(ZNULL);
    }
      
  /* Procedure on rings
  **********************/
  ZENPrcInitTracePtr(R) = NULL;
  ZENPrcCloseTracePtr(R) = NULL;
  
  /* Procedure on matrices
  ****************************/
  /* Matrix conversion */
  if(ZENRingArith(ZENRingOrigin(R))==ZEN_LIB_ZEXT)
    {
      ZENVect2MatPtr(R) = _ZENCloneVect2Mat;
      ZENMat2VectPtr(R) = _ZENCloneMat2Vect;
    }
  else
    {
      ZENVect2MatPtr(R) = ZENMatAssignPtr(R);
      ZENMat2VectPtr(R) = ZENMatAssignPtr(R);
    }
  
  if(__ZENRingInitBuffers(R)) {
    for(i=N;i--;)
      ZENEltFree(ZecRingModinv(R,i),ZENRingOrigin(R));
    ZENFree(ZecRingSubrings(R));
    ZENFree(ZecRingModinvs(R));
    ZBNF(ZENRingP(R));
    ZBNF(ZENRingP1(R));
    ZBNF(ZENRingQ(R));
    ZENRingClose(ZENRingOrigin(R));
    ZENFree(R);
    return(ZNULL);
  }
  
  
  if(ret & POLY_GEN)
    __ZENPolyGeneric(R);
  if(ret & MAT_GEN)
    __ZENMatGeneric(R);
  if(ret & SR_GEN)
    __ZENSrGeneric(R);

  ZENRingCln(ZENRingOrigin(R))++;
  for(i=N;i--;)
    ZENRingCln(ZecRingSubring(R,i))++;
#if CHINESE_VERBOSE
  ZENEltSetRandom(ZENRingHbuf(R),R);
  ZEP(ZENRingHbuf(R),R);
#endif
  return(R);
#else  /* ZEC_AVAILABLE */
  ZENSetError(ZEN_LIB_ZER, ZEN_INIT, ZEN_UNAVAILABLE);
  return(ZNULL);
#endif
}

/**********************************************************************/
int __ZENChineseClose(R)
     ZENRing R;
{
  int i;
  ZENRing Rg;  

  Rg=ZENRingOrigin(R);
  for(i=ZecRingNbS(R);i--;)
    ZENEltFree(ZecRingModinv(R,i),Rg);
  ZENFree(ZecRingModinvs(R));
  if(_ZecClose(R))
    return(ZERR);
  for(i=ZecRingNbS(R);i--;)
    ZENRingCln(ZecRingSubring(R,i))--;
  ZENFree(ZecRingSubrings(R));

  return(ZEN_HAS_INVERSE);
}



