/*
 * zen/zer/c/clone_ring.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   \Section{Clone initialization}
   }
   */
#include <stdio.h>
#include "zer.h"
#include "zed.h"
#include "zeg.h"
#include "zec.h"

/********************************************************/
/* \TEX{   
   \Section{Ring cloning}
   
   There unfortunately exists several way to consider the same finite
   ring. For instance, if this finite ring is a finite field, it is
   well known that its multiplicative group is cyclic and so, an
   element can be seen as the power of a generator |g|. In other
   words, when the finite field is small, it can be sometimes usefull
   to work with logarithms of elements instead of their polynomial
   representation. 
   
   The function |ZENRingClone| enable the programmer to obtain a
   cloned ring from the usual polynomial reprsentation.
   The structure of the clone ring is specified by the arguments.   

   }
   */

/********************************************************/
/* Used by Zem, Zelog, and Zetab */

static ZENRing __ZENRingClone(R)
     ZENRing R;
{ZENRing Rg;
 
 if((Rg=__ZENRingCreate(ZENRingP(R),ZENRingPl(R)))==ZENNULL)
   return(ZNULL);
 
 ZENRingPol(Rg) = ZENRingPol(R);
 ZENRingDef(Rg) = ZENRingDef(R);
 ZENRingOrigin(Rg) =  R;

 ZENRingVar(Rg)=ZENRingVar(R);

 /* The number of elements */
 ZENRingQl(Rg) = ZENRingQl(R);
 if (!(ZENRingQ(Rg) = ZBNC(ZENRingQl(Rg)))) {
   ZBNF(ZENRingP(Rg));
   ZENFree(Rg);
   ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
   return(ZNULL);
 }
 ZBNAssign(ZENRingQ(Rg), ZENRingQ(R), ZENRingQl(Rg));

 ZENRingSizeElt(Rg) = ZENRingSizeElt(R);
 ZENRingSZ(Rg) = 2*ZENRingSizeElt(R)+1;
 if (!(ZENRingP1(Rg) = ZBNC(ZENRingSizeElt(Rg)))) {
   ZBNF(ZENRingP(Rg));
   ZBNF(ZENRingQ(Rg));
   ZENFree(Rg);
   ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
   return(ZNULL);
 }
 ZBNAssign(ZENRingP1(Rg), ZENRingP1(R), ZENRingSizeElt(Rg));

 /************ new pointers of functions are used ***************/

 /* Procedure on elements
  ****************************/
 /* Assigning */
 ZENEltConvertPtr(Rg) = _ZENEltCloneConvert;
 /* Test */
 if ((ZENRingPl(Rg) == 1) && (ZBNDigit(ZENRingP(Rg), 0) == 2)) {
   ZENEltIsASquarePtr(Rg) = _ZENEltIsASquareChar2;
 }
 else {
   ZENEltIsASquarePtr(Rg) = _ZENEltIsASquareCharp;
 }
 
 /* Arithmetic */
 if ((ZENRingPl(Rg) == 1) && (ZBNDigit(ZENRingP(Rg), 0) == 2)) {
   ZENEltSquareRootPtr(Rg) = _ZENEltSquareRootChar2;
 }
 else {
   ZENEltSquareRootPtr(Rg) = _ZENEltSquareRootCharp;
 }

 ZENEltReadFromStringPtr(Rg) = _ZENCloneEltReadFromString;
 ZENEltPrintToStringPtr(Rg) = _ZENCloneEltPrintToString;
 ZENEltReadFromFilePtr(Rg) = _ZENCloneEltReadFromFile;
 ZENEltPrintToFilePtr(Rg) = _ZENCloneEltPrintToFile;
 
 ZENEltGetFromStringPtr(Rg) = _ZENCloneEltGetFromString;
 ZENEltPutToStringPtr(Rg) = _ZENCloneEltPutToString;
 ZENEltGetFromFilePtr(Rg) = _ZENCloneEltGetFromFile;
 ZENEltPutToFilePtr(Rg) = _ZENCloneEltPutToFile;
 

 /* Procedures on polynomials
  *******************************/
 /* Arithmetic */
 ZENPolyDerivePtr(Rg) = _ZENPolyDerive;
 if ((ZENRingPl(Rg) == 1) && (ZBNDigit(ZENRingP(Rg), 0) == 2)) {
   ZENPolyRootsDegree2Ptr(Rg) = _ZENPolyRootsDegree2Char2;
 }
 else {
   ZENPolyRootsDegree2Ptr(Rg) = _ZENPolyRootsDegree2Charp;
 }
 
 /* Updating the origin ring */
 ZENRingCln(ZENRingOrigin(Rg))++;
 
 return(Rg);
}
/********************************************************/


/* \TEX{
   \title{Clone of a ring.}
   \input{A |ZENRing| |Rg| and the type of the clone |cln|.} 
   \output{A ring |R| if everything was succesfully initialized, |ZENNULL| otherwise.}
   \note{Cloning can take time.}   
   }
   */
ZENRing _ZENRingClone(Rg, cln)
     ZENRing Rg;
     ZENCln cln;
{int ret;
 ZENRing R;
 
 switch(ZENRingArith(Rg)) {
   
   /* These rings are already clones */
 case ZEN_LIB_ZEM:
 case ZEN_LIB_ZETAB:
 case ZEN_LIB_ZELOG:
   return(ZENRingClone(ZENRingOrigin(Rg), cln));
   
   /* The other types */
 default:
   /* Tabulation */
   ret=ZENWGINPUT;
#if ZETAB_AVAILABLE
   if ((ret<0)&&(ZENClnIsSet(cln, ZENCLN_TABULATE)))
     {
       if ((ZENRingQl(Rg) == 1)
	   && (ZBNDigit(ZENRingQ(Rg), 0) != 2)
	   && (ZBNDigit(ZENRingQ(Rg), 0) != 0))
	 {
	   if (ZBNDigit(ZENRingQ(Rg), 0) <= (1 << SIZE_CHAR))
	     {
	       if((R=__ZENRingClone(Rg)) == ZENNULL) 
		 return(ZNULL);
	       ret=_ZetabInit(R);
	       if(ret<0)
		 {
		   ZBNF(ZENRingP(R));
		   ZBNF(ZENRingQ(R));
		   ZENFree(R);
		   if(ret==ZENERR)
		     return(ZNULL);
		 }
#if 1
	       else
		 fprintf(stderr,"Tabulation has succeeded...\n");
#endif
	     }
	 }
     }
#endif
#if ZELOG_AVAILABLE
   if ((ret<0)&&(ZENClnIsSet(cln, ZENCLN_LOG)))
     {
       if ((ZENRingQl(Rg) == 1)
	   && (ZBNDigit(ZENRingQ(Rg), 0) != 2)
	   && (ZBNDigit(ZENRingQ(Rg), 0) != 0))
	 {
	   if (ZBNDigit(ZENRingQ(Rg), 0) <= (1 << 2*SIZE_CHAR))
	     {
	       if((R=__ZENRingClone(Rg)) == ZENNULL) 
		 return(ZNULL);
	       ret=_ZelogInit(R);
	       if(ret <= 0)
		 {
		   ZBNF(ZENRingP(R));
		   ZBNF(ZENRingQ(R));
		   ZENFree(R);
		   if(ret==ZENERR)
		     return(ZNULL);
		 }
#if 1
	       else
		 fprintf(stderr,"Generator tabulation has succeeded...\n");
#endif
	     }
	 }
     }
#endif
#if ZEM_AVAILABLE
   if ((ret<0)&&(ZENClnIsSet(cln, ZENCLN_MONTGOMERY)))
     {
       if ((ZENRingArith(Rg)==ZEN_LIB_ZEP)
	   && (ZBNDigit(ZENRingP(Rg), 0) & 1))
	 {
	   if((R=__ZENRingClone(Rg)) == ZENNULL) 
	     return(ZNULL);
	   ret=_ZemInit(R);
	   if(ret<0)
	     {
	       ZBNF(ZENRingP(R));
	       ZBNF(ZENRingQ(R));
	       ZENFree(R);
	       if(ret==ZENERR)
		 return(ZNULL);
	     }
#if 1
	   else
	     fprintf(stderr,"Montgomery's reduction available...\n");
#endif
	 }
     }
#endif

   if(ret>=0)
     {
       if(ret & POLY_GEN)
	 __ZENPolyGeneric(R);
       
       if(ret & MAT_GEN)
	 __ZENMatGeneric(R);

       if(ret & SR_GEN)
	 __ZENSrGeneric(R);
       
       if (__ZENRingInitBuffers(R)) {
	 ZBNF(ZENRingP(R));
	 ZBNF(ZENRingQ(R));
	 ZENFree(R);
	 return(ZNULL);
       }
       return(R);
     }
 }
 
 /* By default, a copy */
 if ((R = ZENRingCopy(Rg))) {
   ZENRingOrigin(R) =  Rg;
   ZENRingCln(ZENRingOrigin(R))++;
 }
 return(R);
}


/* \TEX{
   \title{Copying a ring.}
   \input{A finite field |R|.}
   \output{|ZENNULL| if an error occurred, a copy of the finite Ring |R|
   without its precomputation otherwise.}
   }
   */
ZENRing _ZENRingCopy(R)		
     ZENRing R;
{
  ZENCln clnflags;

  switch(ZENRingArith(R)) {

    /* These rings are clone */
  case ZEN_LIB_ZEC:
    return(ZENChineseRingCreate(ZecRingNbS(R),ZecRingSubrings(R)));
    
  case ZEN_LIB_ZETAB:
  case ZEN_LIB_ZELOG:
  case ZEN_LIB_ZEM:
    ZENClnSetNone(clnflags);
    ZENClnSet(clnflags, ZENCLN_TABULATE);
    return(ZENRingClone(ZENRingOrigin(R), clnflags));

    /* These rings are prime rings */
  case ZEN_LIB_ZE2:
  case ZEN_LIB_ZEF:
  case ZEN_LIB_ZEP:
  case ZEN_LIB_ZEPS:
    return(ZENBaseRingCreate(ZENRingP(R), ZENRingPl(R)));

    /* These rings are extensions */
  case ZEN_LIB_ZEXT:
    return(ZENExtRingCreate(ZENRingPol(R), ZENRingDef(R)));
  default:
    ;
  }
  
  ZENSetError(ZEN_LIB_UNKNOWN, ZEN_UNKNOWN, ZEN_UNKNOWN);
  return(ZNULL);
}

/*************************************************************/
int __ZENCloneClose(R)
     ZENRing R;
{
  switch(ZENRingArith(R)) {
  case ZEN_LIB_ZETAB:
    if(_ZetabClose(R))
      return(ZERR);
    break;
  case ZEN_LIB_ZELOG:
    if(_ZelogClose(R))
      return(ZERR);
    break;
  case ZEN_LIB_ZEM:
    if(_ZemClose(R))
      return(ZERR);
    break;
  default:
    return(ZERR);
  }
  return(0);
}
