/*
 * zen/zer/c/close.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   \Section{Ring closing}
   }
   */
#include <stdio.h>
#include <zen.h>
#include <zer.h>
#include <zec.h>


/********************************************************/
static void __ZENRingFreeBuffers(R)
     ZENRing R;
{
  ZENEltFree(ZENRingHbuf(R), R);
  ZENEltFree(ZENRingIbuf(R), R);
  ZENEltFree(ZENRingJbuf(R), R);
  ZENEltFree(ZENRingFact(R), R);
  ZBNF(ZENRingBNbuf(R));
}
/*************************************************************/

/* \TEX[
   \title{Closing a ring.}
   \input{A finite field |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|R| is completely freed.}
   }
   */
int _ZENRingClose(R)		
     ZENRing R;
{
  /* Can't close an extension if an extension is defined over it */
  if (ZENRingExt(R)) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_CLOSE, ZEN_RGCLOSE);
    return(ZERR);
  }
 
  /* Can't close a finite ring if a clone is defined over it */
  if (ZENRingCln(R)) {
    ZENSetError(ZEN_LIB_ZEG, ZEN_CLOSE, ZEN_RGCLOSE);
    return(ZERR);
  }

  /* Updating the origin ring */
  if (ZENRingOrigin(R)) {
    ZENRingCln(ZENRingOrigin(R))--;
  }

  /* Disallocate precomputations */
  ZENRingRmPrc(R, ZENRingPrcp(R));
 
  /* Disallocate Buffers */
  __ZENRingFreeBuffers(R);

  /* Specific closing */
  switch(ZENRingArith(R)) {
  case ZEN_LIB_ZE2:
  case ZEN_LIB_ZEP:
  case ZEN_LIB_ZEF:
  case ZEN_LIB_ZEPS:
    if(__ZENBaseClose(R))
      return(ZERR);
    break;
  case ZEN_LIB_ZEXT:
    if(__ZENExtClose(R))
      return(ZERR);
    break;
  case ZEN_LIB_ZETAB:
  case ZEN_LIB_ZELOG:
  case ZEN_LIB_ZEM:
    if(__ZENCloneClose(R))
      return(ZERR);
    break;
  case ZEN_LIB_ZEC:
    if(__ZENChineseClose(R))
      return(ZERR);
    break;
  default:
    ZENSetError(ZEN_LIB_UNKNOWN, ZEN_UNKNOWN, ZEN_UNKNOWN);
    return(ZERR);
  }
  
  /* The characteristic */
  ZBNF(ZENRingP(R));

  /* The number of elements */
  ZBNF(ZENRingQ(R));
  ZBNF(ZENRingP1(R));

  ZENFree(R);
  return(0);
}

/* \TEX{
   \title{Closing a clone ring.}
   \input{A ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|R| is freed as well as the underlying |ZENRingOrigin(R)|, if it exists. If |R| is a chinese ring, the list of rings is clone freed, but the under-underlying |ZENRingOrigin(R)| is not freed.}
   }
   */
int _ZENRingCloneClose(R)		
     ZENRing R;
{
  switch(ZENRingArith(R))
    {
    case ZEN_LIB_ZEC:
      {
	int i,N;
	ZENRing *PR;

	N=ZecRingNbS(R);
	PR=(ZENRing *)ZENMalloc(N*sizeof(ZENRing));
	if(PR==NULL)
	  {
	    ZENSetError(ZEN_LIB_ZEC, ZEN_CLOSE, ZEN_ALLOC);
	    return(ZERR);
	  }
	for(i=0;i<N;i++)
	  PR[i]=ZecRingSubring(R,i);
	if(ZENRingClose(R))
	  return(ZERR);
	for(;i--;)
	  {
	    if(ZENRingCloneClose(PR[i]))
	      return(ZERR);
	  }
	ZENFree(PR);
	return(0);
      }
    default:
      {
	ZENRing Rg;

	Rg=ZENRingOrigin(R);
	if(ZENRingClose(R))
	  return(ZERR);
	if(Rg)
	  return(ZENRingClose(Rg));
	return(0);
      }
    }
}
/* \TEX{
   \title{Closing a ring.}
   \input{A finite field |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|R| is completely freed as well as the underlying |ZENRing|s.}
   }
   */
int _ZENRingFullClose(R)		
     ZENRing R;
{
  ZENRing Rg;
  
  switch(ZENRingArith(R))
    {
    case ZEN_LIB_ZEC:
      {
	int i,N;
	ZENRing *PR;

	N=ZecRingNbS(R);
	PR=(ZENRing *)ZENMalloc(N*sizeof(ZENRing));
	if(PR==NULL)
	  {
	    ZENSetError(ZEN_LIB_ZEC, ZEN_CLOSE, ZEN_ALLOC);
	    return(ZERR);
	  }
	for(i=0;i<N;i++)
	  PR[i]=ZecRingSubring(R,i);
	if(ZENRingClose(R))
	  return(ZERR);
	for(i--;i>=1;i--)
	  {
	    if(ZENRingCloneClose(PR[i]))
	      return(ZERR);
	  }
	if(ZENRingFullClose(PR[0]))
	  return(ZERR);
	ZENFree(PR);
	return(0);
      }
    case ZEN_LIB_ZEXT:
      Rg=ZENRingOrigin(R);
      if(!Rg)
	Rg=ZENRingDef(R);
      if(ZENRingClose(R))
	return(ZERR);
      return(ZENRingFullClose(Rg));
      

    default:
      Rg=ZENRingOrigin(R);
      if(ZENRingClose(R))
	return(ZERR);
      if(Rg)	  
	return(ZENRingClose(Rg));
      return(0);      
    }
}
