/*
 * zen/zer/c/ext_ring.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */


/* \TEX{   
   \Section{Extension initialization}
   }
   */
#include "stdio.h"
#include "zer.h"
#include "zext.h"
#include "zeg.h"
/********************************************************/
/* \TEX{
   \title{Initialization of an extension over another ring.}
   \input{A polynomial |P| defined over a ring |Rg|.} 
   \output{A ring if everything was succesfully initialized, |ZENNULL| otherwise.}
   \note{Precomputations are done in in order to speed up
   some arithmetic operations.}	   
   }
   */
ZENRing _ZENExtRingCreate(P, Rg)
     ZENPoly P;
     ZENRing Rg;
{ZENRing R;
 BigNum n;
 int ret, nl, i;
 int var;

#if !ZEXT_AVAILABLE
 ZENSetError(ZEN_LIB_ZEXT, ZEN_INIT, ZEN_UNAVAILABLE);
 return(ZNULL);
#endif
 
 /* Allocation */
 if ((R = __ZENRingCreate(ZENRingP(Rg),ZENRingPl(Rg))) == ZENNULL)
   return(ZNULL);
 
 var=ZENRingVar(Rg);
 switch(var)			/* t u v w x y z a..s A..Z */
				/* better than x y z ... because x y z */
				/* are used with elliptic curves */
   {
   case 0:
     var='t';
     break;
    case 'z':
     var='a';
     break;
   case 's':
     var='A';
     break;
   case 'W':			/* we skip "X" (polynomials) */
     var='Y';
     break;
   case 'Z':
     var='t';
     break;
   default:
     var++;
     break;
   }
 ZENRingVar(R) = var;
 
 /* The number of elements */
 ZENRingQl(R) = ZENRingQl(Rg);
 if (!(ZENRingQ(R) = ZBNC(ZENRingQl(R)))) {
   ZBNF(ZENRingP(R)); ZENFree(R);
   ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
   return(ZNULL);
 }
 ZBNAssign(ZENRingQ(R), ZENRingQ(Rg), ZENRingQl(R));

 for(i = 1; i < ZENPolyDeg(P, Rg); i++) {
   nl = ZENRingQl(R)+ZENRingQl(Rg)+1;
   if (!(n = ZBNC(nl))) {
     ZBNF(ZENRingQ(R)); ZBNF(ZENRingP(R)); ZENFree(R);
     ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
     return(ZNULL);
   }
   ZBNSetToZero(n, nl);
   ZBNMultiply(n, nl, ZENRingQ(R), ZENRingQl(R), ZENRingQ(Rg), ZENRingQl(Rg));
   ZBNF(ZENRingQ(R));
   ZENRingQ(R) = n;
   ZENRingQl(R) = ZBNNumDigits(n, nl);
 }

 nl= ZENRingQl(R);
 if (!(n = ZBNC(nl))) {
   ZBNF(ZENRingQ(R)); ZBNF(ZENRingP(R)); ZENFree(R);
   ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
   return(ZNULL);
 }
 ZBNAssign(n,ZENRingQ(R),nl);
 ZBNSubtractBorrow(n, nl, 0);
 nl = ZBNNumDigits(n, nl);
 ZENRingSizeElt(R) = nl;
 ZENRingSZ(R) = 2*nl+1;
 if (!(ZENRingP1(R) = ZBNC(ZENRingSizeElt(R)))) {
   ZBNF(n); ZBNF(ZENRingQ(R)); ZBNF(ZENRingP(R)); ZENFree(R);
   ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
   return(ZNULL);
 }
 ZBNAssign(ZENRingP1(R), n, nl);
 ZBNF(n);
 
 /* The number of extensions defined over these extensions */
 ZENRingDef(R) = Rg;
 ZENRingOrigin(R) =  NULL;
 
 /* A monic definition polynomial */
 if (!(ZENRingPol(R) = ZENPolyCopy(P, Rg))) {
   ZBNF(ZENRingQ(R));
   ZBNF(ZENRingP(R));
   ZENFree(R);
   ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
   return(ZNULL);
 }
 switch(ZENPolyMakeMonic(ZENRingPol(R), ZENRingPol(R), Rg)) {
 case ZEN_NO_INVERSE:
   ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_NONMONIC);
 case -1:
   ZBNF(ZENRingP(R));
   ZBNF(ZENRingQ(R));
   ZENPolyFree(ZENRingPol(R), Rg);
   ZENFree(R);
   return(ZNULL);
 }

 /* Two BigNum Buffers for EltToZ and ZToElt */
 if (!(ZENRingTmp1(R) = ZBNC(ZENRingQl(R)+1))) {
   ZENPolyFree(ZENRingPol(R), Rg);
   ZBNF(ZENRingQ(R));
   ZBNF(ZENRingP(R));
   ZENFree(R);
   return(ZNULL);
 }
 
 /* Two BigNum Buffers for EltToZ and ZToElt */
 if (!(ZENRingTmp2(R) = ZBNC(ZENRingQl(R)+1))) {
   ZBNF(ZENRingTmp1(R));
   ZENPolyFree(ZENRingPol(R), Rg);
   ZBNF(ZENRingQ(R));
   ZBNF(ZENRingP(R));
   ZENFree(R);
   return(ZNULL);
 }
 
 /* Polynomial buffers */
 if (ZENRingArith(Rg) == ZEN_LIB_ZE2) {
   
   if (!(ZENRingPbuf(R) = ZENPolyCreate(2*(ZENRingDeg(R)+SIZE_BLOC), Rg))) {
     ZBNF(ZENRingTmp2(R));
     ZBNF(ZENRingTmp1(R));
     ZENPolyFree(ZENRingPol(R), Rg);
     ZBNF(ZENRingQ(R));
     ZBNF(ZENRingP(R));
     ZENFree(R);
     return(ZNULL);
   }
 }
 else {
   if (!(ZENRingPbuf(R) = ZENPolyCreateLarge(4*ZENRingDeg(R)+1, Rg))) {
     ZBNF(ZENRingTmp2(R));
     ZBNF(ZENRingTmp1(R));
     ZENPolyFree(ZENRingPol(R), Rg);
     ZBNF(ZENRingQ(R));
     ZBNF(ZENRingP(R));
     ZENFree(R);
     return(ZNULL);
   }
 }
 

  /* The sub degree */
 for(ZENRingSubDeg(R) = ZENRingDeg(R)-1;
     ZENRingSubDeg(R) > 0;
     ZENRingSubDeg(R)--) {
   if (!ZENEltIsZero(ZENPolyGetCoeffPtr(ZENRingPol(R),
					ZENRingSubDeg(R),
					Rg), Rg)) {
     break;
   }
 }

 /* We now complete the construction */

 ret=_ZextInit(R);

 
 if(ret & POLY_GEN)
   __ZENPolyGeneric(R);
 if(ret & MAT_GEN)
   __ZENMatGeneric(R);
 if(ret & SR_GEN)
   __ZENSrGeneric(R);
 
 /* Procedure on elements
  ****************************/
 /* Assignation */
 ZENEltToClonePtr(R) = ZENEltAssignPtr(R);
 ZENCloneToEltPtr(R) = ZENEltAssignPtr(R);

 /* Procedure on matrices
  ****************************/
  /* Matrix conversion */
 switch(ZENRingArith(ZENRingDef(R)))
   {
   case ZEN_LIB_ZE2:
     ZENVect2MatPtr(R) = _Ze2mVect2Mat;
     ZENMat2VectPtr(R) = _Ze2mMat2Vect;
     break;
   case ZEN_LIB_ZEPS:
     ZENVect2MatPtr(R) = _ZepmVect2Mat;
     ZENMat2VectPtr(R) = _ZepmMat2Vect;
     break;
   case ZEN_LIB_ZEF:
   case ZEN_LIB_ZEXT:
   case ZEN_LIB_ZEP:
     ZENVect2MatPtr(R) = _ZextVect2Mat;
     ZENMat2VectPtr(R) = _ZextMat2Vect;
     break;
   default:
     ZENVect2MatPtr(R) = _ZENCloneVect2Mat;
     ZENMat2VectPtr(R) = _ZENCloneMat2Vect;
     break;
   }

 if (ZENEltCreate(ZextRingRbuf(R), R)) {
   ZENPolyFree(ZENRingPbuf(R), ZENRingDef(R));
   ZBNF(ZENRingTmp2(R));
   ZBNF(ZENRingTmp1(R));
   ZENPolyFree(ZENRingPol(R), Rg);
   ZBNF(ZENRingQ(R));
   ZBNF(ZENRingP(R));
   ZENFree(R);
   return(ZNULL); 
 }
 
 if (__ZENRingInitBuffers(R)) {
   ZBNF(ZENRingTmp2(R));
   ZBNF(ZENRingTmp1(R));
   ZENPolyFree(ZENRingPbuf(R), ZENRingDef(R));
   ZBNF(ZENRingP(R));
   ZBNF(ZENRingQ(R));
   ZENPolyFree(ZENRingPol(R), Rg);
   ZENFree(R);
   ZENEltFree(ZextRingRbuf(R), R);
   return(ZNULL);
 }

 ZENRingExt(Rg)++;

 return(R);
}
/*************************************************************/
int __ZENExtClose(R)
     ZENRing R;
{
  ZBNF(ZENRingTmp2(R));
  ZBNF(ZENRingTmp1(R));
  ZENPolyFree(ZENRingPbuf(R), ZENRingDef(R));
  ZENEltFree(ZextRingRbuf(R), R);

  /* The definition polynomial */
  ZENPolyFree(ZENRingPol(R), ZENRingDef(R));

  if (ZENRingDef(R)) 
    ZENRingExt(ZENRingDef(R))--;
  else
    return(ZERR);

  switch(ZENRingArith(R)) {
  case ZEN_LIB_ZEXT:
    if(_ZextClose(R))
      return(ZERR);
    break;
  default:
    return(ZERR);
  }
  
  return(0);
}

