/*
 * zen/zer/c/io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

 /* \TEX{
   We provide two input/output functions on |ZENRing|s.
   These function allow to save the fundamental structures of a |ZENRing| on a file by |ZENRingPrintToFile|.
   Precomputations or cloning procedures are not saved upon file and are performed again if necessary when the call to |ZENRingReadFromFile| occurs.
   }
   */
#include <stdio.h>
#include "zec.h"

/* \TEX{
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |ZENRing| |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |Rg| to |file|.}
   \note{In an extension tower, one only needs to save the last ring as the structure is recursively saved on file.}
   }
   */
int _ZENRingPrintToFile(file, Rg)
     FILE *file;
     ZENRing Rg;  
{
  if(fprintf(file,"%d\n",(int)ZENRingArith(Rg))<0)
    {
      ZENSetError(ZEN_LIB_ZER, ZEN_PRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
  switch(ZENRingArith(Rg))
    {
    case ZEN_LIB_ZETAB:
    case ZEN_LIB_ZELOG:
    case ZEN_LIB_ZEM:
      if(ZENRingPrintToFile(file,ZENRingOrigin(Rg)))
	return(ZERR);
      break;
    case ZEN_LIB_ZEC:
      {
	int i;

	fprintf(file,"%d\n",ZecRingNbS(Rg));
	for(i=0;i<ZecRingNbS(Rg);i++)
	  if(ZENRingPrintToFile(file,ZecRingSubring(Rg,i)))
	    return(ZERR);
      }
    break;
    case ZEN_LIB_ZEXT:
      if(ZENRingPrintToFile(file,ZENRingDef(Rg)))
	return(ZERR);
      if(ZENPolyPrintToFile(file,ZENRingPol(Rg),10,ZENRingDef(Rg)))
	return(ZERR);
      fprintf(file,"\n");
      break;

    default:
      if(ZBNPrintToFile(file,ZENRingP(Rg),ZENRingPl(Rg),10))
	return(ZERR);
      fprintf(file,"\n");
      break;
    }
  if(fprintf(file,"%d\n",(int)ZENRingPrcp(Rg))<0)
    {
      ZENSetError(ZEN_LIB_ZER, ZEN_PRINTTOFILE, ZEN_FPRINTF);
      return(ZERR);
    }
  return(0);
}

/* \TEX{
   \title{Reading from file.}
   \input{A stream |file| and a pointer on a ring |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|Rg| is created according to the datas in |file|.}
   \note{|Rg| can be freed using |ZENRingFullClose|}
   }
   */
int _ZENRingReadFromFile(Rg, file)
     FILE *file;
     ZENRing *Rg;
{
  arith arithmetic;
  ZENPrc prc;

  if( fscanf(file,"%d",&prc)!= 1)
    {
      ZENSetError(ZEN_LIB_ZER, ZEN_READFROMFILE, ZEN_FSCANF);
      return(ZERR);
    }
  arithmetic = (arith)prc;
  switch(arithmetic)
    {
    case ZEN_LIB_ZEC:
      {
	int i,nr;
	ZENRing *Pr;
	
	if( fscanf(file,"%d",&nr)!= 1)
	  {
	    ZENSetError(ZEN_LIB_ZER, ZEN_READFROMFILE, ZEN_FSCANF);
	    return(ZERR);
	  }

	if (!(Pr=(ZENRing *)ZENMalloc(nr*sizeof(ZENRing)))) {
	    ZENSetError(ZEN_LIB_ZER, ZEN_READFROMFILE, ZEN_ALLOC);
	    return(ZERR);
	}
	
	for(i=0;i<nr;i++)
	  if(ZENRingReadFromFile(Pr[i],file))
	    return(ZERR);
	*Rg=ZENChineseRingCreate(nr,Pr);
      }
    break;
    case ZEN_LIB_ZETAB:
    case ZEN_LIB_ZELOG:
    case ZEN_LIB_ZEM:
      {
	ZENRing R;
	ZENCln cln;
	
	if(ZENRingReadFromFile(R,file))
	  return(ZERR);
	ZENClnSetNone(cln);
	switch(arithmetic)
	  {
	  case ZEN_LIB_ZETAB:
	    ZENClnSet(cln,ZENCLN_TABULATE);
	    break;
	  case ZEN_LIB_ZELOG:
	    ZENClnSet(cln,ZENCLN_LOG);
	    break;
	  case ZEN_LIB_ZEM:
	    ZENClnSet(cln,ZENCLN_MONTGOMERY);
	    break;
	  default:
	    ;
	  }
	(*Rg)=ZENRingClone(R,cln);
      }
      break;
      
    case ZEN_LIB_ZEXT:
      {
	ZENRing R;
	ZENPoly P;
	
	if(ZENRingReadFromFile(R,file))
	  return(ZERR);
	if(ZENPolyReadFromFile(P,file,10,R))
	  return(ZERR);
	*Rg=ZENExtRingCreate(P,R);
      }
      break;

    default:
      {
	BigNum n;
	BigNumLength nl;
	
	if(ZBNReadFromFile(&n,&nl,file,10))
	  return(ZERR);
	*Rg=ZENBaseRingCreate(n,nl);
	ZBNF(n);
      }
      break;
    }
  if(!(*Rg))
    return(ZERR);
  if(fscanf(file,"%d",&prc)!= 1)
    {
      ZENSetError(ZEN_LIB_ZER, ZEN_READFROMFILE, ZEN_FSCANF);
      return(ZERR);
    }
  if(ZENRingAddPrc(*Rg,prc))
    return(ZERR);
  return(0);
}
