/*
 * zen/zer/c/ring.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{   
   A finite ring is completely described by a variable of type
   |ZENRing|. At first we can create prime rings with
   |_ZENBaseRingCreate|. This ring is the set of the integers modulo
   another. Then we can create an extension over another ring with
   |_ZENExtRingCreate|. This ring is then the set of polynomials
   modulo another. We can free a ring if no extensions was defined over
   it with |_ZENRingClose|.
   }
   */
#include <stdio.h>
#include <zen.h>
#include <zed.h>
#include <zeg.h>
#include <zer.h>
#include <zef.h>
#include <zext.h>

/********************************************************/
/*
   \TEX{
   \title{Generic polynomial functions}
   \input{A |ZENRing|.}
   \side{The polynomial function pointers of |Rg| are set.}
   \note{These functions are not always optimal.}
   } 
   */

void __ZENPolyGeneric(Rg)
     ZENRing Rg;
{
  /* Allocation */
  ZENPolyCreatePtr(Rg) = _ZedPolyCreate;
  ZENPolyCreateLargePtr(Rg) = _ZedPolyCreate;
  ZENPolyFreePtr(Rg) = _ZedPolyFree;
  
  /* Assigning */
  ZENPolyAssignPtr(Rg) = _ZENPolyAssign;
  ZENPolyAssignLargePtr(Rg) = _ZENPolyAssign;
  ZENPolyLazyToVeryLazyPtr(Rg) = _ZENPolyAssign;
  ZENPolySetToZeroPtr(Rg) = _ZENPolySetToZero;
  ZENPolySetToZeroLargePtr(Rg) = _ZENPolySetToZero;
  ZENPolySetToXiPtr(Rg) = _ZENPolySetToXi;
  ZENPolySetRandomPtr(Rg) = _ZENPolySetRandom;
  ZENPolySetCoeffPtr(Rg) = _ZedPolySetCoeff;
  ZENPolyGetCoeffNewPtr(Rg) = _ZENPolyGetCoeff;
  ZENPolyGetCoeffPtrPtr(Rg) = _ZedPolyGetCoeffPtr;
  ZENPolyUpdateDegreePtr(Rg) = _ZENPolyUpdateDegree;
  
  /* Test */
  ZENPolyAreEqualPtr(Rg) = _ZENPolyAreEqual;
  ZENPolyIsZeroPtr(Rg) = _ZENPolyIsZero;
  ZENPolyIsXiPtr(Rg) = _ZENPolyIsXi;
  
  /* Arithmetic */
  ZENPolyAddPtr(Rg) = _ZENPolyAdd;
  ZENPolyAddLargePtr(Rg) = _ZENPolyAdd;
  ZENPolyNegatePtr(Rg) = _ZENPolyNegate;
  ZENPolyNegateLargePtr(Rg) = _ZENPolyNegate;
  ZENPolySubtractPtr(Rg) = _ZENPolySubtract;
  ZENPolySubtractLargePtr(Rg) = _ZENPolySubtract;

  if ((ZENRingPl(Rg) == 1) && (ZBNDigit(ZENRingP(Rg), 0) == 2)) {
    ZENPolySquarePtr(Rg) = _ZENPolySquareChar2;
    ZENPolySquareLazyPtr(Rg) = _ZENPolySquareChar2;
    ZENPolySquareVeryLazyPtr(Rg) = _ZENPolySquareChar2;
  }
  else {
    ZENPolySquarePtr(Rg) = _ZENPolySquareCharp;
    ZENPolySquareLazyPtr(Rg) = _ZENPolySquareCharp;
    ZENPolySquareVeryLazyPtr(Rg) = _ZENPolySquareCharp;
  }
  
  ZENPolyMultiplyPtr(Rg) = _ZENPolyMultiply;
  ZENPolyMultiplyLazyPtr(Rg) = _ZENPolyMultiply;
  ZENPolyMultiplyVeryLazyPtr(Rg) = _ZENPolyMultiply;
  
  ZENPolyMultiplyScalarPtr(Rg) = _ZENPolyMultiplyScalar;
  ZENPolyDotPtr(Rg) = _ZENPolyDot;
  ZENPolyEvalPtr(Rg) = _ZENPolyEval; /* Evaluation of a polynomial */
  ZENPolyDividePtr(Rg) = _ZENPolyDivide;
  ZENPolyDivideLazyPtr(Rg) = _ZENPolyDivideLazy;
  ZENPolyDivideVeryLazyPtr(Rg) = _ZENPolyDivideLazy;
  ZENPolyModuloLazyPtr(Rg) = _ZENPolyModuloLazy;
  ZENPolyModuloVeryLazyPtr(Rg) = _ZENPolyModuloLazy;
  ZENPolyDerivePtr(Rg) = _ZENPolyDerive;

  if (ZBNDigit(ZENRingP(Rg), 0) == 2) {
    ZENPolyRootsDegree2Ptr(Rg) = _ZENPolyRootsDegree2Char2;
  }
  else {
    ZENPolyRootsDegree2Ptr(Rg) = _ZENPolyRootsDegree2Charp;
  }
}

/********************************************************/

/*
   \TEX{
   \title{Generic matricial functions}
   \input{A |ZENRing|.}
   \side{The matricial function pointers of |Rg| are set.}
   \note{These functions are not always optimal.}
   } 
   */

void __ZENMatGeneric(R)
     ZENRing R;
{
  /* Procedures on matrices
   **************************/
  /* Assigning */
  ZENMatAssignPtr(R) = _ZENMatAssign;	/* Assign a matrix with another */
  ZENPermutation2MatrixPtr(R) = _ZENPermutation2Matrix;	/* Assign a matrix with another */
  ZENMatSetToZeroPtr(R) = _ZENMatSetToZero; /* Set a matrix to zero */
  ZENMatSetToOnePtr(R) = _ZENMatSetToOne; /* Set a matrix to identity */
  ZENMatSetRandomPtr(R) = _ZENMatSetRandom; /* Set randomly a matrix */
  ZENMatSetBlocPtr(R) = _ZedMatSetBloc; /* Set a coefficient of a matrix */
  ZENMatGetBlocNewPtr(R) = _ZedMatGetBloc; /* Get a coefficient of a matrix */
  ZENMatGetBlocPtrPtr(R) = _ZedMatGetBlocPtr; /* Get a coefficient of a matrix */
  ZENMatSetSubMatPtr(R) = _ZedMatSetSubMat; /* Set a submatrix */
  ZENMatGetSubMatPtr(R) = _ZedMatGetSubMat; /* Get a submatrix */
  
  ZENMatPermuteBnPtr(R) = _ZENMatPermuteBn;
  ZENMatPermuteDigitPtr(R) = _ZENMatPermuteDigit;
  
  /* Test */
  ZENMatAreEqualPtr(R) = _ZENMatAreEqual; /* Is a matrix equal to another */
  ZENMatIsZeroPtr(R) = _ZENMatIsZero;	/* Is a matrix equal to zero */
  ZENMatIsOnePtr(R) = _ZENMatIsOne;	/* Is a matrix equal to one */
  
  /* Arithmetic */
  ZENMatAddPtr(R) = _ZENMatAdd;	/* Addition of Matrix */
  ZENMatAddScalarPtr(R) = _ZENMatAddScalar;	/* Addition of Matrix */
  ZENMatMultiplyScalarPtr(R) = _ZENMatMultiplyScalar;	/* Multiplication of Matrix */
  ZENMatSubtractPtr(R) = _ZENMatSubtract; /* Substraction of Matrix */
  ZENMatNegatePtr(R) = _ZENMatNegate;	/* Squaring of Matrix */
  ZENMatMultiplyPtr(R) = _ZENMatWinograd; /* Multiplication of Matrix */
  ZENMatMultiplyPlainPtr(R) = _ZENMatMultiplyPlain; /* Multiplication of Matrix */
  ZENMatMultiplyLeftPermutationPtr(R) = _ZENMatMultiplyLeftPermutation;
  ZENMatMultiplyRightPermutationPtr(R) = _ZENMatMultiplyRightPermutation;
  ZENMatInversePtr(R) = _ZENMatInverse; /* Inverse of a Matrix */
  ZENMatKernelPtr(R) = _ZENMatKernel;	/* Kernel of a Matrix */
  ZENMatGaussPtr(R) = _ZENMatGauss;
  
  /* IO */
  ZENMatGetFromFilePtr(R) = _ZENMatGetFromFile;
  ZENMatPutToFilePtr(R) = _ZENMatPutToFile;
  ZENMatGetFromStringPtr(R) = _ZENMatGetFromString;
  ZENMatPutToStringPtr(R) = _ZENMatPutToString;

}

/********************************************************/

/*
   \TEX{
   \title{Generic series functions}
   \input{A |ZENRing|.}
   \side{The series function pointers of |Rg| are set.}
   \note{These functions are not always optimal.}
   } 
   */
void __ZENSrGeneric(R)
     ZENRing R;
{
  /* Procedures on series
   **************************/
  /* Allocation */
  ZENSrCreatePtr(R) = _ZedSrCreate;	/* Create series */
  ZENSrFreePtr(R) = _ZedSrFree;	/* Free series  */
  
  /* Assigning */
  ZENSrAssignPtr(R) = _ZENSrAssign;	/* Assign a series with another */
  ZENSrSetToZeroPtr(R) = _ZENSrSetToZero; /* Set a series to zero */
  ZENSrSetRandomPtr(R) = _ZENSrSetRandom; /* Set randomly a series */
  ZENSrSetCoeffPtr(R) = _ZedSrSetCoeff; /* Set a coefficient of a series */
  ZENSrGetCoeffNewPtr(R) = _ZENSrGetCoeff; /* Get a coefficient of a series */
  ZENSrGetCoeffPtrPtr(R) = _ZedSrGetCoeffPtr; /* Get pointer to a coefficient of a series */
 ZENSrUpdateValuationPtr(R) = _ZENSrUpdateValuation; /* Update the valuation of a series */
  
  /* Test */
  ZENSrAreEqualPtr(R) = _ZENSrAreEqual; /* Is a series equal to another */
  ZENSrIsZeroPtr(R) = _ZENSrIsZero;	/* Is a series equal to zero */
  
  /* Arithmetic */
  ZENSrAddPtr(R) = _ZENSrAdd;		/* Addition of series */
  ZENSrNegatePtr(R) = _ZENSrNegate; /* Negation of series */
  ZENSrSubtractPtr(R) = _ZENSrSubtract; /* Substraction of series */
  ZENSrMultiplyPtr(R) = _ZENSrMultiply; /* Multiplication of series */
  ZENSrMultiplyScalarPtr(R) = _ZENSrMultiplyScalar; /* Multiplication of series by a scalar*/
  ZENSrDividePtr(R) = _ZENSrDivide; /* Division of series */ 
  if (ZBNDigit(ZENRingP(R), 0) == 2) {
    ZENSrSquarePtr(R) = _ZENSrSquareChar2;
  }
  else {
    ZENSrSquarePtr(R) = _ZENSrSquareCharp;
  }
}

/********************************************************/

/*
   \TEX{
   \title{Allocation of a |ZENRing|.}
   \input{The characteristic of the ring.}
   \output{A partially initialized |ZENRing|.}
   \note{This function is the firs invoked by all the ring creation functions.}
   } 
   */

ZENRing __ZENRingCreate(n,nl)
     BigNum n;
     BigNumLength nl;
{ZENRing R;

 /* Allocation */
 if (!(R = (ZENRing)ZENMalloc(sizeof(struct zenring)))) {
   ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
   return(ZNULL);
  }
 memset(R, 0, sizeof(struct zenring));

 ZENRingArith(R) = ZEN_LIB_ZEG;	/* This will be overwritten further */
 
 /* The characteristic */
 ZENRingPl(R) = nl;
 if (!(ZENRingP(R) = ZBNC(nl))) {
   ZENFree(R);
   ZENSetError(ZEN_LIB_ZEG, ZEN_INIT, ZEN_ALLOC);
   return(ZNULL);
  }
 ZBNAssign(ZENRingP(R), n, nl);

 /* Precomputations */
 ZENPrcSetNone(ZENRingPrcp(R));

 /* Procedure on elements
  ****************************/
 /* Arithmetic */
 ZENEltExpPtr(R) = _ZENEltExpPlain;	/* Exponentiation */

 /* Procedures on matrices
  **************************/
 /* Arithmetic */
 ZENMatWinogradPtr(R) = _ZENMatWinograd; /* Multiplication of Matrix */

 /* Debuging functions
  **************************/
				/* The following is necessary for
				 the linker not to erase the debugging
				 purpose functions from the functions
				 table */
 ZBPPtr(R)=ZBP;
 ZBRPtr(R)=ZBR;
 ZEPPtr(R)=ZEP;
 ZFPPtr(R)=ZFP;
 ZERPtr(R)=ZER;
 ZPPPtr(R)=ZPP;
 ZPRPtr(R)=ZPR;
 ZMPPtr(R)=ZMP;
 ZMRPtr(R)=ZMR;
 ZSPPtr(R)=ZSP;
 ZSRPtr(R)=ZSR;
 
 return(R);
}

/********************************************************/

int __ZENRingInitBuffers(R)
     ZENRing R;
{
  if (ZENEltCreateLarge(ZENRingHbuf(R), R))
    return(ZERR);
  
  if (ZENEltCreate(ZENRingIbuf(R), R)) {
    ZENEltFree(ZENRingHbuf(R), R);
    return(ZERR);
  }
  
  if (ZENEltCreate(ZENRingJbuf(R), R)) {
    ZENEltFree(ZENRingHbuf(R), R);
    ZENEltFree(ZENRingIbuf(R), R);
    return(ZERR);
  }

  if (ZENEltCreate(ZENRingFact(R), R)) {
    ZENEltFree(ZENRingJbuf(R), R);
    ZENEltFree(ZENRingIbuf(R), R);
    ZENEltFree(ZENRingHbuf(R), R);
    return(ZERR);
  }

  if ((ZENRingBNbuf(R)=ZBNC(ZENRingQl(R)))==NULL) {
    ZENEltFree(ZENRingJbuf(R), R);
    ZENEltFree(ZENRingIbuf(R), R);
    ZENEltFree(ZENRingHbuf(R), R);
    ZENEltFree(ZENRingFact(R), R);
    return(ZERR);
  }
  
  return(0); 
}

