/*
 * zen/zetab/c/init.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zetab.h"
#include "zeg.h"
#include "zed.h"

/* \TEX{
   \title{Initialization of tables for small extensions.}
   \input{A |ZENRing| |R| of cardinality less than $2^8$.}
   \output{|ZENERR| if an error occured, $|SR_GEN| \vert |MAT_GEN|$ otherwise.}
   \side{Tabulation of the clone |Rg| is performed.}
   \note{All the arithmetic operations are stored in tables.
   This function is called by |ZENRingClone| and should not be called otherwise.}
   }
   */
int _ZetabInit(Rg)
     ZENRing Rg;
{
  int ne;
  BigNumDigit C[1];
  ZENElt a, b, c;
  int i, j;
  BigNumLength k;

  ZENRingArith(Rg) = ZEN_LIB_ZETAB;

  /* Addition */
  if (!(ZetabAddTablePtr(Rg) = (ZEN_UC_SC2_Table *)ZENMalloc(sizeof(ZEN_UC_SC2_Table)))) {
    ZENSetError(ZEN_LIB_ZETAB, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  /* Negation */
  if (!(ZetabNegTablePtr(Rg) = (ZEN_UC_SC_Table *)ZENMalloc(sizeof(ZEN_UC_SC_Table)))) {
    ZENFree(ZetabAddTablePtr(Rg));
    ZENSetError(ZEN_LIB_ZETAB, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  /* Multiplication */
  if (!(ZetabMultTablePtr(Rg) = (ZEN_UC_SC2_Table *)ZENMalloc(sizeof(ZEN_UC_SC2_Table)))) {
    ZENFree(ZetabAddTablePtr(Rg));
    ZENFree(ZetabNegTablePtr(Rg));
    ZENSetError(ZEN_LIB_ZETAB, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  /* Inversion */
  if (!(ZetabInvTablePtr(Rg) = (ZEN_UC_SC_Table *)ZENMalloc(sizeof(ZEN_UC_SC_Table)))) {
    ZENFree(ZetabAddTablePtr(Rg));
    ZENFree(ZetabNegTablePtr(Rg));
    ZENFree(ZetabMultTablePtr(Rg));
    ZENSetError(ZEN_LIB_ZETAB, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  /* Trace */
  if (!(ZetabTrTablePtr(Rg) = (ZEN_UC_SC_Table *)ZENMalloc(sizeof(ZEN_UC_SC_Table)))) {
    ZENFree(ZetabAddTablePtr(Rg));
    ZENFree(ZetabNegTablePtr(Rg));
    ZENFree(ZetabMultTablePtr(Rg));
    ZENFree(ZetabInvTablePtr(Rg));
    ZENSetError(ZEN_LIB_ZETAB, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  if (ZENEltCreate(a, ZetabRingOrigin(Rg))) {
    ZENFree(ZetabAddTablePtr(Rg));
    ZENFree(ZetabNegTablePtr(Rg));
    ZENFree(ZetabMultTablePtr(Rg));
    ZENFree(ZetabInvTablePtr(Rg));
    ZENFree(ZetabTrTablePtr(Rg));
    ZENSetError(ZEN_LIB_ZETAB, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  if (ZENEltCreate(b, ZetabRingOrigin(Rg))) {
    ZENFree(ZetabAddTablePtr(Rg));
    ZENFree(ZetabNegTablePtr(Rg));
    ZENFree(ZetabMultTablePtr(Rg));
    ZENFree(ZetabInvTablePtr(Rg));
    ZENFree(ZetabTrTablePtr(Rg));
    ZENEltFree(b, ZetabRingOrigin(Rg));
    ZENSetError(ZEN_LIB_ZETAB, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  if (ZENEltCreate(c, ZetabRingOrigin(Rg))) {
    ZENFree(ZetabAddTablePtr(Rg));
    ZENFree(ZetabNegTablePtr(Rg));
    ZENFree(ZetabMultTablePtr(Rg));
    ZENFree(ZetabInvTablePtr(Rg));
    ZENFree(ZetabTrTablePtr(Rg));
    ZENEltFree(a, ZetabRingOrigin(Rg));
    ZENEltFree(b, ZetabRingOrigin(Rg));
    ZENSetError(ZEN_LIB_ZETAB, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  /* The size of the field */
  ne = (int)ZENRingQ(Rg)[0];

  /* Conversion */
  if (!(ZetabConvTablePtr(Rg) = (ZENElt *)ZENMalloc(ne*sizeof(ZENElt)))) {
    ZENFree(ZetabAddTablePtr(Rg));
    ZENFree(ZetabNegTablePtr(Rg));
    ZENFree(ZetabMultTablePtr(Rg));
    ZENFree(ZetabInvTablePtr(Rg));
    ZENFree(ZetabTrTablePtr(Rg));
    ZENEltFree(a, ZetabRingOrigin(Rg));
    ZENEltFree(b, ZetabRingOrigin(Rg));
    ZENEltFree(c, ZetabRingOrigin(Rg));
    ZENSetError(ZEN_LIB_ZETAB, ZEN_INIT, ZEN_ALLOC);
    return(ZERR);
  }

  /************ Filling the tables *********************/
  
  
  for(i = 0; i < ne; i++) {
    C[0] = i;
    ZENZToElt(a, C, 1, ZetabRingOrigin(Rg));

    if (ZENEltCreate(ZetabConvTable(Rg, i), ZetabRingOrigin(Rg))) {
      ZENFree(ZetabAddTablePtr(Rg));
      ZENFree(ZetabNegTablePtr(Rg));
      ZENFree(ZetabMultTablePtr(Rg));
      ZENFree(ZetabInvTablePtr(Rg));
      ZENFree(ZetabTrTablePtr(Rg));
      ZENEltFree(a, ZetabRingOrigin(Rg));
      ZENEltFree(b, ZetabRingOrigin(Rg));
      ZENEltFree(c, ZetabRingOrigin(Rg));
      for(i--;i>=0;i--)
	ZENEltFree(ZetabConvTable(Rg, i), ZetabRingOrigin(Rg));
      ZENFree(ZetabConvTablePtr(Rg));
      ZENSetError(ZEN_LIB_ZETAB, ZEN_INIT, ZEN_ALLOC);
      return(ZERR);
    }
    ZENEltAssign(ZetabConvTable(Rg, i), a, ZetabRingOrigin(Rg));
    
    for(j = 0; j < ne; j++) {

      C[0] = j;
      ZENZToElt(b, C, 1, ZetabRingOrigin(Rg));

      /* Addition */
      ZENEltAssign(c, a, ZetabRingOrigin(Rg));
      ZENEltAdd(c, b, ZetabRingOrigin(Rg));
      ZENEltToZ(C, &k, c, ZetabRingOrigin(Rg));
      ZetabAddTable(Rg, i, j) = C[0];
      
      /* Multiplication */
      ZENEltMultiply(c, a, b, ZetabRingOrigin(Rg));
      ZENEltToZ(C, &k, c, ZetabRingOrigin(Rg));
      ZetabMultTable(Rg, i, j) = C[0];

    }
    /* Negation */
    ZENEltNegate(c, a, ZetabRingOrigin(Rg));
    ZENEltToZ(C, &k, c, ZetabRingOrigin(Rg));
    ZetabNegTable(Rg, i) = C[0];

    ZENEltTrace(c, a, ZetabRingOrigin(Rg));
    ZENEltToZ(C, &k, c, ZetabRingOrigin(Rg));
    ZetabTrTable(Rg, i) = C[0];
    if (ZENEltIsOne(c, ZetabRingOrigin(Rg))) {
      ZetabElt2Int(ZENRingTheta(Rg), Rg) = C[0];
    }

    if (!i) {
      ZetabInvTable(Rg, i) = 0;
    }
    else {
      switch(ZENEltInverse(c, a, ZetabRingOrigin(Rg))) {
      case ZEN_HAS_INVERSE:
	ZENEltToZ(C, &k, c, ZetabRingOrigin(Rg));
	ZetabInvTable(Rg, i) = C[0];
	break;
	
      case ZEN_NO_INVERSE:
	ZetabInvTable(Rg, i) = 0;
	break;

      case ZENERR:
	ZENFree(ZetabAddTablePtr(Rg));
	ZENFree(ZetabNegTablePtr(Rg));
	ZENFree(ZetabMultTablePtr(Rg));
	ZENFree(ZetabInvTablePtr(Rg));
	ZENFree(ZetabTrTablePtr(Rg));
	ZENEltFree(a, ZetabRingOrigin(Rg));
	ZENEltFree(b, ZetabRingOrigin(Rg));
	ZENEltFree(c, ZetabRingOrigin(Rg));
	for(i--;i>=0;i--)
	  ZENEltFree(ZetabConvTable(Rg, i), ZetabRingOrigin(Rg));
	ZENFree(ZetabConvTablePtr(Rg));
	return(ZERR);
      }
    }
  }
  ZENEltFree(a, ZetabRingOrigin(Rg));
  ZENEltFree(b, ZetabRingOrigin(Rg));
  ZENEltFree(c, ZetabRingOrigin(Rg));


  /************ new pointers of functions are used ***************/

  /* Procedure on rings
   ****************************/
  ZENPrcInitEltMultiplyPtr(Rg) = NULL;
  ZENPrcCloseEltMultiplyPtr(Rg) = NULL;

  ZENPrcInitTracePtr(Rg) = NULL;
  ZENPrcCloseTracePtr(Rg) = NULL;

  ZENPrcInitPolyRootsCanonicalPtr(Rg) = _ZetabPrcInitPolyRootsCanonical;
  ZENPrcClosePolyRootsCanonicalPtr(Rg) = _ZetabPrcClosePolyRootsCanonical;

  ZENPrcInitEltExpPtr(Rg) = _ZENPrcInitEltExp;
  ZENPrcCloseEltExpPtr(Rg) = _ZENPrcCloseEltExp;

  /* Procedure on elements
   ****************************/
  /* Allocation */
  ZENEltCreatePtr(Rg) = _ZetabEltCreate;
  ZENEltCreateLargePtr(Rg) = _ZetabEltCreate;
  ZENEltFreePtr(Rg) = _ZetabEltFree;
  
  /* Assigning */
  ZENEltAssignPtr(Rg) = _ZetabEltAssign;
  ZENEltSetToZeroPtr(Rg) = _ZetabEltSetToZero;
  ZENEltSetToOnePtr(Rg) = _ZetabEltSetToOne;
  ZENEltSetRandomPtr(Rg) = _ZetabEltSetRandom;
  ZENEltSetNextPtr(Rg) = _ZetabEltSetNext;

  ZENEltToClonePtr(Rg) = _ZetabEltToClone;
  ZENCloneToEltPtr(Rg) = _ZetabCloneToElt;

  /* Test */
  ZENEltAreEqualPtr(Rg) = _ZetabEltAreEqual;
  ZENEltIsZeroPtr(Rg) = _ZetabEltIsZero;
  ZENEltIsOnePtr(Rg) = _ZetabEltIsOne;

  /* Arithmetic */
  ZENEltAddPtr(Rg) = _ZetabEltAdd;
  ZENEltNegatePtr(Rg) = _ZetabEltNegate;
  ZENEltSubtractPtr(Rg) = _ZetabEltSubtract;
  ZENEltSquarePtr(Rg) = _ZetabEltSquare;
  ZENEltMultiplyPtr(Rg) = _ZetabEltMultiply;
  ZENEltInversePtr(Rg) = _ZetabEltInverse;
  ZENEltTracePtr(Rg) = _ZetabEltTrace;

  /* io */
  ZENZToEltPtr(Rg) = _ZetabZToElt;
  ZENEltToZPtr(Rg) = _ZetabEltToZ;

  ZENEltReadFromStringPtr(Rg) = _ZENCloneEltReadFromString;
  ZENEltPrintToStringPtr(Rg) = _ZENCloneEltPrintToString;
  ZENEltReadFromFilePtr(Rg) = _ZENCloneEltReadFromFile;
  ZENEltPrintToFilePtr(Rg) = _ZENCloneEltPrintToFile;
  
  ZENEltGetFromStringPtr(Rg) = _ZENCloneEltGetFromString;
  ZENEltPutToStringPtr(Rg) = _ZENCloneEltPutToString;
  ZENEltGetFromFilePtr(Rg) = _ZENCloneEltGetFromFile;
  ZENEltPutToFilePtr(Rg) = _ZENCloneEltPutToFile;
  
  /* Procedures on polynomials
   *******************************/
  /* Allocation */
  ZENPolyCreatePtr(Rg) = _ZetabPolyCreate;
  ZENPolyCreateLargePtr(Rg) = _ZetabPolyCreate;
  ZENPolyFreePtr(Rg) = _ZetabPolyFree;

  /* Assigning */
  ZENPolyAssignPtr(Rg) = _ZetabPolyAssign;
  ZENPolyAssignLargePtr(Rg) = _ZetabPolyAssign;
  ZENPolyLazyToVeryLazyPtr(Rg) = _ZetabPolyAssign;
  ZENPolySetToZeroPtr(Rg) = _ZetabPolySetToZero;
  ZENPolySetToZeroLargePtr(Rg) = _ZetabPolySetToZero;
  ZENPolySetToXiPtr(Rg) = _ZetabPolySetToXi;
  ZENPolySetRandomPtr(Rg) = _ZetabPolySetRandom;
  ZENPolySetCoeffPtr(Rg) = _ZetabPolySetCoeff;
  ZENPolyGetCoeffNewPtr(Rg) = _ZetabPolyGetCoeff;
  ZENPolyGetCoeffPtrPtr(Rg) = _ZetabPolyGetCoeffPtr;
  ZENPolyUpdateDegreePtr(Rg) = _ZetabPolyUpdateDegree;

  /* Test */
  ZENPolyAreEqualPtr(Rg) = _ZetabPolyAreEqual;
  ZENPolyIsZeroPtr(Rg) = _ZetabPolyIsZero;
  ZENPolyIsXiPtr(Rg) = _ZetabPolyIsXi;

  /* Arithmetic */
  ZENPolyAddPtr(Rg) = _ZetabPolyAdd;
  ZENPolyAddLargePtr(Rg) = _ZetabPolyAdd;
  ZENPolyNegatePtr(Rg) = _ZetabPolyNegate;
  ZENPolyNegateLargePtr(Rg) = _ZetabPolyNegate;
  ZENPolySubtractPtr(Rg) = _ZetabPolySubtract;
  ZENPolySubtractLargePtr(Rg) = _ZetabPolySubtract;

  if (ZBNDigit(ZENRingP(Rg), 0) == 2) {
    ZENPolySquarePtr(Rg) = _ZetabPolySquareChar2;
    ZENPolySquareLazyPtr(Rg) = _ZetabPolySquareChar2;
    ZENPolySquareVeryLazyPtr(Rg) = _ZetabPolySquareChar2;
  }
  else {
    ZENPolySquarePtr(Rg) = _ZetabPolySquareKaratsuba;
    ZENPolySquareLazyPtr(Rg) = _ZetabPolySquareKaratsubaLazy;
    ZENPolySquareVeryLazyPtr(Rg) = _ZetabPolySquareKaratsubaLazy;
  }

  ZENPolyMultiplyPtr(Rg) = _ZetabPolyMultiply;
  ZENPolyMultiplyLazyPtr(Rg) = _ZetabPolyMultiplyLazy;
  ZENPolyMultiplyVeryLazyPtr(Rg) = _ZetabPolyMultiplyLazy;

  ZENPolyMultiplyScalarPtr(Rg) = _ZetabPolyMultiplyScalar;
  ZENPolyDotPtr(Rg) = _ZetabPolyDot;
  ZENPolyEvalPtr(Rg) = _ZENPolyEval; /* Evaluation of a polynomial */
  ZENPolyDividePtr(Rg) = _ZetabPolyDivide;
  ZENPolyDivideLazyPtr(Rg) = _ZetabPolyDivideLazy;
  ZENPolyDivideVeryLazyPtr(Rg) = _ZetabPolyDivideLazy;
  ZENPolyModuloLazyPtr(Rg) = _ZetabPolyModuloLazy;
  ZENPolyModuloVeryLazyPtr(Rg) = _ZetabPolyModuloLazy;

  /* Procedures on matrices
   **************************/
  /* Allocation */
  ZENMatCreateParamsPtr(Rg) = _ZetabMatCreateParams; /* Create a matrix */
  ZENMatCreatePtr(Rg) = __ZedsMatCreate;	/* Create a matrix */

  /* Assigning */
  ZENMatAssignPtr(Rg) = _ZetabMatAssign; /* Assign a matrix with another */
  ZENPermutation2MatrixPtr(Rg) = _ZENPermutation2Matrix;	/* Assign a matrix with another */
  ZENMatSetToZeroPtr(Rg) = _ZetabMatSetToZero; /* Set a matrix to zero */
  ZENMatSetToOnePtr(Rg) = _ZetabMatSetToOne; /* Set a matrix to identity */
  ZENMatSetRandomPtr(Rg) = _ZetabMatSetRandom; /* Set randomly a matrix */
  ZENMatSetBlocPtr(Rg) = _ZetabMatSetBloc; /* Set a coefficient of a matrix */
  ZENMatGetBlocNewPtr(Rg) = _ZetabMatGetBloc; /* Get a coefficient of a matrix */
  ZENMatGetBlocPtrPtr(Rg) = _ZetabMatGetBlocPtr; /* Get a coefficient of a matrix */
  ZENMatSetSubMatPtr(Rg) = _ZetabMatSetSubMat; /* Set a submatrix */
  ZENMatGetSubMatPtr(Rg) = _ZetabMatGetSubMat; /* Get a submatrix */
  
  ZENMatPermuteBnPtr(Rg) = _ZetabMatPermuteBn;
  ZENMatPermuteDigitPtr(Rg) = _ZetabMatPermuteDigit;
  
  /* Test */
  ZENMatAreEqualPtr(Rg) = _ZetabMatAreEqual; /* Is a matrix equal to another */
  ZENMatIsZeroPtr(Rg) = _ZetabMatIsZero;	/* Is a matrix equal to zero */
  ZENMatIsOnePtr(Rg) = _ZetabMatIsOne;	/* Is a matrix equal to one */
  
  /* Arithmetic */
  ZENMatAddPtr(Rg) = _ZetabMatAdd;	/* Addition of Matrix */
  ZENMatAddScalarPtr(Rg) = _ZetabMatAddScalar;	/* Addition of Matrix */
  ZENMatMultiplyScalarPtr(Rg) = _ZetabMatMultiplyScalar;	/* Multiplication of Matrix */
  ZENMatSubtractPtr(Rg) = _ZetabMatSubtract; /* Substraction of Matrix */
  ZENMatNegatePtr(Rg) = _ZetabMatNegate;	/* Squaring of Matrix */
  ZENMatWinogradPtr(Rg) = _ZENMatWinograd; /* Multiplication of Matrix */
  ZENMatMultiplyPtr(Rg) = _ZetabMatMultiplyPlain; /* Multiplication of Matrix */
  ZENMatMultiplyPlainPtr(Rg) = _ZetabMatMultiplyPlain; /* Multiplication of Matrix */
  ZENMatMultiplyLeftPermutationPtr(Rg) = _ZENMatMultiplyLeftPermutation;
  ZENMatMultiplyRightPermutationPtr(Rg) = _ZENMatMultiplyRightPermutation;
  ZENMatInversePtr(Rg) = _ZetabMatInverse; /* Inverse of a Matrix */
  ZENMatKernelPtr(Rg) = _ZetabMatKernel;	/* Kernel of a Matrix */  
  ZENMatGaussPtr(Rg) = _ZetabMatGauss;
  
  /* IO */
  ZENMatGetFromFilePtr(Rg) = _ZENMatGetFromFile;
  ZENMatPutToFilePtr(Rg) = _ZENMatPutToFile;
  ZENMatGetFromStringPtr(Rg) = _ZENMatGetFromString;
  ZENMatPutToStringPtr(Rg) = _ZENMatPutToString;

  /* Matrix conversion */
	       
  if(ZENRingArith(ZENRingOrigin(Rg))==ZEN_LIB_ZEXT)
    {
      if(ZENRingArith(ZENRingDef(Rg))==ZEN_LIB_ZE2)
	{
	  ZENVect2MatPtr(Rg) = _Zetab2mVect2Mat;
	  ZENMat2VectPtr(Rg) = _Zetab2mMat2Vect;
	}
      else
	{
	  ZENVect2MatPtr(Rg) = _ZENCloneVect2Mat;
	  ZENMat2VectPtr(Rg) = _ZENCloneMat2Vect;
	}
    }
  else
    {
      ZENVect2MatPtr(Rg) = ZENMatAssignPtr(Rg);
      ZENMat2VectPtr(Rg) = ZENMatAssignPtr(Rg);
    }

  return(SR_GEN);
}


/* \TEX{
   \title{Desallocating tables for small extensions.}
   \input{A |ZENRing| of cardinality less than $2^8$.}
   \output{|ZENERR| if an error occured, |ZEN_HAS_INVERSE| otherwise.}
   \note{All tables allocated with |_ZetabInit| are
   freed.}
   }
   */
int _ZetabClose(Rg)
     ZENRing Rg;
{int i;

 i=(int)ZENRingQ(Rg)[0];
 
 for(i--;i>=0;i--)
   ZENEltFree(ZetabConvTable(Rg, i), ZetabRingOrigin(Rg));
 ZENFree(ZetabConvTablePtr(Rg));

 /* Restoring the allocated memory */
 ZENFree(ZetabAddTablePtr(Rg));
 ZENFree(ZetabNegTablePtr(Rg));
 ZENFree(ZetabMultTablePtr(Rg));
 ZENFree(ZetabInvTablePtr(Rg));
 ZENFree(ZetabTrTablePtr(Rg));
 

 return(ZEN_HAS_INVERSE);
}

