/*
 * zen/zetab/c/mat_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matricial assignments}

   }
   */
#include <stdio.h>
#include "zetab.h"

/*
   \TEX{
   \title{Get a bloc}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{|E| is set to the indexed bloc of |A|}
   }
   */
   
int _ZetabMatGetBloc(p_E,A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENElt *p_E;
     ZENRing R;
{
  ZENTest(ZetabMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATGETBLOC,ZEN_LIB_ZETAB, ZERR);
  ZENTest((l<0)||(c<0)||(l>=ZetabMatNbBn(A,R))||(c>=ZetabMatNbDigit(A,R)),
	  ZEN_MAT_DIM0,ZEN_MATGETBLOC,ZEN_LIB_ZETAB, ZERR);
  
  ZetabAnyEltAssign(ZetabElt2Int(*p_E,R),ZetabMatBloc(A,l,c,R),R);
  return(0);
}

/*
   \TEX{
   \title{Set a bloc}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{The indexed bloc of |A| is set to |E|}
   \note{Except for advanced user, |GetCoeff| functions should be preferred.
   This function must be chosen together with |_ZetabMatCreateParams|.}
   }
   
   */
   
int _ZetabMatSetBloc(A,l,c,E,R)
     ZENMat A;
     Dim l,c;
     ZENElt E;
     ZENRing R;
{
  ZENTest(ZetabMatIsPermutation(A,R),
	  ZEN_ISPERM,ZEN_MATSETBLOC,ZEN_LIB_ZETAB, ZERR);
  ZENTest((l<0)||(c<0)||(l>=ZetabMatNbBn(A,R))||(c>=ZetabMatNbDigit(A,R)),
	  ZEN_MAT_DIM0,ZEN_MATSETBLOC,ZEN_LIB_ZETAB, ZERR);

  ZetabAnyEltAssign(ZetabMatBloc(A,l,c,R),ZetabElt2Int(E,R),R);
  return(0);
}

/*
   \TEX{
   \title{Get a bloc pointer}
   \input{A |ZENMat| |A|, two indexes and an element |E| of a finite ring |R|.}
   \output{|ZENERR| if an error occured, 0 otherwise}
   \side{|E| is set to the indexed bloc of |A|}
   \note{This function must be chosen together with |_ZedMatCreateParams|.}
   }
   
   */
   
ZENElt _ZetabMatGetBlocPtr(A,l,c,R)
     ZENMat A;
     Dim l,c;
     ZENRing R;
{
  ZENElt ret;

  ZetabAnyEltAssign(ZetabElt2Int(ret,R),ZetabMatBloc(A,l,c,R),R);
  return(ret);
}

/************************************************/

static void __ZetabMatGetSubMatST(S,A,l,c,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int i,j,k;
 
 for(i=0;i<ZetabMatNbBn(S,R);i++,l++)
   for(j=0,k=c;j<ZetabMatNbDigit(S,R);j++,k++)    
     ZetabAnyEltAssign(ZetabMatBloc(S,i,j,R),ZetabMatBloc(A,l,k,R),R);
}

static void __ZetabMatGetSubMatOT(S,A,l,c,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int i,j,k;
 
 for(i=0;i<ZetabMatNbBn(S,R);i++,l++)
   for(j=0,k=c;j<ZetabMatNbDigit(S,R);j++,k++)    
     ZetabAnyEltAssign(ZetabMatBloc(S,i,j,R),ZetabMatBloc(A,k,l,R),R);
}

static void __ZetabMatSetSubMatST(A,l,c,S,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int i,j,k;
 
 for(i=0;i<ZetabMatNbBn(S,R);i++,l++)
   for(j=0,k=c;j<ZetabMatNbDigit(S,R);j++,k++)    
     ZetabAnyEltAssign(ZetabMatBloc(A,l,k,R),ZetabMatBloc(S,i,j,R),R);
}

static void __ZetabMatSetSubMatOT(A,l,c,S,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int i,j,k;
 
 for(i=0;i<ZetabMatNbBn(S,R);i++,l++)
   for(j=0,k=c;j<ZetabMatNbDigit(S,R);j++,k++)    
     ZetabAnyEltAssign(ZetabMatBloc(A,k,l,R),ZetabMatBloc(S,i,j,R),R);
}

/*
   \TEX{
   \title{Getting a submatrix}
   \input{Two allocated |ZENMat|, two indexes and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The submatrix of |M| at position $(|r|,|c|)$ of size those of |S| is extracted and assigned to |S|.}
   \note{Permutation matrix cannot be used.}
   }
   */

int _ZetabMatGetSubMat(S,A,l,c,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int cas;
 
 ZENTest((ZetabMatIsPermutation(A,R))||(ZetabMatIsPermutation(S,R)),
	 ZEN_ISPERM,ZEN_MATGETSUBMAT,ZEN_LIB_ZETAB, ZERR);
 ZENTest((l<0)||(c<0)||((l+ZetabMatNbRow(S,R))>ZetabMatNbRow(A,R))||((c+ZetabMatNbCol(S,R))>ZetabMatNbCol(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATGETSUBMAT,ZEN_LIB_ZETAB, ZERR);

 cas=(ZetabMatIsRowType(A,R)<<1)|(ZetabMatIsRowType(S,R));

 switch(cas)
   {
   case _RR_:
     __ZetabMatGetSubMatST(S,A,l,c,R);
     break;
   case _RC_:
     __ZetabMatGetSubMatOT(S,A,c,l,R);
     break;
   case _CR_:
     __ZetabMatGetSubMatOT(S,A,l,c,R);
     break;
   case _CC_:
     __ZetabMatGetSubMatST(S,A,c,l,R);
     break;
   }   
 return(0);
}

/*
   \TEX{
   \title{Assigning a submatrix}
   \input{Two |ZENMat|, two indexes, a |ZENElt| and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The matrix |S| is assigned at position $(|r|,|c|)$ in matrix |M|.}
   \note{Permutation matrix cannot be used.}
   }
   */
int _ZetabMatSetSubMat(A,l,c,S,R)
     ZENMat A,S;
     Dim l,c;
     ZENRing R;
{int cas;
 
 ZENTest(ZetabMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETSUBMAT,ZEN_LIB_ZETAB, ZERR);
 ZENTest((l<0)||(c<0)||((l+ZetabMatNbRow(S,R))>ZetabMatNbRow(A,R))||((c+ZetabMatNbCol(S,R))>ZetabMatNbCol(A,R)),
	 ZEN_MAT_DIM0,ZEN_MATSETSUBMAT,ZEN_LIB_ZETAB, ZERR);
 
 cas=(ZetabMatIsRowType(A,R)<<1)|(ZetabMatIsRowType(S,R));

 switch(cas)
   {
   case _RR_:
     __ZetabMatSetSubMatST(A,l,c,S,R);
     break;
   case _RC_:
     __ZetabMatSetSubMatOT(A,c,l,S,R);
     break;
   case _CR_:
     __ZetabMatSetSubMatOT(A,l,c,S,R);
     break;
   case _CC_:
     __ZetabMatSetSubMatST(A,c,l,S,R);
     break;
   }     
  
  return(0);
}

/*
   \TEX{
   
   \title{Assignment.}
   \input{Two |ZENMat| |A| and |B| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{|A| = |B| if no error occured, |zen_errno| is set otherwise.}
   }
   */
   

int _ZetabMatAssign(A,B,R)
     ZENMat A,B;
     ZENRing R;
{int l,c;

 ZENTest(ZetabMatIsPermutation(A,R)
	 ||ZetabMatIsPermutation(B,R),
	 ZEN_ISPERM,ZEN_MATASSIGN,ZEN_LIB_ZETAB, ZERR);
 
 if(ZetabMatAreSameType(A,B,R))
   ZetabMatSameTypeAssign(A,B,R);
 else
   {
     ZENTest((ZetabMatNbBn(A,R) != ZetabMatNbDigit(B,R))
	     ||(ZetabMatNbDigit(A,R) != ZetabMatNbBn(B,R)),
	     ZEN_MAT_DIM,ZEN_MATASSIGN,ZEN_LIB_ZETAB, ZERR);
     
     for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
       for(c=ZetabMatNbDigit(A,R)-1;c>=0;c--)
	 ZetabAnyEltAssign(ZetabMatBloc(A,l,c,R),ZetabMatBloc(B,c,l,R),R);
   }
 return(0);
}

/*
   \TEX{
   \title{Set to zero}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are set to zero.}
   }
   */
   
int _ZetabMatSetToZero(A,R)
     ZENMat A;
     ZENRing R;
{

 ZENTest(ZetabMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETZERO,ZEN_LIB_ZETAB, ZERR);

 ZetabMatSetToZero(A,R);
 
 return(0);
}

/*
   \TEX{
   \title{Set to one}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are set to zero except for the coefficients with same indices on rows and columns that are set to one.}
   \note{There is no error if the matrix is not square.}
   }
   */
   
int _ZetabMatSetToOne(A,R)
     ZENMat A;
     ZENRing R;
{int l,c;

 ZENTest(ZetabMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETONE,ZEN_LIB_ZETAB, ZERR);

 for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZetabMatNbDigit(A,R)-1;c>=0;c--)
     if(l==c)
       ZetabAnyEltSetToOne(ZetabMatBloc(A,l,c,R),R);
     else
       ZetabAnyEltSetToZero(ZetabMatBloc(A,l,c,R),R);
     
 return(0);
}

/*
   \TEX{
   \title{Set to random}
   \input{A |ZENMat| |A| of a finite ring |R|.}
   \output{0 if no error occured, |ZENERR| otherwise}
   \side{All the coefficients of |A| are randomly set.}
   }
   
   */
   
int _ZetabMatSetRandom(A,R)
     ZENMat A;
     ZENRing R;
{int l,c;

 ZENTest(ZetabMatIsPermutation(A,R),
	 ZEN_ISPERM,ZEN_MATSETRANDOM,ZEN_LIB_ZETAB, ZERR);
 for(l=ZetabMatNbBn(A,R)-1;l>=0;l--)
   for(c=ZetabMatNbDigit(A,R)-1;c>=0;c--)
     ZetabAnyEltSetRandom(ZetabMatBloc(A,l,c,R),R);

 return(0);
}

