/*
 * zen/zetab/c/mat_gauss.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Gauss elimination}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions. 
   }
   */
#include <stdio.h>
#include "zetab.h"
#include "zeg.h"

#ifndef min
#  define min(x,y) ((x)<(y)?(x):(y))
#endif

/*
  \TEX{
   \title{General gaussian elimination}
   \input{Three matrices, a pointer on integer, a pointer on a |ZENElt| and on permutation function to use}
   \output{|ZENERR| if an error occurred,  |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned,
   |ZENMAT_HAS_MAXIMAL_RANK| if the matrix is of maximal rank,
   |ZENMAT_HAS_KERNEL| otherwise.}
   \side{Compute |S|, |P|, |p_rk| or |p_det| only if non |NULL| pointers.}
   \note{
   Permute is the operation to use to permute |D|.
   If |P| and |p_rk| are |NULL|, process stops as soon as the matrix is not of maximal rank.
   If |P| and |S| are |NULL|, trigonalization is performed.
   This function should never appear in a program. Use the encapsulated functions instead.
   }
   }
 */
int _ZetabMatGauss(D,S,P,p_rk,p_det,Permute,R)
     ZENMat D,S,P;
     ZENElt *p_det;
     Dim *p_rk;
     int (*Permute) ___((ZENMat,int,int,ZENRing));
     ZENRing R;
{int k,l,c,flag;
 Dim ret,lim,nbbn;
 ZENElt z,g,h;

 nbbn=ZetabMatNbBn(D,R)-(((int)(ZetabMatType(D,R)))>>2); /* Tricky feature:
					       not documented !
					       I need that... */
 ZENTest(ZetabMatIsPermutation(D,R),
	 ZEN_ISPERM,ZEN_MATGAUSS,ZEN_LIB_ZETAB, ZERR);
 if(p_det) ZetabEltSetToOne(*p_det,R);
 if(S)
   {
     ZENTest(ZetabMatIsPermutation(S,R),
	     ZEN_ISPERM,ZEN_MATGAUSS,ZEN_LIB_ZETAB, ZERR);
     ZENTest(!ZetabMatAreSameType(S,D,R),
	     ZEN_MAT_TYPE,ZEN_MATGAUSS,ZEN_LIB_ZETAB, ZERR);
     ZENTest((nbbn != ZetabMatNbBn(S,R))
	     ||(ZetabMatNbBn(S,R) != ZetabMatNbDigit(S,R)),
	     ZEN_MAT_DIM,ZEN_MATGAUSS,ZEN_LIB_ZETAB, ZERR);
   }
   
 if(P)
   {
     ZENTest(ZetabMatAreSameType(P,D,R),
	     ZEN_MAT_TYPE,ZEN_MATGAUSS,ZEN_LIB_ZETAB, ZERR);
     ZENTest((ZetabMatNbDigit(D,R) != ZetabMatNbBn(P,R))
	     ||(ZetabMatNbDigit(P,R) != ZetabMatNbBn(P,R)),
	     ZEN_MAT_DIM,ZEN_MATGAUSS,ZEN_LIB_ZETAB, ZERR);
   }
 
 ret=ZetabMatNbDigit(D,R);
 lim=min(nbbn,ret);
 flag=ZENMAT_HAS_MAXIMAL_RANK;
 
 for(k=0;k<lim;k++)
   {
     while(ZetabAnyEltIsZero(ZetabMatBloc(D,k,k,R),R))
       {
         l=k+1;
	 
         while(l<nbbn)
           if(ZetabAnyEltIsZero(ZetabMatBloc(D,l,k,R),R))
             l++;
           else
             {
               (*Permute)(D,k,l,R);
	       if(S) (*Permute)(S,k,l,R);
	       if(p_det) ZetabEltNegate(*p_det,*p_det,R);
	       goto OK;
             }

	 if((S)&&(!P))			/* GaussInverse */
	   {
	     ZENSetError(ZEN_LIB_ZETAB,ZEN_MATGAUSS,ZEN_MAT_NOTINV);
	     return(ZERR);
	   }

	 if (--ret > k)
	   {
	     if(ret<lim)
	       {
		 if(p_det) ZetabEltSetToZero(*p_det,R);
		 if((!p_rk)&&(!P))		/* GaussTrigInvertible */
		   return(ZENMAT_HAS_KERNEL);
		 flag=ZENMAT_HAS_KERNEL;
		 lim=ret;
	       }
             ZetabMatPermuteDigit(D,k,ret,R);
             if(P) ZENMatPermuteBn(P,k,ret,R); /* ZEN fct for the case of permutation */
	   }
         else
	   {
	     if(p_det) ZetabEltSetToZero(*p_det,R);
	     flag=ZENMAT_HAS_KERNEL;
	     goto FIN;
	   }
       }

     OK :
       switch(ZetabAnyEltInverse(ZetabElt2Int(h,R),ZetabMatBloc(D,k,k,R),R))
	 {
	 case ZEN_HAS_INVERSE:
	   break; 
	 case ZEN_NO_INVERSE:
	   if(p_rk) *p_rk= -1;
	   return(ZEN_NO_INVERSE); 
	 default:
	   return(ZERR);
	 }

     if(p_det) ZetabAnyEltMultiply(ZetabElt2Int(*p_det,R),ZetabMatBloc(D,k,k,R),ZetabElt2Int(*p_det,R),R);

#define nop()
#define ___SubSComputation()                                         \
       for(c=0;c<ZetabMatNbDigit(S,R);c++)                             \
          { ZetabAnyEltMultiply(ZetabElt2Int(g,R),ZetabElt2Int(z,R),ZetabMatBloc(S,k,c,R),R);             \
            ZetabAnyEltAdd(ZetabMatBloc(S,l,c,R),ZetabElt2Int(g,R),R); }
	 

#define ___Computation(foo)                                           \
     if(!(ZetabAnyEltIsZero(ZetabMatBloc(D,l,k,R),R)))                     \
       { ZetabAnyEltNegate(ZetabElt2Int(g,R),ZetabMatBloc(D,l,k,R),R);                     \
	 ZetabEltMultiply(z,h,g,R);                                   \
	 foo();                                                       \
	 ZetabAnyEltSetToZero(ZetabMatBloc(D,l,k,R),R);                    \
         for(c=k+1;c<ZetabMatNbDigit(D,R);c++)                           \
           if(!(ZetabAnyEltIsZero(ZetabMatBloc(D,k,c,R),R)))               \
             { ZetabAnyEltMultiply(ZetabElt2Int(g,R),ZetabElt2Int(z,R),ZetabMatBloc(D,k,c,R),R);           \
               ZetabAnyEltAdd(ZetabMatBloc(D,l,c,R),ZetabElt2Int(g,R),R); }                \
       } else

  if(S)
    for(l=k+1;l<ZetabMatNbBn(D,R);l++)
      ___Computation(___SubSComputation);
  else
    for(l=k+1;l<ZetabMatNbBn(D,R);l++)
      ___Computation(nop);
   }
   
FIN:
 if(p_rk) *p_rk=k;             
 lim=k;
 
#undef ___Computation
#define ___Computation(foo)                                           \
     if(!(ZetabAnyEltIsZero(ZetabMatBloc(D,l,k,R),R)))                     \
       { ZetabAnyEltNegate(ZetabElt2Int(z,R),ZetabMatBloc(D,l,k,R),R);                     \
	 foo();                                                       \
	 ZetabAnyEltSetToZero(ZetabMatBloc(D,l,k,R),R);                    \
         for(c=lim;c<ZetabMatNbDigit(D,R);c++)                           \
           if(!(ZetabAnyEltIsZero(ZetabMatBloc(D,k,c,R),R)))               \
             { ZetabAnyEltMultiply(ZetabElt2Int(g,R),ZetabElt2Int(z,R),ZetabMatBloc(D,k,c,R),R);           \
               ZetabAnyEltAdd(ZetabMatBloc(D,l,c,R),ZetabElt2Int(g,R),R); }                \
       } else

#define ___Normalization \
       ZENCheckError(ZetabAnyEltInverse(ZetabElt2Int(z,R),ZetabMatBloc(D,k,k,R),R),ZERR); \
       ZetabAnyEltSetToOne(ZetabMatBloc(D,k,k,R),R); \
       for(c=lim;c<ZetabMatNbDigit(D,R);c++) \
         ZetabAnyEltMultiply(ZetabMatBloc(D,k,c,R),ZetabElt2Int(z,R),ZetabMatBloc(D,k,c,R),R);
       
 if(S)
   for(;k--;)
     {
       ___Normalization;
       for(c=0;c<ZetabMatNbDigit(S,R);c++)
	 ZetabAnyEltMultiply(ZetabMatBloc(S,k,c,R),ZetabElt2Int(z,R),ZetabMatBloc(S,k,c,R),R);
       for(l=0;l<k;l++)
	 ___Computation(___SubSComputation);
     }
 else if(P)
   for(;k--;)
     {
       ___Normalization;
       for(l=0;l<k;l++)
	 ___Computation(nop);
     }	 
 return(flag);
}


