/*
 * zen/zetab/c/mat_inverse.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Matrix inversion}

   These functions are general ones. They are directly written with |ZENElt| functions and are used in general cases. See the other directories for specially written functions. 
   }
   */
#include <stdio.h>
#include "zetab.h"
#include "zeg.h"

#ifndef min
#  define min(x,y) ((x)<(y)?(x):(y))
#endif

#define FreeReturn(ret) if(1) {    \
  ZetabMatFree(S,R);                 \
  ZetabMatFree(D,R);                 \
  return(ret);                     \
  } else return(0)



#define ZetabAnyCheckInverse(X,A,R) switch(ZetabAnyEltInverse(X,A,R)) { \
	case ZEN_HAS_INVERSE: break; \
	case ZEN_NO_INVERSE: FreeReturn(ZEN_NO_INVERSE); \
	default: ZENTest(ZENErrErrno == ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_MATINVERSE,ZEN_LIB_ZETAB, ZERR); FreeReturn(ZERR);}

#define ZetabCheckGaussInverse(X,A,R) switch(ZetabMatGaussInverse(X,A,R)) { \
	case ZEN_HAS_INVERSE: break; \
	case ZEN_NO_INVERSE: FreeReturn(ZEN_NO_INVERSE); \
	default: ZENTest(ZENErrErrno == ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_MATINVERSE,ZEN_LIB_ZETAB, ZERR); FreeReturn(ZERR);}


/*
   \TEX{
   \title{Matrix inversion}
   \input{Two |ZENMat| over a |ZENRing|}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{$|I|=|M|^{-1}$. If an inverse was impossible to compute,
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */

int _ZetabMatInverse(I,M,R)
     ZENMat I,M;
     ZENRing R;
{ZENMat D,S;
 int j,k;
 ZENElt n;
 
 ZENTest(ZetabMatIsPermutation(I,R)
	 ||ZetabMatIsPermutation(M,R),
	 ZEN_ISPERM,ZEN_MATINVERSE,ZEN_LIB_ZETAB, ZERR);
 ZENTest((ZetabMatNbBn(I,R) != ZetabMatNbBn(M,R))
	 ||(ZetabMatNbDigit(M,R) != ZetabMatNbBn(M,R))
	 ||(ZetabMatNbDigit(I,R) != ZetabMatNbBn(M,R)),
	 ZEN_MAT_DIM,ZEN_MATINVERSE,ZEN_LIB_ZETAB, ZERR);
 
 if(!(D=ZetabMatCopy(M,R)))
   return(ZERR);
 
 if(!(S=(ZetabMatIsRowType(D,R)? ZetabMatRowCreate(ZetabMatNbBn(D,R),ZetabMatNbBn(D,R),R) : ZetabMatColCreate(ZetabMatNbBn(D,R),ZetabMatNbBn(D,R),R))))
   return(ZERR);
 
 ZENCheckError(ZetabMatSetToOne(S,R), ZERR);
 ZetabCheckGaussInverse(D,S,R);

 if(ZetabMatAreSameType(I,D,R))
   for(k=ZetabMatNbDigit(D,R)-1;k>=0;k--)
     {
       ZetabAnyCheckInverse(ZetabElt2Int(n,R),ZetabMatBloc(D,k,k,R),R);
       for(j=ZetabMatNbBn(D,R)-1;j>=0;j--)
	ZetabAnyEltMultiply(ZetabMatBloc(I,k,j,R),ZetabElt2Int(n,R),ZetabMatBloc(S,k,j,R),R);
     }
 else
   for(k=ZetabMatNbDigit(D,R)-1;k>=0;k--)
     {
       ZetabAnyCheckInverse(ZetabElt2Int(n,R),ZetabMatBloc(D,k,k,R),R);
       for(j=ZetabMatNbBn(D,R)-1;j>=0;j--)
	ZetabAnyEltMultiply(ZetabMatBloc(I,j,k,R),ZetabElt2Int(n,R),ZetabMatBloc(S,k,j,R),R);
	 
     }
#undef ZetabCheckInverse
#undef ZetabCheckGaussInverse

 FreeReturn(ZEN_HAS_INVERSE);
#undef FreeReturn
 
}

#define FreeReturn(ret) if(1) {    \
  ZetabMatFree(P,R);                 \
  ZetabMatFree(D,R);                 \
  return(ret);                     \
  } else return(0)


#define ZetabAnyCheckInverse(X,A,R) switch(ZetabAnyEltInverse(X,A,R)) { \
	case ZEN_HAS_INVERSE: break; \
	case ZEN_NO_INVERSE: FreeReturn(ZEN_NO_INVERSE); \
	default: ZENTest(ZENErrErrno == ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_MATINVERSE,ZEN_LIB_ZETAB, ZERR); FreeReturn(ZERR);}

/*
   \TEX{
   \title{Matrix kernel}
   \input{A pointer on an unallocated |ZENMat| and a |ZENMat| over a |ZENRing|}
   \output{|ZENERR| if an error occurred, |ZEN_NO_INVERSE| if an inverse was impossible to compute, |ZEN_HAS_INVERSE| otherwise}
   \side{|*p_K| is allocated and set to a basis of the kernel of |M|.
   |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */

int _ZetabMatKernel(p_K,M,R)
     ZENMat *p_K,M;
     ZENRing R;
{
  int n,rk,i,j;
  Dim l;
  ZENMat D,P;
  ZENElt z,I,elt;
  
  n=ZetabMatNbCol(M,R);
  
  if(!(D=ZetabMatRowCreate(ZetabMatNbRow(M,R),n,R)))
    return(ZERR);
  
  ZENCheckError(ZetabMatAssign(D,M,R), ZERR);
  
  if(!(P=ZENPermutationColCreate(n,R)))
    return(ZERR);
  
  switch(ZetabMatGaussLazy(D,P,&rk,R))
    {
    case ZEN_HAS_INVERSE:
      break; 
    case ZEN_NO_INVERSE:
      FreeReturn(ZEN_NO_INVERSE);
      
    default:
      ZENTest(ZENErrErrno == ZEN_UNKNOWN,ZEN_UNKNOWN,ZEN_MATKERNEL,ZEN_LIB_ZETAB, ZERR);
      FreeReturn(ZERR);
    }
  
#define dim (n<=rk ? 1 : n-rk)
  
  if(!(*p_K=(ZetabMatIsRowType(M,R) ? ZetabMatRowCreate(dim,n,R) : ZetabMatColCreate(dim,n,R))))
    return(ZERR);
  
  ZetabMatSetToZero(*p_K,R);
  
  for(i=0;i<n-rk;i++)
    for(j=0;j<n;j++)
      {
	l=ZENPermutationBn(P,j,R);
	if(j<rk)
	  {
	    ZetabAnyCheckInverse(ZetabElt2Int(I,R),ZetabMatBloc(D,j,j,R),R);
	    ZetabEltNegate(z,I,R);
	    elt = ZetabMatGetCoeffPtr(D,j,i+rk,R);
	    ZetabEltMultiply(I,z,elt,R);
	    ZetabAnyEltAssign(ZetabMatCoeff(*p_K,i,l,R),ZetabElt2Int(I,R),R);
	  }
	else if (j==rk+i)
	  ZetabAnyEltSetToOne(ZetabMatCoeff(*p_K,i,l,R),R);
      }
  
  FreeReturn(ZEN_HAS_INVERSE);
 
}
