/*
 * zen/zetab/c/poly_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */


/* \TEX{
   \Subsection{Polynomial assignment}
   
   One can set a |ZENPoly| to zero with |_ZetabPolySetToZero|. Then we can
   assign a |ZENPoly| with an other with |_ZetabPolyAssign|. Furthermore, we
   can set randomly a |ZENPoly| with |_ZetabPolySetRandom|. 
   
   The functions |_ZetabPolyGetCoeff| and |_ZetabPolySetCoeff|
   enable us to get or set the coefficient of $X^i$,
   |_ZetabPolyUpdateDegree| compute the real degree of a |ZENPoly| if it
   decreased.
   }
   */
#include <stdio.h>
#include "zetab.h"

/* \TEX{
   \title{Assign.}
   \input{Two |ZENPoly| |PX| and |RX| of a finite ring |Rg|.}
   \side{|RX| is filled with |PX|.}
   }
   */
void _ZetabPolyAssign(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZetabPolyDeg(PX, Rg); i++) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, i, Rg),
		ZetabPolyGetCof(PX, i, Rg),
		Rg);
  }
  ZetabPolyDeg(RX, Rg) =  ZetabPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Set to zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{|RX| is set to zero..}
   }
   */
void _ZetabPolySetToZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  ZetabPolyDeg(RX, Rg) =  0;
  ZetabAnyEltSetToZero(ZetabPolyGetCof(RX, 0, Rg), Rg);
}

/* \TEX{
   \title{Set to $X^i$.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg| allocated at least for
   degree $|deg|$.}
   \side{|RX| is set to $X^{|deg|}$.}
   }
   */
void _ZetabPolySetToXi(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  int i;

  ZetabPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    ZetabAnyEltSetToZero(ZetabPolyGetCof(RX, i, Rg), Rg);
  }
  ZetabAnyEltSetToOne(ZetabPolyGetCof(RX, deg, Rg), Rg);
}

/* \TEX{
   \title{Set to random.}
   \input{A |ZENPoly| |RX| of degree |deg| of a finite ring |Rg|.}
   \side{|RX| is set randomly with degree |deg|.}
   }
   */
void _ZetabPolySetRandom(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  int i;

  ZetabPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    ZetabAnyEltSetRandom(ZetabPolyGetCof(RX, i, Rg), Rg);
  }
  ZetabAnyEltSetRandom(ZetabPolyGetCof(RX, deg, Rg), Rg);
  while(ZetabAnyEltIsZero(ZetabPolyGetCof(RX, deg, Rg), Rg)) {
    ZetabAnyEltSetRandom(ZetabPolyGetCof(RX, deg, Rg), Rg);
  }
}

/* \TEX{
   \title{Set a coeff.}
   \input{A |ZENPoly| |RX| and a |ZENElt| |b| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \side{The coeff of $X^d$ in |RX| is set to |b|.}
   }
   */
void _ZetabPolySetCoeff(RX, d, b, Rg)
     ZENPoly RX;
     int d;
     ZENElt b;
     ZENRing Rg;
{
  ZetabPolySetCoeff(RX, d, b, Rg);
}

/* \TEX{
   \title{Get a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The coeff of $X^d$ in |RX|.}
   }
   */
void _ZetabPolyGetCoeff(p_c, RX, d, Rg)
     ZENElt *p_c;
     ZENPoly RX;
     int d;
     ZENRing Rg;
{ 
  ZetabPolyGetCoeff(*p_c, RX, d, Rg);
}

/* \TEX{
   \title{Extract a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The pointer to the coeff of $X^d$ in |RX|.}
   }
   */
ZENElt _ZetabPolyGetCoeffPtr(RX, d, Rg)
     ZENPoly RX;
     int d;
     ZENRing Rg;
{
  ZENElt elt;

  ZetabElt2Int(elt, Rg) = (BigNumDigit)ZetabPolyGetCof(RX,d,Rg);
  return(elt);
}

/* \TEX{
   \title{Update the degree of a polynomial.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{Resetting the degree of |P|, assuming it has decreased.}
   }
   */
void _ZetabPolyUpdateDegree(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  int i;

  for(i = ZetabPolyDeg(RX, Rg); i >= 0; i--) {
    if (!ZetabAnyEltIsZero(ZetabPolyGetCof(RX,i,Rg), Rg)) {
      break;
    }
  }
  if (i < 0) {
    ZetabPolySetToZero(RX, Rg);
  }
  else {
    ZetabPolyDeg(RX, Rg) =  i;
  }
}
