/*
 * zen/zetab/c/poly_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Polynomial multiplication}
   
   Two functions are provided to do multiplications: |_ZetabPolySquare| to square a
   |ZENPoly| and |_ZetabPolyMultily| to compute the multiplication of two |ZENPoly|.
   }
   */
#include <stdio.h>
#include "zetab.h"
#include "zed.h"

#define ZETAB_POLY_MULTIPLY_CUTOFF 1
#define ZETAB_POLY_SQUARE_CUTOFF 4

/* \TEX{
   \title{Squaring in characteristic 2.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZetabPolySquareChar2(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;
  
  for(i = ZetabPolyDeg(PX, Rg); i; i--) {
    ZetabAnyEltSquare(ZetabPolyGetCof(RX, 2*i, Rg),
		  ZetabPolyGetCof(PX, i, Rg),
		  Rg);
    ZetabAnyEltSetToZero(ZetabPolyGetCof(RX, 2*i-1, Rg), Rg);
  }
  ZetabAnyEltSquare(ZetabPolyGetCof(RX, 0, Rg),
		ZetabPolyGetCof(PX, 0, Rg),
		Rg);
  ZetabPolyDeg(RX, Rg) = 2*ZetabPolyDeg(PX, Rg);
  return(0);
}

/* \TEX{
   \title{Squaring when the characteristic is odd.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZetabPolySquareCharp(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, j;

  for(i = ZetabPolyDeg(PX, Rg)+1; i <= 2*ZetabPolyDeg(PX, Rg); i++) {
    ZetabAnyEltSetToZero(ZetabPolyGetCof(RX, i, Rg), Rg);
  }

  for(i = ZetabPolyDeg(PX, Rg); i > 0; i--) {
    
    ZetabAnyEltSquare(ZetabElt2Int(ZENRingHbuf(Rg), Rg), ZetabPolyGetCof(PX, i, Rg), Rg);
    ZetabAnyEltAdd(ZetabPolyGetCof(RX, 2*i, Rg), ZetabElt2Int(ZENRingHbuf(Rg), Rg), Rg);
    
    for(j = i-1; j > 0; j--) {
      ZetabAnyEltMultiply(ZetabElt2Int(ZENRingHbuf(Rg), Rg),
		      ZetabPolyGetCof(PX, j, Rg),
		      ZetabPolyGetCof(PX, i, Rg),
		      Rg);
      ZetabEltAdd(ZENRingHbuf(Rg), ZENRingHbuf(Rg), Rg);
      ZetabAnyEltAdd(ZetabPolyGetCof(RX, i+j, Rg), ZetabElt2Int(ZENRingHbuf(Rg), Rg), Rg);
    }
    ZetabAnyEltMultiply(ZetabElt2Int(ZENRingHbuf(Rg), Rg),
		    ZetabPolyGetCof(PX, j, Rg),
		    ZetabPolyGetCof(PX, i, Rg),
		    Rg);
    ZetabEltAdd(ZENRingHbuf(Rg), ZENRingHbuf(Rg), Rg);
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, i+j, Rg), ZetabElt2Int(ZENRingHbuf(Rg), Rg), Rg);
  }
  ZetabAnyEltSquare(ZetabPolyGetCof(RX, 0, Rg), ZetabPolyGetCof(PX, 0, Rg), Rg);
    
  ZetabPolyDeg(RX, Rg) = 2*ZetabPolyDeg(PX, Rg);
  return(0);
}

/* \TEX{
   \title{Squaring with Karatsuba.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZetabPolySquareKaratsuba(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{  
  int i;
  ZENPoly RBuf;

  if (!(RBuf = ZetabPolyCreateLarge(4*ZetabPolyDeg(PX, Rg)+1, Rg))) {
    return(ZERR);
  }
  ZetabPolySquareKaratsubaLazy(RBuf, PX, Rg);
  ZetabPolyAssign(RX, RBuf, Rg);
  ZetabPolyFree(RBuf, Rg);
  return(ZEN_HAS_INVERSE);
}

static void ZetabPolySquareKaratsubaLazyDegree(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  struct zenpoly UP1, UP3;
  ZENPoly up1, up3;
  int m, n, i, j;
  
  /* Initialization ***********/
  up1 = &UP1; up3 = &UP3;
  m = (ZetabPolyDeg(PX, Rg)+1)/2; 
  n = ZetabPolyDeg(PX, Rg)+1-m;
  
  /* let's go with the recursivity **********/
  
  /* PX- ^ 2 -> RX */
  ZetabPolyDeg(up1, Rg) = m-1;
  ZetabPolyCof(up1, Rg) = ZetabPolyCof(PX, Rg);
  ZetabPolyCof(up3, Rg) = ZetabPolyCof(RX, Rg);
  ZetabPolySquareKaratsubaLazy(up3, up1, Rg);
  
  /* PX- + PX+ -> RX+4*m+2*n-2 */
  for(j = 0, i = 4*m+2*n-2;
      j < m;
      j++, i++) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, i, Rg),
		  ZetabPolyGetCof(PX, j+m, Rg),
		  Rg);
    ZetabAnyEltAdd(ZetabPolyGetCof(RX, i, Rg), ZetabPolyGetCof(PX, j, Rg), Rg);
  }
  if (n > m) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, i, Rg),
		  ZetabPolyGetCof(PX, j+m, Rg),
		  Rg);
  }
  
  /* (PX- + PX+) ^ 2 -> RX+2*m-1 */
  ZetabPolyDeg(up1, Rg) = n-1;
  ZetabPolyCof(up1, Rg) = ZetabPolyCof(RX, Rg)+4*m+2*n-2;
  ZetabPolyCof(up3, Rg) = ZetabPolyCof(RX, Rg)+2*m-1;
  ZetabPolySquareKaratsubaLazy(up3, up1, Rg);
  
  /* (PX- + PX+) ^ 2 - PX- ^ 2  -> RX+2*m-1 */
  for(j = 0, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZetabAnyEltSubtract(ZetabPolyGetCof(RX, i, Rg), ZetabPolyGetCof(RX, j, Rg), Rg);
  }
  
  /* [ PX- ^ 2, (PX- + PX+) ^ 2 - PX- ^ 2 ] -> RX+m */
  for(j = m, i = 2*m-1; j < 2*m-1; j++, i++) {
    ZetabAnyEltAdd(ZetabPolyGetCof(RX, j, Rg), ZetabPolyGetCof(RX, i, Rg), Rg);
  }
  for(;j < n+2*m-1; j++, i++) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, j, Rg), ZetabPolyGetCof(RX, i, Rg), Rg);
  }
  
  /* PX+ ^ 2 -> RX+ */
  ZetabPolyDeg(up1, Rg) = n-1;
  ZetabPolyCof(up1, Rg) = ZetabPolyCof(PX, Rg)+m;
  ZetabPolyCof(up3, Rg) = ZetabPolyCof(RX, Rg)+2*m+n-1;
  ZetabPolySquareKaratsubaLazy(up3, up1, Rg);

  /* [ PX- ^ 2, (PX- + PX+) ^ 2 - PX- ^2 - PX+ ^ 2 , PX+ ^ 2 ] -> RX */
  for(j = m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZetabAnyEltSubtract(ZetabPolyGetCof(RX, j, Rg), ZetabPolyGetCof(RX, i, Rg), Rg);
  }
  for(j = 2*m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZetabAnyEltAdd(ZetabPolyGetCof(RX, j, Rg), ZetabPolyGetCof(RX, i, Rg), Rg);
  }
  for(; j < 2*m+2*n-1; j++, i++) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, j, Rg), ZetabPolyGetCof(RX, i, Rg), Rg);
  }

  ZetabPolyDeg(RX, Rg) = 2*ZetabPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Lazy Squaring with Karatsuba.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0 for compatibylity.}
   \side{$|RX| = |PX|^2$.}
   \note{One must have $|RX|\neq|PX|$. |PX| is destroyed at the return.}
   }
   */
int ZetabPolySquareKaratsubaLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  if (ZetabPolyDeg(PX, Rg) < ZETAB_POLY_SQUARE_CUTOFF) {
    ZetabPolySquareCharp(RX, PX, Rg);
    return(ZEN_HAS_INVERSE);
  }
  ZetabPolySquareKaratsubaLazyDegree(RX, PX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|$.}
   }
   */
int _ZetabPolyMultiply(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZENPoly RBuf;
  int n, m, q, r, L, k;

  if (ZetabPolyDeg(PX, Rg) > ZetabPolyDeg(QX, Rg)) {
    n = ZetabPolyDeg(PX, Rg)+1; m = ZetabPolyDeg(QX, Rg)+1;
    q = n/m; r = n%m;
  }
  else {
    n = ZetabPolyDeg(QX, Rg)+1; m = ZetabPolyDeg(PX, Rg)+1;
    q = n/m; r = n%m;
  }
  switch(q) {
  case 1:
    L = r ? 4*m-4+2*r : 4*m-3;
    break;
  case 2:
    L = 2*r > (m+1) ? 5*m-5+2*r : 6*m-4;
    break;
  default:
    L = 2*r > m ?  (3+q)*m-4+2*r: (4+q)*m-4;
    break;
  }
  if (!(RBuf = ZetabPolyCreateLarge(L, Rg))) {
    return(ZERR);
  }
  ZetabPolyMultiplyLazy(RBuf, PX, QX, Rg);
  ZetabPolyAssign(RX, RBuf, Rg);
  ZetabPolyFree(RBuf, Rg);

  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Lazy multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0 for compatibylity.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|\neq|QX|$.}
   }
   */
int _ZetabPolyMultiplyLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  struct zenpoly sxzp, rxzp, pxzp, qxzp;
  ZENPoly sx, rx, px, qx;
  int i, dp, dq, quo, rem;

  /* Switch
  **********/
  if (ZetabPolyDeg(PX, Rg) == ZetabPolyDeg(QX, Rg)) {
    ZetabPolyMultiplyKaratsubaLazy(RX, PX, QX, Rg);
    return(ZEN_HAS_INVERSE);
  }
  if (ZetabPolyDeg(PX, Rg) > ZetabPolyDeg(QX, Rg)) {
    if ((ZetabPolyDeg(QX, Rg)) <= ZETAB_POLY_MULTIPLY_CUTOFF) {
      ZetabPolyMultiplyPlain(RX, PX, QX, Rg);
      return(ZEN_HAS_INVERSE);
    }
    px = &pxzp;
    ZetabPolyCof(px, Rg) = ZetabPolyCof(PX, Rg);
    dp = ZetabPolyDeg(PX, Rg);

    qx = &qxzp;
    ZetabPolyCof(qx, Rg) = ZetabPolyCof(QX, Rg);
    dq = ZetabPolyDeg(QX, Rg);
  }
  else {
    if (ZetabPolyDeg(PX, Rg) <= ZETAB_POLY_MULTIPLY_CUTOFF) {
      ZetabPolyMultiplyPlain(RX, PX, QX, Rg);
      return(ZEN_HAS_INVERSE);
    }
    px = &pxzp;
    ZetabPolyCof(px, Rg) = ZetabPolyCof(QX, Rg);
    dp = ZetabPolyDeg(QX, Rg);

    qx = &qxzp;
    ZetabPolyCof(qx, Rg) = ZetabPolyCof(PX, Rg);
    dq = ZetabPolyDeg(PX, Rg);
  }

  /* We have to cut px in pieces 
    **********************************/  
  /* Initialization */
  ZetabPolyDeg(RX, Rg) = dp+dq;

  rx = &rxzp;
  ZetabPolyCof(rx, Rg) = ZetabPolyCof(RX, Rg);

  sx = &sxzp;
  ZetabPolyCof(sx, Rg) = ZetabPolyCof(RX, Rg);

  quo = (dp+1)/(dq+1);
  rem = (dp+1)%(dq+1);

  /* First multiplication */
  ZetabPolyDeg(px, Rg) = dq;
  ZetabPolyDeg(qx, Rg) = dq;
  ZetabPolyMultiplyKaratsubaLazy(rx, qx, px, Rg);
  if ((quo == 1) && !rem) {
    return(ZEN_HAS_INVERSE);
  }

  /* Second multiplication */
  ZetabPolyCof(px, Rg) += dq+1;
  ZetabPolyDeg(px, Rg) = (quo == 1 ? rem-1: dq);
  ZetabPolyCof(rx, Rg) += 2*dq+1;
  if (quo == 1) {
    ZetabPolyMultiplyLazy(rx, qx, px, Rg);
  }
  else {
    ZetabPolyMultiplyKaratsubaLazy(rx, qx, px, Rg);
  }
  ZetabPolyDeg(sx, Rg) = dq-1;
  ZetabPolyCof(sx, Rg) += dq+1;
  ZetabPolyAdd(sx, rx, Rg);
  if (quo == 1) {
    return(ZEN_HAS_INVERSE);
  }

  for(i = 2; i < quo; i++) {
    ZetabPolyCof(px, Rg) += (dq+1);
    ZetabPolyCof(rx, Rg) += (dq+1);

    ZetabPolyMultiplyKaratsubaLazy(rx, qx, px, Rg);
    
    ZetabPolyDeg(sx, Rg) = dq-1;
    ZetabPolyCof(sx, Rg) += dq+1;
    ZetabPolyAdd(sx, rx, Rg);
  }

  if (!rem) {
    return(ZEN_HAS_INVERSE);
  }

  ZetabPolyDeg(px, Rg) = rem-1;
  ZetabPolyCof(px, Rg) += (dq+1);
  ZetabPolyCof(rx, Rg) += (dq+1);
  ZetabPolyMultiplyLazy(rx, qx, px, Rg);
    
  ZetabPolyDeg(sx, Rg) = dq-1;
  ZetabPolyCof(sx, Rg) += dq+1;
  ZetabPolyAdd(sx, rx, Rg);

  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|$.}
   }
   */
int ZetabPolyMultiplyPlain(RX, QX, PX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{

  int i, j;

  if (ZetabPolyIsZero(PX, Rg) || ZetabPolyIsZero(QX, Rg)) {
    ZetabPolySetToZero(RX, Rg);
    return(0);
  }

  for(i = ZetabPolyDeg(PX, Rg)+1;
      i <= ZetabPolyDeg(PX, Rg)+ZetabPolyDeg(QX, Rg);
      i++) {
    ZetabAnyEltSetToZero(ZetabPolyGetCof(RX, i, Rg), Rg);
  }

  for(i = ZetabPolyDeg(PX, Rg); i >= 0; i--) {
    for(j = ZetabPolyDeg(QX, Rg); j; j--) {
      ZetabAnyEltMultiply(ZetabElt2Int(ZENRingHbuf(Rg), Rg),
		      ZetabPolyGetCof(QX, j, Rg),
		      ZetabPolyGetCof(PX, i, Rg),
		      Rg);
      ZetabAnyEltAdd(ZetabPolyGetCof(RX, i+j, Rg), ZetabElt2Int(ZENRingHbuf(Rg), Rg), Rg);
    }
    ZetabAnyEltMultiply(ZetabElt2Int(ZENRingHbuf(Rg), Rg),
		    ZetabPolyGetCof(QX, j, Rg),
		    ZetabPolyGetCof(PX, i, Rg),
		    Rg);
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, i+j, Rg), ZetabElt2Int(ZENRingHbuf(Rg), Rg), Rg);
  }
  ZetabPolyDeg(RX, Rg) = ZetabPolyDeg(PX, Rg) + ZetabPolyDeg(QX, Rg);
  return(0);
}

/* \TEX{
   \title{Polynomial multiplication with the Karatsuba method.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an errpr occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|$. If $\deg(|PX|) \ge deg(|QX|)$,
   one polynomial buffer of size |PX| is allocated, otherwise 2
   buffers of size |QX| are allocated.}
   }
   */
int _ZetabPolyMultiplyKaratsuba(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i;
  ZENPoly QBuf, RBuf;
  ZENElt elt;

  if (ZetabPolyDeg(PX, Rg) >= ZetabPolyDeg(QX, Rg)) {
    if (!(QBuf = ZetabPolyCreate(ZetabPolyDeg(PX, Rg), Rg))) {
      return(ZERR);
    }
    ZetabPolyAssign(QBuf, QX, Rg);
    for(i = ZetabPolyDeg(QX, Rg)+1;
	i <= ZetabPolyDeg(PX, Rg);
	i++) {
      elt = ZetabPolyGetCoeffPtr(QBuf, i, Rg);
      ZetabEltSetToZero(elt, Rg);
      ZetabPolySetCoeff(QBuf, i, elt, Rg);
    }

    if (!(RBuf = ZetabPolyCreateLarge(4*ZetabPolyDeg(PX, Rg)+1, Rg))) {
      return(ZERR);
    }

    ZetabPolyMultiplyKaratsubaLazy(RBuf, PX, QBuf, Rg);
  }
  else {
    if (!(QBuf = ZetabPolyCreate(ZetabPolyDeg(QX, Rg), Rg))) {
      return(ZERR);
    }
    ZetabPolyAssign(QBuf, PX, Rg);
    for(i = ZetabPolyDeg(PX, Rg)+1;
	i <= ZetabPolyDeg(QX, Rg);
	i++) {
      elt = ZetabPolyGetCoeffPtr(QBuf, i, Rg);
      ZetabEltSetToZero(elt, Rg);
      ZetabPolySetCoeff(QBuf, i, elt, Rg);
    }

    if (!(RBuf = ZetabPolyCreateLarge(4*ZetabPolyDeg(PX, Rg)+1, Rg))) {
      return(ZERR);
    }

    ZetabPolyMultiplyKaratsubaLazy(RBuf, QX, QBuf, Rg);
  }
  ZetabPolyAssign(RX, RBuf, Rg);
  ZetabPolyFree(QBuf, Rg); ZetabPolyFree(RBuf, Rg);
  return(ZEN_HAS_INVERSE);
}


static void ZetabPolyMultiplyKaratsubaLazyDegree0(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZetabAnyEltMultiply(ZetabPolyGetCof(RX, 0, Rg),
		   ZetabPolyGetCof(PX, 0, Rg),
		   ZetabPolyGetCof(QX, 0, Rg), Rg);
  ZetabPolyDeg(RX, Rg) = 0;
}

static void ZetabPolyMultiplyKaratsubaLazyDegree1(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZetabAnyEltAssign(ZetabPolyGetCof(RX, 0, Rg), ZetabPolyGetCof(PX, 0, Rg), Rg);
  ZetabAnyEltAdd(ZetabPolyGetCof(RX, 0, Rg), ZetabPolyGetCof(PX, 1, Rg), Rg);
  
  ZetabAnyEltAssign(ZetabPolyGetCof(RX, 2, Rg), ZetabPolyGetCof(QX, 0, Rg), Rg);
  ZetabAnyEltAdd(ZetabPolyGetCof(RX, 2, Rg), ZetabPolyGetCof(QX, 1, Rg), Rg);

  ZetabAnyEltMultiply(ZetabPolyGetCof(RX, 1, Rg),
		   ZetabPolyGetCof(RX, 0, Rg),
		   ZetabPolyGetCof(RX, 2, Rg), Rg);
  
  ZetabAnyEltMultiply(ZetabPolyGetCof(RX, 0, Rg),
		   ZetabPolyGetCof(PX, 0, Rg),
		   ZetabPolyGetCof(QX, 0, Rg), Rg);
  
  ZetabAnyEltMultiply(ZetabPolyGetCof(RX, 2, Rg),
		   ZetabPolyGetCof(PX, 1, Rg),
		   ZetabPolyGetCof(QX, 1, Rg), Rg);
  
  ZetabAnyEltSubtract(ZetabPolyGetCof(RX, 1, Rg), ZetabPolyGetCof(RX, 0, Rg), Rg);
  ZetabAnyEltSubtract(ZetabPolyGetCof(RX, 1, Rg), ZetabPolyGetCof(RX, 2, Rg), Rg);

  ZetabPolyDeg(RX, Rg) = 2;
}

static void ZetabPolyMultiplyKaratsubaLazyHeart ___((ZENPoly, ZENPoly, ZENPoly, ZENRing));

static void ZetabPolyMultiplyKaratsubaLazyDegree(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  struct zenpoly UP1, UP2, UP3;
  ZENPoly up1, up2, up3;
  int m, n, i, j;
  
  /* Initialization ***********/
  up1 = &UP1; up2 = &UP2; up3 = &UP3;
  up2->deg = 0;
  m = (ZetabPolyDeg(PX, Rg)+1)/2; 
  n = ZetabPolyDeg(PX, Rg)+1-m;
  
  /* let's go with the recursivity **********/
  
  /* PX- * QX- -> RX */
  ZetabPolyDeg(up1, Rg) = m-1;
  ZetabPolyCof(up1, Rg) = ZetabPolyCof(PX, Rg);
  ZetabPolyCof(up2, Rg) = ZetabPolyCof(QX, Rg);
  ZetabPolyCof(up3, Rg) = ZetabPolyCof(RX, Rg);
  ZetabPolyMultiplyKaratsubaLazyHeart(up3, up1, up2, Rg);
  
  /* PX- + PX+ -> RX+4*m+2*n-2 */
  for(j = 0, i = 4*m+2*n-2;
      j < m;
      j++, i++) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, i, Rg),
		  ZetabPolyGetCof(PX, j+m, Rg),
		  Rg);
    ZetabAnyEltAdd(ZetabPolyGetCof(RX, i, Rg), ZetabPolyGetCof(PX, j, Rg), Rg);
  }
  if (n > m) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, i, Rg),
		  ZetabPolyGetCof(PX, j+m, Rg),
		  Rg);
  }
  
  /* QX- + QX+ -> RX+4*m+3*n-2 */
  for(j = 0, i = 4*m+3*n-2;
      j < m;
      j++, i++) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, i, Rg),
		  ZetabPolyGetCof(QX, j+m, Rg),
		  Rg);
    ZetabAnyEltAdd(ZetabPolyGetCof(RX, i, Rg), ZetabPolyGetCof(QX, j, Rg), Rg);
  }
  if (n > m) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, i, Rg),
		  ZetabPolyGetCof(QX, j+m, Rg),
		  Rg);
  }
  
  /* (PX- + PX+) * (QX- + QX+) -> RX+2*m-1 */
  ZetabPolyDeg(up1, Rg) = n-1;
  ZetabPolyCof(up1, Rg) = ZetabPolyCof(RX, Rg)+4*m+2*n-2;
  ZetabPolyCof(up2, Rg) = ZetabPolyCof(RX, Rg)+4*m+3*n-2;
  ZetabPolyCof(up3, Rg) = ZetabPolyCof(RX, Rg)+2*m-1;
  ZetabPolyMultiplyKaratsubaLazyHeart(up3, up1, up2, Rg);
  
  /* (PX- + PX+) * (QX- + QX+) - PX- * QX-  -> RX+2*m-1 */
  for(j = 0, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZetabAnyEltSubtract(ZetabPolyGetCof(RX, i, Rg), ZetabPolyGetCof(RX, j, Rg), Rg);
  }
  
  /* [ PX- * QX-, (PX- + PX+) * (QX- + QX+) - PX- * QX- ] -> RX+m */
  for(j = m, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZetabAnyEltAdd(ZetabPolyGetCof(RX, j, Rg), ZetabPolyGetCof(RX, i, Rg), Rg);
  }
  for(;j < n+2*m-1; j++, i++) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, j, Rg), ZetabPolyGetCof(RX, i, Rg), Rg);
  }
  
  /* PX+ * QX+ -> RX+ */
  ZetabPolyDeg(up1, Rg) = n-1;
  ZetabPolyCof(up1, Rg) = ZetabPolyCof(PX, Rg)+m;
  ZetabPolyCof(up2, Rg) = ZetabPolyCof(QX, Rg)+m;
  ZetabPolyCof(up3, Rg) = ZetabPolyCof(RX, Rg)+2*m+n-1;
  ZetabPolyMultiplyKaratsubaLazyHeart(up3, up1, up2, Rg);

  /* [ PX- * QX-, (PX- + PX+) * (QX- + QX+) - PX- * QX- - PX+ * QX+ , PX+ * QX+ ] -> RX */
  for(j = m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZetabAnyEltSubtract(ZetabPolyGetCof(RX, j, Rg), ZetabPolyGetCof(RX, i, Rg), Rg);
  }
  for(j = 2*m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZetabAnyEltAdd(ZetabPolyGetCof(RX, j, Rg), ZetabPolyGetCof(RX, i, Rg), Rg);
  }
  for(; j < 2*m+2*n-1; j++, i++) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, j, Rg), ZetabPolyGetCof(RX, i, Rg), Rg);
  }

  ZetabPolyDeg(RX, Rg) = ZetabPolyDeg(PX, Rg)+ZetabPolyDeg(QX, Rg);
}

static void ZetabPolyMultiplyKaratsubaLazyHeart(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  switch (ZetabPolyDeg(PX, Rg)) {

    /* compute the product of 2 polynoms whose degree is 0 */
  case 0:
    ZetabPolyMultiplyKaratsubaLazyDegree0(RX, PX, QX, Rg);
    break;

    /* compute the product of 2 polynoms whose degree is 1 */
  case 1: 
    ZetabPolyMultiplyKaratsubaLazyDegree1(RX, PX, QX, Rg); 
    break;

    /* compute the product of 2 polynoms whose degree is greater than 2*/
  default : 
    ZetabPolyMultiplyKaratsubaLazyDegree(RX, PX, QX, Rg); 
  }
}

int _ZetabPolyMultiplyKaratsubaLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  if (ZetabPolyDeg(PX, Rg) >= ZetabPolyDeg(QX, Rg)) {
    for(i = ZetabPolyDeg(QX, Rg)+1;
	i <= ZetabPolyDeg(PX, Rg);
	i++) {
      elt = ZetabPolyGetCoeffPtr(QX, i, Rg);
      ZetabEltSetToZero(elt, Rg);
      ZetabPolySetCoeff(QX, i, elt, Rg);
    }
  }
  else {
    for(i = ZetabPolyDeg(PX, Rg)+1;
	i <= ZetabPolyDeg(QX, Rg);
	i++) {
      elt = ZetabPolyGetCoeffPtr(PX, i, Rg);
      ZetabEltSetToZero(elt, Rg);
      ZetabPolySetCoeff(PX, i, elt, Rg);
    }
  }
  ZetabPolyMultiplyKaratsubaLazyHeart(RX, PX, QX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Multiplication by a scalar.}
   \input{Two polynomials |PX|, |RX| and a |ZENElt| |e| of a finite ring |Rg|.}
   }
   */
void _ZetabPolyMultiplyScalar(RX, PX, e, Rg)
     ZENPoly RX, PX;
     ZENElt e;
     ZENRing Rg;
{
  int i;
  
  if (ZetabEltIsZero(e, Rg)) {
    ZetabPolySetToZero(RX, Rg);
    return;
  }
  if (ZetabEltIsOne(e, Rg)) {
    ZetabPolyAssign(RX, PX, Rg);
    return;
  }
  for(i = 0; i <= ZetabPolyDeg(PX, Rg); i++) {
    ZetabAnyEltMultiply(ZetabPolyGetCof(RX, i, Rg),
		    ZetabElt2Int(e, Rg),
		    ZetabPolyGetCof(PX, i, Rg),
		    Rg);
  }
  ZENPolyDeg(RX, Rg) = ZENPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Scalar product.}
   \input{Two polynomials |PX|, |RX| and a pointer to a |ZENElt| |e| of a
   finite ring |Rg|.}
   \side{$|*e| = \sum p_iq_i$}
   }
   */
void _ZetabPolyDot(e, PX, QX, Rg)
     ZENElt *e;
     ZENPoly PX, QX;
     ZENRing Rg;
{
  int i, deg;
  
  ZetabAnyEltMultiply(ZetabElt2Int(*e, Rg),
		  ZetabPolyGetCof(PX, 0, Rg),
		  ZetabPolyGetCof(QX, 0, Rg),
		  Rg);

  deg = ZetabPolyDeg(PX, Rg) < ZetabPolyDeg(QX, Rg) ?
    ZetabPolyDeg(PX, Rg) :  ZetabPolyDeg(QX, Rg);

  for(i = 1; i <= deg; i++) {
    ZetabAnyEltMultiply(ZetabElt2Int(ZENRingHbuf(Rg), Rg),
		    ZetabPolyGetCof(PX, i, Rg),
		    ZetabPolyGetCof(QX, i, Rg),
		    Rg);
    ZetabEltAdd(*e, ZENRingHbuf(Rg), Rg);
  }
}

/* \title{Divide.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured,  |ZENRingFact(Rg)| is
   filled with a factor of a modulo if |ZEN_NO_INVERSE| is returned.}
   \note{|PX| is destroyed.}
   */
int _ZetabPolyDivideLazy(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int i, j, deg, IsNotOne;

  if (!ZetabAnyEltIsOne(ZetabPolyGetCof(QX, ZetabPolyDeg(QX, Rg), Rg), Rg)) {
    IsNotOne = 1;
    
    if (ZetabAnyEltInverse(ZetabElt2Int(ZENRingIbuf(Rg), Rg),
		       ZetabPolyGetCof(QX, ZetabPolyDeg(QX, Rg), Rg),
		       Rg) == ZEN_NO_INVERSE) {
      return(ZEN_NO_INVERSE);
    }
  }
  else {
    IsNotOne = 0;
  }
  
  for(i = ZetabPolyDeg(PX, Rg);
      i >= ZetabPolyDeg(QX, Rg);
      i--) {
    if (IsNotOne) {
      ZetabAnyEltMultiply(ZetabPolyGetCof(PX, i, Rg),
		      ZetabElt2Int(ZENRingIbuf(Rg), Rg), ZetabPolyGetCof(PX, i, Rg),
		      Rg);
    }
    
    for(j =  ZetabPolyDeg(QX, Rg)-1;
	j >= 0;
	j--) {
      ZetabAnyEltMultiply(ZetabElt2Int(ZENRingHbuf(Rg), Rg),
		      ZetabPolyGetCof(PX, i, Rg),
		      ZetabPolyGetCof(QX, j, Rg),
		      Rg);
      ZetabAnyEltSubtract(ZetabPolyGetCof(PX,
				     i+j-ZetabPolyDeg(QX, Rg),
				     Rg), ZetabElt2Int(ZENRingHbuf(Rg), Rg), Rg);
    }
  }
  deg =  ZetabPolyDeg(QX, Rg);
  for(i = ZetabPolyDeg(PX, Rg); i >= ZetabPolyDeg(QX, Rg); i--) {
    ZetabAnyEltAssign(ZetabPolyGetCof(RX, i-ZetabPolyDeg(QX, Rg), Rg),
		   ZetabPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZetabPolyDeg(RX, Rg) = ZetabPolyDeg(PX, Rg)-ZetabPolyDeg(QX, Rg);
  if (ZetabPolyDeg(RX, Rg) < 0) {
    ZetabPolySetToZero(RX, Rg);
  }
  for(; i >= 0; i--) {
    if (!ZetabAnyEltIsZero(ZetabPolyGetCof(PX, i, Rg), Rg)) {
      break;
    }
  }
  for(deg = i; i >= 0; i--) {
    ZetabAnyEltAssign(ZetabPolyGetCof(MX, i, Rg),
		   ZetabPolyGetCof(PX, i, Rg),
		   Rg);
  }
  if (deg < 0) {
    ZetabPolySetToZero(MX, Rg);
  }
  else {
    ZetabPolyDeg(MX, Rg) = deg;
  }
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Divide.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZEN_NO_INVERSE| if
   a factor of the modulo was dicovered, |ZENERR| for an error.}
   \side{ |ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZetabPolyDivide(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int ret;
  ZENPoly Pbuf;
  
  if (!(Pbuf = ZetabPolyCopy(PX, Rg))) {
    return(ZERR);
  }
  ret = ZetabPolyDivideLazy(RX, MX, Pbuf, QX, Rg);
  ZetabPolyFree(Pbuf, Rg);
  return(ret);
}


/* \TEX{
   \title{Lazy mudulo computation.}
   \input{A polynomial |PX| to divide by
   |QX|$=X^{\mbox{degree}(QX)}+sum_{i=0}^{|k|}q_iX^i$} in a finite
   ring |Rg|. |MX| will be the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| for an error}
   \note{|PX| is destroyed.}
   */
int _ZetabPolyModuloLazy(MX, PX, QX, k, Rg)
     ZENPoly MX, PX, QX;
     int k;
     ZENRing Rg;
{
  int i, j, deg;
  
  for(i = ZetabPolyDeg(PX, Rg);
      i >= ZetabPolyDeg(QX, Rg);
      i--) {
    for(j =  k;
	j >= 0;
	j--) {
      ZetabAnyEltMultiply(ZetabElt2Int(ZENRingHbuf(Rg), Rg),
		      ZetabPolyGetCof(PX, i, Rg),
		      ZetabPolyGetCof(QX, j, Rg),
		      Rg);
      ZetabAnyEltSubtract(ZetabPolyGetCof(PX,
				     i+j-ZetabPolyDeg(QX, Rg),
				     Rg), ZetabElt2Int(ZENRingHbuf(Rg), Rg), Rg);
    }
  }
  for( ; i >= 0; i--) {
    if (!ZetabAnyEltIsZero(ZetabPolyGetCof(PX, i, Rg), Rg)) {
      break;
    }
  }
  for(deg = i; i >= 0; i--) {
    ZetabPolyGetCof(MX, i, Rg) = ZetabPolyGetCof(PX, i, Rg);
  }
  if (deg < 0) {
    ZetabPolySetToZero(MX, Rg);
  }
  else {
    ZetabPolyDeg(MX, Rg) = deg;
  }

  return(ZEN_HAS_INVERSE);
}
