/*
 * zen/zext/c/elt_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Element assignment}
   
   One can set a |ZENElt| to zero or one with |_ZextEltSetToZero| or
   |_ZextEltSetToOne|. Then we can assign a |ZENElt| with an other with
   |_ZextEltAssign|. Furthermore, we can set randomly a |ZENElt| with
   |_ZextEltSetRandom| or enumerate the elements of a finite ring with
   |_ZextEltSetNext|.
   }
   */
#include <stdio.h>
#include "zext.h"

extern int _ZENRingAreEqual ___((ZENRing, ZENRing));

/* \TEX{
   \title{Assign.}
   \input{A pointer |p_a| on a |ZENElt| and a |ZENElt| |b| of a finite ring |Rg|.}
   \side{|*p_a| is filled with |b|.}
   }
   */
void _ZextEltAssign(p_a, b, Rg)
     ZENElt *p_a, b;
     ZENRing Rg;
{
  ZextEltAssign(*p_a, b, Rg);
}

/* \TEX{
   \title{Set to zero.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set to zero..}
   }
   */
void _ZextEltSetToZero(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZextEltSetToZero(*p_a, Rg);
}

/* \TEX{
   \title{Set to one.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set to one.}
   }
   */
void _ZextEltSetToOne(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZextEltSetToOne(*p_a, Rg);
}

/* \TEX{
   \title{Set to random.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set randomly.}
   }
   */
void _ZextEltSetRandom(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  ZextEltSetRandom(*p_a, Rg);
}

/* \TEX{
   \title{Enumerate.}
   \input{A pointer |p_a| on a |ZENElt| of a finite ring |Rg|.}
   \side{|*p_a| is set randomly.}
   \output{|ZENERR| if an error occurred, 0 otherwise.}
   \note{After a number of call of this function equal to the number of elements in the finite ring, we obtain the same element.}
   }
   */
void _ZextEltSetNext(p_a, Rg)
     ZENElt *p_a;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  for(i = 0;
      i <= ZENPolyDeg(ZextElt2Pol(*p_a, Rg), ZENRingDef(Rg));
      i++) {
    
    elt = ZENPolyGetCoeffPtr(ZextElt2Pol(*p_a, Rg), i, ZENRingDef(Rg));
    ZENEltSetNext(elt, ZENRingDef(Rg));
    ZENPolySetCoeff(ZextElt2Pol(*p_a, Rg), i, elt, ZENRingDef(Rg));
    
    if (!ZENEltIsZero(elt, ZENRingDef(Rg))) {
      break;
    }
  }
  if (i > ZENPolyDeg(ZextElt2Pol(*p_a, Rg), ZENRingDef(Rg))) {
    if (i == ZENRingDeg(Rg)) {
      ZENPolySetToZero(ZextElt2Pol(*p_a, Rg), ZENRingDef(Rg));
    }
    else {
      ZENPolySetToXi(ZextElt2Pol(*p_a, Rg), i, ZENRingDef(Rg));
    }
  }
}

/* \TEX{
   \title{Conversion.}
   \input{A pointer |pa| on a |ZENElt| of a modular ring |Rg|, an element |b| of a modular ring |R|.}
   \side{|*pa| is set to the modular conversion of |b|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   }
   */
int _ZextEltConvert(pa, Rg, b, R)
     ZENElt *pa,b;
     ZENRing Rg,R;
{  
  switch(ZENRingArith(R)) {
    
  /* Fractions
   *************/
  case ZEN_LIB_ZEF:
    {
      BigNum p;
      BigNumLength pl;

      pl = ZENRingSizeElt(Rg);

      if (!(p = ZBNC(pl))) {
	ZENSetError(ZEN_LIB_ZEXT, ZEN_CONVERT, ZEN_ALLOC);
	return(ZERR);
      }
      ZENEltToZ(p, &pl, b, R);
      
      if ((ZENErrErrno == ZEN_INPUT) && (pl == ZENRingSizeElt(Rg))) {
	
	/* This element is a fraction, nothing can be done */
	ZENSetError(ZEN_LIB_ZEXT, ZEN_CONVERT, ZEN_INPUT);
	return(ZERR);
      }
      
      ZBNF(p);
      if (!(p = ZBNC(pl+1))) {
	ZENSetError(ZEN_LIB_ZEXT, ZEN_CONVERT, ZEN_ALLOC);
	return(ZERR);
      }
      ZENEltToZ(p, &pl, b, R);
      switch(ZBNCompareLazy(p, pl,
			    ZENRingP(Rg), ZENRingPl(Rg))) {
      case BN_EQ:
	ZextEltSetToZero(*pa, Rg);
	break;
      case BN_LT:
	ZextZToElt(*pa, p, pl, Rg);
	break;
      default:
	p[pl] = 0;
	ZBNDivide(p, pl+1, ZENRingP(Rg), ZENRingPl(Rg));  
	pl = ZBNNumDigits(p, ZENRingPl(Rg));
	ZextZToElt(*pa, p, pl, Rg);
      }
      ZBNF(p);
    }
    break;
    
    /* Specific cases where things can be done faster
     *************************************************/
  case ZEN_LIB_ZE2:
    if(ZENEltIsZero(b, R))
      ZextEltSetToZero(*pa, Rg);
    else
      ZextEltSetToOne(*pa, Rg);
    return(ZEN_HAS_INVERSE);
    
    /* Default
     **********/
  default:
    {
      BigNumLength pl;
      BigNum p;
      ZENElt c;
      ZENRing r, rg;
      int Rh, Rgh, i;
      
      /* height of Rg */
      Rgh = 1;
      for(rg = ZENRingDef(Rg); rg; rg  = ZENRingDef(rg)) {
	Rgh++;
      }

      /* height of R */
      Rh = 1;
      for(r = ZENRingDef(R); r; r  = ZENRingDef(r)) {
	Rh++;
      }

      if (Rh > Rgh) {
	
	r = R;
	for(; Rh > Rgh; Rh--) {
	  r = ZENRingDef(R);
	}
	pl = ZENRingQl(R);
	if (!(p = ZBNC(pl+1))) {
	  ZENSetError(ZEN_LIB_ZE2,ZEN_CONVERT,ZEN_ALLOC);
	  return(ZERR);
	}
	ZENEltToZ(p, &pl, b, R);
	switch(ZBNCompareLazy(p, pl,
			      ZENRingQ(r), ZENRingQl(r))) {
	case BN_EQ:
	  p[0] = 0; pl = 1;
	  break;
	case BN_GT:
	  p[pl] = 0;
	  ZBNDivide(p, pl+1, ZENRingQ(r), ZENRingQl(r));
	  pl = ZBNNumDigits(p, ZENRingQl(r));
	  break;
	default:
	  break;
	}
	if (ZENEltCreate(c, r)) {
	  ZBNF(p);
	  return(ZERR);
	}
	ZENZToElt(c, p, pl, r);
	ZBNF(p);
	if (ZextEltConvert(*pa, Rg, c, r)) {
	  ZENEltFree(c, R);
	  return(ZERR);
	}
	ZENEltFree(c, r);
	break;
      }
      
      if (Rgh > Rh) {
	rg = Rg;
	for(; Rgh > Rh; Rgh--) {
	  rg = ZENRingDef(rg);
	}
	if (ZENEltCreate(c, rg)) {
	  return(ZERR);
	}
	if (!(p = ZBNC(ZENRingQl(rg)))) {
	  ZENEltFree(c, rg);
	  return(ZERR);
	}
	ZENEltConvert(c, rg, b, R);
	ZENEltToZ(p, &pl, c, rg);
	ZENZToElt(*pa, p, pl, Rg);
	ZENEltFree(c, rg); ZBNF(p);
	break;
      }
      
      if (Rgh == Rh) {
	ZENPoly P, Pg;
	ZENElt e, eg;

	r = ZENRingDef(R); rg = ZENRingDef(Rg); 
	pl = ZENRingQl(R);
	if(!(p = ZBNC(pl))) {
	  ZENSetError(ZEN_LIB_ZEXT,ZEN_CONVERT,ZEN_ALLOC);
	  return(ZERR);
	}
	ZENEltToZ(p, &pl, b, R);  
	P = ZENZToPoly(p,pl,r);
	ZBNF(p);
	if(P==NULL)
	  return(ZERR);
	if (!(Pg = ZENPolyCreate(ZENPolyDeg(P, r), rg))) {
	  ZENSetError(ZEN_LIB_ZEXT,ZEN_CONVERT,ZEN_ALLOC);
	  ZENPolyFree(P, r);
	  return(ZERR);
	}
	for(i = 0; i <= ZENPolyDeg(P, r); i++) {
	  e = ZENPolyGetCoeffPtr(P, i, r);
	  eg = ZENPolyGetCoeffPtr(Pg, i, rg);
	  if (ZENEltConvert(eg, rg, e, r)) {
	    ZENPolyFree(P, r); ZENPolyFree(Pg, rg);
	    return(ZERR);
	  }
	  ZENPolySetCoeff(Pg, i, eg, rg);
	}
	ZENPolyDeg(Pg, rg) = ZENPolyDeg(P, r);
	ZENPolyUpdateDegree(Pg, rg);
	if(ZENPolyDeg(Pg, rg) < ZENRingDeg(Rg)) {
	  ZENPolyAssign(ZENElt2Pol(*pa,ZENRingDef(Rg)), Pg, rg);
	  ZENPolyFree(Pg, rg); ZENPolyFree(P, r);
	}
	else {
	  int ret;
	  
	  ret = ZENPolyDivide(P, ZENElt2Pol(*pa,Rg),
			      Pg, ZENRingPol(Rg), rg);
	  ZENPolyFree(Pg, rg); ZENPolyFree(P, r);
	  if(ret)
	    return(ZERR);
	}
      }
      break;
    }
    
    /* Otherwise, what can we do ? */
    ZENSetError(ZEN_LIB_ZEXT,ZEN_CONVERT,ZEN_INPUT);
    return(ZERR);
  }
  return(ZEN_HAS_INVERSE);
}
