/*
 * zen/zext/c/elt_io.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{input/output}
   
   Conversions between the internal representation and the ASCII
   representation of a |ZENElt| are done by |_ZextEltPrintToString|. |_ZextEltPrintToFile|
   allows to write a |ZENElt| in a file.
   One can moreover write to and then read from a file a |ZENelt| with |_ZextEltPutToFile|
   and |_ZextEltReadFromFile|.
   }
   */
#include <stdio.h>
#include "zext.h"
#include "zeg.h"

/* \TEX{
   \title{Converting from string.}
   \input{A finite field |Rg|, a pointer on a |ZENElt| |G| and a string
   |s| representing a |ZENElt| in base $|base|\in \{2,\ldots,
   16\}$.} 
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|G| is filled with |s| if the output is not |ZENERR|.}
   }
*/
int _ZextEltReadFromString(G, s, base, Rg)
     ZENElt *G;
     char *s;
     int base;
     ZENRing Rg;
{
  int ret;
  ZENPoly PX;

  if ((ret = _ZENPolyReadFromString(&PX, s, ZENRingVar(Rg), base, ZENRingDef(Rg))) != ZENERR) {
    ZENPolyAssign(ZextElt2Pol(*G, Rg), PX, ZENRingDef(Rg));
    ZENPolyFree(PX,  ZENRingDef(Rg));
  }
  return(ret);
}

/* \TEX{
   \title{Converting to string.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{An allocated string representing |G| in base |base| or |ZENNULL| if an error occured.}
   }
   */
char *_ZextEltPrintToString(G, base, Rg)
     ZENElt G;
     int base;
     ZENRing Rg;  
{
  return(ZextEltPrintToString(G, base, Rg));
}


/* \TEX{
   \title{Converting from a string in an internal representation.}
   \input{A finite field |Rg| and a string |s| representing a |ZENElt|
   in an internal representation.}
   \output{The number of character read in |s| or |ZENERR| if an error occured.}
   \side{|G| is filled with s.}
   }
   */
int _ZextEltGetFromString(G, s, Rg)
     char *s;
     ZENElt *G;
     ZENRing Rg;
{
  int ret;
  ZENPoly PX;

  if ((ret = ZENPolyGetFromString(PX, s, ZENRingDef(Rg))) != ZENERR) {
    ZENPolyAssign(ZextElt2Pol(*G, Rg), PX,  ZENRingDef(Rg));
    ZENPolyFree(PX,  ZENRingDef(Rg));
  }
  return(ret);
}

/* \TEX{
   \title{Converting to string in an internal representation.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg|.}
   \output{An allocated string representing |G| in an internal representation or |ZENNULL| if an error occured.}
   }
   */
char *_ZextEltPutToString(G, Rg)
     ZENElt G;
     ZENRing Rg;
{
  return(ZextEltPutToString(G, Rg));
}


/* \TEX{
   \title{Printing to file.}
   \input{A stream |file|, an allocated |ZENElt| |G| of a finite field
   |Rg| and a base $|base|\in \{2,\ldots, 16\}$.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |G| in base |base| to |file|.}
   }
   */
int _ZextEltPrintToFile(file, G, base, Rg)
     FILE *file;
     ZENElt G;
     int base;
     ZENRing Rg;  
{
  return(ZextEltPrintToFile(file, G, base, Rg));
}

/* \TEX{
   \title{Writing to file in an internal representation.}
   \input{A stream |file|, an allocated |ZENElt| |G| of a finite field
   |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |G| to |file|.}
   }
   */
int _ZextEltPutToFile(file, G, Rg)
     FILE *file;
     ZENElt G;
     ZENRing Rg;  
{
  return(ZextEltPutToFile(file, G, Rg));
}

/* \TEX{
   \title{Reading from file.}
   \input{A stream |file| and an element |G| of a finite field |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|G| is filled with the ZENElt read in |file| if no error occurred.}
   }
   */
int _ZextEltReadFromFile(G, file, base, Rg)
     ZENElt *G;
     FILE *file;
     int base;
     ZENRing Rg;
{
  int ret;
  ZENPoly PX;

  if ((ret = _ZENPolyReadFromFile(&PX, file, ZENRingVar(Rg), base, ZENRingDef(Rg))) != ZENERR) {
    ZENPolyAssign(ZextElt2Pol(*G, Rg), PX,  ZENRingDef(Rg));
    ZENPolyFree(PX,  ZENRingDef(Rg));
  }
  return(ret);
}

/* \TEX{
   \title{Getting from file in an internal representation.}
   \input{A stream |file|, and an element |G| of a finite field |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{|G| is filled with the element read in |file|}
   }
   */
int _ZextEltGetFromFile(G, file, Rg)
     ZENElt *G;
     FILE *file;
     ZENRing Rg;
{
  int ret;
  ZENPoly PX;

  if ((ret = ZENPolyGetFromFile(PX, file, ZENRingDef(Rg))) != ZENERR) {
    ZENPolyAssign(ZextElt2Pol(*G, Rg), PX,  ZENRingDef(Rg));
    ZENPolyFree(PX,  ZENRingDef(Rg));
  }
  return(ret);
}

/* \TEX{
   \title{Printing to stdout.}
   \input{An allocated |ZENElt| |G| of a finite field |Rg|}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{Printing a representation of |G| in base |10| to |stdout|.}
   \note{Usefull for debugging ...}
   }
   */
int _ZextEP(G, Rg)
     ZENElt G;
     ZENRing Rg;  
{
  int ret;

  if ((ret = ZextEltPrintToFile(stdout, G, 10, Rg)) != ZENERR) {
    fprintf(stdout, "\n");
  }
  return(ret);
}

/* \TEX{
   \title{Reading from stdin.}
   \input{A pointer to an element |G| of a finite field |Rg|.}
   \output{0 if no error occurred, |ZENERR| otherwise.}
   \side{|G| is filled with the |ZENElt| read in |stdin| in base 10 if no error occurred.}
   \note{Usefull for debugging ...}
   }
   */
int _ZextER(G, Rg)
     ZENElt *G;
     ZENRing Rg;
{
  int ret;

  if ((ret = ZextEltReadFromFile(*G, stdin, 10, Rg)) != ZENERR) {
    fprintf(stdout, "\n");
  }
  return(ret);
}



/* \TEX{
   \title{Evaluating an element as a multivariate polynomial over $Z$.}
   \input{A Bignum |p| of allocated size greater or equal than
   |ZENRingQl(Rg)|, a pointer |p_pl| on the real size of |p|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|p| is filled with the value of |G| considered as a
   multivariate polynomial in which we substitute all the variables
   with the characteristic, |*p_pl| contains the real size of |p|.}
   } */
void _ZextEltToZ(p, p_pl, G, Rg)
     BigNum p;
     BigNumLength *p_pl;
     ZENElt G;
     ZENRing Rg;
{
  int i;

  ZENEltToZ(p, p_pl,
	    ZENPolyGetCoeffPtr(ZextElt2Pol(G, Rg),
			       ZENPolyDeg(ZextElt2Pol(G, Rg), ZENRingDef(Rg)),
			       ZENRingDef(Rg)),
	    ZENRingDef(Rg));
  ZBNSetToZero(p+(*p_pl), ZENRingQl(Rg)-(*p_pl));

  for(i = ZENPolyDeg(ZextElt2Pol(G, Rg), ZENRingDef(Rg))-1;
      i >= 0 ;
      i--) {  
    ZBNSetToZero(ZENRingTmp1(Rg), ZENRingQl(Rg)+1);
    ZBNMultiply(ZENRingTmp1(Rg), ZENRingQl(Rg)+1,
		ZENRingQ(ZENRingDef(Rg)), ZENRingQl(ZENRingDef(Rg)),
		p, *p_pl);
	      
    ZENEltToZ(p, p_pl,
	      ZENPolyGetCoeffPtr(ZextElt2Pol(G, Rg),
				 i, ZENRingDef(Rg)),
	      ZENRingDef(Rg)); 


    ZBNSetToZero(p+(*p_pl), ZENRingQl(Rg)-(*p_pl));
    ZBNAdd(p, ZENRingQl(Rg), ZENRingTmp1(Rg), ZENRingQl(Rg), 0);
    *p_pl = ZBNNumDigits(p, ZENRingQl(Rg));
  }
}

/* \TEX{
   \title{Getting an element from its evaluation as a multivariate
   polynomial over $Z$.}
   \input{A Bignum |p| of size |pl|, an
   allocated element |G| of a finite ring |Rg|.}
   \side{|G| is filled with the element whose evaluation considered as a
   multivariate polynomial in which we substitute all the variables
   with the characteristic is equal to (|p|, |pl|).}
   } */
void _ZextZToElt(G, p, pl, Rg)
     BigNum p;
     BigNumLength pl;
     ZENElt *G;
     ZENRing Rg;
{
  ZENElt elt;
  BigNumLength pbisl;
  int i;

  ZBNAssign(ZENRingTmp2(Rg), p ,pl);
  pbisl = ZBNNumDigits(ZENRingTmp2(Rg), pl);

  for(i = 0; ; i++) {
    
    if (ZBNCompare(ZENRingTmp2(Rg), pbisl,
		   ZENRingQ(ZENRingDef(Rg)), ZENRingQl(ZENRingDef(Rg))) == BN_LT) {
      break;
    }

    ZBNAssign(ZENRingTmp1(Rg), ZENRingTmp2(Rg), pbisl);
    ZBNDigit(ZENRingTmp1(Rg), pbisl) = 0;
    ZBNDivide(ZENRingTmp1(Rg), pbisl+1,
	      ZENRingQ(ZENRingDef(Rg)), ZENRingQl(ZENRingDef(Rg)));
    
    elt = ZENPolyGetCoeffPtr(ZextElt2Pol(*G, Rg), i, ZENRingDef(Rg));
    ZENZToElt(elt, ZENRingTmp1(Rg), ZENRingSizeElt(ZENRingDef(Rg)),
	      ZENRingDef(Rg));
    ZENPolySetCoeff(ZENElt2Pol(*G, Rg),i,elt,ZENRingDef(Rg));
    
    pbisl = ZBNNumDigits(ZENRingTmp1(Rg)+ZENRingQl(ZENRingDef(Rg)),
			 pbisl+1-ZENRingQl(ZENRingDef(Rg)));
    ZBNAssign(ZENRingTmp2(Rg), ZENRingTmp1(Rg)+ZENRingQl(ZENRingDef(Rg)), pbisl);
  }

  elt = ZENPolyGetCoeffPtr(ZextElt2Pol(*G, Rg),
			   i,
			   ZENRingDef(Rg));
  ZENZToElt(elt, ZENRingTmp2(Rg), pbisl, ZENRingDef(Rg));
  ZENPolySetCoeff(ZENElt2Pol(*G, Rg),
		  i,
		  elt,
		  ZENRingDef(Rg));
  ZENPolyDeg(ZENElt2Pol(*G, Rg), ZENRingDef(Rg)) = i;
  ZENPolyUpdateDegree(ZENElt2Pol(*G, Rg), ZENRingDef(Rg));
}
