/*
 * zen/zext/c/elt_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Element multiplication}
   
   Two functions are provided to do multiplications: |_ZextEltSquare| to square a
   |ZENElt| and |_ZextEltMultily| to compute the multiplication of two |ZENElt|.
   }
   */
#include <stdio.h>
#include "zext.h"

#ifndef SILENT
# ifdef ZEN_NO_TEST
#  define VERBOSE_CHOICE 0
# else
#  define VERBOSE_CHOICE 1
# endif
#endif

#define max(a,b) ((a) < (b) ? (b) : (a))

#define ZextRH(R) ZextElt2Pol(ZextRingHbuf(R),(R))
#define ZextRI(R) ZextElt2Pol(ZextRingRbuf(R),(R))

/* \TEX{
   \title{Squaring.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |Rg|.}
   \side{$|*p_c| = a^2$.}
   }
   */
void _ZextEltSquare(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  ZENPolySquareVeryLazy(ZextRingPbuf(Rg),
			ZextElt2Pol(a, Rg),
			ZENRingDef(Rg));
  ZENPolyModuloVeryLazy(ZextElt2Pol(*p_c, Rg),
			ZextRingPbuf(Rg),
			ZENRingPol(Rg),
			ZENRingSubDeg(Rg),
			ZENRingDef(Rg));
}

/* \TEX{
   \title{Multiplication.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |Rg|.}
   \side{$|*p_c| = a\times b$.}
   }
   */
void _ZextEltMultiply(p_c, a, b, Rg)
     ZENElt *p_c, a, b;
     ZENRing Rg;
{
  ZENPolyMultiplyVeryLazy(ZextRingPbuf(Rg),
			  ZextElt2Pol(a, Rg),
			  ZextElt2Pol(b, Rg),
			  ZENRingDef(Rg));
  ZENPolyModuloVeryLazy(ZextElt2Pol(*p_c, Rg),
			ZextRingPbuf(Rg),
			ZENRingPol(Rg),
			ZENRingSubDeg(Rg),
			ZENRingDef(Rg));
}



/* \TEX{
   \title{Adding precomputations for element multiplication in $\GF{2^n}$.}
   \input{A finite ring |R|.} 
   \output{|ZENERR| if an allocation error occured, 0 otherwise.}
   \side{Buffers are allocated and initialized for Ze2m's operations.}
   }
   */
int _Ze2mPrcInitEltMultiply(R)
     ZENRing R;
{
  ZENPoly a, b, c;
  BigNumDigit C[1], t;
  
#define NBSHORTS (1 << (LN2PREZENMULT<<1))
#define RINGDEF ZENRingDef(R)

  Ze2mRingNbBits(R) = (1+ZENRingDeg(R)%SIZE_BLOC);

  if (!(Ze2mRingPreDivFast(R) = (unsigned int *)ZENMalloc(sizeof(unsigned int)*NBSHORTS))) {
    ZENSetError(ZEN_LIB_ZEXT, ZEN_ELTMULTPRC, ZEN_ALLOC);
    return(ZERR);
  }
  if (!(a = ZENPolyCreate(15, RINGDEF))) {
    ZENFree(Ze2mRingPreDivFast(R) );
    return(ZERR);
  }

  C[0] = ZENPolyGetDigit(ZENRingPol(R), 0, RINGDEF);
  if (!(b = ZENZToPoly(C, 1, RINGDEF))) {
    ZENFree(Ze2mRingPreDivFast(R));
    ZENPolyFree(a, RINGDEF);
    return(ZERR);
  }

  if (!(c = ZENPolyCreate(BN_DIGIT_SIZE, RINGDEF))) {
    ZENFree(Ze2mRingPreDivFast(R));
    ZENPolyFree(a, RINGDEF);
    ZENPolyFree(b, RINGDEF);
    return(ZERR);
  }
  
  if (ZENRingDeg(R) < (LN2PREZENMULT<<1)) {
    ZENPolyDeg(a, RINGDEF) = 0;
    for(ZENPolyGetDigit(a, 0, RINGDEF) = 0;
	ZENPolyGetDigit(a, 0, RINGDEF) < NBSHORTS; 
	(ZENPolyGetDigit(a, 0, RINGDEF))++) {
      if ((1<<(ZENPolyDeg(a, RINGDEF)+1)) == ZENPolyGetDigit(a, 0, RINGDEF)) {
	ZENPolyDeg(a, RINGDEF)++;
      }
      ZENPolyMultiply(c, a, b, RINGDEF);
      Ze2mRingPreDivFast(R)[ZENPolyGetDigit(c, 0, RINGDEF)>>Ze2mRingNbBits(R)-1] =
	ZENPolyGetDigit(c, 0, RINGDEF);
    }
  }
  else if (ZENRingDeg(R) < (LN2PREZENMULT<<2)) {
    for(ZENPolyGetDigit(a, 0, RINGDEF) = 0;
	ZENPolyGetDigit(a, 0, RINGDEF) < NBSHORTS; 
	(ZENPolyGetDigit(a, 0, RINGDEF))++) {
      if ((1<<(ZENPolyDeg(a, RINGDEF)+1)) == ZENPolyGetDigit(a, 0, RINGDEF)) {
	ZENPolyDeg(a, RINGDEF)++;
      }
      ZENPolyMultiply(c, a, b, RINGDEF);
      ZENPolyGetDigit(c, 0, RINGDEF) &= (BigNumDigit)(~(unsigned int)0);
      t = ((ZENPolyGetDigit(a, 0, RINGDEF) >> ((LN2PREZENMULT<<2)-Ze2mRingNbBits(R)+1)) << ((LN2PREZENMULT<<2)-Ze2mRingNbBits(R)+1))
	^ (ZENPolyGetDigit(c, 0, RINGDEF) >> (Ze2mRingNbBits(R)-1));
      Ze2mRingPreDivFast(R)[t] = ZENPolyGetDigit(c, 0, RINGDEF);
    }
  }
  else {
    for(ZENPolyGetDigit(a, 0, RINGDEF) = 0;
	ZENPolyGetDigit(a, 0, RINGDEF) < NBSHORTS; 
	(ZENPolyGetDigit(a, 0, RINGDEF))++) {
      if ((1<<(ZENPolyDeg(a, RINGDEF)+1)) == ZENPolyGetDigit(a, 0, RINGDEF)) {
	ZENPolyDeg(a, RINGDEF)++;
      }
      ZENPolyMultiply(c, a, b, RINGDEF);
      Ze2mRingPreDivFast(R)[ZENPolyGetDigit(a, 0, RINGDEF)] = ZENPolyGetDigit(c, 0, RINGDEF);
    }
  }
  ZENEltSquarePtr(R) = _Ze2mEltSquare; /* Multiplication of elements */
  ZENEltMultiplyPtr(R) = _Ze2mEltMultiply; /* Squaring of elements */

  ZENPolyFree(a, RINGDEF);
  ZENPolyFree(b, RINGDEF);
  ZENPolyFree(c, RINGDEF);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Freeing precomputations for element multiplication in $\GF{2^n}$.}
   \input{A finite ring |R|.} 
   \side{Buffers for Ze2m's operations are freed.}
   }
   */
int _Ze2mPrcCloseEltMultiply(R)
     ZENRing R;
{
  ZENFree(Ze2mRingPreDivFast(R));
  ZENEltSquarePtr(R) = _ZextEltSquare; /* Squaring of elements */
  ZENEltMultiplyPtr(R) = _ZextEltMultiply; /* Multiplication of elements */
  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Squaring with the reciprocal polynomial.}
   \input{A pointer |p_c| on a |ZENElt| and a |ZENElt| |a| of a finite ring
   |Rg|.}
   \side{$|*p_c| = a^2$.}
   }
   */
void _ZextEltSquareRecip(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  int i;

  ZENPolySquareLazy(ZextRingPbuf(Rg),
		ZextElt2Pol(a, Rg),
		ZENRingDef(Rg));
  
  if (ZENPolyDeg(ZextRingPbuf(Rg), ZENRingDef(Rg)) >= ZENRingDeg(Rg)) {
  
    ZENPolyDeg(ZextRI(Rg), ZENRingDef(Rg)) =
      ZENPolyDeg(ZextRingPbuf(Rg), ZENRingDef(Rg)) - ZENRingDeg(Rg); 
    for(i = 0; i <= ZENPolyDeg(ZextRI(Rg), ZENRingDef(Rg)); i++) {
      ZENPolySetCoeff(ZextRI(Rg), i,
		      ZENPolyGetCoeffPtr(ZextRingPbuf(Rg),
					 i + ZENRingDeg(Rg),
					 ZENRingDef(Rg)),
		      ZENRingDef(Rg));
    }
    ZENPolyMultiplyLazy(ZextRH(Rg),
			ZextRingRec(Rg),
			ZextRI(Rg),
			ZENRingDef(Rg));
    
    ZENPolyDeg(ZextRI(Rg), Rg) =
      ZENPolyDeg(ZextRH(Rg), Rg) - ZENRingDeg(Rg); 
    for(i = 0; i <= ZENPolyDeg(ZextRI(Rg), Rg); i++) {
      ZENPolySetCoeff(ZextRI(Rg),
		      i,
		      ZENPolyGetCoeffPtr(ZextRH(Rg),
					 i + ZENRingDeg(Rg),
					 ZENRingDef(Rg)),
		      ZENRingDef(Rg));
    }
    ZENPolyMultiplyLazy(ZextRH(Rg),
			ZENRingPol(Rg),
			ZextRI(Rg),
			ZENRingDef(Rg));
    ZENPolySubtract(ZextRingPbuf(Rg), ZextRH(Rg), ZENRingDef(Rg));
  }
  ZENPolyAssign(ZextElt2Pol(*p_c, Rg), ZextRingPbuf(Rg), ZENRingDef(Rg));
}

/* \TEX{
   \title{Multiplication with the reciprocal polynomial.}
   \input{A pointer |p_c| on a |ZENElt| and two |ZENElt| |a| and |b| of a
   finite ring |Rg|.}
   \side{$|*p_c| = a\times b$.}
   }
   */
void _ZextEltMultiplyRecip(p_c, a, b, Rg)
     ZENElt *p_c, a, b;
     ZENRing Rg;
{
  int i, diff;

  ZENPolyMultiplyLazy(ZextRingPbuf(Rg),
		      ZextElt2Pol(a, Rg),
		      ZextElt2Pol(b, Rg),
		      ZENRingDef(Rg));
  diff = ZENPolyDeg(ZextRingPbuf(Rg), ZENRingDef(Rg))-ZENRingDeg(Rg);
  if (diff >= 0) {
    ZENPolyDeg(ZextRI(Rg), ZENRingDef(Rg)) =
      ZENPolyDeg(ZextRingPbuf(Rg), ZENRingDef(Rg)) - ZENRingDeg(Rg); 
    for(i = 0; i <= ZENPolyDeg(ZextRI(Rg), ZENRingDef(Rg)); i++) {
      ZENPolySetCoeff(ZextRI(Rg),
		      i,
		      ZENPolyGetCoeffPtr(ZextRingPbuf(Rg),
					 i + ZENRingDeg(Rg),
					 ZENRingDef(Rg)),
		      ZENRingDef(Rg));
    }
    ZENPolyMultiplyLazy(ZextRH(Rg),
			ZextRingRec(Rg),
			ZextRI(Rg),
			ZENRingDef(Rg));
    
    ZENPolyDeg(ZextRI(Rg), Rg) =
      ZENPolyDeg(ZextRH(Rg), Rg) - ZENRingDeg(Rg); 
    for(i = 0; i <= ZENPolyDeg(ZextRI(Rg), Rg); i++) {
      ZENPolySetCoeff(ZextRI(Rg),
		      i,
		      ZENPolyGetCoeffPtr(ZextRH(Rg),
					 i + ZENRingDeg(Rg),
					 ZENRingDef(Rg)),
		      ZENRingDef(Rg));
    }
    ZENPolyMultiplyLazy(ZextRH(Rg),
			ZENRingPol(Rg),
			ZextRI(Rg),
			ZENRingDef(Rg));
    ZENPolySubtract(ZextRingPbuf(Rg), ZextRH(Rg), ZENRingDef(Rg));
  }
  ZENPolyAssign(ZextElt2Pol(*p_c, Rg), ZextRingPbuf(Rg), ZENRingDef(Rg));
}



/* \TEX{
   \title{Adding precomputations for element multiplication with
   a polynomial reciprocal.}
   \input{A finite ring |R|.} 
   \output{|ZENERR| if an allocation error occured, 0 otherwise.}
   \side{The reciprocal polynomial of the definition polynomial is computed.}
   }
   */
int _ZextPrcInitEltMultiply(R)
     ZENRing R;
{
  int i, ret, nbops;
  ZENElt elt1, elt2, elt3;
  double tt1, tt2;

#define NBOPS 3

  /* Computing the reciprocal of ZENRingPol(R)
  ********************************************/
  if (!(ZextRingRec(R) = ZENPolyCreate(ZENRingDeg(R), ZENRingDef(R)))) {
    return(ZERR);
  }
  if ((ret = ZENPolyReciprocal(ZextRingRec(R), ZENRingPol(R), ZENRingDef(R))) != ZEN_HAS_INVERSE) {
    ZENPolyFree(ZextRingRec(R), ZENRingDef(R));
    return(ret);
  }
  
  /* Comparing the running times
  ******************************/
  if (ZENEltCreate(elt1, R)) {
    ZENPolyFree(ZextRingRec(R), ZENRingDef(R));
    return(ZERR);
  }
  if (ZENEltCreate(elt2, R)) {
    ZENEltFree(elt1, R);
    ZENPolyFree(ZextRingRec(R), ZENRingDef(R));
    return(ZERR);
  }
  ZENEltSetRandom(elt2, R);
  if (ZENEltCreate(elt3, R)) {
    ZENEltFree(elt1, R); ZENEltFree(elt2, R);
    ZENPolyFree(ZextRingRec(R), ZENRingDef(R));
    return(ZERR);
  }

  /* squarings */
  for(tt1 = 0., tt2 = 0., nbops = 1;
      !tt1 && !tt2;
      nbops *= 10) {
    tt1 = runtime();
    for(i = 0; i < nbops; i++) {
      ZENEltSetRandom(elt1, R);
      _ZextEltSquare(&elt3, elt1, R);
    }
    tt1 = runtime()-tt1;
    tt2 = runtime();
    for(i = 0; i < nbops; i++) {
      ZENEltSetRandom(elt1, R);
      _ZextEltSquareRecip(&elt3, elt1, R);
    }
    tt2 = runtime()-tt2;
  }
  if (nbops > 1) {nbops /= 10;}
  if (tt1 > tt2) {
    nbops = (int)((nbops*30.)/tt1);
  }
  else {
    nbops = (int)((nbops*30.)/tt2);
  }
  nbops = nbops < NBOPS ? NBOPS : nbops;

  tt1 = runtime();
  for(i = 0; i < nbops; i++) {
    ZENEltSetRandom(elt1, R);
    _ZextEltSquare(&elt3, elt1, R);
  }
  tt1 = runtime()-tt1;
#if VERBOSE_CHOICE  
  fprintf(stderr,"%d Usual squaring : %f ms/ops\n",nbops,tt1/(double)nbops);
#endif
  tt2 = runtime();
  for(i = 0; i < nbops; i++) {
    ZENEltSetRandom(elt1, R);
    _ZextEltSquareRecip(&elt3, elt1, R);
  }
  tt2 = runtime()-tt2;

#if VERBOSE_CHOICE  
  fprintf(stderr,"%d Reciprocal squaring : %f ms/ops\n",nbops,tt2/(double)nbops);
#endif
  if (tt1 > tt2) {
    ZENEltSquarePtr(R) = _ZextEltSquareRecip; /* Multiplication of elements */
#if VERBOSE_CHOICE  
    fprintf(stderr,"Reciprocal squaring chosen\n");
#endif
  }
#if VERBOSE_CHOICE  
  else {
    fprintf(stderr,"Usual squaring chosen\n");
  }
#endif

  /* multiplications */
  for(tt1 = 0., tt2 = 0., nbops = 1;
      !tt1 && !tt2;
      nbops *= 10) {
    tt1 = runtime();
    for(i = 0; i < nbops; i++) {
      ZENEltSetRandom(elt1, R);
      ZENEltSetRandom(elt2, R);
      _ZextEltMultiply(&elt3, elt1, elt2, R);
    }
    tt1 = runtime()-tt1;
    tt2 = runtime();
    for(i = 0; i < nbops; i++) {
      ZENEltSetRandom(elt1, R);
      ZENEltSetRandom(elt2, R);
      _ZextEltMultiplyRecip(&elt3, elt1, elt2, R);
    }
    tt2 = runtime()-tt2;
  }
  if (nbops > 1) {nbops /= 10;}
  if (tt1 > tt2) {
    nbops = (int)((nbops*30.)/tt1);
  }
  else {
    nbops = (int)((nbops*30.)/tt2);
  }
  nbops = nbops < NBOPS ? NBOPS : nbops;

  tt1 = runtime();
  for(i = 0; i < nbops; i++) {
    ZENEltSetRandom(elt1, R);
    ZENEltSetRandom(elt2, R);
    _ZextEltMultiply(&elt3, elt1, elt2, R);
  }
  tt1 = runtime()-tt1;
#if VERBOSE_CHOICE  
  fprintf(stderr,"%d Usual multiplication : %f ms/ops\n",nbops,tt1/(double)nbops);
#endif
  tt2 = runtime();
  for(i = 0; i < nbops; i++) {
    ZENEltSetRandom(elt1, R);
    ZENEltSetRandom(elt2, R);
    _ZextEltMultiplyRecip(&elt3, elt1, elt2, R);
  }
  tt2 = runtime()-tt2;
#if VERBOSE_CHOICE  
  fprintf(stderr,"%d Reciprocal multiplication : %f ms/ops\n",nbops,tt2/(double)nbops);
#endif
  if (tt1 > tt2) {
    ZENEltMultiplyPtr(R) = _ZextEltMultiplyRecip; /* Multiplication of elements */
#if VERBOSE_CHOICE  
    fprintf(stderr,"Reciprocal multiplication chosen\n");
#endif
  }
#if VERBOSE_CHOICE  
  else {
    fprintf(stderr,"Usual multiplication chosen\n");
  }
#endif

  ZENEltFree(elt1, R); ZENEltFree(elt2, R); ZENEltFree(elt3, R); 
  
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Freeing precomputations for element multiplication  with
   a polynomial reciprocal.}
   \input{A finite ring |R|.} 
   \side{The reciprocal polynomial freed.}
   }
   */
int _ZextPrcCloseEltMultiply(R)
     ZENRing R;
{
  ZENPolyFree(ZextRingRec(R), ZENRingDef(R));
  ZENEltSquarePtr(R) = _ZextEltSquare; /* Usual squaring of elements */
  ZENEltMultiplyPtr(R) = _ZextEltMultiply; /* Usual multiplication of elements */
  return(ZEN_HAS_INVERSE);
}
