/*
 * zen/zext/c/elt_trace.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Element trace}
   
   |_ZextEltTrace| computes the trace of a |ZENElt|.
   }
   */
#include <stdio.h>
#include "zext.h"

/* \TEX{
   \title{Trace with no precomputations.}
   \input{A |ZENElt| |a| of a finite ring |Rg|.}
   \side{|p_c| is filled with the trace of |a|.}
   }
   */
void _ZextEltTraceNaive(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  int j;

  ZextEltAssign(ZENRingJbuf(Rg), a, Rg);
  ZextEltAssign(*p_c, a, Rg);
  for(j = 1; j < ZENRingDeg(Rg); j++) {
    ZextEltExp(ZENRingJbuf(Rg), ZENRingQ(ZENRingDef(Rg)), ZENRingQl(ZENRingDef(Rg)), ZENRingJbuf(Rg), Rg);
    ZextEltAdd(*p_c, ZENRingJbuf(Rg), Rg);
  }
}

/* \TEX{
   \title{Trace with precomputations.}
   \input{A |ZENElt| |a| of a finite ring |Rg|.}
   \side{|p_c| is filled with the trace of |a|.}
   }
   */
void _ZextEltTracePrc(p_c, a, Rg)
     ZENElt *p_c, a;
     ZENRing Rg;
{
  ZENElt elt;

  elt = ZENPolyGetCoeffPtr(ZextElt2Pol(*p_c, Rg), 0, ZENRingDef(Rg)); 
  ZENPolyDot(elt,
	     ZextElt2Pol(a, Rg),
	     ZextRingTbuf(Rg),
	     ZENRingDef(Rg));
  ZENPolySetCoeff(ZextElt2Pol(*p_c, Rg), 0, elt, ZENRingDef(Rg));
  ZextPolyDeg(ZextElt2Pol(*p_c, Rg), Rg) = 0;
}


/* \TEX{
   \title{Precomputations.}
   \input{A finite ring |Rg|.}
   \output{|ZENERR| if an error occured, |ZEN_HAS_INVERSE| otherwise.}
   \side{The finite field |Rg| is filled with precomputations for
   speeding up trace precomputations.}
   }
   */
int _ZextPrcInitTrace(Rg)
     ZENRing Rg;
{
  char *T;
  int i, j, p;
  ZENElt t, v, w, elt, elt1;

  /* Precomputations to compute efficiently traces */
  p = 0;
  if ((ZENRingQl(Rg) == 1) &&
      (ZBNDigit(ZENRingQ(ZENRingDef(Rg)), 0) < ZENRingDeg(Rg))) {
    p = (int)ZBNDigit(ZENRingQ(ZENRingDef(Rg)), 0);
  }
  
  if (!(T = (char *)ZENMalloc(sizeof(char)*ZENRingDeg(Rg)))) {
    return(ZERR);
  }
  for(i = 0; i < ZENRingDeg(Rg); i++) {
    T[i] = 0;
  }
  
  if (!(ZextRingTbuf(Rg) = ZENPolyCreate(ZENRingDeg(Rg)-1, ZENRingDef(Rg)))) {
    ZENFree(T);
    return(ZERR);
  }
  ZENPolyDeg(ZextRingTbuf(Rg), ZENRingDef(Rg)) = ZENRingDeg(Rg)-1;
  
  if (ZextEltCreate(t, Rg)) {
    ZENPolyFree(ZextRingTbuf(Rg), ZENRingDef(Rg));
    ZENFree(T);
    return(ZERR);
  }
    
  if (ZextEltCreate(v, Rg)) {
    ZENPolyFree(ZextRingTbuf(Rg), ZENRingDef(Rg));
    ZextEltFree(t, Rg);
    ZENFree(T);
    return(ZERR);
  }
  
  if (ZextEltCreate(w, Rg)) {
    ZENPolyFree(ZextRingTbuf(Rg), ZENRingDef(Rg));
    ZENFree(T);
    ZextEltFree(t, Rg);
    ZextEltFree(v, Rg);
    return(ZERR);
  }
  
  for(i = 0; i < ZENRingDeg(Rg); i++) {
    if (T[i])
      continue;
    ZENPolySetToXi(ZextElt2Pol(v, Rg), i, ZENRingDef(Rg));
    ZextEltAssign(w, v, Rg);
    for(j = 1; j < ZENRingDeg(Rg); j++) {
      ZextEltExp(t, ZENRingQ(ZENRingDef(Rg)), ZENRingQl(ZENRingDef(Rg)), v, Rg);
      ZextEltAssign(v, t, Rg);
      ZextEltAdd(w, v, Rg);
    }
    if (ZENPolyDeg(ZextElt2Pol(w, Rg),  ZENRingDef(Rg))) {
      ZENPolyFree(ZextRingTbuf(Rg), ZENRingDef(Rg));
      ZENSetError(ZEN_LIB_ZEXT,ZEN_ELTTRACPRC,ZEN_ELT_NOTFF);
      break;
    }
    if (!i) {
      elt = ZENPolyGetCoeffPtr(ZextRingTbuf(Rg), 0, ZENRingDef(Rg));
      elt1 = ZENPolyGetCoeffPtr(ZENElt2Pol(w, Rg), 0, ZENRingDef(Rg));
      ZENEltAssign(elt, elt1, ZENRingDef(Rg));
      ZENPolySetCoeff(ZextRingTbuf(Rg), 0, elt, ZENRingDef(Rg));
      T[0] = 1;
    }
    else {
      for(j = i; j && (j < ZENRingDeg(Rg)); j *= p) {
	elt = ZENPolyGetCoeffPtr(ZextRingTbuf(Rg), j, ZENRingDef(Rg));
	elt1 = ZENPolyGetCoeffPtr(ZENElt2Pol(w, Rg), 0, ZENRingDef(Rg));
	ZENEltAssign(elt, elt1, ZENRingDef(Rg));
	ZENPolySetCoeff(ZextRingTbuf(Rg), j, elt, ZENRingDef(Rg));
	T[j] = 1;
      }
    }
  }
  ZextEltFree(t, Rg);
  ZextEltFree(v, Rg);
  ZextEltFree(w, Rg);
  ZENFree(T);

  if (i == ZENRingDeg(Rg)) {
    ZENEltTracePtr(Rg) = _ZextEltTracePrc;
    return(ZEN_HAS_INVERSE);
  }
  else {
    return(ZERR);
  }
}

/* \TEX{
   \title{Free precomputations for traces.}
   \input{A finite ring |Rg|.}
   \output{|ZEN_HAS_INVERSE|}.
   \side{Memory is desallocated.}
   }
   */
int _ZextPrcCloseTrace(Rg)
     ZENRing Rg;
{
    ZENPolyFree(ZextRingTbuf(Rg), ZENRingDef(Rg));
    ZENEltTracePtr(Rg) = _ZextEltTraceNaive;

    return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Find an element of trace one over an extension.}
   \input{A finite extension |Rg|.}
   \output{|ZENERR| if an error occured, |ZEN_NO_INVERSE| if an inverse was
   inpossible to compute, |ZEN_HAS_INVERSE| otherwise.} 
   \side{The finite field |Rg| is updated. In particular, an element of
   trace one is stored.}
   }
   */
int _ZextPrcInitPolyRootsCanonical(Rg)
     ZENRing Rg;
{
  ZENPrc prc;
  ZENElt w, elt;

  ZENPrcSetNone(prc);
  ZENPrcSet(prc, ZENPRC_POLY_ROOTS_CANONICAL);
  ZENPrcSet(prc, ZENPRC_FINITE_FIELD);
  switch (ZENRingAddPrc(ZENRingDef(Rg), prc)) {
  case ZENERR:
    return(ZERR);
  case ZEN_NO_INVERSE:
    return(ZEN_NO_INVERSE);
  case ZEN_HAS_INVERSE:
    if (!ZENPrcIsSet(ZENRingPrcp(ZENRingDef(Rg)),
		     ZENPRC_POLY_ROOTS_CANONICAL)) {
      return(ZERR);
    }
    break;
  }
  
  if (ZENEltCreate(w, Rg)) {
    return(ZERR);
  }

  if (ZENEltCreate(ZENRingTheta(Rg), Rg)) {
    ZextEltFree(w, Rg);
    return(ZERR);
  }

  do {
    ZENEltSetRandom(ZENRingTheta(Rg), Rg);
    ZENEltTrace(w, ZENRingTheta(Rg), Rg);
  }
  while (ZENEltIsZero(w, Rg));
  
  if (ZENPolyDeg(ZextElt2Pol(w, Rg),  ZENRingDef(Rg))) {
    ZENEltFree(ZENRingTheta(Rg), Rg);
    ZextEltFree(w, Rg);
    return(ZERR);
  }

  elt = ZENPolyGetCoeffPtr(ZextElt2Pol(w, Rg), 0, ZENRingDef(Rg));
  if (ZENEltInverse(elt, elt, ZENRingDef(Rg)) != ZEN_HAS_INVERSE) {
    ZextEltFree(w, Rg);
    ZENEltFree(ZENRingTheta(Rg), Rg);
    return(ZEN_NO_INVERSE);
  }
  ZENEltMultiply(elt, ZENRingTheta(ZENRingDef(Rg)), elt, ZENRingDef(Rg));
  
  ZENPolyMultiplyScalar(ZENElt2Pol(ZENRingTheta(Rg), Rg), ZENElt2Pol(ZENRingTheta(Rg), Rg), elt, ZENRingDef(Rg));
  
  ZextEltFree(w, Rg);
  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Free the element of trace 1.}
   \input{A finite ring |Rg|.}
   \output{|ZEN_HAS_INVERSE|}.
   \side{Memory is desallocated.}
   }
   */
int _ZextPrcClosePolyRootsCanonical(Rg)
     ZENRing Rg;
{
  ZENEltFree(ZENRingTheta(Rg), Rg);

  return(ZEN_HAS_INVERSE);
}
