/*
 * zen/zext/c/mat_create.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Allocating a matrix}
   }
   */


#include <stdio.h>
#include "zext.h"
#include "zeg.h"

/* \TEX{
   \title{Computation of the matrix parameters}
   \input{A |ZENMatTypeStruct|, the dimensions of the desired matrix, a pointer on a |ZENMat| structure and a |ZextRing|.}
   \output{0 if no error occured, |ZENERR| otherwise.}
   \side{The |ZENMat| structure is allocated and the parameters computed according to the type of the desired matrix. Array is not allocated.}
   \note{This function should not be directly used (see |ZENMatCreate|). If chosen in a |ZENRing| initialization  routine, |_ZedGetBloc|, |_ZedGetBlocPtr| and |_ZedSetBloc| must also be chosen.}
   }
   */   

int _ZextMatCreateParams(ts,l,c,p_pms,R)
     Dim l,c;
     ZENMatTypeStruct ts;
     ZENMat *p_pms;
     ZENRing R;
{
  *p_pms=(ZENMat)ZENMalloc(sizeof(struct zenmat));

  ZENTest(*p_pms==NULL,
	  ZEN_ALLOC,ZEN_MATCREPMS,ZEN_LIB_ZEXT, ZERR);
  ZENTest((l<=0)||(c<=0),
	  ZEN_MAT_DIM0,ZEN_MATCREPMS,ZEN_LIB_ZEXT, ZERR);

  ZENMatFreePtr(*p_pms)=_ZextMatFree;
  ZENMatAssignPtr(*p_pms)=__ZENMatAssign;
  ZENMatSetToOnePtr(*p_pms)=__ZENMatSetToOne;
  ZENMatIsOnePtr(*p_pms)=__ZENMatIsOne;
  ZENMatSetRandomPtr(*p_pms)=__ZENMatSetRandom;
  ZENMatPermuteBnPtr(*p_pms)=__ZENMatPermuteBn;     
  switch(ts)
    {
    case ZENMatTypeRow:
      ZENMatType(*p_pms,R)=ZENMatTypeRow;
      ZENMatNbBn(*p_pms,R)=l;
      ZENMatNbDigit(*p_pms,R)=c;
      ZENMatNbBloc(*p_pms,R)=c;

      ZENMatGetCoeffNewPtr(*p_pms)=__ZENMatGetBloc;
      ZENMatGetCoeffPtrPtr(*p_pms)=__ZENMatGetBlocPtr;
      ZENMatSetCoeffPtr(*p_pms)=__ZENMatSetBloc;     
      ZENMatPermuteRowPtr(*p_pms)=__ZENMatPermuteBn;     
      ZENMatPermuteColPtr(*p_pms)=__ZENMatPermuteDigit;     
      break;
      
    case ZENMatTypeCol:
      ZENMatType(*p_pms,R)=ZENMatTypeCol;
      ZENMatNbBn(*p_pms,R)=c;
      ZENMatNbDigit(*p_pms,R)=l;
      ZENMatNbBloc(*p_pms,R)=l;

      ZENMatGetCoeffNewPtr(*p_pms)=__ZENMatGetDigit;
      ZENMatGetCoeffPtrPtr(*p_pms)=__ZENMatGetDigitPtr;
      ZENMatSetCoeffPtr(*p_pms)=__ZENMatSetDigit;     
      ZENMatPermuteRowPtr(*p_pms)=__ZENMatPermuteDigit;     
      ZENMatPermuteColPtr(*p_pms)=__ZENMatPermuteBn;     
      break;

    default:
      ZENTest(1,
	      ZEN_UNKNOWN,ZEN_MATCREPMS,ZEN_LIB_ZEXT, ZERR);
    }

  switch(ZENRingArith(ZENRingDef(R)))
    {
    case ZEN_LIB_ZE2:
      ZENMatSize(*p_pms,R)= divSizeBloc(ZENPolyDeg(ZENRingPol(R),
						   ZENRingDef(R)))+1;
      ZENMatNbPtr(*p_pms,R)=l*c;
      ZENMatNbBnd(*p_pms,R)= ZENMatNbPtr(*p_pms,R)*ZENMatSize(*p_pms,R);
      break; 
    default:
      ZENMatSize(*p_pms,R)= ZENPolyDeg(ZENRingPol(R),ZENRingDef(R));
      ZENMatNbPtr(*p_pms,R)=l*c;
      ZENMatNbBnd(*p_pms,R)= ZENMatNbPtr(*p_pms,R)*ZENMatSize(*p_pms,R);
      break;
    }
  return(0);
}

/*
   \TEX{
   \title{Allocating a matrix}
   \input{A parametrized but unallocated |ZENMat| and a |ZENRing|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{The matrix coefficients are allocated.}
   \note{|ZENMatCreate| calls this function after call to |ZENMatCreateParams|.Should not be used by itself.}
   } 
   */
int __ZextMatCreate(M,R)
     ZENMat M;
     ZENRing R;
{
  ZENElt *ptmat;
  int l,c,k;
  ZENPoly P;
 
  ZENMatMat(M,R)=(ZENElt **)ZENMalloc(ZENMatNbBn(M,R)*sizeof(BigNum *));  
  ZENTest(ZENMatMat(M,R)==NULL,
	  ZEN_ALLOC,ZEN_MAT_CREATE,ZEN_LIB_ZEXT, ZERR);

  ptmat=(ZENElt *)ZENMalloc((ZENMatNbPtr(M,R)+ZENMatNbBnd(M,R))*sizeof(ZENElt));
  ZENTest(ptmat==NULL,
	  ZEN_ALLOC,ZEN_MAT_CREATE,ZEN_LIB_ZEXT, ZERR);
  
  for(l=0;l<ZENMatNbBn(M,R);l++,ptmat += ZENMatNbBloc(M,R))
    ZENMatBn(M,l,R)=ptmat;
  ZENMatFirst(M,R)=(BigNumDigit *)ptmat;  
  
  for(l=0;
      l<ZENMatNbBn(M,R);
      l++)
    for(c=0;
	c<ZENMatNbBloc(M,R);
	c++,ptmat+=ZENMatSize(M,R))
      {
	P=ZENPolyAllocStruct(ZENRingDef(R));
	ZENTest(P==NULL,
		ZEN_ALLOC,ZEN_MAT_CREATE,ZEN_LIB_ZEXT, ZERR);
	ZENPolyDeg(P,ZENRingDef(R))= 0;
	ZENPolyCof(P,ZENRingDef(R))=ptmat;
	switch(ZENRingArith(ZENRingDef(R)))
	  {
	  case ZEN_LIB_ZE2:
	    ZENPolyLgt(P,ZENRingDef(R))= ZENMatSize(M,R)*SIZE_BLOC - 1;
	    break;
	  case ZEN_LIB_ZEPS:
	    ZENPolyLgt(P,ZENRingDef(R))= ZENMatSize(M,R) - 1;
	    break; 
	  case ZEN_LIB_ZELOG:
	    ZENPolyLgt(P,ZENRingDef(R))= ZENMatSize(M,R)*sizeof(BigNumDigit)/sizeof(unsigned short) - 1;
	    break; 
	  case ZEN_LIB_ZETAB:
	    ZENPolyLgt(P,ZENRingDef(R))= ZENMatSize(M,R)*sizeof(BigNumDigit)/sizeof(unsigned char) - 1;
	    break; 
	  default:
	    ZENPolyLgt(P,ZENRingDef(R))= ZENMatSize(M,R) - 1;
	    for(k=0;k< ZENMatSize(M,R);k++)
	      ZENCheckError(
			    ZENEltAlloc(
					ZENPolyGetCof(P,k,ZENRingDef(R)),
					ZENRingDef(R)),
			    ZERR);
	    break;
	  }
	ZENElt2Pol(ZENMatBloc(M,l,c,R),R)=P;
      }
  ZENCheckError(ZENMatSetToZero(M,R), ZERR);
  return(0);
}

/*
   \TEX{
   \title{Freeing a matrix}
   \input{A |ZENMat| over a |ZENRing|.}
   \side{The matrix is completely freed.}
   } 
   */
void _ZextMatFree(M,R)
          ZENMat M;
     ZENRing R;
{
  int l,c,k;
 
  switch(ZENRingArith(ZENRingDef(R)))
    {
    case ZEN_LIB_ZE2:
    case ZEN_LIB_ZEPS:
    case ZEN_LIB_ZELOG:
    case ZEN_LIB_ZETAB:
      for(l=ZENMatNbBn(M,R);l--;)
	for(c=ZENMatNbBloc(M,R);c--;)
	  ZENFree(ZENElt2Pol(ZENMatBloc(M,l,c,R),R));
      break; 
    default:
      for(l=ZENMatNbBn(M,R);l--;)
	for(c=ZENMatNbBloc(M,R);c--;)
	  {
	    for(k=0;k< ZENMatSize(M,R);k++)
	      ZENEltFree(ZENPolyGetCof(ZENElt2Pol(ZENMatBloc(M,l,c,R),R),k,
				       ZENRingDef(R)),
			 ZENRingDef(R));
	    ZENFree(ZENElt2Pol(ZENMatBloc(M,l,c,R),R));
	  }
      break;
    }
  
  ZENMatFirst(M,R)-=ZENMatNbPtr(M,R);
  ZENFree(ZENMatFirst(M,R));
  ZENFree(ZENMatMat(M,R));
  ZENFree(M);
}


