/*
 * zen/zext/c/poly_add.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial additions}
   
   One function is provided to do additions: |_ZextPolyAdd| to add a
   |ZENPoly| to another.
   }
   */
#include <stdio.h>
#include "zext.h"

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))
#define ZextPGC(P,i,R) ZextElt2Pol(ZextPolyGetCof((P),(i),(R)),(R))

/* \TEX{
   \title{Addition.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| += |PX|.}
   }
   */
void _ZextPolyAdd(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int mindeg, i;

  if (ZextPolyDeg(RX, Rg) < ZextPolyDeg(PX, Rg)) {
    mindeg = ZextPolyDeg(RX, Rg);
    ZextPolyDeg(RX, Rg) =  ZextPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZextPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    ZextEltAdd(ZextPolyGetCof(RX, i, Rg), ZextPolyGetCof(PX, i, Rg), Rg);
  }
  for(i = mindeg+1; i <= ZextPolyDeg(PX, Rg); i++) {
    ZextEltAssign(ZextPolyGetCof(RX, i, Rg),
		   ZextPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZextPolyUpdateDegree(RX, Rg);
}

/* For internal use only, cf. poly_multiply. */
void _ZextPolyAddLarge(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int mindeg, i;

  mindeg = min(ZextPolyDeg(RX, Rg), ZextPolyDeg(PX, Rg));
  for(i = 0; i <= mindeg; i++) {
    ZENPolyAddLarge(ZextPGC(RX, i, Rg), ZextPGC(PX, i, Rg), ZENRingDef(Rg));
  }
  for(i = mindeg+1; i <= ZextPolyDeg(PX, Rg); i++) {
    ZENPolyAssignLarge(ZextPGC(RX, i, Rg), ZextPGC(PX, i, Rg), ZENRingDef(Rg));
  }
  for(i = max(i-1, ZextPolyDeg(RX, Rg)); i >= 0; i--) {
    if (!ZENPolyIsZero(ZextPGC(RX, i, Rg), ZENRingDef(Rg))) {
      break;
    }
  }
  ZextPolyDeg(RX, Rg) = i;
}


/* \TEX{
   \title{Negation.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| = -|PX|.}
   }
   */
void _ZextPolyNegate(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZextPolyDeg(PX, Rg); i++) {
    ZextEltNegate(ZextPolyGetCof(RX, i, Rg), ZextPolyGetCof(PX, i, Rg), Rg);
  }
  ZextPolyDeg(RX, Rg) =  ZextPolyDeg(PX, Rg);
}

void _ZextPolyNegateLarge(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int mindeg, i;

  for(i = 0; i <= ZENPolyDeg(PX, Rg); i++) {
    ZENPolyNegateLarge(ZextPGC(RX, i, Rg), ZextPGC(PX, i, Rg), ZENRingDef(Rg));
  }
  ZextPolyDeg(RX, Rg) = ZENPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Subtraction.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \side{|RX| -= |PX|.}
   }
   */
void _ZextPolySubtract(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, mindeg;

  if (ZextPolyDeg(RX, Rg) < ZextPolyDeg(PX, Rg)) {
    mindeg = ZextPolyDeg(RX, Rg);
    ZextPolyDeg(RX, Rg) =  ZextPolyDeg(PX, Rg);
  }
  else {
    mindeg =  ZextPolyDeg(PX, Rg);
  }

  for(i = 0; i <= mindeg; i++) {
    ZextEltSubtract(ZextPolyGetCof(RX, i, Rg), ZextPolyGetCof(PX, i, Rg), Rg);
  }
  for(i = mindeg+1; i <= ZextPolyDeg(PX, Rg); i++) {
    ZextEltNegate(ZextPolyGetCof(RX, i, Rg),
		   ZextPolyGetCof(PX, i, Rg),
		   Rg);
  }
  ZextPolyUpdateDegree(RX, Rg);
}

/* For internal use only, cf. poly_multiply. */
void _ZextPolySubtractLarge(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int mindeg, i;

  mindeg = min(ZextPolyDeg(RX, Rg), ZextPolyDeg(PX, Rg));
  for(i = 0; i <= mindeg; i++) {
    ZENPolySubtractLarge(ZextPGC(RX, i, Rg), ZextPGC(PX, i, Rg), ZENRingDef(Rg));
  }
  for(i = mindeg+1; i <= ZextPolyDeg(PX, Rg); i++) {
    ZENPolyNegateLarge(ZextPGC(RX, i, Rg), ZextPGC(PX, i, Rg), ZENRingDef(Rg));
  }
  for(i = max(i-1, ZextPolyDeg(RX, Rg)); i >= 0; i--) {
    if (!ZENPolyIsZero(ZextPGC(RX, i, Rg), ZENRingDef(Rg))) {
      break;
    }
  }
  ZextPolyDeg(RX, Rg) = i;
}
