/*
 * zen/zext/c/poly_assign.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/* \TEX{
   \Subsection{Polynomial assignment}
   
   One can set a |ZENPoly| to zero with |_ZextPolySetToZero|. Then we can
   assign a |ZENPoly| with an other with |_ZextPolyAssign|. Furthermore, we
   can set randomly a |ZENPoly| with |_ZextPolySetRandom|. 
   
   The functions |_ZextPolyGetCoeff| and |_ZextPolySetCoeff|
   enable us to get or set the coefficient of $X^i$,
   |_ZextPolyUpdateDegree| compute the real degree of a |ZENPoly| if it
   decreased.
   }
   */
#include <stdio.h>
#include "zext.h"

#define min(a,b) ((a) < (b) ? (a) : (b))
#define ZextPGC(P,i,R) ZextElt2Pol(ZextPolyGetCof((P),(i),(R)),(R))

/* \TEX{
   \title{Assign.}
   \input{Two |ZENPoly| |PX| and |RX| of a finite ring |Rg|.}
   \side{|RX| is filled with |PX|.}
   }
   */
void _ZextPolyAssign(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZextPolyDeg(PX, Rg); i++) {
    ZextEltAssign(ZextPolyGetCof(RX, i, Rg),
		  ZextPolyGetCof(PX, i, Rg),
		  Rg);
  }
  ZextPolyDeg(RX, Rg) =  ZextPolyDeg(PX, Rg);
}

/* For internal use only, cf. poly_multiply. */
void _ZextPolyAssignLarge(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZextPolyDeg(PX, Rg); i++) {
    ZENPolyAssignLarge(ZextPGC(RX, i, Rg), ZextPGC(PX, i, Rg), ZENRingDef(Rg));
  }
  ZextPolyDeg(RX, Rg) =  ZextPolyDeg(PX, Rg);
}
/* For internal use only, cf. poly_multiply. */
void _ZextPolyLazyToVeryLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  for(i = 0; i <= ZextPolyDeg(PX, Rg); i++) {
    ZENPolyLazyToVeryLazy(ZextPGC(RX, i, Rg), ZextPGC(PX, i, Rg), ZENRingDef(Rg));
  }
  for(; i <= ZextPolyLgt(RX, Rg); i++) {
    ZENPolySetToZeroLarge(ZextPGC(RX, i, Rg), ZENRingDef(Rg));
  }
  ZextPolyDeg(RX, Rg) =  ZextPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Set to zero.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{|RX| is set to zero..}
   }
   */
void _ZextPolySetToZero(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  ZextPolyDeg(RX, Rg) =  0;
  ZextEltSetToZero(ZextPolyGetCof(RX, 0, Rg), Rg);
}

void _ZextPolySetToZeroLarge(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  int i;
  for(i = 0; i <= ZextPolyLgt(RX, Rg); i++) {
    ZENPolySetToZeroLarge(ZextPGC(RX, i, Rg), ZENRingDef(Rg));
  }
  ZextPolyDeg(RX, Rg) =  0;
}

/* \TEX{
   \title{Set to $X^i$.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg| allocated at least for
   degree $|deg|$.}
   \side{|RX| is set to $X^{|deg|}$.}
   }
   */
void _ZextPolySetToXi(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  int i;

  ZextPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    ZextEltSetToZero(ZextPolyGetCof(RX, i, Rg), Rg);
  }
  ZextEltSetToOne(ZextPolyGetCof(RX, deg, Rg), Rg);
}

/* \TEX{
   \title{Set to random.}
   \input{A |ZENPoly| |RX| of degree |deg| of a finite ring |Rg|.}
   \side{|RX| is set randomly with degree |deg|.}
   }
   */
void _ZextPolySetRandom(RX, deg, Rg)
     ZENPoly RX;
     int deg;
     ZENRing Rg;
{
  int i;

  ZextPolyDeg(RX, Rg) =  deg;

  for(i = 0; i < deg; i++) {
    ZextEltSetRandom(ZextPolyGetCof(RX, i, Rg), Rg);
  }
  ZextEltSetRandom(ZextPolyGetCof(RX, deg, Rg), Rg);
  while(ZextEltIsZero(ZextPolyGetCof(RX, deg, Rg), Rg)) {
    ZextEltSetRandom(ZextPolyGetCof(RX, deg, Rg), Rg);
  }
}

/* \TEX{
   \title{Set a coeff.}
   \input{A |ZENPoly| |RX| and a |ZENElt| |b| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \side{The coeff of $X^d$ in |RX| is set to |b|.}
   }
   */
void _ZextPolySetCoeff(RX, d, b, Rg)
     ZENPoly RX;
     int d;
     ZENElt b;
     ZENRing Rg;
{
  ZextPolySetCoeff(RX, d, b, Rg);
}

/* \TEX{
   \title{Get a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The coeff of $X^d$ in |RX|.}
   }
   */
void _ZextPolyGetCoeff(p_c, RX, d, Rg)
     ZENElt *p_c;
     ZENPoly RX;
     int d;
     ZENRing Rg;
{ 
  ZextPolyGetCoeff(*p_c, RX, d, Rg);
}

/* \TEX{
   \title{Extract a coeff.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|, the degree |d| of the coeff to set.}
   \output{The pointer to the coeff of $X^d$ in |RX|.}
   }
   */
ZENElt _ZextPolyGetCoeffPtr(RX, d, Rg)
     ZENPoly RX;
     int d;
     ZENRing Rg;
{
  return(ZextPolyGetCoeffPtr(RX,d,Rg));
}

/* \TEX{
   \title{Update the degree of a polynomial.}
   \input{A |ZENPoly| |RX| of a finite ring |Rg|.}
   \side{Resetting the degree of |P|, assuming it has decreased.}
   }
   */
void _ZextPolyUpdateDegree(RX, Rg)
     ZENPoly RX;
     ZENRing Rg;
{
  int i;

  for(i = ZextPolyDeg(RX, Rg); i >= 0; i--) {
    if (!ZextEltIsZero(ZextPolyGetCoeffPtr(RX,i,Rg), Rg)) {
      break;
    }
  }
  if (i < 0) {
    ZextPolySetToZero(RX, Rg);
  }
  else {
    ZextPolyDeg(RX, Rg) =  i;
  }
}
