/*
 * zen/zext/c/poly_multiply.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
   \TEX{
   \Subsection{Polynomial multiplication}
   
   Two functions are provided to do multiplications: |_ZextPolySquare| to square a
   |ZENPoly| and |_ZextPolyMultily| to compute the multiplication of two |ZENPoly|.
   }
   */
#include <stdio.h>
#include "zext.h"
#include "zeg.h"

#define max(a,b) ((a) > (b) ? (a) : (b))
#define min(a,b) ((a) < (b) ? (a) : (b))

#define ZEXT_POLY_MULTIPLY_CUTOFF 1
#define ZEXT_POLY_SQUARE_CUTOFF 1

#define ZextPGC(P,i,R) ZextElt2Pol(ZextPolyGetCof((P),(i),(R)),(R))
#define ZextRH(R) ZextElt2Pol(ZextRingHbuf(R),(R))

#define ZextQuickMod(e, n, Rg) \
  ZENPolyModuloVeryLazy(ZENElt2Pol(e, Rg), n, ZENRingPol(Rg), ZENRingSubDeg(Rg), ZENRingDef(Rg))

/* Usual squaring routines
***************************/

/* \TEX{
   \title{Squaring in characteristic 2.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZextPolySquareChar2(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;
  
  for(i = ZextPolyDeg(PX, Rg); i; i--) {
    ZextEltSquare(ZextPolyGetCof(RX, 2*i, Rg),
		  ZextPolyGetCof(PX, i, Rg),
		  Rg);
    ZextEltSetToZero(ZextPolyGetCof(RX, 2*i-1, Rg), Rg);
  }
  ZextEltSquare(ZextPolyGetCof(RX, 0, Rg),
		ZextPolyGetCof(PX, 0, Rg),
		Rg);
  ZextPolyDeg(RX, Rg) = 2*ZextPolyDeg(PX, Rg);
  return(0);
}

int _ZextPolySquareCharpPlainVeryLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i, j, k;
  
  for(k = ZextPolyDeg(PX, Rg); k > 0; k--) {

    /* coefficient of X^(2k) */
#if 0
    ZENPolyDeg(ZextPGC(RX, 2*k, Rg), ZENRingDef(Rg)) = -1;
#else
    ZENPolySetToZeroLarge(ZextPGC(RX, 2*k, Rg), ZENRingDef(Rg));
#endif
    for(i = min(ZextPolyDeg(PX, Rg), 2*k);
	i > k;
	i--) {
      j = 2*k - i;
      ZENPolyMultiplyVeryLazy(ZENRingPbuf(Rg), 
		      ZextPGC(PX, i, Rg), ZextPGC(PX, j, Rg), ZENRingDef(Rg));
      ZENPolyAddLarge(ZextPGC(RX, 2*k, Rg), ZENRingPbuf(Rg), ZENRingDef(Rg));
    }
    ZENPolyAddLarge(ZextPGC(RX, 2*k, Rg), ZextPGC(RX, 2*k, Rg), ZENRingDef(Rg));
    ZENPolySquareVeryLazy(ZENRingPbuf(Rg), ZextPGC(PX, k, Rg), ZENRingDef(Rg));
    ZENPolyAddLarge(ZextPGC(RX, 2*k, Rg), ZENRingPbuf(Rg), ZENRingDef(Rg));

    /* coefficient of X^(2k-1) */
#if 0
    ZENPolyDeg(ZextPGC(RX, 2*k-1, Rg), ZENRingDef(Rg)) = -1;
#else
    ZENPolySetToZeroLarge(ZextPGC(RX, 2*k-1, Rg), ZENRingDef(Rg));
#endif
    for(i = min(ZextPolyDeg(PX, Rg), 2*k-1);
	i >= k;
	i--) {
      j = 2*k - i - 1;
      ZENPolyMultiplyVeryLazy(ZENRingPbuf(Rg), ZextPGC(PX, i, Rg), ZextPGC(PX, j, Rg), ZENRingDef(Rg));
      ZENPolyAddLarge(ZextPGC(RX, 2*k-1, Rg), ZENRingPbuf(Rg), ZENRingDef(Rg));
    }
    ZENPolyAddLarge(ZextPGC(RX, 2*k-1, Rg), ZextPGC(RX, 2*k-1, Rg), ZENRingDef(Rg));
  }

  /* coefficient of X^(0) */
  ZENPolySquareVeryLazy(ZextPGC(RX, 0, Rg), ZextPGC(PX, 0, Rg), ZENRingDef(Rg));

  ZENPolyDeg(RX, Rg) = 2*ZENPolyDeg(PX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Squaring in characteristic $p\neq 2$.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZextPolySquareCharpPlain(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{

  int i, j, k;
  
  for(k = ZextPolyDeg(PX, Rg); k > 0; k--) {

    /* coefficient of X^(2k) */
#if 0
    ZENPolyDeg(ZextRH(Rg), ZENRingDef(Rg)) = -1;
#else
    ZENPolySetToZeroLarge(ZextRH(Rg), ZENRingDef(Rg));
#endif
    for(i = min(ZextPolyDeg(PX, Rg), 2*k);
	i > k;
	i--) {
      j = 2*k - i;
      ZENPolyMultiplyVeryLazy(ZENRingPbuf(Rg), 
		      ZextPGC(PX, i, Rg), ZextPGC(PX, j, Rg), ZENRingDef(Rg));
      ZENPolyAddLarge(ZextRH(Rg), ZENRingPbuf(Rg), ZENRingDef(Rg));
    }
    ZENPolyAddLarge(ZextRH(Rg), ZextRH(Rg), ZENRingDef(Rg));
    ZENPolySquareVeryLazy(ZENRingPbuf(Rg), ZextPGC(PX, k, Rg), ZENRingDef(Rg));
    ZENPolyAddLarge(ZextRH(Rg), ZENRingPbuf(Rg), ZENRingDef(Rg));
    ZextQuickMod(ZextPolyGetCof(RX, 2*k, Rg), ZextRH(Rg), Rg);

    /* coefficient of X^(2k-1) */
#if 0
    ZENPolyDeg(ZextRH(Rg), ZENRingDef(Rg)) = -1;
#else
    ZENPolySetToZeroLarge(ZextRH(Rg), ZENRingDef(Rg));
#endif
    for(i = min(ZextPolyDeg(PX, Rg), 2*k-1);
	i >= k;
	i--) {
      j = 2*k - i - 1;
      ZENPolyMultiplyVeryLazy(ZENRingPbuf(Rg), 
		      ZextPGC(PX, i, Rg), ZextPGC(PX, j, Rg), ZENRingDef(Rg));
      ZENPolyAddLarge(ZextRH(Rg), ZENRingPbuf(Rg), ZENRingDef(Rg));
    }
    ZENPolyAddLarge(ZextRH(Rg), ZextRH(Rg), ZENRingDef(Rg));
    ZextQuickMod(ZextPolyGetCof(RX, 2*k-1, Rg), ZextRH(Rg), Rg);
  }

  /* coefficient of X^(0) */
  ZENPolySquareVeryLazy(ZextRH(Rg), ZextPGC(PX, 0, Rg), ZENRingDef(Rg));
  ZextQuickMod(ZextPolyGetCof(RX, 0, Rg), ZextRH(Rg), Rg);

  ZENPolyDeg(RX, Rg) = 2*ZENPolyDeg(PX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Squaring with Karatsuba.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|^2$.}
   }
   */
int _ZextPolySquareKaratsuba(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;
  ZENPoly RBuf;

  if (!(RBuf = ZextPolyCreateLarge(4*ZextPolyDeg(PX, Rg)+1, Rg))) {
    return(ZERR);
  }
  ZextPolySquareKaratsubaVeryLazy(RBuf, PX, Rg);
  for(i = ZextPolyDeg(RBuf, Rg); i >= 0; i--) {
    ZextQuickMod(ZextPolyGetCof(RBuf, i, Rg), ZextPGC(RBuf, i, Rg), Rg);
  }
  ZextPolyAssign(RX, RBuf, Rg);
  ZextPolyFree(RBuf, Rg);
  return(ZEN_HAS_INVERSE);
}

int _ZextPolySquareKaratsubaLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  int i;

  ZextPolySquareKaratsubaVeryLazy(RX, PX, Rg);
  for(i = ZextPolyDeg(RX, Rg); i >= 0; i--) {
    ZextQuickMod(ZextPolyGetCof(RX, i, Rg), ZextPGC(RX, i, Rg), Rg);
  }
  return(ZEN_HAS_INVERSE);
}


static void ZextPolySquareKaratsubaVeryLazyDegree(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  struct zenpoly UP1, UP3;
  ZENPoly up1, up3;
  int m, n, i, j;
  
  /* Initialization ***********/
  up1 = &UP1; up3 = &UP3;
  m = (ZextPolyDeg(PX, Rg)+1)/2; 
  n = ZextPolyDeg(PX, Rg)+1-m;
  
  /* let's go with the recursivity **********/
  
  /* PX- ^ 2 -> RX */
  ZextPolyDeg(up1, Rg) = m-1;
  ZextPolyCof(up1, Rg) = ZextPolyCof(PX, Rg);
  ZextPolyCof(up3, Rg) = ZextPolyCof(RX, Rg);
  ZextPolySquareKaratsubaVeryLazy(up3, up1, Rg);
  
  /* PX- + PX+ -> RX+4*m+2*n-2 */
  for(j = 0, i = 4*m+2*n-2;
      j < m;
      j++, i++) {
    ZENPolyAssign(ZextPGC(RX, i, Rg),
		  ZextPGC(PX, j+m, Rg),
		  ZENRingDef(Rg));
    ZENPolyAdd(ZextPGC(RX, i, Rg), ZextPGC(PX, j, Rg), ZENRingDef(Rg));
  }
  if (n > m) {
    ZENPolyAssign(ZextPGC(RX, i, Rg),
		  ZextPGC(PX, j+m, Rg),
		  ZENRingDef(Rg));
  }
  
  /* (PX- + PX+) ^ 2 -> RX+2*m-1 */
  ZextPolyDeg(up1, Rg) = n-1;
  ZextPolyCof(up1, Rg) = ZextPolyCof(RX, Rg)+4*m+2*n-2;
  ZextPolyCof(up3, Rg) = ZextPolyCof(RX, Rg)+2*m-1;
  ZextPolySquareKaratsubaVeryLazy(up3, up1, Rg);
  
  /* (PX- + PX+) ^ 2 - PX- ^ 2  -> RX+2*m-1 */
  for(j = 0, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZENPolySubtractLarge(ZextPGC(RX, i, Rg), ZextPGC(RX, j, Rg), ZENRingDef(Rg));
  }
  
  /* [ PX- ^ 2, (PX- + PX+) ^ 2 - PX- ^ 2 ] -> RX+m */
  for(j = m, i = 2*m-1; j < 2*m-1; j++, i++) {
    ZENPolyAddLarge(ZextPGC(RX, j, Rg), ZextPGC(RX, i, Rg), ZENRingDef(Rg));
  }
  for(;j < n+2*m-1; j++, i++) {
    ZENPolyAssignLarge(ZextPGC(RX, j, Rg), ZextPGC(RX, i, Rg), ZENRingDef(Rg));
  }
  
  /* PX+ ^ 2 -> RX+ */
  ZextPolyDeg(up1, Rg) = n-1;
  ZextPolyCof(up1, Rg) = ZextPolyCof(PX, Rg)+m;
  ZextPolyCof(up3, Rg) = ZextPolyCof(RX, Rg)+2*m+n-1;
  ZextPolySquareKaratsubaVeryLazy(up3, up1, Rg);

  /* [ PX- ^ 2, (PX- + PX+) ^ 2 - PX- ^2 - PX+ ^ 2 , PX+ ^ 2 ] -> RX */
  for(j = m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZENPolySubtractLarge(ZextPGC(RX, j, Rg), ZextPGC(RX, i, Rg), ZENRingDef(Rg));
  }
  for(j = 2*m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZENPolyAddLarge(ZextPGC(RX, j, Rg), ZextPGC(RX, i, Rg), ZENRingDef(Rg));
  }
  for(; j < 2*m+2*n-1; j++, i++) {
    ZENPolyAssignLarge(ZextPGC(RX, j, Rg), ZextPGC(RX, i, Rg), ZENRingDef(Rg));
  }

  ZextPolyDeg(RX, Rg) = 2*ZextPolyDeg(PX, Rg);
}

/* \TEX{
   \title{VeryLazy Squaring with Karatsuba.}
   \input{Two |ZENPoly| |RX| and |PX| of a finite ring |Rg|.}
   \output{0 for compatibylity.}
   \side{$|RX| = |PX|^2$.}
   \note{One must have $|RX|\neq|PX|$. |PX| is destroyed at the return.}
   }
   */
int ZextPolySquareKaratsubaVeryLazy(RX, PX, Rg)
     ZENPoly RX, PX;
     ZENRing Rg;
{
  if (ZextPolyDeg(PX, Rg) < ZEXT_POLY_SQUARE_CUTOFF) {
    ZextPolySquareCharpPlainVeryLazy(RX, PX, Rg);
    return(ZEN_HAS_INVERSE);
  }
  ZextPolySquareKaratsubaVeryLazyDegree(RX, PX, Rg);
  return(ZEN_HAS_INVERSE);
}



/* General multiplication routines
**********************************/

/* \TEX{
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an error occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   }
   */
int _ZextPolyMultiply(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZENPoly RBuf;
  int n, m, q, r, L, k;

  if (ZextPolyDeg(PX, Rg) > ZextPolyDeg(QX, Rg)) {
    n = ZextPolyDeg(PX, Rg)+1; m = ZextPolyDeg(QX, Rg)+1;
    q = n/m; r = n%m;
  }
  else {
    n = ZextPolyDeg(QX, Rg)+1; m = ZextPolyDeg(PX, Rg)+1;
    q = n/m; r = n%m;
  }
  switch(q) {
  case 1:
    L = r ? 4*m-4+2*r : 4*m-3;
    break;
  case 2:
    L = 2*r > (m+1) ? 5*m-5+2*r : 6*m-4;
    break;
  default:
    L = 2*r > m ?  (3+q)*m-4+2*r: (4+q)*m-4;
    break;
  }
  if (!(RBuf = ZextPolyCreateLarge(L, Rg))) {
    return(ZERR);
  }
  ZextPolyMultiplyVeryLazy(RBuf, PX, QX, Rg);
  for(k = ZextPolyDeg(RBuf, Rg); k >= 0; k--) {
    ZextQuickMod(ZextPolyGetCof(RBuf, k, Rg), ZextPGC(RBuf, k, Rg), Rg);
  }
  ZextPolyAssign(RX, RBuf, Rg);
  ZextPolyFree(RBuf, Rg);

  return(ZEN_HAS_INVERSE);
}

int _ZextPolyMultiplyLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int k;

  ZextPolyMultiplyVeryLazy(RX, PX, QX, Rg);
  for(k = ZextPolyDeg(RX, Rg); k >= 0; k--) {
    ZextQuickMod(ZextPolyGetCof(RX, k, Rg), ZextPGC(RX, k, Rg), Rg);
  }
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Very Lazy multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0 for compatibylity.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX|\neq|PX|\neq|QX|$.}
   }
   */
int _ZextPolyMultiplyVeryLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  struct zenpoly sxzp, rxzp, pxzp, qxzp;
  ZENPoly sx, rx, px, qx;
  int i, dp, dq, quo, rem;

  /* Switch
  **********/
  if (ZextPolyDeg(PX, Rg) == ZextPolyDeg(QX, Rg)) {
    ZextPolyMultiplyKaratsubaVeryLazy(RX, PX, QX, Rg);
    return(ZEN_HAS_INVERSE);
  }
  if (ZextPolyDeg(PX, Rg) > ZextPolyDeg(QX, Rg)) {
    if ((ZextPolyDeg(QX, Rg)) <= ZEXT_POLY_MULTIPLY_CUTOFF) {
      ZextPolyMultiplyPlainVeryLazy(RX, PX, QX, Rg);
      return(ZEN_HAS_INVERSE);
    }
    px = &pxzp;
    ZextPolyCof(px, Rg) = ZextPolyCof(PX, Rg);
    dp = ZextPolyDeg(PX, Rg);

    qx = &qxzp;
    ZextPolyCof(qx, Rg) = ZextPolyCof(QX, Rg);
    dq = ZextPolyDeg(QX, Rg);
  }
  else {
    if (ZextPolyDeg(PX, Rg) <= ZEXT_POLY_MULTIPLY_CUTOFF) {
      ZextPolyMultiplyPlainVeryLazy(RX, PX, QX, Rg);
      return(ZEN_HAS_INVERSE);
    }
    px = &pxzp;
    ZextPolyCof(px, Rg) = ZextPolyCof(QX, Rg);
    dp = ZextPolyDeg(QX, Rg);

    qx = &qxzp;
    ZextPolyCof(qx, Rg) = ZextPolyCof(PX, Rg);
    dq = ZextPolyDeg(PX, Rg);
  }

  /* We have to cut px in pieces 
    **********************************/  
  /* Initialization */
  ZextPolyDeg(RX, Rg) = dp+dq;

  rx = &rxzp;
  ZextPolyCof(rx, Rg) = ZextPolyCof(RX, Rg);

  sx = &sxzp;
  ZextPolyCof(sx, Rg) = ZextPolyCof(RX, Rg);

  quo = (dp+1)/(dq+1);
  rem = (dp+1)%(dq+1);

  /* First multiplication */
  ZextPolyDeg(px, Rg) = dq;
  ZextPolyDeg(qx, Rg) = dq;
  ZextPolyMultiplyKaratsubaVeryLazy(rx, qx, px, Rg);
  if ((quo == 1) && !rem) {
    return(ZEN_HAS_INVERSE);
  }

  /* Second multiplication */
  ZextPolyCof(px, Rg) += dq+1;
  ZextPolyDeg(px, Rg) = (quo == 1 ? rem-1: dq);
  ZextPolyCof(rx, Rg) += 2*dq+1;
  if (quo == 1) {
    ZextPolyMultiplyVeryLazy(rx, qx, px, Rg);
  }
  else {
    ZextPolyMultiplyKaratsubaVeryLazy(rx, qx, px, Rg);
  }
  ZextPolyDeg(sx, Rg) = dq-1;
  ZextPolyCof(sx, Rg) += dq+1;
  ZextPolyAddLarge(sx, rx, Rg);
  if (quo == 1) {
    return(ZEN_HAS_INVERSE);
  }

  for(i = 2; i < quo; i++) {
    ZextPolyCof(px, Rg) += (dq+1);
    ZextPolyCof(rx, Rg) += (dq+1);

    ZextPolyMultiplyKaratsubaVeryLazy(rx, qx, px, Rg);
    
    ZextPolyDeg(sx, Rg) = dq-1;
    ZextPolyCof(sx, Rg) += dq+1;
    ZextPolyAddLarge(sx, rx, Rg);
  }

  if (!rem) {
    return(ZEN_HAS_INVERSE);
  }

  ZextPolyDeg(px, Rg) = rem-1;
  ZextPolyCof(px, Rg) += (dq+1);
  ZextPolyCof(rx, Rg) += (dq+1);
  ZextPolyMultiplyVeryLazy(rx, qx, px, Rg);
    
  ZextPolyDeg(sx, Rg) = dq-1;
  ZextPolyCof(sx, Rg) += dq+1;
  ZextPolyAddLarge(sx, rx, Rg);

  return(ZEN_HAS_INVERSE);
}

/* Specific classical multiplication routines
*********************************************/

int ZextPolyMultiplyPlainVeryLazy(RX, QX, PX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i, j, k;

  if (ZextPolyIsZero(PX, Rg) || ZextPolyIsZero(QX, Rg)) {
    i = ZextPolyLgt(RX, Rg);
    ZextPolyLgt(RX, Rg) = 0;
    ZextPolySetToZeroLarge(RX, Rg);
    ZextPolyLgt(RX, Rg) = i;
    return(ZEN_HAS_INVERSE);
  }

  for(k = ZextPolyDeg(PX, Rg)+ZextPolyDeg(QX, Rg);
      k >= 0;
      k--) {
#if 0
    ZENPolyDeg(ZextPGC(RX, k, Rg), ZENRingDef(Rg)) = -1;
#else
    ZENPolySetToZeroLarge(ZextPGC(RX, k, Rg), ZENRingDef(Rg));
#endif
    for(i = min(ZextPolyDeg(PX, Rg), k);
	i >= max(0, k - ZextPolyDeg(QX, Rg));
	i--) {
      j = k - i;
      ZENPolyMultiplyVeryLazy(ZENRingPbuf(Rg), 
		      ZextPGC(PX, i, Rg), ZextPGC(QX, j, Rg),
		      ZENRingDef(Rg));
      ZENPolyAddLarge(ZextPGC(RX, k, Rg), ZENRingPbuf(Rg), ZENRingDef(Rg));
    }
  }
  ZextPolyDeg(RX, Rg) = ZextPolyDeg(PX, Rg) + ZextPolyDeg(QX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* \TEX{
   \title{Multiplication.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{0.}
   \side{$|RX| = |PX|\times|QX|$.}
   }
   */
int _ZextPolyMultiplyPlain(RX, QX, PX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i, j, k;

  if (ZextPolyIsZero(PX, Rg) || ZextPolyIsZero(QX, Rg)) {
    ZENPolySetToZero(RX, Rg);
    return(ZEN_HAS_INVERSE);
  }

  for(k = ZextPolyDeg(PX, Rg)+ZextPolyDeg(QX, Rg);
      k >= 0;
      k--) {
#if 0
    ZENPolyDeg(ZextRH(Rg), ZENRingDef(Rg)) = -1;
#else
    ZENPolySetToZeroLarge(ZextRH(Rg), ZENRingDef(Rg));
#endif
    for(i = min(ZextPolyDeg(PX, Rg), k);
	i >= max(0, k - ZextPolyDeg(QX, Rg));
	i--) {
      j = k - i;
      ZENPolyMultiplyVeryLazy(ZENRingPbuf(Rg), 
		      ZextPGC(PX, i, Rg), ZextPGC(PX, j, Rg),
		      ZENRingDef(Rg));
      ZENPolyAddLarge(ZextRH(Rg), ZENRingPbuf(Rg), ZENRingDef(Rg));
    }
    ZextQuickMod(ZextPolyGetCof(RX, k, Rg), ZextRH(Rg), Rg);
  }
  ZextPolyDeg(RX, Rg) = ZextPolyDeg(PX, Rg) + ZextPolyDeg(QX, Rg);
  return(ZEN_HAS_INVERSE);
}

/* Specific Karatsuba multiplication routines
*********************************************/

/* \TEX{
   \title{Polynomial multiplication with the Karatsuba method.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \output{|ZENERR| if an errpr occured, 0 otherwise.}
   \side{$|RX| = |PX|\times|QX|$.}
   }
   */
int _ZextPolyMultiplyKaratsuba(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i;
  ZENPoly QBuf, RBuf;
  ZENElt elt;

  if (ZextPolyDeg(PX, Rg) >= ZextPolyDeg(QX, Rg)) {
    if (!(QBuf = ZextPolyCreate(ZextPolyDeg(PX, Rg), Rg))) {
      return(ZERR);
    }
    ZextPolyAssign(QBuf, QX, Rg);
    for(i = ZextPolyDeg(QX, Rg)+1;
	i <= ZextPolyDeg(PX, Rg);
	i++) {
      elt = ZextPolyGetCoeffPtr(QBuf, i, Rg);
      ZextEltSetToZero(elt, Rg);
      ZextPolySetCoeff(QBuf, i, elt, Rg);
    }

    if (!(RBuf = ZextPolyCreateLarge(4*ZextPolyDeg(PX, Rg)+1, Rg))) {
      return(ZERR);
    }

    ZextPolyMultiplyKaratsubaVeryLazy(RBuf, PX, QBuf, Rg);
  }
  else {
    if (!(QBuf = ZextPolyCreate(ZextPolyDeg(QX, Rg), Rg))) {
      return(ZERR);
    }
    ZextPolyAssign(QBuf, PX, Rg);
    for(i = ZextPolyDeg(PX, Rg)+1;
	i <= ZextPolyDeg(QX, Rg);
	i++) {
      elt = ZextPolyGetCoeffPtr(QBuf, i, Rg);
      ZextEltSetToZero(elt, Rg);
      ZextPolySetCoeff(QBuf, i, elt, Rg);
    }

    if (!(RBuf = ZextPolyCreateLarge(4*ZextPolyDeg(PX, Rg)+1, Rg))) {
      return(ZERR);
    }

    ZextPolyMultiplyKaratsubaVeryLazy(RBuf, QX, QBuf, Rg);
  }
  for(i = ZextPolyDeg(RBuf, Rg); i >= 0; i--) {
    ZextQuickMod(ZextPolyGetCof(RBuf, i, Rg), ZextPGC(RBuf, i, Rg), Rg);
  }
  ZextPolyAssign(RX, RBuf, Rg);
  ZextPolyFree(QBuf, Rg); ZextPolyFree(RBuf, Rg);
  return(ZEN_HAS_INVERSE);
}

int _ZextPolyMultiplyKaratsubaLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  int i;
  ZENElt elt;

  if (ZextPolyDeg(PX, Rg) >= ZextPolyDeg(QX, Rg)) {
    for(i = ZextPolyDeg(QX, Rg)+1;
	i <= ZextPolyDeg(PX, Rg);
	i++) {
      elt = ZextPolyGetCoeffPtr(QX, i, Rg);
      ZextEltSetToZero(elt, Rg);
      ZextPolySetCoeff(QX, i, elt, Rg);
    }
  }
  else {
    for(i = ZextPolyDeg(PX, Rg)+1;
	i <= ZextPolyDeg(QX, Rg);
	i++) {
      elt = ZextPolyGetCoeffPtr(PX, i, Rg);
      ZextEltSetToZero(elt, Rg);
      ZextPolySetCoeff(PX, i, elt, Rg);
    }
  }
  ZextPolyMultiplyKaratsubaVeryLazy(RX, PX, QX, Rg);
  for(i = ZextPolyDeg(RX, Rg); i >= 0; i--) {
    ZextQuickMod(ZextPolyGetCof(RX, i, Rg), ZextPGC(RX, i, Rg), Rg);
  }
  return(ZEN_HAS_INVERSE);
}

static void ZextPolyMultiplyKaratsubaVeryLazyDegree0(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZENPolyMultiplyVeryLazy(ZextPGC(RX, 0, Rg),
		   ZextPGC(PX, 0, Rg),
		   ZextPGC(QX, 0, Rg), ZENRingDef(Rg));
  ZextPolyDeg(RX, Rg) = 0;
}

static void ZextPolyMultiplyKaratsubaVeryLazyDegree1(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  ZENPolyAssign(ZextPGC(RX, 0, Rg), ZextPGC(PX, 0, Rg), ZENRingDef(Rg));
  ZENPolyAdd(ZextPGC(RX, 0, Rg), ZextPGC(PX, 1, Rg), ZENRingDef(Rg));
  
  ZENPolyAssign(ZextPGC(RX, 2, Rg), ZextPGC(QX, 0, Rg), ZENRingDef(Rg));
  ZENPolyAdd(ZextPGC(RX, 2, Rg), ZextPGC(QX, 1, Rg), ZENRingDef(Rg));

  ZENPolyMultiplyVeryLazy(ZextPGC(RX, 1, Rg),
		   ZextPGC(RX, 0, Rg),
		   ZextPGC(RX, 2, Rg), ZENRingDef(Rg));
  
  ZENPolyMultiplyVeryLazy(ZextPGC(RX, 0, Rg),
		   ZextPGC(PX, 0, Rg),
		   ZextPGC(QX, 0, Rg), ZENRingDef(Rg));
  
  ZENPolyMultiplyVeryLazy(ZextPGC(RX, 2, Rg),
		   ZextPGC(PX, 1, Rg),
		   ZextPGC(QX, 1, Rg), ZENRingDef(Rg));
  
  ZENPolySubtractLarge(ZextPGC(RX, 1, Rg), ZextPGC(RX, 0, Rg), ZENRingDef(Rg));
  ZENPolySubtractLarge(ZextPGC(RX, 1, Rg), ZextPGC(RX, 2, Rg), ZENRingDef(Rg));

  ZextPolyDeg(RX, Rg) = 2;
}

static void ZextPolyMultiplyKaratsubaVeryLazyDegree(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  struct zenpoly UP1, UP2, UP3;
  ZENPoly up1, up2, up3;
  int m, n, i, j;
  
  /* Initialization ***********/
  up1 = &UP1; up2 = &UP2; up3 = &UP3;
  up2->deg = 0;
  m = (ZextPolyDeg(PX, Rg)+1)/2; 
  n = ZextPolyDeg(PX, Rg)+1-m;
  
  /* let's go with the recursivity **********/
  
  /* PX- * QX- -> RX */
  ZextPolyDeg(up1, Rg) = m-1;
  ZextPolyCof(up1, Rg) = ZextPolyCof(PX, Rg);
  ZextPolyCof(up2, Rg) = ZextPolyCof(QX, Rg);
  ZextPolyCof(up3, Rg) = ZextPolyCof(RX, Rg);
  ZextPolyMultiplyKaratsubaVeryLazy(up3, up1, up2, Rg);
  
  /* PX- + PX+ -> RX+4*m+2*n-2 */
  for(j = 0, i = 4*m+2*n-2;
      j < m;
      j++, i++) {
    ZENPolyAssign(ZextPGC(RX, i, Rg),
		  ZextPGC(PX, j+m, Rg),
		  ZENRingDef(Rg));
    ZENPolyAdd(ZextPGC(RX, i, Rg), ZextPGC(PX, j, Rg), ZENRingDef(Rg));
  }
  if (n > m) {
    ZENPolyAssign(ZextPGC(RX, i, Rg),
		  ZextPGC(PX, j+m, Rg),
		  ZENRingDef(Rg));
  }
  
  /* QX- + QX+ -> RX+4*m+3*n-2 */
  for(j = 0, i = 4*m+3*n-2;
      j < m;
      j++, i++) {
    ZENPolyAssign(ZextPGC(RX, i, Rg),
		  ZextPGC(QX, j+m, Rg),
		  ZENRingDef(Rg));
    ZENPolyAdd(ZextPGC(RX, i, Rg), ZextPGC(QX, j, Rg), ZENRingDef(Rg));
  }
  if (n > m) {
    ZENPolyAssign(ZextPGC(RX, i, Rg),
		  ZextPGC(QX, j+m, Rg),
		  ZENRingDef(Rg));
  }
  
  /* (PX- + PX+) * (QX- + QX+) -> RX+2*m-1 */
  ZextPolyDeg(up1, Rg) = n-1;
  ZextPolyCof(up1, Rg) = ZextPolyCof(RX, Rg)+4*m+2*n-2;
  ZextPolyCof(up2, Rg) = ZextPolyCof(RX, Rg)+4*m+3*n-2;
  ZextPolyCof(up3, Rg) = ZextPolyCof(RX, Rg)+2*m-1;
  ZextPolyMultiplyKaratsubaVeryLazy(up3, up1, up2, Rg);
  
  /* (PX- + PX+) * (QX- + QX+) - PX- * QX-  -> RX+2*m-1 */
  for(j = 0, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZENPolySubtractLarge(ZextPGC(RX, i, Rg), ZextPGC(RX, j, Rg), ZENRingDef(Rg));
  }
  
  /* [ PX- * QX-, (PX- + PX+) * (QX- + QX+) - PX- * QX- ] -> RX+m */
  for(j = m, i = 2*m-1;
      j < 2*m-1;
      j++, i++) {
    ZENPolyAddLarge(ZextPGC(RX, j, Rg), ZextPGC(RX, i, Rg), ZENRingDef(Rg));
  }
  for(;j < n+2*m-1; j++, i++) {
    ZENPolyAssignLarge(ZextPGC(RX, j, Rg), ZextPGC(RX, i, Rg), ZENRingDef(Rg));
  }
  
  /* PX+ * QX+ -> RX+ */
  ZextPolyDeg(up1, Rg) = n-1;
  ZextPolyCof(up1, Rg) = ZextPolyCof(PX, Rg)+m;
  ZextPolyCof(up2, Rg) = ZextPolyCof(QX, Rg)+m;
  ZextPolyCof(up3, Rg) = ZextPolyCof(RX, Rg)+2*m+n-1;
  ZextPolyMultiplyKaratsubaVeryLazy(up3, up1, up2, Rg);

  /* [ PX- * QX-, (PX- + PX+) * (QX- + QX+) - PX- * QX- - PX+ * QX+ , PX+ * QX+ ] -> RX */
  for(j = m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZENPolySubtractLarge(ZextPGC(RX, j, Rg), ZextPGC(RX, i, Rg), ZENRingDef(Rg));
  }
  for(j = 2*m, i = 2*m+n-1;
      j < 2*m+n-1;
      j++, i++) {
    ZENPolyAddLarge(ZextPGC(RX, j, Rg), ZextPGC(RX, i, Rg), ZENRingDef(Rg));
  }
  for(; j < 2*m+2*n-1; j++, i++) {
    ZENPolyAssignLarge(ZextPGC(RX, j, Rg), ZextPGC(RX, i, Rg), ZENRingDef(Rg));
  }

  ZextPolyDeg(RX, Rg) = ZextPolyDeg(PX, Rg)+ZextPolyDeg(QX, Rg);
}

/* \TEX{
   \title{Very Lazy polynomial multiplication with the Karatsuba method.}
   \input{Three |ZENPoly| |RX|, |PX| and |QX| of a finite ring |Rg|.}
   \side{$|RX| = |PX|\times|QX|$.}
   \note{One must have $|RX| \neq|PX|,|QX|$. The degrees of |PX| and
   |QX| must be equal. |RX| must be four times larger than |PX|.}
   }
   */
int ZextPolyMultiplyKaratsubaVeryLazy(RX, PX, QX, Rg)
     ZENPoly RX, PX, QX;
     ZENRing Rg;
{
  switch (ZextPolyDeg(PX, Rg)) {

    /* compute the product of 2 polynoms whose degree is 0 */
  case 0:
    ZextPolyMultiplyKaratsubaVeryLazyDegree0(RX, PX, QX, Rg);
    break;

    /* compute the product of 2 polynoms whose degree is 1 */
  case 1: 
    ZextPolyMultiplyKaratsubaVeryLazyDegree1(RX, PX, QX, Rg); 
    break;

    /* compute the product of 2 polynoms whose degree is greater than 2*/
  default : 
    ZextPolyMultiplyKaratsubaVeryLazyDegree(RX, PX, QX, Rg); 
  }
  return(ZEN_HAS_INVERSE);
}

/* Division routines
***********************/

/* \TEX{
   \title{VeryLazy division.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZEN_NO_INVERSE| if
   the leading coefficient of |QX| is a factor of a modulo, |ZENERR| for an
   error}
   \note{|PX| is destroyed. |ZENRingFact(Rg)| is filled with a factor
   of a modulo if |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZextPolyDivideLazy(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  ZextPolyLazyToVeryLazy(PX, PX, Rg);
  return(ZextPolyDivideVeryLazy(RX, MX, PX, QX, Rg));
}

int _ZextPolyDivideVeryLazy(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int i, j, deg, IsNotOne;

  if (!ZextEltIsOne(ZextPolyGetCof(QX, ZextPolyDeg(QX, Rg), Rg), Rg)) {
    IsNotOne = 1;
    
    switch(ZextEltInverse(ZENRingIbuf(Rg),ZextPolyGetCof(QX, ZextPolyDeg(QX, Rg), Rg), Rg)) {
    case ZENERR:
      return(ZERR);
    case ZEN_NO_INVERSE:
      return(ZEN_NO_INVERSE);
    }
  }
  else {
    IsNotOne = 0;
  }

  for(i = ZextPolyDeg(PX, Rg);
      i >= ZextPolyDeg(QX, Rg);
      i--) {
    ZextQuickMod(ZextPolyGetCof(PX, i, Rg), ZextPGC(PX, i, Rg), Rg);
      
    if (IsNotOne) {
      ZextEltMultiply(ZextPolyGetCof(PX, i, Rg),
		      ZENRingIbuf(Rg),
		      ZextPolyGetCof(PX, i, Rg),
		      Rg);
    }
    
    for(j =  ZextPolyDeg(QX, Rg)-1;
	j >= 0;
	j--) {
      ZENPolyMultiplyVeryLazy(ZENRingPbuf(Rg),
		      ZextPGC(PX, i, Rg), ZextPGC(QX, j, Rg), ZENRingDef(Rg));
      ZENPolySubtractLarge(ZextPGC(PX, i+j-ZextPolyDeg(QX, Rg), Rg),
		      ZENRingPbuf(Rg), ZENRingDef(Rg));
    }
  }
  deg =  ZextPolyDeg(QX, Rg);
  for(i = ZextPolyDeg(PX, Rg); i >= ZextPolyDeg(QX, Rg); i--) {
    ZextEltAssign(ZextPolyGetCof(RX, i-ZextPolyDeg(QX, Rg), Rg),
		  ZextPolyGetCof(PX, i, Rg),
		  Rg);
  }
  ZextPolyDeg(RX, Rg) = ZextPolyDeg(PX, Rg)-ZextPolyDeg(QX, Rg);
  if (ZextPolyDeg(RX, Rg) < 0) {
    ZextPolySetToZero(RX, Rg);
  }
  deg = -1;
  for(; i >= 0; i--) {
    ZextQuickMod(ZextPolyGetCof(PX, i, Rg), ZextPGC(PX, i, Rg), Rg);
    if (!ZextEltIsZero(ZextPolyGetCof(PX, i, Rg), Rg)) {
      ZextEltAssign(ZextPolyGetCof(MX, i, Rg), ZextPolyGetCof(PX, i, Rg), Rg);
      if (deg < 0) {
	deg = i;
      }
    }
    else if (deg >= 0) {
      ZextEltSetToZero(ZextPolyGetCof(MX, i, Rg), Rg);
    }
  }
  if (deg < 0) {
    ZextPolySetToZero(MX, Rg);
  }
  else {
    ZextPolyDeg(MX, Rg) = deg;
  }
  return(ZEN_HAS_INVERSE);
}


/* \TEX{
   \title{Divide.}
   \input{Two polynomials |PX| and |QX| to divide in a finite ring
   |Rg|. |RX| will be the quotient and |MX| the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZEN_NO_INVERSE| if
   a factor of a modulo was dicovered, |ZENERR| for an error}
   \side{|ZENRingFact(Rg)| is filled with a factor of a modulo if
   |ZEN_NO_INVERSE| is returned.}
   }
   */
int _ZextPolyDivide(RX, MX, PX, QX, Rg)
     ZENPoly RX, MX, PX, QX;
     ZENRing Rg;
{
  int i, ret;
  ZENPoly Pbuf;

  if (!(Pbuf = ZextPolyCreateLarge(ZextPolyDeg(PX, Rg), Rg))) {
    return(ZERR);
  }
  ZextPolyLazyToVeryLazy(Pbuf, PX, Rg);
  ret = ZextPolyDivideVeryLazy(RX, MX, Pbuf, QX, Rg);
  ZextPolyFree(Pbuf, Rg);
  return(ret);
}


/* \TEX{
   \title{VeryLazy mudulo computation.}
   \input{A polynomial |PX| to divide by
   |QX|$=X^{\mbox{degree}(QX)}+sum_{i=0}^{|k|}q_iX^i$} in a finite
   ring |Rg|. |MX| will be the remainder.}
   \output{|ZEN_HAS_INVERSE| if no error occured, |ZENERR| for an error}
   \note{|PX| is destroyed.}
   */
int _ZextPolyModuloVeryLazy(MX, PX, QX, k, Rg)
     ZENPoly MX, PX, QX;
     int k;
     ZENRing Rg;
{
  int i, j, deg, IsNotOne;

  for(i = ZextPolyDeg(PX, Rg);
      i >= ZextPolyDeg(QX, Rg);
      i--) {
    ZextQuickMod(ZextPolyGetCof(PX, i, Rg), ZextPGC(PX, i, Rg), Rg);
    for(j = k;
	j >= 0;
	j--) {
      ZENPolyMultiplyVeryLazy(ZENRingPbuf(Rg),
		      ZextPGC(PX, i, Rg), ZextPGC(QX, j, Rg), ZENRingDef(Rg));
      ZENPolySubtractLarge(ZextPGC(PX, i+j-ZextPolyDeg(QX, Rg), Rg),
		      ZENRingPbuf(Rg), ZENRingDef(Rg));
    }
  }

  deg = -1;
  for(; i >= 0; i--) {
    ZextQuickMod(ZextPolyGetCof(MX, i, Rg), ZextPGC(PX, i, Rg), Rg);
    if (!ZextEltIsZero(ZextPolyGetCof(MX, i, Rg), Rg)) {
      if (deg < 0) {
	deg = i;
      }
    }
  }
  if (deg < 0) {
    ZextPolySetToZero(MX, Rg);
  }
  else {
    ZextPolyDeg(MX, Rg) = deg;
  }
  return(ZEN_HAS_INVERSE);
}

/* Other routines
*********************************************/


/* \TEX{
   \title{Multiplication by a scalar.}
   \input{Two polynomials |PX|, |RX| and a |ZENElt| |e| of a finite ring |Rg|.}
   }
	 */
void _ZextPolyMultiplyScalar(RX, PX, e, Rg)
     ZENPoly RX, PX;
     ZENElt e;
     ZENRing Rg;
{
  int i;

  if (ZextEltIsZero(e, Rg)) {
    ZextPolySetToZero(RX, Rg);
    return;
  }
  if (ZextEltIsOne(e, Rg)) {
    ZextPolyAssign(RX, PX, Rg);
    return;
  }
  for(i = 0; i <= ZextPolyDeg(PX, Rg); i++) {
    ZextEltMultiply(ZextPolyGetCof(RX, i, Rg),
		    e,
		    ZextPolyGetCof(PX, i, Rg),
		    Rg);
  }
  ZextPolyDeg(RX, Rg) = ZextPolyDeg(PX, Rg);
}

/* \TEX{
   \title{Scalar product.}
   \input{Two polynomials |PX|, |RX| and a pointer to a |ZENElt| |e| of a
   finite ring |Rg|.}
   \side{$|*e| = \sum p_iq_i$}
   }
   */
void _ZextPolyDot(e, PX, QX, Rg)
     ZENElt *e;
     ZENPoly PX, QX;
     ZENRing Rg;
{
  int i, deg;

  ZENPolyMultiplyVeryLazy(ZextRH(Rg), ZextPGC(PX, 0, Rg), ZextPGC(QX, 0, Rg), ZENRingDef(Rg));

  deg = ZextPolyDeg(PX, Rg) < ZextPolyDeg(QX, Rg) ?
    ZextPolyDeg(PX, Rg) :  ZextPolyDeg(QX, Rg);

  for(i = 1; i <= deg; i++) {
    ZENPolyMultiplyVeryLazy(ZENRingPbuf(Rg), ZextPGC(PX, i, Rg), ZextPGC(QX, i, Rg), ZENRingDef(Rg));
    ZENPolyAddLarge(ZextRH(Rg), ZENRingPbuf(Rg), ZENRingDef(Rg));
  }
  ZextQuickMod(*e, ZextRH(Rg), Rg);
}


