/*
 * zenfact/c/bench.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zen.h"
#include "zenfact.h"

static char verbose;
extern char *optarg;
extern int optind, opterr;

#define RG_MAX 1000
#define NB_TESTED 1989		/* 1989 */
#define DEG_TESTED 100		/* 100 */

/*
   \TEX{
This program has two goals. First, it shows why the Berlekamp algorithm is used by default in |ZENFACT| for testing irreductibility of a polynomial. Second, it allows users to make benches between several representations of a same ring. The bench function consist in testing irreductibility of polynomials, since this kind of algorithm uses all kinds of functions.
   } 
   */


/***********/

static ZENPoly *ZENPolyRandomArray(n, deg, Rg)
     int n,deg;
     ZENRing Rg;
{
  int i,nnb;
  ZENPoly *a;

  nnb=n+n/10+1;
  a = (ZENPoly *)ZENMalloc(sizeof(ZENPoly)*nnb*2);
  printf("Generating %d random composed polynomials\n",n);
  for(i = 0; i < n; i++) {
    ZENPolyAlloc(a[i], deg, Rg);
    for(;;)
      {
	ZENPolySetRandom(a[i], deg, Rg);
	switch(_ZENPolyIsNotIrreducible(a[i],Berlekamp,Rg))
	  {
	  case ZENPOLY_IS_IRREDUCIBLE:	  
	    if(verbose) printf("*");
	    break;
	  case ZENPOLY_HAS_ZERO_ROOT:
	    if(verbose) printf("0");
	    break;
	  case ZENPOLY_HAS_ONE_ROOT:
	    if(verbose) printf("1");
	    break;
	  case ZENPOLY_HAS_REPEATED_FACTORS:
	    if(verbose) printf("x");
	    break;
	  case ZENPOLY_IS_P_POWER:
	    if(verbose) printf("^");
	    break;
	  case ZENPOLY_IS_COMPOSED :
	    if(verbose) printf("+");
	    goto CONT;
	    
	  default :
	    if(verbose) printf("?");
	    break;
	  }
	if(verbose) fflush(stdout);
      }
  CONT:
    if(verbose) printf("\nComposed polynomial %d\n",i);
  }
  printf("Generating %d random irreducible polynomials\n",nnb-i);
  for(; i < nnb; i++) {
    ZENPolyAlloc(a[i], deg, Rg);
    for(;;)
      {
	ZENPolySetRandom(a[i], deg, Rg);
	switch(_ZENPolyIsNotIrreducible(a[i],Berlekamp,Rg))
	  {
	  case ZENPOLY_IS_IRREDUCIBLE:	  
	    if(verbose) printf("*");
	    goto CONT2;
	  case ZENPOLY_HAS_ZERO_ROOT:
	    if(verbose) printf("0");
	    break;
	  case ZENPOLY_HAS_ONE_ROOT:
	    if(verbose) printf("1");
	    break;
	  case ZENPOLY_HAS_REPEATED_FACTORS:
	    if(verbose) printf("x");
	    break;
	  case ZENPOLY_IS_P_POWER:
	    if(verbose) printf("^");
	    break;
	  case ZENPOLY_IS_COMPOSED :
	    if(verbose) printf("+");
	    break;
	  
	  default :
	    if(verbose) printf("?");
	    break;
	  }
	if(verbose) fflush(stdout);
      }
  CONT2:
    if(verbose) printf("\nIrreducible polynomial %d\n",i);
  }
  return(a);
}
/***********/

static void ZENPolyFreeArray(a, n, Rg)
     ZENPoly *a;
     int n;
     ZENRing Rg;
{
  int i;
  
  n=n+n/10+1;
  for(i = 0; i < n; i++) {
    ZENPolyFree(a[i], Rg);
  }
  ZENFree(a);
}
/***********/

static void Stat(ttab,nb)
     double *ttab[2];
     int nb;
{
  int i,cb,cd,nnb;

  nnb=nb+nb/10+1;
  cb=cd=0;
  for(i=0;i<nb;i++)
    {
      if(ttab[0][i]>ttab[1][i])
	cd++;
      else if(ttab[0][i]<ttab[1][i])
	cb++;
#if 0
      if(verbose) printf("B: %f DDF: %f\n",ttab[0][i],ttab[1][i]);
#endif
    }
  printf("Composed polynomials:\nBerlekamp: %.0f%% DDF: %.0f%%\n",(cb*100.0/nb),(cd*100.0/nb));
  cb=cd=0;
  for(;i<nnb;i++)
    {
      if(ttab[0][i]>ttab[1][i])
	cd++;
      else if(ttab[0][i]<ttab[1][i])
	cb++;
#if 0
      if(verbose) printf("B: %f DDF: %f\n",ttab[0][i],ttab[1][i]);
#endif
    }
  printf("Irreducible polynomials:\nBerlekamp: %.0f%% DDF: %.0f%%\n",(cb*100.0/(nnb-nb)),(cd*100.0/(nnb-nb)));
}
/***********/

int main(argc, argv)
     int argc;
     char *argv[];
{
  char *str;
  int i, nb, rg;
  ZENRing Rg;
  ZENPrc prc;
  int deg;
  double tt;
  int c;
  FILE *fin;
  int degmin=2,degmax;
  double *ttab[2];
  ZENPoly *Ptab;
  int nnb;
	
  
  setbuf(stdout, NULL);

  verbose = 0; nb = NB_TESTED; deg = DEG_TESTED;
  fin = stdin;
  while ((c = getopt(argc, argv, "n:d:vf:")) != EOF)
    switch (c) {
      /* An input file */
    case 'f':
      if (!(fin = fopen(optarg, "r"))) {
	fprintf(stderr, "Can't open %s, reading data on stdin\n", optarg);
	fin = stdin;
      }
      break;

      /* verbose */
    case 'v':
      verbose = 1;
      break;

      /* How many time ? */
    case 'n':
      nb = atoi(optarg);
      break;
      
      /* Max degree */
    case 'd':
      deg = atoi(optarg);
      break;
    }


  Rg=ZENRingDefine(stdin,stdout);
  if(Rg==ZENNULL)
    {
      ZENError();
      exit(-1);
    }
#if 0
  ZENRingPrintDebug(stdout,Rg);
#endif
  ZENPrcSetAll(prc);
  tt=runtime();
  if(ZENRingRecursiveAddPrc(Rg,prc)<0)
    {
      ZENError();
      exit(-1);
    }
  printf("All precomputations done in %.2f s\n",(runtime()-tt)/1000.);
  ZENRingPrintDebug(stdout,Rg);

#if 1
  tt=0.0;
  degmax=deg;
  nnb=nb+nb/10+1;
  ttab[0]=(double *)ZENMalloc(nnb*sizeof(double));
  ttab[1]=(double *)ZENMalloc(nnb*sizeof(double));
  if((ttab[0]==NULL)||(ttab[0]==NULL))
    {
      fprintf(stderr,"Error of allocation\n");
      exit(-1);
    }
#define TEST_ALGO(Algo,ind,tt) 				\
  tt = runtime(); 					\
  for(i=0;i<nb;i++) { 					\
    ttab[ind][i]=runtime(); 				\
    if(!_ZENPolyIsNotIrreducible(Ptab[i],Algo,Rg)) { 	\
      printf("Error !!!\n"); exit(-1); } 		\
    ttab[ind][i]=runtime()-ttab[ind][i]; } 		\
  for(;i<nnb;i++) { 					\
    ttab[ind][i]=runtime(); 				\
    if(_ZENPolyIsNotIrreducible(Ptab[i],Algo,Rg)) { 	\
      printf("Error !!!\n"); exit(-1); } 		\
    ttab[ind][i]=runtime()-ttab[ind][i]; } 		\
  tt = (runtime()-tt)/nnb;
	  
  tt = runtime(); 
  Ptab=ZENPolyRandomArray(nb,deg,Rg); 
  tt = (runtime()-tt)/1000.0; 
  printf("%d polynomials of degree %d (%d irreducible) found in %f s\n",nnb,deg,nnb-nb,tt);
  TEST_ALGO(Berlekamp,0,tt); 
  printf("Berlekamp : %f ms\n",tt); 
  TEST_ALGO(DDF,1,tt); 
  printf("DDF : %f ms\n",tt); 
  Stat(ttab,nb); 
  ZENPolyFreeArray(Ptab,nb,Rg);

#endif    
      
/* Freeing memory */
  ZENRingFullClose(Rg);

#if defined(i486icc) & defined(__DEBUG_ALLOC__)
  _dump_allocated(0);
#endif

  exit(0);
}
