/*
 * zenfact/c/berlekamp.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zen.h"
#include "zenfact.h"

/*
   \TEX{
\subsection{Low-level irreducibility functions}
Two algorithms are implemented to test irreducibility of a polynomial.
   DDF algorithm depends a lot on the shape of the polynomial. Berlekamp's algorithm is usually faster for small degree polynomials, and its behaviour is much more regular. For this reason, Berlekamp's algorithm is used in all the high level functions of the library. It is up to the user to use |_ZENPolyIsNotIrreducible| with |DDF| or any other algorithm he might want to use.

     \title{Testing irreducibility of a polynomial}
     \input{A polynomial |PX| defined over a finite ring |Rg|, and a function pointer on the algorithm to use (|Berlekamp| and |DDF| are available).}
     \output{\begin{description}
       \item[|ZENERR|] if an error occured,
       \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
       \item[|ZENPOLY_HAS_ZERO_ROOT|] if the polynomial has zero as root,
       \item[|ZENPOLY_HAS_ONE_ROOT|] if the polynomial has one as root, 
       \item[|ZENPOLY_HAS_REPEATED_FACTORS|] if the polynomial has repeated factors,
       \item[|ZENPOLY_IS_P_POWER|] if $|P|(X)= g(X)^p$, with $p$ the
         characteristic of the |ZENRing|,
       \item[|ZENPOLY_IS_IRREDUCIBLE|] if the polynomial is irreducible,
       \item[|ZENPOLY_IS_COMPOSED|] otherwise.
       \end{description}}
     \note{This function performs what is usually called the ERF phase (Elimination of Repeated Factors) of a polynomial factorization. This is done before calling the indicated algorithm to complete a DDF phase (Distinct-Degree Factorization). All these operations stop as soon as the polynomial is found composed. The complete factorization is therefore not done. }
     }
   */
int _ZENPolyIsNotIrreducible(P,Algo,R)
     ZENPoly P;
     int (*Algo) ___((ZENPoly, ZENRing));
     ZENRing R;
{ZENPoly D,G;
 int deg;
 ZENElt E,F;
 
 deg=ZENPolyDeg(P,R);

 if(ZENEltIsZero(ZENPolyGetCoeffPtr(P,0,R),R))
   return(ZENPOLY_HAS_ZERO_ROOT);
 if(ZENEltCreate(E,R))
   return(ZERR);
 if(ZENEltCreate(F,R))
   {
     ZENEltFree(E,R);
     return(ZERR);
   }
 ZENEltSetToOne(E,R);
 ZENPolyEval(F,P,E,R);
 if(ZENEltIsZero(F,R))
   {
     ZENEltFree(E,R);
     ZENEltFree(F,R);
     return(ZENPOLY_HAS_ONE_ROOT);
   }
 ZENEltFree(E,R);
 ZENEltFree(F,R);
 
 if(deg <= 1)
   return(ZENPOLY_IS_IRREDUCIBLE);

 D=ZENPolyCreate(deg,R);
 if(!D)
   return(ZERR);
 
 if(ZENPolyDerive(D,P,R))
   {
     ZENPolyFree(D,R);
     return(ZERR);
   }

 if(ZENPolyIsZero(D,R))
   {
     ZENPolyFree(D,R);
     return(ZENPOLY_IS_P_POWER);
   }
 
 G=ZENPolyCreate(deg+1,R);
 if(!(G))
   {
     ZENPolyFree(D,R);
     return(ZERR);
   } 

 switch(ZENPolyGcd(G,P,D,R))
   {
   case ZENERR:
     ZENPolyFree(D,R);
     ZENPolyFree(G,R);
     return(ZERR);
   case ZEN_NO_INVERSE:
     ZENPolyFree(D,R);
     ZENPolyFree(G,R);
     return(ZEN_NO_INVERSE);
   case ZEN_HAS_INVERSE:
     break;
   }
 ZENPolyFree(D,R);

 if(!((ZENEltIsOne(ZENPolyGetCoeffPtr(G,0,R),R))
      &&(ZENPolyDeg(G,R)==0)))
   {
     ZENPolyFree(G,R);
     return(ZENPOLY_HAS_REPEATED_FACTORS);
   }

 ZENPolyFree(G,R);

 return(Algo(P,R));
}

/**********************************************************************/
static int ZENConstructBerlekamp(Q,P,R)
     ZENMat Q;
     ZENPoly P;
     ZENRing R;
{int deg,i,ret;
 ZENElt X,Xq,Y;
 ZENRing R2;
 ZENMat QPE;
#if !ZEN_NO_TEST
 int olddebugflag;
#endif 
 deg=ZENPolyDeg(P,R);
 
 R2=ZENExtRingCreate(P,R);
 
 if(!R2) 
   return(ZERR);
 
 if(ZENEltCreate(X,R2)) {
   ZENRingClose(R2);
   return(ZERR);
 }
 
 if(ZENEltCreate(Xq,R2)) {
   ZENEltFree(X,R2);
   ZENRingClose(R2);
   return(ZERR);
 }

 if(ZENEltCreate(Y,R2)) {
   ZENEltFree(X,R2);
   ZENEltFree(Xq,R2);
   ZENRingClose(R2);
   return(ZERR);
 }

 ZENEltSetToGenerator(Y, R2);
 if(ZENEltExp(Xq,ZENRingQ(R),ZENRingQl(R),Y,R2)) {
   ZENEltFree(X,R2);
   ZENEltFree(Y,R2);
   ZENEltFree(Xq,R2);
   ZENRingClose(R2);
   return(ZERR);
 }

				/* First iteration of DDF */
 ZENEltNegate(Y,Y,R2);
 ZENEltAssign(X,Y,R2);
 ZENEltAdd(X,Xq,R2);
#if !ZEN_NO_TEST
 olddebugflag=debugflag;
 debugflag=0;
#endif
 ret=ZENEltInverse(X,X,R2);
#if !ZEN_NO_TEST
 debugflag=olddebugflag;
#endif
 switch(ret)
   {
   case ZEN_HAS_INVERSE:
     break;
   case ZEN_NO_INVERSE:
     ZENEltFree(X,R2);
     ZENEltFree(Y,R2);
     ZENEltFree(Xq,R2);
     ZENRingClose(R2);
     return(ZENPOLY_IS_COMPOSED);
       
   default:
     if(ZENErrErrno == ZEN_DIVBYZERO)
       ret=ZENPOLY_IS_COMPOSED;
     else
       ret=ZERR;
     ZENEltFree(X,R2);
     ZENEltFree(Y,R2);
     ZENEltFree(Xq,R2);
     ZENRingClose(R2);
     return(ret);     
   }
				/* end of first iteration of DDF */
 if(ZENPolyDeg(P,R)<=3)
   {
     ZENEltFree(X,R2);
     ZENEltFree(Y,R2);
     ZENEltFree(Xq,R2);
     ZENRingClose(R2);
     return(ZENPOLY_IS_SURELY_IRREDUCIBLE);
   }
			/*  second iteration of DDF */
 if(ZENEltExp(X,ZENRingQ(R),ZENRingQl(R),Xq,R2)) {
   ZENEltFree(X,R2);
   ZENEltFree(Y,R2);
   ZENEltFree(Xq,R2);
   ZENRingClose(R2);
   return(ZERR);
 }
 ZENEltAdd(X,Y,R2);
#if !ZEN_NO_TEST
 olddebugflag=debugflag;
 debugflag=0;
#endif
 ret=ZENEltInverse(X,X,R2);
#if !ZEN_NO_TEST
 debugflag=olddebugflag;
#endif
 switch(ret)
   {
   case ZEN_HAS_INVERSE:
     break;
   case ZEN_NO_INVERSE:
     ZENEltFree(X,R2);
     ZENEltFree(Y,R2);
     ZENEltFree(Xq,R2);
     ZENRingClose(R2);
     return(ZENPOLY_IS_COMPOSED);
       
   default:
     if(ZENErrErrno == ZEN_DIVBYZERO)
       ret=ZENPOLY_IS_COMPOSED;
     else
       ret=ZERR;
     ZENEltFree(X,R2);
     ZENEltFree(Y,R2);
     ZENEltFree(Xq,R2);
     ZENRingClose(R2);
     return(ret);     
   }
				/* end of second iteration of DDF */
 if(ZENPolyDeg(P,R)<=5)
   {
     ZENEltFree(X,R2);
     ZENEltFree(Y,R2);
     ZENEltFree(Xq,R2);
     ZENRingClose(R2);
     return(ZENPOLY_IS_SURELY_IRREDUCIBLE);
   }
 
 if(ZENMatAlloc(QPE,deg-1,1,R2)) {
   ZENEltFree(X,R2);
   ZENEltFree(Y,R2);
   ZENEltFree(Xq,R2);
   ZENRingClose(R2);
   return(ZERR);
 }

 ZENEltAssign(X,Xq,R2);
 for(i=0;;)
   {
     ZENPolySetToXi(ZENElt2Pol(Y,R2),i+1,R);
     ZENEltNegate(Y,Y,R2);
     ZENEltAdd(Y,X,R2);
     ZENMatSetCoeff(QPE,i,0,Y,R2);
     i++;
     if(i>=deg-1)
       break;
     
     ZENEltMultiply(X,Xq,X,R2);
   }

#if 0
 printf("ZEN\n");
 ZENMatPrintToFile(stdout,QPE,10,R2);
#endif

 ret=ZENVect2Mat(Q,QPE,R2);
 
 ZENMatFree(QPE,R2);
 ZENEltFree(X,R2);
 ZENEltFree(Y,R2);
 ZENEltFree(Xq,R2);
 ZENRingClose(R2);
 
 return(ret);
}
/**********************************************************************/
/*
   \TEX{
   \title{Distinct-Degree Factorization}
   \input{A |ZENPoly| defined over a |ZENRing|, that must have no repeated factors.}
     \output{\begin{description}
       \item[|ZENERR|] if an error occured,
       \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
       \item[|ZENPOLY_IS_IRREDUCIBLE|] if the polynomial is irreducible,
       \item[|ZENPOLY_IS_COMPOSED|] otherwise.
       \end{description}}
   \note{[description of the algorithm]. We add two tests to this algorithm, which correspond to the first two iterations of |DDF| algorithm. This gives the best results.}
   } 
   */
int Berlekamp(P,R)
     ZENPoly P;
     ZENRing R;
{
  ZENMat Q;
  int ret;
  
  Q=ZENMatRowCreate(ZENPolyDeg(P,R)-1,ZENPolyDeg(P,R),R);
  if(!Q) {
    return(ZERR);
  }
  
  switch(ret=ZENConstructBerlekamp(Q,P,R))
    {
    case ZENPOLY_IS_SURELY_IRREDUCIBLE:
      ZENMatFree(Q,R);
      return(ZENPOLY_IS_IRREDUCIBLE);
    case 0:
      break;
    default:
      ZENMatFree(Q,R);
      return(ret);
    }
  
  ret = ZENMatGaussTrigInvertible(Q,R);
  ZENMatFree(Q,R);
  
  return(ret);
}

/**********************************************************************/
/*
   \TEX{
   \title{Distinct-Degree Factorization}
   \input{A |ZENPoly| defined over a |ZENRing|, that must have no repeated factors.}
     \output{\begin{description}
       \item[|ZENERR|] if an error occured,
       \item[|ZEN_NO_INVERSE|] if an inverse was impossible to compute,
       \item[|ZENPOLY_IS_IRREDUCIBLE|] if the polynomial is irreducible,
       \item[|ZENPOLY_IS_COMPOSED|] otherwise.
       \end{description}}
   \note{[description of the algorithm]}
   } 
   */
int DDF(P,R)
     ZENPoly P;
     ZENRing R;
{
  ZENRing R2;
  ZENPrc prc;
  ZENElt X,Y,Xiq;
  int lim,ret=ZENPOLY_IS_IRREDUCIBLE;
#if !ZEN_NO_TEST
 int olddebugflag;
#endif 
  
  if(ZENExtRingAlloc(R2,P,R))
    return(ZERR);
  if (ZENPrcIsSet(ZENRingPrcp(R), ZENPRC_ELT_EXP)) {
    ZENPrcSetNone(prc); 
    ZENPrcSet(prc, ZENPRC_ELT_EXP);
    ZENRingAddPrc(R2, prc);
  }
  if(ZENEltAlloc(X,R2)) {
    ZENRingClose(R2);
    return(ZERR);
  }
  if(ZENEltAlloc(Y,R2)) {
    ZENEltFree(X,R2);
    ZENRingClose(R2);
    return(ZERR);
  }
  if(ZENEltAlloc(Xiq,R2)) {  
    ZENEltFree(X,R2);
    ZENEltFree(Y,R2);
    ZENRingClose(R2);
    return(ZERR);
  }
  ZENPolySetToXi(ZENElt2Pol(X,R2),1,R);
  ZENEltAssign(Xiq,X,R2);

  lim=ZENPolyDeg(P,R)/2;
  for(;lim--;)
    {
      ZENEltAssign(Y,Xiq,R2);
      if(ZENEltExp(Xiq,ZENRingQ(R),ZENRingQl(R),Y,R2))
	{
	  ret=ZERR;
	  goto OUT;
	}

      ZENEltNegate(Y,X,R2);
      ZENEltAdd(Y,Xiq,R2);
#if !ZEN_NO_TEST
      olddebugflag=debugflag;
      debugflag=0;
#endif
      ret=ZENEltInverse(Y,Y,R2);
#if !ZEN_NO_TEST
      debugflag=olddebugflag;
#endif
      switch(ret)
	{
	case ZEN_HAS_INVERSE:
	  break;
	case ZEN_NO_INVERSE:
	  ret=ZENPOLY_IS_COMPOSED;
	  goto OUT;
	default:
	  if(ZENErrErrno == ZEN_DIVBYZERO)
	    ret=ZENPOLY_IS_COMPOSED;
	  else
	    ret=ZERR;
	  goto OUT;
	}
    }
OUT:
  ZENEltFree(X,R2);
  ZENEltFree(Xiq,R2);
  ZENEltFree(Y,R2);
  ZENRingClose(R2);
  return(ret);
}

