/*
 * zenfact/c/random.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <stdio.h>
#include "zen.h"
#include "ze2.h"
#include "zenfact.h"

/* \TEX{
   \Section{Pseudo-random generator (PRG)}
   
   The following procedures can be used through a call to
   |ZENSetRandomFunctions|.  This pseudo-random generator is intended
   to be used when good quality pseudo-random variables are
   needed. This quality is obtained using a linear feedback shift
   register. The drawback is that it is less efficient than a linear
   congruential generator such as the default one. It is provided with
   the |zenfact| library because high level functions of |ZEN| are
   used in order to implement the PRG.
   
   \Subsection{Description} The core of the PRG is a linear feedback
   shift-register defined by a primitive polynomial of degree 127 over
   $\GF{2}$.  As $2^{127}-1$ is a prime number, every irreducible
   polynomial of degree 127 over $\GF{2}$ is also a primitive
   polynomial. The chosen polynomial is a "good" polynomial, in the
   sense of |ZENPolySetGoodIrreducible|.  } */
#define PRGFIXED 1		/* Can be set to 0 in order to choose the polynomial
				   at initialization */
#define DEBUG 1
#define DEGPRG 127
#if PRGFIXED			/* \CiteC */
# define PolyDefPRG "(1)*X^127+(1)*X^7+(1)*X^6+(1)*X^5+(1)*X^4+(1)*X^3+(1)*X^2+(1)*X+(1)" /* \EndCiteC */
#endif

static ZENRing Ex;
static ZENElt state,adv;

/*
   \TEX{
   \Subsection{Functions}

   \title{Pseudo-random generator initialization}
   \output{0 if no error occured, |ZENERR| otherwise.}
   \side{This function allocates the internal structures of the PRG.}
   \note{This function must be called by the |ZENSetRandomFunctions|.}
   } 
   */
int ZENRandomAlloc()
{
  BigNumDigit d=2;
  ZENRing Rg;
  ZENPoly P;
  ZENPrc prc;
  
  if(ZENBaseRingAlloc(Rg,&d,(BigNumLength)1)<0)    
    return(ZERR);
#if PRGFIXED
  if(ZENPolyReadFromString(P,PolyDefPRG,10,Rg)<0)
    {
      ZENRingClose(Rg);
      return(ZERR);
    }
#else
  if(ZENPolyAlloc(P,DEGPRG,Rg)<0)
    {
      ZENRingClose(Rg);
      return(ZERR);
    }
  if(ZENPolySetGoodIrreducible(P,DEGPRG,Rg)<0)
    {
      ZENPolyFree(P,Rg);
      ZENRingClose(Rg);
      return(ZERR);
    }
# if DEBUG
  ZPP(P,Rg);
# endif
#endif
  if(ZENExtRingAlloc(Ex,P,Rg)<0)
    {
      ZENPolyFree(P,Rg);
      ZENRingClose(Rg);
      return(ZERR);
    }
  ZENPolyFree(P,Rg);
  ZENPrcSetAll(prc);
  ZENPrcSet(prc,ZENPRC_FINITE_FIELD);
  if(ZENRingAddPrc(Ex,prc)<0)
    {
      ZENRingFullClose(Ex);
      return(ZERR);
    }
  if(ZENEltAlloc(state,Ex)<0)
    {
      ZENRingFullClose(Ex);
      return(ZERR);
    }
  ZENEltSetToOne(state,Ex);
  if(ZENEltAlloc(adv,Ex)<0)
    {
      ZENEltFree(state,Ex);
      ZENRingFullClose(Ex);
      return(ZERR);
    }
  ZENPolySetToXi(ZENElt2Pol(adv,Ex),1,Rg);
#if DEBUG
  ZEP(state,Ex);
  ZEP(adv,Ex);
#endif
  return(0);
}

/*
   \TEX{
   \title{Pseudo-random generator freeing memory}
   \output{0 if no error occured, |ZENERR| otherwise.}
   \side{This function frees the internal structures of the PRG.}
   \note{This function can be called by a new call to the |ZENSetRandomFunctions| restoring the initial values of |zbnrandom| and |zbnsrandom|. However, the initialization is pretty long and ahould not be erased, except if no more randomized variable is used.}
   } 
   */
int ZENRandomClose()
{
  ZENEltFree(state,Ex);
  ZENEltFree(adv,Ex);
  return(ZENRingFullClose(Ex));
}

/*
   \TEX{
   \title{Pseudo-random function}
   \output{One pseudo-random |BigNumDigit|}
   } 
   */
BigNumDigit ZENRandom()
{
  BigNumDigit d;
  int i;

  d=(Ze2PolyGetDigit(ZENElt2Pol(state,Ex),0,ZENRingDef(Ex)) & 1);
  for(i=1;i<SIZE_BLOC;i++)
    {
      d <<= 1;
      ZENEltMultiply(state,adv,state,Ex);
      d |= (Ze2PolyGetDigit(ZENElt2Pol(state,Ex),0,ZENRingDef(Ex)) & 1);
    }
#if DEBUG
  ZEP(state,Ex);
  /* test processed at compilation time */
  if (SIZE_BLOC == 64) {
    fprintf(stderr,"Bloc %.16X\n\n",d);
  }
  else {
    fprintf(stderr,"Bloc %.8X\n\n",d);
  }
#endif
  return(d);
}

/*
   \TEX{
   \title{Pseudo-random initialization}
   \input{A non-zero seed for the |ZENRandom| function.}
   \note{If |seed| is zero, it is ignored.}
   } 
   */
void ZENSrandom(seed)
     BigNumDigit seed;
{
  int i;

  if(seed)
    {
      i=divSizeBloc(DEGPRG);
      
      Ze2PolyDeg(ZENElt2Pol(state,Ex),ZENRingDef(Ex))=i*SIZE_BLOC+SIZE_BLOC-1;
      for(;i>=0;i--)    
	Ze2PolyGetDigit(ZENElt2Pol(state,Ex),i,ZENRingDef(Ex))=seed;
      Ze2PolyUpdateDegree(ZENElt2Pol(state,Ex),ZENRingDef(Ex));
    }
#if DEBUG
  ZEP(state,Ex);
#endif  
  for(i=0;i<DEGPRG;i++)
    ZENRandom();
}

