/*-*-LaTeX-*-
 * zmake/doc/compile.c
 *
 * Copyright (C) 2000 Florent Chabaud and Reynald Lercier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

/*
  This is a documentation file of the ZEN library.
  There is no C code in this file, only LaTeX commands.


\TEX{

  All this can be found in \cite{imake}.

  \vspace*{1cm}
  Suppose we're writing a small C program, |myname| consisting of a
  single source file |myname.c|:

  }
\CiteC
*/
#include <stdio.h>

int main()
{
  char *getenv();
  char *p;

  p = getenv ("USER");
  if (p == (char *)NULL) {
    printf("Cannot tell your user name\n");
  }
  else {
    printf("Your user name is %s\n", p);
  }
  exit(0);
}

/*
\EndCiteC
\TEX{
  The |Makefile| for the program might look like this:
  }
\CiteC
*/
myname: myname.o
        cc -o myname myname.o
install: myname
        install myname /usr/local
/*
\EndCiteC
\TEX{
  To build and install myname, we'd use these commands:
  }
\CiteC
*/
% make myname
cc  -c myname.c
cc -o myname myname.o
% make install
install myname /usr/local
/*
\EndCiteC
\TEX{
  
  All our friends are astonished at the use fullness of this program
  and immediately begin asking us for the source
  code. We give it to one of them, together with the |Makefile|, and
  he lopes off to install it on his machine. The next day he sends
  back the following report:
  \begin{itemize}
  \item We don't have any install program, so I used |cp|.
  \item Our local installation directory is |/usr/local/bin| rather
  than |/usr/local|.
  \end{itemize}

  He sends back a revised |Makefile|, having made the necessary
  changes, and also having conveniently parameterized the things he
  found to be different on his system, using the |make| variables
  |INSTALL| and |BINDIR| at the beginning of the |Makefile|:
  }
\CiteC
*/
INSTALL = cp
BINDIR = /usr/local/bin
myname: myname.o
        cc -o myname myname.o
install: myname
        install myname /usr/local
/*
\EndCiteC
\TEX{
  This is a step forward. The installation program and directory as
  specified in this |Makefile| are no longer correct for our machine,
  but now they're parameterized and thus easily located and
  modified. We change the values back to |install| and
  |/usr/local/bin/| so they'll work on our machine, and both we and
  our friend are happy.

  Alas, our idyllic state of mind doesn't last long. We give the
  program to another friend. She builds it and reports back her
  changes:
  \begin{itemize}
  \item I prefer to use |gcc|, not |cc|.
  \item Our C library is broken and doesn't contain |getenv()|, I have
  to link in |-lc_aux| as well.
  \item The environment variable |USER| isn't used on our system, but
  |LOGNAME| is, so I used that instead.
  \end{itemize}

  She, too, sends back a revised |Makefile|, further parameterized:
}
\CiteC
*/
CC = gcc
EXTRA_LIBS = -lc_aux
CFLAGS = -DUSE_LOGNAME
INSTALL = cp
BINDIR = /usr/local/bin
myname: myname.o
        $(CC) $(CFALGS) $(EXTRA_LIBS) -o myname myname.o
install: myname
        install myname /usr/local
/*
\EndCiteC
$
\TEX{
  and a revised file |myname.c|:
}
\CiteC
*/
#include <stdio.h>

int main()
{
  char *getenv();
  char *p;

#ifdef USE_LOGNAME
  p = getenv ("LOGNAME");
#else
  p = getenv ("USER");
#endif
  if (p == (char *)NULL) {
    printf("Cannot tell your user name\n");
  }
  else {
    printf("Your user name is %s\n", p);
  }
  exit(0);
}
/*
\EndCiteC
\TEX{

  Now, with a simple change to the |Makefile|, |myname| can be
  recompiled on machines that use either environment variable. This is
  encouraging. Using our vast and ever increasing porting experience,
  |myname| now works on more machines-three ! And the revised
  |Makefile| is better than the original:
  \begin{itemize}
  \item It parameterizes all the non portability we've encountered so
  far.
  \item Nonportabilities are identified explicitly at the beginning of
  the |Makefile|.
  \item They're easily modified by editing |Makefile|.
  \end{itemize}

  However, we begin to notice uneasily that the number of things that
  might need changing from machine to machine is increasing. They were
  only two differences between machines when we built |myname| on two
  systems. Now we can build it on three systems but the number of
  differences has increased to five. How many differences will we find
  when we attempt to port |myname| to additional systems ? As the
  |Makefile| editing job gets bigger, it becomes more difficult. Each
  nonportability adds another increment to the burden incurred each
  time |myname| is built on a different machine.

  Also, we'd better keep a list of parameters appropriate for each
  type of machine on which the program is known to run, or we'll
  forget them. And we need to distribute the list to other people who
  want to build the program, so they can consult it to see how they
  might need to modify the makefile on their |systems|.

  The previous example outlines mainly four drawbacks of |make|:
  \begin{itemize}
  \item There are no conditionals:\\ 
    If we could assign values to |make| variables conditionally, we
    could parameterize machine dependencies by assigning variable
    values appropriate to a given machine type.  To some extend, we
    can use the shell's |if|-statement syntax to get around lack of
    conditionals in target entries, but it's clunky to do so.

  \item There is no flow control:\\
    Loops and iterators must be done using shell commands to simulate
    them.

  \item It's difficult to make global changes:\\ 
    For example, suppose you use BINDIR to name your installation
    directory. You can assign it the proper value in each |Makefile|
    in your project. But if you decide to change it, you must find and
    change every |Makefile|.  Alternatively, you can assign |BINDIR|,
    once in the top level |Makefile| and pass the value to |make|
    commands in subdirectories using suitable recursive rules. But,
    then you can't run |make| directly in a subdirectory because
    |BINDIR| won't have the proper value. Recursive rules are
    difficult to write correctly, anyway.

  \item Header file dependencies for C source files are inherently
    nonportable:\\ 
    Different systems organize header files differently. You might
    even have multiply sets on systems that support development under
    more than one environment (BSD, System V, \ldots). This makes it
    impossible to list header file dependencies statically in the
    |Makefile|. They must be computed on the target machine at build
    time.
  \end{itemize}
}
*/
